#!/bin/bash

#! It's recommended to run this script only once. If your run it after the 
#! installation completes, your database password will change
#! to update annode later (if needed) run the script annode-upgrade.sh instead.

#- Recommended remote OS - Debian 11, base/minimal/CLI only.

#### read further instructions after the installation

DB_NAME="annedb";
DB_USER_NAME="anneroot";
DB_USER_PASS=$(cat /dev/urandom | tr -dc 'a-zA-Z0-9' | fold -w 21 | head -n 1);
DB_ROOT_USER_PASS="r00tXyZ$DB_USER_PASS";

# Detect OS
is_apt_command_available() {
  which apt > /dev/null 2>&1
}
is_dnf_command_available() {
  which dnf > /dev/null 2>&1
}
is_yum_command_available() {
  which yum > /dev/null 2>&1
}

if is_apt_command_available; then
    command="apt"
elif is_dnf_command_available; then
    command="dnf"
elif is_yum_command_available; then
    command="yum"
else
	echo "This script is running on an unsupported distribution.
      Supported distributions are Debian or RHEL based. 
      You can still install annode manually. It runs anywhere Java runs.
      
      Dependencies:
      mariadb
      unzip
      curl
      java jdk 17"
	exit 1;
fi


export PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/sbin:/usr/local/bin

clear

# Check if user is root
[ $(id -u) != "0" ] && { echo "Error: You must use sudo or be root to run this script"; exit 1; }


run="$command update && $command upgrade -y"
eval "$run"

is_sudo_command_available() {
  which sudo > /dev/null 2>&1
}

is_mysql_command_available() {
  which mysql > /dev/null 2>&1
}

is_unzip_command_available() {
  which unzip > /dev/null 2>&1
}
is_zip_command_available() {
  which zip > /dev/null 2>&1
}

if ! is_sudo_command_available; then
  run="$command install -y sudo"
  eval "$run"
fi

is_nano_command_available() {
  which nano > /dev/null 2>&1
}

if ! is_nano_command_available; then
    run="$command install -y nano"
    eval "$run"
fi


if ! is_zip_command_available; then
  run="sudo $command install -y zip"
  eval "$run"
fi
if ! is_unzip_command_available; then
  run="sudo $command install -y unzip"
  eval "$run"
fi

if ! is_mysql_command_available; then
  run="sudo $command install -y mariadb-server"
  eval "$run"
fi

sudo systemctl restart mariadb

echo "--------------------------------------------------------------"
echo "configuring MariaDB"
echo "--------------------------------------------------------------"
sudo mysqladmin --user=root password "$DB_ROOT_USER_PASS"
printf "$DB_ROOT_USER_PASS\n n\n Y\n Y\n Y\n Y\n" | sudo mysql_secure_installation
# sudo mysql -e "SET PASSWORD FOR root@localhost = PASSWORD('$DB_ROOT_USER_PASS');FLUSH PRIVILEGES;"
mysql --user=root --password="$DB_ROOT_USER_PASS" -e "CREATE DATABASE $DB_NAME DEFAULT CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci;"
mysql --user=root --password="$DB_ROOT_USER_PASS" -e "CREATE USER $DB_USER_NAME@localhost IDENTIFIED BY '$DB_USER_PASS';"
mysql --user=root --password="$DB_ROOT_USER_PASS" -e "GRANT ALL ON $DB_NAME.* TO '$DB_USER_NAME'@'localhost' IDENTIFIED BY '$DB_USER_PASS'; FLUSH PRIVILEGES;"

is_java17_available() {
  java_version=$(java -version 2>&1 | awk -F '"' '/version/ {print $2}')
  [[ $java_version == 17* ]]
}

is_java_command_available() {
  which java > /dev/null 2>&1
}

if ! is_java_command_available || ! is_java17_available; then
  run="sudo $command install -y openjdk-17-jdk"
  # sudo apt-get install openjdk-11-jre
  if is_dnf_command_available || is_yum_command_available; then
      run="sudo $command install -y java-17-openjdk"
  fi
  eval "$run"
fi

is_curl_command_available() {
  which curl > /dev/null 2>&1
}
if ! is_curl_command_available; then
  run="sudo $command install -y curl"
  # sudo apt-get install openjdk-11-jre
  eval "$run"
fi

curl -u radanne:anneforthewin https://anne.network/files/anne-node.zip --output "$HOME/anne-node.zip"

# sudo unzip -o anne-pack.zip -d /root/

sudo mkdir "$HOME/annode";
DIR="$HOME/annode";
sudo unzip -o "$HOME/anne-node.zip" -d $DIR

sudo rm -f "$HOME/anne-node.zip"

sudo chown -R $USER:$USER "$HOME/annode"

if [[ $DESKTOP_SESSION == "" ]]; then
    echo "#!/bin/bash
cd $DIR
/usr/bin/java -jar anne-node.jar --headless" > /usr/bin/annode.sh
    sudo chmod +x /usr/bin/annode.sh
    echo "[Unit]
  Description=Annode
  After=network.target
[Service]
  Type=simple
  User=root
  ExecStart=/usr/bin/annode.sh
  Restart=on-abort
[Install]
  WantedBy=multi-user.target" > /etc/systemd/system/annode.service
  cp "$DIR/conf/node-default.properties" "$DIR/conf/node.properties"
else
    echo "#!/bin/bash
cd $DIR
/usr/bin/java -jar anne-node.jar --headless" > /usr/bin/annode.sh
    sudo chmod +x /usr/bin/annode.sh
fi
if [[ $DESKTOP_SESSION == "" ]]; then
sudo systemctl enable annode.service
fi


echo "--------------------------------------------------------------"
echo "Installation Complete"
echo "--------------------------------------------------------------"
echo "Welcome to...


                @@@@@@@@@@@@@@@@@@@@@@@@@@@@,
            @@@@@@@@@@@@@@&%%&@@@@@@@@@@@@@@@@@@@@.
         @@@@@@@@@,                        @@@@@@@@@@(
        @@@@@@@@@                                .@@@@@@@@
       @@@@@@@@             %@@@@@@@@                @@@@@@@
     @@@@@@@@            (@@@@@@@@@@@@@.               @@@@@@&
   #@@@@@@@             @@@@@@@@@@@@@@@@.               @@@@@@/
  @@@@@@@@              @@@@@@@@@@@@@@@@/               /@@@@@@
 @@@@@@@@                @@@@@@@@@@@@@@%                 @@@@@@@
@@@@@@@@                    @@@@@@@@@                    @@@@@@@"


echo "--------------------------------------------------------------"
echo "Your MariaDB root password is:"
echo "$DB_ROOT_USER_PASS"
echo "--------------------------------------------------------------"
echo "Your anneroot user password for annedb database is:"
echo "$DB_USER_PASS"
echo "--------------------------------------------------------------"
if [[ $DESKTOP_SESSION == "" ]]; then
cp "$DIR/conf/node-default.properties" "$DIR/conf/node.properties"
    echo "In $DIR/conf/node.properties change the following"
    echo "properties to your own"
    echo "anne.nodeAccountId=your NID"
    echo "anne.nodeSecret=your secret/seed"
    echo "DB.User=$DB_USER_NAME"
    echo "DB.Password=$DB_USER_PASS"
    echo "P2P.Port=9115 (change if different than 9115)"
    echo "API.Port=9116 (change if different than 9116)"
    echo "P2P.myAddress=YOURIPADDRESS:YOURP2PPORT"
    echo "--------------------------------------------------------------"
    echo "After modifying the properties run:"
    echo "sudo systemctl start annode.service"
    echo "NB:"
    echo "Because this script was executed in a non-GUI environment, your annode
 will start in headless mode. If you want to run annode GUI later, remove the
 headless parameter from /usr/bin/annode.sh and restart the service."
    echo "--------------------------------------------------------------"
    echo "Check if annode is running"
    echo "systemctl status annode.service"
    echo "--------------------------------------------------------------"
else
    echo "To start annode run..."
    echo "annode.sh"
    echo "--------------------------------------------------------------"
fi
echo "Debug log:"
echo "nano $DIR/annode.log"
echo "--------------------------------------------------------------"
echo "To upgrade annode,"
echo "get:"
echo "https://anne.network/files/annode-upgrade.sh"
echo "and run:"
echo "sudo bash annode-upgrade.sh"
echo "--------------------------------------------------------------"
echo "To upgrade and restore annedb from snap run:"
echo "sudo bash annode-upgrade.sh restore $DB_ROOT_USER_PASS"
echo "--------------------------------------------------------------"