var WIKIDATA = {};

var precision = new arraymap();
precision.setMapKey("i");
precision.addU({i:0, lbl:"billion years", eq:"-5000000000-00-00T00:00:00Z", er:"5 billion years BCE"});
precision.addU({i:3, lbl:"million years", eq:"-13798000000-01-01T00:00:00Z", er:"13,798 million years BCE"});
precision.addU({i:4, lbl:"hundred thousand years", eq:"-2500000-01-01T00:00:00Z", er:"2,500,000 years BCE"});
precision.addU({i:6, lbl:"millennium", eq:"+1500-00-00T00:00:00Z", er:"Second millennium"});
precision.addU({i:7, lbl:"century", eq:"+1801-00-00T00:00:00Z", er:"19th century"});
precision.addU({i:8, lbl:"decade", eq:"+2010-00-00T00:00:00Z", er:"2010s"});
precision.addU({i:9, lbl:"year", eq:"+2025-00-00T00:00:00Z", er:"2025"});
precision.addU({i:10, lbl:"month-year", eq:"+2025-01-00T00:00:00Z", er:"January 2025"});
precision.addU({i:11, lbl:"date", eq:"+2025-01-08T00:00:00Z	", er:"8 January 2025"});

WIKIDATA.parseDateForAnne = function($dObj) {
    console.log($dObj);
    var p = $dObj.precision;
    if (precision.contains(p)==null) return "";
    var pObj = precision.get(p);
    var era = "AD";

    //after, before, time
    //"+1950-09-21T00:00:00Z"
    var time = $dObj.time;
    if (time[0]=="-") era = "BC";
    time = time.substring(1);
    var parts = time.split("T");
    var datestring = parts[0].trim();
    var timestring = parts[1].trim();
    timestring = timestring.substring(0, timestring.length-1);

    var parent = null;
    if (pObj.i==11) parent = ANNE.MEM.gregorian_date;
    else if (pObj.i==10) parent = ANNE.MEM.monthyear;

    //these will only serve as relons, not as neuron-classes.
    else if (pObj.i==9) parent = ANNE.MEM.year;
    else if (pObj.i==8) parent = ANNE.MEM.decade;
    else if (pObj.i==7) parent = ANNE.MEM.century;
    else if (pObj.i==6) parent = ANNE.MEM.millenia;

    //datestring = year, month, day.
    return {date:datestring, time:timestring, parentClassNOBJ:parent};
}


WIKIDATA._neurons = new arraymap();
WIKIDATA._neurons.setMapKey("wd_id");
WIKIDATA.bINIT_MAP=false;

WIKIDATA._prop_map = new arraymap();
WIKIDATA._prop_map.setMapKey("wd_id");
WIKIDATA._inst_of_skips = ["Q208511", "Q208511", "Q1093829", "Q200250", 
                            "Q174844", "Q745456", "Q1549591", "Q13218391", 
                           "Q131338872"];

WIKIDATA._propSkips = ["P131"]


WIKIDATA.testWikiData = async()=> {
    var wdc = new wikidata_connector();
    return await wdc.getAndProcessDataByID();
}

WIKIDATA.extractImgURI=async($rez, $wd_id, $NOBJ)=> {
    if ($rez==null) return null;
    if ($rez.results==undefined) return null;
    var results = $rez.results;
    if (results.bindings==undefined || results.bindings==null) return null;
    var bindings = results.bindings;
    if (bindings.length==0) return null;
    for (var b=0; b < bindings.length; b++) {
        var binding = bindings[b];
        if (binding.image==undefined) continue;
        var img = binding.image;
        if (img.type!="uri") continue;
        var img = img.value;
        console.log($NOBJ.text+" img for id:"+$wd_id+" uri:"+img);
        return img;
    }
    return null;
}



/**
 * We have a known nid.
 * Look for wikidata relon.
 * If exists, use it to get img.
 * 
 * If doesn't exist, look for one.
 * if found, use it to get img.
 * @param {*} $NOBJ 
 */
WIKIDATA.getIMGforNOBJ_naked = async($NOBJ, $altconsole_cb=null)=>{
    //var rez2 = await WIKIDATA.getImagesByID(["Q22686"]);
    //console.log(rez2);
    //return;
    if ($altconsole_cb!=null) $altconsole_cb("get neuron data from anne for "+$NOBJ.nid);

    await ANNE.API.getNeuron($NOBJ.nid, [0,1,2,3,4,5,6,7,8,9,10]);
    var rel = ANNE.MEMG.getHasRelon($NOBJ.outs, ANNE.MEM.wikidata.temp_nid);
    if (rel!=null) {
        if ($altconsole_cb!=null) $altconsole_cb("we have a known wikidata ID "+rel.to_s);
        var id = rel.to_s;
        return await WIKIDATA.getURIByWID([id], $NOBJ, $altconsole_cb);
    }
    else {
        if ($altconsole_cb!=null) $altconsole_cb("no known wikidata id in ANNE");
        var nameWithContext = $NOBJ.getTextWithWikiContext()
        var wdneurons = await WIKIDATA.resolveToID(nameWithContext, $altconsole_cb);
        if (wdneurons && wdneurons.length>0) {
            return await WIKIDATA.getURIByWID([wdneurons[0].wd_id], $NOBJ, $altconsole_cb);
        }
        
        $altconsole_cb("concept not resolved at wikidata"); 
        return {uri:"", match:null, ambigous:false, matches:null}
    }
}


/**
 * 
 * @param {*} $wdid 
 * @param {*} $NOBJ 
 * @param {*} $altconsole_cb 
 * @returns 
 */
WIKIDATA.getURIByWID=async($wdid, $NOBJ, $altconsole_cb)=>{
    var rez2 = await WIKIDATA.getImagesByID($wdid);
    console.log(rez2);
    if (rez2) {
        var uri = await WIKIDATA.extractImgURI(rez2, $wdid, $NOBJ);
        if ($altconsole_cb!=null && uri!=null && uri!=undefined && uri!="") $altconsole_cb("extracted img uri:"+uri);
        return {uri:uri, match:rez2[0], ambiguous:false};
    }
    else {
        if ($altconsole_cb!=null) $altconsole_cb("no img uri found at wikidata for "+$wdid);
        return {uri:"", match:null, matches:null, ambiguous:false};
    }
}


WIKIDATA.getImagesByID = async(ids)=>{
    console.log("GET THE IMAGES, SMOKEY.");
    return new Promise((resolve, reject) => {

    
    const query = `
    SELECT ?entity ?entityLabel ?image WHERE {
    VALUES ?entity { ${ids.map(id => `wd:${id}`).join(' ')} }
    OPTIONAL { ?entity wdt:P18 ?image. }
    }
    LIMIT 10
    `;
    const url = 'https://query.wikidata.org/sparql?query=' + encodeURIComponent(query) + '&format=json';
    fetch(url)
    .then(response => response.json())
    .then(data => {
        console.log(data);
        resolve(data);
    }).catch(error => {
            console.error('Error fetching data:', error);
        });
    });
}


/**
 * 
 * @param {*} $text 
 * @param {*} altconsole_cb - where do do addToConsole
 * @returns 
 */
WIKIDATA.resolveToID = async($text="Bill Murray (actor, comedian)", $altconsole_cb=null)=>{
    var wdc = new wikidata_connector();
    
    if ($altconsole_cb!=null ) wdc.aec.setAltConsole($altconsole_cb);

    wdc.addToConsole("resolving wikidata text '"+$text+"' to a wikidata id");

    var ws = await wdc.exploreForID($text, null);
    if (ws==undefined) {
        return [];
    }
    if (ws.winnerWN!=null) {
        wdc.addToConsole("return winning wikidata resolution wikidata id: "+ws.winnerWN.wd_id);
        return [ws.winnerWN];
    }

    //ambiguous result!
    else if(ws.winnerWNs!=undefined && ws.winnerWNs!=null && ws.winnerWNs.length>0) {
        console.log("ambigous result", ws.winnerWNs);
        wdc.addToConsole("ambiguious results, "+ws.winnerWNs.length);
        return ws.winnerWNs;
    }

    //lookup naked in ANNE, we have nothing in wikidata to go off of.
    else {
        wdc.addToConsole("no results");
        return [];
    }
}



WIKIDATA.getWikidataID = async($text="Bill Murray (actor, comedian)")=>{
    var wdc = new wikidata_connector();
    var ws = await wdc.exploreForID($text, null);
    if (ws==undefined) {
        wdc.addToConsole("no results");
        return [];
    }
    if (ws.winnerWN!=null) {
        console.log("we have a winner", ws.winnerWN);
        
        wdc.addToConsole("we have a winner wikineuron "+ws.winnerWN.wd_id);

        var rels = await ANNE.API.queryForAltID(ANNE.MEM.wikidata.nid, ws.winnerWN.wd_id);
        if (rels!=undefined && rels.length>0) {
            //we have this in ANNE
            console.log("we have this in ANNE! nid: "+rels[0].from_nid);
            ws.winnerWN.setNid(rels[0].from_nid);
        }
        await ws.winnerWN.getWikiData(wdc, ws.inline_context);

        console.log(ws.winnerWN);   
        console.log(ws.winnerWN.wd_relons);  
        return [ws.winnerWN];
    }

    //ambiguous result!
    else if(ws.winnerWNs!=undefined && ws.winnerWNs!=null && ws.winnerWNs.length>0) {
        console.log("ambigous result", ws.winnerWNs);
        return ws.winnerWNs;
    }

    //lookup naked in ANNE, we have nothing in wikidata to go off of.
    else {
        return null;
    }
}


WIKIDATA.addNeuron=function($name, $id, $nObj) {
    var wnobj = new wikidata_neuron($name, $id, $nObj);
    WIKIDATA._neurons.addU(wnobj);
    return wnobj;
}

WIKIDATA.addMapper=function($name, $id, $nObj, $type, $to_type) {
    var mrelnobj = new mappper_relon($name, $id, $nObj, $type, $to_type);
    WIKIDATA._neurons.addU(mrelnobj);
    WIKIDATA._prop_map.addU(mrelnobj);
    return mrelnobj;
}

WIKIDATA.INIT_MAP=function() {
    if (WIKIDATA.bINIT_MAP==true) return;
    WIKIDATA.bINIT_MAP=true;

    //wikidata neurons we map to.
    WIKIDATA.addNeuron("human", "Q5", ANNE.MEM.person);
    WIKIDATA.addNeuron("city", "Q515", ANNE.MEM.city);
    WIKIDATA.addNeuron("Dewey Decimal Classification", "P1036", null);
    WIKIDATA.addNeuron("instance of", "P31", null);
    WIKIDATA.addNeuron("class of", "P279", null);
    WIKIDATA.addNeuron("subclass of", "P1650", null);
    WIKIDATA.addNeuron("name in kana", "P1814", null);
    
    //ids
    WIKIDATA.addNeuron("Library of Congress authority ID", "P244", null);
    WIKIDATA.addNeuron("VIAF ID", "P214", null);
    WIKIDATA.addNeuron("GND ID", "P227", null);
    WIKIDATA.addNeuron("ISNI", "P213", null);
    WIKIDATA.addNeuron("IdRef ID", "P269", null);
    WIKIDATA.addNeuron("Bibliothèque nationale de France ID", "P268", null);
    WIKIDATA.addNeuron("NDL Authority ID", "P349", null);
    WIKIDATA.addNeuron("NL CR AUT ID", "P691", null);
    WIKIDATA.addNeuron("Libraries Australia ID", "P409", null);
    WIKIDATA.addNeuron("SELIBR ID", "P906", null);
    WIKIDATA.addNeuron("National Library of Spain ID", "P950", null);
    WIKIDATA.addNeuron("NSZL (VIAF) ID", "P951", null);
    WIKIDATA.addNeuron("Nationale Thesaurus voor Auteursnamen ID", "P1006", null);
    WIKIDATA.addNeuron("Portuguese National Library author ID", "P1005", null);
    WIKIDATA.addNeuron("Freebase ID", "P646", null);
    WIKIDATA.addNeuron("CANTIC ID (former scheme)", "P1273", null);
    WIKIDATA.addNeuron("Munzinger person ID", "P1284", null);
    WIKIDATA.addNeuron("Perlentaucher ID", "P866", null);
    WIKIDATA.addNeuron("SBN author ID", "P396", null);
    WIKIDATA.addNeuron("FAST ID", "P2163", null);
    WIKIDATA.addNeuron("Gran Enciclopèdia Catalana ID (former scheme)", "P1296", null);
    WIKIDATA.addNeuron("genealogics.org person ID", "P1819", null);
    WIKIDATA.addNeuron("NNDB people ID", "P1263", null);
    WIKIDATA.addNeuron("C-SPAN person string ID (deprecated)", "P2190", null);
    WIKIDATA.addNeuron("Guardian topic ID", "P3106", null);
    WIKIDATA.addNeuron("New York Times topic ID", "P3221", null);
    WIKIDATA.addNeuron("Open Library ID", "P648", null);
    WIKIDATA.addNeuron("Quora topic ID", "P3417", null);
    WIKIDATA.addNeuron("Babelio author ID", "P3630", null);
    WIKIDATA.addNeuron("NLA Trove people ID", "P1315", null);
    WIKIDATA.addNeuron("PIM authority ID", "P3973", null);
    WIKIDATA.addNeuron("Encyclopædia Britannica Online ID", "P1417", null);
    WIKIDATA.addNeuron("Great Russian Encyclopedia Online ID (old version)", "P2924", null);
    WIKIDATA.addNeuron("SHARE Catalogue author ID", "P3987", null);
    WIKIDATA.addNeuron("Libris-URI", "P5587", null);
    WIKIDATA.addNeuron("RKDartists ID", "P650", null);
    WIKIDATA.addNeuron("Treccani ID", "P3365", null);
    WIKIDATA.addNeuron("Treccani's Enciclopedia Italiana ID", "P4223", null);
    WIKIDATA.addNeuron("Open Library subject ID", "P3847", null);
    WIKIDATA.addNeuron("Amazon author ID", "P4862", null);
    WIKIDATA.addNeuron("Apple Music artist ID (U.S. version)", "P2850", null);
    WIKIDATA.addNeuron("Goodreads author ID", "P2963", null);
    WIKIDATA.addNeuron("Google Scholar author ID", "P1960", null);
    WIKIDATA.addNeuron("PhilPapers author ID", "P3232", null);
    WIKIDATA.addNeuron("DBC author ID", "P3846", null);
    WIKIDATA.addNeuron("NSK ID", "P1375", null);
    WIKIDATA.addNeuron("NACSIS-CAT author ID", "P271", null);
    WIKIDATA.addNeuron("NUKAT ID", "P1207", null);
    WIKIDATA.addNeuron("National Library of Korea ID", "P5034", null);
    WIKIDATA.addNeuron("NORAF ID", "P1015", null);
    WIKIDATA.addNeuron("National Library of Latvia ID", "P1368", null);
    WIKIDATA.addNeuron("LittleSis people ID", "P3388", null);
    WIKIDATA.addNeuron("LibraryThing author ID", "P7400", null);
    WIKIDATA.addNeuron("abART person ID", "P6844", null);
    WIKIDATA.addNeuron("The Peerage person ID", "P4638", null);
    WIKIDATA.addNeuron("Brockhaus Enzyklopädie online ID", "P5019", null);
    WIKIDATA.addNeuron("Pontificia Università della Santa Croce ID", "P5739", null);
    WIKIDATA.addNeuron("University of Barcelona authority ID (former scheme)", "P1580", null);
    WIKIDATA.addNeuron("CoBiS author ID", "P7865", null);
    WIKIDATA.addNeuron("Deutsche Biographie (GND) ID", "P7902", null);
    WIKIDATA.addNeuron("GTAA ID", "P1741", null);
    WIKIDATA.addNeuron("EGAXA ID", "P1309", null);
    WIKIDATA.addNeuron("NSZL name authority ID", "P3133", null);
    WIKIDATA.addNeuron("Angelicum ID", "P5731", null);
    WIKIDATA.addNeuron("Geni.com profile ID", "P2600", null);
    WIKIDATA.addNeuron("GeneaStar person ID", "P8094", null);
    WIKIDATA.addNeuron("Hrvatska enciklopedija ID", "P7982", null);
    WIKIDATA.addNeuron("BookBrainz author ID", "P2607", null);
    WIKIDATA.addNeuron("Union Catalog of Armenian Libraries authority ID", "P8833", null);
    WIKIDATA.addNeuron("Companies House officer ID", "P5297", null);
    WIKIDATA.addNeuron("PLWABN ID", "P7293", null);
    WIKIDATA.addNeuron("BHCL UUID", "P9037", null);
    WIKIDATA.addNeuron("COURAGE ID", "P7037", null);
    WIKIDATA.addNeuron("Vienna History Wiki ID", "P7842", null);
    WIKIDATA.addNeuron("Google News topics ID", "P5337", null);
    WIKIDATA.addNeuron("OFDb person ID", "P1375", null);
    WIKIDATA.addNeuron("Visuotinė lietuvių enciklopedija ID", "P7666", null);
    WIKIDATA.addNeuron("National Library of Brazil ID", "P4619", null);
    WIKIDATA.addNeuron("National Library of Chile ID", "P7369", null);
    WIKIDATA.addNeuron("Canadiana Name Authority ID", "P8179", null);
    WIKIDATA.addNeuron("ELNET ID", "P6394", null);
    WIKIDATA.addNeuron("National Library of Greece ID", "P3348", null);
    WIKIDATA.addNeuron("J9U ID", "P8189", null);
    WIKIDATA.addNeuron("National Library of Lithuania ID", "P7699", null);
    WIKIDATA.addNeuron("National Library of Russia ID", "P7029", null);
    WIKIDATA.addNeuron("RERO ID (obsolete)", "P3065", null);
    WIKIDATA.addNeuron("Armeniapedia ID", "P9629", null);
    WIKIDATA.addNeuron("NLP ID (old)", "P1695", null);
    WIKIDATA.addNeuron("Kallías ID", "P9918", null);
    WIKIDATA.addNeuron("NE.se ID", "P3222", null);
    WIKIDATA.addNeuron("CANTIC ID", "P9984", null);
    WIKIDATA.addNeuron("Lex ID", "P8313", null);
    WIKIDATA.addNeuron("CVR person ID", "P7972", null);
    WIKIDATA.addNeuron("Google Arts & Culture entity ID", "0", null);
    WIKIDATA.addNeuron("Webb-site person ID", "P5298", null);
    WIKIDATA.addNeuron("IxTheo authority ID", "P10553", null);
    WIKIDATA.addNeuron("C-SPAN person numeric ID", "P10660", null);
    WIKIDATA.addNeuron("Personality Database profile ID", "P10757", null);
    WIKIDATA.addNeuron("GOLDEN ID", "P7502", null);
    WIKIDATA.addNeuron("Fandango person ID", "P8125", null);
    WIKIDATA.addNeuron("CCAB ID", "P1890", null);
    WIKIDATA.addNeuron("InfluenceWatch ID", "P11615", null);
    WIKIDATA.addNeuron("KBR person ID", "P11249", null);
    WIKIDATA.addNeuron("UK Parliament thesaurus ID", "P4527", null);
    WIKIDATA.addNeuron("University of Barcelona authority ID", "P11686", null);
    WIKIDATA.addNeuron("CiNii Research ID", "P11496", null);
    WIKIDATA.addNeuron("BBC Things ID", "P1617", null);
    WIKIDATA.addNeuron("BBC News topic ID", "P6200", null);
    WIKIDATA.addNeuron("Scopus author ID", "P1153", null);
    WIKIDATA.addNeuron("UOM ID", "P11182", null);
    WIKIDATA.addNeuron("Radio France person ID", "P10780", null);
    WIKIDATA.addNeuron("Hungarian National Namespace person ID (new)", "P11621", null);
    WIKIDATA.addNeuron("WorldCat Entities ID", "P10832", null);
    WIKIDATA.addNeuron("Kinobox person ID", "P12098", null);
    WIKIDATA.addNeuron("Gran Enciclopèdia Catalana ID", "P12385", null);
    WIKIDATA.addNeuron("Parsifal cluster ID", "P12458", null);
    WIKIDATA.addNeuron("Our Campaigns candidate ID", "P8445", null);
    WIKIDATA.addNeuron("DDB person (GND) ID", "P13049", null);
    WIKIDATA.addNeuron("JudaicaLink person (GND) ID", "P13183", null);
    WIKIDATA.addNeuron("France 24 topic ID (French)", "P9346", null);
    WIKIDATA.addNeuron("National Library of Israel ID (old)", "P949", null);
    WIKIDATA.addNeuron("OFDb person ID", "P9404", null);
    WIKIDATA.addNeuron("Trading Card Database person ID", "P5421", null);
    WIKIDATA.addNeuron("Google Arts & Culture entity ID", "P10297", null);
    WIKIDATA.addNeuron("MusicBrainz artist ID", "P434", null);
    WIKIDATA.addNeuron("AlloCiné person ID", "P1266", null);
    WIKIDATA.addNeuron("Discogs artist ID", "P1953", null);
    WIKIDATA.addNeuron("AllMovie person ID", "P2019", null);
    WIKIDATA.addNeuron("Swedish Film Database person ID", "P2168", null);
    WIKIDATA.addNeuron("Elonet person ID", "P2387", null);
    WIKIDATA.addNeuron("PORT person ID", "P2435", null);
    WIKIDATA.addNeuron("Internet Broadway Database person ID", "P1220", null);
    WIKIDATA.addNeuron("Scope.dk person ID", "P2519", null);
    WIKIDATA.addNeuron("ČSFD person ID", "P2605", null);
    WIKIDATA.addNeuron("Danish National Filmography person ID", "P2626", null);
    WIKIDATA.addNeuron("Filmportal ID", "P2639", null);
    WIKIDATA.addNeuron("Kinopoisk person ID", "P2604", null);
    WIKIDATA.addNeuron("TCM Movie Database person ID", "P3056", null);
    WIKIDATA.addNeuron("elFilm person ID", "P3144", null);
    WIKIDATA.addNeuron("EDb person ID", "P3142", null);
    WIKIDATA.addNeuron("elCinema person ID", "P3136", null);
    WIKIDATA.addNeuron("WikiTree person ID", "P2949", null);
    WIKIDATA.addNeuron("KINENOTE person ID", "P3305", null);
    WIKIDATA.addNeuron("Open Plaques subject ID", "P1430", null);
    WIKIDATA.addNeuron("Baseball Reference minor & foreign league player ID", "P1826", null);
    WIKIDATA.addNeuron("Deutsche Synchronkartei actor ID", "P4332", null);
    WIKIDATA.addNeuron("AllMusic artist ID", "P1728", null);
    WIKIDATA.addNeuron("Last.fm ID", "P3192", null);
    WIKIDATA.addNeuron("Internet Off-Broadway Database ID", "P4456", null);
    WIKIDATA.addNeuron("Encyclopædia Universalis ID", "P3219", null);
    WIKIDATA.addNeuron("Open Media Database person ID", "P5534", null);
    WIKIDATA.addNeuron("CineMagia person ID", "P4666", null);
    WIKIDATA.addNeuron("The Numbers person ID", "P4657", null);
    WIKIDATA.addNeuron("U.S. National Archives Identifier", "P1225", null);
    WIKIDATA.addNeuron("Douban movie celebrity ID", "P5284", null);
    WIKIDATA.addNeuron("Giant Bomb ID", "P5247", null);
    WIKIDATA.addNeuron("Comic Vine ID", "P5905", null);
    WIKIDATA.addNeuron("Box Office Mojo person ID (archived)", "P2688", null);
    WIKIDATA.addNeuron("Allcinema person ID", "P7214", null);
    WIKIDATA.addNeuron("Writers Guild of America writer ID", "P7286", null);
    WIKIDATA.addNeuron("Disney A to Z ID", "P6181", null);
    WIKIDATA.addNeuron("Store norske leksikon ID", "P4342", null);
    WIKIDATA.addNeuron("TV.com ID", "P2638", null);
    WIKIDATA.addNeuron("Decine21 person ID", "P7803", null);
    WIKIDATA.addNeuron("Academy Awards Database nominee ID", "P6150", null);
    WIKIDATA.addNeuron("Cinémathèque québécoise person ID", "P8971", null);
    WIKIDATA.addNeuron("Universal Music France artist ID", "P9328", null);
    WIKIDATA.addNeuron("Evene ID", "P5617", null);
    WIKIDATA.addNeuron("American Film Institute person ID", "P5340", null);
    WIKIDATA.addNeuron("BFI Films, TV and people ID", "P4438", null);
    WIKIDATA.addNeuron("MYmovies actor ID (former scheme)", "P4779", null);
    WIKIDATA.addNeuron("Movieplayer person ID", "P4782", null);
    WIKIDATA.addNeuron("FilmTv.it person ID", "P5254", null);
    WIKIDATA.addNeuron("Rotten Tomatoes ID", "P1258", null);
    WIKIDATA.addNeuron("Metacritic ID", "P1712", null);
    WIKIDATA.addNeuron("Behind The Voice Actors person ID", "P5007", null);
    WIKIDATA.addNeuron("cinematografo.it name or company ID", "P4768", null);
    WIKIDATA.addNeuron("LyricsTranslate ID", "P7212", null);
    WIKIDATA.addNeuron("Treccani's Lessico del XXI Secolo ID", "P9775", null);
    WIKIDATA.addNeuron("Treccani's Enciclopedia del Cinema ID", "P9760", null);
    WIKIDATA.addNeuron("Fandom article ID", "P6262", null);
    WIKIDATA.addNeuron("MYmovies person ID", "P9995", null);
    WIKIDATA.addNeuron("MovieMeter person ID", "P9463", null);
    WIKIDATA.addNeuron("ISzDb person ID", "P3114", null);
    WIKIDATA.addNeuron("Tabakalera ID", "P10069", null);
    WIKIDATA.addNeuron("Online PWN Encyclopedia ID", "P7305", null);
    WIKIDATA.addNeuron("Prabook ID", "P3368", null);
    WIKIDATA.addNeuron("George Eastman Museum people ID", "P10018", null);
    WIKIDATA.addNeuron("AdoroCinema person ID", "P10266", null);
    WIKIDATA.addNeuron("Famous Birthdays ID", "P11194", null);
    WIKIDATA.addNeuron("Cinematografo.it name or company ID (new)", "P11345", null);
    WIKIDATA.addNeuron("Movie Walker Press person ID", "P7745", null);
    WIKIDATA.addNeuron("Filmow ID", "P5233", null);
    WIKIDATA.addNeuron("Filmweb.pl person ID", "P5033", null);
    WIKIDATA.addNeuron("Deutsche Synchronkartei person ID", "P11646", null);
    WIKIDATA.addNeuron("Ivi person ID", "P11683", null);
    WIKIDATA.addNeuron("Know Your Meme ID", "P6760", null);
    WIKIDATA.addNeuron("Ciné-Ressources person ID", "P3204", null);
    WIKIDATA.addNeuron("Unifrance person ID", "P3980", null);
    WIKIDATA.addNeuron("Eiga.com person ID", "P7215", null);
    WIKIDATA.addNeuron("Filmfront person ID (archived)", "P8037", null);
    WIKIDATA.addNeuron("ComingSoon.it person ID", "P11434", null);
    WIKIDATA.addNeuron("Süddeutsche Zeitung topic ID", "P10916", null);
    WIKIDATA.addNeuron("Der Spiegel topic ID", "P10234", null);
    WIKIDATA.addNeuron("Laut.de artist ID", "P5356", null);
    WIKIDATA.addNeuron("FilmAffinity person ID", "P12153", null);
    WIKIDATA.addNeuron("Listal ID", "P11386", null);
    WIKIDATA.addNeuron("British Comedy Guide person ID", "P12627", null);
    WIKIDATA.addNeuron("Douban personage ID", "P12836", null);
    WIKIDATA.addNeuron("Emmys person ID", "P8381", null);
    WIKIDATA.addNeuron("Illinois Center for the Book author ID", "P12935", null);
    WIKIDATA.addNeuron("FID performing arts ID", "P10608", null);
    WIKIDATA.addNeuron("KMDb person ID", "P1649", null);
    
    
    //others
    WIKIDATA.addNeuron("Commons category", "P373", null);
    WIKIDATA.addNeuron("Commons gallery", "P935", null);
    WIKIDATA.addNeuron("described by source", "P1343", null);
    WIKIDATA.addNeuron("floruit", "P1317", null);
    WIKIDATA.addNeuron("copyright status as a creator", "P7763", null);
    WIKIDATA.addNeuron("social media followers", "P8687", null);
    WIKIDATA.addNeuron("Krugosvet article", "P6385", null);
    
    //immediate candidates to be props we snag.
    WIKIDATA.addNeuron("languages spoken, written or signed", "P1412", null);
    WIKIDATA.addNeuron("award received", "P166", null);
    WIKIDATA.addNeuron("topic's main category", "P910", null);
    WIKIDATA.addNeuron("net worth", "P2218", null);
    WIKIDATA.addNeuron("convicted of", "P1399", null);
    WIKIDATA.addNeuron("residence", "P551", null);
    WIKIDATA.addNeuron("name in native language", "P1559", null);
    WIKIDATA.addNeuron("has works in the collection", "P6379", null);
    WIKIDATA.addNeuron("work location", "P937", null);
    WIKIDATA.addNeuron("affiliation string", "P6424", null); //??
    WIKIDATA.addNeuron("affiliation", "P1416", null); //??
    WIKIDATA.addNeuron("World Economic Forum ID", "P10347", null); //??
    WIKIDATA.addNeuron("social classification", "P3716", null); //billionaire .. etc
    WIKIDATA.addNeuron("nominated for", "P1411", null);
    WIKIDATA.addNeuron("number of children", "P1971", null); 
    WIKIDATA.addNeuron("filmography", "P1283", null); 
    WIKIDATA.addNeuron("notable work", "P800", null); 
    WIKIDATA.addNeuron("work period (start)", "P2031", null); 
    WIKIDATA.addNeuron("related category", "P7084", null); 
    
    //common values
    WIKIDATA.addNeuron("male", "Q6581097", ANNE.MEM.male);
    WIKIDATA.addNeuron("United States", "Q30", ANNE.MEM.usa);

    WIKIDATA.addNeuron("gregorian date", "Q1985727", ANNE.MEM.date_gregorian);
    
    
    //things we DO MAP...  NOTE: NEED TO ADD PROCESSING TYPE.  YIKES!!!!  (lolz)
    WIKIDATA.addMapper("country of citizenship", "P27", ANNE.MEM.citizenship,  TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("gender", "P21", ANNE.MEM.gender,  TYPE_HAS, TO_TYPE_ANNE);
    
    //this should invoke a clump.
    WIKIDATA.addMapper("start date", "P580", ANNE.MEM.role_start, TYPE_ROLE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("end date", "P582", ANNE.MEM.role_end, TYPE_ROLE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("location", "P1534", ANNE.MEM.role_location, TYPE_ROLE, TO_TYPE_ANNE);

    WIKIDATA.addMapper("birth name","P1477", ANNE.MEM.birthname, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("place of birth", "P19", ANNE.MEM.birthplace, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("date of birth", "P569", ANNE.MEM.birthday, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("date of death", "P570", ANNE.MEM.deathday, TYPE_BE, TO_TYPE_ANNE);
    
    var spouse = WIKIDATA.addMapper("spouse", "P26", ANNE.MEM.spouse, TYPE_HAS, TO_TYPE_ANNE);
    spouse.dual = MEM.spouse;
    spouse.dual_type = TYPE_HAS;

    WIKIDATA.addMapper("education", "P69", ANNE.MEM.alumni_of, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("website", "P856", ANNE.MEM.website, TYPE_HAS, TO_TYPE_URL);
    WIKIDATA.addMapper("employer", "P108", ANNE.MEM.employer, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("occupation", "P106", ANNE.MEM.occupation, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("given name", "P735", ANNE.MEM.firstname, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("family name", "P734", ANNE.MEM.lastname, TYPE_HAS, TO_TYPE_ANNE);

    //kinfolk stuff  (investigate best reverse relons)
    WIKIDATA.addMapper("father", "P22", ANNE.MEM.father, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("mother", "P25", ANNE.MEM.mother, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("child", "P40", ANNE.MEM.kinfolk_child, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("sibling", "P3373", ANNE.MEM.kinfolk_sibbling, TYPE_HAS, TO_TYPE_ANNE);

    //investigate this!  (what is it pointing to??????  corp?  group?  etc)
    var member_of = WIKIDATA.addMapper("member of", "P463", ANNE.MEM.member_of, TYPE_BE, TO_TYPE_ANNE);
    member_of.dual = ANNE.MEM.member;
    member_of.dual_type = TYPE_HAS;

    //investigate this!!!
    var participant_of = WIKIDATA.addMapper("participant in", "P1344",  ANNE.MEM.participant_of, TYPE_BE, TO_TYPE_ANNE);
    participant_of.dual = ANNE.MEM.participant;
    participant_of.dual_type = TYPE_HAS;

    WIKIDATA.addMapper("religion or worldview", "P140", ANNE.MEM.religion, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("image", "P18", ANNE.MEM.neuronImg, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("member of political party", "P102", ANNE.MEM.political_party, TYPE_HAS, TO_TYPE_ANNE);
    WIKIDATA.addMapper("ethnicity", "P172", ANNE.MEM.ethnicity, TYPE_BE, TO_TYPE_ANNE);

    //ids we use
    WIKIDATA.addMapper("IMDB ID", "P345", MEM.IMDB_id, TYPE_HAS, TO_TYPE_STRING);
    WIKIDATA.addMapper("X username", "P2002", ANNE.MEM.twitter_id, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("Ballotpedia ID", "P2390", ANNE.MEM.ballotpedia, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("TMDB person ID", "P4985", ANNE.MEM.TMDB_id, TYPE_BE, TO_TYPE_ANNE);

    //need to make a neuron for these.
    WIKIDATA.addMapper("native language", "P103", null, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("student of", "P1066", null, TYPE_BE, TO_TYPE_ANNE);
    WIKIDATA.addMapper("owner of", "P1830", null, TYPE_BE, TO_TYPE_ANNE);
}

WIKIDATA.resolveANNEClass = function($wd_class_nobj) {
    if ($wd_class_nobj.nid==ANNE.MEM.city.nid) return ANNE.location;
    if ($wd_class_nobj.nid==ANNE.MEM.person.nid) return ANNE.person;
    //date, datepart, group, group event
    //clumps - 
        //marriage, service (how to do this?)
}

class wikidata_neuron {
    constructor($name, $wd_id, $resolvedNOBJ=null) {
        this.name = $name;
        this.wd_id = $wd_id;
        this.wd_lbls = [];                  //wikidata lbls
        this.wd_primary_class="";           //wikidata class (instance of or subclass of)
        this.wd_desc = "";                  //wikidata description
        this.wd_pageid = "";                //wikidata pageid
        this.wd_url = "";                   //wikidata url
        this.wd_et = "";                    //item,
        this.wd_num_id = "";
        this.wd_snaktype = "";              //value, somevalue, novalue, globally-unique-id
        this.wd_type = "";                  //statement,
        this.wd_dtype = "";                 //wikibase-entityid, external-id
        this.inline_context_score = 0;      //matching words (quick)

        //we will need a deeper neuron relon disambig for knowns that we are trying to get the wd_id for.
        //in that disambig we will use a pool that looks up our relons first then matches the context.

        this.classNOBJ = null;              //anne class
        this.resolveNOBJ = $resolvedNOBJ;   //anne neuron

        this.wikidata = null;               //data from the getAndProcessData wikidata query.

        this.ntype = NTYPE_UNKNOWN;

        //when we snag data we will put it in here.
        this.wd_relons = null;
        this.wd_unhandled_props = null;
    }


    /**
     * @param {wikidata_connector} $wdc 
     * @param {*} $inline_context
     * @param {*} $bCHASEINNERS - query other things as needed?  
     * @returns 
     */
    getWikiData=async($wdc, $inline_context=[], $bCHASEINNERS=true)=> {
        if (this.wd_id==null) return;

        this.wikidata = await $wdc.rawGetByIDs([this.wd_id]);
        if (this.wikidata==null) return;
        console.log(this.wikidata);
        
        var entities = this.wikidata.entities;
        if (entities==null) return;
        if (entities[this.wd_id]==null) return;

        var entity = entities[this.wd_id];
        
        if (entity.labels!=undefined) {
            console.log(entity.labels);
        }
        
        //if we don't have a resolvedNOBJ we need to get the classNOBJ
        //in order to create a ghost neuron to hold the new data.
        if (this.classNOBJ==null && this.resolveNOBJ==null) {
            //instance of
            if (entity.claims["P31"]!=null) {
                this.ntype = NTYPE_INST;
                var instance_of = entity.claims["P31"];
                console.log(instance_of);

                //these will all be candidates for be.
                for (var i=0; i < instance_of.length; i++) {
                    var wn = $wdc.processSingleClaim(instance_of[i], WIKIDATA._inst_of_skips);
                    if (wn==null) continue;
                    if (wn.resolveNOBJ==null) {
                        //explore here.
                        var parent_name = await $wdc.getName(wn);
                        ANNE.TALK.printf_console(ANNE.TALK.YELLOW, "inst of ", ANNE.TALK.GREEN, " "+parent_name+" ", ANNE.TALK.CYAN, wn.wd_id);
                    }
                    else {
                        ANNE.TALK.printf_console(ANNE.TALK.YELLOW, "inst of ", ANNE.TALK.GREEN, " "+wn.name+" ");
                    }
                }
            }

            //class of
            else if(entity.claims["P279"]!=null) {
                this.ntype = NTYPE_CLASS;
                var class_of = entity.claims["P279"];
                console.log(class_of);
                //these will all be candidates for be.
                for (var i=0; i < class_of.length; i++) {
                    var wn = $wdc.processSingleClaim(class_of[i]);
                    if (wn==null) continue;
                    if (wn.resolveNOBJ==null) {
                        //explore here.
                        var parent_name = await $wdc.getName(wn);
                        ANNE.TALK.printf_console(ANNE.TALK.YELLOW, "parent name ", ANNE.TALK.GREEN, " "+parent_name+" ", ANNE.TALK.CYAN, wn.wd_id);
                    }
                    else {
                        ANNE.TALK.printf_console(ANNE.TALK.YELLOW, "parent name ", ANNE.TALK.GREEN, " "+parent_name+" ", ANNE.TALK.BLUE, wn.text);
                    }
                }
            }

            //subclass of
            else if(entity.claims["P1650"]!=null) {
                this.ntype = NTYPE_CLASS;
                var subclass_of = entity.claims["P1650"];
                console.log(subclass_of);
                for (var i=0; i < subclass_of.length; i++) {
                    var wn = $wdc.processSingleClaim(subclass_of[i]);
                    if (wn==null) continue;
                    if (wn.resolveNOBJ==null) {
                        //explore here.
                        var parent_name = await $wdc.getName(wn);
                        ANNE.TALK.printf_console(ANNE.TALK.YELLOW, "parent name ", ANNE.TALK.GREEN, " "+parent_name+" ", ANNE.TALK.CYAN, wn.wd_id);
                    }
                    else {
                        ANNE.TALK.printf_console(ANNE.TALK.YELLOW, "parent name ", ANNE.TALK.GREEN, " "+parent_name+" ", ANNE.TALK.BLUE, wn.text);
                    }
                }
            }
        }

        //procesClaims
        await $wdc.processClaims(this, entity.claims);
    }


    /**
     * 
     * @param {*} $propObj 
     * @param {wikidata_neuron} $wn 
     */
    addWdRelon=function($propObj, $wn) {
        if (this.wd_relons==null) this.wd_relons = [];
        this.wd_relons.push({propObj:$propObj, to_wn:$wn});
        if ($wn.name!="") {
            ANNE.TALK.printf_console(ANNE.TALK.CYAN, "addWdRelon", ANNE.TALK.YELLOW, " to value ", ANNE.TALK.GREEN, $wn.name);
        }
        else if ($wn.resolveNOBJ!=null) {
            ANNE.TALK.printf_console(ANNE.TALK.CYAN, "addWdRelon", ANNE.TALK.YELLOW, " to value ", ANNE.TALK.GREEN, $wn.resolveNOBJ.text);
        }

        //but only if it is an entity value.
        else {
            console.log($propObj);
            if ($propObj.to_type==TO_TYPE_ANNE) {
                ANNE.TALK.printf_console(ANNE.TALK.CYAN, "addWdRelon", ANNE.TALK.YELLOW, " to value (UNMAPPED TO_TYPE_ANNE) ", ANNE.TALK.BLUE, $wn.wd_id);
            }
            else {
                ANNE.TALK.printf_console(ANNE.TALK.CYAN, "addWdRelon", ANNE.TALK.YELLOW, " to value", ANNE.TALK.GREEN, $wn.wd_id);
            }
        }
    }

    addUnhandledProp=function($wdobj) {
        if (this.wd_unhandled_props==null) this.wd_unhandled_props = [];
        this.wd_unhandled_props.push($wdobj);
    }


    extraFromSearch=function() {
    }


    /**
     * @param {string} - nidlongstring of nid
     * we have resolve this wikidata neuron via other means and now are setting the resolution.
     */
    setNid = function($nid) {
        if (this.resolvedNOBJ!=null) {
            console.log("already resolved");
            return;
        }
        this.resolvedNOBJ = ANNE.MEMG.getOrMakeIfNew($nid);
    }


    resolveNid=function() {
        if (this.resolvedNOBJ!=null) {
            console.log("already resolved");
            return;
        }
    }
}

class mappper_relon extends wikidata_neuron {
    constructor($name, $wd_id, $resolvedNOBJ, $relon_type, $relon_to_type) {
        super($name, $wd_id, $resolvedNOBJ);
        this.relon_type = $relon_type;
        this.relon_to_type = $relon_to_type;
        //if the relon produces a dual (backlink) relon.  
        //see spouse.   a->has->spouse->b  b->has->spouse->a
        this.dual = null;
        this.dual_type = null;
    }
}

class wikidata_search {
    constructor($searchString) {
        var searchObj = ANNE.SEARCH.parseSearchForContext($searchString);

        this.search = searchObj.search;
        this.inline_context = searchObj.inline_context;

        this.status = "notset";
        this.wd_neurons = new arraymap();
        this.wd_neurons.setMapKey("wd_id");

        this.winnerWN = null;
        this.winnerWNs = [];
        this.high_score_wn = -1;
    }


    /**
     * @param {*} $aRez 
     * @param {*} $knownNOBJ - if set, we known which one we want so match known data in the knownNOBJ to disambig
     */
    addResults=function($aRez, $knownNOBJ=null) {
        console.log("adding "+$aRez.length+" results");
        

        for (var a=0; a < $aRez.length; a++) {
            var rez = $aRez[a];
            if (this.wd_neurons.contains(rez.id)==true) {
                console.log("we already have this");
                continue;
            }

            //use rez.desc to disambig against our inline context?
            var wn = new wikidata_neuron(rez.label, rez.id);
            if (rez.pageid!=undefined) wn.wd_pageid = rez.pageid;
            if (rez.desc!=undefined && rez.desc.length>0) wn.wd_desc = rez.desc;
            if (rez.url!=undefined) wn.wd_url = rez.url;

            if (this.inline_context.length) {
                for (var c=0; c < this.inline_context.length; c++) {
                    var context = this.inline_context[c];
                    if (wn.wd_desc!="" && wn.wd_desc.indexOf(context)!=-1) {
                        wn.inline_context_score+=1;
                    }
                }
            }

            if (wn.inline_context_score > this.high_score_wn) {
                this.winnerWN = wn;
                this.winnerWNs = [];
                this.high_score_wn = wn.inline_context;
            }
            else if (wn.inline_context_score==this.high_score_wn) {
                this.winnerWN=null;
                this.winnerWNs.push(this.winnerWN);
            }
            this.wd_neurons.addU(wn);
        }
    }
}


class wikidata_connector  {
    constructor($CIRCUIT, $nObj=null,  $targetClassNOBJ=null, $callbackDone=null) {
        this.aec = new anne_external_connector("", "wdc", $CIRCUIT);
        
        //resoled NOBJ, known if set.
        this.resolvedNOBJ = $nObj;
        
        this._callback_done = $callbackDone;
        this._new_external_data = 0;

        this.new_akas = [];

        //known target class of the neuron.
        this.targetClassNOBJ = $targetClassNOBJ;
        if (this.targetClassNOBJ==null && this.resolvedNOBJ!=null) {
            this.targetClassNOBJ = this.resolvedNOBJ.getParent();
        }


        this.name = "";

        WIKIDATA.INIT_MAP();
    }


    //https://www.wikidata.org/w/api.php?action=wbgetentities&ids=Q29250&props=claims&format=json&origin=*
    //https://www.wikidata.org/w/api.php?action=wbgetentities&ids=Q29250&format=json&origin=*
    getName = async($wn)=>{
        console.log("get name for wn_id:"+$wn.wd_id);
        var quick = await this.rawGetByIDs($wn.wd_id);
        if (quick==null) return "";
        if (quick.entities==undefined) return "";
        var e = quick.entities[$wn.wd_id];
        if (e==undefined || e==null) return "";
        if (e.labels==undefined || e.labels.en==undefined) return null;
        $wn.name = e.labels.en.value;
        return e.labels.en.value;
    }


    /**
     * First try to resolve ANNE.
     * If ANNE, then get wd_id.  If no wd_id, get it.
     * 
     * If no ANNE, explore via wd.  Make new ghost based on results.
     * @returns 
     */
    resolveAndGet=async($name="")=> {
        var nameObj = ANNE.SEARCH.parseSearchForContext($name);
        if ($name!="") this.name = nameObj.search;

        if ($bes.length==0 && this.targetClassNOBJ!=null) {
            $bes = [this.targetClassNOBJ.nid];
        }

        //nameObj.inline_context;

        //if we already know the neuron, just go to wiki, don't do a search.
        if (this.resolvedNOBJ!=null) {
            if (this.resolvedNOBJ.ghost==false) {
                await ANNE.API.getAllOuts( this.resolvedNOBJ.nid,true, [0,1,2,3,4,5,6,7,8,9,10]);
            }

            //var dataObj = await ANNE.WIKIPEDIADATA.getExternalWikiData(this.name, this.intended_wiki_type, this.aec);
            //await this.neuronFromWiki(dataObj);
            return null;//dataObj;
        }

        var searchObj = ANNE.SEARCH.makeSearch(this.name, $bes);
        searchObj.thresh = 0.9;
        var bDOPRE = true; //ok to end by finding match

        if(nameObj.inline_context.length>0) {
            searchObj.context = nameObj.inline_context;
        }
        
        var aMatches = await CHASE.anne_global_search(searchObj, bDOPRE);

        if (aMatches.length>1) {

            //if we have a context and a BRAIN obj, snag the hottest per context.
            if (ANNE.BRAIN!=null && searchObj.context!=undefined && nameObj.inline_context.length>0) {
                this.resolvedNOBJ = ANNE.BRAIN.getHottestByContextStr(searchObj.context, aMatches);
            }
            else this.resolvedNOBJ = aMatches[0];
            await ANNE.API.getAllOuts( this.resolvedNOBJ.nid,true, [0,1,2,3,4,5,6,7,8,9,10]);
        }
            
        else if (aMatches.length==1) {
            this.resolvedNOBJ = aMatches[0];
            if (this.resolvedNOBJ.ghost==false) await ANNE.API.getAllOuts( this.resolvedNOBJ.nid,true, [0,1,2,3,4,5,6,7,8,9,10]);
        }

        //make a new targetClassNOBJ straight away.
        else {
            this.resolvedNOBJ = await GHOSTDATA.makeGhostNeuron(this.targetClassNOBJ, this.name, NTYPE_INST);
            this.bNEW=true;
            this._new_external_data+=1;
        }
        
        var dataObj = await ANNE.WIKIPEDIADATA.getExternalWikiData(this.name, this.intended_wiki_type, this.aec);
        
        dataObj._targets = $targetsOnly;
        await this.neuronFromWiki(dataObj);

        return dataObj;
    }


    /**
     * @param {wikidata_neuron} $wn
     * @param {*} $claims 
     * @param {boolean} $bEXPLORE - explore the reln values we dont have mapped. 
     */
    processClaims = async($wn, $claims, $bEXPLORE=true)=>{
        var keys = Object.keys($claims);
        for (var k=0; k < keys.length; k++) {
            var key = keys[k];
            var claims = $claims[key];
            
            //we have this mapped and want it.
            if (WIKIDATA._prop_map.get(key)!=null) {
                var prop = WIKIDATA._prop_map.get(key);
                //there is a claim of this prop.
                //process the base neuron with the aClaims and the prop
                await this.processGeneral(key, $wn, claims, prop);
            }

            //for now, this helps me see what is available and i can add it to maps at the top.
            //not fun, as wikidata
            else if ($bEXPLORE==true) {
                if (WIKIDATA._neurons.get(key)==null) {
                    var wn = new wikidata_neuron("", key);
                    var reln_name = await this.getName(wn);
                    ANNE.TALK.printf_console(ANNE.TALK.CYAN, key, ANNE.TALK.YELLOW, " prop name ", ANNE.TALK.GREEN, " "+reln_name);
                    $wn.addUnhandledProp({wd:wn, name:wn.name, claims:claims});
                }

                //don't process these yet, since we don't even know that we want them!  (we don't have prop either!)
                //await this.processGeneral($wn, claims, prop);
            }
        }

        //loop the props we are interested in.
        /*
        for (var p=0; p < WIKIDATA._prop_map.arr.length; p++) {
            
            var prop = WIKIDATA._prop_map.arr[p];

            //there is a claim of this prop.
            if ($claims[prop.wd_id]!=undefined) {
                var aClaims = $claims[prop.wd_id];

                //process the base neuron with the aClaims and the prop
                await this.processGeneral($wn, aClaims, prop);
            }
        }
        */

    }


    /**
     * @param {wikidata_neuron} $fromWN
     * @param {*} $aClaims 
     * @param {*} $propObj 
     */
    processGeneral = async($key, $fromWN, $aClaims, $propObj)=> {
        ANNE.TALK.printf_console(ANNE.TALK.CYAN, "processGeneral", ANNE.TALK.YELLOW, $key, ANNE.TALK.GREEN, " propname: "+$propObj.name);
        
        console.log($aClaims);
        
        for (var a=0; a < $aClaims.length; a++) {
            var claim = $aClaims[a];
            var toWN = this.processSingleClaim(claim);
            $fromWN.addWdRelon($propObj, toWN);
        }
    }


    /**
     * 
     * @param {*} $obj 
     * @returns 
     */
    processSingleClaim=function($obj, $qskips=[]){
        if ($obj.mainsnak==undefined) return;
        var wd_type = $obj.type; //"statement"
        var ms = $obj.mainsnak;
        if (ms.snaktype=="value") {
            var val = ms.datavalue.value;

            if (val.id!=undefined && $qskips.indexOf(val.id)!=-1) {
                return null;
            }

            var wdmapperval = "";
            var wdtextval = "";

            var parentClassNOBJ = null;

            ms.datatype = ms.datatype.toLowerCase();

            //we have this in our mapped wikidata neurons.
            if (ms.datatype=="wikibase-item") {
                if (WIKIDATA._neurons.get(val.id)!=null) return WIKIDATA._neurons.get(val.id);
                wdmapperval = val.id;
                wdtextval = "";
            }
            else if(ms.datatype == "wikibase-entityid") {
                if (WIKIDATA._neurons.get(val.id)!=null) return WIKIDATA._neurons.get(val.id);
                wdmapperval = val.id;
                wdtextval = "";
            }
            else if(ms.datatype == "wikibase-property") {
                if (WIKIDATA._neurons.get(val.id)!=null) return WIKIDATA._neurons.get(val.id);
                wdmapperval = val.id;
                wdtextval = "";
            }
            
            //test how they do dates, as neurons or strings?
            else if (ms.datatype=="time") {
                console.log("datavalue",ms.datavalue);
                console.log("val:", val);
                var parts = val.calendarmodel.split("/");
                if (parts[parts.length-1]!="Q1985727") {
                    console.log("only gregorian dates handled");
                    return null;
                }
                var pObj = WIKIDATA.parseDateForAnne(val);
                if (pObj==null) return null;
                //{date:datestring, time:timestring, parentClassNOBJ:parent}
                wdmapperval = pObj.date;
                wdtextval = pObj.date;

                //if year, century, decade, or millenia, we'll use only integer neuron
                //and then use relon of century, decade, or millenia.
                parentClassNOBJ = pObj.parentClassNOBJ; 
            }

            else if (ms.datatype="external-id") {
                console.log("datavalue",ms.datavalue);
                console.log("val:", val);
                wdmapperval = val;
                wdtextval = val;
            }
            else if (ms.datatype="string") {
                console.log("datavalue",ms.datavalue);
                console.log("val:", val)
                wdmapperval = val;
                wdtextval = val;
            }
            else if (ms.datatype="url") {
                console.log("datavalue",ms.datavalue);
                console.log("val:", val)
                wdmapperval = val;
                wdtextval = val;
            }
            else if (ms.datatype="quantity") {
                console.log("datavalue",ms.datavalue);
                console.log("val:", val);
                wdmapperval = val;
                wdtextval = val;
            }
            else if (ms.datatype=="globecoordinate") {
                console.log("datavalue",ms.datavalue);
                console.log("val:", val);
                wdmapperval = val;
                wdtextval = val;
                return null;
            }
            else {
                console.log("unhandled datatype");
                if (ms.datatype=="wikibase-lexeme") {
                    console.log("skipping wikibase-lexeme", ms);
                    return null;
                }
                else if (ms.datatype=="wikibase-sense") {
                    console.log("skipping wikibase-sense", ms);
                    return null;
                }
                else if (ms.datatype=="wikibase-form") {
                    console.log("skipping wikibase-form", ms);
                    return null;
                }
                else if (ms.datatype=="entity-schema") {
                    console.log("skipping entity-schema", ms);
                    return null;
                }
                else if (ms.datatype=="monolingualtext") {
                    console.log("skipping monolingualtext", ms);
                    return null;
                }
                else if (ms.datatype=="commonsMedia") {
                    console.log("skipping commonsMedia", ms);
                    return null;
                }
                else if (ms.datatype=="geo-shape") {
                    console.log("skipping geo-shape", ms);
                    return null;
                }
                else if (ms.datatype=="tabular-data") {
                    console.log("skipping tabular-data", ms);
                    return null;
                }
                else if (ms.datatype=="math") {
                    console.log("skipping math", ms);
                    return null;
                }
                else if (ms.datatype=="musicalnotation") {
                    console.log("skipping MusicalNotation", ms);
                    return null;
                }
                
            }

            var wn = new wikidata_neuron(wdtextval, wdmapperval);
            wn.wd_snaktype = ms.snaktype;
            wn.wd_dtype = ms.datatype;
            wn.wd_et = val['entity-type'];
            wn.wd_num_id = val['numeric-id'];
            wn.wd_type = wd_type;
            if (parentClassNOBJ!=null) wn.classNOBJ = parentClassNOBJ;
            return wn;
        }

        else if (ms.snaktype=="novalue") {
            console.warn("snaktype: novalue", ms);
            return null;
        }

        else if (ms.snaktype=="somevalue") {
            console.warn("snaktype: somevalue", ms);
            return null;
        }

        else if (ms.snaktype=="globally-unique-id") {
            console.warn("snaktype: globally-unique-id", ms);
            return null;
        }
    }


    addToConsole=function($msg) {
        this.aec.addToConsole($msg);
    }

    /**
     * 
     * @param {*} $somethingtosearch 
     * @param {*} $knownNOBJ - a search to get the wd_id
     * @returns 
     */
    exploreForID=async($somethingtosearch, $knownNOBJ=null)=>{
        if (window.wdk==undefined) return;
        var ws = new wikidata_search($somethingtosearch);
        if ($knownNOBJ==null) this.addToConsole("searching wikidata for "+$somethingtosearch);
        else this.addToConsole("searching wikidata for "+$somethingtosearch+" known nid:"+$knownNOBJ.nid);

        const url = window.wdk.searchEntities({ search: ws.search });
        try {
            console.log("fetching url: "+url);
            const response = await fetch(url);
            if (!response.ok) {
                this.addToConsole("HTTP error! status:"+response.status);
                throw new Error(`HTTP error! status: ${response.status}`);
            }
            const data = await response.json();
            console.log(data);
            if (data.search!=undefined) {
                this.addToConsole("adding results");
                ws.addResults(data.search, $knownNOBJ);
            }
            ws.status="good";
            console.log(ws);
            return ws;
        } 

        catch (error) {
            console.error('Error fetching data:', error);
            ws.status="error";
            console.log(ws);
            return ws;
        }
    }



    /**
     * 
     * @param {*} $ID
     * @returns 
     */
    getAndProcessDataByID=async($ID="Q22686")=>{
        var data = await this.rawGetByIDs([$ID]);
        if (data==null) return;
        var entities = data.entities;
        if (entities==null) return;
        if (entities[$ID]==null) return;
        if (entity.labels!=undefined) {
            console.log(entity.labels);
        }
        var entity = entities[id];
        //if (entity.claims!=null) {
            //await this.process(entity.claims);
        //}
        return;
    }


    /**
     * 
     * @param {*} $IDS 
     * @returns 
     */
    rawGetByIDs=async($IDS=["Q22686"])=>{
        if (window.wdk==undefined) return;
        const url = window.wdk.getEntities({ ids: $IDS });
        console.log(url);
        try {
            const response = await fetch(url);
            if (!response.ok) {
                throw new Error(`HTTP error! status: ${response.status}`);
            }
            const data = await response.json();
            //console.log(data);
            return data;
        } 

        catch (error) {
            console.error('Error fetching data:', error);
        }
    }


}