const PATHTYPE_ANNE = 1;
const PATHTYPE_UDF = 2;
const RENDERTYPE_BASE = 1;
const RENDERTYPE_STANDARD = 2; 
const RENDERTYPE_FLOATER = 3;
const RENDERTYPE_TAB = 4;
const RENDERTYPE_MODAL = 5; 
const RENDERTYPE_INJECT = 6;

const TYPE_NAME = -1;

const TYPE_SYS = 0;             //Sytem info for neuron (sponsored_by, annother, owned_by).  Builds extra indexes.
const TYPE_BE = 1;              //semantic shothand for BE DIMENSION.  CLASS/INST RELNS create a DAG.
const TYPE_PARAM = 2;           //semantic shorthand for param (class attribute/property)
const TYPE_HAS = 3;             //semantic shorthand for param fulfilling has value (attribute/property)
const TYPE_AWARENESS = 4;       //for awareness space (emotions/feelz)

const TYPE_EXPERIENCE = 5;      //for experience (temporal) neuron space

//CLUMPS can function as a TYPE_BE, TYPE_HAS, OR TYPE_SYS.
//those Types are actually short-hand for clumps but they can get away with
//only using one TX to specify the triplet info.
//A full clump is for specifying more than the standard 3 pieces of semantic info
//in a single ANNE TX.
//** For ease of use we can put a TYPE_XXX in the clump
//to clue our interpreter to what dimension the clump lives in.
//The other CLUMPS of episodic, BRAIN and EXPERIENCE are already their own dimension
//so this is only relevant for semantic clumps that are handled in semantic memory space.
const TYPE_CLUMP = 6;

const TYPE_SYS_MSG = 7;        //only relevant to coms, not the neuron
const TYPE_BRAIN = 8;          //for neural-net space wirings a->wire->b
const TYPE_ROLE = 9;           //for specifiying  semantic roles inside of a clump

const TYPE_METAN = 10;         //for specifying meta information about roles used in a clump

const TYPE_EPISODIC = 11;      //for documenting a self-sensory/vitals data at run-time in a clump.
const TYPE_INTERNAL_PAY = 17;  //keyless neurons paying out.
const TYPE_BROADCAST_PAY = 18; //standard broadcast payment.

const TYPE_BRAIN0 = 20;        //20+ are for sequence in the brain dimension

//100+ are available for brain type/custom use.. assumed to be order if more than one in a neuron's outs.
const TYPE_BRAIN120 = 120;     //used in ASDF-> annode relons to store trigger pats. (can use how u like in other neurons)
const TYPE_BRAIN121 = 121;     //myelination measure (a->b->[to_type_number]#c)

const TO_TYPE_SEED = 0;        //relon to is an encrypted seed (all seed are encrypted)
const TO_TYPE_ANNE = 1;        //relon to is a .nid, the idenfier of another neuron.  This is a legit graph triplet.
const TO_TYPE_TXID = 2;        //relon to is a txid.
const TO_TYPE_URL = 3;         //relon to is a url.

//relon to is a number.  Note: our ANNODE stores numerics in a LONG field
//if a non-long/int is needed use the TO_TYPE_STRING

const TO_TYPE_NUMBER = 4;       
const TO_TYPE_DATE = 5;         //relon to is a date in the ANNE date format.    
const TO_TYPE_DATETIME = 6;     //relon to is a datetime in the ANNE datetime format.
const TO_TYPE_STRING = 7;       //relon to is a string

const TO_TYPE_TEXT = 7;         //alias in case u conflate

const TO_TYPE_BLOB = 8;         //relon to is a blob (bad donkey, not welcome in ANNE)
const TO_TYPE_TIME = 9;         //relon to is a time in the ANNE time format.
const TO_TYPE_BOOLEAN = 10;     //relon to is a boolean string which will be converted to true/false at runtime
const TO_TYPE_FILEHASH = 11;    //relon to is a filehash
const TO_TYPE_IPV4 = 12;    //relon to is a IPV4 address
const TO_TYPE_IPV6 = 13;    //relon to is a IPV6 address

//encrypted versions of the above + 30 in value from its respective type.
//expansion to 30 base types possible (don't foresee this as needed
//in a perfect-brain world, all are references to either neurons or other TXIDs

const TO_TYPE_ENCR_ANNE = 31;                
const TO_TYPE_ENCR_TXID = 32;
const TO_TYPE_ENCR_URL = 33;        
const TO_TYPE_ENCR_NUMBER = 34;
const TO_TYPE_ENCR_DATE = 35;
const TO_TYPE_ENCR_DATETIME = 36;
const TO_TYPE_ENCR_STRING = 37; 
const TO_TYPE_ENCR_BLOB = 38;
const TO_TYPE_ENCR_TIME = 39;
const TO_TYPE_ENCR_BOOLEAN = 40;
const TO_TYPE_ENCR_FILEHASH = 41;
const TO_TYPE_ENCR_IPV4 = 42;
const TO_TYPE_ENCR_IPV6 = 43;

const NTYPE_CLASS = 0;
const NTYPE_INST = 1;
const NTYPE_UNKNOWN = 2;
const NTYPE_CLUMP = 3;

const ORIGIN_CHAIN = 0;         //a real TX on-chain
const ORIGIN_NODE = 1;          //an emulated neuron created by the node
const ORIGIN_NODE_PRELOAD = 2;  //an emulated neuron preloaded by the node (with a real seed)
const ORIGIN_OWNERPROOF = 3;    
const ORIGIN_CONTRACTS = 4;     //emulated from contracts
const ORIGIN_KEYLESS = 5;
const ORIGIN_GLOBAL = 6;
const ORIGIN_CUSTODIAL  = 8;
const ORIGIN_EXTERNAL_ML = 12;  //if a ML is joined w/ the DB, that data is ML.

const RORIGIN_ANNODE = 0;         //origination made on the ANNODE in ANNE hypergraph only
const RORIGIN_BROADCAST = 1;    //origination broadcast as a TX
const RORIGIN_COLLAB_BROADCAST = 2;        //collab broadcast (waiting approval)
const RORIGIN_GHOST = 3;        //origination user local

const ANNE_ACCTID = "7777770163980908667";    
const ANNE_ADMIN_ACCTID = "7777771543345157118";
const ANNE_ADMIN_BACKUP_ACCTID = "7777044340734581180";

const TARGET_NID = 1;
const TARGET_TXID = 2;

const TXMODE_GHOST = 0;             //store to local DB only
const TXMODE_ANNE = 1;              //instant send

const CTYPE_COLLAB_ONLY = 0;
const CTYPE_COLLAB_HANAI = 1;
const CTYPE_PENDING_ANJ = 2;
const CTYPE_HANAI_ONLY = 3;
const CTYPE_MERGE_REQUEST = 4;

//Keyless neurons allow for easy creation of child relons w/o the management of another key
//They can be collaborated w/ by request or easy delegation rights set to other approved authors.
//Note: ANNE globals (date, numbers, locations, keywords) are keyless neurons and are created on-demand.
const NEWNEURONMODE_KEYLESS = 0;  

//Key'd neurons provide real a real/key that is only known by the identity that created it.
//Key'd neurons can be collaborated w/ by request but they cannot be delegated since only the
//Key holder can broadcast transactions from the key'd neuron.
//Note: ANNE preloaded neurons are key'd as they need pre-known neuron IDs to reference them in the base-protocol.
const NEWNEURONMODE_KEYED = 1;        

const NEURON_NOT_FROZE = 0;
const NEURON_FROZE = 1;

const LINKS_NEURON_MODAL = 1;            //quick neuron summary page
const LINKS_NEURON_EMBED = 2;            //full neuron summary page
const LINKS_NEURON_DS = 3;               //data-source view of neuron's BE dimension
const LINKS_NEURON_TBL_VIEW = 4;         //table view of neuron's relons
const LINKS_NEURON_TREE = 5;             //tree view of a neuron's BE dimension
const LINKS_NEURON_ANNET = 6;            //graph explore of neurons
const LINKS_NEURON_ANDY = 7;             //ADMIN AREA
const LINKS_NEURON_ALEXANNDRIA = 8;      //library of your neurons  (can also find others)

const LINKS_NEURON_USER = 10;            //user profile page.
const LINKS_NEURON_INJECT_CB = 11;       //injection link w/ callback
const LINKS_NEURON_BLOCKCHAIN = 12;      //a link to the 3d viewer of blockchain data.
const LINKS_NEURON_DOSSIER = 14;         //detailed dossier view
const LINKS_NEURON_BRAIN = 16;           //decide if rememberz/forgetz is on/off
const LINKS_NEURON_ANNULET_OF_TRUST = 17;

const LINKS_RELON_ANNE = 20;             //an ANNE relon of somekind.
const LINKS_TX_BLOCKEXPLORER = 21;       //not an ANNE relon, a standard blockchain TX

//these mirror the value sin the node L1 code.
const BLOCK_HEADER_LENGTH = 232; //TODO: this may change if we add sm in the block...
const MAX_BALANCE_ANNE = 2200000000000; 
const ONE_ANNE = 100;
const MIN_ANNIES_SEND = 1000;                   //not enforced.               
const DELETE_WHEN_MINER_FIXED = ONE_ANNE;
const MAX_BALANCE_NQT = MAX_BALANCE_ANNE * ONE_ANNE;
const INITIAL_BASE_TARGET = 38650387592;        //irrelevant
const BLOCK_REWARD_START = 1000_000;
const BLOCK_REWARD_CYCLE = 21240;
const BLOCK_REWARD_CYCLE_PERCENTAGE = 99;
const BLOCK_REWARD_18500_HEIGHT = 8453521;    //mining reward sunset mode 
const BLOCK_REWARD_SUNSET_AMOUNT = 18500;
const BLOCK_REWARD_LIMIT_HEIGHT = 14655601;   //mining reward end block 
const BLOCK_REWARD_LIMIT_AMOUNT = 0;            //0 reward after that
const SUPER_CYCLE = BLOCK_REWARD_CYCLE * 13;    //every 13 cycles is a super block.
const BLOCK_REWARD_SOLO_CUT = 51;
const BLOCK_REWARD_SHARE_CUT = 0.07;             //7% of the block reward.
const BLOCK_REWARD_SHARE_BROADCAST_FEE = 0.02;   //2% of the 7% cut.

const MAINNET = "mainnet";
const TESTNET = "testnet";

const MEDIA_AUDIT_NOTRUN = 0;
const MEDIA_AUDIT_PASSED = 1;
const MEDIA_AUDIT_PASSED_W_NEWDATA = 2;
const MEDIA_AUDIT_MISSING_DATA = -2; //missing data
const MEDIA_AUDIT_CONFLICTING_TYPE = -3; 
const MEDIA_AUDIT_CONFLICTING_SIZE = -4;
const MEDIA_AUDIT_FINGER_DOES_NOT_MATCH = -5;

const FILEDATASTATUS_NOT_SET = -1;
const FILEDATASTATUS_REQUEST_ERROR = 6;
const FILEDATASTATUS_ACQUIRED = 7;
const FILEDATASTATUS_PENDING_ANTOR = 277;
const FILEDATASTATUS_CANNOT_INIT_ANTOR = 287;
const FILEDATASTATUS_ANTOR_OFF = 297;

const QT_PERCENT = 0;
const QT_SPECIFIC = 1;
const SRC_FUND = 0;
const SRC_OWNER = 1;
const SP_MIN_PAYMENT = 1_000_000;  //new min, 1m annies (10k annecoin)


/************** (from file ./js/base/consts.js)**/ 

class anne_singleton {
    constructor(){
        this._bINIT = false;
    }
    set_init = function ($REF) {
        this._bINIT=true;
        Object.freeze($REF);
    }
    is_init = function () {
        if (this._bINIT==true) return true;
        return false;
    }
}

/************** (from file ./js/base/anne_singleton.js)**/ 

/**
 * Basically, an array of objects, and a corresponding map to help manage their
 * addition/tracking/grouping
 */
class arraymap {
    constructor($mapKey="") {
        this.arr = [];
        this.map = new Object();
        this.mapKey = $mapKey;
        this.arr_other = [];

        this.secondary_mapKey = "";
        this.secondary_map = null;
        this.secondary_mapKey_forceLowerCase = false;
    }

    //good enough...
    copy = function($arraymap) {
        this.mapKey = $arraymap.mapKey;
        this.secondary_mapKey = $arraymap.secondary_mapKey;

        var keys = Object.keys($arraymap.arr);
        for (var k=0; k < keys.length; k++) {
            this.arr[keys[k]] = $arraymap.arr[ keys[k]]
        }
        if ($arraymap.secondary_map!=null) {
            var keys2 = Object.getKeys($arraymap.arr_other);
            for (var k=0; k < keys2.length; k++) {
                this.arr_other[keys2[k]] = $arraymap.arr_other[ keys2[k]]
            }
        }
    }


    setMapKey = function($mk) { this.mapKey = $mk;}
    setSecondaryIndex = function($2ndkey, $bMAKELOWER=false) {
        this.secondary_mapKey = $2ndkey;
        this.secondary_mapKey_forceLowerCase = $bMAKELOWER;
        this.secondary_map = new Object();
    }

    //aliases
    setPrimaryKey = function($mk) { this.setMapKey($mk); }
    setSecondaryKey = function($2ndkey, $bMAKELOWER=false) { this.setSecondaryIndex($2ndkey, $bMAKELOWER);}


    addArray = function($inputArr) {
        for(var i=0; i<$inputArr.length; i++) {
            this.addU($inputArr[i]);
        }
    }

    contains = function($mapKey) {
        if (this.map[$mapKey]==undefined) return false;
        return true;
    }


    //keep obj so we have ptr in mem still in other arrays (like tbls.)
    updateU_oldkey = function($oldObj, $newObj) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }

        var key = $oldObj[this.mapKey];
        if (this.map[key]==undefined) {
            console.log("no mapped object at key '"+key+"' to update")
            return;
        }

        var objToUpdate = this.map[key];
        delete this.map[key];
        var newkey = $newObj[this.mapKey];
        
        var keys = Object.keys($newObj)
        
        //loop through the new object we want to use to update our existing.
        for (var k=0; k < keys.length; k++) {
            if (keys[k]==this.mapKey) continue;
            var key = keys[k];  
            objToUpdate[key] = $newObj[key];
        }
        this.map[newkey]=objToUpdate;
    }


    getSlice=function($from_index, $to_index) {
        var arrx = [];
        for (var f=$from_index; f < $to_index; f++) {
            var obj = this.get(f);
            if (obj==null) {
                console.log("no f:"+f+" obj in arr map");
            }
            arrx.push (obj);
        }

        //we could just try to slice it 
        return arrx;
    }


    getSlice2=function($from_index, $to_index) {
        var arrx = [];
        for (var f=$from_index; f < $to_index; f++) {
            var obj = this.get2(f);
            if (obj==null) {
                //console.log("no f:"+f+" obj in arr map");
                //console.warn("big problem");
                continue;
            }
            arrx.push (obj);
        }

        //we could just try to slice it 
        return arrx;
    }


    /**
     * Use this when mapKey is set.
     * @param {*} $obj 
     */
    updateU = function($obj, $skip_keys=[]) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }

        var key = $obj[this.mapKey];
        if (this.map[key]==undefined) {
            console.log("no mapped object at key '"+key+"' to update")
            return;
        }

        var objToUpdate = this.map[key];

        var keys = Object.keys($obj)
        
        //loop through the new object we want to use to update our existing.
        for (var k=0; k < keys.length; k++) {
            if ($skip_keys.indexOf( keys[k])!=-1) continue;

            var key = keys[k];

            objToUpdate[key] = $obj[key];
        }
    }

    
    updateU2 = function($obj, $skip_keys=[]) {
        if ($obj==undefined) {
            console.warn("dude, what are you doing?");
            return;
        }
        if (this.secondary_mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }
        
        var key2 = $obj[this.secondary_mapKey];
        if (this.secondary_mapKey_forceLowerCase==true) {
            key2 = key2.toLowerCase();
        }

        if (this.secondary_map[key2]==undefined) return;
        
        var objToUpdate = this.secondary_map[key2];

        var keys = Object.keys($obj)
        for (var k=0; k < keys.length; k++) {
            if ($skip_keys.indexOf( keys[k])!=-1) continue;
            var keyx = keys[k];
            objToUpdate[keyx]=$obj[keyx];
        }
    }


    getKeys = function() {
        return Object.getKeys(this.map);
    }

    getU = function($obj) {
        if (this.mapKey==null || this.mapKey=="") {
            console.log("warn must have a mapKey set to use getU");
            return null;
        }
        var key = $obj[this.mapKey];
        return this.get(key);
    }

    get = function($mapKey) {
        if (this.mapKey==null || this.mapKey=="") {
            console.log("warn must have a mapKey set to use getU");
            return null;
        }
        if (this.map[$mapKey]!=undefined) return this.map[$mapKey];
        return null;
    }

    get2 = function($mapKey2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use get2 without a secondary index set!");
            return null;
        }
        if (this.secondary_map[$mapKey2]==undefined) return null;
        return this.secondary_map[$mapKey2];
    }

    get2_partial = function($mapKey2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use get2 without a secondary index set!");
        }
        var txt =$mapKey2;
        if (this.secondary_mapKey_forceLowerCase==true) txt = $mapKey2.toLowerCase();
        var matches = [];
        var keys = Object.keys(this.secondary_map);
        for (var k=0; k < keys.length; k++) {
            if (keys[k].indexOf(txt)==0 && keys[k].length>txt.length){ 
                matches.push( this.secondary_map[ keys[k]]);
            }
        }
        return matches;
    }


    addIndex2 = function($pkOrObj, $key2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use addIndex2 without a secondary index set!");
        }

        //if first param string, use it as a primary keyvalue of the .map
        //to get the obj. Use that object's key2 to set the secondary map.
        if (typeof $pkOrObj==="string") {
            var obj = this.get($pkOrObj);
            if (obj==null) return null;
            this.secondary_map[$key2]=obj;
        }

        //standard, the pkOrObj is an object, and the $key2 is the value of the secondary mapKey
        //for that object.
        else {

            //audit.
            if ($key2!=$pkOrObj[this.secondary_mapKey]) {
                console.log("houston we have a problem!");
                return;
            }
            this.secondary_map[$key2]=$pkOrObj;
        }
    }

    quickAddMK = function($obj) {
        if (this.mapKey=="") {
            console.warn("U can't use quick add without a quickAddMK set.");
            return;
        }
        if (this.map[$obj[this.mapKey]]==undefined) {
            this.map[$obj[this.mapKey]]=$obj;
            this.arr.push($obj);
        }
    }

    quickAdd = function($obj) {
        this.arr.push($obj);
    }
    
    addUnique = function($obj, $mapKeyVal) {
        if ($mapKeyVal==undefined) {
            console.warn("fix this!")
            return;
        }
        if (this.map[$mapKeyVal]==undefined) {
            this.map[$mapKeyVal]=$obj;
            this.arr.push($obj);
            this.addU2($obj);
            return true;
        }
        else this.addU2($obj);
    
        return false;
    }

    /**
     * Use this when mapKey is set.
     * @param {*} $obj 
     */
    addU = function($obj) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use addU");
            return;
        }
        if (this.map[$obj[this.mapKey]]==undefined) {
            this.map[$obj[this.mapKey]]=$obj;
            this.arr.push($obj);
            this.addU2($obj);
            return true;
        }

        //still add secondary key if relevant.
        else this.addU2($obj);
    
        return false;
    }


    addU2=function($obj) {
        if (this.secondary_map!=null) {
            var key2 = $obj[this.secondary_mapKey];
            if (key2==undefined) return true;
            if (this.secondary_mapKey_forceLowerCase==true) key2 = key2.toLowerCase();  
            var val2 = $obj[key2];
            if (val2==undefined) this.addIndex2($obj, key2);
        }
    }

    
    removeNoKey = function($index) {
        this.arr.splice($index, 1); 
    }

    remove = function($mapKey, $prop="") {
        if ($prop!="") this.prop = $prop;
        if (this.map[$mapKey]!=undefined) {
            delete this.map[$mapKey];
            for (var i=0; i < this.arr.length; i++) {
                if (this.arr[i][this.prop]==$mapKey) {
                    this.arr.splice(i, 1);
                    return;
                }
            }
        }
    }

    getArrayFromAltProp = function($altprop, $altpropvalue) {
        var matches = [];
        for (var i=0; i < this.arr.length; i++) {
            if (this.arr[i][$altprop]==$altpropvalue) {
                matches.push(this.arr[i]);
            }
        }
        return matches;
    }

    removeU = function($mapKeyValue) {
        if (this.map[$mapKeyValue]!=undefined) {
            delete this.map[$mapKeyValue];
            for (var i=0; i < this.arr.length; i++) {
                if (this.arr[i][this.mapKey]==$mapKeyValue) {
                    this.arr.splice(i, 1);
                    return true;
                }
            }
        }
        return false;
    }


    /**
     * 
     * @param {*} $obj 
     * @param {*} $objkey 
     * @returns 
     */
    addArrAM = function($obj, $objkey="txid") {
        if (this.mapKey=="") {
            console.warn("must have map key set to use addArrAM");
            return;
        }
        var mk = $obj[this.mapKey];
        if (this.map[mk]==undefined) {
            this.map[mk]=new arraymap();
            if (this.map[mk].addUnique($obj, $objkey)==true) {
                this.arr.push($obj);
                return true;
            }
            return false;
        }
        if (this.map[mk].addUnique($obj, $objkey)==true) {
            this.arr.push($obj);
            return true;
        }
        return false;
    }


    addArrObj = function($obj, $mapKey) {
        if (this.map[$mapKey]==undefined) {
            this.map[$mapKey]=[$obj];
            this.arr.push($obj);
            return true;
        }
        this.map[$mapKey].push($obj);
        this.arr.push($obj);
        return false;
    }

    sortArrDesc = function($prop) {
        if (this.arr.length==0) return;
        this.arr.sort(function($a, $b) {
            if ($a[$prop] < $b[$prop]) return 1;
            else if ($a[$prop] > $b[$prop]) return -1;
            return 0;
        })
    }

    sortArrAsc = function($prop) {
        if (this.arr.length==0) return;
        this.arr.sort(function($a, $b) {
            if ($a[$prop] < $b[$prop]) return -11;
            else if ($a[$prop] > $b[$prop]) return 1;
            return 0;
        })
    }

}

/************** (from file ./js/base/utils/arraymap.js)**/ 


class page_params {
    constructor ($s=null, $c=null, $f=null, $p1=null, $p2=null,$p3=null,$p4=null,$p5=null,$p6=null,$p7=null) {
        this.s_cbk=$s;      //success
        this.c_cbk=$c;      //cancel
        this.f_cbk=$f;      //fail
        this.p1=$p1;
        this.p2=$p2;
        this.p3=$p3;
        this.p4=$p4;
        this.p5=$p5;
        this.p6=$p6;
        this.p7=$p7;
        
        //this allows for tossing things in here to look for specifically.
        //tabber's are set by the pages at runtime which are assigned to be rendered.
        //the tabber is the object "UP" to the parent from the child tab (page_gui)
        this.misc = null;
    }
}


/**
 * helpful obj for callbacks to manipulate the gui/dom/apps after que sends.
 * Typcally the pgui and callback str/funct are used...
 * Othertimes we use this object for more.
 */
class gui_extras {
    constructor($pgui=null, $callback_str="", $callback_funct=null) {
        this._pgui = $pgui;
        this._callback_str = $callback_str;
        this._callback_funct = $callback_funct;
        
        this._status_text = "";
        this._collab_txid = "";

        this._bLASTINQ = false;
        this._passthru = null;

        this._el_display = null;
        this._guitracker = null;

        this._launch = false;

        this._bDONTPULL = false;
        this._bMUSTPULL = false;

        this._guid = null;
    }


    /**
     * @param {string} - just the guid of the pgui
     */   
    set guid($pgui) { this._guid=$guid; }
    get guid() { return this._guid; }


    /**
     * @param {page_gui} - callback function str to be eval'd
     */   
    set pgui($pgui) { this._pgui=$pgui; }
    get pgui() { return this._pgui; }


    /**
     * @param {function} - callback function str to be eval'd
     */  
    set callback_str($cs) { this._callback_str=$cs; }
    get callback_str() { return this._callback_str; }


    /**
     * @param {function} - callback function
     */  
    set callback_funct($cf) { 
        if (typeof $cf!=="function") {
            console.log("(nf_param) callback $cf is not a function");
            return;
        }
        this._callback_funct=$cf; 
    }
    get callback_funct() { return this._callback_funct; }


    /**
     * @param {string} - simple status of the collab (a/d/p)
     * a-approved d-declined p-pending
     */  
    set status_text($st) { this._status_text=$st; }
    get status_text() { return this._status_text; }
    

    /**
     * @param {string} - long txid string
     */  
    set collab_txid($ct) { this._collab_txid=$ct; }
    get collab_txid() { return this._collab_txid; }


    /**
     * @param {boolean} - Is this the last item in the q?
     */  
    set bLASTINQ($liq) { this._bLASTINQ=$liq; }
    get bLASTINQ() { return this._bLASTINQ; }


    /**
     * @param {Object} - A relon
     */  
    set passthru($pt) { this._passthru=$pt; }
    get passthru() { return this._passthru; }


    /**
     * @param {string}   //todo, this may be an actual DOM element
     */    
    set el_display($el) { this._el_display=$el; }
    get el_display() { return this._el_display; }


    /**
     * @param {string}
     */
    set guitracker($pt) { this._guitracker=$pt; }
    get guitracker() { return this._guitracker; }


    /**
     * @param {boolean}
     */
    set launch($launch) { this._launch=$launch; }
    get launch() { return this._launch; }


    /**
     * @param {boolean}
     */
     set bDONTPULL($bDONTPULL) { this._bDONTPULL=$bDONTPULL; }
     get bDONTPULL() { return this._bDONTPULL; }


    /**
     * @param {boolean}
     */
    set bMUSTPULL($bMUSTPULL) { this._bMUSTPULL=$bMUSTPULL; }
    get bMUSTPULL() { return this._bMUSTPULL; }
}

/************** (from file ./js/framework/page_params.js)**/ 

/***********************************************************************************************
                                                                                                                                            
                                     .....                                      
                            @@@@@@@@@@@@@@@@@@@@@@@@@@@@,                       
                       @@@@@@@@@@@@@@&%%&@@@@@@@@@@@@@@@@@@@@.                  
                    @@@@@@@@@,                        @@@@@@@@@@(               
                 @@@@@@@@@                                .@@@@@@@@             
               @@@@@@@@             %@@@@@@@@                @@@@@@@            
             @@@@@@@@            (@@@@@@@@@@@@@.               @@@@@@&          
           #@@@@@@@             @@@@@@@@@@@@@@@@.               @@@@@@/         
          @@@@@@@@              @@@@@@@@@@@@@@@@/               /@@@@@@         
         @@@@@@@@                @@@@@@@@@@@@@@%                 @@@@@@@        
        @@@@@@@@                    @@@@@@@@@                    @@@@@@@        
       @@@@@@@@                                                  @@@@@@@        
      @@@@@@@@@                                                  @@@@@@@*       
      @@@@@@@@#         @@ @    @&@   @.  @,@  @%  @@@@@@        @@@@@@@@       
      @@@@@@@@@        (@  /@   @  @  @.  @  @ @%  @@@(          @@@@@@@@       
      @@@@@@@@@@       @    @@  @   @,@.  @   @@%  @@@@@@        @@@@@@@@       
       @@@@@@@@@                                                  @@@@@@        
        @@@@@@@                                                                 
                   
***********************************************************************************************/

/**
 * Load up of neurons and relons
 */
class mem {
    constructor(){
        this.ANNE = null;
        this.MEMP = null;
        this.USER = null;
        this.ANNE_ACCTID = "7777770163980908667";    
        this.ANNE_ADMIN_ACCTID = "7777771543345157118";
        this.ANNE_ADMIN_BACKUP_ACCTID = "7777044340734581180";
        this.ANNE_DEV_FUND_ACCTID = "7770353147211057540";

        this.BICONOMY_DEPOSIT_NID = "599015992427635231";

        this.KNOWN_MEMO_ACCTS = [this.BICONOMY_DEPOSIT_NID];

        this.ANNE_SUPERS = [this.ANNE_ACCTID, this.ANNE_ADMIN_ACCTID, this.ANNE_ADMIN_BACKUP_ACCTID];
        
        this._aNeurons_hashtbl = [];            //hash of nid w/ prop point to nid.
        this._aNeurons = [];                    //nid indexed nObjs
        this._aRelons_hashtbl = [];             //hash of txid w/ prop pointing to rid
        this._aRelons = [];                     //rid indexed rObj 
        
        this._valid_TI = [];                    //valid to types for our text index
        this.__bDOTEXTINDEX = false;            //flag to turn on general text indexes
        this._bDOPERSONINDEX = false;           //flat to turn on person indexes (not yet made, will be an optimization.)
        
        this._TEXT_INDEX = new Object();  
        this._TEXT_INDEX_PARTIALS = new Object();

        this.__bMINIEDGES = false;

        this.__INDENT = "&#160;└─";              //this will be moved to a website preference neuron :)
        this._emulatedTXIDLength = 20;       
        this._EMULATE_BE_TEXT = true;            //this emulates text neurons for the text strings of our hard-coded neurons.  (overkill for most usecases)
        this._EMULATE_TO_TYPES = true;           //this emulates the non bitcion to types fround in edges.  (a must for advanced utility features)
        this._EMULATE_BE_TEXT_NEURON = false;
        
        this._activeWorldViewTempNids = [];

        this.neuron_necrosis_list = [];
        this.tx_ap_inh_list = []; 
        this._localRunDelete = [];   
        this._thisRunNECROSIS = [];

        this._SYS_TYPE_CODES = [];
        this._SYS_TO_TYPE_CODES = [];

        //any neuron classified as an ANNE sys prop class.
        //these props auto populate javascript .prop values in neurons
        //that supply their value.
        this._SYS_PROPS_TNIDS = [];   

        
        
        this._RESERVED_PROPS = ["parent", 
                                "be", 
                                "bes", 
                                "_gt", 
                                "sponsored_by", 
                                "owned_by", "name", "text",
                                "nid",
                                "txid", "from_nid", "to_s", "reln", "type", "to_type",
                                "outs", "ins", "relon_sponsor"];

        //set to the active language
        //will looks for all text values from this.__TEXT relons as default (english)
        //else, will look for 
        this.__TEXT_ptr = null;     

        
        //confirm emulated boolean maps to existing boolean neuron values.
        this._emulatables = [];
                            
        //as we process local ghosts, look for the ghost_order
        //as we find values, bump this value up to one greater
        this._nextGhostOrder = 0;
        this._anne_ghosts = [];         //local ghost handling (not yet re-enabled)

        //used for loading emulated neurons run-time for apps.
        //big loads can be done then all proccessed instead of processing
        //each relon one at a time.
        this._sand_mark = 0;  
        
        this._active_lang_alpha2 = "";

        this._cascadePropsDownChildrenClasses_cnt = 0;

        this._monthMap=[];
        this._monthLookup=[];
    }


    memoRequired=function($nid) {
        if (this.KNOWN_MEMO_ACCTS.indexOf($nid)!=-1) return true;
        return false;
    }
    
    setPersonIndex=function($bDOPERSONINDEX) {
        this._bDOPERSONINDEX=$bDOPERSONINDEX;
    }

    //this is getting inherited by neuron that is SHOULD NOT BE INHERITED BY???
    setNumber = function ($v) { 
        if ($v==undefined || $v==null) return;
        if (typeof $v != "string") {
            if ($v.text!=undefined) {
                if (ANNE.UTILS.isNum($v.text)==false) {
                    return;
                }  
                $v.setValue( parseFloat($v.text) );
                return;
            }
        }
        $v.setValue( parseFloat($v) );
    }
    

    /** 
     * Boot up the RAM ANNE memory.
     * 
     * Init the MEMP
     * Load up local indexeddb from the browser.
     * Grab preloaded neurons from the ANNODE, if DB mem is empty.
     * 
     * @author {Mr.Scatman} 
     */
     boot_up=async($ANNE)=>{
        this.ANNE = $ANNE;

        //We need a few neurons sys level neurons and some protocol-using specific 
        //concepts when building ANNE local memory in the browser
        //we create them here in bare-bone capacity.  Their definition will build
        //as we process more of their relons.  Here we have the basics to overcome 
        //the chicken-and-egg problem that arrises when using an electronic brain to
        //build a brain.  The language used to build the brain itself is 
        //defined by the neurons and relons that are in it.
        var newSysTypeCode = function ($nObj, $sys_alias, $sys_enum_val, $sys_code="", $validator=null, $set_value=null) {
            $nObj.SYS_ALIAS=$sys_alias;
            $nObj.SYS_ENUM_VAL = $sys_enum_val
            $nObj.SYS_CODE=$sys_code;
            $nObj.format_validator=$validator;      //if exists, will run after set value -- to validate the format.
            $nObj.set_value=$set_value;             //if exists, will run to set the .value of this nObj
            ANNE.MEM._SYS_TYPE_CODES.push($nObj);
        }

        var newSysToTypeCode = function ($nObj, $sys_alias, $sys_enum_val, $sys_code="", $validator=null, $set_value=null) {
            $nObj.SYS_ALIAS=$sys_alias;
            $nObj.SYS_ENUM_VAL = $sys_enum_val
            $nObj.SYS_CODE=$sys_code;
            $nObj.format_validator=$validator;      //if exists, will run after set value -- to validate the format.
            $nObj.set_value=$set_value;             //if exists, will run to set the .value of this nObj
            ANNE.MEM._SYS_TO_TYPE_CODES.push($nObj);
        }

        this.___SELF___ = this.minBootNeuron({nid:"7777089034065106708", text:"___SELF___", ntype:NTYPE_CLASS});
        this.___SELF___.be_lvl = 0;

        this.__ANNE__ = this.minBootNeuron({nid:ANNE_ACCTID, text:"A.N.N.E.", ntype:NTYPE_INST});
        this.__ANNE__.addAkaManual("ANNE");

        //patterns for events and other things can be stored on the app for prefs etc.
        this.ANNE_EXT_APP = this.minBootNeuron({nid:"7772292993542690466", text:"ANNE EXT APP", ntype:NTYPE_INST});
        
        //Neurons will be an instance of a class/sublass or a subclass of a class.
        //All classifications exist in the BE Hierarchy parent/child DAG (TYPE_BE dimension).
        //if the neuron is a class and is intended to only be subclassed classes_only is set.
        //if the neuron is a class and is intended to only be instantiated insts_only is set.
        //if both subclasses and insts are ok, then neither classes_only, or insts_only is set.

        //not intended to be instantiated.
        this._GENERAL_PROP_CLASS = this.minBootNeuron({nid:"7770801601458804038", text:"general prop class", ntype:NTYPE_CLASS});
        this._GENERAL_PROP_CLASS.be_lvl = 1;
        
            this.unicode_start = this.minBootNeuron({nid:"7773282456393979069", text:"unicode start", ntype:NTYPE_CLASS});
            this.unicode_start.be_lvl = 2;
            this.unicode_end = this.minBootNeuron({nid:"7775740209310803806", text:"unicode end", ntype:NTYPE_CLASS});
            this.unicode_end.be_lvl = 2;
        
        //special for the semantic space and protocol enforcement
        this.__ANNE_SYS_CLASS = this.minBootNeuron({nid:"7775402688912060933", text:"ANNE sys prop class", ntype:NTYPE_CLASS});
        this.__ANNE_SYS_CLASS.be_lvl = 1;
            this.rez_pool_priority1 = this.minBootNeuron({nid:"7770223143918999134", text:"rez pool priority 1", ntype:NTYPE_CLASS});    
            this.rez_pool_priority1.be_lvl = 2;
            this.rez_pool_priority2 = this.minBootNeuron({nid:"7772071294239796669", text:"rez pool priority 2", ntype:NTYPE_CLASS});    
            this.rez_pool_priority2.be_lvl = 2;

            this.TYPE = this.minBootNeuron({nid:"7771390251954504665", text:"DATA_TYPE", ntype:NTYPE_CLASS});
            this.TYPE.be_lvl = 2;

                this.DATA_TYPE = this.minBootNeuron({nid:"7776990924864902362", text:"DATA_TYPE", ntype:NTYPE_CLASS});
                this.DATA_TYPE.be_lvl = 3;
            
                        
            this.CMD = this.minBootNeuron({nid:"7772212307781566495", text:"CND", ntype:NTYPE_CLASS});
            this.CMD.be_lvl = 2;
                this.SYS_MSG_CMD = this.minBootNeuron({nid:"7775628525395840248", text:"sys msg cmd", ntype:NTYPE_CLASS});
                this.SYS_MSG_CMD.be_lvl = 3;
                    this.ANNE_SYS_ANNOUNCEMENT = this.minBootNeuron({nid:"7777961960607822223", text:"ANNE system announcement", ntype:NTYPE_CLASS});
                    this.ANNE_SYS_ANNOUNCEMENT.be_lvl = 4;
                    
                    //extreme testing
                    this.ANNE_SUPER_POP = this.minBootNeuron({nid:"777654215432346068", text:"SuperPop", ntype:NTYPE_CLASS});
                    
                    this.ANNE_BRICK_POINT = this.minBootNeuron({nid:"777004069246568712", text:"ANNODE BrickPoint", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_POINT.be_lvl = 4;
                    this.ANNE_BRICK_PT_MSG = this.minBootNeuron({nid:"7771381953594362011", text:"ANNODE BrickPoint message", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_PT_MSG.be_lvl = 4;
                    this.ANNE_BRICK_PT_MIN_V = this.minBootNeuron({nid:"7776219282356411272", text:"ANNODE BrickPoint min version", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_PT_MIN_V.be_lvl = 4;
                    this.ANNE_BRICK_PT_ACTIVATION_BLOCK = this.minBootNeuron({nid:"7770141402154065935", text:"ANNODE BrickPoint activation block", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_PT_ACTIVATION_BLOCK.be_lvl = 4;


        this.SAFUID = this.minBootNeuron({nid:"7774440805627554213", text:"SAFUID", ntype:NTYPE_CLASS});
        this._default_class_nid = this.minBootNeuron({nid:"777178866083313518", text:"default class nid", ntype:NTYPE_CLASS});
        

        //special for the brain firing space
        this.__SYSBRAIN = this.minBootNeuron({nid:"777683243654836824", text:"SYSVAR", ntype:NTYPE_CLASS});
        this.__SYSBRAIN.be_lvl = 1;

            this.brain_assoc_strong = this.minBootNeuron({nid:"777837095094060702", text:"brain:assoc_strong", ntype:NTYPE_CLASS});    
            this.brain_assoc_strong.be_lvl = 2;
            this.brain_assoc_strong.value=2.22;

            this.brain_assoc = this.minBootNeuron({nid:"7772784241541954051", text:"brain:assoc", ntype:NTYPE_CLASS});    
            this.brain_assoc.be_lvl = 2;
            this.brain_assoc_strong.value=1.11;

            this.anbl_assoc = this.minBootNeuron({nid:"7774753656336827036", text:"brain:anbl_assoc", ntype:NTYPE_CLASS});
            
        //this occcurs in the TYPE_BRAIN dimension and throttles firing of the the TO concept it points to.
        this.__NOTACTIVE = this.minBootNeuron({nid:"7776496372488323950", text:"**notactive**", ntype:NTYPE_CLASS});
        this.__NOTACTIVE.be_lvl = 2;
        this.__NOTACTIVE.setValue(-2.22);

        //throttling neuron firing.
        this.__throttlemin = this.minBootNeuron({nid:"7777643683526919133", text:"**throttlemin**", ntype:NTYPE_CLASS});
        this.__throttlemin.be_lvl = 3;
        this.__throttlemin.setValue(-1.11);

        this.__throttlemax = this.minBootNeuron({nid:"7775174343458816319", text:"**throttlemax**", ntype:NTYPE_CLASS});
        this.__throttlemax.be_lvl = 3;
        this.__throttlemin.setValue(-2.22);

        this.__ACTIVE = this.minBootNeuron({nid:"7778810467984387174", text:"**active**", ntype:NTYPE_CLASS});
        this.__ACTIVE.be_lvl = 2;

        this.EC_CLASS = this.minBootNeuron({nid:"777585041016039777", text:"**EC**", ntype:NTYPE_CLASS});
        this.EC_CLASS.be_lvl = 1;

            this.class = this.minBootNeuron({nid:"7777942827590344670", text:"class", ntype:NTYPE_CLASS});
            this.class.be_lvl = 2;
        
            this.inst = this.minBootNeuron({nid:"7777109984382823688", text:"inst", ntype:NTYPE_CLASS});
            this.inst.prompt = "What are you?";
                    
            this.ec_action = this.minBootNeuron({nid:"7774365287013199911", text:"ec:action", ntype:NTYPE_INST});
            this.ec_action.be_lvl = 2;
            this.ec_action._uico = "🏃‍♂️";
            
            this.ec_thing = this.minBootNeuron({nid:"7778260411935216992", text:"ec:thing", ntype:NTYPE_CLASS});
            this.ec_thing.be_lvl = 2;
                this.entity = this.minBootNeuron({nid:"7775965067238290504", text:"entity", ntype:NTYPE_CLASS});
                this.entity.be_lvl = 3;
            
            this.ec_living = this.minBootNeuron({nid:"7772014754426478363", text:"ec:living", ntype:NTYPE_CLASS});
            this.ec_living.be_lvl = 3;
            
            //Person is to document a non-identity definition of a person.
            //Not to be confused with the identity class with are real users of the system.
            this.person = this.minBootNeuron({nid:"7773974326223336303", text:"person", ntype:NTYPE_CLASS});
            this.person.be_lvl = 4;
            this.person._uico = "<span style='color:yellow'>👤</span>";
            this.person._bico = "bi bi-person-fill";
            
            //we can do movies from TMDB with this one...
            this.movie = this.minBootNeuron({nid:"7771769778243524971", text:"movie", ntype:NTYPE_CLASS});  
            //all the TMDB relns we want should be here.  /director/actor/genre etc.

            this.named_after = this.minBootNeuron({nid:"7779919689813420887", text:"named after", ntype:NTYPE_CLASS});
            this.pop_demonym = this.minBootNeuron({nid:"777505779900849159", text:"pop demonym", ntype:NTYPE_CLASS});

            this.ec_nonliving = this.minBootNeuron({nid:"7773981355789251728", text:"ec:nonliving", ntype:NTYPE_CLASS});
            this.ec_nonliving.be_lvl = 3;

            this.ec_location = this.minBootNeuron({nid:"7773349123364275631", text:"ec:location", ntype:NTYPE_CLASS});    
            this.ec_location.be_lvl = 4;

            this.ec_state = this.minBootNeuron({nid:"7777439602338734976", text:"ec:state", ntype:NTYPE_CLASS});
            this.ec_state.be_lvl = 2;
            
            this.ec_measure = this.minBootNeuron({nid:"7770647771955623748", text:"Unit of Measure", ntype:NTYPE_CLASS});
            this.ec_measure.be_lvl = 3;
           
            this.unit_of_measure = this.minBootNeuron({nid:"777700476213440998", text:"Unit of Measure", ntype:NTYPE_CLASS});
            this.unit_of_measure.be_lvl = 4;
             
            this.unit_of_account = this.minBootNeuron({nid:"7774906754160016124", text:"Unit of Account", ntype:NTYPE_CLASS});
            this.unit_of_account.be_lvl = 5;
            
            this.unit_of_crypto= this.minBootNeuron({nid:"7771690081405920648", text:"Unit of Crypto", ntype:NTYPE_CLASS});
            this.unit_of_crypto.be_lvl = 6;

            this.annies = this.minBootNeuron({nid:"7771153305797255960", text:"annies", ntype:NTYPE_CLASS});
            this.annies.be_lvl = 7;

            //aka ROLE_BE.
            this.ec_is = this.minBootNeuron({nid:"7775323663099289436", text:"ec:is", ntype:NTYPE_INST});
            this.ec_is.be_lvl = 2;
            this.role_be = this.ec_is;    //same concept.

            this.ec_posn = this.minBootNeuron({nid:"777042615346522551", text:"ec:posn", ntype:NTYPE_INST});
            this.ec_posn.be_lvl = 2;
            
            this.ec_expection = this.minBootNeuron({nid:"7774919700244959416", text:"ec:expectation", ntype:NTYPE_INST});
            
            this.ec_opposite = this.minBootNeuron({nid:"7777676504812424745", text:"ec.opposite", ntype:NTYPE_CLASS});
        
            //7770250558819326868 mapper state

            this.ec_mode = this.minBootNeuron({nid:"7774567902469413739", text:"ec:mode", ntype:NTYPE_CLASS});    
            this.ec_mode.be_lvl = 2;


                this.location = this.minBootNeuron({nid:"7775670854606187707", text:"location", ntype:NTYPE_CLASS});
                this.location.be_lvl = 3;
                this.location.unit_tests =["Syracuse, New York, U.S."];
                this.location.unitTest=async()=>{
                    for (var u=0; u < ANNE.MEM.location.unit_tests.length; u++) {
                        await ANNE.MEM.location.testParse( ANNE.MEM.location.unit_tests[u]);
                    }
                }
                this.location.testParse=async($loc)=>{
                    this.defaultLocObj = new anne_location(this.misc.default_value);
                    await defaultLocObj.processFull();
                    defaultLoclObj.print();
                    defaultLocObj.printFullDebugString();
                }
                this.location._uico="📍";
                this.location._bico="bi bi-geo-alt-fill";

                this.registrationPolity = this.minBootNeuron({nid:"7775972763622961295", text:"registration polity", ntype:NTYPE_CLASS}); 
                this.hqLocation = this.minBootNeuron({nid:"7775518395372784977", text:"Headquarters", ntype:NTYPE_CLASS}); 
                
            this.ec_theSpeaker = this.minBootNeuron({nid:"6130905652501722840", text:"ec:the speaker", ntype:NTYPE_CLASS});    
            this.ec_theSpeaker.be_lvl = 2;

            //is under ec_idiom
            this.idiom = this.minBootNeuron({nid:"7779579523222965446", text:"idiom", ntype:NTYPE_CLASS});
            this.idiom.be_lvl = 3;

        this.city = this.minBootNeuron({nid:"7779778654320523970", text:"city", ntype:NTYPE_CLASS});
        this.city._bico="bi bi-geo-alt-fill";

        this.anbl = this.minBootNeuron({nid:"7773795520052184088", text:"anbl", ntype:NTYPE_CLASS});

        this.motto = this.minBootNeuron({nid:"7775617504524778859", text:"motto", ntype:NTYPE_CLASS});

        //for testing.  s.b. there already.
        this.ec_anvert_yin = this.minBootNeuron({nid:"7776879633620640959", text:"ec:anvert_yin", ntype:NTYPE_CLASS});    
        this.ec_anvert_yang = this.minBootNeuron({nid:"7773336624326965724", text:"ec:anvert_yang", ntype:NTYPE_CLASS});    
        
        this.eng_subj = this.minBootNeuron({nid:"8557921478923992894", text:"english subject", ntype:NTYPE_CLASS});    
        this.eng_obj = this.minBootNeuron({nid:"16826117856129032282", text:"english object", ntype:NTYPE_CLASS});    
        
        this.eng_grammar_rule = this.minBootNeuron({nid:"7770417256868673002", text:"ENG grammar rule", ntype:NTYPE_CLASS});    
        
        this.eng_action_to_thing = this.minBootNeuron({nid:"7772004807272349660", text:"action->thing", ntype:NTYPE_CLASS});    
        this.eng_action_to_state = this.minBootNeuron({nid:"7770228850164583833", text:"action->state", ntype:NTYPE_CLASS});    
        this.eng_thing_to_state = this.minBootNeuron({nid:"7773065301314646689", text:"thing->state", ntype:NTYPE_CLASS});    
        this.eng_thing_to_thing = this.minBootNeuron({nid:"7774092348316350665", text:"thing->thing", ntype:NTYPE_CLASS});    
        this.eng_state_to_thing = this.minBootNeuron({nid:"7777785038941313166", text:"state->thing", ntype:NTYPE_CLASS});    
        this.eng_state_to_action = this.minBootNeuron({nid:"7773073322491997535", text:"state->action", ntype:NTYPE_CLASS});    
        this.eng_state_to_manner = this.minBootNeuron({nid:"7777057456095646496", text:"state->manner", ntype:NTYPE_CLASS});    
        this.eng_action_to_action = this.minBootNeuron({nid:"7779372114410308475", text:"action->action", ntype:NTYPE_CLASS});    

        //for prounounish resolutions or any language's notion of placeholders that resolve to other concepts
        //that are in context.
        this.ec_resolve = this.minBootNeuron({nid:"777847931831336568", text:"ec:resolvable (thing)", ntype:NTYPE_CLASS});    
        
        this.ec_resolvableThing = this.minBootNeuron({nid:"12811539405663275121", text:"ec:resolvable (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvablePluralThing = this.minBootNeuron({nid:"1229903086758389843", text:"ec:resolvable plural (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvableMale = this.minBootNeuron({nid:"16239629656306601583", text:"ec:resolvable male (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvableFemale = this.minBootNeuron({nid:"13243147462547516359", text:"ec:resolvable female (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvableSelf = this.minBootNeuron({nid:"12040890514035853644", text:"ec:resolvable self (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolves_to_concept = this.minBootNeuron({nid:"7776775212574226990", text:"ec:resolves to concept", ntype:NTYPE_CLASS});    
        this.ec_resolves_to_role = this.minBootNeuron({nid:"7773087588979181787", text:"ec:resolves to role", ntype:NTYPE_CLASS});    
        

        this.ec_negation = this.minBootNeuron({nid:"7772189067884368306", text:"ec:negation", ntype:NTYPE_CLASS});    
        this.ec_plural = this.minBootNeuron({nid:"7775067489908426734", text:"ec:plural", ntype:NTYPE_CLASS});    
        this.ec_singular = this.minBootNeuron({nid:"7778687569286032056", text:"ec:singular", ntype:NTYPE_CLASS});    
        this.ec_direction = this.minBootNeuron({nid:"7775225830148199820", text:"ec:direction", ntype:NTYPE_CLASS}); 
        this.ec_manner = this.minBootNeuron({nid:"7778089761207879407", text:"ec:manner", ntype:NTYPE_CLASS}); 

        this.ec_indefiniteness = this.minBootNeuron({nid:"7772688541621829136", text:"ec:indefiniteness", ntype:NTYPE_CLASS}); 

        this.vb_form = this.minBootNeuron({nid:"5489825729469988645", text:"vb forms", ntype:NTYPE_CLASS}); 
        
        this.vb_v1 = this.minBootNeuron({nid:"7771743005401130790", text:"V1 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v1_infinitive = this.minBootNeuron({nid:"7778016318966708690", text:"V1 - eng infinitive", ntype:NTYPE_CLASS});
        this.vb_v2 = this.minBootNeuron({nid:"7771644245115783595", text:"V2 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v3 = this.minBootNeuron({nid:"7775543337600043609", text:"V3 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v4 = this.minBootNeuron({nid:"7777759725587481842", text:"V4 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v5 = this.minBootNeuron({nid:"7776191686416614280", text:"V5 - eng", ntype:NTYPE_CLASS}); 
        
        this.eng_vb_forms = [this.vb_v1.temp_nid, this.vb_v1_infinitive.temp_nid, this.vb_v2.temp_nid, this.vb_v3.temp_nid, this.vb_v4.temp_nid, this.vb_v5.temp_nid];

        //LC is the language circuitry area of the brain.
        this.LC_prefix = this.minBootNeuron({nid:"7773986200038360248", text:"ec:prefix", ntype:NTYPE_CLASS});    
        this.LC_suffix = this.minBootNeuron({nid:"7778838134500110231", text:"ec:suffix", ntype:NTYPE_CLASS});    
        
        //primary one-schema relon data dimensions.  (relon type)
        //TYPE
        this.relon_dimension = this.minBootNeuron({nid:"7770927391413238961", text:"relon dimension", ntype:NTYPE_CLASS});
        this.ASCII = this.minBootNeuron({nid:"7771796572391639487", text:"ASCII", ntype:NTYPE_CLASS});
        this.ascii_code_num = this.minBootNeuron({nid:"77759961069237519", text:"ASCII code #", ntype:NTYPE_CLASS});
        this.ascii_eval = this.minBootNeuron({nid:"7775975832455259480", text:"ASCII eval", ntype:NTYPE_CLASS});

        this.__SYS = this.minBootNeuron({nid:"7775032170648219182", text:"TYPE_SYS", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__SYS, "TYPE_SYS", TYPE_SYS);
        this.__BE = this.minBootNeuron({nid:"7777873907414567647", text:"TYPE_BE", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BE, "TYPE_BE", TYPE_BE);
        this.__PARAM = this.minBootNeuron({nid:"7777094485082477142", text:"TYPE_PARAM", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__PARAM, "TYPE_PARAM", TYPE_PARAM);
        this.__HAS = this.minBootNeuron({nid:"7776798782476847291", text:"TYPE_HAS", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__HAS, "TYPE_HAS", TYPE_HAS);
        this.__AWARENESS = this.minBootNeuron({nid:"7779521186374438598", text:"TYPE_AWARENESS", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__AWARENESS, "TYPE_AWARENESS", TYPE_AWARENESS);
        this.__EXPERIENCE = this.minBootNeuron({nid:"7776169509680303554", text:"TYPE_EXPERIENCE", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__EXPERIENCE, "TYPE_EXPERIENCE", TYPE_EXPERIENCE);
        this.__CLUMP = this.minBootNeuron({nid:"7770332898798547236", text:"TYPE_CLUMP", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__CLUMP, "TYPE_CLUMP", TYPE_CLUMP);
        this.__SYS_MSG = this.minBootNeuron({nid:"7778077217855587615", text:"TYPE_BRAIN", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__SYS_MSG, "TYPE_SYS_MSG", TYPE_SYS_MSG);
        this.__BRAIN = this.minBootNeuron({nid:"7778077217855587615", text:"TYPE_BRAIN", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BRAIN, "TYPE_BRAIN", TYPE_BRAIN);
        this.__ROLE = this.minBootNeuron({nid:"7779056189210800473", text:"TYPE_ROLE", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BRAIN, "TYPE_ROLE", TYPE_ROLE);
        this.__EPISODIC = this.minBootNeuron({nid:"7770774009406800345", text:"TYPE_EPISODIC", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BRAIN, "TYPE_EPISODIC", TYPE_EPISODIC);

        //THIS IS CURRENTLY NOT BEING USED ANYMORE.
        this.seed = this.minBootNeuron({nid:"7778643893074759184", text:"seed", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.seed, "TO_TYPE_SEED", TO_TYPE_SEED, "", null, null);
        this.seed.bTERMINATING_TYPE = true;  
        this.seed.be_lvl = 4;
        
        this.anne_id = this.minBootNeuron({nid:"7777571073929239436", text:"anne_id", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.anne_id, "TO_TYPE_ANNE", TO_TYPE_ANNE, "", null, null);
        this.anne_id.bTERMINATING_TYPE = false;  
        this.anne_id.be_lvl = 4;
        
        this.e_anne_id = this.minBootNeuron({nid:"7777582090111925141", text:"encrypted_anne_id", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_anne_id, "TO_TYPE_ENCR_ANNE", TO_TYPE_ENCR_ANNE, "", null, null);
        this.e_anne_id.bTERMINATING_TYPE = false;     //we want normal HASSFULL handling.
        this.e_anne_id.be_lvl = 4;
        
        this.blob = this.minBootNeuron({nid:"7772915831532867191", text:"blob", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.blob, "TO_TYPE_BLOB", TO_TYPE_BLOB, "", null, null);
        this.blob.bTERMINATING_TYPE = true; 
        this.blob.be_lvl = 4;

        this.e_blob = this.minBootNeuron({nid:"777644219286194944", text:"encrypted_blob", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_blob, "TO_TYPE_ENCR_BLOB", TO_TYPE_ENCR_BLOB, "", null, null);
        this.e_blob.bTERMINATING_TYPE = true;
        this.e_blob.be_lvl = 4;

        this.date = this.minBootNeuron({nid:"7776322997340577329", text:"date", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.date, "TO_TYPE_DATE", TO_TYPE_DATE, "", null, null);
        this.date.bTERMINATING_TYPE = true;
        this.date.be_lvl = 4;

        this.e_date = this.minBootNeuron({nid:"7770256565764089519", text:"encrypted_date", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_date, "TO_TYPE_ENCR_DATE", TO_TYPE_ENCR_DATE, "", null, null);
        this.e_date.bTERMINATING_TYPE = true;
        this.e_date.be_lvl = 4;

        this.date_gregorian = this.minBootNeuron({nid:"777137291158471827", text:"gregorian date", ntype:NTYPE_CLASS});
        this.date_gregorian.prompt = "Enter date YYYY-DD-MM";
        this.date_gregorian.frenlyPrint=function($parts, $era="AD"){
            if ($parts.length!=3) return "";
            var year = parseInt($parts[0]);
            var month = MEMG.getNeuronNameByClassAndValue(ANNE.MEM.month, parseInt($parts[1]), true);
            
            var day = ANNE.TALK.lazyDayIntEnding(parseInt($parts[2]));
            var era = $era;
            return "<b>"+month+"</b>, <b>"+day+"</b> <b>"+year+"</b> <b>"+era+"</b>";
        }


        //???????  nasty hack, gunna have to get rid of this.
        //this.date_gregorian.formatFromResolve=function($obj){
            //return $obj[this.year.temp_nid+"_text"]+"-"+$obj[this.day.temp_nid+"_text"]+"-"+$obj[this.month.temp_nid+"_text"];
        //}

        this.gregorian_date = this.date_gregorian; //tired of making errors that cost me debug time.
        this.gregorian_date._uico = "🗓️";
        
        
        this.Alma_mater = this.minBootNeuron({nid:"777854422638973861", text:"Alma mater", ntype:NTYPE_CLASS});
        this.religion = this.minBootNeuron({nid:"777925711019642749", text:"religion", ntype:NTYPE_CLASS});
        this.religion._uico = "🙏";

        this.kinfolk = this.minBootNeuron({nid:"7779726195767515267", text:"kinfolk", ntype:NTYPE_CLASS});
        
        
        this.arts = this.minBootNeuron({nid:"1825382717134509033", text:"arts", ntype:NTYPE_CLASS});
        this.mathematics = this.minBootNeuron({nid:"12047036016253144920", text:"mathematics", ntype:NTYPE_CLASS});
        this.military = this.minBootNeuron({nid:"7770621763180151408", text:"miliary", ntype:NTYPE_CLASS});
        this.science_and_medicine = this.minBootNeuron({nid:"17841506035681030284", text:"science and medicine", ntype:NTYPE_CLASS});
        this.sport = this.minBootNeuron({nid:"18255368834152888231", text:"sport", ntype:NTYPE_CLASS});
        this.politics_and_law = this.minBootNeuron({nid:"6474702777226473172", text:"politics and law", ntype:NTYPE_CLASS});

        //wikispooks lists this as religion "jewish" but that is not the right label for religion.
        //we remap it to judiasm.
        this.judaism = this.minBootNeuron({nid:"12609260194840621674", text:"Judaism", ntype:NTYPE_CLASS});
        this.zionism = this.minBootNeuron({nid:"16686116773752269042", text:"Zionism", ntype:NTYPE_CLASS})
        
        //a tribal identity..
        this.jewish = this.minBootNeuron({nid:"17344203630892052405", text:"Jewish", ntype:NTYPE_CLASS});
        
        this.occupation = this.minBootNeuron({nid:"7774593245582491319", text:"occupation", ntype:NTYPE_CLASS});
        this.occupation._uico = "🔨";
        
        this.position = this.minBootNeuron({nid:"7772884086015787050", text:"position", ntype:NTYPE_CLASS}); 
        this.position._uico = "🔨";
        
        this.career = this.minBootNeuron({nid:"10179302147304506923", text:"career", ntype:NTYPE_CLASS}); 

        this.service_term = this.minBootNeuron({nid:"7778128204103734929", text:"service term", ntype:NTYPE_CLASS});
        this.service_term._uico = "🔨";
        
        //used with company type.
        this.higher_education = this.minBootNeuron({nid:"77712777044716550", text:"Higher Education", ntype:NTYPE_CLASS});
        
        this.nationality = this.minBootNeuron({nid:"7772786288596504983", text:"nationality", ntype:NTYPE_CLASS});
        this.political_party = this.minBootNeuron({nid:"7771124117310918099", text:"political party", ntype:NTYPE_CLASS});
        this.tribal = this.minBootNeuron({nid:"7770927790880535113", text:"tribal identity", ntype:NTYPE_CLASS});
        this.citizenship = this.minBootNeuron({nid:"7773415791436217714", text:"citizenship", ntype:NTYPE_CLASS});
        
        //things much more loosely tagged -- ie, from wikipedia last name parse and 
        //tags that are NOT languages but still relevant.  (see "ashkenazi" for epstein)
        //we do not have an alt be parent/hierarchy setup for this yet as we do not want/need it.
        //lateral connections of TYPE_BE is -> ethnicity suffice.  Classification via observation of use.
        this.ethnicity = this.minBootNeuron({nid:"17140987543979601380", text:"ethnicity", ntype:NTYPE_CLASS});

        this.representative = this.minBootNeuron({nid:"6476312658984199463", text:"representative", ntype:NTYPE_CLASS});
        this.governor =this.minBootNeuron({nid:"3567666822242597329", text:"Governor", ntype:NTYPE_CLASS});
        this.lgovernor =this.minBootNeuron({nid:"2050992576214445805", text:"Lieutenant Governor", ntype:NTYPE_CLASS});
        this.president = this.minBootNeuron({nid:"9793621070511313956", text:"President", ntype:NTYPE_CLASS});
        this.mayor =this.minBootNeuron({nid:"13749506479367882199", text:"Mayor", ntype:NTYPE_CLASS});
        this.judge = this.minBootNeuron({nid:"11286563740955259900", text:"Judge", ntype:NTYPE_CLASS});
        //this.cjudge = this.minBootNeuron({nid:"", text:"Circuit Judge", ntype:NTYPE_CLASS});
        
        this.ceo = this.minBootNeuron({nid:"842151752508316689", text:"CEO", ntype:NTYPE_CLASS});
        this.chairman = this.minBootNeuron({nid:"14221162201076302843", text:"chairman", ntype:NTYPE_CLASS});
        
        this.WEF = this.minBootNeuron({nid:"7776951561979368390", text:"World Economic Forum", ntype:NTYPE_INST});
        
        //used to classify board of director and board of trustee neurons.
        //they are "company" neurons that tie back to the corporation or association they are connected to.
        //they themselves have a comany_type of governance board.
        this.governance_board = this.minBootNeuron({nid:"7779663457772195626", text:"governance board", ntype:NTYPE_CLASS});
        this.board_member  = this.minBootNeuron({nid:"7773721544735553188", text:"board member", ntype:NTYPE_CLASS});

        //will have to fine tune these spellings as we dupe them from wikispooks.
        this.us_republican = this.minBootNeuron({nid:"7773678707765331234", text:"US Republican", ntype:NTYPE_CLASS});
        this.us_democrat = this.minBootNeuron({nid:"7772420605042291734", text:"US Democat", ntype:NTYPE_CLASS});
        this.independent = this.minBootNeuron({nid:"7773736162587243688", text:"Independent", ntype:NTYPE_CLASS});
        this.communist = this.minBootNeuron({nid:"7779857133825607397", text:"Communist", ntype:NTYPE_CLASS});
        this.socialist = this.minBootNeuron({nid:"777564380965650179", text:"Socialist", ntype:NTYPE_CLASS});
        this.cpoc = this.minBootNeuron({nid:"7770258420272538620", text:"Communist Party of China", ntype:NTYPE_CLASS});

        //relations
        this.kinfolk_parent = this.minBootNeuron({nid:"7775760258501601282", text:"parent", ntype:NTYPE_CLASS});
        this.kinfolk_child = this.minBootNeuron({nid:"777299714238049539", text:"child", ntype:NTYPE_CLASS});
        this.spouse = this.minBootNeuron({nid:"7778826002852320790", text:"spouse", ntype:NTYPE_CLASS});

        //standout relationship that can be calculated but stands out if relevant to highlight.
        this.kinfolk_cousin = this.minBootNeuron({nid:"1548501709675585821", text:"kinfolk cousin", ntype:NTYPE_CLASS});
        
        this.kinfolk_grandparent = this.minBootNeuron({nid:"150810937878958292", text:"kinfolk grandparent", ntype:NTYPE_CLASS});
        this.grandfather = this.minBootNeuron({nid:"5668674581344300968", text:"grandfather", ntype:NTYPE_CLASS});
        this.grandmother = this.minBootNeuron({nid:"9317028650101357972", text:"grandmother", ntype:NTYPE_CLASS});
        
        //action used in marriage clumps.
        this.marry = this.minBootNeuron({nid:"6385755796053286049", text:"marry", ntype:NTYPE_CLASS});
        this.marriage = this.minBootNeuron({nid:"7772714390270034050", text:"marriage", ntype:NTYPE_CLASS});
        

        this.father = this.minBootNeuron({nid:"7770334697346006910", text:"father", ntype:NTYPE_CLASS});
        this.bio_father = this.minBootNeuron({nid:"7774430960064578792", text:"bilogical father", ntype:NTYPE_CLASS});
        this.adoptive_father = this.minBootNeuron({nid:"777421606944026826", text:"adoptive father", ntype:NTYPE_CLASS});
        this.father_in_law = this.minBootNeuron({nid:"7778846417673009620", text:"father in-lawr", ntype:NTYPE_CLASS});
        this.son = this.minBootNeuron({nid:"7774803688522754022", text:"son", ntype:NTYPE_CLASS});
        this.bio_son = this.minBootNeuron({nid:"7775737496175899538", text:"bio son", ntype:NTYPE_CLASS});
        this.adopted_son = this.minBootNeuron({nid:"7777944849660781543", text:"adopted son", ntype:NTYPE_CLASS});
        this.son_in_law = this.minBootNeuron({nid:"7772106931969635469", text:"son in-law", ntype:NTYPE_CLASS});
        this.daughter = this.minBootNeuron({nid:"7772451463343244417", text:"daughter", ntype:NTYPE_CLASS});
        this.bio_daughter = this.minBootNeuron({nid:"7777495899734158074", text:"bio daughter", ntype:NTYPE_CLASS});
        this.adopted_daughter = this.minBootNeuron({nid:"7778259908422705155", text:"adopted daughter", ntype:NTYPE_CLASS});
        this.daughter_in_law = this.minBootNeuron({nid:"7778965895626708730", text:"daughter in-law", ntype:NTYPE_CLASS});
        this.mother = this.minBootNeuron({nid:"7778367143803659148", text:"mother", ntype:NTYPE_CLASS});
        
        //generic sibbling
        this.kinfolk_sibbling = this.minBootNeuron({nid:"7775340410780712793", text:"sibbling", ntype:NTYPE_CLASS});

        this.brother = this.minBootNeuron({nid:"11056317496902226178", text:"brother", ntype:NTYPE_CLASS});
        this.sister = this.minBootNeuron({nid:"11390471507506307054", text:"sister", ntype:NTYPE_CLASS});

        this.kinfolk_tnids = [
            this.kinfolk_cousin.temp_nid, this.mother.temp_nid,
            this.kinfolk_sibbling.temp_nid, this.kinfolk_parent.temp_nid, this.kinfolk_child.temp_nid, this.spouse.temp_nid, this.father.temp_nid,
            this.bio_father.temp_nid, this.adoptive_father.temp_nid, this.father_in_law.temp_nid,  this.son.temp_nid,
            this.bio_son.temp_nid, this.adopted_son.temp_nid, this.son_in_law.temp_nid,  this.daughter.temp_nid, 
            this.bio_daughter.temp_nid, this.adopted_daughter.temp_nid, this.daughter_in_law.temp_nid
        ];

        this.comp_union = this.minBootNeuron({nid:"7771287958516484511", text:"comp union", ntype:NTYPE_CLASS});

        //0-day 10800BC, every day after is +1.
        //any real gregorian date has an ANNE EPOCH DAY.
        //NOTE: this does not take in to account calendar changes (since gregorian calendar didn't exist in 10800BC!)
        //This is a way to quickly (numerically) compare dates in the annode while searching relons.
        this.ANNE_EPOCH_DAY = this.minBootNeuron({nid:"7778857555786131806", text:"ANNE EPOCH DAY", ntype:NTYPE_CLASS});

        this.month = this.minBootNeuron({nid:"777842179295607978", text:"month", ntype:NTYPE_CLASS});
        this.year = this.minBootNeuron({nid:"7774796274147265857", text:"year", ntype:NTYPE_CLASS});
        this.day = this.minBootNeuron({nid:"7775054193871499071", text:"day", ntype:NTYPE_CLASS});
        this.daymonth = this.minBootNeuron({nid:"7772711710926955713", text:"daymonth", ntype:NTYPE_CLASS});
        this.monthyear = this.minBootNeuron({nid:"7778824928802286657", text:"monthyear", ntype:NTYPE_CLASS});
        
        this.date_any = [MEM.date_gregorian.temp_nid, MEM.daymonth.temp_nid, MEM.monthyear.temp_nid];
        this.date_any_nid = [MEM.date_gregorian.nid, MEM.daymonth.nid, MEM.monthyear.nid];
        this.isDateAny=function($nid) {
            if (this.date_any_nid.indexOf($nid)!=-1) return true;
            return false;
        }
        this.era = this.minBootNeuron({nid:"7775223774408988417", text:"era", ntype:NTYPE_CLASS});
        this.era_bc =  this.minBootNeuron({nid:"7773055704590862597", text:"BC", ntype:NTYPE_CLASS});
        this.era_ad = this.minBootNeuron({nid:"7772300057442531736", text:"AD", ntype:NTYPE_CLASS});

        this.doty = this.minBootNeuron({nid:"7776382262043367717", text:"day of the year", ntype:NTYPE_CLASS});
        this.dotw = this.minBootNeuron({nid:"7773395836100627599", text:"day of the week", ntype:NTYPE_CLASS});


        this.month_jan =  this.minBootNeuron({nid:"7779972248415370744", text:"Januar", ntype:NTYPE_CLASS});
        this.month_jan.value=1;
        this.month_feb =  this.minBootNeuron({nid:"7772796602534780742", text:"February", ntype:NTYPE_CLASS});
        this.month_feb.value=2;
        this.month_mar =  this.minBootNeuron({nid:"7771522067996789646", text:"March", ntype:NTYPE_CLASS});
        this.month_mar.value=3;
        this.month_apr =  this.minBootNeuron({nid:"7774465631312901598", text:"April", ntype:NTYPE_CLASS});
        this.month_apr.value=4;
        this.month_may =  this.minBootNeuron({nid:"7770857201478628723", text:"May", ntype:NTYPE_CLASS});
        this.month_may.value=5;
        this.month_jun =  this.minBootNeuron({nid:"7774339894223286536", text:"June", ntype:NTYPE_CLASS});
        this.month_jun.value=6;
        this.month_jul =  this.minBootNeuron({nid:"7774838671233944689", text:"July", ntype:NTYPE_CLASS});
        this.month_jul.value=7;
        this.month_aug =  this.minBootNeuron({nid:"7779263094123527469", text:"August", ntype:NTYPE_CLASS});
        this.month_aug.value=8;
        this.month_sep =  this.minBootNeuron({nid:"7777384954129664128", text:"September", ntype:NTYPE_CLASS});
        this.month_sep.value=9;
        this.month_oct =  this.minBootNeuron({nid:"7777395188527845949", text:"October", ntype:NTYPE_CLASS});
        this.month_oct.value=10;
        this.month_nov =  this.minBootNeuron({nid:"7775446664476737015", text:"November", ntype:NTYPE_CLASS});
        this.month_nov.value=11;
        this.month_dec =  this.minBootNeuron({nid:"7770036874097137801", text:"December", ntype:NTYPE_CLASS});
        this.month_dec.value=12;
        
        this.gps_lat = this.minBootNeuron({nid:"7773164283122952368", text:"lat", ntype:NTYPE_CLASS});
        this.gps_long = this.minBootNeuron({nid:"7776410985703118068", text:"long", ntype:NTYPE_CLASS});


        /**
         * DoDateFormat event handler.
         * Set any date text fields to use this for their keyup function. 
         * @author {Mr.Scatman}
         */
        this.doDateFormat=function($e) { 
            if (!$e.ctrlKey && !$e.metaKey && ($e.keyCode == 32 || $e.keyCode > 46)) {
                ANNE.MEM.date.formatter($e.target);
            }
        }

        /**
         * this will not work with BC.  Too bad, for now.
         * @param {*} $el 
         */
        this.date.formatter = function($el) {
            var format = "yyyy-mm-dd";
            var match = new RegExp(format
                .replace(/(\w+)\W(\w+)\W(\w+)/, "^\\s*($1)\\W*($2)?\\W*($3)?([0-9]*).*")
                .replace(/m|d|y/g, "\\d"));
            var replace = "$1-$2-$3$4"
                .replace(/\//g, format.match(/\W/));

            $el.value = $el.value.replace(/(^|\W)(?=\d\W)/g, "$10")   // padding
                    .replace(match, replace)             // fields
                    .replace(/(\W)+/g, "$1");            // remove repeats

            var parts = $el.value.split("-");
            if (parts[0].length>4) parts[0].substring(0, 3);
            if (parseInt( parts[1] ) > 12) parts[1] = 12;
            if (parseInt( parts[2] ) > 31) parts[2] = 31;
            $el.value = parts.join("-");
        }
        this.date.be_lvl = 4;
        
        //
        this.start_date = this.minBootNeuron({nid:"7774876474320373351", text:"start date", ntype:NTYPE_CLASS});
        this.end_date = this.minBootNeuron({nid:"7773588826071954087", text:"end date", ntype:NTYPE_CLASS});
        this.launch_date = this.minBootNeuron({nid:"1293921080709290782", text:"launch date", ntype:NTYPE_CLASS});
        
        this.dissolved_on = this.minBootNeuron({nid:"11098414578706519923", text:"dissolved on", ntype:NTYPE_CLASS});

        this.datetime = this.minBootNeuron({nid:"7770829136782284019", text:"datetime", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.datetime, "TO_TYPE_DATETIME", TO_TYPE_DATETIME, "", null, null);
        this.datetime.bTERMINATING_TYPE = true;
        this.datetime.be_lvl = 4;

        this.e_datetime = this.minBootNeuron({nid:"7778516895066405156", text:"encrypted_datetime", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_datetime, "TO_TYPE_ENCR_DATETIME", TO_TYPE_ENCR_DATETIME, "", null, null);
        this.e_datetime.bTERMINATING_TYPE = true;
        this.e_datetime.be_lvl = 4;

        this.number = this.minBootNeuron({nid:"777466536043042154", text:"number", ntype:NTYPE_CLASS});
        this.number.be_lvl = 4;
        newSysToTypeCode(this.number, "TO_TYPE_NUMBER", TO_TYPE_NUMBER,  "", null, null)
        
        
        this.number.set_value = this.setNumber;
        
        this.number.formatter = function ($v) {
            var vtype = typeof $v; 
            if (vtype ==='object') {
                this.number.set_value($v);
                return $v.value.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ",");
            }
            else if (vtype==='number') {
                return $v.toString().replace(/\B(?=(\d{3})+(?!\.\d))/g, ",");
            }
            else return $v.toString().replace(/\B(?=(\d{3})+(?!\.\d))/g, ",");
        }

        this.e_number = this.minBootNeuron({nid:"7770791628311366616", text:"encrypted_number", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_number, "TO_TYPE_ENCR_NUMBER", TO_TYPE_ENCR_NUMBER, "", null, null);
        this.e_number.bTERMINATING_TYPE = true;
        this.e_number.be_lvl = 4;

        this.integer = this.minBootNeuron({nid:"7775505125310503749", text:"integer", ntype:NTYPE_CLASS});
        this.integer.set_value = function($n) { $n.setValue(parseInt($n.text)); }  //overrides the set_value function of number (parseFloat) for integers.
        this.integer.be_lvl = 4;

        this.n0 = this.minBootNeuron({nid:"7772759491040052489", text:"0", ntype:NTYPE_CLASS});
        this.n0.be_lvl = 5;
        this.n1 = this.minBootNeuron({nid:"7779936209861264085", text:"1", ntype:NTYPE_CLASS});
        this.n1.be_lvl = 5;
        this.n2 = this.minBootNeuron({nid:"7777466399639974815", text:"2", ntype:NTYPE_CLASS});
        this.n2.be_lvl = 5;
        this.n3 = this.minBootNeuron({nid:"7770513775335783280", text:"3", ntype:NTYPE_CLASS});
        this.n3.be_lvl = 5;
        this.n4 = this.minBootNeuron({nid:"7779613874588391142", text:"4", ntype:NTYPE_CLASS});
        this.n4.be_lvl = 5;
        this.n5 = this.minBootNeuron({nid:"7773838259673220295", text:"5", ntype:NTYPE_CLASS});
        this.n5.be_lvl = 5;
        this.n6 = this.minBootNeuron({nid:"7772956192305001444", text:"6", ntype:NTYPE_CLASS});
        this.n6.be_lvl = 5;
        this.n7 = this.minBootNeuron({nid:"7779032678869944184", text:"7", ntype:NTYPE_CLASS});
        this.n7.be_lvl = 5;
        this.n8 = this.minBootNeuron({nid:"7777087244990101066", text:"8", ntype:NTYPE_CLASS});
        this.n8.be_lvl = 5;
        this.n9 = this.minBootNeuron({nid:"7778263312619963662", text:"9", ntype:NTYPE_CLASS});
        this.n9.be_lvl = 5;
        this.n10 = this.minBootNeuron({nid:"777749396793604897", text:"10", ntype:NTYPE_CLASS});
        this.n10.be_lvl = 5;
        this.n11 = this.minBootNeuron({nid:"7774857001185551332", text:"11", ntype:NTYPE_CLASS});
        this.n11.be_lvl = 5;
        this.n12 = this.minBootNeuron({nid:"7777727428900691983", text:"12", ntype:NTYPE_CLASS});
        this.n12.be_lvl = 5;

        //referenced for direct "values" of neurons set with the nObj->eVal->xx pattern,
        //where xx is the value.  
        //example.   dozen ->eval ->12
        this.eval = this.minBootNeuron({nid:"7776011636867438465", text:"eval", ntype:NTYPE_CLASS});
        this.eval.set_value=function($v) {
            if (typeof $v !=="object") {
                console.warn("must use neuron to set_value");
                return;
            }
            if ($v["be_"+ANNE.MEM.eval.temp_nid]!=undefined) {   //ez road
                $v.setValue($v["be_"+ANNE.MEM.eval.temp_nid]);
                return $v.value;
            }

            //longer road..
            var evalRelon = MEMG.getRelonByTempR($v.outs, ANNE.MEM.eval.temp_nid);
            if (evalRelon!=null && evalRelon.type==TYPE_BE) {
                if (evalRelon.to_type==TO_TYPE_ANNE) {  //have to get val of this neuron!s
                    var valueNOBJ = ANNE.MEM.getNeuronByTempNid(polarityRelon.t);
                    if (valueNOBJ!=null) $v.polarity=ANNE.MEM.__MASTER_GET_VALUE(valueNOBJ);
                }
                else {
                    $v.setValue(evalRelon.to_s);
                    if (isNaN($v.value)==true) {
                        $v.value=null;
                    }
                    return $v.value;
                }
            }
            return null;
        }

        //Must override the default set_value for parent number class.
        //it converts text to number.
        //proportion has text as a fraction and the value can be eval'd from the text
        //but more accurately (and true) is the eval relon pointing to the decimal value.
        this.proportion = this.minBootNeuron({nid:"7779861280974626916", text:"proportion", ntype:NTYPE_CLASS});
        this.proportion.set_value = function($v) { 
            this.eval.set_value($v);
        }

        this.time = this.minBootNeuron({nid:"7771199207074827487", text:"time", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.time, "TO_TYPE_TIME", TO_TYPE_TIME, "", null, null);
        this.time.bTERMINATING_TYPE = true;
        this.time.be_lvl = 4;

        this.time.formatter = function($el) {
            if ($el==null) return;
            var val = $el.value;
            //20:17:40 GMT+00:00
        }
        this.doTimeFormat=function($e) { 
            if (!$e.ctrlKey && !$e.metaKey && ($e.keyCode == 32 || $e.keyCode > 46)) {
                ANNE.MEM.formatter($e.target);
            }
        }

        this.e_time = this.minBootNeuron({nid:"7776682659772901989", text:"encrypted_time", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_time, "TO_TYPE_ENCR_TIME", TO_TYPE_ENCR_TIME, "", null, null);
        this.e_time.bTERMINATING_TYPE = true;
        this.e_time.be_lvl = 4;

        this.txid = this.minBootNeuron({nid:"7774781823726482570", text:"txid", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.txid, "TO_TYPE_TXID", TO_TYPE_TXID, "",  null, null);
        this.txid.set_value = function ($v) { 
            $v.setValue(MEMG.lookupRelonByTXID($v.text));
        }
        this.txid.be_lvl = 4;

        this.e_txid = this.minBootNeuron({nid:"7774527078408052235", text:"encrypted_txid", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_txid, "TO_TYPE_ENCR_TXID", TO_TYPE_ENCR_TXID, "", null, null);
        this.e_txid.bTERMINATING_TYPE = true;
        this.e_txid.be_lvl = 4;

        this.url = this.minBootNeuron({nid:"7771426118341686899", text:"url", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.url, "TO_TYPE_URL", TO_TYPE_URL, "", null, null);
        this.url.bTERMINATING_TYPE = true;
        this.url.format_validator=ANNE.UTILS.__urlValidation;
        this.url.be_lvl = 4;

        this.e_url = this.minBootNeuron({nid:"7772967135250758561", text:"encrypted_url", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_url, "TO_TYPE_ENCR_URL", TO_TYPE_ENCR_URL, "", null, null);
        this.e_url.bTERMINATING_TYPE = true;
        this.e_url.be_lvl = 4;

        this.boolean = this.minBootNeuron({nid:"7777601580469353964", text:"boolean", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.boolean, "TO_TYPE_BOOLEAN", TO_TYPE_BOOLEAN, "",  this.__setValByBoolPolarity);
        this.boolean.bTERMINATING_TYPE = true;
        this.boolean.be_lvl = 4;

        this.boolean.set_value = function($n) { 
            var rObj = MEMG.getRelonByTempR($n.outs, ANNE.MEM.eval.temp_nid);
            if (rObj!=null) $n.setValue(parseInt(rObj.to_s)); 
        }
        
        this.e_boolean = this.minBootNeuron({nid:"7773554013708022388", text:"encrypted_boolean", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_boolean, "TO_TYPE_ENCR_BOOLEAN", TO_TYPE_ENCR_BOOLEAN);
        this.e_boolean.bTERMINATING_TYPE = true;
        this.e_boolean.be_lvl = 4;

        this.yesno= this.minBootNeuron({nid:"7775762158326148225", text:"yes/no", ntype:NTYPE_CLASS});
        this.yes= this.minBootNeuron({nid:"7773442870220826516", text:"yes", ntype:NTYPE_CLASS});
        this.yes.value=1;
        this.no = this.minBootNeuron({nid:"7775849543552546059", text:"no", ntype:NTYPE_CLASS});
        this.no.value=0;

        this.string = this.minBootNeuron({nid:"7777506478065484896", text:"string", ntype:NTYPE_CLASS});
        this.string.set_value = function($v) { 
            $v.setValue($v.text.trim()); 
        }
        newSysToTypeCode(this.string, "TO_TYPE_STRING", TO_TYPE_STRING, "", null);
        this.string.bTERMINATING_TYPE = true;
        this.string.be_lvl = 4;

        //will point to base of url so we can craft dynamic wikilinks w/o storing the entire string again and again.
        //this.wikispookpart = this.minBootNeuron({nid:"7777667326299864145", text:"wikispook part", ntype:NTYPE_CLASS});
        
        this.e_string = this.minBootNeuron({nid:"777615879896782812", text:"encrypted_string", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_string, "TO_TYPE_ENCR_STRING", TO_TYPE_ENCR_STRING);
        this.e_string.bTERMINATING_TYPE = true;
        this.e_string.be_lvl = 4;

        this.filehash = this.minBootNeuron({nid:"7771966589026475642", text:"filehash", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.filehash, "TO_TYPE_FILEHASH", TO_TYPE_FILEHASH, "", null);
        this.filehash.bTERMINATING_TYPE = true;
        this.filehash.be_lvl = 4;

        this.e_filehash = this.minBootNeuron({nid:"7773418046787332221", text:"encrypted_filehash", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_filehash, "TO_TYPE_ENCR_FILEHASH", TO_TYPE_ENCR_FILEHASH, "", null);
        this.e_filehash.bTERMINATING_TYPE = true;
        this.e_filehash.be_lvl = 4;

        this.ipv4 = this.minBootNeuron({nid:"7770686966970402490", text:"ipv4", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.ipv4, "TO_TYPE_IPV4", TO_TYPE_IPV4, "", null);
        this.ipv4.bTERMINATING_TYPE = true;
        this.ipv4.format_validator=ANNE.UTILS.__ipv4Validation;
        this.ipv4.be_lvl = 4;

        this.e_ipv4 = this.minBootNeuron({nid:"7776380890278424141", text:"encrypted_ipv4", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_filehash, "TO_TYPE_ENCR_IPV4", TO_TYPE_ENCR_IPV4, "", null);
        this.e_ipv4.bTERMINATING_TYPE = true;
        this.e_ipv4.be_lvl = 4;

        this.ipv6 = this.minBootNeuron({nid:"7774737336689250477", text:"ipv6", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.ipv4, "TO_TYPE_IPV6", TO_TYPE_IPV6, "", null);
        this.ipv6.bTERMINATING_TYPE = true;
        this.ipv6.format_validator=ANNE.UTILS.__ipv4Validation;
        this.ipv6.be_lvl = 4;

        this.e_ipv6 = this.minBootNeuron({nid:"7779915290430504", text:"encrypted_ipv4", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_filehash, "TO_TYPE_ENCR_IPV6", TO_TYPE_ENCR_IPV6, "", null);
        this.e_ipv6.bTERMINATING_TYPE = true;
        this.e_ipv6.be_lvl = 4;


        this.sequence = this.minBootNeuron({nid:"7773197197458854483", text:"sequence", ntype:NTYPE_CLASS});
        this.sequence.set_value = function($v) { 
            $v.outs.sort(ANNE.MEM.__sortSeqRelons); 
        }

        this.factor = this.minBootNeuron({nid:"7778025035144069284", text:"factor", ntype:NTYPE_CLASS});
        this.factor.set_value = function($uomNOBJ) {    //the factor class has a set_value function which will correctly set the value of its instances.
            if ($uomNOBJ==null) return;
            
            //if we have done this, don't do it again.
            if ($uomNOBJ.uom!=undefined) return $uomNOBJ.uom;

            var cRelon = MEMG.getRelonByTempR($uomNOBJ.outs, ANNE.MEM.conversion_unit.temp_nid);
            
            //this is just a factor (like a prefix to alter a value without a conversion
            //unit or conversion relon.. it will have a factor relon.
            if (cRelon==null) {
                var uom = {base_unit:$uomNOBJ};

                var unit_lbl_rel = MEMG.getRelonByTempRBeLine($uomNOBJ.outs, [TYPE_HAS], ANNE.MEM.unit_symbol.temp_nid, [ANNE.MEM.html_symbol.temp_nid]);
                if (unit_lbl_rel!=null) {
                    uom.unit_lbl = unit_lbl_rel.to_s;
                }
                else uom.unit_lbl="";
                
                var factor_rel = MEMG.getRelonByTempR($uomNOBJ.outs, ANNE.MEM.factor.temp_nid);
                if (factor_rel==null) return null;
                var parts = factor_rel.to_s.split('^');

                if (parts.length!=2) {
                    uom.factored = uom.factor_rel.to_s;
                    $uomNOBJ.factored = uom.factor_rel.to_s;
                    $uomNOBJ.uom = uom;
                    return uom;
                }
                var base = parseFloat(parts[0]);
                var exponent = parseFloat(parts[1]);
                $uomNOBJ.factored = Math.pow(base, exponent);
                uom.factored = $uomNOBJ.factored;
                $uomNOBJ.uom = uom;
                return uom;
            }
            
            var cUnitNOBJ = MEMG.getNeuronByTempNid(cRelon.t);
            if (cUnitNOBJ==null) return;
            
            var uom = {unit:cUnitNOBJ, base_unit:$uomNOBJ};
    
            var conversionRel = MEMG.getRelonByTempR($uomNOBJ.outs, ANNE.MEM.conversion.temp_nid);
            if (conversionRel==null) return;

            uom.conversion = conversionRel.to_s;

            //this should get "any" child of symbol..  so symbol or also something like si_unit_symbol.. etc
            var unit_lbl_rel = MEMG.getRelonByTempRBeLine(cUnitNOBJ.outs, [TYPE_HAS], ANNE.MEM.unit_symbol.temp_nid, [MEM.html_symbol.temp_nid]);
            if (unit_lbl_rel!=null) {
                uom.unit_lbl = unit_lbl_rel.to_s;
            }
            else uom.unit_lbl="";  

            var parts = uom.conversion.split('^');
            if (parts.length!=2) {
                uom.factored = uom.conversion;
                $uomNOBJ.factored = $uomNOBJ.conversion;
                $uomNOBJ.uom = uom;
                //$uomNOBJ.value = uom.factored;
                return uom;
            }
            var base = parseFloat(parts[0]);
            var exponent = parseFloat(parts[1]);
            $uomNOBJ.factored = Math.pow(base, exponent);
            uom.factored = $uomNOBJ.factored;
            $uomNOBJ.uom = uom;

            var symbolRel = MEMG.getRelonByTempR(cUnitNOBJ.outs, ANNE.MEM.html_symbol.temp_nid);
            if (symbolRel!=null) {
                uom.symbolHTML = symbolRel.to_s;
            }

            return uom;
        }

        this.param_list = this.minBootNeuron({nid:"7773017254938949678", text:"param list", ntype:NTYPE_CLASS});
        this.param_list.set_value = function ($v) {
            console.warn("THIS HAS NOT BEEN TESTED");
            $v.value = this.buildArrOfValueKeyPairs(ANNE.MEM.__PARAM, [TYPE_HAS]);//ACQL.execute_circuit($v, ACQL._RCPgenTextListOfAllRelnsOf_params);
        }

        this.name = this.minBootNeuron({nid:"7770034521431820596", text:"name", ntype:NTYPE_CLASS});
        this.name.prompt = "What is your name?";
        
        //extra info on names (not required for persons, but nice to have for L1 grpah linkage)
        this.firstName = this.minBootNeuron({nid:"7773598701821046892", text:"first name", ntype:NTYPE_CLASS});
        this.lastName = this.minBootNeuron({nid:"7770897545226351159", text:"last name", ntype:NTYPE_CLASS});
        this.middleName = this.minBootNeuron({nid:"7770526629291509703", text:"middle name", ntype:NTYPE_CLASS});
        this.name_variant = this.minBootNeuron({nid:"16280936195756863604", text:"name variant", ntype:NTYPE_CLASS});
        
        this.polarity = this.minBootNeuron({nid:"7778811780002344235", text:"polarity", ntype:NTYPE_CLASS});
        this.polarity.set_value=function($v) {
            //fast version.
            if ($v["be_"+ANNE.MEM.polarity.temp_nid]!=undefined) {
                var pol = MEMG.getNeuronByTempNid($v["be_"+ANNE.MEM.polarity.temp_nid]);
                if (pol) $v.polarity=pol.polarity;
                return $v.polarity;
            }
            //longer road if no index was set.
            var polarityRelon = MEMG.getRelonByTempR($v.outs, ANNE.MEM.polarity.temp_nid);
            if (polarityRelon!=null && polarityRelon.type==TYPE_BE) {
                var polarity = ANNE.MEM.getNeuronByTempNid(polarityRelon.t);
                if (polarity!=null) {
                    $v.polarity=polarity.polarity;
                    return $v.polarity;
                }
            }
            return null;
        }

        this.positive = this.minBootNeuron({nid:"7771641909001981874", text:"positive", ntype:NTYPE_CLASS});
        this.positive.polarity=1;
        this.positive.pol="pos";
        this.negative = this.minBootNeuron({nid:"777797306450052899", text:"negative", ntype:NTYPE_CLASS});
        this.negative.polarity=-1;
        this.negative.pol="neg";
        this.neutral = this.minBootNeuron({nid:"7772223634951560391", text:"neutral", ntype:NTYPE_CLASS});
        this.neutral.polarity=0;
        this.neutral.pol="neu";
        
        this.sponsored_by = this.minBootNeuron({nid:"7777123609427148390", text:"sponsor", ntype:NTYPE_CLASS});
        this.owned_by = this.minBootNeuron({nid:"7770602908121069199", text:"owner", ntype:NTYPE_CLASS});
        
        //this.annother = this.minBootNeuron({nid:"7779834425608376119", text:"annother", ntype:NTYPE_CLASS});
        this.necrosis = this.minBootNeuron({nid:"7773636908287156942", text:"necrosis", ntype:NTYPE_CLASS});
        this.arbor_prune = this.minBootNeuron({nid:"777918832015451554", text:"arbor prune", ntype:NTYPE_CLASS});
        this.inhibit_relon = this.minBootNeuron({nid:"777581383667732382", text:"inhibit relon", ntype:NTYPE_CLASS});
        this.code = this.minBootNeuron({nid:"7770043314931693691", text:"code", ntype:NTYPE_CLASS});
        this.primary_key = this.minBootNeuron({nid:"7770289090923457729", text:"primary key", ntype:NTYPE_CLASS});
        this.primary_key2 = this.minBootNeuron({nid:"7772038105301798922", text:"primary key2", ntype:NTYPE_CLASS});
        this.aka = this.minBootNeuron({nid:"7777180587425253122", text:"aka", ntype:NTYPE_CLASS});
        this.fka = this.minBootNeuron({nid:"7775797982596188653", text:"formerly known as", ntype:NTYPE_CLASS});
         
        this.PARAM_NAME = this.minBootNeuron({nid:"7772502840575856680", text:"PARAM NAME", ntype:NTYPE_CLASS});
        this.REL_PARAM_NAME = this.minBootNeuron({nid:"7774837507134294874", text:"RELN PARAM NAME", ntype:NTYPE_CLASS});
        
        this.ANNE_DDS_SRC = this.minBootNeuron({nid:"7778133976189354085", text:"ANNE DDS SRC", ntype:NTYPE_CLASS});
        this.require = this.minBootNeuron({nid:"7774954569060780571", text:"require", ntype:NTYPE_CLASS});
        this.validation = this.minBootNeuron({nid:"7778126176197655430", text:"validation", ntype:NTYPE_CLASS});
        this._multiples_ok = this.minBootNeuron({nid:"7777251338142322352", text:"multiples ok", ntype:NTYPE_CLASS});
        //praam name for this is "_multiples_ok" and the gui automatically sets this prop for us on the param relon
        //which the TYPE_SYS relon with reln of _multiples_ok points to.  :)

        this.BE_CHILDREN =  this.minBootNeuron({nid:"7775936058919519512", text:"BE_CHILDREN", ntype:NTYPE_CLASS});
        this.BE_PARENT =  this.minBootNeuron({nid:"7775942183860735787", text:"BE_PARENT", ntype:NTYPE_CLASS});
        

        //role mappers to extract facts from clumps via conveyors.
        this.DEFINES =  this.minBootNeuron({nid:"7778557818829959373", text:"defines", ntype:NTYPE_CLASS});
        this.DEFINED_BY =  this.minBootNeuron({nid:"7772056616837510132", text:"defined by", ntype:NTYPE_CLASS});
        
        this.language = this.minBootNeuron({nid:"7777662227658988771", text:"language", ntype:NTYPE_CLASS});
        this.language.be_lvl = 4;
     
        //build a tester based on unicode_start and unicode_end values.
        this.language.ascertainLang=function() {
            var tests = ["Киану Ривз", "키아누 리브스", "キアヌ・リーブス", "เคอานู รีฟส์",
            "基努·里维斯", "קיאנו ריבס", "Keanu Charles Reeves", "Κιάνου Ριβς", 
            "Κιάνου Τσαρλς Ριβς", "Кіану Рівз", "كيانو ريفز", "基努·李維"];
            for (var t=0; t < tests.length; t++) {
                var langNOBJ = ANNE.MEM.LANG.ascertain_lang(tests[t]);
                if (langNOBJ!=null) console.log(tests[t]+" is "+langNOBJ.text);
                else console.log(tests[t]+" is unknown (fix ucode start/finish)");
            }
        }


        this.lang_english = this.minBootNeuron({nid:"777722641707973679", text:"english", ntype:NTYPE_CLASS});
        this.language.be_lvl = 5;

        this.country = this.minBootNeuron({nid:"7773279276859581115", text:"country", ntype:NTYPE_CLASS});
        this.country._bico="bi bi-geo-alt-fill";

        this.territory = this.minBootNeuron({nid:"7779776635715128867", text:"territory", ntype:NTYPE_CLASS}); 
        this.territory._bico="bi bi-geo-alt-fill";

        this.province = this.minBootNeuron({nid:"7773009519142577067", text:"province", ntype:NTYPE_CLASS}); 
        this.province._bico="bi bi-geo-alt-fill";

        this.empire = this.minBootNeuron({nid:"7774987933806776971", text:"empire", ntype:NTYPE_CLASS});
        this.state = this.minBootNeuron({nid:"7772399593363773393", text:"state", ntype:NTYPE_CLASS});
        this.state._bico="bi bi-geo-alt-fill";

        this.cownty = this.minBootNeuron({nid:"7771790616029822741", text:"county", ntype:NTYPE_CLASS});
        this.cownty._bico="bi bi-geo-alt-fill";

        this.district = this.minBootNeuron({nid:"7770117249951034246", text:"district", ntype:NTYPE_CLASS});
        this.district._bico="bi bi-geo-alt-fill";
        
        this.polity =  this.minBootNeuron({nid:"7773730208317207007", text:"polity", ntype:NTYPE_CLASS});
        this.polity.prompt = "What level of goverence is this location known to be?";

        this.gender =  this.minBootNeuron({nid:"7776266836705549230", text:"gender", ntype:NTYPE_CLASS});
        this.gender.be_lvl = 2;
        this.gender.ai_ico_class="ai-gender";
        this.gender._uico="⚥";
        
        this.male =  this.minBootNeuron({nid:"7774389889917145269", text:"male", ntype:NTYPE_CLASS});
        this.male.be_lvl = 3;
        this.female =  this.minBootNeuron({nid:"7775312609915005294", text:"female", ntype:NTYPE_CLASS});
        this.female.be_lvl = 3;

        // These are used in WHITELIST GARDENS (owner-collab)
        // and BLACKLIST GARDENS (communual)
        // They define collabability  (ability to receive collaboration requests)
        // to save time from requests of a neuron owner
        // that are not desired/welcome.
        this.let_wv = this.minBootNeuron({nid:"7779279922094725823", text:"let wv", ntype:NTYPE_CLASS});
        this.deny_wv = this.minBootNeuron({nid:"7770245499442134915", text:"deny wv", ntype:NTYPE_CLASS});
        this.viplet = this.minBootNeuron({nid:"7776626551577333735", text:"vip let", ntype:NTYPE_CLASS});

        // This adds in neurons to a worldview that are otherwise not apart of it.
        // It does not change ownership of the neuron and is permissionless.
        // anyone can adopt any neuron into their worldview
        this.trusted_adopt = this.minBootNeuron({nid:"7771877111711255601", text:"adopt", ntype:NTYPE_CLASS});
        
        //offering ur neuron into another worldview for them to own
        //this builds a consistent more clean (less noisy) world-view based resource.
        //The giver becomes the sponsor, the receiver becomes the owner.
        //ONLY KEYLESS NEURONS CAN BE HANAI'd.
        this.hanai_out = this.minBootNeuron({nid:"7778381533903269856", text:"Hānai out", ntype:NTYPE_CLASS});
        this.hanai_in = this.minBootNeuron({nid:"7773785541375535981", text:"Hānai in", ntype:NTYPE_CLASS});
        this.hanai_in_all = this.minBootNeuron({nid:"7773285486970861931", text:"Hānai in all", ntype:NTYPE_CLASS});
        
        //response to decline a hanai request.
        //to accept the hanai_in is used.
        this.hanai_in_reject = this.minBootNeuron({nid:"7772421775758796352", text:"hanai in - reject", ntype:NTYPE_CLASS});
        
        this.hanai_and_collab= this.minBootNeuron({nid:"7777343524618811798", text:"hanai & Collab", ntype:NTYPE_CLASS});
        this.hanai_and_collab_reject= this.minBootNeuron({nid:"7776217829049139253", text:"hanai & Collab - reject", ntype:NTYPE_CLASS});
        this.hanai_and_collab_accept= this.minBootNeuron({nid:"7776242842114517972", text:"hanai & Collab - accept", ntype:NTYPE_CLASS});
        
        //"a fund" (boosties bucket is a fund
        this.fund=this.minBootNeuron({nid:"7776570505699865118", text:"fund", ntype:NTYPE_CLASS});
        
        this.boosties = this.minBootNeuron({nid:"7772889444404452615", text:"boosties", ntype:NTYPE_CLASS});
        this.boosties_bucket = this.minBootNeuron({nid:"7776774728242910650", text:"boosties bucket", ntype:NTYPE_CLASS});
        this.winner_payout =this.minBootNeuron({nid:"7778589396232770515", text:"winner payout", ntype:NTYPE_CLASS});
        
        this.annex = this.minBootNeuron({nid:"7775093227033351572", text:"annex", ntype:NTYPE_CLASS});
        this.__setImgThumbDefaultClassImg(this.annex, ANNE.___root+"/imgs/annex_white_ico.png");

        this.annexed_by = this.minBootNeuron({nid:"7771428403886879153", text:"annexed by", ntype:NTYPE_CLASS});
        
        // This removes a trusted adopt.  It is essentially an arbor prune.
        this.abandon = this.minBootNeuron({nid:"7777722740177893415", text:"abandon", ntype:NTYPE_CLASS});

        //this is essentially an inst of an inst.  If a neuron is made outside your worldview and 
        //you want to edit it but the owner isn't collaborating -- you can mirror their neuron
        //and get all data values (reflected from the mirror) into a new neuron.  They are not copied.
        //the mirror is a pointer.  You can then add your own values to your own neuron at will.
        this.mirror = this.minBootNeuron({nid:"7771857764028242565", text:"mirror", ntype:NTYPE_CLASS});

        //this is the garden type for the class neuron.  It cascades down to all sub-classes
        this.garden_type = this.minBootNeuron({nid:"7773397126684636358", text:"garden type", ntype:NTYPE_CLASS});
        
        //this is the garden type for instances.  It is defined on classes and cascades down to instances.
        //this will let the protocol L1 know if an instance can be collaborated on.
        this.inst_garden_type = this.minBootNeuron({nid:"7771542651914330289", text:"inst garden type", ntype:NTYPE_CLASS});
        
        //anyone not blacklist can submit collab requests and ADOPT_OUT requests/responses
        //anyone not blacklisted can create children 
        //GUI will show children of class owned by the same owner or the logged in user.
        this.gt_communal = this.minBootNeuron({nid:"7775536037932928919", text:"communual", ntype:NTYPE_CLASS});
        
        //anyone on the whitelist can submit collab requests and ADOPT_OUT requests/responses
        //anyone on the whitelist can create children 
        //GUI will show children of class owned by the same owner or the logged in user.
        this.gt_owner_collab = this.minBootNeuron({nid:"7778367819213636391", text:"owner collab", ntype:NTYPE_CLASS});
        
        //no collab requests, no delegation
        //GUI will show children of class owned by the same owner.
        this.gt_owner_only = this.minBootNeuron({nid:"7776218453394292289", text:"owner only", ntype:NTYPE_CLASS});
        

        this.andy_collab_accepted = this.minBootNeuron({nid:"7776209936855135438", text:"Collab - accepted", ntype:NTYPE_CLASS});
        this.andy_collab_rejected = this.minBootNeuron({nid:"7775459356522660302", text:"Collab - rejected", ntype:NTYPE_CLASS});
        
        this.anj_accepted= this.minBootNeuron({nid:"7774382206911325219", text:"anjucator - accept", ntype:NTYPE_CLASS});
        this.anj_rejected= this.minBootNeuron({nid:"777184939498367203", text:"anjucator - reject", ntype:NTYPE_CLASS});
        
        this.hanai_and_collab_accept_all =this.minBootNeuron({nid:"7773734743483119116", text:"hanai & Collab - accept ALL", ntype:NTYPE_CLASS});
        
        this.collabRespRelns = [
             this.andy_collab_accepted.temp_nid,
             this.andy_collab_rejected.temp_nid,
             this.hanai_and_collab_reject.temp_nid,
             this.hanai_and_collab_accept.temp_nid,
             this.hanai_and_collab_accept_all.temp_nid
        ];

        //delegationrights
        //can alter class or any sub-class with same owner as issuer of rights.
        this.dr_andy_super_class = this.minBootNeuron({nid:"7776049367153763537", text:"ANDY super class", ntype:NTYPE_CLASS});
        this.dr_andy_super_class.ai_ico_class = "ai-andy";

        //can alter the neuron that issued the rights
        this.dr_andy_defn = this.minBootNeuron({nid:"7776053324438397729", text:"ANDY chg definition", ntype:NTYPE_CLASS});
        this.dr_andy_defn.ai_ico_class = "ai-andy";
        
        //can alter any insts in the down be-line of the issuer of rights
        this.dr_andy_super_insts = this.minBootNeuron({nid:"7776120177414977781", text:"ANDY super insts", ntype:NTYPE_CLASS});
        this.dr_andy_super_insts.ai_ico_class = "ai-andy";
        
        //can alter any insts or sub-classes in the down be-line of the issuer of rights
        this.dr_andy_super_all= this.minBootNeuron({nid:"7776131279600795401", text:"ANDY super all", ntype:NTYPE_CLASS});       
        this.dr_andy_super_all.ai_ico_class = "ai-andy";
        
        this.dr_andy_hanai_in = this.minBootNeuron({nid:"7776131985740735420", text:"Hanai admin", ntype:NTYPE_CLASS});
        this.dr_andy_hanai_in.ai_ico_class = "ai-hanai";
        
        //would be nice to have a superall admin.
        //can hanai and collab ANYTHING. (just like ANNE)
        this.dr_super_super = this.minBootNeuron({nid:"7779079235159468841", text:"ANNE super super admin", ntype:NTYPE_CLASS});
        this.dr_super_super.ai_ico_class = "ai-andy";
        

        //restrictions on delegation ability.
        //only specific are granted rights to delegate to
        this.dr_permitted_reln = this.minBootNeuron({nid:"7771471428900947007", text:"delegation permitted reln", ntype:NTYPE_CLASS});
        this.dr_permitted_reln.ai_ico_class = "ai-andy";
        
        this.garden_rule = this.minBootNeuron({nid:"7779338776490778825", text:"garden rule", ntype:NTYPE_CLASS}); 
        
        //restrictions on rights limiting them to the scope of the permitted nid only.
        this.gt_permit_reln = this.minBootNeuron({nid:"7777831065415748606", text:"garden permitted reln", ntype:NTYPE_CLASS});
          
        this.delegation_rights = this.minBootNeuron({nid:"7776042304763512185", text:"delegation rights", ntype:NTYPE_CLASS});
        this.delegation_reln_tnids = [
            this.dr_andy_super_class.temp_nid, 
            this.dr_andy_defn.temp_nid, 
            this.dr_andy_super_all.temp_nid,
            this.dr_andy_super_insts.temp_nid, 
            this.dr_andy_hanai_in.temp_nid, 
            this.dr_permitted_reln.temp_nid,
            this.dr_super_super.temp_nid
        ];


        //used to document un-conceptualized sentences for later,
        //or to audit against our conceptualization and then verbalization.
        this.original_sentence = this.minBootNeuron({nid:"7777509361790960520", text:"original sentence", ntype:NTYPE_CLASS});
        
        this.PENDING_HANAI = this.minBootNeuron({nid:"7776275301750156832", text:"pending hanai", ntype:NTYPE_CLASS});
        
        
        this.ANNE_THEME = this.minBootNeuron({nid:"7778978696584757203", text:"ANNE theme", ntype:NTYPE_CLASS});
            this.THEME_default = this.minBootNeuron({nid:"7779532849877969632", text:"ANNE theme (default)", ntype:NTYPE_CLASS});
        this.preference = this.minBootNeuron({nid:"7777013225235554698", text:"preference", ntype:NTYPE_CLASS});
        
        
        this.meme = this.minBootNeuron({nid:"7779959823859193016", text:"meme", ntype:NTYPE_CLASS});
        
        this.neuronImg = this.minBootNeuron({nid:"7772973406618599102", text:"neuron img", ntype:NTYPE_CLASS});
        this.neuronImg.ai_ico_class="ai-neuron-img";
        this.neuronImg._bico = "bi bi-image";

        //external dataURL
        this.dataUrl = this.minBootNeuron({nid:"7775367751718991425", text:"data url", ntype:NTYPE_CLASS});

        //path to internal/ships-on-annode file location.
        this.dataRelativePath = this.minBootNeuron({nid:"7772478183226777305", text:"data relative path", ntype:NTYPE_CLASS});
        
        this.primaryNeuronImg = this.minBootNeuron({nid:"7776249509320308559", text:"primary neuron img", ntype:NTYPE_CLASS});
        this.primaryNeuronImg._bico = "bi bi-image-fill";

        this.imgThumbDefaultClass = this.minBootNeuron({nid:"7776781300433802741", text:"img Thumb Default Class", ntype:NTYPE_CLASS});
        
        this.fileX = this.minBootNeuron({nid:"7775682115334476760", text:"file", ntype:NTYPE_CLASS});
        this.fileX._bico="bi bi-file-earmark-fill";

        this.photo_taken_on = this.minBootNeuron({nid:"7776305132990229691", text:"photo taken on", ntype:NTYPE_CLASS});
       
        this.fileFingerPrint = this.minBootNeuron({nid:"7774472646258496586", text:"file finger print", ntype:NTYPE_CLASS});
        this.fileSize = this.minBootNeuron({nid:"7770613549369528475", text:"file size", ntype:NTYPE_CLASS});

        this.neuronVid = this.minBootNeuron({nid:"7772892807996216912", text:"neuron video", ntype:NTYPE_CLASS});
        this.neuronVid.ai_ico_class="ai-neuron-video";
        this.neuronVid._bico="bi bi-camera-video-fill";

        this.media_author = this.minBootNeuron({nid:"7779863026234386372", text:"media author", ntype:NTYPE_CLASS});
        
        this.neuronAudio = this.minBootNeuron({nid:"7777637061760581244", text:"neuron audio", ntype:NTYPE_CLASS});
        this.neuronAudio.ai_ico_class="ai-neuron-audio";
        this.neuronAudio._bico="bi bi-file-music-fill";

        //this.neuronAudioDataUrl = this.minBootNeuron({nid:"7774188933350587286", text:"neuron audio data", ntype:NTYPE_CLASS});

        this.url_part = this.minBootNeuron({nid:"777764039944765162", text:"url part", ntype:NTYPE_CLASS});
        this.base_url = this.minBootNeuron({nid:"7779987300332115752", text:"base url", ntype:NTYPE_CLASS});
        this.ai_ico_class = this.minBootNeuron({nid:"7774405233694204252", text:"ai ico class", ntype:NTYPE_CLASS});
        this.logo_path_with_base = this.minBootNeuron({nid:"7772311622096722944", text:"logo path with base", ntype:NTYPE_CLASS});
        
        this.wcImgUriPart_Flag= this.minBootNeuron({nid:"7772262837359918982", text:"wc flag part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_Map= this.minBootNeuron({nid:"7771813649898124264", text:"wc map part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_Seal= this.minBootNeuron({nid:"7776167254082509668", text:"wc seal part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_Skyline= this.minBootNeuron({nid:"7774254145083522907", text:"wc skyline part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_thumbnail= this.minBootNeuron({nid:"777446366696254759", text:"wc thumbnail part", ntype:NTYPE_CLASS});


        //pts to a neuron img.
        this.logo = this.minBootNeuron({nid:"7774775920254990285", text:"logo", ntype:NTYPE_CLASS});
        
        this.website = this.minBootNeuron({nid:"7772702516735048412", text:"website", ntype:NTYPE_CLASS});
        
        this.mimetype = this.minBootNeuron({nid:"7777176012795245449", text:"mime type", ntype:NTYPE_CLASS});

        //sub of entity
        this.identity = this.minBootNeuron({nid:"7778460139233034749", text:"identity", ntype:NTYPE_CLASS});
        this.identity.wvBIPASS = true;
        this.identity.be_lvl = 4;

            //sub of identity
            this.individual_identity = this.minBootNeuron({nid:"7776971213981656361", text:"individual", ntype:NTYPE_CLASS}); 
            this.individual_identity.ai_ico_class = "ai-individual";
            this.individual_identity.wvBIPASS = true;
            this.individual_identity.be_lvl = 5;
            this.individual_identity._bico="bi bi-person-circle";
            
            this.application_identity = this.minBootNeuron({nid:"7778868011102771570", text:"application", ntype:NTYPE_CLASS});
            this.application_identity.ai_ico_class = "ai-application";
            this.application_identity.wvBIPASS = true;
            this.application_identity.be_lvl = 5;
            this.application_identity._uico = "📱";

            this.corporation_identity = this.minBootNeuron({nid:"7775126160437279358", text:"corporation (identity)", ntype:NTYPE_CLASS}); 
            this.corporation_identity.ai_ico_class = "ai-corporation";
            this.corporation_identity.wvBIPASS = true;
            this.corporation_identity.be_lvl = 5;
            this.corporation_identity._uico="🏢";

            //any neuron can be an oracle, in theory with the ASDF
            this.oracle_identity = this.minBootNeuron({nid:"777941170343283715", text:"oracle", ntype:NTYPE_CLASS});
            this.oracle_identity.ai_ico_class = "ai-oracle";
            this.oracle_identity.wvBIPASS = true;
            this.oracle_identity.be_lvl = 5;
            this.oracle_identity._uico = "🧙‍♂️";
            
            this.miner_identity = this.minBootNeuron({nid:"7774067842216408975", text:"miner", ntype:NTYPE_CLASS});
            this.miner_identity.ai_ico_class = "ai-miner";
            this.miner_identity.wvBIPASS = true;
            this.miner_identity.be_lvl = 5;
            this.miner_identity._uico = "⚒️";
            this.miner_identity._bico="bi bi-hammer";
            this.__setImgThumbDefaultClassImg(this.miner_identity, ANNE.___root+"/imgs/miner_ico.png");

            
            this.annode_identity = this.minBootNeuron({nid:"7777777985638153890", text:"annode", ntype:NTYPE_CLASS});
            this.annode_identity.ai_ico_class = "ai-annode";
            this.annode_identity.wvBIPASS = true;
            this.annode_identity.be_lvl = 5;
            this.annode_identity._uico="💻";
            this.annode_identity._bico="bi bi-database-fill";
            
            this.emerging_intelligence = this.minBootNeuron({nid:"7777798527035018972", text:"emerging intelligence", ntype:NTYPE_CLASS});
            this.emerging_intelligence.ai_ico_class = "ai-ei";
            this.emerging_intelligence.wvBIPASS = true;
            this.emerging_intelligence.be_lvl = 5;
            //this.emerging_intelligence._uico = "⭕";
            

        this.established = this.minBootNeuron({nid:"7774165034042346848", text:"established", ntype:NTYPE_CLASS}); 

        //a class used to instaniate events with.
        //this will inherit a ROLE_ACTION of 
        this.group_event =  this.minBootNeuron({nid:"7777943654909778097", text:"group event", ntype:NTYPE_CLASS});
        
        this.group =  this.minBootNeuron({nid:"7774037854818954199", text:"group", ntype:NTYPE_CLASS});
        this.group.be_lvl = 4;

        this.lolita_express_flights = this.minBootNeuron({nid:"7778886709340031497", text:"lolital express flights", ntype:NTYPE_CLASS});
        
        this.key_person = this.minBootNeuron({nid:"7771668729141593676", text:"key person", ntype:NTYPE_CLASS});

        //from dbpedia .honors
        this.military_honours = this.minBootNeuron({nid:"7775417006091441836", text:"military honours", ntype:NTYPE_CLASS});
        this.military_honours._uico="🎖️";
        this.military_rank = this.minBootNeuron({nid:"7770904604329112073", text:"military rank", ntype:NTYPE_CLASS});
        this.military_rank._uico="🪖";
        this.military_office = this.minBootNeuron({nid:"7771182245027462135", text:"military office", ntype:NTYPE_CLASS});
        this.military_office._uico="🪖";

        this.military_academy = this.minBootNeuron({nid:"7774422634022818293", text:"Military Academy", ntype:NTYPE_CLASS});
        this.military_academy._uico="🛡️";
        

        
        //parent of corp_entity
        this.company = this.minBootNeuron({nid:"7772013259370769039", text:"company", ntype:NTYPE_CLASS});
        this.company._uico="🏢";
        this.company._bico="bi bi-building";

        this.__setImgThumbDefaultClassImg(this.company, ANNE.___root+"/imgs/default_location.png");

        this.company_type = this.minBootNeuron({nid:"7771419270983974071", text:"company type", ntype:NTYPE_CLASS});
        
        //parent of company type VALUES... do not instantiate this!
        //this is a the parent class of corp entity types.
        //NOTE: we don't have to do this like this.  We could just create a list of non-living things
        //and then tag them as lateral TYPE_BE corp types  (like we do with polity->country for locations)
        //I want corp entity types to be a bit more tightly controlled within their own heirarchy..
        //so they are more than just a lateral tie-in, they are a tie-in to a structure.
        this.corp_entity_thetype = this.minBootNeuron({nid:"7776097111646937092", text:"corp entity type", ntype:NTYPE_CLASS});
        this.corp_entity_thetype._uico="🏢";
        this.university = this.minBootNeuron({nid:"7771577574214483635", text:"University", ntype:NTYPE_CLASS});
        this.school = this.minBootNeuron({nid:"7770306012760616108", text:"school", ntype:NTYPE_CLASS});
        
        this.athletics_nickname = this.minBootNeuron({nid:"777756177663289842", text:"athletics nickname", ntype:NTYPE_CLASS});

        this.corp_legal_id = this.minBootNeuron({nid:"7779223539556635762", text:"corp legal id", ntype:NTYPE_CLASS});
        this.setCik = this.minBootNeuron({nid:"7778590084799932279", text:"corp legal id", ntype:NTYPE_CLASS});
        this.product = this.minBootNeuron({nid:"7779910224164104021", text:"product", ntype:NTYPE_CLASS});
        this.industry = this.minBootNeuron({nid:"7777507679900547743", text:"industry", ntype:NTYPE_CLASS});
        this.industry._uico="🏭";

        this.service = this.minBootNeuron({nid:"7778977228411484020", text:"service", ntype:NTYPE_CLASS});

        this.valid_identity_classes=[this.application_identity.nid, this.corporation_identity.nid, this.individual_identity.nid, 
                                    this.miner_identity.nid, this.annode_identity.nid, this.emerging_intelligence.nid];
 
        this.opinionz = this.minBootNeuron({nid:"7770982642614293125", text:"opinionz", ntype:NTYPE_CLASS});
        this.feeling = this.minBootNeuron({nid:"7777627441463363769", text:"feeling", ntype:NTYPE_CLASS});
        this.feelz = this.minBootNeuron({nid:"7770855606276797271", text:"feelz", ntype:NTYPE_CLASS});
        
        this.neg_op_bucket = this.minBootNeuron({nid:"7772878965553112082", text:"feelz", ntype:NTYPE_CLASS});
        this.pos_op_bucket = this.minBootNeuron({nid:"7775045073476665086", text:"feelz", ntype:NTYPE_CLASS});

        //this.thought = this.minBootNeuron({nid:"7777402732566502296", text:"thought", ntype:NTYPE_CLASS});
        this.rememberz = this.minBootNeuron({nid:"7779275038752038222", text:"rememberz", ntype:NTYPE_CLASS});

        //not used.  We now just arbor prune the rememberz.  This will unremember it.
        this.forgetz = this.minBootNeuron({nid:"7770910938663002967", text:"forgetz", ntype:NTYPE_CLASS});

        this.email_address = this.minBootNeuron({nid:"7770280080042939437", text:"email address", ntype:NTYPE_CLASS});
        this.email_address.format_validator=ANNE.UTILS.__emailValidation;
        this.email_address.prompt = "What is your email address? (this will be publicly seen)";

        this.KB_PGP_identity =  this.minBootNeuron({nid:"7779531655061632346", text:"KB PGP identity", ntype:NTYPE_CLASS});
        this.KB_PGP_identity.prompt = "What is your kebybase user name?"

        //this.key = this.minBootNeuron({nid:"7772066628629112453", text:"key", ntype:NTYPE_CLASS});

        this.handler = this.minBootNeuron({nid:"7776885682127776948", text:"handler", ntype:NTYPE_CLASS});
        this.clump_semantic = this.minBootNeuron({nid:"7773872021180785983", text:"semantic clump", ntype:NTYPE_CLASS});
        
        //important MAKE cmds
        //ANNODES know what these are.
        //ANY instance can be made without a custom make cmd but 
        //ANNE creates neurons in ANNE worldview in some special cases and these
        //are important to pay special attention to for ANNE.
        this.MAKE_INTEGER = this.minBootNeuron({nid:"7773497986078093746", text:"MAKE INTEGER", ntype:NTYPE_CLASS});
        this.MAKE_DATE = this.minBootNeuron({nid:"7771903652223599868", text:"MAKE DATE", ntype:NTYPE_CLASS});
        this.MAKE_ANNDLE = this.minBootNeuron({nid:"777623943332475085", text:"MAKE ANNDLE", ntype:NTYPE_CLASS});
        this.MAKE_ACTION = this.minBootNeuron({nid:"7779000837012326152", text:"MAKE ACTION (verb)", ntype:NTYPE_CLASS});
        this.MAKE_LOCATION = this.minBootNeuron({nid:"7777564219572819129", text:"MAKE LOCATION", ntype:NTYPE_CLASS});
        this.MAKE_ANORG = this.minBootNeuron({nid:"7771642700241611064", text:"MAKE ANORG", ntype:NTYPE_CLASS});
        this.MAKE_COMPANY = this.minBootNeuron({nid:"777950857929447723", text:"MAKE COMPANY", ntype:NTYPE_CLASS});
        this.MAKE_IMG = this.minBootNeuron({nid:"7779601435233589280", text:"MAKE IMG", ntype:NTYPE_CLASS});
        this.MAKE_PERSON = this.minBootNeuron({nid:"7771801135056150128", text:"MAKE PERSON", ntype:NTYPE_CLASS});
        this.MAKE_BOUNTY = this.minBootNeuron({nid:"7775295870250788931", text:"MAKE BOUNTY", ntype:NTYPE_CLASS});
        
        //note: events will instantiate their action verb.
        this.MAKE_EVENT = this.minBootNeuron({nid:"7776832446666400296", text:"MAKE EVENT", ntype:NTYPE_CLASS});

        
        this.ACTIVATE = this.minBootNeuron({nid:"7770767425934383201", text:"ACTIVATE", ntype:NTYPE_CLASS});
        this.DEACTIVATE = this.minBootNeuron({nid:"7771229050067657498", text:"ACTIVATE", ntype:NTYPE_CLASS});
        

        this.legal_entity = this.minBootNeuron({nid:"7777212189219449042", text:"legal entity", ntype:NTYPE_CLASS});
        this.site = this.minBootNeuron({nid:"7775256219385612317", text:"site", ntype:NTYPE_CLASS});
        this.site.ai_ico_class="ai-site" 
        this.geo_site = this.minBootNeuron({nid:"7772492123548075783", text:"geo site", ntype:NTYPE_CLASS});
        this.geo_site.ai_ico_class = "ai-geo-site";
        
        //all claims (any user-supplied relon) that wishes to invoke a service for validation
        //will use a validation request relon pointing to the txid they want to validate.
        //the service (annode or other 3rd party) will do its magics (black box or using on-chain proof)..
        //and then burn a valid/invalid relon from their service to the txid of the claim.
        //This can be done using the RRV (request relon validation) annvoice..
        //The service burns the valid/invalid relon and also references the annovice txid
        //which paid for the service in the built in reference field. (TODO piggy back signum field)
        this.validation_request = this.minBootNeuron({nid:"7776161597907601765", text:"validation request", ntype:NTYPE_CLASS});

        //set a default location class img url.
        this.__setImgThumbDefaultClassImg(this.location, ANNE.___root+"/imgs/default_location.png");

        this.constituent_of = this.minBootNeuron({nid:"7771195041236647730", text:"constituent of", ntype:NTYPE_CLASS});
        this.constituent = this.minBootNeuron({nid:"7775238842023699728", text:"constituent", ntype:NTYPE_CLASS});

        this.participant =  this.minBootNeuron({nid:"7772308481505259319", text:"participant", ntype:NTYPE_CLASS});
        this.participant_of =  this.minBootNeuron({nid:"7776908840642789745", text:"participant of", ntype:NTYPE_CLASS});
            
        this.founder =  this.minBootNeuron({nid:"7773741333644308987", text:"founder", ntype:NTYPE_CLASS});
        this.founder_of =  this.minBootNeuron({nid:"7778112941848478136", text:"founder of", ntype:NTYPE_CLASS});
        
        //test for ANNE, we don't need diff relns for direction.. we use TYPE_BE and TYPE_HAS for that.
        this.member =  this.minBootNeuron({nid:"7774234737693337904", text:"member", ntype:NTYPE_CLASS});
        this.member_of =  this.minBootNeuron({nid:"7771323162068539738", text:"member of", ntype:NTYPE_CLASS});
        
        this.board_member =  this.minBootNeuron({nid:"7773721544735553188", text:"board member", ntype:NTYPE_CLASS});
        this.board_member_of =  this.minBootNeuron({nid:"7775205884503390304", text:"board member of", ntype:NTYPE_CLASS});
        
        this.corpSponsor =  this.minBootNeuron({nid:"7777670247182851992", text:"corp sponsor", ntype:NTYPE_CLASS});
        this.corpSponsor_of =  this.minBootNeuron({nid:"7772087542288975131", text:"corp sponsor of", ntype:NTYPE_CLASS});
        
        //never instantiated, only a gen prop to point to other companies...
        this.subsidary = this.minBootNeuron({nid:"777364764299065896", text:"subsidary", ntype:NTYPE_CLASS}); 
        this.subsidary_of = this.minBootNeuron({nid:"77754005942822853", text:"subsidary of", ntype:NTYPE_CLASS}); 

        this.division = this.minBootNeuron({nid:"17285996896823940540", text:"division", ntype:NTYPE_CLASS}); 
        this.division_of = this.minBootNeuron({nid:"11438744566110935559", text:"division of", ntype:NTYPE_CLASS}); 

        this.appointed = this.minBootNeuron({nid:"2677740920383363638", text:"appointed", ntype:NTYPE_CLASS}); 
        
        //THIS IS A GOOD CONCEPTUALIZATION EXAMPLE OF HOW TO BUILD THIS CONCEPT FOR LATER NLP USE.
        this.appointed_by = this.minBootNeuron({nid:"18061812464206165953", text:"appointed by", ntype:NTYPE_CLASS}); 

        this.preceded_by = this.minBootNeuron({nid:"18231916307500268961", text:"preceded by", ntype:NTYPE_CLASS});
        this.succeeded_by = this.minBootNeuron({nid:"12775004814305050777", text:"succeeded by", ntype:NTYPE_CLASS});

        //things TYPE_BE relationships of this.
        //it a short-hand clump where the to is really the ROLE_THEME_THING
        //x be founder of y
        //x be participant of y
        this.BE_ofs = [this.participant_of.nid, this.founder_of.nid, this.member_of.nid, this.board_member_of.nid,
            this.corpSponsor_of.nid, this.subsidary_of.nid, this.division_of.nid];

        //things TYPE_HAS relationships of this.
        //y has founder x
        //y has participant x
        this.HAS_no_ofs = [this.participant.nid, this.founder.nid, this.member.nid, this.board_member.nid,
            this.corpSponsor.nid, this.subsidary.nid, this.division.nid];

        //a type of corp
        this.conglomerate = this.minBootNeuron({nid:"8072940268512505445", text:"conglomerate", ntype:NTYPE_CLASS});

        this.ideology = this.minBootNeuron({nid:"17020879412571846679", text:"ideology", ntype:NTYPE_CLASS});
        
        //alumni should makea HAS relon pointing to alumnis.
        //harvard->has->alumni->Bob Smith
        this.alumni = this.minBootNeuron({nid:"7775772515552382303", text:"alumni", ntype:NTYPE_CLASS});

        //alumni_of should make a BE RELON from the alumni to what they are an alumni of
        this.alumni_of = this.minBootNeuron({nid:"7770037193997679017", text:"alumni of", ntype:NTYPE_CLASS});

        this.wikispooks_desc = this.minBootNeuron({nid:"7775709052747670582", text:"wikispooks desc", ntype:NTYPE_CLASS}); 

        //thise are essentially actionvb+when role
        this.foundedOn = this.minBootNeuron({nid:"7772745521469079754", text:"founded on", ntype:NTYPE_CLASS}); 
        
        //for clumps with date OR location roles included.
        this.founded = this.minBootNeuron({nid:"7618523810988349729", text:"founded", ntype:NTYPE_CLASS}); 
        
        this.power_family = this.minBootNeuron({nid:"777842302370981935", text:"power family", ntype:NTYPE_CLASS}); 
        
        this.MAKE_SYMBOL = this.minBootNeuron({nid:"7771149320550437941", text:"MAKE KEYWORD", ntype:NTYPE_CLASS});


        //person id
        this.TMDB_id = this.minBootNeuron({nid:"777811874948779992", text:"TMDB pid", ntype:NTYPE_CLASS});
        this.__setClassBaseURL(this.TMDB_person_id, "https://www.themoviedb.org/");
        this.ai_ico_class="ai-TMDB";

        //
        this.bioguide_id = this.minBootNeuron({nid:"7775032836974177659", text:"buioguide", ntype:NTYPE_CLASS});
        this.thomas_id = this.minBootNeuron({nid:"7778923864354766137", text:"thomas_id", ntype:NTYPE_CLASS});
        this.youtube_id = this.minBootNeuron({nid:"7771610674492367662", text:"youtube_id", ntype:NTYPE_CLASS});
        this.govtrack = this.minBootNeuron({nid:"7770925240349684802", text:"govtrack", ntype:NTYPE_CLASS});
        this.icpsr_prez = this.minBootNeuron({nid:"7773850913328907555", text:"icpsr_prez", ntype:NTYPE_CLASS});
        this.opensecrets = this.minBootNeuron({nid:"7770478812806785802", text:"opensecrets", ntype:NTYPE_CLASS});
        this.fec = this.minBootNeuron({nid:"7779064998469501513", text:"fec", ntype:NTYPE_CLASS});
        this.votesmart = this.minBootNeuron({nid:"7777143372995579784", text:"votesmart", ntype:NTYPE_CLASS});
        this.lis = this.minBootNeuron({nid:"7775005862771329923", text:"lis", ntype:NTYPE_CLASS});
        this.ballotpedia = this.minBootNeuron({nid:"7770151981975829494", text:"ballotpedia", ntype:NTYPE_CLASS});
        this.maplight = this.minBootNeuron({nid:"7770298099837584694", text:"maplight", ntype:NTYPE_CLASS});
        this.cspan = this.minBootNeuron({nid:"7776274437771740534", text:"cspan", ntype:NTYPE_CLASS});
        this.icpsr = this.minBootNeuron({nid:"7776338279415159744", text:"icpsr", ntype:NTYPE_CLASS});
        this.google_entity_id = this.minBootNeuron({nid:"7772247736870247808", text:"google_entity_id", ntype:NTYPE_CLASS});
        this.house_history = this.minBootNeuron({nid:"7772280668716687175", text:"house_history", ntype:NTYPE_CLASS});
        this.wikidata = this.minBootNeuron({nid:"7775289459331965701", text:"wikidata", ntype:NTYPE_CLASS});
        
        //is this a dupe?
        //this.wikipediaDEAD = this.minBootNeuron({nid:"7772905047713397630", text:"wikipedia", ntype:NTYPE_CLASS});
    
        this.wikipart = this.minBootNeuron({nid:"7772333349880064178", text:"wikiurl", ntype:NTYPE_CLASS});
        
        //meaning, from wikipedia.
        this.wiki_meaning = this.minBootNeuron({nid:"17504329198913088809", text:"meaning", ntype:NTYPE_CLASS});
        
        this.dbpedia_id = this.minBootNeuron({nid:"777992912015974318", text:"dbpedia id", ntype:NTYPE_CLASS});
        this.dbpedia_respart = this.minBootNeuron({nid:"7779668581171663468", text:"dbpedia resource part", ntype:NTYPE_CLASS});
        
        this.wikispooks_respart = this.minBootNeuron({nid:"7778158657377162702", text:"wikispooks resource part", ntype:NTYPE_CLASS});
        
        //person id!
        this.IMDB_id = this.minBootNeuron({nid:"7777167264347475422", text:"IMDB pid", ntype:NTYPE_CLASS});
        //https://www.imdb.com/name/
        this.IMDB_id.ai_ico_class="ai-IMDB";
        
        this.instagram_id = this.minBootNeuron({nid:"7771145768785067039", text:"Instagram id", ntype:NTYPE_CLASS});
        this.__setClassBaseURL(this.TMDB_person_id, "https://www.instagram.com/");
        this.instagram_id.ai_ico_class="ai-instagram";

        this.twitter_id = this.minBootNeuron({nid:"7771320139782962995", text:"Twitter id", ntype:NTYPE_CLASS});
        this.__setClassBaseURL(this.twitter_id, "https://twitter.com/");
        this.twitter_id.ai_ico_class="ai-twitter";

        this.facebook_id = this.minBootNeuron({nid:"7779997270324532488", text:"Facebook id", ntype:NTYPE_CLASS});
        this.facebook_id.ai_ico_class="ai-facebook";

        this.birthname = this.minBootNeuron({nid:"7770585205993013954", text:"birth name", ntype:NTYPE_CLASS});
        
        this.birthday = this.minBootNeuron({nid:"7774793041983944904", text:"birthday", ntype:NTYPE_CLASS});
        this.birthday._uico="🎂";
        
        this.birthplace = this.minBootNeuron({nid:"7777813598381855599", text:"birth place", ntype:NTYPE_CLASS});
        
        //used in full clumps -- WHERE is the birthplace, WHEN is the birthday
        this.born = this.minBootNeuron({nid:"7772914512930290384", text:"born", ntype:NTYPE_CLASS});
       
        //for short hand-relons
        this.deathday = this.minBootNeuron({nid:"7775451663389638990", text:"deathday", ntype:NTYPE_CLASS});
        this.deathday.ai_ico_class="ai-death";
        this.deathday._uico="💀";

        this.deathplace = this.minBootNeuron({nid:"7776799915004256366", text:"death place", ntype:NTYPE_CLASS});
        this.deathplace.ai_ico_class="ai-death";
        this.deathplace._uico="🪦";

        this.apprehended = this.minBootNeuron({nid:"7773509819489469268", text:"apprehended", ntype:NTYPE_CLASS});
        this.apprehended._uico = "🚨";

        //used in full clumps -- WHERE is the birthplace, WHEN is the birthday
        this.die  = this.minBootNeuron({nid:"7772869077031626256", text:"die", ntype:NTYPE_CLASS}); 

        this.streampay = this.minBootNeuron({nid:"7779403914189757295", text:"♼ streampay ♼", ntype:NTYPE_CLASS}); 
        this.block_frequency = this.minBootNeuron({nid:"7772686674875797392", text:"block frequency", ntype:NTYPE_CLASS});
        this.min_start_block = this.minBootNeuron({nid:"7776489666160494893", text:"min start block", ntype:NTYPE_CLASS});
        
        this.streampay_src = this.minBootNeuron({nid:"77738031779292675", text:"streampay src", ntype:NTYPE_CLASS});
        this.streampay_src_theFund = this.minBootNeuron({nid:"7773274639808292524", text:"streampay src - the fund", ntype:NTYPE_CLASS});
        this.streampay_src_owner = this.minBootNeuron({nid:"7774646684344886968", text:"streampay src - owner of fund", ntype:NTYPE_CLASS});
        
        this.keep_min_balance = this.minBootNeuron({nid:"777687915780771745", text:"keep min balance", ntype:NTYPE_CLASS});
        this.max_total_payments = this.minBootNeuron({nid:"7778146027628050531", text:"max total payments", ntype:NTYPE_CLASS});
        this.min_streampay_payment = this.minBootNeuron({nid:"7777690415839350242", text:"min streampay payment", ntype:NTYPE_CLASS});
        
        this.ANNE_DDS =  this.minBootNeuron({nid:"7779217783975400919", text:"ANNE DDS", ntype:NTYPE_CLASS});

        //new method to make new neurons keys.  This RELN is used to instruct the ANNODE.
        //Keyless neurons are able to have delegation set.
        //They also can be collab'd with like key'd neurons if their garden type permits
        //collaboration.
        this.MAKE_KEYLESS = this.minBootNeuron({nid:"7776157600356963165", text:"MAKE KEYLESS", ntype:NTYPE_CLASS});
            
        //this.ANNODE_BAN_NID = this.minBootNeuron({nid:"", text:"ANNODE ban nid", ntype:NTYPE_CLASS});
        
        //irrevokable one-time freeze of neuron that will prevent any future OUTS from
        //being broadcast..  and it will also prevent any future ARBOR PRUNES.
        //To extend a froze neuron any further requires a mirror neuron....
        this.CMD_FREEZE = this.minBootNeuron({nid:"7776758007315842114", text:"freeze neuron", ntype:NTYPE_CLASS});
        this.CMD_FREEZE.ai_ico_class="ai-freeze";

        this.CMD_KEYLESS_PAYOUT = this.minBootNeuron({nid:"7771064906243877097", text:"keyless payout", ntype:NTYPE_CLASS});
        
        //will trigger a merge neuron request and must be sent as a collab (not auto-delegatable)
        this.CMD_MERGE_NEURON = this.minBootNeuron({nid:"7775453604074825079", text:"MERGE NEURON", ntype:NTYPE_CLASS});
        this.MERGED_WITH = this.minBootNeuron({nid:"7775681672672970556", text:"MERGED WITH", ntype:NTYPE_CLASS});
        
        //for testing.. in console.. NOT NEEDED IN MEM on release.
        this.usa = this.minBootNeuron({nid:"7774152290984691776", text:"United States of America", ntype:NTYPE_INST});
        this.american = this.minBootNeuron({nid:"10076243270600789032", text:"American", ntype:NTYPE_INST});
        
        this.role = this.minBootNeuron({nid:"7778531759128370850", text:"ec:role", ntype:NTYPE_CLASS});

        //first pass resolution into thing and thing mod
        this.role_thing = this.minBootNeuron({nid:"7770394591629606840", text:"role_thing", ntype:NTYPE_CLASS});
        this.role_thing.ai_ico_class="ai-thing";

        this.role_thing_mod = this.minBootNeuron({nid:"7773772721114312848", text:"role_thing_mod", ntype:NTYPE_CLASS});
        

        this.role_actor = this.minBootNeuron({nid:"7777573229173770374", text:"role_actor", ntype:NTYPE_CLASS}); 
        this.role_actor.ai_ico_class="ai-thing";

        this.role_actor_mod = this.minBootNeuron({nid:"7777519193840188065", text:"role_actor_mod", ntype:NTYPE_CLASS}); 
        this.role_affected_thing = this.minBootNeuron({nid:"777744227568420012", text:"role_affected_thing", ntype:NTYPE_CLASS});
        this.role_affected_thing.ai_ico_class="ai-aff-thing";
        this.role_affected_thing_mod = this.minBootNeuron({nid:"7774964023545941379", text:"role_affected_thing_mod", ntype:NTYPE_CLASS});
        
        //note: in ANNE all verbs are encoded into the R_ACTION role.  They can be thinking, feeling, doing, or being verbs.
        //it doesn't matter.  Be verbs (in english) are children of ec_is (beingness) and they have some differences.
        
        //aux vbs that alter progressive/tense. 
        this.role_aux_action = this.minBootNeuron({nid:"777983066090703356", text:"role_aux_action", ntype:NTYPE_CLASS});
        this.role_mode = this.minBootNeuron({nid:"7774040954807558904", text:"role_mode", ntype:NTYPE_CLASS});
        this.role_tense = this.minBootNeuron({nid:"7770229184388498871", text:"role_tense", ntype:NTYPE_CLASS});
        this.role_action = this.minBootNeuron({nid:"7777055149623754971", text:"role_action", ntype:NTYPE_CLASS});
        this.role_action.ai_ico_class="ai-action";

        this.role_action_thing = this.minBootNeuron({nid:"7776460781235543752", text:"role_action_thing", ntype:NTYPE_CLASS});
        this.role_be_thing = this.minBootNeuron({nid:"7773949836736341630", text:"role_be_thing", ntype:NTYPE_CLASS});
        this.role_be_thing_mod = this.minBootNeuron({nid:"777751343740502205", text:"role_be_thing_mod", ntype:NTYPE_CLASS});
        
        //a role that points to another clump with no particular semantic relationships specified
        this.role_clump = this.minBootNeuron({nid:"7770332898798547236", text:"role_clump", ntype:NTYPE_CLASS});

        this.role_connector = this.minBootNeuron({nid:"777763967723953677", text:"role_connector", ntype:NTYPE_CLASS});
        this.role_posn_thing = this.minBootNeuron({nid:"7773433027241820655", text:"role_posn_thing", ntype:NTYPE_CLASS});
        this.role_posn_thing_mod = this.minBootNeuron({nid:"7776712633779543606", text:"role_posn_thing_mod", ntype:NTYPE_CLASS});
        this.role_manner = this.minBootNeuron({nid:"7777019863026566731", text:"role_manner", ntype:NTYPE_CLASS});
        this.role_theme_thing = this.minBootNeuron({nid:"7777401229711945348", text:"role_theme_thing", ntype:NTYPE_CLASS});
        this.role_theme_thing_mod = this.minBootNeuron({nid:"7776071931241757674", text:"role_theme_thing_mod", ntype:NTYPE_CLASS});
       
        this.role_where = this.minBootNeuron({nid:"7779297186094235437", text:"role_where", ntype:NTYPE_CLASS});
        this.role_where.ai_ico_class="ai-where";
        this.role_where_mod = this.minBootNeuron({nid:"7776640823686512237", text:"role_where_mod", ntype:NTYPE_CLASS});
        this.role_source = this.minBootNeuron({nid:"7779889734760631010", text:"role_source", ntype:NTYPE_CLASS});
        this.role_path = this.minBootNeuron({nid:"7778169513595010118", text:"role_path", ntype:NTYPE_CLASS});
        this.role_goal = this.minBootNeuron({nid:"777522762176315595", text:"role_goal", ntype:NTYPE_CLASS});
       
        this.role_when = this.minBootNeuron({nid:"777581408555193886", text:"role_when", ntype:NTYPE_CLASS});
        this.role_when.ai_ico_class="ai-when";
        this.role_when_mod = this.minBootNeuron({nid:"7774341790656657922", text:"role_when_mod", ntype:NTYPE_CLASS});
        this.role_start = this.minBootNeuron({nid:"7774749544848267976", text:"role_start", ntype:NTYPE_CLASS});
        this.role_end = this.minBootNeuron({nid:"7776097378912616014", text:"role_end", ntype:NTYPE_CLASS});

        this.role_instrument = this.minBootNeuron({nid:"7779418330103123081", text:"role_instrument", ntype:NTYPE_CLASS});
        this.role_instrument_mod = this.minBootNeuron({nid:"7772775616468645341", text:"role_instrument", ntype:NTYPE_CLASS});
        this.role_resulting_new_thing = this.minBootNeuron({nid:"7770394682139219297", text:"role_resulting_new_thing", ntype:NTYPE_CLASS});
        this.role_reason = this.minBootNeuron({nid:"7777005928865157015", text:"role_reason", ntype:NTYPE_CLASS});
        this.role_purpose = this.minBootNeuron({nid:"7772840210245397923", text:"role_purpose", ntype:NTYPE_CLASS});
        this.role_question = this.minBootNeuron({nid:"7778784064660352579", text:"role_question", ntype:NTYPE_CLASS});
                
        //link to other clumps
        this.role_cause = this.minBootNeuron({nid:"777104212581317832", text:"role_cause", ntype:NTYPE_CLASS});
        this.role_cause.ai_ico_class="ai-why";
        this.role_effect = this.minBootNeuron({nid:"7772214179202770375", text:"role_effect", ntype:NTYPE_CLASS});
        this.role_effect.ai_ico_class="ai-effect";
        this.role_next = this.minBootNeuron({nid:"7775884336586790023", text:"role_next", ntype:NTYPE_CLASS});
        this.role_prev = this.minBootNeuron({nid:"7773564479359111744", text:"role_prev", ntype:NTYPE_CLASS});
        this.role_cdx = this.minBootNeuron({nid:"7772055522386399843", text:"role_cdx", ntype:NTYPE_CLASS});
        this.role_cdx_result = this.minBootNeuron({nid:"7776763659152580485", text:"role_cdx_result", ntype:NTYPE_CLASS});
        this.role_why = this.minBootNeuron({nid:"7779965580540643037", text:"role_why", ntype:NTYPE_CLASS});
        this.role_why.ai_ico_class="ai-why";
        
        this.role_feelz = this.minBootNeuron({nid:"777531590761518279", text:"role_feelz", ntype:NTYPE_CLASS});
        this.role_thought = this.minBootNeuron({nid:"7776728107601212876", text:"role_thought", ntype:NTYPE_CLASS});
        this.role_fact = this.minBootNeuron({nid:"7777226313918390844", text:"role_fact", ntype:NTYPE_CLASS});
        this.role_state = this.minBootNeuron({nid:"7777228404354952479", text:"role_state", ntype:NTYPE_CLASS});
        this.role_implies = this.minBootNeuron({nid:"777391174437512597", text:"role_implies", ntype:NTYPE_CLASS});
        this.role_infer = this.minBootNeuron({nid:"7779731604484117151", text:"role_infer", ntype:NTYPE_CLASS});

        //documenting information source
        this.role_src_ref = this.minBootNeuron({nid:"7779273238068191847", text:"role_src_ref", ntype:NTYPE_CLASS});
        this.role_src_ref.ai_ico_class="ai-ref-src";
        
        this.role_meta_meaning = this.minBootNeuron({nid:"7778288624518302407", text:"role_meta_meaning", ntype:NTYPE_CLASS});
        
        this.role_standards=[
            this.role_action.temp_nid, 
            this.role_manner.temp_nid, 
            this.role_actor.temp_nid, 
            this.role_actor_mod.temp_nid, 
            this.role_tense.temp_nid,
            this.role_when.temp_nid, 
            this.role_when_mod.temp_nid, 
            this.role_where.temp_nid, 
            this.role_where_mod.temp_nid
        ];

        //ec patterns use ANNE ops. (see ANNEX)
        this.ec_pattern = this.minBootNeuron({nid:"7770697322248948996", text:"ec.pattern", ntype:NTYPE_CLASS});    
        this.name_pattern = this.minBootNeuron({nid:"7770271905530827053", text:"name_pattern", ntype:NTYPE_CLASS});    
        

        this.__max =  this.minBootNeuron({nid:"7772748822322176774", text:"__max", ntype:NTYPE_CLASS});    
        this.__min =  this.minBootNeuron({nid:"7772077984786233524", text:"__min", ntype:NTYPE_CLASS});    
        this.__min_digits =  this.minBootNeuron({nid:"7771179556152284330", text:"__min_digits", ntype:NTYPE_CLASS});    
        this.__optional =  this.minBootNeuron({nid:"7771382453261270234", text:"__optional", ntype:NTYPE_CLASS});    
        this.__required =  this.minBootNeuron({nid:"7772207950419895534", text:"__required", ntype:NTYPE_CLASS});    
        this.__value =  this.minBootNeuron({nid:"7776251561494105914", text:"__value", ntype:NTYPE_CLASS});    
        this.__next_active =  this.minBootNeuron({nid:"7775608436552493714", text:"__next_active", ntype:NTYPE_CLASS});    
        this.__resolves_to =  this.minBootNeuron({nid:"7776251561494105914", text:"__resolves_to", ntype:NTYPE_CLASS});    
        this.__make_new =  this.minBootNeuron({nid:"7770730581584298096", text:"__make_new", ntype:NTYPE_CLASS});    
        

        this.ec_part =  this.minBootNeuron({nid:"7770730581584298096", text:"ec:part", ntype:NTYPE_CLASS});    
        this.ec_whole =  this.minBootNeuron({nid:"7773902746956577014", text:"ec:whole", ntype:NTYPE_CLASS});    
        
        //these are words/phrases we toss away after we get the signal.  We do not store them.  They help us
        //know what is coming.
        
        this.ec_expect_actor = this.minBootNeuron({nid:"7770746507409045605", text:"x_actor", ntype:NTYPE_CLASS});    
        this.ec_expect_action = this.minBootNeuron({nid:"7778899167440690152", text:"x_action", ntype:NTYPE_CLASS});    
        this.ec_expect_manner = this.minBootNeuron({nid:"7772660770283904989", text:"x_manner", ntype:NTYPE_CLASS});    
        this.ec_expect_affected_thing = this.minBootNeuron({nid:"8150309968616768260", text:"x_affected_thing", ntype:NTYPE_CLASS});    
        this.ec_expect_clump = this.minBootNeuron({nid:"9146769968576781671", text:"x_clump", ntype:NTYPE_CLASS});    
        this.ec_expect_concept = this.minBootNeuron({nid:"13998275032128243027", text:"x_concept", ntype:NTYPE_CLASS});    
        this.ec_expect_condition = this.minBootNeuron({nid:"12319126297907506798", text:"x_condition", ntype:NTYPE_CLASS});    
        this.ec_expect_goal = this.minBootNeuron({nid:"4322029487966010252", text:"x_goal", ntype:NTYPE_CLASS});    
        this.ec_expect_posn = this.minBootNeuron({nid:"13698091595456770706", text:"x_posn", ntype:NTYPE_CLASS});    
        this.ec_expect_question = this.minBootNeuron({nid:"777713567747830967", text:"x_question", ntype:NTYPE_CLASS});    
        this.ec_expect_role = this.minBootNeuron({nid:"3437301647300744555", text:"x_role_*", ntype:NTYPE_CLASS});    
        this.ec_expect_state = this.minBootNeuron({nid:"2704422592051914885", text:"x_state", ntype:NTYPE_CLASS});    
        this.ec_expect_theme_thing = this.minBootNeuron({nid:"3069042795682288556", text:"x_theme_thing", ntype:NTYPE_CLASS});    
        this.ec_expect_thing = this.minBootNeuron({nid:"614649676553071566", text:"x_thing", ntype:NTYPE_CLASS});    
        this.ec_expect_when = this.minBootNeuron({nid:"1703459443289362399", text:"x_when", ntype:NTYPE_CLASS});    
        this.ec_expect_where = this.minBootNeuron({nid:"1875762255298275190", text:"x_where", ntype:NTYPE_CLASS});    
        this.ec_expect_actionthing = this.minBootNeuron({nid:"7774775137559625245", text:"x_actionthing", ntype:NTYPE_CLASS});    
        this.ec_expect_more_of_prev= this.minBootNeuron({nid:"7776848702697001980", text:"x_more_of_prev", ntype:NTYPE_CLASS});    
        this.ec_expect_thing_mod= this.minBootNeuron({nid:"8754326448223055429", text:"x_thing_mod", ntype:NTYPE_CLASS});    


        this.role_mapper = this.minBootNeuron({nid:"7770273269579595019", text:"role mapper", ntype:NTYPE_CLASS});
        this.role_mapper_actor = this.minBootNeuron({nid:"7772508118194483571", text:"role mapper - actor (agent/subj)", ntype:NTYPE_CLASS});
        this.role_mapper_be_tense = this.minBootNeuron({nid:"7778791128313669071", text:"mapper - be tense", ntype:NTYPE_CLASS});
        this.role_mapper_when = this.minBootNeuron({nid:"7779712508649805541", text:"role mapper - when", ntype:NTYPE_CLASS});
        this.role_mapper_where = this.minBootNeuron({nid:"7770345493567182036", text:"role mapper - where", ntype:NTYPE_CLASS});
        this.role_mapper_action = this.minBootNeuron({nid:"7775935060062601273", text:"role mapper - action", ntype:NTYPE_CLASS});
        this.role_mapper_theme = this.minBootNeuron({nid:"7779583700584149129", text:"role mapper - theme", ntype:NTYPE_CLASS});
        this.role_mapper_affected_thing = this.minBootNeuron({nid:"7770417822110584754", text:"role mapper - affected thing", ntype:NTYPE_CLASS});
        this.role_mapper_tense = this.minBootNeuron({nid:"7774822487994741720", text:"role mapper - tense", ntype:NTYPE_CLASS});
        
        this.role_mappers = [
            this.role_mapper_actor.temp_nid,
            this.role_mapper_be_tense.temp_nid,
            this.role_mapper_when.temp_nid,
            this.role_mapper_where.temp_nid,
            this.role_mapper_action.temp_nid,
            this.role_mapper_theme.temp_nid,
            this.role_mapper_affected_thing.temp_nid,
            this.role_mapper_tense.temp_nid
        ];
        this.role_map = [];
        this.role_map[this.role_mapper_actor.temp_nid]= this.role_actor.temp_nid;
        this.role_map[this.role_mapper_when.temp_nid]= this.role_when.temp_nid;
        this.role_map[this.role_mapper_where.temp_nid]= this.role_where.temp_nid;
        this.role_map[this.role_mapper_action.temp_nid]= this.role_action.temp_nid;
        this.role_map[this.role_mapper_theme.temp_nid]= this.role_theme_thing.temp_nid;
        this.role_map[this.role_mapper_affected_thing.temp_nid]= this.role_affected_thing.temp_nid;
        this.role_map[this.role_mapper_tense.temp_nid]= this.role_tense.temp_nid;


        //pts to the from slot of a relon used in role mappers.
        this.relon_from = this.minBootNeuron({nid:"777780868832582872", text:"relon from", ntype:NTYPE_CLASS});
        this.relon_to = this.minBootNeuron({nid:"7772898090768696320", text:"relon to", ntype:NTYPE_CLASS});
        this.relon_reln = this.minBootNeuron({nid:"7773696674962705615", text:"relon reln", ntype:NTYPE_CLASS});
        
        
        this.SMART_SEARCH_ONLY = this.minBootNeuron({nid:"7773734358383597712", text:"smart search only", ntype:NTYPE_CLASS});

        //private data neuron (definitions will be on-chain when preload occurs, as with all other data
        //here in the bare-bones neural pre-load)
        this.ENCRYPED_DATA_TYPE = this.minBootNeuron({nid:"7777437385154819845", text:"encrypted data type", ntype:NTYPE_CLASS});
        this.PDN = this.minBootNeuron({nid:"7776729867696386401", text:"private data neuron", ntype:NTYPE_CLASS});
        this.PDN.wvBIPASS = true;
        this.PDN_PUBLIC_ACCESS = this.minBootNeuron({nid:"7777374627033925008", text:"PDN public access", ntype:NTYPE_CLASS});
        this.PDN_PUBLIC_ACCESS.wvBIPASS = true;
        this.PDN_VIP = this.minBootNeuron({nid:"777865917767872287", text:"PDN VIP", ntype:NTYPE_CLASS});
        this.PDN_VIP.wvBIPASS = true;
        this.PDN_BEARER_TOKEN = this.minBootNeuron({nid:"7772868456433715520", text:"PDN Bearer Token", ntype:NTYPE_CLASS});
        this.PDN_BEARER_TOKEN.wvBIPASS = true;
        //bipass needed ins PDNS are owner only-- so general public won't find
        //them in downline searches (gardenEnforce fails).
        //we want them to stay owner-only for NO COLLAB thus we bipass WV to allow
        //searches to find them w/o user putting them in a white list.
        this.bearer_token = this.minBootNeuron({nid:"7772559250440764390", text:"bearer token", ntype:NTYPE_CLASS});
        this.PDN_OP_QUESTION = this.minBootNeuron({nid:"777461361898715860", text:"PDN OP Q", ntype:NTYPE_CLASS});
        this.PDN_OP_REVEAL_ALL = this.minBootNeuron({nid:"7778580360323145476", text:"PDN OP REVEAL ALL", ntype:NTYPE_CLASS});
        this.PDN_EQUALITY_QUESTION = this.minBootNeuron({nid:"7771390471443803065", text:"PDN EQUALITY Q", ntype:NTYPE_CLASS});
        this.PDN_SECRET = this.minBootNeuron({nid:"7771501455189830305", text:"PDN secret", ntype:NTYPE_CLASS});
        this.ANNODE_URI = this.minBootNeuron({nid:"7775200563510845790", text:"ANNODE URI", ntype:NTYPE_CLASS});

        this.PDN_ANSWER = this.minBootNeuron({nid:"7778569016061145146", text:"PDN answer", ntype:NTYPE_CLASS});


        //limited set of XOPs used in PDNs that are helpful to have a var in MEM memory.
        this.XOP = this.minBootNeuron({nid:"777559007866194998", text:"ANNE X OPERATORS", ntype:NTYPE_CLASS});
        
        this.XOP_equal = this.minBootNeuron({nid:"7775618696518156032", text:"==", ntype:NTYPE_CLASS});
        this.XOP_equal.asdf_date_disp = "on";
        this.XOP_notequal = this.minBootNeuron({nid:"7775628753205217910", text:"!=", ntype:NTYPE_CLASS});
        this.XOP_notequal.asdf_date_disp = "not on";
        this.XOP_lessthan= this.minBootNeuron({nid:"7775667956029190870", text:"<", ntype:NTYPE_CLASS});
        this.XOP_lessthan.asdf_date_disp = "before";
        this.XOP_lessthanequal= this.minBootNeuron({nid:"7775675477949841201", text:"<=", ntype:NTYPE_CLASS});
        this.XOP_lessthanequal.asdf_date_disp = "on or before";
        this.XOP_greaterthan= this.minBootNeuron({nid:"7775682071577396678", text:">", ntype:NTYPE_CLASS});
        this.XOP_greaterthan.asdf_date_disp = "after";
        this.XOP_greaterthanequal = this.minBootNeuron({nid:"7775684994215435705", text:">=", ntype:NTYPE_CLASS});
        this.XOP_greaterthanequal.asdf_date_disp = "on or after";
        
        this.TERMINATING_TYPE = this.minBootNeuron({nid:"7777852852171534168", text:"TERMINATING TYPE", ntype:NTYPE_CLASS});

        this.prompt = this.minBootNeuron({nid:"7777440647376760473", text:"prompt", ntype:NTYPE_CLASS});
        this.conversion_unit = this.minBootNeuron({nid:"7770720868717377200", text:"conversion unit", ntype:NTYPE_CLASS});
        this.conversion = this.minBootNeuron({nid:"7771003415909957137", text:"conversion", ntype:NTYPE_CLASS});
        //this.symbol = this.minBootNeuron({nid:"7771062370252084332", text:"symbol", ntype:NTYPE_CLASS});
        this.unit_symbol = this.minBootNeuron({nid:"7772145224575039225", text:"symbol", ntype:NTYPE_CLASS});
        this.html_symbol = this.minBootNeuron({nid:"7775192422854000995", text:"html symbol", ntype:NTYPE_CLASS});
        
        //used to tune children of a class (if class permissable, if inst permissable).
        //also used in pattern NOBJ->classesonly/instsonly->TO_TYPE_TXID->TXID OF PARAM
        //to inform ANNE gui that the param intends to connect only to insts or classes..
        this.classes_only = this.minBootNeuron({nid:"7775050740207039709", text:"classes only", ntype:NTYPE_CLASS});
        this.insts_only = this.minBootNeuron({nid:"7778060997442299009", text:"insts only", ntype:NTYPE_CLASS});
        this.no_new_children = this.minBootNeuron({nid:"7771418654491356905", text:"no children", ntype:NTYPE_CLASS});
        this.uo = this.minBootNeuron({nid:"7770928393157957794", text:"unique on", ntype:NTYPE_CLASS});

        this.anndle = this.minBootNeuron({nid:"7772143308845525283", text:"anndle", ntype:NTYPE_CLASS});
        this.CMD_MAKE_SUBANNDLE = this.minBootNeuron({nid:"7777257285707664978", text:"MAKE subanndle", ntype:NTYPE_CLASS});
        this.anndle_lvl = this.minBootNeuron({nid:"7775100368007980447", text:"anndle-lvl", ntype:NTYPE_CLASS});
        this.subanndle = this.minBootNeuron({nid:"7775386871234172677", text:"subanndle", ntype:NTYPE_CLASS});
        //pts up to parent anndle/subanndle (lvl-1)
        this.subanndle_of = this.minBootNeuron({nid:"7775674902351449345", text:"subanndle of", ntype:NTYPE_CLASS});
        
        
        //anndle resolution now is equal to the sponsor of the ANNDLE.
        //anndles CANNOT be annexed.
        this.anndle_sellback = this.minBootNeuron({nid:"77793877955602199", text:"sellback anndle", ntype:NTYPE_CLASS});
        this.anndle_renew = this.minBootNeuron({nid:"7777763623704051918", text:"renew anndle", ntype:NTYPE_CLASS});

        this.anndle_offer = this.minBootNeuron({nid:"7775310066328986130", text:"anndle offer", ntype:NTYPE_CLASS});
        this.anndle_offer_refund = this.minBootNeuron({nid:"7777288463269446287", text:"anndle offer", ntype:NTYPE_CLASS});
        this.anndle_accept_offer = this.minBootNeuron({nid:"7771162452567479785", text:"accept anndle offer", ntype:NTYPE_CLASS});
        this.anndle_reject_offer = this.minBootNeuron({nid:"7778362583058494579", text:"reject anndle offer", ntype:NTYPE_CLASS});
        this.anndle_revoke_offer = this.minBootNeuron({nid:"7777984968890672286", text:"revoke anndle offer", ntype:NTYPE_CLASS});
        this.anndle_offer_nullified = this.minBootNeuron({nid:"7775865098209225416", text:"anndle offer nullified", ntype:NTYPE_CLASS});
        
        this.resolves_to_anneid= this.minBootNeuron({nid:"7770557671924077616", text:"resolves to anneid", ntype:NTYPE_CLASS});
        this.resovles_to_uri = this.minBootNeuron({nid:"7778625875487717827", text:"resolves to uri", ntype:NTYPE_CLASS});
        this.resolves_to_ipv4 = this.minBootNeuron({nid:"7770701407567396183", text:"resolves to ipv4", ntype:NTYPE_CLASS});
        this.resolves_to_ipv6 = this.minBootNeuron({nid:"7772083318109703659", text:"resolves to ipv6", ntype:NTYPE_CLASS});

        this.tense = this.minBootNeuron({nid:"7774888866193797726", text:"tense", ntype:NTYPE_CLASS});
        this.tense_present = this.minBootNeuron({nid:"7775362585331918618", text:"present tense", ntype:NTYPE_CLASS});
        this.tense_past = this.minBootNeuron({nid:"7772326017534981884", text:"past tense", ntype:NTYPE_CLASS});
        this.tense_future = this.minBootNeuron({nid:"7770373507125829431", text:"future tense", ntype:NTYPE_CLASS});
        
        //(action verbs)
        this.action_give = this.minBootNeuron({nid:"77705947298070034", text:"give", ntype:NTYPE_CLASS});
        this.action_receive = this.minBootNeuron({nid:"7773006403613657951", text:"receive", ntype:NTYPE_CLASS});
        this.action_delegate = this.minBootNeuron({nid:"7773093116715894615", text:"delegate", ntype:NTYPE_CLASS});
        
        //now is a base identity type..
        this.anorg = this.minBootNeuron({nid:"7777778897441556324", text:"Anorg", ntype:NTYPE_CLASS});
        
        this.description = this.minBootNeuron({nid:"7779025607680638924", text:"description", ntype:NTYPE_CLASS});
        this.class_desc = this.minBootNeuron({nid:"7777732071930766985", text:"class description", ntype:NTYPE_CLASS});
        
        this.anorg_type = this.minBootNeuron({nid:"777725395794844067", text:"anorg type", ntype:NTYPE_CLASS});
        
        //for payouts of keyless neuFrons.
        this.governance = this.minBootNeuron({nid:"7771671873508876461", text:"governance", ntype:NTYPE_CLASS});
        this.governance.ai_ico_class="ai-anjucator";

        this.gov_dictatorship = this.minBootNeuron({nid:"7779412447342618259", text:"dictatorship", ntype:NTYPE_CLASS});
        this.gov_anj_majority = this.minBootNeuron({nid:"777068234629750280", text:"anjucator majority", ntype:NTYPE_CLASS});
        this.gov_anj_super_majority = this.minBootNeuron({nid:"777771743001523175", text:"anjucator super majority", ntype:NTYPE_CLASS});
        this.gov_anj_jury = this.minBootNeuron({nid:"777132582625210184", text:"anjucator jury", ntype:NTYPE_CLASS});
          
        
        this.DIG = this.minBootNeuron({nid:"7778765030709506308", text:"DIG", ntype:NTYPE_CLASS});
        this.annies_per_aon = this.minBootNeuron({nid:"7774922870627231232", text:"annies per aon", ntype:NTYPE_CLASS});

        this.MAKE_CUSTODIAL = this.minBootNeuron({nid:"7778593465311902303", text:"make custodial", ntype:NTYPE_CLASS});
        
        /***** ASDF ************************************************/
        //ASDF is like bounty neurons.  They will be defined, built in the user's worldview and then FROZEN
        //Only a FROZEN ASDF can accept cmds.
        //Funds will be locked in the ASDF neuron.
        this.ASDF = this.minBootNeuron({nid:"7774123821045389572", text:"ASDF", ntype:NTYPE_CLASS});
        this.ASDF.wvBIPASS=true; //we accept other worldviews as default view for ASDFs (ANNE doesn't adopt them)
        this.ASDF_ORACLE = this.minBootNeuron({nid:"7772363726012774091", text:"ASDF-ORACLE", ntype:NTYPE_CLASS});                       //RELN specifying the oracle of an ASDF neuron
        //RELN SIGNAL tht the future event is underway and futures cannot be accepted.
        this.ASDF_ORACLE_NO_MORE_TAKERS = this.minBootNeuron({nid:"777332540881126681", text:"ASDF-NO-MORE-TAKERS", ntype:NTYPE_CLASS});           
        
        this.ASDF_ORACLE_RELN = this.minBootNeuron({nid:"7775619843467308405", text:"ASDF-RELN", ntype:NTYPE_CLASS});                    //the reln of the oracle that will trigger the AUTO SETTLEMENT
        this.ASDF_FUTURE_TO = this.minBootNeuron({nid:"7775674213293630742", text:"ASDF-FUTURE-TO", ntype:NTYPE_CLASS});
        this.ASDF_REWARD = this.minBootNeuron({nid:"7772537879312541472", text:"ASDF-REWARD", ntype:NTYPE_CLASS});           //always relative to 1.
        this.ASDF_STATUS =  this.minBootNeuron({nid:"7776761752802692203", text:"ASDF-STATUS", ntype:NTYPE_CLASS});
        this.ASDF_EXPIRE_AT_BLOCK =  this.minBootNeuron({nid:"7774654388305965960", text:"ASDF-EXPIRE-AT-BLOCK", ntype:NTYPE_CLASS});
        this.ASDF_EXPIRE_AT_BLOCK.ai_ico_class="ai-clock";

        this.ASDF_RISK_FUNDED =  this.minBootNeuron({nid:"7772927452444197703", text:"ASDF-RISK-FUNDED", ntype:NTYPE_CLASS});
        this.ASDF_RISK_REFUNDED =  this.minBootNeuron({nid:"7779055060461771493", text:"ASDF-RISK-REFUNDED", ntype:NTYPE_CLASS});

        this.ASDF_STATUS_EXPIRED =  this.minBootNeuron({nid:"7775414101854878186", text:"ASDF status expired", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_SETTLED =  this.minBootNeuron({nid:"7777642930344078665", text:"ASDF status settled", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_OPEN =  this.minBootNeuron({nid:"7771432588868502834", text:"ASDF status open", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_FULL =  this.minBootNeuron({nid:"777404000421391248", text:"ASDF status full", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_SETTLEMENT_ERR =  this.minBootNeuron({nid:"7773494473412466357", text:"ASDF settlement err", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_NOMORE_RISK = this.minBootNeuron({nid:"7777584087450006238", text:"ASDF status owner cutoff", ntype:NTYPE_CLASS});           
       
        //set to max available risk reamining w/ respect to reward ratio
        //FUTURE TO is the condition which pays out the event reward.
        //If FUTURE TO is not observed on the ORACLE RELN (and something else is), 
        //the funds are distributed to the ASDF owner.
        this.ASDF_RISK_REMAINING = this.minBootNeuron({nid:"7777229642227033086", text:"ASDF-RISK-REMAINING", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_CDX = this.minBootNeuron({nid:"7778014966634263522", text:"ASDF-FUTURE-TO-CDX", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_CDX_NONANNE_VAL = this.minBootNeuron({nid:"777791626878838486", text:"ASDF-FUTURE-TO-NONANNE-CDX-VAL", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_ANNE_CDX_ANNE_DATE = this.minBootNeuron({nid:"7775993685838455770", text:"ASDF-FUTURE-TO-ANNE-CDX-ANNE-DATE", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_NID_ATTR = this.minBootNeuron({nid:"7773689919601064934", text:"ASDF-FUTURE-TO-NID-ATTR", ntype:NTYPE_CLASS});  

        //user taking on the risk specified in the ASDF
        this.ASDF_ACCEPT_RISK = this.minBootNeuron({nid:"7777809231670492713", text:"ASDF-ACCEPT-RISK", ntype:NTYPE_CLASS});  
        this.ASDF_ADD_RISK = this.minBootNeuron({nid:"777390341076576182", text:"ASDF-ADD-RISK", ntype:NTYPE_CLASS});  

        this.ASDF_RESULT = this.minBootNeuron({nid:"7772822327738373678", text:"ASDF-RESULT", ntype:NTYPE_CLASS});  
        this.ASDF_RESULT_FUTURE_OBSERVED = this.minBootNeuron({nid:"7770617469176432515", text:"ASDF-RESULT-FUTURE-OBSERVED", ntype:NTYPE_CLASS});  
        this.ASDF_RESULT_FUTURE_NOT_OBSERVED = this.minBootNeuron({nid:"7770390220377781103", text:"ASDF-RESULT-FUTURE-NOT-OBSERVED", ntype:NTYPE_CLASS});  

        //annode managed neuron to hold block triggers.
        this.ANNE_BLOCK_TRIGGER = this.minBootNeuron({nid:"7770514447679553139", text:"Block Trigger", ntype:NTYPE_CLASS});  

        //these relons are only put in local RAM, never stored in the IDB tbls --- as they are temp (but on the ANNODE)
        //and reflect an intra-block state change pending.  We don't want them stored as they will muck about when they
        //linger and are no longer accurate statuses.  The aplex obj will pull a .has_pending boolean for quicker access.
        this.INTRA_BLOCK_PENDING = this.minBootNeuron({nid:"7772178508525445062", text:"INTRA BLOCK PENDING", ntype:NTYPE_CLASS});  

        this.REDLIST = this.minBootNeuron({nid:"7773260417288127971", text:"REDLIST", ntype:NTYPE_CLASS});  
        
        /**************************************************/

        this.BLOCK_WINNER  = this.minBootNeuron({nid:"7776164025119639648", text:"Block Winner", ntype:NTYPE_CLASS});  
        this.BLOCK_SHARE_WINNER  = this.minBootNeuron({nid:"777072988166742136", text:"Block Share Winner", ntype:NTYPE_CLASS});  
        
        this.achat = this.minBootNeuron({nid:"7772867043812936263", text:"A𝄐chat", ntype:NTYPE_CLASS});
        
        this.bounty = this.minBootNeuron({nid:"7772768880266899814", text:"bounty", ntype:NTYPE_CLASS});
        this.bounty.wvBIPASS=true;
        this.bounty_deadline = this.minBootNeuron({nid:"777443587645819431", text:"bounty deadline", ntype:NTYPE_CLASS});
        
        //bounty types:
        //anyone chase it, complete it, claim it
        //this.bounty_public_race = this.minBootNeuron({nid:"777283655934389925", text:"claim when complete", ntype:NTYPE_CLASS});
        
        //potential bounty shows interest
        //contact offline (later on p2p chat)
        //vendor selected (too big of a project to tackle w/o transparency of selected vendor)
        //this.bounty_vendor_select = this.minBootNeuron({nid:"7779289286905306093", text:"vendor select", ntype:NTYPE_CLASS});

        //the ANNEID chosen to take on the bounty.
        //they must complete the bounty to get the payout.
        //this.bounty_vendor = this.minBootNeuron({nid:"7770289084266469525", text:"vendor", ntype:NTYPE_CLASS});
        
        //for both types, this is the official payment.
        this.bounty_payout = this.minBootNeuron({nid:"7779986301683332147", text:"bounty payout", ntype:NTYPE_CLASS});

        //make a public declaration you have completed a bounty.
        //if the bounty has specified a chatnode, the request will auto switch to the 
        //chat annode, handshake, and create a chat session that can be accessed by the anjucators.
        //there things can be discussed p2p.. (user->annode annode->user)
        //The shared keys are used (not signing/agreement) as to not leak data.
        //A new PGP key could be created for this purpose but as of now is not.
        //The chat is NOT put on-chain.  It is local ANNODE encrypted chat so it is wise
        //for the bounty creator to select an ANNODE they trust or they own.
        this.bounty_claim = this.minBootNeuron({nid:"7777462743773526396", text:"bounty claim", ntype:NTYPE_CLASS});

        //a clump class which already has 2 roles defined..
        //instantiate this and provide the missing 3 roles (of the params) to create a 
        //delegation request.
        this.delegateDistVoteClump =this.minBootNeuron({nid:"7776157359286235395", text:"delegate distribution voting rights", ntype:NTYPE_CLASS});
        
        this.anjucate_vote = this.minBootNeuron({nid:"7771703615799316076", text:"anjucate vote", ntype:NTYPE_CLASS});
        this.anjucate_vote.ai_ico_class="ai-anjucator";
        this.anj_vote_yay = this.minBootNeuron({nid:"7775592309515630471", text:"vote - yay", ntype:NTYPE_CLASS});
        this.anj_vote_nay = this.minBootNeuron({nid:"7770525053411833701", text:"vote - nay", ntype:NTYPE_CLASS});
        this.anj_vote_abstain = this.minBootNeuron({nid:"777916843002801549", text:"vote - abstain", ntype:NTYPE_CLASS});
        this.anj_vote_hasnot = this.minBootNeuron({nid:"7778382797157395681", text:"vote - has not voted", ntype:NTYPE_CLASS});
        
        this.anj_invitation = this.minBootNeuron({nid:"7774437122736557228", text:"anjucator invitation", ntype:NTYPE_CLASS});
        
        //delegate your vote
        this.anj_delegate = this.minBootNeuron({nid:"777888550313830205", text:"anjucate - delegate vote", ntype:NTYPE_CLASS});
        
        this.vote_relns = [this.anj_vote_yay.temp_nid, this.anj_vote_nay.temp_nid, 
                        this.anj_vote_abstain.temp_nid, this.anj_vote_hasnot.temp_nid];
        this.valid_vote = [this.anj_vote_yay.temp_nid, this.anj_vote_nay.temp_nid, 
                            this.anj_vote_abstain.temp_nid];
    
        //helpful calc stored on the vote neuron.
        this.anj_votes_to_pass = this.minBootNeuron({nid:"7775094174782013630", text:"anj votes to pass", ntype:NTYPE_CLASS});
        this.vote_result = this.minBootNeuron({nid:"7770885647556226532", text:"vote result", ntype:NTYPE_CLASS});
        this.vote_passed = this.minBootNeuron({nid:"7771391628585996940", text:"vote passed", ntype:NTYPE_CLASS});
        this.vote_didnt_pass = this.minBootNeuron({nid:"7773739150549833304", text:"vote did not pass", ntype:NTYPE_CLASS});
        
        //distro requests  (call for a vote)
        this.distro_req = this.minBootNeuron({nid:"7778840860841443809", text:"distro request", ntype:NTYPE_CLASS});
        //this.per_d_vote_by_hodler = this.minBootNeuron({nid:"7773920230765932714", text:"% distro vote by hodler", ntype:NTYPE_CLASS});
        //this.per_d_vote_by_list = this.minBootNeuron({nid:"7776254538552572051", text:"% distro vote by list", ntype:NTYPE_CLASS});
        //this.specific_d_vote_by_hodler = this.minBootNeuron({nid:"7772460455121008551", text:"specific distro vote by hodler", ntype:NTYPE_CLASS});
        this.specific_d_vote_by_list = this.minBootNeuron({nid:"7770256274602944478", text:"specific distro vote by list", ntype:NTYPE_CLASS});
        
        this.mr_scatman = this.minBootNeuron({nid:"777776301542347271", text:"Mr.Scatman", ntype:NTYPE_CLASS});
        this.johnny5 = this.minBootNeuron({nid:"7777798716640079675", text:"johnny#5", ntype:NTYPE_CLASS});
        this.nikt = this.minBootNeuron({nid:"7777111323427337713", text:"nikt", ntype:NTYPE_CLASS});
        
        this.distro_src = this.minBootNeuron({nid:"7776409001442599365", text:"distro source", ntype:NTYPE_CLASS});
        this.distro_to_whom = this.minBootNeuron({nid:"7776415552030872607", text:"benefactor", ntype:NTYPE_CLASS});
        this.distro_quantity = this.minBootNeuron({nid:"7773018074680059484", text:"distro quantity", ntype:NTYPE_CLASS});
        this.distro_deadline = this.minBootNeuron({nid:"7774366570419816048", text:"distro deadline", ntype:NTYPE_CLASS});
        this.distro_block_deadline = this.minBootNeuron({nid:"7773505632843152624", text:"distro block deadline", ntype:NTYPE_CLASS});
        this.distro_quantity_type = this.minBootNeuron({nid:"77719030987172936", text:"distro quantity type", ntype:NTYPE_CLASS});
        this.distro_qt_percent = this.minBootNeuron({nid:"7774613566577121605", text:"distro qt percent", ntype:NTYPE_CLASS});
        this.distro_qt_specific = this.minBootNeuron({nid:"7772653274078556397", text:"distro qt specific", ntype:NTYPE_CLASS});
        this.distro_to_type = this.minBootNeuron({nid:"7770704584830358901", text:"distro to type", ntype:NTYPE_CLASS});
        this.distro_tt_hodler = this.minBootNeuron({nid:"7779256405810042395", text:"distro to type hodler", ntype:NTYPE_CLASS});
        this.distro_tt_list = this.minBootNeuron({nid:"7776184053255403454", text:"distro to type list", ntype:NTYPE_CLASS});

        //used to call a vote from a self-neuron, with reln ADV, pointing to a valid distro req nid.
        this.distro_vote = this.minBootNeuron({nid:"7772171285820605524", text:"Annorg distribution vote", ntype:NTYPE_CLASS});
        this.anjucator = this.minBootNeuron({nid:"7775009149036337280", text:"anjucator", ntype:NTYPE_CLASS});
        this.anjucator.ai_ico_class="ai-anjucator";

        this.AOE = this.minBootNeuron({nid:"7774534916375278299", text:"asset of exchange", ntype:NTYPE_CLASS});
        
        //no collab on these when collab is available.
        this.admin_tnids = [];
        this.admin_tnids.push(this.garden_type.temp_nid);
        this.admin_tnids.push(this.inst_garden_type.temp_nid);
        this.admin_tnids.push(this.let_wv.temp_nid);
        this.admin_tnids.push(this.deny_wv.temp_nid);
        this.admin_tnids.push(this.viplet.temp_nid);
        this.admin_tnids.push(this.dr_andy_defn.temp_nid);
        this.admin_tnids.push(this.dr_andy_hanai_in.temp_nid);
        this.admin_tnids.push(this.dr_andy_super_class.temp_nid);
        this.admin_tnids.push(this.dr_andy_super_insts.temp_nid);
        this.admin_tnids.push(this.dr_andy_super_all.temp_nid);
        this.admin_tnids.push(this.dr_permitted_reln.temp_nid);
        this.admin_tnids.push(this.gt_permit_reln.temp_nid);
        this.admin_tnids.push(this.garden_rule.temp_nid);
        
        this.no_hanais = [
            this.anndle.temp_nid,
            this.distro_vote.temp_nid,
            this.streampay.temp_nid,
            this.bounty.temp_nid,
            this.ASDF.temp_nid,
            this.subanndle.temp_nid,
            this.PDN_PUBLIC_ACCESS.temp_nid,
            this.PDN_VIP.temp_nid,
            this.PDN_BEARER_TOKEN.temp_nid
        ];
        
        /******************************************/
        this.HARDCODED_BOOT_SIZE = this._aNeurons.length;
        /******************************************/

        this.USER = ANNE.USER;
        
        this.annovice_tnids = [];

        //make poa (request to annode)
        //->poa-> (burned by ANNODE only) 
        this.nodecmds_tnids = [];

        this.pdnactions_tnids = [
            this.PDN_OP_REVEAL_ALL.temp_nid,
            this.XOP_equal.temp_nid,
            this.XOP_notequal.temp_nid,
            this.XOP_lessthan.temp_nid,
            this.XOP_lessthanequal.temp_nid,
            this.XOP_greaterthan.temp_nid,
            this.XOP_greaterthanequal.temp_nid
        ];
        this.pdnOpTnids =[this.XOP_equal.temp_nid, 
            this.XOP_notequal.temp_nid,
            this.XOP_lessthan.temp_nid,
            this.XOP_lessthanequal.temp_nid,
            this.XOP_greaterthan.temp_nid,
            this.XOP_greaterthanequal.temp_nid];


        //Some ez references so I don't have to type things out so long
        //wen referencing them.
        if (typeof MEMP != "undefined") {
            this.ANNE.MEMP = MEMP;
            this.MEMP = MEMP
            this.MEMP.init(MEM);
            this.ANNE.MEMP = MEMP;
        }
        if (typeof MEMG != "undefined") {
            this.MEMG = MEMG;
            this.MEMG = MEMG;
            this.MEMG.init(MEM);
            this.ANNE.MEMG = MEMG;
        }
        if (typeof MEMZ != "undefined") {
            this.ANNE.MEMZ = MEMZ;
            this.MEMZ = MEMZ;
            this.MEMZ.init(MEM);
            this.ANNE.MEMZ = MEMZ;
        }

        //set alt ptrs in memp/memg/memz
        if (this.MEMP) {
            if (this.MEMG) this.MEMP.MEMG=this.MEMG;
            if (this.MEMZ) this.MEMP.MEMZ=this.MEMZ;
        }
        if (this.MEMG) {
            if (this.MEMP) this.MEMG.MEMP=this.MEMP;
            if (this.MEMZ) this.MEMG.MEMZ=this.MEMZ;
        }
        if (this.MEMZ) {
            if (this.MEMG) this.MEMZ.MEMG=this.MEMG;
            if (this.MEMP) this.MEMZ.MEMP=this.MEMP;
        }

        if (this.ANNE.GUI_STUFF!=null) {
            this.ANNE.GUI_STUFF.init(this.ANNE, true);
        }

        await this._initDBs(); 
        return await this.continueBootUp();
    }


    getNeuronByNid = function($nid) {
        if (ANNE.MEMG!=null) {
            console.warn("user MEMG, not MEM for getNeuronByNid");
            return ANNE.MEMG.getNeuronByNid($nid);
        }
    }

    printPreloadBuilder = function() {

    }


    /**
     * The 3 top admins in ANNE which assign admin rights to the users 
     * so that they can manage ANNE admin duties in various areas of the system.
     * @param {*} $nid 
     * @returns 
     */
    isANNEADMIN=function($nid) {
        if (this.ANNE_SUPERS.indexOf($nid)!=-1) return true;
        return false;
    }

    
    /**
     * Init local DB
     * This will create the indexDB for persistent local memory storage.  
     * If it already exists it will load it in to local RAM. 
     * 
     * @param {function} - callback when done
     * @author {Mr.Scatman}     
     *            
     */              
     _initDBs = async() => {

        //if not INDEXDB available, just call the doneCallback straightaway
        //nothing should be awaiting this function since it uses the callbacks.
        if (!('indexedDB' in window)) {
            console.log('This browser doesn\'t support IndexedDB');
            ANNE.DB.bSUPPORTED=false;
            return false;
        }

        //if we do have INDEXDB supported, we will call the _finishPreInit function
        //and pass throught he donecallback there.
        else {	
            ANNE.DB.bSUPPORTED=true;
            if (this.ANNE.MEM_UDF.bDOINGPRELOAD==false) {
                this.ANNE.DB.editName(this.ANNE.MEM_UDF.appName);
                var result = await ANNE.DB.___init_lite();
                return result;
            }
            else {
                this.ANNE.DB.editName(this.ANNE.MEM_UDF.appName);
                var result = await ANNE.DB.___init();
                if (result==null) return;
                if (result.status!=undefined && result.status=="no-node") {}
                var rez = result.results;
                return await this._finishPreInit(rez.relons, rez.history, rez.settings); 
            }
            
        }
    }

    
    /**
     * aSettings are saved settings local that can be pre-loaded back into
     * the app. you likely have a way to handle this alreayd in the 
     * phone/backoffice.     
     * 
     * @param {array} - an array of relons extracted from indexeddb memory.
     * @param {array} - an array of annode api use history (for optimizations).
     * @param {array} - an array of local (non-saved-onchain OR saved on-chain) settings.
     * @author {Mr.Scatman}      
     *                           
     */
     _finishPreInit=async($aRelons, $aFullPullHistory, $aSettings)=> {

        //pre-load our query history (for optimizing future queries)
        if ($aFullPullHistory!==null) {
            if (this.ANNE.HISTORY!=null) this.ANNE.HISTORY.processPullHistory($aFullPullHistory);
        }

        //pre-load all our graph edges (these are the ANNE TXs we broadcast which made it to the ANNE middle memory layer in Mongo)
        if ($aRelons!=null && $aRelons.length) {
            await MEMP.santizeSavedRelonsForReprocessing($aRelons);
            var rRETURNALL=false;
            var bDOLOCALSAVE=false;

            //do not try to store these in local memory, as we just pulled them from where we store them!
            if ($aRelons.length) {

                //silly test -->
                await MEMP.processNeuronsFromRelons($aRelons, rRETURNALL, bDOLOCALSAVE, "relons", "_finishPreInit");
            }
        }

        //process the ANNE default settings here.
        if ($aSettings.length) {
            if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.aSettingsToProcess = $aSettings;
        }

        return;
    }


    /** 
     * Load up required min memory from the node if we haven't yet.
     *  
     * @author {Mr.Scatman} 
     */
    continueBootUp=async()=> {
        this.ANNE.bBOOTUP_BASE_DONE=true;

        //check memory length.  If none, do the preload.
        //we will have MORE neurons than JS preload if we've pulled preload from ANNODE.
        if (this._aNeurons.length<=this.HARDCODED_BOOT_SIZE) {
            DB.DO_THE_ONETIME_PRELOAD=true;
            if (ANNE.MEM_UDF.bDOINGPRELOAD==true) {
                await ANNE.API.getAllPreloads();
            }
            ANNE.A.setHTML("loader_msg", "hypergraph acquired");
            DB.DO_THE_ONETIME_PRELOAD=false;
        }

        else {
            if (ANNE.AWARE) await ANNE.AWARE.init(); //this will require some nuanced handling.
        }
        await this.specialsBuild();

        //we continued bootup, we're good.
        ANNE.bNEED_TO_CONTINUE_BOOT_UP=false;

        //we are done with bootup.
        //we can now safely check localStorage for login info.
        return ANNE.anne_bootup_complete();

    }


    /** 
     * Wire up important functions and relons.
     * 
     * In past version of the memory manager we preloaded hundreds and even
     * thousands of neurons and relons in files.  This was helpful for testing.
     * In those files we were able to wire up things to setup some functionality
     * of the memory handler because we had direct access to variables to do so.
     * 
     * What we have not done in the preloaded relons we need to do here now..
     * at bootup.
     * 
     * Note: we can still load supporting .js files that have extended helper
     * functions will can be used in conjunction with these dynamic module mem loads.
     * This will add to the module if it is found in this.
     * 
     * NOTE: THIS WILL ONLY WORK IF ALL THESE NEURONS ARE PRESENT AT LOADUP!!!!
     * 
     * @author {Mr.Scatman} 
     */
    specialsBuild = async()=> {

        //ASCII
        //ascii_code_num
        //ascii_eval
        var _n0 = MEMG.getNeuronByNid("7772759491040052489");
        var _n1 = MEMG.getNeuronByNid("7779936209861264085");
        var _n2 = MEMG.getNeuronByNid("7777466399639974815");
        var _n3 = MEMG.getNeuronByNid("7770513775335783280");
        var _n4 = MEMG.getNeuronByNid("7779613874588391142");
        var _n5 = MEMG.getNeuronByNid("7773838259673220295");
        var _n6 = MEMG.getNeuronByNid("7772956192305001444");
        var _n7 = MEMG.getNeuronByNid("7779032678869944184");
        var _n8 = MEMG.getNeuronByNid("7777087244990101066");
        var _n9 = MEMG.getNeuronByNid("7778263312619963662");


        var numHack = [_n0, _n1, _n2, _n3, _n4, _n5, _n6, _n7, _n8, _n9];

        await this.buildAndInjectMod(this.ASCII, "ASCII");

        //nums are be classed as integers and i didn't do a 
        //BE PARENT / BE ASCII for this heirarchy.
        for (var nhack=0; nhack < numHack.length; nhack++) {
            var numNOBJ = numHack[nhack];
            if (numNOBJ==null) continue;
            this.ASCII["p_"+(nhack+48)] = numNOBJ;
            numNOBJ.charCode=(nhack+48)
            numNOBJ.char=numNOBJ.text;
            numNOBJ.pk2_name="char";
            numNOBJ.pk2_val=numNOBJ.text;
            numNOBJ.pk_name="charCode";
            numNOBJ.pk_val=(nhack+48);
        }

        await this.doLangStuff();

        //let's do the creative countries..
        await this.buildAndInjectMod(this.country, "COUNTRY");

        //await this.buildAndInjectMod(this.state, "STATE");
        this.STATE = {"status":"pre-init", "name":"STATE"};
        this.STATE._aChildren=this.getChildrenDynamic(this.state);
        
        for (var a=0; a < this.STATE._aChildren.length;a++) {
            var stateNOBJ = this.STATE._aChildren[a];
            if (stateNOBJ.aka!=null && stateNOBJ.aka.length>0) {
                for (var ak=0; ak < stateNOBJ.aka.length; ak++) {
                    if (stateNOBJ.aka[ak].length==2) this.STATE[ stateNOBJ.aka[ak] ] = stateNOBJ;
                }
            }
        }

        await this.buildAndInjectMod(this.month, "MONTH");

        var months = [this.month_jan, this.month_feb, this.month_mar, this.month_apr, this.month_may, this.month_jun,
                    this.month_jul, this.month_aug, this.month_sep, this.month_oct, this.month_nov, this.month_dec];
        for (var m=0; m < months.length; m++) {
            var monthNOBJ = months[m];
            var str = monthNOBJ.text.toLowerCase();
            this._monthMap[ str ] = monthNOBJ;
            this._monthLookup.push(str);
            var str = str.substring(0,3);
            this._monthMap[ str ] = monthNOBJ;
            this._monthLookup.push(str);
            this._monthMap[ (m+1) ] = monthNOBJ;
        }
        return;
    }

    doLangStuff = async()=>{
        //let's do a language module.
        await this.buildAndInjectMod(this.language, "LANG");
        MEM.LANG.urange = [];

        //set the urange based on known unicode start/ends
        for (var c=0; c<MEM.LANG._aChildren.length; c++) {
            var childNOBJ= MEMG.getNeuronByTempNid(MEM.LANG._aChildren[c]);
            var rel_u_start = MEMG.getRelonByTempR(childNOBJ.outs, MEM.unicode_start.temp_nid);
            if (rel_u_start!=null) {
                var rel_u_end = MEMG.getRelonByTempR(childNOBJ.outs, MEM.unicode_end.temp_nid);
                if (rel_u_end!=null) {
                    MEM.LANG.urange.push({s:rel_u_start.to_s, e:rel_u_end.to_s, lang:childNOBJ.temp_nid, lang_debug:childNOBJ.text});
                }
            }
        }

        //sort the urange by start val
        MEM.LANG.urange.sort( function($a, $b) {
            if ($a.s > $b.s) return 1;
            else if ($a.s < $b.s) return -1;
            return 0;
        });

        //ascertain lang script by ucode
        MEM.LANG.ascertain_lang=function($str, $maxbreak=5){
            var tally = [];
            var max = 0;
            var winning = null;
            for (var s=0; s < $str.length; s++) {
                var test = $str[s].charCodeAt(0);
                for (var u=0; u < ANNE.MEM.LANG.urange.length; u++) {
                    var ur = MEM.LANG.urange[u];
                    if (test >= ur.s && test <= ur.e) {
                        if (tally[ur.lang]==undefined) tally[ur.lang]=1;
                        else tally[ur.lang]+=1;
                        if (tally[ur.lang]>max) {
                            max = tally[ur.lang];
                            winning = MEMG.getNeuronByTempNid(ur.lang);
                        }
                        if (max>=$maxbreak) return winning;
                        break;
                    }
                }
            }
            return winning;
        }
    }


    /**
    * USED FOR MAKING NEW CLASSES AND NEURONS ----> 
    * THIS IS FOR KEY'D NEURONS (OR GHOSTS), NOT KEYLESS.
    * WE MAKE THEM REAL HERE, WITH A SEED, 
    *
    * DO NOT CREATE TEMP NEURONS YOU DO NOT INTEND TO KEEP AROUND.  
    * IF U STORE RELONS ON-CHAIN WITH THESE TEMP NEURONS THEY GO AWAY 
    * AND THEN YOU HAVE ORPHANED RELONS. 
    * @param {string}
    * @param {int}
    * @param {string}
    * @param {anne_neuron}
    * @param {boolean}
    * @param {boolean}
    */
    makeTempNObjOnly = function ($owner, $type, $newText, $parentObj=null, $bPUSH=true){
        if (ANNE.USER==null) {
            console.error("must have a USER object");
            return null;
        }

        var temp_nid = UTILS.randTXID(this._emulatedTXIDLength-2);
        var temp_nObj = new anne_neuron($newText, $type, temp_nid);
        temp_nObj.seed = "";
        temp_nObj.pubKey = "";  
        temp_nObj.ins = [];
        temp_nObj.outs = [];	
        temp_nObj.sponsored_by = "";

        //hash the neuron.
        //mrscatman 04/24/2022 -- monitor this.. could lead to nasty mistakes.
        MEMP.__hashNeuron(temp_nObj);

        if ($owner) temp_nObj.setOwner($owner);

        temp_nObj.temp=true;
        temp_nObj.mine = true;

        if (ANNE.USER!=null && ANNE.USER._userNObj.nid) {
            if(ANNE.USER._userNObj.owns==undefined) ANNE.USER._userNObj.owns=[];
            if (ANNE.USER._userNObj.owns.indexOf(temp_nObj.temp_nid)==-1) ANNE.USER._userNObj.owns.push(temp_nObj.temp_nid);
        }

        if ($type==NTYPE_INST) temp_nObj.bTERMINATING_TYPE=true;
        else if ($parentObj!=null && $parentObj!="") temp_nObj.bTERMINATING_TYPE=$parentObj.bTERMINATING_TYPE;

        if ($parentObj!=null && $parentObj!="" && ($parentObj instanceof anne_neuron) == true) {
            temp_nObj.addUniqueBe_fastIndex($parentObj);
            if ($parentObj.children.indexOf(temp_nObj.temp_nid)==-1) {
                $parentObj.children.push(temp_nObj.temp_nid);
            }

            temp_nObj.__functInherit($parentObj, "mem.makeTempNObjOnly");

            //emulate the BE relon...
            var reln_nid = "";
            if ($type==NTYPE_INST) reln_nid = this.inst.nid;
            else reln_nid = this.class.nid;

            //bPUSH indicates pushing this into the master arrays
            //makeTempRelon=function($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH=true, $bUNIQUE=true) 
                
            var parentRelon = this.makeTempRelon(temp_nObj, TYPE_BE, reln_nid, TO_TYPE_ANNE, $parentObj.nid, $bPUSH);
            parentRelon.parent = $parentObj.nid;
            parentRelon.owned_by = temp_nObj.owned_by;

            //make a temp be text relon for name!
            if ($newText!="") {
                //console.log("Make a new name relon "+$newText);
                var beTextRelon = this.makeTempRelon(temp_nObj, TYPE_BE, this.name.nid, TO_TYPE_STRING, $newText, $bPUSH);
                if ($parentObj==null || $parentObj.nid==undefined) {
                    console.log("WHAAAAAAAAAAAT?");
                }
                beTextRelon.parent = $parentObj.nid;
                beTextRelon.owned_by = temp_nObj.owned_by;    
                //if (this._EMULATE_BE_TEXT_NEURON==true) {
                    //this.makeAndPushIfUniqueEMULATED($newText, TO_TYPE_STRING, beTextRelon);
                //}
            }
        }

        temp_nObj.ghost=false;

        return temp_nObj;
    }


    /**
     * 
     * @param {*} $owner 
     * @param {*} $type 
     * @param {*} $newText 
     * @param {*} $parentObj 
     * @param {*} $bPUSH 
     * @returns 
     */
    makeGhostNObj = async ($owner, $type, $newText, $parentObj=null, $bPUSH=true)=> {
        if (ANNE.USER==null) {
            console.error("must have a USER object");
            return null;
        }

        var temp_nid = UTILS.randTXID(this._emulatedTXIDLength-2);
        var temp_nObj = new anne_neuron($newText, $type, temp_nid);
        temp_nObj.seed = "";
        temp_nObj.pubKey = "";  
        temp_nObj.ins = [];
        temp_nObj.outs = [];	
        temp_nObj.sponsored_by = "";

        //hash the neuron.
        //mrscatman 04/24/2022 -- monitor this.. could lead to nasty mistakes.
        MEMP.__hashNeuron(temp_nObj);

        if ($owner) temp_nObj.setOwner($owner);

        temp_nObj.temp=true;
        temp_nObj.mine = true;

        if (ANNE.USER!=null && ANNE.USER._userNObj.nid) {
            if(ANNE.USER._userNObj.owns==undefined) ANNE.USER._userNObj.owns=[];
            if (ANNE.USER._userNObj.owns.indexOf(temp_nObj.temp_nid)==-1) ANNE.USER._userNObj.owns.push(temp_nObj.temp_nid);
        }

        if ($type==NTYPE_INST) temp_nObj.bTERMINATING_TYPE=true;
        else if ($parentObj!=null && $parentObj!="") temp_nObj.bTERMINATING_TYPE=$parentObj.bTERMINATING_TYPE;

        if ($parentObj!=null && $parentObj!="" && ($parentObj instanceof anne_neuron) == true) {
            temp_nObj.addUniqueBe_fastIndex($parentObj);
            if ($parentObj.children.indexOf(temp_nObj.temp_nid)==-1) {
                $parentObj.children.push(temp_nObj.temp_nid);
            }

            temp_nObj.__functInherit($parentObj, "mem.makeGhostNObj");

            //emulate the BE relon...
            var reln_nid = "";
            if ($type==NTYPE_INST) reln_nid = this.inst.nid;
            else reln_nid = this.class.nid; 
                
            var parentRelon = this.makeTempRelon(temp_nObj, TYPE_BE, reln_nid, TO_TYPE_ANNE, $parentObj.nid, $bPUSH);
            parentRelon.parent = $parentObj.nid;
            parentRelon.owned_by = temp_nObj.owned_by;

            //make a temp be text relon for name!
            if ($newText!="") {
                //console.log("Make a new name relon "+$newText);
                var beTextRelon = this.makeTempRelon(temp_nObj, TYPE_BE, this.name.nid, TO_TYPE_STRING, $newText, $bPUSH);
                if ($parentObj==null || $parentObj.nid==undefined) {
                    console.log("WHAAAAAAAAAAAT?");
                }
                beTextRelon.parent = $parentObj.nid;
                beTextRelon.owned_by = temp_nObj.owned_by;    
            }
        }

        for (var g = 0; g < temp_nObj.outs.length; g++) {
            temp_nObj.outs[g].from_text = $newText;
        }
        await GHOSTDATA.localGhostTheRelons(temp_nObj.outs);
        temp_nObj.ghost = true;
        temp_nObj.ghost_owner = USER._userNObj.nid;

        return temp_nObj;
    }


    /** 
     * Create a new mod in MEM, or snag an existing one.
     * 
     * Existing ones could be loaded from supporting files to have
     * supporting code that does complex things.
     * 
     * @param {anne_neuron} - The classNeuron
     * @param {string} - desired prop name in the MEM space for the MOD.  (ie, this.modeName)
     * @author {Mr.Scatman} 
     * 
     */
    buildAndInjectMod = async($classNObj, $modName="")=> {
        if (($classNObj instanceof anne_neuron) == false) return;
        var MOD = null;
        if ($modName=="") {
            MOD={"status":"pre-init"};
            var modName = $classNObj.text.trim().toUpperCase();
            modName.replace(/[^a-z]/gi, '');
            MOD.name = modName;
            MEM[modName]=MOD;
            MOD._aChildren=[];
        }
        else {
            MOD = {"status":"pre-init", "name":$modName};
            MEM[$modName]=MOD;
            MOD._aChildren=[];
        }
        
        //this feature is ONLY intended for classes that have specified
        //a priary key.  We will use this primary key as the var name
        //in the mod.

        //we should expect to see "iso_639_1" for the language class.
        var pk = $classNObj.pk;
        if (pk==undefined) return;
        var pk_param_name = "";
        var pk2_param_name = "";
        var pkNObj = MEMG.getNeuronByTempNid(pk);
        if (pkNObj!=null) {
            var param_name_relon = MEMG.getParamNameRelon(pkNObj.outs);
            if (param_name_relon!=null) {
                pk_param_name = param_name_relon.to_s;
                //pk_be_lookup = "be_"+pkNObj.temp_nid;
            }
            else return;
        }
        else return;

        var pk2 = $classNObj.pk2;
        if (pk2!=undefined) {
            if(Array.isArray(pk2)==true) {
                for (var a=0; a < pk2.length; a++) {
                    var pk2NObj = MEMG.getNeuronByTempNid(pk2[a]);
                    if (pk2NObj!=null) {
                        var param_name2_relon = MEMG.getParamNameRelon(pk2NObj.outs);
                        if (param_name2_relon!=null) {
                            pk2_param_name = param_name2_relon.to_s;
                            break;
                        }
                    }   
                }
            }
                
            else {
                var pk2NObj = MEMG.getNeuronByTempNid(pk2);
                if (pk2NObj!=null) {
                    var param_name2_relon = MEMG.getParamNameRelon(pk2NObj.outs);
                    if (param_name2_relon!=null) {
                        pk2_param_name = param_name2_relon.to_s;
                    }
                }   
            }
        }
 
        //alt-be quick props..
        var childrenNOBJS = this.getChildrenDynamic($classNObj);
       
        for (var c=0; c < childrenNOBJS.length; c++) {
            var childNOBJ = childrenNOBJS[c];
            MOD._aChildren.push(childNOBJ.temp_nid);
            var pk_quick= childNOBJ[pk_param_name];
            if (pk_quick==undefined) continue;
            MOD['p_'+pk_quick] = childNOBJ;
            childNOBJ["pk_val"] = pk_quick;
            childNOBJ["pk_name"] = pk_param_name;
            if (pk2_param_name!="") {
                var pk2_quick= childNOBJ[pk2_param_name];
                childNOBJ["pk2_val"] = pk2_quick;
                childNOBJ["pk2_name"] = pk2_param_name;
                MOD['p_'+pk2_quick] = childNOBJ;
                var val = parseInt(childNOBJ.value);
            }
        }

        MOD.status="init";
        //console.log("** LOADED dynamic MOD **"+$modName, MOD);
    }


    /** 
     * @author {Mr.Scatman} 
     */
     init=async()=> {
    }


    /** 
     * Add in the user neuron into mem.
     * 
     * @param {string} - The nid longs string of the neuron
     * @param {boolean} - flag indicating this is the logged-in user's neuron. (they own it)
     * @author {Mr.Scatman} 
     * 
     */
     addInUserNeuron=function($nid, $mine=false) {
        var nObj = MEMG.getNeuronByNid($nid);
        if (nObj!=null) {
            nObj.mine=$mine;
            nObj.ntype = NTYPE_INST;

            //sneaky temp add -- audit what this does when we have an
            //actual REAL bes
            //we know this to be true, but don't know which subclass yet.
            //it is important to denote that this IS an idenity tho. 
            if (nObj.bes==undefined) nObj.bes=[];
            //if (nObj.bes.length==0) nObj.bes.push(this.identity.temp_nid);
            //nObj.be = "identity"; 
        
            if (nObj.adopted_with_trust==undefined) nObj.adopted_with_trust = [];
            return nObj;
        }

        nObj = new anne_neuron("", NTYPE_INST, $nid);
        nObj.adopted_with_trust = [];      //a list of adopted worldviews.
        nObj.mine = $mine;

        //default for any user of ANNE if not more specified
        //nObj.setBeTempNid(this.identity.temp_nid);   
        //nObj.be = "identity"; 

        return nObj;
    }


    /** 
     * Add to active worldview.
     * 
     * @param {string} - the nid  long string of the worldview we are adding.
     * @author {Mr.Scatman} 
     */
    addToAppWideActiveWorldViews = function ($worldviewTNid) {
        var wv_tnid = parseInt($worldviewTNid);
        if (this._activeWorldViewTempNids.indexOf(wv_tnid)==-1) this._activeWorldViewTempNids.push(wv_tnid);
    }


    /** 
     * Get children depending on how the class was defined.
     * 
     * Some classes do not have hierarchical children.  Other systems
     * incorrectly pigeon whole their data into a locked-in view.
     * ANNE does not.  Something like a country is not actually a child
     * of Country.  It is a location and a Country is a polity.  The child
     * of location points with a reln of Polity to the neuron Country using
     * a TYPE_BE relon.  This is the practical children of country.
     * 
     * Get the right children and return an array.
     * 
     * @param {anne_neuron} $classNObj - The classNeuron
     * @param {Boolean} $bDOTNIDS - return tnids not nObjs
     * @param {Array} $tnids_existing - an array of tnids we already have as children
     * 
     * @author {Mr.Scatman} 
     */
    getChildrenDynamic = function($classNObj, $bDOTNIDS=false, $tnids_existing=[]) {
        var children = [];
        if ($classNObj['be_'+this.BE_CHILDREN.temp_nid]!=undefined) {
            var be_reln = $classNObj['be_'+this.BE_CHILDREN.temp_nid];
            if ($classNObj['be_'+this.BE_PARENT.temp_nid]!=undefined) {
                var parent_restrict = $classNObj['be_'+this.BE_PARENT.temp_nid];
                var relnNOBJ = ANNE.MEMG.getNeuronByTempNid(be_reln);
                //console.warn("CHILDREN ARE (FROM CHILD) TYPE_BE->reln of '"+relnNOBJ.text+"'");
                if (parent_restrict) {
                    var parentNOBJ = ANNE.MEMG.getNeuronByTempNid(parent_restrict);
                    //console.warn("CHILDREN MUST HAVE REAL PARENT OF '"+parentNOBJ.text+"'");
                    if (parentNOBJ.children==null) return children;
                    for (var c=0; c < parentNOBJ.children.length; c++) {
                        var maybeChildNOBJ = MEMG.getNeuronByTempNid(parentNOBJ.children[c]);
                        if (maybeChildNOBJ==undefined || maybeChildNOBJ==null) continue;
                        if (maybeChildNOBJ.enforceGardenUp()==false) continue;
                        var beAltRelon = ANNE.MEMG.getRelonByTempR(maybeChildNOBJ.outs, relnNOBJ.temp_nid);
                        if (beAltRelon!=null && beAltRelon.to_type==TO_TYPE_ANNE && beAltRelon.t==$classNObj.temp_nid) {
                            if ($tnids_existing.length && $tnids_existing.indexOf(maybeChildNOBJ.temp_nid)!=-1) continue;
                            if ($bDOTNIDS==false) {
                                children.push(maybeChildNOBJ);
                            }
                            else {
                                children.push(maybeChildNOBJ.temp_nid);
                            }
                        }
                    }
                }
            }
            else {
                for (var r=0; r < this._aRelons.length; r++) {
                    var rObj = this._aRelons[r];
                    if (rObj.type!=TYPE_BE) continue;
                    if (rObj.r!=this.polity.temp_nid) continue;
                    if (rObj.t!=$classNObj.temp_nid) continue;
                    if ($tnids_existing.length && $tnids_existing.indexOf(rObj.f)!=-1) continue;                    
                    var nObj2 = ANNE.MEMG.getNeuronByTempNid(rObj.f);
                    if (nObj2.enforceGardenUp()==false) continue;
                    if ($bDOTNIDS==false) children.push(nObj2);
                    else children.push(rObj.f);
                }
            }
        }
        
        //standard hierarchical children.
        else {
            if ($classNObj.children==null) return children;
            for (var c=0; c < $classNObj.children.length; c++) {
                if ($tnids_existing.length && $tnids_existing.indexOf($classNObj.children[c])!=-1) continue;
                var childNOBJ = ANNE.MEMG.getNeuronByTempNid($classNObj.children[c]);
                if (childNOBJ.enforceGardenUp()==false) continue;
                if ($bDOTNIDS==false) children.push(childNOBJ);
                else children.push(rObj.f);
            }
        }
        return children;
    }



    /**
     * 
     * MAY BE DEPRECATED 
     * 
     * 
     * Build a dropdown of valid children.
     * 
     * This needs to build a list of instances of the reln, subclasses, or  *
     * any instances of subclasses.  Basically EVERYTHING below this in the  *
     * BE TREE.   The from is not relevant unless restricting to existing    *
     * params on the class definition.                                       *
     * 
     * @param {anne_neuron} - The from neuron
     * @param {string} - The nid long string of the reln.
     * @param {boolean} - Flag indicating downline instances will also be retrieved.
     * @author {Mr.Scatman}  
     */
    buildValidHasChildrenTempNids=function($fNObj, $reln_nid, $bINSTS=false) {
        var aMatchingParams = MEMG.getParentLineParams($fNObj, $reln_nid);
        var aList = [];
        if (aMatchingParams.length) aList = this.extractTempNidsFromRelonsTo(aMatchingParams);
        var children_of_parent_params = [];
        if (aList.length) children_of_parent_params = this.extractAllChildrenOfTempNidList(aList);
        if (children_of_parent_params.length) aList = aList.concat(children_of_parent_params);
        var relnNObj = MEMG.getNeuronByNid($reln_nid);
        var childrenDownline = [];
        relnNObj.getAllDownlineChildren(childrenDownline, 0, $bINSTS); //this should get all subs, all children of subs.. (all insts of)
        var validClassTempNids = aList.concat(childrenDownline);
        validClassTempNids = validClassTempNids.filter(leg => leg !=$fNObj.temp_nid);
        return validClassTempNids;
    }


    /**
    * Extract all children nids.
    * 
    * This will honor creative hierarchies that use BE_PARENT and BE_CHILDREN
    * to define their primary hierarchy as something non-parent/child.
    *  
    * @param {anne_neuron} - The parent neuron
    * @author {Mr.Scatman} 
    */
    extracAllChildrenNidsDynamic = function($nObj) {
        var childrenNOBJS = this.getChildrenDynamic($nObj);
        var children_nids = [];
        for (var c=0; c < childrenNOBJS.length; c++) {
            children_nids.push(childrenNOBJS[c].nid);
        }
        return children_nids;
    }


    /**
     * 
     * @param {array} - An array of relon objs.
     * @author {Mr.Scatman} 
     */
    extractTempNidsFromRelonsFrom=function($aRelons) {
        if ($aRelons==undefined || $aRelons==null || Array.isArray($aRelons)==false) return [];
        var aTempNids = [];
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.f==undefined || rObj.f==null || rObj.f==-1) {
                var fromNOBJ = MEMG.getNeuronByNid(rObj.from_nid);
                if (fromNOBJ!=null) {
                    if (aTempNids.indexOf(fromNOBJ.temp_nid)===-1) {
                        console.log("extracting "+fromNOBJ.text);
                        aTempNids.push(fromNOBJ.temp_nid);
                    }
                }
                continue;
            }
            if (aTempNids.indexOf(rObj.f)===-1) {
                var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                aTempNids.push(rObj.f);
                console.log("extracting "+fromNOBJ.text);
            }
        }
        return aTempNids;
    }
    extractTempNidsFromRelonsTo=function($aRelons) {
        var aTempNids = [];
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.t==undefined || rObj.t==null || rObj.t==-1) continue;
            if (aTempNids.indexOf(rObj.t)===-1) aTempNids.push(rObj.t);
        }
        return aTempNids;
    }


    /**
     * Get all the children neurons of a list of temp nids.
     * 
     * @param {array} - An array of temp nid ints
     * @author {Mr.Scatman} 
     */
    extractAllChildrenOfTempNidList=function($arr) {
        var aChildren = [];
        for (var i=0; i < $arr.length; i++) {
            var nObj = MEMG.getNeuronByTempNid($arr[i]);
            if (nObj==null) continue;
            var children = nObj.children;
            if (children.length==0) continue;
            for (var c=0; c < children.length; c++) {
                var child = children[c];
                if (aChildren.indexOf(child)!==-1) continue;
                aChildren.push(child);  //unique child for our summary list.
            }
        }
        return aChildren;
    }



                            
    /**
     * This establishes a base URL for a class.
     *
     * @param {anne_neuron} - The class neuron
     * @param {string} - The base url to use as "part" of the url.
     * @author {Mr.Scatman} - 
     */
    __setClassBaseURL=function($classNOBJ, $base_url_string) {
        return this.makeTempRelon($classNOBJ, TYPE_HAS, this.base_url.nid, TO_TYPE_STRING, $base_url_string, true);
    }


    /**
     * 
     * Registers a img for a class to expect to be used as the reln for a given 
     * class.. If a base_url is supplied it will also be created as a relon. 
     * see moviesDB_api::                                                        
     * this.__setImgThumbRelonForClass(this.media_credit, this.profile_path_with_base, "https://image.tmdb.org/t/p/w185");
     *
     * @param {anne_neuron} - The class neuron
     * @param {anne_neuron} - The img thumb reln's full neuron.
     * @param {string} - An optional base url to use as "part" of the url.
     * @author {Mr.Scatman} 
     * 
     */
    __setImgThumbRelonForClass=function($classNOBJ, $imgThumbRelnNOBJ, $base_url="") {
        this.makeTempRelon($classNOBJ, TYPE_HAS, this.imgThumbDefaultClass.nid, TO_TYPE_ANNE, $imgThumbRelnNOBJ.nid, true);
        if ($base_url!="") {
            this.makeTempRelon($classNOBJ, TYPE_HAS, this.base_url.nid, TO_TYPE_STRING, $base_url, true);
        }
    }


    /**
     * local desperado img url for classes and instances that we have no img of.
     * 
     * @param {anne_neuron} - The classneuron neuron.
     * @param {string} - The url.
     * @author {Mr.Scatman} 
     * 
     */
    __setImgThumbDefaultClassImg = function($classNOBJ, $url) {
        $classNOBJ.desperado_img_url=$url;
    }


    /**
     * Set a primary key for a class.
     * 
     * These are properties used by the class in additional to the nid.
     * They are external in nature but are primary properties and can
     * be used to resolve and identify the neuron.
     * 
     * ALL ANNE NEURONS have a true primary key of their .nid.  That is 
     * a given.  This is in addition to that.
     * 
     * @param {anne_neuron} - The target neuron.
     * @param {anne_neuron} - The primary key neuron.
     * @author {Mr.Scatman} 
     * 
     */
    __setPrimaryKey=function($fromNOBJ, $toNOBJ) {
        this.makeTempRelon($fromNOBJ, TYPE_HAS, this.primary_key.nid, TO_TYPE_ANNE, $toNOBJ.nid, true);
    }


    /**
     * Set a secondary key for a class.
     * 
     * This is not a foreign key, it is an alt primary key.
     * Classes such as countries will supply unique ISO codes of 2 digits
     * as well as 3 digits.  Secondary key is used in these such cases.
     * 
     * Note: nid is the ACTUAL ANNE PRIMARY KEY for everything.  Primary and secondary
     * dennote outside ANNE data source data keys.
     * 
     * @param {anne_neuron} - The target neuron.
     * @param {anne_neuron} - The secondary key neuron.
     * @author {Mr.Scatman} 
     * 
     */
    __setSecondaryKey2=function($fromNOBJ, $toNOBJ) {
        this.makeTempRelon($fromNOBJ, TYPE_HAS, this.primary_key2.nid, TO_TYPE_ANNE, $toNOBJ.nid, true);
    }


    /**
     * Make a new neuron and push it, if unique.
     * 
     * @param {Object} - The relon obj.
     * @param {string} - A string of the prop name to extract the nid from.
     * @param {string} - A string name of the prop to create to map to the temp_nid of the new neuron.
     * @param {string} - We will delete this after we use it to make chrome debugging more visually easy.
     * @author {Mr.Scatman} 
     * 
     */
    makeAndPushIfUniqueNid = function ($rObj, $prop='from_nid', $setprop='f', $temptext="") {
        var lookingForNID = $rObj[$prop];
        if ($rObj[$prop+"TNOBJ"]!=undefined) $rObj[$prop+"TNOBJ"] = $rObj[$prop+"TNOBJ"];
        else $rObj[$prop+"TNOBJ"] = MEMG.getNeuronByNid(lookingForNID);

        //(rObj, 'from_nid', 'f', 'from_txt_temp');    

        if ($rObj[$prop+"TNOBJ"]!=null) {
            $rObj[$prop+"TNOBJ"].temp=false;
            if ($rObj[$prop+"TNOBJ"].nid[0]=="z") $rObj[$prop+"TNOBJ"].ghost=true;
            if ($setprop) $rObj[$setprop]=$rObj[$prop+"TNOBJ"].temp_nid;
            if ($temptext!="") $rObj[$temptext]=$rObj[$prop+"TNOBJ"].text;
            return $rObj[$prop+"TNOBJ"];
        }

        //we don't have it -- make it.
        $rObj[$prop+"TNOBJ"] = this.makeNeuronFromANYRelon($rObj, $prop);
        $rObj[$prop+"TNOBJ"].temp=false;
        if ($setprop) $rObj[$setprop]=$rObj[$prop+"TNOBJ"].temp_nid;
        if ($temptext!="") $rObj[$temptext]=$rObj[$prop+"TNOBJ"].text;
        return $rObj[$prop+"TNOBJ"];
    }


    /**
     * Make and push an emulated neuron from a non ANNE ID to.                 
     * 
     * final rObj is for context when the TO_TYPE is TO_TYPE_ENCRYPTED.        
     * we need to look at the RELN to contextualize what type of encrypted   
     * value it is.    
     * TO DO --> use rObj to contextuals the encrypt 
     * (reln will tell us what it is)  
     *
     * @author {Mr.Scatman} 
     */
    makeAndPushIfUniqueEMULATED=function($text, $to_type, $rObj=null, $bFREEPASS=false) {
        if (this._EMULATE_TO_TYPES == false && $bFREEPASS==false) return null;
        if (ANNE.USER==null) return null;

        //not an emulatable to type.
        if ($bFREEPASS==false && (this._emulatables.length==0 || this._emulatables.indexOf($to_type)==-1)) return null;

        //if we are emulating text neurons... only do so in non sys relons//$rObj.type==TYPE_BE || 
        if ($bFREEPASS==false && $to_type==TO_TYPE_STRING && $rObj!=null && ($rObj.type==TYPE_SYS)) return null;


        //we do not emulate neurons that are bigger than 64 chars in length
        //this allows for us to max out with emulation of a TXID as its own neuron. (when it is the to)
        if ($to_type==TO_TYPE_STRING && $text.len > 64) return null;

        var text = ($text+"").trim();
        var type = NTYPE_INST;
        var TTSYSOBJ = MEMG.getSysToTypeObjByTypeAndBe($to_type, this.DATA_TYPE);
        var bes=[];
        var be="";
        var rNObj = null;
        var rNObj_TTOBJ =null;

        if ($rObj!=null) {
            rNObj = MEMG.getNeuronByTempNid($rObj.r);
            if (rNObj==undefined || rNObj==null) {
                console.log($rObj);
                console.log("debug this!!!!");
                rNObj = MEMG.getNeuronByNid($rObj.reln);
                if (rNObj!=null) $rObj.r = rNObj.temp_nid;
            }
            if (rNObj!=null) rNObj_TTOBJ = MEMG.getTerminatingType(rNObj);
        }

        var beNOBJ = null;

        //use the reln if we have it, else use the to_type default.
        if (rNObj_TTOBJ==null) rNObj_TTOBJ = TTSYSOBJ;

        //dates are dates.. that's the end of it.
        if (rNObj_TTOBJ.temp_nid==this.date.temp_nid) {
            bes=[rNObj_TTOBJ.temp_nid];
            beNOBJ = rNObj_TTOBJ;
            be=rNObj_TTOBJ.text;
            var nObj = MEMG.getEmulatedNeuronByClassAndValue(beNOBJ, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        else if (rNObj_TTOBJ!=null && TTSYSOBJ!=null && rNObj_TTOBJ.SYS_ENUM_VAL==TTSYSOBJ.SYS_ENUM_VAL && rNObj.nid!=this.name.nid) {
            bes=[rNObj.temp_nid];
            beNOBJ = rNObj;
            be=rNObj.text;
            var nObj = MEMG.getEmulatedNeuronByClassAndValue(beNOBJ, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }
        
        //super critical here to classify the text neuron as a class of its relon.
        //we accept any existing downliners of the same.
        else if (rNObj_TTOBJ==null && TTSYSOBJ.SYS_ENUM_VAL==TO_TYPE_STRING) {
            bes=[rNObj.temp_nid];
            beNOBJ = rNObj;
            be=rNObj.text;
            nObj = MEMG.getEmulatedNeuronByClassAndValue(rNObj, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        //a number value
        else if (TTSYSOBJ.SYS_ENUM_VAL==TO_TYPE_NUMBER && MEM_NUMBER!=undefined) {
            bes=[this.number.temp_nid];
            beNOBJ = this.number;
            be=this.number.text;
            nObj = MEM_NUMBER.getNumberNeuron($text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        else {
            bes=[TTSYSOBJ.temp_nid];  //bes twe can do.. this to_type was used on a rogue reln which was not defined as sub of a to type
            beNOBJ = TTSYSOBJ;
            be=TTSYSOBJ.text;
            var nObj = this.getEmulatedNeuronByTypeAndText($to_type, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        //we do not make emulated neurons for numbers unless
        //they are special numbers.
        //special numbers are in the downline of number_notation or the conversion class.
        //WE MAY NEED TO RETHINK THIS.... for the LIVE-FIRING-NN version.
        if ($to_type==TO_TYPE_NUMBER) {
            var downline_nids = [this.number_notation.temp_nid];
            this.number_notation.getAllDownlineChildren(downline_nids, 0, false);
            if (downline_nids.indexOf(rNObj.temp_nid)==-1) {
                downline_nids = [this.conversion.temp_nid];
                this.conversion.getAllDownlineChildren(downline_nids, 0, false);
                if (downline_nids.indexOf(rNObj.temp_nid)==-1) return null;
            }
        }
            
        var nid = ANNE.USER.getRandomEmulatedANNEAccountId();

        var emulatedNOBJ = new anne_neuron(text, type, nid);
        emulatedNOBJ.be = be;
        emulatedNOBJ.setBeTempNid(bes[0]);
        emulatedNOBJ.emulated = true;
        emulatedNOBJ.bTERMINATING_TYPE=true;

        //we need to make a new emulated relon pointing to this new emulated neuron
        //we can use this more properly in the all the neuron/relon functions.
        this.emulateEdgeToEmulated($rObj, emulatedNOBJ);

        //emulate the be relon of the emulated to,
        //it points to what we chose as the beNOBJ.
        var emulateBeROBj = this.makeEmulatedRelon(emulatedNOBJ, TYPE_BE, this.class.nid, TO_TYPE_ANNE, beNOBJ.nid, true, false);
        
        emulatedNOBJ.__setValueOfNeuron(beNOBJ, TTSYSOBJ);

        return emulatedNOBJ;
    }



    /**
     * Set the text name with the first valid relon.
     * 
     * This is done when we arbor or inhibit a name relon.  The name is lost.
     * We then try to set the name (rollback) to something else we have in memory.
     * 
     * @param {anne_neuron} - the Neuron object
     * @author {Mr.Scatman} 
     * 
     */
    setTextWithFirstValidAssigner=function($nObj) {
        if ($nObj==undefined || $nObj==null || $nObj.outs.length==0) return "";
        var len = 0;
        var str = "";

        //try for any other BE NAME relons.
        for (var r=0; r < $nObj.outs.length; r++) {
            var rObj = $nObj.outs[r];
            if (rObj.active==false) continue;   //why is this still even here?
            if (rObj.type==TYPE_BE && rObj.reln==this.name.nid && rObj.to_type==TO_TYPE_STRING) {
                if (rObj.to_s.length > len) {
                    len = rObj.to_s.length;
                    str = rObj.to_s;
                }
            }
        }

        if ($nObj.short!=undefined && $nObj.short!="") return $nObj.short;
        if ($nObj.aka!=undefined && $nObj.aka!="") return $nObj.aka;

        //take any be.
        if (str=="") {
            for (var r=0; r < $outs.length; r++) {
                var rObj = $outs[r];
                if (rObj.active==false) continue;   //why is this still even here?
                if (rObj.type==TYPE_BE && rObj.to_type==TO_TYPE_STRING) {
                    if (rObj.to_s.length > len) {
                        len = rObj.to_s.length;
                        str = rObj.to_s;
                    }
                }
            }
        }
        return str;
    }

    /** 
     * sort tnids 
     */
    sortTnidsAlpha=function($tnids) {
        var nObjs = MEMG.getNObjsFromTempNidList($tnids);
        nObjs.sort(ANNE.UTILS.sortByTextNoCase);
        return MEMG.getTnidsFromNObjs(nObjs);
    }


    /** 
     * Create an emulated edge.
     * 
     * Sometimes edges point to non-encoded TERMINATING TYPES and we want to emulate
     * a neuron for that type.  In doing so we also need to then emulate a relon that
     * points to that emulated neuron since the original relon will be in the wrong 
     * TO_TYPE space.
     * 
     * @param {Object} - the original relon we want to create an emulated relon for
     * @param {anne_neuron} - the emulated Neuron we will point to with our emulated relon. 
     * @author {Mr.Scatman} 
     */
    emulateEdgeToEmulated = function ($rObj, $emulatedNOBJ) {
        var fNOBJ = MEMG.getNeuronByTempNid($rObj.f);
        var emulated_relon = this.makeEmulatedRelon(fNOBJ, $rObj.type, $rObj.reln, TO_TYPE_ANNE, $emulatedNOBJ.nid, true, false);
        emulated_relon.emulated_from_txid = $rObj.txid;
        $rObj.active=false;
    }


    /**
     * Make an emulated relon
     * 
     * Emulated relons are NOT ghosts.  They are made up based upon 
     * converting non-neuron values to neurons -- and the emulated relons 
     * support them as-if-they-were-real. 
     * example: turn a date string into a date neuron -- that doesn't 
     * actually exist on-chain, but now exists in mem, as an emulated date 
     * neuron. 
     * 
     * Temp relons are NOT ghosts. but they can become ghosts if they are 
     * sent to the DB.
     * 
     * @author {Mr.Scatman} 
     */
    makeEmulatedRelon=function($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH=true, $bTEMP=false) {
        var rObj = this.makeTempRelon($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH);
        rObj.emulated=true;
        rObj.temp=$bTEMP;
        rObj.ghost=false;
        if ($reln==this.require.nid && $type==TYPE_PARAM) rObj.required=true;
        return rObj;
    }


    /**
     * @author {Mr.Scatman} 
    */
    makeTempRelon=function($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH=true) {
        if ($fNOBJ==null || $fNOBJ.nid==undefined) return null;
        if (typeof $reln !=='string') {
            console.warn("bad data in make temp relon $reln");
            if (typeof $reln ==="object") {
                $reln = $reln.nid;
                if (this.__isValidNid($reln)==true) {
                    console.warn("we fixed it, but u should pass the .nid, not the object!");
                }
                else return null;
            }
            else return null;
        }
        if (typeof $to_s !=='string' && typeof $to_s !=='number') {
            console.log("bad data in make temp relon $to_s");
            if ($to_type==TO_TYPE_ANNE && typeof $to_s ==="object") {
                $to_s = $to_s.nid;
                if (this.__isValidNid($to_s)==true) {
                    console.warn("we fixed it, but u should pass the .nid, not the object!");
                }
                else return null;
            }
            else return null;
        }
        if (parseInt($type)>255 || parseInt($to_type)>255) {
            console.log("bad data in make temp relon $type/$to_type");
            return null;
        }


        var r = MEMG.getNeuronByNid($reln);
        if (r) r=r.temp_nid;
        else r = -1;
        var t = -1;
        var tObj = null;
        if ($to_type==TO_TYPE_ANNE) {
            tObj = MEMG.getNeuronByNid($to_s);
            if (tObj) {
                t = tObj.temp_nid;
                if ($type==TYPE_BE) $fNOBJ["be_"+r]=t;
            }
        }

        $to_type = parseInt($to_type);
        if ($to_type==TO_TYPE_NUMBER) $to_s = parseFloat($to_s);

        //temp relons are 21 length with a leading z.
        var rObj = {
                    "txid":ANNE.UTILS.randTXID(this._emulatedTXIDLength),
                    "type":parseInt($type),
                    "reln":$reln,
                    "to_type":$to_type,
                    "to_s":$to_s,
                    "from_nid":$fNOBJ.nid,
                    "f":$fNOBJ.temp_nid,
                    "r":r,
                    "t":t,
                    "temp":true,
                    "emulated":false,
                    "preloaded":false,
                    "time":-1,
                    "firstseen":0,
                    "rorigin":RORIGIN_GHOST,
                    "ghost":false
                    }
        if ($bPUSH==false) return rObj;

        //puts relons in a master hash array and a master array with rid.
        MEMP.__hashRelon(rObj);
        $fNOBJ.outs.push(rObj);

        if (tObj!=null) {
            if (tObj.ins) tObj.ins.push(rObj);
        }

        //we need to emulate the to as well, if it is a bTERMINATING_TYPE
        else if (rObj.to_type!=TO_TYPE_ANNE &&  rObj.to_type!=TO_TYPE_SEED) {
            var tObj = this.makeAndPushIfUniqueEMULATED(rObj.to_s, rObj.to_type, rObj);
            if (tObj) {
                tObj.ins.push(rObj);  //track that we have an in to this emulated neuron ?
                //we do not emulate the owns and owned relons for emulated to type neurons.
                //we do, however, set the owned_by of the first occurance we observe of any 
                //emulated tos.  We set it equal to the owner of the from that pointed to it.  (if that exists)
                if (tObj.owned_by=="" && $fNOBJ.owned_by!="") tObj.owned_by = $fNOBJ.owned_by;
                rObj.t = tObj.temp_nid;     
            }
        }
        if (tObj!=null && this.require!=undefined && $reln==this.require.nid && tObj.temp_nid==this.validation.temp_nid) {
            $fNOBJ.requires_validation=true; 
        }
        return rObj;
    }


    /**
     * Assess if a relon has met the class PARAM definition.
     * 
     * returns true if the paramRelon we are reviewing is in the array_has
     * we match the be downline....  (not the upline) 
     * example:                                                              
     * identity class has a param of HOUSE
     * instance of identity "Mr.Scatman" has a HAS of house->"house for ants"
     * That "hosue for ants" is an instance of "tiny house", which is a 
     * sub-class of house.   Be looking at the PARAM reln we find out what 
     * was desired to supply.   We can look at its DOWNLINE to gnenerate
     * other viable classes of information.   Tiny House is a type of house,
     * thus the parameter has been supplied in the MR SCATMAN HAS reln. 
     * 
     * @author {Mr.Scatman} 
     */
    hasMetClassDefn=function($relons, $paramRelon, $paramToNObj=null) {
        var paramRelnNObj = MEMG.getNeuronByTempNid($paramRelon.r);
        if (paramRelnNObj==null) return false; //must have this.
        
        //console.log(paramRelnNObj.text);
        
        var paramToNObj = $paramToNObj;
        if ($paramRelon.t && paramToNObj==null) paramToNObj = MEMG.getNeuronByTempNid($paramRelon.t);
        
        var TTOBJ = MEMG.getTerminatingType(paramRelnNObj);

        var TO_TTOBJ = MEMG.getTerminatingType(paramToNObj);

        //loop what we have to see if we have this param!  :)
        //we also satisfy PARAMs with BEs -- so allow these to pass as well.
        for (var a=0; a <$relons.length; a++) {
            var relon = $relons[a];

            //if the relon is a TYPE_ROLE.. 
            //and the reln is a role... it is good whatever the to is.
            if (relon.type==TYPE_ROLE && relon.r==$paramRelon.r) {
                var relNOBJ = MEMG.getNeuronByTempNid(relon.r);
                if (relNOBJ.bes[0]==this.role.temp_nid) {
                    return true;
                }
            }


            //we may have some relon types in this array we don't want to consider.
            //Only a has/sys/be/role can satisfy the PARAM and we already checked the TYPE_ROLE.
            if (relon.type!=TYPE_HAS && relon.type!=TYPE_BE && relon.type!=TYPE_SYS) continue;    //these are supposed to be has onl.. sometimes we are lazy

        
            if (relon.reln!=paramRelnNObj.nid) {
                if (this.matchesBeDownLine(relon.r, paramRelnNObj)!=true) continue;
            }

            //quick exits
            if (TTOBJ!=null) {
                if (relon.to_type==TTOBJ.SYS_ENUM_VAL) return true;
            }

            if (TO_TTOBJ!=null && paramToNObj!=null) {
                if (relon.to_type==TO_TTOBJ.SYS_ENUM_VAL) return true;
            }
            

            //this paramreln is already in the has (or be) ....
            //see if the to matches the param's to class (or a parent);
            //if (paramRelnNObj.bTERMINATING_TYPE == true) EXPECT an emulated neuron in the has...
            if (relon.t!=undefined && relon.t!=-1) {
                var toNObj = MEMG.getNeuronByTempNid(relon.t); //there will be no hasTONObj (unless emulated)
                if (toNObj==null) return false;

                if (toNObj.bes[0] == $paramToNObj.temp_nid) return true;
                else {
                    if ($paramToNObj.temp_nid==this.class.temp_nid) return true;
                    
                    var BENOBJ = MEMG.getNeuronByTempNid(toNObj.bes[0]);
                    //is the has BE a descendent of the param's to nobj?
                    //this.matchesBeDownLine(BENOBJ.temp_nid, $paramToNObj)
                    //does the bas BE upline run into the $paramToNObj?
                    //alt version (testing)
                    if (BENOBJ && this.matchesBeDownLine(BENOBJ.temp_nid, $paramToNObj)==true) return true;  //check the paramTOObj's upline...
                    else {
                        //console.log("param reln matches, to class doesn't...");
                        if (relon.type==TYPE_BE) return true;  //ok for be, not for param/has
                    }
                }
            }

            //if no to class is known, assume this is ok.
            else return true;

        
        }
        return false;
    }


    /**
     * Count how many times a neuron has been used as the reln in relon memory space.
     * 
     * @params {int} - The temp_nid of the relon's reln we want to look for.
     * @author {Mr.Scatman} 
     * 
     */
    countRelnUsage=function($r) {
        var cnt = 0;
        for (var r; r<this._aRelons.length;r++) {
            if (this._aRelons[r].r=$r) cnt+=1;
        }
        return cnt;
    }

    //UNTESTED
    //misc.child_cdx = [{from_nid:for_nOBJ.nid, reln:this.neuronImg.nid, to_s:null, type:null, to_type:TO_TYPE_ANNE, target_part:"to_s"}];
    enforceCDXS=function($cdxs, $nObj) {
        for (var c=0; c <$cdxx.length; c++) {
            var cdx_rel = $cdxs[c];
            if (cdx_rel.target_part=="to_s") {
                var from_nid = cdx_rel.from_nid;
                var reln = cdx_rel.reln;
                var type = cdx_rel.type;
                var to_type = cdx_rel.to_type;
                var exists = this.enforceCDXIn($nObj, type, from_nid, reln, to_type); 
                if (exists==false) return false;
            }
            else if (cdx_rel.target_part=="from_nid") {
                var reln = cdx_rel.reln;
                var type = cdx_rel.type;
                var to_type = cdx_rel.to_type;
                var to_s = cdx_rel.to_s;
                var exists = this.enforceCDXOut($nObj, type, reln, to_type, to_s); 
                if (exists==false) return false;
            }
            else return false;
        }
        return true;
    }

    //UNTESTED
    enforceCDXIn = function ($nObj, $type, $from_nid, $reln, $to_type) {
        for (var r=0; r < $nObj.ins.length; r++) {
            var rel = $nObj.ins[r];
            if ($from_nid!=null && rel.from_nid!=$from_nid) continue;
            if ($reln!=null && rel.reln!=$reln) continue;
            if ($type!=null && rel.type!=$type) continue;
            if ($to_type!=null && rel.to_type!=$to_type) continue;
            return true;
        }
        return false;
    }

    //UNTESTED
    enforceCDXOut = function ($nObj, $type, $reln, $to_type, $to_s) {
        for (var r=0; r < $nObj.ins.length; r++) {
            var rel = $nObj.ins[r];
            if ($reln!=null && rel.reln!=$reln) continue;
            if ($type!=null && rel.type!=$type) continue;
            if ($to_type!=null && rel.to_type!=$to_type) continue;
            if ($to_s!=null && rel.to_s!=$to_s) continue;
            return true;
        }
        return false;
    }


    /**
     * TODO: AUDIT THIS AGAIN
     * 
     * a bit of overkill, we could just walk the list as we build it and     *
     * look for a match..                                                    *
     * (this will handle multiple be lines like GROUP and FEELZ)             *
     * 
     * @author {Mr.Scatman} 
     */
    matchesBeDownLine=function($target_temp_nid, $nObj) {
        if ($nObj==null) return false;
        if ($nObj.temp_nid==this.class.temp_nid) return true;

        if ($nObj.children && $nObj.children.length) {
            var children = $nObj.children;
            for (var c=0; c < children.length; c++) {
                var child = children[c];
                if (child==$target_temp_nid) return true;
                var childNObj = MEMG.getNeuronByTempNid(child);
                if (childNObj.children!=undefined && childNObj.children.length) {
                    var bFOUND = this.matchesBeDownLine($target_temp_nid, childNObj);
                    if (bFOUND==true) return true;
                } 
            }

        }
        return false;
    }

    
    /**
     * Test for inclusion in lineage.
     * 
     * Is this temp_nid or any parent of this be equal to the target_nid ? 
     * In other words, is the $targetNid in the parental TYPE_BE dag lineage of the $temp_nid? 
     * 
     * @param {int} - Temp nid for the neuron we want to test
     * @param {int} - The temp nid of the ancestor we are looking to see if is in the temp_nid line.
     * @Author {Mr.Scatman}
     * 
     */
    tempBeIsA = function($temp_nid, $target_tnid, $ntype=null) {
        var nObj = null;
        nObj = MEMG.getNeuronByTempNid($temp_nid); 
        if ($ntype==null) {
            if ($temp_nid==$target_tnid) return true;       //quick exit
        }
        else {
            if (nObj==null) return false;
            if ($ntype==NTYPE_INST && nObj.ntype!=NTYPE_INST) return false; 
            else if ($ntype==NTYPE_CLASS && nObj.ntype!=NTYPE_CLASS) return false;
            else if ($temp_nid==$target_tnid) return true;  //quick exit
        }

        if (nObj==null) return false;
        if (nObj.bes[0]==$target_tnid) return true;
        var depth=0;
        var parentNObj = MEMG.getNeuronByTempNid(nObj.bes[0]);
        var loops = 0;
        while (depth<10 && parentNObj!=null) {
            loops+=1;
            if (loops>20) break;//safey
            if (parentNObj.be_lvl==0 || parentNObj.be_lvl==-1) break;
            depth++;
            if (parentNObj.temp_nid==$target_tnid) return true;
            parentNObj = MEMG.getNeuronByTempNid(parentNObj.bes[0]);
            if (parentNObj==null) return false;
        }
        return false;
    }


    /** 
     * @author {Mr.Scatman} 
     */
    makeNeuronFromANYRelon = function ($rObj, $prop='from_nid') {
        if ($rObj[$prop] == undefined) {
            console.warn('NO '+$prop+' prop on the relon!!  Process all relons in (processNeuronsFromRelons)');
        }
        return new anne_neuron("", "", $rObj[$prop]);
    }


    /** 
     * Load up a min neuron for use in the brain chicken and egg loadup challenge.
     * note: This only runs once during the js class initialization phase.
     * 
     * @param {Object} - The quick json obj used to preload a neuron to have as a permanent reference
     * used in the chicken & egg loadup challenge.
     * @author {Mr.Scatman} 
     * 
     */
    minBootNeuron=function($nObj_min, $parentNOBJ=null) {
        var nOBJ = new anne_neuron($nObj_min.text, $nObj_min.ntype, $nObj_min.nid);
        if ($parentNOBJ==null) return nOBJ;
        nOBJ.setBeTempNid($parentNOBJ.temp_nid);
        return nOBJ;
    }


    /** 
     * Standard relon sort by value.
     * 
     * @param {Object} - relon a.
     * @param {Object} - relon b.
     * @author {Mr.Scatman} 
     * 
     */
    __sortSeqRelons = function($a, $b) {
        var a = MEMG.getNeuronByTempNid($a.r);
        var b = MEMG.getNeuronByTempNid($b.r);
        if (a.value==undefined) {
            if (b.value!=undefined) return 1;
            else return 0;
        }
        else if (b.value==undefined) return -1;
        if (a.value > b.value) return 1;
        else if (b.value > a.value) return -1;
        return 0;
    }

    /**
     * THIS WILL BE DEPRECATED.
     * Done by the context pool handling of a phrase.
     */
    addModsToGhostNeuron=function($role, $chaseClassNid) {
        if ($role.pre_mods.length==0) return;
        if ($role.new==false) return;
        var nObj =$role.nObj;
        if (nObj.ghost==false) return;
        
        //constituent relons are key here.
        if ($chaseClassNid==this.location.nid) {
            for (var r=0; r < $role.pre_mods.length; r++) {
                var mod = $role.pre_mods[r];
                var nObjmod = mod.nObj;
                if (nObjmod==null) continue;

                //check for relon
                var rel = MEMG.getRelonByTempR(nObj.outs, this.constituent_of.temp_nid);
                if (rel==null) {
                    GHOSTDATA.new_ghost_be(nObj, this.constituent_of.nid, TO_TYPE_ANNE, nObjmod.nid);
                }
                else {
                    //we have a problem!  This location is a constituent of a different location!
                    if (rel.to_s!=nObj.nid) {
                        alert("trouble!!!!");
                        return;
                    }
                    continue;
                }
                nObj = nObjmod;
            }
        }
    }


    /**
     * A class has specified it is a terminating type.
     * We do this when a class wants to piggy-back terminating type logics
     * but it is classified as its own unique thing in the TYPE_BE tree.
     * 
     * Example:  annie(s).
     * They are a unit of measure (which is a measure) but that is a class,
     * not a terminating type.  The GUI will try to look for children of annies
     * when params are used to point to annies, but what we really want is a
     * numerical entry.  
     * 
     * Any class that uses the TERMINATYPE_TYPE relon can specify its actual
     * terminating type.  We can then map that here to the terminating type object.
     */
     mapSysCode=function($nObj, $to_type_temp_nid) {
        for (var s=0; s < this._SYS_TO_TYPE_CODES.length; s++) {
            var sysNOBJ = this._SYS_TO_TYPE_CODES[s];
            if (sysNOBJ.temp_nid==$to_type_temp_nid) {
                $nObj.SYS_ALIAS= sysNOBJ.SYS_ALIAS;
                $nObj.SYS_ENUM_VAL = sysNOBJ.SYS_ENUM_VAL;
                $nObj.SYS_CODE= sysNOBJ.SYS_CODE;
                //console.warn("successfully mapped new TT ", $nObj);
                return;
            }
        }
        console.warn("unsuccessful mapping of new TT ", $nObj);
    }


    /** 
     * THIS WAS DONE MANY VERSIONS AGO WHEN IT WASN'T PREOPTIMIZING..
     * BUT NOW IT IS A JUMBLEDMESS.  IGNORE IT.
     * @author {Mr.Scatman} 
     */
    markChasedIns = function($nObj, $types) {
        return [];
        if ($nObj==null || $nObj.emulated==true) return [];
        if ($nObj.preloaded==true && MEM_UDF._lookupPreloadedIns==false) return [];
        var post = "ins";
        return this.markChasedArr($nObj, $types, post);
    }


    /** 
     * Optimize ANNODE queries by marking what we lookuped up.
     * 
     * THIS WAS DONE MANY VERSIONS AGO WHEN IT WASN'T PREOPTIMIZING..
     * BUT NOW IT IS A JUMBLEDMESS.  IGNORE IT.
     * 
     * @param {anne_neuron} - The neuron we just queried the ANNODE for.
     * @param {array} - An array of ints for the types we just queried.
     * * @author {Mr.Scatman} 
     */ 
    markChasedOuts = function($nObj, $types) {
        return [];
        if ($nObj==null || $nObj.emulated==true) return [];
        if ($nObj.preloaded==true && MEM_UDF._lookupPreloadedOuts==false) return [];
        var post = "outs";
        return this.markChasedArr($nObj, $types, post);
    }


    /** 
     * Optimize ANNODE queries by marking what we lookuped up.
     * 
     * This marks parts of more complex queries that we sometimes cannot
     * optimize.  The pieces can be optimized and prevent us from doing repeat
     * work on future complicated queries that include these smaller parts.
     * 
     * @param {anne_neuron} - The neuron we just queried the ANNODE for.
     * @param {array} - An array of ints for the types we just queried.
     * @param {string} - Either "outs" or "ins" for indicating which direction we chased.
     * @author {Mr.Scatman} 
     * 
     */
    markChasedArr=function($nObj, $types, $post) {
        var arr = $nObj["chased_"+$post];
        if (arr==undefined) {
            var arr = [];
            $nObj["chased_"+$post]=arr;
        }

        //it is new.. update our arr in mem and return the original types array
        if (arr.length==0) {
            for (var s=0; s < 7; s++) {
                if ($types.indexOf(s)!=-1) arr[s]=1;
                else arr[s]=0;  //make a new slot for this type to track it.
            }
            return $types;
        }   

        //we alreayd have done something.. let's see what.
        //return a new array of the net diff of what we have not done.
        var types = [];
        for (var s=0; s < types.length; s++) {
            var sch = types[s];
            var tracker_slot = arr[sch];
            if (tracker_slot==1) continue;
            types.push(s);
        }
        return types;
    }


    /** 
     * Set a neuron's value by its polarity relon.
     * 
     * TODO: revisit this.
     * 
     * @param {anne_neuron} - the neuron you want to set value for.
     * @author {Mr.Scatman} 
     */
    __setValByBoolPolarity = function ($nObj) {
        if ($nObj == null) return;
        var polarity_relon = MEMG.getRelonFromWithRToType($nObj.outs, TYPE_HAS, this.polarity.nid, TO_TYPE_ANNE);
        if (polarity_relon!=null) {
            if (polarity_relon.to_s==this.positive.nid) $nObj.setValue(true);
            else $nObj.setValue(false);
        }
    }

   /** 
     * Potentially call an external function.
     * Maybe it was loaded.  Maybe not.  If loaded, call it.
     * 
     * ALL OF THESE HAVE NO PARAMS.
     * @param {Object} - The mod (ie, MEM_UDF, NF_UDF, LANGUAGE, COUNTRY) 
     * @param {string} - A string of the function name in the mod.
     * 
     * @author {Mr.Scatman} 
     */
    _MOD_FUNC_CALLER = function($MOD, $FUNCSTR="") {
        if ($MOD!=null && $FUNCSTR!="" && $MOD[$FUNCSTR]!=undefined && typeof MOD[$FUNCSTR]==="function") {
            $MOD[$FUNC]();
        }
    }

    /**
    * Get the relns to chase for a set of nObjs.
    * 
    * The nObjs should be classes here.
    * 
    * @param {array} - an array of anne_neuron class neurons (ntype = NTYPE_CLASS).
    * @author {Mr.Scatman} 
    *
    */
    __GET_THUMB_RELNS_TO_CHASE=function($nObjs) {
        var relns = [];
        var handled_class_nids = [];
        for (var n=0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            if (nObj.ntype!=NTYPE_CLASS) continue;
            if (handled_class_nids.indexOf(nObj.nid)!=-1) continue;
            handled_class_nids.push(nObj.temp_nid);
            var thumb_relon = MEMG.getParamRelon(nObj.outs, this.primaryNeuronImg.temp_nid);
            if (thumb_relon!=null) {
                if (relns.indexOf(thumb_relon.to_s)!=-1) relns.push(thumb_relon.to_s);
            }
        }
        return relns;
    }


    
    /**
     * Get the URL
     * 
     * Intended to be generalized for any to_type TO_TYPE_URL to return the url.
     * Sometimes the URL is a URL part and requires a base.  Othertimes it is ready
     * to go as is.
     * 
     * When we look for a prop, store the prop with a p_TEMPNID in the neuron.
     * Use it next time without looking it up. 
     *    
     * @param {Object} - the relon object.                            
     * @author {Mr.Scatman} 
     * 
     */
    __GET_URL=function($relon) {
        if ($relon.to_type!=TO_TYPE_URL) {
            if (typeof $relon ==='string') return $relon;
            else return "";
        }
        var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
        if (fNOBJ['p_'+$relon.r]!=undefined) return fNOBJ['p_'+$relon.r];
        var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
        var base_url_relon = this.__isBaseUrl(relnNOBJ);
        if (base_url_relon!=null) {
            var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
            fNOBJ['p_'+$relon.r] = base_url_relon.to_s+$relon.to_s;
            return fNOBJ['p_'+$relon.r];
        }
        var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
        fNOBJ['p_'+$relon.r] = $relon.to_s;
        return fNOBJ['p_'+$relon.r];
    }


    /**
     * This is for neuron imgs/vid/audio and files that have urls for their data.
     */
    __GET_URL_FROM_TNID=function($temp_nid){
        var assetNOBJ =  MEMG.getNeuronByTempNid($temp_nid);
        if (assetNOBJ==null) return "";
        var fdu = this.dataUrl.temp_nid;
        for (var r=0; r < assetNOBJ.outs.length; r++) {
            if(assetNOBJ.outs[r].to_type!=TO_TYPE_URL) continue;
            if(assetNOBJ.outs[r].r!=fdu) continue;
            return assetNOBJ.outs[r].to_s;
        }
        return "";
    }


    /**
     * ALL images must be vetted for fingerprint to be displayed in the extension.
     * 
     * @param {anne_neuron} - The parent class neuron.
     * @param {anne_neuron} - The target neuron we are trying to get the thumb URL for.
     * @author {Mr.Scatman} 
     * 
     */
     __GET_DEFAULT_IMG_NEURON=function($classNOBJ, $nObj) {
        
        //get it and get out!
        if ($nObj['p_'+this.primaryNeuronImg.temp_nid]!=undefined) {
            return MEMG.getNeuronByNid($nObj['p_'+this.primaryNeuronImg.temp_nid]);
        }

        //set the quick helper ref for next time, get out!
        var defaultSelfImg = MEMG.getRelonByTempR($nObj.outs, this.primaryNeuronImg.temp_nid, [TYPE_PARAM]);
        if (defaultSelfImg!=null) {
            $nObj['p_'+this.primaryNeuronImg.temp_nid] = defaultSelfImg.to_s;
            return MEMG.getNeuronByNid($nObj['p_'+this.primaryNeuronImg.temp_nid]);
        }

        //if no class or if this is the class of class... get out.
        if ($classNOBJ==undefined || $classNOBJ==null || $classNOBJ.temp_nid==this.class.temp_nid) {
            return null;
        }
        
        //check for class-based default image
        var defaultClassImgForAll = MEMG.getRelonByTempR($classNOBJ.outs, this.primaryNeuronImg.temp_nid, [TYPE_PARAM]);
        if (defaultClassImgForAll!=null) {
            $nObj['p_'+this.primaryNeuronImg.temp_nid] = defaultClassImgForAll.to_s;
            return MEMG.getNeuronByNid($nObj['p_'+this.primaryNeuronImg.temp_nid]);
        }
        return null;
    }


    /**
     * 
     * @param {*} $temp_nid 
     * @returns 
     */
    __isAnyReln = function($temp_nid) {
        for (var r=0; r < this._aRelons.length; r++) {
            if(this._aRelons[r].r==$temp_nid) return true;
        }
        return false;
    }


    /**    
     * 
     * Is this a base url relon?
     * 
     * @param {anne_neuron} - the neuron of a relon's reln.
     * @author {Mr.Scatman} 
     * 
     */
    __isBaseUrl=function($relnNOBJ) {
        var base_url_relon = MEMG.getRelonByTempR($relnNOBJ.outs, this.base_url.temp_nid);
        if (base_url_relon!=null) return base_url_relon;
        return null;
    }


    /**   
     * Is this a valid nid? 
     * 
     * This function is not such a big deal now in ANNE CHAIN.
     * In prior iteratons the ID wasn't a long string and it had more robust 
     * checking for conformity.  With a string for a long we can get shorter values
     * and thus cannot check for string length other than knowing it is 20 or less.
     * 
     * @param {string} - The long nid string.
     * @author {Mr.Scatman} 
     */
    __isValidNid = function($nid) {
        if (($nid+"").trim().length>20) return false;
        var test = parseInt($nid);
        if (test!=NaN && test && Number.isInteger(test)==true) return true;
        return false;
    } 


    /**    
     * Is this txid valid?
     * 
     * Since the txid and nid data types in ANNODE are the same (BIGINT(20))
     * this is the same as isValidNid.
     * 
     * @param {string} - The txid long string of a relon.
     * @author {Mr.Scatman} 
     * 
     */
    __isValidTXID = function($txid) {
        var regx = /^[0-9]+$/;
        if (regx.test($txid)==false) return false;

        var len = ($txid+"").trim().length;
        if (len>20 || len <10) return false;  // < 10 possible but rather rare.
        var test = parseInt($txid);
        if (test!=NaN && test && Number.isInteger(test)==true && (test+"").length()==$txid.length) return true;
        return false;
    }

    
	__isMalformedRelon=function($relon) {
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return true;
		}
		if ($relon.from_nid==null || $relon.type==null || $relon.reln==null || $relon.to_type==null || $relon.to_s==null ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return true;
		}
		return false;
	}


    /**    
     * Validate the format of a neuron's value property.
     * 
     * @param {anne_neuron} - The ANNE neuron you want to validate format for.
     * @author {Mr.Scatman} 
     */
    __MASTER_FORMAT_VALIDATOR=function($nObj) {
        $nObj.format_validator($nObj);
    }


    /**  
     * Use the set_value function and get the value.
     *   
     * We call get value when dealing directly with nObjs for various reasons
     * 
     * @param {anne_neuron} - The ANNE neuron you want to get the run-time value for. 
     * @author {Mr.Scatman} 
     */
    __MASTER_GET_VALUE = function ($nObj) {
        if (typeof $nObj!=="object") {
            $nObj = MEMG.getNeuronByNid($nObj);
            if ($nObj==null) return "";
        }
        if ($nObj.bes[0]==this.integer.temp_nid) {
            $nObj.setValue(parseInt($nObj.text));
            return $nObj.value;
        }

        if ($nObj.set_value!=undefined && typeof $nObj.set_value==="function" && $nObj.ntype==NTYPE_INST) {
            this.__MASTER_SET_VALUE($nObj);
        }

        //try for an eval, value.  nOBJ->eval->xyz
        else {
            this.eval.set_value($nObj);
        }

        if ($nObj.value!=undefined) return $nObj.value;
        return $nObj.text;
    }


    /**  
     * Set the value of a neuron.
     * 
     * Use the master set value.  If there is a validator call it.
     * 
     * @param {anne_neuron} - The ANNE neuron you want to set the run-time value for. 
     * @author {Mr.Scatman} 
     */
    __MASTER_SET_VALUE = function ($nObj) {
        //if ($nObj.ntype!=NTYPE_INST) return;
        if ($nObj.set_value!=undefined) $nObj.set_value($nObj);
        if ($nObj.validator!=undefined) {
            this.__MASTER_FORMAT_VALIDATOR($nObj);
        }
    }


    /**  
     * Get a nid array from the worldview temp nid array.
     * 
     * We store the temp_nids array because it is a smaller data footprint.
     * @author {Mr.Scatman} 
     */
     getWorldViewNidList = function () {
        var wvs=[];
        for (var w=0; w < this._activeWorldViewTempNids.length; w++) {
            var wvOBJ = MEMG.getNeuronByTempNid(this._activeWorldViewTempNids[w]);
            wvs.push(wvOBJ.nid);
        }
        return wvs;
    }


    /**
     * just a convienance.. makes it easier than inline
     */
    isECAction=function($nObj, $bINSTONLY=true) {
        if ($nObj==null) return false;
        if ($bINSTONLY==true && $nObj.ntype!=NTYPE_INST) return false;
        return MEMG.besOf($nObj, this.ec_action.temp_nid);
    }
    isVbInst=function($nObj){
        if ($nObj==null) return false;
        if ($nObj.ntype!=NTYPE_INST) return false;
        if ($nObj.is(this.ec_action)==true) return true;
        if ($nObj.is(this.ec_is)==true) return true;
        if ($nObj.is(this.ec_posn)==true) return true;
    }
    isVb=function($nObj){
    }
    isVbClass=function($nObj){
    }

    /**
     */
     isAnorg=function($nObj) {
        if ($nObj==null) return false;
        if ($nObj.ntype!=NTYPE_INST) return false;
        return MEMG.besOf($nObj, this.anorg.temp_nid);
    }

    isRole=function($nObj) {
        if ($nObj==null) return false;
        return MEMG.besOf($nObj, this.role.temp_nid);
    }
    isRoleR=function($r) {
        var nOBJ = MEMG.getNeuronByTempNid($r);
        if (nOBJ==null) return false;
        return MEMG.besOf(nOBJ, this.role.temp_nid);
    }


    /**
     * Some things are just silly to do an ANUM display on....
     * @param {*} $nObj 
     * @returns 
     */
    annumable = function($nObj) {
        if ($nObj==undefined || $nObj==null) return false;
        if (!$nObj.text) return false;
        if (($nObj.is(this.EC_CLASS)==true && $nObj.is(this.ec_thing)==false && $nObj.is(this.ec_action)==false)  
        || $nObj.is(this.__ANNE_SYS_CLASS)==true 
        || $nObj.is(this._GENERAL_PROP_CLASS)==true
        || ($nObj.text.indexOf("USA Rep")!=-1)
        || ($nObj.text.indexOf("USA Sen")!=-1)
        || ($nObj.is(this.service_term)==true)
        || ($nObj.is(this.bounty)==true)
        || ($nObj.is(this.PDN)==true)
        || ($nObj.is(this.ASDF)==true)
        || ($nObj.is(this.distro_req)==true)
        || ($nObj.is(this.service_term))) return false;

        return true;
    }


    /**
     * Check if a nObj is a child of any of the $bes
     */
    isInBeLines=function($nObj, $bes) {
        for (var b=0; b < $bes.length; b++) {
            var be_tnid = $bes[b];
            if (MEMG.besOf($nObj, be_tnid)==true) return true;
        }
        return false;
    }

    /**
     */
    isIdentity=function($nObj) {
        if ($nObj==null) return false;
        return MEMG.besOf($nObj, this.identity.temp_nid);
    }

    /**
     */
    isIdentityINST=function($nObj) {
        if ($nObj==null) return false;
        if ($nObj.ntype!=NTYPE_INST) return false;
        return MEMG.besOf($nObj, this.identity.temp_nid);
    }

    /**
     */
    unitOfMeasure = function($psblUOM) {
        //conversion unit method.. 
        if (MEMG.besOf($psblUOM, this.unit_of_measure.temp_nid)==false) return null;

        //will return a uom object if we can make one.
        return this.factor.set_value($psblUOM);
    }

    convertUOM = function($uom, $uom_start_val, $altToUnits=null) {
        //$uom.unit = nObj of unit
        //$uom.value = factor already done.
        if ($uom==null) return "";
        if ($uom.unit==null) return "";
        if ($uom.factored==null) return "";
        var unit_lbl = "";
        if ($uom.unit_lbl!="") unit_lbl = $uom.unit_lbl;
        return parseFloat($uom_start_val)*parseFloat($uom.factored)+unit_lbl;
    }



    /**
     * 
     * @param {*} $textOld 
     * @param {*} $textNew 
     * @param {*} $nid 
     * @returns 
     */
    editTextIndex = function($textOld, $textNew, $nid) {
        if ($textOld==null) return;
        $textOld = $textOld.trim()    
        if ($textOld=="") return;
        $textOld = $textOld.replace(/[^\w\s]/gi, '').replace(/\s+/g, ' ');
        if ($textOld==" ") return;

        if ($textNew==null) return;
        $textNew = $textNew.trim()    
        if ($textNew=="") return;
        $textNew = $textNew.replace(/[^\w\s]/gi, '').replace(/\s+/g, ' ');
        if ($textNew==" ") return;

        if (this._TEXT_INDEX[ $textOld ] == undefined) {
            console.log("huh???");
        }
        else {
            var entryArr = this._TEXT_INDEX[ $textOld ];
            if (entryArr.length==1) delete this._TEXT_INDEX[ $textOld ];
            else {
                for (var s=0; s < entryArr.length; s++) {
                    if (entryArr[s]==$nid) {
                        entryArr.splice(s, 1);
                        break;
                    }
                }
            }
        }
        this.addToTextIndex($textNew, $nid);
    }


    /**
     * 
     * @param {*} $nids 
     */
    removeTextIndexByNids=async($nids)=> {
        var keys = Object.keys( this._TEXT_INDEX);
        for(var k=0; k < keys.length; k++) {
            var key = keys[k];
            var arr = this._TEXT_INDEX[key];
            for(var a=0; a < arr.length; a++) {
               if ($nids.indexOf(arr[a])!=-1) {
                   if (arr.length>1) {
                      a--;
                      arr.splice(a,1);
                      continue;
                   }
                   else delete this._TEXT_INDEX[key];
               }
            }
        }
    }


    /**
     * 
     * @param {*} $nid 
     */
    removeTextIndexByNid=function($nid) {
        var keys = Object.keys( this._TEXT_INDEX);
        for(var k=0; k < keys.length; k++) {
            var key = keys[k];
            var arr = this._TEXT_INDEX[key];
            for(var a=0; a < arr.length; a++) {
               if (arr[a]==$nid) {
                   if (arr.length>1) {
                      a--;
                      arr.splice(a,1);
                      continue;
                   }
                   else delete this._TEXT_INDEX[key];
               }
            }
        }
    }


        
    /**
     * typically used when deleting ghosts. 
     * @param {*} $text 
     * @param {*} $nid 
     * @returns 
     */
    removeTextIndex = function($text, $nid) {
        if ($text==null) return;
        $text = $text.trim()    
        if ($text=="") return;
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text==" ") return;

        if (this._TEXT_INDEX[ $text ] == undefined) {
            console.log("no text index for: "+$text);    
            return;
        }
        else {
            var entryArr = this._TEXT_INDEX[ $text ];
            if (entryArr.length==1) {
                delete this._TEXT_INDEX[ $text ];
            }
            else {
                for (var s=0; s < entryArr.length; s++) {
                    if (entryArr[s]==$nid) {
                        entryArr.splice(s, 1);
                        break;
                    }
                }
            }
        }
    }


    /**
     * 
     * @param {*} $text 
     * @param {*} $nid 
     * @param {*} $bDOPARTIALS 
     * @returns 
     */
    addToTextIndex = function($text, $nid, $bDOPARTIALS=true) {
        if ($text==null) return;
        $text = $text.trim()    
        if ($text=="") return;
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text==" ") return;
        $text = $text.toLowerCase();
            
        if ($text=="" || $text==" ") return;

        if (ANNE.MEM._TEXT_INDEX[ $text ] == undefined) {
            ANNE.MEM._TEXT_INDEX[ $text ] = [ $nid ];
        }
        else if (ANNE.MEM._TEXT_INDEX[ $text].indexOf($nid) == -1) {
            ANNE.MEM._TEXT_INDEX[ $text ].push($nid);
        }

        //this will blowout ties in our index for any "wd" 
        //experimental for lukat search finding partials..
        //at present it must find full entry in the index (of name/alias or anything referenced)
        if ($bDOPARTIALS==true) {
            var parts = $text.split( /[_\s]+/);
            if (parts.length>1) {
                var partial = ""
                for (var p=0; p < parts.length; p++) {
                    partial+=parts[p];
                    if (partial=="" || partial==" ") {
                        continue;
                    }
                    //if (ANNE.MEM._TEXT_INDEX_PARTIALS[ partial ] == undefined) {
                        //ANNE.MEM._TEXT_INDEX_PARTIALS[ partial ] = [ $nid ];
                    //}
                   // else if (ANNE.MEM._TEXT_INDEX_PARTIALS[ partial].indexOf($nid) == -1) {
                       // ANNE.MEM._TEXT_INDEX_PARTIALS[ partial ].push($nid);
                   // } 

                    if (ANNE.MEM._TEXT_INDEX[ partial ] == undefined) {
                        ANNE.MEM._TEXT_INDEX[ partial ] = [ $nid ];
                    }
                    else if (ANNE.MEM._TEXT_INDEX[ partial].indexOf($nid) == -1) {
                        ANNE.MEM._TEXT_INDEX[ partial ].push($nid);
                    }
                    //partial+=" ";
                }
            }
        }
    }



    /**
     * 
     * @param {*} $text 
     * @returns 
     */
    getTnidsByText = function($text) {
        if ($text==null) return [];
        $text = $text+"";
        if ($text=="") return [];
        $text=$text.trim();
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text=="") return [];
        $text = $text.toLowerCase();
        var arr = ANNE.MEM._TEXT_INDEX [ $text ];
        if (arr==undefined || arr.length==0) return [];
        return ANNE.MEMG.getTnidsFromNids(arr);
    }


    searchTextIndex = async($text, $bPULLIDB=true, $bEXACT=false)=> {
        if ($text==null) return [];
        $text = $text+"";
        if ($text=="") return [];
        $text=$text.trim();
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text=="") return [];
        var exactText = $text;
        $text = $text.toLowerCase();
        var arr = this._TEXT_INDEX [ $text ];
        if (arr==undefined || arr.length==0) return [];

        var nObjs = [];
        var nids_to_pull = [];
        for (var a=0; a < arr.length; a++) {
            let nObj = MEMG.getOrMakeNid(arr[a]);

            if ($bEXACT==true && nObj.text!=exactText) {
                console.log(nObj.text+" is not an exact match, skipping")
                continue;
            }

            if (nObj==null || nObj.text=="" || nObj.bes.length==0) nids_to_pull.push(arr[a]);
            else nObjs.push(nObj);
        }

        if (nids_to_pull.length>0 && $bPULLIDB==true) {
            await ANNE.IDB.lookupNidsGetAllOuts(null, nids_to_pull);
            for(var n=0; n < nids_to_pull.length; n++) {
                let nid = nids_to_pull[n];
                let nObj = MEMG.getNeuronByNid(nid);
                if (nObj!=null) nObjs.push(nObj);
            }
        }
        return nObjs;
    }


    //for loads at run-time of neurons not on-chain.
    //mostly used for testing at present
    _MARKSAND=function() {
        this._sand_mark = this._aRelons.length;
    }
    _CLEANSAND=async()=> {
        await MEMP.reprocessRelonsFrom(this._sand_mark);
        this._sand_mark = 0;
    }

}

const MEM = new mem ();

/************** (from file ./js/mem/mem.js)**/ 

class mem_processor extends anne_singleton {
    constructor(){ 
        super(false);
        this._memProcessRuns=0;
        this._DOINGPROCESSING_BASE_LVL = false;
        this._preloadNids = [];

    }
    init = function($MEM) {
        this._DOINGPROCESSING_BASE_LVL = false;
    }

    /**
     * When we get word of a merge via the neuron that has been merged..
     * We one-time clean up our mess.
     * Nuke the relons before we process the new relons that use the same
     * txids but now map to a different from!
     * WE DO NOT WANT TO FIND THESE IN OUR PROCESSING LOOP SO WE MUST NUKE THEM FIRST!!!
     * 
     * On the execution of the merge the ANNODE moved over (altered froms) of the relons it could use.
     * Any left behind it nuked from relons... so we should not get this neuron once we process
     * the merge info.
     * @param {*} $merge_nids_to_cleanup 
     */
    mergePrep = async($merge_nids_to_cleanup)=> {
        var ars = [];
        var ghost_txids = [];
        var real_txids = [];
        for (var n=0; n < $merge_nids_to_cleanup.length; n++) {
            var mnid = $merge_nids_to_cleanup[n];
            var mergedNOBJ = MEMG.getNeuronByNid(mnid);
            if (mergedNOBJ!=null) MEMP.annihilate_MERGE_RAM(mergedNOBJ, ars);
        }
        for (var r=0; r < ars.length; r++) {
            if (ars[r].ghost==true) ghost_txids.push(ars[r].txid);
            else real_txids.push(ars[r].txids);
        }
        if (ghost_txids.length) await GHOSTDATA.deleteGhostTxids(ghost_txids);
        if (real_txids.length) await ANNE.DB.___delete(real_txids, "relons");
    }


    annihilate_MERGE_RAM = function ($nObj, $annihilated_relons=null) {
        var beNOBJ = MEMG.getNeuronByTempNid($nObj.temp_nid);
        if (beNOBJ!=null) {
            MEMP.deadBeatDad(beNOBJ);
        }
        var annihilated_relons = $annihilated_relons;
        if (annihilated_relons==null) annihilated_relons = [];
        MEMP.annihilateOutRelons_RAM($nObj, annihilated_relons);
        MEMP.annihilateInRelons_RAM($nObj, annihilated_relons);
        MEMP.annihilateRelnRelons_RAM($nObj, annihilated_relons);
        $nObj.bANNIHILATED = true;
        MEMP._resetNeuron_RAM($nObj);
        //this is an array of dead relon objects with the .txid.
        //we can use it to remove data from DB mem as that only uses the .txid index.
        return annihilated_relons;
    }

    processNeuronsFromRelons = async ($aRelons, $bRETURNALL=false, $bDOLOCALSAVE=true, $tbl="relons", $debug="")=> {
        var bNESTED=false;
        if (this._DOINGPROCESSING_BASE_LVL==true && $bDOLOCALSAVE==true) {
            bNESTED = true;
        }
        
        this._DOINGPROCESSING_BASE_LVL = true;

        if ($bDOLOCALSAVE==false) console.log("%c MEMP.processNeuronsFromRelons %c"+$aRelons.length+"%c just processing into RAM %c"+$debug, ANNE.TALK.bglightbluebold, ANNE.TALK.bgred, ANNE.TALK.msgNote, ANNE.TALK.fromMsg);
        else console.log("%c MEMP.processNeuronsFromRelons %c relons:"+$aRelons.length +"%c+"+$debug, ANNE.TALK.bglightbluebold, ANNE.TALK.bgred, ANNE.TALK.fromMsg);

        if ($aRelons==undefined || $aRelons==null) {
            this._DOINGPROCESSING_BASE_LVL=false;
           // console.log("%c [DONE] MEMP.processNeuronsFromRelons %c", ANNE.TALK.bglightbluebold, ANNE.TALK.bgred)
            return [];
        }
        if ($aRelons.length==0) {
            this._DOINGPROCESSING_BASE_LVL=false;
            //console.log("%c [DONE] MEMP.processNeuronsFromRelons %c", ANNE.TALK.bglightbluebold, ANNE.TALK.bgred)
            return [];
        }
        
        MEMP._memProcessRuns+=1;

        ANNE.MEM_UDF.addSubHunt("processNeuronsFromRelons ("+$debug+")");

        //this is relatively light (vetted as fast);
        var merge_nids_to_cleanup = await MEMP.__brainPreops($aRelons, $bDOLOCALSAVE);
        
        if (merge_nids_to_cleanup.length>0) {
            console.log("cleaning up local mem for merged neuron");
            await this.mergePrep(merge_nids_to_cleanup);
        }

        var _memPoolBootups = [];
        
        //a radical new approach..
        var am = new arraymap();
        
        if ($bDOLOCALSAVE==true) {
            am.setMapKey("txid");
            
            for (let r = 0; r < $aRelons.length; r++) {
                let rObj = $aRelons[r];
                rObj.idb_insert=true;
                
                //note: dont save this relon, EVER
                if (rObj.reln==ANNE.MEM.INTRA_BLOCK_PENDING.nid && rObj.to_s=="pending annex chg") {
                    var rtxid = rObj.txid;
                    var rsysreftxid = rObj.sys_ref_txid;
                    console.log("rtxid:"+rtxid+" vs "+rsysreftxid);
                }

                if (merge_nids_to_cleanup.length && merge_nids_to_cleanup.indexOf(rObj.from_nid)!=-1) continue;
                if (rObj.emulated!=undefined && rObj.emulated==true) continue;
                if (rObj.active==false) continue;
                let time = parseInt(rObj.time); //unconf shoudl be -1 now (NOT 0);
    
                //just a helpful marker to know where we are at on the chain (without looking)
                if (rObj.time!=undefined && time > HISTORY._GREATEST_LAST_KNOWN_BLOCK) {
                    if (rObj.temp==undefined || rObj.temp==false) HISTORY.updateGLKB_fromRelons(time);
                }
                am.addU( rObj);
            }

            //avoiding the insert problem when we haven't yet pulled mem for this
            //from IDB.

            //neuorn_necrosis_list is a summary of all necrosis cmds we've hit this 
            //session in our mem manager.  We can skip all references to these nids, safely.
            if (ANNE && ANNE.IDB) await ANNE.IDB.checkForNewTxids_ONLY(am, MEM.neuron_necrosis_list);
        }

        //we already looped/clenaed the doing local save.
        // just loading this part into RAM.
        // bDOLOCALSAVE==true but am.arr_other --> these are the 
        // relons we already had in IDB and moved over to the arr_other arr.
        // They are update candidates if we had net-new info on them.
        // This was already determinced in the checkForNewTx1ds_ONLY function.
        else if ($bDOLOCALSAVE==false) {

            //we **only** want to process our arr_other relons if we are in bDOLOCALSAVE==true here.
            //|| (am!=null && am.arr_other.length)
            //if(am!=null && am.arr_other.length>0) {
                //$aRelons = am.arr_other;
            //}

            for (let r = 0; r < $aRelons.length; r++) {
                let rObj = $aRelons[r];
                if (merge_nids_to_cleanup.length && merge_nids_to_cleanup.indexOf(rObj.from_nid)!=-1) continue;
                if (rObj.emulated!=undefined && rObj.emulated==true) continue;
                if (rObj.active==false) continue;
                
                //just a helpful marker to know where we are at on the chain (without looking)
                if ((rObj.temp==undefined || rObj.temp==false) && rObj.time!=undefined) {
                    let time = parseInt(rObj.time);
                    if (time > HISTORY._GREATEST_LAST_KNOWN_BLOCK) HISTORY.updateGLKB_fromRelons(time);
                }

                if (rObj.dupe!=undefined && rObj.dupe==true) continue;
                
                var duperel = MEMG.lookupRelonByTXID(rObj.txid);
                if (duperel!=null) {
                    rObj.dupe=true;
                    continue;
                }


                //RELONS THAT WE THINK ARE NOT GHOSTS ----
                //BUT WE HAVE NO TIME YET FOR
                //(MEMPOOL)
                if (rObj.time==-1 && rObj.ghost==false) {
                    //ugly.. bad data, really.  likely on a busted send/return cycle.
                    if (rObj.txid[0]=="z") {
                        rObj.ghost=true;
                        continue;
                    }
                    if (bNESTED==false && ANNE.MEM_UDF.bGETDEDSATBOOTUP==true) {
                        _memPoolBootups.push(rObj.txid);
                    }
                }
            }
        }

        //get all the aRelons back.
        //if(am!=null && $bDOLOCALSAVE==true && am.arr.length>0) {
            //$aRelons = am.arr.concat(am.arr_other);
        //}
        //$aRELONS IS NOW BACK TO ALL.
        //2nd pass.
        var _uniqueFromNidsThisRun = [];
        
        //console.time("memp.pnfr stage 4");

        for (let r = 0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.emulated!=undefined && rObj.emulated==true) continue;
            if (rObj.active==false) continue;

            if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ=MEMG.getOrMakeIfNew(rObj.from_nid);
                
            if (rObj.dupe==true) {

                //update in-mem pubKey for this neuron.  (??? why this one?)
                if (rObj.pubKey!=undefined) {
                    rObj.from_nidTNOBJ.pubKey=rObj.pubKey;
                }

                //make sure it is in our text index.
                ANNE.MEM.addToTextIndex(rObj.from_nidTNOBJ.text, rObj.from_nidTNOBJ.nid);
                    
                //extract key props out of dupe
                rObj.from_nidTNOBJ.extractOtherPropsFromRelon(rObj);
                continue;  
            }

            else if (rObj.pubKey!=undefined) {
                //if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ=MEMG.getNeuronByNid(rObj.from_nid);
                rObj.from_nidTNOBJ.pubKey=rObj.pubKey;
            }

            //extract from/reln and make neurons
            //WE WILL DELETE from_txt_temp and reln_txt_temp after testing.
            //it makes it easier to read CHROME DEBUGGER relons to know what they are.
            MEM.makeAndPushIfUniqueNid(rObj, 'from_nid', 'f', 'from_txt_temp');                            //this sets the .f
            if (_uniqueFromNidsThisRun.indexOf(rObj.from_nidTNOBJ.temp_nid)==-1) {
                _uniqueFromNidsThisRun.push(rObj.from_nidTNOBJ.temp_nid);
            }

            //make sure it is in our text index.
            ANNE.MEM.addToTextIndex(rObj.from_nidTNOBJ.text, rObj.from_nidTNOBJ.nid);
            
            if (rObj.from_nidTNOBJ.ghost==true) {
                rObj.ghost=true;
                rObj.from_nidTNOBJ.owned_by=rObj.ghost_owner;
            }

            //this could be a future COLLAB so we can't mark the fNOBJ as a ghost.
            else if (rObj.ghost_owner!=undefined) {
                rObj.ghost=true;
            }
            
            ANNE.MEM.makeAndPushIfUniqueNid(rObj, 'reln', 'r', 'reln_txt_temp');
            if (rObj.from_nidTNOBJ==null || rObj.relnTNOBJ==null) {
                ANNE.TALK.printRelon(rObj);
                console.warn("we should never get here");
                continue;
            }

            //more robust handling in MEM_UDF __handleSingleRelonPostOps.
            if (rObj.type==TYPE_BE && rObj.r==MEM.name.temp_nid) {  
                rObj.from_nidTNOBJ.text = rObj.to_s;
                rObj.from_nidTNOBJ.setLatestNameRelon(rObj);
            }

            //extract to and make a neuron
            if (rObj.to_type==TO_TYPE_ANNE 
                    && rObj.r!=MEM.arbor_prune.temp_nid
                    && rObj.r!=MEM.inhibit_relon.temp_nid 
                    && rObj.r!=MEM.necrosis.temp_nid) {
                MEM.makeAndPushIfUniqueNid(rObj, 'to_s', 't');
            }

            else if (rObj.to_type!=TO_TYPE_ANNE &&  rObj.to_type!=TO_TYPE_SEED) {
                if ((MEM._emulatables.length && MEM._emulatables.indexOf(rObj.to_type)!=-1)) {
                    rObj.to_sTNOBJ = MEM.makeAndPushIfUniqueEMULATED(rObj.to_s, rObj.to_type, rObj);
                    if (rObj.to_sTNOBJ.ghost==true) rObj.ghost=true; 
                }
            }

            //build a text index (experimental)
            if (rObj.to_type==TO_TYPE_STRING) {
                if (rObj.type!=TYPE_SYS_MSG) {
                    ANNE.MEM.addToTextIndex(rObj.to_s, rObj.from_nid);
                }
            }



            /************************* SPECIAL INDEXES ******************************/
            //the new owned_by prop from ANNODE is on all BEs and SYS relons.
            /************************************************************************/
            if (rObj.type!=TYPE_SYS_MSG) {

                //ANNODE .owned_by index
                if (rObj.owned_by!=undefined && (rObj.type==TYPE_BE || rObj.type==TYPE_SYS)) {
                    rObj.from_nidTNOBJ.addOwnedBy(rObj.owned_by);
                }

                /********************/
                //ANNODE .sponsored_by index, (for self only, build on-demand for other user queries)
                if ((rObj.sponsored_by!=undefined && rObj.sponsored_by!="" && (rObj.from_nidTNOBJ.sponsored_by==null || rObj.from_nidTNOBJ.sponsored_by==""))) {
                    rObj.from_nidTNOBJ.addSponsoredBy(rObj.sponsored_by);
                }

                /********************/
                //ANNODE .parent index
                //NOTE: this is the same as bes[0]
                if ((rObj.parent!=undefined && rObj.parent!="") && (rObj.from_nidTNOBJ.bes.length==0 || rObj.from_nidTNOBJ.parent=="")) {
                    let parentNObj = MEMG.getOrMakeIfNew(rObj.parent);  //this may be the only reference to this neuron we have...
                    rObj.from_nidTNOBJ.setBeTempNid(parentNObj);
                }
            }

            if (rObj.origin!=undefined) rObj.from_nidTNOBJ.origin = rObj.origin;

            if (rObj.ntype!=undefined && (rObj.from_nidTNOBJ.ntype!==1 && rObj.from_nidTNOBJ.ntype!=0)) {
                rObj.from_nidTNOBJ.ntype=rObj.ntype;
            }

            //quick set of polarity, system wide can use this on loads.
            if (rObj.r==MEM.polarity.temp_nid && rObj.type==TYPE_BE && rObj.to_sTNOBJ!=null && rObj.from_nidTNOBJ.polarity==undefined) {
                rObj.from_nidTNOBJ.polarity = rObj.to_sTNOBJ.polarity;
            }

            /********************/
            rObj.from_nidTNOBJ.extractOtherPropsFromRelon(rObj);

            //first load of memory only..
            //if a profile is pulled up and pieces validated,
            //we store the ico_display prop in and do a DB.update w/ the be name relon.
            //we reload that into here so we can have it again and then when we see that user
            //on the site they will have their logos (ie, sponsored by) even if we haven't
            //went to their profile this session.
            if ($bDOLOCALSAVE==false && rObj.type==TYPE_BE && (rObj.r==MEM.class.temp_nid || rObj.r==MEM.inst.temp_nid)) {
                if (rObj.ico_display!=undefined) rObj.from_nidTNOBJ.ico_display = rObj.ico_display;
            }

            //if this is OUR remembez relon, set the rememberz property in the nObj.
            //this will help the gui with the brain icon.
            if ((ANNE.USER!=null && ANNE.USER._userNObj.nid && rObj.from_nidTNOBJ.from_nid==ANNE.USER._userNObj.nid) && 
                rObj.r==MEM.rememberz.temp_nid && rObj.temp==false) {
                rObj.to_sTNOBJ.rememberz=rObj;
            }

            MEMP.__handleSingleRelonPostOps (rObj, rObj.from_nidTNOBJ);
        }

        //console.timeEnd("memp.pnfr stage 4");

        //We've now set up all our pointers and quick indexes.
        //Build arrays for DB manipulation (add/delete).
        //This function will do the updates.
        console.time("processRelonsIntoLocalMem");
        var aNewRelons = [];

        //if local save, we only want to return the aNewRelons
        if ($bDOLOCALSAVE==true) {
            await MEMP.processRelonsIntoLocalMem($aRelons, true);
            if (am!=null) {
                aNewRelons = am.arr;
            }
        }
        else {
            aNewRelons = await MEMP.processRelonsIntoLocalMem($aRelons, false);	
        }
        console.timeEnd("processRelonsIntoLocalMem");

        //a quick loop of NEW relons looking for any arbor or necrosis.
        //we go to those, snag the values, and splice them out of their targets
        //the ar and necro arrays stay in place with the TXIDS of what we removed.
        //console.time("__brainPostops");
        MEMP.__brainPostops($aRelons);
        //console.timeEnd("__brainPostops");

        //DO POST OPS FOR ALL UNIQUE NIDs.
        //console.time("nidsPostOps");
        MEMP.nidsPostOps(_uniqueFromNidsThisRun, MEMP._memProcessRuns);
        //console.timeEnd("nidsPostOps");

        //nuke these nids and any connected relons
        //we updated our new relons (old from to new from)
        //whatever is still left n the absorbed needs to be nuked.

        if (MEM._localRunDelete.length) {
            //console.time("_localRunDelete");
            if (ANNE.DB!=null) await DB.___delete(MEM._localRunDelete, "relons");
            
            //now do the mem wipe of txid... (keep the entry to not upset the array indexes in this session)
            for (let d=0; d<MEM._localRunDelete.length; d++) {
                MEMP.__removeByTXID_master(MEM._localRunDelete[d]);
            }
            MEM._localRunDelete=[];
            //console.timeEnd("_localRunDelete");
        }

        if (ANNE.GUI_STUFF!=null) ANNE.GUI_STUFF.updateMemStats();

        //update the localstore with the aNewRelons!
        //Note: we already updated the ones that were existing relons with net new info.
        if ($bDOLOCALSAVE==true) {
            
            if(aNewRelons.length && ANNE.DB!=null) {
                //console.time("___insert"+aNewRelons.length);
                console.log("attemping to add "+aNewRelons.length+" new relons to store");
                
                await DB.___insert(aNewRelons, $tbl);
                //console.timeEnd("___insert"+aNewRelons.length);
            }
        }

        //this comes from local annode useage and stopping an annode (losing the txs in the mempool)
        //when the annnode boots up it will nuke those stranded TXIDS (and neurons) and replace them 
        //if another mempool broadcasts to them OR a block contains them.
        //In either case, we're safe to assume they are dead now, until they come back.
        //This list is a list of potential deads.
        //We need to lookup each one of these against the active mempool..
        //So ---- get mempool txids.
       
        if (bNESTED==false && _memPoolBootups.length > 0 && ANNE.MEM_UDF.bGETDEDSATBOOTUP==true) {
            
            //console.time("_memPoolBootups");
            ANNE.MEM_UDF.bGETDEDSATBOOTUP = false;

            //console.warn("we have potential relons to nuke from this app that were mempool but no longer on the node. INVESTIGATE.");
            //var allDedTxids = [];
            //console.time("DEDS");
            //consider a max here... just get what we can done (a few hundred or so..)
            //We don't want to needlessly chase all TXs from a huge blast out ..  ie, 5k TXs
            //that are all unconfirmed .. no need to panic and chase em.
            //wish i could store a last shutdown variable in the annode.. hmmm.
            //allDedTxids = await ANNE.CHASE.getAllDEDTXIDS(_memPoolBootups);

            //if (allDedTxids.length>0 || ANNE.MEM._localRunDelete.length>0) {
            if (ANNE.MEM._localRunDelete.length>0) {
                //nuke em.
                //note -- we should loop these for "MAKER" pattern TXs and nuke the neurons too.
                //note2 -- a really frenly gesture would be to recreate the ghost neurons AND values.
                //ANNE.MEM._localRunDelete = ANNE.MEM._localRunDelete.concat(allDedTxids);

                if (MEM._localRunDelete.length) {
                    console.log("%cDELETING TXS that are dead %c", TALK.bgred, TALK.bgyellow);
                    if (ANNE.DB!=null) await DB.___delete(MEM._localRunDelete, "relons");
                    //now do the mem wipe of txid... (keep the entry to not upset the array indexes in this session)
                    for (let d=0; d<MEM._localRunDelete.length; d++) {
                        MEMP.__removeByTXID_master(MEM._localRunDelete[d]);
                    }
                    MEM._localRunDelete = []; //reset for next run.
                    if (MEM._thisRunNECROSIS.length>0) {
                        await ANNE.MEM.removeTextIndexByNids(MEM._thisRunNECROSIS);
                        await ANNE.MEMP.necrosisNeuronsRAM(MEM._thisRunNECROSIS);
                    }
                    MEM._thisRunNECROSIS = [];//reset for next run.
                }
            }
            //console.timeEnd("DEDS");
            //console.timeEnd("_memPoolBootups");
            //the other should be GOOD and have their time updated if there is a new time. (mined in a block)
        }

        //on any run to store relons
        //we clean up deletes (from idb)
        //and we clean up ram for anything we deleted
        else if (bNESTED==false) {
             console.time("_localRunDelete");
            if (MEM._localRunDelete.length) {
                console.log("%cDELETING TXS that are dead %c", TALK.bgred, TALK.bgyellow);
                if (ANNE.DB!=null) await DB.___delete(MEM._localRunDelete, "relons");
                for (let d=0; d<MEM._localRunDelete.length; d++) {
                    MEMP.__removeByTXID_master(MEM._localRunDelete[d]);
                }
                MEM._localRunDelete = []; //reset for next run.
            }
            if (MEM._thisRunNECROSIS.length>0) {
                await ANNE.MEM.removeTextIndexByNids(MEM._thisRunNECROSIS);
                await ANNE.MEMP.necrosisNeuronsRAM(MEM._thisRunNECROSIS);
                MEM._thisRunNECROSIS = [];//reset for next run.
            }
             console.timeEnd("_localRunDelete");
        }

        if ($bRETURNALL==true) {
            if (am!=null && am.arr_other.length) return $aRelons.concat(am.arr_other);
            this._DOINGPROCESSING_BASE_LVL=false;
            ANNE.MEM_UDF.closeSubHunt();
            return $aRelons;
        }

        this._DOINGPROCESSING_BASE_LVL=false;
        ANNE.MEM_UDF.closeSubHunt();
        return aNewRelons;  //return aNewRelons just to return the new ones.
    }
    

    /** 
     * 
     * NOTE: not doing be lvl setting / enforcement as we no longer have the entire mem
     * in RAM.  We cannot count on what we have until we chase the BE LINE UP, then we can build
     * the be lvls.
     * 
     * Run from the mem processor for each unique neuron in the processing run.
     * This essentially means any neurons that were in the "from" slot of a relon
     * have their nids post op function run after the processing is done.
     * 
     * @param {array} - list of temp_nids we want to post-process
     * @param {int} - a frenly int of what mem processing run we are on.
     * 
     */
    nidsPostOps($uniqueTNids, $memProcessRuns=0) {
        //console.log("(nidsPostOps) unique from nids in run %c"+$memProcessRuns+": "+$uniqueTNids.length, TALK.bggreen);
        ANNE.MEM._cascadePropsDownChildrenClasses_cnt=0;
        var did_it = [];
        
        //start at every level 1 (with parent of level 0) 
        //be level and cascade its props down.
        //This will do it most efficiently.
        for (var n=0; n <$uniqueTNids.length; n++) {
            var nObj = MEMG.getNeuronByTempNid($uniqueTNids[n]);

            if (nObj.bes.length==0) continue;

            if (nObj.bes[0]==MEM.anndle.temp_nid) {
                if (ANNE.ANNDLE) ANNE.ANNDLE.addQuick(nObj);
            }
 
            //on preload, we only need to chase be_lvl=1.
            //after that, we must do all.
            if (nObj.be_lvl!=2 && $memProcessRuns<=1) continue;
            
            //level 2 inherits from level 1 (beNOBJ).. and so on.
            var beNOBJ = MEMG.getNeuronByTempNid(nObj.bes[0]);
            //console.log("beNOBJ:"+beNOBJ.text+" for neuron:"+nObj.text);
            nObj.__functInherit(beNOBJ, "memp.nidsPostOps");
        }
    }


    /**                                                
     */
     __handleSingleRelonPostOps = function ($rObj, $nObj) {
        if ($nObj==null || $rObj==null) {
            console.log("__handleSingleRelonPostOps cannot have rObj or nObj null", $rObj, $nObj);
            return;
        }

        //classes that don't have a parent class.  (top level classes)
        //or, a class that was created but at the time didn't know of its parent class.
        //both cases will produce a TYPE_SYS reln:class to:class TX.
        else if ($rObj.type == TYPE_SYS) {

        }

        else if ($rObj.type == TYPE_BE) {
            if ($rObj.r == MEM.name.temp_nid) {
                if ($rObj.to_type==TO_TYPE_STRING) {
                    $nObj.text = $rObj.to_s.trim(); //update the text with the name relon.  Always.
                }
            }

            //neuron->class->class, a top.
            else if ($rObj.r==MEM.class.temp_nid) {
                                    
                if ($rObj.t==MEM.__ANNE_SYS_CLASS.temp_nid) {
                    if (MEM._SYS_PROPS_TNIDS.indexOf($rObj.f)==-1) {
                        MEM._SYS_PROPS_TNIDS.push($rObj.f);
                    }
                }

                else if ($rObj.t == MEM.class.temp_nid) {
                    if ($nObj.ntype == NTYPE_UNKNOWN) {
                        if ($nObj.be=="" && $nObj.bes.length==0) {
                            $nObj.setBeTempNid(MEM.class.temp_nid);
                        }
                    }
                }
            }
            else if ($rObj.r==MEM.inst.temp_nid) {
                if ($rObj.t==MEM.__ANNE_SYS_CLASS.temp_nid) {
                    if (MEM._SYS_PROPS_TNIDS.indexOf($rObj.f)==-1) {
                        MEM._SYS_PROPS_TNIDS.push($rObj.f);
                    }
                }
                if ($nObj.ntype=="") {
                    $nObj.ntype=NTYPE_INST;
                    $nObj.bTERMINATING_TYPE=true; 
                }
            }

            //manual override for a class to be a terminating type.
            else if ($rObj.r==MEM.TERMINATING_TYPE.temp_nid) {
                if ($rObj.t && $rObj.to_type==TO_TYPE_ANNE) {
                    ANNE.MEM.mapSysCode($nObj, $rObj.t);
                    $nObj.bTERMINATING_TYPE = true;
                }
            }
        }

        else if ($rObj.type == TYPE_HAS) {
                        
            //these need to be set before post ops which 
            //depend on them.. so we can capture them here and then
            //know they are set when we run the loop again.
            //set pk (primary_prop)
            //this is for neurons that have a special identifier like another key to make 
            //them unique.  NOTE: this is documentation of a be_class' primary prop so it is 
            //not a TYPE_BE relon to specify what the be prop is.
            if ($rObj.r==MEM.primary_key.temp_nid) {
                if ($nObj['pk']==undefined) $nObj['pk']=$rObj.t;
                else {
                    if ($nObj['pk']!=$rObj.t) {
                        console.warn("we already have a pk! This is different! Serious trouble!");
                    }
                }
            }
             
            else if ($rObj.r==MEM.primary_key2.temp_nid) {  //SET fk property
                if ($nObj['pk2']==undefined) $nObj['pk2']=[$rObj.t];
                else {
                    if ($nObj['pk2'].indexOf($rObj.t)==-1) {
                        $nObj['pk2'].push($rObj.t);
                    }
                }
            }
        }
    }


    /** 
     * This is not necessary with ANNODE now, but we will still do it.
     */
    __brainPreops = async($aRelons, $bDOLOCALSAVE=false)=> {
        var bad_audit = 0;
        var bad_data_rels = [];
        var merge_nids_to_cleanup = [];

        for (var r=0; r <$aRelons.length; r++) {
            var rObj = $aRelons[r];

            //we will remove these after we are done with them.                        
            rObj.from_nidTNOBJ = null;
            rObj.relnTNOBJ = null;
            rObj.to_sTNOBJ = null;

            //we get the cmd one-time, we do not store it.
            //we use it run-time to fix our RAM.
            if (rObj.type==TYPE_SYS && rObj.reln==MEM.MERGED_WITH.nid) {
                merge_nids_to_cleanup.push(rObj.from_nid);  
                rObj.active=false;  
                continue;
            }

            //this is for personal ghosts
            if (rObj.ghost==true) {
                //rObj.temp=true;
                if (rObj.ghost_order!=undefined) {
                    if (parseInt(rObj.ghost_order) > MEM._nextGhostOrder) {
                        MEM._nextGhostOrder = parseInt(rObj.ghost_order)+1;
                    }
                }
            }
            if (rObj.emulated!=undefined && rObj.emulated==true) {
                rObj.active=false;
                continue;
            }
            if(rObj.type==undefined) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="type is undefined";
                continue;
            } 
            if (rObj.to_type==undefined) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="to_type is undefined";
                continue;
                
            }
            if (rObj.reln==undefined) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="reln is undefined";
                continue;
            }
            if (rObj.from_nid==undefined) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="from_nid is undefined";
                continue;
            }
            if (rObj.to_s==undefined) {    //not our relon
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="to s is undefined";
                continue;
            }
            rObj.active = true;		
            if (rObj.to_s==null || rObj.to_type==null || rObj.reln==null  || rObj.from_nid==null || rObj.type==null) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="something is null";
                continue;
            }
            if (rObj.to_s === "") {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="to_s is not set";
                continue;
            }
            if (rObj.to_type === "") {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="to_type not set";
                continue;
            }
            if (rObj.reln === "") {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="reln is not set";
                continue;
            }
            if (rObj.from_nid === "") {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="from_nid is not set";
                continue;
            }
            if (rObj.type === "") {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="type is not set";
                continue;
            }

            rObj.type = parseInt(rObj.type);
            if (rObj.type <0 || rObj.type >255) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="type is out of bounds";
                continue;
            }
            rObj.to_type = parseInt(rObj.to_type);
            if (rObj.to_type <0 || rObj.to_type >255) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="to type is out of bounds";
                continue;
            }

            //strings of longs can be any length up until
            if (rObj.from_nid.length > 20) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="length of reln is too long!";
                continue;
            }

            if (rObj.reln.length > 20) {
                rObj.active=false;
                rObj.bad_data=true;
                bad_audit+=1;
                bad_data_rels.push(rObj);
                rObj.bad_reason="length of reln is too long!";
                continue;
            }
           
            if (rObj.to_type==TO_TYPE_ANNE) {
                if (rObj.to_s.length>20) {
                    rObj.active=false;
                    rObj.bad_data=true;
                    bad_audit+=1;
                    bad_data_rels.push(rObj);
                    rObj.bad_reason="length of toS for a TO_TYPE_ANNE is too long!";
                    continue;
                }
            }
            
            //if this was not stored right on chain (as a number) then convert it here.
            else if (rObj.to_type==TO_TYPE_NUMBER) rObj.to_s = parseFloat(rObj.to_s);
            else {
                rObj.to_s = rObj.to_s+"";
                //if (rObj.to_s!=" ") rObj.to_s = rObj.to_s.trim();
            }

            if (rObj.ghost==undefined) rObj.ghost = false;
            var nid = rObj.from_nid;
            
            //add to the master arbor prune and necrosis lists of ANNE.
            //we will also add to this list with any REDLIST items (once we implement that!)  :)
            if (nid==MEM.ANNE_ADMIN_ACCTID) {
                if (rObj.type==TYPE_SYS) {
                    //when ANNE arbors, its gone.
                    if (rObj.reln==MEM.arbor_prune.nid) {       //use full .reln and .nid here.  The .r is not yet set
                        if (MEM.tx_ap_inh_list.indexOf(rObj.to_s)==-1) {
                            MEM.tx_ap_inh_list.push(rObj.to_s);   
                        }
                        if (rObj.ghost==false) {
                            rObj.active=false;
                            rObj.idb_delete=true;
                            if (MEM._localRunDelete.indexOf(rObj.txid)==-1) {
                                MEM._localRunDelete.push(rObj.txid);
                            }
                        }
                    }

                    //when ANNE inhibits, it is an arbor
                    else if (rObj.reln==MEM.inhibit_relon.nid) { //use full .reln and .nid here.  The .r is not yet set
                        MEM.tx_ap_inh_list.push(rObj.to_s);   
                        if (rObj.ghost==false) {
                            rObj.active=false;
                            rObj.idb_delete=true;
                            if (MEM._localRunDelete.indexOf(rObj.txid)==-1) {
                                MEM._localRunDelete.push(rObj.txid);
                            }
                        }
                    }
                }
            } 

            //we onlyl want to process the real necrosis..
            //the SELF-TYPE_SYS->necrosis->SELF relon burned by protocol.
            else {
                
                //TO DO downstream: 
                //we will want to delete anything from db mem as well "from" or "to" the rObj.to_s
                //using _localRunDelete
                if (rObj.from_nid==rObj.to_s && rObj.type==TYPE_SYS && rObj.reln==MEM.necrosis.nid) { //use full .reln and .nid here.  The .r is not yet set   
                    //only interested in the final protocol-created necrosis here.
                    rObj.active=false;
                    rObj.idb_delete=true;
                    if (MEM.neuron_necrosis_list.indexOf(rObj.to_s)===-1) {
                        MEM.neuron_necrosis_list.push(rObj.to_s);
                    }
                    else continue; //done this already.

                    if (MEM._localRunDelete.indexOf(rObj.txid)==-1) {
                        MEM._localRunDelete.push(rObj.txid);
                    }
                    if (MEM._thisRunNECROSIS.indexOf(rObj.to_s)==-1) {
                        MEM._thisRunNECROSIS.push(rObj.to_s);
                    }
                }
            }
        }
        
        //PRELOAD LOGICS
        if ($bDOLOCALSAVE==true) {
            this.setPreloadFlag($aRelons); //set preload value here.
        }
        
        if (bad_audit>0) {
            console.warn("BAD DATA AUDIT", bad_audit);
            console.log(bad_data_rels);
            var bGHOST = false;
            if ($bDOLOCALSAVE==false) {   //load only from mem, fix our mem!
                var bdtxids = [];
                for (var d=0; d < bad_data_rels.length; d++) {
                    bdtxids.push(bad_data_rels[d].txid);
                    if (bad_data_rels[d].ghost==true || bad_data_rels[d].txid[0]=="z") bGHOST=true;
                }
                if (bGHOST==false) await ANNE.DB.___delete(bdtxids);
                else await await ANNE.DB_local.___delete(bdtxids);
            }
        }
        return merge_nids_to_cleanup;
    }


    /**
     * improved and more efficient
     * @param {*} $rels 
     */
    setPreloadFlag=function($rels) {
        var newHere = []
        var addedPs = 0;

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (rel.preload!=undefined && rel.preload==1) continue;
            
            //any relon with a from nid we have in our arrays.
            if (ANNE.MEMP._preloadNids.indexOf(rel.from_nid)!=-1) {
                rel.preload=1;
                addedPs+=1;
                continue;
            }

            //put these in min boot?  yes.
            if (rel.type==TYPE_INTERNAL_PAY || rel.type==TYPE_BROADCAST_PAY) {
                rel.preload=1;
                addedPs+=1;
                continue;
            }

            //countries
            else if (rel.type==TYPE_BE && rel.to_s==MEM.country.nid) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }        
            }

            //states
            else if (rel.type==TYPE_BE && rel.to_s==MEM.state.nid) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }         
            }

            //integers
            else if (rel.type==TYPE_BE && rel.to_s==MEM.integer.nid) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }         
            }

            //dates
            else if (rel.type==TYPE_BE && ANNE.MEM.isDateAny(rel.to_s)==true) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }       
            }

            //ascii
            else if (rel.type==TYPE_BE && rel.to_s==MEM.ASCII.nid) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }          
            }

            
            //preload identitie bes so we have values to work with on login
            else if (rel.type==TYPE_BE && rel.to_type==TO_TYPE_ANNE && ANNE.isIdentity(rel.to_s)==true) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }      
            }
            
            //BRAIN dimension puts feelz in buckets.
            else if (rel.type==TYPE_BRAIN && rel.to_type==TO_TYPE_ANNE && rel.to_s==MEM.feeling.nid) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }   
            }

            //opinionz buckets outs
            else if (ANNE.AWARE!=null && ANNE.AWARE.fzOpNids!=null && ANNE.AWARE!=null && ANNE.AWARE.fzOpNids.length>0 && ANNE.AWARE.fzOpNids.indexOf(rel.from_nid)!=-1) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                } 
            }

            //any anne class
            else if (rel.owned_by!=undefined && ANNE.isANNE(rel.owned_by)==true) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (rel.from_nidTNOBJ.ntype==NTYPE_CLASS) {
                    ANNE.MEMP._preloadNids.push(rel.from_nidTNOBJ.nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }
                else if (rel.from_nidTNOBJ.is(MEM.fileX)==true) {
                    //console.log("minboot file neuron "+fOBJ.nid);
                    ANNE.MEMP._preloadNids.push(rel.from_nidTNOBJ.nid);
                    rel.preload=1;
                    newHere.push(rel.from_nid);
                    addedPs+=1;
                }
            }

            //identities in preloads? //ANNE.isIdentity(rel.to_s)==true

            //altered this to any anne-owned class.
            else if (rel.reln=="7777942827590344670" && rel.type==TYPE_BE) {
                if (rel.from_nidTNOBJ!=null) rel.from_nidTNOBJ = rel.from_nidTNOBJ;
                else rel.from_nidTNOBJ = ANNE.MEMG.getOrMakeIfNew(rel.from_nid);
                if (ANNE.isANNE(rel.from_nidTNOBJ.owned_by)==true) {
                    ANNE.MEMP._preloadNids.push(rel.from_nid);
                    rel.preload=1;
                    newHere.push(rel.to_s);
                    addedPs+=1;
                } 
            }
        }


        //we may have hit preload relons (of the right class) before we knew what that 
        //class was.
        if (newHere.length>0) {
            for (var r=0; r < $rels.length; r++) {
                if ($rels[r].preload!=undefined && $rels[r].preload==1) continue;
                var rel = $rels[r];
                if (ANNE.MEMP._preloadNids.indexOf(rel.from_nid)!=-1) {
                    rel.preload=1;
                    addedPs+=1;
                    continue;
                }
            }
        }

        ANNE.TALK.printf_console("added new preload nids ", TALK.GREEN, newHere.length+"", TALK.WHITE, "relons ", TALK.GREEN, addedPs+"");
    }


    /**
     * Build add/delete arrays and update net-new info into local DB mem.
     * 
     * This will also manipulate the local RAM neurons/relons.
     * @param {array} - the array of Relons from this processing run.
     * @param {boolean} - a flag to indicate if we are saving this run into the local DB
     * @author {Mr.Scatman}
     */
    processRelonsIntoLocalMem = async ($aRelons, $bDOLOCALSAVE=true)=> {
        var aNewRelons = [];

        for (var r=0; r <$aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.bad_data==true) continue;  //we do nothing with bad data.

            if ($bDOLOCALSAVE==true) {
                //ths matches a neuron that underwent necrosis
                if (MEM.neuron_necrosis_list.indexOf(rObj.from_nid)!=-1) {
                    rObj.bad_data=true;
                    rObj.active=false;
                    continue;
                }

                //this has already been arbor pruned or inhibited.
                if (MEM.tx_ap_inh_list.indexOf(rObj.txid)!=-1) {
                    rObj.bad_data=true;
                    rObj.active=false;
                    continue;
                }
            }   

            //DUPES CAN PROVIDE NET-NEW INFO SO THERE COULD BE AN UPDATE.
            if (rObj.active==false || rObj.dupe==true) {  
                if (rObj.emulated!=undefined && rObj.emulated==true) continue;
                if (rObj.ghost==true) continue; //ghost never get touched here.
                
                //we already  marked this as a dupe in our pre-process.
                //snag the real relon and look for updatables.
                //this check is necessary when we jam existing txids from their creation (into mem)
                //as temp objects before we pull them from the blockchain.   In these cases we have
                //a real TXID but it is a .temp relon.
                var inMemRObj = null; 
                if (rObj.dupe!=undefined && rObj.dupe==true && (rObj.idb_update==undefined || rObj.idb_update==true)) {
                    
                    //if we have a dupe flag but no dupe_robj_in_mem.. uh... that should not be possible.
                    if (rObj.dupe_robj_in_mem!=undefined) {
                        //snag a ptr to the dupe in memory.
                        inMemRObj = rObj.dupe_robj_in_mem;
                        delete rObj.dupe_robj_in_mem;   //delete the ptr out of the "new" rObj
    
                        //we are just re-loading to the to_s into the working-memory..
                        if (rObj.to_type==TO_TYPE_BLOB && rObj.to_s!="" && inMemRObj.to_s=="BLOB") {
                            inMemRObj.to_s = rObj.to_s; // just update to_s and get out!  no store
                            continue;
                        }    
                    }
                    else continue;
                }

                //our temp neuron (or an existing one w/ more info) 
                if (rObj.idb_update!=undefined && rObj.idb_update==true) {
                    delete rObj.idb_update; //we don't need this anymore.
                    if ($bDOLOCALSAVE==true) {
                        if (inMemRObj!=null) {
                            //console.log("%c updating relon in RAM "+rObj.txid, TALK.bgpink);
                            await DB.___update(inMemRObj, 'relons');
                            $aRelons[r]=inMemRObj;    
                            rObj = $aRelons[r];         //flip the ptrs back.  We're done with inMemRObj special handling
                        }
                    }
                    else {     
                        console.warn("*** WE SHOULD NOT HAVE GOTTEN HERE!!!! ***");
                        continue;
                    }
                }
                continue;
            }   

            //DUPES HAVE NOW BEEN HANDLED.
            if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);

            //the relon finally makes it in to memory we will keep (in the browser).
            if (rObj.time==undefined) rObj.time=-1;

            //this is necessary for temp relons we created that were given a temp TXID 
            //(not a real one) now we have the real one so this is where finally swap it out.
            var temp_tx_id = "";
            //if ($bDOLOCALSAVE==true && rObj.from_nidTNOBJ.nid!="7777089034065106708" && rObj.type!=TYPE_INTERNAL_PAY) {
                //temp_tx_id = MEMP.killTempRelon(rObj.from_nidTNOBJ, rObj);
            //}

            //this relon and its txid are new.
            //hash it in our relons hash table.
            //if this was a preload it is already hashed in its creation function.
            //if (temp_tx_id=="") {
                if (rObj.temp==undefined) rObj.temp=false;
                if ((rObj.preloaded==undefined || rObj.preloaded==false) 
                   && (rObj.dupe==undefined || rObj.dupe==false) && rObj.active==true) MEMP.__hashRelon(rObj);
            //}


            //update RAM outs and ins for the neuron this relon references
            rObj.from_nidTNOBJ.addUniqueOut(rObj);
            MEMP.addUniqueIns(rObj);          
            
            //we push the new relon to the aNewRelons array to tell our caller to do a DB.update
            //we want this even with the killTempRelon match that updated MEM._aRelon memory
            //and has deleted the old temp xxx txid from DB memory.  It needs to add the new
            //relon object back into DB memory.  (it is already in MEM._aRelon memory and didn't
            //need to be added there again).
            if (rObj.idb_insert==true) {
                aNewRelons.push(rObj);  
            }

            delete rObj.idb_insert;
            delete rObj.idb_update;
            
            //smart update for gui, everywhere.
            //put smart_+"temp_nid" on anything we want the name to update once we get more info.
            if ($bDOLOCALSAVE==true && rObj.r==MEM.name.temp_nid && rObj.type==TYPE_BE) {
                var els = document.querySelectorAll(".smart_"+rObj.f);
                for (var e = 0; e < els.length; e++) {
                    var el = els[e];
                    el.innerHTML=rObj.to_s;
                }
            }
        }

        return aNewRelons;      //MUST RETURN NEW RELONS WE WILL BE ADDING TO THE DB
    }


    /**
     */
    addUniqueIns = function ($rObj) {
        if ($rObj.active==false) return;
        if ($rObj.to_type != TO_TYPE_ANNE) return;
        
        if ($rObj.t==undefined) {
            
            if ($rObj.to_type==TO_TYPE_ANNE && $rObj.to_s!=undefined && $rObj.to_s!="" && $rObj.r!=MEM.necrosis.temp_nid) {
                console.warn("we must have a rObj.t at this point!", $rObj);
                ANNE.TALK.printRelon($rObj);
                if ($rObj.to_sNOBJ==null) $rObj.to_sNOBJ = ANNE.MEMG.getOrMakeIfNew($rObj.to_s);
                $rObj.t=toNOBJ.temp_nid;
            }
                
            else return;
        }
        
        if ($rObj.to_sNOBJ==null) $rObj.to_sNOBJ = MEMG.getNeuronByTempNid($rObj.t);
        if ($rObj.to_sNOBJ==null) {
            console.warn("can't add an in without the to_s resolving to a known neuron!");
            $rObj.active = false;
            return;
        }
        $rObj.to_sNOBJ.addUniqueIns($rObj);
    }


    /** 
     * Do inhibits, arbor prunes, necrosis and some other final indexings.
     * 
     * APPLY inhibits and arbor prunes and necrosis
     * REMINDER:  INHIBITS are local throttles that do not remove relons from the ANNODE.
     * Any user can inhibit any relon they like.  It will be honored in their worldview 
     * via data being removed from their local memory arrays and local DB.
     *
     * ARBOR PRUNES with the right authority alter a neuron for EVERYONE.
     * The self, or owner, or ANNE master account can can ARBOR PRUNE relons.
     * In those cases they are removed and cease to exist in the ANNODE and in all
     * downstream use cases.
     */
    __brainPostops = function($aRelons) {

        //code props (in js) can be set with TYPE_BE or TYPE_SYS relons
        var setCodeProps=function($rObj) {
            if ($rObj.relnTNOBJ==null) $rObj.relnTNOBJ = MEMG.getNeuronByTempNid($rObj.r);
            if ($rObj.relnTNOBJ==null) return;//unfortunately, we may not yet have this reln NOBJ in mem.
            var PNAME = $rObj.relnTNOBJ.PARAM_NAME;
            var REL_PNAME = $rObj.relnTNOBJ.REL_PARAM_NAME;
            if (PNAME!=undefined && PNAME!=null && PNAME.length>1 && PNAME!="text" && PNAME!="name") {
                if ($rObj.from_nidTNOBJ==null) $rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid($rObj.f);
                if ($rObj.from_nidTNOBJ!=null) {
                    var to_val = $rObj.to_s;
                    if ($rObj.TO_TYPE_NUMBER) to_val = parseInt(to_val);
                    else if ($rObj.TO_TYPE_BOOLEAN) to_val = Boolean(to_val);
                    //we could have a prop on the relon to let us know
                    //to store the relon, not the to_s, but that's for later.
                    //no need for it for now.  We want the RELON in the quick
                    //_gt and _gti prop of neurons.
                    if (PNAME=="_gt" || PNAME=="_gti") $rObj.from_nidTNOBJ[PNAME]=$rObj;
                    else $rObj.from_nidTNOBJ[PNAME]=$rObj.to_s;
                }
            }

            //if a reln param val exists, it is true.  It is just a flag.
            //ie, .required, ._multiples_ok.  
            //Non presence of the flag inidcates false.  (not required, multiples NOT ok)
            if (REL_PNAME!=undefined && REL_PNAME!=null && REL_PNAME.length>0) {
                if ($rObj.to_type==TO_TYPE_TXID) {
                    var rel = MEMG.lookupByTXID($rObj.to_s);
                    if (rel!=null) rel[REL_PNAME]=true;
                }
            }
        }

        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.dupe==true) {
                delete rObj.from_nidTNOBJ;
                delete rObj.relnTNOBJ;
                delete rObj.to_sTNOBJ;
                continue;
            }

            var INHBIIT = false;
            var ARBOR = false;
            var NECROSIS = false;

            if (rObj.r==MEM.relon_dimension.temp_nid && rObj.type==TYPE_SYS) {
                if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                rObj.from_nidTNOBJ.target_type = parseInt(rObj.to_s);
            }

            if (rObj.r==MEM.inhibit_relon.temp_nid) {
                if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (rObj.from_nidTNOBJ.inhibit_txs==undefined) rObj.from_nidTNOBJ.inhibit_txs=[];
                else if (rObj.from_nidTNOBJ.inhibit_txs.indexOf(rObj.to_s)!=-1) {
                    delete rObj.from_nidTNOBJ;
                    delete rObj.relnTNOBJ;
                    delete rObj.to_sTNOBJ;
                    continue;    //already done
                }
                rObj.from_nidTNOBJ.inhibit_txs.push(rObj.to_s);
                INHBIIT = true;
                rObj.active = true; 
            }
            else if (rObj.r==MEM.arbor_prune.temp_nid) {
                if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (rObj.from_nidTNOBJ.ap==undefined) rObj.from_nidTNOBJ.ap=[];
                else if (rObj.from_nidTNOBJ.ap.indexOf(rObj.to_s)!=-1) {
                    delete rObj.from_nidTNOBJ;
                    delete rObj.relnTNOBJ;
                    delete rObj.to_sTNOBJ;
                    continue;    //alreayd done
                }
                rObj.from_nidTNOBJ.ap.push(rObj.to_s);
                ARBOR = true;
                rObj.active = true; 
            } 

            //for now, let's only apply necrosis on the PROTOTCOL-CREATED SELF-FROM NEURON.
            //that way we need not muck about it before we need it.
            else if (rObj.r==MEM.necrosis.temp_nid) {
                rObj.active = false; 
            }


            if (INHBIIT || ARBOR) {
                var targetRelon = MEMG.lookupRelonByTXID(rObj.to_s);
                if (targetRelon!=null) {

                    //if this relon is from the neuron owner OR ANNE OR US(self)
                    //we can remove it from the arrays.
                    var targetFromNOBJ = MEMG.getNeuronByTempNid(targetRelon.f);

                    if (targetRelon.from_nid==rObj.from_nid 
                        || rObj.from_nidTNOBJ.temp_nid==MEM.__ANNE__.temp_nid
                        || (ANNE.USER!==null && ANNE.USER._userNObj!=null 
                                && ANNE.USER._userNObj.temp_nid!=undefined 
                                && USER._userNObj.nid!=0 
                                && rObj.from_nidTNOBJ.temp_nid==ANNE.USER._userNObj.temp_nid)) {

                        //turn the target relon off.
                        targetRelon.active=false;

                        //remove it from the outs of the from NOBJ
                        rObj.from_nidTNOBJ.removeByTXID('outs', targetRelon.txid);
                        
                        //if the target Relon was an ANNE TX,
                        //also remove the in of the to.
                        if (targetRelon.to_type==TO_TYPE_ANNE) {
                            var toNOBJ = MEMG.getNeuronByTempNid(rObj.t);
                            if (toNOBJ!=null) toNOBJ.removeByTXID('ins', targetRelon.txid);
                        }
                        
                        //add the txid to the tracker array.  We will block any
                        //new attempts to get this .txid back into RAM.
                        MEM.tx_ap_inh_list.push(targetRelon.txid);

                        //this will need to remove the sponsored_by index
                        //on ALL TYPE_BE and TYPE_SYS relons and also update
                        //DB mem for relons already in DB mem.
                        if (targetRelon.r==MEM.sponsored_by) {
                            if (targetFromNOBJ.sponsored_by!=undefined 
                                && targetFromNOBJ.sponsored_by==targetRelon.to_s) {
                                    targetFromNOBJ.sponsored_by="";
                                console.warn("TO DO FIX TYPE BE/SYS from targetFromNOBJ"); 
                            }
                        }

                        else if (targetRelon.r==MEM.inst.temp_nid || targetRelon.r==MEM.class.temp_nid) {
                            
                            //removal of parent as this parent will remove bes array entries
                            //for this neuron.
                            targetFromNOBJ.childDisavowsParent(targetRelon.t);
                            console.warn("TO DO FIX TYPE BE/SYS from targetFromNOBJ "+targetFromNOBJ.text);
                            ANNE.TALK.printRelon(targetRelon);
                        }

                        else if (targetRelon.r==MEM.name.temp_nid) {
                            if (rObj.from_nidTNOBJ.text==targetRelon.to_s) {
                                rObj.from_nidTNOBJ.text="";
                                rObj.from_nidTNOBJ.text=MEM.setTextWithFirstValidAssigner(rObj.from_nidTNOBJ);
                            }
                        }

                        targetRelon.idb_delete=true;

                        //ADD the target relon to the localRunDelete array.
                        //we batch delete these at the end of this run.
                        //Also, finally destroy it out of the master relons array.
                        if (targetRelon.ghost==false) {
                            MEM._localRunDelete.push(targetRelon.txid); 
                        }  
                    }
                }
            }

            //no other active=false relons get processed any further.
            else if (rObj.active==false) {
                delete rObj.from_nidTNOBJ;
                delete rObj.relnTNOBJ;
                delete rObj.to_sTNOBJ;
                continue;
            } 

            //FINAL PREP WORK
            //super mojo here..
            //some PARAM NAMES are BEs and handled hardcoded.
            //we don't want dynamic props there to interfere.
            //these, however, are dynamic and can be UDF on-chain..
            //to tune the code with available nobj and relon props.
            //IF TYPE_SYS and the RELNOBJ has a PARAM_NAME set (pval)
            //set the fromNOBJ[pval] = rObj.to_s
            //IF TYPE_SYS and the RELNOBJ has a RELON_PARAM_NAME set (pval)
            //set the rObj[pval] = rObj.to_s
            else if (rObj.type==TYPE_SYS) {
                setCodeProps(rObj);  
            }

            //give me a be array (first gen).
            else if (rObj.type==TYPE_BE) {   
                if (rObj.to_type==TO_TYPE_ANNE) {
                    
                    setCodeProps(rObj)

                    if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                    if (rObj.from_nidTNOBJ==null) {
                        console.warn("NEVER SHOULD HAPPEN BUT WITH NEW MEM OPTIMIZERS --- INVESTIGATE");
                        rObj.from_nidTNOBJ = MEMG.getNeuronByNid(rObj.from_nid);
                    }

                    rObj.from_nidTNOBJ.addUniqueBe(rObj);

                    //set identity flag..
                    //this is a strong differentiator in ANNE.
                    //identities are users.. all other neurons are not.
                    rObj.from_nidTNOBJ.bISIDENTITY = MEM.tempBeIsA(rObj.from_nidTNOBJ.temp_nid, MEM.identity.temp_nid, NTYPE_INST);
                }

                //we store both aka and birthname as an aka.
                //this allows us more info than just an aka but the functionality.
                else if (rObj.r==MEM.aka.temp_nid || rObj.r==MEM.birthname.temp_nid) {
                    if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                    rObj.from_nidTNOBJ.addAka(rObj);
                }

                //we didnt cover the TO_TYPE_ANNE cases
                //we waited so we could check for PARAM.
                //Dupe param names WILL NOT WORK and I'm not going 
                //to worry about it.
                else {
                    if (rObj.relnTNOBJ==null) rObj.relnTNOBJ = MEMG.getNeuronByTempNid(rObj.r); //iso_..
                    if (rObj.relnTNOBJ==null || rObj.relnTNOBJ.PARAM_NAME==undefined) {
                        delete rObj.from_nidTNOBJ;
                        delete rObj.relnTNOBJ;
                        delete rObj.to_sTNOBJ;
                        continue;
                    }
                    if (rObj.relnTNOBJ.PARAM_NAME!="text" && rObj.relnTNOBJ.PARAM_NAME!="name") {
                        if(rObj.relnTNOBJ.PARAM_NAME=="_company_type") {
                            console.log("test company type getting set from BE use of company type BE");
                        }
                        //this value is an attribute but no necessarily the PK or PK2.
                        if (rObj.from_nidTNOBJ==null) {
                            //console.log("why?");
                            rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                        }
                        rObj.from_nidTNOBJ[rObj.relnTNOBJ.PARAM_NAME]=rObj.to_s;

                        //if this reln is the pk or pk2 of the parent of the from
                        if (rObj.from_nidTNOBJ.bes.length==0) {
                            delete rObj.from_nidTNOBJ;
                            delete rObj.relnTNOBJ;
                            delete rObj.to_sTNOBJ;
                            continue;
                        }

                        //this won't get creative "BES".. (BE_CHILDREN defined by relon other than parent)
                        //we can get it in a buildAndInjectMod module load if needed in a specific app.
                        //in MEM we do LANG and COUNTRY as examples of how to build dynamic MODULES.
                        var beNOBJ = MEMG.getNeuronByTempNid(rObj.from_nidTNOBJ.bes[0]);
                        if (beNOBJ==null || beNOBJ.pk==undefined) {
                            delete rObj.from_nidTNOBJ;
                            delete rObj.relnTNOBJ;
                            delete rObj.to_sTNOBJ;
                            continue;
                        }
                        //we could put pk_val = or pk2_val = on this nObj.
                        if (rObj.relnTNOBJ.temp_nid==beNOBJ.pk) {
                            rObj.from_nidTNOBJ["pk_val"] = rObj.to_s;
                            rObj.from_nidTNOBJ["pk_name"] = rObj.relnTNOBJ.PARAM_NAME;
                        }                            
                    }
                }
            }

            //TYPE HAS....
            else if (rObj.type==TYPE_HAS) {
                if (rObj.from_nidTNOBJ==null) rObj.from_nidTNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (rObj.from_nidTNOBJ==null) {
                    console.log("%c YUGE PROBLEM fix this", TALK.bgpink);
                    delete rObj.from_nidTNOBJ;
                    delete rObj.relnTNOBJ;
                    delete rObj.to_sTNOBJ;
                    continue;
                }

                if (rObj.from_nidTNOBJ.type==NTYPE_INST) {
                    if(rObj.to_type==TO_TYPE_STRING) {
                        if (rObj.relnTNOBJ==null) rObj.relnTNOBJ = MEMG.getNeuronByTempNid(rObj.r);
                        if (rObj.relnTNOBJ.text.length) {
                            rObj.from_nidTNOBJ[rObj.relnTNOBJ.text.replace(" ","_")]=rObj.to_s;
                        }
                    }
                    else if (rObj.to_type==TO_TYPE_ANNE) {
                        if (rObj.relnTNOBJ==null) rObj.relnTNOBJ = MEMG.getNeuronByTempNid(rObj.r);
                        if (rObj.relnTNOBJ.text.length) {
                            if (rObj.to_sTNOBJ==null) rObj.to_sTNOBJ = MEMG.getNeuronByTempNid(rObj.t);
                            if (rObj.to_sTNOBJ!=null) {
                                rObj.from_nidTNOBJ[relnNOBJ.text.replace(" ","_")]=rObj.to_sTNOBJ.text;
                            }
                        }
                    }
                }
            }

            delete rObj.from_nidTNOBJ;
            delete rObj.relnTNOBJ;
            delete rObj.to_sTNOBJ;
        }

        //MEM._MOD_FUNC_CALLER(ANNE.MEM_UDF, '_udf_handleBrainPostOp');
    }




    //***************** SUPORTING PROCESSING FUNCTIOINS *****************//


    /** 
     * @author {Mr.Scatman} 
     */
     __deleteRelonHash=function($oldtxid, $txid) {
        if ($oldtxid!=$txid) {
            var key = $oldtxid;
            if (key.length == 0) return;

            //safe to delete the old index from the hash table,
            //NEVER DELETE IT FROM THE RELONS TBL
            if (MEM._aRelons_hashtbl[key]!=undefined) {
                var rel = MEM._aRelons[ MEM._aRelons_hashtbl[key] ];
                rel.dead = true;
                delete MEM._aRelons_hashtbl[key]; 
            }
        }
    }



    /** 
     * @author {Mr.Scatman} 
     */
     __hashNeuron = function($nObj){
        if ($nObj==null || $nObj.nid==undefined) { return; }
        
        var index = $nObj.nid;// Math.abs(hash);
        if (MEM._aNeurons_hashtbl[index]!=undefined) {
             //already have this  neuron hash.
             return;
        }
                
        //hash tables provides index to temp_nid.
        else {
            var temp_nid = MEM._aNeurons.length;
            MEM._aNeurons_hashtbl[index]=temp_nid;
            $nObj.temp_nid = temp_nid;
            MEM._aNeurons.push($nObj);
        }
    }


    /**
     * @author {Mr.Scatman}
     */    
     __hashRelon = function($rObj){
        if ($rObj==null) {
            console.log("%c ((YOU HAVE NO ROBJ TO HASH", 'background: white; color: red');
            console.log($rObj);
            return null;
        }
        var key = $rObj.txid;
        if (key==undefined) {
            console.log("%c ((YOU HAVE NO txid on that robj", 'background: white; color: red');
            return null;
        }

        if (key.length == 0) return;

        var hti = MEM._aRelons_hashtbl[key];
        if (hti!=undefined) {
            if (MEM._aRelons[hti].txid==$rObj.txid) {
                //console.warn("fix this -- the relon should have been flagged as already found", $rObj);
                return; //same relon.
            }
            console.log("%c ((has collision:"+$rObj.txid+" vs "+MEM._aRelons[hti].txid+"", 'background: white; color: red');
        }

        //hash tables provides index to rid.
        //don't want the large objects in the hash object.
        else {
            var rid = MEM._aRelons.length;
            MEM._aRelons_hashtbl[key]=rid;
            $rObj.rid = rid;
            MEM._aRelons.push($rObj);
            return;
        }
    }


    /**
     * @author {Mr.Scatman}
     */
     __hashRelonDelete=function($txid) {
        var rid = MEM._aRelons_hashtbl[$txid];
        if (rid==undefined) return null;
        else delete MEM._aRelons_hashtbl[rid];
        return;
    }


    /**
     * Update a temp txid to a real one.
     * @author {Mr.Scatman}
     */
    __hashRelonUpdate=function($rObj_existing, $rObj_new) {
        var rid = -1;
        var oldtxid = -1;
        if ($rObj_new==undefined || $rObj_new==null) return;

        if ($rObj_existing!=null && $rObj_existing!=undefined) {
            rid = $rObj_existing.rid;           
            oldtxid = $rObj_existing.txid;
        }

        //there is no existing, just do a new
        else {
            MEMP.__hashRelon($rObj_new);
            return;
        }

        //we don't have an existing 
        if (rid==undefined || rid==-1) {  
            if ($rObj_new==null) {
                console.log("%c ((YOU HAVE NO new ROBJ TO HASH", 'background: white; color: red');
                return;
            }
            else if ($rObj_new.txid==undefined) {
                console.log("%c ((YOU HAVE NO txid on this robj", 'background: white; color: red');
                console.log($rObj_new);
                return;
            }
            
            MEMP.__hashRelon($rObj_new);    //just hash the new and get out.          
            return;
        }

        //the rid is indexes in to the master relons array already, 
        //we need to set a new hashtbl entry (for the new real txid) to THIS rid.
        else {
            var hash = 0;
            var key = $rObj_new.txid;
            if (key.length == 0) return hash;
            MEM._aRelons_hashtbl[key] = rid;
            $rObj_existing.txid = key;
            MEMP.__deleteRelonHash(oldtxid, key);
        }

        return;
    }


    /**
     * Update the new relon in mem.
     * 
     * We created a temp relon in local memory before we sent a TX.
     * Here we want to take the temp txid and update it with the real TXID.
     * We don't require waiting for processing the new relon to do this. 
     * @param {object} - The existing relon we used to send via the API.
     * @param {string} - The new transaction ID of the actual broadcast relon.
     * 
     * @author {Mr.Scatman} 
     */
    __hashRelonUpdate_AFTER_SEND=function($rObj_existing, $txid="") {
        var rid = -1;
        var oldtxid = -1;
        if ($txid==undefined || $txid==null || $txid=="") {
            console.log("%c ((__hashRelonUpdate_AFTER_SEND) YOU HAVE NO new txid", 'background: white; color: red');
            return;
        }

        if ($rObj_existing!=null && $rObj_existing!=undefined) {
            
            //relon already handled.
            if ($rObj_existing.old_txid!=undefined && $rObj_existing.old_txid.length) return;

            rid = $rObj_existing.rid;           
            oldtxid = $rObj_existing.txid;
            if (oldtxid==undefined) return;
            if (oldtxid[0]!="z") return;    //txid MUST be a z txid here
        }

        //there is no existing, 
        //this relon was never in the hash table.
        //just update it with the new TXID and hash it now.
        else {
            console.log("%c ((__hashRelonUpdate_AFTER_SEND) no existing relon passed in", 'background: white; color: red');
            return;
        }

        $rObj_existing.temp=false;
        $rObj_existing.emulated=false;
        $rObj_existing.ghost=false;
        
        //we don't have an existing
        //just do it new here.
        if (rid==undefined || rid==-1) { 
            $rObj_existing.txid=$txid;
            //console.log("hashing anew");
            MEMP.__hashRelon($rObj_existing);    //just hash the new and get out.     
        }

        //the rid is indexes in to the master relons array already, 
        //we need to set a new hashtbl entry (for the new real txid) to THIS rid.
        else {
            var key = $txid;
            MEM._aRelons_hashtbl[key] = rid;
            $rObj_existing.txid = $txid; 
            $rObj_existing.old_txid = oldtxid.substring(0); //force copy;
            MEMP.__deleteRelonHash(oldtxid, $txid);    
        }
        return;
    }


    /** 
     * Kill a temp relon.
     * 
     *  
     * @author {Mr.Scatman}
     */
     killTempRelon=function($nObj, $rObj) {
        if ($nObj==null || $nObj.outs==undefined || $nObj.outs.length==0) return "";
        var temp_tx_id = "";
        if ($rObj==null) return null;

        for (var r=0; r < $nObj.outs.length; r++) {
            var rObj = $nObj.outs[r];
            if ((rObj.temp==undefined || rObj.temp==false) && rObj.emulated==false) continue;

            //this relon matches another relon
            if (rObj.from_nid==$rObj.from_nid
                && rObj.type==$rObj.type 
                && rObj.reln==$rObj.reln 
                && rObj.to_type==$rObj.to_type
                && rObj.to_s==$rObj.to_s) {
                //console.warn("(killTempRelon) FOUND A .temp DUPE");

                //if the other relon is a temp, switch it to a real.
                //REMOVE the temp relon from the nObj.outs array.  The new one will already
                //have been pushed in to there.
                if ((rObj.temp==true || rObj.emulated==true)) {
                    temp_tx_id = rObj.txid;
                    MEMP.__hashRelonUpdate(rObj, $rObj);
                    if ($rObj.time!=undefined) rObj.time=$rObj.time;
                    else if ($rObj.firstseen!=undefined) rObj.firstseen=$rObj.time;
                    if (rObj.temp==true) rObj.temp=false;
                    if (rObj.emulated==true) rObj.emulated=false;
                    $nObj.outs.splice(r, 1);    //REMOVE THIS OUT.
                    return temp_tx_id;
                }
            }
        }
        return "";
    }




    /**
     * We run this ONCE to catch flags we want to set after login that we      
     * couldn't set because we weren't logged in yet.   
     * @author {Mr.Scatman}
     *    
     * NOTE: we really need to login before we load the memory...uug                    
     */
     __postLoginCleanup = async ()=> {
        if (ANNE.USER._userNObj.owns==null) ANNE.USER._userNObj.owns=[];
        var us = ANNE.USER._userNObj.nid;
        var owns = null;
        var my_sponsored = null;

        if (ANNE.USER._userNObj.owns==undefined || ANNE.USER._userNObj.owns==null) {
            ANNE.USER._userNObj.owns = [];
        }
        owns = ANNE.USER._userNObj.owns;

        if (ANNE.USER._userNObj.my_sponsored==undefined || ANNE.USER._userNObj.my_sponsored==null) {
            ANNE.USER._userNObj.my_sponsored = [];
        }
        my_sponsored = ANNE.USER._userNObj.my_sponsored 

        for (var n=0; n < MEM._aNeurons.length; n++) {
            var nObj = MEM._aNeurons[n];
            if (nObj.owned_by==us) {
                if (owns.indexOf(nObj.temp_nid)==-1) owns.push(nObj.temp_nid);
                nObj.mine = true;
            }
            else if (nObj.sponsored_by==us) {
                if (my_sponsored.indexOf(nObj.temp_nid)==-1) my_sponsored.push(nObj.temp_nid);
            }
        }

        //another post process loop. (lol)
        for (var r=0; r<MEM._aRelons.length; r++) {
            if (MEM._aRelons[r].relon_sponsor!=undefined && MEM._aRelons[r].relon_sponsor==us) {
                ANNE.USER._userNObj.addUsAsRelonSponsor(MEM._aRelons[r]);
            }

            //we have some kind of any rights to either this class or inst.
            //lets tally it in our andy_nids so we can more easily query for andy admin 
            if (MEM._aRelons[r].t==USER._userNObj.temp_nid) {
                if (MEM.delegation_reln_tnids.indexOf(MEM._aRelons[r].r)!=-1) {
                    USER._userNObj.addAndy(MEM._aRelons[r].from_nid);
                }
            }
        }

        //make sure we got our adopted in there (we skipped them in pre-load before we were logged in)
        //we only process our adopted array
        //same with remembers/forgetz (or any cheat props on the nObj)
        var outs = ANNE.USER._userNObj.outs;
        for (var r=0; r < outs.length; r++) {
            var rObj = outs[r];
            if (rObj.r==MEM.trusted_adopt.temp_nid) ANNE.USER._userNObj.addTrustedAdopt(rObj.t);
            if ((rObj.r==MEM.rememberz.temp_nid) && rObj.temp==false) {
                var toNOBJ = MEMG.getNeuronByTempNid(rObj.t);
                toNOBJ.rememberz = rObj;
            }
        }

        //clean up the user's view for relons they've inhibited that are of their own.
        if (ANNE.USER._userNObj.inhibit_txs!=undefined) {
            var inh = ANNE.USER._userNObj.inhibit_txs;
            for (var i=0; i < inh.length; i++) {
                var txid = inh[i];
                var rObj = MEMG.lookupRelonByTXID(txid);
                if (rObj==null) continue;
                var fromNObj = MEMG.getNeuronByTempNid(rObj.f);
                
                //if we own the neuron that has the inhibited relon, we can deactivate it.
                if (fromNObj.mine==true) rObj.active = false; 
            }
        }

        if (ANNE.USER._userNObj.ap!=undefined) {
            var ap = ANNE.USER._userNObj.ap;
            for (var i=0; i < ap.length; i++) {
                var txid = ap[i];
                var rObj = MEMG.lookupRelonByTXID(txid);
                if (rObj==null) continue;


                var fromNObj = MEMG.getNeuronByTempNid(rObj.f);
                
                //we should have this deleted from memory.
                if (fromNObj.mine==true) {
                    rObj.active = false;     
                    await ANNE.DB.___delete([rObj]);
                }
            }
        }

    }
    

    /** 
     * Nuke a relon's data of the master relons array. 
     *
     * NEVER NEVER NEVER DELETE ITEMS FROM _aRelons.  It will shift all
     *  our INDEXES.  BIG BIG BIG NO NO. 
     * @author {Mr.Scatman} 
     */
    __removeByTXID_master = function ($txid) {
        var rel = ANNE.MEMG.lookupByTXID($txid);
        if (rel!=null) {
            rel.active=false;
            rel.from_nid="";
            rel.to_s="";
            rel.reln="";
            rel.outs = [];
            rel.ins = [];
            rel.txid="";
            return true;
        }
        for (var r=0; r < MEM._aRelons.length; r++) {
            if (MEM._aRelons[r].txid==$txid) {
                MEM._aRelons[r].active=false;
                MEM._aRelons[r].from_nid="";
                MEM._aRelons[r].to_s="";
                MEM._aRelons[r].reln="";
                MEM._aRelons[r].outs = [];
                MEM._aRelons[r].ins = [];
                MEM._aRelons[r].txid="";
                return true;
            }
        }
        return false;    
    }


    /** 
     * @author {Mr.Scatman} 
     */
    removeActiveWorldView = function ($worldviewNid) {
        var WV = parseInt($worldviewNid);
        var match = MEM._activeWorldViewNids.indexOf(WV);
        if (match!==-1) MEM._activeWorldViewNids.splice(match, 1);
    }



    /**
     * we've taken it out of outs/ins but it is still in the _aRelons.
     * We do not want to delete from that array as it will totally muck
     * up our .r/.t/.f indexes.. just set the obj vals to null and mark 
     * it as a non ghost, non temp, dead=true relon.
     */
    _resetRelon=function($relon){
        $relon.from_nid=null;
        $relon.f=null;
        $relon.to_s=null;
        $relon.reln=null;
        $relon.r=null;
        $relon.to_type=null;
        $relon.t=null;
        $relon.type=null;
        $relon.ghost=false;
        $relon.temp=false;
        $relon.dead=true;
    }


    /**  
     * @author {Mr.Scatman} 
     */
    _resetNeuron_RAM=function($nObj) {
        if ($nObj.outs && $nObj.outs.length && $nObj.bANNIHILATED==undefined) {
            console.warn("(resetNeuron) must annihilate outs first");
        }
        if ($nObj.ins && $nObj.ins.length) {
            console.warn("(resetNeuron) must annihilate ins first");
        }

        //remove this neuron from its owner's own array
        var owned_by = $nObj.owned_by;
        if (owned_by) {
            var ownerNObj = MEMG.getNeuronByNid(owned_by);
            if (ANNE.UTILS!=null) ANNE.UTILS.removeFromNidArray(ownerNObj.owns, $nObj.temp_nid);
            else alert("TO DO");
        }

        //remove this neuron from any children arrays.
        var bes = $nObj.bes;
        for (var b=0; b < bes.length; b++) {
            var beNObj = MEMG.getNeuronByTempNid(bes[b]);
            if (ANNE.UTILS!=null) ANNE.UTILS.removeFromNidArray(beNObj.children, $nObj.temp_nid);
            else alert("TO DO");
        }

        //DO NOT DELETE THIS ... it will be irrelevant on the next
        //load of the memory manager.
        //$nObj.temp_nid = -1;
        $nObj.dead=true;
        $nObj.text = "**dead**";
        $nObj.ntype = NTYPE_UNKNOWN;
        //delete MEM._aNeurons_hashtbl[$nObj.nid];
        //$nObj.nid = "";
        
        $nObj.ins = [];
        $nObj.outs = [];	
        $nObj.be = "";
        $nObj.bes = [];
        $nObj.ap = [];
        $nObj.necro = [];
        $nObj.owns = [];
        $nObj.emulated=false;	
        $nObj.owned_by = "";
        $nObj.mine = false;
        $nObj.children = [];
        $nObj.bTERMINATING_TYPE=false;
        $nObj.temp=false;
        $nObj._gt=null;
        $nObj._gti=null;
        $nObj.short="";
        $nObj.parent="";
        $nObj.aka=[];
        $nObj.ghost=false;
        $nObj.temp=false;
    }   


    /**
     * Complete nuke a neuron from existence.
     * 
     * This will delete all ins, outs and references to this from the MEM. (RAM)
     * Use the annihilated_relons array to nuke it out of DB mem.
     * 
     * @author {Mr.Scatman}  
     */
    annihilate_RAM = function ($nObj, $annihilated_relons=null, $bREMOVEINDEX=true) {
        
        if ($bREMOVEINDEX==true) MEM.removeTextIndex($nObj.text, $nObj.nid);   //this will kill the text index

        var beNOBJ = MEMG.getNeuronByTempNid($nObj.temp_nid);
        if ($nObj.freeze==1 && $nObj.ghost==false 
            && (USER._userNObj.nid!=MEM.ANNE_ADMIN_ACCTID || 
                USER._userNObj.nid!=MEM.ANNE_ADMIN_BACKUP_ACCTID)) return [];    //can't nuke a frozen neuron.
        
        if (beNOBJ!=null) {
            
            //remove all children of this neuron -- they are now orphans.
            MEMP.deadBeatDad(beNOBJ);
        }
        
        var annihilated_relons = $annihilated_relons;
        if (annihilated_relons==null) annihilated_relons = [];

        MEMP.annihilateOutRelons_RAM($nObj, annihilated_relons);
        MEMP.annihilateInRelons_RAM($nObj, annihilated_relons);
        MEMP.annihilateRelnRelons_RAM($nObj, annihilated_relons);

        $nObj.bANNIHILATED = true;
        MEMP._resetNeuron_RAM($nObj);

        //this is an array of dead relon objects with the .txid.
        //we can use it to remove data from DB mem as that only uses the .txid index.
        return annihilated_relons;

    }

    
    /**
     * 
     * @param {*} $rObj 
     */
    localOnlyArborPrune_RAM=function($rObj) {
        console.log("localOnlyArborPrune_RAM", $rObj);
        var annihilated_txids = [];
        MEMP.annihilateRelon_RAM($rObj, annihilated_txids);
        console.log("localOnlyArborPrune_RAM "+annihilated_txids.length+" has relons");
    }



    /**
     *  Nuke all ins
     * 
     *  This is only used on nObj arrays, safe to do as-is. 
     * @param {object} - the neuron
     * @param {array} - store the nuked relons in the array for downstream DB handling.
     * @author {Mr.Scatman}  
     *
     */   
     annihilateInRelons_RAM = function ($nObj, $annihilated_relons=[]) {
        if ($nObj.ins==undefined || $nObj.ins.length==0) return;

        //loop em all till they are gone.
        while ($nObj.ins.length) {
            var targetRelon = $nObj.ins[0];
            var done = $nObj.insMAP.removeU(targetRelon.txid);
            if (done==true) {
                if ($annihilated_relons.indexOf(targetRelon.txid)!=-1) {
                    $annihilated_relons.push(targetRelon.txid);
                }
            }
            if (targetRelon.from_nid) {
                var fNOBJ = MEMG.getNeuronByNid(targetRelon.from_nid);
                fNOBJ.outsMAP.removeU(targetRelon.txid);
            }  
        }
        if ($nObj.ins.length>0) {
            console.log("%c still ins leftover?  FIX THIS MRSCATMAN NOW!", TALK.bgredbig);
        }
    }


    /**
     *  Nuke all spots a neuron is used as the reln.
     * 
     * @param {object} - the neuron
     * @param {array} - store the nuked relons in the array for downstream DB handling.
     * @author {Mr.Scatman}  
     *
     */   
    annihilateRelnRelons_RAM = function ($nObj, $annihilated_relons=[]) {
        for (var r=0; r < MEM._aRelons.length; r++) {
            if (MEM._aRelons[r].r==$nObj.temp_nid) {
                $annihilated_relons.push(MEM._aRelons[r].txid);
            }
        }
    }


    /**
     *  Nuke all outs
     * 
     *  This is only used on nObj arrays, safe to do as-is. 
     * @param {object} - the neuron
     * @param {array} - store the nuked relons in the array for downstream DB handling.
     * @author {Mr.Scatman}  
     *
     */   
    annihilateOutRelons_RAM = function ($nObj, $annihilated_relons=[]) {
        var arr = $nObj['outs'];
        if (arr==undefined || arr.length==0) return;

        //we want to nuke a neuron's outs.
        for (var r=0; r < $nObj.outs.length; r++) {
            var targetRelon = $nObj.outs[r];
            var done = $nObj.outsMAP.removeU(targetRelon.txid);
            if (done==true) {
                $annihilated_relons.push(targetRelon.txid);
                if (targetRelon.to_s && targetRelon.to_type == TO_TYPE_ANNE) {
                    var tObj = MEMG.getNeuronByNid(targetRelon.to_s);
                    if (tObj==null) continue;
                    //don't need to track this one, we already got it.
                    //this is just removing it from the tObj "ins" array.
                    //tObj.removeByTXID('ins', targetRelon.txid, []);
                    tObj.insMAP.removeU(targetRelon.txid); 
                }  

                r = r-1;
                continue;
            }
        }
    }


    /**
     *  Annihilate a Relon.
     * 
     *  Remove a relon from the master relon array.                      
     *  destroy it from any other associated arrays.  (from, to) etc.     
     * @param {object} - the neuron
     * @param {array} - store the nuked relons in the array for downstream DB handling.
     * @author {Mr.Scatman}  
     *
     */     
    annihilateRelon_RAM = function($targetRelon, $annihilated_txids=[]) {
        if ($targetRelon.pending_from!=undefined) return true;
        if ($targetRelon.pending_to_s!=undefined) return true;
        if ($targetRelon.pending_txid!=undefined) return true;

        //remove it from its from's outs
        var fNObj = MEMG.getNeuronByNid($targetRelon.from_nid);      
        if (fNObj!=null) {
            fNObj.removeByTXID('outs', $targetRelon.txid, $annihilated_txids);
        }

        //remove it from its to's ins
        var tNObj = MEMG.getNeuronByTempNid($targetRelon.t);      
        if (tNObj!=null) {
            tNObj.removeByTXID('ins', $targetRelon.txid, $annihilated_txids);
        }

        this._resetRelon($targetRelon);

        return true;
    }
    

    /**
     * 
     * @param {*} $nids 
     * @returns 
     */
    necrosisNeuronsRAM=async($nids)=>{
        for (var n=0; n < $nids.length; n++) {
            var nOBJ = MEMG.getNeuronByNid($nids[n]);
            if (nOBJ==null) continue;
            MEMP.annihilate_RAM(nOBJ);
        }
        return;
    }


    /**
     *  Parent has went bye bye.
     *  Loop all children and cleanup their in-RAM-MEM connecton to the parent.
     *
     * @author {Mr.Scatman}  
     *
     */
    deadBeatDad = function($parentNOBJ, $bGRANNYTAKESOVER=false) {
        for (var c=0; c<$parentNOBJ.children.length; c++) {
            var childNOBJ = ANNE.MEMG.getNeuronByTempNid($parentNOBJ.children[c]);
            if (childNOBJ==null) continue;
            childNOBJ.removeParentFromChild ($parentNOBJ.temp_nid, $bGRANNYTAKESOVER);
        }
    }


    /**
     *  We marked the sand.  Now we reprocess from where we marked.
     *     
     * @param {int} - the nid mark we had before we did a special local load.
     * @author {Mr.Scatman}  
     *
     */ 
    reprocessRelonsFrom=async($TEMP_NID_SAVE)=> {
        var aRelons = MEM._aRelons.slice($TEMP_NID_SAVE);
        return await MEMP.processNeuronsFromRelons(aRelons, false, false, "relons", "MEMP.reprocessRelonsFrom");
    }

    

    /**
     * Clean up run-time ptrs and indexes that may have gotten into the local DB.
     * 
     * relons stored in local browser db need to be cleaned up a bit before 
     * they are re-run through the local memory processor.  
     *  
     * @author {Mr.Scatman}                 
     */
     santizeSavedRelonsForReprocessing=async($aRelons)=> {
        var fixIts = [];
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            delete rObj['f'];
            delete rObj['t'];
            delete rObj['r'];
            delete rObj['active'];
            delete rObj['temp'];
            delete rObj['outs'];
            delete rObj['ins'];
            if (rObj.ghost!=undefined && rObj.ghost==true ||
                rObj.ghost_owner!=undefined || rObj.ghost_order!=undefined) {
                delete rObj['ghost'];
                delete rObj['ghost_owner'];
                delete rObj['ghost_order'];
                fixIts.push(rObj);
            }
            if (rObj['gui_extras']!=undefined) delete rObj['gui_extras'];
        }

        if (fixIts.length) {
            console.log(fixIts);
            console.log("%c (santizeSavedRelonsForReprocessing) SELF FIX --- BUT WE SHOULD NEVER HAVE TO DO THIS!!!!", TALK.bgredbig);
            for (var f=0; f < fixIts.length; f++) { await DB.___update(fixIts[f]); }
        }
    }
    


    /**
     *  
     * @author {Mr.Scatman}                 
     */
    sanitizeSRFR_Ghosts=function($aRelons) {
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            delete rObj['f'];
            delete rObj['t'];
            delete rObj['r'];
            delete rObj['active'];
            delete rObj['temp'];
            delete rObj['outs'];
            delete rObj['ins'];
        }
    }
}

var MEMP = new mem_processor();
//freeze this


/************** (from file ./js/mem/mem_processors.js)**/ 


/** 
 * ez functions for doing things clean and fast
 */

 class mem_ez extends anne_singleton {
    constructor () {
        super(false);
        this.MEM = null;
    }
    init = function($MEM) {
        this.MEM=$MEM;
    }


    getConstituentLine=function($nObj, $line=[], $depth=0) {
        var line = $line;
        if ($depth > 10) return line;

        for (var r=0; r < $nObj.outs.length; r++) {
            if ($nObj.outs[r].to_type != TO_TYPE_ANNE) continue;
            if ($nObj.outs[r].r != ANNE.MEM.constituent_of.temp_nid) continue;
            if (line.indexOf($nObj.nid)!=-1) {
                console.log("ut oh.");
                return line;
            }
            line.push($nObj.nid);
            var cNOBJ = MEMG.getNeuronByTempNid($nObj.outs[r].t);
            return this.getConstituentLine(cNOBJ, line, ($depth+1));
        }

        return line;
    }


    /**
     * 
     * @param {*} $fromNOBJ 
     * @param {*} $reln 
     * @param {*} $test_loc_to_s 
     * @returns 
     */
    isBestLocation=function($fromNOBJ, $reln, $test_loc_to_s=null) {    
        if ($test_loc_to_s==null) {
            ANNE.consoleError("must have (isBestLocation) $test_loc_to_s here");
            return false;
        }
        
        var tests = [];

        for (var r=0; r < $fromNOBJ.outs.length; r++) {
            if ($fromNOBJ.outs[r].reln!=$reln) continue;
            if ($fromNOBJ.outs[r].to_type!=TO_TYPE_ANNE) continue;
            tests.push($fromNOBJ.outs[r]);
        }

        if (tests.length==0) {
            return true;
        }
        
        var testLOC = MEMG.getNeuronByNid($test_loc_to_s);
        console.log("test for new best location: %c"+testLOC.text, ANNE.TALK.bggreen);
        
        var constituent_line = this.getConstituentLine(testLOC);
        //ANNE.TALK.printNids(constituent_line, "candidate for new best location")

        for (var t=0; t < tests.length; t++) {
            var toNOBJ = MEMG.getNeuronByTempNid(tests[t].t);

            //We have this exact location.
            if (toNOBJ.nid==$test_loc_to_s) {
                return true;    
            }
            
            //our new neuron is better info than the existing... as 
            //our existing location is in the constituent line of our new neuron.
            if (constituent_line.indexOf(toNOBJ.nid)!=-1) {
                console.log("new line is the best, but we already have an inferior one we should be pruning.");
                return true;
            }

            //if the neuron we have already includes the test_loc_to_s line,
            //then it is not better.
            var line = this.getConstituentLine(toNOBJ);
            ANNE.TALK.printNids(line, "existing location line")
            if (line.indexOf($test_loc_to_s)!=-1) {
                console.log("existing line is more refined");
                return false;
            }
        }

        return true;
    }


        
    /********** EZ query/search pattern based ez functions **********/
    /*** ALL EZ FUNCTIONS only use the .r, .f, .t, .owned_by, .sponsored_by
        "from.be", and "to.bes" dennotations.
    ***/

    /**
     * Get nObjs of all the instance neurons for this class.
     * These are children that instantiated this class.
     * @param {Object} - The neuron
     */
    ez_getInst_nObjs = function($nObj) {
        if ($nObj.children.length==0) return [];
        var nObjs = [];
        for (var c=0; c< $nObj.children.length; c++) {
            var childNOBJ = MEMG.getNeuronByTempNid($nObj.children[c]);
            if (childNOBJ.type==NTYPE_INST) nObjs.push(childNOBJ);   
        }
        return nObjs;
    }


    /**
     * Get all nObjs that match in the inbound pattern.
     * return nObjs matching the pat. 
     * @param {Object} - The neuron
     * @param {Object} - The pattern 
     */
    ez_getIns_nObjs = function($nObj, $pat) {
        var relons = MEMZ.ez_getIns($nObj, $pat);
        var nObjs = [];
        if (relons.length) {
            nObjs = MEM.extractFromNObjsFromRelons(relons, false, false); 
            console.log(nObjs);
        }
        return nObjs;
    }


    /**
     * Get all relons that match in the inbound pattern.
     * return relons matching the pat. 
     * 
     * 
     * @param {Object} - The neuron
     * @param {Object} - The pattern
     * 
     * @author {Mr.Scatman}
     */
    ez_getIns = function($nObj, $pat) {
        var ins = $nObj.ins;
        var relons = [];
        for (var i=0; i < ins.length; i++) {
            var rObj = ins[i];
            if ($pat.type!=undefined && rObj.type!=$pat.type) continue;
            if ($pat.r!=undefined && rObj.r!=$pat.r) continue;
            
            var fNOBJ = null;
            if ($pat.owned_by!=undefined && $pat.owned_by!="") {
                fNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (fNOBJ.owned_by!=$pat.owned_by) continue;
            }
            if ($pat['from.be']!=undefined) {
                var from_be = $pat['from.be'];
                if (fNOBJ==null) fNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (MEM.besOf(fNOBJ, from_be)==false) continue;
            }
            relons.push(rObj);
        }
        return relons;
    }


    //this gets all children of a class..
    //matches for a class..
    //then verfies a supporting relon exists.
    //used in getGlobalLocation to get locations,
    //match children,
    //then confirm the location is a constituent of a certain neuron.

    //a better pattern would be: search for the constituent parent..
    //(narrows results)
    //then search for constiuent ofs which match the text. (fewer checks)
    ez_testGraphPat1=function($be_temp_nid, $textMatch, $restrictR, $restrictTo, $to_type=TO_TYPE_ANNE, $bDONOBJS=true) {
        var pats = [];
        var pat1 = {type:"neuron", subtype:"getAllChildren", be:$be_temp_nid};
        pats.push(pat1);
        var pat2 = {type:"neuron", subtype:"textMatch", casesensitive:false, partials:false, target:$textMatch, nObjs:[], temp_nids:[]};
        pats.push(pat2);
        var pat3 = {type:"relon", subtype:"relnTo", reln_temp_nid:$restrictR, to:$restrictTo, to_type:TO_TYPE_ANNE, bDONOBJS:$bDONOBJS, nObjs:[], temp_nids:[]};
        pats.push(pat3);
        return MEMZ.ez_getNeuronByGraphPat(pats);
    }

    ez_testGraphPat2=function($be_temp_nid, $restrictR, $restrictToTnid, $to_type=TO_TYPE_ANNE, $bDONOBJS=true) {
        var pats = [];
        var pat1 = {type:"neuron", subtype:"getAllChildren", be:$be_temp_nid, nObjs:[], temp_nids:[]};
        pats.push(pat1);
        var pat3 = {type:"relon", subtype:"relnTo", reln_temp_nid:$restrictR, to:$restrictToTnid, to_type:TO_TYPE_ANNE, bDONOBJS:$bDONOBJS, nObjs:[], temp_nids:[]};
        pats.push(pat3);
        return MEMZ.ez_getNeuronByGraphPat(pats);
    }


    //resolves multi relons to a single concept.
    //can be made robust to allow for optional values to match
    //(or not to match)
    //or values to not contradict (but not have to be present)
    ez_multiRelResolver=function($be_temp_nid, $relToPartials=[]) {
        var pats = [];
        var pr = $relToPartials[0];
        var pat1 = {type:"relon", subtype:"relnToNid", be:$be_temp_nid, reln_temp_nid:pr.r, to:pr.t, to_type:TO_TYPE_ANNE, nObjs:[], temp_nids:[] };
        pats.push(pat1);
        for (var r=1; r < $relToPartials.length; r++) {
            var pr = $relToPartials[r];
            var patx = {type:"relon", subtype:"relnTo", reln_temp_nid:pr.r, to:pr.t, to_type:TO_TYPE_ANNE, bDONOBJS:true, nObjs:[], temp_nids:[]};
            pats.push(patx);
        }
        return MEMZ.ez_getNeuronByGraphPat(pats);
    }

    //WORK IN PROGRESS..
    //a new implementation of the concept of searching by pat..
    //This is a baby step towards a full ACQL implementation.
    ez_getNeuronByGraphPat = function($pats) {
        var temp_nids = [];
        var active_nids = [];
        var last_pat = null;
        for (var p=0; p < $pats.length; p++) {
            var pat = $pats[p];
            if (pat.type==undefined) {
                console.log("MALFORMED PAT. Must have type.");
                return;
            }
            //search for neurons
            if (pat.type=="neuron") {
                if (pat.subtype=="getAllChildren") {
                    var be = pat.be;
                    pat.temp_nids = MEMG.getNeuronByTempNid(be).children;
                    last_pat = pat;
                }
                else if (pat.subtype=="textMatch") {
                    if (pat.partials!=undefined && pat.partials==true) {}
                    else {
                        var target = "";
                        if (pat.casesensitive==undefined && pat.casesensitive==true) {
                            target = pat.target;
                        }
                        else target = pat.target.toUpperCase();
                        
                        pat.temp_nids = [];

                        //optimize by looping nObjs instead, if we have them.
                        if (last_pat.nObjs!=undefined) {
                            for (var n=0; n < last_pat.nObjs.length; n++) {
                                var nObj = last_pat.nObjs[n];
                                if (nObj.text.toUpperCase()==target) {
                                    pat.temp_nids.push(nObj.temp_nid);
                                }
                            }
                        }
                        else {
                            for (var n=0; n < last_pat.temp_nids.length; n++) {
                                var nObj = MEMG.getNeuronByTempNid(last_pat.temp_nids[n]);
                                if (nObj.text.toUpperCase()==target) {
                                    pat.temp_nids.push(nObj.temp_nid);
                                }
                            }
                        }
                        last_pat = pat;
                    }
                }
            }

            //seach for relons filters
            else {
                if (pat.subtype=="relnToNid"){
                    var reln_temp_nid = pat.reln_temp_nid;
                    var to = pat.to;
                    var to_type = pat.to_type;
                    var be = pat.be;
                    if (be==undefined) be = null;
                    var bDONOBJS = false;
                    if (pat.bDONOBJS==true) bDONOBJS=true;
                    pat.temp_nids = [];
                    var toNOBJ = MEMG.getNeuronByTempNid(to);
                    var ins = toNOBJ.ins;
                    for (var r=0; r < ins.length; r++) {
                        if (ins[r].r!=reln_temp_nid) continue;
                        if (ins[r].to_type!=to_type) continue;
                        var nObj = MEMG.getNeuronByTempNid(ins[r].f);
                        if (be!=null) {
                            if(nObj.isT(be)==false) continue;
                        }
                        if (bDONOBJS==true) {
                            pat.nObjs.push(nObj);
                        }
                        pat.temp_nids.push(nObj.temp_nid);
                    }
                    last_pat = pat;
                }
                else if (pat.subtype=="relnTo") {
                    var reln_temp_nid = pat.reln_temp_nid;
                    var to = pat.to;
                    var to_type = pat.to_type;
                    var bDONOBJS = false;
                    if (pat.bDONOBJS==true) bDONOBJS=true;
                    pat.temp_nids = [];
                    if (bDONOBJS==true) pat.nObjs = [];
                    for (var n=0; n < last_pat.temp_nids.length; n++) {
                        var nObj = MEMG.getNeuronByTempNid(last_pat.temp_nids[n]);
                        for (var r=0; r < nObj.outs.length; r++) {
                            var rObj = nObj.outs[r];
                            if (rObj.to_type!=to_type) continue;
                            if (rObj.r!=reln_temp_nid) continue;
                            if (to_type==TO_TYPE_ANNE) {
                                if(rObj.t!=to) continue;
                            }
                            else if (to!=rObj.to_s) continue;
                            if (bDONOBJS==true) {
                                pat.nObjs.push(nObj);
                            }
                            pat.temp_nids.push(nObj.temp_nid);
                        }
                    }
                    last_pat = pat;
                }
            }
        }

        return last_pat;
    }



    /**
    * @author {Mr.Scatman}
    */
    ez_setNARRPtrsForNew=function($rObj) {
        if ($rObj==null) return;
        if (govenor.isCollabRespReln($rObj.r)==true && $rObj.type==TYPE_SYS) {
            return; //we can do this AFTER the send and confirmation of receipt.
        }

        var nObj = MEMG.getNeuronByTempNid($rObj.f);
        
        if (nObj==null) return;
        if (nObj.parent!="") $rObj.parent = nObj.parent;
        if (nObj.owned_by!="") $rObj.owned_by = nObj.owned_by;
        if (nObj.sponsored_by!="") $rObj.sponsored_by = nObj.sponsored_by;
        //if (nObj.annother!="") $rObj.annother = nObj.annother;
        $rObj.ntype = nObj.ntype;
        if (nObj.text!=null && nObj.text && $rObj.type==TYPE_BE) $rObj.text = nObj.text;
        
    }


    /**
    * @author {Mr.Scatman}
    */
    ez_new_param=async($fromNOBJ, $reln, $to_type, $to_s, $bSINGLESEND=true, $gui_extras=null, $callback=null, $ghostModeOveride=false)=> {
        if (ANNE._txMode==TXMODE_GHOST || $ghostModeOveride==true) {
            if ($fromNOBJ==null) {
                console.log("we need a neuron (temp or real) here");
                return null;
            }
            var rObj = MEM.makeTempRelon($fromNOBJ, TYPE_PARAM, $reln, $to_type, $to_s, true);
            if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
            var done = await ANNE.GHOSTDATA.localGhostTheRelons([rObj]);
            return rObj;
        }
        else {
            //we are in ANNE TX mode but this neuron is not real yet.
            //It needs some help.
            if ($fromNOBJ==null || $fromNOBJ.temp==true) {
                await SENDQ.pendingFromRelon(TYPE_PARAM, $reln, $to_type, $to_s, TXB.TXTYPE_CHILD_RELON, $gui_extras, $bSINGLESEND, $callback);
            }

            else {
                var rObj = MEM.makeTempRelon($fromNOBJ, TYPE_PARAM, $reln, $to_type, $to_s, true);
                if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
                var TXTTYPE = ANNE.TXB.calcTXTTYPE(rObj);    
                await SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, TXTTYPE, $callback);
            }
        }
    }



    /**
     * This does not create a clump, it creates a TYPE_CLUMP relon pointing to one.
     * @author {Mr.Scatman}
     */
    ez_new_clump=async($fromNOBJ, $reln, $to_type, $to_s, $bSINGLESEND=true, $gui_extras=null, $callback=null, $ghostModeOveride=false)=> {
        if (ANNE._txMode==TXMODE_GHOST || $ghostModeOveride==true) {
            if ($fromNOBJ==null) {
                console.log("we need a neuron (temp or real) here");
                return null;
            }
            var rObj = MEM.makeTempRelon($fromNOBJ, TYPE_HAS, $reln, $to_type, $to_s, true);
            if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
            var done = await ANNE.GHOSTDATA.localGhostTheRelons([rObj]);
            return rObj;
        }
        
        else {
            //we are in ANNE TX mode but this neuron is not real yet.
            //It needs some help.
            if ($fromNOBJ==null || $fromNOBJ.temp==true) {
                await SENDQ.pendingFromRelon(TYPE_CLUMP, $reln, $to_type, $to_s, TXB.TXTYPE_CHILD_RELON, $gui_extras, $bSINGLESEND, $callback);
            }

            else {
                var rObj = MEM.makeTempRelon($fromNOBJ, TYPE_CLUMP, $reln, $to_type, $to_s, true);
                if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
                var TXTTYPE = ANNE.TXB.calcTXTTYPE(rObj);    
                await SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, TXTTYPE, $callback);
            }
        }
        return rObj;
    }


    /**
     * no ghost for ez new sys.  Just real relons.
     * @author {Mr.Scatman}
     */
    ez_new_sys_msg = async($fromNOBJ, $reln, $to_type, $to_s, $gui_extras=null, $callback=null)=> {
        if ($reln==null) return null;
        if ($to_s==null) return null;
        if ($fromNOBJ==null) {
            console.log("we need a neuron (temp or real) here");
            return null;
        }
        if (typeof $fromNOBJ === "string") {
            $fromNOBJ = MEMG.getNeuronByNid($fromNOBJ);
            if ($fromNOBJ instanceof anne_neuron == false) {
                console.warn("bad data in fromNOBJ, needs to be an anne_neuron");
                return null;
            }
        }
        var rObj = MEM.makeTempRelon($fromNOBJ, TYPE_SYS_MSG, $reln, $to_type, $to_s, true);
        if (rObj==null) {
            console.warn("skipping relon, rObj from makeTemprelon is null");
            return null;
        }
        var bDUPEOK = false;

        if ($gui_extras!=null) {
            rObj.gui_extras=$gui_extras;
            if ($gui_extras.dupe_ok!=undefined && $gui_extras.dupe_ok==true) {
                bDUPEOK = true;
            }
            if ($gui_extras.payment!=undefined) rObj.annies = $gui_extras.payment;
            if ($gui_extras.sys_ref_txid!=undefined) rObj.sys_ref_txid = $gui_extras.sys_ref_txid;
        }
        ANNE.TALK.printRelon(rObj);
        var TXTTYPE = ANNE.TXB.calcTXTTYPE(rObj);    
        var cmdObj = await SENDQ.addRelonToSendQ(rObj, false, TXTTYPE, $callback, bDUPEOK);
        if ($gui_extras!=null && $gui_extras.annies_offset!=undefined) cmdObj.annies += $gui_extras.annies_offset; //adjust.
        return rObj;
    }



    /**
     * Send an awareness edge.
     * 
     * Awareness edges will only ever be sent from the identity.  They cannot be collab'd on
     * as they are temporal representations of live run-time feelz and opinions.
     * 
     * FEELZ are meant to capture what is "felt".  
     * They are used to indicate how the target (relon.to_s) makes them feel.
     * IT IS NOT HOW THEY FEEL ABOUT IT.  That linguistic pattern is often 
     * short-hand for how one thinks/believes.  We want FEELZ to focus on HOW THEY MAKE YOU FEELZ.
     * 
     * Opinionz are held.  They indicate what one thinks/believes about the target.
     * 
     * We have semantic short-hand relons to capture these in a single edge without doing a full
     * semantic clump with roles.
     * 
     * "That guy is a jerk" would be an OPINIONZ. USER-> (opinionz:jerk) -> That guy.
     * If you read an article and feelz HATE because of it, that's a USER-> (feelz:HATE)->the article.
     * 
     * Note: technically, the full clump for "that is a jerk" would be:
     * NEW CLUMP.
     * NEW CLUMP-> role:who -> that guy (nid of the guy)
     * NEW CLUMP-> role:be -> is/be
     * NEW CLUMP-> role:state -> jerk
     * 
     * @author {Mr.Scatman}
     */
    ez_new_awareness=async($reln, $to_type, $to_s, $bSINGLESEND=true, $gui_extras=null, $callback)=> {
        //if (ANNE.AWARE==null) {
            //alert("you must include the awareness.js to use awareness edges");
        //}

        if ($to_type!=TO_TYPE_ANNE && $to_type!=TO_TYPE_TXID) {
            alert("awareness edges must be to_type BITCOIN or TXID");
            return;
        }

        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_AWARENESS, $reln, $to_type, $to_s, true);
        if ($gui_extras!=null) rObj.gui_extras=$gui_extras;

        if (ANNE._txMode==TXMODE_GHOST) {
            var done = await ANNE.GHOSTDATA.localGhostTheRelons([rObj]);
            return rObj;
        }
        else {
            var tempCmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, ANNE.TXB.TXTYPE_SELF_RELON, $callback);
        }

        return rObj;
    }




    /************************ EZ GUI STUFF ********************/
    
    /** 
     * @author {Mr.Scatman}
     */
     ezDisplayfrom_tnidList = function($el, $lbl, $list, $icoclass="", $linkType=LINKS_NEURON_TBL_VIEW) {
        if ($list==null) return;
        var links=[];
        for(var l=0; l<$list.length; l++) {
            var nObj = ANNE.MEMG.getNeuronByTempNid($list[l]);
            var link = null;//LINKS.makeNeuronLink(nObj, $linkType);
            link = "FIXTHIS ezDisplayFrom_tnidList";
            links.push(link);
        }

        var hrefs="";
        if (links.length > 1) hrefs = $lbl+" "+links.join(",");
        else hrefs = $lbl+" "+links[0];

        ANNE.A.ezIcoSpan($el, "solo", $icoclass, hrefs);
    }



    /** 
     * @author {Mr.Scatman}
     */
    ezOwnerSpan=function($nObj,$divid,$lbl="owned by",$linkType=LINKS_NEURON_USER) {
        if($nObj.owned_by=="") return;
        var obNOBJ=ANNE.MEMG.getOrMakeIfNew($nObj.owned_by);
        var link = "FIXTHIS ezOwnerSpan";//LINKS.makeNeuronLink(obNOBJ, $linkType);
        ANNE.A.ezlblSpan($divid,$lbl,link,"solo");
    }


    /** 
     * @author {Mr.Scatman}
     */
    ez_propDisplaySpan=function($parentdivid,$lbl,$htmlVal,$sep=": ") {
        var parent=ANNE.A.__getEl($parentdivid);
        if(parent==null) return;
        var span=document.createElement("span");
        span.style.display="inline";
        var innerspanhtml="<span style='display:inline; font-weight:bold'>"+$htmlVal+"</span>"
        span.innerHTML=$lbl+$sep+innerspanhtml;
        parent.append(span);
    }

    /**
     * @author {Mr.Scatman}
     */
    ezBesLinkPath=function($el, $nObj, $bINCLUDECLASS=false) {
        if ($nObj == null) return;
        if ($nObj.bes.length == 0) return;
        var safetyfirsteverybody = 0;
        var bes = [];
        while ($nObj.bes.length) {
            bes.push($nObj.bes[0]);
            $nObj = ANNE.MEMG.getNeuronByTempNid($nObj.bes[0]);
            if ($bINCLUDECLASS==false && $nObj.temp_nid == MEM.class.temp_nid) break;
            if (safetyfirsteverybody++ > 20) break;
        }
        var revbes = bes.reverse();
        ANNE.MEMZ.linkify_neurons($el, revbes);
        return;
    }
    
    linkify_neurons = function($el, $tnids, $delim="&nbsp;/&nbsp;", $max=20, $linkclass="", $nukeclass=false) {
        for (var b=0; b < $tnids.length; b++) {
            var nObj = ANNE.MEMG.getNeuronByTempNid($tnids[b]);
            var linkel = ANNE.EXT_BTNS.makeNeuronLink($el, nObj.nid);
            
            if ($nukeclass && $linkclass!="") linkel.classList = "";

            if (linkel!=null && $linkclass!="") linkel.classList.add($linkclass);
            var slash = document.createElement("span");
            slash.style='display:inline-block';
            slash.innerHTML = $delim;
            $el.append(slash);
            if (b>=$max) {
                slash.innerHTML = slash.innerHTML+" ...";
                return;
            }
        }
    }


    /**
     * @author {Mr.Scatman}
     */
    ezBesCdxLinkPathWSchema=function($el, $nObj, $bINCLUDECLASS=false, $linkType=LINKS_NEURON_ANNET) {
        var el=ANNE.A.__getEl($el);
        if (el == null) return null;
        if ($nObj == null) return;
        if ($nObj.bes.length == 0) return;
        var nObj = $nObj;
        var p_bes = [];
        var safetyfirsteverybody = 0;
        while ($nObj.bes.length) {
            $nObj = MEMG.getNeuronByTempNid($nObj.bes[0]);
            if ($bINCLUDECLASS==false && $nObj.temp_nid == MEM.class.temp_nid) break;
            //var link = LINKS.makeNeuronLink($nObj, $linkType);
            var link = "FIXTHIS ezBesCdxLinkPathWSchema";
            p_bes.push(link);
            if (safetyfirsteverybody++>20) break;
        }
        p_bes.reverse();
        var html="";
        if (nObj.ntype=="class") {
            if (p_bes.length>1) html="subclass of "+p_bes.join("->");
            else if (p_bes.length==1) html="subclass of "+p_bes[0];
        }
        else if (p_bes.length) {
            if (p_bes.length>1) html="instance of "+p_bes.join("->");
            else if(p_bes.length==1) html="instance of "+p_bes[0];
        }
        ANNE.A.setHTML(el,html,true);
    }


    /**
     * @author {Mr.Scatman}
     */
    ezBesToLinksSpan=function($nObj,$divid,$lbl="Bes") {
        if ($nObj==null) return;
        var bes = $nObj.bes;
        if (bes.length==0) return;
        var links = [];
        for(var b=0; b<bes.length; b++) {
            var nObj = MEMG.getNeuronByTempNid(bes[b]);
            //var link = LINKS.makeNeuronLink(nObj, LINKS_NEURON_TBL_VIEW);
            var link = "FIXTHIS ezBesToLinksSpan";
            if (b==0) link = "<b>"+link+"</b>";
            links.push(link);
        }
        ANNE.A.ezlblSpan($divid, $lbl, links.join(", "));
    }

    /**
     * @author {Mr.Scatman}
     */
    ez_neuronSchemaSpan_set=function($divid,$type) {
        var el = ANNE.A.__getEl($divid);
        if (el == null) {
            console.log($divid+" doesn't exist in the DOM");
            return;
        }
        el.classList.add("ai-"+$type);
        el.setAttribute("title", "this is a "+$type+" neuron");
    }

    /**
     * @author {Mr.Scatman}
     */
    ez_neuronSchemaSpan_make=function($divid,$type,$html,$classAdd) {
        var el = ANNE.A.__getEl($divid);
        if(el == null) {
            console.log($divid+" doesn't exist in the DOM");
            return;
        }
        el.innerHTML = $html;
        var span = document.createElement("span");
        span.classList.add("ai-"+$type);
        if( $classAdd!="") span.classList.add($classAdd);
        span.setAttribute("title","this is a "+$type+" neuron");
        el.append(span);
    }
    
    /**
     * @author {Mr.Scatman}
     */
    ez_relon_checks=function($relons,$uniqueid="jj",$pre_msg) {
        var div=document.createElement("div");
        div.setAttribute("id","ezchks_"+$uniqueid);
        if ($pre_msg) {
            var span=document.createElement("span");
            span.innerHTML = $pre_msg;
            div.append(span);
            div.append(br);
        }
        for (var r=0; r<$relons.length; r++) {
            var relon = $relons[r];
            var talkmsg=ANNE.TALK.humanReadableRelon(relon,false);
            var checked=true;
            var disabled=true;
            if (relon.type == TYPE_HAS) disabled=false;
            A.ezCheckMark(div, "chk_"+$uniqueid+relon.txid, relon.txid, talkmsg, checked, disabled);
            var br=document.createElement("br");
            div.append(br);
        }
        div.setAttribute("id","ezchks_"+$uniqueid);
        return div;
    }



    /**
     * @author {Mr.Scatman}
     */
    ez_propDisplayRow=function($parentdivid,$lbl,$htmlVal) {
        var parent=ANNE.A.__getEl($parentdivid);
        if(parent==null) return;
        var row=document.createElement("div");
        row.classList.add("row");
        var span=document.createElement("span");
        row.append(span);
        span.style.display="inline";
        var innerspanhtml="<span style='display:inline; font-weight:bold'>"+$htmlVal+"</span>"
        span.innerHTML=$lbl+": "+innerspanhtml;
        parent.append(row);
    }


    /**
     * @author {Mr.Scatman}
     */
    ez_propsDisplayRow=function($parentdivid, $objs) {
        var parent=ANNE.A.__getEl($parentdivid);
        if(parent==null) return;
        var row=document.createElement("div");
        row.classList.add("row");
        var span=document.createElement("span");
        row.append(span);
        span.style.display="inline";
        var html="";
        for(var o=0;o<$objs.length;o++) {
            var innerspanhtml="<span style='display:inline; font-weight:bold'>"+$objs[o].value+"</span>"
            html=$objs[o].name+": "+innerspanhtml;
        }

        var innerspanhtml="<span style='display:inline; font-weight:bold'>"+html+"</span>"
        span.innerHTML=html;
        parent.append(row);
    }


    /**
     * Ez display props row
     * 
     * Intended for use with array of json objsct with .name and .value
     * 
     * @params {string} - div to render the results in
     * @params {array} - array of objects with prop .name and .value
     * @author {Mr.Scatman}
     * 
     */ 
    ez_propsDisplayRow2 = function($parentdivid, $objs) {
        var parent = ANNE.A.__getEl($parentdivid);
        if (parent==null) return;
        var row = document.createElement("div");
        row.classList.add("row");
        var span = document.createElement("span");
        row.append(span);
        span.style.display="inline";
        var html = "";
        for (var o=0; o < $objs.length; o++) {
            var innerspanhtml = "<span style='font-weight:bold'>"+$objs[o].value+"</span>"
            html += "<div padding-left:4px;'>"+$objs[o].name+": "+innerspanhtml+"</div>";
        }
        span.innerHTML = html;
        parent.append(row);
    }


    /**
     * @author {Mr.Scatman}
     */ 
    ez_img = function($el,$ref,$w=32,$h=0,$imgClassAdd="") {
        $el.append(this.html_img($ref,$w,$h,$imgClassAdd));
    }

    html_img=function($ref,$w=32,$h=0,$imgClassAdd="") {
        if($w&&$h) return '<img src="'+$ref+'" alt="" id="'+$imgElName+'" class="img-thumbnail'+$imgClassAdd+'" height="'+$h+'" width="'+$w+'">';
        else if($h) return '<img src="'+$ref+'" alt=""  class="img-thumbnail'+$imgClassAdd+'" height="'+$h+'">';
        else if($w) return '<img src="'+$ref+'" alt="" class="img-thumbnail'+$imgClassAdd+'" width="'+$w+'">';
    }

    /**
     * @author {Mr.Scatman}
     */ 
    ez_delete_relon_span = function($el,$txid) {
        var span=document.createElement("span");
        $el.append(span);
        var deleteIt='<a href="javascript:relon_table.deleteThisRelon(\''+$txid+'\')"><span style="color:red"><i class="far fa-times-circle"></i></span></a>';
        span.innerHTML=deleteIt;
        return {span: span,deleteIt: deleteIt};
    }


    /**
     * @author {Mr.Scatman}
     */ 
    ezNeuronGarden=async($el, $nObj, $linkType=LINKS_NEURON_USER, $bDOLETDENY=true)=> {
        if ($nObj.bISIDENTITY!=null && $nObj.bISIDENTITY==true) return;
        var el = ANNE.A.__getEl($el);
        if (el==null) return null;
        var srcgarden = document.createElement("div");
        srcgarden.classList = "src-garden";
        srcgarden.id = "garden_"+$nObj.temp_nid;
        el.append(srcgarden);
        await MEMZ.ezDSGardenInfo(srcgarden, $nObj, $linkType, $bDOLETDENY);
    }


    /**
     * @author {Mr.Scatman}
     */ 
    ezNeuronSignal = function($el, $nObj, $linkType=LINKS_NEURON_USER, $guid="") {
        if ($nObj._napl!=undefined && $nObj._napl!=null && $nObj._napl.sell_status!=0) {
            $nObj._napl.sell_status_txt = MEMG.getNeuronByNid($nObj._napl.sell_status).text;
            var el = ANNE.A.__getEl($el);
            if (el==null) return null;
            var srcsignal = document.createElement("div");
            srcsignal.classList = "src-sell-signal";
            $el.append(srcsignal);
            //var link = LINKS.makeNeuronLink($nObj, $linkType, {text:$nObj._napl.sell_status_txt});
            var link = "FIXTHIS ezNeuronsignal";
            srcsignal.innerHTML="<b>sell signal</b>&nbsp;"+link;
            return srcsignal;
        }
        else if ($nObj._napl) $nObj._napl.sell_status_txt = "none";

        return null;
    }


    /**
     * @author {Mr.Scatman}
     */ 
    ezNeuronSponsor = function($el, $nObj, $linkType=LINKS_NEURON_USER, $guid="") {
        if ($nObj.sponsored_by!=undefined && $nObj.sponsored_by!=null && $nObj.sponsored_by!="") {
            var el = ANNE.A.__getEl($el);
            if (el==null) return null;
            var srcsponsoredby = document.createElement("span");
            srcsponsoredby.classList = "src-sponsoredBy";
            $el.append(srcsponsoredby);
            var sponsoredByNOBJ = MEMG.getOrMakeIfNew($nObj.sponsored_by);
            EXT_BTNS.makeNeuronLink(srcsponsoredby, sponsoredByNOBJ.nid, "", "sponsor&nbsp;");
            srcsponsoredby.title = "sponsor";
            if (sponsoredByNOBJ.nid==USER._userNObj.nid) srcsponsoredby.append(ANNE.A.span("(you!)", "text-smallest-you"));
            return srcsponsoredby;
        }
        return null;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $instChildren 
     * @returns 
     */
    ezInstChildren = function($nObj, $instChildren=[]) {
        for (var c=0; c  < $nObj.children.length; c++) {
            var childNObj = MEMG.getNeuronByTempNid($nObj.children[c]);
            if (childNObj.ntype==NTYPE_CLASS) {
                this.ezInstChildren(childNObj, $instChildren);
            }
            else if (childNObj.ntype==NTYPE_INST) $instChildren.push(childNObj);
        }
        return $instChildren;
    }


    /**
     * @author {Mr.Scatman}
     */ 
    ezKnownChildren=function($el, $nObj) {
        if ($nObj.ntype==NTYPE_INST) return null;

        //get normal children.
        var subclassesTnids = [];
        var instTnids = [];
        for (var c=0; c  < $nObj.children.length; c++) {
            var childNObj = MEMG.getNeuronByTempNid($nObj.children[c]);
            if (childNObj.ntype==NTYPE_CLASS) subclassesTnids.push(childNObj.temp_nid);
            else if (childNObj.ntype==NTYPE_INST) instTnids.push(childNObj.temp_nid);
        }

        //functional children (instances of another class but still a child)
        if ($nObj.functional_childparent_parent!=undefined && 
            $nObj.functional_childparent_parent!=null) {
            var childrenNOBJS = ANNE.MEM.getChildrenDynamic($nObj);
            for (var c=0; c  < childrenNOBJS.length; c++) {
                var childNObj = childrenNOBJS[c];
                if (childNObj.ntype==NTYPE_CLASS) subclassesTnids.push(childNObj.temp_nid);
                else if (childNObj.ntype==NTYPE_INST) instTnids.push(childNObj.temp_nid);
            }
        }

        if ($el!=null) {
            var el_known_children = null;
            if (subclassesTnids.length || instTnids.length) {
                el_known_children = document.createElement("div");
                el_known_children.classList = "known-children";
                $el.append(el_known_children);
            }
    
            if (subclassesTnids.length) {
                var ellbl2 = document.createElement("div");
                ellbl2.innerHTML="<b>known subclasses</b>";
                ellbl2.display="flex";
                ellbl2.classList = "known-subclasses-title";
                
                var el = ANNE.A.ezEl(el_known_children, "div", "known-subclasses");
                el.classList.add("dib");
                el.appendChild(ellbl2);
                MEMZ.headerlist(el, subclassesTnids, 20, "nobgunderline", true);
            }
            
            if (instTnids.length) {
                var ellbl3 = document.createElement("div");
                ellbl3.innerHTML="<b>known insts</b>";
                ellbl3.display="flex";
                ellbl3.classList = "known-insts-title";
                
                var el = ANNE.A.ezEl(el_known_children, "div", "known-insts");
                el.classList.add("dib");
                el.appendChild(ellbl3);

                MEMZ.headerlist(el, instTnids, 20, "nobgunderline", true);
            }
            return el_known_children;
        }

        return instTnids;
    }

    headerlist = function($el, $tnids,  $max=20, $linkclass="", $nukeclass=false) {
        for (var b=0; b < $tnids.length; b++) {
            var nObj = ANNE.MEMG.getNeuronByTempNid($tnids[b]);
            var linkel = ANNE.EXT_BTNS.makeNeuronLink($el, nObj.nid);
            if ($nukeclass && $linkclass!="") linkel.classList = "";
            if (linkel!=null && $linkclass!="") linkel.classList.add($linkclass);
            if (b>=$max) return;
        }
    }



    /**
     * @author {Mr.Scatman}
     */ 
    ezNeuronOwner = function($el, $nObj, $linkType=LINKS_NEURON_USER, $guid="", $accolor="black") {
        if ($nObj.bISIDENTITY!=null && $nObj.bISIDENTITY==true)  return;    //no owner displays for identity
        var el = ANNE.A.__getEl($el);
        if (el==null) return null;
        var srcOwnedBy = document.createElement("span");
        srcOwnedBy.classList = "src-ownedby";
        el.append(srcOwnedBy);
        if ($nObj.owned_by!="") {
            var owned_byNOBJ = MEMG.getOrMakeIfNew($nObj.owned_by);
            EXT_BTNS.makeNeuronLink(srcOwnedBy, owned_byNOBJ.nid, "", "<b>owner</b>&nbsp;", false, "");
            if (owned_byNOBJ.nid==USER._userNObj.nid) {
                srcOwnedBy.append(ANNE.A.span("(you!)", "text-smallest-you"));
            }
            else if (ANNE.isANNE_display(owned_byNOBJ.nid)==true) {
                if ($accolor=="black")  srcOwnedBy.append(ANNE.A.span("&nbsp;"+ANNE.ico_black, "text-smallest"));
                else  srcOwnedBy.append(ANNE.A.span("&nbsp;"+ANNE.ico, "text-smallest"));
            }
            else {
                srcOwnedBy.append(ANNE.A.span("&nbsp;<span class='tiny-text bg-danger'>non-anne owned</span>", "text-smallest"));
                console.log("owned_by:"+$nObj.owned_by);
            }
        }
        else srcOwnedBy.innerHTML = 'no owner on record';
        return srcOwnedBy;
    }
    

    ezVIPLET = function($el, $nObj, $linkType=LINKS_NEURON_TBL_VIEW) {
        var el = ANNE.A.__getEl($el);
        if (el==null) return null;
        if ($nObj._vip_let==null || $nObj._vip_let.length==0) return;
        var srcLets = document.createElement("div");
        srcLets.classList = "vip-let";
        el.append(srcLets);
        srcLets.innerHTML = "<b>vip let</b> "+LINKS.linkifyTnids($nObj._vip_let);
    }

    

    /**
     * $linkType is no longer relevant in the ANNODE APP..
     * all linkTypes launch the dossier page for that neuron.
     * (in prev iterations it could launch different widgets)
     * @author {Mr.Scatman}
     */ 
    ezNeuronBes = function($el, $nObj, $linkType=LINKS_NEURON_TBL_VIEW, $bDOREALBE=true, $classPre="neuron") {
        var el = ANNE.A.__getEl($el);
        if (el==null) return null;
        var srcBes = document.createElement("div");
        srcBes.classList = $classPre+"-bes";
        el.append(srcBes);

        if ($nObj.bes.length) {
            var ico = "";
            var ntype = "";
            if ($nObj.ntype==NTYPE_CLASS) {
                if ($nObj.be_lvl==0) ntype = "class";
                else if ($nObj.be_lvl>=1) ntype = "subclass";
            }
            else if ($nObj.ntype==NTYPE_INST) ntype = "instance";
            else ntype = "unk if class or inst";

            if ($nObj && $nObj.ai_ico!=undefined && $nObj.ai_ico!="" && $nObj.ai_ico!=null) ico = '<i class="'+$nObj.ai_ico+'"></i>';
            var span = document.createElement("span");
            span.style.display="inline-block";
            span.innerHTML=ico+'<b>be</b>:&nbsp;<span class="ntype-'+$nObj.ntype+'">'+ntype+'</span>&nbsp;of&nbsp;';
            srcBes.append(span);

            //new ext frenly.
            MEMZ.ezBesLinkPath(srcBes, $nObj, $linkType);
        }
        else srcBes.innerHTML="parent not set";

        //TO DO: fix for EXT.. no inline js links
        if ($nObj.functional_childparent_reln!=null) {
            var cpr = MEMG.getNeuronByTempNid($nObj.functional_childparent_reln);
            if ($nObj.functional_childparent_parent!=null) {
                var pa = MEMG.getNeuronByTempNid($nObj.functional_childparent_parent);
                var realBe = document.createElement("div");
                realBe.classList = "neuron-real-be-msg";
                var msg = "<b>not instantiable<</b> "+pa.bestNameHTML();
                msg = msg + " with a "+cpr.bestNameHTML()+" relationship to "+$nObj.bestNameHTML();
                realBe.innerHTML = msg
                el.append(realBe);
            }
            
            else {
                var realBe = document.createElement("div");
                realBe.classList = $classPre+"-real-be-msg";
                var msg = "children have "+cpr.bestNameHTML()+" relationship to "+$nObj.bestNameHTML();
                realBe.innerHTML = msg
                el.append(realBe);
            }
        }
        
        //check if this is an alt child a definition on its heirarchical 
        //standard parent.
        //the standard parent would have a xx->BE_PARENT-> (parent)
        else {
            if ($nObj.functional_be!=null) {
                var pasNOBJ = MEMG.getNeuronByTempNid($nObj.functional_be);
                var realBe = document.createElement("div");
                realBe.classList = $classPre+"-real-be";
                realBe.innerHTML = "<b>functional be</b> ";
                EXT_BTNS.makeNeuronLink(realBe, pasNOBJ.nid);//, $linkType);
                el.append(realBe);
            }
            else {
                var potentialAlts = $nObj.getParentAlts();
                if(potentialAlts==undefined || potentialAlts==null) potentialAlts=[];
                for (var pas=0; pas < potentialAlts.length; pas++) {
                    var pasNOBJ = MEMG.getNeuronByTempNid(potentialAlts[pas]);
                    if (pasNOBJ==null) continue;
                    //console.log("psbl alt parent:"+pasNOBJ.text);
                    if (pasNOBJ.functional_childparent_reln!=null) {
                        //console.log("psbl alt parent has functional_childparent_reln listed:"+pasNOBJ.functional_childparent_reln);

                        //country->REAL_CHILLPARENT_RELN->polity
                        //look for this reln on THIS NEURON...
                        var trobj = MEMG.getRelonByTempR($nObj.outs, pasNOBJ.functional_childparent_reln);
                        
                        //if we have this relon..
                        if (trobj!=null && trobj.t==pasNOBJ.temp_nid) {
                            $nObj.functional_be = pasNOBJ.temp_nid;
                            var realBe = document.createElement("div");
                            realBe.classList = $classPre+"-real-be";
                            realBe.innerHTML = "<b>functional be<b> ";
                            EXT_BTNS.makeNeuronLink(realBe, pasNOBJ.nid);
                            el.append(realBe);
                        }
                    }
                }
            }
        }

        return srcBes;
    }


    /**
     * Display garden and collab infos for a neuron.
     * 
     * Note: For datasources you pass the lowest (final) DS.. 
     * it has the wv_let and wv_deny and also sideways-be-inherits the _gt of the
     * last datasource.
     * 
     * @author {Mr.Scatman}
     * 
     */
    ezDSGardenInfo = async($el, $nObj, $linkType=LINKS_NEURON_DOSSIER, $bDOLETDENY=true)=> {
        var el=ANNE.A.__getEl($el);
        if (el==null) return null;

        //if they haven't been set for the user for this neuron, set them.
        await $nObj.setGovRights(USER._userNObj);

        var span = ANNE.A.createElement(el, "span", "garden-msg");
        var gardenmsg = "default";
        if ($nObj.collab_ACTION==govenor.THIS_IS_YOU) {
            gardenmsg = "No Collab on identites"
            span.classList.add("gt-nocollab");
        }
        else if ($nObj.collab_ACTION==govenor.COLLAB_NOCOLLABIDENTITY) {
            gardenmsg = "No Collab on identites"
            span.classList.add("gt-nocollab");
        }

        if ($nObj.ntype==NTYPE_CLASS) {
            if ($nObj._gt!=null) {
                var gtNOBJ = MEMG.getNeuronByTempNid($nObj._gt.t);
                if (gtNOBJ!=null) {
                    gardenmsg = gtNOBJ.text;
                    if ($nObj._gt.t==MEM.gt_owner_only.temp_nid) {
                        span.classList.add("gt-owner-only");
                    }
                }
            }
        }
        else if ($nObj.ntype==NTYPE_INST) {
            if ($nObj._gti!=null) {
                var gtiNOBJ = MEMG.getNeuronByTempNid($nObj._gti.t);
                if (gtiNOBJ!=null) {
                    gardenmsg = gtiNOBJ.text;
                    if ($nObj._gti.t==MEM.gt_owner_only.temp_nid) {
                        span.classList.add("gt-owner-only");
                    }
                }
            }
        }

        var el_gt = ANNE.A.createElement(span, "span", "", gardenmsg);
        var ico = ANNE.A.createElement(el_gt, "img", "seed-ico");
        ico.src = "./imgs/seedling.png";
        ico.setAttribute("height", 20);
        ico.setAttribute("width", 20); 

        var collab_msg = await $nObj.getCollabStatusMsgHTML();
        var divx = document.createElement("div");
        divx.classList = "collab-status";
        divx.innerHTML = collab_msg;
        el.append(divx);

        if ($bDOLETDENY==false) return divx;

        //do the let/deny list.
        //this DISPLAYS the lists, it doesn't enforce them.
        if ($nObj._wv_let!=undefined && $nObj._wv_let.length) {
            if ($nObj._wv_let) MEMZ.ezDisplayfrom_tnidList($el, "<b>let</b>", $nObj._wv_let, "ai-check", $linkType);
        }
        if ($nObj._wv_deny!=undefined && $nObj._wv_deny.length) {
            if ($nObj._wv_deny) MEMZ.ezDisplayfrom_tnidList($el, "<b>deny</b>", $nObj._wv_deny, "ai-ban", $linkType);
        }
        return divx;
    }

        
    /**
     * 
     */
    genDisplayListOtherBes = function($nObj) {
        var outs = $nObj.outs;
        var other_bes = [];
        for (var r=0; r < outs.length; r++) {
            var rel = outs[r];
            if (rel.type==TYPE_BE && (rel.r!=MEM.class.temp_nid && rel.r!=MEM.inst.temp_nid && rel.r!=MEM.name.temp_nid)) {
                other_bes.push(rel);
            }
        }
        if (other_bes.length) {
            return this.genKeyPairDisplayListFromRelons(other_bes);
        }

        return "";
    }


    /**
     * 
     */
    genKeyPairDisplayListFromRelons=function($arr) {
        var kps = [];
        if  ($arr.length==0) return "";
        for (var r=0; r < $arr.length; r++) {
            var relon = $arr[r];
            var rNOBJ = MEMG.getNeuronByTempNid($arr[r].r);
            //var link = LINKS.makeNeuronLink(rNOBJ, LINKS_NEURON_ANNET);
            var link = "FIXTHIS genKeyPair";
            var reln_link = "";
            var toNOBJ = MEMG.getNeuronByTempNid(relon.t);
            if (toNOBJ && relon.type==TO_TYPE_ANNE) {
                //reln_link = LINKS.makeNeuronLink(toNOBJ, LINKS_NEURON_ANNET, {linkclass:"link-green"});
                reln_link = "FIXTHIS genKeyPairDisplayListFromRelons";
            }

            else reln_link = MEMG.__MASTER_DISPLAY_RELON_TO(relon);
            var final = link+":"+reln_link;
            kps.push(final);
        }
        return kps.join(" | ");
    }


}

var MEMZ = new mem_ez();
//freeze this ?

/************** (from file ./js/mem/mem_ez.js)**/ 

class search_obj {
    constructor($defaults) {
        if ($defaults.text!=undefined) this.text=$defaults.text;
        else this.text="";

        if ($defaults.original!=undefined) this.original=$defaults.original;
        else this.original="";

        if ($defaults.context!=undefined) this.context=$defaults.context;
        else this.context=[];

        if ($defaults.context!=undefined) this.context=$defaults.context;
        else this.context=[];

        if ($defaults.bes!=undefined) this.bes=$defaults.bes;
        else this.bes=[];

        if ($defaults.bes_tnids!=undefined) this.bes_tnids=$defaults.bes_tnids;
        else this.bes_tnids=[];

        if ($defaults.exact!=undefined) this.exact=$defaults.exact;
        else this.exact=true;

        if ($defaults.applywvtobeline!=undefined) this.applywvtobeline=$defaults.applywvtobeline;
        else this.applywvtobeline=true;

        if ($defaults.inbeline!=undefined) this.inbeline=$defaults.inbeline;
        else this.inbeline=true;

        if ($defaults.chaseIt!=undefined) this.chaseIt=$defaults.chaseIt;
        else this.chaseIt=true;

        if ($defaults.wvs!=undefined) this.wvs = $defaults.wvs;
        else this.wvs=[];

        this.req_extras = [];
        this.bEXTRAS_REQUIRED=false;
    }


    /**
     * Set a requirement of extra relons that match.
     * This will most often be an ALT BE relon specializing the standard/generic thing.
     * @param {*} $req_extras 
     * @param {*} $bREQUIRED 
     */
    setRequiredExtras=function($req_extras=[], $bREQUIRED=true) {
        this.req_extras=$req_extras;
        this.bEXTRAS_REQUIRED = $bREQUIRED;
    }


    /**
     * Some extra relon required to match..
     * most often an ALT lateral BE relon.
     * @param {*} $aMatches 
     * @returns 
     */
    filterExtras=function($aMatches) {
        if (this.req_extras.length>0) {
            var final_matches = [];
            for (var m=0; m < $aMatches.length; m++) {
                var match = $aMatches[m];
                var bHASALL=false;
                for (var e=0; e < this.req_extras.length; e++) {
                    var ex = this.req_extras[e];
                    var test_rel = ANNE.MEMG.getRelonFromWithTo(match.outs, ex.type, ex.reln, ex.to_s);
                    if (test_rel==null) {
                        bHASALL==false; //turn it off, exit
                        break;
                    }
                    bHASALL=true;   //set it on match of 1, meaning we looked
                }
                if (bHASALL==true) final_matches.push(match);
            }

            return final_matches;
        }
        return $aMatches;
    }


    /**
     * 
     * @param {*} $bDOPRE 
     * @param {*} $bACCEPTRAMONLY 
     * @returns 
     */
    search = async ($bDOPRE=false, $bACCEPTRAMONLY=true)=> {
        if ($bDOPRE==true) {
            var matches = await ANNE.SEARCH.searchNeuronText (this.text, this.bes_tnids, this.wvs, this.exact, this.applywvtobeline, this.inbeline); 
            if (this.chaseIt==false) {
                if (matches.length>0) {
                    if (this.bEXTRAS_REQUIRED==true) return this.filterExtras(matches);
                    else return matches;
                }
                return [];
            }
            if ($bACCEPTRAMONLY==true && matches.length>0) {
                if (this.bEXTRAS_REQUIRED==true) return this.filterExtras(matches);
                else return matches;
            }
        }
        if (this.chaseIt==true) {
            var done = await ANNE.API.queryForText (this.text, this.wv, this.bes, this.exact, this.applywvtobeline, this.inbeline);
            console.log(done);
        }

        var matches = await ANNE.SEARCH.searchNeuronText (this.text, this.bes_tnids, this.wvs, this.exact, this.applywvtobeline, this.inbeline);   
        if (this.bEXTRAS_REQUIRED==true) return this.filterExtras(matches);
        else return matches;
    }
    
}


class mem_search {
    constructor() {
        this.SS = null;
    }

    init=function($SS=null) { 
        this.SS = $SS; 
    }

    //ANNE.SEARCH.parseSearchForContext("tom cruise[person]")
    //ANNE.SEARCH.parseSearchForContext("red[movie]")
    //ANNE.SEARCH.parseSearchForContext("bank[action, flight]")
    parseSearchForContext = function($str) {
        var find = 0;
        var i_left = -1;
        i_left = $str.indexOf("(");
        if (i_left!==-1) find=")";
        else {
            i_left = $str.indexOf("[");
            if(i_left!=-1) find="]";
            else {
                i_left = $str.indexOf("{")
                if (i_left!=-1) find="}";
            }
        }

        if (find!="") {
            var i_right = $str.indexOf(find);
            if (i_right!=-1) {
                var extract = $str.substring(i_left+1, i_right).trim();
                console.log(extract);
                var str = $str.substring(0, i_left).trim()+$str.substring(i_right+1).trim();
                var rez= {search:str, inline_context:extract.split(",")};
                return rez;
            }
        }

        return {search:$str, inline_context:[]};
    }


    /**
     * Get the last n_items of the class and order by most recent.
     */
    getRecents=function($nids, $maxFireWeight=0.01, $max_results=200) {
        var downlineNOBJS = [];

        //recent INSTS (only)
        var bDOINSTS = true;
        var bTNIDLIST = false;
        var bDOALLBES = false;
        var bINSTSONLY = true;

        for (var n=0; n < $nids.length; n++) {
            var nOBJ = MEMG.getNeuronByNid($nids[n], $maxFireWeight);
            console.log("get recents for "+nOBJ.text);
            console.log("children test:"+nOBJ.children.length);
            nOBJ.getAllDownlineChildren(downlineNOBJS, 0, bDOINSTS, bTNIDLIST, bDOALLBES, bINSTSONLY);
        }
        downlineNOBJS.sort(UTILS.sortByBeTime);
        return downlineNOBJS.slice(0, Math.min(downlineNOBJS.length, $max_results));
    }



    /**
     * Search the text, case-insensitive, of all neurons in local memory. 
     * Returns neurons   
     *    
     * Rethink this for our new indexes.
     * Our .text comes in on relons but is not stored in the local DB.
     * We need to search neuron memmory for text.
     * 
     * @author {Mr.Scatman}
     */
    personSearch = function($textString, $bEXACT=true) {
        console.log($textString);
        if ($textString==undefined || $textString=="" || $textString==null) {
            console.log("fix this");
            return [];
        }
        var aMatches = [];
        var aMatchesNids = [];
        $textString=$textString+"";
        var lower = $textString.toLowerCase();

        var parts = lower.split(" ");//use cap
        var first = parts[0];
        
        for (var n=0; n < MEM._aNeurons.length; n++) {
            var nObj = MEM._aNeurons[n];
            
            if (nObj.bes[0]!=MEM.person.temp_nid) continue;
            if (nObj.owned_by!=MEM.ANNE_ACCTID && nObj.owned_by!=USER._userNObj.nid) continue;
            

            if ($bEXACT==false) {
                nObj.fire = UTILS.similarity($textString, nObj.text, 1000);
                if (nObj.fire < .8 && nObj.aka!=undefined && nObj.aka.length>0) {
                    for (var a=0; a < nObj.aka.length; a++) {
                        var fi = UTILS.similarity($textString, nObj.aka[a], 1000);
                        if (fi > nObj.fire) {
                            nObj.fire = fi;
                        }
                    }
                }

                //if we the target string is contained with this potential string, 
                //let's bump it up more.  (don't bother if already above .8)
                if (nObj.fire>=.8) {
                    aMatches.push(nObj)
                    aMatchesNids.push(nObj.temp_nid);
                }
                else if (nObj.fire>=.4) {
                    if (nObj.text.toLowerCase().indexOf(lower)!=-1) {
                        nObj.fire = Math.min(1, nObj.fire+(nObj.text.length*.05));
                    }
                    nObj.fire = nObj.fire - ((1-nObj.fire)*(1-nObj.fire)/2);
                    if (nObj.fire<=.4) continue;
                    nObj.fire = Math.round(nObj.fire*100)/100;
                    if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                    aMatchesNids.push(nObj.temp_nid);
                    aMatches.push(nObj);
                } 
            }

            
            else {
                if (nObj.text.toLowerCase() == lower) {
                    if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                    aMatchesNids.push(nObj.temp_nid);
                    aMatches.push(nObj);
                }

                //if the first name is right...
                else if (this.lower.indexOf(first)==0) {

                    //check the aka_lower array.
                    if (nObj.aka_lower && nObj.aka_lower.indexOf(lower)!=-1) {
                        if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                        aMatchesNids.push(nObj.temp_nid);
                        aMatches.push(nObj);
                    }         
                }
            }
        }

        return aMatches; 
    }



    /**
     * 
     * @param {*} $text 
     * @returns 
     */
    textSearchNaked = async($text)=> {
        return await ANNE.MEM.searchTextIndex($text);
    }


    searchNeuronTextSO = async($so)=>{
        return await this.searchNeuronText($so.text, $so.bes, $so.wv, $so.exact, $so.applywvtobeline);
    }

    /**
     * Search the text, case-insensitive, of all neurons in local memory. 
     * Returns neurons   
     *    
     * Rethink this for our new indexes.
     * Our .text comes in on relons but is not stored in the local DB.
     * We need to search neuron memmory for text.
     * 
     * @author {Mr.Scatman}
     */
    searchNeuronText = async($textString, $beTnids=[], $WVs=[], $bEXACT=false, $bAPPLYWVTOBELINEMATCH=false)=> {
        console.log($textString);
        if ($textString==undefined || $textString=="" || $textString==null) {
            return [];
        }
        
        if ($WVs.indexOf(USER._userNObj.nid)==-1) $WVs.push(USER._userNObj.nid);
        
        //test search.
        await ANNE.IDB.queryForText($textString, null, $bEXACT);

        var aMatches = [];
        var aMatchesNids = [];
        $textString=$textString+"";
        var lower = $textString.toLowerCase();

        for (var n=0; n < MEM._aNeurons.length; n++) {
            var nObj = MEM._aNeurons[n];

            //this is to restrictive and not considerate of BEMS.
            if ($beTnids.length && $beTnids.indexOf(nObj.bes[0])==-1) continue;
            
            if ($WVs.length && nObj.owned_by!="") {
                if ($bAPPLYWVTOBELINEMATCH==false && $WVs.indexOf(nObj.owned_by)==-1) continue;
                else if ($bAPPLYWVTOBELINEMATCH==true) {
                    var beNOBJ = MEMG.getNeuronByTempNid(nObj.bes[0]);
                    if (beNOBJ==null) continue;
                    if($WVs.indexOf(beNOBJ.owned_by)==-1) continue;
                }
            }


            //var test = false;
            if ($bEXACT==false) {

                //if char is set, it must match, exactly.
                if (nObj.char!=undefined) {
                    if ($textString.length==1) {
                        nObj.fire = UTILS.similarity($textString, nObj.text, 1000);
                    }
                    else continue;
                }
                else nObj.fire = UTILS.similarity($textString, nObj.text, 1000);
                
                if (nObj.fire < .8 && nObj.aka!=undefined && nObj.aka.length>0) {
                    for (var a=0; a < nObj.aka.length; a++) {
                        var fi = UTILS.similarity($textString, nObj.aka[a], 1000);
                        if (fi > nObj.fire) {
                            nObj.fire = fi;
                        }
                    }
                }

                //if we the target string is contained with this potential string, 
                //let's bump it up more.  (don't bother if already above .8)
                
                if (nObj.fire>=.8) {
                    aMatches.push(nObj)
                    aMatchesNids.push(nObj.temp_nid);
                }
                else if (nObj.fire>=.4) {
                    if (nObj.text.toLowerCase().indexOf(lower)!=-1) {
                        nObj.fire = Math.min(1, nObj.fire+(nObj.text.length*.05));
                    }
                    if(nObj.text.length>lower.length) {
                        if (nObj.text.toLowerCase().indexOf(lower)==0) {
                            if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                            aMatchesNids.push(nObj.temp_nid);
                            aMatches.push(nObj);
                        }
                    }
                    nObj.fire = nObj.fire - ((1-nObj.fire)*(1-nObj.fire)/2);
                    if (nObj.fire<=.4) continue;
                    nObj.fire = Math.round(nObj.fire*100)/100;
                    if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                    aMatchesNids.push(nObj.temp_nid);
                    aMatches.push(nObj);
                }
                else if (nObj.fire>.25 && nObj.text.length>lower.length && nObj.text.toLowerCase().indexOf(lower)==0) {
                    if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                    aMatchesNids.push(nObj.temp_nid);
                    aMatches.push(nObj);
                }
                
            }
            else if (nObj.text.toLowerCase() == lower) {
                if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                aMatchesNids.push(nObj.temp_nid);
                aMatches.push(nObj);
            }
            else if (nObj.char!=undefined && $textString.length==1 && $textString==nObj.char) {
                if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                aMatchesNids.push(nObj.temp_nid);
                aMatches.push(nObj);
            }
            else {
                //check the aka_lower array.
                if (nObj.aka_lower && nObj.aka_lower.indexOf(lower)!=-1) {
                    if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                    aMatchesNids.push(nObj.temp_nid);
                    aMatches.push(nObj);
                }         
            }
        }

        //look for exact TYPE_BE relons to non-neuron TO_TYPE_STRINGs.
        //this will get primary keys, aliases, and any other text string short hands.
        //we *SHOULD* make an effort to make a keyword neuron anytime a TYPE_BE
        //relon to a TO_TYPE_STRING is put into the ANNE worldview (on the annode)
        for (var r=0; r < MEM._aRelons.length; r++) {
            if (MEM._aRelons[r].to_type!=TO_TYPE_STRING) continue;
            if (MEM._aRelons[r].type!=TYPE_BE && MEM._aRelons[r].type!=TYPE_HAS) continue;
            if (MEM._aRelons[r].r==MEM.name.temp_nid) continue; //we already checked names
            if (MEM._aRelons[r].to_s.toLowerCase()==lower) {
                var nObj = MEMG.getNeuronByTempNid(MEM._aRelons[r].f);
                if (nObj==null) continue;

                //if this is a char it must be exact.  X != x.
                if (nObj.char!=undefined && nObj.char!=$textString) continue;
                
                if (aMatchesNids.indexOf(nObj.temp_nid)!=-1) continue;
                aMatchesNids.push(nObj.temp_nid);
                aMatches.push(nObj);
            }
        }

        return aMatches; 
    }



    /**
     * Search the text, case-insensitive, of all neurons in local memory. 
     * This is for an ajax like text entry search.  We must match the start of the word to get a partial match.
     * Returns neurons 
     * 
     * @author {Mr.Scatman} 
     */
    searchTextStart = function($textString, $typesToEnforce=[], $beNid=-1, $WVs=[]) {
        var aMatches=[];
        var beNOBJ = null;
        
        $textString = $textString.toLowerCase();

        if ($beNid>=0) beNOBJ = MEMG.getNeuronByTempNid($beNid);

        for (var n=0; n < MEM._aNeurons.length; n++) {
            var nObj = MEM._aNeurons[n];

            //if must match type, enforce first. (faster)
            if ($typesToEnforce.length && $typesToEnforce.indexOf(nObj.ntype)==-1) continue;
        
            if (nObj.text=="") continue;

            var text = nObj.text.toLowerCase();
            
            if (text.indexOf($textString)!=0) continue;  //only looking for matches to the start of a word here.
            
            //be class enforced  (note, this should be decendents, not just be)
            if (beNOBJ!=null && nObj.bes.indexOf(beNOBJ.temp_nid)==-1) continue;

            if ($WVs.length && $WVs.indexOf(nObj.owned_by)==-1) continue;
            aMatches.push(nObj);

            nObj.fire = UTILS.similarity($textString, nObj.text, 1000);
        }
        return aMatches; 
    }


    /**
     * Search the text, case-insensitive, of all neurons in local memory. 
     * Returns neurons 
     * 
     * @author {Mr.Scatman} 
     */
    searchText = function($textString, $typesToEnforce=[], $bEXACT=false, 
                                $bDOAKAS=true, $beNid=-1, $emulatedOk=false, $bDOWEIGHT=true, $WVs=[]) {
        var regex = new RegExp($textString, 'i');
        var aMatches=[];
        var beNOBJ = null;

        if ($beNid>=0) beNOBJ = MEMG.getNeuronByTempNid($beNid);

        for (var n=0; n < MEM._aNeurons.length; n++) {
            var nObj = MEM._aNeurons[n];
            
            //if must match type, enforce first. (faster)
            if ($typesToEnforce.length && $typesToEnforce.indexOf(nObj.ntype)==-1) continue;
            
            //watch this.. we may want emulated in some searches..
            if ($emulatedOk==false && nObj.emulated == true) continue;

            var text = nObj.text;
            if (nObj.text=="") continue;

            //match the text
            var test = false;
            if ($bEXACT==false) test = regex.test(text);
            else if (nObj.text == $textString) test = true; 
            
            if (test == false) {

                //quick aka check.
                if ($bDOAKAS==true && nObj.aka.indexOf(text)!=-1) {
                    if (beNOBJ!=null && nObj.bes.indexOf(beNOBJ.temp_nid)==-1) continue;
                    if ($WVs.length && $WVs.indexOf(nObj.owned_by)==-1) continue;
                    if ($bDOWEIGHT==true) nObj.fire = nObj.fire-1;
                    aMatches.push(nObj);
                    continue;
                }	

                continue;
            }

            //be class enforced  (note, this should be decendents, not just be)
            if (beNOBJ!=null && nObj.bes.indexOf(beNOBJ.temp_nid)==-1) continue;

            if ($WVs.length && $WVs.indexOf(nObj.owned_by)==-1) continue;
            aMatches.push(nObj);
            if ($bDOWEIGHT==true) {
                nObj.fire = UTILS.similarity($textString, nObj.text, 1000);
            }
        }
        return aMatches; 
    }

}

var SEARCH = new mem_search();

/************** (from file ./js/mem/mem_search.js)**/ 


/** 
 * Functions that get and lookup stuff. 
 * 
 * getaaa,  lookupaaa, 'aaa'Lookup, extract..
 * 
 * @author {Mr.Scatman} 
 */

class mem_gets extends anne_singleton {
    constructor () {
        super(false);
    }
    init = function($MEM) {
    }


    /**
     * Is this neuron ($nObj) in the be downline of the neuron with temp_nid $temp_nid?
     * 
     * @param {anne_neuron} - the neuron you are checking
     * @param {int} - the temp nid of the class you are trying to see if the neuron is within.
     * @author {Mr.Scatman}  
     */
    besOf = function ($nObj, $target_temp_nid_class) {
        if (typeof $target_temp_nid_class !=="number") {
            if ($target_temp_nid_class.temp_nid!=undefined) $target_temp_nid_class = $target_temp_nid_class.temp_nid;
            alert("fix your bad (besOf) caller! 2nd param should be a integer, not a neuron!");
            return false;
        }
        
        //quick hit direct parent class.
        if ($nObj.bes[0]==$target_temp_nid_class) return true;

        return MEMG.chaseUpLineForParentMatch($nObj, $target_temp_nid_class);
    }

    //aliases
    isBes = function($nObj, $target_temp_nid_class) { return MEMG.besOf($nObj, $target_temp_nid_class); }
    besIn = function($nObj, $target_temp_nid_class) { return MEMG.besOf($nObj, $target_temp_nid_class); }
    isBeA = function($nObj, $target_temp_nid_class) { return MEMG.besOf($nObj, $target_temp_nid_class); }
    isBeOf = function($nObj, $target_temp_nid_class) { return MEMG.besOf($nObj, $target_temp_nid_class); }
    beIsA = function($nObj, $target_temp_nid_class) { return MEMG.besOf($nObj, $target_temp_nid_class); }
    

    /**
     * 
     * @param {anne_neuron} $nObj 
     * @param {Array} $bes_tnids 
     * @returns 
     */
    isBesOfs = function($nObj, $bes_tnids) {
        for (var b=0; b <$bes_tnids.length; b++) {
            var tnid = $bes_tnids[b];
            if (this.isBeOf($nObj, tnid)==true) return true;
        }
        return false;
    }


    /**
     * these are .tnid, .ntype objs not just an array of tnids.
     * @param {*} $nObj 
     * @param {*} $tnidntypeobs 
     * @returns 
     */
    isBesOfs2 = function($nObj, $tnidntypeobs) {
        for (var b=0; b <$tnidntypeobs.length; b++) {
            var obj = $tnidntypeobs[b];
            if ($nObj.ntype!=obj.ntype) continue;
            if (this.isBeOf($nObj, obj.tnid)==true) return true;
        }
        return false;
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    buildParentLine=function($nObj, $bDOSELF=true, $bNIDARRAY=false, $tnidsunique=null, $ignoreToTnids=[]) {
        var tnids = $tnidsunique;
        if (tnids==null) tnids = [];
        if ($nObj==null) return tnids;
        if ($bDOSELF==true) tnids.push($nObj.temp_nid);

        if ($nObj.bes.length) {
            var bes = $nObj.bes;
            var lines = [];

            //not all bes are worth of be line chasing..
            //best to not PUT them in the bes array to begin with? (ie, be->eval->four
            for (var b=0; b < bes.length; b++) {
                var be_temp_nid = $nObj.bes[b];
                if (tnids.length && tnids.indexOf(be_temp_nid)!=-1) break;
                var nObj = MEMG.getNeuronByTempNid(be_temp_nid);
                if (nObj==null) return tnids;

                if ($ignoreToTnids.length>0 && $ignoreToTnids.indexOf(nObj.temp_nid)!=-1) break;

                //temp hack to honor polity and BE relns.. what we really need to honor
                //is any pre-defined BE_CHILDREN BE_PARENT relons which define
                //special be parentage for any given class.  These we can honor.. the other
                if (b!=0 && be_temp_nid!=MEM.polity.temp_nid && be_temp_nid!=MEM.ec_is.temp_nid) continue;
                lines.push(nObj);
                tnids.push(nObj.temp_nid);
            }

            //chase BE lines.  We inherit params from all BES.  (reconsidering this..)
            
            //we use tnids here to make sure we never chase any dupe lines.
            for (var ab=0; ab < lines.length; ab++) {
                MEMG.buildParentLine(lines[ab], false, true, tnids);
            }
        }

        if ($bNIDARRAY==true) return tnids;
        else return MEMG.getNObjsFromTempNidList(tnids);
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    buildUpAndDownline=function($nObj, $bNIDARRAY=false, $bINSTS=false, $bDOSELF=true, $bINCLUDECLASS=true) {
        var childrenLines = [];
        $nObj.getAllDownlineChildren(childrenLines, 0, $bINSTS, $bNIDARRAY); 
        var parentLine = MEM.buildParentLine($nObj, $bDOSELF, $bNIDARRAY);
        if ($bINCLUDECLASS==true) {
            if ($bNIDARRAY==true) parentLine.push(MEM.class.temp_nid);
            else parentLine.push(MEM.class);
        }
        if (parentLine.length>1) parentLine = parentLine.reverse();
        return parentLine.concat(childrenLines);
    }


    /*** concatUplineRelons *************************************************
    * This expects the buildParentLineExtendedLive was used in the case     *
    * of chasing parents and assumes the array of nObjs is ordered from     *
    * highest on the tree (lowest be lvl) down (TOP DOWN).  It can then     *
    * override dupes params with the one it finds further DOWN.             *
    ******************************************************** Mr.Scatman ****/
    concatUplineRelons = function($arr, $aNewRels) {
        for (var r=0; r < $aNewRels.length; r++) {
            var newRel = $aNewRels[r];
            if ($arr.length) {
                var bDUPE=false;
                for (var x=0; x < $arr.length; x++) {
                    var existingRel = $arr[x];
                    if (existingRel.r==newRel.r && existingRel.to_s == newRel.to_s) {
                        if (existingRel.type==TYPE_PARAM) {
                            //take the new relon, it came later or was at same
                            //lvl but a primary be (not a sideways be)
                            $arr[x]=newRel; 
                            bDUPE=true;
                        }
                    }
                }
                if (bDUPE==false) $arr.push(newRel);
            }
            else $arr.push(newRel);
        }
    }


    buildParentLineExtendedLive=async($nObj, $retArr=null)=> {
        if ($nObj==null) return;
        $retArr.push($nObj);    //don't push inst's actual params
        if ($nObj.be && $nObj.bes.length) {
            var bes = $nObj.bes;
    
            if (bes.length && bes[0]!=MEM.class.temp_nid) {
                await CHASE.lookupMinTNids(bes);
            }
    
            //do this for each of our bes..  (chase the respective be lines..
            //chase the PRIMARY BE last.
            for (var b=bes.length-1; b >= 0; b--) {
                var be_nid = bes[b];
                var nObj = MEMG.getNeuronByTempNid(be_nid);
                await MEMG.buildParentLineExtendedLive(nObj, $retArr);
            }
        }
        return;
    }



    /**
     *   
     * @author {Mr.Scatman}  
     */
    chaseUpLineForParentMatch=function($nObj, $target_temp_nid) {
        if (typeof $target_temp_nid !=="number") {
            if ($target_temp_nid.temp_nid!=undefined) $target_temp_nid = $target_temp_nid.temp_nid;
            alert("fix your bad (chaseUpLineForParentMatch) caller! 2nd param should be a integer, not a neuron!");
            return false;
        }
        if ($target_temp_nid==$nObj.temp_nid) return true;
        var parentTempNid=$nObj.bes[0];
        var loops = 20;     //safety first. 
        while (loops>=0 && parentTempNid && parentTempNid!=MEM.___SELF___.temp_nid) {
            if (parentTempNid == $target_temp_nid) {
                //console.log("found target_temp_nid:"+$target_temp_nid);
                return true;
            }
            var thisNOBJ = MEMG.getNeuronByTempNid(parentTempNid);
            if (thisNOBJ.be_lvl==0) break;
            parentTempNid=thisNOBJ.bes[0];
            //console.log("(culfpm) "+thisNOBJ.text+" parentTempNid:"+parentTempNid);
            loops--;
        }
        var targetNOBJ = MEMG.getNeuronByTempNid($target_temp_nid);
        //console.log("did not find target_temp_nid:"+$target_temp_nid+" text:"+targetNOBJ.text+" nid:"+targetNOBJ.nid+" in be line of '"+$nObj.text+"' nid:"+$nObj.nid);
        return false;
    }


    /**
     * @author {Mr.Scatman}  
     */
    concatUniqueRelnToPrioritize=function($arr, $aNewRelons, $nidPriority, $bINHERITOK=true) {
        for (var r=0; r < $aNewRelons.length; r++) {
            var newRel = $aNewRelons[r];
            if ($arr.length) {
                var bDUPE=false;
    
                //could use a prototype search here, but meh
                for (var x=0; x < $arr.length; x++) {
                    var existingRel = $arr[x];
                    
                    //no inherits for these.
                    if ($bINHERITOK==false && existingRel.f!=$nidPriority) continue;

                    //we have an inherited dupe.. take the nidPriority one!
                    if (existingRel.r==newRel.r && existingRel.to_s == newRel.to_s) {
                        if (existingRel.f==$nidPriority) {
                            bDUPE=true;
                            break;
                        }
                        
                        //swap out the existing with this one!
                        else {
                            $arr[x]=newRel;
                            bDUPE = true;
                            break;
                        }
                    }
                }
                if (bDUPE==false) $arr.push(newRel);
                else continue;
            }
            else $arr.push(newRel);
        }
    }
    

    /**
     * Search the text, case-insensitive, of all neurons in local memory.
     * 
     * @author {Mr.Scatman}  
     */
    extractAllFromRelons=function($aRelons, $bSKIPIFHASTEXT=true, 
                                    $bDOEXTRAS=false, $skipRelnNids=[], $bDONIDS=false) {
        var aList = [];
        var aNidsList = [];
        var chase_nids = [];
        
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.to_type!=TO_TYPE_ANNE) continue;
            
            //we aren't extracting from these....
            if ($skipRelnNids.length && $skipRelnNids.indexOf(rObj.r)!=-1) continue;    

            var fNObj = MEMG.getNeuronByNid(rObj.from_nid);
            if (fNObj==null) {
                if (chase_nids.indexOf(rObj.from_nid)==-1) chase_nids.push(rObj.from_nid);
                continue;
            }

            var rNObj = MEMG.getNeuronByNid(rObj.reln);
            if(rNObj==null) {
                if (chase_nids.indexOf(rObj.reln)==-1) chase_nids.push(rObj.reln);
                continue;
            }

            var toNObj = MEMG.getNeuronByNid(rObj.to_s);
            if(toNObj==null) {
                if (chase_nids.indexOf(rObj.to_s)==-1) chase_nids.push(rObj.to_s);
                continue;
            }
        }

        
        //chase_nids if we have them
        //await ANNE.CHASE.lookupMinList(chase_nids, [0,1,2,3,7]);
            
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.to_type!=TO_TYPE_ANNE) continue;
            
            //we aren't extracting from these....
            if ($skipRelnNids.length && $skipRelnNids.indexOf(rObj.r)!=-1) continue;    

            var fNObj = MEMG.getNeuronByNid(rObj.from_nid);
            if (fNObj==null) continue;
            
            if ($bSKIPIFHASTEXT==true && fNObj.text!=""); 
            else if (fNObj.bCHASED_min==undefined && aNidsList.indexOf(fNObj.temp_nid)==-1) {
                aList.push(fNObj.nid);
                aNidsList.push(fNObj.temp_nid);
            }
            var rNObj = MEMG.getNeuronByNid(rObj.reln);
            if ($bSKIPIFHASTEXT==true && rNObj.text!="");
            else if (rNObj.bCHASED_min==undefined && aNidsList.indexOf(rNObj.temp_nid)==-1) {
                aList.push(rNObj.nid);
                aNidsList.push(rNObj.temp_nid);
            }
            
            
            var tNObj = null;
            if (rObj.t!=undefined) tNObj = MEMG.getNeuronByNid(rObj.to_s);
            else {
                tNObj = MEMG.getOrMakeIfNew(rObj.to_s);
                if (tNObj!=null) {
                    rObj.t = tNObj.temp_nid;   //fix it while we are here.
                    if ($bSKIPIFHASTEXT==true && tNObj!=null && tNObj.text!="");
                    else if (tNObj.bCHASED_min==undefined && aNidsList.indexOf(tNObj.temp_nid)==-1) {
                        aList.push(tNObj.nid);
                        aNidsList.push(tNObj.temp_nid);
                    }
                }
            }

            if ($bDOEXTRAS==true) {
                if (rObj.owned_by!=undefined && rObj.owned_by!="") {
                    var ownedByNObj = MEMG.getOrMakeIfNew(rObj.owned_by);
                    if ($bSKIPIFHASTEXT==true && ownedByNObj.text!="");
                    else if (ownedByNObj.bCHASED_min==undefined && aNidsList.indexOf(ownedByNObj.temp_nid)==-1) {
                        aList.push(ownedByNObj.nid);
                        aNidsList.push(ownedByNObj.temp_nid);
                    }
                }
                if (rObj.sponsored_by!=undefined && rObj.sponsored_by!="") {
                    var sponsoredByNObj = MEMG.getOrMakeIfNew(rObj.sponsored_by);
                    if ($bSKIPIFHASTEXT==true && sponsoredByNObj.text!="");
                    else if (sponsoredByNObj.bCHASED_min==undefined && aNidsList.indexOf(sponsoredByNObj.temp_nid)==-1) {
                        aList.push(sponsoredByNObj.nid);
                        aNidsList.push(sponsoredByNObj.temp_nid);
                    }
                }
            }

        }
        if ($bDONIDS) return aNidsList;
        return aList;
    }


    /**
     * Search the text, case-insensitive, of all neurons in local memory.
     * 
     * @author {Mr.Scatman}  
     */
    extractAllFromRelonsAndChaseMin=async($aRelons, $SHARD=20, $bDOEXTRAS=false)=> {
        var temp_nids = MEMG.extractAllFromRelons($aRelons, true, $bDOEXTRAS, [], true);
        if (temp_nids.length==0) return;

        //chase out the min info on these nid addresses,
        var SHARD = $SHARD;
        if (temp_nids.length) {
            var lookup = [];
            for (var i=0; i  < temp_nids.length; i++) {
                if (lookup.length < SHARD) lookup.push(temp_nids[i]);
                else {
                    var done = await ANNE.CHASE.lookupMinTNids(lookup);
                    lookup = [];
                }
            }
            if (lookup.length) {
                var done = await ANNE.CHASE.lookupMinTNids(lookup);
                lookup = [];
            }
        }
        return;
    }


    /**
     * 
     */
    getDefines = function($nObj) {
        var nObjs = [];
        for (var r=0; r < $nObj.outs.length; r++) {
            if($nObj.outs[r].r==MEM.DEFINES.temp_nid) {
                var nObj = MEMG.getNeuronByTempNid($nObj.outs[r].t);
                nObjs.push(nObj);
            }
        }
        return nObjs;
    }


    /**
     *
     */
    getDefinedBy = function($nObj) {
        if ($nObj==undefined || $nObj==null) return null;
        for (var r=0; r < $nObj.outs.length; r++) {
            if($nObj.outs[r].r==MEM.DEFINED_BY.temp_nid) {
                var nObj = MEMG.getNeuronByTempNid($nObj.outs[r].t);
                return nObj;
            }
        }
        return null;
    }

    


    /**
     * Runtime relons for context.
     * This is not for display -- it is for dpool context resolutions.
     * Upline relons that are otherwise ignored (grandparents) in the standard inheritance
     * use are included in the results here to enable better context pool matching.
     */
    getRuntimeRelonsForContext = async($nid, $types)=> {
        var nObj = MEMG.getNeuronByNid($nid);
        if ($types==null) $types = [TYPE_SYS, TYPE_BE, TYPE_PARAM, TYPE_HAS, TYPE_CLUMP, TYPE_ROLE, TYPE_BRAIN];
        var bADDSELF = true;
        var bEXTRAPRETURN = true;
        
        //we want all dupes and all uplines.. we're not building a context pool.
        var bFORINHERITANCE = false;    

        var acceptable_outs = MEMG.getParentLineRelons(nObj, $types, [], bADDSELF, bEXTRAPRETURN, bFORINHERITANCE);

        var bUS = USER.isUsOrOwnedByUs(nObj.nid);
        var outs = acceptable_outs;
        if (outs.length==0) outs = nObj.outs;
        if (outs.length==0) return outs;

        //do our best to have basic info for all the parts of the relons.
        await MEMG.extractAllFromRelonsAndChaseMin(outs,true);
        
        var finalRelons=[];
        var _usedTXIDs=[];
        //var types=[0,1,2,3,5,6,7,8,9];

        for (var r=0; r < outs.length; r++) {
            var relon = outs[r];
            if (_usedTXIDs.indexOf(relon.txid)!=-1) continue;
            _usedTXIDs.push(relon.txid);
            finalRelons.push(relon);
        }

        return finalRelons;
    }


    /**
     * Get a comprehensive list of relons (including inherited) according to ANNE L1 protocol.
     * 
     * We do not get the special relon types here.. they apply only to the neuron they are on.
     * 
     * @param {string} - nid longstring 
     * @author {Mr.Scatman}
     */
    getRuntimeRelons = async($nid, $types=null)=> {
        if (($nid instanceof anne_neuron) == true) {
            if ($nid.nid!=undefined) {
                console.warn("(getRuntimerelons should take a nid, not a neuron!");
                $nid = $nid.nid;
            }
            else {
                console.warn("(getRuntimerelons) bad param");
                return [];
            }
        }

        var nObj = MEMG.getNeuronByNid($nid);
        
        //quick annode lookup??
        if (nObj==null) {
            alert("no nobj for getRuntimerelons");
            return;
        }

        var acceptable_outs = [];
        if ($nid!=USER._userNObj.nid) {
            if (MEM.tempBeIsA(nObj.temp_nid, MEM.identity.temp_nid, NTYPE_INST)==true) {
                var basic_other_identity_types  = [TYPE_HAS, TYPE_BE, TYPE_CLUMP, TYPE_ROLE];
                acceptable_outs = MEMG.prepForIdentity(nObj, basic_other_identity_types);
            }
            else {
                if ($types==null) $types = [TYPE_SYS, TYPE_BE, TYPE_PARAM, TYPE_HAS, TYPE_CLUMP, TYPE_ROLE, TYPE_BRAIN];
                var bADDSELF = true;
                var bEXTRAPRETURN = true;
                var bINHERITANCE = true;
                acceptable_outs = MEMG.getParentLineRelons(nObj, $types, [], bADDSELF, bEXTRAPRETURN, bINHERITANCE);
            }
        }

        //used for self-identity
        else {
            acceptable_outs = MEMG.prepForIdentity(nObj, $types);
        }

        //var bUS = USER.isUsOrOwnedByUs(nObj.nid);
        var outs = acceptable_outs;
        if (outs.length==0) outs = nObj.outs;
        if (outs.length==0) return outs;

        //do our best to have basic info for all the parts of the relons.
        await MEMG.extractAllFromRelonsAndChaseMin(outs,true);

        var finalRelons=[];
        var _usedTXIDs=[];

        for (var r=0; r < outs.length; r++) {
            var relon = outs[r];
           
            if (_usedTXIDs.indexOf(relon.txid)!=-1) continue;
            if (relon.type==TYPE_PARAM) {
                if (relon._multiples_ok==undefined || relon._multiples_ok==false) {
                    var toNObj = MEMG.getNeuronByNid(relon.to_s);
                    var bWEHAVEIT=false;
                    var relNOBJ = MEMG.getNeuronByTempNid(relon.r);

                    //roles are THE DEFINITION of the class concept.
                    //they are now bound by the PARAM restrictions.
                    //the params are for the roles IT assigns.
                    if (relNOBJ.ntype==NTYPE_CLASS && MEM.isRoleR(relon.r)==false) {
                        bWEHAVEIT = MEM.hasMetClassDefn(outs, relon, toNObj);
                        if (bWEHAVEIT == true) continue;
                    }
                }
            }
            
            _usedTXIDs.push(relon.txid);
            finalRelons.push(relon);
        }

        //gets relons from collab space (MEM_UDF._collab_relons)
        //this injects relons that are not in _aRelon space so we can
        //use them as if they were.  Only the ADMIN or submitter
        var andy_relons = ANNE.MEM_UDF.getAndyTblRelonsForNid($nid);
        if (andy_relons!=null) {
            for (var a=0; a < andy_relons.length; a++) {
                if (_usedTXIDs.indexOf(andy_relons[a].txid)!=-1) continue;
                finalRelons.push(andy_relons[a]);
                _usedTXIDs.push(andy_relons[a].txid);
            }
        }
        return finalRelons;
    }
    
    
    /**
     */
     prepForIdentity = function($targetNOBJ, $types=null) {
        var inherit_relons = MEMG.getParentLineRelons($targetNOBJ,[TYPE_PARAM]);
        var acceptable_outs = [];
        var params = inherit_relons.params;
        var txids = [];

        //these are only params.
        for (var i=0; i < params.length; i++) {
            var rel=params[i];

            //identities are instances.  We will inherit the default class param of class->class and 
            //we don't want that in here.
            if (rel.r == MEM.class.temp_nid) continue;

            var fObj = MEMG.getNeuronByTempNid(rel.f);
            var rnOBJ = MEMG.getNeuronByTempNid(rel.r);
        
            //no params of sys classes.
            //these are set by sys processes.
            if (rnOBJ.bes[0] == MEM.__ANNE_SYS_CLASS.temp_nid) continue;

            //var dimen = MEMG.getRelonByTempR(rnOBJ.outs, MEM.relon_dimension.temp_nid);
            var target_relon_type = TYPE_HAS;
            if (rnOBJ.target_type!=undefined) {
                target_relon_type = rnOBJ.target_type;
            }

            //if (dimen!=null) target_relon_type = dimen.to_s;
            rel.target_type = target_relon_type;
            
            if (fObj.prompt!=undefined) rel.prompt = fObj.prompt;// :"";

            if (txids.indexOf(rel.txid)==-1) {
                acceptable_outs.push(rel);
                txids.push(rel.txid);
            }
        }
            
        var types = [];
        if ($types==null) types = [TYPE_HAS, TYPE_BE, TYPE_CLUMP, TYPE_ROLE];
        else types = $types;
        var hasbe_outs = MEMG.getAllRelonsOfType($targetNOBJ.outs, types);
        //ANNE.TALK.printRelons(hasbe_outs);
        if (hasbe_outs.length) {
            for (var h=0; h < hasbe_outs.length; h++) {
                var rel = hasbe_outs[h];
                if (txids.indexOf(rel.txid)==-1) {
                    acceptable_outs.push(rel);
                    txids.push(rel.txid);
                }
            }
        }
        return acceptable_outs;
    }


    /** 
     * @author {Mr.Scatman} 
     */
     getToTextOfRelon=function($rObj, $bSHORTOK=false) {
        if ($rObj.to_type==TO_TYPE_ANNE) {
            var tObj = MEMG.getNeuronByTempNid($rObj.t);
            if (tObj) {
                if ($bSHORTOK==true && tObj.short!=undefined) return tObj.short;
                return tObj.text;
            }
            else return "";
        }
        else return $rObj.to_s;
    }

    getAllRelonsWithRelnOnly=function($reln){
        var matches=[];
        for (var r=0; r < ANNE.MEM._aRelons.length; r++) {
            var rObj = ANNE.MEM._aRelons[r];
            if (rObj.reln==$reln) {
                matches.push(rObj);
            }
        }
        return matches;
    }

    /** 
     * @author {Mr.Scatman} 
     */
    getRelonsFromWith_fast = function ($outs, $type, $r, $to_type) {
        var matches=[];
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r==$r && rObj.type==$type && rObj.to_type==$to_type) {
                matches.push(rObj);
            }
        }
        return matches;
    }


    getRelonsFromWith2 = function ($outs, $type, $r) {
        var matches=[];
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r==$r && rObj.type==$type) {
                matches.push(rObj);
            }
        }
        return matches;
    }

    /**
     * This function is intended to be run and chase all relevant max info   
     * on parentLine, honoring garden and WV.                                
     * The chase for parents does not yet honor garden and WV because we     
     * first have to acquire it to know what the garden and WV is!          
     * Note: this was used in the DATA SOURCE building and not yet in the EXT
     * 
     * @author {Mr.Scatman} 
     */
    getParentLineParamsLive=async($nObj)=> {
        var selfAndUpNObjs = [];
        await MEMG.buildParentLineExtendedLive($nObj, selfAndUpNObjs);

        console.log(selfAndUpNObjs);

        //now we can honor garden/wv and get the params array.
        var allParamRelons = [];
        if (selfAndUpNObjs.length==0) return allParamRelons;
        selfAndUpNObjs.reverse();

        //for each item in our selfAndUp -- snag the PARAM relons.
        //process then from TOP DOWN to override w/ later relons.
        //TREES are defined bottom up but inherit top down -- so the child 
        //can override any inherited parent BE.
        for (var  n=0; n < selfAndUpNObjs.length; n++) {
            var lineNOBJ = selfAndUpNObjs[n];
            var aRelons = MEMG.getAllParamsOf(lineNOBJ.outs);
            if (aRelons) {
                MEMG.concatUplineRelons(allParamRelons, aRelons);
            }
        }
        return allParamRelons;
    }


    /** 
     * @author {Mr.Scatman} 
     */
     getSysTypeObjByTypeAndBe=function($enum_val, $beNObj) {
        var sObj=null;
        for (var s=0; s < MEM._SYS_TYPE_CODES.length; s++) {
            sObj = MEM._SYS_TYPE_CODES[s];
            if (sObj.SYS_ENUM_VAL==$enum_val && sObj.bes.indexOf($beNObj.temp_nid)!==-1) return sObj;
        }
        return null;
    }

    /**
     * 
     * @param {*} $enum_val 
     * @returns 
     */
    getSysTypeObjByType=function($enum_val) {
        var sObj=null;
        for (var s=0; s < MEM._SYS_TYPE_CODES.length; s++) {
            sObj = MEM._SYS_TYPE_CODES[s];
            if (sObj.SYS_ENUM_VAL==$enum_val) return sObj;
        }
        return null;    
    }

    /** 
     * @author {Mr.Scatman} 
     */
    getSysToTypeObjByTypeAndBe=function($enum_val, $beNObj) {
        var sObj=null;
        for (var s=0; s < MEM._SYS_TO_TYPE_CODES.length; s++) {
            sObj = MEM._SYS_TO_TYPE_CODES[s];
            if (sObj.SYS_ENUM_VAL==$enum_val && sObj.bes.indexOf($beNObj.temp_nid)!==-1) return sObj;
        }
        return null;
    }
    getSysToTypeObjByType=function($enum_val) {
        var sObj=null;
        for (var s=0; s < MEM._SYS_TO_TYPE_CODES.length; s++) {
            sObj = MEM._SYS_TO_TYPE_CODES[s];
            if (sObj.SYS_ENUM_VAL==$enum_val) return sObj;
        }
        return null;    
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
     getAllBesOf = function ($arr, $iggy_temps=[]) {
        var aMatches = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_BE) aMatches.push(rObj);
        }
        return aMatches;
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    getAllSysMsgOf = function ($arr, $iggy_temps=[]) {
        var aMatches = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_SYS_MSG) aMatches.push(rObj);
        }
        return aMatches;
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    getAllBrainsOf = function ($arr, $iggy_temps=[]) {
        var aMatches = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_BRAIN) aMatches.push(rObj);
        }
        return aMatches;
    }

        

    /**
     *   
     * @author {Mr.Scatman}  
     */
    getAllClasses = function () {
        var children = MEM.class.children;
        return MEMG.getNObjsFromTempNidList(children); 
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    getALLNIDSINLOCALMEM = function($bINSTONLY=false) {
        var tnids = [];
        for (var n=0; n < MEM._aNeurons.length; n++) {
            if ($bINSTONLY==true) if (MEM_.aNeurons[n].ntype!=NTYPE_INST) continue;
            tnids.push(MEM._aNeurons[n].temp_nid);
        }
        return tnids;
    }


   /**
     */
    getAllRelonsOfType=function($outs, $types) {
        var matches=[];
        for (var r=0; r < $outs.length; r++) { if ($types.indexOf($outs[r].type)!==-1) matches.push($outs[r]);} 
        return matches;
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    getAllSysOf = function ($arr, $iggy_temps=[]) {
        var aMatches = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_SYS) aMatches.push(rObj);
        }
        return aMatches;
    }
    

    /**
     *   
     * @author {Mr.Scatman}  
     */
    getAllUplineBes = function($nObj, $bes, $loops=0) {
        if ($loops>30) return "STOP";
        for (var b=0; b < $nObj.bes.length; b++) {
            var beNOBJ = MEMG.getNeuronByTempNid($nObj.bes[b]);
            if (beNOBJ.temp_nid==MEM.class.temp_nid) return;
            if ($bes.indexOf(beNOBJ.temp_nid)==-1) $bes.push(beNOBJ.temp_nid);
            var test = MEMG.getAllUplineBes(beNOBJ, $bes, $loops++);   //recurse up.. (shouldn't be long here..);  
            if (test=="STOP") return "STOP";
        }
        return;
    }


    /**
     * Extract all clump relons from a neuron.
     * These clump relons point to clumps so we can expect to handle them accordingly.
     * @author {Mr.Scatman}  
     */
     getAllClumpsOf = function ($arr, $iggy_temps=[]) {
        var aRelons = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_CLUMP) aRelons.push(rObj);
        }
        return aRelons;
    }

    getAllClumpNids = function ($arr, $iggy_temps=[]) {
        var clump_nids = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_CLUMP) {
                if (clump_nids.indexOf(rObj.to_s)==-1) clump_nids.push(rObj.to_s);
            }
        }
        return clump_nids;
    }


    /**
     * Extract all has relons from a neuron.
     * 
     * @author {Mr.Scatman}  
     */
    getAllHasOf = function ($arr, $iggy_temps=[]) {
        var aMatches = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_HAS) aMatches.push(rObj);
        }
        return aMatches;
    }


    /**
     * 
     */
     getAllInstNOBJS = function($skips=[]) {
        var nOBJS = [];
        for (var n=0; n < MEM._aNeurons.length; n++) {
            if(MEM._aNeurons[n].ntype==NTYPE_INST) {
                if ($skips.length && $skips.indexOf(MEM._aNeurons[n].bes[0])!=-1) continue;
                nOBJS.push(MEM._aNeurons[n]);
            }
        }
        return nOBJS;
    }


    /**
     * 
     */
    getAllInstTnids = function() {
        var tnids = [];
        for (var n=0; n < MEM._aNeurons.length; n++) {
            if(MEM._aNeurons[n].ntype==NTYPE_INST) tnids.push(MEM._aNeurons[n].temp_nid);
        }
        return tnids;
    }


    /**
     * Extract all param relons from a neuron.
     * 
     * @author {Mr.Scatman}  
     */
    getAllParamsOf = function ($arr, $relnMatching="", $iggy_temps=[]) {
        var aMatches = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_PARAM) {
                if ($relnMatching!="" && rObj.reln!=$relnMatching) continue;
                else aMatches.push(rObj);
            }
        }
        return aMatches;
    }

    /**
     * Extract all the semantic roles (proper) from a neuron.
     * The HAS/BE/PARAM are technically semantc in nature but they are
     * short-hand roles of the triplet format in their respective dimensions.   
     * @author {Mr.Scatman}  
     */
     getAllRolesOf = function ($arr, $iggy_temps=[]) {
        var aMatches = [];
        var txids = [];
        for (var r=0; r < $arr.length; r++) {
            var rObj = $arr[r];
            if ($iggy_temps.length && $iggy_temps.indexOf(rObj.r)!=-1) continue;
            if (rObj.type==TYPE_ROLE) {
                if (txids.indexOf(rObj.txid)==-1) {
                    aMatches.push(rObj);
                    txids.push(rObj.txid);
                }
            }
        }
        return aMatches;
    }
    

    /**
     * Get all relons with a reln of given types.
     * 
     * @param {array} - The outs
     * @param {array} - The acceptable reln temp_nids.
     * @param {array} - The acceptable types
     * 
     * @author {Mr.Scatman} 
    */
     getAllRelonsWithRelnTempNids=function($outs, $reln_temp_nids, $types) {
        if ($reln_temp_nids==undefined || $reln_temp_nids==null) {
            console.warn("fix this getAllRelonsWithRelnTempNids");
            return [];
        }
        var aMatches = [];
        if ($outs.length==0) return aMatches;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if ($reln_temp_nids.indexOf(rObj.r)!=-1) {
                if ($types.length==0) aMatches.push(rObj);
                else if ($types.indexOf(rObj.type)!==-1) aMatches.push(rObj);
            }
        }
        return aMatches;
    }

    /**
     */
    getBELINEDelegation = function($nObj) {
        var dg = $nObj.delegation;
        if (dg==undefined) dg = [];
        var safety = 0;
        var parentNOBJ = MEMG.getNeuronByTempNid($nObj.bes[0]);
        while(parentNOBJ!=null && safety < 15) {
            if (parentNOBJ.delegation!=null && parentNOBJ.delegation.length) {
                dg = dg.concat(parentNOBJ.delegation);
            }
            if (parentNOBJ.be_lvl==0) break;
            if (parentNOBJ.bes.length==0) break;
            safety+=1;
            parentNOBJ = MEMG.getNeuronByTempNid(parentNOBJ.bes[0]);
        }
        return dg;
    }


    /**
     *   
     * @author {Mr.Scatman}  
     */
    getBeNOBJ=function($outs, $type) {
        if ($outs.length==0) return null;
        var temp_nid =-1;
        if ($type==NTYPE_INST) temp_nid = MEM.inst.temp_nid;
        else if ($type==NTYPE_CLASS) temp_nid = MEM.class.temp_nid;
        else return null;
    
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==temp_nid) return MEMG.getNeuronByTempNid($outs[r].t);
        }
        return null;
    }

    /**
     */
    getBeRelon=function($outs, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_BE) return $outs[r];
        }
        return null;
    }
    getBeOrHasRelon=function($outs, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && ($outs[r].type==TYPE_BE || $outs[r].type==TYPE_HAS)) return $outs[r];
        }
        return null;
    }


    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getBeRelons=function($outs, $relnTnid) {
        var rels=[];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_BE) rels.push($outs[r]);
        }
        return rels;
    }
    getBeOrHasRelons=function($outs, $relnTnid) {
        var rels=[];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && ($outs[r].type==TYPE_BE || $outs[r].type==TYPE_HAS)) rels.push($outs[r]);
        }
        return rels;
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @param {*} $toS 
     * @returns 
     */
    getBeRelonTo=function($outs, $relnTnid, $toS) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_BE && $outs[r].to_s==$toS) return $outs[r];
        }
        return null;
    }

    /**
     * @author {Mr.Scatman}  
     */
    getBeParentRelon = function ($nObj) {
        for (var r=0; r < $nObj.outs.length; r++) {
            var rObj = $nObj.outs[r];
            if (rObj.type==TYPE_BE && (rObj.r==MEM.class.temp_nid || rObj.r==MEM.inst.temp_nid)) return $nObj.outs[r];
        }
        return null
    }
    
    /**
     * Get a bool object's options and supply it in a bool obj for gui use.
     * 
     * @param {anne_neuron} - the bool neuron
     * @author {Mr.Scatman} 
     * 
     */
     getBoolPair=function($boolNOBJ) {
        if ($boolNOBJ.children<2) return null;
        var boolObj = {yesNOBJ:null, noNOBJ:null}
        for (var c=0; c < $boolNOBJ.children.length; c++) {
            var childNOBJ = MEMG.getNeuronByTempNid($boolNOBJ.children[c]);
            if (childNOBJ.value==undefined || childNOBJ.value==null) {
                MEM.boolean.set_value(childNOBJ);
                if (childNOBJ.value==undefined || childNOBJ.value==null) return null;
                if (childNOBJ.value==1) boolObj.yesNOBJ = childNOBJ;
                else if (childNOBJ.value==0) boolObj.noNOBJ = childNOBJ;
            }
            else if (childNOBJ.value==1) boolObj.yesNOBJ = childNOBJ;
            else if (childNOBJ.value==0) boolObj.noNOBJ = childNOBJ;

        }

        if (boolObj.yesNOBJ!=null && boolObj.noNOBJ!=null) return boolObj;
        return null;
    }

    /** 
     * @author {Mr.Scatman} 
     */
     getEmulatedNeuronByClassAndValue=function($beNOBJ, $text) {
        var bEXACT = true;
        var bDOAKAS = false;
        var bEMULATEDOK = true;
        if ($beNOBJ==null) {
            console.log("rut roh");
            return null;
        }
        var aMatches = MEMS.searchText($text, [NTYPE_INST], bEXACT, bDOAKAS, $beNOBJ.temp_nid, bEMULATEDOK);
        if (aMatches.length) {
            return aMatches[0];
        }
        return null;
    }

    /**
     * 
     * @returns 
     */
    getGhosts = function() {
        var aGhosts = [];
        for (var n=0; n < MEM._aNeurons.length; n++) {
            if (MEM._aNeurons[n].ghost==true) {
                aGhosts.push(MEM._aNeurons[n]);
            }
        }
        ANNE.TALK.printNObjs(aGhosts);
        return aGhosts;
    }


    /** 
     * @author {Mr.Scatman} 
     */
    getTextListByRelonTos = function($relons, $sep=", ") {
        var list = [];
        for (var r=0; r < $relons.length; r++) {
            var rel = $relons[r];
            list.push( MEMG.__MASTER_DISPLAY_RELON_TO(rel) );
        }
        if (list.length==0) return list;
        else if (list.length==1) return list[0];
        return list.join($sep);
    }


    /** 
     * @author {Mr.Scatman} 
     */
    getGlobalNeuron=function($classNObj, $val) {
        //check for number.
        if ($classNObj.temp_nid==ANNE.MEM.number.temp_nid) {
            for (var n=0; n < ANNE.MEM.number.ins.length; n++) {
                var rObj = ANNE.MEM.number.ins[n];
                if (rObj.r == MEM.eval.temp_nid && rObj.reln.to_s==$val) {
                    return MEMG.getNeuronByTempNid(ANNE.MEM.number.ins[n].f);
                }
            }
        }
        else if ($classNObj.temp_nid==ANNE.MEM.date.temp_nid) {
            //alert("to do date get global")
        }
        else if ($classNObj.temp_nid==ANNE.MEM.location.temp_nid) {
            //alert("to do get location global");
        }
        else if ($classNObj.temp_nid==ANNE.MEM.neuronImg.temp_nid) {
            var in_mem_now = MEMG.getHasTo(ANNE.MEM.fileFingerPrint.temp_nid, $val);
            if (in_mem_now!=null) {
                return ANNE.MEMG.getNeuronByTempNid(in_mem_now.f);
            }
            for (var n=0; n < ANNE.MEM.neuronImg.ins.length; n++) {
                var rObj = ANNE.MEM.neuronImg.ins[n];
                var imgNOBJ = ANNE.MEMG.getNeuronByTempNid(rObj.f);
                var finger = ANNE.MEMG.getHasRelonTo(imgNOBJ.outs, ANNE.MEM.fileFingerPrint.temp_nid, $val);
                if (finger!=null) {
                    return imgNOBJ;
                }
            }
        }
        return null;
    }

    
    /**
     * No longer used?  (like a gazillion of these old functions...)
     * 
     * @param {*} $classNObj 
     * @param {*} $val 
     * @param {*} $reln 
     * @param {*} $to_s 
     * @returns 
     */
    getNeuronWithRelon=async($classNObj, $val, $reln, $to_s, $bEXACT=true)=> {
        
        //search local memmory for this thing by class and text
        var wvs = [];
        var bes_tnids = [$classNObj.temp_nid];
        if (ANNE.USER._bWORLDVIEWBIPASS==false) wvs = MEM.getWorldViewNidList();
        var searchObj = { 
            text:$val, 
            wvs:wvs, 
            bes:bes_tnids,
            bes_tnids:bes_tnids,
            exact:$bEXACT,
            applywvtobeline:$bAPPLYWVTOBELINE,
            inbeline:0,
            chaseIt:true
        };
        var bDOPRE = false; //ok to end by finding match
        var aMatches = await CHASE.anne_global_search(searchObj, bDOPRE);
        
        //verify if the matching things also have the desired relon (reln/to_s);
        var final_matches = [];
        if (aMatches.length) {
            for (var m=0; m < aMatches.length; m++) {
                var matchNOBJ = aMatches[m];
                if (MEMG.hasRelon(matchNOBJ.outs, $reln, $to_s)==true) {
                    final_matches.push(matchNOBJ);
                }
            }
        }

        //return an array(always)
        return final_matches;
    }



    /**
     * Get a subanndle in memory for a given parent anndle/subanndle.
     * @param {*} $parentNOBJ 
     * @param {*} $name 
     * @returns 
     */
    getSubAnndle = function($parentNOBJ, $name) {
        $name = $name.toLowerCase();
        for (var r=0; r< $parentNOBJ.outs.length; r++) {
            var rel = $parentNOBJ.outs[r];
            if (rel.r!=MEM.subanndle_of.temp_nid) continue;
            var nOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (nOBJ.text.toLowerCase()==$name) return nOBJ;
        }
        return null;
    }


    /** 
     * @author {Mr.Scatman} 
     */
    getMineNObjs = function($mustBeNids=[]) {
        if (ANNE.USER==null) return;
        var owns = ANNE.USER._userNObj.owns;
        var nObjs = [];
        for (var n=0; n < owns.length; n++) {
            var nObj = MEMG.getNeuronByTempNid(owns[n]);
            if (nObj.mine==true) {
                if ($mustBeNids.length) {
                    if($mustBeNids.indexOf(nObj.bes[0])!==-1) nObjs.push(nObj);
                }
                else nObjs.push(nObj);
            }
        }
        return nObjs;
    }


    /** 
     * @author {Mr.Scatman} 
     * @param {anne_neuron} - the user
     * 
     * We will not pull rememberz here.. it only works with what is in local mem.
     * This is primarily intended for the logged-in user thus they will have all their
     * relons in mem already.
     */
    getRememberzNObjs = function($nObj) {
        var outs = $nObj.outs;
        var nObjs = [];
        for (var r=0; r < outs.length; r++) {
            var rel = outs[r];
            if (rel.type!=TYPE_AWARENESS) continue;
            if (rel.r!=MEM.rememberz.temp_nid) continue;
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            var nObj = MEMG.getNeuronByTempNid(rel.t);
            nObjs.push(nObj);
        }
        return nObjs;
    }


    /**
     * 
     * @param {String} $sponsored_by - nid longstring 
     * @param {anne_neuron} $classNOBJ 
     * @returns 
     */
    getSponsoredBy=function($sponsored_by, $classNOBJ) {
        var sp_nObjs = [];
        var sp_audit=0;
        
        if ($classNOBJ instanceof anne_neuron == false) {
            console.log("pass nOBJ please, fix this");
            $classNOBJ=MEMG.getNeuronByNid($classNOBJ);
        }
        
        for (var n=0; n  < MEM._aNeurons.length; n++) {
            if (MEM._aNeurons[n].sponsored_by!=$sponsored_by) continue;
            sp_audit+=1;
            
            if (MEM._aNeurons[n].is($classNOBJ)==false) {
                continue;
            }
            
            sp_nObjs.push(MEM._aNeurons[n]);
        }
        console.log("(getSponsoredBy) sp_audit %c"+sp_audit, ANNE.TALK.bgyellow);
        return sp_nObjs;
    }
    
    
    /**
     * 
     * @param {*} $sponsored_by 
     * @param {anne_neuron} $classNOBJ
     * @returns 
     */
    getSponsoredByClasses=function($sponsored_by, $classNOBJ) {
        var sp_nObjs = [];
        var sp_audit=0;
        if ($classNOBJ instanceof anne_neuron == false) {
            console.log("pass nOBJ please, fix this");
            $classNOBJ=MEMG.getNeuronByNid($classNOBJ);
        }
        for (var n=0; n  < MEM._aNeurons.length; n++) {
            if (MEM._aNeurons[n].sponsored_by!=$sponsored_by) continue;
            sp_audit+=1;
            if (MEM._aNeurons[n].ntype!=NTYPE_CLASS) continue;
            if (MEM._aNeurons[n].is($classNOBJ)==false) continue;
            sp_nObjs.push(MEM._aNeurons[n]);
        }
        console.log("(getSponsoredByClasses) sp_audit %c"+sp_audit, ANNE.TALK.bgyellow);
        return sp_nObjs;
    }
    
    
    /**
     * 
     * @param {*} $str 
     * @param {*} $wvs 
     * @returns 
     */
    getDateNeuron=async($str, $wvs=[])=> {
        var arr = await ANNE.MEM.searchTextIndex($str);
        for (var a=0; a < arr.length; a++) {
            var nObj = arr[a];
            if ($wvs.length==0) {
                if (ANNE.isANNE(nObj.owned_by)==false && nObj.owned_by!=ANNE.USER._userNObj.nid) {
                    continue;
                }
            }

            if (nObj.bes.indexOf(ANNE.MEM.date_gregorian.temp_nid)!=-1) {
                return nObj;
            }
            else if (nObj.bes.indexOf(ANNE.MEM.monthyear.temp_nid)!=-1) {
                return nObj;
            }
            else {
                console.log("Matching nObj.text "+nObj.text+" but not a date.");
            }
        }
        return null;
    }

    

    /**
     * This is not going to be in RAM unless we lookup TYPE_SYS_MSGs..
     * Also problematic is these getting stuck in MEM (not deleted)...
     * as they are just temp relons.. They should NOT be stored in db mem, only local RAM!
     * @param {*} $outs 
     * @returns 
     */
    getUnconfirmedAnnex = function($outs){
        for (var r=0; r < $outs.length; r++) {
            var rel = $outs[r];
            if (rel.type!=TYPE_SYS_MSG) continue;

            //note: dont save this relon, EVER
            if (rel.reln!=ANNE.MEM.INTRA_BLOCK_PENDING.nid) continue; //intra block pending.
            
            if (rel.to_s!="pending annex chg") continue;
            if (rel.sys_ref_txid==undefined) continue;

            if (rel.sys_ref_txid==rel.txid) {
                console.log("WHY?????????");
                console.log(rel);
                ANNE.TALK.printRelon(rel);
                continue;
            }
            var annex_rel = MEMG.lookupByTXID(rel.sys_ref_txid);
            if (annex_rel && annex_rel.r==MEM.annex.temp_nid && annex_rel.time<=0) {
                return annex_rel;//s.b -1 for unconf, fixing that now.
            }
            else {
                console.log("what is up with this????");
            }
        }

        //lookup here with API.. it is worth the trouble to save the user
        //adding info to their queue and getting a failed send.
        
        return null;
    }


    /** 
     * Can filter for specific types of things
     * @author {Mr.Scatman} 
     */
     getIdentityStuff = function($identity_nid, $bDOINGOWNED=true, $bDOINGSPONSORED=true, $bDOINGREMEMBERZ=true, $bGHOSTS=true, $filter_tnids=[]) {
        var identity_nObj = ANNE.MEMG.getOrMakeIfNew($identity_nid);
        if (identity_nObj.owns==null) identity_nObj.owns=[];
        var nObjs = [];
        var tnids = []; //faster to check for dupes.. doesn't return out.
        var missing = [];
        var ghost_skips = 0;
         
        if ($bGHOSTS==true) {
            for (var n=0; n < identity_nObj.owns.length; n++) {
                var nObj = ANNE.MEMG.getNeuronByTempNid(identity_nObj.owns[n]);
                if (nObj==null) {
                    if (missing.indexOf(nObj.nid)!=-1) missing.push(nObj.nid);
                    continue;
                }
                if (nObj.ghost==false) {
                    continue;
                }
                if ($filter_tnids.length && ANNE.MEM.isInBeLines(nObj, $filter_tnids)==true) continue;
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }

        if (($bDOINGOWNED==true) && identity_nObj.owns!=null) {
            for (var n=0; n < identity_nObj.owns.length; n++) {
                var nObj = ANNE.MEMG.getNeuronByTempNid(identity_nObj.owns[n]);
                if ($bGHOSTS==false && nObj.ghost==true) {
                    ghost_skips+1;
                    continue;
                }
                if (nObj==null) {
                    if (missing.indexOf(nObj.nid)!=-1) missing.push(nObj.nid);
                    continue;
                }
                if ($filter_tnids.length && MEM.isInBeLines(nObj, $filter_tnids)==true) continue;
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }

        //this should work for any user (now) if we have the relons.
        if ($bDOINGSPONSORED==true) {
            ANNE.MEMG.setMySponsored(identity_nObj.nid);
            for (var s=0; s < identity_nObj.my_sponsored.length; s++) {
                var nObj = ANNE.MEMG.getNeuronByTempNid(identity_nObj.my_sponsored[s]);
                if (nObj==null) {
                    if (missing.indexOf(nObj.nid)!=-1) missing.push(nObj.nid);
                    continue;
                }
                //shouldn't be remembering neurons if they have a sponsored by.. this is likely ptless
                if ($bGHOSTS==false && nObj.ghost==true) {  
                    ghost_skips+1;
                    continue;
                }
                if ($filter_tnids.length && ANNE.MEM.isInBeLines(nObj, $filter_tnids)==true) continue;
                
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }

        //this will work with any user if we have the data..
        //typically we WON'T be pulling all their awareness relons.
        if ($bDOINGREMEMBERZ==true) {
            var nObjRememberz = ANNE.MEMG.getRememberzNObjs(identity_nObj);
            for (var n=0; n < nObjRememberz.length; n++) {
                var nObj = nObjRememberz[n];
                if (nObj==null) {
                    if (missing.indexOf(nObj.nid)!=-1) missing.push(nObj.nid);
                    continue;
                }

                //shouldn't be remembering ghosts.. this is likely ptless
                if ($bGHOSTS==false && nObj.ghost==true) { 
                    ghost_skips+1;
                    continue;
                }
                if ($filter_tnids.length && MEM.isInBeLines(nObj, $filter_tnids)==true) continue;
                
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }
        return nObjs;
    }


    /** 
     * @author {Mr.Scatman} 
     */
     getMyStuffByClass = function($identity_nid, $bDOINGOWNED=true, $bDOINGSPONSORED=true, $bDOINGREMEMBERZ=true, $bGHOSTS=true, $class_tnids=[], $ntypes=[]) {
        var identity_nObj = MEMG.getOrMakeIfNew($identity_nid);
        if (identity_nObj.owns==null) identity_nObj.owns=[];

        //pull mystuff from local db first.

        var nObjs = [];
        var tnids = []; //faster to check for dupes.. doesn't return out.
        if ($bDOINGOWNED==true && identity_nObj.owns!=null) {
            for (var n=0; n < identity_nObj.owns.length; n++) {
                var nObj = MEMG.getNeuronByTempNid(identity_nObj.owns[n]);
                if ($ntypes.length && $ntypes.indexOf(nObj.ntype)==-1) continue;
                if ($class_tnids.length>0 && MEM.isInBeLines(nObj, $class_tnids)==false) continue;
                if ($bGHOSTS==false && nObj.ghost==true) continue;
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }

        //this should work for any user (now) if we have the relons.
        if ($bDOINGSPONSORED==true) {
            
            //do it again anyways.
            MEMG.setMySponsored(identity_nObj.nid);

            for (var s=0; s < identity_nObj.my_sponsored.length; s++) {
                var nObj = MEMG.getNeuronByTempNid(identity_nObj.my_sponsored[s]);
                if ($ntypes.length && $ntypes.indexOf(nObj.ntype)==-1) continue;
                if ($class_tnids.length && MEM.isInBeLines(nObj, $class_tnids)==false) continue;
                if ($bGHOSTS==false && nObj.ghost==true) continue;
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }

        if ($bDOINGREMEMBERZ==true) {
            var nObjRememberz = MEMG.getRememberzNObjs(identity_nObj);
            for (var n=0; n < nObjRememberz.length; n++) {
                var nObj = nObjRememberz[n];
                if (nObj==null) {
                    if (missing.indexOf(nObj.nid)!=-1) missing.push(nObj.nid);
                    continue;
                }
                //shouldn't be remembering ghosts.. this is likely ptless
                if ($bGHOSTS==false && nObj.ghost==true) continue;

                if ($class_tnids.length && MEM.isInBeLines(nObj, $class_tnids)==false) continue;
                
                if (tnids.indexOf(nObj.temp_nid)==-1) {
                    nObjs.push(nObj);
                    tnids.push(nObj.temp_nid);
                }
            }
        }

        return nObjs;
    }
    

    /** 
     * @author {Mr.Scatman} 
     */
     getNeuronByNid = function ($nid, $maxFireWeight=0.03) {
        if ($nid == undefined || $nid == null || $nid == '') {
            console.log("(getNeuronByNid)nid address was not provided (null)");
            return null;
        }
        if (($nid+"").length<10) {
            console.warn("find the violators");
            return null;
        }
    
        return MEMG.__hashNeuronLookup($nid, $maxFireWeight);
    }


    /** 
     * @author {Mr.Scatman} 
     */
    getNeuronByPendingNid=function($pending_nid) {
        var parts = $pending_nid.split("_");
        if (parts[0]=="new") return MEMG.__hashNeuronLookup(parts[1]);
        console.log("trouble with pending nid");
        return null;
    }


    /** 
     * @author {Mr.Scatman} 
     */
     getNeuronByTempNid = function ($temp_nid, $maxFireWeight=0.01) {
        if ($temp_nid==undefined) return null;
        //signa anne chain convert TEST.. REMOVE THIS CHECK AFTER VETTING.
        if ((""+$temp_nid).length>8) {
            console.warn("error in getNeuronByTempNid "+$temp_nid+"(intended getNeuronByNid??)");
            return MEMG.getNeuronByNid($temp_nid);
        }
        var temp = parseInt($temp_nid);
        if (isNaN(temp)==true) return null;

        if (ANNE.BRAIN!=null) {
            ANNE.BRAIN.fireNeuron(ANNE.MEM._aNeurons[temp], $maxFireWeight);
        }

        return ANNE.MEM._aNeurons[temp];
    }

    
    /**
     * TODO: reroute this if text indexes are not turned on.
     * They can be memory intense.
     * @param {*} $str 
     * @returns an array of tnids for the neurons
     */
    getTnidsByAnbl = function($str) {
        var str = $str.toLowerCase();
        str = str.trim();
        return ANNE.BRAIN.getTnidsByAnbl(str);
    }
    

    /** 
     * NOT very fast, in any use of the word.  MUCH MORE IDEAL to use the 
     * getNObjsFromTempNidList now that bes, adopted, and children arrays 
     * are all temp_nid index arrays.       
     *  
     * @author {Mr.Scatman} 
     */
    getNidsFromNObjs=function($aList) { return this.getNObjsFromNidList($aList);}
    getNObjsFromNidList=function($aList) {
        var aNObjs = [];
        for (var n=0; n<$aList.length;n++) {
            if ($aList[n]==undefined || $aList[n]==null) {
                console.warn("(getNObjsFromNidList) bad data in arr")
                continue;
            }
            var nObj = MEMG.getNeuronByNid($aList[n]);
            aNObjs.push(nObj);
        }
        return aNObjs;
    }

    getNObjsFromRelonsToS=function($relons) {
        var aNObjs = [];
        for (var r=0; r<$relons.length;r++) {
            if ($relons[r]==undefined || $relons[r]==null) {
                console.warn("(getNObjsFromRelonsToS) bad data in arr")
                continue;
            }
            if ($relons[r].to_type!=TO_TYPE_ANNE) continue;
            var nObj = MEMG.getNeuronByTempNid($relons[r].t);
            aNObjs.push(nObj);
        }
        return aNObjs;
    }

    /** 
     * @author {Mr.Scatman} 
     */    
    getNObjsFromTempNidList=function($aList, $ignoreList=[], $bINSTSONLY=false) {
        var aNObjs = [];
        for (var n=0; n<$aList.length;n++) {
            var val = $aList[n];

            //sometimes we accidently pass in neurons here and not temp_nids.
            if ((val instanceof anne_neuron) == true) {
                if ($ignoreList.length && $ignoreList.indexOf(val.nid)!=-1) continue;
                aNObjs.push(val);
            }
            else {
                var nObj = MEMG.getNeuronByTempNid($aList[n]);
                if ($ignoreList.length && $ignoreList.indexOf(nObj.nid)!=-1) continue;
                if ($bINSTSONLY==true && nObj.ntype==NTYPE_INST) continue;
                aNObjs.push(nObj);
            }
        }
        return aNObjs;
    }


    /**
     * Get or make a new integer.
     * @param {*} $int 
     * @returns 
     */
    getOrMakeIfNewInteger=async($aec, $int)=> {
        var intyr = $int+"";
        
        for (var i=0; i < ANNE.MEM.integer.ins.length; i++) {
            var intNOBJ = ANNE.MEMG.getNeuronByTempNid( ANNE.MEM.integer.ins[i].f);
            if (intNOBJ==null) continue;
            if (intNOBJ.text==intyr) {
                return intNOBJ;
            } 
        }
        var result = await CHASE.getGlobal(ANNE.MEM.integer, intyr);
        if (result.status=="good" && result.message=="available") {
            if ($aec!=null) $aec.printf("added new integer neuron", TALK.BLUE, intyr);
            var ghostNOBJ = await GHOSTDATA.makeGhostNeuron(ANNE.MEM.integer, intyr, NTYPE_CLASS);
            return ghostNOBJ;
        }
        else {
            if ($aec!=null) $aec.printf("found existing integer", TALK.YELLOW, intyr);
            console.log(result);
            console.log("already exists");
            return result.result;//this should be the nOBJ
        }
    }


    /**
     * 
     * @param {anne_external_connector} $aec 
     * @param {*} $name 
     * @param {*} $extraReln 
     * @param {*} $extraTo 
     * @returns 
     */
    getOrMakeIfNewPerson=async($aec, $name, $extraReln=null, $extraToType=null, $extraTo=null, $bEARLYEXIT=false)=> {        
        if ($extraReln==null && $aec!=null && $aec.hasPerson($name)==true) {
            return {status:"good", match:$aec.getPerson($name)};
        }

        var name = $name.toLowerCase();
        
        var extraRelnNOBJ = null;
        if ($extraReln!="" && $extraReln!=null) extraRelnNOBJ = MEMG.getNeuronByNid($extraReln);
        
        for (var i=0; i < MEM.person.ins.length; i++) {
            if (MEM.person.ins[i].type!=TYPE_BE) continue;
            if (MEM.person.ins[i].r!=MEM.inst.temp_nid) continue;
            var personNOBJ  = MEMG.getNeuronByTempNid( MEM.person.ins[i].f);
            if (personNOBJ==null) continue;

            if (personNOBJ.text.toLowerCase()==name) {
                if (extraRelnNOBJ!=null) {
                    //var wikirel = MEMG.getHasRelon(nObj.outs, ANNE.MEM.wikipart.temp_nid);
                    var rel = MEMG.getRelonByTempR(personNOBJ.outs, extraRelnNOBJ.temp_nid);
                    if (rel!=null) {
                        if (rel.to_s!=$extraTo) continue;
                    }
                }

                $aec.printf("found existing person (by alias) of", TALK.YELLOW, $name);
                return {status:"good", match:personNOBJ};
            }
            if (personNOBJ.aka_lower!=null && personNOBJ.aka_lower.length>0) {
                if (personNOBJ.aka_lower.indexOf(name)!=-1) {
                    if (extraRelnNOBJ!=null) {
                        var rel = MEMG.getRelonByTempR(personNOBJ.outs, extraRelnNOBJ.temp_nid);
                        if (rel!=null) {
                            if (rel.to_s!=$extraTo) continue;
                        }
                    }
                    if ($aec!=null) $aec.printf("found existing person (by alias) of", TALK.YELLOW_i, $name);
                    return {status:"good", match:personNOBJ};
                }
            }
        }

        if ($bEARLYEXIT==true) return {status:"early_exit"};


        //would be lovely if we could have snagged the 0/1 relons here..
        //hmmmm.
        var NOBJS = await CHASE.queryFoNeuronByClassAndRelon(MEM.person, $name, $extraReln, $extraToType, $extraTo);
        
        if (NOBJS!=undefined && NOBJS!=null && NOBJS.length) {
            if (NOBJS.length==1) {
                var bAUDITFAIL=false;

                //audit
                if (extraRelnNOBJ!=null) {
                    var rel = MEMG.getRelonByTempR(NOBJS[0].outs, extraRelnNOBJ.temp_nid);
                    if (rel!=null) {
                        if (rel.to_s!=$extraTo) {
                            console.log("AUDIT FAIL SMOKEY.");
                            bAUDITFAIL=true;
                        }
                    }
                }
                
                if (bAUDITFAIL==true) {
                    if ($aec!=null)$aec.printf("AUDIT FAIL, ambiguous persons for", TALK.GREEN, $name);
                    return {status:"ambiguous", matches:NOBJS};
                }

                if ($aec!=null)$aec.printf("found existing person", TALK.YELLOW, $name);
                return {status:"good", match:NOBJS[0]};
            }
            else {

                //figure out why we have ambiguous results here.
                for (var n=0; n < NOBJS.length; n++) {
                    console.log(NOBJS[n].text+" "+NOBJS[n].nid);
                    ANNE.TALK.printRelons(NOBJS[n].outs);
                }
            }
            if ($aec!=null) $aec.printf("ambiguous persons for", TALK.GREEN, $name);
            return {status:"ambiguous", matches:NOBJS};
        }

        //this will make a first name + last name alias if there is no suffix
        var nObj = await GHOSTDATA.makeGhostNeuron(MEM.person, $name, NTYPE_INST);
        if ($aec!=null) $aec.addPerson(nObj);
        if ($aec!=null) $aec.printf("added new person", TALK.YELLOW, $name);
        return {status:"good", match:nObj};
    }


    /**
     * 
     * @param {*} $parentClass 
     * @returns 
     */
    estimateClassNames = function($parentClass) {
        var downline =[];
        var names = [];
        $parentClass.getAllDownlineChildren(downline,0, false, false, false, false); 
        for (var d=0; d <downline.length; d++) {
            var ctNOBJ = downline[d];
            names.push(ctNOBJ.text);
        }
        return names;
    }
    

    /**
     * Easy and quick way to get or make a new ghost corp.
     * @param {*} $name 
     * @param {*} $companyTypeNOBJ 
     * @returns 
     */
    getOrMakeIfNewCorp=async($aec, $name, $companyTypeNOBJ=null)=> {
        if ($aec.hasCorp($name)==true) {
            return {status:"good", match:$aec.getCorp($name)};
        }

        var desperados = [];

        //use the new text index, better
        var aMatches = await CHASE.nakeSearchForResolution($name, MEM.company.temp_nid);
        //console.log(aMatches);    

        if (aMatches.length>0) {
            if ($companyTypeNOBJ==null) {
                $aec.printf("found existing corp of",TALK.GREEN);
                return {status:"good", match:aMatches[0]};
            }

            //if we have a companyTypeNOBJ requirement..
            for(var m=0; m < aMatches.length; m++) {
                var ct = "";
                if (aMatches[m]._company_type==undefined) {
                    var ct_rel = MEMG.getRelonFromWithR(aMatches[m].outs, TYPE_BE, MEM.company_type.temp_nid);
                    if (ct_rel!=null) {
                        ct = ct_rel.to_s;
                        aMatches[m]._company_type=ct_rel.to_s;
                    }
                    else {
                        desperados.push(aMatches[m]);
                        continue;
                    }
                }
                else ct = aMatches[m]._company_type;

                var ctypeNOBJ = MEMG.getNeuronByNid(ct);

                if (ctypeNOBJ && ctypeNOBJ.is($companyTypeNOBJ)==true) {
                    $aec.printf("found existing corp of", TALK.YELLOW, $name,TALK.WHITE, "of type", TALK.BLUE_i,$companyTypeNOBJ.text);
                    return {status:"good", match:aMatches[m]};
                }

                //we changed the company type logic, so just take the damn thing.
                //we don't want dupes based on reparses of this.
                else desperados.push(aMatches[m]);
            }
        }


        if (desperados.length) {
            if (desperados.length==1){
                $aec.printf("found existing corp of", TALK.YELLOW,$name);
                return {status:"good", match:desperados[0]};
            }
            else {
                $aec.printf("found ambiguous corps of", TALK.GREEN,$name);
                return {status:"ambiguous", matches:desperados};
            }
        }

        
        var ghostNOBJ = await GHOSTDATA.makeGhostNeuron(MEM.company, $name, NTYPE_INST);
        $aec.addCorp(ghostNOBJ);
        if ($companyTypeNOBJ!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_BE, ghostNOBJ, MEM.company_type.nid, TO_TYPE_ANNE, $companyTypeNOBJ.nid);
            ghostNOBJ._company_type = $companyTypeNOBJ.nid;
            $aec.printf("added new corp", TALK.GREEN, $name, TALK.WHITE, "of type", TALK.BLUE_i,$companyTypeNOBJ.text);
        }
        else {
            $aec.printf("added new corp", TALK.GREEN, $name);
        }
        return {status:"good", match:ghostNOBJ};
    }



    /**
     * Easy and quick way to get or make a new ghost corp.
     * @param {*} $name 
     * @returns 
     */
    getOrMakeIfNewGroup=async($aec, $name, $bCOMPANYOK=false)=> {
        if ($aec.hasGroup($name)==true) {
            return {status:"good", match:$aec.getGroup($name)};
        }
        var name = $name.toLowerCase();
        for (var i=0; i < MEM.group.ins.length; i++) {
            if (MEM.group.ins[i].type!=TYPE_BE) continue;
            if (MEM.group.ins[i].r!=MEM.inst.temp_nid) continue;
            var thisGroupNOBJ  = MEMG.getNeuronByTempNid( MEM.group.ins[i].f);
            if (thisGroupNOBJ==null) continue;
            if (thisGroupNOBJ.text.toLowerCase()==name) {
                $aec.printf("found existing group of", TALK.YELLOW, $name);
                return {status:"good", match:thisGroupNOBJ};
            }
            if (thisGroupNOBJ.aka_lower!=null && thisGroupNOBJ.aka_lower.length>0) {
                if (thisGroupNOBJ.aka_lower.indexOf(name)!=-1) {
                    $aec.printf("found existing group (by alias) of", TALK.YELLOW_i, $name);
                    return {status:"good", match:thisGroupNOBJ};
                }
            }
        }
        
        var NOBJS = await  CHASE.queryFoNeuronByClassAndRelon(MEM.group, $name, null, null, null);
        if (NOBJS.length>0) {
            if (NOBJS.length==1) {
                $aec.printf("found existing group of", TALK.YELLOW, $name);
                return {status:"good", match:NOBJS[0]};
            }
            $aec.printf("ambiguous group for", TALK.GREEN, $name);
            return {status:"ambiguous", matches:NOBJS};
        }

        
        //could be a company already! (like cfr)
        if ($bCOMPANYOK==true) {
             for (var i=0; i < MEM.company.ins.length; i++) {
                if (MEM.company.ins[i].type!=TYPE_BE) continue;
                if (MEM.company.ins[i].r!=MEM.inst.temp_nid) continue;
                var thisCompanyNOBJ  = MEMG.getNeuronByTempNid( MEM.company.ins[i].f);
                if (thisCompanyNOBJ==null) continue;
                 
                if (thisCompanyNOBJ.text.toLowerCase()==name) {
                    $aec.printf("found existing corp (for group) of", TALK.YELLOW, $name);
                    return {status:"good", match:thisCompanyNOBJ};
                }
            }   
        }
        
        
        var ghostNOBJ = await GHOSTDATA.makeGhostNeuron(MEM.group, $name, NTYPE_INST);
        $aec.addGroup(ghostNOBJ);
        $aec.printf("added new group", TALK.BLUE, $name);
        return {status:"good", match:ghostNOBJ};
    }

    
    /**
     * 
     * @param {*} $aec 
     * @param {*} $name 
     * @returns 
     */
    getOrMakeIfNewReligion=async($aec, $name)=> {

        //converting zionism in parses, if a religion..
        //to judaism.
        
        if (name=="zionsim") {
            console.warn("DO NOT KEEP CREATING THIS!!!!");
            name = "judaism";    
            return {status:"good", match:ANNE.MEM.judaism};
        }

        if ($aec.hasReligion($name)==true) {
            return {status:"good", match:$aec.getReligion($name)};
        }
        var name = $name.toLowerCase();
        for (var i=0; i < MEM.religion.ins.length; i++) {
            if (MEM.religion.ins[i].type!=TYPE_BE) continue;
            if (MEM.religion.ins[i].r!=MEM.inst.temp_nid) continue;
            var thisReligionNOBJ  = MEMG.getNeuronByTempNid( MEM.religion.ins[i].f);
            if (thisReligionNOBJ==null) continue;
            if (thisReligionNOBJ.text.toLowerCase()==name) {
                $aec.printf("found existing religion of", TALK.YELLOW, name);
                return {status:"good", match:thisReligionNOBJ};
            }
            if (thisReligionNOBJ.aka_lower!=null && thisReligionNOBJ.aka_lower.length>0) {
                if (thisReligionNOBJ.aka_lower.indexOf(name)!=-1) {
                    $aec.printf("found existing religion (by alias) of", TALK.YELLOW_i, name);
                    return {status:"good", match:thisReligionNOBJ};
                }
            }
        }
        
        var NOBJS = await  CHASE.queryFoNeuronByClassAndRelon(MEM.religion, $name, null, null, null);
        if (NOBJS.length>0) {
            if (NOBJS.length==1) {
                $aec.printf("found existing religion of", TALK.YELLOW, name);
                return {status:"good", match:NOBJS[0]};
            }
            $aec.printf("ambiguous religion for", TALK.GREEN, name);
            return {status:"ambiguous", matches:NOBJS};
        }
        return {status:"no-new", matches:[], match:null};

        var ghostNOBJ = await GHOSTDATA.makeGhostNeuron(MEM.religion, $name, NTYPE_INST);
        $aec.addReligion(ghostNOBJ);
        return {status:"good", match:ghostNOBJ};
    }

    
    
    /**
     * Make a group event.
     * If we have a year in the group event name, extract it (keep it in the name)
     * and use that for a year relon from the group event.
     * 
     * If $bSTRIPSPOOKSNUMB=true and we have a number < 10 strip it off the neuron name.  We don't want that in the name
     * as it is from wikispooks to split up the pages.
     * @param {*} $groupEventName 
     * @returns 
     */
    getOrMakeIfNewGroupEvent=async($aec, $name, $bSTRIPSPOOKSNUMB=true)=>{
        var parts = $name.split("/");
        if (parts.length==0) parts = $name.split(" ");
        var makeYrRelonForYear = 0;
        var name = $name;
        if (parts.length>1) {
            var yrtag = parts[ parts.length-1];
            var yr = parseInt(yrtag);
            if (Number.isInteger(yr)==true) {
                //this is a tag they use to split up the page.  Ignore it!
                if ($bSTRIPSPOOKSNUMB==true && yr<10) {
                    parts.pop();
                    name = parts.join(" ");
                }
                
                //this will need a yr relon!
                else {
                   // console.log("tack on a yr "+yr+"relon to this event -- or wait until it is queried??");
                    name = parts.join(" ");
                    makeYrRelonForYear=yr;
                }
            }
            else name = parts.join(" ");
        }

        var nameLower = name.toLowerCase();
        if ($aec.hasGroupEvent(nameLower)==true) {
            return {status:"good", match:$aec.getGroupEvent(nameLower)};
        }
        for (var i=0; i < MEM.group_event.ins.length; i++) {
            if (MEM.group_event.ins[i].type!=TYPE_BE) continue;
            if (MEM.group_event.ins[i].r!=MEM.inst.temp_nid) continue;
            var groupEventNOBJ  = MEMG.getNeuronByTempNid( MEM.group_event.ins[i].f);
            if (groupEventNOBJ==null) continue;
            if (groupEventNOBJ.text.toLowerCase()==nameLower) {
                $aec.printf("found existing group event of", TALK.YELLOW, name);
                return {status:"good", match:groupEventNOBJ};
            }
            if (groupEventNOBJ.aka_lower!=null && groupEventNOBJ.aka_lower.length>0) {
                if (groupEventNOBJ.aka_lower.indexOf(nameLower)!=-1) {
                    $aec.printf("found existing group event (by alias) of", TALK.YELLOW_i, name);
                    return {status:"good", match:groupEventNOBJ};
                }
            }
        }

        //ANY RESULT MEANS THIS ALREADY EXISTS.
        var NOBJS = await  CHASE.queryFoNeuronByClassAndRelon(MEM.group_event, name, null, null, null);
        if (NOBJS.length>0) {
            var NOBJSFILTERED=[];

            //if we have year, we can have a match with a non-year..
            //but not a match with a different year.
            if (makeYrRelonForYear>0) {

                //how can this ever be null?
                var yrNOBJ = await MEMG.getOrMakeIfNewInteger($aec, makeYrRelonForYear);
                NOBJSFILTERED = MEMG.getFilteredRelnToCANTCONTRADICT(NOBJS, [{reln:MEM.role_when.nid, to_s:yrNOBJ.nid}]);
            }
            else NOBJSFILTERED = NOBJS;
            
            if (NOBJSFILTERED.length==1) {
                $aec.printf("found existing group event of", TALK.YELLOW, name);
                return {status:"good", match:NOBJSFILTERED[0]};
            }
            $aec.printf("ambiguous group event for", TALK.GREEN, name);
            return {status:"ambiguous", matches:NOBJSFILTERED};
        }

        var ghostNOBJ = await GHOSTDATA.makeGhostNeuron(MEM.group_event, name, NTYPE_INST);
        //console.log("%cmade new group event %c"+ghostNOBJ.text+ "%c"+ghostNOBJ.nid, TALK.bggreen, TALK.bgblack, TALK.bggred);
        if (makeYrRelonForYear>0) {
            var yrNOBJ = await MEMG.getOrMakeIfNewInteger($aec, makeYrRelonForYear);
            if ((yrNOBJ instanceof anne_neuron) == true) {  
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, ghostNOBJ, MEM.role_when.nid, TO_TYPE_ANNE, yrNOBJ.nid);
            }
            else console.log("not a integer neuron");
        }
        $aec.addGroupEvent(ghostNOBJ);
        $aec.printf("added new group event", TALK.BLUE, name);
        return {status:"good", match:ghostNOBJ};
    }


    //confirms a NOBJ doesn't contradict a value.  (no value is ok)
    //if non-contradicting confirming value exists, will return on first full match if flag is set.
    //example.. WEF MEETING 2012, WEF MEETING,  WEF MEETING 2013
    
    //we get 3 neurons  (all above) and are looking to confirm which has a yr relon pting to 2013.
    //2nd is a non-contradicting hit, but 3rd is a direct hit.
    getFilteredRelnToCANTCONTRADICT=function($NOBJS, $partRels, $bRETURNFIRSTMATCH=true) {
        var finalNOBJS = [];

        for (var n=0; n < $NOBJS.length; n++) {
            var nObj = $NOBJS[n];
            var bNOBJ_NO_GOOD = true;

            var found_direct = 0;
            //loke, role_when-> to a specific year nid.
            for (var p=0; p < $partRels.length; p++) {
                var pRel = $partRels[p];
                var bREL_CONTRADICTS=false;
                for (var r=0; r < nObj.outs.length; r++) {
                    if (nObj.outs[r].reln==pRel.reln) {
                        if (nObj.outs[r].to_s == pRel.to_s) {
                            bREL_CONTRADICTS=false;
                            found_direct+=1;
                            break;
                        }
                        else {
                            bREL_CONTRADICTS = true;
                            break;
                        }
                    }    
                }
                if (bREL_CONTRADICTS==true) {
                    bNOBJ_NO_GOOD = true;
                    break;
                }

                //continue checking any other prel requirements that can't contradict.
                else {
                    bNOBJ_NO_GOOD = false;
                }
            }

            if (bNOBJ_NO_GOOD==false) {
                if ($bRETURNFIRSTMATCH==true && found_direct==$partRels.length) {
                    return [nObj];
                }
                finalNOBJS.push(nObj);
            }
            continue;
        }
        return finalNOBJS;
    }
    

    
    /** 
     * No allowing for ambiguous occupations.
     * We shouldn't have an issue here.
     * 
     * Creating a class for occupation is a very silly idea.
     * Just use the generalized concept classified under EC class.
     * And wire it up in a more flat way from there.
     * 
     * This saves us loads of taxonomy dramas (same with locations and other classes)
     * 
     * FLAT LINEAGES are far superior in ANNE.
     * 
     * SHOULD NEVER HAVE CLASSIFIED ANYTHIGN AS OCCUPATION.  NO NO NO NO NO NO NO.
     * Lateral BE, ok.
     */
    getOrMakeIfNewOccupationClass=async($aec, $occupation)=> {
        if($aec.hasOccupation($occupation)==true) {
            return {status:"good", match:$aec.getOccupation($occupation)};
        }

        var occupation = $occupation.toLowerCase();
        
        var all = MEM.ec_living.ins;
        all = all.concat(MEM.ec_nonliving.ins);
        all = all.concat(MEM.ec_thing.ins);

        for (var i=0; i < all.length; i++) {
            if (all[i].type!=TYPE_BE) continue;
            if (all[i].r!=MEM.ec_is.temp_nid) continue;
            var livingThingClass = MEMG.getNeuronByTempNid( all[i].f);
            if (livingThingClass==null) continue;
            if (livingThingClass.text.toLowerCase()==occupation) {
                $aec.printf("found existing occupation of", TALK.YELLOW, $occupation);
                return {status:"good", match:livingThingClass};
            }
            if (livingThingClass.aka_lower!=null && livingThingClass.aka_lower.length>0) {
                if (groupEventNOBJ.aka_lower.indexOf(nameLower)!=-1) {
                    $aec.printf("found existing occupation (by alias) of", TALK.YELLOW_i, $occupation);
                    return {status:"good", match:livingThingClass};
                }
            }
        }

        //$classNObj, $text, $alt_be_reln, $alt_be_to
        var NOBJS = await CHASE.queryFoNeuronByAltBeTexAndBe(MEM.ec_living, $occupation, MEM.ec_is.nid, MEM.occupation.nid);
        if (NOBJS.length) {
            if (NOBJS.length==1) {
                $aec.printf("found existing occupation of", TALK.YELLOW, $occupation);
                return {status:"good", match:NOBJS[0]};
            }
            $aec.printf("ambiguous occupation for", TALK.GREEN, $occupation);
            return {status:"ambiguous", matches:NOBJS};
        }

        //this will make a first name + last name alias if there is no suffix
        var ghostNOBJ = await GHOSTDATA.makeNewOccupation($occupation);
        $aec.addOccupation(ghostNOBJ);
        $aec.printf("added new occupation", TALK.BLUE, $occupation);
        return {status:"good", match:ghostNOBJ};
    }
    
    
    getOrMakeIfNewLocation=async($name, $po)=> {
    }

    getOrMakeIfNewEventWithYear=async($name, $year)=> {
    }

    getOrMakeIfNewEventWithStartDate=async($name, $startDate)=> {
    }

    getOrMakeIfNewEventWithEndDate=async($name, $endDate)=> {
    }

    //we need to pass an array of roles here.
    //then on the annode we process the roles into an array,
    //extract them..
    //then search for them in the order the come.
    //first role should be the most refining to give us the fewest hits.
    getOrMakeIfNewClumpWithRoles=async($roles)=>{
    }
    

    //alias function.
    getOrMakeNid = function($nid, $name="", $ntype=NTYPE_UNKNOWN, $fireWeight=.03) {
        return this.getOrMakeIfNew($nid, $name, $ntype, $fireWeight);
    }

    /**
     * Get a neuron by nid or make it if it doesn't exist.
     *
     * @author {Mr.Scatman} 
     * 
     */
    getOrMakeIfNew = function ($nid, $name="", $ntype=NTYPE_UNKNOWN, $fireWeight=.03) {
        if ($nid==null || $nid=="") return null;
        if (parseInt($nid) < 100000000) {
            console.warn("HIGHLY UNLIKELY THIS IS A NID");
        }
        var nObj = MEMG.getNeuronByNid($nid, $fireWeight);
        if (nObj==null) nObj = new anne_neuron($name, $ntype, $nid);
        return nObj;
    }


    getFutureTriggers=function($outs) {
        var rels = [];
        var h = ANNE.HISTORY.getGLKB();
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].type==TYPE_BRAIN120 && parseInt($outs[r].sys_ref_txid)>h) {
                rels.push($outs[r]);
            }
        }
        return rels;
    }

    getRelonsByType=function($outs, $type) {
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].type==$type) rels.push($outs[r]);
        }
        //ANNE.TALK.printRelons($outs);
        return rels;
    }

    getPatternRels = function($outs){
        var rels = [];
        for(var r=0; r < $outs.length; r++) {
            if ($outs[r].r!=MEM.ec_pattern.temp_nid) continue;
            rels.push($outs[r])
        }
        rels.sort(ANNE.UTILS.sortByDimension);
        return rels;
    }


    /** 
     * Wrapper to to geta quick parameter value.
     * 
     * @author {Mr.Scatman} 
     */
    getParamVal = function($outs) {
        var rel = MEMG.getParamNameRelon($outs);
        if (rel!=null) return rel.to_s;
        return "";
    }

    /** 
     * Wrapper to to geta quick parameter relon.
     * This was moved to the SYS dimension (no longer in the TYPE_HAS)
     * 
     * @author {Mr.Scatman} 
     */
    getParamNameRelon = function($outs) {
        return MEMG.getRelonFromWithRToType($outs, TYPE_SYS, MEM.PARAM_NAME.temp_nid, TO_TYPE_STRING);
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getParamRelon = function($outs, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_PARAM) return $outs[r];
        }
        return null;
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getParamRelons = function($outs, $relnTnid) {
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_PARAM) rels.push($outs[r]);
        }
        return rels;
    }

    getAllParamRelons = function($outs) {
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].type==TYPE_PARAM) rels.push($outs[r]);
        }
        return rels;
    }


    /**
     * A quicker way to get any kinfolk relations..
     * as some of the relns for kinfolk are sub-classes of higher level generalized classes (like parent or child)
     * @param {*} $outs 
     * @returns 
     */
    getKinfolk=function($outs){
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            if (MEM.kinfolk_tnids.indexOf($outs[r].r)==-1) continue;
            rels.push($outs[r]);
        }
        return rels;
    }


    getHasTo = function($r, $to_s) {
        for (var r=0; r < ANNE.MEM._aRelons.length; r++) {
            if(ANNE.MEM._aRelons[r].r!=$r) continue;
            if (ANNE.MEM._aRelons[r].to_s!=$to_s) continue;
            return ANNE.MEM._aRelons[r];
        }
        return null;
    }
    

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getHasRelon=function($outs, $relnTnid) { return this.getHas($outs, $relnTnid);}

    getEarliestHasRelon=function($outs, $relnTnid) { 
        var relon = null;    
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_HAS)  {
                if (relon!=null && $outs[r].time < relon.time) relon = $outs[r];
                relon = $outs[r];
            }
        }
        return relon;
    }


    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getHas=function($outs, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_HAS) return $outs[r];
        }
        return null;
    }
    getRelon=function($outs, $type, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].type==$type && $outs[r].r==$relnTnid && $outs[r].type==TYPE_HAS) return $outs[r];
        }
        return null;
    }
    getHasNObj=function($outs, $relnTnid) {
        var rel = this.getRelon($outs, TYPE_HAS, $relnTnid);
        if (rel==null) return null;
        if (rel.to_type!=TO_TYPE_ANNE) return null;
        return ANNE.MEMG.getNeuronByNid(rel.to_s);
    }
    getHasInt=function($outs, $relnTnid) {
        var rel = this.getRelon($outs, TYPE_HAS, $relnTnid);
        if (rel==null) return null;
        if (rel.to_type==TO_TYPE_ANNE) return null;
        return parseInt(rel.to_s);
    }
    

    /**
     * typically the $rels is the .ins of the $to_s, but not always.
     * @param {*} $rels 
     * @param {*} $relnTnid 
     * @param {*} $toS 
     * @returns 
     */
    getHasRelonTo=function($rels, $relnTnid, $toS) {
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].r==$relnTnid && $rels[r].type==TYPE_HAS && $rels[r].to_s==$toS) return $rels[r];
        }
        return null;
    }

    //typically the $rels is the .ins of the $to_s, but not always.
    getHasRelonsTo=function($rels, $relnTnid, $toS) {
        var rels = [];
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].r==$relnTnid && $rels[r].type==TYPE_HAS && $rels[r].to_s==$toS) rels.push($rels[r]);
        }
        return rels;
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getHasRelons=function($outs, $relnTnid) { 
        var matches = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_HAS) matches.push($outs[r]);
        }
        return matches;
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getSysMsgRelon=function($outs, $relnTnid) { 
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS_MSG) return $outs[r];
        }
        return null;
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getSysMsgRelon_HIGHEST=function($outs, $relnTnid) { 
        var highest_height = -1;
        var latest_rel = null;
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS_MSG) {
                if ($outs[r].time==0) return $outs[r].time;
                if ($outs[r].time>highest_height) {
                    latest_rel = $outs[r];
                    highest_height = latest_rel.time;
                }
            }
        }
        return latest_rel;
    } 

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getSysMsgRelons=function($outs, $relnTnid) { 
        var matches = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS_MSG) matches.push($outs[r]);
        }
        return matches;
    }

    
    /**
     * 
     * @param {*} $rels 
     * @returns 
     */
    getInternalPayments = function($rels) {
        var rels = [];
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].type==TYPE_INTERNAL_PAY) rels.push($rels[r]);
        }
        return rels;
    }


    /**
     * For streampays with an owner src --
     * The payment comes out of the main identity account but is tagged with a sysreftxid of 
     * relon that froze the fund. (and activated it)
     * @param {*} $rels 
     * @param {*} $sysreftxid 
     * @returns 
     */
    getInternalPaymentsOutWithSYSREFID = function($rels, $sysreftxid) {
        var rels = [];
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].type==TYPE_INTERNAL_PAY && $rels[r].sys_ref_txid==$sysreftxid) rels.push($rels[r]);
        }
        return rels;
    }
    

    /**
     * typically the $rels is the .ins of the $to_s, but not always.
     * @param {*} $rels 
     * @param {*} $relnTnid 
     * @param {*} $toS 
     * @returns 
     */
    getSysMsgRelonsTo=function($rels, $relnTnid, $toS) {
        var rels = [];
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].r==$relnTnid && $rels[r].type==TYPE_SYS_MSG && $rels[r].to_s==$toS) rels.push($rels[r]);
        }
        return rels;
    }

    /**
     * 
     * @param {*} $rels 
     * @param {*} $relnTnid 
     * @param {*} $toS 
     * @returns 
     */
    getSysMsgRelonTo=function($rels, $relnTnid, $toS) {
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].r==$relnTnid && $rels[r].type==TYPE_SYS_MSG && $rels[r].to_s==$toS) return $rels[r];
        }
        return null;
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $relnTnid 
     * @returns 
     */
    getAwarenessRelons=function($outs, $relnTnid) { 
        var matches = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_AWARENESS) matches.push($outs[r]);
        }
        return matches;
    }

    
    /**
     * Get the to value of a TYPE_HAS relon.
     * @param {array} - outs
     * @param {integer} - int of temp_nid for the reln  (in js it is just a number)
     */
    getHasValue=function($outs, $relnTnid) {
        var rel = MEMG.getRelonFromWithRToType($outs, TYPE_HAS, $relnTnid, TO_TYPE_ANNE);
        if (rel==null) return null;
        var val = this.__MASTER_DISPLAY_RELON_TO(rel);
        if (val=="") val=null;
        return val;
    }

    /**
     * 
     * @param {*} $altBe 
     */
    isAltBe=function($altBe){
        //if (this['be_'+$altBeNid]!=undefined && 
    }

    /**
     * 
     * @param {*} $outs 
     * @param {*} $toTnid 
     * @returns 
     */
    hasAltBe=function($outs, $toTnid) {
        var rels = MEMG.getRelonsFromWithRToType($outs, TYPE_BE, MEM.ec_is.temp_nid, TO_TYPE_ANNE);
        if (rels.length==0) return null;
        for (var r=0; r<rels.length; r++) { if (resl[r].t==$toTnid) return true; }
        return false;
    }


    /**
     * this is just the parent-line be-params.
     * This is going to have to CHASE data too.. (aaarg)  
     * We have no guarantee the user has the parent line relons in memory. 
     * We must query for them first (all the way on the way up). 
     * 
     * @author {Mr.Scatman} 
     */
     getParentLineParams=function($nObj, $relnMatching="", $bDOSELF=true) {
        var nObjs = MEMG.buildParentLine($nObj, $bDOSELF); //just the be line
        ANNE.TALK.printNeurons(nObjs);

        var allParamRelons = [];
        if (nObjs.length==0) return allParamRelons;
        for (var  n=0; n <nObjs.length; n++) {
            var nObj = nObjs[n];
            var aRelons = MEMG.getAllParamsOf(nObj.outs, $relnMatching);
            if (aRelons) {
                MEMG.concatUniqueRelnToPrioritize(allParamRelons, aRelons, $nObj.temp_nid);
            }
        }
        return allParamRelons;
    }


    /** 
     * Heavily used to build relon tables.
     * 
     * Parentline, inherit. (for searching to find this function since I keep forgetting the name)
     * 
     * @author {Mr.Scatman} 
     */
    getParentLineRelons = function($nObj, 
                                    $types=[TYPE_SYS, TYPE_BE, TYPE_PARAM, TYPE_HAS, TYPE_ROLE, TYPE_CLUMP, TYPE_BRAIN], 
                                    $ignoreRelnTnids=[],
                                    $bGETALLSELF=false,
                                    $bEXTRAPRETURN=false,
                                    $bFORINHERITANCE=false,
                                    $ignoreToTnids=[]) {
        
        if ($nObj==null) {
            console.log("nObj shouldn't be null here");
            return [];
        }
        
        var nObjs = MEMG.buildParentLine($nObj, true, false, null, $ignoreToTnids); 

        if ($bGETALLSELF==true) nObjs.unshift($nObj);
        
        var all = new Object();
        all.sys = [];
        all.bes = [];
        all.params = [];
        all.has = [];
        all.clumps = [];
        all.roles = [];
        all.brain = [];
        all.sys_msg = [];

        all.recurse = [];

        if (nObjs.length==0) return all;    //an array of arrays indexed by TYPE

        var bDOINGSYS = false;
        var bDOINGBE  = false;
        var bDOINGPARAM = false;
        var bDOINGHAS = false;
        var bDOINGCLUMPS = false;
        var bDOINGROLES = false;
        var bDOINGBRAIN = false;
        var bDOINGSYSMSG = false;

        if ($types.indexOf(TYPE_SYS)!=-1) bDOINGSYS=true;
        if ($types.indexOf(TYPE_BE)!=-1) bDOINGBE=true;
        if ($types.indexOf(TYPE_PARAM)!=-1) bDOINGPARAM=true;
        if ($types.indexOf(TYPE_HAS)!=-1) bDOINGHAS=true;
        if ($types.indexOf(TYPE_CLUMP)!=-1) bDOINGCLUMPS=true;
        if ($types.indexOf(TYPE_ROLE)!=-1) bDOINGROLES=true;
        if ($types.indexOf(TYPE_BRAIN)!=-1) bDOINGBRAIN=true;
        if ($types.indexOf(TYPE_SYS_MSG)!=-1) bDOINGSYSMSG=true;
        
        //we must get final garden first -- we inherit its params...
        //the ONLY time we inherit from a non-TYPE_BE relon's line.
        for (var  n=0; n <nObjs.length; n++) {
            var nObj = nObjs[n];
            if (bDOINGSYS==true) {
                MEMG.gplr_final_sys(all.sys, nObj, $ignoreRelnTnids, $nObj);
            }
        }

        var gt = $nObj._gt;

        //we need to snag the PARAMS of this known garden type.
        //we also need to snag any SYS relons that apply to those params  (like multiples ok)
        //they apply to the neuron and help users best manage their garden (by type)
        if (gt!=null) {
            var finalGardenNOBJ = MEMG.getNeuronByTempNid( gt.t );
            if (finalGardenNOBJ!=null) {
                var gardenParams = [];
                gardenParams = MEMG.getAllParamsOf(finalGardenNOBJ.outs);
                MEMG.concatUniqueRelnToPrioritize(all.params, gardenParams, gt.f);
                var gardenSys = MEMG.getAllSysOf(finalGardenNOBJ.outs);
                var finalGardenSys = [];

                //this sys rel references a param rel.
                for (var gs=0; gs < gardenSys.length; gs++) {
                    var gsrel=gardenSys[gs];
                    if (gsrel.to_type!=TO_TYPE_TXID) continue;
                    if (ANNE.UTILS.findValInObjnArr(gardenParams, "txid", gsrel.to_s)!=null) {
                        finalGardenSys.push(gsrel);
                    }
                }
                MEMG.concatUniqueRelnToPrioritize(all.sys, finalGardenSys, finalGardenNOBJ.temp_nid);
            }
        }

        //sys msg is only on self (not inherited any up) and it is RARE we want to 
        //even see them.
        if (bDOINGSYSMSG==true) {
            var aRelons8 = MEMG.getAllSysMsgOf($nObj.outs, $ignoreRelnTnids);
            if (aRelons8.length) MEMG.concatUniqueRelnToPrioritize(all.sys_msg, aRelons8, $nObj.temp_nid);
        }

        //snag then process the rest off the relons (all the non-SYS relons)
        //remember, our nObjs here are parent line neurons... so we have to get all
        //inherited info from them and tabulate it according to ANNE DATA PROTOCOL RULES.
        for (var  n=0; n <nObjs.length; n++) {
            var nObj = nObjs[n];
            //console.log("chasing upline "+nObj.text+" nid:"+nObj.nid);


            //we will have to further re-filter the bes after we get them all.
            if (bDOINGBE==true) {
                var aRelons2 = MEMG.getAllBesOf(nObj.outs, $ignoreRelnTnids);
                if (aRelons2.length) MEMG.concatUniqueRelnToPrioritize(all.bes, aRelons2, $nObj.temp_nid);
            }

            //we will have to further re-filter the bes after we get them all.
            if (bDOINGPARAM==true) {   
                var aRelons3 = MEMG.getAllParamsOf(nObj.outs, "", $ignoreRelnTnids);

                if (ANNE.MEM.ec_thing.nid == nObj.nid) {
                    //console.log(nObj.text);
                    //ANNE.TALK.printRelon(aRelons3);
                }
                
                if (aRelons3.length) MEMG.concatUniqueRelnToPrioritize(all.params, aRelons3, $nObj.temp_nid);
            }

            if (bDOINGHAS==true) {
                //we should skip this entirely if the nObj is 
                //not a type_be -> polity -> (here) or type_be -> MEM.BE -> (here)
                //or a special defined BE_CHILDREN
                //FOR NOW, skip it if not the main neuron.
                if (nObj.temp_nid==$nObj.temp_nid) {
                    if ($bFORINHERITANCE==true) MEMG.gplr_final_has(all.has, nObj, $ignoreRelnTnids, $nObj);
                    else  MEMG.gplr_final_has_FOR_CONTEXT(all.has, nObj, $ignoreRelnTnids, $nObj);
                }
            }

            if (bDOINGCLUMPS==true) {
                var aRelons5 = MEMG.getAllClumpsOf(nObj.outs, $ignoreRelnTnids);
                if (aRelons5.length) MEMG.concatUniqueRelnToPrioritize(all.clumps, aRelons5, $nObj.temp_nid);
            }

            //we do not inherit roles (makes no sense)
            //we DO inherit clumps.. HOWEVER, a CLUMP dimension (type) should be abandoned and 
            //HAS/BE/PARAM relns point to the clump with the desired reln.
            //this will help with easier handling since we have HAS/BE/PARAM shorthand.
            if (bDOINGROLES==true) {
                var aRelons6 = MEMG.getAllRolesOf(nObj.outs, $ignoreRelnTnids);
                if (aRelons6.length) MEMG.concatUniqueRelnToPrioritize(all.roles, aRelons6, $nObj.temp_nid, false);
            }

            if (bDOINGBRAIN==true) {
                var aRelons7 = MEMG.getAllBrainsOf(nObj.outs, $ignoreRelnTnids);
                if (aRelons7.length) MEMG.concatUniqueRelnToPrioritize(all.brain, aRelons7, $nObj.temp_nid);
            }
        }   

        //final pass filter to apply ANNE protocol tree/inheritance.
        if (bDOINGBE==true) {
            if ($bFORINHERITANCE==true) MEMG.gplr_final_bes(all, $nObj);
            //else do nothing if doing context (keep all bes)
        }
        if (bDOINGPARAM==true) MEMG.gplr_final_params(all, $nObj);


        //turn this off for now.
        if (all.recurse.length>0 && 2==1) {
            var recurse = all.recurse.concat();
            all.recurse = [];
            for (var r=0; r < recurse.length; r++) {
                var rtnidskips = MEM.delegation_reln_tnids.concat();
                rtnidskips.push(MEM.name.temp_nid);

                rtnidskips.push(MEM.constituent_of.temp_nid);
                rtnidskips.push(MEM.eval.temp_nid);
                rtnidskips.push(MEM.BE_CHILDREN.temp_nid);
                rtnidskips.push(MEM.BE_PARENT.temp_nid);
                rtnidskips.push(MEM.aka.temp_nid);
                rtnidskips.push(MEM.fka.temp_nid);

                //we want to recurse on this TYPE_BE ec_is alt class but we don't want to 
                //chase ANNE SYS CLASSES, GENERAL PROPS or SYSBRAIN heirarchies.
                //We want anything below those lvls, as they apply to alt be connections.
                var all2 = this.getParentLineRelons(recurse[r],[TYPE_PARAM], 
                    rtnidskips,
                    false,
                    false,
                    false,
                    [MEM.__ANNE_SYS_CLASS.temp_nid, MEM._GENERAL_PROP_CLASS.temp_nid, MEM.__SYSBRAIN.temp_nid]);

                //if (bDOINGSYS==true && all2.sys.length) MEMG.concatUniqueRelnToPrioritize(all.sys, all2.sys, $nObj.temp_nid, true);
                //if (bDOINGBE==true && all2.bes.length) MEMG.concatUniqueRelnToPrioritize(all.bes, all2.bes, $nObj.temp_nid, true);
                if (bDOINGPARAM==true && all2.params.length) MEMG.concatUniqueRelnToPrioritize(all.params, all2.params, $nObj.temp_nid, true);
                //if (bDOINGHAS==true && all2.has.length) final = MEMG.concatUniqueRelnToPrioritize(all.has, all2.has, $nObj.temp_nid, true);
                //if (bDOINGCLUMPS==true && all2.clumps.length) MEMG.concatUniqueRelnToPrioritize(all.clumps, all2.clumps, $nObj.temp_nid, true);
                //if (bDOINGROLES==true && all2.roles.length) MEMG.concatUniqueRelnToPrioritize(all.roles, all2.roles, $nObj.temp_nid, true);
                //if (bDOINGBRAIN==true && all2.brain.length) MEMG.concatUniqueRelnToPrioritize(all.brain, all2.brain, $nObj.temp_nid, true);
            }    
        }

        if ($bEXTRAPRETURN==true) {
            var final = [];
            if (bDOINGSYS==true) final = all.sys;
            if (bDOINGBE==true && all.bes.length) final = final.concat(all.bes);
            if (bDOINGPARAM==true && all.params.length) final = final.concat(all.params);
            if (bDOINGHAS==true && all.has.length) final = final.concat(all.has);
            if (bDOINGCLUMPS==true && all.clumps.length) final = final.concat(all.clumps);
            if (bDOINGROLES==true && all.roles.length) final = final.concat(all.roles);
            if (bDOINGBRAIN==true && all.brain.length) final = final.concat(all.brain);
            if (bDOINGSYSMSG==true && all.sys_msg.length) final = final.concat(all.sys_msg);
            return final;
        }

        
        return all;
    }

    getPrimaryParentRel = function($nObj) {
        for (var r=0; r < $nObj.outs.length; r++) {
            if ($nObj.outs[r].type!=TYPE_BE) continue;
            if ($nObj.outs[r].r!=MEM.class.temp_nid && $nObj.outs[r].r!=MEM.inst.temp_nid) continue;
            return $nObj.outs[r];
        }
        return null;
    }

    
    /**
     */
    getUserRevealsFor = async($pdnNOBJ)=>{
        var reveals = [];
        var lookups = [];
        var urels =[];

        //get reveals, look at the sys ref txid FROM to ascertain what PDN this was for.
        for(var u=0; u < USER._userNObj.outs.length; u++) {
            var urel=USER._userNObj.outs[u];
            if (urel.type!=TYPE_SYS_MSG) continue;
            if (urel.sys_ref_txid==undefined) continue;
            if (urel.r!=MEM.PDN_ANSWER.temp_nid) continue;
            var refRel = MEMG.lookupByTXID(urel.sys_ref_txid);
            
            if (refRel==null) {
                if (lookups.indexOf(urel.sys_ref_txid)==-1) {
                    lookups.push(urel.sys_ref_txid); 
                    urels.push(urel);
                }
                continue;
            }
            if (refRel.f!=$pdnNOBJ.temp_nid) continue;
            urel.sys_ref_relon = refRel;
            reveals.push(urel);
        }

        if (lookups.length) await CHASE.lookupTxids(lookups);
        for (var u=0; urels<urels.length; u++) {
            var urel=USER._userNObj.outs[u];
            var refRel = MEMG.lookupByTXID(urel.sys_ref_txid);
            if (refRel==null) continue;
            if (refRel.f!=$pdnNOBJ.temp_nid) continue;
            urel.sys_ref_relon = refRel;
            reveals.push(urel);
        }

        if (reveals.length) {
            return await USER.decryptPDNAnswerRelons(reveals);
        }
        return reveals;
    }


    /**
     * @author {Mr.Scatman} 
     */
    getVoteData=function($nid) {
        var nObj = MEMG.getNeuronByNid($nid);
        var votes = [];
        for (var r=0; r < MEM._aRelons.length; r++) {
            if (MEM._aRelons[r].r==nObj.temp_nid) { //&& MEM.valid_vote.indexOf(MEM._aRelons[r].t)!=-1
                var vNobj = MEMG.getNeuronByTempNid(MEM._aRelons[r].f);
                votes.push(vNobj);
            }
        }
        return votes;
    }


    /**
     * Critical final prep to params array of a neuron applying ANNE protocol rules.
     * @author {Mr.Scatman} 
     */
    gplr_final_params = function($allObj, $nObj) {

        //is temp_nid the parent of ANY of the tempNids?
        function matchAnyParent($temp_nid, $tempNids) {
            var tempNOBJ = MEMG.getNeuronByTempNid($temp_nid);
            
            //we want this relon to make it through!
            for (var t=0; t < $tempNids.length; t++) {
                var testNOBJ = MEMG.getNeuronByTempNid($tempNids[t]);
                if (testNOBJ==null) continue;
                if (testNOBJ.be_lvl==-1) continue;  //no parent info on this
                //our temp is a high lvl # (lower on the tree) than this tempNid  (whatever it is)..
                //so it cannot be an ancestor parent. 
                if (tempNOBJ.be_lvl>= testNOBJ.be_lvl) continue; 
                var bes = testNOBJ.bes;
                if (bes==null) continue;
                var bRETURNTRUE=false;
                while (bes) {
                    var be = bes[0];
                    if ($temp_nid==be) {
                        bRETURNTRUE=true;
                        break;
                    }
                    testNOBJ = MEMG.getNeuronByTempNid(be);
                    if (testNOBJ==null) break;
                    if (testNOBJ.be_lvl==0 || testNOBJ.be_lvl==-1) break;
                    bes = testNOBJ.bes;
                }
                if(bRETURNTRUE==true) return true;
            }
            return false;
        }

        var params = $allObj.params;
        var final_params = [];
        var param_tnids = [];
        for (var p=0; p < params.length; p++) {
            if (param_tnids.indexOf(params[p].r)==-1) param_tnids.push(params[p].r);
        }

        //is this param reln NOBJ a hierarchical parent of any of the others?
        //if so, skip it.  We will keep the lowest (most refined) specification.
        //example:  MEME vs neuron img vs flag.  (keep flag)
        for (var p=0; p < params.length; p++) {
            var paramRelon = params[p];
            //ANNE.TALK.printRelon(paramRelon);

            //this is a PARAM that depends on a TYPE_HAS being
            //created from ANOTHER PARAM.  THAT TYPE_HAS will specify
            //the actual TO of this PARAM.
            if (paramRelon.to_type==TO_TYPE_TXID) {
                var cdxRelon = MEMG.lookupByTXID(paramRelon.to_s);
                if (cdxRelon==null) continue;
                
                //cdx relons are like variables.. they point to a value
                //we are attempting to get from another relon.
                if (cdxRelon.type!=TYPE_SYS) continue;

                var cdx_to_s = cdxRelon.to_s;
            
                //get the PARAM which points to this same to_s;
                var otherRelon = MEMG.getRelonFromWith(params, TYPE_PARAM, cdx_to_s);
                if (otherRelon==null) continue;

                paramRelon.depends_on = otherRelon.txid;
                if (otherRelon.affects==undefined) otherRelon.affects=[paramRelon.txid];
                else if (otherRelon.affects.indexOf(paramRelon.txid)==-1) otherRelon.affects.push(paramRelon.txid);
            }

            if (paramRelon.r!=paramRelon.t && matchAnyParent(paramRelon.r, param_tnids)==true) continue;
            
            final_params.push(paramRelon);
        }

        $allObj.params = final_params; //alter the params array to be the final_params array.
    }


    /**
     * Critical final prep to BE array of a neuron applying ANNE protocol rules.
     * @author {Mr.Scatman} 
     */
    gplr_final_bes = function($allObj, $nObj) {
        var bes = $allObj.bes;
        var final_bes = [];
        var skips = [];
        var temp_bes = [];
        for (var b=0; b < bes.length; b++) {
            var beRelon = bes[b];

            //on a BE ec_is, we want to inherit EVERYTHING.
            //this is a true alt heirarchy.
            if (beRelon.r==MEM.ec_is.temp_nid) {
                var beNOBJ = MEMG.getNeuronByTempNid(beRelon.t);
                $allObj.recurse.push(beNOBJ);
            }
            
            //if this is an inherited be.. we have some skips.
            if (beRelon.f!=$nObj.temp_nid) {

                continue;   //for now, skip all bes if not from target nobj
                if (beRelon.r==MEM.constituent_of.temp_nid) continue;   //NEVER INHERIT THIS.
                if (beRelon.r==MEM.class.temp_nid) continue;            //NEVER INHERIT THIS.
                if (beRelon.r==MEM.name.temp_nid) continue;             //NEVER INHERIT THIS.
                if (beRelon.r==MEM.eval.temp_nid) continue;             //NEVER INHERIT THIS.
                if (beRelon.r==MEM.BE_CHILDREN.temp_nid) {
                    skips.push(beRelon.t);
                    continue;
                }
                if (beRelon.r==MEM.BE_PARENT.temp_nid) continue;
                if (beRelon.r==MEM.aka.temp_nid) continue;
                if (beRelon.r==MEM.fka.temp_nid) continue;
                if (beRelon.r==MEM.ec_is.temp_nid) continue;
            }
            temp_bes.push(beRelon);
        }

        //we don't want the BE_CHILDREN relon of a alt parent..
        //ie, the be->polity->country is only relevant to the neuron it is
        //on.  It is NOT inherited by locations with an alt be parent of a country.
        //They will get their own polity assigned (or have none).
        for (var b=0; b<temp_bes.length;b++) {
            var beRelon = temp_bes[b];
            if (beRelon.f!=$nObj.temp_nid) {
                if (skips.indexOf(beRelon.r)!==-1) continue;
            }
            final_bes.push(beRelon);
        }
        $allObj.bes = final_bes;    //alter the bes array to be the final_bes array.
    }


    /**
     * Critical final prep to HAS array of a neuron applying ANNE protocol rules.
     * @author {Mr.Scatman} 
     */
    gplr_final_has=function($arr, $nObj, $ignoreRelnTnids, $mainNObj) {
        var aRelons = MEMG.getAllHasOf($nObj.outs, $ignoreRelnTnids);
        for (var r=0; r < aRelons.length; r++) {
            var newRel = aRelons[r];
            if (newRel.f != $mainNObj.temp_nid) {                             //if not the priority neuron (ie, an inherited relon)
                //no has from alt bes unless reln of type_be was BE or BE_CHILDREN/BE_PARENT
                //continue;   
                
                if (newRel.r==MEM.SMART_SEARCH_ONLY.temp_nid) continue; //do not inherit smart search
                if (newRel.r==MEM.primaryNeuronImg.temp_nid) continue;  //do not inherit neuron img
                if (newRel.r==MEM.neuronImg.temp_nid) continue;  //do not inherit neuron img
                
            }
            if ($arr.length) {
                var bDUPE=false;
                for (var x=0; x < $arr.length; x++) {
                    var existingRel = $arr[x];
                    //we have an inherited dupe.. take the priority one!
                    if (existingRel.r==newRel.r && existingRel.to_s == newRel.to_s) {
                        if (existingRel.f==$mainNObj.temp_nid) {
                            bDUPE = true;
                            break;
                        }
                        else { //swap out the existing with this one!
                            $arr[x]=newRel;
                            bDUPE = true;
                            break;
                        }
                    }
                }
                if (bDUPE==false) $arr.push(newRel);
                else continue;
            }
            else $arr.push(newRel);
        }
    }


    gplr_final_has_FOR_CONTEXT=function($arr, $nObj, $ignoreRelnTnids, $mainNObj) {
        var aRelons = MEMG.getAllHasOf($nObj.outs, $ignoreRelnTnids);
        for (var r=0; r < aRelons.length; r++) {
            var newRel = aRelons[r];
            if ($arr.length) $arr.push(newRel);
            else $arr.push(newRel);
        }
    }



    /**
     * Critical final prep to SYS array of a neuron applying ANNE protocol rules.
     * 
     * NOTE: most SYS relons do NOT get inherited by children insts and classes.
     * SYS is not a semantic dimension to define conceptual information  --- it is
     * intended to govern the use of the neuron in ANNE.
     * 
     * As such, most SYS will not be inherited.
     * 
     * Those that ARE inherited:
     *      GARDEN TYPE (with caveats to override inheritance)
     *      DATA DELEGATION RIGHTS  (some varieties)
     *      wv_let, wv_dney
     *      SYS->_dimension
     * 
     * @author {Mr.Scatman} 
     */
    gplr_final_sys=function($arr, $fromNOBJ, $ignoreRelnTnids, $mainNObj) {
        var aRelons = MEMG.getAllSysOf($fromNOBJ.outs, $ignoreRelnTnids);
        //ANNE.TALK.printRelons(aRelons);
        //console.log("%c -------------------------------------------------------------------", TALK.bgyellow);
        for (var r=0; r < aRelons.length; r++) {
            var newRel = aRelons[r];

            //main object takes precedence
            if (newRel.f==$mainNObj.temp_nid) {            

                //this must be the gt.
                //NOTE: _gt is only inherited from clases and relevant to itself & its sub-classes.
                //remove any other inherited gt relons.
                if ($mainNObj.ntype==NTYPE_CLASS && newRel.r == MEM.garden_type.temp_nid) {
                    $arr = ANNE.UTILS.removeFromRelonsArray($arr, MEM.garden_type.temp_nid); 
                    if ($mainNObj._gti && $mainNObj._gt.txid!=newRel.txid) {
                        console.log($mainNObj.text+" garden type relon "+newRel.txid);
                        ANNE.TALK.printRelon($mainNObj._gt);
                        ANNE.TALK.printRelon(newRel._gt);
                        
                        alert("gt audit problem, this should have already been set MEMP");
                    }
                    $mainNObj.setGtRelon(newRel);
                }
                
                //inhert inst_garden_type all the way down the BE branch.
                else if (newRel.r == MEM.inst_garden_type.temp_nid) {
                    $arr = ANNE.UTILS.removeFromRelonsArray($arr, MEM.inst_garden_type.temp_nid); 
                    if ($mainNObj._gti && $mainNObj._gti.txid!=newRel.txid) {
                        console.log($mainNObj.text+" garden type relon "+newRel.txid);
                        ANNE.TALK.printRelon($mainNObj._gti);
                        ANNE.TALK.printRelon(newRel._gti);
                        alert("gti audit problem, this should have already been set in the MEMP");
                    }
                    $mainNObj.setGtiRelon(newRel);
                }
            }

            else {                    
                //gt inheritance 
                //last (lowest on tree, highest be_lvl) NON-ALT-BE value.
                if (newRel.r == MEM.garden_type.temp_nid) {
                    if ($mainNObj._gt==null) {
                        //console.warn("why wasn't this already set?");
                        $mainNObj.setGtRelon(newRel);
                    }

                    //we have a gt already.  Is this more accurate?
                    else {
                        if ($mainNObj._gt.txid==newRel.txid) continue;  //already have it.
                        if ($mainNObj.getGtLvl() > $fromNOBJ.getGtLvl()) {
                            //console.log("new gt is higher up the tree (lower in lvl), not as accurate");
                            $arr = ANNE.UTILS.removeRelonByTxid($arr, newRel); //remove it from our $arr if it is in there
                            continue;   
                        }

                        //update gt.
                        else {
                            $mainNObj.setGtRelon(newRel);
                        }
                    } 
                }

                else if (newRel.r == MEM.inst_garden_type.temp_nid) {
                    if ($mainNObj._gti==null) {

                        //when we don't preload all neurons then gti is 
                        //is not going to necessarily be set.
                        //console.warn("why wasn't this already set?");
                        
                        
                        $mainNObj.setGtiRelon(newRel);
                    }

                    //we have a gti already.  Is this more accurate?
                    else {
                        if ($mainNObj._gti.txid==newRel.txid) continue;  //already have it.
                        if ($mainNObj.getGtiLvl() > $fromNOBJ.getGtiLvl()) {
                            //console.log("new gti is higher up the tree (lower in lvl), not as accurate");
                            $arr = ANNE.UTILS.removeRelonByTxid($arr, newRel);   //remove it from our $arr if it is in there
                            continue;   
                        }

                        //update gti.
                        else {
                            $mainNObj.setGtiRelon(newRel);
                        }
                    } 
                }

                else if (MEM.delegation_reln_tnids.indexOf(newRel.r)!=-1) { 
                }

                //applies to an inherited param so we can inherit this too.
                else if (newRel.r==MEM._multiples_ok.temp_nid) {
                }

                //roles should be in the role dimension... not in the SYS
                //fix this.
                else if (MEM.role.children.indexOf(newRel.r)!=-1) {
                    ANNE.TALK.printRelon(newRel);
                    console.log(newRel);
                    console.log("%c debug", TALK.bgbigred);
                    alert("R_ROLE in SYS dimension");
                } 
                else continue;

                if (newRel.r==MEM.insts_only.temp_nid && newRel.to_type==TO_TYPE_TXID) {
                    var paramRel = MEMG.lookupRelonByTXID(newRel.to_s);
                    if (paramRel!=null) paramRel.insts_only=true;
                }
            }

            if ($arr.length) {
                var bDUPE=false;
                for (var x=0; x < $arr.length; x++) {
                    var existingRel = $arr[x];
                    //we have an inherited dupe.. take the priority one!
                    if (existingRel.r==newRel.r && existingRel.to_s == newRel.to_s) {
                        if (existingRel.f==$mainNObj.temp_nid) {
                            bDUPE = true;
                            break;
                        }
                        else { //swap out the existing with this one!
                            $arr[x]=newRel;
                            bDUPE = true;
                            break;
                        }
                    }
                }
                if (bDUPE==false) $arr.push(newRel);
                else continue;
            }
            else $arr.push(newRel);
        }

        return;
    }


    /**
     * Get the the neuron of a given eval (.value) and class.
     * 
     * Return the best string or shortest (if bSHORTIF is on)
     * 
     * @param {anne_neuron} - the parent
     * @param {int} - the eval we want to find for a matching child
     * @author {Mr.Scatman} 
     * 
     */
    getNeuronNameByClassAndValue=function($classNOBJ, $eval, $bSHORTIF=true) {
        var nObj = MEMG.getNeuronByClassAndValue($classNOBJ, $eval);
        if ((nObj instanceof anne_neuron) == true) {
            if ($bSHORTIF==true) return nObj.bestNameHTML();
            else return nObj.shortName();
        }
        else return $eval;
    }
    
    getNeuronByClassAndValue=function($classNOBJ, $eval) {
        var ins = $classNOBJ.ins;
        if (ins.length==0) return $eval;
        for (var i=0; i < ins.length; i++) {
            var fromNOBJ = MEMG.getNeuronByTempNid(ins[i].f);
            if (fromNOBJ==null) continue;
            if (fromNOBJ.value==null || fromNOBJ.value=="") {
                var val = fromNOBJ.bestValue(fromNOBJ);
                if (val=="") continue;
                if (fromNOBJ.value==$eval) return fromNOBJ;
                else continue;
            }
            else if(fromNOBJ.value==$eval) return fromNOBJ;
        }
        return $eval;
    }


    /**
     */
    getBrainRels = function($outs) {
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            var rel = $outs[r];
            if (rel.type==TYPE_BRAIN) rels.push(rel);
            else if (rel.type >=20 && rel.type<120) rels.push(rel);
        }
        return rels;
    }

    /**
     */
    getPats = function($nObj, $bGETPARENTPATS=true) {
        var patNOBJs = [];
        var pats = MEMG.getRelonsFromWith_fast($nObj.outs, TYPE_BRAIN, MEM.ec_pattern.temp_nid, TO_TYPE_ANNE);
        if (pats==null) return [];

        for (var p=0; p < pats.length; p++) {
            var patNOBJ = MEMG.getNeuronByTempNid(pats[p].t);
            patNOBJs.push(patNOBJ);
        }

        //We aren't chasing up here
        //this is intended for one gen down or same-gen pats.
        if ($bGETPARENTPATS==true) {
            if ($nObj.bes.length==0) return patNOBJs;
            var parentNOBJ=MEMG.getNeuronByTempNid($nObj.bes[0]);
            pats = MEMG.getRelonsFromWith_fast(parentNOBJ.outs, TYPE_BRAIN, MEM.ec_pattern.temp_nid, TO_TYPE_ANNE);
            if (pats==null) return patNOBJs
            for (var p=0; p < pats.length; p++) {
                var patNOBJ = MEMG.getNeuronByTempNid(pats[p].t);
                patNOBJs.push(patNOBJ);
            }
        }
        return patNOBJs;
    }

    /**
     */
    getTextListFromNObjs= function($nObjs, $sep=" :: ") {
        var list = [];
        for (var n=0; n < $nObjs.length; n++) list.push ($nObjs[n].text);
        return list;
    }

    /**
     */
    getTextListFromTnids= function($tnids, $sep=" :: ") {
        var list = [];
        for (var t=0; t < $tnids.length; t++) {
            var nObj = MEMG.getNeuronByTempNid($tnids[t]);
            if (nObj.text!="") list.push (nObj.text);
            else list.push(nObj.nid);
        }
        return list;
    }

    /**
     */
    getTextListStrFromTnids= function($tnids, $sep=" :: ") {
        return MEMG.getTextListFromTnids($tnids, $sep).join($sep);
    }

    /**
     */
    getTextListStrFromNObjs= function($nObjs, $sep=" :: ") {
        return MEMG.getTextListFromNObjs($nObjs, $sep).join($sep);
    }

    hasRoleRelon=function($outs, $reln_nid, $role_value) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].reln==$reln_nid) {
                if ($outs[r].to_s==$role_value) return true;
            }
        }
        return false;
    }

    /**
     */
    getRoleRelon=function($outs, $temp_reln_nid) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].r==$temp_reln_nid) {
                if ($outs[r].type==TYPE_ROLE) return $outs[r];
            }
        }
        return null;
    }
    getRoleRelons=function($outs, $temp_reln_nid) {
        var rels = [];
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].r==$temp_reln_nid) {
                if ($outs[r].type==TYPE_ROLE) rels.push($outs[r]);
            }
        }
        return rels;
    }


    /**
     * This function probably exists in 10 other places.
     * @param {*} $outs 
     * @param {*} $temp_reln_nid 
     * @param {*} $type 
     * @returns 
     */
    getRelonByTypeAndR=function($outs, $temp_reln_nid, $type) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].r==$temp_reln_nid) {
                if ($outs[r].type!=$type) continue;
                return $outs[r];
            }
        }
        return null;
    }


    /**
     * Look for an in with a specific reln (unknown from)
     * @param {*} $ins 
     * @param {*} $temp_reln_nid 
     * @param {*} $excludeTypes 
     * @returns 
     */
    getRelonInByTempR=function($ins, $temp_reln_nid, $excludeTypes=[]) {
        for (var r=0; r <$ins.length; r++) {
            if ($ins[r].r==$temp_reln_nid) {
                if ($excludeTypes.length && $excludeTypes.indexOf($outs[r].type)!=-1) continue;
                return $outs[r];
            }
        }
        return null;
    }


    /**
     * Snag a quick relon by its reln.
     * 
     * @param {array} - the array of outs to limit the search to
     * @param {string} - the temp_nid of the reln
     * @author {Mr.Scatman} 
     * 
     */
     getRelonByTempR=function($outs, $temp_reln_nid, $excludeTypes=[]) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].r==$temp_reln_nid) {
                if ($excludeTypes.length && $excludeTypes.indexOf($outs[r].type)!=-1) continue;
                return $outs[r];
            }
        }
        return null;
    }
    getRelonByTempRTo=function($outs, $temp_reln_nid, $to_s, $excludeTypes=[]) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].r==$temp_reln_nid) {
                if ($outs[r].to_s!=$to_s) continue;
                if ($excludeTypes.length && $excludeTypes.indexOf($outs[r].type)!=-1) continue;
                return $outs[r];
            }
        }
        return null;
    }


    getRelonsByTempR=function($outs, $temp_reln_nid, $excludeTypes=[]) {
        if ($outs==undefined || $outs==null) return [];
        var rels=[];
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].r==$temp_reln_nid) {
                if ($excludeTypes.length && $excludeTypes.indexOf($outs[r].type)!=-1) continue;
                rels.push($outs[r]);
            }
        }
        return rels;
    }

    //find an out of a given type that has a relon that falls within a lineage.
    //we can look "up" the tree line to see if we hit hte parentLineTempNid.
    getRelonByTempRBeLine=function($outs, $types, $parentLineTempNid, $skipTnids=[]) {
        for (var r=0; r < $outs.length; r++) {  
            var relon = $outs[r];
            if ($types.indexOf(relon.type)==-1) continue;
            if (relon.r==$parentLineTempNid) return relon;
            var relNOBJ = MEMG.getNeuronByTempNid(relon.r);
            if ($skipTnids.length && $skipTnids.indexOf(relon.r)!=-1) continue;
            if (MEMG.besOf(relNOBJ, $parentLineTempNid)==true) return relon;
        }
        return null;
    }

    _getFuncObj=function($params) {
        var str = "autoget";
        var params=[];
        if ($params[0].type!=undefined) {
            str="Type";
            params.push($params[0].type);
        }
        if ($params[0].from_nid!=undefined) {
            str=str+"From";
            params.push($params[0].from_nid);
        }
        if ($params[0].reln!=undefined) {
            str=str+"Reln";
            params.push($params[0].reln);
        }
        if ($params[0].to_type!=undefined) {
            str=str+"ToType";
            params.push($params[0].to_type);
        }
        if ($params[0].to_s!=undefined) {
            str=str+"ToS";
            params.push($params[0].to_s);
        }
        return {functName:str, params:params};
    }


    get=function($outs, ...$objs) {
        var functObj = this._getFuncObj($objs);
        console.log(functObj);
        //return this[functObj.funcName]($outs, params);
    }

    gets=function($outs, $getObj){
    }


    getRelonFromTo=function($outs, $type, $to_s) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].type!=$type) continue;
            if ($outs[r].to_s==$to_s) return $outs[r];
        }
        return null;
    }
    getRelonsFromTo=function($outs, $type, $to_s) {
        var relons = [];
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].type!=$type) continue;
            if ($outs[r].to_s==$to_s) relons.push($outs[r]);
        }
        return relons;
    }
    
   /**
    * Get a relon downline.
    * 
    * For an array of outs, and a reln_nid -- get the first to_s value that
    * matches any downloine of the reln_nid in the reln.  
    * IE, looking for a thumbnail image of a certain class or subclass.
    *  
    * This is essentially looking for any qualifying reln (property) value 
    * that is a subclass of the actual value we were expecting.
    * 
    * note: do not use insts here.
    * 
    * @author {Mr.Scatman} 
    */
    getRelnByRelonAndValueAnyDownline=function($outs, $reln_nid) {
        var childrenDownline = [$reln_nid];
        var relnNOBJ = MEMG.getNeuronByTempNid($reln_nid);
        var bDOINSTS = false;
        relnNOBJ.getAllDownlineChildren(childrenDownline, 0, bDOINSTS); //this should get all subs, all children of subs.. (all insts of)
        for (var r=0; r <$outs.length; r++) {
            console.log($outs[r].r+" vs "+$reln_nid);
            if (childrenDownline.indexOf($outs[r].r)!=-1) return $outs[r].to_s;
        }
        return "";
    }

    /**
     */
    getAnyRelonWithR = function($outs, $r) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r==$r) return rObj;
        }
        return null;
    }


    /**
     * 
     * @param {*} $outs 
     * @param {*} $type 
     * @param {*} $r 
     * @author {Mr.Scatman}
     */
    getRelonFromWithR = function ($outs, $type, $r) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r==$r && rObj.type==$type) return rObj;
        }
        return null;
    }

    /** 
     * Get a relon by type, reln and to_type
     * 
     * @param {array} - The array of outs to look in
     * @param {int} - The dimension/TYPE we want.
     * @param {int} - The temp_nid or the reln.
     * @param {int} - The TO_TYPE
     * 
     * @author {Mr.Scatman}  
     */
     getRelonFromWithRToType = function ($outs, $type, $r, $to_type) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r==$r && rObj.type==$type && rObj.to_type==$to_type) return rObj;
        }
        return null;
    }
    getRelonsFromWithRToType = function ($outs, $type, $r, $to_type) {
        if (ANNE.UTILS.isset($outs)==false) return [];
        var matches = [];
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r==$r && rObj.type==$type && rObj.to_type==$to_type) matches.push(rObj);
        }
        return matches;
    }

    /**
     * @param {array} - an array of relons to search
     * @param {int} - the dimension (TYPE) of the relon
     * @param {string} - nide longstring of the reln
     * 
     * @author {Mr.Scatman} 
     */
    getRelonFromWith = function ($outs, $type, $reln) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.reln==$reln && rObj.type==$type) return rObj;
        }
        return null;
    }


    /**
     * @param {array} - an array of relons to search
     * @param {int} - the dimension (TYPE) of the relon
     * @param {int} - temp_nid of the neuron you want in the reln slot of the relon.
     * @param {string} - string of what to expect in the to_s (nidlongstring or any value)
     * 
     * @author {Mr.Scatman} 
     */
    getRelonFromWithTo = function ($outs, $type, $reln, $to_s) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.type!=$type) continue;
            if (rObj.reln!=$reln) continue;
            if (rObj.to_s!=$to_s) continue;
            return rObj;
        }
        return null;
    }

    getRelonFromWithTo_orParentLine1 = function ($outs, $type, $reln, $to_s) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        var relNOBJ = ANNE.MEMG.getNeuronByNid($reln);
        var parentRELN = relNOBJ.getParent();
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.type!=$type) continue;
            if (rObj.reln!=$reln && rObj.reln!=parentRELN.nid) continue;
            if (rObj.to_s!=$to_s) continue;
            return rObj;
        }
        return null;
    }

    hasRelon = function ($outs, $r, $to_s) {
        if (ANNE.UTILS.isset($outs)==false) return null;
        if ($r > 100000000) {
            console.warn("r should NOT be a nid, it should be a temp_nid! (index in the aNeuronsHash tbl)");
        }

        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.r!=$r) continue;
            if (rObj.to_s!=$to_s) continue;
            return rObj;
        }
        return null;
    }

    //used for known relon patterns to pre-encoded values
    //ie, data import from wiki company of "founders" pointing to a "string" for a person
    getRelonFromWithToText = function ($outs, $type, $reln, $text) {
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.type!=$type) continue;
            if (rObj.reln!=$reln) continue;
            
            if (rObj.to_type==TO_TYPE_ANNE) {
                var nObj = MEMG.getNeuronByTempNid(rObj.t);
                if (nObj.text!=$text) continue;
            }
            else continue;
            return rObj;
        }
        return null;
    }
    getRelonsFromWithTo = function ($outs, $type, $r, $to_s) {
        var relons = [];
        if (ANNE.UTILS.isset($outs)==false) return relons;
        if ($r > 100000000) {
            console.warn("r should NOT be a nid, it should be a temp_nid! (index in the aNeuronsHash tbl)");
        }
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.type!=$type) continue;
            if (rObj.r!=$r) continue;
            if (rObj.to_s!=$to_s) continue;
            relons.push(rObj);
        }
        return relons;
    }

    /**
     * @param {array} - an array of relons to search
     * @param {int} - the dimension (TYPE) of the relon
     * @param {string} - nide longstring of the reln
     * 
     * @author {Mr.Scatman} 
     */
    getRelonsFromWith = function ($outs, $type=null, $reln) {
        var relons = [];
        if (ANNE.UTILS.isset($outs)==false) return relons;
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if (rObj.reln!=$reln) continue;
            if ($type!=null && rObj.type!=$type) continue;
            relons.push(rObj);
        }
        return relons;
    }

    
    getRelonsFromWiths = function ($outs, $types, $relns) {
        var relons = [];
        for (var r=0; r < $outs.length; r++) {
            var rObj = $outs[r];
            if ($relns.indexOf(rObj.r)!=-1 && $types.indexOf(rObj.type)!=-1) {
                relons.push(rObj);
            }
        }
        return relons;
    }


    getRelonsTo = function ($ins, $type, $to_type, $to_s) {
        var relons = [];
        for (var r=0; r < $ins.length; r++) {
            var rObj = $ins[r];
            if (rObj.type!=$type) continue;
            if ($rObj.to_type!=to_type) continue;
            if ($rObj.to_s!=$to_s) continue;
            relons.push($rObj);
        }
        return relons;
    }

    /** 
     * Brutal search through all relons.. can be avoided if we create
     * an emulated neuron for any txid that shows up in the to_s.
     * Then it would be a neuron that could have ins that are easily searched
     * by the getRelonsTo function above.
     * @author {Mr.Scatman} 
     */
     getRelonsToNaked = function ($type, $to_type, $to_s) {
        var relons = [];
        for (var r=0; r < MEM._aRelons.length; r++) {
            var rObj = $ins[r];
            if (rObj.type!=$type) continue;
            if ($rObj.to_type!=to_type) continue;
            if ($rObj.to_s!=$to_s) continue;
            relons.push($rObj);
        }
        return relons;
    }


    /** 
     * @author {Mr.Scatman} 
     */
    getSysRelon=function($outs, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS) return $outs[r];
        }
        return null;
    }

    getSysRelonWithNoSysRefTxid=function($outs, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS) {
                if ($outs[r].sys_ref_txid==null || $outs[r].sys_ref_txid=="") return $outs[r];
            }
        }
        return null;
    }

    /** 
     * @author {Mr.Scatman} 
     */
    getSysRelons=function($outs, $relnTnid) {
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS) rels.push($outs[r]);
        }
        return rels;
    }


    /** 
     * @author {Mr.Scatman} 
     */
    getSysToTypes=function() {
        return MEM._SYS_TO_TYPE_CODES;
    }
    getSysTypes=function() {
        return MEM._SYS_TYPE_CODES;
    }

    //


    /** 
     * 
     * @author {Mr.Scatman} 
     */
    getTerminatingType=function($nObj) {
        if ($nObj==null) return null;
        if ($nObj.SYS_ENUM_VAL!=undefined) return $nObj;
        if ($nObj.be_lvl==-1) return null;
        var bes = $nObj.bes;
        if (bes.length==0) return null;
        var beObj = MEMG.getNeuronByTempNid(bes[0]);
        var safety = 20;
        while (beObj) {
            safety=safety-1;
            if (safety<=0) return null;
            if (beObj.SYS_ENUM_VAL!=undefined) return beObj;
            if (beObj.bes.length==0) return null;   //not a terminating type.
            if (beObj.be_lvl==1) return null;       //safety check.
            if (beObj.temp_nid==beObj.bes[0]) {
                console.warn("catastrophic self-reference in the BE dimension!!!");
                return null;
            }
            beObj = MEMG.getNeuronByTempNid(beObj.bes[0]);
            if (beObj.be_lvl==0) return null;       //safety check.
        }
        return null;
    }


    /** 
     * @author {Mr.Scatman} 
     */  
    getToNObjPropFromRelon=function($fulfillmenRelon, $prop="SYS_ENUM_VAL") {
        if ($fulfillmenRelon==null) return "";
        if ($fulfillmenRelon.t==undefined || $fulfillmenRelon.t==-1) return "";
        var nOBJ = MEMG.getNeuronByTempNid($fulfillmenRelon.t);
        if (nOBJ==null) return "";
        if (nOBJ[$prop]==undefined) return "";
        return nOBJ[$prop];
    }
    

    /**
     * Sometimes i need some help.. remembering function names.
     * @param {*} $nObjs 
     * @returns 
     */
    getTnidList=function($nObjs) {
        return this.getTnidsFromNObjs($nObjs);
    }

    
    /** 
     * @author {Mr.Scatman} 
     */    
     getTnidsFromNObjs=function($nObjs) {
        var tnids = [];
        for (var n=0; n<$nObjs.length;n++) { tnids.push($nObjs[n].temp_nid); }
        return tnids;
    }

    getTnidsFromNids=function($nids) {
        var tnids = [];
        for (var n=0; n<$nids.length;n++) { 
            var nObj = ANNE.MEMG.getOrMakeIfNew($nids[n]);
            tnids.push(nObj.temp_nid); 
        }
        return tnids;
    }
    
    
    /**
     * 
     * @param {*} $relons 
     * @param {*} $r 
     * @returns 
     */
    getFromArrayMap=async($relons, $r=-1)=>{
        var am = new arraymap();
        var chase_froms = [];

        for (var r=0; r < $relons.length; r++) {
            var rel = $relons[r];
            if ($r!=-1 && rel.r!=$r) continue;

            if (rel.to_type!=TO_TYPE_ANNE) continue;
            var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
            if (fromNOBJ.text=="") {
                if (chase_froms.indexOf(rel.from_nid)==-1) chase_froms.push(rel.from_nid);
            }
            am.addUnique(fromNOBJ, fromNOBJ.nid);
        }

        if (chase_froms.length) {
            await ANNE.API.getNames(chase_froms);
        }

        return am;
    }


    getNObjsFromRels=function($rels) {
        return this.getToNObjsFromRelon($rels);
    }
    /**
     * 
     * @param {*} $rels 
     * @returns 
     */
    getToNObjsFromRelon=function($rels) {
        var nObjs = [];
        for (var r=0; r < $rels.length; r++) {
            if ($rels[r].to_type!=TO_TYPE_ANNE) continue;
            var toNOBJ = ANNE.MEMG.getNeuronByNid($rels[r].to_s);
            nObjs.push(toNOBJ);
        }
        return nObjs;
    }


    /**
     * 
     * @param {*} $countryNOBJ 
     * @param {*} $bdoIDB 
     * @returns 
     */
    getNationalityFromNation=async($countryNOBJ, $bdoIDB=true)=> {
        if ($countryNOBJ.nid==ANNE.MEM.usa.nid) return ANNE.MEM.american.nid;
        
        //onetime pull of the ins of this country.
        if ($bdoIDB==true) await ANNE.IDB.loadIDBNidIns_single($countryNOBJ.nid);

        for (var r=0; r < $countryNOBJ.ins.length; r++) {
            var rel = $countryNOBJ.ins[r];
            if (rel.type!=TYPE_BE) continue;
            if (rel.reln!=ANNE.MEM.country.nid) continue;
            return rel.to_s;
        }

        return "";
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $reln 
     * @param {*} $bCHASEIDB 
     * @returns 
     */
    getToNObjsByReln=async($nObj=null, $reln="", $bCHASEIDB=false)=>{
        if ($nObj==null) return [];
        if ($reln=="" || $reln==null) return [];
        var nObjs = [];
        var from_nids = [];
        
        if ($reln!="") {
            var relnNOBJ = ANNE.MEMG.getNeuronByNid($reln);
            console.log("looking for relnNOBJ "+relnNOBJ.text+" on "+$nObj.text);
        }
        
        //ANNE.TALK.printRelons($nObj.outs);

        await ANNE.IDB.getNid(null, $nObj.nid);

        ANNE.TALK.printRelons($nObj.outs);
            
        for (var r=0; r < $nObj.outs.length; r++) {
            var rel = $nObj.outs[r];
            if (rel.reln!=$reln) continue;
            var nObj = MEMG.getNeuronByNid(rel.to_s);
            nObjs.push(nObj);
            if (from_nids.indexOf(nObj.nid)!=-1) {
                from_nids.push(nObj.nid);
            }
        }

        if ($bCHASEIDB==true) {
            var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", from_nids.join(""));
		    await ANNE.IDB.lookupNids(phObj, from_nids, [0,1,2,3,4,5,6,7,8,9,10])
        }

        return nObjs;
    }


    getBirthDeath=function($nObj) {
        var birthday = null;
        var deathday = null;
        for (var r=0; r < $nObj.outs.length; r++) {
            if ($nObj.outs[r].type!=TYPE_BE && $nObj.outs[r].type!=TYPE_HAS) continue;
            var rel = $nObj.outs[r];
            if (rel.r==ANNE.MEM.birthday.temp_nid) {
                birthday=rel.to_s;
            }
            else if (rel.r==ANNE.MEM.deathday.temp_nid) {
                deathday=rel.to_s;
            }
        }
        
        return {date1:birthday, date2:deathday};
    }
    

    /**
     * We could build an auto alt be parent/child heirarchy here, but meh.
     * It is only for one use in the threejs app.
     */
    getTribesArr=async()=>{
        
        var tribal_nids = [];
        for (var r=0; r < MEM._aRelons.length; r++) {
            if (MEM._aRelons[r].to_type!=TO_TYPE_ANNE) continue;
            if (MEM._aRelons[r].type!=TYPE_BE) continue;

            if (MEM._aRelons[r].r==MEM.ec_is.temp_nid
                && MEM._aRelons[r].t==MEM.tribal.temp_nid) {
                if (tribal_nids.indexOf(MEM._aRelons[r].from_nid)==-1) {
                    tribal_nids.push(MEM._aRelons[r].from_nid);
                }
            }
            
            else if (MEM._aRelons[r].r==MEM.tribal.temp_nid) {
                if (tribal_nids.indexOf(MEM._aRelons[r].to_s)==-1) {
                    tribal_nids.push(MEM._aRelons[r].to_s);
                } 
            }
        }
        
        //religions
        for (var i=0; i < MEM.religion.ins.length; i++) {
            if (MEM.religion.ins[i].type!=TYPE_BE) continue;
            if (MEM.religion.ins[i].r!=MEM.inst.temp_nid) continue;
            if (tribal_nids.indexOf(MEM.religion.ins[i].from_nid)==-1) {
                tribal_nids.push(MEM.religion.ins[i].from_nid);
            }
        }

        //political parties.
        for (var i=0; i < MEM.political_party.ins.length; i++) {
            if (MEM.political_party.ins[i].type!=TYPE_BE) continue;
            if (MEM.political_party.ins[i].r!=MEM.inst.temp_nid) continue;
            if (tribal_nids.indexOf(MEM.political_party.ins[i].from_nid)==-1) {
                tribal_nids.push(MEM.political_party.ins[i].from_nid);
            }
        }

        //awards/groups/alma maters?  (some identify with those as their primary identification)
        //and it is relevant.  (we have the data already in the hypergraph,  TO DO this later.
        if (tribal_nids.length>0) return this.getNObjsFromNidList(tribal_nids);
        return [];
    }


    getNeuronsFromTnids=function($tnids) { 
        return this.getNObjsFromTnids($tnids);
    }

    //lets write another one since I can never find things.
    getNObjsFromTnids=function($tnids) {
        var nObjs = [];
        for (var t=0; t<$tnids.length; t++) { 
            var nObj = MEMG.getNeuronByTempNid($tnids[t]);
            nObjs.push(nObj); 
        }
        return nObjs;
    }

    getNidsListFromTnidsList=function($tnids) {
        var nids = [];
        for (var t=0; t<$tnids.length; t++) { 
            var nObj = MEMG.getNeuronByTempNid($tnids[t]);
            nids.push(nObj.nid); 
        }
        return nids;
    }
 

    //probably already did this, as usual.. doing it again.
    getNidsListFromNObjs = function($nObjs) {
        var nids = [];
        for (var n=0; n < $nObjs.length; n++) {
            if (nids.indexOf($nObjs[n].nid)==-1) nids.push($nObjs[n].nid);
        }
        return nids;
    }

    getNObjsFromList = function($nids) {
        var nObjs = [];
        for (var n=0; n < $nids.length; n++) {
            var nObj = MEMG.getNeuronByNid($nids[n]);
            if (nObj==null) continue;
            nObjs.push(nObj);
        }
        return nObjs;
    }
    

   /** 
    * Lookup up a neuron in our hash table.
    * 
    * In ANNE chain the neurons are already .nid strings and they 
    * ARE the hash.  Previous implementations had longer strings
    * and needed to hash them into a table to prevent slowdowns.
    * 
    * ANNE chain neurons are max 20 chars long and they the associative
    * key in the hash table pointing to the the temp_nid which is 
    * the index of the neuron in numeric order in the _aNeurons table.
    * 
    * .nid = the full id (in string form of a signed long) as it exists on the blockchain.
    * .temp_nid = the numeric index into the _aNeurons.
    * 
    * @author {Mr.Scatman} 
    */    
    __hashNeuronLookup=function($nid, $maxFireWeight=0.03) {
        if ($nid.length == 0) return null;
        if ($nid==undefined || $nid==null) return null;          
        if (typeof $nid === "object") {
            return null;
        }
        var temp_nid = ANNE.MEM._aNeurons_hashtbl[$nid];
        if (temp_nid!=undefined) {
            if (ANNE.BRAIN!=null && $maxFireWeight!=0) {
                ANNE.BRAIN.fireNeuron(ANNE.MEM._aNeurons[temp_nid], $maxFireWeight);
            }
            return ANNE.MEM._aNeurons[temp_nid];
        }
        return null;
    }

    /**
     * Lookup relon (Transaction) by txid using our hash table.
     * For use when you are looking up a relon by txid. 
     * Get the rid index in the hash table, and use it to snag the relon!
     *                                                          
     *
     * @param {string}  - This is a JAVA Long as a string for the TransactionId
     * @author Mr.Scatman
     */
    __hashRelonLookup=function($txid) {
        if ($txid==undefined || $txid==null || typeof $txid!=="string") {
            console.log("this should never happen"); 
            return null;
        }

        $txid = $txid.trim();

        if ($txid.length == 0 || $txid.length>21) {
            console.log("this should never happen"); 
            return null;
        }
        if (typeof $txid !== "string") return null;
        var rid = MEM._aRelons_hashtbl[$txid];
        if (rid!=undefined) return MEM._aRelons[rid];
        return null;
    }


    /**
     * Get alt ec_is TYPE_BE relon to the temp_t 
     */
     isAltBeOf = function($outs, $temp_t) {
        for (var r=0; r <$outs.length; r++) {
            if ($outs[r].type!=TYPE_BE) continue;
            if ($outs[r].to_type!=TO_TYPE_ANNE) continue;
            if ($outs[r].t!=$temp_t) continue;
            return true;
        }
        return false;
    }
    
    /** 
     * Lookup a relon by TXID.
     * 
     * tired of mispelling the other one.
     * 
     * @author {Mr.Scatman} 
     */
    lookupByTXID = function ($txid, $arr=null) { 
        var rel =  MEMG.lookupRelonByTXID($txid, $arr);
        return rel;
    }

    lookupByTXID_chase = async ($txid, $arr=null)=> { 
        var rel =  MEMG.lookupRelonByTXID($txid, $arr);
        if (rel==null) return await CHASE.lookupTxid($txid);
        return rel;
    }

    lookupTXID = function($txid, $arr=null) {
        return this.lookupRelonByTXID($txid, $arr);
    }

    /**
     */
    lookupRelonByTXID = function($txid, $arr=null) {
         var arr = null;
         
         //if array is passed in use it the old fashioned looper way
         if ($arr!=null) arr = $arr;
 
         //if no array, we are searching the entire relon space. Use our hash table method.
         else return MEMG.__hashRelonLookup($txid);
 
         //if $arr, then just loop it for our txid.
         for (var r=0; r < arr.length; r++) {
             var rObj = arr[r];
             if (rObj.txid==$txid) return rObj;
         }
         return null;
     }


     lookupRelonByRid = function($rid) {
        var relon = MEM._aRelons[$rid];
        return relon;
    }

    
    /**    
     * Display the to_s value of a relon the best we can.
     * If a decrypted value exists, display it.
     * 
     * TODO: Should have clear ways to show some of the other TO_TYPEs..
     * (formatters)...
     * 
     * @author {Mr.Scatman} 
     */
     __MASTER_DISPLAY_RELON_TO = function($relon, $to_class="") {
        var to_s = $relon.to_s;
        var to_type = $relon.to_type;
        var pre_ico_add = "";

        if (to_type>=TO_TYPE_ENCR_ANNE) {
            if ($relon.decrypted!=undefined) {
                to_s = $relon.decrypted;
                to_type = parseInt(to_type)-30;
            }

            else if ($relon.ghost==true) {
                var fNOBJ = MEMG.getNeuronByTempNid($relon.f);

                //non encrypted value is in the to_s still since we 
                //do not do encryption until we have a real FROM NOBJ
                if (fNOBJ.ghost==false) to_type = parseInt(to_type)-20;

                //for ghost relons FROM REAL NEURONS we will have the encryption
                //already done.
            }
            else {
                pre_ico_add = '<span title="encrypted" style="filter: grayscale(100%); font-size:.8rem">🔒</span>';
            }
            //too nasty to alter all the calls to await this funct..
            //each session the encryped to will remain encrypted until
            //decrypted.  That app session it will then be available.
            //else {
                //await USER.ascertainEncryptionAndDecrypt($relon);
            //}
        }

        if (to_type==TO_TYPE_ANNE) {  
            var toNOBJ = null;
            if ($relon.t!=undefined && $relon.t!=-1) toNOBJ = MEMG.getNeuronByTempNid($relon.t);
            else toNOBJ = MEMG.getNeuronByNid(to_s);

            if (toNOBJ!=null) {
                if (toNOBJ.text!="") return pre_ico_add+toNOBJ.text;
                return pre_ico_add+toNOBJ.nid;
            }
            return pre_ico_add+to_s;
        }

        else if (to_type==TO_TYPE_NUMBER) {
            var to_s_str = to_s+"";
            if (to_s_str.indexOf(".")!=-1) to_s_str = ANNE.UTILS.comma(to_s)+"";

            var post = this.doPostANNIESHackFix($relon);
            if (post.length>0) {
                return pre_ico_add+post;
            }
            else {
                if (to_s_str.indexOf(".")!=-1) return pre_ico_add+to_s_str;
                else if (parseInt(to_s)>=1000) return pre_ico_add+ANNE.UTILS.commas(to_s);
                return pre_ico_add+to_s_str;
            }

        }

        else if (to_type==TO_TYPE_STRING) {
            var relNOBJ = MEMG.getNeuronByTempNid($relon.r);

            //if this is a url part with a url base, blow it out full format.
            if (relNOBJ && relNOBJ.is(MEM.url_part)==true) {
                var base_rel = MEMG.getHasRelon(relNOBJ.outs, MEM.base_url.temp_nid);
                if (base_rel!=null) {
                    var parts = to_s.split(":::");//anne protocol for src:::display.
                    var disp = "";
                    var url = base_rel.to_s+parts[0];
                    if (parts.length>1) disp = parts[1];
                    else disp = url;
                   
                    return pre_ico_add+'<a title="'+url+'" href="'+url+'" target="_blank">'+disp+'</a>';
                }
                else return pre_ico_add+to_s;
            }

            //var post = this.doPostANNIESHackFix($relon);
            if ($to_class!="") {
                return  pre_ico_add+"<span class='"+$to_class+"'>"+$relon.to_s+"</span>";
            }
            return pre_ico_add+$relon.to_s;
        }

        else if (to_type==TO_TYPE_BOOLEAN) {
            var val = "";
            if (to_s==0 || to_s=="0") val = false;
            else if (to_s==1 || to_s=="1") val = true;
            else val = to_s;
            return pre_ico_add+val;
        }

        else if (to_type==TO_TYPE_URL) {
            var relNOBJ = MEMG.getNeuronByTempNid($relon.r);

            //if this is a url part with a url base, blow it out full format.
            if (relNOBJ && relNOBJ.is(MEM.url_part)==true) {
                var base_rel = MEMG.getHasRelon(relNOBJ.outs, MEM.base_url.temp_nid);
                if (base_rel==null) {
                    return pre_ico_add+val;
                }
                else {
                    var parts = to_s.split(":::");//anne protocol for src:::display.
                    var disp = "";
                    if (parts.length>1) disp = parts[1];
                    else disp = to_s;
                    
                    //https://commons.wikimedia.org/wiki/
                    var url = base_rel.to_s+parts[0];
                    
                    //if (base_rel.to_s=="https://commons.wikimedia.org/wiki/") {
                        //url = base_rel.to_s+"File:"+parts[0];   
                    //}
                    return pre_ico_add+'<a class="external" title="'+url+'" href="'+url+'" target="_blank">'+disp+'</a>';
                } 
            }

            var parts = to_s.split(":::");//anne protocol for src vs display.
            var disp = "";
            if (parts.length>1) disp = parts[1];
            else {
                if (to_s.length<40) disp = parts[0];
                else disp = parts[0];
            }
            return pre_ico_add+'<a class="external" title="'+parts[0]+'" href="'+parts[0]+'" target="_blank">'+disp+'</a>';
        }

        //anything else
        var val = to_s;
        var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
        var TTOBJ = MEMG.getTerminatingType(relnNOBJ);
        if (TTOBJ!=null) {
            var to_sNOBJ = null;
            if ($relon.t!=undefined && $relon.t!=-1) to_sNOBJ = MEMG.getNeuronByTempNid($relon.t);
            if (TTOBJ.formatter!=undefined) {
                if (to_sNOBJ!=null) return pre_ico_add+TTOBJ.formatter(to_sNOBJ);
                else return pre_ico_add+TTOBJ.formatter(val);
            }
            else return pre_ico_add+val;
        }
        else return pre_ico_add+val;
    }


    /**
     * 
     * @param {*} $relon 
     * @returns 
     */
    doPostANNIESHackFix = function($relon) {
        var val = parseInt(parseInt($relon.to_s)/100);
        var fixval = "<img class='ibi icofix' width='16' height='16' src='../imgs/anne_ico_small.png'>";
        if ($relon.r==ANNE.MEM.ASDF_RISK_FUNDED.temp_nid) return ANNE.UTILS.commas(val)+fixval;
        else if ($relon.r==ANNE.MEM.ASDF_RISK_REFUNDED.temp_nid) return ANNE.UTILS.commas(val)+fixval;
        else if ($relon.r==ANNE.MEM.ASDF_RISK_REMAINING.temp_nid) return ANNE.UTILS.commas(val)+fixval;
        else if ($relon.r==ANNE.MEM.min_streampay_payment.temp_nid) return ANNE.UTILS.commas(val)+fixval;
        else if ($relon.r==ANNE.MEM.keep_min_balance.temp_nid) return ANNE.UTILS.commas(val)+fixval;
        else if ($relon.r==ANNE.MEM.annies.temp_nid) return ANNE.UTILS.commas(val)+fixval;
        else if ($relon.r==ANNE.MEM.distro_quantity.temp_nid) {
            var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
            
            //a temp hack..
            if (fNOBJ.bes[0]==ANNE.MEM.streampay.temp_nid) {
                var rel_qt = MEMG.getHasRelon(fNOBJ.outs, MEM.distro_quantity_type.temp_nid);
                if (rel_qt!=null && rel_qt.t==ANNE.MEM.distro_qt_specific.temp_nid) {
                    return ANNE.UTILS.commas(val)+fixval;
                }
            }
        }
        
        return val;
    }
    isAFixer=function($relon) {
        if ($relon.r==ANNE.MEM.ASDF_RISK_FUNDED.temp_nid) return true;
        else if ($relon.r==ANNE.MEM.ASDF_RISK_REFUNDED.temp_nid) return true;
        else if ($relon.r==ANNE.MEM.ASDF_RISK_REMAINING.temp_nid) return true;
        else if ($relon.r==ANNE.MEM.min_streampay_payment.temp_nid) return true;
        else if ($relon.r==ANNE.MEM.keep_min_balance.temp_nid) return true;
        else if ($relon.r==ANNE.MEM.annies.temp_nid) return true;
        else if ($relon.r==ANNE.MEM.distro_quantity.temp_nid) {
            var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
            
            //a temp hack..
            if (fNOBJ.bes[0]==ANNE.MEM.streampay.temp_nid) {
                var rel_qt = MEMG.getHasRelon(fNOBJ.outs, MEM.distro_quantity_type.temp_nid);
                if (rel_qt!=null && rel_qt.t==ANNE.MEM.distro_qt_specific.temp_nid) {
                    return true;
                }
            }
        }
    }


    /**
     * Set the my_sponsored array.
     *  
     * We can't get these on initial mem load.. since USer is not logged in.
     * I don't want to do another full loop of relon mem just for this var
     * so we do it on-demand.
     * 
     * Run-time after mem load this will be set properly as the relons come in
     * to the mem processor.
     * 
     * @author {Mr.Scatman} 
     */
    setMySponsored = function($nid) {
        var userNOBJ = MEMG.getNeuronByNid($nid);
        if (userNOBJ.my_sponsored==undefined) userNOBJ.my_sponsored=[];
        for (var r=0; r<ANNE.MEM._aRelons.length; r++) {
            var rObj = ANNE.MEM._aRelons[r];
            if (rObj.sponsored_by!="" && rObj.sponsored_by==$nid) {
                var fNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (userNOBJ.my_sponsored.indexOf(fNOBJ.temp_nid)==-1) {
                    userNOBJ.my_sponsored.push(fNOBJ.temp_nid);
                }
            }
        }
    }

    getRelPat = function($prop, $val) {
        return {prop:$prop, val:$val};
    }

    
    /** May be able to switch a BULK of these functions to simply doing this 
            var relPatArr = [{prop:'type', val:TYPE_PARAM}, 
                            {prop:'r', val:$relon_pat.r}, 
                            {prop:'t', val:to_t}]; 
    */
    getRelByPat = function($outs, $relPatArr) {
        for (var r=0; r < $outs.length; r++) {
            var rel = $outs[r];
            var bNO = false;
            for (var rp=0; rp < $relPatArr.length; rp++) {
                var pat = $relPatArr[rp];
                if (rel[ pat.prop] != pat.val) {
                    bNO=true;
                    break;
                }
            }
            if (bNO==false) return rel;
        }
    }

    getRelsByPat = function ($outs, $relPatArr) {
        var rels = [];
        for (var r=0; r < $outs.length; r++) {
            var rel = $outs[r];
            var bNO = false;
            for (var rp=0; rp < $relPatArr.length; rp++) {
                var pat = $relPatArr[rp];
                if (rel[ pat.prop] != pat.val) {
                    bNO=true;
                    break;
                }
            }
            if (bNO==false) rels.push(rel);
        }
        return rels;
    }
    
    
    getFreezeRel = function($outs) {
        for (var r=0; r < $outs.length; r++) {
            var rel = $outs[r];
            if (rel.r == ANNE.MEM.CMD_FREEZE.temp_nid) {
                return rel;
            }
        }
        return null;
    }

    nidsToTnidsList=function($tnids) {
        var nids = [];
        for (var t=0; t < $tnids.length; t++) {
            nids.push( MEMG.getNeuronByTempNid($tnids[t]).nid);
        }
        return nids;
    }



}


var MEMG = new mem_gets();
//freeze this


/************** (from file ./js/mem/mem_gets.js)**/ 


var ANNEX = {};

//relon sponsors get paid each time a neuron is ANNEX'd.  This incentivizes building
//good data and COLLAB with ANNE on ANNE-owned USER-sponsored neurons.  Anyone can collab 
//and provide new usable data for an ANNE-owned neuron.  As sponsor they get the fire fee 
//for any feedback on THAT RELON (if someone "loves" the fact Tom Cruise has a birthday of July 3, 1962).
//Note, if they LOVE "July 3, 1962" then the sponsor of that DATE neuron gets the fire fee.
//Any sponsors of relons accumulate the fees as the relon is rated -- but each time the neuron
//is annexed (changes sponsor rights holder) they get paid 3 additional coins.
ANNEX.RELON_SPONSOR_FIRE = 300;

//all relons caculate the same value (regardless of their individual feedback)
//or their type (TEXT/URLS cost more to make but do not generate higher value here)
//if a relon gains feedback ratings that is more valuable but that value is given to the 
//relon sponsor, not the neuron.  This can measure/inform us how popular/active a neural
//circuit is -- but that shows us total "value" of that circuit and relon-feedback doesn't affect the cost of ANNEX.
//Also, NEURON feedback is a value that shows how much this neuron generates.  It does not alter
//the price --- it alters the VALUE.  If a neuron pulls in alot of awareness relons holding its
//sponsorship is lucrative thus it has a high value compared to the annex price (thus is a good deal).
ANNEX.RELON_VALUE = 2000;

ANNEX.no_annex_classes = [];

ANNEX.init=function() {
    ANNEX.no_annex_classes.push(MEM.bounty.nid);
}


/**
 * 
 * @param {*} $nObj 
 * @returns 
 */
ANNEX.assessAnnexabiltyNoChase=function($nObj) {
    if ($nObj.owned_by!=ANNE_ACCTID) return false;
    if ($nObj.is(MEM.identity)==true && ($nObj.ntype==NTYPE_INST || $nObj.ntype==NTYPE_UNKNOWN)) return false;
    if ($nObj.is(MEM.__ANNE_SYS_CLASS)==true && $nObj.is(MEM.gregorian_date)==false) return false;
    if ($nObj.bes[0]==MEM.anndle.temp_nid) return false;
    if ($nObj.bes[0]==MEM.subanndle.temp_nid) return false;
    if ($nObj.is(MEM.PDN)==true) return false;
    if ($nObj.is(MEM.ASDF)==true) return false;
    if ($nObj.is(MEM.bounty)==true) return false;
    return true;
}


/**
 * We need to get BEM line be/sys relons for this nid to figure this out.
 */
ANNEX.assessAnnexability=async($nObj, $bCHASE=false)=> {
    
    if ($nObj.nid==MEM.___SELF___.nid) return false;
    
    if ($nObj.owned_by!=ANNE_ACCTID) return false;
    
    if ($nObj.is(MEM.__ANNE_SYS_CLASS)==true && $nObj.ntype==NTYPE_INST) return false;
    if ($nObj.bes[0]==MEM.anndle.temp_nid) return false;
    if ($nObj.bes[0]==MEM.subanndle.temp_nid) return false;
    if ($nObj.is(MEM.PDN)==true) return false;
    if ($nObj.is(MEM.ASDF)==true) return false;
    if ($nObj.is(MEM.bounty)==true) return false;
    return true;
}


/**
 * 
 */
class anne_annex_gui {
    constructor($el, $guid, $nObj=null, $send_cb=null) {
        this.el = $el;
        this._ID = $guid;
        if (this._ID==null) this._ID="";
        else this._ID = "_"+this._ID;
        
        this.nObj = $nObj;
        this.backlink = null;
        this.new_boosties_in_annecoin = 0;
        this.send_cb = $send_cb;
        this.dossierPTR = null;
        if ($nObj._aplex!=undefined && $nObj._aplex!=null && $nObj._aplex.has_pending==true) {
            this.has_pending=true;
        }
        else this.has_pending=false;
    }


    setParentPtr = function($ptr) {
        this.dossierPTR = $ptr;
    }


    /**
     */
    annecoin_entered = async($e)=>{
        var el_input = $e.target;
        var val = ANNE.A.getValue(el_input)+"";
        var newval = "";
        var lastChar = "";

        if (val.length==1) lastChar = val;
        else lastChar = val[val.length-1];
        
        if (lastChar=="k" || lastChar=="K") {
            val = val.substring(0, val.length-1);
            this.new_boosties_in_annecoin = Number.parseFloat(val) * 1000;
            el_input.value = this.new_boosties_in_annecoin;
        }
        else if (lastChar=="m" || lastChar=="M") {
            val = val.substring(0, val.length-1);
            this.new_boosties_in_annecoin = Number.parseFloat(val) * 1000000;
            el_input.value = this.new_boosties_in_annecoin;
        }
        else if (lastChar=="b" || lastChar=="B") {
            val = val.substring(0, val.length-1);
            this.new_boosties_in_annecoin = Number.parseFloat(val) * 1000000000;
            el_input.value = this.new_boosties_in_annecoin;
        }

        else {
            for (var v=0; v < val.length; v++) {
                if ("0123456789.,".indexOf(val[v])==-1) continue;
                newval = newval+val[v];
            }
            this.new_boosties_in_annecoin = newval;
        }

        ANNE.A.setValue(el_input, this.new_boosties_in_annecoin);

        var el_boosties_final_btn = document.getElementById("annex_boosties_final_btn"+this._ID);
        this.new_boosties_in_annecoin = parseInt(this.new_boosties_in_annecoin);
        
        if (this.new_boosties_in_annecoin>=400) {
            ANNE.A.show(el_boosties_final_btn);
            ANNE.A.displayAnnecoinWithCommas("annex_boosties_annies"+this._ID, this.new_boosties_in_annecoin, true);
        }
        else {
            ANNE.A.hide(el_boosties_final_btn);
            ANNE.A.setHTML("annex_boosties_annies"+this._ID, "");
        }
    }
    

    /**
     * Boosties are a sys message sent from the neuron being boosted
     * It works like a collab relon in the SYS dimension.
     * If the user is not the current sponsor on record the TX will be rejected.
     */
    final_boosties_submit = async($e)=>{
        var el_boosties_final_btn= document.getElementById("annex_boosties_final_btn"+this._ID);   
        var el_bootsies_input= document.getElementById("annex_bootsies_input"+this._ID);
        ANNE.A.hide(el_boosties_final_btn);
        if (parseInt(this.new_boosties_in_annecoin)<=0) {
            el_bootsies_input.focus();
            return;
        }

        ANNE.A.hide(el_bootsies_input);

        //this could just be a payment but we want to document it with a relon.
        var aplex = this.nObj._aplex;
        var gui_extras = new Object();
        gui_extras.payment = parseInt(this.new_boosties_in_annecoin * ANNE.ANNE_TO_ANNIES);

        //this is already put in the sendQ.
        var sys_msg_rel = ANNE.MEMZ.ez_new_sys_msg(this.nObj, MEM.boosties.nid, TO_TYPE_NUMBER, gui_extras.payment, gui_extras);
        //var sys_msg_rel = await ANNE.GHOSTDATA.new_ghost_sys_msg(this.nObj, MEM.boosties.nid, TO_TYPE_NUMBER, gui_extras.payment, gui_extras.payment);
        console.log(sys_msg_rel.annies);

        if (aplex.last_fullhash!=undefined && aplex.last_fullhash!="") {
            sys_msg_rel.referencedTransactionFullHash = aplex.last_fullhash;
        }

        sys_msg_rel.annies = gui_extras.payment;

        ANNE.MEM_UDF.sendQueue();
        ANNE.MEM_UDF.popupElCancel();
    }


    /**
     * 
     */
    soloUseEvents =function($btn, $amtDiv, $input, $finalBtn) {
        $finalBtn.addEventListener("click", e=>this.final_boosties_submit(e));
        $input.addEventListener("input", e=>this.annecoin_entered(e));
        const btn = $btn;
        const amtDiv = $amtDiv;
        const input = $input;
        $btn.addEventListener("click", function($e){
            ANNE.A.show(amtDiv);
            ANNE.A.hide(btn);
            input.focus();
        });
    }


    /**
     * ANNEXes must be confirmed (mined) before they are active.
     * 
     * @param {*} $bDOEVENTS 
     * @param {Object} $backlink - pending annex
     * @returns 
     */
    populateANNEX_div = async($bDOEVENTS=true, $backlink=null)=> {
        this.bDOEVENTS = $bDOEVENTS;
        this.backlink = $backlink;

        if (this.nObj.ghost==true) return;

        var aplex = null;

        //get a new aplex every time now.
        if (this.nObj._aplex==null) {
            aplex = await ANNE.API.getAplex(this.nObj);
            if (aplex==undefined || aplex==null) {
                console.warn("THIS SHOULD NEVER HAPPEN");
                return false;
            }
        } 
        else aplex = this.nObj._aplex;

        if (this.backlink==null) {
            this.backlink = ANNE.MEMG.getUnconfirmedAnnex(this.nObj.outs);
        }

        
        //try for backlink now..
        if (this.nObj._aplex.pending_relon!=null) {
            this.backlink=this.nObj._aplex.pending_relon;
        }
        
        console.log(aplex);
        
        var ret = await aplex.calculate_annex_vars(this.nObj);
        if (ret==undefined) {
            console.log("PROBLEM.");
            alert("fix this annex display issue mr.scatman.");
        }
        if (ret==-1) {
            ANNE.A.hide(el_annex_btn);
            ANNE.A.hide(el_boosties_btn);
            ANNE.A.hide("anneflap");
            return false;
        }
        ANNE.A.setHTML("annex_lvl"+this._ID, aplex.price_lvl);
        ANNE.A.setHTML("annexlvltitle"+this._ID, aplex.price_lvl);
        ANNE.A.empty("annex_cost_to_annex"+this._ID);
        ANNE.A.ezAnniesToAnneDisplay("annex_cost_to_annex"+this._ID, aplex.calculated_cost);
        ANNE.A.setHTML("annex_refs_to_profit"+this._ID, UTILS.formatNumberCommas(Math.ceil(aplex.calculated_cost/200)));

        ANNE.A.empty("annex_boosties"+this._ID);
        ANNE.A.ezAnniesToAnneDisplay("annex_boosties"+this._ID, aplex.boosties);

        ANNE.A.empty("annex_last_price"+this._ID);  
        ANNE.A.ezAnniesToAnneDisplay("annex_last_price"+this._ID, aplex.prev_annex_price);

        
        ANNE.A.empty("annex_annies_in"+this._ID);  
        ANNE.A.ezAnniesToAnneDisplay("annex_annies_in"+this._ID, aplex.annies_in);
        
        
        ANNE.A.setHTML("annex_fire"+this._ID, UTILS.formatNumberCommas(aplex.fire));
        ANNE.A.setHTML("annex_non_awareness_outs"+this._ID, UTILS.formatNumberCommas(aplex.non_awareness_outs));
        ANNE.A.setHTML("annex_sponsored_relons"+this._ID, UTILS.formatNumberCommas(aplex.annex_sponsored_relons));
        ANNE.A.setHTML("annex_total_awareness"+this._ID, UTILS.formatNumberCommas(aplex.total));
        ANNE.A.fadeIn(this.el);
        var el_annex_btn= document.getElementById("annex_btn"+this._ID);
        var el_boosties_btn = document.getElementById("annex_boosties_btn"+this._ID);
        var el_boosties_amt_div= document.getElementById("annex_boosties_amt_div"+this._ID);
        ANNE.A.hide(el_boosties_amt_div);
        var el_bootsies_input= document.getElementById("annex_bootsies_input"+this._ID);
        var el_boosties_final_btn= document.getElementById("annex_boosties_final_btn"+this._ID);

 
        //no self-annex.
        //boosties ok.
        if (this.nObj.sponsored_by==USER._userNObj.nid || (this.nObj.sponsored_by=="" && this.nObj.owned_by==USER._userNObj.nid)) {
            ANNE.A.hide(el_annex_btn);
            ANNE.A.show(el_boosties_btn);

            //we need a boosties modal with boosty amount
            if ($bDOEVENTS) {
                if (el_boosties_final_btn!=null) el_boosties_final_btn.addEventListener("click", e=>this.final_boosties_submit(e));
                if (el_bootsies_input!=null) {
                    el_bootsies_input.addEventListener("input", e=>this.annecoin_entered(e));
                }
                if (el_boosties_btn!=null) el_boosties_btn.addEventListener("click", function($e){
                    ANNE.A.show(el_boosties_amt_div);
                    ANNE.A.hide(el_boosties_btn);
                    el_bootsies_input.focus();
                });
            }
        }

        else {
            if (el_annex_btn!=null) {
                if (ANNE.USER.isActivated()==true) {
                    ANNE.A.show(el_annex_btn);
                }
                else {
                    ANNE.A.show("getanne_btn")
                }
                
            }
            if (el_boosties_btn!=null) ANNE.A.hide(el_boosties_btn);
            if ($bDOEVENTS==true) {
                if (ANNE.USER.isActivated()==true) {
                    if (el_annex_btn!==null) {
                        el_annex_btn.addEventListener("click", (e)=>{
                            this.annexNow(e);
                            ANNE.A.hide(el_annex_btn);
                        });
                    }
                }
                else {
                    var el_get = document.getElementById("getanne_btn"+this._ID);
                    el_get.addEventListener("click", ($e)=>{
                        $e.stopImmediatePropagation();
                        $e.preventDefault();
                        ANNE.MEM_UDF.popupElCancel();
                        ANNE.playSound("readytorumble");
                        ANNE.MEM_UDF.loadSwaps("mineORavailable");
                    });   
                }
            }
        }

        if (this.backlink!=null) {
            var el_backlink = document.getElementById("annex_pending"+this._ID);
            if (el_backlink==null) {
                return true;
            }
            var what = "";
            if (this.backlink.r==ANNE.MEM.annex.temp_nid) {
                what="annex ";
            }
            else if (this.backlink.r==ANNE.MEM.boosties.temp_nid) {
                what="boosties ";
            }
            el_backlink.innerHTML="MEMPOOL HAS PENDING "+what+"CHANGE";
            ANNE.A.show(el_backlink);
            ANNE.A.hide(el_annex_btn);
            ANNE.A.hide(el_boosties_btn);
            var el_refresh = ANNE.A.show("annex_pending_refresh"+this._ID);
            if (el_refresh!=null) el_refresh.addEventListener("click", e=>this.refreshThis(e));
        }
        
        else if (this.has_pending==true) {
            console.log(this.nObj._aplex);
            var el_backlink = document.getElementById("annex_pending"+this._ID);
            if (el_backlink==null) {
                return true;
            }
            el_backlink.innerHTML="MEMPOOL HAS PENDING CHANGE";
            ANNE.A.show(el_backlink);
            ANNE.A.hide(el_annex_btn);
            ANNE.A.hide(el_boosties_btn);
            var el_refresh = ANNE.A.show("annex_pending_refresh"+this._ID);
            if (el_refresh!=null) {
                el_refresh.addEventListener("click", e=>this.refreshThis(e));
            }
        }

        //do we freeze on annex pending still?
        //have to check the annode.. arrrrg.
        else if (this.nObj.freeze==1) {
            el_annex_btn.setAttribute("disabled", true);
            el_annex_btn.setAttribute("title", "When prev TXs are mined the neuron will be unfrozen.");
            ANNE.A.hide(el_annex_btn);
            el_boosties_btn.setAttribute("disabled", true);
            el_boosties_btn.setAttribute("title", "When prev TXs are mined the neuron will be unfrozen.");
            ANNE.A.hide(el_boosties_btn);
        }

        return true;
    }
    

    /**
     * Refresh this..
     * @param {*} $e 
     */
    refreshThis = async($e)=>{
        if (this.dossierPTR!=null) {
            this.dossierPTR.refreshAnnex();
        }
    }


    /**
     * in TXB handler we will set the .has_pending for this session.
     * It will get updated when a new _annex obj is pulled.
     */
    annexNow = async($e)=>{
        var aplex = this.nObj._aplex;
        var gui_extras = new Object();
        gui_extras.payment = aplex.calculated_cost;
        
        //this is already put in the sendQ.
        var sysmsg_rel = await ANNE.MEMZ.ez_new_sys_msg(USER._userNObj, MEM.annex.nid, TO_TYPE_ANNE, this.nObj.nid, gui_extras, null);
        //var sysmsg_rel = await ANNE.GHOSTDATA.new_ghost_sys_msg(USER._userNObj, MEM.annex.nid, TO_TYPE_ANNE, this.nObj.nid, aplex.calculated_cost);
        console.log(sysmsg_rel.annies);

        //critical for this to get accepted via the node if it is anything other than 
        //the first lvl of annexing.  All non-first lvls must reference the previous annex
        //transaction to create a chain of annexings that flow for annex documentation in time.
        //We have this data without the chain but it is an extra helper that is useful.
        if (aplex.last_fullhash!=undefined && aplex.last_fullhash!="") {
            sysmsg_rel.referencedTransactionFullHash = aplex.last_fullhash;
        }

        sysmsg_rel.annies = aplex.calculated_cost;

        ANNE.MEM_UDF.sendQueue();
        ANNE.MEM_UDF.popupElCancel();
    }
}


/************** (from file ./js/base/annex.js)**/ 

"use strict";

(function(root) {

    function checkInt(value) {
        return (parseInt(value) === value);
    }

    function checkInts(arrayish) {
        if (!checkInt(arrayish.length)) { return false; }

        for (var i = 0; i < arrayish.length; i++) {
            if (!checkInt(arrayish[i]) || arrayish[i] < 0 || arrayish[i] > 255) {
                return false;
            }
        }

        return true;
    }

    function coerceArray(arg, copy) {

        // ArrayBuffer view
        if (arg.buffer && ArrayBuffer.isView(arg) && arg.name === 'Uint8Array') {

            if (copy) {
                if (arg.slice) {
                    arg = arg.slice();
                } else {
                    arg = Array.prototype.slice.call(arg);
                }
            }

            return arg;
        }

        // It's an array; check it is a valid representation of a byte
        if (Array.isArray(arg)) {
            if (!checkInts(arg)) {
                throw new Error('Array contains invalid value: ' + arg);
            }

            return new Uint8Array(arg);
        }

        // Something else, but behaves like an array (maybe a Buffer? Arguments?)
        if (checkInt(arg.length) && checkInts(arg)) {
            return new Uint8Array(arg);
        }

        throw new Error('unsupported array-like object');
    }

    function createArray(length) {
        return new Uint8Array(length);
    }

    function copyArray(sourceArray, targetArray, targetStart, sourceStart, sourceEnd) {
        if (sourceStart != null || sourceEnd != null) {
            if (sourceArray.slice) {
                sourceArray = sourceArray.slice(sourceStart, sourceEnd);
            } else {
                sourceArray = Array.prototype.slice.call(sourceArray, sourceStart, sourceEnd);
            }
        }
        targetArray.set(sourceArray, targetStart);
    }



    var convertUtf8 = (function() {
        function toBytes(text) {
            var result = [], i = 0;
            text = encodeURI(text);
            while (i < text.length) {
                var c = text.charCodeAt(i++);

                // if it is a % sign, encode the following 2 bytes as a hex value
                if (c === 37) {
                    result.push(parseInt(text.substr(i, 2), 16))
                    i += 2;

                // otherwise, just the actual byte
                } else {
                    result.push(c)
                }
            }

            return coerceArray(result);
        }

        function fromBytes(bytes) {
            var result = [], i = 0;

            while (i < bytes.length) {
                var c = bytes[i];

                if (c < 128) {
                    result.push(String.fromCharCode(c));
                    i++;
                } else if (c > 191 && c < 224) {
                    result.push(String.fromCharCode(((c & 0x1f) << 6) | (bytes[i + 1] & 0x3f)));
                    i += 2;
                } else {
                    result.push(String.fromCharCode(((c & 0x0f) << 12) | ((bytes[i + 1] & 0x3f) << 6) | (bytes[i + 2] & 0x3f)));
                    i += 3;
                }
            }

            return result.join('');
        }

        return {
            toBytes: toBytes,
            fromBytes: fromBytes,
        }
    })();

    var convertHex = (function() {
        function toBytes(text) {
            var result = [];
            for (var i = 0; i < text.length; i += 2) {
                result.push(parseInt(text.substr(i, 2), 16));
            }

            return result;
        }

        // http://ixti.net/development/javascript/2011/11/11/base64-encodedecode-of-utf8-in-browser-with-js.html
        var Hex = '0123456789abcdef';

        function fromBytes(bytes) {
                var result = [];
                for (var i = 0; i < bytes.length; i++) {
                    var v = bytes[i];
                    result.push(Hex[(v & 0xf0) >> 4] + Hex[v & 0x0f]);
                }
                return result.join('');
        }

        return {
            toBytes: toBytes,
            fromBytes: fromBytes,
        }
    })();


    // Number of rounds by keysize
    var numberOfRounds = {16: 10, 24: 12, 32: 14}

    // Round constant words
    var rcon = [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91];

    // S-box and Inverse S-box (S is for Substitution)
    var S = [0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76, 0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0, 0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15, 0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75, 0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84, 0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf, 0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8, 0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2, 0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73, 0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb, 0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79, 0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08, 0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a, 0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e, 0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf, 0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16];
    var Si =[0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb, 0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb, 0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e, 0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25, 0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92, 0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84, 0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06, 0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b, 0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73, 0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e, 0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b, 0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4, 0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f, 0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef, 0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61, 0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d];

    // Transformations for encryption
    var T1 = [0xc66363a5, 0xf87c7c84, 0xee777799, 0xf67b7b8d, 0xfff2f20d, 0xd66b6bbd, 0xde6f6fb1, 0x91c5c554, 0x60303050, 0x02010103, 0xce6767a9, 0x562b2b7d, 0xe7fefe19, 0xb5d7d762, 0x4dababe6, 0xec76769a, 0x8fcaca45, 0x1f82829d, 0x89c9c940, 0xfa7d7d87, 0xeffafa15, 0xb25959eb, 0x8e4747c9, 0xfbf0f00b, 0x41adadec, 0xb3d4d467, 0x5fa2a2fd, 0x45afafea, 0x239c9cbf, 0x53a4a4f7, 0xe4727296, 0x9bc0c05b, 0x75b7b7c2, 0xe1fdfd1c, 0x3d9393ae, 0x4c26266a, 0x6c36365a, 0x7e3f3f41, 0xf5f7f702, 0x83cccc4f, 0x6834345c, 0x51a5a5f4, 0xd1e5e534, 0xf9f1f108, 0xe2717193, 0xabd8d873, 0x62313153, 0x2a15153f, 0x0804040c, 0x95c7c752, 0x46232365, 0x9dc3c35e, 0x30181828, 0x379696a1, 0x0a05050f, 0x2f9a9ab5, 0x0e070709, 0x24121236, 0x1b80809b, 0xdfe2e23d, 0xcdebeb26, 0x4e272769, 0x7fb2b2cd, 0xea75759f, 0x1209091b, 0x1d83839e, 0x582c2c74, 0x341a1a2e, 0x361b1b2d, 0xdc6e6eb2, 0xb45a5aee, 0x5ba0a0fb, 0xa45252f6, 0x763b3b4d, 0xb7d6d661, 0x7db3b3ce, 0x5229297b, 0xdde3e33e, 0x5e2f2f71, 0x13848497, 0xa65353f5, 0xb9d1d168, 0x00000000, 0xc1eded2c, 0x40202060, 0xe3fcfc1f, 0x79b1b1c8, 0xb65b5bed, 0xd46a6abe, 0x8dcbcb46, 0x67bebed9, 0x7239394b, 0x944a4ade, 0x984c4cd4, 0xb05858e8, 0x85cfcf4a, 0xbbd0d06b, 0xc5efef2a, 0x4faaaae5, 0xedfbfb16, 0x864343c5, 0x9a4d4dd7, 0x66333355, 0x11858594, 0x8a4545cf, 0xe9f9f910, 0x04020206, 0xfe7f7f81, 0xa05050f0, 0x783c3c44, 0x259f9fba, 0x4ba8a8e3, 0xa25151f3, 0x5da3a3fe, 0x804040c0, 0x058f8f8a, 0x3f9292ad, 0x219d9dbc, 0x70383848, 0xf1f5f504, 0x63bcbcdf, 0x77b6b6c1, 0xafdada75, 0x42212163, 0x20101030, 0xe5ffff1a, 0xfdf3f30e, 0xbfd2d26d, 0x81cdcd4c, 0x180c0c14, 0x26131335, 0xc3ecec2f, 0xbe5f5fe1, 0x359797a2, 0x884444cc, 0x2e171739, 0x93c4c457, 0x55a7a7f2, 0xfc7e7e82, 0x7a3d3d47, 0xc86464ac, 0xba5d5de7, 0x3219192b, 0xe6737395, 0xc06060a0, 0x19818198, 0x9e4f4fd1, 0xa3dcdc7f, 0x44222266, 0x542a2a7e, 0x3b9090ab, 0x0b888883, 0x8c4646ca, 0xc7eeee29, 0x6bb8b8d3, 0x2814143c, 0xa7dede79, 0xbc5e5ee2, 0x160b0b1d, 0xaddbdb76, 0xdbe0e03b, 0x64323256, 0x743a3a4e, 0x140a0a1e, 0x924949db, 0x0c06060a, 0x4824246c, 0xb85c5ce4, 0x9fc2c25d, 0xbdd3d36e, 0x43acacef, 0xc46262a6, 0x399191a8, 0x319595a4, 0xd3e4e437, 0xf279798b, 0xd5e7e732, 0x8bc8c843, 0x6e373759, 0xda6d6db7, 0x018d8d8c, 0xb1d5d564, 0x9c4e4ed2, 0x49a9a9e0, 0xd86c6cb4, 0xac5656fa, 0xf3f4f407, 0xcfeaea25, 0xca6565af, 0xf47a7a8e, 0x47aeaee9, 0x10080818, 0x6fbabad5, 0xf0787888, 0x4a25256f, 0x5c2e2e72, 0x381c1c24, 0x57a6a6f1, 0x73b4b4c7, 0x97c6c651, 0xcbe8e823, 0xa1dddd7c, 0xe874749c, 0x3e1f1f21, 0x964b4bdd, 0x61bdbddc, 0x0d8b8b86, 0x0f8a8a85, 0xe0707090, 0x7c3e3e42, 0x71b5b5c4, 0xcc6666aa, 0x904848d8, 0x06030305, 0xf7f6f601, 0x1c0e0e12, 0xc26161a3, 0x6a35355f, 0xae5757f9, 0x69b9b9d0, 0x17868691, 0x99c1c158, 0x3a1d1d27, 0x279e9eb9, 0xd9e1e138, 0xebf8f813, 0x2b9898b3, 0x22111133, 0xd26969bb, 0xa9d9d970, 0x078e8e89, 0x339494a7, 0x2d9b9bb6, 0x3c1e1e22, 0x15878792, 0xc9e9e920, 0x87cece49, 0xaa5555ff, 0x50282878, 0xa5dfdf7a, 0x038c8c8f, 0x59a1a1f8, 0x09898980, 0x1a0d0d17, 0x65bfbfda, 0xd7e6e631, 0x844242c6, 0xd06868b8, 0x824141c3, 0x299999b0, 0x5a2d2d77, 0x1e0f0f11, 0x7bb0b0cb, 0xa85454fc, 0x6dbbbbd6, 0x2c16163a];
    var T2 = [0xa5c66363, 0x84f87c7c, 0x99ee7777, 0x8df67b7b, 0x0dfff2f2, 0xbdd66b6b, 0xb1de6f6f, 0x5491c5c5, 0x50603030, 0x03020101, 0xa9ce6767, 0x7d562b2b, 0x19e7fefe, 0x62b5d7d7, 0xe64dabab, 0x9aec7676, 0x458fcaca, 0x9d1f8282, 0x4089c9c9, 0x87fa7d7d, 0x15effafa, 0xebb25959, 0xc98e4747, 0x0bfbf0f0, 0xec41adad, 0x67b3d4d4, 0xfd5fa2a2, 0xea45afaf, 0xbf239c9c, 0xf753a4a4, 0x96e47272, 0x5b9bc0c0, 0xc275b7b7, 0x1ce1fdfd, 0xae3d9393, 0x6a4c2626, 0x5a6c3636, 0x417e3f3f, 0x02f5f7f7, 0x4f83cccc, 0x5c683434, 0xf451a5a5, 0x34d1e5e5, 0x08f9f1f1, 0x93e27171, 0x73abd8d8, 0x53623131, 0x3f2a1515, 0x0c080404, 0x5295c7c7, 0x65462323, 0x5e9dc3c3, 0x28301818, 0xa1379696, 0x0f0a0505, 0xb52f9a9a, 0x090e0707, 0x36241212, 0x9b1b8080, 0x3ddfe2e2, 0x26cdebeb, 0x694e2727, 0xcd7fb2b2, 0x9fea7575, 0x1b120909, 0x9e1d8383, 0x74582c2c, 0x2e341a1a, 0x2d361b1b, 0xb2dc6e6e, 0xeeb45a5a, 0xfb5ba0a0, 0xf6a45252, 0x4d763b3b, 0x61b7d6d6, 0xce7db3b3, 0x7b522929, 0x3edde3e3, 0x715e2f2f, 0x97138484, 0xf5a65353, 0x68b9d1d1, 0x00000000, 0x2cc1eded, 0x60402020, 0x1fe3fcfc, 0xc879b1b1, 0xedb65b5b, 0xbed46a6a, 0x468dcbcb, 0xd967bebe, 0x4b723939, 0xde944a4a, 0xd4984c4c, 0xe8b05858, 0x4a85cfcf, 0x6bbbd0d0, 0x2ac5efef, 0xe54faaaa, 0x16edfbfb, 0xc5864343, 0xd79a4d4d, 0x55663333, 0x94118585, 0xcf8a4545, 0x10e9f9f9, 0x06040202, 0x81fe7f7f, 0xf0a05050, 0x44783c3c, 0xba259f9f, 0xe34ba8a8, 0xf3a25151, 0xfe5da3a3, 0xc0804040, 0x8a058f8f, 0xad3f9292, 0xbc219d9d, 0x48703838, 0x04f1f5f5, 0xdf63bcbc, 0xc177b6b6, 0x75afdada, 0x63422121, 0x30201010, 0x1ae5ffff, 0x0efdf3f3, 0x6dbfd2d2, 0x4c81cdcd, 0x14180c0c, 0x35261313, 0x2fc3ecec, 0xe1be5f5f, 0xa2359797, 0xcc884444, 0x392e1717, 0x5793c4c4, 0xf255a7a7, 0x82fc7e7e, 0x477a3d3d, 0xacc86464, 0xe7ba5d5d, 0x2b321919, 0x95e67373, 0xa0c06060, 0x98198181, 0xd19e4f4f, 0x7fa3dcdc, 0x66442222, 0x7e542a2a, 0xab3b9090, 0x830b8888, 0xca8c4646, 0x29c7eeee, 0xd36bb8b8, 0x3c281414, 0x79a7dede, 0xe2bc5e5e, 0x1d160b0b, 0x76addbdb, 0x3bdbe0e0, 0x56643232, 0x4e743a3a, 0x1e140a0a, 0xdb924949, 0x0a0c0606, 0x6c482424, 0xe4b85c5c, 0x5d9fc2c2, 0x6ebdd3d3, 0xef43acac, 0xa6c46262, 0xa8399191, 0xa4319595, 0x37d3e4e4, 0x8bf27979, 0x32d5e7e7, 0x438bc8c8, 0x596e3737, 0xb7da6d6d, 0x8c018d8d, 0x64b1d5d5, 0xd29c4e4e, 0xe049a9a9, 0xb4d86c6c, 0xfaac5656, 0x07f3f4f4, 0x25cfeaea, 0xafca6565, 0x8ef47a7a, 0xe947aeae, 0x18100808, 0xd56fbaba, 0x88f07878, 0x6f4a2525, 0x725c2e2e, 0x24381c1c, 0xf157a6a6, 0xc773b4b4, 0x5197c6c6, 0x23cbe8e8, 0x7ca1dddd, 0x9ce87474, 0x213e1f1f, 0xdd964b4b, 0xdc61bdbd, 0x860d8b8b, 0x850f8a8a, 0x90e07070, 0x427c3e3e, 0xc471b5b5, 0xaacc6666, 0xd8904848, 0x05060303, 0x01f7f6f6, 0x121c0e0e, 0xa3c26161, 0x5f6a3535, 0xf9ae5757, 0xd069b9b9, 0x91178686, 0x5899c1c1, 0x273a1d1d, 0xb9279e9e, 0x38d9e1e1, 0x13ebf8f8, 0xb32b9898, 0x33221111, 0xbbd26969, 0x70a9d9d9, 0x89078e8e, 0xa7339494, 0xb62d9b9b, 0x223c1e1e, 0x92158787, 0x20c9e9e9, 0x4987cece, 0xffaa5555, 0x78502828, 0x7aa5dfdf, 0x8f038c8c, 0xf859a1a1, 0x80098989, 0x171a0d0d, 0xda65bfbf, 0x31d7e6e6, 0xc6844242, 0xb8d06868, 0xc3824141, 0xb0299999, 0x775a2d2d, 0x111e0f0f, 0xcb7bb0b0, 0xfca85454, 0xd66dbbbb, 0x3a2c1616];
    var T3 = [0x63a5c663, 0x7c84f87c, 0x7799ee77, 0x7b8df67b, 0xf20dfff2, 0x6bbdd66b, 0x6fb1de6f, 0xc55491c5, 0x30506030, 0x01030201, 0x67a9ce67, 0x2b7d562b, 0xfe19e7fe, 0xd762b5d7, 0xabe64dab, 0x769aec76, 0xca458fca, 0x829d1f82, 0xc94089c9, 0x7d87fa7d, 0xfa15effa, 0x59ebb259, 0x47c98e47, 0xf00bfbf0, 0xadec41ad, 0xd467b3d4, 0xa2fd5fa2, 0xafea45af, 0x9cbf239c, 0xa4f753a4, 0x7296e472, 0xc05b9bc0, 0xb7c275b7, 0xfd1ce1fd, 0x93ae3d93, 0x266a4c26, 0x365a6c36, 0x3f417e3f, 0xf702f5f7, 0xcc4f83cc, 0x345c6834, 0xa5f451a5, 0xe534d1e5, 0xf108f9f1, 0x7193e271, 0xd873abd8, 0x31536231, 0x153f2a15, 0x040c0804, 0xc75295c7, 0x23654623, 0xc35e9dc3, 0x18283018, 0x96a13796, 0x050f0a05, 0x9ab52f9a, 0x07090e07, 0x12362412, 0x809b1b80, 0xe23ddfe2, 0xeb26cdeb, 0x27694e27, 0xb2cd7fb2, 0x759fea75, 0x091b1209, 0x839e1d83, 0x2c74582c, 0x1a2e341a, 0x1b2d361b, 0x6eb2dc6e, 0x5aeeb45a, 0xa0fb5ba0, 0x52f6a452, 0x3b4d763b, 0xd661b7d6, 0xb3ce7db3, 0x297b5229, 0xe33edde3, 0x2f715e2f, 0x84971384, 0x53f5a653, 0xd168b9d1, 0x00000000, 0xed2cc1ed, 0x20604020, 0xfc1fe3fc, 0xb1c879b1, 0x5bedb65b, 0x6abed46a, 0xcb468dcb, 0xbed967be, 0x394b7239, 0x4ade944a, 0x4cd4984c, 0x58e8b058, 0xcf4a85cf, 0xd06bbbd0, 0xef2ac5ef, 0xaae54faa, 0xfb16edfb, 0x43c58643, 0x4dd79a4d, 0x33556633, 0x85941185, 0x45cf8a45, 0xf910e9f9, 0x02060402, 0x7f81fe7f, 0x50f0a050, 0x3c44783c, 0x9fba259f, 0xa8e34ba8, 0x51f3a251, 0xa3fe5da3, 0x40c08040, 0x8f8a058f, 0x92ad3f92, 0x9dbc219d, 0x38487038, 0xf504f1f5, 0xbcdf63bc, 0xb6c177b6, 0xda75afda, 0x21634221, 0x10302010, 0xff1ae5ff, 0xf30efdf3, 0xd26dbfd2, 0xcd4c81cd, 0x0c14180c, 0x13352613, 0xec2fc3ec, 0x5fe1be5f, 0x97a23597, 0x44cc8844, 0x17392e17, 0xc45793c4, 0xa7f255a7, 0x7e82fc7e, 0x3d477a3d, 0x64acc864, 0x5de7ba5d, 0x192b3219, 0x7395e673, 0x60a0c060, 0x81981981, 0x4fd19e4f, 0xdc7fa3dc, 0x22664422, 0x2a7e542a, 0x90ab3b90, 0x88830b88, 0x46ca8c46, 0xee29c7ee, 0xb8d36bb8, 0x143c2814, 0xde79a7de, 0x5ee2bc5e, 0x0b1d160b, 0xdb76addb, 0xe03bdbe0, 0x32566432, 0x3a4e743a, 0x0a1e140a, 0x49db9249, 0x060a0c06, 0x246c4824, 0x5ce4b85c, 0xc25d9fc2, 0xd36ebdd3, 0xacef43ac, 0x62a6c462, 0x91a83991, 0x95a43195, 0xe437d3e4, 0x798bf279, 0xe732d5e7, 0xc8438bc8, 0x37596e37, 0x6db7da6d, 0x8d8c018d, 0xd564b1d5, 0x4ed29c4e, 0xa9e049a9, 0x6cb4d86c, 0x56faac56, 0xf407f3f4, 0xea25cfea, 0x65afca65, 0x7a8ef47a, 0xaee947ae, 0x08181008, 0xbad56fba, 0x7888f078, 0x256f4a25, 0x2e725c2e, 0x1c24381c, 0xa6f157a6, 0xb4c773b4, 0xc65197c6, 0xe823cbe8, 0xdd7ca1dd, 0x749ce874, 0x1f213e1f, 0x4bdd964b, 0xbddc61bd, 0x8b860d8b, 0x8a850f8a, 0x7090e070, 0x3e427c3e, 0xb5c471b5, 0x66aacc66, 0x48d89048, 0x03050603, 0xf601f7f6, 0x0e121c0e, 0x61a3c261, 0x355f6a35, 0x57f9ae57, 0xb9d069b9, 0x86911786, 0xc15899c1, 0x1d273a1d, 0x9eb9279e, 0xe138d9e1, 0xf813ebf8, 0x98b32b98, 0x11332211, 0x69bbd269, 0xd970a9d9, 0x8e89078e, 0x94a73394, 0x9bb62d9b, 0x1e223c1e, 0x87921587, 0xe920c9e9, 0xce4987ce, 0x55ffaa55, 0x28785028, 0xdf7aa5df, 0x8c8f038c, 0xa1f859a1, 0x89800989, 0x0d171a0d, 0xbfda65bf, 0xe631d7e6, 0x42c68442, 0x68b8d068, 0x41c38241, 0x99b02999, 0x2d775a2d, 0x0f111e0f, 0xb0cb7bb0, 0x54fca854, 0xbbd66dbb, 0x163a2c16];
    var T4 = [0x6363a5c6, 0x7c7c84f8, 0x777799ee, 0x7b7b8df6, 0xf2f20dff, 0x6b6bbdd6, 0x6f6fb1de, 0xc5c55491, 0x30305060, 0x01010302, 0x6767a9ce, 0x2b2b7d56, 0xfefe19e7, 0xd7d762b5, 0xababe64d, 0x76769aec, 0xcaca458f, 0x82829d1f, 0xc9c94089, 0x7d7d87fa, 0xfafa15ef, 0x5959ebb2, 0x4747c98e, 0xf0f00bfb, 0xadadec41, 0xd4d467b3, 0xa2a2fd5f, 0xafafea45, 0x9c9cbf23, 0xa4a4f753, 0x727296e4, 0xc0c05b9b, 0xb7b7c275, 0xfdfd1ce1, 0x9393ae3d, 0x26266a4c, 0x36365a6c, 0x3f3f417e, 0xf7f702f5, 0xcccc4f83, 0x34345c68, 0xa5a5f451, 0xe5e534d1, 0xf1f108f9, 0x717193e2, 0xd8d873ab, 0x31315362, 0x15153f2a, 0x04040c08, 0xc7c75295, 0x23236546, 0xc3c35e9d, 0x18182830, 0x9696a137, 0x05050f0a, 0x9a9ab52f, 0x0707090e, 0x12123624, 0x80809b1b, 0xe2e23ddf, 0xebeb26cd, 0x2727694e, 0xb2b2cd7f, 0x75759fea, 0x09091b12, 0x83839e1d, 0x2c2c7458, 0x1a1a2e34, 0x1b1b2d36, 0x6e6eb2dc, 0x5a5aeeb4, 0xa0a0fb5b, 0x5252f6a4, 0x3b3b4d76, 0xd6d661b7, 0xb3b3ce7d, 0x29297b52, 0xe3e33edd, 0x2f2f715e, 0x84849713, 0x5353f5a6, 0xd1d168b9, 0x00000000, 0xeded2cc1, 0x20206040, 0xfcfc1fe3, 0xb1b1c879, 0x5b5bedb6, 0x6a6abed4, 0xcbcb468d, 0xbebed967, 0x39394b72, 0x4a4ade94, 0x4c4cd498, 0x5858e8b0, 0xcfcf4a85, 0xd0d06bbb, 0xefef2ac5, 0xaaaae54f, 0xfbfb16ed, 0x4343c586, 0x4d4dd79a, 0x33335566, 0x85859411, 0x4545cf8a, 0xf9f910e9, 0x02020604, 0x7f7f81fe, 0x5050f0a0, 0x3c3c4478, 0x9f9fba25, 0xa8a8e34b, 0x5151f3a2, 0xa3a3fe5d, 0x4040c080, 0x8f8f8a05, 0x9292ad3f, 0x9d9dbc21, 0x38384870, 0xf5f504f1, 0xbcbcdf63, 0xb6b6c177, 0xdada75af, 0x21216342, 0x10103020, 0xffff1ae5, 0xf3f30efd, 0xd2d26dbf, 0xcdcd4c81, 0x0c0c1418, 0x13133526, 0xecec2fc3, 0x5f5fe1be, 0x9797a235, 0x4444cc88, 0x1717392e, 0xc4c45793, 0xa7a7f255, 0x7e7e82fc, 0x3d3d477a, 0x6464acc8, 0x5d5de7ba, 0x19192b32, 0x737395e6, 0x6060a0c0, 0x81819819, 0x4f4fd19e, 0xdcdc7fa3, 0x22226644, 0x2a2a7e54, 0x9090ab3b, 0x8888830b, 0x4646ca8c, 0xeeee29c7, 0xb8b8d36b, 0x14143c28, 0xdede79a7, 0x5e5ee2bc, 0x0b0b1d16, 0xdbdb76ad, 0xe0e03bdb, 0x32325664, 0x3a3a4e74, 0x0a0a1e14, 0x4949db92, 0x06060a0c, 0x24246c48, 0x5c5ce4b8, 0xc2c25d9f, 0xd3d36ebd, 0xacacef43, 0x6262a6c4, 0x9191a839, 0x9595a431, 0xe4e437d3, 0x79798bf2, 0xe7e732d5, 0xc8c8438b, 0x3737596e, 0x6d6db7da, 0x8d8d8c01, 0xd5d564b1, 0x4e4ed29c, 0xa9a9e049, 0x6c6cb4d8, 0x5656faac, 0xf4f407f3, 0xeaea25cf, 0x6565afca, 0x7a7a8ef4, 0xaeaee947, 0x08081810, 0xbabad56f, 0x787888f0, 0x25256f4a, 0x2e2e725c, 0x1c1c2438, 0xa6a6f157, 0xb4b4c773, 0xc6c65197, 0xe8e823cb, 0xdddd7ca1, 0x74749ce8, 0x1f1f213e, 0x4b4bdd96, 0xbdbddc61, 0x8b8b860d, 0x8a8a850f, 0x707090e0, 0x3e3e427c, 0xb5b5c471, 0x6666aacc, 0x4848d890, 0x03030506, 0xf6f601f7, 0x0e0e121c, 0x6161a3c2, 0x35355f6a, 0x5757f9ae, 0xb9b9d069, 0x86869117, 0xc1c15899, 0x1d1d273a, 0x9e9eb927, 0xe1e138d9, 0xf8f813eb, 0x9898b32b, 0x11113322, 0x6969bbd2, 0xd9d970a9, 0x8e8e8907, 0x9494a733, 0x9b9bb62d, 0x1e1e223c, 0x87879215, 0xe9e920c9, 0xcece4987, 0x5555ffaa, 0x28287850, 0xdfdf7aa5, 0x8c8c8f03, 0xa1a1f859, 0x89898009, 0x0d0d171a, 0xbfbfda65, 0xe6e631d7, 0x4242c684, 0x6868b8d0, 0x4141c382, 0x9999b029, 0x2d2d775a, 0x0f0f111e, 0xb0b0cb7b, 0x5454fca8, 0xbbbbd66d, 0x16163a2c];

    // Transformations for decryption
    var T5 = [0x51f4a750, 0x7e416553, 0x1a17a4c3, 0x3a275e96, 0x3bab6bcb, 0x1f9d45f1, 0xacfa58ab, 0x4be30393, 0x2030fa55, 0xad766df6, 0x88cc7691, 0xf5024c25, 0x4fe5d7fc, 0xc52acbd7, 0x26354480, 0xb562a38f, 0xdeb15a49, 0x25ba1b67, 0x45ea0e98, 0x5dfec0e1, 0xc32f7502, 0x814cf012, 0x8d4697a3, 0x6bd3f9c6, 0x038f5fe7, 0x15929c95, 0xbf6d7aeb, 0x955259da, 0xd4be832d, 0x587421d3, 0x49e06929, 0x8ec9c844, 0x75c2896a, 0xf48e7978, 0x99583e6b, 0x27b971dd, 0xbee14fb6, 0xf088ad17, 0xc920ac66, 0x7dce3ab4, 0x63df4a18, 0xe51a3182, 0x97513360, 0x62537f45, 0xb16477e0, 0xbb6bae84, 0xfe81a01c, 0xf9082b94, 0x70486858, 0x8f45fd19, 0x94de6c87, 0x527bf8b7, 0xab73d323, 0x724b02e2, 0xe31f8f57, 0x6655ab2a, 0xb2eb2807, 0x2fb5c203, 0x86c57b9a, 0xd33708a5, 0x302887f2, 0x23bfa5b2, 0x02036aba, 0xed16825c, 0x8acf1c2b, 0xa779b492, 0xf307f2f0, 0x4e69e2a1, 0x65daf4cd, 0x0605bed5, 0xd134621f, 0xc4a6fe8a, 0x342e539d, 0xa2f355a0, 0x058ae132, 0xa4f6eb75, 0x0b83ec39, 0x4060efaa, 0x5e719f06, 0xbd6e1051, 0x3e218af9, 0x96dd063d, 0xdd3e05ae, 0x4de6bd46, 0x91548db5, 0x71c45d05, 0x0406d46f, 0x605015ff, 0x1998fb24, 0xd6bde997, 0x894043cc, 0x67d99e77, 0xb0e842bd, 0x07898b88, 0xe7195b38, 0x79c8eedb, 0xa17c0a47, 0x7c420fe9, 0xf8841ec9, 0x00000000, 0x09808683, 0x322bed48, 0x1e1170ac, 0x6c5a724e, 0xfd0efffb, 0x0f853856, 0x3daed51e, 0x362d3927, 0x0a0fd964, 0x685ca621, 0x9b5b54d1, 0x24362e3a, 0x0c0a67b1, 0x9357e70f, 0xb4ee96d2, 0x1b9b919e, 0x80c0c54f, 0x61dc20a2, 0x5a774b69, 0x1c121a16, 0xe293ba0a, 0xc0a02ae5, 0x3c22e043, 0x121b171d, 0x0e090d0b, 0xf28bc7ad, 0x2db6a8b9, 0x141ea9c8, 0x57f11985, 0xaf75074c, 0xee99ddbb, 0xa37f60fd, 0xf701269f, 0x5c72f5bc, 0x44663bc5, 0x5bfb7e34, 0x8b432976, 0xcb23c6dc, 0xb6edfc68, 0xb8e4f163, 0xd731dcca, 0x42638510, 0x13972240, 0x84c61120, 0x854a247d, 0xd2bb3df8, 0xaef93211, 0xc729a16d, 0x1d9e2f4b, 0xdcb230f3, 0x0d8652ec, 0x77c1e3d0, 0x2bb3166c, 0xa970b999, 0x119448fa, 0x47e96422, 0xa8fc8cc4, 0xa0f03f1a, 0x567d2cd8, 0x223390ef, 0x87494ec7, 0xd938d1c1, 0x8ccaa2fe, 0x98d40b36, 0xa6f581cf, 0xa57ade28, 0xdab78e26, 0x3fadbfa4, 0x2c3a9de4, 0x5078920d, 0x6a5fcc9b, 0x547e4662, 0xf68d13c2, 0x90d8b8e8, 0x2e39f75e, 0x82c3aff5, 0x9f5d80be, 0x69d0937c, 0x6fd52da9, 0xcf2512b3, 0xc8ac993b, 0x10187da7, 0xe89c636e, 0xdb3bbb7b, 0xcd267809, 0x6e5918f4, 0xec9ab701, 0x834f9aa8, 0xe6956e65, 0xaaffe67e, 0x21bccf08, 0xef15e8e6, 0xbae79bd9, 0x4a6f36ce, 0xea9f09d4, 0x29b07cd6, 0x31a4b2af, 0x2a3f2331, 0xc6a59430, 0x35a266c0, 0x744ebc37, 0xfc82caa6, 0xe090d0b0, 0x33a7d815, 0xf104984a, 0x41ecdaf7, 0x7fcd500e, 0x1791f62f, 0x764dd68d, 0x43efb04d, 0xccaa4d54, 0xe49604df, 0x9ed1b5e3, 0x4c6a881b, 0xc12c1fb8, 0x4665517f, 0x9d5eea04, 0x018c355d, 0xfa877473, 0xfb0b412e, 0xb3671d5a, 0x92dbd252, 0xe9105633, 0x6dd64713, 0x9ad7618c, 0x37a10c7a, 0x59f8148e, 0xeb133c89, 0xcea927ee, 0xb761c935, 0xe11ce5ed, 0x7a47b13c, 0x9cd2df59, 0x55f2733f, 0x1814ce79, 0x73c737bf, 0x53f7cdea, 0x5ffdaa5b, 0xdf3d6f14, 0x7844db86, 0xcaaff381, 0xb968c43e, 0x3824342c, 0xc2a3405f, 0x161dc372, 0xbce2250c, 0x283c498b, 0xff0d9541, 0x39a80171, 0x080cb3de, 0xd8b4e49c, 0x6456c190, 0x7bcb8461, 0xd532b670, 0x486c5c74, 0xd0b85742];
    var T6 = [0x5051f4a7, 0x537e4165, 0xc31a17a4, 0x963a275e, 0xcb3bab6b, 0xf11f9d45, 0xabacfa58, 0x934be303, 0x552030fa, 0xf6ad766d, 0x9188cc76, 0x25f5024c, 0xfc4fe5d7, 0xd7c52acb, 0x80263544, 0x8fb562a3, 0x49deb15a, 0x6725ba1b, 0x9845ea0e, 0xe15dfec0, 0x02c32f75, 0x12814cf0, 0xa38d4697, 0xc66bd3f9, 0xe7038f5f, 0x9515929c, 0xebbf6d7a, 0xda955259, 0x2dd4be83, 0xd3587421, 0x2949e069, 0x448ec9c8, 0x6a75c289, 0x78f48e79, 0x6b99583e, 0xdd27b971, 0xb6bee14f, 0x17f088ad, 0x66c920ac, 0xb47dce3a, 0x1863df4a, 0x82e51a31, 0x60975133, 0x4562537f, 0xe0b16477, 0x84bb6bae, 0x1cfe81a0, 0x94f9082b, 0x58704868, 0x198f45fd, 0x8794de6c, 0xb7527bf8, 0x23ab73d3, 0xe2724b02, 0x57e31f8f, 0x2a6655ab, 0x07b2eb28, 0x032fb5c2, 0x9a86c57b, 0xa5d33708, 0xf2302887, 0xb223bfa5, 0xba02036a, 0x5ced1682, 0x2b8acf1c, 0x92a779b4, 0xf0f307f2, 0xa14e69e2, 0xcd65daf4, 0xd50605be, 0x1fd13462, 0x8ac4a6fe, 0x9d342e53, 0xa0a2f355, 0x32058ae1, 0x75a4f6eb, 0x390b83ec, 0xaa4060ef, 0x065e719f, 0x51bd6e10, 0xf93e218a, 0x3d96dd06, 0xaedd3e05, 0x464de6bd, 0xb591548d, 0x0571c45d, 0x6f0406d4, 0xff605015, 0x241998fb, 0x97d6bde9, 0xcc894043, 0x7767d99e, 0xbdb0e842, 0x8807898b, 0x38e7195b, 0xdb79c8ee, 0x47a17c0a, 0xe97c420f, 0xc9f8841e, 0x00000000, 0x83098086, 0x48322bed, 0xac1e1170, 0x4e6c5a72, 0xfbfd0eff, 0x560f8538, 0x1e3daed5, 0x27362d39, 0x640a0fd9, 0x21685ca6, 0xd19b5b54, 0x3a24362e, 0xb10c0a67, 0x0f9357e7, 0xd2b4ee96, 0x9e1b9b91, 0x4f80c0c5, 0xa261dc20, 0x695a774b, 0x161c121a, 0x0ae293ba, 0xe5c0a02a, 0x433c22e0, 0x1d121b17, 0x0b0e090d, 0xadf28bc7, 0xb92db6a8, 0xc8141ea9, 0x8557f119, 0x4caf7507, 0xbbee99dd, 0xfda37f60, 0x9ff70126, 0xbc5c72f5, 0xc544663b, 0x345bfb7e, 0x768b4329, 0xdccb23c6, 0x68b6edfc, 0x63b8e4f1, 0xcad731dc, 0x10426385, 0x40139722, 0x2084c611, 0x7d854a24, 0xf8d2bb3d, 0x11aef932, 0x6dc729a1, 0x4b1d9e2f, 0xf3dcb230, 0xec0d8652, 0xd077c1e3, 0x6c2bb316, 0x99a970b9, 0xfa119448, 0x2247e964, 0xc4a8fc8c, 0x1aa0f03f, 0xd8567d2c, 0xef223390, 0xc787494e, 0xc1d938d1, 0xfe8ccaa2, 0x3698d40b, 0xcfa6f581, 0x28a57ade, 0x26dab78e, 0xa43fadbf, 0xe42c3a9d, 0x0d507892, 0x9b6a5fcc, 0x62547e46, 0xc2f68d13, 0xe890d8b8, 0x5e2e39f7, 0xf582c3af, 0xbe9f5d80, 0x7c69d093, 0xa96fd52d, 0xb3cf2512, 0x3bc8ac99, 0xa710187d, 0x6ee89c63, 0x7bdb3bbb, 0x09cd2678, 0xf46e5918, 0x01ec9ab7, 0xa8834f9a, 0x65e6956e, 0x7eaaffe6, 0x0821bccf, 0xe6ef15e8, 0xd9bae79b, 0xce4a6f36, 0xd4ea9f09, 0xd629b07c, 0xaf31a4b2, 0x312a3f23, 0x30c6a594, 0xc035a266, 0x37744ebc, 0xa6fc82ca, 0xb0e090d0, 0x1533a7d8, 0x4af10498, 0xf741ecda, 0x0e7fcd50, 0x2f1791f6, 0x8d764dd6, 0x4d43efb0, 0x54ccaa4d, 0xdfe49604, 0xe39ed1b5, 0x1b4c6a88, 0xb8c12c1f, 0x7f466551, 0x049d5eea, 0x5d018c35, 0x73fa8774, 0x2efb0b41, 0x5ab3671d, 0x5292dbd2, 0x33e91056, 0x136dd647, 0x8c9ad761, 0x7a37a10c, 0x8e59f814, 0x89eb133c, 0xeecea927, 0x35b761c9, 0xede11ce5, 0x3c7a47b1, 0x599cd2df, 0x3f55f273, 0x791814ce, 0xbf73c737, 0xea53f7cd, 0x5b5ffdaa, 0x14df3d6f, 0x867844db, 0x81caaff3, 0x3eb968c4, 0x2c382434, 0x5fc2a340, 0x72161dc3, 0x0cbce225, 0x8b283c49, 0x41ff0d95, 0x7139a801, 0xde080cb3, 0x9cd8b4e4, 0x906456c1, 0x617bcb84, 0x70d532b6, 0x74486c5c, 0x42d0b857];
    var T7 = [0xa75051f4, 0x65537e41, 0xa4c31a17, 0x5e963a27, 0x6bcb3bab, 0x45f11f9d, 0x58abacfa, 0x03934be3, 0xfa552030, 0x6df6ad76, 0x769188cc, 0x4c25f502, 0xd7fc4fe5, 0xcbd7c52a, 0x44802635, 0xa38fb562, 0x5a49deb1, 0x1b6725ba, 0x0e9845ea, 0xc0e15dfe, 0x7502c32f, 0xf012814c, 0x97a38d46, 0xf9c66bd3, 0x5fe7038f, 0x9c951592, 0x7aebbf6d, 0x59da9552, 0x832dd4be, 0x21d35874, 0x692949e0, 0xc8448ec9, 0x896a75c2, 0x7978f48e, 0x3e6b9958, 0x71dd27b9, 0x4fb6bee1, 0xad17f088, 0xac66c920, 0x3ab47dce, 0x4a1863df, 0x3182e51a, 0x33609751, 0x7f456253, 0x77e0b164, 0xae84bb6b, 0xa01cfe81, 0x2b94f908, 0x68587048, 0xfd198f45, 0x6c8794de, 0xf8b7527b, 0xd323ab73, 0x02e2724b, 0x8f57e31f, 0xab2a6655, 0x2807b2eb, 0xc2032fb5, 0x7b9a86c5, 0x08a5d337, 0x87f23028, 0xa5b223bf, 0x6aba0203, 0x825ced16, 0x1c2b8acf, 0xb492a779, 0xf2f0f307, 0xe2a14e69, 0xf4cd65da, 0xbed50605, 0x621fd134, 0xfe8ac4a6, 0x539d342e, 0x55a0a2f3, 0xe132058a, 0xeb75a4f6, 0xec390b83, 0xefaa4060, 0x9f065e71, 0x1051bd6e, 0x8af93e21, 0x063d96dd, 0x05aedd3e, 0xbd464de6, 0x8db59154, 0x5d0571c4, 0xd46f0406, 0x15ff6050, 0xfb241998, 0xe997d6bd, 0x43cc8940, 0x9e7767d9, 0x42bdb0e8, 0x8b880789, 0x5b38e719, 0xeedb79c8, 0x0a47a17c, 0x0fe97c42, 0x1ec9f884, 0x00000000, 0x86830980, 0xed48322b, 0x70ac1e11, 0x724e6c5a, 0xfffbfd0e, 0x38560f85, 0xd51e3dae, 0x3927362d, 0xd9640a0f, 0xa621685c, 0x54d19b5b, 0x2e3a2436, 0x67b10c0a, 0xe70f9357, 0x96d2b4ee, 0x919e1b9b, 0xc54f80c0, 0x20a261dc, 0x4b695a77, 0x1a161c12, 0xba0ae293, 0x2ae5c0a0, 0xe0433c22, 0x171d121b, 0x0d0b0e09, 0xc7adf28b, 0xa8b92db6, 0xa9c8141e, 0x198557f1, 0x074caf75, 0xddbbee99, 0x60fda37f, 0x269ff701, 0xf5bc5c72, 0x3bc54466, 0x7e345bfb, 0x29768b43, 0xc6dccb23, 0xfc68b6ed, 0xf163b8e4, 0xdccad731, 0x85104263, 0x22401397, 0x112084c6, 0x247d854a, 0x3df8d2bb, 0x3211aef9, 0xa16dc729, 0x2f4b1d9e, 0x30f3dcb2, 0x52ec0d86, 0xe3d077c1, 0x166c2bb3, 0xb999a970, 0x48fa1194, 0x642247e9, 0x8cc4a8fc, 0x3f1aa0f0, 0x2cd8567d, 0x90ef2233, 0x4ec78749, 0xd1c1d938, 0xa2fe8cca, 0x0b3698d4, 0x81cfa6f5, 0xde28a57a, 0x8e26dab7, 0xbfa43fad, 0x9de42c3a, 0x920d5078, 0xcc9b6a5f, 0x4662547e, 0x13c2f68d, 0xb8e890d8, 0xf75e2e39, 0xaff582c3, 0x80be9f5d, 0x937c69d0, 0x2da96fd5, 0x12b3cf25, 0x993bc8ac, 0x7da71018, 0x636ee89c, 0xbb7bdb3b, 0x7809cd26, 0x18f46e59, 0xb701ec9a, 0x9aa8834f, 0x6e65e695, 0xe67eaaff, 0xcf0821bc, 0xe8e6ef15, 0x9bd9bae7, 0x36ce4a6f, 0x09d4ea9f, 0x7cd629b0, 0xb2af31a4, 0x23312a3f, 0x9430c6a5, 0x66c035a2, 0xbc37744e, 0xcaa6fc82, 0xd0b0e090, 0xd81533a7, 0x984af104, 0xdaf741ec, 0x500e7fcd, 0xf62f1791, 0xd68d764d, 0xb04d43ef, 0x4d54ccaa, 0x04dfe496, 0xb5e39ed1, 0x881b4c6a, 0x1fb8c12c, 0x517f4665, 0xea049d5e, 0x355d018c, 0x7473fa87, 0x412efb0b, 0x1d5ab367, 0xd25292db, 0x5633e910, 0x47136dd6, 0x618c9ad7, 0x0c7a37a1, 0x148e59f8, 0x3c89eb13, 0x27eecea9, 0xc935b761, 0xe5ede11c, 0xb13c7a47, 0xdf599cd2, 0x733f55f2, 0xce791814, 0x37bf73c7, 0xcdea53f7, 0xaa5b5ffd, 0x6f14df3d, 0xdb867844, 0xf381caaf, 0xc43eb968, 0x342c3824, 0x405fc2a3, 0xc372161d, 0x250cbce2, 0x498b283c, 0x9541ff0d, 0x017139a8, 0xb3de080c, 0xe49cd8b4, 0xc1906456, 0x84617bcb, 0xb670d532, 0x5c74486c, 0x5742d0b8];
    var T8 = [0xf4a75051, 0x4165537e, 0x17a4c31a, 0x275e963a, 0xab6bcb3b, 0x9d45f11f, 0xfa58abac, 0xe303934b, 0x30fa5520, 0x766df6ad, 0xcc769188, 0x024c25f5, 0xe5d7fc4f, 0x2acbd7c5, 0x35448026, 0x62a38fb5, 0xb15a49de, 0xba1b6725, 0xea0e9845, 0xfec0e15d, 0x2f7502c3, 0x4cf01281, 0x4697a38d, 0xd3f9c66b, 0x8f5fe703, 0x929c9515, 0x6d7aebbf, 0x5259da95, 0xbe832dd4, 0x7421d358, 0xe0692949, 0xc9c8448e, 0xc2896a75, 0x8e7978f4, 0x583e6b99, 0xb971dd27, 0xe14fb6be, 0x88ad17f0, 0x20ac66c9, 0xce3ab47d, 0xdf4a1863, 0x1a3182e5, 0x51336097, 0x537f4562, 0x6477e0b1, 0x6bae84bb, 0x81a01cfe, 0x082b94f9, 0x48685870, 0x45fd198f, 0xde6c8794, 0x7bf8b752, 0x73d323ab, 0x4b02e272, 0x1f8f57e3, 0x55ab2a66, 0xeb2807b2, 0xb5c2032f, 0xc57b9a86, 0x3708a5d3, 0x2887f230, 0xbfa5b223, 0x036aba02, 0x16825ced, 0xcf1c2b8a, 0x79b492a7, 0x07f2f0f3, 0x69e2a14e, 0xdaf4cd65, 0x05bed506, 0x34621fd1, 0xa6fe8ac4, 0x2e539d34, 0xf355a0a2, 0x8ae13205, 0xf6eb75a4, 0x83ec390b, 0x60efaa40, 0x719f065e, 0x6e1051bd, 0x218af93e, 0xdd063d96, 0x3e05aedd, 0xe6bd464d, 0x548db591, 0xc45d0571, 0x06d46f04, 0x5015ff60, 0x98fb2419, 0xbde997d6, 0x4043cc89, 0xd99e7767, 0xe842bdb0, 0x898b8807, 0x195b38e7, 0xc8eedb79, 0x7c0a47a1, 0x420fe97c, 0x841ec9f8, 0x00000000, 0x80868309, 0x2bed4832, 0x1170ac1e, 0x5a724e6c, 0x0efffbfd, 0x8538560f, 0xaed51e3d, 0x2d392736, 0x0fd9640a, 0x5ca62168, 0x5b54d19b, 0x362e3a24, 0x0a67b10c, 0x57e70f93, 0xee96d2b4, 0x9b919e1b, 0xc0c54f80, 0xdc20a261, 0x774b695a, 0x121a161c, 0x93ba0ae2, 0xa02ae5c0, 0x22e0433c, 0x1b171d12, 0x090d0b0e, 0x8bc7adf2, 0xb6a8b92d, 0x1ea9c814, 0xf1198557, 0x75074caf, 0x99ddbbee, 0x7f60fda3, 0x01269ff7, 0x72f5bc5c, 0x663bc544, 0xfb7e345b, 0x4329768b, 0x23c6dccb, 0xedfc68b6, 0xe4f163b8, 0x31dccad7, 0x63851042, 0x97224013, 0xc6112084, 0x4a247d85, 0xbb3df8d2, 0xf93211ae, 0x29a16dc7, 0x9e2f4b1d, 0xb230f3dc, 0x8652ec0d, 0xc1e3d077, 0xb3166c2b, 0x70b999a9, 0x9448fa11, 0xe9642247, 0xfc8cc4a8, 0xf03f1aa0, 0x7d2cd856, 0x3390ef22, 0x494ec787, 0x38d1c1d9, 0xcaa2fe8c, 0xd40b3698, 0xf581cfa6, 0x7ade28a5, 0xb78e26da, 0xadbfa43f, 0x3a9de42c, 0x78920d50, 0x5fcc9b6a, 0x7e466254, 0x8d13c2f6, 0xd8b8e890, 0x39f75e2e, 0xc3aff582, 0x5d80be9f, 0xd0937c69, 0xd52da96f, 0x2512b3cf, 0xac993bc8, 0x187da710, 0x9c636ee8, 0x3bbb7bdb, 0x267809cd, 0x5918f46e, 0x9ab701ec, 0x4f9aa883, 0x956e65e6, 0xffe67eaa, 0xbccf0821, 0x15e8e6ef, 0xe79bd9ba, 0x6f36ce4a, 0x9f09d4ea, 0xb07cd629, 0xa4b2af31, 0x3f23312a, 0xa59430c6, 0xa266c035, 0x4ebc3774, 0x82caa6fc, 0x90d0b0e0, 0xa7d81533, 0x04984af1, 0xecdaf741, 0xcd500e7f, 0x91f62f17, 0x4dd68d76, 0xefb04d43, 0xaa4d54cc, 0x9604dfe4, 0xd1b5e39e, 0x6a881b4c, 0x2c1fb8c1, 0x65517f46, 0x5eea049d, 0x8c355d01, 0x877473fa, 0x0b412efb, 0x671d5ab3, 0xdbd25292, 0x105633e9, 0xd647136d, 0xd7618c9a, 0xa10c7a37, 0xf8148e59, 0x133c89eb, 0xa927eece, 0x61c935b7, 0x1ce5ede1, 0x47b13c7a, 0xd2df599c, 0xf2733f55, 0x14ce7918, 0xc737bf73, 0xf7cdea53, 0xfdaa5b5f, 0x3d6f14df, 0x44db8678, 0xaff381ca, 0x68c43eb9, 0x24342c38, 0xa3405fc2, 0x1dc37216, 0xe2250cbc, 0x3c498b28, 0x0d9541ff, 0xa8017139, 0x0cb3de08, 0xb4e49cd8, 0x56c19064, 0xcb84617b, 0x32b670d5, 0x6c5c7448, 0xb85742d0];

    // Transformations for decryption key expansion
    var U1 = [0x00000000, 0x0e090d0b, 0x1c121a16, 0x121b171d, 0x3824342c, 0x362d3927, 0x24362e3a, 0x2a3f2331, 0x70486858, 0x7e416553, 0x6c5a724e, 0x62537f45, 0x486c5c74, 0x4665517f, 0x547e4662, 0x5a774b69, 0xe090d0b0, 0xee99ddbb, 0xfc82caa6, 0xf28bc7ad, 0xd8b4e49c, 0xd6bde997, 0xc4a6fe8a, 0xcaaff381, 0x90d8b8e8, 0x9ed1b5e3, 0x8ccaa2fe, 0x82c3aff5, 0xa8fc8cc4, 0xa6f581cf, 0xb4ee96d2, 0xbae79bd9, 0xdb3bbb7b, 0xd532b670, 0xc729a16d, 0xc920ac66, 0xe31f8f57, 0xed16825c, 0xff0d9541, 0xf104984a, 0xab73d323, 0xa57ade28, 0xb761c935, 0xb968c43e, 0x9357e70f, 0x9d5eea04, 0x8f45fd19, 0x814cf012, 0x3bab6bcb, 0x35a266c0, 0x27b971dd, 0x29b07cd6, 0x038f5fe7, 0x0d8652ec, 0x1f9d45f1, 0x119448fa, 0x4be30393, 0x45ea0e98, 0x57f11985, 0x59f8148e, 0x73c737bf, 0x7dce3ab4, 0x6fd52da9, 0x61dc20a2, 0xad766df6, 0xa37f60fd, 0xb16477e0, 0xbf6d7aeb, 0x955259da, 0x9b5b54d1, 0x894043cc, 0x87494ec7, 0xdd3e05ae, 0xd33708a5, 0xc12c1fb8, 0xcf2512b3, 0xe51a3182, 0xeb133c89, 0xf9082b94, 0xf701269f, 0x4de6bd46, 0x43efb04d, 0x51f4a750, 0x5ffdaa5b, 0x75c2896a, 0x7bcb8461, 0x69d0937c, 0x67d99e77, 0x3daed51e, 0x33a7d815, 0x21bccf08, 0x2fb5c203, 0x058ae132, 0x0b83ec39, 0x1998fb24, 0x1791f62f, 0x764dd68d, 0x7844db86, 0x6a5fcc9b, 0x6456c190, 0x4e69e2a1, 0x4060efaa, 0x527bf8b7, 0x5c72f5bc, 0x0605bed5, 0x080cb3de, 0x1a17a4c3, 0x141ea9c8, 0x3e218af9, 0x302887f2, 0x223390ef, 0x2c3a9de4, 0x96dd063d, 0x98d40b36, 0x8acf1c2b, 0x84c61120, 0xaef93211, 0xa0f03f1a, 0xb2eb2807, 0xbce2250c, 0xe6956e65, 0xe89c636e, 0xfa877473, 0xf48e7978, 0xdeb15a49, 0xd0b85742, 0xc2a3405f, 0xccaa4d54, 0x41ecdaf7, 0x4fe5d7fc, 0x5dfec0e1, 0x53f7cdea, 0x79c8eedb, 0x77c1e3d0, 0x65daf4cd, 0x6bd3f9c6, 0x31a4b2af, 0x3fadbfa4, 0x2db6a8b9, 0x23bfa5b2, 0x09808683, 0x07898b88, 0x15929c95, 0x1b9b919e, 0xa17c0a47, 0xaf75074c, 0xbd6e1051, 0xb3671d5a, 0x99583e6b, 0x97513360, 0x854a247d, 0x8b432976, 0xd134621f, 0xdf3d6f14, 0xcd267809, 0xc32f7502, 0xe9105633, 0xe7195b38, 0xf5024c25, 0xfb0b412e, 0x9ad7618c, 0x94de6c87, 0x86c57b9a, 0x88cc7691, 0xa2f355a0, 0xacfa58ab, 0xbee14fb6, 0xb0e842bd, 0xea9f09d4, 0xe49604df, 0xf68d13c2, 0xf8841ec9, 0xd2bb3df8, 0xdcb230f3, 0xcea927ee, 0xc0a02ae5, 0x7a47b13c, 0x744ebc37, 0x6655ab2a, 0x685ca621, 0x42638510, 0x4c6a881b, 0x5e719f06, 0x5078920d, 0x0a0fd964, 0x0406d46f, 0x161dc372, 0x1814ce79, 0x322bed48, 0x3c22e043, 0x2e39f75e, 0x2030fa55, 0xec9ab701, 0xe293ba0a, 0xf088ad17, 0xfe81a01c, 0xd4be832d, 0xdab78e26, 0xc8ac993b, 0xc6a59430, 0x9cd2df59, 0x92dbd252, 0x80c0c54f, 0x8ec9c844, 0xa4f6eb75, 0xaaffe67e, 0xb8e4f163, 0xb6edfc68, 0x0c0a67b1, 0x02036aba, 0x10187da7, 0x1e1170ac, 0x342e539d, 0x3a275e96, 0x283c498b, 0x26354480, 0x7c420fe9, 0x724b02e2, 0x605015ff, 0x6e5918f4, 0x44663bc5, 0x4a6f36ce, 0x587421d3, 0x567d2cd8, 0x37a10c7a, 0x39a80171, 0x2bb3166c, 0x25ba1b67, 0x0f853856, 0x018c355d, 0x13972240, 0x1d9e2f4b, 0x47e96422, 0x49e06929, 0x5bfb7e34, 0x55f2733f, 0x7fcd500e, 0x71c45d05, 0x63df4a18, 0x6dd64713, 0xd731dcca, 0xd938d1c1, 0xcb23c6dc, 0xc52acbd7, 0xef15e8e6, 0xe11ce5ed, 0xf307f2f0, 0xfd0efffb, 0xa779b492, 0xa970b999, 0xbb6bae84, 0xb562a38f, 0x9f5d80be, 0x91548db5, 0x834f9aa8, 0x8d4697a3];
    var U2 = [0x00000000, 0x0b0e090d, 0x161c121a, 0x1d121b17, 0x2c382434, 0x27362d39, 0x3a24362e, 0x312a3f23, 0x58704868, 0x537e4165, 0x4e6c5a72, 0x4562537f, 0x74486c5c, 0x7f466551, 0x62547e46, 0x695a774b, 0xb0e090d0, 0xbbee99dd, 0xa6fc82ca, 0xadf28bc7, 0x9cd8b4e4, 0x97d6bde9, 0x8ac4a6fe, 0x81caaff3, 0xe890d8b8, 0xe39ed1b5, 0xfe8ccaa2, 0xf582c3af, 0xc4a8fc8c, 0xcfa6f581, 0xd2b4ee96, 0xd9bae79b, 0x7bdb3bbb, 0x70d532b6, 0x6dc729a1, 0x66c920ac, 0x57e31f8f, 0x5ced1682, 0x41ff0d95, 0x4af10498, 0x23ab73d3, 0x28a57ade, 0x35b761c9, 0x3eb968c4, 0x0f9357e7, 0x049d5eea, 0x198f45fd, 0x12814cf0, 0xcb3bab6b, 0xc035a266, 0xdd27b971, 0xd629b07c, 0xe7038f5f, 0xec0d8652, 0xf11f9d45, 0xfa119448, 0x934be303, 0x9845ea0e, 0x8557f119, 0x8e59f814, 0xbf73c737, 0xb47dce3a, 0xa96fd52d, 0xa261dc20, 0xf6ad766d, 0xfda37f60, 0xe0b16477, 0xebbf6d7a, 0xda955259, 0xd19b5b54, 0xcc894043, 0xc787494e, 0xaedd3e05, 0xa5d33708, 0xb8c12c1f, 0xb3cf2512, 0x82e51a31, 0x89eb133c, 0x94f9082b, 0x9ff70126, 0x464de6bd, 0x4d43efb0, 0x5051f4a7, 0x5b5ffdaa, 0x6a75c289, 0x617bcb84, 0x7c69d093, 0x7767d99e, 0x1e3daed5, 0x1533a7d8, 0x0821bccf, 0x032fb5c2, 0x32058ae1, 0x390b83ec, 0x241998fb, 0x2f1791f6, 0x8d764dd6, 0x867844db, 0x9b6a5fcc, 0x906456c1, 0xa14e69e2, 0xaa4060ef, 0xb7527bf8, 0xbc5c72f5, 0xd50605be, 0xde080cb3, 0xc31a17a4, 0xc8141ea9, 0xf93e218a, 0xf2302887, 0xef223390, 0xe42c3a9d, 0x3d96dd06, 0x3698d40b, 0x2b8acf1c, 0x2084c611, 0x11aef932, 0x1aa0f03f, 0x07b2eb28, 0x0cbce225, 0x65e6956e, 0x6ee89c63, 0x73fa8774, 0x78f48e79, 0x49deb15a, 0x42d0b857, 0x5fc2a340, 0x54ccaa4d, 0xf741ecda, 0xfc4fe5d7, 0xe15dfec0, 0xea53f7cd, 0xdb79c8ee, 0xd077c1e3, 0xcd65daf4, 0xc66bd3f9, 0xaf31a4b2, 0xa43fadbf, 0xb92db6a8, 0xb223bfa5, 0x83098086, 0x8807898b, 0x9515929c, 0x9e1b9b91, 0x47a17c0a, 0x4caf7507, 0x51bd6e10, 0x5ab3671d, 0x6b99583e, 0x60975133, 0x7d854a24, 0x768b4329, 0x1fd13462, 0x14df3d6f, 0x09cd2678, 0x02c32f75, 0x33e91056, 0x38e7195b, 0x25f5024c, 0x2efb0b41, 0x8c9ad761, 0x8794de6c, 0x9a86c57b, 0x9188cc76, 0xa0a2f355, 0xabacfa58, 0xb6bee14f, 0xbdb0e842, 0xd4ea9f09, 0xdfe49604, 0xc2f68d13, 0xc9f8841e, 0xf8d2bb3d, 0xf3dcb230, 0xeecea927, 0xe5c0a02a, 0x3c7a47b1, 0x37744ebc, 0x2a6655ab, 0x21685ca6, 0x10426385, 0x1b4c6a88, 0x065e719f, 0x0d507892, 0x640a0fd9, 0x6f0406d4, 0x72161dc3, 0x791814ce, 0x48322bed, 0x433c22e0, 0x5e2e39f7, 0x552030fa, 0x01ec9ab7, 0x0ae293ba, 0x17f088ad, 0x1cfe81a0, 0x2dd4be83, 0x26dab78e, 0x3bc8ac99, 0x30c6a594, 0x599cd2df, 0x5292dbd2, 0x4f80c0c5, 0x448ec9c8, 0x75a4f6eb, 0x7eaaffe6, 0x63b8e4f1, 0x68b6edfc, 0xb10c0a67, 0xba02036a, 0xa710187d, 0xac1e1170, 0x9d342e53, 0x963a275e, 0x8b283c49, 0x80263544, 0xe97c420f, 0xe2724b02, 0xff605015, 0xf46e5918, 0xc544663b, 0xce4a6f36, 0xd3587421, 0xd8567d2c, 0x7a37a10c, 0x7139a801, 0x6c2bb316, 0x6725ba1b, 0x560f8538, 0x5d018c35, 0x40139722, 0x4b1d9e2f, 0x2247e964, 0x2949e069, 0x345bfb7e, 0x3f55f273, 0x0e7fcd50, 0x0571c45d, 0x1863df4a, 0x136dd647, 0xcad731dc, 0xc1d938d1, 0xdccb23c6, 0xd7c52acb, 0xe6ef15e8, 0xede11ce5, 0xf0f307f2, 0xfbfd0eff, 0x92a779b4, 0x99a970b9, 0x84bb6bae, 0x8fb562a3, 0xbe9f5d80, 0xb591548d, 0xa8834f9a, 0xa38d4697];
    var U3 = [0x00000000, 0x0d0b0e09, 0x1a161c12, 0x171d121b, 0x342c3824, 0x3927362d, 0x2e3a2436, 0x23312a3f, 0x68587048, 0x65537e41, 0x724e6c5a, 0x7f456253, 0x5c74486c, 0x517f4665, 0x4662547e, 0x4b695a77, 0xd0b0e090, 0xddbbee99, 0xcaa6fc82, 0xc7adf28b, 0xe49cd8b4, 0xe997d6bd, 0xfe8ac4a6, 0xf381caaf, 0xb8e890d8, 0xb5e39ed1, 0xa2fe8cca, 0xaff582c3, 0x8cc4a8fc, 0x81cfa6f5, 0x96d2b4ee, 0x9bd9bae7, 0xbb7bdb3b, 0xb670d532, 0xa16dc729, 0xac66c920, 0x8f57e31f, 0x825ced16, 0x9541ff0d, 0x984af104, 0xd323ab73, 0xde28a57a, 0xc935b761, 0xc43eb968, 0xe70f9357, 0xea049d5e, 0xfd198f45, 0xf012814c, 0x6bcb3bab, 0x66c035a2, 0x71dd27b9, 0x7cd629b0, 0x5fe7038f, 0x52ec0d86, 0x45f11f9d, 0x48fa1194, 0x03934be3, 0x0e9845ea, 0x198557f1, 0x148e59f8, 0x37bf73c7, 0x3ab47dce, 0x2da96fd5, 0x20a261dc, 0x6df6ad76, 0x60fda37f, 0x77e0b164, 0x7aebbf6d, 0x59da9552, 0x54d19b5b, 0x43cc8940, 0x4ec78749, 0x05aedd3e, 0x08a5d337, 0x1fb8c12c, 0x12b3cf25, 0x3182e51a, 0x3c89eb13, 0x2b94f908, 0x269ff701, 0xbd464de6, 0xb04d43ef, 0xa75051f4, 0xaa5b5ffd, 0x896a75c2, 0x84617bcb, 0x937c69d0, 0x9e7767d9, 0xd51e3dae, 0xd81533a7, 0xcf0821bc, 0xc2032fb5, 0xe132058a, 0xec390b83, 0xfb241998, 0xf62f1791, 0xd68d764d, 0xdb867844, 0xcc9b6a5f, 0xc1906456, 0xe2a14e69, 0xefaa4060, 0xf8b7527b, 0xf5bc5c72, 0xbed50605, 0xb3de080c, 0xa4c31a17, 0xa9c8141e, 0x8af93e21, 0x87f23028, 0x90ef2233, 0x9de42c3a, 0x063d96dd, 0x0b3698d4, 0x1c2b8acf, 0x112084c6, 0x3211aef9, 0x3f1aa0f0, 0x2807b2eb, 0x250cbce2, 0x6e65e695, 0x636ee89c, 0x7473fa87, 0x7978f48e, 0x5a49deb1, 0x5742d0b8, 0x405fc2a3, 0x4d54ccaa, 0xdaf741ec, 0xd7fc4fe5, 0xc0e15dfe, 0xcdea53f7, 0xeedb79c8, 0xe3d077c1, 0xf4cd65da, 0xf9c66bd3, 0xb2af31a4, 0xbfa43fad, 0xa8b92db6, 0xa5b223bf, 0x86830980, 0x8b880789, 0x9c951592, 0x919e1b9b, 0x0a47a17c, 0x074caf75, 0x1051bd6e, 0x1d5ab367, 0x3e6b9958, 0x33609751, 0x247d854a, 0x29768b43, 0x621fd134, 0x6f14df3d, 0x7809cd26, 0x7502c32f, 0x5633e910, 0x5b38e719, 0x4c25f502, 0x412efb0b, 0x618c9ad7, 0x6c8794de, 0x7b9a86c5, 0x769188cc, 0x55a0a2f3, 0x58abacfa, 0x4fb6bee1, 0x42bdb0e8, 0x09d4ea9f, 0x04dfe496, 0x13c2f68d, 0x1ec9f884, 0x3df8d2bb, 0x30f3dcb2, 0x27eecea9, 0x2ae5c0a0, 0xb13c7a47, 0xbc37744e, 0xab2a6655, 0xa621685c, 0x85104263, 0x881b4c6a, 0x9f065e71, 0x920d5078, 0xd9640a0f, 0xd46f0406, 0xc372161d, 0xce791814, 0xed48322b, 0xe0433c22, 0xf75e2e39, 0xfa552030, 0xb701ec9a, 0xba0ae293, 0xad17f088, 0xa01cfe81, 0x832dd4be, 0x8e26dab7, 0x993bc8ac, 0x9430c6a5, 0xdf599cd2, 0xd25292db, 0xc54f80c0, 0xc8448ec9, 0xeb75a4f6, 0xe67eaaff, 0xf163b8e4, 0xfc68b6ed, 0x67b10c0a, 0x6aba0203, 0x7da71018, 0x70ac1e11, 0x539d342e, 0x5e963a27, 0x498b283c, 0x44802635, 0x0fe97c42, 0x02e2724b, 0x15ff6050, 0x18f46e59, 0x3bc54466, 0x36ce4a6f, 0x21d35874, 0x2cd8567d, 0x0c7a37a1, 0x017139a8, 0x166c2bb3, 0x1b6725ba, 0x38560f85, 0x355d018c, 0x22401397, 0x2f4b1d9e, 0x642247e9, 0x692949e0, 0x7e345bfb, 0x733f55f2, 0x500e7fcd, 0x5d0571c4, 0x4a1863df, 0x47136dd6, 0xdccad731, 0xd1c1d938, 0xc6dccb23, 0xcbd7c52a, 0xe8e6ef15, 0xe5ede11c, 0xf2f0f307, 0xfffbfd0e, 0xb492a779, 0xb999a970, 0xae84bb6b, 0xa38fb562, 0x80be9f5d, 0x8db59154, 0x9aa8834f, 0x97a38d46];
    var U4 = [0x00000000, 0x090d0b0e, 0x121a161c, 0x1b171d12, 0x24342c38, 0x2d392736, 0x362e3a24, 0x3f23312a, 0x48685870, 0x4165537e, 0x5a724e6c, 0x537f4562, 0x6c5c7448, 0x65517f46, 0x7e466254, 0x774b695a, 0x90d0b0e0, 0x99ddbbee, 0x82caa6fc, 0x8bc7adf2, 0xb4e49cd8, 0xbde997d6, 0xa6fe8ac4, 0xaff381ca, 0xd8b8e890, 0xd1b5e39e, 0xcaa2fe8c, 0xc3aff582, 0xfc8cc4a8, 0xf581cfa6, 0xee96d2b4, 0xe79bd9ba, 0x3bbb7bdb, 0x32b670d5, 0x29a16dc7, 0x20ac66c9, 0x1f8f57e3, 0x16825ced, 0x0d9541ff, 0x04984af1, 0x73d323ab, 0x7ade28a5, 0x61c935b7, 0x68c43eb9, 0x57e70f93, 0x5eea049d, 0x45fd198f, 0x4cf01281, 0xab6bcb3b, 0xa266c035, 0xb971dd27, 0xb07cd629, 0x8f5fe703, 0x8652ec0d, 0x9d45f11f, 0x9448fa11, 0xe303934b, 0xea0e9845, 0xf1198557, 0xf8148e59, 0xc737bf73, 0xce3ab47d, 0xd52da96f, 0xdc20a261, 0x766df6ad, 0x7f60fda3, 0x6477e0b1, 0x6d7aebbf, 0x5259da95, 0x5b54d19b, 0x4043cc89, 0x494ec787, 0x3e05aedd, 0x3708a5d3, 0x2c1fb8c1, 0x2512b3cf, 0x1a3182e5, 0x133c89eb, 0x082b94f9, 0x01269ff7, 0xe6bd464d, 0xefb04d43, 0xf4a75051, 0xfdaa5b5f, 0xc2896a75, 0xcb84617b, 0xd0937c69, 0xd99e7767, 0xaed51e3d, 0xa7d81533, 0xbccf0821, 0xb5c2032f, 0x8ae13205, 0x83ec390b, 0x98fb2419, 0x91f62f17, 0x4dd68d76, 0x44db8678, 0x5fcc9b6a, 0x56c19064, 0x69e2a14e, 0x60efaa40, 0x7bf8b752, 0x72f5bc5c, 0x05bed506, 0x0cb3de08, 0x17a4c31a, 0x1ea9c814, 0x218af93e, 0x2887f230, 0x3390ef22, 0x3a9de42c, 0xdd063d96, 0xd40b3698, 0xcf1c2b8a, 0xc6112084, 0xf93211ae, 0xf03f1aa0, 0xeb2807b2, 0xe2250cbc, 0x956e65e6, 0x9c636ee8, 0x877473fa, 0x8e7978f4, 0xb15a49de, 0xb85742d0, 0xa3405fc2, 0xaa4d54cc, 0xecdaf741, 0xe5d7fc4f, 0xfec0e15d, 0xf7cdea53, 0xc8eedb79, 0xc1e3d077, 0xdaf4cd65, 0xd3f9c66b, 0xa4b2af31, 0xadbfa43f, 0xb6a8b92d, 0xbfa5b223, 0x80868309, 0x898b8807, 0x929c9515, 0x9b919e1b, 0x7c0a47a1, 0x75074caf, 0x6e1051bd, 0x671d5ab3, 0x583e6b99, 0x51336097, 0x4a247d85, 0x4329768b, 0x34621fd1, 0x3d6f14df, 0x267809cd, 0x2f7502c3, 0x105633e9, 0x195b38e7, 0x024c25f5, 0x0b412efb, 0xd7618c9a, 0xde6c8794, 0xc57b9a86, 0xcc769188, 0xf355a0a2, 0xfa58abac, 0xe14fb6be, 0xe842bdb0, 0x9f09d4ea, 0x9604dfe4, 0x8d13c2f6, 0x841ec9f8, 0xbb3df8d2, 0xb230f3dc, 0xa927eece, 0xa02ae5c0, 0x47b13c7a, 0x4ebc3774, 0x55ab2a66, 0x5ca62168, 0x63851042, 0x6a881b4c, 0x719f065e, 0x78920d50, 0x0fd9640a, 0x06d46f04, 0x1dc37216, 0x14ce7918, 0x2bed4832, 0x22e0433c, 0x39f75e2e, 0x30fa5520, 0x9ab701ec, 0x93ba0ae2, 0x88ad17f0, 0x81a01cfe, 0xbe832dd4, 0xb78e26da, 0xac993bc8, 0xa59430c6, 0xd2df599c, 0xdbd25292, 0xc0c54f80, 0xc9c8448e, 0xf6eb75a4, 0xffe67eaa, 0xe4f163b8, 0xedfc68b6, 0x0a67b10c, 0x036aba02, 0x187da710, 0x1170ac1e, 0x2e539d34, 0x275e963a, 0x3c498b28, 0x35448026, 0x420fe97c, 0x4b02e272, 0x5015ff60, 0x5918f46e, 0x663bc544, 0x6f36ce4a, 0x7421d358, 0x7d2cd856, 0xa10c7a37, 0xa8017139, 0xb3166c2b, 0xba1b6725, 0x8538560f, 0x8c355d01, 0x97224013, 0x9e2f4b1d, 0xe9642247, 0xe0692949, 0xfb7e345b, 0xf2733f55, 0xcd500e7f, 0xc45d0571, 0xdf4a1863, 0xd647136d, 0x31dccad7, 0x38d1c1d9, 0x23c6dccb, 0x2acbd7c5, 0x15e8e6ef, 0x1ce5ede1, 0x07f2f0f3, 0x0efffbfd, 0x79b492a7, 0x70b999a9, 0x6bae84bb, 0x62a38fb5, 0x5d80be9f, 0x548db591, 0x4f9aa883, 0x4697a38d];

    function convertToInt32(bytes) {
        var result = [];
        for (var i = 0; i < bytes.length; i += 4) {
            result.push(
                (bytes[i    ] << 24) |
                (bytes[i + 1] << 16) |
                (bytes[i + 2] <<  8) |
                 bytes[i + 3]
            );
        }
        return result;
    }

    var AES = function(key) {
        if (!(this instanceof AES)) {
            throw Error('AES must be instanitated with `new`');
        }

        Object.defineProperty(this, 'key', {
            value: coerceArray(key, true)
        });

        this._prepare();
    }


    AES.prototype._prepare = function() {

        var rounds = numberOfRounds[this.key.length];
        if (rounds == null) {
            throw new Error('invalid key size (must be 16, 24 or 32 bytes)');
        }

        // encryption round keys
        this._Ke = [];

        // decryption round keys
        this._Kd = [];

        for (var i = 0; i <= rounds; i++) {
            this._Ke.push([0, 0, 0, 0]);
            this._Kd.push([0, 0, 0, 0]);
        }

        var roundKeyCount = (rounds + 1) * 4;
        var KC = this.key.length / 4;

        // convert the key into ints
        var tk = convertToInt32(this.key);

        // copy values into round key arrays
        var index;
        for (var i = 0; i < KC; i++) {
            index = i >> 2;
            this._Ke[index][i % 4] = tk[i];
            this._Kd[rounds - index][i % 4] = tk[i];
        }

        // key expansion (fips-197 section 5.2)
        var rconpointer = 0;
        var t = KC, tt;
        while (t < roundKeyCount) {
            tt = tk[KC - 1];
            tk[0] ^= ((S[(tt >> 16) & 0xFF] << 24) ^
                      (S[(tt >>  8) & 0xFF] << 16) ^
                      (S[ tt        & 0xFF] <<  8) ^
                       S[(tt >> 24) & 0xFF]        ^
                      (rcon[rconpointer] << 24));
            rconpointer += 1;

            // key expansion (for non-256 bit)
            if (KC != 8) {
                for (var i = 1; i < KC; i++) {
                    tk[i] ^= tk[i - 1];
                }

            // key expansion for 256-bit keys is "slightly different" (fips-197)
            } else {
                for (var i = 1; i < (KC / 2); i++) {
                    tk[i] ^= tk[i - 1];
                }
                tt = tk[(KC / 2) - 1];

                tk[KC / 2] ^= (S[ tt        & 0xFF]        ^
                              (S[(tt >>  8) & 0xFF] <<  8) ^
                              (S[(tt >> 16) & 0xFF] << 16) ^
                              (S[(tt >> 24) & 0xFF] << 24));

                for (var i = (KC / 2) + 1; i < KC; i++) {
                    tk[i] ^= tk[i - 1];
                }
            }

            // copy values into round key arrays
            var i = 0, r, c;
            while (i < KC && t < roundKeyCount) {
                r = t >> 2;
                c = t % 4;
                this._Ke[r][c] = tk[i];
                this._Kd[rounds - r][c] = tk[i++];
                t++;
            }
        }

        // inverse-cipher-ify the decryption round key (fips-197 section 5.3)
        for (var r = 1; r < rounds; r++) {
            for (var c = 0; c < 4; c++) {
                tt = this._Kd[r][c];
                this._Kd[r][c] = (U1[(tt >> 24) & 0xFF] ^
                                  U2[(tt >> 16) & 0xFF] ^
                                  U3[(tt >>  8) & 0xFF] ^
                                  U4[ tt        & 0xFF]);
            }
        }
    }

    AES.prototype.encrypt = function(plaintext) {
        if (plaintext.length != 16) {
            throw new Error('invalid plaintext size (must be 16 bytes)');
        }

        var rounds = this._Ke.length - 1;
        var a = [0, 0, 0, 0];

        // convert plaintext to (ints ^ key)
        var t = convertToInt32(plaintext);
        for (var i = 0; i < 4; i++) {
            t[i] ^= this._Ke[0][i];
        }

        // apply round transforms
        for (var r = 1; r < rounds; r++) {
            for (var i = 0; i < 4; i++) {
                a[i] = (T1[(t[ i         ] >> 24) & 0xff] ^
                        T2[(t[(i + 1) % 4] >> 16) & 0xff] ^
                        T3[(t[(i + 2) % 4] >>  8) & 0xff] ^
                        T4[ t[(i + 3) % 4]        & 0xff] ^
                        this._Ke[r][i]);
            }
            t = a.slice();
        }

        // the last round is special
        var result = createArray(16), tt;
        for (var i = 0; i < 4; i++) {
            tt = this._Ke[rounds][i];
            result[4 * i    ] = (S[(t[ i         ] >> 24) & 0xff] ^ (tt >> 24)) & 0xff;
            result[4 * i + 1] = (S[(t[(i + 1) % 4] >> 16) & 0xff] ^ (tt >> 16)) & 0xff;
            result[4 * i + 2] = (S[(t[(i + 2) % 4] >>  8) & 0xff] ^ (tt >>  8)) & 0xff;
            result[4 * i + 3] = (S[ t[(i + 3) % 4]        & 0xff] ^  tt       ) & 0xff;
        }

        return result;
    }

    AES.prototype.decrypt = function(ciphertext) {
        if (ciphertext.length != 16) {
            throw new Error('invalid ciphertext size (must be 16 bytes)');
        }

        var rounds = this._Kd.length - 1;
        var a = [0, 0, 0, 0];

        // convert plaintext to (ints ^ key)
        var t = convertToInt32(ciphertext);
        for (var i = 0; i < 4; i++) {
            t[i] ^= this._Kd[0][i];
        }

        // apply round transforms
        for (var r = 1; r < rounds; r++) {
            for (var i = 0; i < 4; i++) {
                a[i] = (T5[(t[ i          ] >> 24) & 0xff] ^
                        T6[(t[(i + 3) % 4] >> 16) & 0xff] ^
                        T7[(t[(i + 2) % 4] >>  8) & 0xff] ^
                        T8[ t[(i + 1) % 4]        & 0xff] ^
                        this._Kd[r][i]);
            }
            t = a.slice();
        }

        // the last round is special
        var result = createArray(16), tt;
        for (var i = 0; i < 4; i++) {
            tt = this._Kd[rounds][i];
            result[4 * i    ] = (Si[(t[ i         ] >> 24) & 0xff] ^ (tt >> 24)) & 0xff;
            result[4 * i + 1] = (Si[(t[(i + 3) % 4] >> 16) & 0xff] ^ (tt >> 16)) & 0xff;
            result[4 * i + 2] = (Si[(t[(i + 2) % 4] >>  8) & 0xff] ^ (tt >>  8)) & 0xff;
            result[4 * i + 3] = (Si[ t[(i + 1) % 4]        & 0xff] ^  tt       ) & 0xff;
        }

        return result;
    }


    /**
     *  Mode Of Operation - Electonic Codebook (ECB)
     */
    var ModeOfOperationECB = function(key) {
        if (!(this instanceof ModeOfOperationECB)) {
            throw Error('AES must be instanitated with `new`');
        }

        this.description = "Electronic Code Block";
        this.name = "ecb";

        this._aes = new AES(key);
    }

    ModeOfOperationECB.prototype.encrypt = function(plaintext) {
        plaintext = coerceArray(plaintext);

        if ((plaintext.length % 16) !== 0) {
            throw new Error('invalid plaintext size (must be multiple of 16 bytes)');
        }

        var ciphertext = createArray(plaintext.length);
        var block = createArray(16);

        for (var i = 0; i < plaintext.length; i += 16) {
            copyArray(plaintext, block, 0, i, i + 16);
            block = this._aes.encrypt(block);
            copyArray(block, ciphertext, i);
        }

        return ciphertext;
    }

    ModeOfOperationECB.prototype.decrypt = function(ciphertext) {
        ciphertext = coerceArray(ciphertext);

        if ((ciphertext.length % 16) !== 0) {
            throw new Error('invalid ciphertext size (must be multiple of 16 bytes)');
        }

        var plaintext = createArray(ciphertext.length);
        var block = createArray(16);

        for (var i = 0; i < ciphertext.length; i += 16) {
            copyArray(ciphertext, block, 0, i, i + 16);
            block = this._aes.decrypt(block);
            copyArray(block, plaintext, i);
        }

        return plaintext;
    }


    /**
     *  Mode Of Operation - Cipher Block Chaining (CBC)
     */
    var ModeOfOperationCBC = function(key, iv) {
        if (!(this instanceof ModeOfOperationCBC)) {
            throw Error('AES must be instanitated with `new`');
        }

        this.description = "Cipher Block Chaining";
        this.name = "cbc";

        if (!iv) {
            iv = createArray(16);

        } else if (iv.length != 16) {
            throw new Error('invalid initialation vector size (must be 16 bytes)');
        }

        this._lastCipherblock = coerceArray(iv, true);

        this._aes = new AES(key);
    }

    ModeOfOperationCBC.prototype.encrypt = function(plaintext) {
        plaintext = coerceArray(plaintext);

        if ((plaintext.length % 16) !== 0) {
            throw new Error('invalid plaintext size (must be multiple of 16 bytes)');
        }

        var ciphertext = createArray(plaintext.length);
        var block = createArray(16);

        for (var i = 0; i < plaintext.length; i += 16) {
            copyArray(plaintext, block, 0, i, i + 16);

            for (var j = 0; j < 16; j++) {
                block[j] ^= this._lastCipherblock[j];
            }

            this._lastCipherblock = this._aes.encrypt(block);
            copyArray(this._lastCipherblock, ciphertext, i);
        }

        return ciphertext;
    }

    ModeOfOperationCBC.prototype.decrypt = function(ciphertext) {
        ciphertext = coerceArray(ciphertext);

        if ((ciphertext.length % 16) !== 0) {
            throw new Error('invalid ciphertext size (must be multiple of 16 bytes)');
        }

        var plaintext = createArray(ciphertext.length);
        var block = createArray(16);

        for (var i = 0; i < ciphertext.length; i += 16) {
            copyArray(ciphertext, block, 0, i, i + 16);
            block = this._aes.decrypt(block);

            for (var j = 0; j < 16; j++) {
                plaintext[i + j] = block[j] ^ this._lastCipherblock[j];
            }

            copyArray(ciphertext, this._lastCipherblock, 0, i, i + 16);
        }

        return plaintext;
    }


    /**
     *  Mode Of Operation - Cipher Feedback (CFB)
     */
    var ModeOfOperationCFB = function(key, iv, segmentSize) {
        if (!(this instanceof ModeOfOperationCFB)) {
            throw Error('AES must be instanitated with `new`');
        }

        this.description = "Cipher Feedback";
        this.name = "cfb";

        if (!iv) {
            iv = createArray(16);

        } else if (iv.length != 16) {
            throw new Error('invalid initialation vector size (must be 16 size)');
        }

        if (!segmentSize) { segmentSize = 1; }

        this.segmentSize = segmentSize;

        this._shiftRegister = coerceArray(iv, true);

        this._aes = new AES(key);
    }

    ModeOfOperationCFB.prototype.encrypt = function(plaintext) {
        if ((plaintext.length % this.segmentSize) != 0) {
            throw new Error('invalid plaintext size (must be segmentSize bytes)');
        }

        var encrypted = coerceArray(plaintext, true);

        var xorSegment;
        for (var i = 0; i < encrypted.length; i += this.segmentSize) {
            xorSegment = this._aes.encrypt(this._shiftRegister);
            for (var j = 0; j < this.segmentSize; j++) {
                encrypted[i + j] ^= xorSegment[j];
            }

            // Shift the register
            copyArray(this._shiftRegister, this._shiftRegister, 0, this.segmentSize);
            copyArray(encrypted, this._shiftRegister, 16 - this.segmentSize, i, i + this.segmentSize);
        }

        return encrypted;
    }

    ModeOfOperationCFB.prototype.decrypt = function(ciphertext) {
        if ((ciphertext.length % this.segmentSize) != 0) {
            throw new Error('invalid ciphertext size (must be segmentSize bytes)');
        }

        var plaintext = coerceArray(ciphertext, true);

        var xorSegment;
        for (var i = 0; i < plaintext.length; i += this.segmentSize) {
            xorSegment = this._aes.encrypt(this._shiftRegister);

            for (var j = 0; j < this.segmentSize; j++) {
                plaintext[i + j] ^= xorSegment[j];
            }

            // Shift the register
            copyArray(this._shiftRegister, this._shiftRegister, 0, this.segmentSize);
            copyArray(ciphertext, this._shiftRegister, 16 - this.segmentSize, i, i + this.segmentSize);
        }

        return plaintext;
    }

    /**
     *  Mode Of Operation - Output Feedback (OFB)
     */
    var ModeOfOperationOFB = function(key, iv) {
        if (!(this instanceof ModeOfOperationOFB)) {
            throw Error('AES must be instanitated with `new`');
        }

        this.description = "Output Feedback";
        this.name = "ofb";

        if (!iv) {
            iv = createArray(16);

        } else if (iv.length != 16) {
            throw new Error('invalid initialation vector size (must be 16 bytes)');
        }

        this._lastPrecipher = coerceArray(iv, true);
        this._lastPrecipherIndex = 16;

        this._aes = new AES(key);
    }

    ModeOfOperationOFB.prototype.encrypt = function(plaintext) {
        var encrypted = coerceArray(plaintext, true);

        for (var i = 0; i < encrypted.length; i++) {
            if (this._lastPrecipherIndex === 16) {
                this._lastPrecipher = this._aes.encrypt(this._lastPrecipher);
                this._lastPrecipherIndex = 0;
            }
            encrypted[i] ^= this._lastPrecipher[this._lastPrecipherIndex++];
        }

        return encrypted;
    }

    // Decryption is symetric
    ModeOfOperationOFB.prototype.decrypt = ModeOfOperationOFB.prototype.encrypt;


    /**
     *  Counter object for CTR common mode of operation
     */
    var Counter = function(initialValue) {
        if (!(this instanceof Counter)) {
            throw Error('Counter must be instanitated with `new`');
        }

        // We allow 0, but anything false-ish uses the default 1
        if (initialValue !== 0 && !initialValue) { initialValue = 1; }

        if (typeof(initialValue) === 'number') {
            this._counter = createArray(16);
            this.setValue(initialValue);

        } else {
            this.setBytes(initialValue);
        }
    }

    Counter.prototype.setValue = function(value) {
        if (typeof(value) !== 'number' || parseInt(value) != value) {
            throw new Error('invalid counter value (must be an integer)');
        }

        for (var index = 15; index >= 0; --index) {
            this._counter[index] = value % 256;
            value = value >> 8;
        }
    }

    Counter.prototype.setBytes = function(bytes) {
        bytes = coerceArray(bytes, true);

        if (bytes.length != 16) {
            throw new Error('invalid counter bytes size (must be 16 bytes)');
        }

        this._counter = bytes;
    };

    Counter.prototype.increment = function() {
        for (var i = 15; i >= 0; i--) {
            if (this._counter[i] === 255) {
                this._counter[i] = 0;
            } else {
                this._counter[i]++;
                break;
            }
        }
    }


    /**
     *  Mode Of Operation - Counter (CTR)
     */
    var ModeOfOperationCTR = function(key, counter) {
        if (!(this instanceof ModeOfOperationCTR)) {
            throw Error('AES must be instanitated with `new`');
        }

        this.description = "Counter";
        this.name = "ctr";

        if (!(counter instanceof Counter)) {
            counter = new Counter(counter)
        }

        this._counter = counter;

        this._remainingCounter = null;
        this._remainingCounterIndex = 16;

        this._aes = new AES(key);
    }

    ModeOfOperationCTR.prototype.encrypt = function(plaintext) {
        var encrypted = coerceArray(plaintext, true);

        for (var i = 0; i < encrypted.length; i++) {
            if (this._remainingCounterIndex === 16) {
                this._remainingCounter = this._aes.encrypt(this._counter._counter);
                this._remainingCounterIndex = 0;
                this._counter.increment();
            }
            encrypted[i] ^= this._remainingCounter[this._remainingCounterIndex++];
        }

        return encrypted;
    }

    // Decryption is symetric
    ModeOfOperationCTR.prototype.decrypt = ModeOfOperationCTR.prototype.encrypt;


    ///////////////////////
    // Padding

    // See:https://tools.ietf.org/html/rfc2315
    function pkcs7pad(data) {
        data = coerceArray(data, true);
        var padder = 16 - (data.length % 16);
        var result = createArray(data.length + padder);
        copyArray(data, result);
        for (var i = data.length; i < result.length; i++) {
            result[i] = padder;
        }
        return result;
    }

    function pkcs7strip(data) {
        data = coerceArray(data, true);
        if (data.length < 16) { throw new Error('PKCS#7 invalid length'); }

        var padder = data[data.length - 1];
        if (padder > 16) { throw new Error('PKCS#7 padding byte out of range'); }

        var length = data.length - padder;
        for (var i = 0; i < padder; i++) {
            if (data[length + i] !== padder) {
                throw new Error('PKCS#7 invalid padding byte');
            }
        }

        var result = createArray(length);
        copyArray(data, result, 0, 0, length);
        return result;
    }

    ///////////////////////
    // Exporting


    // The block cipher
    var aesjs = {
        AES: AES,
        Counter: Counter,

        ModeOfOperation: {
            ecb: ModeOfOperationECB,
            cbc: ModeOfOperationCBC,
            cfb: ModeOfOperationCFB,
            ofb: ModeOfOperationOFB,
            ctr: ModeOfOperationCTR
        },

        utils: {
            hex: convertHex,
            utf8: convertUtf8
        },

        padding: {
            pkcs7: {
                pad: pkcs7pad,
                strip: pkcs7strip
            }
        },

        _arrayTest: {
            coerceArray: coerceArray,
            createArray: createArray,
            copyArray: copyArray,
        }
    };


    // node.js
    if (typeof exports !== 'undefined') {
        module.exports = aesjs

    // RequireJS/AMD
    // http://www.requirejs.org/docs/api.html
    // https://github.com/amdjs/amdjs-api/wiki/AMD
    } else if (typeof(define) === 'function' && define.amd) {
        define(aesjs);

    // Web Browsers
    } else {

        // If there was an existing library at "aesjs" make sure it's still available
        if (root.aesjs) {
            aesjs._aesjs = root.aesjs;
        }

        root.aesjs = aesjs;
    }


})(this);


/************** (from file ./js/other/aes2.js)**/ 

var DB = {};
DB.db = null;
DB.dbName = "ANNODE";
DB.bSUPPORTED = false;
DB.ERROR_INSERTS = 0;
DB.DO_THE_ONETIME_PRELOAD=false;
DB.index_relonsminboot_txid = null;
DB.index_relon_txid = null;
DB.version = 4;

DB.editName = function($POST) {
    if (DB.dbName.indexOf("_"+$POST)!=-1) return;
    DB.dbName = DB.dbName+"_"+$POST;
}

/**
 * 
 * NOTE:  we could just store a col flag of MINBOOT, in the same table to avoid the drama of different tables
 * 
 * 
 * The relons table is accumulated over all app usages and will get VERY large.
 * We strategically load relons out of this table into RAM on-demand in key places throughout the apps.
 * Typically this is along-side the API queries to go get info from the ANNODE.
 * If relons return we save them into the relons table along with updating our history optimizer which is 
 * intended as a query filter for future ANNODE queries.  It documents the last time that query was used
 * and saves the last block (height) for which results were found.  The next time it queries it will use
 * that historic information as a filter to avoid redundant data pulls from the ANNODE.
 */


/**
 * @param {*} $tbls 
 * @returns 
 */
DB.getCnt=async($tbls)=> {
    var total = 0;
    for (var t=0; t < $tbls.length; t++) {
        var tbl = $tbls[t];
        var objs = await DB.loadAll(tbl)
        total+=objs.length;
    }
    return total;
}


/**
 * @author {Mr.Scatman}
 */
DB.___init_lite=async()=> {
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB.editName("testnet");
    }
    if (DB.bSUPPORTED==false) {
        return false;
    }
    return await DB.openConnection(DB.dbName, DB.version, null, null);
}



/**
 * @author {Mr.Scatman}
 */
 DB.___init=async($onLoadCallback=null, $callbackCallback=null)=> {
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB.editName("testnet");
    }

    //big trouble for this app...
    if (DB.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return;
    }

    //open existing (or new, if first time) DB.
    var done = await DB.openConnection(DB.dbName, DB.version, $onLoadCallback, $callbackCallback);
    if (done==true) {
        var aFullPullHistory = await DB.loadAll("fullPullHistoryANNE");
        var aSiteSettings = await DB.loadAll("settings");

        //initial load of RAM is limited to hypergraph_preload.
        var aR = await DB.loadAll("relons");
        var aRelonsToProcess = [];

        //just get it done, all now, in-mem.
        if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) {
            for (var a=0; a < aR.length; a++) {
                if (aR[a].dead!=undefined) {
                    console.warn("HOW DID THIS GET IN HERE????");
                    delete aR[a].dead;
                }
                if (aR[a].temp!=undefined) {
                    console.warn("HOW DID THIS GET IN HERE????");
                    delete aR[a].temp; 
                }
                aRelonsToProcess.push(aR[a]);
                if (aR[a].preload!=undefined && aR[a].preload==1) {
                    if(ANNE.MEMP._preloadNids.indexOf(aR[a].from_nid)==1) {
                        ANNE.MEMP._preloadNids.push(aR[a].from_nid)
                    }
                }
                if (aR[a].to_type!=TO_TYPE_STRING) continue;
                if (aR[a].type==TYPE_SYS_MSG) continue;
                ANNE.MEM.addToTextIndex(aR[a].to_s, aR[a].from_nid);
            }
        }
        else {
            for (var a=0; a < aR.length; a++) {
                if (aR[a].dead!=undefined) {
                    console.warn("HOW DID THIS GET IN HERE????");
                    delete aR[a].dead;
                }
                if (aR[a].temp!=undefined) {
                    console.warn("HOW DID THIS GET IN HERE????");
                    delete aR[a].temp; 
                }
                if (aR[a].preload!=undefined && aR[a].preload==1) {
                    aRelonsToProcess.push(aR[a]);
                    if(ANNE.MEMP._preloadNids.indexOf(aR[a].from_nid)==1) {
                        ANNE.MEMP._preloadNids.push(aR[a].from_nid)
                    }
                }
                if (aR[a].to_type!=TO_TYPE_STRING) continue;
                if (aR[a].type==TYPE_SYS_MSG) continue;
                ANNE.MEM.addToTextIndex(aR[a].to_s, aR[a].from_nid);
            }
        }
        

        if ($onLoadCallback!=null) {
            $onLoadCallback(aRelonsToProcess, aFullPullHistory, aSiteSettings, $callbackCallback);
        }
        
        else {
            var response = {};
            response.relons = aRelonsToProcess;
            response.history = aFullPullHistory;
            response.settings = aSiteSettings;
            response.status = "ok";
            var retObj = {status:'ok', func:"init_db", fromclaim:"annelocalstorage", results:response};
            return retObj;
        }
    }
    else if ($onLoadCallback!=null) {
        $onLoadCallback([], [], [], $callbackCallback);
    }

    else {
        var response = {};
        response.relons = [];
        response.history = [];
        response.settings = [];
        response.status = "ok";
        var retObj = {status:'ok', func:"init_db", fromclaim:"annelocalstorage", results:response};
        return retObj;
    }
}


/**
 * ???????????????????????????????????????
 * 
 * We can run this query to load hypergraph MEM anytime we go external
 * to load new ANNE relons.  If we aren't getting new ANNE relons we don't have
 * to look to local indexdb for more relons.
 * 
 * Note: We will want to do this, however, even when there are no external ANNODE relons in our query.
 * 
 * Filters are 
 * .type
 * .from_nid 
 * .reln  
 * .to_type
 * .to_s
 * .time
 * .be
 * .sponsored_by
 * .relon_sponsor
 * @param {*} $filters 
 */
DB.___geRelonsQuery=async()=>{
    console.time("getRelonsQuery");
    var aRelonsToProcess = await DB.loadAll("relons");
    console.timeEnd("getRelonsQuery");
    return aRelonsToProcess;
}


/**
 * @author {Mr.Scatman}
 */
DB.___deleteEntireDB=async()=> {
    if (DB.bSUPPORTED==false) return;
    if (DB.db==null) return;
    return new Promise((resolve, reject) => {
        DB.db.close();
        var request = indexedDB.deleteDatabase(DB.dbName, 2);
        request.onblocked = function () {
            console.log("Delete blocked.");
            return resolve(false);
        }
        request.onerror = function () {
            console.log("Error deleting the DB");
            return resolve(false);
        };
        request.onsuccess = function () {
            console.log("Deleted OK.");
            return resolve(true);
        }   
    });             
}


DB.getDBSize=async()=> {
    var size = await DB.getDBTblSize("fullPullHistoryANNE");
    size+= await DB.getDBTblSize("settings");
    size+= await DB.getDBTblSize("relons");
    return size;
}

DB.getDBTblSize=async($tbl, $method="readwrite")=> {
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                var arr = event.target.result;
                var json = JSON.stringify(arr);
                size = json.length;
                return resolve(size);
            };
        } 
    });
}



/**
 * @author {Mr.Scatman}
 */
DB.openConnection=async($dbName, $version)=> {
    if (DB.db!=null) return true;
    return new Promise((resolve, reject) => {
        var request = indexedDB.open($dbName, $version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("****************************");
            console.warn("openConnection request.onUpradeneeded");
            console.log("****************************");
        
            DB.db = event.target.result;
        
            console.warn(DB.db.objectStoreNames);

            if (!DB.db.objectStoreNames.contains('relons')) {
                console.log("ADD THE RELONS STORE");
                var relonObjStore = DB.db.createObjectStore("relons", { keyPath: "txid" });
                DB.index_relon_txid = relonObjStore.createIndex("txid", "txid", { unique: true });
            }
        
            if (!DB.db.objectStoreNames.contains('fullPullHistoryANNE')) {
                console.log("ADD THE fullPullHistoryANNE");
                var fullPullHistoryStore = DB.db.createObjectStore("fullPullHistoryANNE", { keyPath: "pullid" });
                fullPullHistoryStore.createIndex("pullid", "pullid", { unique: true });
            }
            else console.log("we already have this table.");
        
            if (!DB.db.objectStoreNames.contains('settings')) {
                console.log("ADD THE settings");
                var settingsObjStore = DB.db.createObjectStore("settings", { keyPath: ["uniqueID", "misc_value", "user_nid"] });
                settingsObjStore.createIndex("valclassuser", ["uniqueID", "misc_value", "user_nid"], { unique: true });
            }
        };
    });
}



DB.getTransaction=async($txid)=> {
    return new Promise((resolve, reject) => {
        let getRequest = DB.index_relon_txid.get($txid);
        getRequest.onsuccess = function($e) {
            let rez = $e.target.result;
            if (rez!=null) {
                console.log("tx found:", rez.txid);
                return resolve(tx);
            } 
            else {
                console.log("No tx found with txid:", $txid);
                return resolve(null);
            }
        };
        getRequest.onerror = function($e) {
            console.error("Error retrieving record:", $e.target.error);
            return resolve(null);
        };
    });
}


DB.getByRelonsMinBootTxid=function($txid) {
    let getRequest = DB.index_relonsminboot_txid.get(txidToLookup);
    getRequest.onsuccess = function(event) {
        let record = event.target.result;
        if (record) {
            console.log("Record found:", record);
        } else {
            console.log("No record found with txid:", txidToLookup);
        }
    };
    getRequest.onerror = function(event) {
        console.error("Error retrieving record:", event.target.error);
    };
}



/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.loadAll=async($tbl, $method="readwrite")=> {
    if (DB.db==null) {
        console.log("DB.db is null");
        return;
    }
    //console.log("DB.loadAll "+$tbl);
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                return resolve(event.target.result);
            };
        } 
        else {
            var arr = [];

            // Fallback to the traditional cursor approach if getAll isn't supported.
            objectStore.openCursor().onsuccess = function(event) {
                var cursor = event.target.result;
                if (cursor) {     
                    arr.push(cursor.value);       
                    cursor.continue();
                } 
                else return resolve(arr);
            };
        }
    });
}


/**
 * @param {*} $objs 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.___insert = async($objs, $tbl="relons", $method="readwrite")=> {
    if (DB.bSUPPORTED==false) return;
    if ($objs.length==0) return;

    if (DB.db==null) {
        console.log("db was not initiated!");
        return;
    }

    if (DB.DO_THE_ONETIME_PRELOAD==true) {
        //init aware here.
        if (ANNE.AWARE) await ANNE.AWARE.init(); //this will require some nuanced handling.
    }

    //do the normal relons insert.
    if ($objs.length) {
        return await DB.____finalInsert($objs, $tbl, $method);
    }
    else return [];
}



/**
 * @param {*} $objs 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.____finalInsert=async($objs, $tbl, $method)=> {
    DB.ERROR_INSERTS = 0;
    DB.reroutes_needed = [];

    //some willy-wonkas are required here.
    //we try to optimize RAM by pulling only relons we think we need.
    //Sometimes the ANNODE returns relons we already have in indexedb but we 
    //HAVE NOT YET LOADED in RAM, even with our RAM pull prior to the indexedb pull.
    //This can happen when we pull the getAllMy$tuff and look in indexedb for our sponsors.
    //We may not be documented yet as the sponsor of a neuron so we won't get that relon...
    //yet the ANNODE API call gets that relon which is already in indexedb.  Since we don't have it 
    //yet in RAM we will think it is new and try to insert it into indexed
    //In ADD NEW mode, the MEM processor will attempt to save in IDB all relons we didn't have in ram.
    //This will cause a key error..  We capture that key error and then reroute it through the ___update
    DB.reroutes_needed = [];

    //normal processing here...
    if (DB.DO_THE_ONETIME_PRELOAD==false) {
        for (var o=0; o < $objs.length; ++o) {
            
            //we never insert his relon.
            if ($tbl=="relons" && $objs[o].reln==MEM.INTRA_BLOCK_PENDING.nid) continue;

            var updateReroute = await DB.____insertRecord($tbl, $method, $objs[o]);
            if (updateReroute!=null) {
                if ($objs[o].type==TYPE_INTERNAL_PAY || $objs[o].type==TYPE_BROADCAST_PAY) {
                    console.log("no need to update internal/broadcast pay relon "+$objs[o].txid);
                }
                else await DB.___update($objs[o], $tbl);
            }
        }
    }


    //one-time preload batch insert.. fast.
    else {
        
        DB.insertPRELOAD_ONETIME_Batch($tbl, $method, $objs, 0);
        DB.DO_THE_ONETIME_PRELOAD = false;
    }

    if (DB.ERROR_INSERTS>0) {
        console.log("%c error inserts requiring updates: %c"+DB.ERROR_INSERTS, ANNE.TALK.bgblack, ANNE.TALK.bgred);
    }

    DB.ERROR_INSERTS=0;
    return;
}



/**
 * We need to update the proper place when we update a relon.
 * If we updated a min-boot record, update the relon in that table.
 * Otherwise, just update our normal relons table.
 */
DB.__handleRelonUpdate=async($obj, $method="readwrite")=>{
    if (DB.bSUPPORTED==false) return null;
    var transaction  = DB.db.transaction(["relons"], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore("relons");
    if (objectStore==null) return;

    var sanitized_rObj = {
        txid:$obj.txid, 
        type:parseInt($obj.type),
        from_nid:$obj.from_nid,
        reln:$obj.reln,
        to_type:parseInt($obj.to_type),
        to_s:$obj.to_s,
        time:$obj.time,
        firstseen:$obj.firstseen,
        freeze:$obj.freeze,
        preload:$obj.preload 
    }

    if ($obj.old_txid!=undefined) sanitized_rObj.old_txid = $obj.old_txid;
    if (sanitized_rObj.txid[0]=="z") {
        console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
        console.warn(sanitized_rObj);
        alert("hunt this down!");
        return null;
    }

    if ($obj.owned_by!=undefined && $obj.owned_by!=null) {
        sanitized_rObj.owned_by=$obj.owned_by;
    }
    else $obj.owned_by="";
    if ($obj.sponsored_by!=undefined && $obj.sponsored_by!=null) {
        sanitized_rObj.sponsored_by=$obj.sponsored_by;
    }
    else $obj.sponsored_by="";

    if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies;
    if ($obj.parent!=undefined) sanitized_rObj.parent=$obj.parent;
    if ($obj.origin!=undefined) sanitized_rObj.origin=$obj.origin;
    if ($obj.rorigin!=undefined) sanitized_rObj.rorigin=$obj.rorigin;
    if ($obj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$obj.relon_sponsor;
    if ($obj.from_text!=undefined) sanitized_rObj.from_text=$obj.from_text;
    if ($obj.ntype!=undefined) sanitized_rObj.ntype=$obj.ntype;
    if ($obj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$obj.sys_ref_txid;
    if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies; 
    if ($obj.decrypted!=undefined) sanitized_rObj.decrypted=$obj.decrypted;
    if ($obj.flathexint!=undefined) {
        sanitized_rObj.flathexint = $obj.flathexint;
    }

    return new Promise((resolve, reject) => {
        request = objectStore.put(sanitized_rObj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}




/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @param {*} $rObjs 
 * @param {*} $startIndex 
 * @returns 
 */
DB.insertPRELOAD_ONETIME_Batch=function($tbl, $method, $rObjs, $startIndex=0) {
    console.log("%c insertPRELOAD_ONTIME_Batch %c"+$tbl, ANNE.TALK.bgblack, ANNE.TALK.bgyellow)
    const batchSize = 40000;  
    const endIndex = Math.min($startIndex + batchSize, $rObjs.length);

    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    var txids = [];
    
    for (var i = $startIndex; i < endIndex; i++) {
        var $rObj = $rObjs[i];

        //wiillllllllllllly wonka, yet again.
        if (txids.indexOf($rObj.txid)!=-1) continue;
        txids.push($rObj.txid);

        var sanitized_rObj = {
            txid:$rObj.txid, 
            type:parseInt($rObj.type),
            from_nid:$rObj.from_nid,
            reln:$rObj.reln,
            to_type:parseInt($rObj.to_type),
            to_s:$rObj.to_s,
            time:$rObj.time,
            firstseen:$rObj.firstseen,
            freeze:$rObj.freeze,
            preload:$rObj.preload
        }

        //DONT SAVE THIS, EVER
        if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

        //for bad ghost cleanup fixing
        if ($rObj.old_txid!=undefined) sanitized_rObj.old_txid = $rObj.old_txid;
        if (sanitized_rObj.txid==undefined || sanitized_rObj.txid==null) {
            console.warn("bad data, no txid");
            alert("hunt this down!");
            resolve(null);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            resolve(null);
        }

        if ($rObj.anbl_id!=undefined) sanitized_rObj.anbl_id=$rObj.anbl_id;

        if ($rObj.owned_by!=undefined && $rObj.owned_by!=null) {
            sanitized_rObj.owned_by=$rObj.owned_by;
        }
        else $rObj.owned_by="";
        if ($rObj.sponsored_by!=undefined && $rObj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$rObj.sponsored_by;
        }
        else $rObj.sponsored_by="";
        if ($rObj.parent!=undefined) sanitized_rObj.parent=$rObj.parent;
        if ($rObj.origin!=undefined) sanitized_rObj.origin=$rObj.origin;
        if ($rObj.rorigin!=undefined) sanitized_rObj.rorigin=$rObj.rorigin;
        if ($rObj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$rObj.relon_sponsor;
        if ($rObj.from_text!=undefined) sanitized_rObj.from_text=$rObj.from_text;
        if ($rObj.ntype!=undefined) sanitized_rObj.ntype=$rObj.ntype;
        if ($rObj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$rObj.sys_ref_txid;
        if ($rObj.annies!=undefined) sanitized_rObj.annies=$rObj.annies;
        if ($rObj.decrypted!=undefined) sanitized_rObj.decrypted=$rObj.decrypted;
        if ($rObj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $rObj.flathexint;
        }
        objectStore.add(sanitized_rObj);
    }

    transaction.oncomplete = function() {
        //console.log("WHEN DOES THIS GET PRINTED??");
        if (endIndex < $rObjs.length) {
            DB.insertBatch($tbl, $method, $rObjs, $endIndex);
        } 
        else {
            //console.log('All records inserted successfully!');
            return true;
        }
    };

    transaction.onerror = function(event) {
        console.error('Error inserting records:', event.target.error);
        return null;
    };
}

/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @param {*} $rObj 
 * @returns 
 */
DB.____insertRecord_RR=async($tbl, $method, $rObj)=> {

    return new Promise((resolve, reject) => {   
        var transaction  = DB.db.transaction([$tbl], $method);
        if (transaction==null) return;
        var objectStore = transaction.objectStore($tbl);
        if (objectStore==null) return;
        var sanitized_rObj = {
            txid:$rObj.txid, 
            type:parseInt($rObj.type),
            from_nid:$rObj.from_nid,
            reln:$rObj.reln,
            to_type:parseInt($rObj.to_type),
            to_s:$rObj.to_s,
            time:$rObj.time,
            firstseen:$rObj.firstseen,
            freeze:$rObj.freeze,
            preload:$rObj.preload 
        }

        //DONT SAVE THIS, EVER
        if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

        //for bad ghost cleanup fixing
        if ($rObj.old_txid!=undefined) sanitized_rObj.old_txid = $rObj.old_txid;
        if (sanitized_rObj.txid==undefined || sanitized_rObj.txid==null) {
            console.log("test");
            resolve(null);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            resolve(null);
        }

        if ($rObj.anbl_id!=undefined) sanitized_rObj.anbl_id=$rObj.anbl_id;

        if ($rObj.owned_by!=undefined && $rObj.owned_by!=null) {
            sanitized_rObj.owned_by=$rObj.owned_by;
        }
        else $rObj.owned_by="";
        if ($rObj.sponsored_by!=undefined && $rObj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$rObj.sponsored_by;
        }
        else $rObj.sponsored_by="";
        if ($rObj.parent!=undefined) sanitized_rObj.parent=$rObj.parent;
        if ($rObj.origin!=undefined) sanitized_rObj.origin=$rObj.origin;
        if ($rObj.rorigin!=undefined) sanitized_rObj.rorigin=$rObj.rorigin;
        if ($rObj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$rObj.relon_sponsor;
        if ($rObj.from_text!=undefined) sanitized_rObj.from_text=$rObj.from_text;
        if ($rObj.ntype!=undefined) sanitized_rObj.ntype=$rObj.ntype;
        if ($rObj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$rObj.sys_ref_txid;
        if ($rObj.annies!=undefined) sanitized_rObj.annies=$rObj.annies;
        if ($rObj.decrypted!=undefined) sanitized_rObj.decrypted=$rObj.decrypted;
        if ($rObj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $rObj.flathexint;
        }
        var request = objectStore.add(sanitized_rObj);
        request.onsuccess = function(event) {
            resolve(null);
        };
        request.onerror = function(event) {
            DB.ERROR_INSERTS+=1;
            console.warn("insert error", $rObj);
            DB.reroutes_needed.push(sanitized_rObj)
            reject(null);
        }
    });
}


/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @param {*} $rObj 
 * @returns 
 */
DB.____insertRecord=async($tbl, $method, $rObj)=> {
    
    //never insert this relon.
    if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
    
    return new Promise((resolve, reject) => {   
        var transaction  = DB.db.transaction([$tbl], $method);
        if (transaction==null) return;
        var objectStore = transaction.objectStore($tbl);
        if (objectStore==null) return;
        var sanitized_rObj = {
            txid:$rObj.txid, 
            type:parseInt($rObj.type),
            from_nid:$rObj.from_nid,
            reln:$rObj.reln,
            to_type:parseInt($rObj.to_type),
            to_s:$rObj.to_s,
            time:$rObj.time,
            firstseen:$rObj.firstseen,
            freeze:$rObj.freeze,
            preload:$rObj.preload 
        }

        //DONT SAVE THIS, EVER
        if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return resolve(null);

        //for bad ghost cleanup fixing
        if ($rObj.old_txid!=undefined) sanitized_rObj.old_txid = $rObj.old_txid;
        if (sanitized_rObj.txid==undefined || sanitized_rObj.txid==null) {
            console.log("test");
            return resolve(null);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            return resolve(null);
        }

        if ($rObj.anbl_id!=undefined) sanitized_rObj.anbl_id=$rObj.anbl_id;

        if ($rObj.owned_by!=undefined && $rObj.owned_by!=null) {
            sanitized_rObj.owned_by=$rObj.owned_by;
        }
        else $rObj.owned_by="";
        if ($rObj.sponsored_by!=undefined && $rObj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$rObj.sponsored_by;
        }
        else $rObj.sponsored_by="";
        if ($rObj.parent!=undefined) sanitized_rObj.parent=$rObj.parent;
        if ($rObj.origin!=undefined) sanitized_rObj.origin=$rObj.origin;
        if ($rObj.rorigin!=undefined) sanitized_rObj.rorigin=$rObj.rorigin;
        if ($rObj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$rObj.relon_sponsor;
        if ($rObj.from_text!=undefined) sanitized_rObj.from_text=$rObj.from_text;
        if ($rObj.ntype!=undefined) sanitized_rObj.ntype=$rObj.ntype;
        if ($rObj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$rObj.sys_ref_txid;
        if ($rObj.annies!=undefined) sanitized_rObj.annies=$rObj.annies;
        if ($rObj.decrypted!=undefined) sanitized_rObj.decrypted=$rObj.decrypted;
        if ($rObj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $rObj.flathexint;
        }
        var request = objectStore.add(sanitized_rObj);
        request.onsuccess = function(event) {
            return resolve(null);
        };
        request.onerror = function(event) {
            DB.ERROR_INSERTS+=1;
            console.warn("insert error tbl:''"+$tbl+"''", sanitized_rObj);
            ANNE.TALK.printRelon($rObj);
            return resolve(sanitized_rObj.txid);
        }
    });
}


DB.___insert_single = async($obj, $tbl="relons", $method="readwrite", $callaback=null)=> {
    if (DB.bSUPPORTED==false) return;
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    var sanitized_rObj = {}

    if ($tbl=="relons") { 
        sanitized_rObj = {
            txid:$obj.txid, 
            type:parseInt($obj.type),
            from_nid:$obj.from_nid,
            reln:$obj.reln,
            to_type:parseInt($obj.to_type),
            to_s:$obj.to_s,
            time:$obj.time,
            firstseen:$obj.firstseen,
            freeze:$obj.freeze,
            preload:$obj.preload 
        }
    
        if ($obj.old_txid!=undefined) sanitized_rObj.old_txid = $obj.old_txid;

        //DO NOT SAVE THIS, EVER
        if (sanitized_rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) {
            return resolve(true);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            return resolve(true);
        }

        if ($obj.anbl_id!=undefined) sanitized_rObj.anbl_id=$obj.anbl_id;

        if ($obj.owned_by!=undefined && $obj.owned_by!=null) {
            sanitized_rObj.owned_by=$obj.owned_by;
        }
        else $obj.owned_by="";
        if ($obj.sponsored_by!=undefined && $obj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$obj.sponsored_by;
        }
        else $obj.sponsored_by="";

        if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies;
        if ($obj.parent!=undefined) sanitized_rObj.parent=$obj.parent;
        if ($obj.origin!=undefined) sanitized_rObj.origin=$obj.origin;
        if ($obj.rorigin!=undefined) sanitized_rObj.rorigin=$obj.rorigin;
        if ($obj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$obj.relon_sponsor;
        if ($obj.from_text!=undefined) sanitized_rObj.from_text=$obj.from_text;
        if ($obj.ntype!=undefined) sanitized_rObj.ntype=$obj.ntype;
        if ($obj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$obj.sys_ref_txid;
        
        if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies; 
        if ($obj.decrypted!=undefined) sanitized_rObj.decrypted=$obj.decrypted;
       
        if ($obj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $obj.flathexint;
        }
    }
    else sanitized_rObj = $obj;

    return new Promise((resolve, reject) => {
        var request = objectStore.add(sanitized_rObj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror == function(event) {
            console.log(event);
            console.log(sanitized_rObj);
            return resolve(false);
        }       
    });
}


/**
 * 
 * @param {*} $objs 
 * @param {*} $tbl 
 * @param {*} $method 
 */
DB.___updateAll = async($objs, $tbl="relons", $method="readwrite")=> {
    for (var r=0; r < $objs.length; r++) {
        await DB.___update($objs[r], $tbl, $method);
    }
}



/**
 */
DB.___update = async($obj, $tbl="relons", $method="readwrite")=> {
    if ($tbl=="relons") {
        return await DB.__handleRelonUpdate($obj, $method);
    }
    if (DB.bSUPPORTED==false) {
        console.log("IDB not supported");
        return;
    }
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) {
        console.log("can get $tbl:"+$tbl+" with method:"+$method);
        return;
    }
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) {
        console.log("no object store for $tbl:"+$tbl);
        return;
    }
    
    return new Promise((resolve, reject) => {

        if ($obj==null || $obj==undefined ) {
            console.log("obj empty;");
            return resolve(false);
        }
        var request = null;
        var sanitized_rObj = {}

        var keys = Object.keys($obj);
        for (var k=0; k < keys.length; k++) {
            if (typeof ($obj[keys[k]])=="function") {
                alert("woh horsey!");
                delete $obj[keys[k]];
            }
                
            else if (typeof ($obj[keys[k]])=="object") {
                 delete $obj[keys[k]];
            }
                
            else if ($obj.to_sNOBJ!=null) {
                delete $obj.to_sNOBJ;
            }
                
            else if ($obj.from_nidTNOBJ!=null) {
                delete $obj.from_nidTNOBJ;
            }
            
        }
        
        request = objectStore.put($obj);
        
        request.onsuccess = function(event) {
            return resolve(true);
        };
        
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}


/**
 * 
 * @param {*} $unique 
 * @param {*} $miscval 
 * @param {*} $user_nid 
 * @returns 
 */
DB.___deleteSetting = async($unique, $miscval, $user_nid)=>{
    if (DB.bSUPPORTED==false) return false;
    var transaction  = DB.db.transaction(["settings"],"readwrite");
    if (transaction==null) return;
    var objectStore = transaction.objectStore("settings");
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        console.log("delete setting "+$unique+" miscval "+$miscval+" user nid:"+$user_nid);
        var request = objectStore.delete([$unique, $miscval, $user_nid]);
        request.onsuccess = function(event) { return; }
        request.onerror = function(event) { return; }
        return resolve(true);
    });
}


/**
 * 
 * @param {*} $txids 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.___delete = async ($txids, $tbl="relons", $method="readwrite")=> {
    if (DB.bSUPPORTED==false) return false;
    if ($txids.length==0) return false;
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        //we only delete one at time
        $txids.forEach(function($txid) {
            if ($txid==undefined || typeof $txid!=='string') return;   //just like continue in the forEach
            var request = objectStore.delete($txid);
            request.onsuccess = function(event) { return; }
            request.onerror = function(event) { return; }
        });

        return resolve(true);
    });
  }

/************** (from file ./js/db.js)**/ 

//extras

class annode_auth  {
    constructor($resultObj){
        if ($resultObj==undefined || $resultObj==null) return;
        this.annode_time = $resultObj.annode_time;
        this.annode_pubKey = $resultObj.annode_pubKey;
        this.annode_nid = $resultObj.annode_nid;
        this.annode_drl = $resultObj.annode_drl;
        this.next_proof_id = $resultObj.next_proof_id;
        if(this.annode_drl==undefined) this.annode_drl=1000;

        //we can show these in a gui of some-kind.
        this.network = $resultObj.network;
        if (this.network==undefined) this.network = false;

		this.ANNDLE_active = $resultObj.ANNDLE_active;
        if (this.ANNDLE_active==undefined) this.ANNDLE_active = false;
        
		this.ASDF_active = $resultObj.ASDF_active;
		if (this.ASDF_active==undefined) this.ASDF_active = false;
        
        this.BOUNTY_active = $resultObj.BOUNTY_active;
        if (this.BOUNTY_active==undefined) this.BOUNTY_active = false;
        
        this.PDNS_active = $resultObj.PDNS_active;
        if (this.PDNS_active==undefined) this.PDNS_active = false;
        
        this._nonce = "";
        this._encrypted_sk = "";

        //used to prove annode identity.
        this._nonce2 = "";
        this._encrypted_proof = "";
    }


    //These are hex strings from the ANNODE we are handshook with.
    //NOTE: We will neeed to change the LS user storage object to allow also for multiple
    //annodes for the same origin/app/user if we want to store more than the "last"
    //annode handshook with for the origin/app/user.
    //At present this is location in the annode code folder html/ui/annewallet/js/localstorage.js.

    //We already decrypted the shared key from these values and 
    //then encrypted them with a simpler encryption for "us" in our LS User object.
    //This is a backup for testing.
    set encrypted_sk ($esk) { this._encrypted_sk=$esk};
    get encrypted_sk(){ return this.encrypted_sk;}
    set encrypted_proof ($ep) { this._encrypted_proof=$ep};
    get encrypted_proof(){ return this._encrypted_proof;}
    set nonce ($nonce) { this._nonce=$nonce};
    get nonce () { return this._nonce; }
    set nonce2 ($nonce2) { this._nonce2=$nonce2};
    get nonce2 () { return this._nonce2; }
    get time() { return this.annode_time; }
    get pubKey() { return this.annode_pubKey; }
    get id() { return this.annode_nid; }
    drl=function() { 
        return this.annode_drl; 
    }
}

/************** (from file ./js/network/ANNODE_AUTH.js)**/ 

class anne_utils extends anne_singleton {
    constructor(){
        super(false);
	}

    formatBitcoinPrice=function($price) {
        if (isNaN($price) || $price < 0) {
            throw new Error("Invalid input: Please provide a non-negative number.");
        }
        return $price.toFixed(8);
    }

    exists = function($val) {
        if ($val==undefined) return false;
        if ($val==null) return false;
        if ($val==="") return false;
        return true;
    }
    existsNonZero() {
        if ($val==undefined) return false;
        if ($val==null) return false;
        if ($val==="") return false;
        if ($val==0) return false;
    }
    ne($a,$b) {
        if (this.exists($a)==false) {
            if (this.exists($b)==true) return true;
            else return false;
        }
        if (this.exists($b)==false) return true;
        if ($a==$b) return false;
        return true;
     }

    quickAbbreviateName = function($text) {
        $text =  $text.trim();
        if ($text.length<10) return $text;
        const suffix = ["jr", "jr.", "sr", "sr.", "iii", "ii", "iv", "i", "v", "vi", "vii", "viii", "ix"];
        var parts = $text.split(" ");

        if (parts.length==2 && $text.length<17) return $text;

        var abbr = "";
        var first = "";
        var lastpart = parts[parts.length-1];
        var addsuffix = "";
        var pmax = parts.length;

        if (suffix.indexOf(lastpart.toLowerCase())!=-1) {
            addsuffix = lastpart;
            pmax = pmax-1;
        }

        for (var p=0; p < pmax; p++) {
            if (p==0) {
                abbr+=parts[p][0];
                first=parts[p];
            }
            else if (p < pmax-1){
                abbr+=parts[p][0]; //JD Rockfeller Jr
            }
            else if (p==(pmax-1)) abbr+=" "+parts[p];
            
        }
        return abbr;
    }

    stripAllNonAlphaChar=function($str, $charSlot) {
        var char = $str[$charSlot];
        if("abcdefghijklmnopqrstuvwzyz".indexOf(char)==-1) {
            $str = $str.substring(1);
            return this.stripAllNonAlphaChar($str, $charSlot);
        }
        return $str;
    }
    stripNonAlphaNum = function($str) {
        return this.stripNonIn($str, "abcdefghijklmnopqrstuvwzyz1234567890");
    }

    isMixedCase = function($str) {
        return /[a-z]/.test($str) && /[A-Z]/.test($str);
    }

    /**
     * note to self: comment this so i remember what it is.
     * @param {*} $str 
     * @param {*} $ifstr 
     * @returns 
     */
    extractNumberIf = async($str, $ifstr)=>{
        if ($str.indexOf($ifstr)!=-1) {
            if ($afterONLY==true) {
                var i = $str.indexOf($ifstr)+$ifstr.length;
                var newstr = $str.substr(i);
                return this.extractNumber($str);
            }
            else {
                return this.extractNumber($str);
            }
        }
        return null;
    }

    isSameYear = function ($nObj1, $nObj2) {
        var year1 = -1;
        var year2 = -1;
        
        if ($nObj1.bes.indexOf(ANNE.MEM.integer.temp_nid)!=-1) year1 = parseInt($nObj1.text);
        else {
            var year_relon = ANNE.MEM.getHasRelon($nOb1.outs, MEM.year.temp_nid);
            if (year_relon!=null) {
                var yearNOBJ = MEMG.getNeuronByNid(year_relon.to_s);
                if (yearNOBJ!=null) year1 = parseInt(yearNOBJ.text);
            }
        }

        if ($nObj2.bes.indexOf(ANNE.MEM.integer.temp_nid)!=-1) year2 = parseInt($nObj2.text);
        else {
            var year_relon = ANNE.MEMG.getHasRelon($nObj2.outs, MEM.year.temp_nid);
            if (year_relon!=null) {
                var yearNOBJ = MEMG.getNeuronByNid(year_relon.to_s);
                if (yearNOBJ!=null) year2 = parseInt(yearNOBJ.text);
            }
        }

        if(year1!=-1 && year2!=-1 && year1==year2) return true;
        return false;
        
    }


    extractNumber2 = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    var numNOBJ = await MEMG.getOrMakeIfNewInteger(null, number);
                    if (numNOBJ!=null) {
                        parts.splice(p, 1);
                        return {nObj:numNOBJ, str:part, remaining:parts.join(" ")};
                    }
                    break;
                }
            }
        }
        return null;
    }


    extractNumber = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    var numNOBJ = await MEMG.getOrMakeIfNewInteger(null, number);
                    if (numNOBJ!=null) {
                        return numNOBJ;
                    }
                    break;
                }
            }
        }
        return null;
    }

    extractNumberStr = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    return number+"";
                }
            }
        }
        return "";
    }


    replaceWeirdLetters = function($str) {
        const regex = /[À-ÖØ-öø-ÿ]/g;
        const AtoZ = {
          À: "A",
          Á: "A",
          Â: "A",
          Ã: "A",
          Ä: "A",
          Å: "A",
          Æ: "AE",
          Ç: "C",
          È: "E",
          É: "E",
          Ê: "E",
          Ë: "E",
          Ì: "I",
          Í: "I",
          Î: "I",
          Ï: "I",
          Ð: "D",
          Ñ: "N",
          Ò: "O",
          Ó: "O",
          Ô: "O",
          Õ: "O",
          Ö: "O",
          Ø: "O",
          Ù: "U",
          Ú: "U",
          Û: "U",
          Ü: "U",
          Ý: "Y",
          Þ: "TH",
          ß: "ss",
          à: "a",
          á: "a",
          â: "a",
          ã: "a",
          ä: "a",
          å: "a",
          æ: "ae",
          ç: "c",
          è: "e",
          é: "e",
          ê: "e",
          ë: "e",
          ì: "i",
          í: "i",
          î: "i",
          ï: "i",
          ð: "d",
          ñ: "n",
          ò: "o",
          ó: "o",
          ô: "o",
          õ: "o",
          ö: "o",
          ø: "o",
          ù: "u",
          ú: "u",
          û: "u",
          ü: "u",
          ý: "y",
          þ: "th",
          ÿ: "y",
        };
        var result = $str.replace(regex, (match) => AtoZ[match]);
        console.log(result); // Output: "AAA EEE III OOO UUU"
        return result;
    }


    containsTXID=function($rels, $txid) {
        for (var r=0; r <$rels.length; r++) {
            if ($rels[r].txid==$txid) return true;
        }
        return false;
    }
    

    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    fixHack=function($str=null, $replace_underscores="") {
        if ($str==undefined || $str==null || typeof $str !="string") {
            console.log("don't call this, sir, and non-existant $str");
            return "";
        }

        //dont muck with date looking strings.
        var aParts = $str.split("-");
        if (aParts.length>=2) {
            var numbtest = parseInt($str.replaceAll("-", ""));
            if (numbtest > 0) {
                var parenObj = ANNE.UTILS.stripParens($str);
                console.log(parenObj);
                return parenObj;
            }
        }
        
        $str = $str.replaceAll("-2D","-");
        $str = $str.replaceAll("-28", "(");
        $str = $str.replaceAll("-29", ")");

        //too tired to chase it.  lazy kill
        //we already yank info from wikispooks in a hacky non rdf way..
        //so it comes with drama
        var apostrophe = $str.indexOf("-27");
        if (apostrophe!=-1 && apostrophe!=0 && apostrophe!=$str.length-1) {
            $str = $str.replaceAll("-27", "'");
        }
        
        var fix = $str.indexOf("-28unidentified-29");
        if (fix!=-1 && fix!=0) {
            $str = $str.replaceAll("-28unidentified-29", "");
        }

        $str = $str.replaceAll("-26", "&");
        $str = $str.replaceAll("27-", "");

        var parenObj = ANNE.UTILS.stripParensMulti($str);

        if ($replace_underscores!="") {
            parenObj.str = parenObj.str.replaceAll("_", $replace_underscores).trim();
        }
        else parenObj.str.trim();
        
        //console.log(parenObj);
        return parenObj;
    }

    
    blowOutAmpersands = function($arr) {
        var finalArray = [];
        for (var a=0; a < $arr.length; a++) {
            var val = $arr[a];
            if (val.indexOf("&")==-1) finalArray.push(val);
            else {
                var parts = val.split(" & ");
                for (var p=0; p < parts.length; p++) {
                    var val2 = parts[p].trim();
                    if (finalArray.indexOf(val2)==-1) finalArray.push( parts[p].trim() );
                }
            }
        }
        return finalArray;
    }


    //bob and suzy smith
    //s.b.  "bob smith", "suzy smith"
    extrapAmperStr = function($str) {
        var parts = $str.split(" & ");
        if (parts.length==1) part2 = $str.split(" and ");
        if (parts.length!=2) return null;
        
        var last_part = parts[ parts.length-1];
        var first_part = parts [0];
        if (last_part.split(" ").length > first_part.split(" ").length) {
            var last = last_part.split(" ");
            var last_name = last [ last.length-1];
            var first_person = first_part+" "+last_name;
            return [first_person, last_part];
        }
        return null;
    }

    /**
     * Strips a string of its paren and returns the new string with the stripped value in the context.
     * @param {*} $str 
     * @returns 
     */
    stripParens = function($str){
        const regex = /\s?\(([^)]+)\)/;
        var match = $str.match(regex);
        var newStr = $str.replace(/\s?\([^)]+\)/, '');
        var context = "";
        if (match && match.length>0) context = match[1];
        if (newStr==null || newStr=="" || newStr.length==0) newStr = $str;
        return {str:newStr.trim(), context:context};
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    prepNakedSearch = function($str) {
        var pOBJ = this.stripParens($str);
        pOBJ.inputStr = $str;
        const inputString = pOBJ.str;
        const regexPattern = /(['"])(.*?)\1/;

        let QUOTED = false;
        let extractedString = '';

        const match = inputString.match(regexPattern);
        if (match) {
            QUOTED = true;
            extractedString = match[2]; 
        }

        if (QUOTED==true) {
            pOBJ.bEXACT=true;
            pOBJ.str = extractedString;
        }
        else pOBJ.bEXACT=false;

        console.log("naked search:", pOBJ);
        return pOBJ;
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    stripParensMulti=function($str) {
        const regex = /\(([^)]+)\)/;
        const match = $str.match(regex);
        let newStr = $str.replace(/\([^)]+\)/, '');
        let context = [];
        if (match && match.length > 0) {
            context = match[1].split(',').map(item => item.trim());
        }
        if (newStr == null || newStr == "" || newStr.length == 0) {
            newStr = $str;
        }
        return {str: newStr.trim(), context: context};
    }


    /**
     * 
     * @param {*} $str 
     * @param {*} $stripifnotthis 
     * @returns 
     */
    stripNonIn=function($str, $stripifnotthis) {
        var newstr = $str.concat();
        for (var a=0; a < newstr.length; a++)  {
            var chara = newstr[a];
            if ($stripifnotthis.indexOf(chara)==-1) {
                newstr.slice(a, 1);
                a--;
            }
        }
        return newstr;
    }


    /**
     * For testing purposes, the firstseen datestamp marks
     * when the annode first saw a relon.
     * 
     * If the relon was in the mempool (time/block=-1) but the annode restarted..
     * and the relon never made it back to the mempool...
     * 
     * We have a faux relon in our localDB that the ANNE network doesn't actually know
     * about.  Our annode may know of it, but the the network does not.
     * 
     * We need to fix our indexDB.
     * 
     * We need to solve this problem in the annode as well.
     * Shutdown mid-mempool will leave neuron indexes created, neurons froze, etc..
     * which will reject the next attempt to create the same neuron indexes
     * (and will actually resolve the nid index conflict by creating a new one)
     * 
     * This will be a HUGE problem as that annode will be out of concensus.
     * @param {*} $sec 
     * @returns 
     */
    shouldBeMined = function($sec) {
        var se = this.timeElapsed($sec);
        if (se > 600) return true;
        return false;
    }

    timeElapsed = function($secs) {
        var currentTimestamp = Date.now();
        var timeElapsed = currentTimestamp - parseInt($secs);
        var secondsElapsed = Math.floor(timeElapsed / 1000);
        console.log("Time elapsed in seconds: " + secondsElapsed);
        return secondsElapsed;
    }


    secsElapsedToDHMS = function($secs) {
        const secmin = 60;
        const minshr = 60;
        const sechr = secmin*60;
        const secsday = sechr*24;
        var hrsday = 24;
        var minutes = Math.floor($secs/60);

        if (minutes > secmin) {
            var hours = Math.floor(minutes/60);
            if (hours > hrsday) {
                var days = Math.floor(hours / hrsday);
                var hrs_left = hours - (days*hrsday);
                var mins_left = Math.round(minutes % 60);
                var secs_left = Math.round($secs % 60);
                var dayswd = ANNE.TALK.pluralizeNounIf(days, "day", "s");
                var hrswd = ANNE.TALK.pluralizeNounIf(hrs_left, "hr", "s");
                var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
                var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s")
                return days+dayswd+" "+hrs_left+hrswd+" "+mins_left+minswd+" "+secs_left+secsswd;
            }
            else {
                hours = Math.floor(hours);
                var mins_left = Math.round(minutes % 60);
                var secs_left = Math.round($secs % 60);
                var hrswd = ANNE.TALK.pluralizeNounIf(hours, "hr", "s");
                var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
                var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s");
                return hours+hrswd+" "+mins_left+minswd+" "+secs_left+secsswd;
            }
        }
        else if (minutes > 0) {
            var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
            var secs_left = Math.round($secs % 60);
            var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s")
            return minutes +minswd+" "+secs_left+secsswd;
        }
        else {
            var secsswd = ANNE.TALK.pluralizeNounIf($secs, "sec", "s")
            return $secs+secsswd;
        }
    }

    forceInt = function($int) {
        if ($int==undefined || $int==null || $int=="") return 0;
        var int2 = parseInt($int);
        if (isNaN(int2)==true) return 0;
        return int2;
    }

    capFirstLetters = function($str) {
        var wds = $str.split(" ");
        for (var p=0; p < wds.length; p++) {
            wds[p] = wds[p][0].toUpperCase() + wds[p].substr(1);
        }
        return wds.join(" ");
    }

    concatUniqueNObjs=function($a1, $a2) {
        if ($a1.length==0) return $a2;
        if ($a2.length==0) return $a1;
        var a_temp_nid_list=[];
        for (var a=0; a < $a1.length; a++) {
            a_temp_nid_list.push($a1[a].temp_nid);
        }
        var a2_uniques = [];
        for (var a2=0; a2 < $a2.length; a2++) {
        var a2_temp_nid = $a2[a2].temp_nid;
            if (a_temp_nid_list.indexOf(a2_temp_nid)!=-1) continue;
            a2_uniques.push($a2[a2]);
        }
        return $a1.concat(a2_uniques);
    }


    copyToClipboard=function($str) {
        var el = document.createElement('textarea');
        el.value = $str;
        el.setAttribute('readonly', '');
        el.style = {position: 'absolute', left: '-9999px'};
        document.body.appendChild(el);
        el.select();
        document.execCommand('copy');
        document.body.removeChild(el);
    }

    concatUniqueRels = function ($arr1, $arr2) {
        var txids = [];
        var rels = [];
        for (var t=0; t < $arr1.length; t++) {
            var txid = $arr1[t].txid;
            if (txids.indexOf(txid)==-1) {
                txids.push(txid);
                rels.push($arr1[t]);
            }
        }
        for (var t=0; t < $arr2.length; t++) {
            var txid = $arr2[t].txid;
            if (txids.indexOf(txid)==-1) {
                txids.push(txid);
                rels.push($arr2[t]);
            }
        }
        return rels;
    }
    

    convertToCSV = function ($objArray, $bGENHEADER=false) {
        var array = typeof $objArray != 'object' ? JSON.parse($objArray) : $objArray;
        var str = '';

        if ($bGENHEADER==true) {
            var keys = Object.keys(array[0]);
            var header = new Object();
            for (var k=0; k < keys.length; k++) {
                header[keys[k]]=keys[k];
            }
            array.unshift(header);
        }

        for (var i = 0; i < array.length; i++) {
            var line = '';
            for (var index in array[i]) {
                if (line != '') line += ','

                line += array[i][index];
            }
            str += line + '\r\n';
        }
        return str;
    }

    l33t_checkForce = function($str) {
        var l33ts_num = [];
        var l33ts_alpha = [];
        
        //1234567890
        //1=l
        for (var s=0; s < $str.length; s++) {
            var char = $str[s]=$str[s];
            if (char=="0") {
                l33ts_num.push("0");
                l33ts_alpha.push("O");
            }
            else if (char=="1") {
                l33ts_num.push("1");
                l33ts_alpha.push("l");
            }
            else if (char=="2") {
                l33ts_num.push("1");
                l33ts_alpha.push("l");
            }
            else {
            } 
        }

    }


    /**
     * Alpha num.
     * single spaces ok.
     */
    cleanAnorgName = function($propName) {
        var str = $propName.replace(/ +(?= )/g,'');
        var cleanstr = str.replace(/[^a-z0-9 ]/gi,'');
        return cleanstr;
    }


    cleanPropName = function($propName) {
        var b = $propName.replace(/[^a-z0-9_]/gi,'');
        if (b.match(/^\d/)==true) {
            console.log("(utils.cleanPropname) result: "+"p_"+b);
            return "p_"+b;
        }
        return b;
    }

    roundThenCommas=function($x) {
        var rounded = Math.round($x);
        return this.commas($x);
    }

    //.toFixed(8);
    abbreviateNumber=function($num) {
        $num = new Number($num);
        if ($num >= 1e12) {
            return (Math.round($num / 1e11) / 10) + 't'; 
        } 
        else if ($num >= 1e9) {
            return (Math.round($num / 1e8) / 10) + 'b'; 
        } 
        else if ($num >= 1e6) {
            return (Math.round($num / 1e5) / 10) + 'm'; 
        } 
        else if ($num >= 1e3) {
            return (Math.round($num / 1e2) / 10) + 'k'; 
        } 
        else {
            return $num.toString(); 
        }
    }

    commas=function($x, $bDOSPACES=false) {
        if ($x === undefined || $x === null) {
            console.log("huh?");
            return;
        }

        let parts = $x.toString().split('.');
        let integerPart = parts[0];
        let decimalPart = parts.length > 1 ? '.' + parts[1] : '';
    
        if ($bDOSPACES === true) {
            integerPart = integerPart.replace(/\B(?=(\d{3})+(?!\d))/g, " ");
        } 
        else {
            integerPart = integerPart.replace(/\B(?=(\d{3})+(?!\d))/g, ",");
        }
    
        return integerPart + decimalPart;
    }


    estimateTimeByBlocks = function($blocksMore) {
        var minsMore = $blocksMore*1.31;
        var secsMore = minsMore*60;
        const currentDate = new Date();
        const futureDate = new Date(currentDate.getTime() + secsMore * 1000);
        const options = { month: 'long', day: 'numeric', year: 'numeric' };
        const formattedDate = futureDate.toLocaleDateString('en-US', options);
        console.log(formattedDate);
        return formattedDate;
    }   

    estimateTimeByBlocksLeft = function($blocksMore) {
        var minsMore = $blocksMore*1.31;
        var secsMore = minsMore*60;
        return this.secsElapsedToDHMS(secsMore);
    }

    
    firstCap = function($str, $bALL=true) {
        if ($bALL==true) {
            var newstr = [];
            var parts = $str.split(" ");
            if (parts.length==1) return parts[0][0].toUpperCase()+parts[0].substring(1);
            for (var p=0; p < parts.length; p++) {
                var str = parts[p][0].toUpperCase()+parts[p].substring(1);
                newstr.push(str);
            }
            return newstr.join(" ");
        }
        else {
            var parts = $str.split(" ");
            if (parts.length==1) return parts[0][0].toUpperCase()+parts[0].substring(1);
        } 
    }


    //won't get fully spelled numbers like "one"
    isNum = function($n) {
        if (!isNaN(parseFloat($n)) && isFinite($n)==true) return true;
        const regex = /^\d+(st|nd|rd|th)?$/;
        return regex.test($n);
    }
    

    /**
     * 
     * from https://gist.github.com/tralves/9e5de2bd9f582007a52708d7d4209865
     * @param {*} bytes 
     * @returns 
     */
    frenlySize = function (bytes) {
        var thresh = 1024;
        if (Math.abs(bytes) < thresh) return bytes + ' B';
        var units = ['KB','MB','GB','TB','PB','EB','ZB','YB'];
        var u = -1;
        do {
            bytes /= thresh;
            ++u;
        } 
        while(Math.abs(bytes) >= thresh && u < units.length - 1);
        return bytes.toFixed(1)+' '+units[u];
    }



    /**
     * Calculate annex price lvl factor via nth fib.
     * 
     * could be done with matrix math but we won't be getting high lvl numbers here
     * and that optimization is just showing off.
     * this suffices.
     * 
     * DO NOT CHANGE THIS FUNCTION UNLESS YOU ALSO CHANGE THE FUNCTION ON THE ANNODE.
     * 
     * @param {*} $nth 
     * @returns 
     */
    nthFib=function($nth){
        if ($nth==1 || $nth==2) return 1;
        var seq = [1, 1];
        var nextval = 0;
        for (var n=2; n < $nth; n++) {
            nextval = seq[n-1] + seq[n-2];
            seq.push(nextval);
        }
        return nextval;
    }


    /**
     * Only 2 decimal places
     * @param {*} $annies 
     * @param {*} $gap 
     * @returns 
     */
    displayAnneValue = function($annies, $gap=" ") {
        var parts = ((parseInt($annies)/ANNE.ANNE_TO_ANNIES)+"").split(".");
        if (parts.length==2) {
            var whole = UTILS.commas(parts[0]);
            var deci = parts[1];
        }
        else if (parts.length==1) {
            if ($annies<ANNE.ANNE_TO_ANNIES) {
                var whole = 0;
                var deci = parts[0];
            }
            else {
                var whole = UTILS.commas(parts[0]);
                var deci = "0";
            }
        }
        if (deci==undefined) return whole;
        deci = deci+"";
        
        if (deci.length==2) deci = "."+$gap+deci;
        else if (deci.length==1) deci = "."+$gap+deci+"0";
        else if (deci.length==0) return whole;
        
        return whole;//+deci;
    }


    /**
     * 
     * @param {*} $val 
     * @param {*} $minDigits 
     * @returns 
     */
    formatDigits = function($val, $minDigits=4) {
         if ($val==null) {
            console.log("%cERROR val can't be null here", TALK.bgred);
            return "";
        }
        var final = $val+"";
       
        if (final.length<$minDigits) {
            var dif = $minDigits-final.length;
            for (var d=1; d <= dif; d++) { final = "0"+final; }
        }
        return final;
    }

    
    //For calculating edit distance
    editDistance = function (s1, s2) {
        s1 = s1.toLowerCase();
        s2 = s2.toLowerCase();
        var costs = new Array();
        for (var i = 0; i <= s1.length; i++) {
            var lastValue = i;
            for (var j = 0; j <= s2.length; j++) {
                if (i == 0)
                    costs[j] = j;
                else {
                if (j > 0) {
                    var newValue = costs[j - 1];
                    if (s1.charAt(i - 1) != s2.charAt(j - 1)) newValue = Math.min(Math.min(newValue, lastValue), costs[j]) + 1;
                    costs[j - 1] = lastValue;
                    lastValue = newValue;
                }
            }
        }
        if (i > 0)
            costs[s2.length] = lastValue;
        }
        return costs[s2.length];
    }


    /**
     * 
     * @param {*} $dataEl 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportJSONFileViaID=function($dataEl, $fileTitle, $el="", $bGENHEADER=true) {
        console.log("'"+$dataEl+"'");
        var jsonStringData = document.getElementById($dataEl).innerHTML;
        if (jsonStringData) {
            this.exportJSONFile(jsonStringData, $fileTitle, $el, $bGENHEADER);
        }
    }

    /**
     * 
     * @param {*} $object 
     * @param {*} $fileTitle 
     */
    exportJSONFile=function($object, $fileTitle) {
        var exportedFilenmae = $fileTitle + '.json' || 'export.json';
        var jsonstring = JSON.stringify($object);
        const bytes = new TextEncoder().encode(jsonstring);
        const json = new Blob([bytes], {
            type: "application/json;charset=utf-8"
        });

        var link = document.createElement("a");
        if (link.download !== undefined) { // feature detection
            // Browsers that support HTML5 download attribute
            var url = URL.createObjectURL(json);
            link.setAttribute("href", url);
            link.setAttribute("download", exportedFilenmae);
            link.style.visibility = 'hidden';
            var el = null;
            el = document.body;
            el.appendChild(link);
            link.click();
            el.removeChild(link);
        }
    }

    /**
     * 
     * @param {*} $dataEl 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportCSVFileViaID=function($dataEl, $fileTitle, $bGENHEADER=true) {
        console.log("'"+$dataEl+"'");
        var jsonStringData = document.getElementById($dataEl).innerHTML;
        if (jsonStringData) {
            this.exportCSVFile(jsonStringData, $fileTitle, $bGENHEADER);
        }
    }

    /**
     * 
     * @param {*} $jsonStringOrObjectArray 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportCSVFile=function($jsonStringOrObjectArray, $fileTitle,  $bGENHEADER=true) {
        var csv = this.convertToCSV($jsonStringOrObjectArray, $bGENHEADER);
        this.exportCSVFile_inner(csv, $fileTitle);
    }

    /**
     * 
     * @param {*} $csv 
     * @param {*} $fileTitle 
     */
    exportCSVFile_inner=function($csv, $fileTitle) {
        var exportedFilenmae = $fileTitle + '.csv' || 'export.csv';
        var blob = new Blob([$csv], { type: 'text/csv;charset=utf-8;' });
        if (navigator.msSaveBlob) { // IE 10+
            navigator.msSaveBlob(blob, exportedFilenmae);
        } 
        else {
            var link = document.createElement("a");
            if (link.download !== undefined) { // feature detection
                // Browsers that support HTML5 download attribute
                var url = URL.createObjectURL(blob);
                link.setAttribute("href", url);
                link.setAttribute("download", exportedFilenmae);
                link.style.visibility = 'hidden';
                var el = null;
                el = document.body;
                el.appendChild(link);
                link.click();
                el.removeChild(link);
            }
        }
    }

    posNegRand = function($fact) {
        return (Math.random()-.5)*$fact;
    }

    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    findValAndRemoveInObjnArr = function ($arr, $prop, $val) {
        for (var a=0; a<$arr.length; a++) {
            if ($arr[a][$prop]==$val){
                $arr.splice(a, 1);
                return true;
            }
        }
        return false;
    }

    /**
     * 
     * @param {*} $rels 
     * @param {*} $skip_tnids 
     * @returns 
     */
    getLatest = function($rels, $skip_tnids=[]) {
        var lastValid = null;
        for (var sr=0; sr < $rels.length; sr++) {
            var rel = $rels[sr];
            if ($skip_tnids.length && $skip_tnids.indexOf(rel.r)!=-1) continue;
            if (lastValid==null) lastValid = rel;

            //if we have a bigger block height OR a 0 (indicating mempool not updated yet)
            else if ((lastValid.time) > 0 && (rel.time > 0 || rel.time==0)) {
                if (rel.time > lastValid.time) lastValid = rel;
            }

            //this has been firstseen later, meaning it is newer
            else if (lastValid.firstseen < rel.firstseen) {
                lastValid = rel;
            }
        }
        return lastValid;
    }


    /**
     * 
     * @param {*} $rels 
     * @returns 
     */
    getLastSeen = function($rels) {
        if ($rels==null) return null;
        if ($rels.length==0) return null;
        if ($rels.length==1) return $rels[0];
        var last_seen = 0;
        var lastseenRel = null;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (rel.firstseen!=undefined && rel.firstseen > last_seen) {
                lastseenRel = rel;
                last_seen = rel.firstseen;
            }
        }
        if (lastseenRel!=null) return lastseenRel;
        if ($rels.length) return $rels [ $rels.length-1 ];
    }

        
    /**
     * 
     * @param {*} $rels 
     * @param {*} $annie 
     * @returns 
     */
    getFirst = function($rels, $annie=0) {
        if ($annies>0) return this.getFirstWithGTEAnnies($rel, $annies);
        var first = null;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (first==null) first=rel;
            else if ((rel.time < first.time) || (rel.time==0 && first.time!=0)) first=rel;
            else if (rel.time==first.time) {
                if (rel.firstseen < first.firstseen) first=rel;
            }
        }
        return first;
    }


    /**
     * attemps to handle multi payments if there was underpayment.
     * will use time/first seen from the last payment in the bunch to qualify
     * the payment as being the first seen (winner) of the FCFS 
     * 
     * @param {*} $rels 
     * @param {*} $annies 
     * @returns 
     */
    getFirstWithGTEAnnies = function($rels, $annies) {
        var first = null;
        var from_multi = new Object();

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var annies = rel.annies; 
            var rel_fs = rel.firstseen;
            var rel_time = rel.time;

            if (annies==undefined || annies==0) continue;
            if (annies < $annies) {
                var fm_this = from_multi[ rel.f ];
                if (fm_this==undefined) {
                    from_multi[ rel.f ] = {annies:annies, time:rel.time, firstseen:rel.firstseen, rel:rel};
                }
                else {
                    fm_this.annies += annies; 
                    if (rel.time > fm_this.time) {
                        fm_this.time = rel_time;
                        fm_this.firstseen = rel_fs;
                        fm_this.rel = rel;
                    }

                    if (fm_this.annies < $annies) continue; //no enough, still!
                }
            }

            if (first==null) first=rel;
            else if ((rel_time < first.time) || (rel_time==0 && first.time!=0)) first=rel;
            else if (rel_time==first.time) {
                if (rel_fs < first.firstseen) first=rel;
            }
        }
        return first;
    }


    /**
     * 
     * @param {*} $txt 
     * @param {*} $bDECIMALOK 
     * @returns 
     */
    inputNumberOnly = function ($txt, $bDECIMALOK=false) {
        var newString = "";
        for (var c=0; c < $txt.length; c++) {
            var charCode = $txt.charCodeAt(c);
            //console.log(charCode);
            if ($bDECIMALOK==false) {
                if (charCode < 48 || charCode> 57) continue;
                newString = newString + $txt[c];
            }
            else if ($bDECIMALOK==true) {
                if (charCode < 48 || (charCode> 57 && charCode!=190)) continue;
                newString = newString + $txt[c]; 
            }  
        }
        return newString;
    }


    /**
     */
    addUniqueRel=function($arr, $rel) {
        UTILS.addUnique($arr, "txid", $rel);
    }


    /**
     * 
     * @param {Array} $objArr 
     * @param {Array} $destination_arr 
     * @param {String} $uniqueObjProp 
     */
    addUniqueObjArray=function($objArr, $destination_arr, $uniqueObjProp="txid") {
        for (var r=0; r < $objArr.length; r++) {
            var newObj = $objArr[r];
            UTILS.addUnique($destination_arr, $uniqueObjProp, newObj);
        }
    }

    /**
     * 
     * @param {*} $arr 
     * @param {*} $tnids 
     */
    addUniqueTnids=function($arr, $tnids) {
        for (var t=0; t < $tnids.length; t++) {
            var tnid=$tnids[t];
            if ($arr.indexOf(tnid)==-1) $arr.push(tnid);
        }
    }



    /**
     * 
     * @param {Array} $arr 
     * @param {String} $uniqueProp 
     * @param {Object} $newRelon 
     * @returns 
     */
    addUnique=function($arr, $uniqueProp="txid", $newRelon) {
        if ($arr==null) return false;
        if ($newRelon==null) return false;

        if (this.findValInObjnArr($arr, $uniqueProp, $newRelon[$uniqueProp])==null) {
            $arr.push($newRelon);
            return true;
        }
        return false;
    }


    /**
     * 
     */
    deepCopy=function($obj) {
        if ($obj === null || typeof $obj !== 'object') {
            return $obj;
        }
    
        let copy = new $obj.constructor();
    
        for (let key in $obj) {
            if ($obj.hasOwnProperty(key)) {
                copy[key] = this.deepCopy($obj[key]);
            }
        }
        return copy;
    }
    
    
    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    findValInObjnArr = function ($arr, $prop, $val) {
        if ($prop!="") {
            for (var a=0; a < $arr.length; a++) if ($arr[a][$prop]==$val) return $arr[a];
            return null;
        }
        for (var a=0; a < $arr.length; a++) if ($arr[a]==$val) return $arr[a];
        return null;
    }


    /**
     * Does an item exist in the array with the existing prop value set?
     * @param {*} $arr 
     * @param {*} $prop 
     * @returns 
     */
    findValInObjnArr2 = function ($arr, $prop, $val) {
        for (var a=0; a < $arr.length; a++) if ($arr[a][$prop]==$val) return true;
        return false;
    }


    sortByProp = function($arr, $prop) {
        $arr.sort( 
            function($a, $b) {
                var val_a = $a[$prop];
                if (val_a==undefined) return 0;

                var val_b = $b[$prop];
                if (val_b==undefined) return 0;

                if (val_a < val_b) return -1;
                else if (val_a > val_b) return 1;
                return 0;
            }
        );
        return $arr;
    }


    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    removeObjInObjArr = function($arr, $prop, $val) {
        for (var a=0; a<$arr.length; a++) {
            if ($arr[a][$prop]==$val){
                $arr.splice(a, 1);
                return;
            }
        }
    }


    //weak attempt to add .png to end of bad imgur url
    fixImgURL=function($imgUrl) {
        if ($imgUrl.indexOf("imgur")==-1) return $imgUrl;   //do nothing
        var aParts = $imgUrl.split(".");
        aParts.reverse();
        var ext = aParts[0];
        aParts.reverse();
        if (ext.length<3 || ext.length>4) $imgUrl = $imgUrl+".png";
        else if (ext!="png" && ext!="jpg") $imgUrl = $imgUrl+".png";
        return $imgUrl;
    }


    /**
     * try to format the bigger number.
     * if it has a deci and we can't short-hand it with a B or M or K,
     * then just format the number with commas.
     */
    formatNumber=function($num, $bROUND=false) {
        var num_parts = ($num+"").toString().split(".");
        var num = parseInt(num_parts[0]);
        if (num >0 && num_parts.length>1 && $bROUND==false) return this.formatNumberCommas($num);

        var b = num / 1000000000;
        if (b>=1) {
            if ($bROUND==true) {
                var b100=Math.floor(b*100)/100;
                num_parts[0]=((b100+"").split(".").join("."))+" B";
            }
            else num_parts[0]=b+"B";
        }
        else {
            var m = num / 1000000;
            if (m>=1) {
                if ($bROUND==true) {
                    var m100=Math.floor(m*100)/100;
                    num_parts[0]=((m100+"").split(".").join("."))+" M";
                }
                else num_parts[0]=m+"M";
            }
            else {
                var k = num / 1000;
                if (k>=1) {
                    if ($bROUND==true) {
                        var k100=Math.floor(k*100)/100;
                        num_parts[0]=((k100+"").split(".").join("."))+" K";
                    }
                    else num_parts[0]=k+" K";
                }
                else num_parts[0]=num_parts[0];
            }
        }
        if ($bROUND==true) return num_parts[0];
        else if (num_parts.length==1) return num_parts[0];
        return num_parts[0]+"."+num_parts[1];
    }


    formatNumberCommas=function($num) {
        var num_parts = ($num+"").toString().split(".");
        num_parts[0] = num_parts[0].replace(/\B(?=(\d{3})+(?!\d))/g, ",");
        return num_parts.join(".");
    }

    convArr=function($arr) {
        var final_arr = [];
        var keys = Object.keys($arr);
        for (var k=0; k < keys.length; k++) {
            final_arr.push ( $arr [ keys[k]]);
        }
        return final_arr;
    }
    

    getSelectedOptionFromDD=function($dd_el) {
        var sel = document.getElementById($dd_el);
        return this.getSelectedOption(sel);
    }

    getSelectedOption = function($sel) {
        var opt;
        for ( var i = 0, len = $sel.options.length; i < len; i++ ) {
            opt = $sel.options[i];
            if ( opt.selected === true ) {
                break;
            }
        }
        return opt;
    }

    getUrlVars = function () {
        var vars = {};
        var parts = window.location.href.replace(/[?&]+([^=&]+)=([^&]*)/gi, function (m, key, value) {
            vars[key] = value;
        });
        return vars;
    }

    hex_to_ascii=function($str1) {
        var hex  = $str1.toString();
        var str = '';
        for (var n = 0; n < hex.length; n += 2) {
            str += String.fromCharCode(parseInt(hex.substr(n, 2), 16));
        }
        return str;
    }

    htmlEntities=function(str) {
        return String(str).replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/"/g, '&quot;').replace(/'/g, '&apos;');
    }

    
    isPrime($x){     
        var d = $x-1;
        while (d > 1){
          if (($x % d) == 0) return false;
          d--;
        }
        return true;
      }


    /**
     * 
     * @param {*} $map 
     * @param {*} $prop 
     * @returns 
     */
    __sortNumObjMapByProp = function($map, $prop="other") {
        var keys = Object.keys($map);
        keys.sort( function($a, $b) {
            var val_a = $map[$a][$prop];
            if (val_a==undefined) return 0;

            var val_b = $map[$b][$prop];
            if (val_b==undefined) return 0;

            if (val_a < val_b) return -1;
            else if (val_a > val_b) return 1;
            return 0;
        });
        
        var sortedMapObj = new Object();
        keys.forEach(key => {
            sortedMapObj[key] = $map[key];
          });
        return sortedMapObj;
    }


    __sortRelnKeyMap = function($relnMAP) {
        var keys = Object.keys($relnMAP);
        keys.sort( function($a, $b) {
            var rels_a = $relnMAP[$a];
            var text_a = MEMG.getNeuronByTempNid(rels_a[0].r).text.toLowerCase();
            var rels_b = $relnMAP[$b];
            var text_b = MEMG.getNeuronByTempNid(rels_b[0].r).text.toLowerCase();
            if (text_a < text_b) return -1;
            else if (text_a > text_b) return 1;
            return 0;
        });
        return keys;
    }

    endsWithStrip = function($str, $char=",") {
        if ($str[$str.length-1]==$char) {
            return $str.substring(0, $str.length-1);
        }
        return $str;
    }
    
    /**
     * 
     * @param {*} $val 
     * @returns 
     */
    isset = function($val) {
        if ($val==undefined || $val==null) return false;
        var _type = typeof $val;
        if (_type =="boolean") return $val;
        
        if (_type =="string") { 
            if($val.length==0) return false;
            return true;
        }
        if (_type=="number") {
            if($val==0) return false;
        }
        if (Array.isArray($val)==true && $val.length==0) return false;
        return true;
    }


    /**
     * move over unique props from one object to another
     * This is helpful when merging objects that are not in a custom class..
     * (basic objs or json being used)
     */
    moveUniquePropsToObj=function($obj_from, $obj_to) {
        if ($obj_from == null) return;
        if ($obj_to == null) return;
        var keys = Object.keys($obj_from);
        for (var k=0; k < keys.length; k++) {
            var prop = keys[k];
            if ($obj_to[prop]!=undefined) continue;
            var val = $obj_from[ prop ];
            $obj_to[prop]=val;
        }
    }


    printunixds=function($unixds) { 
        var d = new Date($unixds); //, timeZone:'UTC', timeZoneName:'short'
        const options = { year: 'numeric', month: '2-digit', day: '2-digit'};
        return d.toLocaleDateString('fr-CA', options)+" "+d.toLocaleTimeString();
    }

    //too tired to right now to implement the listsplitter.
    //this is for lazy sloppy jackass .csv files that include multiple
    //values in the column with a list.
    //we can split them into their own column and then give the user
    //the option to map it to an AKA.
    //for now, it gets nuked.
    processCSV_withHeader=async($csv, $delim, $listsplitter=";")=> {
        var allTextLines = $csv.split(/\r\n|\n/);
        var headers = [];
        var headerRow = allTextLines[0].split($delim);
        for (var h=0; h<headerRow.length; h++) {
            var clean = headerRow[h].replace(/['"]+/g, '');
            headers.push(clean.trim());
        }
        console.log(headers);
        var csvRowObjs = [];
        for (var n=1; n < allTextLines.length-1; n++) {
            var row = allTextLines[n].split($delim);
            var rowObj = [];
            for (var j=0; j<row.length; j++) {
                var clean2 = row[j].replace(/['"]+/g, '');
                var colName = headers[j];
                if (clean2.indexOf($listsplitter)!=-1) {
                    var aParts = clean2.split(";");
                    clean2 = aParts[0].trim();
                }
                if (headers!=null) rowObj[ colName ] = clean2;
                else rowObj.push(clean2.trim());
            }
            rowObj.__ROW_ACTIVE=true;
            csvRowObjs.push(rowObj);
        }
        return {'headers':headers, 'data': csvRowObjs};
    }

    processJsonKeyPairs=async($json, $delim)=> {
        var headers = [];
        var finaldata = [];
        for (var n=0; n < $json.length; n++) {
            var row = $json[n];
            //first loop snag all props from first row and make headers.
            if (n==0) {
                for (var prop in row) {
                    var header = prop;//.replace(/['"]+/g, '');
                    headers.push(header);
                }
            }
            var rowObj = [];
            //loop the headers, snag the objects
            for (var h=0; h<headers.length; h++) {
                var prop = headers[h];
                var value = row[prop];
                if (value=="null") value=="";
                else if (typeof value ==='string') value = value.replace(/['"]+/g, '');
                rowObj[ prop ] = value;
            }
            rowObj.__ROW_ACTIVE=true;
            finaldata.push(rowObj);
        }
        return {'headers':headers, 'data': finaldata};
    }
    
    pushUniqueProp=function($obj, $prop, $val) {
        if (Array.isArray($obj[$prop]) == true) { if ($obj[$prop].indexOf($val)==-1) $obj[$prop].push($val); }
        else {
            $obj[$prop]=[];
            $obj[$prop].push($val);
        }
    }

    fisherYatesShuffle=function($array) {
        for (let i = $array.length - 1; i > 0; i--) {
            const j = Math.floor(Math.random() * (i + 1));
            [$array[i], $array[j]] = [$array[j], $array[i]];
        }
        return $array;
    }

    range = function ($min,$max) {
        return Math.floor(Math.random()*($max-$min+1)+$min);
    }

    inclusive_range($min, $max) {
        return Math.floor(Math.random() * ($max - $min + 1)) + $min;
    }

    randTXID = function($len=20) {
        var txid = "";
        for (var s=1; s <= $len; s++) {
            txid = txid+Math.floor(Math.random()*10);
        }
        return "z"+txid;
    }

    formatDeci = function($num, $benchmark) {
        var decis = this.decis($benchmark);
        return $num.toFixed(decis);
    }

    getRadioValue=function($radelid) {
        const radios = document.getElementsByName($radelid);
        for (let i = 0; i < radios.length; i++) {
            if (radios[i].checked) {
                return radios[i].value; 
            }
        }
        return null;
    }

    /**
     * 
     * @param {*} $base 
     * @param {*} $per - in the form of a whole number.  4 = 4%. 
     */
    addPercent=function($base, $per) {
        var decis =  this.decis($base);
        var base = $base;
        console.log(typeof base);
        if ((base instanceof Number) == false) base = new Number($base);
        base = base + (base * $per/100);
        return base.toFixed(decis+1);
    }

    setRadioValue=function($radelid, $value) {
        const radios = document.getElementsByName($radelid);
        for (let i = 0; i < radios.length; i++) {
            if (radios[i].value != $value) {
                radios[i].checked=false;
                continue;
            }
            radios[i].checked=true; 
            return;
        }
        return;
    }

    decis = function($num) {
        var numStr = $num;
        if (typeof $num == "Number") numStr = $num.toString();
        else numStr = $num+"";
        if (numStr.includes('.')) {
            return numStr.split('.')[1].length;
        }
        return 0;
    }


    /**
     * ghosts start with z.
     * @param {*} $len 
     * @returns 
     */
    randNID = function($len=20) {
        var txid = "";
        for (var s=1; s <= $len; s++) {
            txid = txid+Math.floor(Math.random()*10);
        }
        return "z"+tx 
    }

    convertDMSLatToDec($degrees, $minutes, $seconds) {
        return $degrees + ($minutes / 60) + ($seconds / 3600);
    }
    convertDMSLongToDec($degrees, $minutes, $seconds, $direction) {
        var dd = $degrees + ($minutes / 60) + ($seconds / 3600);
        if ($direction === 'W') {
            return dd * -1;
        } 
        else if ($direction === 'E') {
            return dd;
        }
    }
    testGeo=function() {
        //giza great pyramid
        var lat = this.convertDMSLatToDD(40, 42, 31.0, "N"); // 40.70861°N
        var long = this.convertDMSLongToDD(74, 0, 32.0, "W"); // -74.00889°W
        console.log(lat+", "+long)
    }
    


    
    getRandHex = function($len) { return UTILS.randomHex($len);}
    randHex = function($len) { return UTILS.randomHex($len); }
    randomHex = function($len) {
        var maxlen = 8;
        var min = Math.pow(16,Math.min($len,maxlen)-1) 
        var max = Math.pow(16,Math.min($len,maxlen)) - 1;
        var n   = Math.floor( Math.random() * (max-min+1) ) + min;
        var r   = n.toString(16);
        while ( r.length < $len ) {
            r = r + this.randomHex( $len - maxlen );
        }
        return r;
    };

    randomSpherePoint=function($x0,$y0,$z0,$radius){
        var u = Math.random()+.01;
        var v = Math.random()+.01;
        var theta = 2 * Math.PI * u;
        var phi = Math.acos(2 * v - 1);
        var pt = new Object();
        pt.x = $x0 + ($radius * Math.sin(phi) * Math.cos(theta));
        pt.y = $y0 + ($radius * Math.sin(phi) * Math.sin(theta));
        pt.z = $z0 + ($radius * Math.cos(phi));
        return pt;
    }

    //nid or tnid array works here
    removeFromNidArray=function($arr, $nid) {
        var index = $arr.indexOf($nid)
        if (index!==-1)  $arr.splice(index,1);
    }

    removeFromTnidsArray=function($tnids_arr, $temp_nid) {
        var index = $tnids_arr.indexOf($temp_nid)
        if (index!==-1)  $tnids_arr.splice(index,1);
    }

    removeFromRelonsArray=function($relons_arr, $r) {
        for (var r=0; r < $relons_arr.length; r++) {
            var rel = $relons_arr[r];
            if (rel.r==$r) {
                $relons_arr.splice(r,1);
                return $relons_arr;
            }
        }
        return $relons_arr;
    }

    removeRelonByTxid=function($relons_arr, $txid) {
        for (var r=0; r < $relons_arr.length; r++) {
            var rel = $relons_arr[r];
            if (rel.txid==$txid) {
                $relons_arr.splice(r,1);
                return $relons_arr;
            }
        }
        return $relons_arr;
    }

    
    getHighestSimilarity = function($NOBJs, $lookupval) {
        var highestSim = 0;
        var highestNOBJ = null;
        for (var c=0; c <$NOBJs.length; c++) {
            var nObj = $NOBJs[c];
            var sim = ANNE.UTILS.similarity(nObj.text, $lookupval);
            if (sim>highestSim) {
                highestSim = sim;
                highestNOBJ = nObj;
            }
        }
        return {nObj:highestNOBJ, sim:highestSim};
    }


    //levenshtein distance, altered
    similarity = function(s1, s2, $factor=1) {
        var longer = s1;
        var shorter = s2;
        if (s1==undefined || s1==null) {
            alert("woh");
            return 0;
        }
        if (s2==undefined || s2==null) {
            alert("woh");
            return 0;
        }
        if (s1.length < s2.length) {
            longer = s2;
            shorter = s1;
        }
        var longerLength = longer.length;
        
        if (longerLength == 0) return 1.0; //s1 and s2 have no length? lolz.

        var lenFact = shorter.length/longer.length;
        if ($factor>1) {
            var final = Math.floor($factor * (longerLength - this.editDistance(longer, shorter)) / parseFloat(longerLength));
            return (final/$factor)*lenFact;
        }
        else return ((longerLength - this.editDistance(longer, shorter)) / parseFloat(longerLength))*$factor;
    }


    sortByNid=function($a, $b) {
        if ($a==undefined || $a==null || $b.nid==undefined || $b.nid==null) return 0;
        if ($a.nid < $b.nid) return -1;
        else if ($a.nid > $b.nid) return 1;
        return 0;
    }


    //targetting the earliest here..
    sortByBeTime=function($a, $b) {
        if ($a==undefined || $a.be_time==null || $b.be_time==undefined || $b.be_time==null) return 0;
        if ($a.be_time < $b.be_time) return 1;
        else if ($a.be_time > $b.be_time) return -1;
        return 0;
    }


    //sort clump neurons by their WHEN role.
    sortClumpsByWhen=function($events) {
        if ($events==undefined || $events==null || $events.length<=1) return;
        $events.sort( UTILS.sortByWhenRole );
        return;
    }
    


    //these are role_when arrays.  We will only use the first
    //in 99.99% of cases there will only be one of a given role per sentence.
    sortByWhenRole=function($a, $b, $bFWD=true) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.role_when==undefined || $a.role_when==null) {
            if ($b.role_when!=undefined && $b.role_when!=null) return 1; 
            else return 0;
        }
        else if ($b.role_when==undefined && $b.role_when==null) return -1; 

        var a_when = MEMG.getNeuronByTempNid($a.role_when[0].t);
        var b_when = MEMG.getNeuronByTempNid($b.role_when[0].t);
        
        if ($bFWD) return UTILS.compareDates(b_when, a_when);
        else return UTILS.compareDates(a_when, b_when);
    }

    tnidsUnion = function($a, $b) {
        var union = [];
        for (var a=0; a <$a.length; a++) {
            var atnid = $a[a];
            for (var b=0; b <$b.length; b++) {
                if ($b[b]==atnid) {
                    if (union.indexOf(atnid)==-1) {
                        union.push(atnid);
                        break;
                    }
                }
            }   
        }
        return union;
    }


    //we should have an ANNE epoch day number for all dates.
    //that would be a better (more efficient) date compare.

    compareDates = function($a, $b) {
        $a.processAsDateNeuron();
        $b.processAsDateNeuron();
        if ($a.year < $b.year) return -1;
        else if ($a.year > $b.year) return 1;
        else {
            if ($a.month < $b.month) return -1;
            else if ($a.month > $b.month) return 1;
            else {
                if ($a.day < $b.day) return -1;
                else if ($a.day > $b.day) return 1;
                else return 0;
            }
        }
        return 0;
    }

    isUSSuffix=function($test) {
        var suffix = ["JR", "SR", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X"];
        if ($test[0]==".") $test = $test.substring(1);
        $test = $test.toUpperCase();
        if (suffix.indexOf($test)!=-1) return true;
        return false;
    }

    removeOption = function($el, $val) {
        var el = document.getElementById($el);
        if (el==undefined || el==null) return;
        var options = document.querySelector("#"+$el).options;
        if (options==undefined || options==null) return;
        for (var o=0; o < options.length; o++) {
            if (options[o].value==$val) {
                options.remove(o);
                break;
            }
        }
    }


    roughSizeOfObject($object) {
        var objectList = [];
        var stack = [$object];
        var bytes = 0;
        while (stack.length) {
            var value = stack.pop();
            if (typeof value === 'boolean') bytes += 4;
            else if (typeof value === 'string') bytes += value.length * 2;
            else if (typeof value === 'number') bytes += 8;
            else if (typeof value === 'object' && objectList.indexOf( value ) === -1) {
                objectList.push( value );
                for( var i in value ) stack.push( value[ i ] );
            }
        }
        return bytes;
    }

    sleep=function($ms) {
        return new Promise(resolve => setTimeout(resolve, $ms));
    }

    //test this with DOM converstion...
    sortDD = function ($el) {
        var sel = docuement.getElementById($el);
        var selected = sel.value; // cache selected value, before reordering
        var opts_list = sel.querySelector('option');
        opts_list.sort(function (a, b) { return document.getElementById(a).text > document.getElementById(b).text ? 1 : -1; });
        sel.innerHTML="";
        sel.append(opts_list);  //append children back
        sel.value=selected; 
    }

    sortByFire=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null || $a.fire==undefined || $b.fire==undefined) return 0;
        if ($a.fire < $b.fire) return -1;
        if ($a.fire > $b.fire) return 1;
        return 0;
    }

    sortByFireThenIns=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null || $a.fire==undefined || $b.fire==undefined) return 0;
        if ($a.fire > $b.fire) return -1;
        if ($a.fire < $b.fire) return 1;
        if ($a.ins.length > $b.ins.length) return -1;
        if ($a.ins.length < $b.ins.length) return 1;
        return 0;
    }

    sortByDimension=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.type < $b.type) return -1;
        if ($a.type > $b.type) return 1;
        return 0;
    }

    sortByText=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.text < $b.text) return -1;
        if ($a.text > $b.text) return 1;
        return 0;
    }

    sortTnidsByAlpha=function($tnids) {
        var nObjs = MEMG.getNeuronsFromTnids($tnids);
        nObjs.sort(this.sortByTextNoCase);
        return MEMG.getTnidList(nObjs);
    }

    sortByTextNoCase=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.text==undefined || $b.text==undefined || $a.text==null || $b.text==null) {
            console.log("alsdfkjalsdfjlafjadkfj;alsdfj");
            return 0;
        }
        if ($a.text.toUpperCase() < $b.text.toUpperCase()) return -1;
        if ($a.text.toUpperCase() > $b.text.toUpperCase()) return 1;
        return 0;
    }


    sortRelonsByFromNObjText=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.to_type!=TO_TYPE_ANNE) return -1;
        if ($b.to_type!=TO_TYPE_ANNE) return -1;
        var aNOBJ = MEMG.getNeuronByTempNid($a.f);
        if (aNOBJ==null || aNOBJ.text==undefined) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.f);
        if (bNOBJ==null || bNOBJ.text==undefined) return;
        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }

    sortRelonsByToNObjText=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.to_type!=TO_TYPE_ANNE) return -1;
        if ($b.to_type!=TO_TYPE_ANNE) return -1;
        var aNOBJ = MEMG.getNeuronByTempNid($a.t);
        if (aNOBJ==null) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.t);
        if (bNOBJ==null) return;
        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }


    sortGhostsLast=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.ghost==true) {
            if ($b.ghost==false) return 1;
            else return 0;
        }
        return 0;
    }

    sortByRelnTextNoCase=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        var aNOBJ = MEMG.getNeuronByTempNid($a.r);
        if (aNOBJ==null) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.r);
        if (bNOBJ==null) return;

        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }

    sortByTime=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return -1;
        if ($b.time && $a.emulated==false <=0) return 1;
        if ($a.time < $b.time) return 1;
        if ($a.time > $b.time) return -1;
        return 0;
    }
    sortByTimeAsc=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return 1;
        if ($b.time && $a.emulated==false <=0) return -1;
        if ($a.time < $b.time) return -1;
        if ($a.time > $b.time) return 1;
        return 0;
    }
    sortByTimeASC=function($b,$a) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return -1;
        if ($b.time && $a.emulated==false <=0) return 1;
        if ($a.time < $b.time) return 1;
        if ($a.time > $b.time) return -1;
        return 0;
    }
    sortByFirstSeen=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.firstseen > $b.firstseen) return 1;
        if ($a.firstseen < $b.firstseen) return -1;
        return 0;
    }




    syntaxHighlight=function($json) {
        json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) {
                if (/:$/.test(match)) {
                    cls = 'key';
                } else {
                    cls = 'string';
                }
            } else if (/true|false/.test(match)) {
                cls = 'boolean';
            } else if (/null/.test(match)) {
                cls = 'null';
            }
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }

        

    /**
     * validate an ipv4 addressed
     * @param {*} $nObj 
     * @returns {Boolean} 
     */
     __ipv4Validation=function($nObj) {
        console.log("(__ipv4Validation) to do!");
        return true;
    }


    /**
     * validate an ipv6 address
     * @param {*} $nObj 
     * @returns {Boolean}
     */
    __ipv6Validation=function($nObj) {
        console.log("(__ipv6Validation) to do!");
        return true;
    }


    /**
     * Validate an email string conforms to known email strings.
     * 
     * @param {anne_neuron} - The neuron with .value OR string we want to test
     * @author {Mr.Scatman} 
     *
     */
    __emailValidation = function($nObj) {
        var mailformat = /(?!.*\.{2})^([a-z\d!#$%&'*+\-\/=?^_`{|}~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+(\.[a-z\d!#$%&'*+\-\/=?^_`{|}~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+)*|"((([ \t]*\r\n)?[ \t]+)?([\x01-\x08\x0b\x0c\x0e-\x1f\x7f\x21\x23-\x5b\x5d-\x7e\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|\\[\x01-\x09\x0b\x0c\x0d-\x7f\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))*(([ \t]*\r\n)?[ \t]+)?")@(([a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|[a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF][a-z\d\-._~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]*[a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])\.)+([a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|[a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF][a-z\d\-._~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]*[a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])\.?$/i
        
        //just a string (not an nObj);
        if (typeof $nObj === "string") {
            var test = mailformat.test($nObj.toLowerCase());
            //console.log(test);
            if(test==undefined || test==null) return "invalid format";
            else if (test==true) return true;
            else return false;
        }

        if ($nObj.value==undefined || $nObj.value==null || $nObj.value=="") {
            $nObj.value = $nObj.text.trim();    //default nObj .value is the .text if none was provided by a .set_value function.
        }
        if(mailformat.test($nObj.value.toLowerCase())==true) {
            $nObj.valid = true;
            return true;
        }
        else {
            $nObj.value = "";
            $nObj.valid = false;
            return false;
        }
    }

    //from devshed
    __urlValidation=function($str) {
        var pattern = new RegExp('^(https?:\\/\\/)?'+ // protocol
        '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|'+ // domain name
        '((\\d{1,3}\\.){3}\\d{1,3}))'+ // OR ip (v4) address
        '(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*'+ // port and path
        '(\\?[;&a-z\\d%_.~+=-]*)?'+ // query string
        '(\\#[-a-z\\d_]*)?$','i'); // fragment locator
      return !!pattern.test($str);
    }

    
    /**
     * 
     * @param {*} $aMatches 
     * @param {*} $aMatches1 
     * @param {*} $prop 
     * @returns 
     */
    concatUnique=function($aMatches, $aMatches1, $prop="nid") {
        if($aMatches.length==0) return $aMatches1;
        if($aMatches1.length==0) return $aMatches;
        var arr = new arraymap();
        arr.setMapKey($prop);
        arr.addArray($aMatches);
        arr.addArray($aMatches1);
        return arr.arr;
    }


    uniqueConcat=function($a, $b, $prop="nid") {
        if ($a.length==0) return $b.concat();
        var a2 = $a.concat();
        for (var b=0; b < $b.length; b++) { 
            if ($a.indexOf($b[b])==-1) a2.push($b[b]);
        }
        return a2;
    }

    uniqueConcatRels = function($a, $b) {
        var a_rels_txids = [];
        var rels = [];
        for (var a=0; a < $a.length; a++) {
            a_rels_txids.push($a[a].txid);
            rels.push($a[a]);
        }
        for (var b=0; b < $b.length; b++) {
            if (a_rels_txids.indexOf($b[b].txid)!=-1) continue;
            a_rels_txids.push($b[b].txid);
            rels.push($b[b]);
        }
        return rels;
    }

    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    poorMansAbbreviate=function($str) {
        $str = $str.trim();
        const words = $str.split(' ');
        
        if (words.length === 1) {
            return $str; 
        } 
        else {
            var abbr = '';    
            words.forEach((word, index) => {
                if (index === 0 || index === words.length - 1) {
                    abbr += word.charAt(0).toUpperCase() + '.';
                } 
                else {
                    abbr += word.charAt(0).toUpperCase();
                }
            });
            return abbr;
        }
    }

    
    /**
     * Check for a relon in a relon array.
     * If it is unique, add it to the array and return true.
     * if it is not unique, return false.
     * @param {array} $rels 
     * @param {Object} $rel 
     * @returns 
     */
    pushIfUniqueRel = function($rels, $rel) {
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if(rel.txid==$rel.txid) {
                return false;
            }
        }
        $rels.push($rel);
        return true;
    }

    seededRandom=function() {
        var x = Math.sin(Math.random()*1000) * 10000;
        return x - Math.floor(x);
    }

    getScatModSeedArrItem=function($seed, $arr) {
        var per = UTILS.getScatModSeedPer($seed);
        var max_usable = Math.round($arr.length * per);
        return $arr [ Math.floor(Math.random()*max_usable)];
    }
    

    getScatModSeedPer=function($seed) {
        $seed = Math.round($seed);
        var per = .5;
        if ($seed % 3 !=0) per +.1;
        if ($seed % 2 !=0) per +.15;
        if ($seed % 7 ==0) per +.25;
        return per;
    }
    
    isWholeNumber=function($number) {
        return $number % 1 === 0;
    }

    HEXtoRGB = function($HEX) {
        var isNumb = Number.isInteger($HEX);
        if (isNumb) var bigint = $HEX;
        else var bigint = parseInt($HEX.slice(1), 16);
        const r = (bigint >> 16) & 255;
        const g = (bigint >> 8) & 255;
        const b = bigint & 255;
        return { r, g, b };
    }

    RGBtoHEX = function($rgb) {
        const { r, g, b } = $rgb;
        const hex = `0x${((1 << 24) + (r << 16) + (g << 8) + b).toString(16).slice(1)}`;
        return hex;
    };

    shiftRGB = function($RGB, $incr) {
        const shiftedColor = {
          r: $RGB.r + $incr,
          g: $RGB.g + $incr,
          b: $RGB.b + $incr,
        };
        shiftedColor.r = Math.max(0, Math.min(255, shiftedColor.r));
        shiftedColor.g = Math.max(0, Math.min(255, shiftedColor.g));
        shiftedColor.b = Math.max(0, Math.min(255, shiftedColor.b));
        return shiftedColor;
    };
}


const UTILS = new anne_utils ();
UTILS.set_init(UTILS);

/************** (from file ./js/base/utils.js)**/ 


class anne_babytalk {
    constructor($output, $order, $part_of_anne, $tense_lbl="") {
        this.output = $output;
        this.order = $order;
        this.poa = $part_of_anne;
        this.tense_lbl = $tense_lbl;
    }
}

class v1actionx {
    constructor($nid){
        this.nid=$nid;
        this.v2 = null;
        this.v3 = null;
        this.v4 = null;
        this.v5 = null;
        
         //inherited TYPE_ROLE assigned to any action vb parent in the BE line.
        //this is the "meaning" of the action, in full, with def of all BEM line roles
        this._meaning_roles = null;

        //roles this action assigns.
        //This will be stored on both inst and class action neurons.
        //It is inherited from the BEM line.
        this._param_roles = null;
    }
}

class clump_lookup {
    constructor() {
        this.nObj = null;
        this.parent = null;
        this._non_action_roles = [];
        this._rolesMap = new Object();
        this._ROLES_BY_NAME = [];
    }

    addRole=function($roleNid, $roleValueNid) {
        var ROLE_NOBJ = ANNE.MEMG.getNeuronByNid($roleNid);

        //the clump will instantiate this neuron.
        if (ROLE_NOBJ.nid==ANNE.MEM.role_action.nid) {
            this.parent = ANNE.MEMG.getNeuronByNid($roleValueNid);
        }

        //do not put action in here.
        else this._non_action_roles.push({role:$roleNid, role_value:$roleValueNid}) 
        
        if (this._rolesMap[$roleNid]==undefined) {
            this._rolesMap[$roleNid]=new arraymap();
            this._rolesMap[$roleNid].setMapKey("nid");
            this._ROLES_BY_NAME[ROLE_NOBJ.text.toUpperCase()] = [];
        }
        var roleValNOBJ = MEMG.getNeuronByNid($roleValueNid);
        this._rolesMap[$roleNid].addU(roleValNOBJ);
        this._ROLES_BY_NAME[ROLE_NOBJ.text.toUpperCase()].push(roleValNOBJ);
    }

    testNeuronForRoles = function($clumpNOBJ, $roles) {
        if ($clumpNOBJ==null) {
            console.warn("fix this, no null clump here");
            return false;
        }
        
        if ($roles==undefined || $roles==null) {
            console.warn("fix this, no null/undefined roles here");
            return false;
        }
        
        for (var r=0; r < $roles.length; r++) {
            var role = $roles[r];
            var roleNOBJ = ANNE.MEMG.getNeuronByNid(role.role);
            var valueNOBJ = ANNE.MEMG.getNeuronByNid(role.role_value);
            console.log("testing '"+$clumpNOBJ.text+"' for role ["+roleNOBJ.text.toLowerCase()+"] value "+valueNOBJ.text);
            var bGOOD = ANNE.MEMG.hasRoleRelon($clumpNOBJ.outs, role.role, role.role_value);
            if (bGOOD==false) {
                console.log("DO NOT HAVE THIS ROLE to value:"+valueNOBJ.nid+" ::: not a match");
                return false;
            }
        }
        return true;
    }


    //with a proper action, fidn the clump.
    findClump=async($bMAKE=true)=> {
        if (this.parent==null) {
            return await this.findClumpPartial();
        }

        //quick local test (we may have a ghost and we don't want dupes since
        //it won't be yet found on the node)
        await ANNE.IDB.queryForIns(null, this.parent.nid);
        
        var matches = [];
        var candidates = [];
        var pull_froms = [];
        
        for (var r=0; r < this.parent.ins.length; r++) {

            //action role OR parent of clump. is what we are looking for here.
            if (this.parent.ins[r].type!=TYPE_ROLE && this.parent.ins[r].type!=TYPE_BE) continue;  
            
            var clumpNOBJ = ANNE.MEMG.getNeuronByNid(this.parent.ins[r].from_nid);
            
            console.log("candidate match: "+clumpNOBJ.text);
            if (pull_froms.indexOf(clumpNOBJ.nid)==-1) {
                pull_froms.push(clumpNOBJ.nid);
                candidates.push(clumpNOBJ);
            }
        }

        //get all the roles so we can try to match the already-in-our-local-mem clump.
        if (pull_froms.length) {
            await ANNE.IDB.lookupNids(null, pull_froms);
        }

        for (var c=0; c < candidates.length; c++) {           
            if (this.testNeuronForRoles(candidates[c], this._non_action_roles)==true) {
                matches.push(candidates[c]);
            }
            else {
                console.log("NOT A MATCH");
            }
        }

        //no luck in mem, look it up in the ANNODE.
        if (matches.length==0) { //TEST THE API.
            var results = await ANNE.API.findClump(this.parent.nid, this._non_action_roles);   
        }
        else {
            var results = matches;
        }

        //handle results from annode (or local mem)
        if (results.length==1) {
            this.nObj = results[0];
            console.log(this.nObj.text);
            return {status:"good", match:[this.nObj]};
        }
        else if (results.length>1) {
            console.log("ambiguous results");
            this.nObj=null;
            return {status:"ambiguous", matches:results}
        }
        else if ($bMAKE==true) {
            this.setName();
            var ghostNOBJ = await GHOSTDATA.makeGhostNeuron(this.parent, this.name, NTYPE_INST);
            for (var r=0; r < this._non_action_roles.length; r++) {
                var role = this._non_action_roles[r];
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, ghostNOBJ, role.role, TO_TYPE_ANNE, role.role_value);
            }
            return {status:"good", match:ghostNOBJ};
        }

        return {status:"new", match:[]};
    }

    setName = function() {
        var parts = [];
        if (this._rolesMap[ANNE.MEM.role_actor.nid]!=undefined) {
            var subj = this._rolesMap[ANNE.MEM.role_actor.nid].arr[0];
            parts.push(subj.text);
        }
        parts.push(this.parent.text);

        //one of these is fine.
        if (this._rolesMap[ANNE.MEM.role_theme_thing.nid]!=undefined){
            var theme = this._rolesMap[ANNE.MEM.role_theme_thing.nid].arr[0];
            parts.push(theme.text);
        }
        else if (this._rolesMap[ANNE.MEM.role_posn_thing.nid]!=undefined) {
            var posn_thing = this._rolesMap[ANNE.MEM.role_posn_thing.nid].arr[0];
            parts.push(posn_thing.text);
        }
        else if (this._rolesMap[ANNE.MEM.role_be_thing.nid]!=undefined) {
            var role_be_thing = this._rolesMap[ANNE.MEM.role_be_thing.nid].arr[0];
            parts.push(role_be_thing.text);
        }

        if (this._rolesMap[ANNE.MEM.role_when.nid]!=undefined){
            var when = this._rolesMap[ANNE.MEM.role_when.nid].arr[0];
            parts.push(when.text);
        }
        if (this._rolesMap[ANNE.MEM.role_where.nid]!=undefined) {
            var where = this._rolesMap[ANNE.MEM.role_where.nid].arr[0];
            parts.push(where.text);
        }

        this.name = parts.join(" ");
        console.log(this.name);
    }


    findClumpPartial=async()=>{
        return;
    }
}




/**
    Deault garden - no delegation, no LET, no DENY.
    Defaults to restricting collab requests to PARAMs in the parent class.

    Communal garden - Honors owner set delegation to allow collab approvals.  
    Defaults to BLACKLIST mode for denys (ignores lets, all are let).
    Anyone can make collab relon requests.
    Delegated admins can auto make approved requests equal to their delegation rights
    or approved pending collab req equal or less than their granted rights.
    gui - Neurons owned by the owner or adopted will be displayed

    Owner collab - Honors owner set delegation to allow collab approvals.
    Defaults to WHITELIST mode for lets.  (ignores denies, all are default denied)
    Only those listed in lets can make collab relons.
    Delegated admins can auto make approved requests equal to their delegation rights
    or approved pending collab req equal or less than their granted rights.
    gui - Neurons owned by the owner or adopted will be displayed

    Owner Only - No delegation, no lets, no denys, no one can make collab relons,
    gui - Neurons owned by the owner will be displayed
*/
class anne_neuron {

    /**
     * @param {string} - default name of neuron
     * @param {int} - ntype int of neuron type
     * @param {string} - Long string of nid.
     */
    constructor ($text, $ntype=NTYPE_UNKNOWN, $nid="") {
        this.text = $text;
        this.short = null;      //the shortest be alt name, aka, abbreviation or text.

        this.ntype = $ntype;
        if (this.ntype==="") this.ntype=NTYPE_UNKNOWN;
        this.nid = $nid;
        this.be = ""; 
        this.be_lvl = -1;
        this.bNEW=false;
        this.dead = false;

        this.owned_by = "";
        this.sponsored_by = "";
        this.parent = "";  //same as bes[0]
        this.annother = null;
        this.polarity = null;
        this.origin = null; //important to know for key'd vs not key'd.

        //allocated arrays
        this.bes = [];
  
        this.insMAP = new arraymap();
        this.insMAP.setMapKey("txid");
        this.ins = this.insMAP.arr;      //map the ins ptr to the map.arr 

        this.outsMAP = new arraymap();
        this.outsMAP.setMapKey("txid");
        this.outs = this.outsMAP.arr;    //map the outs ptr to the map.arr

        this.children = [];
        
        //assigned TYPE_ROLE relons for any ation vb instance neuron.
        //this is a unique semantic sentence/thought
        //All sentences/statements/thoughts/clumps are store by instantiating the 
        //action vb which they are using.
        //If there are dual action vbs.  (he punched and kicked the man) we can either
        //create a new compound action neuron "punched and kicked"
        //OR we can create two clumps "he punched the man" -> AND -> "he kicked the man".
        //Both semantically work.

        //for classes this is the specific meaning from THIS neuron (not inherited meanings)
        this._roles = null;

        //inherited TYPE_ROLE assigned to any action vb parent in the BE line.
        //this is the "meaning" of the action, in full, with def of all BEM line roles
        this._meaning_roles = null;

        //roles this action assigns.
        //This will be stored on both inst and class action neurons.
        //It is inherited from the BEM line.
        this._param_roles = null;

        this.be_time = -1; //the time of the be relon (used to sort new neurons)

        //entire feelz summary w/ APLEX status
        this._napl = null;

        //this is a TYPE_SYS relon 
        //we inherit gt from classes down to subclasses.
        //it is only relevant to classes
        this._gt = null;
        
        //this is a TYPE_SYS relon
        //we inherit gti from classes down through subclasses down to instances
        //it is only relevant to instances
        this._gti = null;

        //don't allocate for these until we need them
        //even an empty array hits us bigly.
        this._wv_let = null;          //wvs that are let to collab (relevant in WHITELIST owner-collab garden)
        this._wv_deny = null;         //wvs that are denied to collab (relevant in BLACKLIST communual garden)
        this._vip_let = null;         //misc context (like VIP neurons)
        this.trusted_adopt = null;    //relevant in all gardens.
        this.abandon = null;          //will delete at protocol layer.  Array relevant for session.
        this.delegation = null;       //relevant to all keyless neurons with garden of owner-collab or communual
        this.ap = null;

        this.inhibit_txs = null;
        this.owns = null;
        this.aka = [];

        this.no_new_children = null;
        this.classes_only = null;
        this.insts_only = null;
        
        //used on the logged-in identity to tally nids (not tnids) we have delgation
        //rights to.  We use this to do the andy query to setup the andy area of ANNE
        //for managing collab requests.
        this.andy_nids = null;          

        //only relevant for identities, for US (logged in) and anyone we chase for this purpose
        //like in a profile view
        this.my_sponsored = null;  

        this.PARAM_NAME = null;
        this.REL_PARAM_NAME = null;

        //bools
        this.emulated = false
        this.bMINE = false;
        this.preloaded = false;
        this.temp = false;
        this.ghost = false;

        if (this.nid=="" || (this.nid.length && this.nid[0]=="z")) {
            this.ghost=true;
            this.mine=true; //ghosts that are not ghost_owner=USER are not processed.
        }

        this.bTERMINATING_TYPE = false;

        this.requires_validation = false;
        this.bISIDENTITY=null;

        this.ai_ico = null;
        
        //inheritable sys functions.
        this.set_value = null;
        this.set_value2 = null;
        this.formatter = null;
        this.format_validator = null;
        this.requires_validation = null;
        this.fire_circuit = null;

        this.temp_nid=-1;

        this.value = null;

        //collab/delegation, only set when encountered.
        this.collab_ACTION = null;      // COLLAB_REQ_OK etc.
        this.egrs_tnids = null;         // extra garden relns

        //hack prop for queue rollbacks of TYPE_BE MEM.name sends.
        //we should consider putting this elsewhere.
        this.rollbackText = null;   

        //json data for pos/neg/neu summary last_sell_price and annnies_in
        this._aplex = null;

        this.SYS_ALIAS = null;
        this.SYS_ENUM_VAL = null;
        this.SYS_CODE= null;

        //see country, state, city, town, township, borrough, neighborhood. etc
        //all are sub-classes of POLITY. all have children, however, which are 
        //instances of LOCATION.
        this.functional_childparent_reln = null;
        this.functional_childparent_parent = null; 
        this.functional_be = null;                  //for children with a functional parent

        this.freeze = NEURON_NOT_FROZE;

        this._anndle = null;
        this._anndle_resolution = null;
        
        this.ec_type_tnid = null;    //time saver if we calc it, save it.

        this.latestNameRelon =  null;
        this.adate = null;              //for neurons as dates
        this._liferange = null;         //for neurons with liferanges (company/corp, person, nation etc)

        if (MEMP!=undefined) MEMP.__hashNeuron(this);
    }


    setValue = function($v) {
        if (isNaN($v)==true) {
            console.log(this);
            console.log(this.text);
            alert("tell mr scatman about this bullocks");
            return;
        }
        this.value = $v;
    }

    getECTop=function() {
        return this.getParentOfLvl(2);
    }


    getTextWithWikiContext=function() {
        if (this.text=="") return "";
        var txt = this.text;
        var context = [];
        var parent = this.getParent();
        if (parent!=null && parent.text!="") context.push(parent.text);
        for (var b=0; b < this.outs.length; b++) {
            if (this.outs[b].type!=TYPE_BE) continue;
            var rel = this.outs[b];
            if (rel.reln==ANNE.MEM.class.nid && rel.reln==ANNE.MEM.inst.nid) continue;
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            if (rel.reln!=ANNE.MEM.ec_is.nid) continue;
            var toNOBJ = ANNE.MEMG.getNeuronByNid(rel.to_s);
            if (toNOBJ!=null && toNOBJ.text!="") context.push(toNOBJ.text);
        }
        if (context.length==0) return txt;

        var namewithcontext = txt + "("+context.join(", ")+")";
        ANNE.TALK.printf_console(ANNE.TALK.GREEN, namewithcontext);
        return namewithcontext;
    }


    //this is same as getTextWithWikiContext but it prints class/parent and this context.
    getParentTextWithChildContext = function() {
        var context = [];
        var parent = this.getParent();
        var ptxt = "";
        if (parent!=null && parent.text!="") ptxt = parent.text; 
        for (var b=0; b < this.outs.length; b++) {
            if (this.outs[b].type!=TYPE_BE) continue;
            var rel = this.outs[b];
            if (rel.reln==ANNE.MEM.class.nid && rel.reln==ANNE.MEM.inst.nid) continue;
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            if (rel.reln!=ANNE.MEM.ec_is.nid) continue;
            var toNOBJ = ANNE.MEMG.getNeuronByNid(rel.to_s);
            if (toNOBJ!=null && toNOBJ.text!="") context.push(toNOBJ.text);
        }
        if (context.length==0) return ptxt;

        var pnamewithcontext = ptxt + "("+context.join(", ")+")";
        ANNE.TALK.printf_console(ANNE.TALK.GREEN, pnamewithcontext);
        return pnamewithcontext;
    }


    /**
     * Go up the be tree until a specific lvl.
     * @param {*} $lvl 
     * @returns 
     */
    getParentOfLvl=function($lvl) {
        if (this.be_lvl>$lvl) {
            var parent = this.getParent();
            if (parent==null) return null;
            return parent.getParentOfLvl($lvl);
        }
        else if (this.be_lvl==$lvl) return this;
        return null;
    }

    getConstituentOf=function() {
        var rel = ANNE.MEMG.getBeRelon(this.outs, ANNE.MEM.constituent_of.temp_nid);
        if (rel!=null) return ANNE.MEMG.getNeuronByTempNid(rel.t);
        return null;
    }

    hasConstitParentOf=function($nid) {
        var rel = ANNE.MEMG.getBeRelon(this.outs, ANNE.MEM.constituent_of.temp_nid);
        if (rel==null) return null;
        if (rel.to_s==$nid) return true;
    }


    /**
     * Optimized to calc once, then store.
     * Stores ec_type_tnid
     * This type is crucial for FIRST-PASS-CONCEPTUALIZATION.
     * The types we handle are all the types used in communication.
     * 
     * NOTE: this should match static nlp_roles.getRoleFromECTop
     * 
     * @returns {anne_neuron}
     */
    setECConceptualizingParent=function() {
        if (this.dead==true) return;
        if (this.ec_type_tnid!=null) {
            return this.ec_type_tnid;
        }

        //nouns
        if (this.anyParentHasBeOf(MEM.ec_thing.temp_nid)==true) this.ec_type_tnid = MEM.ec_thing.temp_nid;
        
        //adjs
        else if (this.anyParentHasBeOf(MEM.ec_state.temp_nid)==true) this.ec_type_tnid = MEM.ec_state.temp_nid;
        
        //vbs
        else if (this.anyParentHasBeOf(MEM.ec_action.temp_nid)==true) this.ec_type_tnid = MEM.ec_action.temp_nid;
        
        //advs
        else if (this.anyParentHasBeOf(MEM.ec_manner.temp_nid)==true) this.ec_type_tnid = MEM.ec_manner.temp_nid;
        
        //modals
        else if (this.anyParentHasBeOf(MEM.ec_mode.temp_nid)==true) this.ec_type_tnid = MEM.ec_mode.temp_nid;
        
        //be vbs
        else if (this.anyParentHasBeOf(MEM.ec_is.temp_nid)==true) this.ec_type_tnid = MEM.ec_is.temp_nid;
        
        //posn vbs
        else if (this.anyParentHasBeOf(MEM.ec_posn.temp_nid)==true) this.ec_type_tnid = MEM.ec_posn.temp_nid;
        
        //when signals
        else if (this.anyParentHasBeOf(MEM.ec_expect_when.temp_nid)==true) this.ec_type_tnid = MEM.ec_expect_when.temp_nid;
        
        //where signals
        else if (this.anyParentHasBeOf(MEM.ec_expect_where.temp_nid)==true) this.ec_type_tnid = MEM.ec_expect_where.temp_nid;
        
        //cjs
        else if (this.anyParentHasBeOf(MEM.ec_expect_more_of_prev.temp_nid)==true) this.ec_type_tnid = MEM.ec_expect_more_of_prev.temp_nid;
        
        //infinitive
        else if (this.anyParentHasBeOf(MEM.ec_expect_actionthing.temp_nid)==true) this.ec_type_tnid = MEM.ec_expect_actionthing.temp_nid;
        
        //question wds
        else if (this.anyParentHasBeOf(MEM.ec_expect_question.temp_nid)==true) this.ec_type_tnid = MEM.ec_expect_question.temp_nid;
        
        //aux
        else if (this.anyParentHasBeOf(MEM.ec_expect_action.temp_nid)==true) this.ec_type_tnid = MEM.ec_expect_action.temp_nid;
        
        //preps and other sig wds
        else if (this.anyParentHasBeOf(MEM.ec_expection.temp_nid)==true) this.ec_type_tnid = MEM.ec_expection.temp_nid;
        
        //pns, countables etc
        else if (this.anyParentHasBeOf(MEM.ec_resolve.temp_nid)==true) this.ec_type_tnid = MEM.ec_resolve.temp_nid;
        
        //anne sys classes are things.
        else if (this.anyParentHasBeOf(MEM.__ANNE_SYS_CLASS.temp_nid)==true) {
            this.ec_type_tnid = MEM.ec_thing.temp_nid;
        }

        if (this.ec_type_tnid==null) {
            console.log("%c NO EC TYPE FOR NEURON '"+this.text+"' "+this.nid, TALK.bgred);
            this.printParentLine();
            return null;
        }

        //if ($matchObj!=null) $matchObj.ec_type_tnid = this.ec_type_tnid;
        return this.ec_type_tnid;
    }

    //TODO convert this to an array of annothers.
    //even though the NARR in ANNODE only quick indexes one,
    //we will still get ANNOTHER relons for multiple relons..
    //we just won't get them in quick indexes.  We will have to go look for them.
    /** 
     * 
     * @author {Mr.Scatman} 
     */
    addAnother=function($annother) {
        if (this.dead==true) return;
        this.annother = $annother;
    }


    /** 
     * This is adopting someone else's neuron.
     * You do not need their permission but they can change it, at will.
     * It is best to figure out which worldview will be the resource and get the 
     * neurons into that worldview with a HANAI_IN or HANAI_OUT
     * @author {Mr.Scatman} 
     */
    addTrustedAdopt=function($temp_nid) {
        if (this.dead==true) return;
        if (this.trusted_adopt==null) this.trusted_adopt = [$temp_nid];
        else if (this.trusted_adopt.indexOf($temp_nid)==-1) this.trusted_adopt.push($temp_nid);
    }


    /** 
     * 
     * @author {Mr.Scatman} 
     */
    addAbandom=function() {
        if (this.dead==true) return;
        if (this.abandon==null) this.abandon = [$temp_nid];
        else if (this.abandon.indexOf($temp_nid)==-1) this.abandon.push($temp_nid);
    }


    /**
     * Add a non-dag (not a inst/class reln) BE to a neuron.
     * 
     * NON-CLASS/INST BES DO NOT GO INTO THEIR BE's CHILD ARRAY.
     * THAT ARRAY IS RESERVED FOR LEGIT ANNE PARENT/CHILDREN. 
     * THE .bes array is for ANY BE relationship with any reln, but the [0] 
     * slot in the bes array is reserved for the class/be RELON! 
     * 
     * DONE
     * 
     * @author {Mr.Scatman} 
     * 
     */
     addAltBE=function($rObj) {

        if (this.dead==true) return;
        //do not add be->eval relons to the bes
        //this should be sys or has anyhow!
        if ($rObj.r==MEM.eval.temp_nid) return;

        var bes = this.bes;
        if (bes==undefined) {
            this.bes = [];
            this.be="";
        }

        if ($rObj.to_type==TO_TYPE_ANNE) {    
            this["be_"+$rObj.r] = $rObj.t;  //quick be tracker for direct reference
            
            //the children of country are actually polity relns pointing to country.
            //country->be_child->polity
            if ($rObj.r==MEM.BE_CHILDREN.temp_nid) {
                this.functional_childparent_reln=$rObj.t;
            }

            //the children of country with polity relns pointing to country
            // must be instances of BE_PARENT reln: to.  (in this case, location.)
            //country->be_parent->location
            else if ($rObj.r==MEM.BE_PARENT.temp_nid) {
                this.functional_childparent_parent=$rObj.t; //this is ACTUAL parent class
            }

            if (this.bes.indexOf($rObj.t)!=-1) return;
            
            if (this.bes.length && $rObj.t==MEM.class.temp_nid) return; 

            else if (this.bes.length>0 && this.bes[0]==MEM.class.nid) {
                this.bes[0]=$rObj.t;  //hijack it.
                var matching_index = MEM.class.children.indexOf($rObj.t);
                if (matching_index!=-1) MEM.class.children.splice (matching_index,1); 
            }
            else {
                if (this.bes.length && ($rObj.r==MEM.class.temp_nid || $rObj.r==MEM.inst.temp_nid)) {
                    if (this.bes.indexOf($rObj.t)==-1) {
                        this.bes.unshift($rObj.t);
                        this.be_time = $rObj.time;
                    }
                }
                else if (this.bes.indexOf($rObj.t)==-1) {

                    this.bes.push($rObj.t);
                }
            }
        }
    }

    
    /** 
     * cascade wv_lets inline here in this function.
     * they do not cascade to instances.  Instances look at the parent class
     * to enforce the deny.
     * @author {Mr.Scatman} 
     */
    addLet=function($temp_nid) {
        if (this.dead==true) return;
        if (this._wv_let==null) this._wv_let=[$temp_nid];
        else if (this._wv_let.indexOf($temp_nid)==-1) this._wv_let.push($temp_nid);
        for (var c=0; c < this.children.length; c++) {
            var nObj = this.children[c];
            if (nObj.ntype!=NTYPE_CLASS) continue;
            if (nObj._wv_let==null) nObj.addLet($temp_nid);
        }
    }
    

    addVipLet=function($temp_nid) {
        if (this.dead==true) return;
        if (this._vip_let==null) this._vip_let=[$temp_nid];
        else if (this._vip_let.indexOf($temp_nid)==-1) this._vip_let.push($temp_nid);
    }


    /** 
     * cascade wv_denies inline here in this function.
     * they do not cascade to instances.  Instances look at the parent class
     * to enforce the deny.
     * @author {Mr.Scatman} 
     */
    addDeny=function($temp_nid) {
        if (this.dead==true) return;
        if (this._wv_deny==null) this._wv_deny=[$temp_nid];
        else if (this._wv_deny.indexOf($temp_nid)==-1) this._wv_deny.push($temp_nid);
        for (var c=0; c < this.children.length; c++) {
            var nObj = this.children[c];
            if (nObj.ntype!=NTYPE_CLASS) continue;
            if (nObj._wv_deny==null) nObj.addDeny($temp_nid);
        }
    }


    /** 
     * Process delegated rights for a neuron.
     * 
     * Delegation clump relons contain delegation param rights.  These are specific columns
     * that the owner authorizes the delegated owner to make or approve of at their own discretion
     * without requiring workflow/approval from the author.  The delegated rights ARE the approval.
     * 
     * A full semantic clump is required since there are more than the 3 short-hand roles
     * required to store the params w/ respect to both a nid and an identity.
     * These clumps follow the form:
     * 
     * [what_role] - the nid which is providing the rights
     * [action_role] - the rights being granted (in our case this will only be the permits reln right)
     * [obj_who_role] - the identity who is getting the rights
     * [obj_what_role] - the param (relns/columns) which are permitted.  Multiples are 
     * ok here -- or a new clump can be made for each individual column.  
     *
     * Note:  we can model any sentence in any language using semantic roles.
     * In ANNE we want to define ONLY THE MOST BASIC ROLES of who/what/when/where. etc.
     * and the attempt (a lofty one) is to learn more refined roles through observation 
     * and experience of language (in any language).   Roles such as target, receiver, giver, instrument etc are hotly debated
     * in linguistic circles.  We can avoid all such debates by starting only with a minimal base
     * (since we don't have language acquisition from zero point we must have "a" base of something).  
     * In ANNE we have the "early concepts" neurons which stand in the place of foundational "pegs" for which
     * all alive entities acquire via experience...  which all other learnings use as a tangible reference..
     * (avoding the relative definition problem where everything is defined by something that is defined
     * by something else).
     * 
     * From a minimal "early concepts" base we can
     * achieve a learning/experience based circuit building model --- the best we can.
     * by programming the least possible and allowing the experiences to program.
     *
     * We can push the line back further wo what the ANNE SI learns via experience as we 
     * gain access to more ANNE sensory input channels.  (text, sound, IP, bytestreams. etc)
     *  
     * The R&D side of ANNE, the ANNE SI... builds her brain by using her brain.
     * The semantic side stores semantic information without having to "understand" what it actually is.
     * (even with that we can pull of more impressive AI feats than the current market does with 
     * their smoke-n-mirror charrades).
     * 
     * THING->delegationrights->to some identity
     * 
     * @author {Mr.Scatman} 
     */
     addDelegation=function($rObj) {
        if (this.dead==true) return;
        if (this.delegation==null) this.delegation=[];
        for (var d=0; d<this.delegation.length; d++) {
            if (this.delegation[d].txid==$rObj.txid) return;
        }
        this.delegation.push($rObj);
    }

    /**
     * we only track these for the logged in user.
     */
    addAndy=function($nid) {
        if (this.dead==true) return;
        if (this.andy_nids==null) this.andy_nids = [];
        if (this.andy_nids.indexOf($nid)==-1) {
            this.andy_nids.push($nid); 
        }
    }


    resetAndyNids=function() {    
        if (this.dead==true) return;
        this.andy_nids= [];
        for (var i=0; i < this.ins.length; i++) {
            var rObj = this.ins[i];
            if (rObj.type!=TYPE_SYS) continue;
            if (this.andy_nids.indexOf(rObj.from_nid)==-1) {
                this.andy_nids.push(rObj.from_nid);
            }
        }

        return this.andy_nids;
    }


    addAka = function($rObj) {
        if (this.dead==true) return;
        var akaval = "";
        if ($rObj.to_type==TO_TYPE_ANNE) {
            var akaNeuron = MEMG.getNeuronByTepNid($rObj.t);
            if (akaNeuron!=null) {
                akaval = akaNeuron.to_s;   
            }                     
        } 
        else akaval = $rObj.to_s;

        if (this.aka.length==0) {
            this.aka = [akaval];
            this.aka_lower = [akaval.toLowerCase()];
        }
        else {
            if (this.aka.indexOf(akaval)==-1) {
                this.aka.push(akaval);
                this.aka_lower.push(akaval.toLowerCase());
            }
        }

        //dont set short by other lang akas.  just standard aka. 
        if ($rObj.r==MEM.aka.temp_nid && akaval.length && akaval.length < this.text.length) {
            if (this.short==null || (this.short!=undefined && akaval.length < this.short.length)) this.short = akaval;
        }
    }

    anyNamesEqual = function($name) {
        $name = $name.toLowerCase();
        if (this.aka_lower!=undefined) {
            if (this.aka_lower.indexOf($name)!=-1) return true;
        }
        if ($name.toLowerCase()==this.text.toLowerCase()) return true;
        return false;
    }

    addAkaManual = function($aka) {
        if (this.aka.length==0) {
            this.aka = [$aka];
            this.aka_lower = [$aka.toLowerCase()];
            if ($aka.length<this.short) this.short=$aka;
        }
        else {
            if (this.aka.indexOf($aka)==-1) {
                this.aka.push($aka);
                if ($aka.length<this.short) this.short=$aka;
                this.aka_lower.push($aka.toLowerCase());
            }
        }
    }


    /** 
     * 
     * @author {Mr.Scatman} 
     */
    addGardenRule=function($rObj) {
        if (this.dead==true) return;
        if (this.egrs_tnids==null) this.egrs_tnids=[$rObj.t];
        else if (this.egrs_tnids.indexOf($rObj.t)===-1) {
            this.egrs_tnids.push($rObj.t);
        }
    }
   

     /** 
     * 
     * @author {Mr.Scatman} 
     */
    addOwnedBy=function($nid) {
        if (this.dead==true) return;
        
        if (this.owned_by=="") {
            this.owned_by = $nid; 
            var ob = MEMG.getOrMakeIfNew($nid);
            if (ob) ob.addOwns(this.temp_nid);
        }
        else {
            if (this.owned_by!=$nid) {
                //console.log("%cwe have a conflict in owned_by indexes", TALK.bgred);
                if (this.ghost==false) {
                    if (ANNE.isANNE(this.owned_by)==false && ANNE.isANNE($nid)==true) {
                        this.owned_by = $nid;
                    }
                }
                else {
                    if (ANNE.isANNE($nid)==false) {
                        this.owned_by = $nid;
                    }
                }
            }
            else {
                var ob = MEMG.getOrMakeIfNew($nid);
                if (ob) ob.addOwns(this.temp_nid);
            }
        }
        if (typeof USER!="undefined" && USER._userNObj!=null && this.owned_by==USER._userNObj.nid) this.mine=true;
    }

    addOwns=function($temp_nid) {
        if (this.dead==true) return;
        if (this.owns==null) this.owns=[];
        if (this.owns.length==0) this.owns.push($temp_nid);
        else if (this.owns.indexOf($temp_nid)===-1) this.owns.push($temp_nid);
    }


    /** 
     * 
     * @author {Mr.Scatman} 
     */
     addSponsoredBy=function($sponsored_by) {
        if (typeof ANNE.USER=="undefined" || ANNE.USER==null) return;
        if ($sponsored_by==null || $sponsored_by==undefined) {
            alert("fix this messed up sponsored by");
            this.sponsored_by="";
            return;
        }
        if (this.dead==true) return;               
        if (this.sponsored_by=="") ANNE.MEMG.getOrMakeIfNew($sponsored_by);  //this may be the only reference to this neuron we have..
        this.sponsored_by = $sponsored_by;
        
        if ($sponsored_by==ANNE.USER._userNObj.nid) {
            if (ANNE.USER._userNObj.my_sponsored==null) ANNE.USER._userNObj.my_sponsored=[this.temp_nid];
            else if (ANNE.USER._userNObj.my_sponsored.indexOf(this.temp_nid)==-1) {
                ANNE.USER._userNObj.my_sponsored.push(this.temp_nid);
            }
        }
    }


    /**
     * Only tracked for logged-in user.
     * @param {*} $relon_sponsor 
     * @returns 
     */
    addUsAsRelonSponsor=function($rel) {
        if ($rel==undefined || $rel==null) {
            alert('fix this');
            return;
        }
        if (this.dead==true) return;               
        if (this.relon_sponsored==undefined) {
            this.relon_sponsored=[$rel];
            return;
        }
        var txid = $rel.txid;
        for (var r=0; r < this.relon_sponsored.length; r++) {
            if (this.relon_sponsored[r].txid==txid) return;
        }
        this.relon_sponsored.push($rel);
    }
    

    
    /**
     * Add a unique be to the bes array.
     * 
     * This is called for all TYPE_BE relons.
     * Parent/hierarchical bes will be routed through setBeTempNid.
     * All other BEs are routed through the addAltBe handler.
     * 
     * @param {Object} - The TYPE_BE relon.
     * @author {Mr.Scatman} 
     */
     addUniqueBe = function($rObj) {
        if (this.dead==true) return;
        if ($rObj == null) return;
        if ($rObj.r == MEM.class.temp_nid) {
            if ($rObj.ntype!=1 && $rObj.ntype!=2) {
                this.ntype = NTYPE_CLASS;
            }
        }
        else if ($rObj.r == MEM.inst.temp_nid) {
            if ($rObj.ntype!=1 && $rObj.ntype!=2) {
                this.ntype = NTYPE_INST;
            }
            this.be_time = $rObj.time;
        }
        else {
            this.addAltBE($rObj);
            return;
        }
        var parentNOBJ = MEMG.getNeuronByTempNid($rObj.t);
        this.setBeTempNid(parentNOBJ.temp_nid);
    }

    /**
     * Add a unique Be.
     * @author {Mr.Scatman} 
     */
     addUniqueBe_fastIndex = function($parentNObj) {
        if (this.dead==true) return;
        if (this.temp_nid==-1) {
            alert("fix this (addUniqueBe_fastIndex) temp_nid = -1");
            return;
        }

        if ($parentNObj==null) {
            alert("What?"); 
            return;
        }
        if (($parentNObj instanceof anne_neuron) == false) {
            alert("you must pass a neuron into addUniqueBe_fastINdex");
        }

        var bes = this.bes;
        if (bes==null) {
            this.bes = [];
            this.be="";
        }

        //might be overkill as this is the nid of the temp_nid in bes[0]
        //the notation comes from the NARR record in the ANNODE.
        this.parent = $parentNObj.nid;  

        //now we know we need the array.
        if ($parentNObj.children==null) $parentNObj.children=[];

        //this["be_"+MEM.name.temp_nid] = $parentNObj.temp_nid;

        //we have a more refined be class, override it.
        if (this.bes.length && this.bes[0]==MEM.class.nid) {
            this.bes[0]=$parentNObj.temp_nid;
            
            //remove it from the children of the "class" neuron
            var matching_index = MEM.class.children.indexOf($parentNObj.temp_nid);
            if (matching_index!=-1) MEM.class.children.splice (matching_index,1); 
        }
        else {

            //class/inst hierarchical BES must be in the slot [0] of the bes array!
            if (this.bes.length && ($parentNObj.temp_nid==MEM.class.temp_nid || $parentNObj.temp_nid==MEM.inst.temp_nid)) {
                if (this.bes.indexOf($parentNObj.temp_nid)==-1) {
                    this.bes.unshift($parentNObj.temp_nid);
                }
            }
            else if (this.bes.indexOf($parentNObj.temp_nid)==-1) {
                //this be parent is a sub class of what we had before. this is more specific.
                this.bes.unshift($parentNObj.temp_nid);
            }
        }

        this.setBeLevel($parentNObj);
        if ($parentNObj.bTERMINATING_TYPE==true) {
            this.bTERMINATING_TYPE=true;
            //ensure any children are also bTERMINATING_TYPE just in case 
            //we preloaded or found them in an order that missed this.
            //ie.  ticker2->subclasses ticker1-> subclasses text.
            //if we see ticker2 first we don't get the BE info of it being a descendant of text.
            if ($parentNObj.children==null) $parentNObj.children=[];

            if ($parentNObj.children) {   //obviously should have them..
                for (var c=0; c <$parentNObj.children.length; c++) {
                    MEMG.getNeuronByTempNid($parentNObj.children[c]).bTERMINATING_TYPE=true;
                }
            }
        }
                
        if ($parentNObj.children.indexOf(this.temp_nid)==-1) $parentNObj.children.push(this.temp_nid);
        if ($parentNObj.text!="") this.be = $parentNObj.text;

        //fix lvls and text while we are at it.
        //iterate through be children.
        if ($parentNObj.children.length>0) { 
            if ($parentNObj.bes && $parentNObj.bes.length>0) {
                var beBeNOBJ = MEMG.getNeuronByTempNid($parentNObj.bes[0]);
                if (beBeNOBJ) $parentNObj.setBeLevel(beBeNOBJ);
            }
        }
    }


    /**
     * Add a unique in to a neurons ins array.
     * 
     * @param {Object} - the in
     * 
     * @author {Mr.Scatman}
     */
    addUniqueIns=function($rObj) {
        if (this.dead==true) return;
        if (this.ins==null) this.ins=[];
        else {
            if (this.insMAP.get($rObj.txid)!=null) return;

            this.insMAP.addU($rObj);
        }
        return;
    }

        
    /**
     * 
     * @author {Mr.Scatman}
     */
     addUniqueLocalArborPrune=function($rObj) {
        if (this.dead==true) return;
        if ($rObj.from_nid!=this.nid) {
            console.log("nid doesn't match. big trouble little china")
            return;
        }
        if ($rObj.reln!=MEM.arbor_prune.nid) return; //local arbor prune
        $rObj.active=false;
        var txid = $rObj.to_s.trim();
        if (this.ap==null) this.ap=[];
        else {
            for (var r=0; r < this.ap.length; r++) {
                var a = this.ap[r];
                var ap_txid = this.ap[r].trim();
                if (ap_txid == txid) return;
            }
        }
        this.ap.push($rObj.to_s);    //the TXID we want to prune!
    }



    
    /**
     * Add a unique out to a neurons outs array.
     * 
     * @param {Object} - the out
     * 
     * @author {Mr.Scatman}
     */
     addUniqueOut=function($rObj) {
        if (this.dead==true) return;
        if ($rObj.from_nid!=this.nid) {
            console.log("nid doesn't match. big trouble little china")
            return;
        }

        if (this.outs==null) this.outs = [];
        
        //this txid is already in our arbor prune list for THIS local neuron.
        //do not allow this txid in.
        if (this.ap!=null && this.ap.length) {
            if (this.ap.indexOf($rObj.txid.trim())>=0) {
                $rObj.active=false;
                return;
            }
        }
        if (this.inhibit_txs!=null && this.inhibit_txs.length) {
            if (this.inhibit_txs.indexOf($rObj.txid.trim())>=0) {
                $rObj.active=false;
                return;
            }
        }

        //this IS a new local arbor prune or inhibit
        if ($rObj.r==MEM.arbor_prune.temp_nid || $rObj.r==MEM.inhibit_relon.temp_nid) { 
            
            //locals only are allowed to prune TXIDS from their neuron
            var killObj = MEMG.lookupRelonByTXID($rObj.to_s, this.outs);
            
            //turn it off before it gets in any arrays (or else we have to go find it in post ops);
            if (killObj != null) {
                this.addUniqueLocalArborPrune($rObj);
                killObj.active = false;
                this.removeByTXID('outs', killObj.txid); //remove it if we already have it in 
            }
            return;
        }

        //this is a necrosis 
        if ($rObj.r==MEM.necrosis.temp_nid) {
            return;
        }

        if ($rObj.active==false) return;

        if (this.outsMAP.get($rObj.txid)!=null) return;

        this.outsMAP.addU($rObj);
    }


    anyParentHasBeOf=function($temp_nid) {
        if (this.dead==true) return false;
        var nObj = this;
        var safety=16;
        if (nObj.temp_nid==$temp_nid) return true;

        //var testToMatchNOBJ = MEMG.getNeuronByTempNid($temp_nid);
        //console.log("testing for '"+nObj.text+"' having any parent or w/ alt be of: '"+testToMatchNOBJ.text+"' nid:"+testToMatchNOBJ.nid);
        while (nObj!=null && nObj.be_lvl>0 && safety>0) {
            if(nObj.bes.indexOf($temp_nid)!=-1) {
                return true;
            }
            nObj = nObj.getParent();
            safety=safety-1;
            if (nObj==null) break;
        }
        return false;
    }


    //we can display disambigs in here.
    bestSelectDisplay = function($pre="", $bOWNEDBY=false, $bPARENT=false) {
        if (this.dead==true) return "";
        var parts = [];
        if ($bOWNEDBY==true) {
            var lbl = this.owned_by;
            if (this.owned_by==ANNE.MEM.ANNE_ACCTID) lbl = "ANNE";
            parts.push("owner:"+lbl+"");
        }
        if ($bPARENT==true && this.getParent()!=null) {
            parts.push("parent:"+this.getParent().text+"");
        }
        else {
           parts.push("parent:?");  
        }
        if (parts.length) return $pre+this.text+" ("+parts.join(" ")+")";
        return $pre+this.text;
    }


    uicoWithName = function() {
        var name = "";
        this._uico = undefined;

        if (this.text==undefined || this.text==null) this.text="";
        
        if (this.text.length) name = this.text;
        else name = this.nid;

        return name;
    }
    


    bestName = function($default="", $bPOSTFIXNID=false) {
        if (this.dead==true) return "";
        var prepend = "";
        if (this.ghost==true) prepend = '(gh)';
    
        if (this.bTERMINATING_TYPE==true) {
            if(this.value!=null && this.value!="") return prepend+this.value;
            else if (this.set_value!=null) this.set_value(this);
            if (this.value!=null && this.value!="") return prepend+this.value;
        }
        if (this.text!="") {
            if ($bPOSTFIXNID==false) return prepend+this.text;
            else return prepend + this.text+" - "+this.nid;
        }

        if ($default!="") return prepend+$default;
        //check if be-line is a role_action, if so put "actionevent: '[name of role]'"
        if (MEMG.besOf(this, MEM.ec_action.temp_nid)==true) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            return "actionevent: "+prepend+beNOBJ.text;
        }
        return prepend+this.nid;
    }
    

    //unlike shortName, this attempts to find the shortest alias 
    //to use in an alt-search of external data when we have no viable options found
    //on our first attempt.
    getShortestAlias = function() {
        if (this.aka==undefined || this.aka==null) return this.shortName;
        var shortest = this.aka[0];
        for (var a=1; a < this.aka.length; a++) {
            if (this.aka[a].length < shortest.length) {
                shortest = this.aka[a];
            }
        }
        return shortest;
    }


    /**
     * 
     * @param {*} $default 
     * @param {*} $bPOSTFIXNID 
     * @returns 
     */
    bestNameHTML = function($default="", $bPOSTFIXNID=false, $icobump="rem14") {
        if (this.dead==true) return "";
        var prepend = "";
        var post = "";
        if (this.ghost==true) prepend = '<img class="dib" src="./imgs/ghost_ico.png" width="12px" title="Ghost (local neuron).">&nbsp;';

        //distinguish image neurons with an img ico to avoid confusion.
        //if (this.bes.length && this.bes.indexOf(MEM.neuronImg.temp_nid)!=-1) post = "📷";

        var parent = this.getParent();

        if (this._bico!=undefined && this._bico!="") {
            post = "<i class='"+this._bico+" "+$icobump+"'></i>&nbsp;";
        }
        else if (parent!=null) {
            if (parent!=null && parent._bico!=undefined && parent._bico!="") {
                post = "<i class='"+parent._bico+" "+$icobump+"'></i>&nbsp;";
            }
            //else if (parent!=null && parent._uico!=undefined && parent._uico!="") {
                //post = parent._uico;
            //}
        }
        //else if (this._uico!=undefined && this._uico!="") {
            //post = this._uico;
        //}
       
        
        if (this.bTERMINATING_TYPE==true) {
            if(this.value!=null && this.value!=isNaN(this.value) && this.value!="") return prepend+this.value+post;
            else if (this.set_value!=null) this.set_value(this);
            if (this.value!=null && this.value!="") return prepend+post+this.value;
        }

        //use short name here, if we have it.
        //ie, for District of Columbia we can use DC.
        var str = this.text;
        if (str!="") {
            if ($bPOSTFIXNID==false) return post+prepend+str;
            else return prepend+post+str+" - "+this.nid;
        }

        if ($default!="") return prepend+$default;

        //check if be-line is a role_action, if so put "actionevent: '[name of role]'"
        if (MEMG.besOf(this, MEM.ec_action.temp_nid)==true) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            return "actionevent: "+prepend+post+beNOBJ.text;
        }

        return prepend+post+this.nid;
    }


    /**
     * Try a custom set_value function first.
     * If that doesn't exist, try for an eval relon.
     */ 
    bestValue = function() {
        if (this.dead==true) return null;
        if (this.set_value!=undefined && typeof this.set_value === 'function') {
            return this.set_value(this);
        }
        return MEM.eval.set_value(this);
    }

    getSpouses = function() {
        var spouse_rels = ANNE.MEMG.getHasRelons(this.outs, ANNE.MEM.spouse.temp_nid);
        if (spouse_rels==null || spouse_rels.length==0) return [];
        return ANNE.MEMG.getNObjsFromRels(spouse_rels);
    }

    //blow out the disambig however we like here..
    //setDisambigForLightContextResolution=function() {
    //}


    /**
     * Text search of a disambig list associated with this neuron in RAM.
     * Score it if there are any hits to the context array passed in.
     * @param {*} $context 
     * @returns 
     */
    contextAdditive=function($context=[]) {
        if ($context.length==0) return 0;
        
        var additive = 0;
        for (var d=0; d < this.disambig.length; d++) {
            if ($context.includes(this.disambig[d])==true) {
                additive+=1;
            }
        }
        return additive;
    }


    printDisambig=function($bSET=false, $max=3) {
        if ($bSET==true) this.setDisambigArrStr($max);
        return this.bestName()+"["+this.disambig.join(",")+"]";
    }

    /**
     * 
     */
    setDisambigArrStr = function($max=3, $skipsTNIDS=[]) {
        this.disambig = [];
        for (var r=0; r < this.outs.length; r++) {
            if (this.disambig.length==$max) {
                return;
            }
            var rel = this.outs[r];
            if ($skipsTNIDS.length>0 && $skipsTNIDS.indexOf(rel.r)!=-1) continue;
            if (rel.type!=TYPE_BE && rel.type!=TYPE_BRAIN && rel.type!=TYPE_HAS) continue;
            if (rel.type==TYPE_HAS && rel.r!=MEM.brain_assoc.temp_nid && rel.r!=MEM.brain_assoc_strong.temp_nid) continue;
            if (rel.r==ANNE.MEM.class.temp_nid) continue;
            if (rel.r==ANNE.MEM.inst.temp_nid) continue;
            if (rel.r==ANNE.MEM.birthplace.temp_nid) continue;
            if (rel.r==ANNE.MEM.birthday.temp_nid) continue;
            if (rel.r==ANNE.MEM.deathplace.temp_nid) continue;
            if (rel.r==ANNE.MEM.deathday.temp_nid) continue;
            if (rel.r==ANNE.MEM.nationality.temp_nid) continue;
            if (rel.r==ANNE.MEM.name.temp_nid) continue;
            if (rel.to_type==TO_TYPE_ANNE && rel.type==TYPE_BE) {
                var toNOBJ = MEMG.getNeuronByNid(rel.to_s);
                if (toNOBJ==null || toNOBJ.text==null) {
                    console.log("?????");
                    continue;
                }
                if (toNOBJ.tex==null || toNOBJ.text=="") continue;
                if (this.text==null || this.text==undefined) {
                    console.log("WHAT IN THE HECK???");
                }
                if (toNOBJ.text==undefined || toNOBJ.text==null) {
                    console.log("toNOBJ probs");
                }
                if (toNOBJ && toNOBJ.text && toNOBJ.bestName().toLowerCase()==this.bestName().toLowerCase()) continue;
                if (toNOBJ!=null && this.disambig.includes(toNOBJ.text.toLowerCase())==false) {
                    this.disambig.push(toNOBJ.text.toLowerCase());
                }
            }
            if (this.disambig.length==$max) {
                return;
            }
        }
        

        for (var r=0; r < this.outs.length; r++) {
            if (this.disambig.length==$max) {
                return;
            }
            var rel = this.outs[r];
            if ($skipsTNIDS.length>0 && $skipsTNIDS.indexOf(rel.r)!=-1) continue;
            if (rel.type!=TYPE_BE && rel.type!=TYPE_BRAIN) continue;
            if (rel.r==ANNE.MEM.class.temp_nid) continue;
            if (rel.r==ANNE.MEM.inst.temp_nid) continue;
            if (rel.r==ANNE.MEM.birthplace.temp_nid) continue;
            if (rel.r==ANNE.MEM.birthday.temp_nid) continue;
            if (rel.r==ANNE.MEM.deathplace.temp_nid) continue;
            if (rel.r==ANNE.MEM.deathday.temp_nid) continue;
            if (rel.r==ANNE.MEM.nationality.temp_nid) continue;
            if (rel.r==ANNE.MEM.name.temp_nid) continue;
            if (rel.to_type==TO_TYPE_STRING && rel.type==TYPE_BE) {
                if (rel.to_s.toLowerCase()==this.text.toLowerCase()) continue;
                if (this.disambig.includes(rel.to_s.toLowerCase())==false) this.disambig.push(rel.to_s.toLowerCase());
            }
            if (this.disambig.length==$max) {
                return;
            }
        }

        if (this.disambig.length==0) {
            var parentNOBJ = this.getParent();
            if (parentNOBJ!=null && parentNOBJ.text!="") this.disambig.push(parentNOBJ.text);
        }

        return;
    }



    //get a short name, if we have it,
    //otherwise fall back on the bestName.
    shortName = function() {
        if (this.dead==true) return "";
        if (this.short!=null && this.short!="") return this.short;
        else if (this.text!="") return this.text;
        else return this.nid;
    }

    
    /**
     * Cascade properties down to children.
     * 
     * NOTE: now that we use this heavily we will have to check
     * for circular paths in our TYPE_BE DAG structure.
     * 
     * This would cause the MEM manager to hang.
     *  
     * @author {Mr.Scatman} 
     */
    cascadePropsDownChildrenClasses=function($propsObj, $chain=[]) {
        ANNE.MEM._cascadePropsDownChildrenClasses_cnt+=1;

        if (this.dead==true) return;
        if (this.children==null) return;

        for (var n=0; n < this.children.length; n++) {
            var childNOBJ = MEMG.getNeuronByTempNid(this.children[n]);
            
            if (childNOBJ.bes.indexOf(this.temp_nid)==-1) {
                alert("this should never happen");
                continue;
            }
            
            //func inherits
            if (this.ntype==NTYPE_CLASS) {
                if (childNOBJ.set_value!=undefined && childNOBJ.set_value!=null) {
                    $propsObj["set_value"]=childNOBJ.set_value;
                }
                if (childNOBJ.set_value2!=undefined && childNOBJ.set_value2!=null) {
                    $propsObj["set_value2"]=childNOBJ.set_value2;
                }
                if (ANNE.UTILS.isset(childNOBJ.formatter==true)) {
                    console.log("%c"+childNOBJ.text+" has formatter to inherit", TALK.bggreen);
                    $propsObj["formatter"]=childNOBJ.formatter;
                    if (this.nid==MEM.string.nid || childNOBJ.nid==MEM.string.nid) {
                        console.log("%c (cascadePr0psDownChildrenClasses) hunting string bug", TALK.bggreen);
                    }
                }
                if (childNOBJ.format_validator!=undefined && childNOBJ.format_validator!=null) {
                    $propsObj["format_validator"]=childNOBJ.format_validator;
                    
                }
                if (childNOBJ.requires_validation!=undefined && childNOBJ.requires_validation!=null) {
                    $propsObj["requires_validation"]=childNOBJ.requires_validation;
                }
                if (childNOBJ.fire_circuit!=undefined && childNOBJ.fire_circuit!=null) {
                    $propsObj["fire_circuit"]=childNOBJ.fire_circuit;
                }

                //TODO:
                //not a func, but inheritable.  (CHECK THIS WITH OUR NEW CLUMP HANDLING)
                if (childNOBJ.role_action!=undefined && childNOBJ.role_action!=null) {
                    console.log("%c REVIEW THIS", TALK.bgred);
                    $propsObj["role_action"]=childNOBJ.role_action;
                }
                
                if (childNOBJ.target_type!=undefined && childNOBJ.target_type!=null) {
                    $propsObj["target_type"]=childNOBJ.target_type;
                }
            }
    
            var props = Object.keys($propsObj);

            for (var p=0; p < props.length; p++) {
                var prop = props[p];

                if (prop=="set_value" && $propsObj[prop].name!=undefined && $propsObj[prop].name=="setNumber") {
                    console.log(childNOBJ.text+" lvl:"+childNOBJ.be_lvl+" is inheriting set_value: setNumber!");
                }

                //only set the child prop with the parent prop if the child
                //prop isn't already set.
                if (childNOBJ[prop]==undefined || childNOBJ[prop]==null) {
                    childNOBJ[prop] = $propsObj[prop];//this[prop];       //this prop OR  $propsObj[prop] ?
                }

                //if we had a child prop already set, replace that in our props
                //array as we will cascade this value down to the next gen of children,
                //not the less accurate gparent value.
                else $propsObj[prop]=childNOBJ[prop];
            
            }

            //no need to chase inst children.
            //no such thing, other than our creative "MIRROR" neuron which is an
            //inst of an inst.. but we will handle that run-time on demand.
            if (childNOBJ.ntype==NTYPE_INST) continue;

            $chain.push(this.text+"->"+childNOBJ.text+" lvl:"+this.be_lvl);

            var propsObj = $propsObj
            var keys = Object.keys($propsObj);
            
            //we need a copy to traverse as we may change a value from another downline
            //and we don't want that in a higher branch.
            if (keys.length>0) {
                propsObj = [];
                for (var k=0; k < keys.length; k++) {
                    propsObj[ keys[k] ] = $propsObj[ keys[k] ];
                }     
            }
            
            //childNOBJ.cascadePropsDownChildrenClasses(propsObj, $chain);
        }
        return;
    }


    /**
     * The child has severed its ties with its parent.
     * @author {Mr.Scatman}
     */
    childDisavowsParent=function($parent_temp_nid) {
        if (this.dead==true) return;
        this.removeParentFromChild($parent_temp_nid);
        var parentNOBJ = MEMG.getNeuronByTempNid($parent_temp_nid);
        if (parentNOBJ==null) return;
        parentNOBJ.removeChild(this.temp_nid);
        return;
    }


    /** 
     * 
     * @author {Mr.Scatman} 
     */
     disambigIDHTMLStr($el) {
        if (this.dead==true) return;
        if ($el==undefined || $el==null) return;
        $el = document.getElementById($el);
        ANNE.A.empty($el);
        ANNE.A.show($el);
        var span = ANNE.A.ezEl($el, "span", "text-success-lbl", this.text);
        var pieces = 0;
        if (this.bes.length) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            if (beNOBJ.text!="") ANNE.MEM_UDF.makeNeuronLink($el, beNOBJ.nid, "", "<b>parent</b>&nbsp;");
            pieces+=1;
        }
        if (this.owned_by!="") {
            var ownedBYNOBJ = MEMG.getNeuronByNid(this.owned_by);
            if (ownedBYNOBJ) ANNE.MEM_UDF.makeNeuronLink($el, ownedBYNOBJ.nid, "", "<b>owner</b>&nbsp;");
            pieces+=1;
        }
        if (this.sponsored_by!="") {
            var sponsoredBYNOBJ = MEMG.getNeuronByNid(this.sponsored_by);
            if (sponsoredBYNOBJ) ANNE.MEM_UDF.makeNeuronLink($el, sponsoredBYNOBJ.nid, "", "<b>sponsor</b>&nbsp;") ;
            pieces+=1;
        }

        //to do, make sure aka full neurons are put in the aka array as text strings.
        if (this.aka!=null) {
            var span = document.createElement("span");
            span.innnerHTML = "also known as <b>"+this.aka.join("/")+"</b>";
            $el.append(span);
            pieces+=1;
        }

        if (this.pubKey!=undefined) {
            var span = document.createElement("span");
            span.innnerHTML = "pubkey <b>"+this.pubKey+"</b>";
            $el.append(span);
            pieces+=1;
        }

        if (pieces==0) {
            ANNE.A.ezEl($el, "span", "bg-danger roundify", "<b>WARNING</b>: this is an unnamed unclassified NID. This is treated as an onboarding.");
        }
    }


    /**
     * Using this in extension -- on send page to disambugate neuron
     * account we are sendign to.
     * @param {*} $el 
     * @param {*} $bSHOWNIDIFNOTEXT 
     * @returns 
     */
    disambigIDStr=async($el, $bSHOWNIDIFNOTEXT=true)=> {
        if (this.dead==true) return;
        if ($el==undefined || $el==null) return;
        $el = document.getElementById($el);
        ANNE.A.empty($el);
        ANNE.A.show($el);

        var pieces = 0;

        if (this.text=="" || this.bes.length==0) {
            await ANNE.API.getNeuronByNid(this.nid);
        }

        if (this.text!="") {
            var span = document.createElement("span");
            span.classList = "disambig-item";
            span.innerHTML = "<b>name</b><span class='disambig-value name'>"+this.text+"</span>";
            $el.append(span);  
        }
        else if ($bSHOWNIDIFNOTEXT==true) {
            var span = document.createElement("span");
            span.classList = "disambig-item";
            span.innerHTML = "<b>nid</b><span class='disambig-value nid'>"+this.nid+"</span>";
            $el.append(span);    
        }
        
        if (this.bes.length) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            if (beNOBJ.text!="") {
                var p0 = document.createElement("span");
                p0.classList = "disambig-item";
                p0.innerHTML = "<b>parent</b><span class='disambig-value'>"+beNOBJ.bestName()+"</span>";
                $el.appendChild(p0);
                pieces+=1;
            }
        }
        if (this.owned_by!="") {
            var ownedBYNOBJ = MEMG.getNeuronByNid(this.owned_by);
            if (ownedBYNOBJ!=null) {
                var p0 = document.createElement("span");
                p0.classList = "disambig-item";
                p0.innerHTML = "<b>owner</b><span class='disambig-value'>"+ownedBYNOBJ.bestName()+"</span>";
                $el.appendChild(p0);
                pieces+=1;
            }
        }
        if (this.sponsored_by!="") {
            var sponsoredBYNOBJ = MEMG.getNeuronByNid(this.sponsored_by);
            if (sponsoredBYNOBJ!=null) {
                var p0 = document.createElement("span");
                p0.classList = "disambig-item";
                p0.innerHTML = "<b>sponsor</b><span class='disambig-value'>"+sponsoredBYNOBJ.bestName()+"</span>";
                $el.appendChild(p0);
                pieces+=1;
            }
        }

        //to do, make sure aka full neurons are put in the aka array as text strings.
        if (this.aka!=null && this.aka.length>0) {
            var p0 = document.createElement("span");
            p0.classList = "disambig-item";
            p0.innerHTML = "<b>aka</b><span class='disambig-value'>"+this.aka.join("/")+"</span>";
            $el.appendChild(p0);
            pieces+=1;
        }

        if (pieces.length==0) {
            if (this.pubKey!=undefined) {
                var p0 = document.createElement("span");
                p0.classList = "disambig-item";
                p0.innerHTML = "<b>nid is real and exists onchain</b>";
                $el.appendChild(p0);
                pieces+=1;
            }
            else {
                var span2 = document.createElement("span");
                span2.innnerHTML = "<b>WARNING</b><span class='disambig-value'>this is an unnamed unclassified NID. This is treated as an onboarding.</span>";
                $el.append(span2);
            }
        }
    }


    /** 
     * Show disambig (with clickable links) or just text to help differentiate what this is.
     * @author {Mr.Scatman} 
     */
    disambigHTMLStr($el, $bSHOWNID=false, $bSHOWSELF=false, $selfPre="<b>selected&nbsp;</b", $lblClass="", $disambig_text_only=false) {
        if (this.dead==true) return;
        console.log(this);
        if ($el==undefined || $el==null) return;
        var el = ANNE.A.empty($el);

        if ($bSHOWSELF) {
            if ($lblClass!="") {
                var span = document.createElement("span");
                span.classList.add($lblClass);
                el.append(span);
                span.innerHTML=$selfPre;

                var span2 = document.createElement("span");
                el.append(span2);
                span2.classList="disambig-neuron";

                if ($disambig_text_only==true) span2.innerHTML+="["+this.text+"]";
                else EXT_BTNS.makeNeuronLink(span2, this.nid, "");
                el = span2; //for the rest of the additions.
                
            }
            else EXT_BTNS.makeNeuronLink(el, this.nid, "", $selfPre);
        }

        if (this.bes.length) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            if (beNOBJ.text!="") {
                 if ($disambig_text_only==true) el.innerHTML+="[<b>parent&nbsp;</b>"+beNOBJ.text+"]";
                 else EXT_BTNS.makeNeuronLink(el, beNOBJ.nid, "", "<b>parent&nbsp;</b>");
            }
        }
        if (this.owned_by!="") {
            var ownedBYNOBJ = MEMG.getNeuronByNid(this.owned_by);
            if (ownedBYNOBJ) {
                if ($disambig_text_only==true)el.innerHTML+="[<b>owner&nbsp;</b>"+ownedBYNOBJ.text+"]";
                else EXT_BTNS.makeNeuronLink(el, ownedBYNOBJ.nid, "", "<b>owner</b>&nbsp;");
            }
        }
        if (this.sponsored_by!="") {
            var sponsoredBYNOBJ = MEMG.getNeuronByNid(this.sponsored_by);
            if (sponsoredBYNOBJ) {
                if ($disambig_text_only==true)el.innerHTML+="[<b>sponsor&nbsp;</b>"+sponsoredBYNOBJ.text+"]";
                else EXT_BTNS.makeNeuronLink(el, sponsoredBYNOBJ.nid, "", "<b>sponsor</b>&nbsp;") ;
            }
        }
        if ($bSHOWNID==true) {
            var span = document.createElement("span");
            if ($disambig_text_only==true) span.innerHTML = "[<b class='show-nid'>nid</b> "+this.nid+"]";
            else span.innerHTML = "<b class='show-nid'>nid</b> "+this.nid;
            el.append(span);
        }

        //to do, make sure aka full neurons are put in the aka array as text strings.
        if (this.aka!=null) {
            var span = document.createElement("span");
            span.innnerHTML = "also known as <b>"+this.aka.join("/")+"</b>";
            el.append(span);
        }

        ANNE.A.show(el);
    }



    /** 
     * Ensure this child is a legit child of the class and not a rogue unsanctioned creation.
     * 
     * Reminder: (cheap) Adopt get a neuron made in another wordlview into the acceptable
     * worldview of the class.
     * 
     * @author {Mr.Scatman} 
     */


    //TODO: REDO THIS FOR  gt and gti
    enforceGardenUp=function() {
        if (this.dead==true) return;
        var parentNOBJ = null;
        if (this.bes.length==0) return true;
        if (this.mine==true) return true;       //anything we own is ok to show in our gui
        if (this._gt==null) {
            if (this.bes.length==0) return true;
            //quick check for parent _gt if we missed it.
            var parentNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            if (parentNOBJ==null) return true;
            if (parentNOBJ._gt!=null) this.setGtRelon(parentNOBJ._gt);
            else return true;
        }

        //Communal garden - Honors owner set delegation to allow collab approvals.  
        //Deny entry will OVERRIDE delegated rights
        //Defaults to BLACKLIST mode for denys (ignores those on the LET list).
        //Anyone can make collab relon requests.
        //Anyone can make keyless children (that they own)
        //gui - Neurons owned by the owner or anyone not in the deny list will be displayed.
        else if (this._gt.reln==MEM.gt_communal.nid) {
            //if ($nObj._wv_let!=undefined && $nObj._wv_let.length) {
               // if ($nObj._wv_let) MEMZ.ezDisplayfrom_tnidList($el, "<b>let</b>", $nObj._wv_let, "ai-check", $linkType);
           // }
            //if (this._wv_deny!=undefined && this._wv_deny.length) {
               // if (this._wv_deny
            //}
            return true;
        }
        
        
        //Owner collab - Honors owner set delegation to allow collab approvals.
        //Defaults to WHITELIST mode for LETS.
        //Only those listed in LETS can make collab relon request and keyless children (that they own).
        //gui - Only those neurons owned by the owner, or the LETS, will be displayed.
        else if (this._gt.reln==MEM.gt_owner_collab.nid) {
            return true;
        }

        //for owner only:
        parentNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
        if (parentNOBJ.owned_by==this.owned_by) return true;
        if (parentNOBJ.trusted_adopt!=null && parentNOBJ.trusted_adopt.length && parentNOBJ.trusted_adopt.indexOf(this.temp_nid)) return true;
        
        //see PDN classes
        if (parentNOBJ.wvBIPASS!=undefined && parentNOBJ.wvBIPASS==true) return true;

        return false;
    }

    /**
     * other neuron props extracted from relons.
     */
    extractOtherPropsFromRelon=function($rObj) {
        if (this.dead==true) return;
        
        if ($rObj.pubKey!=undefined && (this.pubKey==undefined || this.pubKey==null || this.pubKey=="")) {
            this.pubKey=$rObj.pubKey;
        }

        if ($rObj.freeze!=undefined) {
            this.freeze = $rObj.freeze;
        }

        //essentially a TYPE_SYS clump.
        if ($rObj.type==TYPE_CLUMP) {
            //we are pointing to a delegation rights clump
            if (MEM.delegation_rights.temp_nid==$rObj.r) {
                this.addDelegation($rObj);
            }
        }

        //do some quick mojos here.
        else if ($rObj.type==TYPE_SYS) {

            //set the eval value
            //we don't do this for the TYPE BE evals.
            //they have their type_be value already.
            if($rObj.r==MEM.eval.temp_nid) {
                MEM.eval.set_value(this);
            }
            else if ($rObj.r==MEM.garden_type.temp_nid) {
                this.setGtRelon($rObj);     //we will want this to cascade down the be tree
            }
            else if ($rObj.r==MEM.inst_garden_type.temp_nid) {
                this.setGtiRelon($rObj);     //we will want this to cascade down the be tree
            }
            else if ($rObj.r==MEM.let_wv.temp_nid) {
                this.addLet($rObj.t);       //we will want this to cascade down the be tree
            }
            else if ($rObj.r==MEM.deny_wv.temp_nid) {
                this.addDeny($rObj.t);      //we will want this to cascade down the be tree
            }
            else if ($rObj.r==MEM.viplet.temp_nid) {
                this.addVipLet($rObj.t);
            }
            else if ($rObj.r==MEM.trusted_adopt.temp_nid) this.addAdopt($rObj.t);
            else if ($rObj.r==MEM.abandon.temp_nid) this.addAbandon($rObj.t);
            
            else if ($rObj.r==MEM.gt_permit_reln.temp_nid) {
                //ANNE.TALK.printRelon($rObj);
                //if (MEM.location.temp_nid==$rObj.f) {
                    //console.log("%c CONFIRM EGRS HERE", 'background: white; color: red');
                //}
                this.addGardenRule($rObj);
            }
            else if (MEM.delegation_reln_tnids.indexOf($rObj.r)!==-1) {
                this.addDelegation($rObj);
            
                //all delegation relons stored in: this.delegation
                //note, these are not yet cascaded.
                //that happens run-time when needed.

                //store this for quick test/helper.
                
                if ($rObj.r==MEM.dr_andy_hanai_in.temp_nid) {
                    //set some helpful arrays for IDENTITIES.hanai_admin
                    //and nonidentityneuron.hanai_admins
                    var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);
                    if (fromNOBJ!=null) {
                        //non-identity neurons get hanai_admins array.
                        if (fromNOBJ.hanai_admins==undefined) fromNOBJ.hanai_admins=[$rObj.t];
                        else if (fromNOBJ.hanai_admins.indexOf($rObj.t)==-1) fromNOBJ.hanai_admins.push($rObj.t);
                    }
                    
                    //identities get a hanai_admin array.  (mostly only relevant for USER._userNObj)
                    var toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
                    if (toNOBJ!=null) {                    
                        if (toNOBJ.hanai_admin==undefined) toNOBJ.hanai_admin=[$rObj.f];
                        else if (toNOBJ.hanai_admin.indexOf($rObj.f)==-1) toNOBJ.hanai_admin.push($rObj.f);
                    }
                }
            }

            //unique on.. to help w/ disambig.
            else if ($rObj.r==MEM.uo.temp_nid) {
                if(this.uo==undefined) this.uo=[$rObj.t];
                else this.uo.push($rObj.t);
            }

            
            //mark this to_class requiring validation.
            //if it is used as a reln, the to will have to be validated
            //according to the set validator for that type.
            else if ($rObj.r==MEM.require.temp_nid && $rObj.t==MEM.validation.temp_nid) this.requires_validation=true; 
            
            //clean these up for param. must be alpha start and alphanum only
            //if non_alpha start prepend "p_";
            else if ($rObj.r==MEM.PARAM_NAME.temp_nid) {
                var cleanPropName = UTILS.cleanPropName($rObj.to_s);
                if (MEM._RESERVED_PROPS.indexOf(cleanPropName)==-1) {
                    this.PARAM_NAME=$rObj.to_s;
                }
            }

            //relon param name for this nObj
            else if ($rObj.r==MEM.REL_PARAM_NAME.temp_nid) {
                var cleanPropName = UTILS.cleanPropName($rObj.to_s);
                if (MEM._RESERVED_PROPS.indexOf(cleanPropName)==-1) {
                    this.REL_PARAM_NAME=$rObj.to_s;
                }
            }
            else if ($rObj.r==MEM.no_new_children.temp_nid && Boolean($rObj.to_s)==true) this.no_new_children=true;
            else if ($rObj.r==MEM.classes_only.temp_nid && Boolean($rObj.to_s)==true) this.classes_only=true;
            
        }

        else if ($rObj.type==TYPE_BE) {


            /********************/
            //ANNODE .from_text index
            //Our neurons tbl in ANNODE now stores the .text.
            //We send it over as from_text since it is a view tied to relons.
            //Relons do too, but for some node-made-neurons we don't make
            //the MEM.name-> relon.  We make a sponsor, or a BE class, etc.
            //Text comes through with the NARR record so process it here.
            //We only want to bother with TYPE_BE returns for the TEXT.. as it
            //will be on ALL RELONS in the NARR view. 
            if ($rObj.r!=MEM.name.temp_nid && $rObj.from_text!=undefined && $rObj.from_text.length) {
                if (this.text.length < $rObj.from_text.length) {
                    if (this.text=="") {
                        this.text = $rObj.from_text;
                    }
                    if (this.text==undefined) { 
                        this.text=""; //???
                    }

                    //we really should be storing "latest name" for MEM.name relons.
                    else {
                        //if (this.latest_name_relon.time > $rObj.time) {
                          //  this.addAkaManual(this.latest_name_relon.to_s);
                           // this.text = $rObj.from_text;
                        //}
                    }
                    
                }
            }

            if ($rObj.r==MEM.ANNE_DDS_SRC.temp_nid) {
                var toNObj = MEMG.getNeuronByTempNid($rObj.t);
                if (toNOBJ._gt!=null && this._gt==null) this.setGtRelon(toNObj._gt);    
                this.addAltBE($rObj);   
            }
        }

        else if ($rObj.type==TYPE_HAS) {
            //local memmanger ico propert
            // (easy access prop) for this version of the memory manager.
            if ($rObj.r==MEM.ai_ico_class.temp_nid) this.ai_ico=$rObj.to_s;
        }

        else if ($rObj.type==TYPE_SYS_MSG) {
            
            //pending relon to freeze a neuron 
            if ($rObj.r==MEM.CMD_FREEZE.temp_nid && $rObj.ghost==true) {
                var fromNOBJ = MEMG.getNeuronByTempNid($rObj.f);
                fromNOBJ.freeze=1;
            }
        }
    }


    /**
     * 
     * @param {*} $nameRelon 
     * @returns 
     */
    setLatestNameRelon = function($nameRelon) {
        if (this.latestNameRelon==null) {
            this.latestNameRelon = $nameRelon;
            return;
        }

        if ($nameRelon.time > this.latestNameRelon.time) {
            this.addAkaManual(this.latestNameRelon.to_s);
            this.latestNameRelon = $nameRelon;
        }
        else {
            this.addAkaManual($nameRelon.to_s);
        }
    }


    /**
     * Inherit important functions by class.
     * 
     * .requires_validation  -- this is for external relons that validate    
     * claims.                                                                  
     *
     * .set_value  -- this is to set the default .value of a neuron  
     * (or class of neurons) 
     *  
     * .formatter  -- this is to help format the value of an obj
     *  
     * .format_validator  -- this is for the value being good or bad according  
     * to desired form.  
     * 
     * These can be manually or programmatically set, and if set, they will
     * be inherited down the child tree.   As of now they cascade only by the
     * proper TYPE_BE reln: class/inst type parent/child relationshps, not the creative
     * TYPE_BE other relations using BE_CHILD and BE_PARENT handling (to build dynamic
     * trees and datasources).
     * 
     * NOTE: we also inherit _gt (a relon pointing to the garden type) here to cascade to all children.
     * There are quite a few other things inherited in a hierarchical structure
     * (including lateral BEs on INSTS which inherit params/has from the lateral BE).
     * Those are done run-time when needed.  These are setup ready to be used on the class
     * (not the instances) since the instances will call their class function.
     *  
     * GT is an exception --- we do that here (tho it is not a function) and it is cascaded
     * to all neurons in the BE parent/child tree.
     * 
     * NOTE2: The GT cascades ONLY TO HEIRARCHICAL CHILDREN not dynamic parent/child trees such
     * as the country tree... It uses the BE_PARENT and BE_CHILDREN pattern to inform us that
     * it makes a tree of "locations" which have a BE polity of "country".  This allows for the 
     * most flexible of heirarchies since we can generate the live and transform to various 
     * be tree structures without being locked in.   "AMERICA" or "AUSTRALIA" are countries, per say,
     * but in ANNE they are children of location.  Their parent is the location neuron.  We know
     * they are countries because they have a TYPE_BE polity-> country relon.
     * 
     * @params {Object} - the BE neuron object we inherit from.
     *  
     * @author {Mr.Scatman} 
     */
     __functInherit=function($beObj, $whatCalledThis="") {
        this._uico = undefined;

        //if (ANNE.SETTINGS && ANNE.SETTINGS.bUICOS==false) {
            
        //}
        
        //console.log("(__functInherit)")
        if (this.dead==true) return;
        if ($beObj==null) return;

        var child_props = new Object();
        if (this.ntype==NTYPE_CLASS) {
            if ($beObj.set_value!=undefined && this.set_value==undefined) {
                this.set_value=$beObj.set_value;
                child_props["set_value"]=$beObj.set_value;
            }
            if ($beObj.set_value2!=undefined && this.set_value2==undefined) {
                this.set_value2=$beObj.set_value2;
                child_props["set_value2"]=$beObj.set_value2;
            }
            if ($beObj.formatter!=undefined && this.formatter==undefined) {
                //console.log("%c"+$beObj.text+" formatter to push down to child", TALK.bggreen);
                if ($beObj.nid==this.nid && MEM.string.nid==this.nid){
                    //console.log("SETTING STRING FORMATTER");
                }
                this.formatter=$beObj.formatter;
                child_props["formatter"]=$beObj.formatter;
            }
            if ($beObj.format_validator!=undefined && this.format_validator==undefined) {
                this.format_validator=$beObj.format_validator;
                child_props["format_validator"]=$beObj.format_validator;
            }
            if ($beObj.requires_validation!=undefined) {
                this.requires_validation=true;
                child_props["requires_validation"]=this.requires_validation;
            }
            if ($beObj.fire_circuit!=undefined) {
                this.fire_circuit=$beObj.fire_circuit;
                child_props["fire_circuit"]=$beObj.fire_circuit;
            }

            //review this..
            if ($beObj.role_action!=undefined) {
                console.log("%c REVIEW THIS", TALK.bgred);
                this.role_action=$beObj.role_action;
                child_props["role_action"] = $beObj.role_action;
            }

            //will inform the platform what relon type to make with params using
            //this concept as a reln
            if ($beObj.target_type!=undefined) {
                this.target_type=$beObj.target_type;
                child_props["target_type"] = $beObj.target_type;
            }
        }

        //not a funct but we want to inherit these props too.
        if (this._gt==null) {
            if ($beObj._gt!=null) {
                this.setGtRelon($beObj._gt);
                child_props["_gt"]=$beObj._gt;
            }
        }
        else child_props["_gt"]=$beObj._gt;

        //gti
        if (this._gti==null) {
            if ($beObj._gti!=null) {
                this.setGtRelon($beObj._gti);
                child_props["_gti"]=$beObj._gti;
            }
        }
        else child_props["_gti"]=$beObj._gti;

        //if (this._uico!=undefined) {
            //child_props["_uico"]=this._uico;
        //}
        //else {
           // if ($beObj._uico!=undefined) {
                //this._uico=$beObj._uico;
                //child_props["_uico"]=this._uico;
            //}
        //}

        //nothing key functs (or gt) to inherit.
        if (this.children==null || this.children.length==0) return;

        var keys = Object.keys(child_props);
        if (keys.length>0) {
            var chain = [];
            this.cascadePropsDownChildrenClasses(child_props, chain);
            
            //console.log("chain of props for *****"+this.text+"***** ("+$whatCalledThis+")", chain);
        }
    }


    /**
     * recurse up the be line to generate the BEM nids.
     * @param {*} $arr 
     * @returns 
     */
    getBemNids = function($arr=[]) {
        var bem_nids = $arr;
        if (this.bes.length) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            bem_nids.push(beNOBJ.nid);
            return beNOBJ.getBemNids(bem_nids);
        }
        return bem_nids;
    }


    /**
     * Look for the greatest firstseen value for a given reln (relationship).
     * Example:  find the most recent 'status' relon (a relon with reln:SL1C_status.nid).
     */
    getLastFirstSeen=function($r) {
        if (this.dead==true) return;
        var biggestFS = null;
        for (var r=0; r < this.outs.length; r++) {
            if (this.outs[r].r!=$r) continue;
            var rel = this.outs[r];
            if (rel.firstseen==undefined || rel.firstseen==null) continue;
            if (biggestFS==null) biggestFS = rel;
            if (rel.firstseen > biggestFS.firstseen) biggestFS = rel;
        }
        if (biggestFS!=null) return biggestFS.firstseen;
        return 0;
    }


    /**
     * @param {anne_neuron}
     * @returns {boolean}
     */
    getNameRelon=function($relNOBJ) {
        for (var r=0; r < this.outs.length; r++) {
            if (this.outs[r].r==$relNOBJ.temp_nid) return this.outs[r];
        }
        return null;
    }
    

    /**
     * Get the default class a param relon is looking to assign.
     * 
     * In normal params we point to the default class via using the reln and or the to.  Either can be used
     * to help us know what the default class of the param is.
     * 
     * With semantic clumps we use the to_s value.
     * 
     * Default class for a param using a ROLE in the reln is used with the to_s, but that is generalized at a high level.
     * 
     * To update or specify a more refined class we can use a TYPE_SYS relon and reference the param.
     * The referenced param the default class applies to is in the sys_ref_txid slot of the relon.
     * 
     * Here we lookup a default class nid for a given class and a given param.
     * This is needed for inherited params that cannot specify this info in the to_s because they are high level and the
     * downstream this value can change.  A param for ROLE_AFFECTED_THING is part of top-level action.  At the top-level it points
     * to an ec_thing.
     * 
     * That top-level action is the parent of a lower-level actions that inherits this parameter,
     * However, at lower levels the action may want/need to specify that the default class of value it points to is something 
     * different than just ec_thing.  
     * 
     * Example: In the case of marry the default class of the ROLE_AFFECTED_THING should be person.  People marry people.
     * We could just store again the relon for PARAM->ROLE_AFFECTED_THING to update it (make it more specific) in the marry 
     * neuron.  We can use either method to get the default.  This function is for the sys_ref_txid method.
     * 
     * @param {string} $ptxid - txid long string of the parameter txid. 
     * @returns 
     */
    getDefaultClassForParam=function($ptxid) {
        for (var r=0; r < this.outs.length; r++) {
            var rel = this.outs[r];
            if (rel.type!=TYPE_SYS) continue;
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            if (rel.r!=MEM._default_class_nid.temp_nid) continue;
            if (rel.sys_ref_txid==undefined || rel.sys_ref_txid=="" || rel.sys_ref_txid!=$ptxid) continue;
            return rel.to_s;
        }
        var prel = MEMG.lookupRelonByTXID($ptxid);
        if (prel!=null) return prel.to_s;   //default param relon
        return null;
    }

    getRelnUsageChildren=function($type) {
        var to_nids = []
        for (var r=0; r < ANNE.MEM._aRelons.length; r++) {
            if (ANNE.MEM._aRelons[r].type!=$type) continue;
            if (ANNE.MEM._aRelons[r].reln!=this.nid) continue;
            if (ANNE.MEM._aRelons[r].to_type!=TO_TYPE_ANNE) continue;
            if (to_nids.indexOf(ANNE.MEM._aRelons[r].to_s)==-1) {
                to_nids.push(ANNE.MEM._aRelons[r].to_s);
            }
        }
        return ANNE.MEMG.getNObjsFromList(to_nids);
    }


    getNameRelon=function(){
        return MEMG.getBeRelon(this.outs, MEM.name.temp_nid);
    }

    /**
     */
    makeNameTempRelon=function($relNOBJ, $name) {
        if (this.dead==true) return;
        return MEM.makeTempRelon(nOBJ, TYPE_BE, $relNOBJ.nid, TO_TYPE_STRING, $name, true);
    } 
    
    /**
     * Is this neuorn an alt parent requirement for any other class?
     * 
     * Hierarchical parents in the BE tree  are not always the real (functional) 
     * hierarchical parent.  All countries, states, towns, villages (etc) are hierarchical 
     * children of "location". State, town, country .. etc.. however, are not SUB-CLASSES of 
     * location.  They are polities.  An instance of a country or state or city is not actually
     * and instance of those classes. They are instances of "location" but their functional parent 
     * is state, town or country.  This is specified with a BE_CHILD from country to "polity" and 
     * then the actual instances of location which ARE countries specify their polity with 
     * polity->country.  This tells ANNE the parent/child reln used is polity to define the actual
     * function parent.  
     * 
     * The BE_PARENT relon specifies that to be a country you must not only have the polity->country relon 
     * but also the neuron must be a child of location.   That relon, the "must be a child of location" is 
     * what we are looking for here. We want to find out if this neuron is being used for ANY 
     * alt be parental lineages. We can use that information to see reclassify the children of 
     * this neuron to their intended (functional) parent.
     */
    getParentAlts = function() {
        if (this.dead==true) return;
        var parent_alts = [];
        if (this.bes.length==0) return parent_alts;
        var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
        if (beNOBJ==null) return;
        //console.log("assessing ''"+this.text+"'' parent ''"+beNOBJ.text+"'' for BE PARENT in relons");
        //var audit = MEMG.getNeuronByNid("7775942183860735787");
        //console.log("audit "+audit.temp_nid+" == "+MEM.BE_PARENT.temp_nid)
        for (var i=0; i < beNOBJ.ins.length; i++) {
            if (beNOBJ.ins[i].r!=MEM.BE_PARENT.temp_nid) continue;
            parent_alts.push(beNOBJ.ins[i].f);
        }
        return parent_alts;
    }

    /**
     * We want this to honor the alt be parent line setup..
     * TODO
     */
    getFunctionalParentName = function() {
        if (this.bes.length==0) return "";
        var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
        if (beNOBJ==null) return;
        for (var i=0; i < beNOBJ.ins.length; i++) {
            if (beNOBJ.ins[i].r!=MEM.BE_PARENT.temp_nid) continue;
            var fBE = MEMG.getNeuronByTempNid(beNOBJ.ins[i].f);
            if (fBE!=null) return fBE.text;
        }
        return beNOBJ.text;
    }

    //if a user (any user) adopts a class -- they get ALL CHILDREN
    //of the class in their worldview.
    //This may already work with our children search and handling..
    //i will some cases.
    /** 
     * 
     * @author {Mr.Scatman} 
     */
    getAdoptClassChildren=function() {
        alert("TO DO");
    }


    /**
     * UNTESTED
     * 
     * We have a requirement of a relon.. to build a list of neurons.
     * Instead of building  the list of neurons then filtering..
     * We build the list via the filter.
     * user case:  we want a dropdown of all image ---- THAT a user is connected to.
     * I don't want to build the list of all images the filter. 
     * Better to just go to the user and connect to the images and build the list.
     * This could just be a simple function but this is called in a standardized area that 
     * most otherwise DOES want to just provide a list of children (all children).
     * 
     * Test this by connecting to some pictures in a persons neurons.
     * Then -- set the primaryNeuronImg and we should only be presented with a dropdown of 
     * the images that are tied to the neuron already.
     */
    getAllDownlineChildren_cdxFAST=function($cdxs, $downline, $bINSTS=true, $bTNIDLIST=true, $bDOALLBES=false, $bINSTSONLY=false) {
        console.log("%c doing optimized getAllDownlineChildren_cdxFAST", TALK.bgpurple);
        console.log($cdxs); 
        var cdx = $cdxs[0];
        var fromNOBJ = MEMG.getNeuronByNid(cdx.from_nid);
        var nObjs = [];
        var reln = cdx.reln;
        var type = cdx.type;
        var to_type = cdx.to_type;
        if (to_type!=TO_TYPE_ANNE) return [];
        var ntype = null;
        if ($bINSTSONLY) ntype=NTYPE_INST;
        for (var r=0; r < fromNOBJ.outs.length; r++) {
            var rel = fromNOBJ.outs[r];
            if (type!=null && rel.type!=type) continue;
            if (to_type!=null && rel.to_type!=to_type) continue;
            if (reln!=null && rel.reln!=reln) continue;
            var nObj = MEMG.getNeuronByTempNid(rel.t);
            if (MEM.tempBeIsA(nObj.temp_nid, this.temp_nid, ntype)==true) {
                if ($bTNIDLIST==true) $downline.push(nObj.temp_nid);
                else $downline.push(nObj);
            } 
        }
        return nObjs;
    }


    /**
     * UNTESTED;.
     */
    getAllDownlineChildren_cdx=function($cdxs, $downline, $bINSTS=true, $bTNIDLIST=true, $bDOALLBES=false, $bINSTSONLY=false) {
        if ($cdxs.length==1 && $cdxs[0].target_part=="to_s") {
            return this.getAllDownlineChildren_cdxFAST($cdxs, $downline, $bINSTS, $bTNIDLIST, $bDOALLBES, $bINSTSONLY);
        }
        var nObjs = this.getAllDownlineChildren($downline, $lvl, $bINSTS, false, $bDOALLBES, $bINSTSONLY);
        var nObjs_final = [];
        var tnids=[];
        for (var n=0; n<nObjs.length; n++) {
            var nObj = nObjs[n];
            //misc.child_cdx = [{from_nid:for_nOBJ.nid, reln:MEM.neuronImg.nid, to_s:null, type:null, to_type:TO_TYPE_ANNE, target_part:"to_s"}];
            var passed = MEM.enforceCDXS($cdxs, nObj);
            if (passed==false) continue;
            if (tnids.indexOf(nObj.temp_nid)!=-1) continue
            tnids.push(nObj.temp_nid);
            nObjs_final.push(nObj);
        }
        if ($bTNIDLIST==true) return tnids;
        return nObjs_final;
    }


    getClassChildren=function() {
        var downline = [];
        this.getAllDownlineChildren(downline, 0, false, true, false, false);
        return this.children;
    }


   /**
    * Get downline chilrden of a neuron.
    * 
    * We want to chase all the downline of a neuron and build an array..   
    * We want to chase all paths.   Also, if we have the bDOALLBES set we   
    * will snag any inbound relns to this neuron with a TYPE_BE that are   
    * not hierarchical children.                                         
    *                                                                     
    * @author {Mr.Scatman}
    * 
    * @param {*} $downline 
    * @param {*} $lvl 
    * @param {*} $bINSTS 
    * @param {*} $bTNIDLIST 
    * @param {*} $bDOALLBES 
    * @param {*} $bINSTSONLY 
    * @returns 
    */
    getAllDownlineChildren=function($downline, $lvl=0, $bINSTS=true, $bTNIDLIST=true, $bDOALLBES=false, $bINSTSONLY=false) {
        if ($lvl>16) return;
        if ($lvl==0) {
            //get other bes of this thing that are not hierarchical children.
            if ($bDOALLBES==true) {
                for (var i=0; i < this.ins.length; i++) {
                    var inRel = this.ins[i];
                    var childNOBJ = MEMG.getNeuronByTempNid(inRel.f);
                    if ($bINSTS==false && childNOBJ.ntype!=NTYPE_CLASS) continue;
                    if (childNOBJ.enforceGardenUp()==false) continue;
                    
                    if (inRel.type==TYPE_BE && inRel.r!=MEM.class.temp_nid && inRel.r!=MEM.inst.temp_nid) {
                        if ($bTNIDLIST==true) {
                            if ($downline.indexOf(inRel.f)==-1) $downline.push(inRel.f);
                        }
                        else $downline.push(childNOBJ);
                    }
                }
            }
        }

        //add these in if they haven't been done yet.
        //only add unique children.
        if (this['be_'+MEM.BE_CHILDREN.temp_nid]!=undefined) {
            var new_children = MEM.getChildrenDynamic(this, $bTNIDLIST, this.children);
            console.log(new_children);
        }

        if (this.children==null) return; //no more work to be done

        for (var c=0; c < this.children.length; c++) {
            var child_temp_nid = this.children[c];
            var childNObj = MEMG.getNeuronByTempNid(child_temp_nid);
            if (childNObj.enforceGardenUp()==false) continue;
            if ($bINSTS==false && childNObj.ntype!=NTYPE_CLASS) continue;
            if ($bTNIDLIST==true) {
                if ($downline.indexOf(child_temp_nid)==-1) $downline.push(child_temp_nid);
            }
            else $downline.push(childNObj);
            childNObj.getAllDownlineChildren($downline, $lvl+1, $bINSTS, $bTNIDLIST);
        }
        
        //to get inst only we still had to traverse the class tree.
        if ($lvl==0 && $bINSTSONLY==true) {
            var downline = $downline.concat();
            console.log("old downline "+downline.length);
            $downline.length=0;
            for (var d=0; d < downline.length; d++) {
                var val = downline[d];
                //if we were in tnidlist mode, get the nObj and check
                //for inst only.
                if ($bTNIDLIST==true) {
                    var nObj = MEMG.getNeuronByTempNid(val);
                    if (nObj.ntype!=NTYPE_INST) continue;
                }
                else if (val.ntype!=NTYPE_INST) continue;  //not an instances
                $downline.push(val);    //a valid instance..
            }
            
            console.log("new downline length:"+$downline.length);
            //$downline = downline.concat();   //swap out the array ref. (i miss C++ ptrs but this is close)
        }
        return;
    }



    getWikiPart = function() {
        return MEMG.getSysValue(this.outs, MEM.wikipart.temp_nid);
    }

    /**
     * We know the base of the wikipart but here we still look it up
     * to demonstrate how this can be done for any URL based + part.
     */
    getWikiURL = function() {
        var part = MEMG.getSysValue(this.outs, MEM.wikipart.temp_nid);
        if (part!=null) {
            var base = MEMG.getHasValue(MEM.wikipart.outs, MEM.base_url.temp_nid)
            if (base!=null) return base+part;
        }
        return null;
    }


    /**
     * Does this neuron have any ghosts relons?
     */
    hasGhostRelons=function(){
        for (var r=0; r < this.outs.length; r++) {
            if (this.outs[r].ghost==true) return true;
        }
        return false;
    }

    
    hasRelon=function($reln, $to_s) {
        for (var r=0; r < this.outs.length; r++) {
            if (this.outs[r].reln==$reln && this.outs[r].to_s==$to_s) return true;
        }
        return false;
    }


    /**
     * look for smart search relon .. or garden of free for all..
     */
    hasRelonTo=function($type, $to_s, $target) {
        if ($target==TARGET_NID) {
            var toNOBJ = MEMG.getNeuronByTempNid($to);
            var rels = MEMG.getRelonsTo(toNOBJ, $type, TO_TYPE_ANNE, $to_s);
            if (rels.length>0) return true;
        }
        else if ($target==TARGET_TXID) {
            var rels = MEMG.getRelonsToNaked($type, TO_TYPE_TXID, $to_s);
        }

        return false;
    }


    /**
     * look for smart search relon .. or garden of free for all..
     */
    isSmartSearch = function() {
        return true;
    }



    /***** MAKE functions for easy relon creation.
     * These support collab by using the from (this neuron) *****/
    
    /**
     * Make an abandon SYS relon
     * 
     * Abandon is only relevent on previously adopted classes/inst.  This severs
     * the adoption. This is only permitted by the 
     * owner OR delegated owner.  It is not permitted as a collaboration.
     *  
     * @author {Mr.Scatman}
     */
     _makeAbandon  = async($abandonNid)=> {
        if (this.dead==true) return;
        return await GHOSTDATA.new_ghost_sys(this, MEM.abandon, TO_TYPE_ANNE, $wvToDeny);
    }

    /**
     * Make an adopt SYS relon
     * 
     * This adopts a class or inst into this worldview.  This is only permitted by the 
     * owner OR delegated owner.  It is not permitted as a collaboration.
     * 
     * @author {Mr.Scatman}
     */
     _makeAdopt  = async($adoptNid)=> {
        if (this.dead==true) return;
        return await GHOSTDATA.new_ghost_sys(this, MEM.trusted_adopt, TO_TYPE_ANNE, $wvToDeny);
    }
    


    /**
     * Make an arbor prune SYS relon
     * 
     * This prunes a relon from your worldview.   If the relon being pruned is owned
     * by the sender or the sender has delegated owner rights, the relon will be
     * removed from all ANNE indexes and wiped from existence.
     * 
     * If you want to block/hide/ or remove something from view that is not yours,
     * use the inhibit relon.  This will instruct the memory handler to do a run-time
     * block of any relon you inhibit.   It is also useful when MIRRORING an inst with  
     * a MIRROR inst but wanting to negate a few data points (but otherwise keep all the others).
     * 
     * 
     * @author {Mr.Scatman}
     */
    _makeArborPrune=async($txid, $remove=true)=> {
        if (this.dead==true) return;
        if (typeof $txid === "object") {
            $txid = $txid.txid;
            console.log("fix this reference!  pass the txid, not the object!");
        }
        if ($txid==undefined || $txid==null || $txid=="") {
            console.log("bad txid");
            return;
        }
        var rel_to_delete = MEMG.lookupByTXID($txid);
        if (rel_to_delete==null) return;

        //no need for arbor prune if this is a ghost.
        //just delete the ghost.
        if (rel_to_delete.ghost==true) {
            await GHOSTDATA.deleteSingleGhostRelon($txid);
            return;
        }

        //we need an arbor prune..
        //we will make a GHOST arbor prune.
        //the user can broadcast it whenever they like. (or delete it and chg their mind.)
        //the arbor pruned relon is removed from RAM and the local DB.
        //If they change their mind they will delete the ghost and then 
        //upon refresh of the app they will get their relon back in local DB mem (and then RAM) again.
        var rel = await GHOSTDATA.new_ghost_sys(this, MEM.arbor_prune.nid, TO_TYPE_TXID, $txid);
        if ($remove==false) return rel;
        var removed_txids = [];
        this.removeByTXID("outs", $txid, removed_txids);

        var rel_to_ap = MEMG.lookupRelonByTXID($txid);
        if (rel_to_ap!=null) {
            if (rel_to_ap.to_type==TO_TYPE_ANNE) {
                var toNOBJ = MEMG.getNeuronByTempNid(rel_to_ap.t);
                if (toNOBJ!=null) toNOBJ.removeByTXID("ins", $txid, removed_txids);
            }
        }

        await ANNE.DB.___delete([$txid], "relons");
        return rel;
    }

   
    /**
     * Make a deny SYS relon.
     * 
     * Denies ability to send collab requests to the worldview.  This is relevant in the BLACKLIST
     * GARDEN of "communal" as the general public is whitelisted.  
     * 
     * Only relevant when supplying a let on a owned class or a class one has
     * delegation rights to.  Otherwise, collabs of the SYS variety are NOT permitted.
     * 
     * 
     * @author {Mr.Scatman}
     */
     _makeDeny  = async($wvToDeny)=> {
        if (this.dead==true) return;
        return await GHOSTDATA.new_ghost_sys(this, MEM.deny_wv, TO_TYPE_ANNE, $wvToDeny);
    }

    
    /**
     * Make an inhibit SYS relon.
     * 
     * A lighter version of Arbor Prune which does not blot things out of existence.
     * This will handle the inhibit of a relon at run-time in the memory handler.  It is a way
     * to cleanup your view of other worldviews without altering their worldview.
     *
     * @author {Mr.Scatman}
     */
     _makeInhibit=async($txid)=> {
        if (this.dead==true) return;
        //we could do some validation here to make sure the $txid exists
        var rObj = await GHOSTDATA.new_ghost_sys(this, MEM.inhibit_relon.nid, TO_TYPE_TXID, $txid);
    }


    /**
     * Make a let SYS relon.
     * 
     * Only relevant when supplying a LET on a owned class or a class one has
     * delegation super to.  Otherwise, collabs of the SYS variety are NOT permitted.
     * 
     * LETS are relevant in the WHITELIST of COLLAB-OWNER.  Only worldviews LET will have
     * the ability to send collab requests.
     * 
     * @author {Mr.Scatman}
     */
     _makeLet = async($wvToLet)=> {
        if (this.dead==true) return;
        return await GHOSTDATA.new_ghost_sys(this, MEM.let_wv, TO_TYPE_ANNE, $wvToLet);
    }


    /**
    * @author {Mr.Scatman}
    */
    _makeSelfName = async($name)=> {
        if (this.dead==true) return;
        if ($name==undefined || $name==null || $name=="") {
            console.log("name cannot be empty");
            console.log("%c NAME CANNOT BE EMPTY", TALK.bgred);
            return null;
        }
        var rObj = await ANNE.GHOSTDATA.new_ghost_be(this, MEM.name.nid, TO_TYPE_STRING, $name);
    }


    /**
     * This will only create a ghost.  The user has to send that ghost to ANNE to make it real.
     * @author {Mr.Scatman}
     */
     _makeSelfIdentity = async ($identity_class_of)=> {
        if (this.dead==true) return;
        if ($identity_class_of=="") {
            console.log("identity class cannot be empty");
            return null;
        }
        var futureBE = ANNE.MEMG.getNeuronByNid($identity_class_of);

        if (this.bes.length>0 && this.bes[0]==futureBE.temp_nid) {
            return null; //this is already done.
        }
         
        return await ANNE.GHOSTDATA.new_ghost_be(USER._userNObj, MEM.inst.nid, TO_TYPE_ANNE, futureBE.nid);
        //await ANNE.MEMG.processNeuronsFromRelons([ghost_identity], false, true, "ghost_relons");
    }


    /**
     * Make a necrosis SYS relon.
     * 
     * Completely nuke a neuron and all its outs and ins from existence.  Sender must be
     * the author or approved delegation author (with super rights) to submit necrosis SYS relons.
     * 
     * Collab does not otherwise accept SYS relons.
     * 
     * @author {Mr.Scatman}
     */
    _makeNecrosis=async($nObj)=> {
        if (this.dead==true) return;
        //we can safely nuke this in local mem after it returns back from broadcast.
        return await GHOSTDATA.new_ghost_sys(this, MEM.necrosis.nid, TO_TYPE_ANNE, $nObj.nid);
    }
    
    //_fixName("13664565417915800120", "oracle data run")
    _fixName=async($nid, $newName)=>{
        var fNOBJ = ANNE.MEMG.getNeuronByNid($nid);

        var rel_name = fNOBJ.getNameRelon(MEM.name);
        if (rel_name==null) return;
        
        var rel = await ANNE.GHOSTDATA.new_ghost_sys(fNOBJ, MEM.arbor_prune.nid, TO_TYPE_TXID, rel_name.txid);
        await SENDQ.addRelonToSendQ(rel, false, TXB.TXTYPE_CHILD_RELON);
        
        var newrel = await ANNE.GHOSTDATA.new_ghost_be(fNOBJ, MEM.name.nid, TO_TYPE_STRING, $newName);
        await SENDQ.addRelonToSendQ(newrel, false, TXB.TXTYPE_CHILD_RELON);

        await ANNE.SENDQ.___sendQueue(($msg="update name")=>{
        });
    }

    printBes = function() {
        var bes_text = [];
        for (var b=0; b < this.bes.length; b++) {
            var beNOBJ = MEMG.getNeuronByTempNid(this.bes[b]);
            bes_text.push(beNOBJ.text+" nid:"+beNOBJ.nid);
        }    
        var str =  bes_text.join(" :: ");
        console.log(str);
        return str;
    }
    

    /**
     * 
     */
    printRights = async()=> {
        if (this._govenor==null) await this.setGovRights(USER._userNObj);
        this._govenor.printRights();
    }


    /**
     */
    printBeParentage = function() {
        //console.log("print be parentage for "+this.text+" nid:"+this.nid+" be_lvl:"+this.be_lvl);
        var beNOBJ = null;
        if (this.bes.length==0) return;
        beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
        while(beNOBJ) {
            //console.log("be_lvl:"+beNOBJ.be_lvl+" "+beNOBJ.text+" nid:"+beNOBJ.nid);
            beNOBJ = MEMG.getNeuronByTempNid(beNOBJ.bes[0]);
        }
    }
    

    fixBeLvls = function() {
        return; //this doesn't make sense when we don't have the full graph with the new JS optimized attempt.
        if (this.dead==true) return false;
        if (this.be_lvl!=-1) return false;
        var beNOBJ = null;
        if (this.bes.length==0) return false;
        beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
        while(beNOBJ.be_lvl==-1) {
            //console.log(beNOBJ.text+" nid:"+beNOBJ.nid+" be_lvl:"+beNOBJ.be_lvl);
            beNOBJ = MEMG.getNeuronByTempNid(beNOBJ.bes[0]);
            if (beNOBJ==null) {
                //console.log("vat is zis? "+this.nid+ " "+this.text);
                break;
            }
        }

        if (beNOBJ && beNOBJ.be_lvl!=-1) {
            beNOBJ.cascadeChildrenBeLevels();
            return true;
        }
        
        return false;
    }


    /**
     * take a child out of the children array
     * 
     * DONE
     */
     removeChild=function($child_temp_nid) {
        if (this.children==null) return;
        for (var c=0; c<this.children.length; c++) {
            if (this.children[c]==$child_temp_nid) {
                this.children.splice(c,1);
                return;
            }
        } 
    }
    

    /**
     * Remove a parent nid from the bes in-RAM-MEM array and unset the .parent prop.
     * 
     * @author {Mr.Scatman}
     *    
     */
     removeParentFromChild=function($parent_temp_nid, $bGRANNYTAKESOVER=false) {
        this.parent = "";  //quick ptr set by ANNODE on relons.
        if (this.bes.length<=1) return;

        //remove the be out of the bes array.
        for (var b=0; b < this.bes.length;b++) {
            if (this.bes[b]==$parent_temp_nid) {
                this.bes.splice(b, 1);
                break;
            }
        }
        this.be="";

        //RETEST THIS !!!!!

        if ($bGRANNYTAKESOVER==true) {
            var parentNOBJ = MEMG.getNeuronByTempNid($parent_temp_nid);
            if (parentNOBJ!=null) {
                var grandpaNOBJ = MEMG.getNeuronByTempNid($parent_temp_nid);
                if (grandpaNOBJ==null) {
                    this.be_lvl = 0;
                    this.cascadeChildrenBeLevels();
                }

                //set the children be levels properly.
                else {
                    this.be_lvl = grandpaNOBJ+1;
                    this.cascadeChildrenBeLevels();
                }
            }
            else {
                this.be_lvl = 0;
                this.cascadeChildrenBeLevels();
            }
        }
        else {
            this.be_lvl = 0;
            this.cascadeChildrenBeLevels();
        }
    }


    /**  
     * Remove a relon from an array.
     * In our outs/ins we now use the MAP so we need the MAP and 
     * we can use remove functions of the map.s
     * This is only used on nObj arrays (not the master MEM._aRelons), safe to do as-is. 
     * 
     * @author {Mr.Scatman} 
     */
     removeByTXID=function($arrName, $txid, $removed_txids=[]) {
        console.log("removeByTXID "+$arrName+" txid:"+$txid);
        
        var map = this[$arrName+"MAP"];
        if (map==undefined || map==null) return;
        
        console.log("map type: is arraymap? "+(map instanceof arraymap));

        var removed = map.removeU($txid);
        if (removed==true) {
            if ($removed_txids.indexOf($txid)==-1) {
                $removed_txids.push($txid);
                return true;
            }
        } 
        console.log("%c (removeByTXID) didn't find "+$txid, TALK.bgred);
        return false;
    }


    /** 
     * Set the primary hierarchical INST/CLASS be.
     * 
     * Critical to building our primary DAG TYPE_BE class/inst parent/child structure.
     * 
     * ANNE is multi-dimensions as this is merely one dimension, albeit an imortant one.
     * In ANNE there is ONE and **ONLY** ONE hierarchical parent.  This grafts the neuron into the
     * primary BE tree.  Other TYPE_BE relons exist and further define neurons and give them
     * access to "alt" structures.  The primary TYPE_BE relon has a reln of MEM.class.nid or 
     * MEM.inst.nid and the relon.to_s of that relon (which is the hierarchical parent) will 
     * be put in the .parent of the nObj. The temp_nid of the parent will be put into slot 0 of 
     * the nObj.bes array.  All other bes can be found in the .bes array, but after slot [0].
     * 
     * Example.  You are an identity.  Maybe u are a corporation, or an individual, a miner, or a silicon-based intelligence.
     * Whatever the case, your identity neuron points with a TYPE_BE relon reln:MEM.inst.nid to the value of your
     * identity class.  You are an instance of that class (which itself is a sub-class of MEM.identity).  
     * 
     * You may also be many other things.
     * 
     * Maybe you are a father, a mother, a programmer, an ice skater, an ornithologist or maybe
     * a sommelier!  These are also defininig about you but they are not your primary BE class.  These 
     * are TYPE_BE relons too in ANNE but they are alt TYPE BES since their reln will NOT be MEM.class.nid or
     * MEM.inst.nid.  The relon.reln will be something else... (even the generic MEM.ec_is.nid).
     * 
     * There are javascript BE .props made following the syntax nObj."be_%temp_nid%".  (see nObj.addAltBe) 
     * This is to support property/json key/pair notation for easy access in apps.  
     * If you want the polity BE value of a location (which has a relon of some location->polity->some polity value)
     * you can grab the relon or look for  nObj["be_"+MEM.polity.temp_nid].  For the country 
     * United States of America that value will be MEM.country.temp_nid.
     * 
     * nObj.bes[0] will always be the temp_nid of the hierarchical parent.
     * .parent will always be the .nid of the heirarchcal parent. 
     * .be will be a quick text of the hierarchical parent.  (we could nix this for RAM saving)
     * 
     * 
     * @author {Mr.Scatman} 
     */
    setBeTempNid=function($be_tnid) {
        if (this.dead==true) return;
        if (typeof $be_tnid==="object") $be_tnid = $be_tnid.temp_nid;
        if (typeof $be_tnid!=="number") {
            console.log("error with be_nid"); 
            return;
        }

        if ($be_tnid==MEM.class.temp_nid) {
            this.be_lvl=3;
            this.be = "class";
            this.ntype = NTYPE_CLASS; //ideally we'd take the text of the neuron
            this.addUniqueBe_fastIndex(MEM.class);
            return;
        }

        var parentNOBJ = MEMG.getNeuronByTempNid($be_tnid);
        this.addUniqueBe_fastIndex(parentNOBJ);
    }


    /** 
     * 
     * @param {*} $beObj 
     * @author {Mr.Scatman} 
     * @returns 
     */
     setBeLevel=function($beObj) {
        if (this.dead==true) return;
        if ($beObj.be_lvl>-1) this.be_lvl = $beObj.be_lvl+1;       //set the children levels off of the parent
        else if (this.be_lvl>-1) { 
            if ($beObj.be_lvl==-1) $beObj.be_lvl = this.be_lvl-1;  //set the parent from the child, do the grandchildren
        }
        else { //we have no reliable lvl information yet.
            this.be_lvl = -1; 
            return;
        }
        if (this.children==null) this.children=[];
        var child_lvl =  this.be_lvl+1;
        for (var c=0; c < this.children.length; c++) {
            var child = MEMG.getNeuronByTempNid(this.children[c]);

            if (child==undefined) continue;
            child.be_lvl = child_lvl;
            child.cascadeChildrenBeLevels();
        }
    }


    /** 
     * Set the BE TREE BE_LVL of this neuron.
     * 
     * @author {Mr.Scatman} 
     */
    setChildrenBeLevels=function() {
        if (this.dead==true) return;
        if (this.children!=null && this.children.length) {
            var child_lvl =  this.be_lvl+1;
            for (var c=0; c < this.children.length; c++) {
                var child = MEMG.getNeuronByTempNid(this.children[c]);
                child.be_lvl = child_lvl;
            }
        }
    }


    /**
     * 
     * @returns 
     */
    setIsIdentity = function() {
        if (this.dead==true) return;
        this.bISIDENTITY = MEM.tempBeIsA(this.temp_nid, MEM.identity.temp_nid, NTYPE_INST);
    }


    /**
     * 
     * @returns 
     */
    cascadeChildrenBeLevels=function() {
        if (this.dead==true) return;
        if (this.children!=null && this.children.length) {
            var child_lvl =  this.be_lvl+1;
            for (var c=0; c < this.children.length; c++) {
                var child = MEMG.getNeuronByTempNid(this.children[c]);
                child.be_lvl = child_lvl;
                child.setChildrenBeLevels();
            }
        }
    }


    /** 
     * 
     * @author {Mr.Scatman} 
     */
     setGtRelon=function($gtRelon) {
        if (this.dead==true) return;
        //safety check for GT inherit rule.
        var realGTFromNOBJ = MEMG.getNeuronByTempNid($gtRelon.f);
        if (realGTFromNOBJ==null) {
            alert("bad");
            return;
        }

        //test if this is lower or higher
        if (this._gt!=null) {
            var existingBeLVL = this.getGtLvl();
            if (existingBeLVL > realGTFromNOBJ.be_lvl) {
                //alert("exit");
                return;
            }
        }
        this._gt = $gtRelon; //accept new one (lowest most accurate down the tree).
    }

    /**
     * 
     * @param {*} $gtRelon 
     * @returns 
     */
    setGtiRelon=function($gtRelon) {
        if (this.dead==true) return;
        //safety check for GT inherit rule.
        var realGTiFromNOBJ = MEMG.getNeuronByTempNid($gtRelon.f);
        if (realGTiFromNOBJ==null) {
            alert("bad");
            return;
        }

        //test if this is lower or higher
        if (this._gti!=null) {
            var existingBeLVL = this.getGtiLvl();
            if (existingBeLVL > realGTiFromNOBJ.be_lvl) {
                //alert("exit");
                return;
            }
        }
        this._gti = $gtRelon; //accept new one (lowest most accurate down the tree).
    }


    getGtRelon=function() { return this._gt; }
    getGtiRelon=function() { return this._gti; }
    getGtLvl=function() {
        if (this._gt==null) return -1;
        var nObj = MEMG.getNeuronByTempNid(this._gt.f);
        if (nObj==null) return -1;
        return nObj.be_lvl;
    }


    /**  
     */
    getGtiLvl=function() {
        if (this._gti==null) return -1;
        var nObj = MEMG.getNeuronByTempNid(this._gti.f);
        if (nObj==null) return -1;
        return nObj.be_lvl;
    }


    /**  
     */
    getParent = function() { 
        if(this.bes.length==0) return null;
        else return MEMG.getNeuronByTempNid(this.bes[0]); 
    }
    

    /**
     * Get runtime relons according to ANNE protocol.
     * Default types used if no types passed
     */
    getAllRelons = async($bPRINT=true, $types=null)=>{
        return await this.getRuntimeRelons($bPRINT, $types);
    }


    /**
     * Get the run-time relons according to protocol rules.
     * @param {Boolean} $bPRINT 
     * @param {Array} $types 
     * @returns 
     */
    getRuntimeRelons = async($bPRINT=true, $types=null)=> {
        var all_relons = await MEMG.getRuntimeRelons(this.nid, $types);
        var txids = [];
        for (var a=0; a < all_relons.length; a++) {
            if (txids.indexOf(all_relons[a].txid)==-1) txids.push(all_relons[a].txid);
        }

        //make sure our outs are in this runtimerelons  (all types, not just $types)
        for (var r=0; r < this.outs.length; r++) {
            if (txids.indexOf(this.outs[r].txid)==-1) {
                all_relons.push(this.outs[r]);
                txids.push(this.outs[r].r);
            }
        }  
        if ($bPRINT==true) ANNE.TALK.printRelons(all_relons);
        return all_relons;
    }


    /**
     * 
     * @param {*} $owned_tnid 
     * @returns 
     */
    _removeOwns=function($owned_tnid){
        if (this.dead==true) return;
        if (this.owns==undefined) return;
        var ii = this.owns.indexOf($owned_tnid);
        if (ii>-1) this.owns.splice(ii,1);
    }


    /**
     * 
     * @param {*} $sponsored_by_tnid 
     * @returns 
     */
    _removeSponsoredBy=function($sponsored_by_tnid) {
        if (this.dead==true) return;
        if (this.my_sponsored==undefined) return;
        var ii = this.my_sponsored.indexOf($sponsored_by_tnid);
        if (ii>-1) this.my_sponsored.splice(ii,1);
    }


    /** 
     * @param {string}
     * @author {Mr.Scatman} 
     */
    setOwner=function($owned_by) {
        if (this.dead==true) return;
        if (typeof $owned_by==="string") this.owned_by = $owned_by;
        else if (typeof $owner==="object") {
            if ($owned_by.nid!=undefined) {
                this.owned_by=$owner.nid;
            }
        }
        else if (ANNE.USER._userNObj.nid) this.owned_by = USER._userNObj.nid;
        else this.owned_by="";
    }


    /**
     * Set the value of a neuron with its set_value function
     * run-time funct to utilize any set_value functions we have on our TT
     * @author {Mr.Scatman}
     *
     */
     __setValueOfNeuron=function($beNOBJ, $TTSYSOBJ=null) {
        if (this.dead==true) return;
        if (this.ntype==NTYPE_CLASS) return;

        //set the value first..  if there is a be class set_value function.
        if (this.set_value!=undefined) this.set_value(this);
        
        else if ($beNOBJ!=null && $beNOBJ.set_value!=undefined) {
            $beNOBJ.set_value(this);
            this.set_value = $beNOBJ.set_value;
        }

        //set the value first.. if there is a TTSYSOBJ set_value function.  
        else if ($TTSYSOBJ!=null && $TTSYSOBJ.set_value!=undefined) {
            $TTSYSOBJ.set_value(this);
            $beNOBJ.set_value = $TTSYSOBJ.set_value;
            this.set_value = $beNOBJ.set_value;
        }

        //next do the validator
        if (this.format_validator!=undefined) this.format_validator(this); 
        else if ($beNOBJ!=null && $beNOBJ.format_validator!=undefined) {
            $beNOBJ.format_validator(this);    
            this.format_validator=$beNOBJ.format_validator;
            if (this.nid=="7777506478065484896" || $beNOBJ.nid=="7777506478065484896") {
                console.log("TEST HERE");
            }
        }
        else if ($TTSYSOBJ!=null && $TTSYSOBJ.format_validator!=undefined) {
            $TTSYSOBJ.format_validator(this);
            this.format_validator=$TTSYSOBJ.format_validator;
            if (this.nid=="7777506478065484896") {
                console.log("TEST HERE");
            }
        }
        return;
    }

    /**
     */
    topClass = function() {
        if (this.bes.length) {
            var nObj = MEMG.getNeuronByTempNid(this.bes[0]);
            while(nObj.be_lvl>0) {
                nObj = MEMG.getNeuronByTempNid(nObj.bes[0]);
            }
            return nObj.nid;
        }
        return this.nid;
    }



/***** COLLAB AND GARDEN STUFF ******/



    /**
     * Get the WV for a query.
     * 
     * The WV scope for a neuron is the owner of the nid.
     * This can only be extended by the nid in question having
     * adopts registered.  Adopted nids are grafted in to the worldview.
     * 
     * The exception is FREE FOR ALL.  If This is a FREE FOR ALL garden then 
     * there is no WV enforced for queries.
     * 
     * @author {Mr.Scatman}  
     */
     getWVForQuery=function() {
        if (this.emulated==true) {
            console.log("There is no results when querying emulated neurons.  They aren't real.  They were .... emulated.");
            return null;
        }
        if (this.ntype==NTYPE_INST) {
            console.log("this function is not needed to directly query instances");
            return null;
        }

        //if (this._gt!=null && this._gt.t==MEM.gt_free_for_all.temp_nid) {
           // console.log("free for all!  no worldview enforced");
           // return [];
       // }

        if (this.bCHASED_min==false && this.preloaded==false) {
            console.log("FRENLY WARNING :: YOU SHOULD HAVE ATLEAST CHASED THE MIN ON THIS NOBJ BEFORE THIS");
        }

        if (this.trusted_adopt!=null && this.trusted_adopt.length) return this.trusted_adopt.concat(this.owned_by);
        return [this.owned_by];
    }
       

    /** 
     * 
     * NOT IN USE IN EXT??? 
     * 
     * Get the valid params for collab..
     * 
     * This can help buiid guis.
     * 
     * @param {anne_neuron} - The identity we are assesssing access rights for.
     * @param {array} - An array of neurons which are the params (relns of relons) we want to check for access to.
     * @author {Mr.Scatman}
     * 
     */
     getParamsValidForCollab=async($identityNObj, $param_nids)=> {
        await this.setGovRights($identityNObj);

        //any is fine.
        if (this.egrs_tnids==null || this.egrs_tnids.length==0) return $param_nids;

        var valid_param_temp_nids = [];
        for (var p=0; p < $param_nids.length; p++) {
            if (this.egrs_tnids.indexOf($param_nids[p])!=-1) valid_param_temp_nids.push(param);
        }
        return valid_param_temp_nids;
    }


    /** 
     * Return a friendly message for collab status.
     */
    getCollabStatusMsgHTML =async()=> {
        if (this._govenor==null) await this.setGovRights(USER._userNObj);
        var statusStr = govenor.actionText[this.collab_ACTION];
        if (this.collab_ACTION==govenor.COLLAB_REQ_AUTOREJECT) {
            return "<div class='red-garden'>"+statusStr+"</div>";
        }
        else if (this.collab_ACTION==govenor.COLLAB_NOCOLLABIDENTITY) {
            return "<div class='red-garden'>"+statusStr+"</div>";
        }
        else if (this.collab_ACTION==govenor.COLLAB_REQ_NO_COLLAB_RIGHTS) {
            return "<div class='red-garden'>"+statusStr+"</div>";
        }
        else if (this.collab_ACTION==govenor.COLLAB_REQ_OK) {
            return "<div class='green-garden'>"+statusStr+"</div>";
        }
        else if (this.collab_ACTION==govenor.DELEGATE_NOW) {
            return "<div class='yellow-garden'>"+statusStr+"</div>";
        }
        else if (this.collab_ACTION==govenor.SELF_OWNED) {
            if (this.ghost==true) {
                return "<div class='ghost-garden'>"+statusStr+"</div>";
            }
            return "<div class='green-garden'>"+statusStr+"</div>";
        }
        else if (this.collab_ACTION==govenor.THIS_IS_YOU) {
            if (this.ghost==true) {
                return "<div class='ghost-garden'>"+statusStr+"</div>";
            }
            return "<div class='green-garden'>"+statusStr+"</div>";
        }
    }

    isAltBe=function($temp_nid) {
        if (this['be_'+$temp_nid]!=undefined) return true;

        //TEMPORARY AUDIT!!
        if (this.bes.indexOf($temp_nid)!=-1) {
            console.log("fix this -- why is nt there a quick be_?");
            return true;
        }

        return false;
    }

    //convienance for debugging.
    printGhosts=function(){
        var ghosts = [];
        for(var g=0; g < this.outs.length; g++) {
            var rel = this.outs[g];
            if (rel.ghost==true) ghosts.push(ghost);
        }
        ANNE.TALK.printRelons(ghosts);
    }

    printBes=function($outs=null) {
        var bes = [];
        if ($outs==null) $outs = this.outs;
        for(var p=0; p < this.outs.length; p++) {
            var rel = this.outs[p];
            if (rel.type==TYPE_BE) bes.push(rel);
        }
        ANNE.TALK.printRelons(bes);
    }

    /**
     * 
     * @param {Array} $outs - optional, if not provided, uses only outs from this neuron 
     */
    printParams=function($outs=null) {
        var params = [];
        if ($outs==null) $outs = this.outs;
        for(var p=0; p < this.outs.length; p++) {
            var rel = this.outs[p];
            if (rel.type==TYPE_PARAM) params.push(rel);
        }
        ANNE.TALK.printRelons(params);
    }


    printChildren=function(){
        for (var c=0; c < this.children.length; c++) {
            var child = MEMG.getNeuronByTempNid(this.children[c]);
            console.log("child - "+child.text);
        }
    }

    
    /**
     * For debugging specific relon types+relns.
     * @param {*} $type - the relone type
     * @param {*} $r - the temp nid for the reln (not the nid)
     */
    printRelonsTypeAndReln=function($type, $r) {
        for (var r=0; r < this.outs.length; r++) {
            if (this.outs[r].r!=$r) continue;
            if (this.outs[r].type!=$type) continue;
            ANNE.TALK.printRelon(this.outs[r]);
        }
    }


    printSysRelons=function($plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
        this.printOnlyRelons(TYPE_SYS, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1)
    }
    printSysMsgRelons=function($plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
        this.printOnlyRelons(TYPE_SYS_MSG, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1)
    }
    printOnlyRelons=function($TYPE, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
        ANNE.TALK.printOnlyRelons($TYPE, this.outs, $plainText, $bDOTEMPNIDS, $consoleColor);
    }

    printRelons = function() {
        ANNE.TALK.printRelons(this.outs);
    }


    printParentLine=function() {
        var safety=16;
        console.log("%cPRINT PARENT LINE", TALK.bgblack);
        var nObj = this;
        while(nObj!=null && ((safety--)>0)) {
            console.log("%c"+nObj.text+" %c > be lvl "+nObj.be_lvl, TALK.green, TALK.black);
            nObj = nObj.getParent();
        }
    }

    printInsAndOuts=function() {
        this.printIns();
        this.printRelons();
    }

    printIns=function() {
        ANNE.TALK.printRelons(this.ins);
    }
    printOwns=function() {
        ANNE.TALK.printFromTempNidArray(this.owns);
    }



    /** 
     * NOT IN USE FOR EXT, use via console.
     * 
     * For quick debugging of this most epic adventure.
     * 
     * @author {Mr.Scatman} 
     */
    printGardenDebug=async()=> {
        this._govenor = null;   //redo this everytime to help us test.

        await this.setGovRights(USER._userNObj, true);
        console.log("govenor action for "+this.text+" ::"+govenor.actionText[this.collab_ACTION]);
        
        if (this.delegation) ANNE.TALK.printRelons(this.delegation);

        if (this.collab_ACTION >= govenor.COLLAB_REQ_OK) {
            if( this.egrs_tnids!=null && this.egrs_tnids.length) {
                for (var c=0; c < this.egrs_tnids.length; c++) {
                    var nObj = MEMG.getNeuronByTempNid(this.egrs_tnids[c]);
                    console.log("collab param ok if param: "+nObj.text);
                }
            }
            else console.log("any reln ok for collab req");
        }
       
        if (this.collab_ACTION >= govenor.DELEGATE_NOW) {
            if (this._govenor.dr_restrict_reln_tnids!=null && this._govenor.dr_restrict_reln_tnids.length) {
                for (var p=0; p < this._govenor.dr_restrict_reln_tnids.length; p++) {
                    var nObj = MEMG.getNeuronByTempNid(this._govenor.dr_restrict_reln_tnids[p]);
                    console.log("delegation param ok if param: "+nObj.text);
                }
            }
            else {
                console.log("all relns ok for delegation");
            }
        }
    }

            
    /**
     * 
     * Can this identity collab/delegate on this nObj?
     * 
     * NOTE: used by ANNET.initInfoTabData
     * 
     * @param {anne_neuron} - The identity neuron we are assessing access for.
     * @param {boolean} - Force a redo of the govenor even if we've done it for this user.
     * 
     * @author {Mr.Scatman}
     */
     setGovRights=async($identityNObj, $bFORCEREDO=false)=> {
        if (($identityNObj instanceof anne_neuron) == false) {
            throw "you must pass an anne_neuron as the identityNObj";
        }
        return await this._setGovenorForUser($identityNObj);
    }


    /**
     * If this user is the last applied for identity in the govenor, redo it.
     * If it isn't done yet, do it.
     * 
     * @param {anne_neuron} - user neuron to apply this govenor to
     * @return {integer} - action code
     */
    _setGovenorForUser=async($userNOBJ)=> {
        this._govenor = new govenor(this);
        await this._govenor.applygardenRulesForIdentity($userNOBJ.nid);
        return this._govenor.action;
    }


    /**
     * snag the class description relon (if it exists)
     * @returns 
     */
    getClassDesc = function() {
        if (this.ntype==NTYPE_INST) return ""; //we could look at the parent for this
        var rel = ANNE.MEMG.getRelonFromWithRToType(this.outs, TYPE_SYS, ANNE.MEM.class_desc.temp_nid, TO_TYPE_BLOB);
        if (rel!=null) return rel.to_s;
    }


    anyBe = function($classNOBJ) {
        if ($classNOBJ==null) return false;
        if (this.bes.indexOf($classNOBJ.temp_nid)!=-1) return true;
        return false;
    }
    
    /**
     * 
     * @param {*} $classNOBJ 
     * @returns 
     */
    is = function($classNOBJ) {
        if ($classNOBJ==undefined || $classNOBJ==null) {
            console.log("hunt this down");
            return false;
        }
        var nid = "";

        //derp mistakes.
        if (($classNOBJ instanceof anne_neuron) == false) {
            console.log("fix this -- pass the nOBJ not the nid");
            if (typeof $classNOBJ==="string") $classNOBJ = MEMG.getNeuronByNid(nid);
            else if (typeof $classNOBJ==="object") {
                if ($classNOBJ.nObj!=undefined) {
                    $classNOBJ = $classNOBJ.nObj;
                }
            }            
            if ($classNOBJ==undefined || $classNOBJ==null) return false;
        }

        return MEMG.beIsA(this, $classNOBJ.temp_nid);
    }


    isT = function($temp_nid) {
        return MEMG.beIsA(this, $temp_nid);
    }
    

    /**
     * 
     * @param {*} $tnids 
     * @returns 
     */
    isMultiT = function ($tnids) {
        for (var t=0; t < $tnids.length; t++) {
            if (this.isT($tnids[t])==true) return true;
        }
        return false;
    }
    

    /**
     * called for the identity using ANNE...
     */
    isAndyForThisRelon=async($relon)=> {
        var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
        if (fNOBJ==null) return false;
        return await fNOBJ.userIsAndyForThisRelon(this, $relon);
    }


    /**
     * We have an ANDY with ANDY rights to a neuron..
     * but can they administer THIS particular relon?
     */
    userIsAndyForThisRelon=async($identityNOBJ, $relon)=> {
        var ttxtype =await this.validateCollabReq($identityNOBJ, $relon.r);
    
        if (ttxtype==TXB.TXTYPE_COLLAB_REQ_EXP_DEL) return true;    //we're delegated admin for this
        if (ttxtype==TXB.TXTYPE_CHILD_RELON) return true;           //we own this.
            
        return false;
    }


    /**
     * @author {Mr.Scatman}
     */ 
    validateCollabReq=async($identityNObj, $target_reln_temp_nid)=> {
        if (this._govenor == null) this._govenor = new govenor(this);
        if (this._govenor._applied_for_identity_nid==null) await this._govenor.applygardenRulesForIdentity($identityNObj.nid);
        return await this._govenor.isValidCollabReq($identityNObj.nid, $target_reln_temp_nid);    
    }



    /** 
     * EXT NO USING THIS???
     * 
     * Validate a collaboration request response.
     * 
     * Use the govenor to determine if this admin is actually and admin
     * and if the collab  request is legit.
     * 
     * @author {Mr.Scatman} 
     */
    validCollabRespAction=async($admin_nid, $tnid_action_taken, $target_reln_temp_nid)=> {
        if (this._govenor == null || this._govenor._applied_for_identity_nid==null) {
            await this.setGovRights($admin_nid);
        }
        return await this._govenor.isValidCollabResponse($admin_nid, $tnid_action_taken, $target_reln_temp_nid);    
    }


    /**
     * TO DO, what about the ANNE EPOCH DAY NUMBER? 
     */
    processAsDateNeuron=function(){
        if (this.year!=undefined || this.month!=undefined || this.day!=undefined) return;//already done.

        var rel_year = MEMG.getRelonFromWith(this.outs, TYPE_HAS, MEM.year.nid);
        if (rel_year==null) this.year = null;
        else this.year = rel_year.to_s;
        
        var rel_month = MEMG.getRelonFromWith(this.outs, TYPE_HAS, MEM.month.nid);
        if (rel_month==null) this.month=null;
        else this.month = rel_month.to_s;
        
        var rel_day = MEMG.getRelonFromWith(this.outs, TYPE_HAS, MEM.day.nid);
        if (rel_day==null) this.day=null;
        else rel_day = rel_day.to_s;

        var rel_era = MEMG.getRelonFromWith(this.outs, TYPE_HAS, MEM.era.nid);
        if (rel_era==null) this.era = MEM.era_ad.nid;
        else this.era = rel_era.to_s;
    }


    /**
     * Delete a role from a ghost clump.
     * Clean up the quick prop for the ROLE_NAME that points to an array of those relons.
     * These quick objects are created (or updated) when we run the processAsClumpNeuron
     * @param {String} $txid - txid long string 
     * @returns {Array} - the relon array of roles that are remaining in that specific role bucket after this deletion.
     */
    deleteRolefromClump=async($txid)=>{
        if (this.ghost==false) return;
        var rel_to_delete = MEMG.lookupByTXID($txid);
        var relnNOBJ = MEMG.getNeuronByTempNid(rel_to_delete.r);
        var rel_arr = this[relnNOBJ.text.toLowerCase()];
        for (var r=0; r < rel_arr.length; r++) {
            if (rel_arr[r].txid==$txid) {
                rel_arr.splice(r, 1);
                break;
            }
        }

        //delete it, there are no more here.
        if (rel_arr.length==0) delete this[relnNOBJ.text.toLowerCase()];
        await ANNE.GHOSTDATA.deleteSingleGhostRelon($txid);

        //note if in the dossier, you need to delete from all_relons to cleanup that page view.
        //it cleans up on next view auto if you don't do that.

        return rel_arr;
    }


    
    defaultRolesForClumpForNLP=async()=> {
        if (this.dead==true) return;
        //pull param roles for this action.. so we can default some extra buttons.
        var role_params = [];
        var all_relons = await MEMG.getRuntimeRelons(this.nid);

        for (var p=0; p < all_relons.length; p++) {
            var paramRelon = all_relons[p];
            if (paramRelon.type!=TYPE_PARAM) continue;
            if (MEM.isRoleR(paramRelon.r)==false) continue;
            var relNOBJ = MEMG.getNeuronByTempNid(paramRelon.r);
            var toNOBJ = MEMG.getNeuronByTempNid(paramRelon.t);
            if (relNOBJ.r==MEM.role_action.temp_nid) continue;

            var roleParam = {
                                roleNOBJ:relNOBJ, 
                                roleTempNid:relNOBJ.temp_nid,
                                role_name:relNOBJ.text, 
                                toNOBJ:toNOBJ, 
                                to_text:toNOBJ.text, 
                                found:false
                            };
            
            role_params[relNOBJ.temp_nid]=roleParam;

            if (MEM.role_standards.indexOf(relNOBJ.temp_nid)==-1) {
                roleParam.priority=1;
            }

            else roleParam.priority=0;

            //note, roles themselves can assign roles.. 
            //WHERE is connected to WHERE_POS_MOD
            //WHEN is connected to WHEN_MOD
            //we don't want to go too deep into a hole, 
            //so just look at default we already have. 
            //only TYPE_PARAMS qualify here
            for (var pp=0; pp < relNOBJ.outs.length; pp++) {
                if (relNOBJ.outs[pp].type==TYPE_PARAM) {
                    if (MEM.isRoleR(relNOBJ.outs[pp].r)==false) continue;
                    var ppRelnNOBJ = MEMG.getNeuronByTempNid(relNOBJ.outs[pp].r);  //the role assigned by the role
                    if (relNOBJ.outs[pp].r==MEM.role_action.temp_nid) continue;
                    var ppToNOBJ = MEMG.getNeuronByTempNid(paramRelon.t);
                    //add the role assigned by the role if we don't have that slot taken already by THIS
                    //clump/vb which would takes precedent.
                    if (role_params[ppRelnNOBJ.temp_nid]==undefined) {
                        var roleParam = {found:false, roleNOBJ:ppRelnNOBJ, roleTempNid:ppRelnNOBJ.temp_nid, role_name:ppRelnNOBJ.text, toNOBJ:ppToNOBJ, to_text:ppToNOBJ.text};
                        role_params[ppRelnNOBJ.temp_nid]=roleParam;
                    }
                }
            }
        }

        return role_params;
    }


    /**
     * 
     * @param {nlp_parse} $nlpP 
     * @returns 
     */
    setTense = function($nlpP=null) {
        this.tense = "present tense";
        this.tenseNOBJ = MEM.tense_present;
        this.shift_lang = false;

        if (this.ntype==NTYPE_INST) {
            
            //insts can shift the concept of a v1 with the tense relon if there
            //is no tensified concept in memory.  (eg, an action like "divorce" does not have a "divorced" subclass)
            //if the user wants "divorced" then can shift divorce with a R_TENSE relon in the instance.
            //They could also create a "divorced" as a sub-class of "divorce" and "divorced" would have a vb_form of v2.
            var rel_tense = MEMG.getRoleRelon(this.outs, MEM.role_tense.temp_nid);
            if (rel_tense!=null) {
                this.tenseNOBJ = MEMG.getNeuronByTempNid(rel_tense.t);
                this.tense = this.tenseNOBJ.text;
                this.shift_lang = true;
                if ($nlpP!=null) $nlpP.setTense(this.tenseNOBJ);
                return this.tense;
            }
            
            if (this.bes.length==0) return "";
            var parentNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
            this.tense = parentNOBJ.setTense();
            this.tenseNOBJ = parentNOBJ.tenseNOBJ;
            this.shift_lang = false;
            if ($nlpP!=null) $nlpP.setTense(this.tenseNOBJ);
            return this.tense;
        }

        //actual tense relon on this concept.
        //this will be a ROLE relon.
        var rel_tense = MEMG.getRoleRelon(this.outs, MEM.role_tense.temp_nid);
        if (rel_tense!=null) {
            this.tenseNOBJ = MEMG.getNeuronByTempNid(rel_tense.t);
            this.tense = this.tenseNOBJ.text;
            this.shift_lang = false;
            if ($nlpP!=null) $nlpP.setTense(this.tenseNOBJ);
            this.setVs();
            return this.tense;
        }

        //forms (v1/v2/v3/v4/v5) are TYPE_BE relons.
        //OTHER forms (not the one this is) are TYPE_HAS
        var rel_vbForm = MEMG.getBeRelon(this.outs, MEM.vb_form.temp_nid);
        if (rel_vbForm!=null) {
            var toNOBJ = MEMG.getNeuronByTempNid(rel_vbForm.t); //the vb form..

            //the tense role is on the vb form.
            var tense_role = MEMG.getRoleRelon(toNOBJ.outs, MEM.role_tense.temp_nid);
            if (tense_role!=null) {
                this.tenseNOBJ = MEMG.getNeuronByTempNid(tense_role.t);
                this.tense = this.tenseNOBJ.text;
                this.shift_lang = false;
                if ($nlpP!=null) $nlpP.setTense(this.tenseNOBJ);
                return this.tense;
            }
        }

        //simple tense relon (not v1,v2,v3,v4,v5)
        var rel_tense = MEMG.getBeRelon(this.outs, MEM.tense.temp_nid);
        if (rel_tense) {
            this.tenseNOBJ = MEMG.getNeuronByTempNid(rel_tense.t);
            this.tense = this.tenseNOBJ.text;
            this.shift_lang = false;
            if ($nlpP!=null) $nlpP.setTense(this.tenseNOBJ);
            return this.tense;
        }

        this.setVs();
        
        return this.tense;
    }


    /**
     * 
     */
    processAsDate=function(){
        if (this.adate!=null) return;
        this.adate = new anne_date();
        this.adate.loadFromNeuron(this, false);
    }


    /**
     * 
     * @param {Array} $all_relons - if passsed in, use it.  Otherwise we gen it in dprc 
     * @param {Boolean} $bDODEBUG - print debug
     * @returns 
     */
    processActionNeuron=async($all_relons=[], $bDODEBUG=false, $skipRoles=[], $el_printouts=null)=>{
        this._roles = [];
        this._param_roles = [];
        this._meaning_roles = [];

        $all_relons = await this.defaultParamRolesForClump($all_relons);
        
        var talkobj = await this.processAsClumpNeuron(null, $skipRoles, $el_printouts);
        
        if ($bDODEBUG==true) this.printRoleDebug();
                
        //nestd meta meaning is a meaning defined by another clump and not by embedded roles.
        //this is done when a concept's definition already exists in another concept.
        //ie, a synonym-like concept where this concept has no definition and relies upon the other for its defintion.
        //This can often be the case in idioms that reference other idioms or already-defined concepts.
        var meta_meaning = MEMG.getRelonByTypeAndR(this.outs, MEM.role_meta_meaning.temp_nid, TYPE_CLUMP);
        if (meta_meaning) {
            var MM = MEMG.getNeuronByTempNid(meta_meaning.t);
            var all_mm_relons = await MM.getRuntimeRelons(false);
            var meta_talkobj = await MM.processActionNeuron(all_mm_relons, $bDODEBUG, $skipRoles, $el_printouts);
            talkobj.meta_meaning_talk = meta_talkobj;
            talkobj.meta_meaning_NOBJ = MM;
            talkobj.meta_meaning_rel = meta_meaning;
        }

        return talkobj;
    }


    /**
     * 
     * @param {*} $all_relons 
     * @param {*} $bDODEBUG 
     * @param {*} $skipRoles 
     * @param {*} $el_printouts 
     * @returns 
     */
    processIdiom=async($all_relons=[], $bDODEBUG=true, $skipRoles=[], $el_printouts=null)=>{
        this._roles = [];
        this._param_roles = [];
        this._meaning_roles = [];
        $all_relons = await this.defaultParamRolesForClump($all_relons);
        var talkobj = await this.processAsClumpNeuron(null, $skipRoles, $el_printouts);
        if ($bDODEBUG==true) this.printRoleDebug();
        return talkobj;
    }


    /**
     * 
     */
    printRoleDebug = async()=> {
        var typelbl = "class";
        if (this.ntype==NTYPE_INST) typelbl = "inst";
        else if (this.ntype==NTYPE_UNKNOWN) typelbl = "unknown type";

        //console.log("_roles", this._roles);
        if (this._roles!=null) ANNE.TALK.printRelons(this._roles);
        //console.log("_param_roles", this._param_roles);
        if (this._param_roles!=null) ANNE.TALK.printRelons(this._param_roles);
        //console.log("_meaning_roles", this._meaning_roles);
        if (this._meaning_roles!=null) ANNE.TALK.printRelons(this._meaning_roles);
        //console.log("%c '"+this.text+"'%c ntype:"+typelbl+"%c tense:"+this.tense+" %c belvl:"+this.be_lvl, TALK.bggreenbig, TALK.bgyellow, TALK.bggreen, TALK.bgred);    
        //console.log("_nested_meaning_roles", this._nested_meaning_roles);

        if (this.is(MEM.ec_action)==true) {
            this.printVs();
        }
    }


    /**
     * set the verb forms to easiy reference/map back to others.
     */
    setVs = function() {
        var setV = function($this, $var, $tnid) {
            var vNOBJ = MEMG.getNeuronByTempNid($tnid);
            $this[$var]=vNOBJ;
            $this[$var+"_txt"]=vNOBJ.text;
        }

        //the form of this neuron. (could be any of the 5)
        if (this.vb_form_rel==undefined) {
            this.vb_form_rel = MEMG.getBeRelon(this.outs, MEM.vb_form.temp_nid);
            if (this.vb_form_rel!=null) {
                if (this.vb_form_rel.t==MEM.vb_v1.temp_nid) setV(this, "v1", this.temp_nid);
                else if (this.vb_form_rel.t==MEM.vb_v2.temp_nid) setV(this, "v2", this.temp_nid);
                else if (this.vb_form_rel.t==MEM.vb_v3.temp_nid) setV(this, "v3", this.temp_nid);
                else if (this.vb_form_rel.t==MEM.vb_v4.temp_nid) setV(this, "v4", this.temp_nid);
                else if (this.vb_form_rel.t==MEM.vb_v5.temp_nid) setV(this, "v5", this.temp_nid);
            }
        }

        //creates .v1, .v2, .v3, .v4 and .v5 (with _txt)
        for (var v=1; v <=5; v++) {
            if (this["v"+v]==undefined || this["v"+v]==null) {
                this["v"+v] = null;
                var vrel = MEMG.getHasRelon(this.outs, MEM['vb_v'+v].temp_nid);
                if (vrel!=null) setV(this, "v"+v, vrel.t);
            }
        }

        //if this wasnt v1, we can get the vs from v1.
        if (this.vb_form_rel!=null && this.vb_form_rel.t!=MEM.vb_v1.temp_nid) {
            if(this.v1!=null) {
                this.setVsFromParent(this.v1)
            }
        }
    }


    /**
     * 
     * @param {*} $parentNOBJ 
     */
    setVsFromParent=function($parentNOBJ) {
        $parentNOBJ.setVs();
        for (var v=1; v <=5; v++) {
            if ($parentNOBJ["v"+v]!=undefined && $parentNOBJ["v"+v]!=null) {
               this["v"+v]=$parentNOBJ["v"+v];
               this["v"+v+"_txt"]=$parentNOBJ["v"+v+"_txt"];
            }
        }
    }


    /**
     * 
     */
    printVs = function() {
        var obj = {v1:this.v1, v1_txt:this.v1_txt, v2:this.v2, v2_txt:this.v2_txt, 
                    v3:this.v3, v3_txt:this.v3_txt, v4:this.v4, v4_txt:this.v4_txt, 
                    v5:this.v5, v5_txt:this.v5_txt};
        if (this.v1) console.log("v1 "+this.v1+" "+this.v1_txt);
        if (this.v2) console.log("v2 "+this.v2+" "+this.v2_txt);
        if (this.v3) console.log("v3 "+this.v3+" "+this.v3_txt);
        if (this.v4) console.log("v4 "+this.v4+" "+this.v4_txt);
        if (this.v5) console.log("v5"+this.v5+" "+this.v5_txt);
        console.log(obj);
        return obj;
    }


    /**
     * Here we get PARAM relons with ROLE reln values.
     * These are telligng us what roles this action/posn/is assigns.
     * 
     * Huge note:
     * Instances are actual clumps (not classes) that will most often be a sub-class of an action neuron.
     * The TYPE_ROLEs specified in the upline are conceptually defining for the concept.
     * They are NOT inherited by the instance.  ONLY TYPE_ROLE  ROLE_TENSE is inherited by the instance.
     * While the TYPE_ROLEs of the parent do applyl to understanding the concept, the new instance clump will be 
     * providing its own ROLES to specify the semantic parts of the information it is encoding.
     * 
     * Instance DO INHERIT ALL THE TYPE_PARAMS.  These are critical to informing us what ROLES this action typically assigns.
     * 
     * @param {Array} $all_relons 
     * @returns 
     */
     defaultParamRolesForClump=async($all_relons=[])=> {
        if (this.dead==true) return;
        var all_relons = null;
        
        if ($all_relons.length==0) {
            all_relons = await MEMG.getRuntimeRelons(this.nid);
        }
        else all_relons = $all_relons;

        var wanted_outs = [];

        for (var p=0; p < all_relons.length; p++) {
            var relon = all_relons[p];
            if (relon.temp==true && relon.ghost==false) continue;   //temp relon for gui help only
            if (relon.type!=TYPE_ROLE && relon.type!=TYPE_PARAM) continue; 
            if (MEM.isRoleR(relon.r)==false) continue;

            var relNOBJ = MEMG.getNeuronByTempNid(relon.r);

            //class neuron assigned roles are MEANING roles.
            if (relon.type==TYPE_ROLE) {
                if (relon.flathexint==undefined) {
                    relon.flathexint=0;
                    relon.flathex="0";
                }
                wanted_outs.push(relon);
                continue;
            }
            else if (relon.type==TYPE_PARAM) {


                this.addParamRole(relon);

                //note, roles themselves can assign roles.. 
                //WHERE is connected to WHERE_POS_MOD
                //WHEN is connected to WHEN_MOD
                //we don't want to go too deep into a hole, 
                //so just look at default we already have. 
                //only TYPE_PARAMS qualify here
                for (var pp=0; pp < relNOBJ.outs.length; pp++) {
                    var pprel  = relNOBJ.outs[pp];
                    if (pprel.type==TYPE_PARAM) {
                        if (MEM.isRoleR(pprel.r)==false) continue;
                        if (pprel.r==MEM.role_action.temp_nid) continue;
                        if (this.getParamRole(pprel.f, pprel.r)==null) this.addParamRole(pprel);
                    }
                }
            }
        }

        //audit (delete when good)
        wanted_outs.sort(this.__sortRolesByFlathexint);

        for (var p=0; p < wanted_outs.length; p++) {
            var relon = wanted_outs[p];
            
            var fromNOBJ = MEMG.getNeuronByTempNid(relon.f);
            var relNOBJ = MEMG.getNeuronByTempNid(relon.r);

            //if this neuron we are looking at is a class neuron..
            //All TYPE_ROLEs are meaning roles.
            if (fromNOBJ.ntype==NTYPE_CLASS) {
                //console.log(relNOBJ.text);
                this.addMeaningRole(relNOBJ.text, relon);
                continue;
            }

            //if this neuron we are looking at is an inst neuron
            else {

                //this relon is inherited
                //it is a meaning role for the action vb.
                if (relon.f!=this.temp_nid) {
                    this.addMeaningRole(relNOBJ.text, relon);
                    
                    //tense is the ONLY role that inherits as a ._role for 
                    //instances.  An instance of a past tense action is a sentence/thought
                    //that is in the past tense.
                    if (relNOBJ.r==MEM.role_tense.temp_nid) {
                        await this.addRole(relNOBJ.text, relon);
                    }
                }

                //this role relon is supplied on the inst neuron itself.
                //it is semantic expression of the meaning of the inst clump.
                //this is what we store in _roles.
                else await this.addRole(relNOBJ.text, relon);
            }

            //only PARAM relons below here.
            MEMG.getNeuronByTempNid(relon.t);

            //if we have role_tense set at this point (_roles set the this.["role_xxx"] property)
            //we do not add in the PARAM ROLE.
            if (relon.r==MEM.role_tense.temp_nid && ANNE.UTILS.isset(this.role_tense)==true) {
                continue;
            }

            //the toNOBJ can now get upgraded/overriden... by the TYPE_SYS _default_class relon.
            //this can be the case when downline subclasses are more refined and specific with their params.
            //var default_class_nid = this.getDefaultClassForParam(relon.txid);
            //if (default_class_nid!=null) {
                //toNOBJ = MEMG.getNeuronByNid(default_class_nid);
            //}

        }

        return all_relons;
    }


    /**
     * 
     * @param {*} $a 
     * @param {*} $b 
     * @returns 
     */
    __sortRolesByFlathexint=function($a, $b) {
        if($a.flathexint < $b.flathexint) return -1;
        else if($a.flathexint > $b.flathexint) return 1;
        return 0;
    }
    

    getRolesArr = async($gpat=null, $skipRoles=[])=>{
        await this.processAsClumpNeuron($gpat, $skipRoles);
        return this._roles;
    }

    /**
     * 
     * @param {Object} $gpat 
     * @param {Array} $skipRoles 
     * @returns {anne_babytalk}
     */
     processAsClumpNeuron = async($gpat=null, $skipRoles=[])=> {  
        if (this.dead==true) return;

        if ($gpat==null) $gpat = TALK.eng_pat1;
        this._roles = [];
         
        //ANNE.TALK.printRelons(this.outs);
         
        var wanted_outs = [];
        for (var r=0; r < this.outs.length; r++) {
            var rObj = this.outs[r];
            if (rObj.type!=TYPE_ROLE && rObj.type!=TYPE_PARAM) continue; //hmm relook at && rObj.type!=TYPE_CLUMP

            //we do not add ROLES inherited from parents here if we have an instance neuron
            //those roles are conceptually defining for the concept (action) being used.
            //they are very important in understanding but for our inst we will supply our own values for the clump's roles.
            if (this.ntype==NTYPE_INST && rObj.r!=MEM.role_tense.temp_nid  && rObj.f!=this.temp_nid) continue;
 
            if (MEM.isRoleR(rObj.r)!=-1) {
                if (rObj.flathexint==undefined) {
                    rObj.flathexint=0;
                    rObj.flathex="0";
                }
                wanted_outs.push(rObj);
            }
        }

        wanted_outs.sort(this.__sortRolesByFlathexint);
        for (var r=0; r < wanted_outs.length; r++) {
            var rObj = wanted_outs[r];
            var roleNOBJ = MEMG.getNeuronByTempNid(rObj.r);

            var rAUDIT = ANNE.MEMG.getNeuronByNid(rObj.reln);
            //console.log("audit "+rAUDIT.text);
           
            //param roles are inherited from uplines.
            //note, to get full params you should run the defaultParamRolesForClump!
            if (rObj.type==TYPE_PARAM) this.addParamRole(rObj);
            else {
                if (rObj.f==this.temp_nid) {
                    await this.addRole(roleNOBJ.text, rObj);
                }
                else {
                    this.addMeaningRole(roleNOBJ.text, rObj);
                }
            }
        }
        

        this.bPROCESSEDCLUMP = true;

        this.setTense();

        //set the default action role if it isnt set in the clump.
        //we can inherit from the parent since that is the verb class.
        //INSTANCE CLUMPS are using the PARENT CLASS ACTION ALREADY.

        //NOTE:
        //CLASS CLUMPS use TYPE_ROLE->ROLE_ACTION to provide the verb of the definition of the clump.
        //Any role that a class clump (action vb) sets is the semantic DEFINITION of that clump!
        //Instances should NOT inherit their parent action's set TYPE_ROLE relons, only the class PARAM roles!!!
        //The inherited TYPE_ROLE relons will go into the .meaning_roles array.   
        if (ANNE.UTILS.isset(this.role_action)==false) {
            if(this.bes.length>0) {
                var beNOBJ = MEMG.getNeuronByTempNid(this.bes[0]);
                if (beNOBJ && MEM.isECAction(beNOBJ, false)==true) {
                    
                    //if beNOBJ is a v1 and this is a v2/v3/v4/v5, we want the ROLE_ACTION of 
                    //the parent vb of the beNOBJ!
                    beNOBJ.setTense();
                    if (beNOBJ.tense=="present tense" && this.tense!="present tense") {
                        //not for saving, only made on-demand when viewing this action.
                        var gparent = beNOBJ.getParent();
                        var temp_rel = MEM.makeTempRelon(this, TYPE_ROLE, MEM.role_action.nid, TO_TYPE_ANNE, gparent.nid);
                        temp_rel.flathexint="0";                   
                    }            
                    else {
                        //not for saving, only made on-demand when viewing this action.
                        var temp_rel = MEM.makeTempRelon(this, TYPE_ROLE, MEM.role_action.nid, TO_TYPE_ANNE, beNOBJ.nid);
                        temp_rel.flathexint="0";
                    }
                    await this.addRole("role_action", temp_rel);
                    this.addMeaningRole("role_action", temp_rel);
                }
            }
        }
    

        //walk this back if this has a tense on it. We want the present tense
        //root verb.  Any non-stems are sub-classes of their parent stem.
        //ie, elected has a parent of elect.  They are both classes still..
        //action clumps (or posn/fact/feelz/thought clumps) will all be instances of the 
        //verb they are using.  If an instance of a non stem (that is ok, we get tense).
        //stem can be used with tense role too for the same conceptual result.
        //if (this.role_action!=undefined && this.role_action!=null && this.role_action.length) {
            //this.stem = MEMG.getNeuronByTempNid( this.role_action[0].t );
        //}
        //else this.stem = null;
        if (this.ntype==NTYPE_CLASS) return ANNE.TALK.verbalizeClassActionDef(this);

        return ANNE.TALK.verbalizeInstActionClump(this, $gpat, $skipRoles, true, false);
    }



    /**
     */
    updateGhostParent=async($newToNid, $instOrClass=MEM.inst)=> {
        if (this.dead==true) return null;
        var be_relon  = MEMG.getRelonFromWithRToType(this.outs, TYPE_BE, $instOrClass.temp_nid, TO_TYPE_ANNE);
        if (be_relon!=null) {
            var annihilated_txids = [];
            MEMP.annihilateRelon_RAM(be_relon, annihilated_txids);
            await DB_local.___delete(annihilated_txids, "ghost_relons");
            console.log(this.bes);
            console.log(this.be);
            this.bes=[];
            this.be="";
        }
        return await ANNE.GHOSTDATA.new_ghost_be(this, $instOrClass.nid, TO_TYPE_ANNE, $newToNid);
    }

    /**
     * not in use in the EXT
     */
    ghostEditName=async($text)=>{
        if (this.dead==true) return;
        var be_relon  = MEMG.getRelonFromWithRToType(this.outs, TYPE_BE, MEM.name.temp_nid, TO_TYPE_STRING);
        if (be_relon!=null) {
            var annihilated_txids = [];
            MEMP.annihilateRelon_RAM(be_relon, annihilated_txids);
            await DB_local.___delete(annihilated_txids, "ghost_relons");
            console.log(this.bes);
            console.log(this.be);
        }
        await ANNE.GHOSTDATA.new_ghost_be(this, MEM.name.nid, TO_TYPE_STRING, $text);
        this.text=$text;
    }


    //cleanup our _roles and [role_name] arrays...
    //even though in most cases we just regen them as needed.
    removeRole=function($p, $r){
        if (this.dead==true) return;
        if (this._roles==null) return;
        UTILS.removeRelonByTxid(this._roles, $r.txid);
        if (this[$p]!=undefined) {
            UTILS.removeRelonByTxid(this[$p], $r.txid);
        }
    }


    /**
     * Find the role in the $arr that has a .flathex = to $mod_of_hex
     * @param {*} $arr 
     * @param {*} $mod_of_hex 
     */
    _getHexRole = function($arr, $mod_of_hex) {
        for (var a=0; a < $arr.length; a++) {
            if ($arr[a].flathex==undefined) continue; //never should be th case!
            if ($arr[a].flathex == $mod_of_hex) return $arr[a];
        }
        return null;
    }


    /**
     * add a role to a helpful prop holder and also the roles array.
     * 
     * @param {*} $role_text 
     * @param {*} $rel 
     * @returns 
     */
    addRole=async($role_text, $rel)=> {
        if (this.dead==true) return;    //what is this??? lolz
        if ($rel==null) return;
        if (this._roles==null) this._roles = [];
        if ($rel.flathexint!=undefined && $rel.flathex==undefined) {
            $rel.flathex = $rel.flathexint.toString(16);
            $rel.semantic_lvl = $rel.flathex.length;
            //console.log("flathexint:"+$rel.flathexint+" flathex:"+$rel.flathex+" lvl:"+$rel.semantic_lvl);
            ANNE.TALK.printRelon($rel);
        }
        var parent_role = null;
        if ($rel.semantic_lvl!=undefined && $rel.semantic_lvl !=1) {
            if (this._nested_roles==undefined) this._nested_roles = [];
            UTILS.addUnique(this._nested_roles, "txid", $rel);
            $rel.mod_of_hex = $rel.flathex.substr(1);//(peel off left most val -- this is a mode of what is remaining.)
            if ($rel.mod_of_hex.length==1) parent_role  = this._getHexRole(this._roles, $rel.mod_of_hex);
            else parent_role = this._getHexRole(this._nested_roles, $rel.mod_of_hex);
            
            if (parent_role==null) {
                var Gmod_of_hex = $rel.mod_of_hex.substr(1);
                if (Gmod_of_hex.length==1) {
                    var gparent_role = this._getHexRole(this._roles, Gmod_of_hex);
                    var emu_mod_role_name = MEMG.getNeuronByTempNid(gparent_role.r).text+"_mod".toLowerCase(); 
                }
                else {
                    var gparent_role = this._getHexRole(this._nested_roles, Gmod_of_hex);
                    var emu_mod_role_name = MEMG.getNeuronByTempNid(gparent_role.r).text+"_mod".toLowerCase();
                }

                //parent role will need to point to a new emulated neuron (only local RAM)
                var gparent_toNOBJ = MEMG.getNeuronByTempNid(gparent_role.t);
                
                var emu_compound = MEM.makeTempNObjOnly(MEM.ANNE_ACCTID, NTYPE_CLASS, "", gparent_toNOBJ, true, false);
                
                var emu_role = MEM[emu_mod_role_name];
                var fromNOBJ = MEMG.getNeuronByTempNid(gparent_role.f);
                var emu_rel = MEM.makeTempRelon(fromNOBJ, TYPE_ROLE, emu_role.nid, TO_TYPE_ANNE, emu_compound.nid, false);
                emu_rel.flathex = $rel.mod_of_hex;
                emu_rel.flathexint = parseInt($rel.mod_of_hex, 16);
                emu_rel.semantic_lvl = emu_rel.flathex.length;
                emu_rel.nameless = true;
                console.log("(EMULATED) flathexint:"+emu_rel.flathexint+" flathex:"+emu_rel.flathex+" lvl:"+emu_rel.semantic_lvl);
                if ($rel.mod_of_hex.length==1) UTILS.addUnique(this._roles, "txid", emu_rel);
                else UTILS.addUnique(this._nested_roles, "txid", emu_rel);
                emu_rel.mods = new Object();
                emu_rel.mods[$rel.flathex]=$rel;   //yay!

                //make sure we link from gparent to the new MOD role we created 
                //(that was not made in the ANNODE mem since the mods themselves in combination ARE the mod!
                //note -- this allows for simple use of a ROLE_X_MOD in annode mem OR for compound mods the use of the flathexint!
                if (gparent_role.mods==undefined) gparent_role.mods=new Object();
                if (gparent_role.mods[emu_rel.flathex]==undefined) gparent_role.mods[emu_rel.flathex]=emu_rel;
                return;
            }

            if (parent_role.mods==undefined) parent_role.mods = new Object();
            if (parent_role.mods[$rel.flathex]==undefined) parent_role.mods[$rel.flathex]=$rel;
            return;
        }

        UTILS.addUnique(this._roles, "txid", $rel);
        
        //only the _role relons will store objects in the neuron for quick access.
        //param and meaning arrays can be iterated through.
        if (this[$role_text]!=undefined) UTILS.addUnique(this[$role_text], "txid", $rel);
        else this[$role_text]=[$rel];
        return;
    }

    
    /**
     * -1, role, but no match.
     * 0, no role
     * 1, match.
     * @param {*} $role_text 
     * @param {*} $nid 
     * @returns 
     */
    hasRole=function($role_text, $nid) {
        $role_text = $role_text.toLowerCase();
        if (this[$role_text]==undefined) return 0;
        
        //if any in the array match, it is a match.
        for (var r=0; r < this[$role_text].length; r++) {
            if (this[$role_text][r].to_s==$nid) {
                return 1;
            }
        }

        //if roles exist but don't match, it is a NO MATCH (contradict)
        return -1;
    }


    /**
     * Note, this is not conceptually testing different date neurons
     * for the same value.  I don't need to support that.
     * Here we are testing for min-union matches with no contraction.
     * 
     * @param {*} $nid 
     * @returns 
     */
    minDateUnionMatch = function($nid) {

        //load up our adate vals
        this.processAsDate();

        var otherTimeValueNOBJ = ANNE.MEMG.getNeuronByNid($nid);
        otherTimeValueNOBJ.processAsDate();
        var other_ad = otherTimeValueNOBJ.adate;

        //find out the lowest common date/type we have to compare.
        //we want a "year" to give a positive match
        var lowest_common_lev = ANNE.MEM.gregorian_date.temp_nid;

        var getLowerLev=function($temp_nid) {
            if ($temp_nid==ANNE.MEM.monthyear.temp_nid) {
                lowest_common_lev = ANNE.MEM.monthyear.temp_nid;
            }
            else if ($temp_nid==ANNE.MEM.daymonth.temp_nid) {
                lowest_common_lev = ANNE.MEM.daymonth.temp_nid;
            }
            else if ($temp_nid==ANNE.MEM.integer.temp_nid) {
                lowest_common_lev = ANNE.MEM.integer.temp_nid;
            }
        }

        getLowerLev(this.bes[0]);
        getLowerLev(otherTimeValueNOBJ.bes[0]);
        
        if (lowest_common_lev==ANNE.MEM.gregorian_date.temp_nid) {
            //comparing date vs date -- must be exact date match.
            if (otherTimeValueNOBJ.bes[0]==ANNE.MEM.gregorian_date.temp_nid) {
                if ($nid!=this.nid) return false;
                return true;
            }
        }

        //comparing date vs date part.
        else if (lowest_common_lev==ANNE.MEM.monthyear.temp_nid) {
            if (other_ad.sameMonthYear(this.adate)==true) return true;
            return false;
        }

        else if (lowest_common_lev==ANNE.MEM.daymonth.temp_nid) {
            if (other_ad.sameDayMonth(this.adate)==true) return true;
            return false;
        }

        else if (lowest_common_lev===ANNE.MEM.integer.temp_nid) {
            if (other_ad.sameYear(this.adate)==true) return true;
            return false;
        }
        else return false;
    }


    /**
     * 
     * @param {*} $role_text 
     * @param {*} $nid 
     * @returns 
     */
    hasTimeRoleNoContradict=async($role_text, $nid)=> {
        $role_text = $role_text.toLowerCase();
        if (this[$role_text]==undefined) return 0;

        //if any in the array match, it is a match.
        for (var r=0; r < this[$role_text].length; r++) {
            if (this[$role_text][r].to_s==$nid) {
                return 1;
            }

            //see if one is a year, and one is a date
            else {
                var toNOBJ = ANNE.MEMG.getNeuronByNid(this[$role_text][r].to_s);
                if (toNOBJ.bes[0].length==0 || (toNOBJ.bes[0]!=ANNE.MEM.integer.temp_nid && toNOBJ.outs.length<5)) {
                    await ANNE.API.getNid(toNOBJ.nid);
                }
                if (toNOBJ.minDateUnionMatch($nid)==true) {
                    return 2;   //2 = new info.
                }
            }
        }

        //if roles exist but don't match, it is a NO MATCH (contradict)
        return -1;
    }



    /**
     * Add a meaning role.
     *
     * TYPE_ROLE relons are meaning roles.  They document the semantic meaning
     * of a neuron concept.
     * 
     * @param {*} $role_text 
     * @param {*} $rel 
     */
    addMeaningRole=function($role_text, $rel) {
        if (this.dead==true) return; //what is this??? lolz
        if ($rel==null) return;
        if (this._meaning_roles==null) this._meaning_roles = [];

        if ($rel.flathexint!=undefined && $rel.flathex==undefined) {
            $rel.flathex = $rel.flathexint.toString(16);
            $rel.semantic_lvl = $rel.flathex.length;
            console.log("flathexint:"+$rel.flathexint+" flathex:"+$rel.flathex+" lvl:"+$rel.semantic_lvl);
            ANNE.TALK.printRelon($rel);
        }
        if ($rel.semantic_lvl!=undefined && $rel.semantic_lvl !=1) {
            if(this._nested_meaning_roles==undefined) this._nested_meaning_roles = [];
            UTILS.addUnique(this._nested_meaning_roles, "txid", $rel);
            $rel.mod_of_hex = $rel.flathex.substr(1);//(peel off left most val -- this is a mode of what is remaining.)
            return;
        }

        UTILS.addUnique(this._meaning_roles, "txid", $rel);

        if (this["_meaning_"+$role_text]!=undefined) {
            var arr = this["_meaning_"+$role_text];
            for (var a=0; a < arr.length; a++) {
                
                //dupe role  R_XX-> NID .. from same from?
                //sometimes when we define things we repeat a role that the 
                //parent already had on it.
                if (arr[a].r==$rel.r && arr[a].t==$rel.t) {
                    if($rel.f==arr[a].f) continue;
                    
                    //keep the lower level one.
                    else {
                        var relFromNOBJ = MEMG.getNeuronByTempNid($rel.f);
                        var relArrNOBJ = MEMG.getNeuronByTempNid(arr[a].f);
                        if (relFromNOBJ.be_lvl>relArrNOBJ.be_lvl) {
                            arr[a] = $rel;
                            return;
                        }
                        else {
                            return; //keep the one we have, no need for $rel, it is same relon but higher up the tree. 
                        }
                    }
                }
            }
            UTILS.addUnique(this["_meaning_"+$role_text], "txid", $rel);
        }
        else this["_meaning_"+$role_text]=[$rel];
    }


    /**
     * Looking for a specific param relon from a specific from.
     * @param {int} $f - temp_nid of from 
     * @param {ing} $r - temp_nid of reln
     * @returns {Object} - return Object, if exists, or null.
     */
    getParamRole=function($f,$r) {
        var fromNOBJ = MEMG.getNeuronByTempNid($f);
        return MEMG.getParamRelon(fromNOBJ.outs, $r);
    }

    addParamRole=function($rel) {
        if (this.dead==true) return;//what is this??? lolz
        if ($rel==null) return;
        if (this._param_roles==null) this._param_roles = [];
        UTILS.addUnique(this._param_roles, "txid", $rel);    
    }


    extractRoles = function($roleNOBJ) {
        if (this.dead==true) return;
        var roles = [];
        if (this._roles==null) return roles;
        for (var r=0; r < this._roles.length; r++) {  
            var role = this._roles[r];
            if (role.r==$roleNOBJ.temp_nid) {
                var toNOBJ = MEMG.getNeuronByTempNid(role.t);
                roles.push(toNOBJ);
            }
        } 
        return roles;
    }


    //allow for multiple roles of the same thng in a clump (optimizing storage for rights).
    //loop all roles looking for the target role and target value.
    validateEquals = function($prop, $val) {
        if (this[$prop]==undefined) return false;
        var role_rObjs = this[$prop];
        ANNE.TALK.printRelons(role_rObjs, "(validateEquals)");
        for (var r=0; r < role_rObjs.length; r++) {  
            var to_s = role_rObjs[r].to_s;
            if (to_s==$val) return true; 
        } 
        return false;
    }


    /**
     * @param {*} $bLOAD 
     * @returns 
     */
    setLifeRange = async($bLOAD=true)=>{
        if (this._liferange!=null) return;
        var birthdeath = ANNE.MEMG.getBirthDeath(this);
        this._liferange = new anne_date_range();
        if ($bLOAD==true)  await this._liferange.loadFromDates(birthdeath.date1, birthdeath.date2)
        return;
    }
    
    printNameWithLR=async($class="")=> {
        if (this._liferange==null) {
            await this.setLifeRange(true);
        }
        
        if (this._liferange==null) {
            return this.bestName();
        }
        
        if ($class!="") return this.bestName()+"<span class='"+$class+"'>"+this._liferange.simpleDisplay+"</span>";
        else return this.bestName()+this._liferange.simpleDisplay;
    }
    
    getLR=function(){
        if (this._liferange==null) return "";
        return this._liferange.simpleDisplay;
    }


    /**
     * 
     * @param {*} $bCHASE 
     * @param {*} $bMUSTPULL 
     * @param {*} $types 
     * @returns 
     */
    loadOuts = async($bCHASE=false, $bMUSTPULL=false, $types=[0,1,2,3,4,5,6,7,8,9,10])=> {
        if ($bCHASE==true) {
            return await ANNE.API.getOuts(this.nid, $bMUSTPULL, $types);
        }
        var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts", $nid);
		await ANNE.IDB.queryForOuts(phObj, this.nid, $types);
        return;
    }

    
    /**
     * Quick attempt to match some (any) text.
     * @param {*} $lookupval 
     * @returns 
     */
    textMatch=function($lookupval) {
        var lower = $lookupval.toLowerCase();
        if (lower==this.text) return true;
        
        if(this.aka_lower==undefined || this.aka_lower.length==0) return false;
        for (var a=0; a < this.aka_lower.length; a++) {
            if (lower==this.aka_lower[a]) return true;
        }
        return false;
    }
}



/**
 * This started a clump for service neurons only..
 * but i am extending it for marriage as well.
 * 
 * 
 */

class wiki_clump_neuron {
    constructor($rand, $aec=null)  {
        this.rand = $rand;
        this.aec = $aec;
        this.nObj = null;
        this.term_start = "";   //text for processing
        this.term_end = "";     //text for processing
        this.term_themes = [];   //text for processing
        this.term_affected_things = [];
        this.term_themes_mod = [];

        //all exepcted to be people, if set.
        this.preceded_by_text = null; //role_prev
        this.preceded_by_nObj = null;
        this.ROLE_PREV = null;
        
        this.succeeded_by_text = null; //role_next
        this.succeeded_by_nObj = null;
        this.ROLE_NEXT = null;

        this.appointed_by_text = null; 
        this.appointed_by_nObj = null;

        //resolved to neurons -- as to make relons.
        this.start_year = "";
        this.ROLE_START = null;         //the start of the term
        this.end_year = "";
        this.ROLE_END = null;           //the end of the term
        this.ROLE_ACTORS = [];           //the office holder, or in marriage, the people married.
        this.ROLE_THEME_THINGS = [];    //the office/role held
        this.ROLE_AFFECTED_THINGS = [];
        this.ROLE_THEME_THINGS_MOD = [];    //the office/role held
        this.text_assocs = [];
        this.nObj_political_party=null; //the party (if applicable)
        
        this.nObj_country = null;      //whatever level of polity applicable.  (city, state, district, country etc)
        
        this.nObj_location = null;      //whatever level of polity applicable.  (city, state, district, country etc)
        this.location_relnNOBJ = ANNE.MEM.location;

        //for house of rep, for example, district -> (14th)
        this.nObj_constituency = null;
        this.constituency_relnNOBJ = null;
        
        this.bCUSTOMWIKIOBJECT = true;
        this.text = "";
        this.to_s = "";
        this.term_length_years = null;
        this.extras2 = [];
        this.pre_text = "";
        this.pt_index=-1;
        this.post_text = "";
    }

    addExtra = function($type, $reln, $to_type, $to_s) {
        this.extras2.push({type:$type, reln:$reln, to_type:$to_type, to_s:$to_s});
    }


    getTermLength = function() {
        if (parseInt(this.start_year)>0 && parseInt(this.end_year)>0) {
            return this.end_year - this.start_year;
        }
        return 0;
    }

    setTermStartManual=async($term_start)=> {
        console.log("setTermStartManual "+$term_start);
        this.ROLE_START = null;
        this.term_start = "";
        this.start_year = "";
        this.term_start = $term_start;
        var ad = new anne_date(this.term_start);
        if (ad==null) {
            return; //problem with parse (skip it)
        }
        var bMAKEGHOSTIFNOTFOUND=true;
        await ad.testDate(bMAKEGHOSTIFNOTFOUND);
        this.ROLE_START = ad.nObj;
        if (this.ROLE_START.bes[0]==ANNE.MEM.gregorian_date.temp_nid) {
            this.start_year = ad.year;
        }
        else {
            this.start_year = parseInt($term_start);
        }
    }


    /**
     * 
     * @param {*} $term_end 
     * @returns 
     */
    setTermEndManual=async($term_end)=> {
        console.log("setTermEndManual "+$term_end);
        this.ROLE_END = null;
        this.term_end = "";
        this.end_year = "";
        this.term_end = $term_end;
        var ad = new anne_date(this.term_end);
        if (ad==null) {
            return;
        } 
        var bMAKEGHOSTIFNOTFOUND=true;
        await ad.testDate(bMAKEGHOSTIFNOTFOUND);
        this.ROLE_END = ad.nObj;
        
        if (this.ROLE_END.bes[0]==ANNE.MEM.gregorian_date.temp_nid) {
            this.end_year = ad.year;
        }
        else {
            this.end_year = parseInt($term_end);
        }
    }


    /**
     * code this to resolve and set location.
     * @param {*} $location 
     */
    setLocation = async($location)=> {
        var al = new anne_location();
        var anne_loc_test = new anne_location($str);
        var stateNOBJ = await anne_loc_test.checkForRegionOrState($str);
        if (stateNOBJ!=null) {
            this.nObj_location = stateNOBJ;
            this.location_relnNOBJ = ANNE.MEM.state;
            this.nObj_country = ANNE.MEM.usa;
            return true;
        }
        else {
            var country = await al.checkCountriesNeuronDirect($location);
            if (country!=null) {
                this.nObj_country = country;
            }
        }
    }

    setGenericMarryText = function($bRETURNONLY=false) {
        var parts = [];
        for (var t=0; t < this.ROLE_ACTORS.length; t++) {
            if (t>0) parts.push("and");
            var txt = this.ROLE_ACTORS[t].text;
            parts.push(txt);
        }
        parts.push("marriage");
        
        if ($bRETURNONLY==false) {
            this.text = parts.join(" ").trim();
        }

        //if (this.end_year!=0)  parts.push(this.start_year+"-"+this.end_year);

        return parts.join(" ").trim();
    }


    /**
     * 
     */
    setGenericText=function($bRETURNONLY=false) {
        var parts = [];
        if (this.pre_text!="") {
            parts.push(this.pre_text);
        }
        const THIS = this;
        var noRolesWithThis=function($txt) {
            for (var t=0; t < THIS.ROLE_THEME_THINGS_MOD.length; t++) {
                if (THIS.ROLE_THEME_THINGS_MOD[t].text.indexOf($txt)!=-1) return false;
            }
            for (var t=0; t < THIS.ROLE_THEME_THINGS.length; t++) {
                if (THIS.ROLE_THEME_THINGS[t].text.indexOf($txt)!=-1) return false;
            }
            return true;
        }

        //dont repeat location/country if it is embedded in a theme/thememod..
        //we don't necessarily conceptualize all roles out into their parts.
        //ie,  "Justice of the Wisconsin Supreme court"
        if (this.nObj_location!=null) {
            if (noRolesWithThis(this.nObj_location.text)==true) {
                if (this.nObj_location.alpha2!=undefined) parts.push(this.nObj_location.alpha2);
                else if (this.nObj_location.shortName()) parts.push(this.nObj_location.shortName());
            }
        }
        else if (this.nObj_country!=null) {
            if (noRolesWithThis(this.nObj_country.text)==true) {
                if (this.nObj_country.alpha3!=undefined) parts.push(this.nObj_country.alpha3);
                else parts.push(this.nObj_country.shortName());
            }
        }

        //we don't want to repeat mod in name if we already had it in pre.
        if (this.ROLE_THEME_THINGS_MOD.length) {
            if (this.pt_index>-1 && this.ROLE_THEME_THINGS_MOD.length>1) {
                parts.push(this.ROLE_THEME_THINGS_MOD[this.ROLE_THEME_THINGS_MOD.length-1].text);
            }
            else if (this.pt_index==-1) {
                parts.push(this.ROLE_THEME_THINGS_MOD[this.ROLE_THEME_THINGS_MOD.length-1].text);
            }
        }

        for (var t=0; t < this.ROLE_THEME_THINGS.length; t++) {
            if (t>0) parts.push("and");
            var txt = this.ROLE_THEME_THINGS[t].text;
            txt = txt.replaceAll("White House", "WH");
            if (txt.endsWith(" of")==true) {
                txt = txt.substring(0, txt.length-3).trim();
            }
            parts.push(txt);
        }
        
        parts.push(this.ROLE_ACTORS[0].text);

        if (this.post_text!="") parts.push(this.post_text);  
        if(this.end_year!=0)  parts.push(this.start_year+"-"+this.end_year);
        var test_text = parts.join(" ").trim();
        if ($bRETURNONLY==false) {
            this.text = parts.join(" ").trim();
        }

        //try it again without the actor name.
        //if we can get this in an anbl, that is ideal.
        if (test_text.length>63) {
            test_text = test_text.replace(this.ROLE_ACTORS[0].text+" ", "").trim();
        }

        return test_text;
    }


    setActor = async($actorNOBJ)=>{
        if ($actorNOBJ==null) return;
        this.ROLE_ACTORS.push($actorNOBJ);
    }
    setThemeThing = function($ttNOBJ) {
        if ($ttNOBJ==null) return;
        this.ROLE_THEME_THINGS.push($ttNOBJ);
    }
    setThemeThingMod = function($ttmodNOBJ) {
        if ($ttmodNOBJ==null) return;
        this.ROLE_THEME_THINGS_MOD.push($ttmodNOBJ);
    }


    /**
     * We can use the update for any of the our clumps..
     * It will process max new info (irrelevant for some clumps)
     * @param {*} $newRESOLVEDNOBJ 
     * @param {*} $MOD 
     */
    updateGenericMax=async($newRESOLVEDNOBJ, $MOD)=>{
        this.nObj=$newRESOLVEDNOBJ;
        var ptb = null;
        if ($MOD!=null && $MOD.printToBest!=undefined && typeof $MOD.printToBest=="functoin") {
            ptb = $MOD.printToBest;
        }
        var pb = function($msg) {
            if (ptb==null) return;
            ptb($msg);
        }

        //for service terms
        if (this.ROLE_THEME_THINGS!=null && this.ROLE_THEME_THINGS.length) {
            for (var r=0; r < this.ROLE_THEME_THINGS.length; r++) {
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_theme_thing.nid, this.ROLE_THEME_THINGS[r].nid);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_theme_thing.nid, TO_TYPE_ANNE, this.ROLE_THEME_THINGS[r].nid);
                pb("adding role theme thing", TALK.YELLOW, this.ROLE_THEME_THINGS[r].text);
            }   
        }

        //for service terms
        if (this.ROLE_AFFECTED_THINGS!=null && this.ROLE_AFFECTED_THINGS.length) {
            for (var r=0; r < this.ROLE_AFFECTED_THINGS.length; r++) {
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_affected_thing.nid, this.ROLE_AFFECTED_THINGS[r].nid);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_affected_thing.nid, TO_TYPE_ANNE, this.ROLE_AFFECTED_THINGS[r].nid);
                pb("adding role affected thing", TALK.YELLOW, this.ROLE_AFFECTED_THINGS[r].text);
            }   
        }

        //for service terms
        if (this.ROLE_THEME_THINGS_MOD!=null && this.ROLE_THEME_THINGS_MOD.length) {
            for (var r=0; r < this.ROLE_THEME_THINGS_MOD.length; r++) {
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_theme_thing_mod.nid, this.ROLE_THEME_THINGS_MOD[r].nid);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_theme_thing_mod.nid, TO_TYPE_ANNE, this.ROLE_THEME_THINGS_MOD[r].nid);
                pb("adding role theme thing mod", TALK.YELLOW, this.ROLE_THEME_THINGS_MOD[r].text);
            }   
        }

        //for service terms, marriages
        if (this.ROLE_START!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_start.nid, this.ROLE_START.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_start.nid, TO_TYPE_ANNE, this.ROLE_START.nid);
                pb("adding role start", TALK.YELLOW, this.ROLE_START.text);
            }
        }

        //for service terms, marriages
        if (this.ROLE_END!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_end.nid, this.ROLE_END.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_end.nid, TO_TYPE_ANNE, this.ROLE_END.nid);
                pb("adding role end", TALK.YELLOW, this.ROLE_END.text);
            }
        }

        //for service terms
        if (this.preceded_by_nObj!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.preceded_by.nid, this.preceded_by_nObj.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.preceded_by.nid, TO_TYPE_ANNE, this.preceded_by_nObj.nid);
                pb("adding preceded by", TALK.YELLOW, this.preceded_by_nObj.text);
            }
        }

        //this should point to the RESOLVED SERVICE-NEURON OF THE PREV holder of the position.
        //the preceeded by value points to the person, wherase this points to the entire term
        //for service terms, marriages
        if (this.ROLE_PREV!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_prev.nid, this.ROLE_PREV.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_prev.nid, TO_TYPE_ANNE, this.ROLE_PREV.nid);
                pb("adding role prev", TALK.YELLOW, this.ROLE_PREV.text);
            }
        }

        //for service terms.
        if (this.succeeded_by_nObj!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.succeeded_by.nid, this.succeeded_by_nObj.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.succeeded_by.nid, TO_TYPE_ANNE, this.succeeded_by_nObj.nid);
                pb("adding succeeded by", TALK.YELLOW, this.succeeded_by_nObj.text);
            }
        }


        //this should point to the RESOLVED SERVICE-NEURON OF THE NEXT holder of the position.
        //the succeeded by value points to the person, wherase this points to the entire term
        //for service terms, marriages.
        if (this.ROLE_NEXT!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_ROLE, MEM.role_next.nid, this.ROLE_NEXT.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_next.nid, TO_TYPE_ANNE, this.ROLE_NEXT.nid);
                pb("adding role next", TALK.YELLOW, this.ROLE_NEXT.text);
            }
        }
        
        //for appointed we want to do ANOTHER CLUMP
        //for the one who appointed..
        //X appointed Y for THIS (this clump)
        //TODO: build this after the othre stuff.
        //for service terms
        if (this.appointed_by_nObj!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.appointed_by.nid, this.appointed_by_nObj.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.appointed_by.nid, TO_TYPE_ANNE, this.appointed_by_nObj.nid);
                pb("adding appointed by", TALK.YELLOW, this.appointed_by_nObj.text);
            }

            //only do this if it is new.
            //await this.makeAppointedClump(this.appointed_by_nObj, this.nObj);
        }

        //for country only
        //for service terms, marriages
        if (this.nObj_country!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.country.nid, this.nObj_country.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, ANNE.MEM.country.nid, TO_TYPE_ANNE, this.nObj_country.nid);
                pb("adding country", TALK.YELLOW, this.nObj_country.text);
            }
        }

        //for region (like State, ie, Texas)
        //for service terms, marriages
        if (this.nObj_location!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.location_relnNOBJ.nid, this.nObj_location.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, this.location_relnNOBJ.nid, TO_TYPE_ANNE, this.nObj_location.nid);
                pb("adding location", TALK.YELLOW, this.nObj_location.text);
            }
        }

        //for service terms
        if (this.nObj_political_party!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.political_party.nid, this.nObj_political_party.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.political_party.nid, TO_TYPE_ANNE, this.nObj_political_party.nid);
                pb("adding political party", TALK.YELLOW, this.nObj_political_party.text);
            }
        }

        //congressional district etc.
        //for service terms
        if (this.nObj_constituency!=null && this.constituency_relnNOBJ!=null) {
            var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, TYPE_HAS, MEM.constituency_relnNOBJ.nid, this.nObj_constituency.nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, this.constituency_relnNOBJ.nid, TO_TYPE_ANNE, this.nObj_constituency.nid);
                pb("adding constituency", TALK.YELLOW, this.nObj_constituency.text);
            }
        }

        //for all.
        if (this.extras2!=undefined && this.extras2.length>0) {
            for (var e=0; e < this.extras2.length; e++) {
                var erel =  this.extras2[e];
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, erel.type, erel.reln, erel.to_s);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(erel.type, this.nObj, erel.reln, erel.to_type, erel.to_s);
                if (erel.to_type==TO_TYPE_ANNE) {
                    var TONOBJ = ANNE.MEMG.getNeuronByNid(erel.to_s);
                    pb("adding extra", TALK.YELLOW, TONOBJ.text);
                }
                else {
                    pb("adding extra", TALK.YELLOW, erel.to_s); 
                }
            }
        }
    }
    

    /**
     * make a quick appointed clump.
     * the actor here is the found/created "person" (likely a president)
     * that appointed this judge
     * (that's our primary use case for this build out here)
     * @param {*} $actorNOBJ 
     * @param {*} $affected_thingNOBJ 
     * @returns 
     */
    makeAppointedClump = async($actorNOBJ, $affected_thingNOBJ)=>{
        var clumplookup = new clump_lookup();
        clumplookup.addRole(ANNE.MEM.role_actor.nid, $actorNOBJ.nid);
        clumplookup.addRole(ANNE.MEM.role_action.nid, ANNE.MEM.appointed.nid);
        clumplookup.addRole(ANNE.MEM.role_affected_thing.nid, $affected_thingNOBJ.nid);
        var rez = await clumplookup.findClump(true);
        if (rez.status=="good") {
            var clump_nid = rez.match.nid;
            var x_relon = MEMG.getRelonFromWithTo($actorNOBJ.outs, TYPE_CLUMP, ANNE.MEM.appointed.nid, clump_nid);
            if (x_relon==null) {
                await GHOSTDATA.makeGhostRelon(TYPE_CLUMP, $actorNOBJ.nObj, ANNE.MEM.appointed.nid, TO_TYPE_ANNE, clump_nid);
            }
            return rez.match;
        }
        else if (rez.status=="ambiguious") {
            console.log("don't do anything here.");
            return null;
        }
    }
    

    /**
     * Make a new service term neuron based on
     * all the values that have been set.
     * @returns 
     */
    makeMarryNeuron = async()=> {
        var action = MEM.marry;
        this.setGenericMarryText();
        this.nObj = await GHOSTDATA.makeGhostNeuron(action, this.text, NTYPE_INST);
        if (this.ROLE_ACTORS.length) {
            for (var a=0; a < this.ROLE_ACTORS.length; a++) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_actor.nid, TO_TYPE_ANNE, this.ROLE_ACTORS[a].nid);
            }
        }
        if (this.ROLE_THEME_THINGS!=null && this.ROLE_THEME_THINGS.length) {
            for (var r=0; r < this.ROLE_THEME_THINGS.length; r++) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_theme_thing.nid, TO_TYPE_ANNE, this.ROLE_THEME_THINGS[r].nid);
            }   
        }
        if (this.ROLE_START!=null) {
            console.log("audit start date:"+this.ROLE_START.text);
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_start.nid, TO_TYPE_ANNE, this.ROLE_START.nid);
        }
        if (this.ROLE_END!=null) {
            console.log("audit end date:"+this.ROLE_END.text);
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_end.nid, TO_TYPE_ANNE, this.ROLE_END.nid);
        }
        if (this.ROLE_PREV!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_prev.nid, TO_TYPE_ANNE, this.ROLE_PREV.nid);
        }
        if (this.ROLE_NEXT!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_next.nid, TO_TYPE_ANNE, this.ROLE_NEXT.nid);
        }
        //for country only
        if (this.nObj_country!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, ANNE.MEM.country.nid, TO_TYPE_ANNE, this.nObj_country.nid);
        }
        //for region (like State, ie, Texas)
        if (this.nObj_location!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, this.location_relnNOBJ.nid, TO_TYPE_ANNE, this.nObj_location.nid);
        }
        if (this.extras2!=undefined && this.extras2.length>0) {
            for (var e=0; e < this.extras2.length; e++) {
                var erel =  this.extras2[e];
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, erel.type, erel.reln, erel.to_s);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(erel.type, this.nObj, erel.reln, erel.to_type, erel.to_s);
            }
        }
        return this.nObj;
    }


    /**
     * Make a new service term neuron based on
     * all the values that have been set.
     * @returns 
     */
    makeServiceTermNeuron = async()=> {
        var action = MEM.service_term;
        this.setGenericText();
        this.nObj = await GHOSTDATA.makeGhostNeuron(action, this.text, NTYPE_INST);
        if (this.ROLE_ACTORS.length) {
            for (var a=0; a < this.ROLE_ACTORS.length; a++) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_actor.nid, TO_TYPE_ANNE, this.ROLE_ACTORS[a].nid);
            }
        }
        if (this.ROLE_THEME_THINGS!=null && this.ROLE_THEME_THINGS.length) {
            for (var r=0; r < this.ROLE_THEME_THINGS.length; r++) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_theme_thing.nid, TO_TYPE_ANNE, this.ROLE_THEME_THINGS[r].nid);
            }   
        }
        if (this.ROLE_AFFECTED_THINGS!=null && this.ROLE_AFFECTED_THINGS.length) {
            for (var r=0; r < this.ROLE_AFFECTED_THINGS.length; r++) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_affected_thing.nid, TO_TYPE_ANNE, this.ROLE_AFFECTED_THINGS[r].nid);
            }   
        }

        if (this.ROLE_THEME_THINGS_MOD!=null && this.ROLE_THEME_THINGS_MOD.length) {
            for (var r=0; r < this.ROLE_THEME_THINGS_MOD.length; r++) {
                await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_theme_thing_mod.nid, TO_TYPE_ANNE, this.ROLE_THEME_THINGS_MOD[r].nid);
            }   
        }

        //not in the brain -- as we intended to later find out what this is and deal with it.
        if (this.text_assocs!=null && this.text_assocs.length) {
            for (var r=0; r < this.text_assocs.length; r++) {
                console.log("%c text_assoc %c "+this.text_assocs[r], ANNE.TALK.bgyellow, ANNE.TALK.bggreen);    
                await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.anbl_assoc.nid, TO_TYPE_STRING, this.text_assocs[r]);
            }   
        }

        if (this.ROLE_START!=null) {
            console.log("audit start date:"+this.ROLE_START.text);
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_start.nid, TO_TYPE_ANNE, this.ROLE_START.nid);
        }
        if (this.ROLE_END!=null) {
            console.log("audit end date:"+this.ROLE_END.text);
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_end.nid, TO_TYPE_ANNE, this.ROLE_END.nid);
        }

        if (this.ROLE_PREV!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_prev.nid, TO_TYPE_ANNE, this.ROLE_PREV.nid);
        }

        if (this.ROLE_NEXT!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_ROLE, this.nObj, MEM.role_next.nid, TO_TYPE_ANNE, this.ROLE_NEXT.nid);
        }

        if (this.appointed_by_nObj!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.appointed_by.nid, TO_TYPE_ANNE, this.appointed_by_nObj.nid);
            await this.makeAppointedClump(this.appointed_by_nObj, this.nObj);
        }

        if (this.succeeded_by_nObj!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.succeeded_by.nid, TO_TYPE_ANNE, this.succeeded_by_nObj.nid);
        }

        //for country only
        if (this.nObj_country!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, ANNE.MEM.country.nid, TO_TYPE_ANNE, this.nObj_country.nid);
        }

        //for region (like State, ie, Texas)
        if (this.nObj_location!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, this.location_relnNOBJ.nid, TO_TYPE_ANNE, this.nObj_location.nid);
        }
        if (this.nObj_political_party!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, MEM.political_party.nid, TO_TYPE_ANNE, this.nObj_political_party.nid);
        }

        //congressional district etc.
        if (this.nObj_constituency!=null && this.constituency_relnNOBJ!=null) {
            await GHOSTDATA.makeGhostRelon(TYPE_HAS, this.nObj, this.constituency_relnNOBJ.nid, TO_TYPE_ANNE, this.nObj_constituency.nid);
        }
        if (this.extras2!=undefined && this.extras2.length>0) {
            for (var e=0; e < this.extras2.length; e++) {
                var erel =  this.extras2[e];
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, erel.type, erel.reln, erel.to_s);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(erel.type, this.nObj, erel.reln, erel.to_type, erel.to_s);
            }
        }
        return this.nObj;
    }


    //picked up some new non-contradicting info along the way...
    doExtras = async()=> {
        if (this.nObj==null) return;
        if (this.extras2!=undefined && this.extras2.length>0) {
            for (var e=0; e < this.extras2.length; e++) {
                var erel =  this.extras2[e];
                var x_relon = MEMG.getRelonFromWithTo(this.nObj.outs, erel.type, erel.reln, erel.to_s);
                if (x_relon!=null) continue;
                await GHOSTDATA.makeGhostRelon(erel.type, this.nObj, erel.reln, erel.to_type, erel.to_s);
            }
        }
    }
    

    /**
     * This is done quick and dirty... sloppy.. and not comprehensive.
     * The proper way would be to store these concepts in neurons (already)
     * and then fire them in a pool when "parsing" --- the winning concepts would
     * have definitions (like how long their terms are) on-data-chain.. and we (uhh, I) wouldn't
     * have to try and code as much.
     *
     * @param {*} $roleparts 
     */
    conceptualizeRoles=async($roleparts)=> {
        var rep = "house of representatives";
        var sen = "senator";
        var prez = "president";
        var govn = "governor";
        var mayor = "mayor";
        var judge = "judge";
        var lgovn = "lieutenant governor";
        var memberof = "member of";
        var roles_arr = null;
        var roles_text_arr = null;
        var anyKnownCountry = async($str)=>{
            var al = new anne_location();
            var country = await al.checkCountriesNeuronDirect($str);
            if (country!=null) {
                this.nObj_country = country;
            }
        }
        var anyKnownCity = async($str)=>{
            var bMAKEGHOSTIFNOTFOUND = false;
            var anne_loc = new anne_location($str);
            await anne_loc.processFull([], false, bMAKEGHOSTIFNOTFOUND);
            if (anne_loc.nObj!=null) {
                this.nObj_location = anne_loc.nObj;
                this.location_relnNOBJ = ANNE.MEM.city;
                return true;
            }
            return false;
        }
        var StateOrRegion = async($str)=> {
            var anne_loc_test = new anne_location($str);
            var stateNOBJ = await anne_loc_test.checkForRegionOrState($str);
            if (stateNOBJ!=null) {
                this.nObj_location = stateNOBJ;
                this.location_relnNOBJ = ANNE.MEM.state;
                this.nObj_country = ANNE.MEM.usa;
                return true;
            }
            return false;
        }
        var extractCaseMirror = function($str1, $str2) {
            var str1lower = $str1.toLowerCase();
            var str2lower = $str2.toLowerCase();
            var i = str2lower.indexOf(str1lower);
            if (i==-1) return $str1;
            return $str2.substring(i, $str1.length).trim();
        }
        var extractCase=function($str, $str2) {
            var i = $str2.toLowerCase().indexOf($str.toLowerCase());
            if (i!=-1) {
                return $str2.slice(0, i) + $str2.slice(i + $str.length);
            }
            return $str;
        }

        const RTTM = this.ROLE_THEME_THINGS_MOD;
        
        var checkPreClues = async($role_part, $mapp=1)=> {
            var pc = $role_part.pre_clues;
            if(pc=="") return;

            //extract ordinal here.
            var numbObj = await ANNE.UTILS.extractNumber2(pc, $mapp);
            if (numbObj!=null) {
                RTTM.push(numbObj.nObj);
                if(t==0) {
                    this.pre_text = numbObj.str;    //this was the ord or num str.
                    this.pt_index = RTTM.length-1;
                }
                $role_part.pre_clues = numbObj.remaining;
                
                if ($role_part.pre_clues.length==0) $role_part.pre_clues_handled=true;
                else $role_part.pre_clues_handled=false;
                return;
            }

            //if this is an ordinalizd number, resolve the number,
            //store in ROLE_THEME_THINGS_MOD.
            $role_part.pre_clues_handled = false;
        }

        //test to parse out the THEME_THING and THEME_THING_MOD
        //var testRoles = await ANNE.wikiThemeModPhraseParse($roleparts[0].arr[0], $roleparts[0].links);
        //console.log(testRoles);

        //clean them to replace the willy wonka indexOf lists ...
        //would be to look at the root phrase (before any signals) in the THEME_THING from the phraseParse..
        
        for (var t=0; t < $roleparts.length; t++) {
            var role_part = $roleparts[t];
            var target_role = role_part.target_role;

            if (target_role==undefined || target_role==null) {
                continue;
            }
           
            if (target_role.nid==ANNE.MEM.role_theme_thing.nid) {
                roles_arr = this.ROLE_THEME_THINGS;
                roles_text_arr = this.term_themes;
            }
            else if (target_role.nid==ANNE.MEM.role_affected_thing.nid) {
                roles_arr = this.ROLE_AFFECTED_THINGS;
                roles_text_arr = this.term_affected_things;
            }

            var bUSA=false;

            if ($roleparts[t].arr.length>1) {
                alert("we found one bigger than one!");
                console.log("testing roleparts[t].arr.length "+$roleparts[t].arr.length);
            }


            //we will set this below if any of the inner parts handle the pre_clues.
            role_part.pre_clues_handled = false;

            
            //loop the inner array.
            //first strip out USA location.
            //would obviously be lovely if we parsed this then handled locations (all)
            //they either come as [LOCATION] [THING]  or some [THING] [LOCATION]
            for (var a=0; a < $roleparts[t].arr.length; a++) {
                
                var value = role_part.arr[a].toLowerCase();
                var valueCaps = role_part.arr[a];

                //we're going to have to upgrade this for other countries..
                //had to start somewhere tho.. 
                if (value.indexOf("u.s.")!=-1 || value.indexOf("usa")!=-1 || value.indexOf("united states")!=-1) {
                    this.nObj_country = ANNE.MEM.usa;
                    bUSA=true;
                    var bHANDLED=false;
                    var starters = ["u.s.", "usa", "united states"];
                    for (var s=0; s < starters.length; s++) {
                        var starter = starters[s];
                        if (value.indexOf(starter)==0) {
                            role_part.remaining = value.substring(starter.length+1).trim();
                            console.log("remaining: %c"+role_part.remaining, ANNE.TALK.bggreen);
                            valueCaps = role_part.arr[a].substring(starter.length+1);
                            console.log("valueCaps: %c"+valueCaps, ANNE.TALK.bggreen);
                            value = valueCaps.toLowerCase();
                            var bFOUND = StateOrRegion(role_part.remaining);
                            if (bFOUND) role_part.remaning="";
                            bHANDLED=true;
                            break;
                        }
                    }

                    if (bHANDLED==false) {
                        var enders = ["of the u.s.", "of the usa", "of the united states"];
                        for (var e=0; e < enders.length; e++) {
                            var ender = enders[e];
                            var indx = value.indexOf(ender);
                            if (indx!=-1) {
                                role_part.remaining = value.substring(0, indx-1).trim();
                                console.log("remaining: %c"+role_part.remaining, ANNE.TALK.bggreen);
                                valueCaps = role_part.arr[a].substring(0, role_part.remaining.length);
                                console.log("valueCaps: %c"+valueCaps, ANNE.TALK.bggreen);
                                value = valueCaps.toLowerCase();
                                var bFOUND = StateOrRegion(role_part.remaining);
                                if (bFOUND==true) role_part.remaning="";
                                break;
                            }
                        }
                    }
                }
            
                
                if (value.indexOf(rep)!=-1) {
                    if (bUSA==true) this.term_length_years = 2; 
                    roles_arr.push(ANNE.MEM.representative);
                    role_part.remaining=extractCase(rep, valueCaps);
                    await checkPreClues(role_part);
                }
                else if (value.indexOf(sen)!=-1) {
                    if (bUSA==true) this.term_length_years = 4;
                    roles_arr.push(ANNE.MEM.senator);
                    role_part.remaining=extractCase(sen, valueCaps);
                    await checkPreClues(role_part);
                }
                else if (value.indexOf(lgovn)!=-1) {
                    if (bUSA==true) this.term_length_years = 4;
                    roles_arr.push(ANNE.MEM.lgovernor);
                    await checkPreClues(role_part);
                    if (value.indexOf("lieutenant governor of")==0) {
                         if (value.indexOf("lieutenant governor of")==0) {
                            role_part.remaining = valueCaps.substring(23).trim();
                            var found = await StateOrRegion(role_part.remaining);
                            if (found==true) role_part.remaining="";
                        }
                    }
                    else {
                        role_part.remaining=extractCase(lgovn, valueCaps);
                    }
                }
                else if (value.indexOf(prez)!=-1) {
                    if (bUSA==true) this.term_length_years = 4;
                    roles_arr.push(ANNE.MEM.president);
                    await checkPreClues(role_part);
                    if (value.indexOf("president of")==0) {
                        role_part.remaining = valueCaps.substring(12).trim();
                        var found = await StateOrRegion(role_part.remaining);
                        if(found==true) role_part.remaining="";
                    }
                    else {
                        role_part.remaining=extractCase(prez, valueCaps);
                    }
                }
                
                else if (value.indexOf(govn)!=-1) {
                    if (bUSA==true) this.term_length_years = 4;
                    roles_arr.push(ANNE.MEM.governor);
                    await checkPreClues(role_part);
                    if (value.indexOf("governor of")==0) {
                         if (value.indexOf("governor of")==0) {
                            role_part.remaining = valueCaps.substring(12).trim();
                            var found = await StateOrRegion(role_part.remaining);
                            if (found==true) role_part.remaining="";
                        }
                    }
                    else {
                        role_part.remaining=extractCase(govn, valueCaps);
                    }
                }
                else if (value.indexOf(mayor)!=-1) {
                    if (bUSA==true) this.term_length_years = 4;
                    roles_arr.push(ANNE.MEM.mayor);
                    if (value.indexOf("mayor of")==0) {
                        role_part.remaining = valueCaps.substring(9).trim();
                        var cities = await ANNE.getQuickCity(role_part.remaining);
                        if (cities.length==1) {
                            this.nObj_location = cities[0];
                            this.location_relnNOBJ = ANNE.MEM.city;
                            role_part.remaining="";
                        }
                        else if (cities.length>1 && ANNE.BRAIN!=null) {
                            this.nObj_location = ANNE.BRAIN.applyPoolWeights(cities, this.ROLE_ACTORS, true);
                            this.location_relnNOBJ = ANNE.MEM.city;
                            role_part.remaining="";
                        }
                        else if (bUSA==true) {
                            var bFOUND = await anyKnownCity(role_part.remaining+", USA");
                            if (bFOUND==true) role_part.remaining="";
                        }
                        else {
                            var bFOUND = await anyKnownCity(role_part.remaining);
                            if (bFOUND==true) role_part.remaining = "";
                        }
                    }
                    else {
                        role_part.remaining=extractCase(mayor, valueCaps);
                    }
                    await checkPreClues(role_part);
                }

                else if (value.indexOf(memberof)!=-1) {
                    roles_arr.push(ANNE.MEM.member_of);
                    if (value.indexOf("member of the")==0) {
                        role_part.remaining = valueCaps.substring(14).trim();
                        var cities = await ANNE.getQuickCity(role_part.remaining);
                        if (cities.length==1) {
                            this.nObj_location = cities[0];
                            this.location_relnNOBJ = ANNE.MEM.city;
                             role_part.remaining="";
                        }
                        else if (cities.length>1 && ANNE.BRAIN!=null) {
                            this.nObj_location = ANNE.BRAIN.applyPoolWeights(cities, this.ROLE_ACTOR, true);
                            this.location_relnNOBJ = ANNE.MEM.city;
                             role_part.remaining="";
                        }
                        else if (bUSA==true) {
                            var bFOUND = await anyKnownCity(role_part.remaining+", USA");
                            if (bFOUND==true)  role_part.remaining="";
                        }
                        else {
                            var bFOUND = await anyKnownCity(role_part.remaining);
                            if (bFOUND==true)  role_part.remaining="";
                        }
                    }
                    else {
                        role_part.remaining=extractCase(memberof, valueCaps);
                    }
                    await checkPreClues(role_part);
                }


                //take the original value. (resolve later)
                else {
                    if (role_part.pre_clues!="") {

                        var pc = role_part.pre_clues.toLowerCase();
                        if (pc.indexOf(rep)!=-1) {
                            if (bUSA==true) this.term_length_years = 2; 
                            roles_arr.push(ANNE.MEM.representative);
                            role_part.pre_clues_handled = true;
                        }
                        else if (pc.indexOf(sen)!=-1) {
                            if (bUSA==true) this.term_length_years = 4;
                            roles_arr.push(ANNE.MEM.senator);
                            role_part.pre_clues_handled = true;
                        }
                        else if (pc.indexOf(lgovn)!=-1) { //lgovn = "lieutenant governor"
                            if (bUSA==true) this.term_length_years = 4;
                            roles_arr.push(ANNE.MEM.lgovernor);
                            role_part.pre_clues_handled = true;
                        }

                        else if (pc.indexOf(prez)!=-1) {
                            if (bUSA==true) this.term_length_years = 4;
                            roles_arr.push(ANNE.MEM.president);
                            role_part.pre_clues_handled = true;
                        }
                                                
                        else if (pc.indexOf(govn)!=-1) {
                            if (bUSA==true) this.term_length_years = 4;
                            roles_arr.push(ANNE.MEM.governor);
                            role_part.pre_clues_handled = true;
                        }
                        else if (pc.indexOf(mayor)!=-1) {
                            if (bUSA==true) this.term_length_years = 4;
                            roles_arr.push(ANNE.MEM.mayor);
                            role_part.pre_clues_handled = true;
                        }
                        else if (pc.indexOf(judge)!=-1) {;
                            roles_arr.push(ANNE.MEM.judge);
                            role_part.pre_clues_handled = true;
                        }
                        else if (pc.indexOf(memberof)!=-1) {;
                            role_part.pre_clues_handled = true;
                                   
                            if (pc.indexOf("member of the")==0) {
                                if (value.indexOf("board")!=-1) {
                                    this.ROLE_THEME_THINGS.push(ANNE.MEM.board_member_of);
                                    var prepre = value.substring(0, value.indexOf("board")-1).trim();
                                    console.log(prepre);
                                    var cities = await ANNE.getQuickCity(prepre);
                                    if (cities.length==1) {
                                        this.nObj_location = cities[0];
                                        this.location_relnNOBJ = ANNE.MEM.city;
                                    }
                                    else if (cities.length>1 && ANNE.BRAIN!=null) {
                                        this.nObj_location = ANNE.BRAIN.applyPoolWeights(cities, this.ROLE_ACTOR, true);
                                        this.location_relnNOBJ = ANNE.MEM.city;
                                    }
                                    else if (bUSA==true) {
                                        if (await anyKnownCity(role_part.remaining+", USA")==false) {
                                            this.ROLE_THEME_THINGS.push(ANNE.MEM.member_of);
                                            this.term_themes_mod.push({text:valueCaps.trim(), ntype:NTYPE_INST});
                                        }
                                    }
                                    else if (await anyKnownCity(role_part.remaining)==false) {
                                        this.ROLE_THEME_THINGS.push(ANNE.MEM.member_of);
                                        this.term_themes_mod.push({text:valueCaps.trim(), ntype:NTYPE_INST});
                                    }
                                }
                                else {
                                     this.ROLE_THEME_THINGS.push(ANNE.MEM.member_of);
                                     this.term_themes_mod.push({text:valueCaps.trim(), ntype:NTYPE_INST});

                                     if (valueCaps.indexOf("Assembly of")==0) {
                                        var remaining = valueCaps.substring(12).trim();
                                        await anyKnownCountry(remaining);
                                     }
                                }
                            }
                            else {
                                this.ROLE_THEME_THINGS.push(ANNE.MEM.member_of);
                            }
                        }
                                
                        else {
                            //unique entries
                            if (roles_text_arr.indexOf(valueCaps.trim())==-1) {

                                //try to extract state (if so we know USA)
                                var parts2 = valueCaps.split(" ");
                                if (parts2.length>0) {
                                    var found = await StateOrRegion(parts2[0]);
                                    if (found==true) {
                                        //valueCaps = valueCaps.replace(parts2[0]+" ", "");
                                    }
                                }
 
                                await checkPreClues(role_part, 8);

                                if (role_part.pre_clues_handled==false) {
                                    if (role_part.pre_clues.endsWith("of the")==true) {
                                        var ot = role_part.pre_clues.indexOf("of the");
                                        var val = role_part.pre_clues.substring(0, ot);
                                    
                                        //chief justice.. or title name..  OF THE.
                                        this.term_themes_mod.push({text:valueCaps.trim(), ntype:NTYPE_CLASS});
                                    
                                        //OF THE ---  indicative of inst.
                                        this.term_themes.push(val.trim());
                                    }
                                }
                                else roles_text_arr.push(valueCaps.trim());
                            }
                            else await checkPreClues(role_part);
                        }
                    }

                    //remaining was just handled here..
                    else {
                        //unique entries.
                        if (roles_text_arr.indexOf(valueCaps.trim())==-1) {
                            roles_text_arr.push(valueCaps.trim());
                            role_part.remaining="";
                        }
                    }

                    //this could be.. any number of nouns. (location/company/?)
                    //"from the 2nd district"
                    if (role_part.post_test!="") {
                        //this would be lovely to parse in a phrase.
                        if (role_part.post_test.indexOf("from")==0) {
                            role_part.likelyLocation = true;

                            if (role_part.post_test.indexOf("from the")==0) {
                                var remaining = role_part.post_test.substring(9);
                                console.log(remaining);
                                role_part.post_test = remaining;

                                remaining = remaining.replace("district","");
                                remaining = remaining.replace("  "," ").trim();
                                if (remaining.length!=role_part.post_test.length) {
                                    var numbNOBJ = await ANNE.UTILS.extractNumber(remaining, 1);
                                    if (numbNOBJ!=null) {
                                        this.addExtra(TYPE_HAS, ANNE.MEM.district.nid, TO_TYPE_ANNE, numbNOBJ.nid); 
                                        role_part.post_test = ""; 
                                    }
                                }
                            }
                            else {
                                var remaining = role_part.post_test.substring(4);
                                console.log(remaining);
                                role_part.post_test = remaining;
                                remaining.replace("district","");
                                remaining.replace("  "," ");
                                if (remaining.length!=role_part.post_test) {
                                    var numbNOBJ = await ANNE.UTILS.extractNumber(pc, 1);
                                    if (numbNOBJ!=null) {
                                        this.addExtra(TYPE_HAS, ANNE.MEM.district.nid, TO_TYPE_ANNE, numbNOBJ.nid);  
                                        role_part.post_test = "";
                                    }
                                }
                            }
                        }

                        //put the post test in the afftected things.
                        if (role_part.post_test!="") {
                            this.term_affected_things.push(role_part.post_test);
                        }
                    }
                    
                }   

                if(role_part.pre_clues_handled==false) {
                    console.log("%c PRE CLUES NOT HANDLED %c "+role_part.pre_clues, ANNE.TALK.bgbigred, ANNE.TALK.bgyellow);    
                }

                if (role_part.remaining!=undefined && role_part.remaining.length>0) {
                    console.log("still remaining "+role_part.remaining);
                    role_part.remaining = extractCaseMirror(role_part.remaining, valueCaps);
                    if (role_part.remaining.indexOf("the ")!=0) {
                        role_part.remaining = role_part.remaining.replace("the ", "").trim();
                    }
                    else role_part.remaining = extractCaseMirror(role_part.remaining, role_part.arr[a]);
                    this.term_themes_mod.push({text:role_part.remaining, ntype:NTYPE_INST});
                }

            }
        }    
        this.nameFunc = this.setGenericText;
    }


    /**
     * 
     * @returns 
     */
    quickPerson=async($person)=> {
        var GMOBJ = await MEMG.getOrMakeIfNewPerson(this.aec, $person);
        if (GMOBJ.status=="good") {
            if (this.aec!=null) {
                if (GMOBJ.match.bNEW==true) this.aec.printf("new person", TALK.BLUE, $person);
                else this.aec.printf("resolved person", TALK.YELLOW, $person);
            }
            return GMOBJ.match;
        }
        
        //resolve this somehow. or do nothing.
        else if (GMOBJ.status=="ambiguous") {
            this.aec.printf("resolved person", TALK.YELLOW, $person);
            return GMOBJ.matches[0];
        }
    }

    
    /**
     * A term neuron from wikipedia. 
     * @param {*} $obj 
     * @param {*} $postfix 
     */
    processServiceTermFromWikipedia=async($obj, $postfix="", $sameOffice=null, $aec=null)=>{
        if ($obj["order"+$postfix]!=undefined) {
            if ($sameOffice[0].pre_clues=="") {
                $sameOffice[0].pre_clues=$obj["order"+$postfix].data.text;
            }
            else {
                if ($sameOffice[0].pre_clues!=$obj["order"+$postfix].data.text) {
                    console.warn("WHAT IS GOING ON HERE????? ");
                    console.log($sameOffice[0].pre_clues+" vs "+$obj["order"+$postfix].data.text);
                }
            }
        }
        
        if ($sameOffice!=null) {
            await this.conceptualizeRoles($sameOffice, $aec);
        }

        //extras.
        if($obj['predecessor'+$postfix]!=undefined) {
            this.preceded_by_text = $obj['predecessor'+$postfix].data.text;
            this.preceded_by_nObj = await this.quickPerson(this.preceded_by_text);
        }
        
        if($obj['successor'+$postfix]!=undefined) {
            this.succeeded_by_text = $obj['successor'+$postfix].data.text;
            this.succeeded_by_nObj = await this.quickPerson(this.succeeded_by_text);
        }
        
        if($obj['appointer'+$postfix]!=undefined) {
            this.appointed_by_text = $obj['appointer'+$postfix].data.text;
            this.appointed_by_nObj = await this.quickPerson(this.appointed_by_text);
        }
        
        //$infobox.data["constituency"+postfix]==undefined
        if($obj['constituency'+$postfix]!=undefined) {
            //"TX's 14th congressional district"
            var constituency = $obj['constituency'+$postfix].data.text;
            console.log(constituency);
            var num = await ANNE.UTILS.extractNumber(constituency);
            if (num!=null) {
                if (constituency.indexOf("district")!=-1) {
                    this.nObj_constituency = num;
                    this.constituency_relnNOBJ = ANNE.MEM.district;
                }
            }
        }

        if ($obj["term_start"+$postfix]!=undefined) {
            this.term_start = $obj["term_start"+$postfix].data.text;
            var ad = new anne_date(this.term_start);
            if (ad==null) {
                return; //problem with parse (skip it)
            }
            var bMAKEGHOSTIFNOTFOUND=true;
            await ad.testDate(bMAKEGHOSTIFNOTFOUND);
            this.ROLE_START = ad.nObj;
            this.start_year = ad.year;
        }

        if ($obj["term_end"+$postfix]!=undefined) {
            this.term_end = $obj["term_end"+$postfix].data.text;
            var ad = new anne_date(this.term_end);
            if (ad==null) {
                return;
            } 
            var bMAKEGHOSTIFNOTFOUND=true;
            await ad.testDate(bMAKEGHOSTIFNOTFOUND);
            this.ROLE_END = ad.nObj;
            this.end_year = ad.year;
        }
    }
}

/************** (from file ./js/base/neuron.js)**/ 

var DB_local = {};
DB_local.db = null;
DB_local.dbName = "ANNODE_local";
DB_local.bSUPPORTED = false;
DB_local.version = 6;

DB_local.editName = function($POST) {
    if (DB_local.dbName.indexOf("_"+$POST)!=-1) return;
    DB_local.dbName = DB_local.dbName+"_"+$POST;
}

/**
 * Do we even use full pull history at all anymore in DB LOCAL????
 * @author {Mr.Scatman}
 */
 DB_local.___init=async()=> {
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB_local.editName("testnet");
    }
    
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback();
        return;
    }
    var done = await DB_local.openConnection(DB_local.dbName, DB_local.version);
    return;
}

DB_local.___initMMOnly=async()=>{
    if (DB_local.db!=null) return true;
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB_local.editName("testnet");
    }
    return new Promise((resolve, reject) => {
        var request = indexedDB.open(DB_local.dbName, DB_local.version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB_local.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB_local.db = event.target.result;
            if (DB_local.db.objectStoreNames.contains('otc_commish')==false) {
                var commish = DB_local.db.createObjectStore("otc_commish", { keyPath:"midmanne_commishid" });
                commish.createIndex("midmanne_commishid", "midmanne_commishid", { unique: true });
            }
            return resolve(false);  //no load needed
        };
    });
}


DB_local.getDBSize=async()=> {
    var size = await DB_local.getDBTblSize("collab_relons");
    size+= await DB_local.getDBTblSize("fullPullHistory");
    size+= await DB_local.getDBTblSize("ghost_relons");
    size+= await DB_local.getDBTblSize("traditional_txs");
    size+= await DB_local.getDBTblSize("external_requests");
    size+= await DB_local.getDBTblSize("adf_otc_swaps");
    size+= await DB_local.getDBTblSize("bot_trade_history");
    size+= await DB_local.getDBTblSize("otc_commish");
    return size;
}

DB_local.getDBTblSize=async($tbl, $method="readwrite")=> {
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                var arr = event.target.result;
                var json = JSON.stringify(arr);
                size = json.length;
                return resolve(size);
            };
        } 
    });
}

/**
 * We wait until we know what user we are dealing with to get these ones
 */
DB_local.__getGhostsOnly=async($onLoadCallback=null)=>{
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return [];
    }
    var aGhostRelons = await DB_local.loadAll("ghost_relons");
    return aGhostRelons;
}


/**
 * Table for exensions to store a history of the requests they saw.
 * 
 * Use.  subtype, msg, exernaldata and datestamp.
 * @param {*} $onLoadCallback 
 * @returns 
 */
DB_local.__getExternalRequestsOnly=async($onLoadCallback=null)=>{
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return [];
    }
    var aGhostRelons = await DB_local.loadAll("external_requests");
    return aGhostRelons;
}


/**
 */
DB_local.__getStoredTxs=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("traditional_txs");
}


DB_local.loadADFSwaps=async($network="mainnet")=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aSwaps = await DB_local.loadAll("adf_otc_swaps");
    var aSwapsFinal = [];
    for (var s=0; s < aSwaps.length; s++) {
        if (aSwaps[s].network==$network) aSwapsFinal.push(aSwaps[s]);
    }
    return aSwapsFinal;
}


/**
 * 
 * @param {*} $network 
 * @returns 
 */
DB_local.loadCommishHistory=async($network="mainnet")=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aCommishHistory = await DB_local.loadAll("otc_commish");
    var aComisHistoryFinal = [];
    for (var s=0; s < aCommishHistory.length; s++) {
        if (aCommishHistory[s].network==$network) aComisHistoryFinal.push(aSwaps[s]);
    }
    return aComisHistoryFinal;
}



/**
 * 
 * @returns 
 */
DB_local.__getFullPullHistory=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("fullPullHistory");
}


DB_local.__getBotTradeHistory=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("botcoinbalance");
}

/**
 */
 DB_local.__getCollabRelons=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aCollabs = await DB_local.loadAll("collab_relons");
    var ours = [];
    for (var c=0; c < aCollabs.length; c++) {
        if(aCollabs[c].from_nid==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else if(aCollabs[c].to_s==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else if(aCollabs[c].collaborator==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else {
            //do we have right to approve this?
        }
    }
    return ours;
}


/**
 * @author {Mr.Scatman}
 */
DB_local.___deleteEntireDB=async()=> {
    if (DB_local.bSUPPORTED==false) return;
    if (DB_local.db==null) return;
    return new Promise((resolve, reject) => {
        DB_local.db.close();
        var request = indexedDB.deleteDatabase(DB_local.dbName, 2);
        request.onblocked = function () {
            console.log("Delete blocked.");
            return resolve(false);
        }
        request.onerror = function () {
            console.log("Error deleting the DB");
            return resolve(false);
        };
        request.onsuccess = function () {
            console.log("Deleted OK.");
            return resolve(true);
        }   
    });             
}

/**
 * @author {Mr.Scatman}
 */
DB_local.openConnection=async($dbName, $version)=> {
    if (DB_local.db!=null) return true;
    return new Promise((resolve, reject) => {
        var request = indexedDB.open($dbName, $version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB_local.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB_local.db = event.target.result;

            if (DB_local.db.objectStoreNames.contains('ghost_relons')==false) {
                var ghostObjStore = DB_local.db.createObjectStore("ghost_relons", { keyPath:"txid" });
                ghostObjStore.createIndex("txid", "txid", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('fullPullHistory')==false) {
                var fullPullHistoryStore = DB_local.db.createObjectStore("fullPullHistory", { keyPath:"pullid" });
                fullPullHistoryStore.createIndex("pullid", "pullid", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('collab_relons')==false) {
                var collabRelonStore = DB_local.db.createObjectStore("collab_relons", { keyPath:"txid" });
                collabRelonStore.createIndex("txid", "txid", { unique: true });
            }
            
            if (DB_local.db.objectStoreNames.contains('traditional_txs')==false) {
                var ttStore = DB_local.db.createObjectStore("traditional_txs", { keyPath:["txid", "anne_app_user"] });
                ttStore.createIndex("usertxid", ["txid", "anne_app_user", "confirmed"], { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('external_requests')==false) {
                var extReqStore = DB_local.db.createObjectStore("external_requests", { keyPath:"req_id" });
                extReqStore.createIndex("req_id", "req_id", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('adf_otc_swaps')==false) {
                var swaps = DB_local.db.createObjectStore("adf_otc_swaps", { keyPath: ["swap_id", "network"] });
                swaps.createIndex("swap_id", "swap_id", { unique: false });
                swaps.createIndex("network", "network", { unique: false });
            }

            //ticker + currentTime = unique key
            if (DB_local.db.objectStoreNames.contains('botcoinbalance')==false) {
                var bottradehistory = DB_local.db.createObjectStore("botcoinbalance", {keyPath:["ticker", "currentTime"]});
                bottradehistory.createIndex("ticker", "ticker", { unique: false });
                bottradehistory.createIndex("currentTime", "currentTime", { unique: false });
            }

            if (DB_local.db.objectStoreNames.contains('otc_commish')==false) {
                var commish = DB_local.db.createObjectStore("otc_commish", { keyPath:"midmanne_commishid" });
                commish.createIndex("midmanne_commishid", "midmanne_commishid", { unique: true });
            }

            return resolve(false);  //no load needed
        };
    });
}




DB_local.loadAll=async($tbl, $method="readwrite")=> {
    if (DB_local.db==null) return;
    if (DB_local.db.transaction==undefined) {
        return;
    }
    
    console.log("DB.loadAll "+$tbl);
    
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                return resolve(event.target.result);
            };
        } 
        else {
            var arr = [];

            // Fallback to the traditional cursor approach if getAll isn't supported.
            objectStore.openCursor().onsuccess = function(event) {
                var cursor = event.target.result;
                if (cursor) {
                    arr.push(cursor.value);
                    cursor.continue();
                } 
                else return resolve(arr);
            };
        }
    });
}


DB_local.___insert = async($objs, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;
    if ($objs.length==0) return;

    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {    
        $objs.forEach(function($obj) {
            var request = null;

            if ($tbl=="ghost_relons") {
                //dont save this relon, EVER
                if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
                if ($obj.txid==undefined || $obj.txid==null || $obj.txid=="") return;

                //used in the gui but never broadcast.
                var TEMP_ONLY = false;
                if ($obj.TEMP_ONLY==true) {
                    TEMP_ONLY=true;
                }

                var sanitized_obj = {
                    from_nid: $obj.from_nid,
                    reln: $obj.reln,
                    rorigin: $obj.rorigin,
                    time:$obj.time,
                    to_s: $obj.to_s,
                    to_type: $obj.to_type,
                    txid: $obj.txid,
                    type: $obj.type,
                    ghost_order: $obj.ghost_order,
                    ghost_owner: $obj.ghost_owner,
                    owned_by:$obj.ghost_owner,
                    TEMP_ONLY: TEMP_ONLY
                }
                if ($obj.from_text!=undefined) sanitized_obj.from_text = $obj.from_text;

                request = objectStore.add(sanitized_obj);
            }
   
            //will need collab sanitized stuff too. (TO DO)
            else if ($tbl=="collab_relons") {
                //dont save this relon, EVER
                if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
                request = objectStore.add($obj);
            }
            
            else {
                console.log("adding INDEXDB:"+$obj);
                request = objectStore.add($obj);
            }
        });

        transaction.onsuccess = function(event) {
            return resolve(true);
        };
        transaction.onerror = function(event) {
            return resolve(false);
        }
        return resolve(true);
    });
}


DB_local.___insert_single = async($obj, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;

    //dont save this relon, EVER
    if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return objectStore.add($obj);    
}

DB_local.___updateAll = async($objs, $tbl="collab_relons", $method="readwrite")=> {
    for (var r=0; r < $objs.length; r++) {
        await DB_local.___update($objs[r], $tbl, $method);
    }
}

DB_local.___update = async($obj, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    var sanitized_obj  = {};

    //for ghost relons we santitize the obj first.
    if ($tbl=="ghost_relons") {
        var TEMP_ONLY = false;
        if ($obj.TEMP_ONLY==true) {
            TEMP_ONLY=true;
        }
        sanitized_obj = {
            from_nid: $obj.from_nid,
            reln: $obj.reln,
            rorigin: $obj.rorigin,
            time:$obj.time,
            to_s: $obj.to_s,
            to_type: $obj.to_type,
            txid: $obj.txid,
            type: $obj.type,
            ghost_order: $obj.ghost_order,
            ghost_owner: $obj.ghost_owner,
            owned_by:$obj.ghost_owner,
            TEMP_ONLY:TEMP_ONLY
        }
        if ($obj.flathexint!=undefined) sanitized_obj=$obj.fathexint;
        if ($obj.from_text!=undefined) sanitized_obj.from_text = $obj.from_text;
    }

    //all other tbl objects are fine as-is.
    else sanitized_obj = $obj;

    return new Promise((resolve, reject) => {
        var request = objectStore.put(sanitized_obj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}


DB_local.___delete = async ($ids, $tbl="collab_relons", $method="readwrite", $callback=null)=> {
    if (DB_local.bSUPPORTED==false) return false;
    if ($ids.length==0) return false;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    for (var t=0; t < $ids.length; t++) {
        if ($ids[t]==undefined || (typeof $ids[t]!=='string' && typeof $ids[t]!=='number')) continue;
        await DB_local.__deleteFinal(objectStore, $ids[t]);
    }
}

DB_local.___deleteWithMultiKey = async ($multikey, $tbl="adf_otc_swaps", $method="readwrite", $callback=null)=> {
    if (DB_local.bSUPPORTED==false) return false;
    if ($multikey.length==0) return false;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    for (var t=0; t < $multikey.length; t++) {
        await DB_local.__deleteFinal(objectStore, $multikey[t]);
    }
}


DB_local.__deleteFinal = async($objectStore, $key)=>{
    return new Promise((resolve, reject) => {
        var request = $objectStore.delete($key);
        request.onsuccess = function(event) { 
            return resolve(true); 
        }
        request.onerror = function(event) { 
            return resolve(false); 
        } 
    });
}

/************** (from file ./js/db/db_local.js)**/ 

const LINGO_TYPE_ANNE = 0;
const LINGO_TYPE_GRAPH = 1;
const LINGO_TYPE_COMMON = 2;

const TENSE_PRESENT = 1;
const TENSE_PAST = 2;
const TENSE_PAST_COMPLETE = 3;


//This is not semantic langauge via clumps.  This is just a basic output
//class for debugging or quick english display of things.
class anne_talk extends anne_singleton {
	constructor(){
		super();
		
		//THIS IS A SIMPLE WAY TO PROGRAM GRAMMAR utilizing base roles as storage.
		//different languages can have different grammar patterns.
		//one could spend a very long time in here building a robust langauge handler..
		//it can be done for AI/AGI use cases.
		
		//WE CAN STORE THE ROLES PROPERLY.
		//WE CAN VERBALIZE THEM.
		//In time, (with a fuller vocab) we can fully parse and conceptualize into the ANNE semantics.

		//ANNE training is done by experience/use..  The very basics are put in at inception to have something 
		//to work with.  The idea is parts-of-speachless (we use a higher concept of part of anne)
		//Language is a map into a higher-order brain protocol.
		//The Language may change, and with it the ways to communicate, but the concepts are stored using the 
		//higher-order unified brain protocol.  The same brain can speak 4 different languages and doesn't rqeuire
		//4 unique concepts repeating what a "dog" is defined as.  There are 4 differeent lang word maps to the 
		//same concepts (if it is indeed the same concept) 

		//clump definition pattern.
		this.def = new Object();
		this.def.infini = {pat_part_name:"infini", roles:[{role:"role_action", p:"to"}]};	//no use ot do doInfinitive w/ rule as it is just a "to"+
		this.def.mode = {pat_part_name:"mode", roles:["role_mode"]};
		this.def.obj = {pat_part_name:"obj", roles:[{role:'role_affected_thing', mod:'role_affected_thing_mod', det_func:this.___doThingDet}, 
													{role:'role_resulting_new_thing', mod:'role_resulting_new_thing_mod', det_func:this.___doThingDet},
													{role:'role_theme_thing', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_instrument_mod',
													{role:'role_instrument', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_state', 'role_feelz', 'role_thought', 'role_fact']};

		this.def.manner = {pat_part_name:"manner", roles:[{role:"role_manner", modfunc:this.___doMannerEngShifter}]}
													
		this.def.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"to"}, 
				{role:'role_when', mod:'role_when_mod'},
				'role_start','role_end']
		};

		this.eng_pat_def = [
			this.def.infini,
			this.def.mode,
			this.def.obj,
			this.def.manner,
			this.def.wherewhen2
		]


		//action inst pattern
		this.actionInst = new Object();
		this.actionInst.subj = {pat_part_name:"subj", roles:['role_actor']};
		this.actionInst.mode = {pat_part_name:"mode", roles:['role_manner', 'role_mode']};
		this.actionInst.pred = {pat_part_name:"pred", roles:['role_action'], mod:'role_tense', modfunc:this.englishTensify}; //'role_reason'
		this.actionInst.obj = {pat_part_name:"obj", roles:['role_affected_thing', 'role_resulting_new_thing', 
										'role_state', 'role_feelz', 'role_thought', 'role_fact',
										'role_instrument_mod','role_instrument']};

		this.actionInst.theme_thing = {pat_part_name:"theme_thing", roles:['role_theme_thing_mod', 'role_theme_thing']};
		
		//to use role_source/path/goal in the same clump you will likely want 
		//more than one role_where_mod.  if so, tuff beans.  Create a nested neuron.
		this.actionInst.wherewhen = {
				pat_part_name:"wherewhen",
				roles:['role_source', 'role_path', 'role_goal',
				'role_where_mod', 'role_where', 'role_when_mod', 'role_when', 
				'role_start', 'role_end']
			};
		this.actionInst.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"at", det_func:this.___doWhereDet}, 
				{role:'role_when', mod:'role_when_mod', p:"on"}
				   ,'role_start', 'role_end']
		}
		this.actionInst.pre_connectors = {
			pat_part_name:"pre_connectors",
			recurse:true,
			roles:['role_cause', 'role_cdx', 'role_prev']};
		this.actionInst.post_connectors = {
			recurse:true,
			pat_part_name:"post_connectors",
			roles:['role_effect', 'role_result', 'role_next']};

		//this.actionInst.tense,
		this.eng_pat1 = [
			this.actionInst.subj, 
			this.actionInst.mode,
			this.actionInst.pred, 
			this.actionInst.theme_thing, 
			this.actionInst.obj, 
			this.actionInst.wherewhen2,
			this.actionInst.pre_connectors,
			this.actionInst.post_connectors
		];


		this.type_pre = ["sys", "be", "param", "has", "awareness", "exper", "clump", "sys msg", "brain", "role", "episodic"];

		this.vowel=["a", "e", "i", "o", "u"];  //and sometimes y? lolz.
		
		this.init();
	}


	init = function() {
		this.bgbigred = this.redbig;
		this.bgblack = 'background:black; color:white';
		this.bgblackbold = 'background:black; color:white; font-weight:bold';
		this.bgblackyellow = 'background:#171717; color: #fefe17';
		this.bgblue = 'background:blue; color: white';
		this.bgbluebig = 'background:blue; color: white; font-size:1.1rem';
		this.bgerror = 'background: red; color: white; font-weight:bold; font-size:1.1rem;';
		this.bgfire = 'background:#F7C24A; color: #171717';
		this.bggray = 'background:#ccc; color: #171717';
		this.bggreen = 'background: green; color: white';
		this.bggreenbig = 'background: green; color: white; font-size:1.1rem;';
		this.bgivoryblue = 'background:ivory; color: blue';
		this.bglightblue = 'background:lightblue; color: #171717';
		this.bglightbluebold = 'background:lightblue; color: #171717; font-weight:bold';
		this.bglightgreen = 'background:lightgreen; color: #171717';
		this.bglightred = 'background:lightgreen; color: #171717';
		this.bgneg = this.bgred;
		this.bgneu = this.bgyellow;
		this.bgorange = 'background: orange; color: #171717';
		this.bgpink = 'background:pink; color: #171717';
		this.bgcyan = 'background:cyan; color: #171717';
		this.bgpos = 'background: #00ff00; color: #171717';
		this.bgpurple = 'background:purple; color: white';
		this.bgred = 'background: red; color: white';
		this.bgredbig = 'background: red; color: white; font-size:1.1rem;';
		this.bgyellow = 'background:#fefe17; color: black';
		
		this.bgheading_yellow = 'background:black; color: #fefe17; font-size:1.1rem';
		this.bgheading_black = 'background:whitesmoke; color: black; font-size:1.1rem';
		this.bgheading = 'background:ivory; color: #171717; font-size:1.4; font-weight:bold';

		this.fromMsg = 'background:ivory; color: #171717; font-weight:bold';
		this.funct = 'background:pink; color: #171717; font-style:italic';
		this.ghosts = 'background:white; color: blue';
		this.msg = 'background: green; color: white';
		this.msgNote = 'background:ivory; color:green; font-style:italic';
		this.pooltitle = 'background:ivory; color: black; font-weigt:bold; font-size:1.2rem;';
		this.title = this.pooltitle;
		
		this.bold_black = 'background:ivory; color: #171717; font-weight:bold';
		this.black = 'background:ivory; color: #171717';
		this.green = 'background: ivory; color: green';
		this.orange = 'background: ivory; color: orange;';
		this.redbig = 'background: red; color: white';
		this.yellow = 'background: ivory; color: yellow';
		
		this.part_cnt = this.bgpink;
		this.part_depth = this.bgyellow;
		this.part_id = this.bglightblue;
		this.part_full = this.bggreen;
		this.part_hex = this.bglightblue;
		this.part_mod = this.bglightred;
		this.part_mod_ph = this.bgorange;
		this.part_part = this.bglightgreen;
		this.part_sig = this.bglightblue;
		
		this.action = 'background:red;color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.aux = 'background:salmon; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.connector = 'background:purple; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.manner = 'background:indianred; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.mode = 'background:firebrick; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
			
		this.tab0 ='';
		this.tab1 ='██';
		this.tab2 ='████';
		this.tab3 ='██████';
		this.tab4 ='████████';
		this.tab5 ='██████████';
		this.tab6 ='████████████';
		this.tab7 ='██████████████';
		this.tab8 ='████████████████';
		this.tab9 ='██████████████████';
		this.tab10='████████████████████';


		this.BLACK = 30;

		this.RED = 31;
		this.TRADE_RED = 311;
		
		this.GREEN = 32;
		this.TRADE_GREEN = 321;
		
		this.YELLOW = 33;
		this.YELLOW_i = 330;
		this.TRADE_YELLOW = 331;
		
		this.BLUE = 34;
		this.BLUE_i = 340;
		
		this.PINK = 35;
		this.CYAN = 36;
		this.WHITE = 37;

	}


	printf=function(...$args) {
		var lastColor = TALK.WHITE;
		var html = "";
		$args.forEach($arg => {
			if(typeof $arg==="string") html+="<span class='copr-"+lastColor+"'>"+$arg+"&nbsp;</span>";
			else {
				lastColor = parseInt($arg);
				if (lastColor==undefined) lastColor = TALK.WHITE;	
			}
		});
		return html;
	}

	/**
	 * 
	 * @param {*} $int 
	 * @returns 
	 */
	console_color_map = function($int) {
		$int = parseInt($int);
		if ($int==ANNE.TALK.BLACK) return this.bgblack;
		else if ($int==ANNE.TALK.RED) return this.bgred;
		else if ($int==ANNE.TALK.GREEN) return this.bggreen;
		else if ($int==ANNE.TALK.YELLOW) return this.bgyellow;
		else if ($int==ANNE.TALK.BLUE) return this.bgblue;
		else if ($int==ANNE.TALK.PINK) return this.bgpink;
		else if ($int==ANNE.TALK.CYAN) return this.bgcyan;
		else if ($int==ANNE.TALK.WHITE) return this.bgwhite;
		return this.bgblack;
	}
	
	printf_console=function(...$args) {
		var lastColor = this.bgblack;
		var strs = [];
		var colors = [];
		$args.forEach($arg => {
			if(typeof $arg==="string") {
				strs.push("%c"+$arg);
				colors.push(lastColor);
			}
			else {
				lastColor = this.console_color_map(parseInt($arg));
				if (lastColor==undefined) lastColor = this.bgwhite;	
			}
		});
		console.log(strs.join(""),...colors);
	}



	/**
	 * 
	 * @param {*} $vb 
	 * @param {*} $tense 
	 * @returns 
	 */
	 englishTensify = function($vb, $tense) {
		if ($tense=="past tense") {
			var rev = $vb.split("").reverse().join("");
			var c0 = rev[0];
			var c1 = "";
			if (rev.length>1) c1 = rev[1];
			if (c0=="e") return $vb+"d";
			if (c0=="y" && $text.length>1) return $vb.substring(0, $vb.length-1)+"ied";
			if (this.vowel.indexOf(c1)!=-1 && this.vowel.indexOf(c0)==-1) return $vb+c0+"ed";
			if (c0=="c") return $vb+"ked"; 
			return $vb+"ed";
		}
		else return $vb;
	}


	/**
	 * For a given role, what is the EC type we expect to fill this role?
	 * @param {String} $role_nid - nid long string 
	 * @returns 
	 */
	__getTopMatchRoleFromRole=function($role_nid) {
        if ($role_nid==MEM.role_action_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_theme_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_action.temp_nid) return MEM.ec_action.temp_nid;
        else if ($role_nid==MEM.role_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_actor.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_affected_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_posn_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_be_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_state.temp_nid) return MEM.ec_state.temp_nid;
        else if ($role_nid==MEM.role_when.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_where.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_manner.temp_nid) return MEM.ec_manner.temp_nid;
        else if ($role_nid==MEM.role_mode.temp_nid) return MEM.ec_mode.temp_nid;
        else if ($role_nid==MEM.role_connector.temp_nid) return MEM.ec_expect_more_of_prev.temp_nid;
        else if ($role_nid==MEM.role_aux_action.temp_nid) return MEM.ec_expect_action.temp_nid;
        else if ($role_nid==MEM.role_thing_mod.temp_nid) return EM.ec_expect_thing_mod.temp_nid;
        else if ($role_nid==MEM.role_question.temp_nid) return MEM.role_question.temp_nid;
        else {
            var nOBJ = MEMG.getNeuronByTempNid($temp_nid);
            console.log("%c unhandled (__getTopMatchRoleFromRole) %c "+nOBJ.text, TALK.bgred, TALK.green);
        }
    }

	/**
	 * 
	 * @param {*} $balanceObj 
	 */
	printBalanceObj=function($balanceObj) {
		console.log($balanceObj);
		console.log("%c annecoin "+ANNE.UTILS.commas($balanceObj.annecoin), TALK.bgblackyellow);
		console.log("%c annies "+ANNE.UTILS.commas($balanceObj.annies), TALK.bglightblue);

		if ($balanceObj.avail_now_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_now_annies/100), TALK.bglightblue);
		}
		if ($balanceObj.avail_soon_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_soon_annies/100), TALK.bglightblue);
		}
	}

	/**
	 * Shift an english word from the stored EC type to the desired EC type.
	 * 
	 * @param {anne_neuron} $roleNOBJ - the role we are shifting from. 
	 * @param {*} $valueNOBJ - the value of the role we are shiting from.
	 * @returns 
	 */
	__engShifter = function($roleNOBJ, $valueNOBJ) {
		console.log("(TALK.__engShifter)", $roleNOBJ, $valueNOBJ);

		//the top EC type of the role we are shifting from.
		var ec_top_desired_tnid = ANNE.TALK.__getTopMatchRoleFromRole($roleNOBJ.temp_nid);
		if (ec_top_desired_tnid==null) return;

		//the ec type we want to shift to.
		var ec_top_observed_tnid = $valueNOBJ.setECConceptualizingParent();
			
		//if the type is already right, return the text of the from.  No shift is needed.
		if (ec_top_desired_tnid==ec_top_observed_tnid) return $valueNOBJ.text;
		
		//we need to find the rule that shifts this concept from ec_top_observed_nid to the ec_top_desired_nid!
		for (var g=0; g < MEM.eng_grammar_rule.children.length; g++) {
			var ruleNOBJ = MEMG.getNeuronByTempNid( MEM.eng_grammar_rule.children[g]);
			var role_src = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_source.temp_nid);
			if (role_src==null) continue;
			if (role_src.t!=ec_top_observed_tnid) continue;	

			//we have a rule that starts from where our concept starts.
			//now we need to see if it converts to our desired
			var role_goal = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_goal.temp_nid);
			if (role_goal==null) continue;
			if (role_goal.t!=ec_top_desired_tnid) continue;

			//we found a rule!
			//we need to now look for the suffixes that use this rule.
			for (var i=0; i < ruleNOBJ.ins.length; i++) {
				var in_rel = ruleNOBJ.ins[i];
				if (in_rel.type!=TYPE_HAS) continue;
				if (in_rel.r!=MEM.eng_grammar_rule.temp_nid) continue;

				//we found a suffix that uses this rule.
				var suffixNOBJ = MEMG.getNeuronByTempNid(in_rel.f);
				if (suffixNOBJ==null) continue;
				return TALK.__applyOutEngSuffix(suffixNOBJ, $valueNOBJ);
			}
		}

		console.log("%c could not find rule.");
	}


	/**
	 * 
	 * @param {anne_neuron} $suffixNOBJ 
	 * @param {anne_neuron} $valueNOBJ 
	 * @returns 
	 */
	__applyOutEngSuffix=function($suffixNOBJ, $valueNOBJ) {
		console.log("(TALK.__applyOutEngSuffix)");
		var suffixOBJ  = ANNE.BRAIN.NLP.LC._suffixMap.getByText($suffixNOBJ.text);
		if (suffixOBJ==null || suffixOBJ.cdx_rule_out==undefined) {
			//console.log("%c NO KNOWN OUTBOUND ENGLISH RULE FOR SUFFIX "+$suffixNOBJ.text, TALK.bgredbig);
			return $valueNOBJ.text+$suffixNOBJ.text;
		}
		return $suffixNOBJ.cdx_rule_out($valueNOBJ.text);
	}



	/**
	 * Take a manner role or roles and convert them to be proper adverbs if they are not adverbs.
	 * @param {Array} $manner_roles 
	 * @returns 
	 */
	___doMannerEngShifter = function($manner_roles) {
		console.log("(TALK.___doMannerEngShifter)", $manner_roles);
		var manner_text = [];
		for (var w=0; w < $manner_roles.length; w++) {
			var rel_manner = $manner_roles[w];
			var toNOBJ = MEMG.getNeuronByTempNid(rel_manner.t);
			var manner = ANNE.TALK.__engShifter(MEM.role_manner, toNOBJ);
			manner_text.push(manner);
		}
		return TALK.listifyStrArr(manner_text);
	}

	___doThingDet = function($thing_roles) {
		console.log("(TALK.___doWhereDet)", $thing_roles);
		var thing_text = [];
		for (var w=0; w < $thing_roles.length; w++) {
			var rel_where = $thing_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				thing_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					thing_text.push("the "+toNObj.text);
				}
				else thing_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(thing_text);
	}

	___doWhereDet = function($where_roles) {
		//console.log("(TALK.___doWhereDet)", $where_roles);
		var where_text = [];
		for (var w=0; w < $where_roles.length; w++) {
			var rel_where = $where_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				where_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					where_text.push("the "+toNObj.text);
				}
				else where_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(where_text);
	}

	/**
	 * This print out the meaning (infinitive form definition) of an action neuron.
	 * Whatever ROLES are stored in the clump will define the clump.
	 * 
	 * We can also show connotations via TYPE_BRAIN and assocs but it will be 
	 * better to do that in the dossier file for this concept where we can display them
	 * more cleanly.
	 * 
	 * @param {anne_neuron} $clumpNOBJ 
	 * @param {Object} $gpat 
	 * @param {Boolean} $bDEBUG 
	 * @returns 
	 */
	verbalizeClassActionDef = function($clumpNOBJ, $gpat=null, $bDEBUG=true, $tenseOverride="", $lvl=0) {
		if  ($gpat==null) $gpat = this.eng_pat_def;
		if ($clumpNOBJ==null) return;
		var output = "";
		var parts = [];
		var order = [];

        if ($bDEBUG==true) {
			console.log("%c (verbalizeClassActionDef) lvl:"+$lvl+" for clump"+$clumpNOBJ.text, TALK.bggreenbig);
		}

		var part_of_anne = null;
		var role_pre_var = "";

		//if this is a class and not present tense, rewire to verbalize the parent action instead.
		if ($clumpNOBJ.ntype==NTYPE_CLASS && $clumpNOBJ.tense!="present tense" && $tenseOverride=="") {
			role_pre_var = "_meaning_";
		}
		
		console.log("role_pre_var: "+role_pre_var);

		//will have to fix roles here for EC TYPE shifts.. hmm.
		var doModNest=function($mods, $lvl=0, $det_func=null, $bDETFIRSTONLY=false) {
			console.log("(doModNest) lvl"+$lvl);
			var modTxt = "";
			var keys = Object.keys($mods)
			for (var k=0; k < keys.length; k++) {
				var rObj = $mods[ keys[k] ];
				if (rObj.mods!=undefined) {
					modTxt += doModNest(rObj.mods, $lvl+1, $det_func, true)+" ";
					$det_func = null;
				}
				if (rObj.nameless!=undefined) continue;
				var this_text =  "";
				if (rObj.r!=MEM.role_connector.temp_nid && $det_func!=null) {
					if ($bDETFIRSTONLY==true && k==0) this_text = $det_func([rObj]);
					else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				} 
				else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				modTxt+=this_text+" ";
			}
			return modTxt;
		}

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var pre_default = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre = "";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				var arr = $clumpNOBJ[role_pre_var+role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null) {
					mod_arr = $clumpNOBJ[role_pre_var+mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";
				order = order.concat(arr);


				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {

						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}

					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				//mod mods the WHOLE phrase if there are multiples.
				else if (role=="role_manner" && modfunc!=null) {
					to = modfunc(arr);
				}

				//{role:'role_where', mod:'role_where_mod', p:"to"},
				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}


				//new method of mods..
				else if (arr[0].mods!=undefined) {

					console.log("mods in role "+role, arr[0].mods);
					mod_arr = arr[0].mods;
					var modTxt = doModNest(mod_arr, 0, det_func);

					//the base role that has these mods (in eng pat we put the mod BEFORE)

					//note, this should not be an array, only one with our mods method.
					//will pinch it off here -- (will write a single verbalizeRole in a bit)
					to = modTxt+TALK.verbalizeRolesArr([ arr[0] ]);	

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						//var modNOBJ = MEMG.getNeuronByTempNid(mod_arr[0].t); //should also be role_text already..
						//console.log(modNOBJ);
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else {
					if (det_func!=null) to = det_func(arr);
					else to = TALK.verbalizeRolesArr(arr);
				}
				

				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		console.log(output);
		$clumpNOBJ.eval = output;
		if ($lvl!=0) return output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		var tense = $clumpNOBJ.tense;
		if ($tenseOverride!="") tense = $tenseOverride;

		return new anne_babytalk(output, order, part_of_anne, $clumpNOBJ.tense);
	}

	

    /**
     * Rudimentary verbalizer of _roles via a gpat (for lang or maodality specific orderings)
	 *
	 * @param {anne_neuron} $clumpNOBJ - the clump we want to print 
	 * @param {Object} $gpat - the language pattern we want to use
	 * @param {Array} $skipRoleNames - any roles we want to skip
     * @param {Boolean} $bDOSENTENCE - print in sentence mode
	 * @param {Boolean} $bDEBUG - degug the printing.
	 *
	 * @returns {anne_babytalk} - returns .output of sentence and .order of role tnids in the order they were printed from the gpat.
	 */
	 verbalizeInstActionClump=function($clumpNOBJ, $gpat=null, $skipRoleNames=[], $bDOSENTENCE=true, $bDEBUG=true){
        if ($gpat==null) $gpat = this.eng_pat1;
        var output = "";
		var parts = [];
		var order = [];
        if ($bDEBUG==true) console.log("%c print for clump"+$clumpNOBJ.text+" %cdosentence:"+$bDOSENTENCE, TALK.bggreenbig, TALK.bgblackyellow);
			
		var part_of_anne = null;	//EC type

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				var pre_default = "";
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre="";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				if ($skipRoleNames.length && $skipRoleNames.indexOf(role)!=-1) continue;

				var arr = $clumpNOBJ[role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null && mod!="") {
					mod_arr = $clumpNOBJ[mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";

				order = order.concat(arr);

				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {
						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}
					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);
				}

				else if (det_func!=null) {
					to = det_func(arr);
				}
		 
				else {
					to = TALK.verbalizeRolesArr(arr);
				}
				
				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		//make a sentence out of it?
		if ($bDOSENTENCE==true) {
			output = output.charAt(0).toUpperCase() + output.slice(1);
		}	
		console.log(output);
		$clumpNOBJ.eval = output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		return new anne_babytalk(output, order, part_of_anne);
    }


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	verbalizeRolesArr = function($arr) {
		if ($arr==undefined || $arr==null || $arr.length==0) return "[ERROR]";
		if ($arr.length==1) {
			var rObj = $arr[0];
			if (rObj.nameless!=undefined) return "";
			if (rObj.role_text!=undefined) return rObj.role_text;
			rObj.role_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			return rObj.role_text;
		}
		var str = "";
		var froms = [];
		var conj = "and";

		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			if (froms.length==0) froms.push(rObj.f);
			else if (froms.indexOf(rObj.f)!=-1) froms.push(rObj.f);
			else conj="or";
		}
		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			var role_text =  MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			str += role_text;
			rObj.role_text = role_text;
			if (r==$arr.length-2) str=str+" "+conj+" ";
			else if (r==$arr.length-1) ;
			else str=str+", ";
		}
		return str;
	}


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	listifyStrArr = function($arr) {
		if (ANNE.UTILS.isset($arr)==false) return "";
		if ($arr.length==1) return $arr[0];
		if ($arr.length==2) return $arr[0]+" and "+$arr[1];
		var finalstr = $arr[0];
		for (var a=1; a <$arr.length; a++) {
			if (a==$arr.length-1) finalstr = finalstr+", "+$arr[a];
			else finalstr = finalstr+" and "+$arr[a];
		}
		return finalstr;
	}


	
	/** 
	 * print simple relon tags.
	 * 
	 * This TALK class is an utter disaster.. changed a gazillion times.
	 * Use this to do basic printing.  We have lots of rework to do just
	 * for this basic talk.  Would be better to get full PAS understanding.
	 */ 
	 simpleRelonTags = function ($rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}
		if ($rObj.type==TYPE_AWARENESS) {
			var tt = TARGET_NID;
			if ($rObj.to_type==TO_TYPE_TXID) tt = TARGET_TXID;
			return this.verbalizeForAwareness($rObj, true);
		}

		var typeText = this.getTypeName($rObj.type, true, "rel-type");

		//willy wonka from name
		var fromNOBJ = null;
		var fromText = "";
		if ($rObj.from_nid!=null) {
			fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		}
		else if ($rObj.pending_from!=null) {
			fromNOBJ = MEMG.getNeuronByPendingNid($rObj.pending_from);
			if ($rObj.pending_from_debug!=undefined && $rObj.pending_from_debug.length>0) fromText = $rObj.pending_from_debug;//+"("+$rObj.pending_from+")";
			else fromText = $rObj.pending_from;
		}
		if (fromNOBJ==null) {
			console.log("NO FROM OBJ??");
			console.log($rObj);
			$rObj.error = true;
			return "[[error, bad relon]]";
		}
		else {
			if (fromNOBJ.emulated) {
				console.log("NO RELONS FROM EMULATED NEURONS -- they are not real!");
				return;
			}
			fromText = fromNOBJ.bestNameHTML();
		}
		var fromClass = this.getNeuronDisplayClass(fromNOBJ);

		//we want the name of the here even tho the from is actually
		//our nid (in the cmd relon)
		if ($rObj.debug_name!=undefined) {
			fromText = "(MAKE) "+$rObj.debug_name;
			fromClass = "neuron-ghost";
		}

		//reln text
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		if (relnNOBJ==null) {
		    console.log("problem with the reln. It must exist in local mem");
			$rObj.error = true;
		    return "[[ERROR in RELN "+$rObj.reln+"]]"
		}
		var relonText = relnNOBJ.text;
		var relnClass = "";
		if (relnNOBJ.nid==MEM.hanai_in.nid || relnNOBJ.nid==MEM.hanai_in_all.nid) relnClass = "neuron-hanai";
		else if (relnNOBJ.nid==MEM.hanai_out.nid)  relnClass = "neuron-hanai";
		else relnClass = this.getNeuronDisplayClass(relnNOBJ);
		

		if ($showfromNid==true) {
			fromText+="("+fromNOBJ.nid+")";
		}

		//to type
		var toText = this.verbalizeRelonToByToType($rObj, $bCOLORIZEBYTYPE);
		var ttypemsg = "";
		if ($ttype!=-1) ttypemsg = TXB.printType($ttype);
		if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" "+relonText+"&nbsp;<b>"+toText+"</b></span>";
		}
		else if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" MAKE LOCATION <b>&nbsp;"+toText+"</b>&nbsp;in&nbsp;"+fromText+"</span>";
		}

		//to stuff
		var toClass = "rel-to";
		var toNOBJ = null;
		if ($rObj.to_type==TO_TYPE_ANNE) {
			if ($rObj.pending_to_s!=undefined) {
				var parts = $rObj.pending_to_s.split("_");
				if (parts.length>1) {
					toNOBJ = MEMG.getNeuronByNid(parts[1].trim());
					if (toNOBJ!=null) {
						toClass = this.getNeuronDisplayClass(toNOBJ);
					}
				}
			}
			else { 
				toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
				if (toNOBJ!=null) {
					toClass = this.getNeuronDisplayClass(toNOBJ);
				}
			}
		}

		return "<span class='q-part "+fromClass+"' title='"+fromNOBJ.nid+"'>"+fromText+"</span>"+typeText+"<span class='q-part "+relnClass+"'>"+relonText+"</span><span class='q-part "+toClass+"'>"+toText+"</span><span class='q-part anne-tx-type'>*"+ttypemsg+"TX*</span>";
	}

	/**
	 * Helpful display to colorize things consistently thruout the app.
	 * This took a bit of time to arrive at -- for some time I tried to colorize
	 * class neurons red, instances blue, text green...
	 * but have now settled on ghost blue, anne yellow, text green, self red (maybe orange later), and hanai pink.
	 */
	getNeuronDisplayClass=function($nObj) {
		if ($nObj==null) {
			return "neuron-ghost";
		}
		
		if ($nObj.nid==MEM.USER._userNObj.nid) return "neuron-self";
		if ($nObj==undefined || $nObj==null) {
			return "neuron-ghost";
		}
		if ($nObj.ghost==true) return "neuron-ghost";
		else if (ANNE.isANNE($nObj.owned_by)==true) return "neuron-anne";
		else {
			if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
				return "neuron-self";
			}
			else {
				return "neuron-other"  //black for someone else's neuron (not ANNE's).
			}
		}
	}



	/**
	 * 
	 * @param {*} $realTXID 
	 * @param {*} $relon 
	 * @param {*} $sep 
	 * @param {*} $bCOLORIZE 
	 * @returns 
	 */
	genShortRelonTextByTXID=function($realTXID, $relon=null, $sep="&nbsp;»&nbsp;", $bCOLORIZE=false) {
		var relon = $relon;
		if ($relon==null) relon = MEMG.lookupRelonByTXID($realTXID);
		if (relon==null) return "";
		var fNObj = MEMG.getNeuronByNid(relon.from_nid);
		var rNObj = MEMG.getNeuronByNid(relon.reln);
		var tNOBJ = null;
		if (relon.t) tNOBJ = MEMG.getNeuronByNid(relon.to_s);
		var to = "";
		if (tNOBJ!=null) to = tNOBJ.bestNameHTML();
		else if ($relon.to_type==TO_TYPE_TXID) to="{url}";
		else if ($relon.to_type==TO_TYPE_BLOB) to="{blob}";
		else to = relon.to_s;
		var from = fNObj.text;
		if (from=="") from = fNObj.nid;
		if ($bCOLORIZE==true) {
			return "<b>"+from+"</b>"+$sep+"<em class='text-red'>"+rNObj.text+"</em>"+$sep+"<b class='text-green'>"+to+"</b>";
		} 
		return "<b>"+from+"</b>"+$sep+"<em>"+rNObj.text+"</em>"+$sep+"<b>"+to+"</b>";
	}


	/**
	 * Uses the Q to print a friendly gui.
	 */
	printSanitizedAlertCmds=function($updateElID, $aQ, $bPRINTBALANCE=true) {
		var msgDiv = ANNE.A.setHTML($updateElID, ""); //qrb_txs_list_div
		
		console.log($aQ);
		
		//append bar to qrb_progress div
		var progDiv = document.createElement("div");
		progDiv.classList = "bg-light tx-proggy";
		progDiv.setAttribute("id", "tx_totals");
		ANNE.A.append('qrb_progress', progDiv);

		var progressBar = document.createElement("div");
		progressBar.setAttribute("id", "tx_progress_bar");
		progressBar.classList = "proggy-bar positive";
		progressBar.style.display = "inline-block!important";

		ANNE.A.append(progDiv, progressBar, true);

		for (var c=0; c < $aQ.length; c++) {
			var cmdObj = $aQ[c];
			var txtlbl = "";

			if (cmdObj.cmds_special.length && cmdObj.cmds_special.indexOf(TXB.CMD_POPULATE_WITH_VARS_ONLY_NO_SEND)!=-1) {
				continue;//no need to display this in the gui.
			}
			
			if (cmdObj.relon!=undefined && cmdObj.relon.warning!=undefined) {
				ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q bg-warning", "tx", c+"x_warn",  cmdObj.relon.warning);
			}

			var msg = "";
			if (cmdObj.msg!="" && cmdObj.relon!=null && cmdObj.relon.debug_name==undefined) {
				msg = cmdObj.msg;
			}
			else if (cmdObj.relon!=undefined) {
				//$rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false
				msg = ANNE.TALK.simpleRelonTags(cmdObj.relon, cmdObj.TXTYPE, false);
			}

			
			if (cmdObj.relon!=undefined && cmdObj.relon!==null) {
				if (cmdObj.relon.type==TYPE_PARAM) {
					msg = msg +" <span class='tiny-text'>PARAM setter</span>";
				}
			}

			if (cmdObj.relon!=undefined && cmdObj.relon.sys_ref_txid!=undefined) {
				msg = msg +" <span class='tiny-text'>&nbsp;depends on "+cmdObj.relon.sys_ref_txid+"</span>";
			}

			if (cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_CHILD_TO_CMD)!=-1 
			|| cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_SELF_TO)!=-1) {
				msg = msg +"<span class='tiny-text'>&nbsp;ENCRYPTED AFTER NID IS ACQUIRED</span>";
			}

			var ico = "";
			ico = UTILS.formatNumber(cmdObj.totalAnnies/100)+ANNE.GUI_STUFF.getAnneImgHtml('16px');

			//friendly ico display in the q
			var ico3 = "";
			if (cmdObj.relon!=undefined) {
				ico3 = ANNE.TXB.getTxICOByType(cmdObj.TXTYPE, cmdObj.relon, "16");
			}

			var txtlbl = "<span class='q-relon-link'>tx #"+(c+1)+"</span>"+ico3;
			var total = "<span class='q-flexright' style='text-align:right'>"+ico+"</span>";

			var finalmsg = txtlbl+msg+total;

			ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q", "tx", c+"x",  finalmsg);
		}
	}


	/**
	*/
	printNids = function($nids, $title="") {
		if ($title!="") console.log("%c"+$title, ANNE.TALK.bgbigred);
		
		for (var n=0; n < $nids.length; n++) {
			var nObj = MEMG.getNeuronByNid($nids[n]);
			console.log("n="+n+" "+nObj.text);
		}	
	}

	
	/**
	 *
	 */
	updateSantizedAlertCmd=function($updateElID, $cmdObj, $current_index, $total, $badgeEl="") {
		var txid = $cmdObj.broadcast_txid;
		var percentFloat = ($current_index+1)/$total;
		var bFINAL = false;
		var remain = $total-($current_index+1);
		
		if (remain>0) ANNE.A.setHTML($badgeEl, remain);
		else ANNE.A.setHTML($badgeEl, "");
		
		if ($current_index+1==$total) bFINAL=true;

		var percent = Math.ceil(percentFloat*100);
		ANNE.A.setStyle("tx_progress_bar", {width:percent+"%"}, true);
		var msgDiv = document.getElementById($updateElID);

		//try to track with the element getting updated.
		
		if (($cmdObj.skipped==undefined || $cmdObj.skipped==false) && txid!=undefined && txid.length>0) {
			var msg = "";
			if ($cmdObj.msg=="") msg = "success! ";
			else msg = $cmdObj.msg;
		
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-green-themeDARK", "bg-txid-alert-q");
			
			//update the txid #n with the txid ico.
			var txidcol = tx_alert.querySelector(".q-relon-link");
			if (txidcol!=null && txid!="") {
				ANNE.MEM_UDF.makeRelonViewLink(txidcol, txid, true, 12, true);
			}

			//if this was the last $cmdObj, print the complete transactions.  :)
			if (bFINAL==true) {
				var ezal = ANNE.A.ezAlert(msgDiv, "bg-green-themeDARK", "", "",  "<b>"+$total+"</b>&nbsp;Transactions Complete");
				ezal.classList.add("txs-complete");
				
			}
			return;
		}

		else if ($cmdObj.skipped!=undefined && $cmdObj.skipped==true) {
			var msg = "skipped::";
			if ($cmdObj.skipped_reason!=undefined) msg += $cmdObj.skipped_reason;
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-red", "bg-txid-alert-q");
		}
	}



	/*** printRelonByID ************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/ 
	printRelonByID=function($id) {
		var rObj = MEMG.lookupRelonByTXID($id);
		if (rObj) this.printRelon(rObj);
	}

	//here we just want the names.
	printNeurons = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log($nObjs[n].bestNameHTML());
		}
	}

	printNObjsList = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) { console.log($nObjs[n].bestName());}
	}

	/*** printNObjs ****************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/
	printNObjs = function ($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log("***********************************");
			var nObj = $nObjs[n];
			console.log("for neuron: "+nObj.nid);
			if (nObj.mine==true) console.log("mine=true!");
			else console.log("not mine");

			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printFromTempNidArray=function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log("for neuron: "+nObj.nid);
			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printOnlyRelons = function($TYPE, $aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		for (var r = 0; r < $aRelons.length; r++) {
			if ($aRelons[r].type!=$TYPE) continue;
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
	}

	/**
	 * Print an array of relons.
	 * 
	 * @param {Array} $aRelons 
	 * @param {Boolean} $plainText 
	 * @param {Boolean} $bDOTEMPNIDS 
	 * @returns 
	 */
	printRelons = function ($aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if (ANNE.UTILS.isset($aRelons)==false) {
			console.log("no relons to print");
			return;
		} 

		console.log($aRelons);

		for (var r = 0; r < $aRelons.length; r++) {
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
		return;
	}


	printTNIDS = function($tnids){
		var texts=[];
		for (var t=0; t < $tnids.length; t++) {
			var nOBJ = MEMG.getNeuronByTempNid($tnids[t]);
			console.log(nOBJ.text);
			texts.push(nOBJ.text);
		}
		console.log(texts.join(","));
		return texts.join(",");
	}



	/*** selfRelon ******************************************* xx/xx/2020 ***
	******************************************************** Mr.Scatman ****/ 
	selfRelon = function ($relon, $bYOU=true) {
		var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var toObj = MEMG.getNeuronByTempNid($relon.t);
		var totxt="";
		if (toObj!=null) totxt = toObj.text;
		else totxt = $relon.to_s;

		var your = "Your";
		if ($bYOU==false) your = "Their";
		var you = "You";
		if ($bYOU==false) you = "They";

		var msg = "<span class='text-primary'> "+you+"</span>";
		if ($relon.type==TYPE_AWARENESS) {
			
			var toTxt = "";
			if ($relon.to_type==TO_TYPE_ANNE) {
				toTxt = this.verbalizeForAwareness($relon);
			}
			else toTxt = $relon.to_s;
			return your + " "+relnTxt + " "+toTxt;
		}

		else if ($relon.type==TYPE_BE) {
			if ($relon.to_type==TO_TYPE_STRING) {
				var your = "<span class='text-primary'> "+your+"</span>";
				if ($relon.r==MEM.name.temp_nid) return your+" <span class='text-danger'>name </span>is <span class='text-success-lbl'>"+totxt+"</span>";
				else if ($relon.r==MEM.aka.temp_nid) return msg+" are <span class='text-danger'>also known as </span>is <span class='text-success-lbl'>"+totxt+"</span>";
			}
			msg = msg +" are ";
			var determiner = this.getDet(totxt);
			return msg +determiner+" <span class='text-danger'>"+totxt+"</span>";
		}
		else if ($relon.type==TYPE_SYS) {
			if (toObj!=null) {
				if ($relon.r==MEM.necrosis.temp_nid) {
					"<span class='text-danger-lbl'>"+relnNOBJ.text+" (delete) </span> <span class='text-success'>"+totxt+"</span>";;
				}
				else return relnNOBJ.text+" = "+totxt;
			}
			else return relnNOBJ.text+" = "+$relon.to_s; 
		}
		else if ($relon.type==TYPE_HAS) {
			msg = msg+" have ";	
			var determiner = this.getDet(relnNOBJ.text);
			var reln_txt = "<span class='text-danger'>"+relnNOBJ.text+"</span>";
			msg = msg+determiner+" "+reln_txt+" of ";

			var toStr = "";
			if (toObj!=null) toStr = totxt;
			else toStr = $relon.to_s;

			var to_txt = ""
			if (toObj!=null && toObj.bTERMINATING_TYPE==true) to_txt = "<span class='text-success'>"+toStr+"</span>";
			else to_txt = "<span class='text-danger'>"+toStr+"</span>";
			return msg + to_txt;
		}
	}


	printListByReln = function($outs, $r, $excludeTypes=[TYPE_PARAM], $sep=", ") {
		var relons = MEMG.getRelonsByTempR($outs, $r, $excludeTypes);
		if (relons.length==0) return "";
		return MEMG.getTextListByRelonTos(relons, $sep);
	}


	colorizeByPol=function($nObj) {
		if ($nObj.polarity==undefined) MEM.polarity.set_value($nObj);
		var fclass = "";
		if ($nObj.polarity==1) fclass = "text-success-lbl";
		else if ($nObj.polarity==-1) fclass = "text-danger-lbl";
		else if ($nObj.polarity==0) fclass = "text-warning-lbl";
		return "<span class='q-part "+fclass+"'>"+$nObj.text+"</span>";
	}


	printEzName = function($nObj) {
		var text = $nObj.text;
		var be = $nObj.be;
		if (text=="") {
			return "<em>(unnamed) "+be+"</em> <span class='raw-address'>"+$nObj.nid+"</span>";
		}

		return "<em>"+be+"</em> <span style='color:black; font-weight:bold'>"+text+"</span>&nbsp;(<span class='raw-address'>"+$nObj.nid+"</span>)";
	}



	/**
	 * Get a determiner for a given string.
	 * 
	 * Have to go to phonemes to get accurate with this.
	 * Not doing that now.  going the cheap and inaccurate road.
	 * 
	 * @author {Mr.Scatman}
	 */ 
	getDet=function ($str) {
		var str =$str.toLowerCase();
		var char = str.substring(0,1);
		var aeiou = "aeiou";
		if (aeiou.indexOf(char)!=-1) return "an";
		return "a";	
	}

	/**
	 * Another rudimentary det function in lieu of using NLP in this app.
	 * (that comes later with a more robust vocab and understanding of lang in the circuits..
	 * ...tho.. technically... we can use the NLP classes to parse now)
	 * @param {*} $nObj 
	 * @returns 
	 */
	getDet2=function ($nObj) {
		var str = $nObj.text.toLowerCase();

		if ($nObj.ntype==NTYPE_INST) {
			if ($nObj.bes.indexOf(MEM.person.temp_nid)!=-1 || $nObj.bes.indexOf(MEM.location.temp_nid)!=-1) return $nObj.text;
			return "the "+$nObj.text;
		}

		//an elephant, an apple, a peach, a car, a dog.
		else {
			var char = str.substring(0,1);
			var aeiou = "aeiou";
			if (aeiou.indexOf(char)!=-1) return "an "+$nObj.text;
			return "a "+$nObj.text;		
		}
	
	}
	


	/** for later.. */
	toSoundEncoding(word) {
		return word.toLowerCase().split('')
		.map(function(c) { return encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */ 
	soundsLike=function($w1, $w2) {
		return this.toSoundEncoding($w2) === this.toSoundEncoding($w2);
	}


	/** for later.. */ 
	toSoundEncoding=function($w) {
		return $w.toLowerCase().split('')
		.map(function(c) {return this.encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */
	encodeCharacter=function(c) {
		if ('kgq'.indexOf(c) >= 0) return 'k'; // velar stops
		if ('bpfv'.indexOf(c) >= 0) return 'p'; // labials
		if('td'.indexOf(c) >= 0) return 't'; // alveolars
		if ('szjc'.indexOf(c) >= 0) return 's'; // shushes
		if ('mn'.indexOf(c) >= 0) return 'm'; // nasals
		if ('lr'.indexOf(c) >= 0) return 'l'; // glides
		if ('aeiou'.indexOf(c) >= 0) return 'o'; // vowels
		if ('yhw'.indexOf(c) >= 0) return '';
		return '';
	}
	

	/**
	 * Quick n dirty ending printer for frenly day printing of dates.
	 */
	lazyDayIntEnding = function($int) {
		if ($int==1 || $int==21 || $int==31) return $int+"st";
		else if ($int==2 || $int==22) return $int+"nd";
		else if ($int==3 || $int==23) return $int+"rd";
		else return $int+"th";
	}


	/**
	 * poor man's printing -- no intelligence involved.	
	 * quite rudimentary -- but for now, it works.
	 */
	humanReadableRelon = function ($relon, $bNOFONT=false) {
		//redo this.. for now, just print relon.
		return this.printRelon($relon, true, false);
	}


	/**
	 */
	whoOwnsPhrase = function($who, $nObj=null) {
		if ($who=="anne") return "A.N.N.E. owns this neuron";
		else if ($who=="app") return "This app owns this neuron";
		else if ($who=="me") {
			if ($nObj.temp_nid==USER._userNObj.temp_nid) return "<span class='text-primary'>This IS you!</span>";
			else return "<span class='text-primary'>You owns this neuron!</span>";
		}
		else if ($who=="worldview") return "This is in your <span class='text-primary'>worldview</span>";
		
		return "**This neuron is not in your worldview**";
	}


	
	/**
	 */ 
	 verbalizeForAwareness = function ($relon, $bTAGIFY=true) {
		if ($relon.to_type!=TO_TYPE_ANNE) return " [[[don't use this w/ non neuron tos]]] ";
		var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
		var rNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var relnTxt = this.colorizeByPol(rNOBJ);

		//opinionz are just keywords..
		//if (rNOBJ.bes[0]==MEM.symbol.temp_nid) {
			//var determiner = this.getDet(relnTxt);
			//if ($bTAGIFY==false) return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+relnTxt;	
			//return "<span class='type-aware'>AWARE</span><span class='q-part rel-from'>I</span> <span class='q-part'>think </span><span class='q-part'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>is</span> "+relnTxt;	
		//}

		//else 
		if (rNOBJ.nid==MEM.rememberz.nid) {
			if ($bTAGIFY==false) return "rememberz "+toNOBJ.bestNameHTML();
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-reln'>rememberz</span><span class='q-part reln-to'>"+toNOBJ.bestNameHTML()+"</span>";
		}

		//the rest are pos/neg/neu feelings (buckets) or feelz (most refined expression of feeling)
		else {
			if ($bTAGIFY==false) return ""+toNOBJ.bestNameHTML()+" makes me FEELZ "+relnTxt;
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-to'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>makes</span><span class='q-part rel-from'>me</span>"+relnTxt;
		}
	}


	/**
	 */
	 printAwarenessForDisplay = function($relnNOBJ, $tnid) {
		var tnid = parseInt($tnid);
		var targetNOBJ = MEMG.getNeuronByTempNid(tnid);
		if (targetNOBJ==null) return "fix this";
		if (targetNOBJ.bes[0]==MEM.opinionz.temp_nid) {
			var determiner = this.getDet($relnNOBJ.text);
			return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+$relnNOBJ.text;	
		}
		else if ($relnNOBJ.nid==MEM.rememberz.nid) return "rememberz "+targetNOBJ.bestNameHTML();	
		return targetNOBJ.bestNameHTML()+" makes me FEELZ "+$relnNOBJ.bestNameHTML();
	}


	//tells us what something is from a text search.
	//this function is used as a priority when searching for things
	//we prioritize YOU first, then ANNE, then temp-worldview-pass (review this), then outside the network.
	verbalizePriorityFromTallySummary=function($tally, $textVal, $parentNObj=null, $prop="name") {
		var who = ""
		var has = "has";
		if ($tally.me) {
			who = "YOU";
			has = "have";
		}
		else if ($tally.anne) who = "A.N.N.E.";
		else if ($tally.app) who = "THIS '"+MEM_UDF.THIS_APP.text.toUpperCase()+"' APP";
		else if ($tally.worldview) who = "An adopted neuron in your network";
		else if ($tally.wv_pass) who = "A temporary-worldview-pass neuron"; 
		else if ($tally.other) who = "Someone outside of your network";
		who = "<span style='color:white'>"+who+"</span>"+"&nbsp;"+has;
		var search_msg = who+"&nbsp;already created a neuron with the same '"+$prop+"' <span class='text-success'>&nbsp;'"+$textVal+"'</span>";
		if ($parentNObj != null) search_msg=search_msg+"&nbsp;with parent class <span class='text-success'>&nbsp;'"+$parentNObj.text+"'</span>";
		msg = "As a friendly reminder;&nbsp;"+search_msg+"!  Are you sure you want to create another one?"; 
		return msg;
	}


	/**
	 */
	 typeTalk=function($type, $bTEXTONLY=false, $invertBG=false) {
		if ($bTEXTONLY) return this.type_pre[parseInt($type)];
		var invert = "";
		if ($invertBG==true) invert="-invert";
        var type_string="";
        if ($type==TYPE_SYS_MSG) type_string="<span class='type-sys"+invert+"'>sys msg</span>";
        else if ($type==TYPE_SYS) type_string="<span class='type-sys"+invert+"'>sys</span>";
        else if ($type==TYPE_BE) type_string="<span class='type-be"+invert+"'>be</span>";
        else if ($type==TYPE_PARAM) type_string="<span class='type-param"+invert+"'>param</span>";
        else if ($type==TYPE_HAS) type_string="<span class='type-has"+invert+"'>has</span>";
        else if ($type==TYPE_AWARENESS) type_string="<span class='type-aware"+invert+"'>aware</span>";
        else if ($type==TYPE_CLUMP) type_string="<span class='type-clump"+invert+"'>clump</span>";
		else if ($type==TYPE_ROLE) type_string="<span class='type-role"+invert+"'>role</span>";
		else if ($type==TYPE_BRAIN) type_string="<span class='type-brain"+invert+"'>brain</span>";
		else if ($type==TYPE_EPISODIC) type_string="<span class='type-episodic"+invert+"'>episodic</span>";
		else if ($type==TYPE_EXPERIENCE) type_string="<span class='type-experience"+invert+"'>experience</span>";
		else if ($type==TYPE_INTERNAL_PAY) type_string="<span class='type-pay"+invert+"'>internal payment</span>";
		else if ($type==TYPE_BROADCAST_PAY) type_string="<span class='type-pay"+invert+"'>broadcast payment</span>";
		else if ($type==TYPE_BRAIN120) type_string="<span class='type-brain120"+invert+"'>brain120 backlink</span>";
		else type_string = "unhandled ("+$type+")";
		return type_string;
    }


	/**
	 * 
	 * @param {*} $type 
	 * @param {*} $fromNOBJ 
	 * @param {*} $relnNOBJ 
	 * @param {*} $toType 
	 * @param {*} $toText 
	 */
	quickRelonLinePrint=function($type=null, $fromNOBJ=null, $relnNOBJ=null, $toType=null, $toText=null) {
		var parts = [];
		if ($type!=null) parts.push("["+ANNE.TALK.typeTalk($type)+"]");
		if ($fromNOBJ!=null) parts.push($fromNOBJ.bestNameHTML()+" ->");
		if ($relnNOBJ!=null) parts.push($relnNOBJ.bestNameHTML());
		if ($toType!=null) parts.push("["+TALK.toTypeTalk($toType)+"] ->");
		if ($toText!=null) parts.push($toText);
        return parts.join(" ");
	}

	
	/**
	 */
    toTypeTalk=function($to_type, $bI=false) {
        var to_type_string="";
        if($to_type==TO_TYPE_SEED) to_type_string="encrypted seed";
        else if($to_type==TO_TYPE_ANNE) to_type_string="address";
        else if($to_type==TO_TYPE_TXID) to_type_string="txid";
        else if($to_type==TO_TYPE_URL) to_type_string="url";
        else if($to_type==TO_TYPE_NUMBER) to_type_string="number";
        else if($to_type==TO_TYPE_DATE) to_type_string="date";
        else if($to_type==TO_TYPE_DATETIME) to_type_string="datetime";
        else if($to_type==TO_TYPE_STRING) to_type_string="string";
        else if($to_type==TO_TYPE_BLOB) to_type_string="blob";
        else if($to_type==TO_TYPE_TIME) to_type_string="time";
        else if($to_type==TO_TYPE_BOOLEAN) to_type_string="boolean";
        else if($to_type==TO_TYPE_ENCR_ANNE) to_type_string="encr addr";
        else if($to_type==TO_TYPE_ENCR_TXID) to_type_string="encr txid";
        else if($to_type==TO_TYPE_ENCR_URL) to_type_string="encr url";
        else if($to_type==TO_TYPE_ENCR_NUMBER) to_type_string="encr #";
        else if($to_type==TO_TYPE_ENCR_DATE) to_type_string="encr date";
        else if($to_type==TO_TYPE_ENCR_DATETIME) to_type_string="encr dtime";
        else if($to_type==TO_TYPE_ENCR_STRING) to_type_string="encr str";
        else if($to_type==TO_TYPE_ENCR_BLOB) to_type_string="encr blob";
        else if($to_type==TO_TYPE_ENCR_TIME) to_type_string="encr time";
        else if($to_type==TO_TYPE_ENCR_BOOLEAN) to_type_string="encr bool";
        

        if($bI==true) return "<i>"+to_type_string+"</i>";
        else return to_type_string;
    }


	/**
	 */
	use_d_pat=function($d_pat, $nObj, $bDOLINKS=true, $linkType=LINKS_NEURON_ANNET) {
		if ($d_pat==undefined || $d_pat==null || $d_pat.length==0) return;
		var msg = "";
		for (var d=0; d < $d_pat.length; d++) {
			var pat = $d_pat[d];
			var prop = pat.prop;
			var pre = pat.pre;
			var post = pat.post;
			var bALT = false;
			if (prop==undefined) continue;
			var val = $nObj[prop];
			if (val==undefined || val==null) {
				if (pat.alt_lbl!=undefined) {
					val = pat.alt_lbl;
					bALT = true;
				}
				else continue;
			}
			if ($bDOLINKS==true && bALT==false) {
				var valNOBJ = MEMG.getNeuronByTempNid(val);
				if (valNOBJ==null) {
					if (pat.cancel_prev!=undefined) return "";
					continue;
				}
				val = LINKS.makeNeuronLink(valNOBJ, $linkType);
				if (pat.show_aid!=undefined && pat.show_aid==true) {
					val = val +"&nbsp;("+valNOBJ.nid+")";
				}
			}

			if (pre!="") {
				if (msg!="") msg = msg+"&nbsp;"+pre;
				else msg = pre;
			}
			if (msg=="") msg = val;
			else msg = msg+"&nbsp;"+val;
			
			if (post!="") msg = msg+"&nbsp;"+post;
		}

		return msg.trim();
	}

	printTempNids = function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log(nObj.text, nObj);
		}
	}

	/**
	 * plaintext no longer does anything.
	 * this function is a DISASTER, fay.  -dlm
	 */
	printRelon = function ($rObj, $plainText=false, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}

		//console.log($rObj);
		var typeText = this.getTypeName($rObj.type);
		var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		if (fromNOBJ==null) {
			return;
		}

		var emulated = "";
		var fromText = fromNOBJ.text;
		if (fromText=="") fromText = fromNOBJ.nid;
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		var relonText = relnNOBJ.text;
		if ( fromNOBJ.emulated) emulated = "(emulated) ";
		var relonStr = emulated+typeText + " from:" + fromText+ "  reln:"+relonText;
		var toText = this.verbalizeRelonToByToType($rObj);
		relonStr += " to:"+toText;

		var toSPrintOnly = this.toSPrint($rObj);

		var freeze = fromNOBJ.freeze;
		if (freeze==1) console.log("%cfrom nid froze", TALK.bgivoryblue);

		var ghost = "";
		if ($rObj.ghost==true) ghost="(gh)";

		var sysreftxid=""
		if ($rObj.sys_ref_txid!=undefined) sysreftxid=" sysref:"+$rObj.sys_ref_txid;

		var annies=""
		if ($rObj.annies!=undefined) annies=" annies:"+$rObj.annies;
		
		var fhi="";
		if ($rObj.flathexint!=undefined) fhi=" fhi:"+$rObj.flathexint;
		
		if ($consoleColor!=-1) {
			console.log("%cg"+ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi, $consoleColor);
		}
		else {
			console.log(ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi);
		}

		if ($plainText==true) return emulated+"[" + typeText+"]"+fromText+relonText+toSPrintOnly;
		return relonStr;
	}


	toSPrint=function($rObj){
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, false);
		return toText;
	}


	/**
	 */
	verbalizeRelonToByToType = function ($rObj, $bCOLORIZEBYTYPE=true) {
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, $bCOLORIZEBYTYPE);
		return toText;
	}


	/**
	 */ 
	getToTextByToType = function ($to_type, $to_s, $colorize=true) {
		var text = $to_s;
		if ($to_type == TO_TYPE_ANNE) {
			var nObj = MEMG.getNeuronByNid($to_s);
			if (nObj) {
				if ($colorize==true) text = nObj.bestNameHTML();
				else text = nObj.bestName();
			}	
		}
		if ($colorize) return "<span class='to-type tt-"+$to_type+"'>"+text+"</span>";
		else return text;
	}


	/**
	 */
	getTypeName = function ($type, $bDOCLASSHTML=false, $addClass="") {
		if ($bDOCLASSHTML==false) {
			if ($type == TYPE_SYS || $type=="sys") return 'SYS';
			if ($type == TYPE_BE || $type=="be") return 'BE';
			if ($type == TYPE_PARAM || $type=="para") return 'PARAM';
			if ($type == TYPE_HAS || $type=="has") return 'HAS';
			if ($type == TYPE_AWARENESS || $type=="awareness") return 'AWARENESS';
			if ($type == TYPE_SYS_MSG || $type=="msg") return  'SYS MSG';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return  'EXPER';
			if ($type == TYPE_CLUMP || $type=="clump") return  'CLUMP';
			if ($type == TYPE_ROLE || $type=="role") return  'ROLE';
			if ($type == TYPE_BRAIN || $type=="brain") return  'BRAIN';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return  'INTERNAL PAYMENT';
			if ($type == TYPE_BROADCAST_PAY || $type=="broadcast payment") return  'BROADCAST PAYMENT';
			if ($type == TYPE_BRAIN120 || $type=="brain120") return  'BRAIN120';
			
			
		}
		else {
			if ($type == TYPE_SYS || $type=="sys") return '<span class="type-sys '+$addClass+'">SYS</span>';
			if ($type == TYPE_BE || $type=="be") return '<span class="type-be '+$addClass+'">BE</span>';
			if ($type == TYPE_PARAM || $type=="param") return '<span class="type-param '+$addClass+'">PARAM</span>';
			if ($type == TYPE_HAS || $type=="has") return '<span class="type-has '+$addClass+'">HAS</span>';
			if ($type == TYPE_AWARENESS || $type=="awareness") return '<span class="type-awareness '+$addClass+'">AWARE</span>';
			if ($type == TYPE_SYS_MSG || $type=="msg") return '<span class="type-sys-msg '+$addClass+'">SYS MSG</span>';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return '<span class="type-experience '+$addClass+'">EXPER</span>';
			if ($type == TYPE_CLUMP || $type=="clump") return '<span class="type-clump '+$addClass+'">CLUMP</span>';
			if ($type == TYPE_ROLE || $type=="role") return '<span class="type-role '+$addClass+'">ROLE</span>';
			if ($type == TYPE_BRAIN || $type=="brain") return '<span class="type-brain '+$addClass+'">BRAIN</span>';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return '<span class="type-brain '+$addClass+'">INTERNAL PAYMENT</span>';
			if ($type == TYPE_BROADCAST_PAY || $type=="broadcast payment") return '<span class="type-brain '+$addClass+'">BROADCAST PAYMENT</span>';
			if ($type == TYPE_BRAIN120 || $type=="brain120") return '<span class="type-brain '+$addClass+'">BRAIN120</span>';
			
		}

		//anything 100 and over is typebrain 0, 1, 2 etc
		if ($type>=20 && $type<120) return "BRAIN"+(parseInt($type));
		return 'unknown '+$type;
	}


	/**
	 */
	pluralizeNounIf = function($mixed, $baseWord, $plural, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $baseWord+$plural;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $baseWord+$plural;
		return $baseWord+$singularPostFix;
	}

	pluralizeNounIf2 = function($mixed, $baseWord, $pluralwd, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $pluralwd;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $pluralwd;
		return $baseWord+$singularPostFix;
	}

	/**
	 */
	pluralizeNP = function ($arr, $baseWord, $plural, $conj="or", $nestedProp="content") {
		if ($arr.length>1) return $baseWord+$plural+" "+this.listify($arr, $conj, $nestedProp);
		else  {
			if ($nestedProp!="") return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0][$nestedProp].text+"</span>";
			else return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0].text+"</span>";
		}
	}


	/**
	 */
	listify=function($arr, $conj, $nestedProp="content"){
		var list = "";
		for (var a=0; a < $arr.length; a++) {
			var sep = "";
			if ((a+1)==$arr.length) sep = "</span>&nbsp"+$conj+"&nbsp;<span class='text-success-lbl'>";
			else if (a>=1) sep = ",&nbsp;";
			else list = "<span class='text-success-lbl'>";
			
			if ($nestedProp) list += sep+""+$arr[a][$nestedProp].text+"";
			else list += sep+"<b>"+$arr[a].text+"</b>";
		}
		return list+"</span>";
	}


	/**
	 */
	prounounPick=function($arr, $pr, $prpl) {
		if ($arr.length>1) return $prpl;
		else return $pr;
	}


	//To make regular nouns plural, add ‑s to the end. ...
	//If the singular noun ends in ‑s, -ss, -sh, -ch, -x, or -z, add ‑es to the end to make it plural. ...
	//In some cases, singular nouns ending in -s or -z, require that you double the -s or -z prior to adding the -es for pluralization.
	pluralizeNoun=function($text) {
		var rev = $text.split("").reverse().join("");
		var c0 = rev[0];
		var c1 = "";
		if (rev.length>1) c1 = rev[1];
		if (c0=="s" || c0=="x" || c0=="z") {
			if (c0=="s" && c1=="s") return $text+"s";
			if (c0=="s") return $text+"ses";
			if (c0=="z") return $text+"zes";
			return $text+"es";
		}
		if (c0=="h" && (c1=="s" || c1=="c")) return $text+"es";
		return $text + "s";
	}


	/**
	 */
	beState = function($array) {
		if ($array.length==0) return "are";
		else if ($array.length==1) return "is";
		return "are";
	}

	/**
	 * make an english list phrase of "x, y, z and a" from an array of [x, y, z, a]
	 */
	listifyArray=function($arr) {
		if ($arr.length==0) return "";
		if ($arr.length==1) return $arr[0];
		var str = $arr[0];
		for (var a=1; a < $arr.length-1; a++) {str = str+", "+$arr[a];}
		str = str+ " and "+$arr[($arr.length-1)];
		return str;
	}


	/**
	 * Just some helpful dumb msgs popped in here for certain relons.
	 * We can display these in the relons page to inform the user a bit more
	 * about what this relons is/was for.
	 */
	basicExplain = async($rObj)=> {
		var rel_sponsor_msg="";

		if ($rObj.relon_sponsor!=undefined && $rObj.relon_sponsor!=null && $rObj.relon_sponsor!="") {
			rel_sponsor_msg = "  This was originally a collab";
		}

		if ($rObj.from_nid==MEM.___SELF___ && $rObj.reln==MEM.BLOCK_WINNER.nid) {
			return "This is a block reward winner.";
		}
		if ($rObj.type==TYPE_BROADCAST_PAY && $rObj.reln==MEM.annies.nid && $rObj.to_type==TO_TYPE_ANNE) {
			return "This was a broadcast payment relon.";
		}
		if ($rObj.ghost==true && $rObj.to_type>=TO_TYPE_ENCR_ANNE) {
			var fromNOBJ = MEMG.getNeuronByTempNid($rObj.f);
			if (fromNOBJ.ghost==true) return "This to value will be encrypted at run-time when broadcast.";
		}
		if ($rObj.r==MEM.anj_accepted.temp_nid) return "This is an accepted anjucator invite.";
		if ($rObj.r==MEM.anj_rejected.temp_nid) return "This is an declined anjucator invite.";
		if ($rObj.reln==MEM.anjucator.nid) {
			if ($rObj.sys_ref_txid!=undefined && $rObj.sys_ref_txid.length) {
				if ($rObj.to_type==TO_TYPE_ANNE) {
					var srtRel=MEMG.lookupByTXID($rObj.sys_ref_txid);
					if (srtRel!=null) {
						if (srtRel.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to add an anjucator to a neuron.";
						else if (srtRel.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to add an anjucator to a neuron.";
					}
				}
				else if ($rObj.to_type==TO_TYPE_TXID) {
					if ($rObj.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to be added as an anjucator to a neuron.";
					else if ($rObj.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to be added as an anjucator to a neuron.";
				}
			}
			else return "This is a request to add an anjucator to a neuron.";
		}
		if ($rObj.reln==MEM.winner_payout.nid) return "This is a winning payout for the boosties bucket draw."
		if ($rObj.reln==MEM.hanai_in.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is approval of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is approval of a HANAI request.";
		}
		if ($rObj.reln==MEM.hanai_in_reject.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is rejection of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is rejection of a HANAI request.";
		}
        if ($rObj.r==MEM.andy_collab_accepted.temp_nid) return "This is an accepted collab request.";
        if ($rObj.r==MEM.andy_collab_rejected.temp_nid) return "This is a rejected collab request.";
        if ($rObj.r==MEM.hanai_out.temp_nid) {
            var whoNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			return "This is a request to HANAI out a neuron into the <b>"+whoNOBJ.bestNameHTML()+"</b> worldview.";
        }
		if ($rObj.r==MEM.anndle_offer.temp_nid) {
			var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			var pending = "is a pending";
			if (ANNE.UTILS.isset($rObj.sys_ref_txid)==true) pending = "was an ";
			return "This "+pending+" ANNDLE offer that paid "+UTILS.formatNumber($rObj.annies)+" annies ("+UTILS.formatNumber($rObj.annies/100)+" ANNE coin) to control the '"+toNOBJ.text+"' ANNDLE.";
		}

        if ($rObj.type==TYPE_SYS_MSG) {
			if ($rObj.r==MEM.annexed_by.temp_nid) {

				var arel = null;
				var ANNEXED_THING = null;

				if ($rObj.sys_ref_txid!=null) arel = ANNE.MEMG.lookupByTXID($rObj.sys_ref_txid);
				if (arel!=null) {
					ANNEXED_THING = MEMG.getNeuronByTempNid(arel.t);
				}

				if ($rObj.f==USER._userNObj.nid) {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Your neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "Your neuron was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
				else {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "A neuron from user "+MEMG.getOrMakeIfNew($rObj.to_s)+" was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
			}

            //if ($rObj.r==MEM.aon.temp_nid) return "This is an onboarding relon with "+$rObj.annies+" annies given to the recipient.";
            return "This is a relon sent as a <b>message</b> to an ANNODE for instructions to take action.  It can also be generated by the annode on behalf of a keyless neuron and serve as a status update or receipt.";
        }

        else if ($rObj.type==TYPE_SYS) return "This is a <b>system</b> relon used to engage with ANNE protocol rules.";
        else if ($rObj.type==TYPE_BE) {
            if ($rObj.r==MEM.class.temp_nid || $rObj.r==MEM.inst.temp_nid) {
                return "This is a relon that specifies be-hierarchical parent information."+rel_sponsor_msg;
            }
            else if ($rObj.r==MEM.name.temp_nid) return "This is a relon that sets the neuron's name."+rel_sponsor_msg;
            else if ($rObj.r==MEM.aka.temp_nid) return "This is a relon that sets an <b>also known as</b> for this neuron."+rel_sponsor_msg;
            else if ($rObj.r==MEM.fka.temp_nid) return "This is a relon that sets a <b>formerly known as</b> for this neuron."+rel_sponsor_msg;
        	return "This is a relon that specifies additional be-defining information."+rel_sponsor_msg; 
        }
        else if ($rObj.type==TYPE_PARAM) {
            return "This is a <b>parameter</b> relon defining a relationship that instances can provide."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_HAS) {
            if (this.bCOLLAB==true) {
                if (this.bISPENDING==true || ($rObj.status_text!=undefined && $rObj.status_text=="p")) {
					return "This is a collab relon request.";
                }
                return "This is a collab relon request.";
            }
            return "This is a relon which specifies a parameter value a neuron <b>has</b>."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_AWARENESS) {
            return "This is a relon for run-time feelz and opinionz of an identity using ANNE.";
        }
        else if ($rObj.type==TYPE_BRAIN) {
            return "This is a relon used for wiring neurons together in the non-semantic <b>brain</b> dimension."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_CLUMP) return "This is a relon connecting to a more clump neuron."+rel_sponsor_msg;
        else if ($rObj.type==TYPE_ROLE) {
            return "This is a semantic role relon that exists in a clump neuron and defines a semantic role within the clump."+rel_sponsor_msg;
        }
		else if ($rObj.type==TYPE_INTERNAL_PAY) {
			if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block reward won via solo mining by "+toNOBJ.text;
			}
			else if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_SHARE_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block share reward won via share mining by "+toNOBJ.text;
			}
			else if ($rObj.r==ANNE.MEM.boosties.temp_nid) {
				return "This added boosties to this neuron.";
			}
            else return "This is an internal payment from a keyless neuron made by L1 protocol rules.";
        }
		else if ($rObj.type>=TYPE_BRAIN0 && $rObj.type<TYPE_BRAIN120) {
			return "This is a special type to indicate sorting order #"+$rObj.type;
		}
		else if($rObj.type==TYPE_BRAIN120) {
			return "This is a special BRAIN120 relon created by the ANNODE to document backlinks and provide helpful summaries.";
		}
		else if ($rObj.type==TYPE_BRAIN121) {
			return "myelination";
		}
	}


	logFunc=function($funct, $str="", $str2="", $str3="") {
		if ($str3!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2+" %c"+$str3, TALK.funct, TALK.bggreen, TALK.bgblue, TALK.bgyellow);
		else if ($str2!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2, TALK.funct, TALK.bggreen, TALK.bgblue);
		else if ($str!="") console.log("%c("+$funct+") %c"+$str, TALK.funct, TALK.bggreen);
	}
        
       
        

}


const TALK = new anne_talk ();
TALK.set_init(TALK);




/************** (from file ./js/base/TALK.js)**/ 

class ANNE_NET {
    constructor($base_uri="http://localhost:9116") {
		this._nodeConnectedCallback = null;
		this._consecutive_network_errors = 0;
		this.bSSEINIT=false;
		this.last_msg_id = 0;
		this.TODOS = [];
		this.active_coms = false;
		//this.bDOINGROUTERPP=false;
		this.setBaseURI($base_uri);
    }


	/**
	 * This is a RAM-only relon that will only be in RAM if the neuron is in RAM
	 * and was pending a change in the mempool.
	 * 
	 * Since this app is "open" it will receive said pending change and thus we 
	 * can count on being able to remove the pending RAM-only relon here.
	 * @param {*} $nid 
	 */
	cleanupIntraPending = function($nid) {
		console.log("(cleanupIntraPending)"+$nid);

		if ($nid==undefined || $nid==null || $nid=="") return;
		if (parseInt($nid)==0) return;
		
		var nOBJ = MEMG.getOrMakeIfNew($nid);
		if (nOBJ._aplex!=null) nOBJ._aplex.has_pending=false;
		
		var intrablockpending_relon = MEMG.getRelonFromWith(nOBJ.outs, TYPE_SYS_MSG, ANNE.MEM.INTRA_BLOCK_PENDING.nid);
		if (intrablockpending_relon!=null) MEMP.localOnlyArborPrune_RAM(intrablockpending_relon);
		return nOBJ;
	}


	/**
	 * 
	 * @param {*} $uri 
	 */
	setBaseURI=function($uri="") {
		console.log("setting BaseURI:"+$uri);
		this.base_uri = $uri;
		if (this.base_uri!="", this.base_uri[this.base_uri.length-1]=="/") {
			this.base_uri = this.base_uri.substring(0, this.base_uri.length-1);
		}
	}


	/**
	 * 
	 * @param {*} $f 
	 */
	setNodeConnectedCallback=function($f) { 
		this._nodeConnectedCallback = $f; 
	}


	/** 
	 * 
	*/
	setNodeConnected=function($bNODECONNECTED=true) {
		if (this._nodeConnectedCallback!=null) {
			this._nodeConnectedCallback($bNODECONNECTED);
		}
	}



	//***************************************************** */
	//////  TRANACTION CREATION / SIGNING / BROADCAST ////////
	//***************************************************** */

	/**
	 * @param {String} $functName 
	 * @param {Object} $params 
	 * @param {String} $uriOverride - the node uri to use, if not the current node ur connected to. 
	 * @returns 
	 */
	makeOnNodeSignOnJSThenBroadcast=async($functName, $params=null, $uriOverride="")=> {
		var bCHASE_FROM=false;
		var bDO_EXTRAS=false;

		if ($params!=null) {
			if ($params.secretPhrase!=undefined) delete $params.secretPhrase;
			
			//chase the froms in the broadcastTransaction so we can get
			//crucial return data that we need in the processing of the next TXs in 
			//the queue.
			if ($params.chase_from!=undefined) {
				bCHASE_FROM = $params.chase_from;
				delete $params.chase_from;
			}
			if ($params.do_extras!=undefined) {
				bDO_EXTRAS = $params.do_extras;
				delete $params.do_extras;
			}

			//we need to make the TX first (using the annode)
			//then sign it here in the EXT,
			//then broadcast it.
			if ($params.makeSignSend!=undefined && $params.makeSignSend==true) {
				$params.publicKey = USER.pubKey;
				delete $params.makeSignSend;
			}

			//this wasn't set up right and it is just a normal query/send.
			else {
				return await this.query($functName, $params);
			}

			try {
			
				var response = await this.post($functName, $params, $uriOverride);
				console.log('%capi call for '+$functName, TALK.bggreenbig);
				
				this._consecutive_network_errors = 0;

				if (response && response.height!=undefined) {
					HISTORY._GREATEST_LAST_KNOWN_BLOCK = response.height;
				}

				if (response.status!=undefined && response.status!="good") {
					console.log(response);
					var retObj =  {status:"fail", reason:response.status};
					if (response.recourse!=undefined) retObj.recourse = response.recourse;
					if (response.result!=undefined) retObj.result = response.result;
					if (response.results!=undefined) retObj.results = response.results;					
					return retObj;
				}

				if (response.unsignedTransactionBytes==undefined || response.unsignedTransactionBytes==null) {
					return {status:"fail", reason:"problem with unsignedTransactionBytes"};
				}
				var finalResp = await this.signAndBroadcastTransaction(response, bCHASE_FROM, bDO_EXTRAS, $uriOverride);
				
				//don't do this from RELONS!!!
				//we want to remove relons that are beyond this point..
				//for rollback situations.
				if (finalResp.block_height!=undefined) {
					ANNE.HISTORY.updateGLKB_SOLID(finalResp.block_height);
				}
				ANNE.API.setNodeConnected(true);

 				return finalResp;
			}

			catch(e) {
				console.warn(e);
				if (e.data=="Network Error") ANNE.API.setNodeConnected(false);
				var reason = "problem with transaction";
				if (e.data!=undefined && e.data.errorCode!=undefined) {
					if(e.reason!=undefined && e.message==undefined) e.message = e.reason;
					e.data.status="bad";
					return e.data;
				}
				if (e.message!=undefined) reason = e.message;
				var retObj ={status:"bad", reason:reason};
				if (e.recourse!=undefined) retObj.recourse = e.recourse;
				return retObj;
			}
		}

		//this was just a simple query (without params)
		return await this.query($functName);
	}



	/**
	 * 
	 * @param {*} $resp 
	 * @param {*} $bCHASEFROM 
	 * @param {*} $bDOEXTRAS 
	 * @param {*} $uriOverride 
	 * @returns 
	 */
	signAndBroadcastTransaction=async($resp, $bCHASEFROM=false, $bDOEXTRAS=false, $uriOverride="")=> {
		this.active_coms = true;
		if (USER.privKey=="" || USER.privKey==undefined) {
			if (USER.seedObj==null) {
				var aO = await USER.getPrivateKeyHex();
			}
			else var aO = USER.prepANNEAuthObjFromSeed(USER.seedObj.seed);
			USER.privKey = aO.sign_key;
		}

		var sig=null;
		
		//gen the signature
		try{
			sig = ANNE.SIGNA.crypto.generateSignature($resp.unsignedTransactionBytes, USER.privKey);
		}
		catch(e) {
			this.active_coms = false;
			return {status:"fail", reason:"(generateSignature) problem"};
		}

		//verify the signature
		try{
			var rez = ANNE.SIGNA.crypto.verifySignature(sig, $resp.unsignedTransactionBytes, USER.pubKey);
			if (rez==false) {
				this.active_coms = false;
				return {status:"fail", reason:"(verifySignature) did not pass"};
			}
		}
		catch(e){
			console.log(e);
			this.active_coms = false;
			return {status:"fail", reason:"(verifySignature) problem"};
		}

		//generate signed bytes
		var signedTransactionBytes=null;
		try {
			signedTransactionBytes = ANNE.SIGNA.crypto.generateSignedTransactionBytes($resp.unsignedTransactionBytes, sig);
		}
		catch(e){
			console.log(e);
			this.active_coms = false;
			return {status:"fail", reason:"(generateSignedTransactionBytes) problem"};
		}
		
		//broadcast the transcation.
		var broadcastParams = {
			transactionBytes:signedTransactionBytes,
			transactionJSON:$resp.transactionJSON
		}
		if ($bCHASEFROM==true) {
			broadcastParams.chase_from = true;
		}
		else if ($bDOEXTRAS==true) {
			broadcastParams.do_extras = true;
		}

		try {
			var response2 = await this.broadcastSigned(broadcastParams, $uriOverride);
			if (response2.status!=undefined && response.status!="good") {
				console.log(response2);
			}
			this.active_coms = false;
			return response2;
		}	

		catch(e) {
			console.log('ERROR: api call for broadcastTransaction', e);
			if (e.data=="Network Error") {
				this._consecutive_network_errors+=1;
			}

			var rez = {status:"fail", msg:""};
			
			if (e.data!=undefined && e.data.message!=undefined) {
				rez.msg = e.data.message;
			}
			else {
				rez.msg = e.message;
			}
			if (e.data!=undefined) {
				if (e.data.error!=undefined) rez.error = e.data.error;
				if (e.data.errorCode!=undefined) rez.errorCode = e.data.errorCode;
				if (e.data.errorDescription!=undefined) rez.errorDescription = e.data.errorDescription;
				if (e.data.errorRecourse!=undefined) rez.errorRecourse = e.data.errorRecourse;
				if (e.data.result) rez.result=e.data.result;
			}
			
			this.active_coms = false;
			return rez;
		}
	}



	/**
	 * Last step of process to gen a TX on the annode, sign it on the edge device
	 * and broadcast transaction bytes back to the annode
	 * @param {Object} $params
	 * @param {String} $uriOverride  
	 * @returns 
	 */
	broadcastSigned = async($params, $uriOverride="")=>{
		ANNE.playSound("broadcast");
		return await this.post("broadcastTransaction", $params, $uriOverride);
	}



	/**
	 * Post data (used for TX broadcasts)
	 * @param {*} $functName 
	 * @param {*} $params
	 * @param {String} $uriOverride  
	 * @returns 
	 */
	post = async($functName, $params, $uriOverride="")=>{
		this.active_coms = true;

		var uri = this.base_uri;

		if (uri==undefined) {
			ANNE.TALK.printf_console(ANNE.TALK.RED, "(post) NET.base_uri not set, setting to defaults");
			uri = ANNE.setAnnodeByURI();
		}

		if ($uriOverride!="") {
			ANNE.TALK.printf_console(ANNE.TALK.BLUE, "(post) using URL override", ANNE.TALK.GREEN, $uriOverride);
			uri = $uriOverride;
		}

		var cs = ANNE.SIGNA.sig.ChainService;
		const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
		var response = null;
		if ($params!=null) response = await service.send($functName, $params);
		else response = await service.send($functName);
	

		
		this.active_coms = false;
		return response;
	}

	
	proxyRouter=async($uri, $functName, $params=null)=>{
		var uri = $uri+"?requestType="+$functName;
		if ($params != null) {
			const queryStringComponents = [];
			for (const key in $params) {
				if ($params.hasOwnProperty(key)) {
					const encodedKey = encodeURIComponent(key);
					const encodedValue = encodeURIComponent($params[key]);
					queryStringComponents.push(`${encodedKey}=${encodedValue}`);
				}
			}
			uri += '&' + queryStringComponents.join('&');
		}

		console.log(uri); 

		let resp =  await fetch(uri).catch(function($e){
			console.log("error:"+$e);
			return null;
		});

		if (!resp) {
            console.warn("HTTP error! -- no response");
			return null;
        }

        else if (!resp.ok) {
            console.warn(`HTTP error! status: ${resp}`);
			return null;
        }

        let data = await resp.json();
        console.log(data); 

		return data;
	}


	apiHTTPNOJSON=async($functName, $params)=> {
		this.active_coms = true;
		
		var uri = this.base_uri;
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}

		
		var cs = ANNE.SIGNA.sig.ChainService;
		const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
		if ($params!=null) {
			var rez = await service.send($functName, $params);
			this.active_coms = false;
			return rez;
		}
		else {
			var rez = await service.send($functName);
			this.active_coms = false;
			return rez;
		}
		

	}


    /**
     * Perhaps we lack here too from oversimplification of the signa method...
	 * may have to revert back to using the chainService here too.
	 * 
     * @param {String} $functName 
     * @param {Object} $params 
	 * @param {String} $uriOverride
     * @returns 
     */
    query = async($functName, $params=null, $uriOverride="")=>{
		//if ($functName!="getLatestMsgs") {
			if ($params!=null) console.log("%cAPI.query."+$functName, ANNE.TALK.bgpurple, $params);
			else console.log("%cAPI.query."+$functName, ANNE.TALK.bgpurple);
		//}
		//if ($params!=null && $params.height!=undefined) console.log("for height:"+$params.height);
		
		this.active_coms = true;

		var uri = this.base_uri;
		
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}
		if ($uriOverride!="" && $uriOverride!=null) uri = $uriOverride;
		
	
		var cs = ANNE.SIGNA.sig.ChainService;
		if (cs==undefined) console.throw("Must have ANNE.SIGNA.sig.Chainservice");
		
		try {
			const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
			//console.log('api call for '+$functName);

			if ($params!=null) var response = await service.send($functName, $params);
			else var response = await service.send($functName);
		}
			
		catch(e) {
			ANNE.TALK.printf(TALK.RED, " connection to annode error (no annode) ");
			this.active_coms = false;
			if ($functName=="annodeDetails") {
				if (ANNE && ANNE.badNode!=undefined) ANNE.badNode();
			}
			return this.makeCatchReturn(e);
		}
		

		//get height if we passed it.. helpful for other processes.
		if (response!=null && response.result!=undefined && response.result.block_height!=undefined) {
			ANNE.HISTORY.updateGLKB_SOLID(response.result.block_height);
		}

		if ($functName=="getNeuronBalance" && response.status!=undefined && response.status!="bad") {
			await this.__balanceHack(response, $params);
		}
		this.active_coms = false;
		return response;
    }


	/**
	 * 
	 * @param {*} $e 
	 * @returns 
	 */
	makeCatchReturn = function($e) {
		if ($e==null || $e==undefined) return {status:"fail", reasons:"totally hosed everything"};

		console.log($e);
		var reason = "problem with net query";
		if ($e!=null && $e.status==405) {
			ANNE.API.setNodeConnected(false);
			reason = "no connected annoded.";
		}
		
		else if ($e.data!=undefined && $e.data!=null && $e.data=="Network Error") {
			ANNE.API.setNodeConnected(false); //are we sure we wnat to do this here????
		}

		else if ($e.data!=undefined && $e.data!=null && $e.data.errorCode!=undefined && $e.data.errorCode=="777.2.0") {
			var retObj = {status:"good-but", message:"taken"};
			if ($e.data.result!=undefined) retObj.result = $e.data.result;
			if ($e.data.results!=undefined) retObj.results = $e.data.results;
			return retObj;
		}
		
		return {status:"fail", reason:reason};
	}


	/**
	 * TODO: review this.
	 * @param {*} $response 
	 * @param {*} $params 
	 * @returns 
	 */
	__balanceHack=async($response, $params)=> {
		if ($response==undefined) { return;}
		
		//this will be active sooooon.
		if ($response.activated_in_mempool!=undefined) {
            if (this._userNObj!=null) this._userNObj.activated_in_mempool=true;
        }

		if ($response.activated=="y") {
			if ($params.nid==USER._userNObj.nid && ANNE.USER.isActivated()==false) {
				await ANNE.USER.updateLSActivated(true); 
			}
		}

		//somehow we have this as activated (on this network) but it isn't.
		//correct that.
		else if ($response.activated=="n" && ANNE.USER.isActivated()==true) {
			if ($params.nid==USER._userNObj.nid) {
				await ANNE.USER.updateLSActivated(false); 
			}
		}
		return $response;
	}


	fetchJSON = async (uri ="https://www.biconomy.com/api/v1/depth?symbol=ANNE_USDT") => {
		try {
			const response = await fetch(uri);
			if (!response.ok) {
				throw new Error(`HTTP error! Status: ${response.status}`);
			}
			const data = await response.json();
			return data;
		}
		catch (error) {
			console.error("Error fetching JSON:", error);
			throw error; 
		}
	};
	
}

/************** (from file ./js/network/net.js)**/ 

class ANNODE_API extends ANNE_NET {
    constructor($base_uri){	//expects url:port like http://localhost:9116
		super($base_uri);	//converthing ANNODE_API to use ANNE_NET (standalone networking class)
		this.client = null;
		this.db = null;
		this.anon_user_id = null;	//for use when AUTHing anon externally.
		this._AAUTH = null;
		this._bNODECONNECTED=false;
		this._consecutive_network_errors = 0;
		this.privateFolderAbsolute = "";
		this._annodeDetailsObj = null;
		this.OPT_CHECKER = [];

	}

	opwipe = function() {
		this.OPT_CHECKER=[];
	}
    opcheckerPass = function($funct, $secs=10) {
		if (this.OPT_CHECKER[$funct]!=undefined) {
			var lasttime = this.OPT_CHECKER[$funct];
			let currentTime = Math.floor(Date.now() / 1000);
			var diff = currentTime - lasttime;
			if (diff < $secs) {
				console.log("skip, "+diff+" "+$funct+" is too soon. Find out who is calling this.");
				return false; 
			}
		}
		return true;
	}

	//tracking the seconds marker of last uses of functions.
	//avoiding hitting them too frequently for no good reason.
	opcheckerTrackIt = function($funct) {
		this.OPT_CHECKER[$funct] = Math.floor(Date.now() / 1000);
	}

	getAnnodeProp=function($prop) {
		var val = this._annodeDetailsObj[$prop];
		if (val==undefined) return null;
		return val;
	}

	
	/**
	 * 
	 * @param {*} $start_index 
	 * @param {*} $end_index 
	 * @param {*} $bDOTRANSACTIONS 
	 * @returns 
	 */
	getLastBlocks=async($start_index, $end_index, $bDOTRANSACTIONS)=> {
		var rez = await this.query("getLastBlocks", {
			from_height:$start_index,
			to_height:$end_index,
			includeTransactions:$bDOTRANSACTIONS
		});

		return rez;
	}

	getBlockByHeight = async($height)=>{
		var rez = await this.query("getBlockByHeightOrID", {
			height:$height
		});
		return rez;
	}

	getBlockById = async($blockId)=>{
		var rez = await this.query("getBlockByHeightOrID", {
			blockId:$blockId
		});
		return rez;
	}



	/**
	 * 
	 * @param {*} $prop 
	 * @returns 
	 */
	getAnnodeDetail=async($prop="circulating_supply", $bREFRESHSPECIAL=true)=> {
		if (this._annodeDetailsObj==null) {
			await this.annodeDetails();
			if (this._annodeDetailsObj==null) return;
		}
		var val = this._annodeDetailsObj[$prop];
		if (val==undefined) return null;

		if ($prop=="annode_mode" && val=="SYNC" && $bREFRESHSPECIAL==true) {
			this._annodeDetailsObj = null;
			return await this.getAnnodeDetail($prop, false);
		}
		return val;
	}


	/**
	 * Used when getting new annode details.
	 * We potentially have new _active vars.
	 * @param {*} $obj 
	 */
	processAuxAAUTHData=async($obj)=>{
		var bCHANGE = false;
		this._annodeDetailsObj = $obj;
		//console.log(this._annodeDetailsObj);

		if ($obj==null) return;
		if ($obj.height!=undefined) {
			if (ANNE.HISTORY!=null) ANNE.HISTORY.updateGLKB_SOLID($obj.height);
		}
		
        if (ANNE.USER==null || ANNE.USER._userLS==null || ANNE.USER._userLS.aauth==null) return;
		
		if (USER._userLS.aauth.ANNDLE_active!=$obj.anndle_active) {
            ANNE.USER._userLS.aauth.ANNDLE_active = $obj.anndle_active;
			ANNE.API._AAUTH.ANNDLE_active = $obj.anndle_active;
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.ANNDLE_active = $obj.anndle_active;
            bCHANGE=true;
        }
        if (ANNE.API._AAUTH.ASDF_active!=$obj.asdf_active) {
            ANNE.USER._userLS.aauth.ASDF_active = $obj.asdf_active;
			ANNE.API._AAUTH.ASDF_active = $obj.asdf_active;
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.ASDF_active = $obj.asdf_active;
            bCHANGE=true;
        }
        if (ANNE.API._AAUTH.PDNS_active!=$obj.pdn_active) {
            ANNE.USER._userLS.aauth.PDNS_active = $obj.pdn_active;
			ANNE.API._AAUTH.PDNS_active = $obj.pdn_active;
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.PDNS_active = $obj.pdn_active;
            bCHANGE=true;
        }
        if (ANNE.API._AAUTH.BOUNTY_active!=$obj.bounty_active) {
            ANNE.USER._userLS.aauth.BOUNTY_active = $obj.bounty_active;
			ANNE.API._AAUTH.BOUNTY_active = $obj.bounty_active; 
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.BOUNTY_active = $obj.bounty_active;
            bCHANGE=true;
        }
        if (bCHANGE==true) {	//will save updates to _AAUTH.
			if (USER.seedObj!=null) {
				ANNE.API._AAUTH = ANNE.USER._userLS.aauth;
				await ANNE.USER.__saveANNE_JO(ANNE.USER._userLS);
			}
		}
	}


	setAAUTHFromLS = function($resp) {
		this._AAUTH = new annode_auth($resp);
		this._bNODECONNECTED=true;
		USER.aauth = this._AAUTH;
		
	}

	getNextProofId = function() {
		if (this._AAUTH==null) return -1;
		return this._AAUTH.next_proof_id;
	}

	updateAndSaveNextProofId = async($id)=> {
		if (this._AAUTH==null) return -1;
		this._AAUTH.next_proof_id = $id;
		USER._userLS.aauth = this._AAUTH;
		console.log("%c (updateAndSaveNextProofId) id:"+$id, TALK.bgbigred);
		return await ANNE.USER.updateNextProofId($id);
	}

	getRelonLimit = function() {
		if (this._AAUTH == null) return 0;
		return this._AAUTH.drl();
	}


	/**
	 * 
	 * @param {*} $functName 
	 * @param {*} $params 
	 * @returns 
	 */
	apiSimpleQueryWithParams=async($functName, $params, $bDOLOADER=true)=> {
		var response = await this.query($functName, $params);
		return response;
	}
	apiSimpleQueryWithParamsNS=async($functName, $params)=> {
		var response = await this.query($functName, $params);
		return response;
	}
	apiSimpleQuery=async($functName)=> {
		var response = await this.query($functName, null);
		return response;
	}


	//***************************************************** */
	//////  TRANACTION CREATION / SIGNING / BROADCAST ////////
	//***************************************************** */



	//***************************************************** */
	////////////////  TRANSACT WITH AN ANNDOE ////////////////
	//***************************************************** */
	

	/**
	 * Add boosties to de-centivize annexes against your sponsored neuron!
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	addBoosties = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=>{
		 console.log("(adddBoosties) sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}
		else if ($relon.sys_ref_txid!=undefined) {
			alert("THIS SHOULDNT BE NEEDED HERE");
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}
		
		var json = await this.makeOnNodeSignOnJSThenBroadcast("addBoosties",params);
		return json;
	}
	

	/**
	 * Send a relon where the FROM is the sender (you).
	 * 
	 * If the .sys_ref_txid is set in the sending relon the API will lookup 
	 * the full hash of the transaction.  When broadcast as part of the TX it serves
	 * as a dependency to mine first.
	 * 
	 * In our ANNE GRAPH we also use this variable after-the-fact to link TXs to other
	 * TXs they are in reference (or response) to.
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	sendSelfRelon = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;

		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};

		if ($relon.r==ANNE.MEM.annex.temp_nid) {
			params.chase_from=true;
		}

		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		//with annexes this is already stored and set as part of the annex chain.
		//just pass it through here.
		if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("sendSelfRelon",params);
		console.log(json);
		return json;
	}

	
	/**
	 * Call for a distribution vote for an annorg (technically this will work on ANY keyless neuron)
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	anorgVoteCall = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;

		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		
		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
			else if ($relon.r==MEM.distro_vote.nid) {
				alert("error, couldn't get ref sys txid's hash for DISTRO VOTE (must tie to FREEZE)");
				return;
			}
		}
		
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anorgVoteCall",params);
		return json;
	}
	

	/**
	 * Claim a bounty.
	 * This will invoke an ACHAT session on the annode it is broadcast through.
	 * The bounty will specify the annode that it wants to ACHAT on.
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	bountyClaim = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeBountyClaim",params);
		return json;
	}

	/**
	 * Accept some risk being offered by an ASDF.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	asdfAcceptRisk = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("ASDFacceptRisk",params);
		return json;
	}


	/**
	 * Add more risk to an ASDF you own.
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	asdfAddRisk = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("ASDFaddRisk",params);
		return json;
	}


	/**
	 * A way to stop accepting risk takers for an ASDF.
	 * AKin to no-more-bets or the real-world event has started.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	asdfNomoreRisk = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("ASDFnoMoreRisk",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	anndleSellback = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleSellback",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	anndleRenew = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleRenew",params);
		return json;
	}
	

	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	annexNeuron = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("annexNeuron",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makeAnndleOffer = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleBuysideOffer",params);
		return json;
	}



	
	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	acceptAnndleOffer = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_froms:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleOfferAccept",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	revokeAnndleOffer = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleOfferRevoke",params);
		return json;
	}

	/**
	 * Make an ANNE admin for a specific class you have access to grant rights to.
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	makeAnneAdmin = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeAnneAdmin",params);
		return json;
	}


	/**
	 * Add to the data of an ANNDLE you have sponsored.
	 * Only class Params are relns that will be accepted by the ANNODE.
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	makeAnndleEdit = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeAnndleEdit",params);
		return json;
	}
	

	/**
	 * Make a subanndle under an anndle that you control (are the sponsor of).
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	makeSubAnndle = async($relon,  $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeSubAnndle",params);
		return json;
	}


	/**
	 * Accept some amount of an ANNE OTC offer.
	 * This is specific to the SIGNA integration since we can 
	 * broadcast the SIGNA transaction from the ANNODE to the SIGNA node.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	acceptAOTCOffer = async($relon,  $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		if (ANNE.UTILS.isset($relon.signedTransactionBytes)==false) {
			alert("we must have the signed transaction bytes from the AOE CHAIN TX!")
			return;
		}
		else params.signedAOETransactionBytes = $relon.signedTransactionBytes;

		var json = await this.makeOnNodeSignOnJSThenBroadcast("AOTCacceptOfferSIGNA",params);
		return json;
	}

	

	/**
	 * Response to a collab request.
	 * 
	 * NOTE: This must be a TO_TYPE_TXID with the collab request as the to
	 * 
	 * It also must reference the full hash of the collab request (even if only in blockchain and not in hypergraph).
	 * This ensures payment for admin is available prior to the admin's response being mined.
     * 
	 * The only time amountNQT is relevant is when approving a merge that has a required annex to settle it.
     * No other collab response patterns require annies to respond.  (in fact, admins gain annies for responding)
	 */
	 makeCollabResp = async($relon, $annies=0, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		
		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeCollabResp", params);

		return json;
	}

	/**
	 * Respond to the invitation to be an Anjucator.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makeAnjInviteResp = async($relon, $annies=0, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			deadline:1440
		};
		
		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeAnjInviteResp", params);
		return json;
	}


	/**
	 * Make a keyless class
	 * 
	 * Keyless class makers are the owner of the neuron class.
	 * 
	 * The RELN should be the standard CLASS or INST to denote the nType (neuron type).
	 * The TO should be the parent class.  This is a standard TYPE_BE parent relon,
	 * We send it through as a TYPE_SYS_MSG but on the annode a TYPE BE will be created.
	 * 
	 * @param {Object}
	 * @param {string}
	 * @param {string}
	 * @param {integer}
	 * @author Mr.Scatman
	 */
	 makeKeylessNeuron = async($relon,  $annies=900, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeKeylessNeuron",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		//console.log(json);
		return json;
	}


	/**
	 * Make a new Private Data Neuron.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makePDN = async($relon,  $annies=2000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makePDN",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}


	/**
	 * Make a custodial Neuron that this particular ANNODE will manage.
	 * 
	 * DO EXTRAS FLAG HERE IS CRUCIAL!!!!!
	 * WE MUST RETURN THE CUSTODIAL NEW NID AND KEY AFTER SIGNING AND BROADCAST!!!
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makeCustodialNeuron = async($relon,  $annies=800, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeCustodialNeuron",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			do_extras:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}


	/**
	 * We default to 1 coin in the custodial neuron.
	 * THere are NO fees so these remain in the faucet as an initial bump.
	 * @param {*} $relon 
	 * @param {*} $recipientPublicKey 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	announceCustodial = async($relon, $recipientPublicKey, $referencedTransactionFullHash, $recipient, $annies=100, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if ($recipientPublicKey=="") {
			console.log("MUST HAVE A RECIPIENT PUBLIC KEY!");
			return false;
		}
		if ($recipient=="") {
			console.log("WE MUST HAVE RECIPIENT HERE!");
			return false;
		}
		if (MEM.__isMalformedRelon($relon)==true) return false;


		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440,
			recipientPublicKey:$recipientPublicKey,
			recipient:$recipient
		}

		if ($referencedTransactionFullHash!="") params.referencedTransactionFullHash = $referencedTransactionFullHash;
		else if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}
		else if ($relon.sys_ref_txid!=undefined) {
			alert("THIS SHOULDNT BE NEEDED HERE");
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("announceCustodialNeuron", params);

		console.log(json);
		return json;
	}

	

	/**
	 * Make (sponsor) a new global neuron.
	 * 
	 * @param {Object}
	 * @param {string}
	 * @param {string}
	 * @param {integer}
	 * @author Mr.Scatman
	 */
	 makeNewGlobal = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeGlobal",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}

	/**
	 * Make (sponsor) a new global neuron. (location)
	 * 
	 * @param {Object}
	 * @param {string}
	 * @param {string}
	 * @param {integer}
	 * @author Mr.Scatman
	 */
	makeGlobalLocation = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeGlobalLocation",{
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}	


	/**
	 * 
	 * @param {Object} $relon 
	 * @param {int} $amount   (2,147,483,647 billion annies)  going to likely have to do a sys overhaul of all amounts and use bigint.
	 * @param {int} $fee 
	 * @returns 
	 */
	 sendChildRelon = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES)=> {
    	if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			console.log($relon);
			return {status:"fail", reason:"missing relon info"};
		}
		//console.log("(sendChildRelon) sending");
		ANNE.TALK.printRelon($relon);

		var fhi=0;
		if ($relon.flathexint!=undefined) fhi = $relon.flathexint;

		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			flathexint:fhi,
			deadline:1440
		};

		//we want the relons back on a freeze.
		if ($relon.nid==MEM.CMD_FREEZE.nid) {
			params.chase_from=true;
		}

		if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}

		else {
			//if future devs use this API this will ensure any NEW uses of custodials will
			//not throw any errors on the annodes.
			var fromNOBJ = MEMG.getNeuronByNid($relon.from_nid);
			if (fromNOBJ!=null && ANNE.UTILS.isset(fromNOBJ.origin)==true && fromNOBJ.origin==ORIGIN_CUSTODIAL) {
				//we need to make sure all child relon TXs of custodials refer back to their
				//TXTYPE_ANNOUNCE   reln:MAKE_CUSTODIAL TO_TYPE_TXID transaction.

				//we can find our make custodial request as an in of the from (the cusodial)
				//pubkey was broadcast on MAKE_CUSTODIAL tx.. we can get that txid
				//and lookup the pubky if we don't have it in local mem.

				var rel = null;
				var the_txid_we_want_the_fullhash_of = "";
				var fullHash = "";

				for (var o=0; o < fromNOBJ.outs.length; o++) {
					if (fromNOBJ.outs[o].r==MEM.class.temp_nid || fromNOBJ.outs[o].r==MEM.inst.temp_nid) {
						the_txid_we_want_the_fullhash_of = fromNOBJ.outs[o].sys_ref_txid;
						rel  = MEMG.lookupByTXID(fromNOBJ.outs[o].sys_ref_txid);
						if (rel!=null && rel.fullHash!=undefined) fullHash = rel.fullHash;
						break;
					}
				}

				if (fullHash.length==0) {

					//we need to get the txid.. otherwise we can't continue.
					if (the_txid_we_want_the_fullhash_of=="") {
						await CHASE.lookupMinNid(fromNOBJ.nid);
						for (var o=0; o < fromNOBJ.outs.length; o++) {
							if (fromNOBJ.outs[o].r==MEM.class.temp_nid || fromNOBJ.outs[o].r==MEM.inst.temp_nid) {
								the_txid_we_want_the_fullhash_of = fromNOBJ.outs[o].sys_ref_txid;
								break;
							}
						}
					}

					//if we don't have the txid, we can't continue.
					if (the_txid_we_want_the_fullhash_of=="") {
						console.log("problem with finding custoidal reference.");
						return;
					}


					//go get the hash.  We must reference it.
					if (the_txid_we_want_the_fullhash_of.length>0) {	//this is a BE relon.. just do min chase
						var json = await this.getBlockchainTransaction(the_txid_we_want_the_fullhash_of);
						console.log(json);
						if (json!=null && json.fullHash!=undefined) {
							params.referencedTransactionFullHash = json.fullHash;
						}
						else if ($relon.r==MEM.action_delegate.temp_nid) {
							alert("error, couldn't get ref sys txid's hash for delegation request");
							return;
						}
					}
				}

				else params.referencedTransactionFullHash = fullHash;
			}
		}


		var json = await this.makeOnNodeSignOnJSThenBroadcast("sendChildRelon", params);
		return json;
	}

	
	
	/**
	 * Send a viewPDN relon transaction.
	 * This is the same pattern as child relon and it 
	 * will use the child relon TX in the ANNODE.
	 * 
	 * @param {Object}
	 * @param {string}
	 * @author Mr.Scatman
	 */
	viewPDN = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending viewPDN relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		console.log("(viewPDN) sending");
		ANNE.TALK.printRelon($relon);

		var json = await this.makeOnNodeSignOnJSThenBroadcast("viewPDN", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			deadline:1440
		});
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $amount 
	 * @param {*} $fee 
	 * @param {*} $bEXPECTED_DELEGATION 
	 * @author Mr.Scatman
	 */
	makeCollabReq = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES, $bEXPECTED_DELEGATION=false)=> {
    	console.log("sending child relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		ANNE.TALK.printRelon($relon);

		var fhi=0;
		if ($relon.flathexint!=undefined) fhi = $relon.flathexint;

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeCollabReq", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			deadline:1440,
			flathexint:fhi,
			expect_delegation:$bEXPECTED_DELEGATION
		});
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $amount 
	 * @param {*} $fee 
	 * @param {*} $bEXPECTED_DELEGATION 
	 * @author Mr.Scatman
	 */
	makeHanaiCollabReq = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES, $bEXPECTED_DELEGATION=false)=> {
    	console.log("sending child relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		ANNE.TALK.printRelon($relon);
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeHanaiCollabReq", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true, 
			feeNQT:$fee,
			amountNQT:$amount,
			deadline:1440
		});
		return json;
	}	


	/**
	 */
	makeHanaiOutReq = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending hanai out relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		ANNE.TALK.printRelon($relon);
		
		var fhi=0;
		if ($relon.flathexint!=undefined) fhi = $relon.flathexint;

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeHanaiOutReq", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			flathexint:fhi,
			deadline:1440
		});
		return json;
	}	



	
	/**
	 * Send a standard payment TX to another Account.
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @author Mr.Scatman
	 */
	 sendAnne = async($recipient, $recipientPublicKey, $amount, $fee=ANNE.ANNE_TO_ANNIES, $deadline=1440, $msg="", $encrypted=false, $edata=null, $ndata=null)=> {
		console.log("(sendAnne) recipient:"+$recipient+" amount:"+$amount+" fee:"+$fee);
		var paramsObj = {
			recipient:$recipient, 
			amountNQT:$amount, 
			feeNQT:$fee,
			deadline:$deadline,
			makeSignSend:true,
		} 

		//onboarding a pubKey with send..
		//this should be auto with the sendAnne functionality and the PublicKeyAnnouncement appendix
		//also required when sending encrypted message to recipient (even when their key is known)
		if (ANNE.UTILS.isset($recipientPublicKey)==true) paramsObj.recipientPublicKey=$recipientPublicKey;
		else if ($encrypted==true) return {status:"bad", reason:"must have recipientPublicKey to encrypt"};
		 
		if ($msg!="") {
			console.log("%c (sendAnne) with plaintext msg", TALK.bgpink);
			paramsObj.message = $msg;
			paramsObj.messageIsText = true;
		}
		else {
			if ($encrypted==true) {
				if (ANNE.UTILS.isset($edata)==false || ANNE.UTILS.isset($ndata)==false) {
					return {status:"bad", reason:"user wants to send encrypted data but the provided encrypted data is missing."};
				}
				console.log("%c (sendAnne) with encrypted msg", TALK.bgpink);
				paramsObj.encryptedMessageData = $edata;
		        paramsObj.encryptedMessageNonce = $ndata;
			}
		}
		var rez = await this.makeOnNodeSignOnJSThenBroadcast("sendAnne", paramsObj); 
		if (rez.broadcasted==undefined) rez.status = "bad";
		else rez.status = "good";
		return rez;
	}


	//***************************************************** */
	///////////////////  PING & HANDSHAKE ////////////////////
	//***************************************************** */

	/**
	 * Initialize the ANNODE auth handshake.
	 * 
	 * If already "logged" in, just ping node w/ ANNODEAuthHandshake and set the js _AAUTH object.
	 * This will be required to get the nodeId for certain types of transactions that payout the annode that processes the TX.
	*/
	ANNODEPING=async()=> {

		var uri = ANNE._annodesLS.last_uri;
		if (ANNE._annodesLS.last_uri[ ANNE._annodesLS.last_uri.length-1]=="/") {
			uri = ANNE._annodesLS.last_uri.substring(0, ANNE._annodesLS.last_uri.length-1);
		}

		var response = await ANNE.API.apiSimpleQueryWithParams("ANNODEAuthHandshake", {
			annode_claimed_uri:uri} 
			);//"https://www.anne.network:9116"
		if (response.status=="good") {
			var AAUTHOBJ = new annode_auth(response.result);
			this._AAUTH = AAUTHOBJ;
		}
		else {
			console.warn("we have a problem houston");
			return;
		}
	}


	/**
	 * Handshake with an annode.
	 * 
	 * Set up the active aauth object.
	 * This does not persist session to session.
	 * If another handshake is made w/ the same or different node
	 * the API uses it.
	 * 
	 * @param {*} $aO 
	 * @param {*} $uniqueReqRandId 
	 * @param {*} $height 
	 * @returns 
	 */
	ANNODEHANDSHAKE=async($aO, $uniqueReqRandId, $height=0) => {
		//handshake with annode
		var uri = ANNE._annodesLS.last_uri;
		if (ANNE._annodesLS.last_uri[ ANNE._annodesLS.last_uri.length-1]=="/") {
			uri = ANNE._annodesLS.last_uri.substring(0, ANNE._annodesLS.last_uri.length-1);
		}
		var response = await ANNE.API.apiSimpleQueryWithParams("ANNODEAuthHandshake", {
			annode_claimed_uri:uri}); //"https://www.anne.network:9116"


		if (response.status=="good") {
			var AAUTHOBJ = new annode_auth(response.result);
			this._AAUTH = AAUTHOBJ;
			if (USER._userNObj!=null) USER._userNObj.aauth = this._AAUTH;
		}
		else {
			//roll back to last successful node.
			console.warn("we have a problem houston");
			return;
		}

		//returned from the NODE.
		var annode_time = this._AAUTH.time;
		var annode_pubKey = this._AAUTH.pubKey;

		//We will encrypt a string of the user's public nid and the annode time
		//The handling node will verify this time is within a reasonable ping range 
		//of milliseconds.
		var encryptThis = USER.public_nid+"::"+annode_time;

        var eObj = await ANNE.USER.encryptMsgForPUBKEY(encryptThis, annode_pubKey, $aO.priv_key, true);
        //console.log(eObj);

        
		//we send info back to the server to close the ANNODEAUTH
        var result = await ANNE.API.apiSimpleQueryWithParams("ANNODEAuthFinish", {
												user_nid:USER.public_nid, 
												user_pubKey:$aO.pub_key, 
												annode_time:annode_time, 
												user_sig_data:eObj.data_string,
												user_sig_nonce:eObj.nonce_string,
												height:$height,
												uid:$uniqueReqRandId}
												);

		//console.log(result);
		result.original_uid = $uniqueReqRandId;
		if (result.lnots!=undefined) {
			if (result.lnots.length) {
				console.log("%c lnots present", TALK.bgbigred);
				console.log(result.lnots);
			}
		}
		return result;
	}
	
	//***************************************************** */
	//////////////////  QUERY THE l1 ANNODE //////////////////
	//***************************************************** */
	


	/**
	 * Match condition and do join to chase values to prevent further lookups
	 * 
	 * normal space relons contain awareness relons so they must be filtered
	 * for type to avoid clutter.				
	 * @param {string}
	 * @param {array}
	 * @param {array} 
	 * @param {boolean}
	 * @param {array}  
	 * @param {array}   
	 * 
	 * @author Mr.Scatman								
	 */
	chaseInsGWV = async($nid, $adopt, $bMUSTPULL, $types=[1,3], $chasetypes=[0,1,2])=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.chaseInsGWV ", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {	//means it is a one-time query per session.
			console.log("super optimizer");
			return [];
		}

		var response = await this.apiSimpleQueryWithParams("chaseInsGWV", 
				{
					nid:$nid, 
					adopt:$adopt,
					types:$types,
					chase_types:$chasetypes,
					height:lastBlock

				});
		console.log(response);
		phObj.bDONETHISSESSION=true;
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.chaseInsGWV");
		}
		return aRelons;
	}
	
	
	/**
	 * Match condition and do join to chase values to prevent further lookups
	 * 
	 * normal space relons contain awareness relons so they must be filtere
	 * for type to avoid clutter.				
	 * @param {string}
	 * @param {string}
	 * @param {array}
	 * @param {array} 
	 * @param {boolean}
	 * @param {array}  
	 * @param {array}   
	 * @author Mr.Scatman								
	 */
	chaseInsWithRelnGWV = async($nid, $reln, $adopt, $types=[1,3], $chasetypes=[0,1,2], $parentNOBJ)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.chaseInsWithRelnGWV ", $nid);
		lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("chaseInsWithRelnGWV", 
			{
				nid:$nid, 
				reln:$reln,
				adopt:$adopt,
				types:$types,
				chase_types:$chasetypes,
				be:$parentNOBJ.nid,
				height:lastBlock

			});

		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.chaseInsWithRelnGWV");
		}
		return aRelons;
	}
	
		
	/**
	 * Get account by accountId (ANNE neurons in the memory handler call this nid).
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	getAccount = async($accountId)=> {
		var json = await this.apiSimpleQueryWithParams("getAccount",{account:$accountId});
		if (json.status!=undefined && json.status=="fail") return null;
		return json;
	}


	/**
	 * 
	 * @param {*} $account 
	 * @param {*} $ts 
	 * @param {*} $firstIndex 
	 * @param {*} $lastIndex 
	 * @returns 
	 */
	getAccountBlocks = async($account, $firstIndex=0, $lastIndex=100000000)=>{
		var response = await this.apiSimpleQueryWithParams("getAccountBlocks", 
		{	account:$account,
			timestamp:0,
			firstIndex:$firstIndex,
			lastIndex:$lastIndex
		});
		if (response.blocks!=undefined) {
			console.log(response.blocks);
			return response.blocks;
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $adtype 
	 * @param {*} $adsubtype 
	 * @param {*} $adpayloadjsonstr 
	 * @param {*} $adpayload_sig 
	 * @returns 
	 */
	sendAltData = async($adtype, $adsubtype, $adpayloadjsonstr="", $adpayload_sig="")=>{
		var response = await this.apiSimpleQueryWithParams("sendAltData", 
		{	adtype:$adtype,
			adsubtype:$adsubtype,
			adpayload:$adpayloadjsonstr,
			adpayloadsig:$adpayload_sig
		});
		return response;
	}



	
	/**
	 * 
	 * @param {*} $alt_id_reln 
	 * @param {*} $alt_id_value 
	 * @param {*} $wvs 
	 * @returns 
	 */
	queryForAltID=async($alt_id_reln, $alt_id_value, $wvs=null)=>{
		if ($alt_id_reln==undefined || $alt_id_reln==null) return;
		if ($alt_id_value==undefined || $alt_id_value==null) return;
		
		if ($wvs==null) {
			$wvs = [ANNE.MEM.ANNE_ACCTID];
			if (ANNE.USER!=null && ANNE.USER._userNObj!=null) {
				if ($wvs.indexOf(ANNE.USER._userNObj.nid)==-1) $wvs.push(ANNE.USER._userNObj.nid);
			}
		}
		var response = await this.apiSimpleQueryWithParams("queryForAltID", {
			alt_id_reln:$alt_id_reln,
			alt_id_value:$alt_id_value,
			wvs:$wvs
		});
		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForAltID");
				return aRelons;
			}
		}
		return [];
	}

	
	/**
	 * firstIndex is the literal list index of the TXs for that account.
	 * 1st TX is 0, 2nd is 1, etc.
	 * 
	 * TO DO: we can store these to opimize.. OR just accept the last 500 and 
	 * do a more robust transaction area later.
	 * 
	 * @param {string} - account we want to query TXs for.
	 * 
	 * @author Mr.Scatman								
	 */
	getAccountTransactions = async($account, $firstIndex=0)=> {

		//todo first the first index.

		var response = await this.apiSimpleQueryWithParams("getAccountTransactions", 
			{	account:$account,
				timestamp:0,
				type:-1,
				subtype:-1,
				firstIndex:0,
				lastIndex:10000000,
				numberOfConfirmations:0,
				includeIndirect:true
			});
		console.log(response);
		if (response.transactions!=undefined) {
			console.log(response.transactions);
			console.log(response.transactions.length);//153
			return response.transactions;
		}
		return [];
	}


	/**
	 * Get all preloaded relons from the ANNODE.
     * 
     * This is a one-time preload done to establish a base in the browser's indexeddb.
	 * This is NOW (updated) stored with a .preload flag for optimized preload.
	 * 
     * The data is retained as long as the indexeddb is not wiped.
     * When it is wiped this function will run again the next time the app loads.
     * 
	 * @author Mr.Scatman
	 */
	 getAllPreloads=async()=> {
		console.log("%c GET ALL PRELOADS", ANNE.TALK.bgbigred);
		ANNE.A.setHTML("loader_msg", "getting all preloads (one-time)");
		if (ANNE._annodesLS.last_uri=="") {
			ANNE._bNEED_TO_PRELOAD_DATA=true;
			console.log("Can't continue with preload until node is set.");
			return;
		}
		var aMinBootRelons = [];
		const LIMIT = 9000;
		var offset = 0;
		var response = await this.apiSimpleQueryWithParams("getAllPreloads", {
			limit:LIMIT,
			offset:offset
		});

		var safety=4;

		while (response.status=="good" && --safety>=0) {
			aMinBootRelons = aMinBootRelons.concat(response.result);
			ANNE.A.setHTML("loader_msg", ANNE.UTILS.commas(aMinBootRelons.length)+" relons acquired");
			if (response.result.length==LIMIT) {
				offset+=LIMIT;
				response = await this.apiSimpleQueryWithParams("getAllPreloads", {
					limit:LIMIT,
					offset:offset
				});
			}
		}

		//switching to preload=1 for preload vars.
		console.log("final process of all relons: "+aMinBootRelons.length);
		for (var r=0; r < aMinBootRelons.length;r++) {
			aMinBootRelons[r].preload=1;
		}
		return await MEMP.processNeuronsFromRelons(aMinBootRelons, true, true, "relons", "API.get@llPreloads");
	 }



	/**
	 * We have payment relons from sender->receiver with .annies.
	 * @param {*} $nid 
	 * @returns 
	 */
	getAllPaymentsIn=async($nid)=>{
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllPaymentsIn", $nid);
		lastBlock = phObj.lastBlock;
		console.log("%c(getAllPaymentsIn) with lastBlock:"+lastBlock, TALK.bgred);
        var aNewRelons = [];
		var response = await this.apiSimpleQueryWithParams("getAllPaymentsIn", {
			nid:$nid,
			height:lastBlock,
		});
		if (response.status=="good") {
			var relons_to_process = response.result;
            aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getAllPaymentsIn");
			await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
        }
        return aNewRelons;
	}

	
	/**
	 * TYPE_BROADCAST and TYPE_INTERNAL_PAY are now in relons_payment tbl.
	 * The API will pull from them.  Once in this app they are all treated the same
	 * as they are still just relons.
	 * @param {*} $nid 
	 * @param {*} $fromMsg 
	 * @returns  
	 */
	getAllPayments=async($nid, $fromMsg="")=>{
		if (this.opcheckerPass("getAllPayments", 30)==false) {
			return [];
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.openSubHunt("getAllPayments");
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllPayments", $nid);
		var aNewRelons = [];
		var lastHeight = phObj.lastBlock;

		var response = await this.apiSimpleQueryWithParams("getAllPayments", {
			nid:$nid,
			height:lastHeight
		});
		
		if (response.status=="good") {
			var relons_to_process = response.result;

			if (relons_to_process.length==0) {
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
				return [];
			}
			
            aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getAllPayments <-"+$fromMsg);
			await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
        }

		this.opcheckerTrackIt("getAllPayments");
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
        return aNewRelons;
	}


	/**
	 * 
	 * @param {*} $el 
	 * @param {*} $annecoin 
	 * @param {*} $fee 
	 * @param {*} $msg 
	 * @returns 
	 */
	generateSendTransactionQRCode=async($el, $annies, $fee=100, $msg="")=>{
		const elImg = document.createElement("img");
		$el.appendChild(elImg);
        elImg.classList.add("qrsquare");
		var uri = this.base_uri;
		if (uri==undefined) {
			alert("?????");
			console.log("error, fix this generateSendTransactionQRCode")
			return null;
		}
		elImg.src = uri+"/anne?requestType=generateSendTransactionQRCode&immutable=true&receiverId="+USER._userNObj.nid+"&amountNQT="+$annies+"&feeNQT="+$fee+"&message="+$msg;
		return {requestType:'generateSendTransactionQRCode',
				immutable:true,
				receiverId:USER._userNObj.nid,
				amountNQT:$annies,
				feeNQT:$fee,
				message:$msg
		}
	}


	/**
	 * 
	 * @returns 
	 */
	getREDLIST=async()=>{
		var response = await this.apiSimpleQueryWithParams("getRedlist", {});
		if (response.status=="good") {
			return response;
        }
		return null;
	}
	


	/**
	 * Get All my neurons
	 */
	getAllNeuronsFor=async($nid, $bSPONSOREDTOO=true)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllNeuronsFor", $nid+$bSPONSOREDTOO);
		lastBlock = phObj.lastBlock;

		console.log("%c(getAllNeuronsFor) with lastBlock:"+lastBlock, TALK.bgred);
        var aNewRelons = [];
		var response = await this.apiSimpleQueryWithParams("getAllNeuronsFor", {
			nid:$nid,
			height:lastBlock,
			sponsored:$bSPONSOREDTOO
		});
	
		console.time("processganf");
		if (response.status=="good") {
			var relons_to_process = response.result;
            aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getAllNeuronsFor");
			await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
			console.log(aNewRelons.length);
        }
		console.timeEnd("processganf");

        return aNewRelons;
	}


	/**
	 * 
	 * @param {*} $txids_to_check 
	 */
	getTXIDSThatDontExistInMEMPOOOL=async($txids_to_check)=>{
		return;
		console.log("%c(getTXIDSThatDontExistInMEMPOOOL) with "+$txids_to_check.length+" txids", TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("getTXIDSThatDontExistInMEMPOOOL", {
			txids:$txids_to_check,
		});
	
		if (response.status=="good") {
			console.log(response.IDS_NOT_IN_MEMPOOL);
			return response.IDS_NOT_IN_MEMPOOL;
        }
        return [];
	}

	
	verifyPayloadSig=async($user_nid, $sighex, $jsonstring)=>{
		console.log("%c(verifyPayloadSig)", TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("verifyPayloadSig", {
			user_nid:$user_nid,
			sighex:$sighex,
			jsonstring:$jsonstring
		});
		console.log(response);
		if (response.status=="good") {
			return response.verified;
        }
        return false;
	}


	verifyLogin=async($user_nid, $user_pub_key, $proof_sig, $msg_that_was_signed)=>{
		console.log("%c(verifyLogin_EX)", TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("verifyLogin", {
			user_nid:$user_nid,
			user_pub_key:$user_pub_key,
			proof_sig:$proof_sig,
			msg_that_was_signed:$msg_that_was_signed
		});
	
		if (response.status=="good") {
			console.log(response.IDS_NOT_IN_MEMPOOL);
			return response.IDS_NOT_IN_MEMPOOL;
        }
        return [];
	}

	/**
	 * An example how to use this QUERY LANGUAGE.
	 * 
	 * from:{
	 * 		val:"{}" || "somevalue", 
	 * 		op: "=" | "!="" | "LIKE", 
	 * 		use_results_from_pool:"unique_name_of_pool_to_use_for_filter"
	 * 		slot: "toA" || "from"
	 * 		},
	 * type: TYPE,
	 * reln: {val:nid, op:"="},
	 * to_type: TO_TYPE,
	 * to: {
	 * 		val:"{}" || "somevalue", 
	 * 		op: "=" | "!="" | "LIKE" | "DOWNLINE" | "DATE_GREATERTHAN" | "DATE_LESSTHAN" | "DATE_EQUAL" | "DATE_BETWEEN", 
	 * 		use_results_from_pool:"unique_name_of_pool_to_use_for_filter" 
	 * 		slot: "toA" || "from" (defaults to "from if not set",
	 * 		reln_down: "(set if using op "DOWNLINE") nid of reln if using DOWNLINE",
	 *      max_gens: "(optional) numbber of gens to chase down building the pool"
	 * 		},
	 * pool_type: "match" || "lookup_pool"
	 * results_pool_name:"unique_name_for_results_in_this_pool",
	 * debug: "somenameforprinting"
	 * 
	 * @returns 
	 */
	test_queryByRelonJSON_archerPool = async()=>{

		//test dynamic lang
		//person named trump
		var relons = [
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.inst.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:ANNE.MEM.person.nid, op:"="},
				pool_type:"match"
			},
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.name.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"%archer%", op:"LIKE"},
				results_pool_name:"archerpool",
				pool_type:"match",
				debug:"people with name archer in their name"
			},

			//do a lookup pool of all TEXAS locations.
			{
				from:{val:"", op:""},
				type:-1,
				reln:{val:"", op:""},
				to_type:-1,
				to:{val:"9008074888561760466", op:"DOWNLINE", reln_down:MEM.constituent.nid, max_gens:4},
				results_pool_name:"texas locations",
				pool_type:"lookup_pool",
				debug:"texas locations"
			},

			//find a relon to ANY LOCATION IN THE ABOVE POOL.
			//birthplace IN Texas (anywhere)
			{
				from:{val:"", op:"", use_results_from_pool:"archerpool"},
				type:-1,
				reln:{val:"7777813598381855599", op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"{}", op:"=", use_results_from_pool:"texas locations", slot:"toA"},
				pool_type:"match",
				debug:"match users to birth place in texas"
			},
		];

		return await this.queryByRelonJSON(relons);
	}



	test_queryByRelonJSON_dates = async()=>{
		var relons = [
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.inst.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:ANNE.MEM.person.nid, op:"="},
				pool_type:"match",
				debug:"all people pool (not efficient, lol)"
			},
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.name.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"%trump%", op:"LIKE"},
				results_pool_name:"trumppool",
				pool_type:"match",
				debug:"people with name trump in their name"
			},

			//do a lookup pool of all TEXAS locations.
			{
				from:{val:"", op:""},
				type:-1,
				reln:{val:"", op:""},
				to_type:-1,
				to:{val:"", op:"DATE_BETWEEN", date_min:"4655170", date_max:"4655190"},
				results_pool_name:"datepool",
				pool_type:"lookup_pool",
				debug:"dates equal to 1946-06-14"
			},

			//find a relon to ANY LOCATION IN THE ABOVE POOL.
			//birthplace IN Texas (anywhere)
			{
				from:{val:"", op:"", use_results_from_pool:"trumppool"},
				type:-1,
				reln:{val:MEM.birthday.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"{}", op:"=", use_results_from_pool:"datepool", slot:"from"},
				pool_type:"match",
				debug:"persons born on date 4655179"
			},
		];

		return await this.queryByRelonJSON(relons);
	}


	/**
	 * find a clump by its roles (however many you like.)  Daisy chain of pools..
	 * @param {*} $action_nid 
	 * @param {*} $roles 
	 */
	findClump=async($action_nid, $roles)=> {
		var role_cmds = [];
		var actionNOBJ = MEMG.getNeuronByNid($action_nid);
		if (actionNOBJ.ghost==true) {
			console.warn("ghost neuron detected in to "+actionNOBJ.text);
			return [];
		}

		//action nid is the BE/parent.
		//this will build a pool of ALL clumps with this action.
		var action_role = {
			from:{val:"", op:""},
			type:TYPE_BE,
			reln:{val:ANNE.MEM.inst.nid, op:"="},
			to_type:TO_TYPE_ANNE,
			to:{val:$action_nid, op:"="},
			pool_type:"match"
		};
		console.log("audit clump lookup: [BE] "+actionNOBJ.text);
		role_cmds.push(action_role);

		//{role:$roleNid, role_value:$roleValueNid}
		//these cmds will serve as additional filters for the pool of 
		//actions that match the initial pool.
		for (var r=0; r < $roles.length; r++) {
			var role = $roles[r];
			
			var toNOBJ = ANNE.MEMG.getNeuronByNid(role.role_value);
			if (toNOBJ.ghost==true) {
				console.warn("ghost neuron detected in to "+toNOBJ.text);
				return [];	
			}
			
			var rcmd = 
			{
				from:{val:"", op:""},
				type:TYPE_ROLE,
				reln:{val:role.role, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:role.role_value, op:"="},
				pool_type:"match"
			};
			
			if (r==0) {
				rcmd.results_pool_name="r0";
			}
			else {
				if (r!=$roles.length-1) {
					rcmd.results_pool_name="r"+r;
				}
				else rcmd.use_results_from_pool = "r"+(r-1);
			}
			
			var roleNOBJ = ANNE.MEMG.getNeuronByNid(role.role);
			console.log("audit clump lookup: ["+roleNOBJ.text.toLowerCase()+"] "+toNOBJ.text);
			role_cmds.push(rcmd);
		}
		
		var relons = await this.queryByRelonJSON(role_cmds);
		
		if (relons.length>0) {
			var froms = [];
			for (var r=0; r < relons.length; r++) {
				if (froms.indexOf(relons[r].from_nid)==-1) {
					froms.push(relons[r].from_nid);
				}
			}
			return ANNE.MEMG.getNObjsFromList(froms);
		}
		
		else return [];
	}


	test_queryByRelonJSON_clump = async()=>{

		//test dynamic lang
		//person named trump
		var relons = [
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.inst.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"7618523810988349729", op:"="},
				pool_type:"match"
			},
			{
				from:{val:"", op:""},
				type:TYPE_ROLE,
				reln:{val:ANNE.MEM.role_when.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"2579424936603799988", op:"="},
				pool_type:"match",
				debug:"when role to 2579424936603799988"
			},
			{
				from:{val:"", op:""},
				type:TYPE_ROLE,
				reln:{val:ANNE.MEM.role_actor.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"3482471539687014039", op:"="},
				pool_type:"match",
				debug:"role actor should be x.com"
			}
		];

		console.log("debug");
		console.log(relons);
		return await this.queryByRelonJSON(relons);
	}


	/**
	 * 
	 * @param {*} $rObjs 
	 * @returns 
	 */
	queryByRelonJSON = async($rObjs)=>{
		console.log($rObjs);
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryByRelonJSON");
		lastBlock = 0;// phObj.lastBlock;
		console.log("%c(queryByRelonJSON) with lastBlock:"+lastBlock, TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("queryByRelonJSON", {
			jsonstring:JSON.stringify($rObjs),
			height:lastBlock
		});

		console.log(response);
		var aRelons = [];

		if (response.status=="good") {
			if(response.result.length) {
				await HISTORY.processResultsForNewPullHistory(response.result, phObj);
				aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryByRelonJSON");
			}
		}
		return aRelons;
	}


	/**
	 * 
	 * @param {*} $nid 
	 * @param {*} $forceNEW 
	 * @returns 
	 */
	getAllMyStuff_lite=async($nid, $forceNEW=false)=> {
		if (USER._userNObj.nid==MEM.ANNE_ACCTID) return [];
		if (USER._userNObj.ghost==true) return [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllMyStuff_lite", $nid);
		
		lastBlock = Math.max(0, phObj.lastBlock-10);
		
		if ($forceNEW==true) lastBlock=0;
		console.log("%c(getAllMyStuff_lite) with lastBlock:"+lastBlock, TALK.bgred);
		
		//moving to an IDB helper to keep it cleaner.
		if (ANNE.SETTINGS && ANNE.SETTINGS.bPRELOADALLDATA==true) {
			await ANNE.IDB.getAllMyStuff(phObj, $nid, true);
		}
		
		var aRelons = [];
		var LIMIT = 8000;
		var currentOffset = 0;
		var response = await this.apiSimpleQueryWithParams("getAllMyStuff_lite", {
			nid:$nid,
			height:lastBlock,
			limit:LIMIT,
			offset:currentOffset
		});
		
		//abandoned the limit/offset as we have multiple functions (AAAAR)
		if (response.status=="good") {
			if (response.result.length>0) {
				await HISTORY.processResultsForNewPullHistory(response.result, phObj);
				aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getAllMyStuff");
			}
		}
		return;
	}


	/**
	 * Get all neurons, outbound payments and inbound payments
	 * First load data from IDB so we have it in RAM.
	 * @param {*} $nid 
	 * @param {*} $bSPONSOREDTOO 
	 * @returns 
	 */
	getAllMyStuff=async($nid, $bSPONSOREDTOO=true, $forceNEW=false)=> {
		if (USER._userNObj.nid==MEM.ANNE_ACCTID) return [];
		if (USER._userNObj.ghost==true) return [];
	
		if (this.opcheckerPass("getAllMyStuff")==false) {
			return []; 
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("getAllMyStuff");
		
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllMyStuff", $nid+$bSPONSOREDTOO);
		
		lastBlock = Math.max(0, phObj.lastBlock-10);
		
		if ($forceNEW==true) lastBlock=0;
		console.log("%c(getAllMyStuff) with lastBlock:"+lastBlock, TALK.bgred);

		//moving to an IDB helper to keep it cleaner.
		//if we didn't preload all data, we need to get all our stuff from IDB atleast once
		//NOTE:  our stuff should be marked as preload already, but .. this is fine to do.
		if (ANNE.SETTINGS && ANNE.SETTINGS.bPRELOADALLDATA==false) {
			await ANNE.IDB.getAllMyStuff(phObj, $nid, $bSPONSOREDTOO);
		}

		var LIMIT = 8000;
		var currentOffset = 0;
		var response = await this.apiSimpleQueryWithParams("getAllMyStuff", {
			nid:$nid,
			height:lastBlock,
			sponsored:$bSPONSOREDTOO,
			limit:LIMIT,
			offset:currentOffset
		});

		
		//abandoned the limit/offset as we have multiple functions (AAAAR)
		if (response.status=="good") {
			var rels = [];
			if (response.narrs && response.narrs.result && response.narrs.result.length>0) {
				//console.log("we have nars!")
				if (response.narrs.result) {
					rels = response.narrs.result;
				}
			}
			if (response.awareness && response.awareness.result && response.awareness.result.length>0) {
				if (response.awareness.result) {
					rels = rels.concat(response.awareness.result);
				}
			}
			if (rels.length>0) {
				await MEMP.processNeuronsFromRelons(rels, true, true, "relons", "API.getAllMyStuff");
			}
				
			await HISTORY.processResultsForNewPullHistory(rels, phObj);
		}
		
		this.opcheckerTrackIt("getAllMyStuff");
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt("getAllMyStuff");
		
		return;
	}
	


	/** 
	 * Get the anneid summary level data for an ANNE identity.
	 * this is used for tallies, summaries, of a user's andy_actions, relons_sponsored, hanais, and other metrics.
	 * @author Mr.Scatman
	 */
	getANNEID=async($nObj, $bFORCEREDO=false)=> {
		if ($nObj==null) return;
		if ($bFORCEREDO==false && $nObj._anneid) return $nObj._anneid;

		var response = await this.apiSimpleQueryWithParams("getANNEID", {
			nid:$nObj.nid
		});

		if (response.status=="good") {
			if (response.result!=null && response.results) $nObj._anneid = new anneid($nObj, response.result);
			else $nObj._anneid = null;
		}
		return $nObj._anneid;
	}


	/**
	 * 
	 * @param {*} $nObj 
	 */
	getANNEID_ranks=async($nObj, $bFORCEREDO=false)=>{
		if ($nObj==null) return;
		if ($bFORCEREDO==false && $nObj._anneid!=null && $nObj._anneid.bRANKS==true) return $nObj._anneid;

		var response = await this.apiSimpleQueryWithParams("getANNEID_ranks", {
			nid:$nObj.nid
		});

		if (response.status=="good") {
			if (response.result!=null && response.results) {
				$nObj._anneid = new anneid($nObj);
				$nObj._anneid.processFromObj(response.result[0]);
			}
			else $nObj._anneid = null;
			
			if (response.extra_relons!=undefined) {
				await MEMP.processNeuronsFromRelons(response.extra_relons, false, true, "relons", "");
			}
		}
		return $nObj._anneid;
	}


	/**
	 * Get the annode pubKey from the annode auth object stored from the handshake.
	 */
	getAnnodePk = function() {
		if (this._AAUTH==null) return null;
		else return this._AAUTH.pubKey;
	}


	/**
	 * Get balance by nid.
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	getBalance = async($nid)=> {
		var json = await this.apiSimpleQueryWithParams("getNeuronBalance",{nid:$nid});
		if (json.status!=undefined && json.status=="fail") return null;
		return json;
	}

	
	/**
	 * Get a merge analysis object for helpful info.
	 * @param {*} $nid1 - nid of a neuron.
	 * @param {*} $nid2 - nid of another neuron.
	 * @returns 
	 */
	getMergeReconByNids = async($nid1, $nid2)=> {
		if ($nid1==$nid2) return {status:"bad", reason:"can't merge a neuron into itself"};
		var json = await this.apiSimpleQueryWithParams("getMergeReconByNids",{nid1:$nid1, nid2:$nid2});
		if (json.status!=undefined && json.status=="fail") return null;
		console.log(json);
		return json;
	}

	getMergeReconByTXID = async($txid)=> {
		var json = await this.apiSimpleQueryWithParams("getMergeReconByTXID",{txid:$txid});
		if (json.status!=undefined && json.status=="fail") return null;
		console.log(json);
		return json;
	}


	/**
	 * query the NAPLEX of a given sponsor and class.
	 * Build the aplex objects so we have access to their data in our viz.
	 * @param {*} $sponsored_by_nid 
	 * @param {*} $class_nid 
	 * @returns 
	 */
	queryNAPLEX=async($sponsored_by, $class_nid, $offset=0, $nObjs=[])=> {
		const LIMIT = 2000;
		var json = await this.apiSimpleQueryWithParams("getNaplexByClass", 
				{sponsored_by:$sponsored_by, class_nid:$class_nid, limit:LIMIT, offset:$offset}
		);
		if (json.status!=undefined && json.status=="fail") return null;
		console.log(json);
		var results = json.result;
		var nObjs = $nObjs;

		for (var r=0; r < results.length; r ++) {
			var rez = results[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;

			if ($sponsored_by!=undefined && $sponsored_by!=null) {
				nObj.sponsored_by = $sponsored_by;
			}
			
			nObj._aplex = new aplex(nObj, nObj.nid, rez);	//here

			if (results.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(results.pending, false, true, "relons", "API.queryNAPLEX");
            }

			nObjs.push(nObj);
		}

		if (nObjs.length==LIMIT+$offset) {
			return await this.queryNAPLEX($sponsored_by, $class_nid, $offset+LIMIT, nObjs);
		}
		return nObjs;
	}


	/**
	 * We get a set of aplex objects here.
	 * @returns {Array} nObjs - an array of neurons that are the top by level.
	 */
	getTopXAnnexByLevel = async($limit=20)=>{
		var response = await this.apiSimpleQueryWithParams("getTopXAnnexByLevel", {limit:$limit});
		//if (json.status!=undefined && json.status=="fail") return null;
		//console.log(json);
		//
		return await this.processNaplexOptimized(response, "API.queryForRecents");

		var results = json.result;
		var nObjs = [];
		for (var r=0; r < results.length; r ++) {
			var rez = results[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;
			nObj._aplex = new aplex(nObj, nObj.nid, rez);

			if (results.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(results.pending, false, true, "relons", "API.getTopXAnnexByLevel");
            }
			nObjs.push(nObj);
			if (nObjs.length==$limit) break;
		}
		return nObjs;
	}



	
	getAwarvuSummary = async($since_height=0)=>{
		var response = await this.apiSimpleQueryWithParams("getAwarvuSummary", {height:$since_height});
		var annexerWhales = [];
		if (response.status=="good") {
			var results = response.result;
			for (var r=0; r < results.length; r++) {
				var anneid_obj = results[r];
				var nid = anneid_obj.nid;
				var nObj = MEMG.getOrMakeIfNew(nid);

				nObj._anneid = new anneid(nObj, anneid_obj);
				
				annexerWhales.push(nObj);
			}
		}
		return annexerWhales;
	}


	/**
	 * we get an array of anneid records here.
	 * @param {*} $cnt 
	 * @returns {array} annexWhalers - an array of nObjs with thie respective ._anneid objects set.
	 */
	getAnnexerWhales = async($limit=10)=>{
		var response = await this.apiSimpleQueryWithParams("getANNEXWhalesNaplex", {limit:$limit});
		var anneids = response.anneids;
		var naplex = response.naplex;

		var nObjs = [];

		if (anneids && anneids.status=="good") {
			var results = anneids.result;
			for (var r=0; r < results.length; r++) {
				var anneid_obj = results[r];
				var nid = anneid_obj.nid;
				var nObj = MEMG.getOrMakeIfNew(nid);
				console.log(nObj.nid+" "+nObj.text+" (has _anneid)");
				nObj._anneid = new anneid(nObj, anneid_obj);
				if (nObj._anneid==null) {
					console.warn("WHAAAAAAT?");
				}
				nObjs.push(nObj);
			}
		}

		if (naplex==undefined || naplex==null) return nObjs;
		console.log(naplex);
		var result = naplex.result;
        for (var r=0; r < result.length; r ++) {
			var rez = result[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;
			nObj._aplex = new aplex(nObj, nObj.nid, rez);	//here
			console.log(nObj.nid+" "+nObj.text+" (has _aplex)");
			if (rez.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(rez.pending, false, true, "relons", "API.queryForNakedText");
            }
		}

		return nObjs;
	}


	/**
	 * 
	 * @param {*} $nObj 
	 * @param {*} $height 
	 * @returns 
	 */
	getAwarvuSummaryFor = async($nObj, $since_height=0) => {
		if ($nObj==null) return null;
		if ($nObj.ghost==true) return null;
		var response = await this.apiSimpleQueryWithParams("getAwarvuSummaryFor", {
			nid:$nObj.nid,
			since_height:$since_height
		});
		if (response.status=="good") {
			var apxs = new aplex_sum($nObj, APLEX_SUMMARY_SPECIFIC_NEURON, response);
			$nObj._aplex_sum = apxs;
			return apxs;
        }
		return null;
	}


	/**
	 * Will summarize ALL AWAREZ over the entire system.
	 * @param {*} $since_height 
	 * @returns 
	 */
	getAwarvuMetaSummary = async($since_height=0) => {
		var response = await this.apiSimpleQueryWithParams("getAwarvuMetaSummary", {
			since_height:$since_height
		});
		if (response.status=="good") {
			if (response.result!=null && parseInt(response.results)>0) {
				var apxs = new aplex_sum(MEM.___SELF___, APLEX_SUMMARY_NEURON_LINE, response);
				return apxs;
			}
			else {
				var apxs = new aplex_sum(MEM.___SELF___, APLEX_SUMMARY_NEURON_LINE, null);
				return apxs;
			}
        }
		return null;
	}

	
	/**
	 * Get the aplex data for neuron or hypergraph edge (relon).
	 * 
	 * This will get a feelz summary for any neuron or edge.
	 */
	getAplex=async($nObj=null, $bFORCENEW=false)=> {
		if ($bFORCENEW==false && $nObj._aplex!=undefined && $nObj._aplex!=null) return $nObj._aplex;

		if ($nObj==null) return null;
		if ($nObj.ghost==true) return null;
		var response = await this.apiSimpleQueryWithParams("getAplex", {
			id:$nObj.nid,
			id_type:0
		});
		if (response.status=="good") {
			if (response.result!=null && response.results!=0 && response.results!="") {
				if (response.result.text!="" && $nObj.text=="") $nObj.text = response.result.text;
				
				var extra_out = null;
				
				if (response.result.pending!=undefined) {
					if (response.result.pending.extra_outs!=undefined) {
						var extras = response.result.pending.extra_outs;
						extra_out = extras[0];
						ANNE.TALK.printRelons(extras);
						await MEMP.processNeuronsFromRelons(extras, false, true, "relons", "API.getAplex pending.extra_outs");
					}
				}
				else if (response.result && response.result.sponsored_by && response.result.sponsored_by!=$nObj.sponsored_by) {
					//update neuron sponsored by
					console.log("update neuron sponsored by.");
					$nObj.sponsored_by = response.result.sponsored_by;
				}

				var updates = [];
				var ghost_updates = [];

				if($nObj.sponsored_by!=null && $nObj.sponsored_by!=undefined) {
					for (var r=0; r < $nObj.outs.length; r++) {
						if ($nObj.outs[r].sponsored_by!=$nObj.sponsored_by) {
							$nObj.outs[r].sponsored_by = $nObj.sponsored_by;
							if ($nObj.ghost==true || $nObj.outs[r].ghost==true || $nObj.outs[r].txid[0]=="z") {
								ghost_updates.push($nObj.outs[r]);
							}
							else {
								updates.push($nObj.outs[r]);
							}
						}
					}
				}
				
				if (updates.length) {
					await ANNE.DB.___updateAll(updates);
				}
				if (ghost_updates.length) {
					await ANNE.DB_local.___updateAll(ghost_updates, "ghost_relons");
				}

				$nObj._aplex = new aplex($nObj, $nObj.nid, response.result);
			}
			else {
				$nObj._aplex = new aplex($nObj, $nObj.nid, null);
			}
        }

        //augment with our ghost tally. (????)
        ANNE.AWARE.augmentGhosts($nObj);

        return $nObj._aplex;
	}


	/**
	 * 
	 * @param {*} $yearNid 
	 * @param {*} $monthNid 
	 * @param {*} $dayNid 
	 */
	getDates = async($yearNid, $monthNid, $dayNid)=>{
		var searchRels = [];
		var target="from_nid";

		if ($yearNid!=0) {
			var yrRel = {target:target, be:[ANNE.MEM.date_gregorian.nid], type:[TYPE_HAS], reln:ANNE.MEM.year.nid, to_a:$yearNid};
			searchRels.push(yrRel);
			target="pool";
		}

		if ($monthNid!=0) {
			var monthRel = {target:target, type:[TYPE_HAS], reln:ANNE.MEM.month.nid, to_a:$monthNid};
			if (searchRels.length==0) monthRel.be=[ANNE.MEM.date_gregorian.nid];
			searchRels.push(monthRel);
			target="pool";
		}

		if ($dayNid!=0) {
			var dayRel = {target:target, type:[TYPE_HAS], reln:ANNE.MEM.day.nid, to_a:$dayNid};
			if (searchRels.length==0) dayRel.be=[ANNE.MEM.date_gregorian.nid];
			searchRels.push(dayRel);
		}

		
		var response = await this.apiSimpleQueryWithParams("getAplex", {
			json_relon:searchRels
		});


		console.log(response);
		return;
	}

	
	/**
	 * for relons (pass id_type 1)
	 */
	 getAplexRelon=async($relon=null)=> {
		if ($relon==null) return null;
		if ($relon.ghost==true) return null;
		var response = await this.apiSimpleQueryWithParams("getAplex", {
			id:$relon.txid,
			id_type:1
		});
		if (response.status=="good") {
			if (response.result!=null && response.results!=0) new aplex($relon, $relon.txid, response.result);
			else new aplex($relon, $relon.txid, null);
        }

		//augment with our ghost tally.
		ANNE.AWARE.augmentGhosts($relon);

        return $relon._aplex;
	}


	/** 
	 * Get the blockchain status info. 
     */
	getBlockchainStatus = async()=> {
		var json = await this.apiSimpleQuery("getBlockchainStatus");
		console.log(json);
		return json;
	}


	/**
	 * @param {String} $nid - nid longstring 
	 * @author Mr.Scatman
	 */
	getLNOTS = async($nid)=> {
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getLnots", $nid);
		var lastBlock = 0;//phObj.lastBlock;	//optimize this later after testing.
		var json = await this.apiSimpleQueryWithParams("getLnots",{nid:$nid});// height:lastBlock});
		if (json.status!=undefined && json.status=="fail") return null;
		await HISTORY.processResultsForNewPullHistory(json.lnots, phObj);
		return json.lnots;
	}

	
	/**
	 * 
	 * @param {*} $reln 
	 * @param {*} $to 
	 * @param {*} $to_type 
	 * @returns 
	 */
	getRelonsWithTo = async($reln, $to, $to_type)=> { 
		var response = await this.apiSimpleQueryWithParams("getRelonsWithTo",{reln:$reln, to:$to, to_type:$to_type});// height:lastBlock});
		if (response.status=="good") return response.result;
		return [];
	}
	


	/**
	 * 
	 * @param {String} $parentNid - nid longstring 
	 * @param {*} $value - proposed new subanndle under the parent anndle/subanndle
	 * @returns 
	 */
	getSubAnndle = async($parentNid, $value)=> {
		var json = await this.apiSimpleQueryWithParams("getSubAnndle",{parent_nid:$parentNid, value:$value});
		return json;
	}

	/**
	 * 
	 * @param {String} $anndle_str 
	 * @returns 
	 */
	getAnndlePricing = async($anndle_str)=>{
		var response = await this.apiSimpleQueryWithParams("getAnndlePricing", {anndle:$anndle_str});
		console.log(response);
		if (response.status=="good") {
			return response.result;
		}
		return null;
	}

	/**
	 * 
	 * @param {*} $nid 
	 * @returns 
	 */
	calculateAnnexPrice = async($nid)=>{
		var response = await this.apiSimpleQueryWithParams("calculateAnnexPrice", {nid:$nid});
		console.log(response);
		if (response.status=="good") {
			return response.result;
		}
		return null;
	}


	/**
	 * 
	 * @param {string} - account we want to query TXs for.
	 * 
	 * @author Mr.Scatman								
	 */
	getUnconfirmedTransactions = async($account)=> {
		var response = await this.apiSimpleQueryWithParams("getUnconfirmedTransactions", 
			{	
				account:$account
			}
			);
		console.log(response);
		if (response.unconfirmedTransactions!=undefined
			&& response.unconfirmedTransactions.length) {
			console.log(response.unconfirmedTransactions);
			return response.unconfirmedTransactions;
		}
		return [];
	}




	/**
	 * @param {*} $txid 
	 * @returns 
	 */
	getRelon = async($txid)=> {
    	var response = await this.apiSimpleQueryWithParams("getRelon",{
			txid:$txid
		});
		if (response.status=="good") {
			if (response.result!=null) {
				response.result.bMUSTINSERT=true;
				await MEMP.processNeuronsFromRelons([response.result], false, true, "relons", "API.getRelon");
			}
		}
		return ANNE.MEMG.lookupByTXID($txid);
	}

	getRelonsBySYSREFTXID = async($txid)=> {
    	var response = await this.apiSimpleQueryWithParams("getRelonsWithSysRefTxid",{
			txid:$txid
		});
		if (response.status=="good") {
			if (response.result.length) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getRelonsBySYSREFTXID");
				return aRelons;
			}
		}
		return [];
	}

	//accepts blockid or height
	getTransactionIdsInBlock = async($blockid)=> {
		var json = await this.apiSimpleQueryWithParams("getTransactionIdsInBlock",{
			blockId:$blockid,
			height:0
		});
		return json;
	}
	getTransactionIdsInBlockByHeight = async($height=0)=> {
		var json = await this.apiSimpleQueryWithParams("getTransactionIdsInBlock",{
			height:$height
		});
		return json;
	}


	/**
	 * Get the ANNODE tranasction (traditional bockchain format), not the hypergraph relon.
	 * @param {*} $txid 
	 * @returns 
	 */
	 getBlockchainTransaction = async($txid)=> {
    	var json = await this.apiSimpleQueryWithParams("getBlockchainTransaction",{
			transaction:$txid
		});

		console.log(json);

		if(ANNE.UTILS.isset(json.attachment)==true) {
			if (json.attachment.reln!=undefined) {
				var from_nid = "";
				var attachment = json.attachment;
				if (json.type==40) from_nid=json.sender;
				else from_nid = attachment.from;
	
				var temp_rel = {
					txid:json.transaction,
					from_nid:from_nid,
					type:attachment.type,
					reln:attachment.reln,
					to_type:attachment.to_type,
					to_s:attachment.to,
					height:json.height,
					annies:json.amountNQT,
					force_update:true
				}
	
				//put this in mem (or update existing relon in mem)
				await MEMP.processNeuronsFromRelons([temp_rel], false, true, "relons", "API.getBlockchainTransaction");
			}
		}

		return json;
	}


	/**
	 * Get the public key for a neuron.
	 * This is only relevant with key'd neurons, not keyless.
	 * @param {*} $nid 
	 */
	getAccountPublicKey = async($nid, $uri="")=>{
		var nObj = MEMG.getOrMakeIfNew($nid);
		if (nObj.pubKey!=undefined) return nObj.pubKey;	
		
		var json = await this.apiSimpleQueryWithParams("getAccountPublicKey",{account:$nid}, $uri);
		console.log("json:",json);
		
		if (json.publicKey!=undefined && json.publicKey.length>0) {

			//this is a keyless neuron, not onboardable.
			if (json.non_zero_pub_key!=undefined && json.non_zero_pub_key==false) {
				nObj.non_zero_pub_key = false;
				return "";
			}

			//this is the pubkey onchain for this neuron.
			nObj.non_zero_pub_key = true;
			nObj.pubKey = json.publicKey; 
			return json.publicKey;
		}

		//no acct on record.
		else if (json.error_wtf!=undefined) {
			return null;
		}

		return null;
	}


	getNeuronRecon = async($nid)=> {
		console.log("API.getNeuronRecon");
		var json = await this.apiSimpleQueryWithParams("getNeuron",{nid:$nid});
		console.log("json:",json);
		if (json && json.result) {
			var rez = json.result;
			if (rez.acct_public_key!=undefined) {
				var nObj = MEMG.getOrMakeIfNew($nid);
				nObj.pubKey = json.acct_public_key; 
				return rez;
			}
		}
		return null;
	}


	/**
	 * No saving in local db here.. just a quick pull
	 */
	 getDirectFundsOf = async($account)=>{
		var response = await this.apiSimpleQueryWithParams("getDirectFundsOf", {nid:$account,});
		console.log(response);
		if (response.status=="good") {
			return response.result;
		}
		return [];
	}

	
	//get a specific feelz/opinion reln (from a list) that points into a nid.
	getSpecificAwarenesses = async($to_nid, $relns)=> {
		var lastBlock = 0;
		if ($to_nid[ $to_nid.length-1]=="z") {
			console.log("DONT CALL THIS WITH TEMP NEURONS!");
			return [];
		}
		var hash = ANNE.other.sha256($to_nid+$relns.join());
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getSpecificAwarenesses ", hash);
		lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("getSpecificAwarenesses", 
			{
				nid:$to_nid, 
				relns:$relns,
				height:lastBlock

			});
		console.log(response);
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.getSpecificAwarenesses");
		}
		return aRelons;
	}
	
	

	/**
	 * 
	 */
	getVoteData = async ($nid)=> {
		var lastBlock = 0;
		var response = await this.apiSimpleQueryWithParams("getVoteData", 
				{
					nid:$nid,
					height:0
				});
		console.log(response);
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.getVoteData");
		}
		return aRelons;					
	}


	/**
	 * We pull this everytime on this app, and also get the aplex records for display.
	 */
	queryForRecents=async($classNids, $max_results=200)=> {
		var lastBlock = ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK;
		if ($max_results==true) $max_results=100;
		var response = await this.apiSimpleQueryWithParams("queryForRecentsNaplex", {
			class_nids:$classNids,
			height:lastBlock-100000,
			max_results:$max_results
		});
		
		//make sure we have updated sponsored_by info here.
		if (response && response.narr && response.narr.result) {
			ANNE.TALK.printRelons(response.narr.result);
		}
		else return [];

		return await this.processNaplexOptimized(response, "API.queryForRecents");
	}



	/**
	 * Query for awareness relons.
	 * There could potentially be alot of these for any given $nid.
	 * These are outbound awareness relons, or "feelz" and "opinionz" expressed.
	 * @param {*} $nid 
	 * @param {*} $lastBlock 
	 * @param {*} $offset 
	 * @returns {Array}
	 */
	queryForAwareness=async($nid, $lastBlock, $offset)=> {
	   var aNewRelons = [];
	   var response = await this.apiSimpleQueryWithParams("queryForAwareness", {
		   nid:$nid,
		   height:$lastBlock,
		   offset:$offset
	   });

	   console.log("(queryForAwareness)");
	   if (response.status=="good") {
		   aNewRelons = response.result;
	   }

	   return aNewRelons;
   }


	/**
	 * get the global for this class and value
	 * 
	 * @param {String} $classNid - nid longstring
 	 * @param {*} $value
	 * @author Mr.Scatman
	 */
	queryForGlobal = async($classNid, $value)=> {
		var json = await this.apiSimpleQueryWithParams("getGlobal",{class_nid:$classNid, value:$value});
		return json;
	}


	/**
	 * Query for a neuron of a given class with a given value.
	 * Optionally provide a required reln and to_s value of that reln.
	 * The ANNODE will first match the class/value and then filter for reln_nid/to_s (if provided).
	 * 
	 * This does not check BEM line (be line class).  The neuron must have a be value of the value searched for 
	 * in classNid.  This is intended for flat value classes like persons, company, thing etc.
	 * @param {*} $classNid 
	 * @param {*} $value 
	 * @param {*} $reln_nid 
	 * @param {*} $to_type 
	 * @param {*} $to_s 
	 * @returns 
	 */
	queryFoNeuronByClassAndRelon = async($classNid, $value, $reln_nid, $to_type=TO_TYPE_ANNE, $to_s)=> {
		var json = await this.apiSimpleQueryWithParams("getByClassAndRelon",
			{class_nid:$classNid, 
				value:$value,
				reln:$reln_nid,
				to_type:$to_type,
				to:$to_s});
		return json;
	}


	queryFoNeuronByAltBeTexAndBe = async($classNid, $text, $alt_be_reln, $alt_be_to)=> {
		var json = await this.apiSimpleQueryWithParams("getByAltBeTextAndBe",
			{class_nid:$classNid, 
				text:$text,
				alt_be_reln:$alt_be_reln,
				alt_be_to:$alt_be_to});
		return json;
	}


	/**
	 * get a person by name and bday
	 * 
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForPersonWithBday=async($bday, $namestr, $alias="")=> {
		var json = await this.apiSimpleQueryWithParams("getPersonWithBday",
				{	
					bday:$bday, 
					name:$namestr,
					alias:$alias
				});
		return json;
	}

	getPerson=async($namestr)=> {
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.openSubHunt("getPerson");
		var json = await this.apiSimpleQueryWithParams("getPerson",{name:$namestr});
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return json;
	}


	/**
	 * 
	 * @param {*} $foundedOnStr 
	 * @param {*} $namestr 
	 * @param {*} $foundedAtNID 
	 * @returns 
	 */
	getCompany=async($namestr, $registratonPolityNid=null)=> {
		var json = await this.apiSimpleQueryWithParams("getCompany",
		{	 
			name:$namestr,
			registration_polity_nid:$registratonPolityNid
		});
		return json;
	}


	/**
	 * Query for a global location.
	 * We pass along the sender nid so we can check for user-owned neurons too..
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForLocation = async($name, $constituent_of_nid, $alt_be_nid)=> {
		var json = await this.apiSimpleQueryWithParams("getLocation",{
					name:$name, 
					constituent_of_nid:$constituent_of_nid,
					alt_be_nid:$alt_be_nid,
					sender_nid:ANNE.USER._userNObj.nid
				});
		return json;
	}


	/**
	 * Recon on the parts of a location that we are looking to potentially resolve.
	 * @param {*} $name 
	 * @returns 
	 */
	getLocationParts = async($name)=> {
		var response = await this.apiSimpleQueryWithParams("getLocationParts",{
			name:$name
		});
		
		if (response.status=="good") {
			/************************************************************** */
			//OPTIMIZED and far cleaner.
			//this should be done, everywhere we pull in new data.
			var relons_to_process = response.result;
			//var am = new arraymap();
			//am.setMapKey("txid");
			//for(var r=0; r < relons_to_process.length; r++) {
				//am.addU( relons_to_process[r]);
			//}
			//await ANNE.IDB.checkForNewTxids(am);
			//those left in the relons map are new.
            if(relons_to_process.length>0) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getLocationParts");
				return aNewRelons;
			}
			/************************************************************** */
		}
		return;
	}


	/**
	 * 
	 * @param {*} $anums 
	 * @returns 
	 */
	getNidsByAnums = async($anums)=> {
		var json = await this.apiSimpleQueryWithParams("getNidsByAnums",{
			anums:$anums
		});
		if (json && json.status=="good") {
			for (var a=0; a < json.result.length; a++) {
				var nid = json.result[a].nid;
				var nObj = ANNE.MEMG.getOrMakeIfNew(nid);
				if (nObj.text=="") nObj.text = json.result[a].anblstr;
			}
			return json.result.concat();
		}
		return null;
	}

	

	/**
	 * 
	 * @param {*} $name 
	 * @returns 
	 */
	queryForNamedLocation = async($name)=> {
		console.log("queryForNamedLocation: "+$name);
		var response = await this.apiSimpleQueryWithParams("getNamedLocation",{
					name:$name
				});

		if (response.status=="good") {
			var relons_to_process = response.result;
			if (relons_to_process.length) {
				aRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.queryForNamedLocation");
				//await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
				return aRelons;
			}
		}
		return [];
	}

	/**
	 * Query for a hanai.
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForHanai = async($nid, $height=0)=> {
		$height = Math.max(0, $height);
		//var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForHanai", $nid);
		//lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("queryForHanai", {
			nid:$nid, 
			height:$height
		});
		console.log(response);
		if (response.status=="good") {
			var relons_to_process = response.result;
			if (relons_to_process.length) {
				var aRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.queryForHanai");
				//await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
				return aRelons;
			}
		}
		return [];
	}


	/**
	 * User is an admin for an open hanai request that doesnt have a sys_ref_txid set.
	 * That is set when the repsonse is officially processed.
	 * Look for an admin response to the request that is pending mining.
	 */
	queryForHanaiResponseTo = async($txid)=>{
		var response = await this.apiSimpleQueryWithParams("queryForHanaiResponseTo", {
			txid:$txid
		});
		console.log(response);
		if (response.status=="good" && response.result.length>0) {
			var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForHanaiResponseTo");
			return aRelons[0];
		}
		return null;
	}
	

	/**
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	 queryForAndy=async($nid, $nids_i_am_delegated_admin_for=[])=> {
		//var lastBlock = 0;	//optimize later
		if ($nid==null) return;
		if ($nids_i_am_delegated_admin_for==null) $nids_i_am_delegated_admin_for=[];

		var height = 0;
		if (ANNE.MEM_UDF!=null) height = ANNE.MEM_UDF._collab_hlc;
		 
height = 0;

		//de-optimize until we get this right....
		var response = await this.apiSimpleQueryWithParams("queryForAndy", {
				from:$nid, 
				nids:$nids_i_am_delegated_admin_for, 
				height:height
			});

		//_collab_hlc is managed in MEM_UDF _processCollabs.
		//when a user is "unlocked" it preloads from local DB mem.
		//run-time it updates through here.
		console.log(response);
		if (response.status=="good") {
			if (response.result.length) {
				await MEM_UDF._processCollabs(response.result);
			}
			return MEM_UDF._collab_relons;
		}
		return MEM_UDF._collab_relons;
	}


	/**
	 * Query the node for this text name.
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForNakedText = async ($searchString, $bEXACT=false) => {
		if ($searchString=="") return;
		var lastBlock = 0;
		//var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForNakedTextNaplex", $searchString);
		//lastBlock = 0;//phObj.lastBlock;
		var exact = "no";
		if ($bEXACT==true) exact = "yes";

		var response = await this.apiSimpleQueryWithParams("queryForNakedTextNaplex", {
				text:$searchString,
				exact:exact, 
				height:lastBlock
		});

		//now returns an array of 
		return await this.processNaplexOptimized(response, "API.queryForNakedText");
	}


	/**
	 * we get annex sys relons back here of the most recent annex relons pointing to TO nids (the annexed neuron)
	 * @param {*} $cnt 
	 * @returns {Array} an array of {.nObj, .cnt}
	 */
	getLatestXAnnexes = async($limit=20)=>{
		var response = await this.apiSimpleQueryWithParams("queryForLatestXANNEXNaplex", {limit:$limit});
		return await this.processNaplexOptimized(response, "API.getLatestXAnnexes");
	}


	/**
	 * not optimizing for height, as we want all the latest napplex data.
	 * When using sponsored_by index, we get all relons unless we filter them out on the L1 query,
	 * as in the NARR view the .sponsored_by index is on all relons coming from the from_nid join.
	 * @param {*} $nid 
	 * @returns 
	 */
	getSponsored = async ($nid) => {
		var lastBlock = 0;
		var response = await this.apiSimpleQueryWithParams("getMySponsoredNaplex", {
				nid:$nid,				
				height:lastBlock
		});
		//now returns an array of 
		return await this.processNaplexOptimized(response, "API.getSponsored");
	}


	/**
	 * 
	 * @param {*} $response 
	 * @returns 
	 */
	processNaplexOptimized = async($response, $debug="")=>{
        var narr = $response.narr;
		if (narr && narr.status=="good") {
			if (narr.result.length>0) {
				await MEMP.processNeuronsFromRelons(narr.result, true, true, "relons", $debug);
			}
		}
		var naplex = $response.naplex;
		var nObjs = [];
        if (naplex==undefined || naplex==null) return nObjs;
		console.log(naplex);
		var result = naplex.result;
		for (var r=0; r < result.length; r ++) {
			var rez = result[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;
			nObj._aplex = new aplex(nObj, nObj.nid, rez);	//here
			nObjs.push(nObj);
			if (rez.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(rez.pending, false, false, "relons", "API.queryForNakedText");
            }
		}
		return nObjs;
	}


		
	/**
	 * 
	 * @param {*} $searchString 
	 * @param {*} $wvs 
	 * @param {*} $bes 
	 * @param {*} $bEXACT 
	 * @param {*} $bAPPLYTOBELINE ---must be true to invoke enforement of $bes array OR bem line via inbeline
	 * @param {*} $inbeline  --- left at 0 (or null) when not wanting a BEM LINE seearch of that class nid
	 * @returns 
	 */
	queryForText = async ($searchString, $wvs=[], $bes=[], $bEXACT=false, $bAPPLYTOBELINE=false, $inbeline=0) => {
		if ($searchString=="") return;
		$searchString = $searchString+"";
		$searchString = $searchString.trim();
		var aRelons = [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForText", $searchString+$bEXACT+$bAPPLYTOBELINE);
		lastBlock = 0;//phObj.lastBlock;

		console.log("(API.queryForText) "+$searchString);
		var response = await this.apiSimpleQueryWithParams("queryForText", {
				text:$searchString, 
				wvs:$wvs, 
				bes:$bes,
				applytobeline:$bAPPLYTOBELINE,
				inbeline:$inbeline, 
				height:lastBlock
			});

		if (response.status=="good") {
			if (response.result.length>0) {
				if(response.result.length>100) {
					console.log("what is this??  so many results for "+$searchString);
				}
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForText");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				return aRelons;
			}
		}
		return [];
	}


	/**
	 * Much more efficient to use this when we know we are looking for a valid identity (by text string)
	 * @param {*} $searchString 
	 * @returns 
	 */
	queryForIdentity = async ($searchString) => {
		if ($searchString=="") return;
		var aRelons = [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForIdentity", $searchString);
		lastBlock = 0;//phObj.lastBlock;

		console.log("(API.queryForIdentity)");
		var response = await this.apiSimpleQueryWithParams("queryForIdentity", {
				text:$searchString, 
				height:lastBlock
			});

		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForIdentity");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				return aRelons;
			}
		}
		return [];
	}
	


	/**
	 * Test rollback handling.
	 * REMOVE THIS IN MAINNET -- AS WELL AS ALTER THE APIKEY (if we want to use it at all!)
	 * @returns 
	 */
	popOffMocknet = async()=> {
		return await this.apiSimpleQueryWithParams("popOff",{numBlocks:1, height:0, apiKey:"aaaa"}); 
	}

	
	/**
	 * 
	 * @returns 
	 */
	mempoolOk = async()=>{
		var rez = await this.annodeDetails();
		if (rez.mempool_ok==true) return true;
		return false;
	}


	/**
	 * reset any warings.. give it a go.
	 * @returns 
	 */
	annodeDetails = async($uri_override="", $frenlyname="")=> {

		//if ($frenlyname!="") console.log("frenlyname:"+$frenlyname);
		//else console.log("no frenly name");
		
		var el = document.getElementById("current_uri");
		if (el!=null) el.classList.remove("bg-disconnect-warning");

		var rez = await this.query("annodeDetails", null, $uri_override);
		//console.log(rez);
		if (rez==null) {
			ANNE.API._bNODECONNECTED=false;
			ANNE.A.setHTML("current_uri", "?? no response", true);	
			return {};
		}

		//this was a test..
		if ($uri_override!="") {
			if (rez.status=="fail") {
				return rez;
			}
			rez.result.status="good";
			await this.processAuxAAUTHData(rez.result);
			await ANNE.setAnnodeByDetails(rez.result);
			return rez.result;
		}
		
		if (rez.status=="good") {
			if (rez.result.network=="ANNE") ANNE.setMainnet();
			else if (rez.result.network=="ANNE-TESTNET") ANNE.setTestnet();

			if ($frenlyname=="") $frenlyname = this.base_uri;
			else $frenlyname = $frenlyname+" - "+this.base_uri;

			ANNE.A.setHTML("current_uri", this.base_uri, true); 
			ANNE.API._bNODECONNECTED=true;
			await this.processAuxAAUTHData(rez.result);
			await ANNE.setAnnodeByDetails(rez.result);
			rez.result.status="good";
			return rez.result;
		}
		
		return {};
	}
	

	/** 
	 * Get outs for an nid, and types w/ filters.
	 * Note: This gets the BEM outs on the ANNODE (the entire be-line)
	 *
 	 * @param {string}
	 * @param {Boolean}
	 * @param {Array}
	 * @author Mr.Scatman
	*/
	queryForOuts_not = async($nid, $bMUSTPULL=false, $types=[4])=> {
		var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}
		
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.openSubHunt("queryForOuts_not");
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_not", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		await ANNE.IDB.queryForOuts_not(phObj, $nid, $types);

		var response = await this.apiSimpleQueryWithParams("queryForOuts_not", {
				from:$nid, 
				type:$types, 
				height:lastBlock
		});

		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOuts_not");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
				return aRelons;
			}
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return [];
	}


	getANNEBasicDemogrpahics = async($nids, $bDOLOCALONLY=false, $bEXTERNALONLY=false)=>{
		return await this.queryTosFromRelns($nids, 0, null, $bDOLOCALONLY, $bEXTERNALONLY);
        //await ANNE.API.queryTosFromRelns(MEM.tribal.nid, 0, null, bDOLOCALONLY);
	}


	/**
	 * 
	 * @param {*} $reln_nid 
	 * @param {*} $offset 
	 * @param {*} $phObj 
	 * @param {*} $bDOLOCALONLY 
	 * @param {*} $bEXTERNALONLY 
	 * @returns 
	 */
	queryTosFromRelns=async($relns, $offset=0, $phObj=null, $bDOLOCALONLY=false, $bEXTERNALONLY=false)=>{
		var nObjs = [];
		var nids = [];
		for (var r=0; r < $relns.length; r++) {
			var relNOBJ = MEMG.getOrMakeIfNew($relns[r]);
			if (relNOBJ.ghost==true) {
				continue;
			}
			nObjs.push(relNOBJ);
			nids.push(relNOBJ.nid);
		}
			
		var lastBlock = 0;
		var hash = ANNE.other.sha256(nids.join(""));
		
		var phObj = $phObj;
		if (phObj==null && $bEXTERNALONLY==false) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryTosFromRelns", hash);
			await ANNE.IDB.queryTosFromRelns(phObj, nids);
		}

		if ($bDOLOCALONLY==true) return;

		const LIMIT = 2000;

		var response = await this.apiSimpleQueryWithParams("queryTosFromRelns", {
				nids:nids, 
				height:lastBlock,
				limit:LIMIT,
				offset:$offset
		});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryTosFromRelns");
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				return await this.getTosFromRelns(nids, $offset+LIMIT, phObj);
			}
		}

		if (response.result.length && $offset==0) {
			await HISTORY.processResultsForNewPullHistory(response.result, phObj);
		}

		return;
	}


	//this will NOT get awareness relons..
    getIns = async($nid, $bMUSTPULL=false)=>{
        return await this.queryForIns($nid, $bMUSTPULL);
    }


	/**
	 * THIS WILL NOT GET AWARENESS RELONS.
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @returns 
	 */
    queryForIns = async($nid, $bMUSTPULL=false)=>{
        var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("queryForIns");

		var aRelons = [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForIns", $nid);
		lastBlock = phObj.lastBlock;

		await ANNE.IDB.queryForIns(phObj, $nid);

		if ($bMUSTPULL==true) lastBlock = 0; //phObj.lastBlock;

		var response = await this.apiSimpleQueryWithParams("queryForIns", {
				to:$nid, 
				height:0
			});

		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForIns");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();

				return aRelons;
			}
		}
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return [];
    }

	
	/** 
	 * We want to chase updated sponsored by values to prevent us drama on the annode
	 * by submitting the wrong TXs.
	 * 
	 * 
	 * @author Mr.Scatman
	 */
	queryForGhostPrep = async($nids)=> {
		var hash = ANNE.other.sha256($nids.join(""));
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForGhostPrep", hash);
		
		var nids = [];
		for (var n=0; n < $nids.length; n++) {
			var nid = $nids[n];
			var nOBJ = null;
			if ((nid instanceof anne_neuron)==true) {
				console.warn("FIX THIS, s.b. a nid, not a neuron");
				nOBJ = nid;
			}
			else nOBJ = MEMG.getOrMakeIfNew($nids[n]);
			
			if (nOBJ.ghost==true) continue;
			nids.push(nOBJ.nid);
		}

		await ANNE.IDB.queryForOutsFroms_not(phObj, nids, [4]);
		var nids2 = [];
		for (var n=0; n < nids.length; n++) {
			var nObj=MEMG.getNeuronByNid(nids[n]);
			if (nObj.sponsored_by=="" || nObj.sponsored_by==ANNE.MEM.USER._userNObj.nid) {
				nids2.push(nObj.nid);
			}
		}
		nids = nids2;
		

		if (nids.length==0) return []; //we're good now with the IDB pull.

		var aRelons = [];
		var lastBlock = phObj.lastBlock;
		
		//too many nids..
		const nid_SHARD=30;
		if (nids.length>nid_SHARD) {
			var loops = Math.ceil(nids/nid_SHARD);
			var aRelsAll = [];
			for (var n=0; n < loops; n++) {
				var nids2 = nids.slice(n*nid_SHARD, Math.min( (n*nid_SHARD)+nid_SHARD, nids.length-1) );
				var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
					froms:nids2, 
					types:$types, 
					height:lastBlock
				});
				if (response.status=="good") {
					if (response.result.length>0) {
						var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForGhostPrep");
						aRelsAll = aRelsAll.concat(aRelons);
					}
				}
			}
			if (aRelsAll.length>0) {
				await HISTORY.processResultsForNewPullHistory(aRelsAll, phObj);
				return aRelsAll;
			}
		}

		else {
			//we only need 0/1 here.
			var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
				froms:nids, 
				types:[2,3,4,5,6,7,8,9,10], 
				height:lastBlock
			});
			if (response.status=="good") {
				if (response.result.length>0) {
					var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForGhostPrep");
					await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
					return aRelons;
				}
			}
		}
		return [];
	}

	
	/** 
	 * Get outs for a list of nids, and types w/ filters.
	 * 
	 * This is helpful when wanting all the data but not relons like feelz/awareness..
	 * as we could eventually get millions of those edges on any given neuron.
	 * 
	 * @author Mr.Scatman
	*/
	queryForOutsFroms_not = async($nids, $bMUSTPULL=false, $types=[4], $bDOIDB=true, $bREBASE01=false)=> {
		if ($nids==undefined || $nids==null) return [];
		
		//use this instead.
		if ($nids.length==1) {
			return await this.queryForOuts_not($nids[0], $bMUSTPULL, $types);
		}
		
		var hash = ANNE.other.sha256($nids.join("")+$types.join(""));
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_not", hash);
		
		var nids = [];
		for (var n=0; n < $nids.length; n++) {
			var nid = $nids[n];
			var nOBJ = null;
			if ((nid instanceof anne_neuron)==true) {
				console.warn("FIX THIS, s.b. a nid, not a neuron");
				nOBJ = nid;
			}
			else nOBJ = MEMG.getOrMakeIfNew($nids[n]);
			
			if (nOBJ.ghost==true) continue;
			nids.push(nOBJ.nid);
		}

		if ($bDOIDB==true) await ANNE.IDB.queryForOutsFroms_not(phObj, nids, $types);

		//we only are using this function to chase for name and be.
		//if we now have those, don't chase them.
		if ($bREBASE01==true) {
			var nids2 = [];
			for (var n=0; n < nids.length; n++) {
				var nObj=MEMG.getNeuronByNid(nids[n]);
				if (nObj.text.length>0 && nObj.bes.length>0) {
					nids2.push(nObj.nid);
				}
			}
			nids = nids2;
		}

		if (nids.length==0) return []; //we're good now with the IDB pull.

		var aRelons = [];

		//intended for multi nids
		var lastBlock = 0;// phObj.lastBlock;
		
		//too many nids..
		const nid_SHARD=30;
		if (nids.length>nid_SHARD) {
			var loops = Math.ceil(nids/nid_SHARD);
			var aRelsAll = [];
			for (var n=0; n < loops; n++) {
				var nids2 = nids.slice(n*nid_SHARD, Math.min( (n*nid_SHARD)+nid_SHARD, nids.length-1) );
				var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
					froms:nids2, 
					types:$types, 
					height:lastBlock
				});
				if (response.status=="good") {
					if (response.result.length>0) {
						var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsFroms_not");
						aRelsAll = aRelsAll.concat(aRelons);
					}
				}
			}
			if (aRelsAll.length>0) {
				await HISTORY.processResultsForNewPullHistory(aRelsAll, phObj);
				return aRelsAll;
			}
		}

		else {
			var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
				froms:nids, 
				types:$types, 
				height:lastBlock
			});
			if (response.status=="good") {
				if (response.result.length>0) {
					var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsFroms_not");
					await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
					return aRelons;
				}
			}
		}
		return [];
	}


	/** 
	 * Get outs for an nid, and types w/ filters.
	 * Note: This gets the BEM outs on the ANNODE (the entire be-line)
	 *
 	 * @param {string} $nid - nidlongstr
	 * @param {Boolean}
	 * @param {Array} $types - passing none means all types
	 * @author Mr.Scatman
	*/
	getAllOuts = async($nid, $bMUSTPULL=false, $types=[])=> {
		if (this.opcheckerPass("getOuts"+$nid)==false) return []
		var rels = await this.queryForOuts($nid, $bMUSTPULL, $types);
		this.opcheckerTrackIt("getOuts"+$nid);
		return rels;
	}

	getOuts = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		return await this.queryForOuts($nid, $bMUSTPULL, $types);
	}

	chaseOuts = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		return await this.queryForOuts($nid, $bMUSTPULL, $types);
	}


	/**
	 * This now returns arrays of 
	 * .narrs
	 * .payments
	 * .awareness
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOuts = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		if ($nid==undefined || $nid==null) return [];
		
		var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}

		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts", $nid+""+$types.join(":"));
		lastBlock = phObj.lastBlock;

		await ANNE.IDB.queryForOuts(phObj, $nid, $types);

		if ($bMUSTPULL==true) lastBlock = 0; //phObj.lastBlock;
		
		var response = await this.apiSimpleQueryWithParams("queryForOuts", {
				from:$nid, 
				type:$types, 
				height:lastBlock
			});

		if (response.status=="good") {
			var rels = [];
			if (response.narrs.result && response.narrs.result.length>0) {
				//console.log("we have nars!")
				if (response.narrs.result) {
					rels = response.narrs.result;
				}
			}
			if (response.awareness && response.awareness.result.length>0) {
				if (response.awareness.result) {
					rels = rels.concat(response.awareness.result);
				}
			}
			if (response.payments && response.payments.result.length>0) {
				if (response.payments.result) {
					rels = rels.concat(response.payments.result);
				}
			}

			if (rels.length>0) {
				await MEMP.processNeuronsFromRelons(rels, true, true, "relons", "API.queryForOuts");
			}
				
			await HISTORY.processResultsForNewPullHistory(rels, phObj);
		}
		return [];
	}
	
	


	/**
	 * Used in emulation when we need more identities
	 * @param {*} $limit 
	 * @returns 
	 */
	getIdentities=async($limit=100, $offset=0)=>{
		return await this.getChildren(MEM.identity.nid, $limit, $offset);
	}
	
	/**
	 * We should do an IDB query here too... dagnab.
	 * @param {*} $parent_nid 
	 * @param {*} $limit 
	 * @param {*} $offset 
	 * @returns 
	 */
	getChildren=async($parent_nid, $limit=100, $offset=0, $phObj=null, $bDOLOCALONLY=false) => {
		if ($parent_nid==undefined || $parent_nid==null) return [];
		
		$limit = Math.min($limit, 2000);

		var phObj = $phObj;
		if (phObj==null) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForInstChildren", $nid);
			if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
				return [];
			}
			await ANNE.IDB.getChldren(phObj, $parent_nid, NTYPE_INST);
		}

		if ($bDOLOCALONLY==true) return;

		var response = await this.apiSimpleQueryWithParams("queryForInstChildren", {
			nid:$parent_nid,
			offset:$offset,
			limit:$limit
		});
		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getChildren");
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				return await this.getChildren($parent_nid, $limit, $offset+$limit, phObj);
			}
		}

		if (response.result.length && $offset==0) {
			await HISTORY.processResultsForNewPullHistory(response.result, phObj);
		}
		return;
	}



	/**
	 * USE THIS WISELY.  If done against a big class the return is not desirable.
	 * It should not be used to query large classes.  It is intended for instance neurons.
	 * 
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $type 
	 * @returns 
	 */
	queryForOutsAndIns = async($nid, $bMUSTPULL=false, $types=[0,1,2,3], $offset=0, $phObj=null)=> {
		if ($nid==undefined || $nid==null) return [];
		
		var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}
		if (nOBJ.ntype==NTYPE_CLASS) {
			console.warn("You are going to to query with queryForOutsAndIns with a class neuron.  NOT IDEAL.")
		}

		var lastBlock = 0;
		var phObj = $phObj;
		if (phObj==null) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOutsAndIns", $nid);
			if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
				return [];
			}
			await ANNE.IDB.queryForOutsAndIns(phObj, $nid, $types);
		}

		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		const LIMIT = 4000;

		console.log("(API.queryForOutsAndIns) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForOutsAndIns", {
				from:$nid, 
				type:$types, 
				height:lastBlock,
				limit:LIMIT,
				offset:$offset
			});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result && response.result.length) {
				await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsAndIns");
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				return await this.queryForOutsAndIns($nid, $bMUSTPULL, $types, $offset+LIMIT, phObj);
			}
		}

		if (response.result && response.result.length && $offset==0) {
			await HISTORY.processResultsForNewPullHistory(response.result, phObj);
		}
	}


	/**
	 * This can get CRAZY big.. if ur not calling it wisely.
	 * 
	 * @param {*} $nids 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOutsAndInsFroms = async($nids, $bMUSTPULL=false, $types=[0,1,2,3], $offset=0, $phObj=null, $bDOIDB=true, $rels=[])=> {
		var nids = [];
		for (var n=0; n < $nids.length; n++) {
			var nObj = MEMG.getOrMakeIfNew($nids[n]);
			if (nObj.ghost==true) continue;
			nids.push(nObj.nid);
		}

		if (nids.length==0) return [];

		var aRelons = [];
		var lastBlock = 0;

		var hash = ANNE.other.sha256(nids.join(""));
		var phObj = $phObj;
		
		//only first pass (when phObj is null) is optimizer.. 
		if (phObj==null) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOutsAndInsFroms", hash);
			if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
				console.log("super optimizer");
				return [];
			}
			if($bDOIDB==true) await ANNE.IDB.queryForOutsAndInsFroms(phObj, nids, $types);
		}
		
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		const LIMIT = 4000;
		
		var response = await this.apiSimpleQueryWithParams("queryForOutsAndInsFroms", {
				froms:nids, 
				type:$types, 
				height:lastBlock,
				limit:LIMIT,
				offset:$offset
			});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				$rels = $rels.concat(response.result);
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				aRelons = await this.queryForOutsAndInsFroms($nids, $bMUSTPULL, $types, $offset+LIMIT, phObj, $bDOIDB, $rels);
			}
		}

		if (aRelons.length>0 && $offset==0) {
			await MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "API.queryForOutsAndInsFroms");
			await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
			return aRelons;
		}

		return $rels;
	}

	/**
	 * This one does the entire BEM line for the outs (chasing up the parents and getting their outs too)
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOutsAndIns_BEM = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		var lastBlock = 0;
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("queryForOutsAndIns_BEM");

		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOutsAndIns_BEM", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
			return [];
		}

		var nObj = MEMG.getNeuronByTempNid($nid);
		var bem_nids = nObj.getBemNids();
		if (bem_nids.length) {
			await ANNE.IDB.queryForOutsAndIns_BEM(phObj, $nid, $types);
		}

		console.log("(API.queryForOutsAndIns) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForOutsAndIns_BEM", {
				from:$nid, 
				type:$types, 
				height:lastBlock
		});

		var aNewRelons = [];
		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsAndIns_BEM");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt("queryForOutsAndIns_BEM");
				return aNewRelons;
			}
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt("queryForOutsAndIns_BEM");
		return [];
	}


	/**
	 * This one does the entire BEM line for the outs (chasing up the parents and getting their outs too)
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOuts_BEM = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_BEM", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
			return [];
		}

		var nObj = MEMG.getNeuronByNid($nid);
		if (nObj==null) return [];
		
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("queryForOuts_BEM");

		var bem_nids = nObj.getBemNids();
		
		await ANNE.IDB.queryForOuts_BEM(phObj, bem_nids, $types);
		
		  console.log("(API.queryForOuts_BEM) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForOuts_BEM", {
				from:$nid, 
				type:$types, 
				height:lastBlock
			});

		var aNewRelons = [];
		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				try {
					//ANNE.TALK.printRelons(response.result);
					aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOuts_BEM");
				}
				catch(e) {
					console.log(e);
				}
				
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();

				return aNewRelons;
			}
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return [];
	}


	/**
	 * 
	 * @param {*} $last_id 
	 */
	getLatestMsgs = async($last_id)=> {
		console.log("getLatestMsgs");
		var response = await this.apiSimpleQueryWithParamsNS("getLatestMsgs", {
			last_id:$last_id
		});
		//this.processBlockCycleMsgs(response);
	}


	/**
	 * graph explore neuron query
	 * Get all outs/ins other than awareness relons
	 */
	queryForCircuitOutsAndIns = async($nid, $bMUSTPULL=false)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForCircuitOutsAndIns", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
			return [];
		}
		
		var aRelonsAudit = await ANNE.IDB.queryForCircuitOutsAndIns(phObj, $nid);
		console.log(aRelonsAudit.length);
		
		console.log("(API.queryForCircuitOutsAndIns) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForCircuitOutsAndIns", {
				from:$nid, 
				height:lastBlock
			});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForCircuitOutsAndIns");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				return aNewRelons;
			}
		}
		return [];
	}


	/**
	 */
	queryForLocationParts = async($aParts)=>{
		var response = await this.apiSimpleQueryWithParams("queryForLocationParts", {
			parts:$aParts
		});
		console.log(response);
		if (response.status=="good") {
			var relons_to_process = response.result;
			if (relons_to_process.length) {
				return await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.queryForLocationParts");
			}
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $text 
	 * @param {*} $be_nid 
	 * @param {*} $WVs 
	 * @param {*} $bMUSTPULL 
	 * @returns 
	 */
	queryForDate=async($text, $WVs=[], $bMUSTPULL=false)=>{
		console.log("(API.queryForDate) with text:"+$text);
		if (ANNE.UTILS.isset($text)==false) return [];
		var hash = ANNE.other.sha256($text+$WVs.join(""));
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForTextAndBe", hash);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		//if we have results for date, bail.
		//we're trying to resolve here only.
		var aRels = await ANNE.IDB.queryForTextAndBe(phObj, $text, MEM.date_gregorian.nid, $WVs);
		if (aRels.length>0) {
			return aNewRelons;
		}

		var response = await this.apiSimpleQueryWithParams("queryForTextAndBe", {
				text:$text, 
				class_nid:MEM.date_gregorian.nid, 
				wvs:$WVs,
				height:lastBlock
			});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForDate");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
			}
		}

		var dateNOBJ = await ANNE.MEMG.getDateNeuron($text);
		if (dateNOBJ!=null) {
			return dateNOBJ;
		}
		
		return null;
	}



	queryForMonthYear=async($text, $WVs=[], $bMUSTPULL=false)=>{
		console.log("(API.queryForMonthYear) with text:"+$text);
		if (ANNE.UTILS.isset($text)==false) return [];
		var hash = ANNE.other.sha256($text+$WVs.join(""));
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForMonthYear", hash);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		//if we have results for date, bail.
		//we're trying to resolve here only.
		var aRels = await ANNE.IDB.queryForTextAndBe(phObj, $text, MEM.monthyear.nid, $WVs);
		if (aRels.length>0) {
			return aNewRelons;
		}

		var response = await this.apiSimpleQueryWithParams("queryForTextAndBe", {
				text:$text, 
				class_nid:MEM.monthyear.nid, 
				wvs:$WVs,
				height:lastBlock
			});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForMonthYear");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
			}
		}

		//will get dates or date parts.
		var dateNOBJ = await ANNE.MEMG.getDateNeuron($text);
		if (dateNOBJ!=null) {
			return dateNOBJ;
		}
		
		return null;
	}

	
	/** 
 	 * @param {string}
	 * @param {string} - nid long string of the be class
	 * @author Mr.Scatman
	*/
	queryForTextAndBe = async($text, $be_nid, $WVs=[], $bMUSTPULL=false)=> {
		console.log("(API.queryForTextAndBe) with text:"+$text);
		if (ANNE.UTILS.isset($text)==false) return [];

		var hash = ANNE.other.sha256($text+$be_nid+$WVs.join(""));
		
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForTextAndBe", hash);
		
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		await ANNE.IDB.queryForTextAndBe(phObj, $text, $be_nid, $WVs);

		var response = await this.apiSimpleQueryWithParams("queryForTextAndBe", {
				text:$text, 
				class_nid:$be_nid, 
				wvs:$WVs,
				height:lastBlock
			});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForTextAndBe");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				return aNewRelons;
			}
		}
		return [];
	}


	/**
	 * This returns the beline object of props .nid, .text, .be_nid, .owned_by and .sponsored_by 
	 * 
	 * Here we don't know where a possible match may fall in a be line that has multiple sub-classes.
	 * Primary use case here is for ec_action.   Actions will subclass and their parent and can be more than 2 levels deep.
	 * When a user is search for an existing actions we can't look for the simple parent because the parent may 
	 * be another action.  (and will not match)
	 * @param {*} $text 
	 * @param {*} $be_nid 
	 * @param {*} $WVs 
	 * @returns 
	 */
	queryForText_BEM = async($text, $class_nid, $wvs=[])=> {
		console.log("(API.queryForText_BEM) with text:"+$text);

		//do we need to do this here?
		//var hash = ANNE.other.sha256($text+$class_nid+$wvs.join(""));
		//var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForText_BEM", hash);
		//await ANNE.IDB.queryForText_BEM(phObj, $text, $class_nid, $wvs);
		
		var response = await this.apiSimpleQueryWithParams("queryForText_BEM", {
				text:$text, 
				class_nid:$class_nid, 
				wvs:$wvs
			});

		var beline_objs = [];
		if (response.status=="good") {
			beline_objs = response.result;
			if (beline_objs.length) {
				//if we do this again in any other function, we can build a processor function in the MEMP.
				for (var b=0; b < beline_objs.length; b++) {
					var bemObj = beline_objs[b];
					var nOBJ = MEMG.getNeuronByNid(bemObj.nid.trim());
					if (nOBJ==null) {
						var nOBJ = new anne_neuron(bemObj.text, parseInt(bemObj.ntype), bemObj.nid);
						if (bemObj.owned_by!="") {
							nOBJ.addOwnedBy(bemObj.owned_by);
						}
						if (bemObj.sponsored_by!="") {
							nOBJ.addSponsoredBy(bemObj.sponsored_by);
						}
					}
					else {
						nOBJ.text = bemObj.text;
						bemObj.ntype = parseInt(bemObj.ntype);
					}
				}
			}
		}
		return beline_objs;
	}


	//because i have a hard time remembering things.
	queryNidBasic = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	queryNeuron = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	lookupNeuron = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	getNeuronByNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	getNeuron = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	getNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	queryNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	

	/**
	 * Use to find get relons of to neurons a from neuron connects to.
	 * ie, service terms.
	 * @param {*} $nid 
	 * @param {*} $reln 
	 * @returns 
	 */
	getAllToNeuronRelonsGeneric = async($from_nid, $reln)=> {
		var fromNOBJ = ANNE.MEMG.getNeuronByNid($from_nid);
		var relnNOBJ = ANNE.MEMG.getNeuronByNid($reln);
		console.log("(getAllToNeuronRelonsGeneric) from "+fromNOBJ.text+" with reln:"+relnNOBJ.text);
		var response = await this.apiSimpleQueryWithParams("getAllToNeuronRelonsGeneric", {
			from_nid:$from_nid, 
			reln:$reln
		});
		if (response.status=="good" && response.result.length) {
			return await this.returnNOBJS(response.result, "relons", "API.getAllToNeuronRelonsGeneric");
		}
		return [];
	}

	getAllToNeuronRelonsGeneric2 = async($nids, $reln)=> {
		if ($nids.length==1) return await this.getAllToNeuronRelonsGeneric($nids[0], $reln);
		var response = await this.apiSimpleQueryWithParams("getAllToNeuronRelonsGeneric2", {
			nids:$nids, 
			reln:$reln
		});
		if (response.status=="good" && response.result.length) {
			return await this.returnNOBJS(response.result, "relons", "API.getAllToNeuronRelonsGeneric");
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $reln 
	 * @param {*} $to_s 
	 * @returns 
	 */
	getAllFromNeuronRelonsGeneric = async($reln, $to_s)=> {
		 console.log("(API.getAllFromNeuronRelonsGeneric) with to_s:"+$to_s+" reln:"+$reln);
		var response = await this.apiSimpleQueryWithParams("getAllFromNeuronRelonsGeneric", {
			reln:$reln,
			to_nid:$to_s
		});

		if (response.status=="good" && response.result.length) {
			return await this.returnNOBJS(response.result, "relons", "API.getAllFromNeuronRelonsGeneric");
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $relons 
	 * @param {*} $tbl 
	 * @param {*} $msg 
	 * @returns 
	 */
	returnNOBJS=async($relons, $tbl="relons", $msg="")=> {
		if ($relons==undefined || $relons==null || $relons.length==0) return [];
		var froms = [];
		for (var r=0; r < $relons.length; r++) {
			var rel = $relons[r];
			if (froms.indexOf(rel.from_nid)==-1) froms.push(rel.from_nid);
		}
		await MEMP.processNeuronsFromRelons($relons, true, true, $tbl, $msg);
		return ANNE.MEMG.getNObjsFromList(froms);
	}


	/**
	 * Process these ones as they come in..
	 * We can monitor if it is better to concat the relons and process into our local MEM in one batch
	 * or to do them as they come in the shards...
	 * @param {*} $to_nids 
	 * @returns 
	 */
	getAllInsToTheseNids = async($to_nids, $SHARD_SIZE=30)=> {
		 
		var SHARD=$SHARD_SIZE;
		var handled_nids = [];
		var to_nids = [];

		console.log("(API.getAllInsToTheseNids) with $to_nids:"+$to_nids+" shard size:"+$SHARD_SIZE);
		console.time("getAllInsToTheseNids");

		//loop into shards to reduce strain on annode w/ too many nids in list search.
		for (var n=0; n<$to_nids.length; n++) {
			var nid = $to_nids[n];
			if (handled_nids.indexOf(nid)==-1) {
				handled_nids.push(nid);
				to_nids.push(nid);
			}
			else continue;

			if (to_nids.length==SHARD) {
				var response = await this.apiSimpleQueryWithParams("getAllInsToTheseNids", {
					to_nids:to_nids
				});
				if (response.status=="good" && response.result.length) {
					await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getAllInsToTheseNids");
				}
				to_nids = [];
			}
		}

		//get any left overs that didn't hit a shard limit (the last bits)
		if (to_nids.length) {
			var response = await this.apiSimpleQueryWithParams("getAllFromNeuronRelonsGeneric", {
				to_nids:to_nids
			});
			if (response.status=="good" && response.result.length) {
				await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getAllFromNeuronRelonsGeneric");
			}
			to_nids = [];
		}

		console.timeEnd("getAllFromNeuronRelonsGeneric");

		return;
	}



	/**
	 * Lookup a nid (neuron id) to get relons of types.	
	 * no height filter on this??? why??
	 * 
	 * @param {string} - The longstring of the neuron id.
	 * @param {array} - Array of types to filter to
	 */
	 lookupNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		 console.log("(API.lookupNid) with nid:"+$nid);
		
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNid", $nid+$types.join(""));
		await ANNE.IDB.lookupNid(phObj, $nid, $types);

		var response = await this.apiSimpleQueryWithParams("lookupNid", {
			nid:$nid, 
			types:$types,
			dopubkey:$bDOPUBKEY
		});

		if (response.status=="good" && response.result.length) {
			return await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.lookupNid :"+$nid+" types:"+$types.join("/"));
		}

		return [];
	}


	/**
	 * 
	 * @param {*} $nids 
	 * @param {*} $types 
	 * @param {*} $sysMsgRelns 
	 * @returns 
	 */
	lookupForNidsAndSysMsgRelns = async($nids, $types=[0,1,2], $sysMsgRelns=[])=> {
		
		var hash = ANNE.other.sha256($nids+$types.join()+$sysMsgRelns.join(""));
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupForNidsAndSysMsgRelns", hash);
		
		await ANNE.IDB.lookupForNidsAndSysMsgRelns(phObj, $nids, $types, $sysMsgRelns);
		
		var response = await this.apiSimpleQueryWithParams("lookupNidsAndSysMsgRelns", {nids:$nids, types:$types, relns:$sysMsgRelns});
		console.log(response);

		if (response.status=="good" && response.result.length>0) {
			return await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.lookupForNidsAndSysMsgRelns");
		}
		return [];
	}

	getNames =async($nids)=>{
		return this.lookupNids($nids, [0,1], true, false);
	}

	/**
	 * Lookup a list of ANNE Ids. 
	 * We have missing info and need to get the minimums to display a prop	
	 * or neuron.  If it is ever clicked on we will chase it out a bit more.
	 * Once this is done once, it
	 * will not need to be done again as long as
	 * the user has the data in their indexDB.
	 * @param {Array}
	 * @param {Array}
	 * @author Mr.Scatman								
	 */
	lookupNids = async($nids, $types=[0,1,2], $bDOQUERY=true, $bDOLOADER=true, $from="", $fixAfterIDB=false)=> {
		console.log("%clookupNids nid cnt:"+$nids.length+" %c"+$from, ANNE.TALK.funct, ANNE.TALK.bgred, ANNE.TALK.msg)
		const SHARD = 30;

		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", $nids.join("")+$types.join(""));
		
		await ANNE.IDB.lookupNids(phObj, $nids, $types);

		if ($bDOQUERY==true) {
			var nids = [];
			var rezRelons = [];
			for (var n=0; n < $nids.length; n++) {
				
				if ($fixAfterIDB==true) {
					var NOBJ = ANNE.MEMG.getNeuronByNid($nids[n]);
					if (NOBJ.text!="" && NOBJ.bes.length>0) continue;
				}
				
				nids.push($nids[n]);	
				if (nids.length==SHARD) {
					var response = await this.apiSimpleQueryWithParams("lookupNids", {nids:nids, types:$types});
					if (response.status=="good" && response.result.length) {
						if (rezRelons.length==0) rezRelons = response.result;
						else rezRelons = rezRelons.concat(response.result);
					}
					nids = [];
				}
			}

			if (nids.length) {
				var response = await this.apiSimpleQueryWithParams("lookupNids", {nids:nids, types:$types}, $bDOLOADER);
				if (response.status=="good" && response.result.length) {
					if (rezRelons.length==0) rezRelons = response.result;
					else rezRelons = rezRelons.concat(response.result);
				}
				nids = [];
			}
			
			if (rezRelons.length) {
				return await MEMP.processNeuronsFromRelons(rezRelons, true, true, "relons", "API.lookupNids");
			}	
		}

		return [];
	}



	/**
	 * limit is a node param used with the offset. (where did it go?  LOLZ) 
	 * @param {*} $nids 
	 * @param {*} $offset 
	 * @returns 
	 */
	lookupNidsGetAllOuts = async($nids, $offset=0, $bDOIDB=true)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNidsGetAllOuts ", $nids);
		lastBlock = phObj.lastBlock;
	
		//we get them all on offset=0
		if ($offset==0 && $bDOIDB==true) {
			await ANNE.IDB.lookupNidsGetAllOuts(phObj, $nids);
		}
		
		var response = await this.apiSimpleQueryWithParams("lookupNidsGetAllOuts", {
			nids:$nids,
			height:lastBlock,
			offset:$offset
		});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.lookupNidsGetAllOuts");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				return aNewRelons;
			}
		}
	}


	/**
	 * lookup max 30 txids at a time.
	 */
	lookupTxids = async ($txids)=> {

		await IDB.lookupTxids($txids);
		
		var txids = [];
		var shard=100;	//is this too big?  (100 * int(20)+ a few extras?)
		var results = [];
		for (var t=0; t < $txids.length; t++) {
			txids.push($txids[t]);
			if (txids.length==shard) {
				var response = await this.apiSimpleQueryWithParams("lookupTxids", 
					{
						txids:txids
					});
				if (response.status=="good") {
					results = results.concat(response.result);
				}
				txids=[];
			}
		}

		//left overs
		if (txids.length>0) {
			var response = await this.apiSimpleQueryWithParams("lookupTxids", 
				{
					txids:txids
				});
			if (response.status=="good") {
				results = results.concat(response.result);
			}
			txids=[];
		}

		var aRelons = [];
		if (results.length) {
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.lookupTxids");
		}
		return aRelons;					
	}
	

	/**
	 * 
	 * @param {*} $nids 
	 * @param {*} $types 
	 * @param {*} $relns 
	 * @param {*} $height 
	 * @returns 
	 */
	participantChase = async($nids, $types=[0,1,2], $relns=[], $height=0)=> {
		var response = await this.apiSimpleQueryWithParams("participantChase", {nids:$nids, types:$types, relns:$relns, height:$height});
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.participantChase");
		}
		return aRelons;
	}


	javscriptErrorReporting = async($message, $source, $lineno, $colno, $error)=> {
		var response = await this.apiSimpleQueryWithParams("javscriptErrorReporting", 
			{
				message:$message, 
				source:$source, 
				lineno:$lineno, 
				colno:$colno,
				error:$error
			});
		console.log(response);
		return;
	}

		
	/**
	 * NOTE: not in use yet in the EXT.
	 * 
	 * Query for delete
	 * Do a super query on every spot we see this in from, reln, to.
	 * This helps us understand the impact of a delete (necrosis).
	 * 
	 * @param {string} - nid we want to potentially dlete
	 * @author Mr.Scatman							
	 */
	 queryForDelete = async($nid)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForDelete ", $nid);
		lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("queryForDelete", {nid:$nid, height:lastBlock});

		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.queryForDelete");
		}
		return aRelons;
	}


	/**			
	 * @param {int}
	 * @param {string}
	 * @param {string}
	 * @param {int} 
	 * @param {int}
	 * 
	 * @author Mr.Scatman								
	 */
	 queryForInsWithTo = async($type, $reln, $to, $to_type, $offset=0)=> {
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForInsWithTo ", ANNE.other.sha256($type+$reln+$to+$to_type+$offset));
		var lastBlock = phObj.lastBlock;
		
		if ($offset==0) {
			await ANNE.IDB.queryForInsWithTo(phObj, $type, $reln, $to, $to_type);
		}
		
		var response = await this.apiSimpleQueryWithParams("queryForInsWithTo", 
			{
				type:$type, 
				reln:$reln,
				to:$to,
				to_type:$to_type,
				offset:$offset,
				height:lastBlock

			});
		
		if (response.status=="good") {
			if (aRelons.length) {
				return await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForInsWithTo");
			}
		}
		return [];
	}


	/**
	 * Delete an lnot.
	 * We must prove who we are to the ANNODE in order for the ANNODE to honor our request.
	 * @param {*} $txid 
	 */
	deleteLNOT=async($txid, $el_error=null)=> {
		var encrypted_msg = ANNE.USER.genGeneralProveIt($txid);
		console.log(encrypted_msg);
		var response = await this.apiSimpleQueryWithParams("sendMsgForSesh", {
			claimed_sender_nid:USER._userNObj.nid,
			encrypted_msg:encrypted_msg
		});
		console.log(response);
		if (response.status=="good") {
			var next_proof_id = response.next_proof_id;
			if (next_proof_id && next_proof_id!=-1) {
				await this.updateAndSaveNextProofId(next_proof_id);
			}
			return;
        }
		else {
			if (response.reason!=undefined) {
				console.log("%c "+response.reason, TALK.bgbigred);
				if ($el_error!=null) ANNE.A.setHTML($el_error, response.reason);
			}
			else if ($el_error!=null) ANNE.A.setHTML($el_error, "unknown error");
			return [];
		}
	}



	/**
	 * TODO: get this to work after timeouts or disconnects.
	 * 
	 * Get life info from node.
	 * 
	 * We will use this when we get to registering public ANNODES which are 
	 * competing for public use and service fees.
	 * 
	 * Client-side apps will ping the registry of ANNODES to see if the one they
	 * want to handshake with is alive.
	 */
	ping = async($node_uri_override="")=> {
		try {
			var response = await this.query("ping", null, $node_uri_override);
			if (response.status!=undefined && response.status!="good") {
				console.log(response);
			}
			return response;
		} 
		catch (e) {
			console.log('node not alive', e);
			return {status:"fail", reason:e};
		}
	}


	/**
	 * A willy-wonka CORS workaround, but hey, I like chocolate.  (actually, I don't)
	 * @param {*} $formatted_name 
	 * @returns 
	 */
	avoidCorsWIKISPOOKS = async($formatted_name="")=> {
		if ($formatted_name=="" || $formatted_name.indexOf(" ")!==-1) {
			return {status:"fail", reason:"must provide a wikispooks formatted name"};
		}
		try {
			var response = await this.apiSimpleQueryWithParams("avoidCorsWIKISPOOKS", {formatted_name:$formatted_name});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}
	

	/**
	 * 
	 * @param {*} $formatted_name 
	 * @returns 
	 */
	avoidCorsDBPEDIA = async($formatted_name="")=> {
		if ($formatted_name=="" || $formatted_name.indexOf(" ")!==-1) {
			return {status:"fail", reason:"must provide a dbpedia formatted name"};
		}
		try {
			var response = await this.apiSimpleQueryWithParams("avoidCorsDBPEDIA", {formatted_name:$formatted_name});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}


	/**
	 * 
	 * @param {*} $relative 
	 * @returns 
	 */
	loadLocalOnlyFile=async($relative)=> {
		let FILERESP =  await fetch($relative).catch(function($e){
			console.log("error:"+$e);
			return;
		});
		//console.log(FILERESP);
		if (FILERESP==undefined) return null; 
		let blob = await FILERESP.blob();
		var response = {};
		response.mimetype2 = blob.type;
		
		return new Promise((resolve, reject) => {
			let reader = new FileReader();
			reader.onloadend = function() {
				let b64data = reader.result;
				response.b64data = b64data;
				return resolve(response);
			};
			reader.readAsDataURL(blob);
		});
	}


	/**
	 * Extract actual src from page..
	 * get image.. store on annode, for later upgrade to real folder.
	 * @param {*} $url 
	 * @param {*} $ext
	 * @param {*} $baseurl  
	 * @param {*} $elid 
	 * @returns 
	 */
	getAndCacheFileFromElement = async($url, $ext, $baseurl="", $elid="")=>{
		if ($url=="") {
			return {status:"fail", reason:"must provide a url"};
		}
		$url = $url.replaceAll(" ", "_");
		try {
			var response = await this.apiSimpleQueryWithParams("getAndCacheFileFromElement",
						 {url:$url, ext:$ext, baseurl:$baseurl, elid:$elid});
			
			if (response.status=="good" && response.finger!=undefined) {
				
				//the files are in the html/ui/  a2a_cache... || a2a_neurons...  
				//let's go get it directly from our folder.
				var relativepath = response.relativepath;
				//ANNE.API.base_uri+
				let FILERESP =  await fetch(relativepath+"/"+response.filename).catch(function($e){
					console.log("error:"+e);
					return;
				}
				);
		
				 console.log(FILERESP);
				if (FILERESP==undefined) return response; 
				
				let blob = await FILERESP.blob();
				response.mimetype2 = blob.type;

				return new Promise((resolve, reject) => {
					let reader = new FileReader();
					reader.onloadend = function() {
						let b64data = reader.result;
						response.b64data = b64data;
						return resolve(response);
					};
					reader.readAsDataURL(blob);
				});
				
			}
			else {
				console.log("bad??  why?");
				console.log(response);
				return response;
			}
		} 

		catch (e) {
			return {status:"fail", reason:e};
		}
	}



	/**
	 * Get file by nid.
	 * If annode permits, an antor will be triggered if this file is not yet on THIS annode.
	 */
	getFileByNid = async($nid="")=>{
		if ($nid=="") {
			return {status:"fail", reason:"must provide a nid"};
		}
		try {
			var fileserve_uri = ANNE.API.base_uri+"/fileserve/nid/"+$nid;
			let FILERESP =  await fetch(fileserve_uri).catch(function($e) {
				console.log("catch error:"+e);
				return {status:FILEDATASTATUS_REQUEST_ERROR};
			}
			);
	
			if (FILERESP==undefined) {
				console.lpg("no FILERESPONSE")
				return {status:FILEDATASTATUS_REQUEST_ERROR}; 
			}

			if (FILERESP.status==FILEDATASTATUS_ANTOR_OFF) {
				console.log("ANTOR IS OFF");
				return FILERESP;
			}

			else if (FILERESP.status==FILEDATASTATUS_PENDING_ANTOR) {
				console.log("pending antor!");
				return FILERESP;
			}

			else if (FILERESP.status==FILEDATASTATUS_CANNOT_INIT_ANTOR) {
				console.log("cannot init antor");
				return FILERESP;
			}

			console.log(FILERESP);

			let blob = await FILERESP.blob();
			var response = {};
			response.mimetype2 = blob.type;
			
			return new Promise((resolve, reject) => {
				let reader = new FileReader();
				reader.onloadend = function() {
					let b64data = reader.result;
					response.b64data = b64data;
					response.status=FILEDATASTATUS_ACQUIRED;
					return resolve(response);
				};
				reader.readAsDataURL(blob);
			});
		} 
		catch (e) {
			return {status:FILEDATASTATUS_REQUEST_ERROR, reason:e};
		}
	}


	/**
	 * @param {*} $url 
	 * @param {*} $ext 
	 * @param {*} $nid 
	 * @param {*} $bDOTHUMB - return a small img thumb instead of the full one
	 * @returns 
	 */
	getAndANNODECacheTheFile = async($url, $ext="", $nid="", $bDOTHUMB=true)=>{
		if ($url=="") {
			return {status:"fail", reason:"must provide a url"};
		}

		if ($ext==null || $ext=="") {
			alert("can't work with this!  MUST HAVE EXTENSION!");
			return;
		}

		try {
			var response = await this.apiSimpleQueryWithParams("getAndCacheFile", {url:$url, ext:$ext, nid:$nid});
			if (response.status=="good" && response.finger!=undefined) {
				
				var relativepath = response.relativepath;
				//console.log("relative path:"+relativepath);

				var fileserve_uri = "";
				if ($bDOTHUMB==true) {
					var parts = response.filename.split(".");
					parts[ parts.length-2] = parts[ parts.length-2]+"_thumb";
					var filename = parts.join(".");
					fileserve_uri = ANNE.API.base_uri+"/fileserve/"+relativepath+"/"+filename;
				}
				else fileserve_uri = ANNE.API.base_uri+"/fileserve/"+relativepath+"/"+response.filename;

				//console.log("using fileserve_uri: "+fileserve_uri);

				let FILERESP =  await fetch(fileserve_uri).catch(function($e){
					console.log("error:"+e);
					return;
				}
				);
		
				//console.log(FILERESP);
				if (FILERESP==undefined) return response; 
				
				let blob = await FILERESP.blob();
				response.mimetype2 = blob.type;
				
				return new Promise((resolve, reject) => {
					let reader = new FileReader();
					reader.onloadend = function() {
						let b64data = reader.result;
						response.b64data = b64data;
						return resolve(response);
					};
					reader.readAsDataURL(blob);
				});
				
			}
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}
	

	avoidCorsWikimediaImg = async($name="bill_clinton.png", $width=200)=> {
		if ($name=="") {
			return {status:"fail", reason:"must provide a wikimedia name"};
		}
		try {
			var response = await this.apiSimpleQueryWithParams("avoidCorsWikimediaImg", {img_name:$name, img_width:$width});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}
	

	/**
	 * 
	 * @param {*} $txid 
	 * @param {*} $nid 
	 * @param {*} $type 
	 * @param {*} $reln 
	 * @param {*} $to_type 
	 * @param {*} $to_s 
	 * @returns 
	 */
	saveLocalRelon = async($txid, $nid, $type, $reln, $to_type, $to_s)=> {
		if ($txid=="" || $nid=="" || $type=="" || $reln=="" || $to_type=="" || $to_s=="") {
			return;
		}
		try {
			var response = await this.apiSimpleQueryWithParams("saveLocalRelon", 
				{txid:$txid, from:$nid, type:$type, reln:$reln, to_type:$to_type, to:$to_s});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}


	/**
	 * 
	 * @returns 
	 */
	getPrivateFolderAbsolute = async()=>{
		try {
			var response = await this.apiSimpleQueryWithParams("getPrivateFolderAbsolute");
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}



	/**
	 * 
	 * @param {*} $nid 
	 * @param {*} $b64data 
	 * @param {*} $ext 
	 * @returns 
	 */
	sendB64Data = async($nid, $b64data, $ext)=> {
		try {
			var response = await this.apiSimpleQueryWithParams("sendB64Data", 
					{nid:$nid, b64data:$b64data, ext:$ext});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}

}


/************** (from file ./js/network/API.js)**/ 

class history_slim  {
    constructor(){
        //tracks our space-wide FULL PULLS by the most recent block id in the results.
        this._aHistory = []; 

        //if we ever want to load memory from a snapshot or post-boot
        //we can set this number higher than 0 to optimize.
        this._ANNE_EARLIEST_BLOCK = 0;  

        //just a helpful heads up to us.
        this._GREATEST_LAST_KNOWN_BLOCK = this._ANNE_EARLIEST_BLOCK;
        this._GLKB_SOLID = 0;
        
        //have to make my own.  cant get the built in auto key figured out (don't want to waste any more time)
        this.pullid = 1;    

        this.bhid = "chain_height";

        this.update_h_cb = null;
    }

    setBlockUpdateCb=function($cb) {
        this.update_h_cb = $cb;
    }

    getGLKB = function() {
        if (this._GLKB_SOLID > this._GREATEST_LAST_KNOWN_BLOCK) return this._GLKB_SOLID;
        return this._GREATEST_LAST_KNOWN_BLOCK;
    }

    updateGLKB_SOLID = function($b) {
        this._GREATEST_LAST_KNOWN_BLOCK = $b;
        this._GLKB_SOLID = $b;
        ANNE.A.setHTML(this.bhid, "#"+ANNE.UTILS.commas($b));
        if (ANNE._annodeLS!=null && ANNE._annodesLS.network!=null) {
            localStorage.setItem("last_height_"+ANNE._annodesLS.network, $b);
        }
        if (this.update_h_cb!=null) {
            this.update_h_cb($b);
        }
    }
    updateGLKB_fromRelons = function($b) {
        if ($b<=this._GREATEST_LAST_KNOWN_BLOCK) return;
        this._GREATEST_LAST_KNOWN_BLOCK = $b;   
        ANNE.A.setHTML(this.bhid, "#"+ANNE.UTILS.commas($b));
        if (ANNE._annodeLS!=null && ANNE._annodesLS.network!=null) {
            localStorage.setItem("last_height_"+ANNE._annodesLS.network, $b);
        }
    }

    getNextPullID = function () {
        var pullid = HISTORY.pullid;
        HISTORY.pullid +=1;
        return pullid;
    }


    /**
     * Process the pull history (optimizer).
     * 
     * We store this in a DB in localStorage                                             
     * 
     * @author {Mr.Scatman}
     */
    processPullHistory = function ($aFullPullHistory) {
        var maxpullid = 0;
        if ($aFullPullHistory!=undefined && $aFullPullHistory!=null) {
            HISTORY._aHistory = $aFullPullHistory;
            for (var h=0; h < HISTORY._aHistory.length; h++) {
                var hObj = HISTORY._aHistory[h];
                if (hObj.pullid > maxpullid) maxpullid = hObj.pullid;
                hObj.bDIDIDB = false;
            }
        }

        HISTORY.pullid = maxpullid + 1;
    }
    

    /**
     * Get the last block we found results using this exact query.
     * 
     * Try to save processing/pull attempts by looking for this pull obj.  
     * We will still try to pull from the blockchain but with a min block  
     * number from the last known spot we already pulled in data for this  
     * type of pull.                                                       
     * 
     * @author {Mr.Scatman}
     */
    getLastBlockFromPullHistory = function($prefix, $functionName, $target, $extra=null) {
        var extra="";
        if ($extra!=null) extra=$extra;

        if (HISTORY._aHistory.length==0) {
            //console.log("%c NO HISTORY ---- U SHOULD SEE THIS ONCE for "+$functionName+" target:"+$target+" "+$extra, 'background: white; color: red');
            return HISTORY.__makeDefaultPullHistoryObj($prefix, $functionName, $target, $extra);
        }

        for (var h=0; h < HISTORY._aHistory.length; h++) {
            var phObj = HISTORY._aHistory[h];
            if (phObj.user_nid!=USER._userNObj.nid) continue;
            if (phObj.prefix!=$prefix || phObj.functionName!=$functionName || phObj.target!=$target) continue;
            if (extra!="") {
                if (phObj.extra!=$extra) {
                    continue;
                }
            }
            //console.log("%c history - we found a pull history object for function: "+$functionName+" target:"+$target, 'background: green; color: white');
            //console.log(phObj);
            return phObj;
        }

        //we had history in our  history object, but not for THIS entry.
        //console.warn("temp new HISTORY object (will save if we find results) for ", $prefix, $functionName, $target);
        return HISTORY.__makeDefaultPullHistoryObj($prefix, $functionName, $target, $extra);
    }




    /**
     * Update our history trackers for latest goodies found.
     * 
     * Look for more recent .time (block int number).   Update existing 
     * tracker records or push a new one. 
     * block index is only relevant within the scope of the searrch, not 
     * for the highest block we've seen elsewhere. 
     * 
     * @author {Mr.Scatman}
     * 
     */ 
    processResultsForNewPullHistory=async($arr, $phObj)=> {
        var localMaxBlock = $phObj.lastBlock;     
        var bUPDATINGEXISTINGRECORD = false;

        if ($arr==undefined || $arr==null) {
            return;
        }
        
        //process results for max time. 
        for (var r=0; r  < $arr.length; r++) {
            var obj = $arr[r];
            if (obj.time > localMaxBlock) localMaxBlock = obj.time;
        }
        
        //greatest of either the localmax (observed, or lastrun)
        //and the greatest last known block
        localMaxBlock = Math.max(localMaxBlock, HISTORY._GREATEST_LAST_KNOWN_BLOCK);
        
        //our results have a greater block than our previous history and we have an existing pull history object
        //we want to update this record with the new block. 
        if (localMaxBlock > $phObj.lastBlock  && $phObj.pullid>0) {
            bUPDATINGEXISTINGRECORD=true;
            if(localMaxBlock>=1500000000) {
                console.log("%c HISTORY localMAX BLOCK TOO BIG THIS SHOULD NEVER HAPPEN!!!", 'background: red; color: white');
            }
            else $phObj.lastBlock = localMaxBlock; 
        }    

        //this is a first time new entry, not new info.
        else if ($phObj.pullid==0) {
            bUPDATINGEXISTINGRECORD = false;
            if(localMaxBlock>=1500000000) {
                console.log("%c HISTORY localMAX BLOCK TOO BIG THIS SHOULD NEVER HAPPEN!!!", 'background: red; color: white');
            }
            else $phObj.lastBlock = localMaxBlock;
        }

        //assign it a real ID (next in line)
        if ($phObj.pullid==0) {
            $phObj.lastBlock = $phObj.lastBlock+1;
            $phObj.pullid = HISTORY.getNextPullID();    //this accepts the existing pullid and then incriments it by one
             console.log("%c ((inserting new history record with pullid:"+$phObj.pullid+" "+$phObj.functionName+" target:"+$phObj.functionName+"))", 'background: white; color: green');
            var done = await DB.___update($phObj, "fullPullHistoryANNE");    //insert a new record
        }

        //if we had new info, update it.
        else if (bUPDATINGEXISTINGRECORD) {
            $phObj.lastBlock = $phObj.lastBlock+1;
            console.log("%c ((updating existing record with new data))" , 'background: green; color: white');
            var done = DB.___update($phObj, "fullPullHistoryANNE");    //update the localDB record
        }

        //existing pull history object and no new results.
        //else console.warn("((no new info from block > "+$phObj.lastBlock+" info to add to pull history))")
        return;
    }


    /**
     * @author {Mr.Scatman}
     */ 
    __makeDefaultPullHistoryObj=function($prefix, $functionName, $target, $extra=null) {
        var phObj = new Object();
        phObj.user_nid = USER._userNObj.nid;
        phObj.prefix = $prefix;
        phObj.functionName = $functionName;
        phObj.target = $target;
        phObj.extra = $extra;
        phObj.lastBlock = 0;//HISTORY._GREATEST_LAST_KNOWN_BLOCK;
        phObj.pullid = 0;  //we will set this to the next one in line if we are going to insert the record
        phObj.bDONETHISSESSION = false;
        phObj.bDIDIDB = false;          //for RAM optimization.
        HISTORY._aHistory.push(phObj);
        return phObj;
    }
}

const HISTORY = new history_slim ();

/************** (from file ./js/db/history.js)**/ 


const LSM_DIRECT = 1;
const LSM_RELATIVE = 2;

//location searches are not normal searches.
class loc_search_pool {
    constructor($str, $constituent_of_nOBJ, $altBeNid=null, $fireThresh=1) {
        this.str = $str;
        this.constituent_of_nOBJ = $constituent_of_nOBJ;
        this.altBeNid = $altBeNid;
        this.fireThresh = $fireThresh;
        this.search_matches = [];
        this.direct_full_matches = 0;
        this.direct_partial_matches = 0;
        this.relative_full_matches = 0;
        this.relative_partial_matches = 0;
    }


    /**
     * 
     * Add a match to the search match pool
     * @param {anne_neuron} $nObj 
     * @param {Number} $fire 
     * @param {int} $co - constiuent of match type (LSM_DIRECT, as requested  or LSM_RELATIVE, a near neighbor)
     * @param {int} $co_tnid - if LSM_RELATIVE, the tnid of the relative constiuent_of we found.
     * @returns {loc_search_match}
     */
    addMatch = function($nObj, $fire=1, $co=0, $co_tnid=null) {
        if ($co==LSM_DIRECT) {
            if ($fire==1) this.direct_full_matches+=1;
            else this.direct_partial_matches+=1;
        }
        else {
            $fire = $fire*.9;
            if ($fire==1) this.relative_full_matches+=1;
            else this.relative_partial_matches+=1;
        }

        var lsm  = new loc_search_match($nObj, $fire, $co, $co_tnid);
        this.search_matches.push(lsm);
        return lsm;
    }



    /**
     * 
     * @param {*} $bSORT 
     * @returns 
     */
    getMatches = function($bSORT=true) {
        if (this.search_matches.length==0) return this.search_matches;
        if ($bSORT==true) {
            this.search_matches.sort(function($a, $b) {
                if ($a.fire>$b.fire) return -1;
                else if ($a.fire<$b.fire) return 1;
                return 0;
            });
        }
        return this.search_matches; 
    }
}


/**
 * 
 */
class loc_search_match {
    constructor($nObj, $fire=1, $co=0, $co_tnid=null) {
        this.nObj = $nObj;
        this.fire = $fire;
        this.co = $co;
        this.co_tnid = $co_tnid;
    }
}


class anne_chase extends anne_singleton {
    constructor(){
        super(false);
	}

    /**
     * we want be class and text info for anything this nObj is connecting to.
     * @param {*} $nObj 
     * @returns 
     */
    oneGenQueryable=async($nObj, $addTypes=[])=>{
        var queryable_nids = [];
        
        if ($nObj.bes.length==0 || $nObj.text=="") queryable_nids.push($nObj.nid);

        for (var r=0; r < $nObj.outs.length; r++) {
            if ($nObj.outs[r].to_type!=TO_TYPE_ANNE) continue;
            var toNOBJ = MEMG.getNeuronByNid($nObj.outs[r].to_s);
            if (toNOBJ==null) continue;
            if (toNOBJ.ghost==true) continue;
            if (toNOBJ.text=="" || toNOBJ.bes.length==0) {
                if (queryable_nids.indexOf(toNOBJ.nid)==-1) queryable_nids.push(toNOBJ.nid);
            }
        }

        var types = [0,1];
        if ($addTypes.length>0) {
            for (var t=0; t < $addTypes.length; t++) {
                if (types.indexOf($addTypes[t]==-1)) types.push($addTypes[t]);
            }
        }

        //we don't want to pull all in names -- here..
        //for classes this could be crazy huge.. like..
        //millions, for "person" class.
        if ($nObj.ntype==NTYPE_INST) {
            for (var r=0; r < $nObj.ins.length; r++) {
                if ($nObj.ins[r].to_type!=TO_TYPE_ANNE) continue;
                var fromNOBJ = MEMG.getNeuronByNid($nObj.ins[r].from_nid);
                if (fromNOBJ==null) continue;
                if (fromNOBJ.ghost==true) continue;
                if (fromNOBJ.text=="" || fromNOBJ.bes.length==0) {
					if (queryable_nids.indexOf(fromNOBJ.nid)==-1) queryable_nids.push(fromNOBJ.nid);
                }
            }
        }

        if (queryable_nids.length>0) {
            await ANNE.CHASE.lookupMinList(queryable_nids, types, "onegenQ", true);
        }
    }


    /**
     * Get all outs not of types
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOuts_not = async($nObj, $bMUSTPULL=false, $types=[], $debugJsonEl="", $callback=null)=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
        }

        var nid = $nObj.nid;
        
        if ($types.length) {
            var aRelons = await ANNE.API.queryForOuts_not(nid, $bMUSTPULL, $types);
        }
        else {
            //alert("optimizer skip query");
            console.log("%c queryForOuts %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);

        if($callback!=null) $callback($nObj.outs);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    queryMinBem = async($nObj) =>{
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
        }
        return await ANNE.API.queryForOuts_BEM($nObj.nid, true, [0,1]);
        
    }

    /**
     * note:  this does no filtering for garden/worldview. 
     * @param {anne_neuron}  - Neuron Object to query
     * @param {Boolean} - Require pull despite flags
     * @param {Array}   - Types (dimensions) of relons desired
     * @param {string}  - Optional debug el for json printing
     * @author Mr.Scatman                                 
    */
    queryForOuts = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
        }

        var nid = $nObj.nid;
        if ($type.length) {
            var aRelons = await ANNE.API.queryForOuts(nid, $bMUSTPULL, $type);
        }
        else {
            //alert("optimizer skip query");
            console.log("%c queryForOuts %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOutsAndIns = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
            return;
        }
        var nid = $nObj.nid;
        
        //PROCEED ONLY IF WE HAVE SOMETHING NEW TO CHASE.
        var aRelons = await ANNE.API.queryForOutsAndIns(nid, $bMUSTPULL, $type);
        
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nids 
     * @param {*} $bMUSTPULL 
     * @param {*} $types 
     * @returns 
     */
    queryForOutsAndInsFromNids = async($nids, $bMUSTPULL=false, $types=[])=>{
        var nids = [];
        for (var n=0; n < $nids.length; n++) {
            if (($nObjs[n] instanceof anne_neuron) == false) {
                console.warn("FIX UR CALL --> needs an object for nObj");
                continue;
            }
            if ($nObjs[n].ghost==true || $nObjs[n].temp==true) {
                console.warn("no ghosts or temps found in the ANNODE layer");
                continue;
            }
        }
        
        //PROCEED ONLY IF WE HAVE SOMETHING NEW TO CHASE.
        return await ANNE.API.queryForOutsAndInsFroms(nids, $bMUSTPULL, $types);
    }


    /**
     * 
     * @param {*} $nObjs 
     * @param {*} $bMUSTPULL 
     * @param {*} $types 
     * @returns 
     */
    queryForOutsAndInsFroms = async($nObjs, $bMUSTPULL=false, $types=[])=>{
        var nids = [];
        for (var n=0; n < $nObjs.length; n++) {
            if (($nObjs[n] instanceof anne_neuron) == false) {
                console.warn("FIX UR CALL --> needs an object for nObj");
                continue;
            }
            if ($nObjs[n].ghost==true || $nObjs[n].temp==true) {
                console.warn("no ghosts or temps found in the ANNODE layer");
                continue;
            }
            nids.push($nObjs[n].nid);
        }
        
        //PROCEED ONLY IF WE HAVE SOMETHING NEW TO CHASE.
        return await ANNE.API.queryForOutsAndInsFroms(nids, $bMUSTPULL, $types);
    }



    /**
     * Chase outs and ins including the BEM line of the nObj.nid  (its parent line outs)
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOutsAndIns_BEM = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
            return;
        }
        var nid = $nObj.nid;
        if ($type.length) {
            var aRelons = await ANNE.API.queryForOutsAndIns_BEM(nid, $bMUSTPULL, $type);
        }
        else {
            console.log("%c queryForOutsAndIns_BEM %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOuts_BEM = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            alert("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
            return;
        }
        var nid = $nObj.nid;
        if ($type.length) {
            var aRelons = await ANNE.API.queryForOuts_BEM(nid, $bMUSTPULL, $type);
        }
        else {
            console.log("%c queryForOuts_BEM %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * Get a global value.
     * Numbers, Dates, and Keywords are globals and created by the nodes,
     * sponsored by identities.  Get a global from local memory first.  If it exists
     * that is it.  If not, go check the node for the global.
     * If it exists, that's it.
     * If not, the user can trigger the creation of the global by using a 
     * globalNeuron TX type.  This will pay the node operator to create a global.
     * 
     * Locations are too but they are handled a bit differently.
     * 
     * @author Mr.Scatman             
    */
    getGlobal=async($classNOBJ, $val)=>{
        var nObj = ANNE.MEMG.getGlobalNeuron($classNOBJ, $val);
        if (nObj!=null) {
			
			if (nObj.ghost==false) {
	            console.warn("*** WE ALREADY HAVE THIS GLOBAL IN LOCAL MEM! **");
	            return {status:"good-but", result:nObj};
	        }
			else {
				console.warn("*** WE ALREADY HAVE THIS GLOBAL GHOST IN LOCAL MEM! **");
	            return {status:"good-but", result:nObj};
			}
		}
        
        var response = await ANNE.API.queryForGlobal($classNOBJ.nid, $val);
        
        if (response.status=="good" && response.message=="available") {
            //console.log(response);
            return response;
        }

        //sb.b good but.
        else if (response.status=="good-but" && response.result) {
            var rObj = response.result;
            if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "CHASE.getGlobal");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                response.result = nObj;
                return response;
            } 
            else {
                console.log(response);
                return [];
            }
        }
        else return [];
    }


    /**
     * 
     * @param {*} $nid 
     */
    loadChildrenFromIDB=async($nids)=> {
		console.time("loadChildrenFromIDB");
		var aRelonsToProcess = await ANNE.DB.loadAll("relons");
		var aRelons = [];
		for (var r=0; r < aRelonsToProcess.length; r++) {
			if ($nids.indexOf(aRelonsToProcess[r].be)==-1) {
				continue;
			}
			aRelons.push(aRelonsToProcess[r]);
		}
        if (aRelons.length) await ANNE.MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "CHASE.loadChildrenFromIDB");
		console.timeEnd("loadChildrenFromIDB");
        return aRelons;
    }

    
    /**
     * With a list of nids...
     * load their BEM outs from local IDB into RAM.
     * @param {*} $nids 
     * @returns 
     */
    loadLocalBEMs=async($nids)=>{
        var nids = [];
        for (var n=0; n < $nids.length; n++) {
            var nid = $nids[n];
            if (nids.indexOf(nid)!==-1) continue;
            nids.push(nid);
            var nObj = MEMG.getNeuronByNid(nid);
            var parent = nObj.getParent();
            while(parent!=null) {
                if (nids.indexOf(parent.nid)!=-1) break;
                nids.push(parent.nid);
                parent = parent.getParent();
            }
        }
		var aRelons = await ANNE.IDB.loadIDBNidOuts(nids);
	    console.log("%c loaded "+aRelons.length+" from localBEMs", ANNE.TALK.bggreen);
	}

    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    loadLocalBEM=async($nObj)=>{
        var nids = [$nObj.nid];
        var parent = $nObj.getParent();
        while(parent!=null) {
            nids.push(parent.nid);
            parent = parent.getParent();
        }
        return await ANNE.IDB.loadIDBNidOuts(nids);
    }



    /**
     * 
     * THIS DOES NOT SEARCH LOCAL MEMORY FIRST.
     * USE A SPECIALIZED FUNCTION FIRST like MEMG.getOrMakeIfNewPerson.
     * 
     * This returns multiple results in nObjs.
     * no array value means we do not know of this neuron.
     * 
     * Note: we could do a multiple relon version of this too (if needed)
     * @param {*} $classNObj 
     * @param {*} $val 
     * @returns 
     */
    queryFoNeuronByClassAndRelon=async($classNObj, $val, $reln_nid, $to_type, $to_s)=>{
        var response = await ANNE.API.queryFoNeuronByClassAndRelon($classNObj.nid, $val, $reln_nid, $to_type, $to_s);
        if (response.status=="good" && response.message=="available") {
            return [];
        }
        else if (response.status=="good-but" && response.results!=undefined) {
			var relons = response.result;
			if (Array.isArray(relons)==false || relons.length==0) {
				relons = response.results;
			}
            if (relons!=undefined && relons!=null && relons.length) {
                relons = await ANNE.MEMP.processNeuronsFromRelons(relons, true, true, "relons", "CHASE.queryFoNeuronByClassAndRelon");
                ANNE.TALK.printRelons(relons);
                var matches = [];
                var unique_tnids = [];
                for (var r=0; r < relons.length; r++) {
                    var nObj = MEMG.getOrMakeIfNew(relons[r].from_nid);
                    if (unique_tnids.indexOf(nObj.temp_nid)==-1) {
                        matches.push(nObj);
                        unique_tnids.push(nObj.temp_nid);
                    }
                }
                return matches;
            } 
            else {
                //"helper"
                //"matching_text_nids" -- nids that matched w/ no matching reln/to_s
                //"contradictors" -- nids that contradicted
                if (response.matching_text_nids!=undefined) {
                    if (response.contradictions!=undefined) {
                        var matches = [];
                        for (var m=0; response.matching_text_nids.length; m++) {
                            var match_nid = response.matching_text_nids[m];
                            if (response.contradictions.indexOf(match_nid)!=-1) continue;
                            var nOBJ = MEMG.getOrMakeIfNew(match_nid);
                            matches.push(nOBJ);
                        }
                        return matches;
                    }
                }

				console.log(response);
                return [];              //matches exist but contradict, we have to create a new thing.
            }
        }
    }


    /**
     * Used as a psuedo optimization for searching by ALT BE CLASS first, then text.
     * Classes like "occupation" are not instantiated.  This would be a large classication error.
     * We simply create a new ec_living "thing" and laterally tag it with an alt TYPE_BE of "occupation".
     * To find an occupation we can snag lateral TYPE_BE occupation relons that are ec_things...
     * THEN we can validate the TEXT string (more expensive) instead of forcing the annode to check
     * against all text strings for all things.
     * @param {*} $classNObj 
     * @param {*} $text 
     * @param {*} $alt_be_reln 
     * @param {*} $alt_be_to 
     * @returns 
     */
    queryFoNeuronByAltBeTexAndBe=async($classNObj, $text, $alt_be_reln, $alt_be_to)=>{
        var response = await ANNE.API.queryFoNeuronByAltBeTexAndBe($classNObj.nid, $text, $alt_be_reln, $alt_be_to);
        if (response.status=="good" && response.message=="available") {
            console.log(response);
            return [];
        }
        
        else if (response.status=="good-but" && response.results!=undefined) {
            
            var relons = response.results;
            if (relons!=null && relons.length) {
                relons = await ANNE.MEMP.processNeuronsFromRelons(relons, true, true, "relons", "queryFoNeuronByAltBeTexAndBe");
                ANNE.TALK.printRelons(relons);
                matches = [];
                for (var r=0; r < relons.length; r++) {
                    var nObj = MEMG.getNeuronByTempNid(relons[r].f);
                    matches.push(nObj);
                }
                return matches;
            } 

            else {
                console.log(response);
                return [];
            }
        }

        return [];
    }


    
    /**
     * Look for a subanndle under a givne parent anndle/subanndle.
     * @param {anne_neuron} $parentNOBJ 
     * @param {String} $val 
     * @returns 
     */
    getSubAnndle=async($parentNOBJ, $val)=>{
        console.log("(CHASE.getSubAnndle) for class:"+$parentNOBJ.text+" value:"+$val);
        var nObj = MEMG.getSubAnndle($parentNOBJ, $val);
        if (nObj!=null) {
            console.warn("*** WE ALREADY HAVE THIS GLOBAL IN LOCAL MEM! **");
            return nObj;
        }
        var response = await ANNE.API.getSubAnndle($classNObj.nid, $val);
        if (response.status=="good" && response.message=="available") {
            console.log(response);
            return response;
        }
        else {
            var rObj = response.result;
            if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "getSubAnndle");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                response.result = nObj;
                return response;
            } 
            else {
                console.log(response);
                return response;
            }
        }
    }

    
    /**
     * Get open hanais that pertain to me.
     * note: LNOTS (local notifications) will be available from the ANNODES that have handshoot
     * w/ this userID within the last 7 days.
     * 
     * This function can be used to get hanais that are either outside of this range or were
     * not captured (at the time they came in) by an the present ANNODE that the user is handshook with (using).
     * 
     * @param {String} - Nid long string
     * @param {int} - block height min
     * @author Mr.Scatman             
    */
    queryForHanai = async($nid, $height=0)=> {
        return await ANNE.API.queryForHanai($nid, $height);
    }

    //see use in dossier
    getPendingHanaiResponse=async($txid)=> {
        var rez =  await ANNE.API.queryForHanaiResponseTo($txid);
        console.log("(getPendingHanaiResponse)",rez);
        return rez;
    }


    /**
     * Get granparent (one), uncles/aunties, brother/sister and cousin locations relative to a location neuron.
     * Any of these can match when we are trying to find a location in memory.
     * 
     * We want to avoid creating a Statue of Liberty site in NYC, then another one in the Bourough of Manhattan.
     * 
     * If we already store it in NYC then we need not create another Statue of Liberty.  We can always alter
     * the parent location (Statue of Liberty->[TYPE_BE]->constituent_of->NYC) to be more refined.
     * 
     * NOTE: This is intended to be processed by the getGlobalLocation_localonly which looks at the ins
     * irrespective of the FROM and takes the FROM as a legit potential match to string match against.
     * 
     * @param {*} $parentLocation 
     * @returns 
     */
    getSpotlightConstituentIns = function($parentLocation) {
        var ins_spotlight = [];

        for (var i=0; i <  $parentLocation.ins.length; i++) {
            var rel =  $parentLocation.ins[i];
            if (rel.r == ANNE.MEM.constituent_of.temp_nid) {

                //sibblings to the level we are testing a new alt location (brothers/sisters)
                ins_spotlight.push(rel);

                //children locations under the sibblings of our parent location.
                //ie, newphews/nieces.
                var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
                ins_spotlight = ins_spotlight.concat(fromNOBJ.ins);
            }
        }

        //grandparent location
        var co = $parentLocation["be_"+ANNE.MEM.constituent_of.temp_nid];
        if (co!=undefined) {
            var grandParentLocation = MEMG.getNeuronByTempNid(co);
            var gpco = MEMG.getBeRelon(grandParentLocation.outs, MEM.constituent_of.temp_nid);
            if (gpco!=null) ins_spotlight.push(gpco); //this gets gramps in the array.
            for (var i=0; i <  grandParentLocation.ins.length; i++) {
                var rel =  grandParentLocation.ins[i];
                if (rel.r == ANNE.MEM.constituent_of.temp_nid) {

                    //sibblings to our parent location. (uncles/aunties)
                    ins_spotlight.push(rel);
                    
                    //children of our uncles/aunties  (cousins)
                    var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
                    var co2 = MEMG.getBeRelons(fromNOBJ.ins, ANNE.MEM.constituent_of.temp_nid);
                    ins_spotlight = ins_spotlight.concat(co2);
                }                
            }
        }

        return ins_spotlight;
    }


    /**
     * Search local mem for a location within a consitiuent of a specific altBe.
     * 
     * @param {loc_search_pool} $lsp 
     */
    getGlobalLocation_localonly=function($lsp) {
        var altBeNOBJ = MEMG.getNeuronByNid($lsp.altBeNid);
        var ins = [];
        var lower = $lsp.str.toLowerCase();

        //we should be a bit more flexible on constituent location and look up and down their constituent tree.
        //ie,    NYC -> up to New York..  NYC -> down to the Bronx, Brooklyn, Manhattan, Queens and Staten Island.
        ins = this.getSpotlightConstituentIns($lsp.constituent_of_nOBJ);
        for (var r=0; r < ins.length; r++) {
            var rel = ins[r];
        
            var fromLocNOBJ = MEMG.getNeuronByTempNid(rel.f);   //this nObj is a constituent of our location.
            if (ANNE.isANNE(fromLocNOBJ.owned_by)==false && fromLocNOBJ.owned_by!=USER._userNObj.nid) continue;

            //this is the proper alt-be  (ie, a site in NEW YORK CITY)
            if (altBeNOBJ!=null && MEMG.isAltBeOf(fromLocNOBJ.outs, altBeNOBJ.temp_nid)==false) continue;

            //see if the text matches or is close!
            var fire = ANNE.UTILS.similarity(fromLocNOBJ.text.toLowerCase(), lower, 1000);

            //look at what we want to do here with near hits..
            //if we reutrn a match object we have to handle that in all cases down stream from
            //our search.
            if (fire >= $lsp.fire) {
                var co = 0;
                var co_tnid = null;
                if (rel.t==$lsp.constituent_of_nOBJ.temp_nid) co=$lsp.LSM_DIRECT;
                else {
                    co=$lsp.LSM_RELATIVE;
                    co_tnid = rel.t;
                }
                $lsp.addMatch(fromLocNOBJ, fire, co, co_tnid);
            }
            else if (fire>.8) console.log("(close) partial "+fromLocNOBJ.text+" fire:"+fire);
        }
    }


    /**
     * Don't use this if you can use the location search with a known constituent parent.
     * This search can return more than one result because there is often locations
     * that are named the same.  (ie, multiple cities in the US are named the same thing)
     * 
     * We use this function when we don't have anything else to go on.
     * @param {*} $location 
     * @returns 
     */
    getNamedLocation=async($location) => {
        var matches = [];
        var matches_tnids = [];
        var location = $location.toLowerCase();
        for (var c = 0; c < MEM.location.children.length; c++) {
            var nObjLoc = MEMG.getNeuronByTempNid(MEM.location.children[c]);
            if (nObjLoc.text.toLowerCase()==location) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
            else if (nObjLoc.aka_lower==undefined) continue;
            else if (nObjLoc.aka_lower.indexOf(location)!=-1) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
        }
        if (matches.length>0) return matches;

        //search for it in the hypergraph.
        var rels = await ANNE.API.queryForNamedLocation($location);

        if (rels.length==0) return [];

        //search the returned relons.
        for (var r = 0; r < rels.length; r++) {
            var nObjLoc = MEMG.getNeuronByTempNid(rels[r].from_nid);
            if (nObjLoc.text.toLowerCase()==location) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
            else if (nObjLoc.aka_lower==undefined) continue;
            else if (nObjLoc.aka_lower.indexOf(location)!=-1) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
        }
        if (matches.length>0) return matches;
    }


    /**
     * Get a location in a constituent location.
     * 
     * If altbe is a site or geo site we want to be a bit more flexible on the constituent location.
     * We should, technically, allow for finding the Statue of Liberty in New York... even if it is in New York City.
     * 
     * @param {Array}
     * @returns {location_search_pool}
     * @author Mr.Scatman             
    */
    getGlobalLocation=async($location, $constituent_locationNOBJ, $altBeNid, $exactThresh=1)=>{
        var lsp = new loc_search_pool($location, $constituent_locationNOBJ, $altBeNid, $exactThresh);
        this.getGlobalLocation_localonly(lsp);
        var matches = lsp.getMatches();
        var nObjs = [];
        if (matches.length>0) {
            for (var m=0; m < matches.length; m++) {
                var match = matches[n];
                var nObj = match.nObj;
                if (nObj.owned_by==ANNE.USER._userNObj.nid || ANNE.isANNE(nObj.owned_by)==true) {
                    if (match.co==LSM_DIRECT) nObjs.push(nObj);
                }
            }
        }
        
        if (nObjs.length>0) return lsp; //bail without a query of ANNE.

        var response = await ANNE.API.queryForLocation($location, $constituent_locationNOBJ.nid, $altBeNid);
        console.log("response:"+response);
        if (response.status=="good" && response.message=="available") return lsp;
        else {
            var rObj = response.result;
            if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "getGlobalLocation");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                lsp.addMatch(nObj, 1, LSM_DIRECT);
                return lsp;
            }
            else {
                alert("fix this.. it should have returned back the getGlobalLocation exiting value");
            }
        }
        return lsp;
    }
    


    /**
     * Make a global location
     * 
     * @param {string} - the name of the location
     * @param {string} - The nid longstring of a KNOWN location that
     * this new location will be a constituent of.
     * 
     * @author Mr.Scatman             
     */
    makeGlobalLocation=async($location, $constituent_locationNOBJ, $bSINGLESEND=true, $gui_extras=null)=>{
        if ($location==undefined || $location==null || $location=="") return;
        var msg = "<span class='q-part rel-reln'>make global location</span> <span class='q-part rel-to'>"+$location+"</span>";
        var rObj = ANNE.MEM.makeTempRelon($constituent_locationNOBJ, TYPE_SYS_MSG, MEM.constituent.nid, TO_TYPE_STRING, $location, true);
        rObj.msg = msg;
        if ($gui_extras) rObj.gui_extras=$gui_extras;
        var cmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION);
        if (cmdObj!=null && $gui_extras!=null && $gui_extras.znid!=undefined) cmdObj.znid=$gui_extras.znid;
        return cmdObj;
    }


    //we have a pending globalLocation that requires another location be made first.
    makePendingGlobalLocation=async($location, $constituent_locationNOBJ, $bSINGLESEND=true, $gui_extras=null)=>{
        if ($location==undefined || $location==null || $location=="") return;
        //var rObj = MEM.makeTempRelon($constituent_locationNOBJ, TYPE_SYS_MSG, MEM.constituent.nid, TO_TYPE_STRING, $location, true);
        var msg = "<span class='q-part rel-reln'>make global location</span> <span class='q-part rel-to'>"+$location+"</span>";
        var rObj = {type:TYPE_SYS_MSG, from_nid:null, pending_from:"new_"+$constituent_locationNOBJ.nid, reln:MEM.constituent.nid, to_type:TO_TYPE_STRING, to_s:$location}
        rObj.msg = msg;
        if ($gui_extras) rObj.gui_extras=$gui_extras;
        await ANNE.SENDQ.pendingDynamicVarRelon(rObj, true, false, false, ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION, $gui_extras, false);
        return rObj;
    }

    
    /**
     * Make a global neuron via ANNODE request.
     * NOTE: globals are just classes of neurons.  They could be 
     * created with keyless request (no special handling needed)
     * HOWEVER, we want to handle certain classes with extra care to 
     * as we intend to get them into the ANNE worldview. (some of them)
     * 
     * There DOES NOT NEED TO BE EXPANSION to annode handling to handle 
     * new classes.  They can be created and simply made.
     * 
     * @param {anne_neuron} - the class of global neuron we are making
     * @param {string} - value of the global
     * @param {boolean} - send now?
     * @param {Object} - the guiextras return object we use to do stuffs w/.
     * @author Mr.Scatman             
     */
    makeGlobal=async($classNObj, $val, $bSINGLESEND=true, $gui_extras=null)=>{
        var cmd_nid=0;
        if ($val==undefined || $val==null || $val=="") return;
        var to_type = TO_TYPE_STRING;
        var msg = "";
        //we actually are making integers (not just numbers)...
        if ($classNObj.temp_nid==MEM.integer.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_INTEGER.nid;
            to_type = TO_TYPE_NUMBER;
            msg = "<span class='q-part rel-reln'>make global number</span>";
        }
        else if ($classNObj.temp_nid==MEM.anndle.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_ANNDLE.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make anndle</span>";
        }
        //this will create in THEIR worldview.. (not ANNE's)
        else if ($classNObj.temp_nid==MEM.ec_action.temp_nid && $classNObj.ntype==NTYPE_CLASS) {
            cmd_nid = ANNE.MEM.MAKE_ACTION.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make global action (verb)</span>";
        }
        //YYYY/MM/DD
        else if ($classNObj.temp_nid==MEM.date_gregorian.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_DATE.nid;
            to_type = TO_TYPE_DATE;
            msg = "<span class='q-part rel-reln'>make global date</span>";
        }
        else if ($classNObj.temp_nid==MEM.anorg.temp_nid) {
            cmd_nid = MEM.MAKE_ANORG.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make an anorg</span>";
        }
        else if ($classNObj.temp_nid==MEM.company.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_COMPANY.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make global company</span>";
        }
        //this is the url to get the img data.
        else if ($classNObj.temp_nid==MEM.neuronImg.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_IMG.nid;
            to_type = TO_TYPE_URL;
            msg = "<span class='q-part rel-reln'>make global IMG reference</span>";
        }
        else if ($classNObj.temp_nid==MEM.person.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_PERSON.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make person</span>";
        }
        else if ($classNObj.temp_nid==MEM.bounty.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_BOUNTY.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make bounty</span>";
        }
        else {
            alert("unhandled global create attempt");
            console.warn("unhandled global create attempt");
            return;
        }

        msg = msg + " <span class='q-part rel-to'>"+$val+"</span>";
        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_SYS_MSG, cmd_nid, to_type, $val, true);
        if (rObj==null) {
			console.warn("rObj failed to make temp CMD_NID relon!!! BIG TROUBLE WILL HAPPEN!!!");
		}
		rObj.msg = msg;
        
        if ($gui_extras) rObj.gui_extras=$gui_extras;   //single send is safer to use rObj.gui_extras not cmdObj.gui_extras

        var tempCmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, TXB.TXTYPE_MAKE_GLOBAL);
		
        if (tempCmdObj!=null) {
			tempCmdObj.delay = 500;
		}
		else {
			console.warn("big trouble will happen here!!!  Investigate why no tempCmdObj was created!!", rObj);
		}
		
        if (tempCmdObj && $gui_extras!=null && $gui_extras.znid!=undefined) tempCmdObj.znid = $gui_extras.znid;

        return tempCmdObj;
    }


   /**
     * Make a keylesss neuron via NODE request.
     * Keylesss neurons are made via standard sendPayData relons.
     * The payTo must be set to the node that the request it broadcast to.
     * 
     * NOTE: The ANNODES will reject any request to make a keyless class with a parent class
     * that does not yet exist.  In the case of a new top-class use the parent class of 
     * MEM.class (7777942827590344670) 
     * 
     * or 
     * 
     * MEM.gen_class (7770801601458804038) if you intend
     * this new class to be used in a property-like capacity (if it will be used mostly 
     * as the reln: in a relon, as opposed to being a class
     * that ultimately has children that are instances).
     * 
     * @param {anne_neuron} - the parent/be.
     * @param {anne_neuron} - MEM.inst or MEM.class
     * @param {function} - callback
     * @param {boolean} - trigger sending now (don't queue up further requests)
     * @author Mr.Scatman             
     */
    makeKeylessNeuron=async($classNObj, $instOrClassNOBJ, $callbackSingleSend=null, $bINSTANTSEND=true, $gui_extras=null)=>{
        var msg = "new keyless "+$instOrClassNOBJ.text+" of '"+$classNObj.text+"'";
        //console.log("%cKEYLESS %c"+msg, ANNE.TALK.bgred, ANNE.TALK.bggreen);
        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_SYS_MSG, $instOrClassNOBJ.nid, TO_TYPE_ANNE, $classNObj.nid, true);
        rObj.msg = msg;
        if ($gui_extras!=null) rObj.gui_extras = $gui_extras;
        var cmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bINSTANTSEND, TXB.TXTYPE_KEYLESS_NEURON, $callbackSingleSend);
        if (cmdObj!=null) {
            cmdObj.delay = 500; //we need to delay this to ensure a different epoch sec timestamp lest big troubles.
        }
        if ($gui_extras!=null && $gui_extras.znid!=undefined) cmdObj.znid=$gui_extras.znid;
        //console.log(cmdObj);
        return rObj;
    }

    /**
     * 
     * @param {anne_neuron} $classNObj 
     * @param {anne_neuron} $instOrClassNOBJ 
     * @param {function} $callbackSingleSend 
     * @param {Boolean} $bINSTANTSEND 
     * @param {Object} $gui_extras 
     */
    makePDNNeuron=async($classNObj, $instOrClassNOBJ, $callbackSingleSend=null, $bINSTANTSEND=true, $gui_extras=null)=>{
        var msg = "new PDN "+$instOrClassNOBJ.text+" of '"+$classNObj.text+"'";
        var rObj = ANNE.MEM.makeTempRelon(USER._userNObj, TYPE_SYS_MSG, $instOrClassNOBJ.nid, TO_TYPE_ANNE, $classNObj.nid, true);
        rObj.msg = msg;
        if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
        await ANNE.SENDQ.addRelonToSendQ(rObj, $bINSTANTSEND, TXB.TXTYPE_PDN, $callbackSingleSend);
    }

    /**
     * Send a custodial neuron.
     * This will receive back critical data required to finish the custodial.
     * The from is a dynamic neuron (the custodial nid) but the to of the ANNOUNCE is the txid of the CUSTODIAL request.
     * The pubKey from the annode is passed back through and is must be set 
     * @param {anne_neuron} $classNObj 
     * @param {anne_neuron} $instOrClassNOBJ 
     * @param {function} $callbackSingleSend 
     * @param {Boolean} $bINSTANTSEND 
     * @param {Object} $gui_extras 
     */
    makeCustodialNeuron=async($classNObj, $instOrClassNOBJ, $callbackSingleSend=null, $bINSTANTSEND=true, $gui_extras=null)=>{
        var msg = "new custodial inst of '"+$classNObj.text+"'";
        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_SYS_MSG, $instOrClassNOBJ.nid, TO_TYPE_ANNE, $classNObj.nid, true);
        rObj.msg = msg;
        if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
        await ANNE.SENDQ.addRelonToSendQ(rObj, $bINSTANTSEND, TXB.TXTYPE_CUSTODIAL, $callbackSingleSend);

        //we need to snag the last txid (among a few other vars in the response handler).
        //pop in 120 coins (12,000 annies) so start it off with.
        //NOTE -- ALL USAGE OF THE FAUCET OR CUSTODIAL SHOULD USE A referencedTransactionFullHash
        //to THIS ANNOUNCEMENT --- to ensure is mined in the same block or PRIOR to the usage of this.
        //We have to do this because this TX will be itself using a sys_ref_txid to the custodial creation transaction!
        //This means the ANNOUNCE may get delayed till the next block and then the fromception would fail for the TXs.

        //NOTE:  let's tell the node to wait on MEMPOOL TXS if they fail fromception.  Can we do that?  Hmm.
        //200 annies, 100 for fee, 100 init fund.
        var announce_relon = {annies:200, type:TYPE_SYS_MSG, from_nid:USER._userNObj.nid, reln:MEM.MAKE_CUSTODIAL.nid, to_type:TO_TYPE_TXID, to_s:null, pending_to_s:"new_"+rObj.txid};
        await ANNE.SENDQ.pendingDynamicVarRelon(announce_relon, false, false, true, TXB.TXTYPE_ANNOUNCE_CUSTODIAL_NEURON, null, false);
        return rObj;
    }


    
    /**
     * WIth a list of nids, check for min info chased.  If not, chase it.    
     * This function should be called over the chaseMin and chaseGhostMin as 
     * it smartly ascertains how to best chase the neuron.                   
     * bCHASED_min indicates we've looked up the 0/1/2 relons.  
     * @param {Array}
     * @author Mr.Scatman             
     */
     lookupMinTNids=async($temp_nids)=> {
        var nid_list = [];
        var nObjs = [];
        for (var n=0; n < $temp_nids.length; n++) {
            var nObj = MEMG.getNeuronByTempNid($temp_nids[n]);
            if (nObj.nid==USER.public_nid) continue;    //don't look for our own name.
            if (nObj.ghost==true) continue;
            if (nObj.bCHASED_min==true) continue;
            if (nObj.preloaded==true) continue;      
            nObjs.push(nObj);
            nid_list.push(nObj.nid);
            nObj.bCHASED_min=true;  //don't do this again :)
        }
        if (nid_list.length!=0) await this.lookupMinList(nid_list, [0,1,2]);
        return;
    }


    /**
    * Lookup a list of nids.
    * Typically used on classess that show up as relns.  We do not need
    * to get everything for these, until we do.  Then we do.  Otherwise,    
    * just get the basic min info.                                          
    * This will get any type:1 relons (including the be defining parent   
    * class/inst and the BE mem.name (for .text property)                   
    * bCHASED_min was set coming in to this function so neurons will not be 
    * chased again for mins.
    * @param {Array}
    * @author Mr.Scatman                                                
    */
    lookupMinList=async($nidslist, $types=[0,1,2], $from="", $bFIXAFTERIDB=false)=>{
        var aRelons = await ANNE.API.lookupNids($nidslist, $types, true, true, $from, $bFIXAFTERIDB);
        return aRelons;
    }
    lookupNonAwareList=async($nidslist)=>{
        var aRelons = await ANNE.API.lookupNidsNonAwareness($nidslist);
        return aRelons;
    }
    lookupNidsGetAllOuts=async($nidslist, $offset=0)=>{
        var aRelons = await ANNE.API.lookupNidsGetAllOuts($nidslist);
        if (aRelons.length==ANNE.API.getRelonLimit()) {
            console.warn("IMPLEMENT OFFSET RETURN LISTS");
            console.warn("THERE ARE MORE RELONS TO GET");
        }
        return aRelons;
    }



    /**
     * Get all the TXs that touch the USER.
     * outs/ins and relon_sponsored.
     * Note: all childRelPat relons broadcast will have the user as the relon_sponsor.
     * Note2: we have to make sure to optimize that query to efficiently grab all that data.
     * @param {*} $nid 
     */
    getAllUserPayments = async($nid, $bDOSPONSORED=false, $bDOIDB=true)=>{
        //console.log("(getAllUserPayments)");
        
        //get all annies in/out from broadcast/internal payments.
        //also snag any in/out relons with .annies on it.
        await ANNE.API.getAllPayments($nid, "getAllUserPayments", $bDOIDB);
        
        var nOBJ = MEMG.getNeuronByNid($nid);
        var rels = nOBJ.outs.concat();

        //debug --- WE WILL REMOVE THIS LATER.
        for (var r=0; r < rels.length; r++) { rels[r].debug="from outs";}
        for (var r=0; r < nOBJ.ins.length; r++) { nOBJ.ins[r].debug="from ins";}
        
        if (ANNE.isANNE(ANNE.USER._userNObj.nid)==false) {       
            if (nOBJ.relon_sponsored!=undefined) {
                for (var r=0; r < nOBJ.relon_sponsored.length; r++) { 
                    nOBJ.relon_sponsored[r].debug="from relon sponsor";
                }
            }
        }
        
        //so... outs and ins..
        rels = rels.concat(nOBJ.ins);
        
        if ($bDOSPONSORED==true && ANNE.UTILS.isset(nOBJ.relon_sponsored)==true) {
            return ANNE.UTILS.uniqueConcatRels(rels, nOBJ.relon_sponsored);
        }
        return rels;
    }


    /**
     * This does not get all TXs.. (as of our new RAM optimizer IDB mashup)
     * @param {*} $nid 
     * @returns 
     */
    getAllUserTransactions3 = async($nid)=>{
        await ANNE.API.getAllPaymentsIn($nid);
        var nOBJ = MEMG.getNeuronByNid($nid);
        var rels = nOBJ.outs.concat();
        rels = rels.concat(nOBJ.ins);
        return rels;
    }


    /**
     * chase will pull IDB info too.
     */
    getAnnexerWhales=async()=>{
        var nObjs = await ANNE.API.getAnnexerWhales(20);
        await ANNE.IDB.lookupNids_fast(nObjs); //explore this new fix.
        return nObjs;
    }
    

    /**
     * chase will pull IDB info too.
     */
    getSponsored=async()=>{
        var nObjs = await ANNE.API.getSponsored(ANNE.USER._userNObj.nid);
        await ANNE.IDB.lookupNids_fast(nObjs); //explore this new fix.
        return nObjs;
    }


    /**
     * chase will pull IDB info too.
     */
    getLatestXAnnexes=async($cnt=20)=>{
        var nObjs = await ANNE.API.getLatestXAnnexes($cnt);
        await ANNE.IDB.lookupNids_fast(nObjs); //explore this new fix.
        return nObjs;
    }

    /**
     * chase will pull IDB info too.
     */
    getTopXAnnexByLevel=async($cnt=20)=>{
        var nObjs = await ANNE.API.getTopXAnnexByLevel($cnt);
        await ANNE.IDB.lookupNids_fast(nObjs); //explore this new fix.
        return nObjs;
    }

    /**
     */
    search_recents_by_class=async($nids=[], $maxResults=200)=>{
        var aNeurons = await ANNE.API.queryForRecents($nids, $maxResults);
        return aNeurons;
    }

    
    /**
     * 
     * @param {*} $searchVal 
     * @param {*} $class_tnid 
     * @param {*} $wvs 
     */
    anne_bem_search=async($searchVal, $class_tnid, $wvs=[])=> {
        var done = await ANNE.API.queryForText_BEM ($searchVal, $class_tnid, $wvs);
        console.log(done);

        //we need to make sure these are unique.. (TODO);
        var matches = await ANNE.SEARCH.searchNeuronText ($searchVal, $class_tnid, $wvs, false);
        if (matches.length) return matches;
    }


    /**
     * 
     * @param {String} $searchVal 
     * @param {Boolean} $bLOCALONLY 
     * @param {Array} $ntypes 
     * @param {Boolean} $bDOAKAS 
     * @returns 
     */
    anne_naked_neuron_search=async($searchVal, $bLOCALONLY=false, $ntypes=[NTYPE_INST, NTYPE_CLASS], $bDOAKAS=false)=> {
        
        //search all spaces
        if ($bLOCALONLY==false) await ANNE.API.queryForText ($searchVal, [], [], true);

        //search only local.
        return ANNE.SEARCH.searchTextStart($searchVal, $ntypes, false, $bDOAKAS);
    }


    /**
     * for app boot up to clean up some messes that may have bene caused.
     * for large TX sends where the user closes the app.. this could be no so frenly..
     * 6k unconfirmed TXs would not exactly be fun to chase out.  Maybe we max it out 
     * and chip away?
     * 
     * ANY TXIDS returned from this function will be deleted from IDB mem.
     * @param {*} $txids_to_check 
     * @returns 
     */
    getAllDEDTXIDS=async($txids_to_check)=> {
        return;
        var all_ded = await this.getTXIDSThatDontExistInMEMPOOOL($txids_to_check);

        //these TXIDS do not exist in the MEMPOOL...  were they mined or do they not actually exist??
        if (all_ded.length) {
			return await this.getTXIDSThatDontExistInANNE(all_ded);
		}
        return [];
    }


	/**
	 * 
	 * @param {*} $memPoolBootups 
	 */
	getTXIDSThatDontExistInMEMPOOOL=async($memPoolBootups)=>{
		var SHARD = 30;
		var txids = [];
        var all_deds = [];
		for (var t=0; t < $memPoolBootups.length; t++) {
            if ($memPoolBootups[t][0]=="z") continue;
			if (txids.length==SHARD) {
                var ded_txids = await ANNE.API.getTXIDSThatDontExistInMEMPOOOL(txids);
                txids=[];
                if (ded_txids.length) all_deds = all_deds.concat(ded_txids);
			}
            txids.push($memPoolBootups[t]);
		}
        if (txids.length) {
            var ded_txids = await ANNE.API.getTXIDSThatDontExistInMEMPOOOL(txids);
            if (ded_txids.length) all_deds = all_deds.concat(ded_txids);
        }
        return all_deds;
	}

    
    /**
     * 
     * @param {*} $txids_to_check 
     * @returns 
     */
	getTXIDSThatDontExistInANNE=async($txids_to_check)=> {
        var all_deds = [];
		var aRels = await this.lookupTxids($txids_to_check);

		if (aRels.length > 0) {
			var exist_txids = [];
			
			for (var a=0; a < aRels.length; a++) {
				exist_txids.push(aRels[a].txid);
			}
			
			for (var r=0; r < $txids_to_check.length; r++) {
                if (exist_txids.indexOf($txids_to_check[r])!=-1) {
					continue;
				}

                //not in real space.
                all_deds.push($txids_to_check[r]);
            }
		}
        else return $txids_to_check;

        return all_deds;
	}


    /**
     */
    queryForOutsFroms_not=async($arr, $bMUSTPULL=false, $types=[4], $bDOIDB=true)=> {
        var SHARD = 30;
        var this_shard = [];

        //one big one instead of repeated pulls of the entire IDB space to loop.
        var hash = ANNE.other.sha256($arr.join("")+$types.join(""));
		var phObj = ANNE.HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_not", hash);
		if ($bDOIDB==true) await ANNE.IDB.queryForOutsFroms_not(phObj, $arr, $types);

        const bDOIDB = false;
        
        for (var s=0; s < $arr.length; s++) {
            if (this_shard.length==SHARD) {
                await ANNE.API.queryForOutsFroms_not(this_shard, $bMUSTPULL, $types, bDOIDB);
                this_shard = [];
            }
            this_shard.push($arr[s]);
        }

        //if any left over.
        if (this_shard.length) {
            await ANNE.API.queryForOutsFroms_not(this_shard, $bMUSTPULL, $types, bDOIDB);
        }
        
        return;
    }
	


    /**
     * simple and fast.
     * @param {*} $text 
     * @param {*} $bEXACT - we turn this on if a string is quoted when typed in. 
     * @returns 
     */
    nakedSearch = async($text, $bEXACT=false)=>{

        //in this app, we query it everytime, as we want the A lvl.
        var nObjs = await ANNE.API.queryForNakedText($text, $bEXACT);
        if (nObjs==undefined) return [];
        
        if (nObjs.length) {
            await ANNE.IDB.lookupNids_fast(nObjs);
        }

        return nObjs;
    }
    


    /**
     * We are trying to figure out if we have something or not
     * @param {*} $text 
     */
    nakeSearchForResolution = async($text, $classTempNid)=>{
        var matches = [];
        
        var nObjs = await ANNE.MEM.searchTextIndex($text);
        
        var nids = MEMG.getNidsListFromNObjs(nObjs);
        await ANNE.IDB.lookupNids(null, nids, []);

        for (var n=0; n < nObjs.length; n++) {
            if (nObjs[n].bes.indexOf($classTempNid)!=-1) {
                matches.push(nObjs[n]);
            }
        }

        if (matches.length>0) return matches;

        //hit the annode.
        await ANNE.API.queryForNakedText($text);
        nObjs = await ANNE.MEM.searchTextIndex($text);
        if (nObjs.length) {
            for (var n=0; n < nObjs.length; n++) {
                if (nObjs[n].bes.indexOf($classTempNid)!=-1) {
                    matches.push(nObjs[n]);
                }
            }    
        }

        return matches;
    }


    /**
     * Do a global search for the text string.
     * 
     * After the global search, do a local search.
     * Return those results.
     * 
     * NOTE: use the $so.context field to resolve ambiguities (multiple matches)
     * if possible.
     * 
     * @param {Array}
     * @author Mr.Scatman 
     */
    anne_global_search = async ($so, $bDOPRE=false, $bACCEPTRAMONLY=true)=> {

        //text searches are not fun when **not** preloading all of IDB.
        //await ANNE.IDB.textSearch($so.text, $so.exact);
        //await ANNE.IDB.queryForText($so.text, null);
    
        //we're ok here if we find a match before searching the ANNODE space.
        if ($bDOPRE==true) {
            var matches = await ANNE.SEARCH.searchNeuronText ($so.text, $so.bes_tnids, $so.wvs, $so.exact, $so.applywvtobeline, $so.inbeline); 
            
            //not chasing it, then return matches or no matches here.  Sorry.
            if ($so.chaseIt==false) {
                if (matches.length) return matches;
                return [];
            }
            if ($bACCEPTRAMONLY==true && matches.length>0) {
                return matches;
            }
        }

        if ($so.chaseIt==true) {
			if ($so.wv==undefined) $so.wv = [];
            var done = await ANNE.API.queryForText ($so.text, $so.wv, $so.bes, $so.exact, $so.applywvtobeline, $so.inbeline);
            console.log(done);
        }

        //to do, fix for applywvtobeline and inbeline logics (???)
        return await ANNE.SEARCH.searchNeuronText ($so.text, $so.bes_tnids, $so.wvs, $so.exact, $so.applywvtobeline, $so.inbeline);   
    }


    /**
     * We reroute a text search to a nid search based on nid size and all numerics.
     * @param {Array}
     * @author Mr.Scatman 
     */
    anne_nid_search = async ($nid, $bMUSTPULL=false)=> {
        var done = await ANNE.API.queryForOuts ($nid, $bMUSTPULL);
        console.log(done);
        var matches = [];
        var nObj = MEMG.getNeuronByNid($nid);
        if (nObj!=null) matches.push(nObj);
        return matches;
    }


    /**
     */
    getAllMyACHATSessions = async()=> {
        var done = await ANNE.API.queryForOuts ($nid);
        console.log(done);
    }


    /**
     * We just need to ensure we have data from the ANNODE in the local memory
     * so we can process a string that we know to be a location.
     */
    locationParts=async($aParts)=>{
        var done = await ANNE.API.queryForLocationParts ($aParts);
    }


    /**
     * @param {String} - String we are looking for
     * @param {String} - Be nid string of the parent class
     * @author Mr.Scatman 
     */
    queryForTextAndBe=async($text, $be_nid, $WVs=[MEM.ANNE_ACCTID], $bMUSTPULL=false)=>{
        if ($text==undefined || $text==null || $text=="") return [];
        
        //note: for specific classes with known properties this is not the efficient way
        //to find them.  We should target their optimized lookups stored in the class
        //.rzpp1 and .rzpp2 props (set by the optimizing relons 7770223143918999134 and/or 7772071294239796669)

        var done = await ANNE.API.queryForTextAndBe ($text, $be_nid, $WVs, $bMUSTPULL);
        console.log(done);
        var beNOBJ =  MEMG.getNeuronByNid($be_nid);
        var aMatches = [];
        var textlower = $text.toLowerCase();
        for (var i=0; i < beNOBJ.ins.length; i++) {
            var childNOBJ=MEMG.getNeuronByTempNid(beNOBJ.ins[i].f);
            if ($WVs.indexOf(childNOBJ.owned_by)==-1) continue;
            if (childNOBJ.text.toLowerCase()==textlower) aMatches.push(childNOBJ);
        }
        return aMatches;
    }

    queryForCustom=async($beNOBJ)=>{
        var rzpp1 = $beNOBJ.rzpp1;
        var rzpp2 = $beNOBJ.rzpp2;
    }

 
    /**
     * 
     * @param {*} $reln 
     * @param {*} $to 
     * @param {*} $to_type 
     * @returns 
     */
    getRelonsWithTo= async($reln, $to, $to_type)=> {
        var aRelons = await ANNE.API.getRelonsWithTo($reln, $to, $to_type);
        await ANNE.MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "getRelonsWithTo");
        return aRelons;
    }


    /**
     * This is a bit willy-wonka.
     * It used to be tight with forced registratoin polity nid.
     * Now it can be used for just a basic company lookup (no required registration polity)
     * This can return multiple values back.
     */
     getCompany=async($namestr, $registrationPolityNid=null)=>{
        console.log("(CHASE.getCompany)");
        var response = await ANNE.API.getCompany($namestr, $registrationPolityNid);
        //console.log("response:", response);
        if (response.status=="good" && response.message=="available") {
            return response;
        }
        else {
            var rObj = response.result;
            if (Array.isArray(rObj)==true) {
                var aRelons = rObj;
                var arelons2 = await ANNE.MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "getCompany");
                ANNE.TALK.printRelons(aRelons);
                var nObj = MEMG.getNeuronByNid(aRelons[0].from_nid);
                response.results = [nObj];
                response.result = nObj;
                return response;
            }
            else if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "getCompany");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                response.results = [nObj];
                response.result = nObj;
                return response;
            } 
            else {
                alert("fix this.. it should have returned back the getCompany exiting value");
            }
        }
        return null;
     }


    /**
     */
    getPersonWithBday=async($bday, $namestr, $alias="")=> {
        console.log("(CHASE.getPersonWithBday)");
        var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, ANNE.MEM.ANNE_ACCTID], true);
        if (aMatches.length) {
            console.log(aMatches);
            return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
        }
        var response = await ANNE.API.queryForPersonWithBday($bday, $namestr, $alias);
        console.log("response:", response);
        if (response.status=="good" && response.message=="available") {
            console.log("GO TIME");
            return response;
        }
        else {
            var results = response.result;
            if (results!=null) {
                ANNE.TALK.printRelons(results);
                await ANNE.MEMP.processNeuronsFromRelons(results, true, true, "relons", "getPersonWithBday");

                //relons could hvae just been date relons..
                var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, MEM.ANNE_ACCTID], true);
                if (aMatches.length) {
                    console.log(aMatches);
                    return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
                }
                else return {status:"good",message:"available"};
            } 
            else {
                alert("fix this.. it should have returned back the getPersonWithBday exiting value");
            }
        }        
        return null;
    }

   
    
    /**
     */
    getPerson=async($namestr)=> {
        console.log("(CHASE.getPerson)");
		//ANNE.MEM_UDF.openSubHunt("getPerson_searchNeuronText");
        //var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, ANNE.MEM.ANNE_ACCTID], true);
        //ANNE.MEM_UDF.closeSubHunt();
		//if (aMatches.length) {
            //console.log(aMatches);
            //return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
        //}

        var response = await ANNE.API.getPerson($namestr);
        console.log("response:", response);
        if (response.status=="good" && response.message=="available") {
            console.log("GO TIME");
            return response;
        }
        
        var results = response.result;
        if (results!=null) {
            ANNE.TALK.printRelons(results);
            await ANNE.MEMP.processNeuronsFromRelons(results, true, true, "relons", "getPerson");
            
            //we want ALL matches in memory, not just net-new ones we may have picked up
            //in our getPerson search (which may have height filter)
            var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, ANNE.MEM.ANNE_ACCTID], true);
            if (aMatches.length) {
                console.log(aMatches);
                return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
            }
            else return {status:"good",message:"available"};
        } 
        else {
            alert("fix this.. it should have returned back the getPerson exiting value");
            return {status:"bad", message:"unknown problem"};
        }
    }


    /**
     * Get aplex data (annies_in, pos/neu/neg feedback, sell status).
     * @param {string} - nid long string for the neuron you want to get aplex data for.
     * @author Mr.Scatman 
     */
     get_aplex_data = async ($nid, $bMUSTPULL=false, $from="")=> {
        if ($nid[0]=="z") return null;
        var nObj = MEMG.getOrMakeIfNew($nid);
        if (nObj.ghost==true) return null;
        if($bMUSTPULL==false && nObj._aplex!=undefined) return nObj._aplex;
        console.log("%c NEURON APLEX bug hunt! (CHASE.get_aplex_data) from"+$from, ANNE.TALK.bgbigred);  
        nObj._aplex = await ANNE.API.getAplex (nObj);
        return nObj._aplex;
    }


    lookupTxid = async($txid)=> {
        var rels = await this.lookupTxids([$txid]);
        if (rels.length>0) return rels[0];
        return null;
    }


    lookupTxids = async($txids)=> {
        return await ANNE.API.lookupTxids($txids);
    }
    

    /**
     * 
     * @param {*} $nid 
     * @param {*} $bDOPUBKEY 
     * @returns 
     */
    lookupMinNid=async($nid, $bDOPUBKEY=false)=>{
        if (typeof $nid==="object") {
            console.warn("you should pass a number into lookupMinNid, not an object");
            if ($nid.nid!=undefined) {
                $nid = $nid.nid;
                console.warn("in this instance we were able to fix it for you.  PLEASE ALTER YOUR CODE");
            }
            else return null;
        }
        return await ANNE.API.lookupNid($nid, [0,1,2], $bDOPUBKEY);
    }


    /**
     * ARE WE STILL USING THIS?? lolz
     * Returns an array of vote neurons.
     */
     getVoteData = async($nObj)=> { 
        var newVoteData = await ANNE.API.getVoteData($nObj.nid);
        console.log("new vote data", newVoteData);
        return MEMG.getVoteData($nObj.nid);
    }



    /*** chaseDDS_Src ***************************************** x.x.20xx ****
    * Chase out a data source until we reach a non-DDS neuron..  This is    *
    * the root query/neuron and we can get its children.                    *
    * The rest of the downstream DDS wrappers were filters off of that      *
    * original source.   We will gather all WV/garden/deny/lets and filters *
    * for the entire ANNE DDS path...  This will allow us to re-create the  *
    * datasource, live.                                                     *
    * THIS WILL ULTIMATELY RETURN OUT THE SOURCE NEURON.                    *
    ********************************************************* Mr.Scatman ***/
    chaseDDS_Src = async($nObj)=> {
        console.warn("TO DO");
        await ANNE.CHASE.lookupMinNid($nObj.nid);
        if ($nObj.bes[0]!=ANNE.MEM.ANNE_DDS.temp_nid) {
            await ANNE.CHASE.chaseIns_GWV($nObj);
            return $nObj;   
        }
        else {
            var DDS_SRC_NID = $nObj["be_"+ANNE.MEM.ANNE_DDS_SRC.temp_nid];
            if (DDS_SRC_NID==undefined) {
                var ddsRelon = MEMG.getRelonFromWithRToType($nObj.outs, TYPE_BE, ANNE.MEM.ANNE_DDS_SRC.nid, TO_TYPE_ANNE); 
                if (ddsRelon==null) return;
                DDS_SRC_NID = ddsRelon.t;
            }
            console.log(DDS_SRC_NID);
            var DDSNOBJ = MEMG.getNeuronByTempNid(DDS_SRC_NID); 
            return await ANNE.CHASE.chaseDDS_Src(DDSNOBJ);
        }
    }


    /*** chaseNewInfo ***************************************** x.x.2019 ****
    * User on tree/folder clicks and graph neuron clicks.                   *
    ********************************************************* Mr.Scatman ***/
     chaseNewInfo=async($nObj, $type=[0,1,2,3])=> {
        if ($nObj==null || $nObj==undefined) {
            console.warn("(chaseNewInfo) hunt this down");
            return [];
        }

        //get 0/1/2/3 outs..  and INS
        if ($nObj.ntype==NTYPE_CLASS) {
            var aRelons2 = await ANNE.CHASE.queryForOuts($nObj, false, $type);
            var aRelons = await ANNE.CHASE.chaseIns_GWV($nObj, false);
        }

        //get 0/1/2/3 outs only
        else {
            var aRelons = await ANNE.CHASE.queryForOuts($nObj, false, $type);
        }
        return;
    }


    //READRESS THIS NOW THAT WE USE THE ANNODE AS A PERSONAL WEB SERVER.
    //THESE OPTIMIZATIONS ARE A PAIN IN THE ASS AND LESS NEEDED/USEFUL
    //WHEN WE ARE NOT TRYING TO SERVICE MANY USERS WITH ONE ANNDOE.

    /*** chaseIns_GWV ***************************************** x.x.20xx ****
    * THIS FUNCTION MUST HONOR THE GARDEN TYPE AND WORLDVIEW RULES.  IT     *
    * CAN ONLY BE RUN IF THE nOBJ.bCHASED_min is set to true, indicating    *
    * we have gotten all the data we need to properly chase the ins.        *
    ********************************************************* Mr.Scatman ***/
    chaseIns_GWV = async($nObj, $bMUSTPULL=false, $intypes=[1,3], $chasetypes=[0,1,2], $bOVERRIDE_CHECK=false)=>{
        if ($bOVERRIDE_CHECK==false && $nObj.bCHASED_min==false) {
            await ANNE.CHASE.lookupMinNid($nObj.nid);
        }
        var nid = $nObj.nid;
        
        //rules object looks like:  
        //{let:[nid addesses], deny:[nid addesses]};
        //NONE indicates DO NOT apply that filter to the query.

        //only chase it (call the middle layer) if we haven't already done so this (bCHASE_ins)
        //note, it is complex to try and optimize by intypes and chasetypes thus we will not
        //try to limit queries based on tracking that.  The caller can determine how to limit
        //these types of queries.  (ie, like with bCHASED_Ins flag used for dates in the threejs viewer)
        if ($bMUSTPULL==true) {
            var aRelons = await ANNE.API.chaseInsGWV(nid, $nObj.trusted_adopt, $bMUSTPULL, $intypes, $chasetypes);
        }

        return $nObj.outs;
    }



    /**
     */
    getAllMyAwarnesses = async($identity_nid, $bCHASEREMEMBERZ=true)=> {
        var offset = 0;
        var max_results = 1000;     //this is set in the annode code.
        var final_relons = [];
        var maxloops = 5;
        var lastBlock = 0;
        //var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForAwareness", $identity_nid);
        lastBlock = 0;//phObj.lastBlock;
        var rememberz_nids = [];

        while (maxloops>0) {
            maxloops=maxloops-1;
            var relons_to_process = await ANNE.API.queryForAwareness($identity_nid, lastBlock, offset);
            if (relons_to_process.length) {
                var aNewRelons = await ANNE.MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "getAllMyAwarnesses");
                ANNE.TALK.printRelons(aNewRelons);
                //await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
                console.log(relons_to_process);
                if (relons_to_process.length!=max_results) {
                    final_relons = final_relons.concat(aNewRelons);
                    break;
                }
                final_relons = final_relons.concat(aNewRelons);
                offset = offset+max_results;
            }
        }

        for (var f=0; f < final_relons.length; f++) {
            if (final_relons[f].reln==MEM.rememberz.nid) {
                rememberz_nids.push(final_relons[f].to_s);
            }
        }
        if (rememberz_nids.length) {
            await this.lookupMinList(rememberz_nids, [0,1,2]);
        }

        return final_relons;
    }



    /**
     * No garden or wv here..
     * It thus uses an offset if max results are reached (1000 max)
     */
     queryForInsWithRelnTo = async($type, $reln, $to, $to_type, $offset=0)=>{
        return await ANNE.API.queryForInsWithTo($type, $reln, $to, $to_type, $offset);
    }


    /**
     * 
     * @param {*} $nidOrTxid 
     * @param {*} $rObj 
     * @returns 
     */
     getAwarenessSummary = async ($nidOrTxid, $rObj=null) => {
        alert("(CHASE.getAwarenessSummary) do we still use this?");
        console.log("(CHASE.getAwarenessSummary) nidOrTxid:"+$nidOrTxid);
        var nObj = null;

        //feelz on a txid.
        if ($rObj!=null) {
            nObj = MEMG.getNeuronByTempNid($rObj.t); //emulated neuron (for the results, but use txid to get data)
            var id = $rObj.txid;
            var aRez = await ANNE.API.___getAwarenessSummary(id);
            var awarenessMongoDoc = {empty:true};
            if (aRez.length) {
                awarenessMongoDoc = aRez[0];
                var objkeys = Object.keys(awarenessMongoDoc);
                var total = 0;
                for (var k=0; k < objkeys.length; k++) {
                    if (objkeys[k]=="pos") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neg") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neu") total+=awarenessMongoDoc[ objkeys[k] ];
                }
                awarenessMongoDoc.total=total;
                if ($rObj) $rObj.awarenessMongoDoc = awarenessMongoDoc;
            }
            else {
                if ($rObj) $rObj.awarenessMongoDoc = {empty:true};
            }
            return awarenessMongoDoc;
        }

        //standard neuron feelz
        else {
            nObj = MEMG.getOrMakeIfNew($nidOrTxid);
            if (nObj.ghost==true ||  nObj.emulated==true || nObj.preloaded==true || nObj.temp==true) {
                return {empty:true};
            }
            var id=$nidOrTxid;
            var aRez = await ANNE.API.___getAwarenessSummary(id);
            var awarenessMongoDoc = {empty:true};

            if (aRez.length) {
                awarenessMongoDoc = aRez[0];
                var objkeys = Object.keys(awarenessMongoDoc);
                var total = 0;
                for (var k=0; k < objkeys.length; k++) {
                    if (objkeys[k]=="pos") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neg") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neu") total+=awarenessMongoDoc[ objkeys[k] ];
                }
                awarenessMongoDoc.total=total;

                if (nObj) nObj.awarenessMongoDoc = awarenessMongoDoc;
            }
            else {
                if (nObj) nObj.awarenessMongoDoc = {empty:true};
            }
            return awarenessMongoDoc;
        }
    }


    /**
     * 
     * @param {*} $to_nid 
     * @param {*} $relns 
     * @returns 
     */
    getSpecificAwarenesses=async($to_nid, $relns)=>{
        var nOBJ = MEMG.getNeuronByNid($to_nid);
        if (nOBJ.ghost==true) return [];
        var aRelons = await ANNE.API.getSpecificAwarenesses($to_nid, $relns);
        return aRelons;
    }


    /**
     * 
     * **** REDO THIS FOR ANNODE CHANGES ****
     * 
     * 
     * We are looking for a specifc type of inbound relon here..           
     * Used in dynamic tree generation when looking for a non-standard      
     * child.. like a Ford150 which is a model that has a TYPE_BE of make to
     * ford.                                                               
     * We cannot use our ins type optimizer since this is qualified by the 
     * reln.  We must depend upon our HISTORY tracker to optimize this type
     * of function.  If, however, we've already chased the ins for these  
     * types, then it is presumed we'd already have the reln for that    
     * schmema too.  SO, we can throttle back on queries but not alter the  
     * ins tracker from it.                                                 
     */
     chaseInsWithReln=async($nObj, $reln, $intype, $chasetypes=[0,1,2], $desiredParentNOBJ=null)=>{
        var intypes = [parseInt($intype)];
        if ($nObj.trusted_adopt==null) $nObj.trusted_adopt = [];
        var aRelons = await ANNE.API.chaseInsWithRelnGWV($nObj.nid, $reln, $nObj.trusted_adopt, intypes, $chasetypes, $desiredParentNOBJ);
        return aRelons;
    } 


    /**
     * @author {Mr.Scatman}
     */
    searchTMDBByString=async($searchVal, $cb=null, $validMatchTypes=["person"])=> {
        $searchVal = $searchVal.trim();
        //var validMatchTypes = ["person"]; //"company" 
        var matches = [];
        console.log("search moviedb for "+$searchVal);
        this.searchTMDBMulti($searchVal, function($rez) {
            console.log($rez);
            if ($rez.length==0) {
                if ($cb!=null) $cb([], $searchVal);
            }
            else {
                var aMatches = [];
                for (var r=0; r <$rez.length; r++) {
                    var rez = $rez[r];
                    if ($validMatchTypes.indexOf(rez.media_type)!==-1) {
                        aMatches.push(rez);
                    }
                }
                if ($cb!=null) $cb(aMatches, $searchVal);
            }
        });
    }



    /**
     * @author {Mr.Scatman}
     */
    searchTMDBMulti = async ($textString, $callback, $append=null)=> {
        var params = {
            "lang": "en",
            "timezone": "America/New_York",
            "adult" : false
        };
        console.log($textString);
        params.query=$textString;
        this.tmdb.call("/search/multi", params, 
            function(data) {if($callback) $callback(data.results);}, 
            function (err) {if($callback) $callback(data);}
        );
    }


    /**
     * @author {Mr.Scatman}
     */
    getMovieDBPerson= async ($personID, $callback, $appended=null) => {
        console.log("(getMovieDBPerson) person:"+$personID);
        var params = {
            "lang": "en",
            "timezone": "America/New_York",
            "adult" : false
        };
        if ($appended==null) {
            var append = ['movie_credits', 'tv_credits', 'combined_credits', 'external_ids', 'images', 'tagged_images'];
            params['append_to_response'] = [];
            params['append_to_response'] = append;
        }
        else params['append_to_response'] = $appended;
        this.tmdb.call("/person/"+$personID, params, 
            function(data) {if($callback) $callback(data, $personID, "person");}, 
            function (err) {if($callback) $callback(err, $personID, "person");}
        );
    }


    //tmdb caller.
    //"f6c6c0a4bff37a87365e3099e7caf7ff"  (ANNE's key)
    tmdb = {
        "api_key": "enter-your-tmdb-key",
        "base_uri": "https://api.themoviedb.org/3",
        "images_uri": "https://image.tmdb.org/t/p",
        "timeout": 7000,
        call: function($url, $params, $success, $error){
            var params_str ="api_key=f6c6c0a4bff37a87365e3099e7caf7ff";
            for (var key in $params) {
                if ($params.hasOwnProperty(key)) {
                    params_str+="&"+key+"="+encodeURIComponent($params[key]);
                }
            }
            var xhr = new XMLHttpRequest();
            xhr.timeout = 7000;
            xhr.ontimeout = function () {
                console.log("Request timed out: " + $url +" "+ params_str);
                console.log("plz try again");
                if ($error) $error({success:false, error_msg:"request timed out, please try again"});
            };
            var url = "https://api.themoviedb.org/3" + $url + "?" + params_str;
            console.log("url:"+url);
            xhr.open("GET", url, true);
            xhr.setRequestHeader('Accept', 'application/json');
            xhr.responseType = "text";
            xhr.onreadystatechange = function () {
                if (this.readyState === 4) {
                    if (this.status === 200){
                        if (typeof $success == "function") {
                            $success(JSON.parse(this.response));
                        }	
                        else {
                            console.log("success, but no callback");

                            return;
                        }
                    }
                    else {
                        var errorstr = this.response.toString();
                        if (typeof $error == "function") {    
                            if (this.status==404) {
                                $error({success:false, error_msg:"bad internet connection to source data"});
                            }
                            else if (this.status==404) {
                                $error({success:false, error_msg:"bad internet connection to source data"});
                            }
                            else $error(errorstr);
                        }
                        else {
                            console.log("no callback, error in api call");
                            return;
                        }
                    }
                }
            };
            xhr.send();
        }
    }
    

    /**
     * 
     * @param {*} $url 
     * @param {*} $method 
     * @returns 
     */
    simpleExternalCaller = async($url, $method="GET")=> {
        return new Promise((resolve, reject) => {
            var request = new XMLHttpRequest();
            request.open($method, $url, true);
            request.onload = function () {
                var data = JSON.parse(this.response)
                if (request.status >= 200 && request.status < 400) {
                    return resolve(data);
                } 
                else {
                    return resolve("");
                }
            }
            request.onerror = function() {
                console.log("there has been an error in the send to url:"+$url);
            }
            request.send()
        });
    }
}

const CHASE = new anne_chase ();
CHASE.is_init(CHASE);

/************** (from file ./js/network/CHASE.js)**/ 

class anne_alert{
    constructor($title, $body, $cb_confirm=null, $cb_no=null, $cb_close=null){
        var el_alert_header = document.getElementById("alerta_header");
        el_alert_header.innerHTML = $title;

        var el_alert_text = document.getElementById("alerta_text");
        el_alert_text.innerHTML = $body;

        var el_alert_left_btns =  ANNE.A.empty("alerta_left_btns")
        var el_alert_right_btns =  ANNE.A.empty("alerta_right_btns")
        
        var el_alert_close = document.getElementById("alerta_header_close");
        el_alert_close.addEventListener("click", function($e){
            if ($cb_close!=null) $cb_close();
            ANNE.A.hide("alerta");
        })

        var btns = [];

        //confirm callback.  (if exists make confirm/yes button)
        if ($cb_confirm!=null) {
            var btnconf = document.createElement("button");
            btns.push(btnconf);
            btnconf.classList = "alerta-btn-confirm";
            btnconf.innerHTML = "yes";
            el_alert_left_btns.append(btnconf);
            el_alert_left_btns.addEventListener("click", function($e){
                $e.stopPropagation();
                $e.preventDefault();
                for (var b=0; b<btns.length;b++){
                    ANNE.A.removeSelf(btns[b]);
                }
                el_alert_text.innerHTML="";
                el_alert_header.innerHTML="";
                $cb_confirm();
                ANNE.A.hide("alerta");
            });
            ANNE.A.show(el_alert_left_btns);
        }

        //no callback (if exists make no button)
        if ($cb_no!=null) {
            var btnno = document.createElement("button");
            btns.push(btnno);
            btnno.classList = "alerta-btn-no";
            btnno.innerHTML = "no";
            el_alert_right_btns.append(btnno);
            el_alert_right_btns.addEventListener("click", function($e){
                $e.stopPropagation();
                $e.preventDefault();
                for (var b=0; b<btns.length;b++){
                    ANNE.A.removeSelf(btns[b]);
                }
                el_alert_text.innerHTML="";
                el_alert_header.innerHTML="";
                $cb_no();
                ANNE.A.hide("alerta");
            });
            ANNE.A.show(el_alert_left_btns);
        }

        ANNE.A.fadeIn("alerta");
    }
}


class anne_toggle_obj {
    constructor($divObj) {
        var divObj = $divObj.div;
        var btnObj = $divObj.btn;

        this.div_el = null; //will be set to the el that is made for this anne_toggle_obj

        if (divObj.id!=undefined) this.id = divObj.id;
        else this.id="";

        if (divObj.title!=undefined) this.title = divObj.title;
        else this.title="";
        
        if (divObj.style!=undefined) this.style = divObj.style;
        else this.style="";

        if (divObj.text!=undefined) this.text = divObj.text;
        else this.text="";
        
        if (divObj.div_class!=undefined) this.div_class = divObj.div_class;
        else this.divclass="";
        
        if (divObj.cb!=undefined && typeof divObj.cb==="function") this.cb = divObj.cb;
        else this.cb=null;

        //btn for the toggling.
        this.btn_el = null;
        this.btn_toggle_id = "";
        this.btn_text = "";
        this.btn_class = "";
        this.btn_tnid = "";

        if (btnObj!=undefined) {
            if (btnObj.id!=undefined) this.btn_toggle_id = btnObj.id;
            if (btnObj.text!=undefined) this.btn_text = btnObj.text;
            if (btnObj.btn_class!=undefined) this.btn_class = btnObj.btn_class;
            if (btnObj.btn_tnid!=undefined) this.btn_tnid = btnObj.btn_tnid;   
        }
    }
}


/**
 * 
 */
class anne_toggle_set {
    constructor($objs) {
        this.atos = [];
        if ($objs.length) {
            for (var a=0; a < $objs.length; a++) {
                var ato = new anne_toggle_obj($objs[a]);
                this.atos.push(ato);
            }
        }
    }

    /**
     * 
     */
    makeToggleSetFromAwarePol=function(){
    }

    /**
     * 
     */
    makeToggleSetFromAwareBucket=function(){
    }

}


/**
 * 
 */
class anne_select_option {
    constructor($selVal="", $selDisplayVal="", $titleHTML="", $bodyHTML="", $val2=""){
        this.selval = $selVal;
        this.selDispVal = $selDisplayVal;
        this.titleHTML = $titleHTML;
        this.bodyHTML = $bodyHTML;
        this.val2 = $val2;
        this.gui_id = UTILS.randHex(8);
        this.el_opt = null;
    }
}

class anne_select {
    constructor($w="400px", $el, $options, $misc) {
        this.width = $w;
        this.el = ANNE.A.__getEl($el);
        
        if (this.el == null) return null;
        
        this.options = $options;

        if ($misc.defaultLbl!=undefined) this.defaultLbl = $misc.defaultLbl;
        else this.deafultLbl = "";

        if ($misc.addNewCb!=undefined) this.addNewCb=$misc.addNewCb;
        else this.addNewCb = null;

        if ($misc.cb!=undefined) this.cb = $misc.cb;
        else this.cb=null;

        if ($misc.cb_ddopen!=undefined) this.cb_ddopen = $misc.cb_ddopen;
        else this.cb_ddopen = null;

        if ($misc.cb_ddclose!=undefined) this.cb_ddclose = $misc.cb_ddclose;
        else this.cb_ddclose = null;

        if ($misc.ddclass!=undefined) this.ddclass = $misc.ddclass;
        else this.ddclass = "";

        if ($misc.max_height!=undefined) this.max_height=$misc.max_height;
        //default 660px
    }


}


/******************* anne_dom - a vanilla GUI/DOM stuff *********************
*           lighter, faster, and simpler for my style of coding.            *
****************************************************************************/
class anne_dom extends anne_singleton {
    constructor() {
        super();
    }

    /**
     * Create a button that launches a dialogue w/ various details and callbacks.
     * @param {*} $el 
     * @param {*} $btnClass 
     * @param {*} $title 
     * @param {*} $msg 
     * @param {*} $cbYes 
     * @param {*} $cbNo 
     * @param {*} $cbOk 
     */
    btnToLaunchConfirm=async($el, $btnClass="anne-btn2", $title, $msg, $cbYes=null, $cbNo=null, $cbOk=null)=>{
        this.ezButton($el, "", $title, $btnClass, async($e)=>{
            new anne_alert($title, $msg, $cbYes, $cbNo, $cbOk);
        });
    }


    /**
     * 
     * @param {*} $wPx 
     * @param {function} $clickCB
     * @returns 
     */
    memeframer = function($wPx="900px", $clickCB=null, $name="testjpg") {
        const el_meme = document.createElement("div");
        el_meme.classList="meme-container";
        el_meme.style['max-width']=$wPx;
        el_meme.style['width']=$wPx;
        
        var row = document.createElement("div");
        row.style['min-width']='100%';
        row.classList = "meme-row";
        el_meme.appendChild(row);

        el_meme.setAttribute("file_name", $name);

        var img = document.createElement("img");
        img.src="./imgs/anne_logo.png";
        img.setAttribute("alt", "logo");
        img.classList="meme-logo";
        row.appendChild(img);

        const btn = document.createElement("button");
        btn.classList="meme-btn-make";
        btn.innerHTML="make meme!"
        row.append(btn);
        
        var row2 = document.createElement("div");
        row2.style['min-width']='100%';
        row2.classList = "meme-body";
        el_meme.appendChild(row2);


        if ($clickCB!=null && typeof $clickCB==="function") {
            btn.addEventListener("click", (e)=>{
                ANNE.A.hide(btn);
                var name = el_meme.getAttribute("file_name");
				$clickCB(el_meme, name);
			});
        }

		return {el:el_meme, body:row2};
    }


    buildGridNeuron = function($el, $test, $class, $afile=null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("grid-neuron-holder");
        var neuron = document.createElement("div");
        neuron.classList.add("grid-neuron");
        
        if ($afile!=null) {
            ANNE.A.doImageB64(neuron, "", $afile, 0, 99, "", 99);
        }
        else {
			neuron.innerHTML = $test;
			neuron.classList.add($class);
		}
        el.appendChild(neuron);
    }


    buildGridArrow=function($el, $arrowClass, $lbl="test") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("grid-"+$arrowClass); 
        var lbl = document.createElement("div");
        lbl.innerHTML = $lbl;
        lbl.classList.add("grid-arrow-lbl");
        lbl.classList.add($arrowClass+"-text");
        el.appendChild(lbl);
        return;
    }


    /**
     * [x][->][x][<-][x]
     * @param {*} $el 
     */
    build1con1con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }
        el.style['max-width']=$w;
        el.style['width']=$w;
        var grid = this.buildDOMGrid(el, 5, 1);
        console.log(grid);
        
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow", $m0[0].afile);
        this.buildGridArrow(grid["i1"], "right", $m0[0].reln_text);
        this.buildGridNeuron(grid["i2"], $m0[1].text, "ggreen", $m0[1].afile);
        this.buildGridArrow(grid["i3"], "left", $m1[0].reln_text);
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue", $m1[0].afile);
        grid.parent_el = el;
        return grid;
    }
    

    build1con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "ne", "");
        this.buildGridArrow(grid["i6"], "right", $m0[0].reln_text);
        this.buildGridNeuron(grid["i7"], $m0[1].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");

        this.buildGridArrow(grid["i13"], "horizontal", $m1[1].reln_text);
        this.buildGridArrow(grid["i12"], "left-up", "");
        
        grid.parent_el = el;
        return grid;
    }



    build2con1con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gblue");
        this.buildGridArrow(grid["i11"], "horizontal", $m0[1].reln_text);
        this.buildGridArrow(grid["i12"], "right-up", "");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "nw", $m1[0].reln_text);
        this.buildGridArrow(grid["i8"], "left", "");
        this.buildGridNeuron(grid["i7"], $m0[2].text, "ggreen");
        
        grid.parent_el = el;
        return grid;
    }

    /**
     * [a]     [b]
     *  |       |
     * [c]     [d]
     *  |       |   
     *   ->[e]<-
     * @param {*} $el 
     * @param {*} $w 
     * @returns 
     */
    build2con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "ne", "");
        this.buildGridArrow(grid["i16"], "right", $m0[1].reln_text);
        this.buildGridNeuron(grid["i17"], $m0[2].text, "ggreen");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "nw", "");
        this.buildGridArrow(grid["i18"], "left", $m0[1].reln_text);
        

        grid.parent_el = el;
        return grid;
    }

    build3con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "horizontal", $m0[2].reln_text);
        this.buildGridArrow(grid["i22"], "right-up", "");
        this.buildGridNeuron(grid["i17"], $m0[3].text, "ggreen");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i13"], "left", "");
        this.buildGridArrow(grid["i12"], "left-down", "");
        
        
        grid.parent_el = el;
        return grid;
    }


    build2con3con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i11"], "right", $m0[1].reln_text);
        this.buildGridArrow(grid["i12"], "right-down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i17"], $m0[2].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[1].reln_text);
        this.buildGridNeuron(grid["i24"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i23"], "left", "");
        this.buildGridArrow(grid["i22"], "left-up", "");
    
        grid.parent_el = el;
        return grid;
    }


    build3con3con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "right", $m0[2].reln_text);
        this.buildGridNeuron(grid["i22"], $m0[3].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[1].reln_text);
        this.buildGridNeuron(grid["i24"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i23"], "left", "");
    
        grid.parent_el = el;
        return grid;
    }


    /**
     * [x0]   [y1]-->[y3]
     *  |      |      |
     * [x1]   [y0]   [y4]
     *  |             |
     * [x2]-->[x3]->[match]  
     * @param {*} $el 
     * @param {*} $m0 
     * @param {*} $m1 
     * @param {*} $w 
     * @returns 
     */
    build4con4con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "right", $m0[2].reln_text);
        this.buildGridNeuron(grid["i22"], $m0[3].text, "gyellow");
        this.buildGridArrow(grid["i23"], "right", $m0[3].reln_text);
        this.buildGridNeuron(grid["i24"], $m0[4].text, "ggreen");

        
        this.buildGridNeuron(grid["i12"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i7"], "up", $m1[0].reln_text);
        this.buildGridNeuron(grid["i2"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i3"], "right", $m1[1].reln_text);
        this.buildGridNeuron(grid["i4"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[2].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[3].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[3].reln_text);

        grid.parent_el = el;
        return grid;
    }



    buildDOMGrid=function($el, $flexWidth = 5, $rows = 5, $pre="gridx") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var grid = new Object();
        el.classList.add("grid-container");
        for (let i = 0; i < $flexWidth * $rows; i++) {
            const gridSquare = document.createElement('div');
            gridSquare.classList.add('gridSquare');
            gridSquare.style.setProperty('--flexWidth', $flexWidth);
            el.appendChild(gridSquare);
            var r = Math.floor((i+$flexWidth)/$flexWidth)-1;
            var row = grid["r"+r];
            if (row==undefined) {
                row = new Object();
                grid["r"+r]=row;
            }
            var c = (i+$flexWidth) % $flexWidth;
            el.setAttribute("id",$pre+"_"+r+"_"+c)
            row["c"+c] = gridSquare;
            grid["i"+i]=gridSquare;
        }

        return grid;
    }


    
    /**
     * Make a nifty little HTML DD for the users DD.
     * //works with .text/.nickname and .nid/.public_nid
     * @param {*} $el 
     * @param {*} $users 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     */
    makeHTMLDD_fromUserObjs = function($w="400px", $el, $users, $configs) {
        var options = [];
        for (var n=0; n < $users.length; n++) {
            var user = $users[n];
            var nid = user.public_nid;
            if (nid==undefined) nid = user.nid;
            if (nid==undefined || nid==null) {
                continue;
            }
            var userNOBJ = MEMG.getOrMakeIfNew(nid);
            var lastKnownAnnies = user.last_known_annies;
            if (lastKnownAnnies==undefined || lastKnownAnnies==null) lastKnownAnnies = "0";
            //else 
            var titleHTML = "";
            var bodyHTML = "";
            var sel_disp_val = "";
            var text = user.text;
            if (text==undefined || text==null || text == "") {
                text = user.nickname;
                if (text==undefined || text==null) text="";
            }
            if (text!="") {
                titleHTML ='<span class="acct-name">'+text+'</span>';
                sel_disp_val = text;
            }
            else {
                titleHTML ="<span class='acct-name'>"+nid+"</span>";
                sel_disp_val = nid;
            }

            bodyHTML = "<span class='bal-small bal-"+userNOBJ.temp_nid+"'>"+UTILS.displayAnneValue(lastKnownAnnies)+"<span style='font-size:1.6rem;'>&#x1d110;</span></span>";
            
            var opt = new anne_select_option(nid, sel_disp_val, titleHTML, bodyHTML);
            options.push(opt);
            user.gui_id = opt.gui_id;
        }
        var ASELECT  = new anne_select($w, $el, options, $configs);
        this.makeHtmlDD(ASELECT);
    }


    
    /**
     * make a dropdown of annodes
     */
    makeANNODEDD = async($el="main_annodes", $configs=null)=> {
        var configs = $configs;
        if (configs==null) configs = {
            defaultLbl:ANNE._annodesLS.last_uri,
            ddclass:"annodedd",
            addNewCb:function(){
                console.log("add new item")
            },
            cb:async($annode_nid, $uri)=>{
                console.log("select new annode! "+$annode_nid);
                if ($annode_nid!=undefined && $annode_nid!=null && $uri!=undefined && $uri!=null) {
                    if (typeof ANNE.setNewAnnodeURI==="function") {
                        ANNE.setNewAnnodeURI($annode_nid, $uri);
                    }
                }
            //app_main
            }, cb_ddopen:function(){
                ANNE.A.smartDefocus(true);
            }, cb_ddclose:function() {
                ANNE.A.smartDefocus(false);
            }
        };
        return ANNE.A.makeHTMLDD_fromAnnodeObjs("300px", $el, ANNE._annodesLS.annodes, configs);
    }

    removeAllButTopOption = function($el) {
        var el = this.__getEl($el);
        if (el==null) return;
        const optionsCount = el.options.length;
        for (let i = optionsCount - 1; i > 0; i--) {
            el.remove(i);
        }
    }

    getDDSelectedValue=function($el) {
        var el = this.__getEl($el);
        if (el==null) return null;
        if (el.options==undefined) return null;
        var opt = el.options[ el.selectedIndex];
        if (opt==null) return null;
        if (opt==undefined) return null;
        return opt.value;
    }
    getDDSelectedAttr=function($el, $attr) {
        var el = this.__getEl($el);
        if (el==null) return null;
        var opt = el.options[ el.selectedIndex];
        if (opt==null) return null;
        if (opt==undefined) return null;
        return opt.getAttribute($attr);
    }

    setDDToIndex=function($el, $index=0) {
        var el = this.__getEl($el);
        if (el==null) return;
        el.selectedIndex = $index;
    }
    setDDToValue=function($el, $value) {
        var el = this.__getEl($el);
        if (el == null) return;
        for (var i = 0; i < el.options.length; i++) {
            if (el.options[i].value == $value) {
                el.selectedIndex = i;
                return;
            }
        }    
    }


    spinnerInject = function($el, $bWIPE=false) {
        var el = this.__getEl($el);
        if (el==null) return;
        if (el.id==null || el.id=="") el.id = ANNE.UTILS.randHex(4);
        var el_spinner = document.createElement("span");
        el_spinner.id = el.id+"_s";
        el_spinner.classList = "inlinespinner";
        if ($bWIPE==true) ANNE.A.emptySelf(el);
        el.appendChild(el_spinner);
        ANNE.A.show(el);
    }
    
    spinnerRemove = function($el, $bREMOVEPARENT=false) {
		var el = this.__getEl($el);
        if (el==null) return;
        var el_spinner = document.getElementById(el.id+"_s");
        if(el_spinner==null) {
            var inlinespinner = el.querySelectorAll(".inlinespinner");
            if (inlinespinner!=null) ANNE.A.removeSelf(inlinespinner);
        }
        else {
            ANNE.A.removeSelf(el_spinner);
        }
        if ($bREMOVEPARENT==true) {
            ANNE.A.removeSelf(el);
        }
    }

    makeANNODEDD_first_time = async($el="main_annodes", $configs=null)=> {
        var configs = $configs;
        if (configs==null) configs = {
            defaultLbl:"select annode",
            ddclass:"annodedd",
            addNewCb:function(){
                console.log("add new item");
            },
            cb:async($annode_nid, $uri)=>{
                console.log("select new annode! "+$annode_nid);
            //app_main
            }, cb_ddopen:function(){
                ANNE.A.smartDefocus(true);
            }, cb_ddclose:function() {
                ANNE.A.smartDefocus(false);
            }
        };

        return ANNE.A.makeHTMLDD_fromAnnodeObjs("300px", $el, ANNE._annodesLS.annodes, configs);

    }



    /**
     * 
     * @param {*} $bDEFOCUS 
     */
    smartDefocus = async($bDEFOCUS=true)=> {
        if ($bDEFOCUS==true) {
            var el = ANNE.A.addClass("scrollable_"+this.ID, "defocus");
			if (el==null) return;
            el.setAttribute("disabled",true);
        }
        else {
            var el = ANNE.A.removeClass("scrollable_"+this.ID, "defocus");
			if (el==null) return;
            el.removeAttribute("disabled",false);
        }
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $annodes 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     */
    makeHTMLDD_fromAnnodeObjs = function($w="400px", $el, $annodes, $configs) { 
        var options = [];
        for (var n=0; n < $annodes.length; n++) {
            var annode = $annodes[n];
            var titleHTML = annode.nickname+' - '+annode.nid;
            var bodyHTML = annode.uri;
            var sel_disp_val = annode.nickname+" "+annode.uri;
            var opt = new anne_select_option(annode.nid, sel_disp_val, titleHTML, bodyHTML, annode.uri);
            options.push(opt);
        }
        var ASELECT  = new anne_select($w, $el, options, $configs);
        this.makeHtmlDD(ASELECT);
        return ASELECT;
    }



    /**
     * 
     * @param {*} $el 
     * @param {*} $options
     * @param {*} config $defaultLbl, $cb
     * 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     * @returns 
     */
    makeHtmlDD=function($ASELECT) {
        if ($ASELECT.el==null) return;
        var sb = null;  //select box

        if ($ASELECT.el.classList.contains("selectbox")==false) {
            sb = document.createElement("div");
            sb.classList = "selectbox";
            $ASELECT.el.appendChild(sb);
        }
        else sb = $ASELECT.el;

        sb.style.width = $ASELECT.width;
    
        if ($ASELECT.addNewCb!=null) {
            //var plusSpan = document.createElement("span");
            //$ASELECT.el.parentNode.appendChild(plusSpan);
            //plusSpan.innerHTML = "[+]";
            //plusSpan.classList = "btn-select-add-option"; 
        }
        
        const optionsContainer = document.createElement("div");
        optionsContainer.classList = "options-containersb";
        sb.appendChild(optionsContainer);

        if ($ASELECT.max_height!=undefined) {
            sb.style['max-height']=$ASELECT.max_height;
            optionsContainer.style['max-height']=$ASELECT.max_height;
            optionsContainer.setAttribute("mh_backup", $ASELECT.max_height);
        }
		else {
			sb.style['max-height']='50vh';
            optionsContainer.style['max-height']='50vh';
            optionsContainer.setAttribute("mh_backup", '50vh');
		}

        const selectedsb = document.createElement("div");
        selectedsb.classList = "selectedsb";

        if ($ASELECT.ddclass!="") {
            selectedsb.classList.add($ASELECT.ddclass);
        }
        selectedsb.style.width = $ASELECT.width;

        sb.appendChild(selectedsb);
        if ($ASELECT.defaultLbl!="") {
            var lbl_sp3 = document.createElement("span");
            lbl_sp3.classList = "sbh3 pr30";
            lbl_sp3.innerHTML = $ASELECT.defaultLbl;
            selectedsb.append(lbl_sp3);
        }

        //create options
        for (var i=0; i  < $ASELECT.options.length; i++) {
            var opt = $ASELECT.options[i];
            var selval = opt.selval;
            var val2 = opt.val2;
            const divopt = document.createElement("div");
            optionsContainer.appendChild(divopt);
            divopt.classList = "optionsb";
            if (selval!=null) divopt.setAttribute("selval", selval);
            if (val2!=null) divopt.setAttribute("val2", val2);
            if (opt.selDispVal!=null) divopt.setAttribute("dispval", opt.selDispVal);
            
            var divopt_rad = document.createElement("input");
            divopt_rad.setAttribute("type", "radio");
            divopt_rad.classList = "radiosb";
            divopt_rad.id = opt.gui_id;
            divopt.appendChild(divopt_rad);
            
            var divopt_lbl = document.createElement("label");
            divopt_lbl.setAttribute("for", divopt_rad.id);
            divopt_lbl.classList = "opt-lbl";
            divopt.appendChild(divopt_lbl);
            
            var lbl_sp3 = document.createElement("span");
            lbl_sp3.classList = "sbh3";
            lbl_sp3.innerHTML = opt.titleHTML;
            divopt_lbl.append(lbl_sp3);
            
            var lbl_sp5 = document.createElement("span");
            lbl_sp5.classList = "sbh5";
            lbl_sp5.innerHTML = opt.bodyHTML;
            divopt_lbl.append(lbl_sp5);
            opt.el_opt = divopt;

            opt.el_opt.addEventListener("click", ($e) => {
				console.log("(option clicked!)");
                $e.stopPropagation();
                $e.preventDefault();
				var selval = divopt.getAttribute("selval");
                var val2 = divopt.getAttribute("val2");
                if (val2==undefined) val2 = null;
                var dispval =  divopt.getAttribute("dispval");
				//selectedsb.innerHTML = divopt.querySelector("label").innerHTML;
				//selectedsb.innerHTML = selval;
                ANNE.A.empty(selectedsb);
                var lbl_sp3 = document.createElement("span");
                lbl_sp3.classList = "sbh3 pr30";
                if (dispval) lbl_sp3.innerHTML = dispval;
                else lbl_sp3.innerHTML = selval;
                selectedsb.append(lbl_sp3);

                document.removeEventListener("click", clickout);

                optionsContainer.classList.add("hide");
                optionsContainer.style['z-index']=-1;
				optionsContainer.classList.remove("active");
                optionsContainer.style['pointer-events']='none';

                if ($ASELECT.cb!=null) {
                    if ($ASELECT.cb_ddclose!=null) {
                        $ASELECT.cb_ddclose();
                    }
                    $ASELECT.cb(selval, val2);
                }
			});
        }

        const clickout = function(){
            console.log("clickout");
            optionsContainer.style['z-index']=-1;
            optionsContainer.style['max-height']=0;
            optionsContainer.style['pointer-events']='none';
            optionsContainer.classList.add("hide");
            document.removeEventListener("click", clickout);
            optionsContainer.classList.toggle("active");
            if ($ASELECT.cb_ddclose!=null) {
				$ASELECT.cb_ddclose();
	        }
		}

        optionsContainer.addEventListener("mouseout", $e=> {
            if (optionsContainer.classList.contains("active")==true) {
                const fromElement = $e.relatedTarget;
                if (!optionsContainer.contains(fromElement)) {
                    clickout();
                }
            }
        });

        selectedsb.addEventListener("click", ($e) => {
            $e.stopPropagation();
            $e.preventDefault();

            optionsContainer.classList.remove("hide");
            optionsContainer.style['pointer-events']='all';

            //this will add hide.
			if (optionsContainer.classList.contains("active")==true) {
                console.log("oc active, make inactive");
                clickout();
            }

            //this will have no hide.
            else {
                console.log("oc inactive, make active");
                document.addEventListener("click", clickout);
                optionsContainer.style['z-index']=9999;
                optionsContainer.style['opacity']=1;
                var mh = optionsContainer.getAttribute("mh_backup");
                optionsContainer.style['max-height']=mh;
                optionsContainer.classList.toggle("active");
                if ($ASELECT.cb_ddopen!=null) {
					$ASELECT.cb_ddopen();
				}
            }   
           
		});
    }

    

    /**
     * Standardized link color maker based on the neuron (not the relon)
     * This will be used to generated links for neurons.
     */
    getLinkColorClass = function($nObj) {
        if ($nObj==undefined || $nObj==null) return "link-black";
        if ($nObj.ghost==true) return "link-blue";

        //presume ANNE ownership if not set (in the js)
        if (ANNE.isANNE($nObj.owned_by)==true 
            || $nObj.owned_by==ANNE.MEM.ANNE_ACCTID) return "link-yellow";  //anne yellow
        
        else {
            if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
                //possible hanai could be here.  Look for pending hanai.
                if ($nObj.bPENDING_HANAI==undefined) {
                    var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
                    if (pending_hanai!=null) {
                        $nObj.bPENDING_HANAI=true;
                        return "link-pink";  //hanai pink
                    }
                }
                return "link-red";   //red for self owned
            }
            else {
                return "link-grey";  //black for someone else's neuron (not ANNE's).
            }
        }
    }


    /**
     */
    getNeuronColorClass = function($nObj, $pre="title-", $post, $default="other") {
        if ($pre!="" && $pre[ $pre.length-1]!="-") $pre = $pre+"-";
        if ($post!="" && $post[ $post.length-1]!="-") $post = "-"+$post;

        if (ANNE.isANNE($nObj.owned_by)==false && $nObj.ghost==false && $nObj.bPENDING_HANAI==undefined) {
            var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
            if (pending_hanai!=null) {
                $nObj.bPENDING_HANAI=true;
                return $pre+"hanai"+$post;  
            }
        }
        else if ($nObj.bPENDING_HANAI==true) return $pre+"hanai"+$post; 

        else if ($nObj.ghost==true) {
            return $pre+"ghost"+$post;  
        }
        else if (ANNE.isANNE($nObj.owned_by)==true) {
            return $pre+"anne"+$post;  
        }
        else if ($nObj.owned_by==USER._userNObj.nid) {
            return $pre+"self"+$post; 
        }
        return $pre+$default+$post;  //hanai pink
    }


    /**
     */
    addClass = function ($el, $classname) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($classname) == false) el.classList.add($classname);
        return el;
    }

    
    /**
     * Quickly add an ai-ico class image (see anne-icos.css).
     * 
     * @param {*} $el 
     * @param {*} $icoClass 
     * @returns 
     */
    addIco = function ($el, $icoClass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.classList = $icoClass;
        el.appendChild(span);
    }


    addRemoveClass = function ($el, $addclass, $removeclass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        if (el.classList.contains($removeclass) == true) el.classList.remove($removeclass);
        return el;
    }

    addRemoveClasses = function ($el, $addclass, $removeClasses) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var r = 0; r < $removeClasses.length; r++) {
            if (el.classList.contains($removeClasses[r]) == true) el.classList.remove($removeClasses[r]);
        }
        if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        return el;
    }

    /**
     * Same as addRemoveClass but checks if the removeclass is on.
     */
    addRemoveClassIf = function ($el, $addclass, $removeclass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($addclass) == false) {
            if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        }
        if (el.classList.contains($removeclass) == true) el.classList.remove($removeclass);
        return el;
    }


    animateProp = function ($el, $prop, $propVal, $timestring = "0.5s ease-out", $removeClass = "hide") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($removeClass) el.classList.remove($removeClass);
        el.style[$prop] = $propVal;
        el.style.transition = $prop + " " + $timestring;
    }
    animateProps = function ($el, $jsonprops, $dur = ".5s", $delay = "0", $ease = "ease-out", $removeClass = "hide") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($removeClass) el.classList.remove($removeClass);
        var keys = Object.keys($jsonprops);
        var propstrs = [];
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsonprops[prop];
            propstrs.push(prop);
        }
        el.style['transition-property'] = propstrs.join(",");
        el.style['transition-duration'] = $dur;
        el.style['transition-timing-function'] = $ease;
        if ($delay != "") el.style['animation-delay'] = $delay;

        //set the props after the transition styling.
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsonprops[prop];
            el.style[prop] = val;
        }
    }





    /**
     * Disambig chases of information from external lookups.
     * @param {*} $neurons 
     * @param {*} $val 
     * @param {*} $ambigObj 
     * @param {*} $el_console 
     * @param {*} $resolveAmbigs 
     * @returns 
     */
    disambigInner=async($neurons, $val, $ambigObj, $el_console, $resolveAmbigs)=> {
        if ($neurons==undefined || $neurons==null || $neurons.length==0) {
            return $resolveAmbigs(); //call the next one.
        }
        var el_dchoice = document.createElement("div");
        var sp = document.createElement("div");
        sp.innerHTML="original text: <b>"+$val+"</b>";
        sp.style.display="block";
        el_dchoice.appendChild(sp);

        //the choice (and all options)
        $el_console.appendChild(el_dchoice);

        for (var n=0; n < $neurons.length; n++) {
            var el_choice = document.createElement("div");
            el_choice.classList = "neuron-choice";
            el_dchoice.appendChild(el_choice);
            
            var el_choice_info = document.createElement("div");
            el_choice_info.classList = "neuron-choice-info";
            
            var nObj = $neurons[n];
            await ANNE.GUI_STUFF.disambigHTMLStr(el_choice_info, nObj, ["sp"]);
            el_choice.appendChild(el_choice_info);
            
            var el_choice_btn = document.createElement("div");
            el_choice_btn.classList = "neuron-choice-btn-holder";  
            el_choice.appendChild(el_choice_btn);
            
            var el_btn = ANNE.A.ezButton(el_choice_btn, "", "choose match "+(n+1), "anne-tiny-btn btn-purple-sq btn-maxH", async(e)=>{
                var nid_x = e.target.getAttribute("nid");
                var nObjx = MEMG.getNeuronByNid(nid_x);
                
                ANNE.A.empty(e.target.parentElement, true);//empty and hide choice div.
                
                //key to set this.
                $ambigObj.jsonobj.to_s = nObjx.nid;
                return await $resolveAmbigs(); //call the next one.
            });
            el_btn.setAttribute("nid", nObj.nid);
            el_btn.setAttribute("val", $val);
            $el_console.appendChild(el_dchoice);
        }

        //skip option.
        var el_skip = document.createElement("div");
        el_skip.classList = "neuron-choice";
        el_dchoice.appendChild(el_skip);

        var el_skip_info = document.createElement("div");
        el_skip_info.classList = "neuron-choice-skip";
        el_skip.appendChild(el_skip_info);     
        var el_skip_btn = document.createElement("div");
        el_skip_btn.classList = "neuron-choice-btn-holder";  
        el_skip.appendChild(el_skip_btn);
        var el_btn = ANNE.A.ezButton(el_skip_btn, "", "SKIP THIS", "anne-tiny-btn btn-red-sq btn-maxH", async(e)=>{
            ANNE.A.empty(el_dchoice. true);//empty and hide choice div.
            $ambigObj.jsonobj.to_s = null;
            return await $resolveAmbigs(); //call the next one.
        });
        

        //make new
        var el_new = document.createElement("div");
        el_new.classList = "neuron-choice";
        el_dchoice.appendChild(el_new);
        var el_new_info = document.createElement("div");
        el_new_info.classList = "neuron-choice-new";
        el_new.appendChild(el_new_info);     
        var el_new_btn = document.createElement("div");
        el_new_btn.classList = "neuron-choice-btn-holder";  
        el_new.appendChild(el_choice_btn);
        var el_btn = ANNE.A.ezButton(el_choice_btn, "", "NEW", "anne-tiny-btn btn-blue-sq btn-maxH", async(e)=>{
            ANNE.A.empty(el_dchoice. true); //empty and hide choice div.
            console.log($ambigObj);
            $ambigObj.jsonobj.to_s = null;
            //this is going to need a callback to make a "NEW" correctly!!!!
            return await $resolveAmbigs(); //call the next one.
        });
        $el_console.appendChild(el_dchoice);
    }


    /**
     * Generalized ez annies entry.
     * i_min_val, i_max_val and i_placeholder attrs can be added to the outter button.
     * They will passthrough to the input field.
     * 
     * @param {*} $el 
     * @param {*} $cb_submit - submit callback
     * @param {*} $btnTitle - title of outter button
     * @param {*} $cb_entry - callback on entry of data (if more than just formatting of annies is desired)
     * @param {*} $idpre - Used to create a unique ezAnnecoin_entry if there is more than one in the DOM at once.
     * @returns 
     */
    ezAnnecoin_entry = function($el, $cb_submit=null, $btnTitle="accept available risk", $btn2Title="accept risk", 
                                $cb_entry=null, $idpre="unique", $btnClassStr = "noround btn-someanniesaction") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_holder = ANNE.A.ezEl(el, "div", "flexrow");
        const idpre = $idpre;
        const btn_aar = ANNE.A.ezButton(el_holder, "btn_main_"+$idpre, $btnTitle, $btnClassStr, async(e)=>{
            
            //snag passthrough attrs we put on button AFTER we created here.
            //these will be ready for this event since we created them outside of this funct on the return
            //of the gui creation. (optional attrs)
            var min_val = btn_aar.getAttribute("i_min_val");
            var max_val = btn_aar.getAttribute("i_max_val");
            var placeholder = btn_aar.getAttribute("i_placeholder");

            //create the btn clone so we get the button input button functionality
            var btnClone = btn_aar.cloneNode(true);
            const disp = document.createElement("div");
            btnClone.id = "btn_"+idpre+"_final";
            btnClone.setAttribute("tied_input", "input_"+idpre);
            ANNE.A.hide(btn_aar);
    
            const input_field = document.createElement("input");
            input_field.setAttribute("size", 24);
            input_field.setAttribute("type", "text");
            input_field.id="input_"+idpre;
            input_field.setAttribute("el_display", "annies_"+idpre+"_disp");
            input_field.setAttribute("el_submit", "btn_"+idpre+"_final");
            if (min_val!=undefined) input_field.setAttribute("min_val", min_val);
            if (max_val!=undefined) input_field.setAttribute("max_val", max_val);
            if (placeholder!=undefined) input_field.setAttribute("placeholder", placeholder);
            
            //add th input field after the btn_aar
            btn_aar.after(input_field);

            var max_btn = null;
            var min_btn = null;


            //add max btn
            if (max_val!=undefined) {
                max_btn = document.createElement("button");
                max_btn.innerHTML="max";
                max_btn.classList = 'btn-minmax';
                max_btn.addEventListener("click", function($e){
                    input_field.value = max_val;
                    ANNE.A.hide(max_btn);
                    ANNE.A.show(btnClone);
                    if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
                })
                input_field.after(max_btn)
            }

            //add max btn
            if (min_val!=undefined) {
                min_btn = document.createElement("button");
                min_btn.classList = 'btn-minmax';
                min_btn.innerHTML="min";
                min_btn.addEventListener("click", function($e){
                    input_field.value = min_val;
                    ANNE.A.hide(min_btn);
                    ANNE.A.show(btnClone);
                    if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
                })
                input_field.after(min_btn)
            }


            //add input input event
            input_field.addEventListener("input", async($e)=>{
                await ANNE.A.anne_entered($e);
                if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
            });


            //a display for the formatted/summary version of what the user typed.
            disp.classList="flexrow";
            disp.id="annies_"+idpre+"_disp";

            //if we have max, add disp after it.  else add disp after input field.
            if (max_btn!=null) max_btn.after(disp);
            else if (min_btn!=null) min_btn.after(disp);
            else input_field.after(disp);

            disp.after(btnClone);
            btnClone.innerHTML=$btn2Title;
            ANNE.A.hide(btnClone);

            //add submit callback.
            if (typeof $cb_submit=="function") {
                btnClone.addEventListener("click", async($e)=>{
                    var val = input_field.value;
                    $cb_submit(val)
                });
            }
        });
        return el_holder;
    }


    /**
     * annies entered
     * This can handle el-display (and display value entered)
     * This can handle el-submit (and show a submit btn after valid entry)
     * @param {*} $e 
     */
    anne_entered = async($e, $target=null)=>{
        var el_input = $target;
        if (el_input==null && $e!=null && $e.target!=null) {
            el_input = $e.target;
        }
        if (el_input==null) return;
        
        var val = el_input.value+"";
        var lastChar = val[val.length-1];

        var el_display = el_input.getAttribute("el_display");
        if (el_display!=null) el_display = document.getElementById(el_display);

        var el_submit = el_input.getAttribute("el_submit");
        if (el_submit!=null) el_submit = document.getElementById(el_submit);

        var minval = el_input.getAttribute("min_val");
        var maxval = el_input.getAttribute("max_val");
        
        var final_val = 0;

        if (lastChar=="k" || lastChar=="K") {         //shorthand for thousands
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000;
            el_input.value = newval;
        }
        else if (lastChar=="m" || lastChar=="M") {    //shorthand for millions
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000000;
            el_input.value = newval;
        }
        else if (lastChar=="b" || lastChar=="B") {    //shorthand for billions
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000000000;
            el_input.value = newval;
        }
        else if ("0123456789.,".indexOf(lastChar)==-1) {
            var newval = "";
            var amt = el_input.value;
            for (var a=0; a < amt.length; a++) {
                if ((a!=(amt.length-1)) && ("kKmMbB".indexOf(amt[a])!=-1)) continue;
                else if ("0123456789.,".indexOf(amt[a])!=-1) continue;
                newval = newval+amt[a];
            }
            el_input.value = newval;
        }

        var val  = el_input.value;
        if (val=="") {
            return;
        }
        final_val = parseInt(val);

        if (minval!=undefined && final_val < minval) {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.setHTML(el_display, "<span class='min-msg'>minimum "+minval+" annecoin</span>"); 
        }

        else if (maxval!=undefined && final_val > maxval) {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.setHTML(el_display, "<span class='max-msg'>max "+maxval+" annecoin</span>"); 
        }

        else if (final_val>0) {
            if (el_submit!=null) ANNE.A.show(el_submit);
            if (el_display!=null) ANNE.A.ezAnnecoinDisplay(el_display, final_val, true); 
        }
        else {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.ezAnnecoinDisplay(el_display, 0, true); 
        }
    }

    appendHTML = function($el, $html) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.innerHTML = el.innerHTML+$html;
    }

    append = function ($el, $misc, $bSHOW = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.append($misc);
        if ($bSHOW == true) this.show(el);
        return el;
    }

    appendStringList=function($el, $strArr, $size="10px", $class="") {
        var el = this.__getEl($el);
        if (el==null) {
            console.log("missing el");
            return null;
        }
        if ($strArr==undefined || $strArr==null || Array.isArray($strArr)==false || $strArr.length==0) return;
        var spanlist=document.createElement("span");
        spanlist.style['display']="inline-block";
        spanlist.style['font-size']=$size;
        if ($class!="") spanlist.classList.add($class);
        if ($class!="") {
            spanlist.innerHTML="("+$strArr.join(", ").trim()+")";
        }
        else spanlist.innerHTML="("+$strArr.join(", ").trim()+")";
        el.append(spanlist);
    }

    addEmojisFromRelTos=function($el, $rels, $title) {
        var el = this.__getEl($el);
        if (el == null) return null;

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (toNOBJ._uico!=undefined) {
                var sp = document.createElement("span");
                sp.innerHTML = toNOBJ._uico;
                sp.setAttribute("title", $title+" "+toNOBJ.bestNameHTML());
                sp.setAttribute("txid", rel.txid);
                el.appendChild(sp);
            }
        }
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    getAnum=function($str) {
        let annum = 0;
        const normalizedText = $str.normalize("NFD").replace(/[\u0300-\u036f]/g, "").toLowerCase().trim();
        for (let i = 0; i < normalizedText.length; i++) {
            const c = normalizedText.charAt(i);
            const asciiValue = c.charCodeAt(0);
            if (asciiValue < 49) continue;
            if (asciiValue <= 57) {
                annum += asciiValue - 48;
                continue;
            }
            else if (asciiValue < 97 || asciiValue > 123) {
                continue;
            }
            else annum += asciiValue - 96;
        }
        return annum;
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $text 
     * @param {*} $classpost 
     */
    doAnumWithCalc = function($el, $text, $classpost="") {
        var num = this.getAnum($text+"");
        this.doAnum($el, num+"", $classpost);
        return num;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $num 
     * @param {*} $classpost 
     * @returns 
     */
    doAnum=function($el, $num, $classpost="") {
        var el = this.__getEl($el);
        if (el == null) return null;

        if(el.classList.contains("anumbox"+$classpost)==false) el.classList.add("anumbox"+$classpost);
        el.innerHTML="";

        $num=$num+"";
        
        for (var n=0; n <$num.length; n++) {
            var sp = document.createElement("span");
            sp.classList = "anum"+$classpost;
            sp.innerHTML = $num[n];
            el.appendChild(sp);
        }
        return;
    }


    /**
     * 
     * @param {*} $rels 
     * @param {*} $title 
     * @param {*} $skips 
     * @returns 
     */
    getEmojisFromRelTos_html=function($rels, $title, $skips=[]) {
        if (ANNE.SETTINGS && ANNE.SETTINGS.bUICOS==false) return "";
        var html = "";
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if ($skips.indexOf(rel.txid)!==-1) continue;
            var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (toNOBJ._uico!=undefined) {
                var span = "<span title='"+$title+" "+toNOBJ.bestNameHTML()+"' txid='"+rel.txid+"'>"+toNOBJ._uico+"</span>";
                html+=span;
                $skips.push(rel.txid);
            }
        }
        return html;
    }


    click = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.click();
        return el;
    }

    //this clones the DOM node.
    //DOM node doesn't have to be in the DOM.. it can just be a temp made
    //el that is not yet appended.
    cloneAndAppend = function ($getel, $putel, $bCLONE = true, $bSHOWPUTEL = true) {
        if ($bCLONE == true) {
            var getel = this.__getEl($getel);
            if (getel == null) return null;
            var clone = getel.cloneNode(true);
        }
        else clone = this.__getEl($getel);
        if (clone == null) return null;
        var putel = this.__getEl($putel);
        if (putel == null) return null;
        putel.appendChild(clone);
        if ($bSHOWPUTEL == true) this.show($putel);
        return clone;
    }

    //this clones the .content of the template
    cloneAndAppendTemplate = function ($getel, $putel, $bSHOWPUTEL = true) {
        var getel = this.__getEl($getel);
        if (getel == null) return null;
        var clone = getel.content.cloneNode(true);
        var putel = this.__getEl($putel);
        if (putel == null) return null;
        putel.appendChild(clone);
        if ($bSHOWPUTEL == true) this.show($putel);
        return clone;
    }



    createElement = function ($el, $tag, $class = "", $html = "", $value = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newtag = document.createElement($tag);
        if ($class != "") newtag.classList = $class;
        if ($html != "") newtag.innerHTML = $html;
        if ($value != "") newtag.value = $value;
        el.append(newtag);
        return newtag;
    }

    /**
     */
    doImageDyno = function ($div, $imgElId = "", $src = "", $imgClass = "left_dyno_w") {
        if ($src == "") return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {anne_file} $afile - ANNE FILE
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDynoB64 = function ($div, $imgElId = "", $afile=null, $imgClass = "left_dyno_w") {
        
        if ($afile == null) return null;
        if ($afile.b64data==null) return null;

        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;
        if ($afile.b64data.indexOf("data:")!=-1) newImgEl.setAttribute('src', $afile.b64data); 
        else newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
        return newImgEl;
    }


    /**
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $src 
     * @param {*} $biggest 
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDyno2 = function ($div, $imgElId = "", $src = "", $biggest = 200, $imgClass = "left_dyno_w") {
        if ($src == "") return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('error', function () {
            console.log('Error loading image');
            return;
        });
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            var biggestD = realW;
            var bgd = "w";
            if (realH > realW) {
                biggestD = realH;
                bgd = "h";
            }
            if (biggestD > $biggest) {
                if (bgd == "w") {
                    newImgEl.setAttribute("width", $biggest);
                    var newH = $biggest / ratio;  //(>1 ratio)
                    newImgEl.setAttribute("height", newH);
                }
                else {
                    newImgEl.setAttribute("height", $biggest);
                    var newW = $biggest * ratio; //<1 ratio
                    newImgEl.setAttribute("width", newW);
                }
            }
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $afile 
     * @param {*} $biggest 
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDyno2_B64 = function ($div, $imgElId = "", $afile = null, $biggest = 200, $imgClass = "left_dyno_w") {
        if ($afile == null) return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('error', function () {
            console.log('Error loading image');
            return;
        });
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            var biggestD = realW;
            var bgd = "w";
            if (realH > realW) {
                biggestD = realH;
                bgd = "h";
            }
            if (biggestD > $biggest) {
                if (bgd == "w") {
                    newImgEl.setAttribute("width", $biggest);
                    var newH = $biggest / ratio;  //(>1 ratio)
                    newImgEl.setAttribute("height", newH);
                }
                else {
                    newImgEl.setAttribute("height", $biggest);
                    var newW = $biggest * ratio; //<1 ratio
                    newImgEl.setAttribute("width", newW);
                }

            }
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;
        if ($afile.b64data.indexOf("data:")!=-1) newImgEl.setAttribute('src', $afile.b64data); 
        else newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
        return newImgEl;
    }


    /**
     */
    doBadImg = function ($div, $imgElId = "", $src = "", $imgClass = "left_dyno_w") {
        var newImgEl = this.ezEl($div, "span", $imgClass + " img-bad-audit", "FAILED<BR>AUDIT");
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        newImgEl.style.top = "0px";
        if ($src != undefined && $src != null && $src != "") {
            newImgEl.setAttribute('bad-audit-url', $src);
        }
        var attr_fix = newImgEl.getAttribute("ssfixpostload");
        if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        return newImgEl;
    }



    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $src 
     * @param {*} $w 
     * @param {*} $h 
     * @param {*} $imgClassAdd 
     * @param {*} $max 
     * @returns 
     */
    doImage = function ($div, $imgElId, $src, $w = 600, $h = 900, $imgClassAdd = "", $max = 0) {
        console.log("src "+$src);
        
        const newImgEl = document.createElement('img');
        newImgEl.setAttribute("id", $imgElId);
        if ($imgClassAdd != "") this.__classListParse(newImgEl, $imgClassAdd);  //need to do this elsewhere
        newImgEl.onerror = function(image) {
            newImgEl.src = "";
            newImgEl.onerror = null; // Prevent infinite loop if fallback image also fails to load
            return;
        }
        
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            if ($w && $h) {
                newImgEl.setAttribute("width", $w);
                newImgEl.setAttribute("height", $h);
            }
            else if ($w) {
                if ($w < realW) {
                    if ($max && ($w / ratio) > $max) {
                        newImgEl.setAttribute("height", $max);
                    }
                    else newImgEl.setAttribute("width", $w);
                }
                else {
                    if ($max!=0 && realH > $max) {
                        newImgEl.setAttribute("height", $max);
                    }
                    else newImgEl.setAttribute("width", realW);
                }
            }
            else if ($h) {
                if ($h < realH) {
                    if ($max && (ratio * $h) > $max) {
                        newImgEl.setAttribute("width", $max);
                    }
                    newImgEl.setAttribute("height", $h);
                }
                else newImgEl.setAttribute("height", realH);
            }
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {anne_file} $afile 
     * @param {int} $w 
     * @param {int} $h 
     * @param {string} $imgClassAdd 
     * @param {int} $maxOtherD 
     * @returns 
     */
    doImageB64 = function ($div, $imgElId, $afile=null, $w = 600, $h = 900, $imgClassAdd = "", $maxOtherD = 0) {      
		const newImgEl = document.createElement('img');
        newImgEl.setAttribute("id", $imgElId);

		if ($afile==null) {
			console.log("must have an afile here");
			return newImgEl;
		}

        if ($imgClassAdd != "") this.__classListParse(newImgEl, $imgClassAdd);  //need to do this elsewhere
		
        newImgEl.onerror = function($img) {
            newImgEl.src = "";
            newImgEl.onerror = null; // Prevent infinite loop if fallback image also fails to load
            return;
        }
        
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            if ($w && $h) {
                newImgEl.setAttribute("width", $w);
                newImgEl.setAttribute("height", $h);
            }
            else if ($w) {
                if ($w < realW) {
                    if ($maxOtherD && ($w / ratio) > $maxOtherD) {
                        newImgEl.setAttribute("height", $maxOtherD);
                    } 
                    else {
                        newImgEl.setAttribute("width", $w);
                        newImgEl.setAttribute("height", $w / ratio); 
                    }
                } 
                else {
                    if ($maxOtherD != 0 && realH > $maxOtherD) {
                        newImgEl.setAttribute("height", $maxOtherD);
                        newImgEl.setAttribute("width", $maxOtherD * ratio); 
                    } 
                    else {
                        newImgEl.setAttribute("width", realW);
                        newImgEl.setAttribute("height", realH);
                    }
                }
            } 

            else if ($h) {
                if ($h < realH) {
                    if ($maxOtherD && (ratio * $h) > $maxOtherD) {
                        newImgEl.setAttribute("width", $maxOtherD);
                        newImgEl.setAttribute("height", $maxOtherD / ratio); // Adjust height to maintain aspect ratio
                    } 
                    else {
                        newImgEl.setAttribute("height", $h);
                        newImgEl.setAttribute("width", ratio * $h); // Adjust width to maintain aspect ratio
                    }
                } 
                else {
                    // If the specified height is greater than or equal to the original height
                    newImgEl.setAttribute("height", realH);
                    newImgEl.setAttribute("width", realW);
                }
            }
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;

		if ($afile.b64data==undefined || $afile.b64data==null) {
			console.log("problem with b64 data");
			return newImgEl;
		}
		
        if ($afile.b64data.indexOf("data:")!=-1) {
			newImgEl.src = $afile.b64data;
		}
        else {
			newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
		}
        return newImgEl;
    }

    emptySelfAfterChild=function($el, $childIndex=0, $bHIDE=false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var c=0; c < el.children.length; c++) {
            if (c<=$childIndex) continue;
            el.removeChild( el.children[c]);
            c=c-1;
        }
        if ($bHIDE == true) this.hide($el);
        return el;
    }

    emptySelf = function($el, $bHIDE=false){ return this.empty($el, $bHIDE);}   //i keep thinking it is named this.
    empty = function ($el, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        while (el.firstChild) el.removeChild(el.firstChild);
        el.innerHTML = "";
        if ($bHIDE == true) this.hide($el);
        return el;
    }
    emptyParent = function ($el, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_parent = el.parentNode;
        if (el_parent) this.empty(el_parent);
        else return el;
        if ($bHIDE == true) this.hide(el_parent);
        return el_parent;
    }

    emptyAllBut=function($el, $skipEl, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_parent = el.parentNode;
        for (var c=0; c < el_parent.children.length; c++) {
            var child = el_parent.children[c];
            if (child==$skipEl) continue;
            el_parent.removeChild(child);
        }
        if ($bHIDE == true) this.hide(el);
        return;
    }


    ezAlert = function ($el, $class, $id_pre, $id, $msg = "", $insertAfter = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        const alert = document.createElement("div");
        var classList = "alert fade show";
        if ($class) classList = classList + " " + $class; //could be multiple ones.
        alert.classList = classList;
        alert.setAttribute("role", "alert");
        if ($id_pre != "" && $id != "") alert.setAttribute("id", $id_pre + "_" + $id);
        else if ($id != "") alert.setAttribute("id", $id);
        if ($msg != "") alert.innerHTML = $msg;
        if ($insertAfter != null) $insertAfter.after(alert);
        else el.append(alert);
        return alert;
    }


    /**
     * HTML is silly.
     * @param {*} $el - the select el or string of the id
     * @param {String} $attr - property to extract 
     * @returns 
     */
    getSelectedProp=function($el, $attr) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectedOption = el.options[el.selectedIndex];
        if (selectedOption==null) return null;
        return selectedOption.getAttribute($attr);
    }

    //var el_lvl = ANNE.A.setHTML("annex_lvl", aplex.price_lvl);
    ezAnniesToAnneDisplay = function ($el, $val = 0, $bdclass="balance-display", $ico="fermatta1rem", $valClass="annecoin-balance-value", $bRESET=true) {
        var el = this.__getEl($el);
        if (el == null) return null;

		if ($bRESET==true) el.innerHTML="";
		
        if ($val == 0) {
            el.classList.add("text-danger-lbl");
            el.innerHTML = "0";
            return;
        }
        var annecoin = $val / ANNE.ANNE_TO_ANNIES;
        var span = ANNE.A.ezSpan(el, "", $bdclass);
        var val = UTILS.formatNumberCommas(annecoin);
        //console.log(el.id + " val:" + val);
        ANNE.A.ezSpan(span, "", $valClass, val + "");
        ANNE.A.ezSpan(span, "", $ico);
    }

    /**
    * annies are never decimal, only whole
    */
    validateAndFormatAnnecoin = function ($el_input, $displayEl, $annecoinstring, $bRESET = true) {
        var str = "";
        var lastchar = $annecoinstring[$annecoinstring.length - 1];
        var multiplier = 0;
        if (lastchar == "k") {
            multiplier = 1000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        else if (lastchar == "m") {
            multiplier = 1000000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        else if (lastchar == "b") {
            multiplier = 1000000000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        for (var s = 0; s < $annecoinstring.length; s++) {
            if ("0123456789".indexOf($annecoinstring[s]) == -1) continue;
            str = str + $annecoinstring[s];
        }
        if (multiplier != 0) str = (parseInt(str) * multiplier) + "";
        $el_input.value = str;

        if (str.length > 0) {
            this.ezAnnecoinDisplay($displayEl, str, $bRESET);
        }
    }

    
    /**
     * 
     */
    ezAddNote = function($el, $noteText, $class="small-note") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.innerHTML = $noteText;
        span.classList = $class;
        el.append(span);
    }

    displayAnnecoinWithCommas = function ($el, $annecoin = 0, $bRESET=false, $pre = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($annecoin=="" || $annecoin==undefined) $annecoin=0;
        if ($bRESET == true) ANNE.A.empty(el);
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = ANNE.UTILS.commas($annecoin);
        if ($pre != "") $pre = "&nbsp;" + $pre;
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", $pre + annecoin + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico");
    }

    ezAnnecoinDisplay = function ($el, $annecoin = 0, $bRESET = false, $pre = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($annecoin=="" || $annecoin==undefined) $annecoin=0;
        if ($bRESET == true) ANNE.A.empty(el);
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = Math.round(parseInt($annecoin)*100)/100;
        var acval = UTILS.formatNumber(annecoin);
        if ($pre != "") $pre = "&nbsp;" + $pre;
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", $pre + acval + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico");
    }

    /**
     * will round for better display
     */
    ezAnneDisplayApprox = function ($el, $val = 0, $bRESET = false, $addon=null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($bRESET == true) ANNE.A.empty(el);
        if ($val == 0) {
            var span = ANNE.A.ezSpan(el, "", "balance-display");
            ANNE.A.ezSpan(span, "", "annecoin-balance-value", 0 + "");
            ANNE.A.ezSpan(span, "", "annecoin-small-ico")
            return;
        }
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = $val / ANNE.ANNE_TO_ANNIES;
        var acval = UTILS.formatNumber(annecoin, true);
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", acval + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico")
        if ($addon!=null) ANNE.A.ezSpan(span, "", "", $addon);
    }
    


    /**
     * 
     * Get and display our annies balance.
     * 
     * Get and display the annies balance in both ANNE coin and annies.
     */
    ezDisplayAnnecoinWithNeuron = async ($el, $nObj, $lbl = "") => {
        var balanceObj = $nObj.balanceObj;
        if (balanceObj == null && $nObj.ghost == false && $nObj.temp == false) {
            balanceObj = await USER.checkBalance($nObj.nid);
        }
        if (balanceObj!=null) {
            if (balanceObj.annies != undefined) {
                var annies = parseInt(balanceObj.annies);
                ANNE.GUI_STUFF.doAnniesToAnneSpanFull($el, $lbl, annies);
            }
        }
    }


    ezDisplayPendingAnnies = async ($el, $nObj, $lbl = "") => {
        var balanceObj = $nObj.balanceObj;
        if (balanceObj == null && $nObj.ghost == false && $nObj.temp == false) {
            balanceObj = await USER.checkBalance($nObj.nid);
        }
        if (balanceObj) {
            if (balanceObj.avail_soon_annies != undefined) {
                var annies = parseInt(balanceObj.avail_soon_annies);
                ANNE.GUI_STUFF.doAnniesToAnneSpanFull($el, $lbl, annies);
            }
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $id 
     * @param {*} $value 
     * @param {*} $lbl 
     * @param {*} $checked 
     * @param {*} $disabled 
     * @param {*} $truncateClass 
     * @param {*} $cb 
     * @returns 
     */
    ezCheckMark = function ($el, $id, $value, $lbl, $checked = true, $disabled = false, 
                            $truncateClass = "", $cb=null, $chkclass="chk-parent") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($id=="" || $id==null) $id = ANNE.UTILS.randHex(3);
        const div = document.createElement("div");
        if ($chkclass!="") div.classList.add($chkclass);
        if ($truncateClass != "") div.classList.add($truncateClass);
        el.append(div);
        const chk = document.createElement("input");
        chk.setAttribute("type", "checkbox");
        if ($id!="") chk.setAttribute("id", $id);
        if ($id!="") chk.setAttribute("name", $id);
        chk.setAttribute("value", $value);
        if ($checked) chk.setAttribute("checked", true);
        if ($disabled) chk.setAttribute("disabled", true);
        div.append(chk);
        const lbl = document.createElement("label");
        lbl.classList="chk-label";
        lbl.setAttribute("for", $id);
        lbl.innerHTML = $lbl;
        div.append(lbl);

        //simple callback
        if ($cb) {
            chk.addEventListener("change", $e=>$cb($e));
        }
        return chk;
    }


    /**
     */
    ezInputText = function ($el, $id = "", $classList = "", $placeholder = "", $inputEventCallback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_input = document.createElement("input");
        el_input.setAttribute("type", "text");
        if ($classList != "") this.__classListParse(el_input, $classList);
        if ($id != "") el_input.setAttribute("id", $id);
        if ($placeholder) el_input.setAttribute("placeholder", $placeholder);
        el.append(el_input);
        if ($inputEventCallback != null && typeof $inputEventCallback === "function") {
            el_input.addEventListener("input", async ($e) => {
                var val = $e.target.value;
                $inputEventCallback(val);
            });
        }
        return el_input;
    }


    /** 
     * @param {element or string} - parent element to append to
     * @param {anne_neuron} - class to apply uom to.
     */
    ezInput = function ($el, $classNOBJ = null, $id = "", $class = "",
        $placeholder = "", $inputEventCallback = null, $inputFilterCallback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($class != "") {
            var div = document.createElement("div");
            div.classList = $class;
            el.append(div);
        }
        else div = el;

        const input = document.createElement("input");
        input.setAttribute("type", "text");
        if ($id != "") input.setAttribute("id", $id);
        if ($placeholder) input.setAttribute("placeholder", $placeholder);
        div.append(input);
        var uom = null;
        if ($classNOBJ != null) uom = MEM.unitOfMeasure($classNOBJ);
        if (uom != null) {
            var span = document.createElement("span");
            span.style.display = 'inline-block';
            span.classList = 'uom-symbol';
            div.appendChild(span);
        }
        if ($inputEventCallback != null && typeof $inputEventCallback === "function") {
            div.addEventListener("input", async ($e) => {
                var val = $e.target.value;
                if ($inputFilterCallback != null && typeof $inputFilterCallback === "function") {
                    val = await $inputFilterCallback(val);
                    $e.target.value = val;
                }
                if (uom != null) {
                    var thisconv = null;
                    if (uom.conversions != undefined && uom.conversions["_" + val] != undefined) {
                        thisconv = uom.conversions["_" + val];
                    }
                    else thisconv = selector_number.convert(val, uom);
                    span.innerHTML = thisconv.total + "<span>&nbsp;" + thisconv.post + "</span>";
                }
                $inputEventCallback(val);
            });
        }
        return div;
    }


    //should do radios.
    ezRadio = function ($el, $id, $value, $lbl, $checked = true, $disabled = false, $truncateClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        const div = document.createElement("div");
        if ($truncateClass != "") div.classList.add($truncateClass);
        el.append(div);
        const chk = document.createElement("input");
        chk.setAttribute("type", "radio");
        chk.setAttribute("id", $id);
        chk.setAttribute("name", $id);
        chk.setAttribute("value", $value);
        if ($checked) chk.setAttribute("checked", true);
        if ($disabled) chk.setAttribute("disabled", true);
        div.append(chk);
        const lbl = document.createElement("label");
        lbl.setAttribute("for", $id);
        lbl.innerHTML = $lbl;
        div.append(lbl);
        return chk;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $class 
     * @param {*} $cb 
     */
    injectInputAfter=function($el, $wrapperclass, $inputclass, $cb, $bDOCANCEL=true){
        var el = this.__getEl($el);
        if (el == null) return null;
        var wrapperdiv=document.createElement("div");
        if ($wrapperclass!="") wrapperdiv.classList = $wrapperclass;
        var newel = document.createElement("input");
        if ($inputclass!="") newel.classList = $inputclass;
        newel.style['flex-grow']=1;
        newel.style.cursor='auto';

        wrapperdiv.append(newel);
        this.appendAfter(el, wrapperdiv);
        ANNE.A.makeReturnPressEvent(newel, $cb);
        ANNE.A.injectEnterAfter(wrapperdiv, async($e)=>{
            $cb(newel);
            ANNE.A.removeSelf(wrapperdiv);
        })
        if ($bDOCANCEL==true) ANNE.A.injectCancelAfter(wrapperdiv, async($e)=>{
            $cb(newel);
            ANNE.A.removeSelf(wrapperdiv);
        });
        return newel;
    }

    injectEnterAfter = function($el, $cb, $class="") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var btnEnter = document.createElement("button");
        btnEnter.innerHTML = "enter";
        if ($class!="") btnEnter.classList = $class;
        el.append(btnEnter);
        btnEnter.addEventListener("click", $cb);
    }

    injectCancelAfter = function($el, $cb, $class="btn-red-sq") {
        var el = this.__getEl($el);
        if (el == null) return null;
        const btnCancel = document.createElement("button");  
        if ($class!="") btnCancel.classList=$class;         
        btnCancel.innerHTML = "[x]";
        el.appendChild(btnCancel);
        btnCancel.addEventListener("click", $cb);
    }


    /**
     * Make a return press event to callback the $cb.
     * @param {*} $el 
     * @param {*} $cb 
     * @returns 
     */
    makeReturnPressEvent = async($el, $cb)=>{
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($cb!=null) {
            el.addEventListener("keyup", async($e)=>{
                if ($e.keyCode === 13) {
                    $cb(el);
                }
            });    
        }
        return;
    }

    
    /**
     * Make a new element just like the $el and inject it after the $el as a sibbling.
     * @param {*} $el 
     * @returns 
     */
    makeSibAfter=function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var clone = $el.cloneNode(true);
        clone.innerHTML="";
        clone.id="newidclon_"+Math.floor(Math.random()*1000);
        this.appendAfter(el, clone);
        return clone;
    }


    /**
     * 
     * @param {anne_gui_role} $agr - the gui role object 
     */
    ezRoleEdit = function ($el, $clump_nid, $rel, $guid, $bEDIT = false, $bMOD = false, $el_inject = "", $newcb = null, $cancelcb = null, $deletecb = null, $span2 = null) {
        var span2 = $span2;
        if (span2 == null) span2 = this.ezRoleDisplay($el, $rel, $guid);
        else span2 = this.__getEl($span2);

        var el_inject = this.__getEl($el_inject);
        if (el_inject) {
            span2.setAttribute("el_inject", $el_inject.id);
        }

        span2.setAttribute("clump_nid", $clump_nid);
        span2.setAttribute("guid", $guid);

        if ($bEDIT == true) {
            var btnAdd = document.createElement("button");
            btnAdd.innerHTML = "+add";
            btnAdd.id = "add_" + $rel.rid + "_" + $guid;
            span2.append(btnAdd);
            btnAdd.addEventListener("click", e => ANNE.A.editRoleLaunchSelector(span2, $newcb, $cancelcb));
            span2.setAttribute("delete_ok", 1);
            btnAdd.style['margin-left'] = '4px';

            if ($rel.type == TYPE_ROLE) {
                var btndelete = document.createElement("button");
                btndelete.innerHTML = "[remove]";
                btndelete.id = "delete_" + $rel.rid + "_" + $guid;
                span2.append(btndelete);
                btndelete.setAttribute("txid", $rel.txid);
                btndelete.addEventListener("click", e => $deletecb(span2, btndelete));
                btndelete.style['margin-left'] = '4px';
            }

        }

        if ($bMOD == true) {
            var btnmod = document.createElement("button");
            btnmod.innerHTML = "+mod";
            btnmod.id = "mod_" + $rel.rid + "_" + $guid;
            btnmod.style['margin-left'] = '4px';
            btnAdd.addEventListener("click", e => function (e) { alert("todo"); });
            div.append(btnmod);
        }

        var ptxid = "";
        var to_s = "";
        if ($rel != null) {
            ptxid = $rel.txid;
            to_s = $rel.to_s;
        }

        var default_class_nid = null;
        var clumpNOBJ = MEMG.getNeuronByNid($clump_nid);

        if (ptxid != "") default_class_nid = clumpNOBJ.getDefaultClassForParam(ptxid);
        if (default_class_nid != null) span2.setAttribute("classRestrictNid", default_class_nid);
        else if (to_s != "") span2.setAttribute("classRestrictNid", to_s);
        else if ($default_class_nid != null) span2.setAttribute("classRestrictNid", $default_class_nid);
    }


    /**
     * An editable role has been clicked on, launch the selector
     * @param {*} $span2 
     */
    editRoleLaunchSelector = async ($span2, $cb_new = null, $cb_cancel = null, $cb_defocus = null) => {
        var tnid = $span2.getAttribute("role_tnid");
        var el_inject = $span2.getAttribute("el_inject");
        var roleNOBJ = MEMG.getNeuronByTempNid(tnid);
        var misc = {};
        misc.smartDefocus = $cb_defocus;
        misc.bCALLBACKONLY = true;
        var pObj = new page_params($cb_new, $cb_cancel, null, roleNOBJ.nid);
        pObj.misc = misc;
        var contextNOBJ = null;
        var classRestrictNid = $span2.getAttribute("classRestrictNid");
        if (classRestrictNid != undefined) {
            pObj.misc.classRestrictNid = classRestrictNid;
            contextNOBJ = MEMG.getNeuronByNid(classRestrictNid);
        }
        var page = ANNE.MEM_UDF.getDefaultSelectorPageForClass(roleNOBJ, null, contextNOBJ);
        console.log("%cPAGE "+page, TALK.bggreen);
        var clump_nid = $span2.getAttribute("clump_nid");
        pObj.misc.from_nid = clump_nid;   //all TYPE_ROLE relons will be put on the clump.
        pObj.misc.type = TYPE_ROLE;
        pObj.misc.reln_nid = roleNOBJ.nid;
        ANNE.A.empty(el_inject);
        pObj.injection_div = el_inject;
        var guid = $span2.getAttribute("guid");
        if (guid == null) return;
        
        var pgui = ANNE.MEM_UDF.getPguiByGuid(guid);
        if (pgui.INST && typeof pgui.INST.smartDefocus === "function") {
            pgui.INST.smartDefocus(true);
        }
        if (pgui == null) return;
        await pgui.loadPart(page, pObj);
    }


    /**
     * @param {*} $rel 
     */
    ezRoleDisplay = function ($el, $rel, $guid) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var roleNOBJ = MEMG.getNeuronByTempNid($rel.r);
        var _roletname = roleNOBJ.text.replaceAll("_", "-");

        //frienly replacement of role name from affected_thing to who
        //when dealing with persons in the that slot.
        if ($rel.to_type == TO_TYPE_ANNE) {
            if (_roletname == "role-affected-thing") {
                var toNOBJ = MEMG.getNeuronByTempNid($rel.t);
                if (toNOBJ.is(MEM.person) == true) _roletname = "role-affected-who";
            }
            else if (_roletname == "role-actor") {
                var toNOBJ = MEMG.getNeuronByTempNid($rel.t);
                if (toNOBJ.is(MEM.person) == true) _roletname = "role-actor-who";
            }
        }

        var div = document.createElement("div");
        div.classList = "role2 " + _roletname + "-outter";
        el.append(div);
        var span1 = document.createElement("span");
        span1.classList = _roletname + "-" + "lbl";
        div.append(span1);
        var span2 = document.createElement("span");
        span2.classList = "r-inner";
        span2.id = "val_" + roleNOBJ.temp_nid + "_" + $guid;
        span2.setAttribute("role_tnid", roleNOBJ.temp_nid);
        div.append(span2);

        //display only if a role relon
        if ($rel.type == TYPE_ROLE) {
            if ($rel.to_type == TO_TYPE_ANNE) {
                ANNE.MEM_UDF.makeNeuronLink(div, $rel.to_s);
            }
            else if ($rel.to_type == TO_TYPE_TXID) {
                ANNE.MEM_UDF.makeRelonViewLink(div, $rel.to_s);
            }
            else {
                var totxt = MEMG.__MASTER_DISPLAY_RELON_TO($rel);
                ANNE.A.setHTML(div, totxt);
            }
        }

        return span2;
    }


    ezSQDWrapper = function ($el, $class, $display = "", $callback, $callbackParams, $size = "36px", $title = "remove member") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var termdelete = document.createElement("div");
        el.append(termdelete);
        if ($class != "") termdelete.classList = $class;
        if ($display != "") termdelete.style.display = $display;
        ANNE.A.ezSquareDelete(termdelete, $callback, $callbackParams, $size, $title);
        return termdelete;
    }

    ezSquareDelete = function ($el, $callback, $callbackParams, $size = "36px", $title = "remove member") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var dsquare = document.createElement("div");
        dsquare.classList = "delete-square";
        el.append(dsquare);
        var boxX = ANNE.A.ezEl(dsquare, "span", "boxX", "<img src='../imgs/boxX.png' height='" + $size + "' width='" + $size + "'>", "", $title);
        var sl1c = this.sl1c;
        if (typeof $callback !== "function") {
            console.log("callback jacked up in ezSquareDelete");
            return;
        }
        boxX.addEventListener("click", e => $callback(...$callbackParams));
    }
    
    /**
     * 
     * @param {*} $el 
     * @param {*} $prop 
     * @param {*} $maxgens 
     * @returns 
     */
    extractNearestObjByProp=function($el, $prop="tnid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var attr_val = $el.getAttribute($prop);
            if (attr_val != null) return $el;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }


    /**
     * Extract the prop of a nearest-ancestor
     * @param {*} $el 
     * @param {*} $prop 
     * @param {*} $maxgens 
     * @returns 
     */
    extractNearest=function($el, $prop="tnid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var attr_val = $el.getAttribute($prop);
            if (attr_val != null) return attr_val;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }



    /**
     * @param {Element} $el
     */
    ez_getNearestParentWithProp = function ($el, $prop="nid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var nid = $el.getAttribute($prop);
            if (nid != null) return nid;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }

    /**
     */
    ez_floatActionIco = async ($nid, $ainame, $titleOverride = "")=> {
        var title = ""
        var actionz = "";

        if ($ainame == "mine") {
            title = "This is your data.  Click on it to delete it.";
            actionz = "delete";
        }

        else if ($ainame == "collab") {
            title = "This is data you can collab on.  Click on it to delete it.";
            actionz = "delete";
        }

        //this is locally ghosted data..
        else if ($ainame == "ghost") {

            if (ANNE._txMode == TXMODE_GHOST) {
                title = "This is your ghost.  Only you see it.";
            }

            //if in ANNE mode...
            else if (ANNE._txMode == TXMODE_ANNE) {

                //if it is upgradable...
                var nObj = MEMG.getNeuronByTempNid($nid);
                if (nObj._govenor == null) await nObj.setGovRights(USER._userNObj, false);
                if (nObj.collab_ACTION >= govenor.COLLAB_REQ_OK) {
                    title = "Save your ghost to ANNE.";
                    $ainame = "ghost-to-anne";
                    actionz = "upgrade-ghost-to-anne";
                }
                else {
                    title = "This class of local ghost data cannot be upgraded to ANNE Ghost data.";
                }
            }
        }

        var newspan = document.createElement("span");
        newspan.classList = "hoverable";
        newspan.setAttribute("title", title);
        var itag = document.createElement("i");
        itag.classList = "ai-" + $ainame + " actions-ico";

        //this will trigger DOM events on class clicks to optimize (minimize) events.
        //the widget calling this helper will have to be listening.
        if (actionz) itag.setAttribute("actionz", actionz);

        newspan.append(itag);
        return newspan;
    }

    /**
     */
    ez_auto_inject_close_btn = function ($el, $class, $hrefid, $spanClass = "hoverable", $icoclass = "ai-times-circle ai-2x", $bPREPEND = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var betterEl = el.querySelector(".inject-close-here");

        //always prepend into the 
        if (betterEl) el = betterEl;

        var bPREPEND = true;    //if we didn't find a better el, PREPEND (to injection div)
        return ANNE.A.ezHrefBtn(el, $class, $hrefid, $spanClass, $icoclass, bPREPEND);
    }


    /**
     * EXT - NOT IN USE IN EXT...
     * 
     * This will render a gui ico for deleting or updgrading ur own records.
     */
    ez_floatActionIcos = async ($nObj, $el, $ownStatus = "collab")=> {
        var el = this.__getEl($el);
        if (el == null) return null;
        var bSHOW = false;

        if ($nObj.mine == true) {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "mine");
            floatico.style = "cursor:default";
            el.append(floatico);
            bSHOW = true;
        }

        //only one of these...
        else if ($nObj.ghost == true) {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "ghost");
            el.append(floatico);
            bSHOW = true;
        }
        else if ($ownStatus == "collab") {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "collab");
            el.append(floatico);
            bSHOW = true;
        }

        if (bSHOW == true) ANNE.A.show(el);
    }


    /**
     * text dd
     */
    ezDD_basic = function ($el, $items, $prompt = "select", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        ANNE.A.empty(selectel);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(selectel.value));
        }
        var opt = document.createElement("option");
        opt.setAttribute("value", "");
        opt.innerHTML = $prompt;
        selectel.append(opt);
        for (var i = 0; i < $items.length; i++) {
            var item = $items[i];
            var opt = document.createElement("option");
            opt.setAttribute("value", item);
            opt.innerHTML = item;
            selectel.append(opt);
        }
        ANNE.A.show(selectel);
        return selectel;
    }


    /** 
     */
    ezDD = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $bSHOWNID = false, $skiptnids = []) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            if ($skiptnids.length && $skiptnids.indexOf(nObj.temp_nid) != -1) continue;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);

            opt.innerHTML = nObj.bestNameHTML("", $bSHOWNID);

            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }


    /**
     * used in pref dds 
     */
    ezDD_masterDDObj = function ($el, $masterOBJS, $callback = null, $msg = "easy access acct nids", $ddClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else selectel = this.ezEl(selectec, "select");

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);

        for (var n = 0; n < $masterOBJS.length; n++) {
            var masterOBJ = $masterOBJS[n];
            var nObj = masterOBJ.nObj;
            var type = masterOBJ.type;
            var nick = masterOBJ.nick;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            var text = "";
            if (nick == nObj.nid) opt.innerHTML = "(" + type + ") " + nObj.nid;
            else opt.innerHTML = "(" + type + ") " + nick + " - " + nObj.nid;
            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     */
    ezDD_nObjs = function ($el, $nObjs, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $bSHOWNID = false) {
        var el = this.__getEl($el);
        if (el == null) {
            console.log("(ezDD_nObjs) el "+$el+" doesnt exist");
            return null;
        }
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => {
				$callback(e);
			});
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        //var aNObjs=MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) $nObjs.sort(ANNE.UTILS.sortByTextNoCase);
        for (var n = 0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            if (nObj == null) continue;
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            if (nObj.local_nickname != undefined) opt.innerHTML = nObj.local_nickname;
            else opt.innerHTML = nObj.bestNameHTML("", $bSHOWNID);

            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    doGoogleMap = function ($el, $lat = -34.397, $long = 150.644) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var map = new google.maps.Map(el, {
            center: { lat: $lat, lng: $long },
            zoom: 8,
        });
        return map;
    }

    /**
     */
    ezDDWithGens = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);

        //aNOBJS are the are the tops.  
        //We want to dig down and get their children.
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            var downline = [];
            nObj.getAllDownlineChildren(downline, 0, false, false, true, false);
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            opt.innerHTML = nObj.bestNameHTML();
            opt.classList = "option-top";
            selectel.append(opt);

            for (var a = 0; a < downline.length; a++) {
                var childNOBJ = downline[a];
                //doHeading(nObj);
                var indent = "&#9494;";
                var space = '&#160;';
                if (childNOBJ.be_lvl == 0) space = "";
                else {
                    for (var i = 0; i < childNOBJ.be_lvl - 1; i++) { space += '&#160;'; }
                    space = space + indent;
                }
                var opt2 = document.createElement("option");
                opt2.setAttribute("value", childNOBJ.nid);
                opt2.innerHTML = space + childNOBJ.bestNameHTML();
                selectel.append(opt2);
                //items.push({id: nObj.temp_nid,text: '<div>'+space+nObj.text+' ('+nObj.nid+')</div>',value: nObj.nid});
            }
            //results.push({text: $heading,children: items});
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     */
    ezDDWithGens_ACTIONS = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $V1ONLY = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);

        //aNOBJS are the are the tops.  
        //We want to dig down and get their children.
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;

            //we only want V1 vbs.
            if ($V1ONLY == true) {
                var rel_vbForm = MEMG.getBeRelon(nObj.outs, MEM.vb_form.temp_nid);
                if (rel_vbForm != null) {
                    if (rel_vbForm.t != MEM.vb_v1.temp_nid) {
                        console.log("skipping non v1");
                        continue;
                    }
                }
            }

            var downline = [];
            nObj.getAllDownlineChildren(downline, 0, false, false, true, false);
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            opt.innerHTML = nObj.bestNameHTML();
            opt.classList = "option-top";
            selectel.append(opt);

            for (var a = 0; a < downline.length; a++) {
                var childNOBJ = downline[a];

                //we only want v1 vbs.
                if ($V1ONLY == true) {
                    var rel_vbForm = MEMG.getBeRelon(childNOBJ.outs, MEM.vb_form.temp_nid);
                    if (rel_vbForm != null) {
                        if (rel_vbForm.t != MEM.vb_v1.temp_nid) {
                            console.log("skipping non v1");
                            continue;
                        }
                    }
                }

                var indent = "&#9494;";
                var space = '&#160;&#160;&#160;';
                if (childNOBJ.be_lvl == 0) space = "";
                else {
                    for (var i = 0; i < childNOBJ.be_lvl - 1; i++) { space += '&#160;&#160;&#160;'; }
                    space = space + indent;
                }
                var opt2 = document.createElement("option");
                opt2.setAttribute("value", childNOBJ.nid);
                opt2.innerHTML = space + childNOBJ.bestNameHTML();
                selectel.append(opt2);
            }
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $val 
     * @returns 
     */
    setSelectedIndexByValue=function($el, $val) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var i=0; i<el.options.length; i++) {
            if (el.options[i].value==$val) {
                selectel.setAttribute('selectedIndex', i);
                break;
            }
        }
    }


    /**
     * ez row with 2 col
     */
    ez2Col = function ($el, $v1 = "", $v2 = "", $v1extraclass = "", $v2extraclass = "") {
        var v1class = "simple-col";
        if ($v1extraclass != "") v1class = v1class + " " + $v1extraclass;
        var v2class = "simple-col";
        if ($v2extraclass != "") v2class = v2class + " " + $v2extraclass;
        var row = ANNE.A.ezEl($el, "div", "simple-row");
        ANNE.A.ezEl(row, "div", v1class, $v1);
        ANNE.A.ezEl(row, "div", v2class, $v2);
        return row;
    }


    //intended for new divs, spans, and other elements that
    //you want to also givev a class, id and style.
    ezEl = function ($el, $type, $classList = "", $text = "", $id = "", $title = "", $style = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newEl = document.createElement($type);
        if ($classList != "") this.__classListParse(newEl, $classList);
        if ($text != "") newEl.innerHTML = $text;
        if ($title != "") newEl.setAttribute("title", $title);
        if ($id != "") newEl.setAttribute("id", $id);
        if ($style != null) ANNE.A.setStyle($el, $style);
        el.append(newEl);
        return newEl;
    }
    ezElPrepend = function ($el, $type, $classList = "", $text = "", $id = "", $title = "", $style = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newEl = document.createElement($type);
        if ($classList != "") this.__classListParse(newEl, $classList);
        if ($text != "") newEl.innerHTML = $text;
        if ($title != "") newEl.setAttribute("title", $title);
        if ($id != "") newEl.setAttribute("id", $id);
        if ($style != null) ANNE.A.setStyle($el, $style);
        el.prepend(newEl);
        return newEl;
    }


    doImgB64 = function($el, $nObj, $icostyle="dossier", $w=84, $h=0) {     
        var el = this.__getEl($el);
        if (el == null) return null;
        var url = ANNE.A.getThumbImgUrl($nObj, $icostyle);
        if (url.length) {
            ANNE.A.doImageB64(el, "", $afile, $w, $h);
        }
    }



    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    getThumbImgUrl=function($nObj) {
        var parentNOBJ = $nObj.getParent();
        var neuronImgNOBJ = MEM.__GET_DEFAULT_IMG_NEURON(parentNOBJ, $nObj);
        if (neuronImgNOBJ!=null) {
            var url = IMAGES.getUrl(neuronImgNOBJ);
            if (url!="") return url;
        }

        //we're going to nixx all this drama.
        return ANNE.A.getThumbImgUrlFromRelons($nObj);
    }


    /**
     * 
     * @returns 
     */
    getThumbImgUrlFromRelons = function($nObj, $icostyle=null) {
        var test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_thumbnail.temp_nid);
        if (test==null) {
            test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Skyline.temp_nid);
            if (test==null) {
                test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Map.temp_nid);
                if (test==null) {
                    test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Flag.temp_nid); 
                    if (test==null) test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Seal.temp_nid); 
                }
            }
        }
        if (test!=null) { 
            return "http://commons.wikimedia.org/wiki/Special:FilePath/"+test.to_s+"?width=200";
        }

        //console.log("not enough info on "+$nObj.text+" for img or ico");
        return "";
    }


    focusTextFix = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.addEventListener("focus", function () {
            el.setSelectionRange(0, 0);
        });
    }

    
    __classListParse = function ($el, $cl) {
        if ($cl == "") return;

        if ($cl != "") {
            if ($cl.indexOf(" ") != -1) {
                var cl = $cl.split(" ");
                for (var c = 0; c < cl.length; c++) {
                    var clstr = cl[c].trim();
                    if (clstr == "") return;
                    $el.classList.add(clstr);
                }
            }
            else $el.classList = $cl;
        }
    }

    //we can inject a close button into a model with this..
    //and then wire it to its pgui.close function, automatically. :)
    ezHrefBtn = function ($el, $hrefClass, $hrefid=null, $spanClass = "hoverable", $icoclass = "ai-times-circle ai-2x", $bPREPEND = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var href = document.createElement("a");
        if ($hrefid!=null) href.setAttribute("id", $hrefid);
        if ($hrefClass) href.classList = $hrefClass;
        if ($bPREPEND == false) $el.append(href);
        else $el.prepend(href);
        ANNE.A.ezIcoSpan(href, $spanClass, $icoclass);
        return href;
    }

    /**
     * ez ico img using the css ico classes
     */
    ezIcoImgSpan = function ($el, $img_class, $msg, $bEMPTY = true, $bSHOW = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($bEMPTY == true) ANNE.A.empty(el);
        var span = document.createElement("span");
        span.classList = $img_class;
        el.append(span);
        if ($msg != "") {
            var msg_span = document.createElement("span");
            msg_span.innerHTML = $msg;
            el.append(msg_span);
        }
        if ($bSHOW == true) ANNE.A.show(el);
    }


    ezIcoLink = function ($el, $href = "#", $ico_class = "ai-anne") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var link = document.createElement("a");
        link.setAttribute("href", $href);
        el.append(link);
        var link_i = document.createElement("i");
        link_i.classList = $ico_class;
        link.append(link_i);
    }


    ezHref = function ($el, $href = "#", $target = "_blank", $txt = "", $ico_class = "") {
        var el = this.__getEl($el);
        var link = document.createElement("a");
        link.setAttribute("href", $href);
        link.setAttribute("target", $target);
        if ($txt == "") link.innerHTML = $href;
        else link.innerHTML = $txt;
        el.append(link);
        if ($ico_class != "") {
            var link_i = document.createElement("i");
            link_i.classList = $ico_class;
            link.append(link_i);
        }
    }


    ezIcoSpan = function ($el, $spanClass, $icoClass, $spanMsg = "", $stylejson = null) {
        var el = this.__getEl($el);
        //if (el==null) return null;
        var newel = document.createElement("span");
        if ($spanClass) newel.classList = $spanClass;
        if (el != null) el.append(newel); //append if we have it, else we just return the new one
        if ($stylejson != null) el.style = $stylejson;
        var ico = document.createElement("i");
        ico.classList = $icoClass;
        newel.append(ico);
        if ($spanMsg != "") newel.innerHTML = newel.innerHTML + $spanMsg;
        return newel;
    }

    ezlblSpan = function ($el, $lbl, $value = "", $spanclass = "", $valclass = "", $sep = "&nbsp;") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newel = document.createElement("span");
        if ($spanclass != "") newel.classList = $spanclass;
        if ($valclass == "") newel.innerHTML = $lbl + $sep + $value;
        else newel.innerHTML = $lbl + $sep + '<span class="' + $valclass + '">' + $value + '</span>';
        el.append(newel);
        return newel;
    }

    ezTextList = function ($el, $outs, $relnNOBJ, $lbl = "", $class = "simple-row", $exclude = [TYPE_PARAM], $sep = ", ") {
        var list = TALK.printListByReln($outs, $relnNOBJ.temp_nid, $exclude, $sep);
        if (list.length > 0) {
            var eltextlist = document.createElement("div");
            if ($class != "") eltextlist.classList = $class;
            if ($lbl != "") eltextlist.innerHTML = $lbl + list;   //lbl should have trailing space if it is needed.
            $el.append(eltextlist);
        }
    }

    
    ezTwoColTextList = function ($el, $outs, $relnNOBJ, $lbl = "", $exclude = [TYPE_PARAM], $lblClass = "span-lbl", $dataClass = "span-data") {
        var relons = MEMG.getRelonsByTempR($outs, $relnNOBJ.temp_nid, $exclude);
        if (relons.length == 0) return "";

        var elrow = this.ezEl($el, "div", "simple-row");
        var lblcol = this.ezEl(elrow, "div", "simple-col tal");
        this.ezSpan(lblcol, "", $lblClass, $lbl);

        var datacol = this.ezEl(elrow, "div", "simple-col tal");

        for (var r = 0; r < relons.length; r++) {
            var rel = relons[r];
            var totxt = MEMG.__MASTER_DISPLAY_RELON_TO(rel);
            var el_link = EXT_BTNS.makeRelonViewLink(datacol, rel.txid, false, 8, true, "", totxt);
            el_link.style.display = "block";
        }
    }

    ezSortCol = function ($el, $colObj, $tbldivid, $sort) {
        var el = this.__getEl($el);
        if (el == null) return null;

        var classList = ["sortico", "hoverable-green"];

        if ($colObj.ttype == "number") classList.push("ai-sort-amount-down");
        else if ($colObj.ttype == "string" || $colObj.ttype == "anne") classList.push("ai-sort-alpha-down");
        else if ($colObj.ttype == "date") classList.push("ai-sort-down");

        var span = document.createElement("span");
        span.innerHTML = $colObj.text + "&nbsp;";
        el.append(span);
        var ico = document.createElement("i");
        ico.classList = classList.join(" ");
        ico.setAttribute("prop", $colObj.ttype);
        ico.setAttribute("sort", $sort);
        ico.setAttribute("tbl", $tbldivid);
        span.append(ico);

        return ico; //return the sortico (so we can easily put events on it)
    }


    ezSpan = function ($el, $id, $class, $innermsg, $styleJson = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = this.ezSpanSolo($id, $class, $innermsg, $styleJson);
        el.append(span);
        return span;
    }

    //make it even if empty.
    ezSpan2 = function ($el, $id, $class, $innermsg, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = this.ezSpanSolo($id, $class, $innermsg, null);
        el.append(span);
        if ($callback) span.addEventListener("click", $e=>$callback($e));
        return span;
    }

    /**
     * does not append, just creates
     */
    ezSpanSolo = function ($id = "", $class = "", $innermsg = "", $styleJson = null) {
        var newel = document.createElement("span");
        if ($class) newel.classList = $class;
        if ($id != "") newel.setAttribute("id", $id);
        newel.innerHTML = $innermsg;
        if ($styleJson != null) newel.style = $styleJson;   //not working as i thought
        return newel;
    }

    ezButton = function ($el, $id = "", $text = "save", $class = "anne-btn btn-success", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = $class;
        btn.innerHTML = $text;
        if ($id != "") btn.setAttribute("id", $id);
        el.appendChild(btn);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", function ($e) { 
                $e.preventDefault();
                $callback($e); 
            });
        }
        return btn;
    }

    ezButtonWithIcoClass = function ($el, $id = "", $text = "save", $class = "anne-btn btn-success", $callback = null, $icoClass="") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = $class;
        var icoEl = document.createElement("span");
        icoEl.classList = $icoClass;
        btn.appendChild(icoEl);
        icoEl.style['pointer-events']= 'none';
        var textEl = document.createElement("span");
        textEl.innerHTML = $text;
        textEl.style['pointer-events']= 'none';
        btn.appendChild(textEl);
        if ($id != "") btn.setAttribute("id", $id);
        el.appendChild(btn);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", function ($e) { 
                $e.preventDefault();
                $callback($e); 
            });
        }
        
        return btn;
    }

    ezFbwBtn = function ($el, $tnid, $guid, $classList = "", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        if ($classList != "") this.__classListParse(btn, $classList);
        if ($tnid) btn.setAttribute("tnid", $tnid);
        if ($tnid && $guid) btn.setAttribute("id", $guid + "_" + $tnid);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", $callback);
        }
        el.appendChild(btn);
        return btn;
    }

    ezFbwBtn_prepend = function ($el, $tnid, $guid, $classList = "", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        if ($classList != "") this.__classListParse(btn, $classList);
        if ($tnid) btn.setAttribute("tnid", $tnid);
        if ($tnid && $guid) btn.setAttribute("id", $guid + "_" + $tnid);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", $callback);
        }
        el.prepend(btn);
        return btn;
    }

    /**
     */
    getPrintNameIcoSpan = function ($nObj) {
        if ($nObj.temp == true) return "<span class='temp-neuron'>" + $nObj.bestName() + "</span>";
        else if ($nObj.ghost == true) return "<span class='ghost-neuron'><span class='ghost-small-blue-ico'></span>" + $nObj.bestName() + "</span>";
        else if (ANNE.isANNE($nObj.owned_by)==true) return "<span class='real-neuron'><span class='anne-small-black-ico'></span>" + $nObj.bestName() + "</span>";
        else if ($nObj.owned_by == USER._userNObj.nid) return "<span class='real-neuron'><span class='mine-small-ico'></span>" + $nObj.bestName() + "</span>";
        return $nObj.text;
    }


    imgBadAudit = function ($el, $id = "", $imgref = "", $imgclass = "search-img-small", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var spanEl = this.doBadImg(el, $id, $imgref, $imgclass)
        if ($callback != null && typeof $callback === "function") {
            spanEl.addEventListener("click", function (e) { $callback(e); });
        }
        return spanEl;
    }


    imgButton = function ($el, $id = "", $imgref, $imgclass = "search-img-small", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var imgEl = this.doImageDyno(el, $id, $imgref, $imgclass)
        if ($id != "") imgEl.setAttribute("id", $id);
        el.appendChild(imgEl);
        if ($callback != null && typeof $callback === "function") {
            imgEl.addEventListener("click", function (e) { $callback(e); });
        }
        return imgEl;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $id 
     * @param {anne_file} $afile 
     * @param {*} $imgclass 
     * @param {function} $callback 
     * @returns {element}
     */
    imgButtonB64 = function ($el, $id = "", $afile=null, $imgclass = "search-img-small", $callback = null) {
        if ($afile==null || $afile==undefined) {
			return null;
		}
        var el = this.__getEl($el);
        if (el == null) {
			return null;
		}
        var imgEl = this.doImageDynoB64(el, $id, $afile, $imgclass)
        if (imgEl==undefined || imgEl==null) {
			return null;
		}
		if ($id != "") imgEl.setAttribute("id", $id);
        el.appendChild(imgEl);
        if ($callback != null && typeof $callback === "function") {
            imgEl.addEventListener("click", function (e) { $callback(e); });
        }
        return imgEl;
    }

    
    /**
     * Insert an element into the DOM after another element.
     * lolz, as this is the same as the native .after, so we should just use that.
     * @param {Element or String} $existingNode 
     * @param {Element} $newNode 
     * @returns 
     */
    insertAfter = function($existingNode, $newNode) {
        var el = this.__getEl($existingNode);
        if (el == null) return;
        //if (el.parentNode==null) return;
        //el.parentNode.insertBefore($newNode, el.nextSibling);
        $existingNode.after($newNode);
        return;
    }

    /**
     * Move an el to be after another el.
     */
    moveElAfter=function($el1, $insertAfterThisEl) {
        var el1 = this.__getEl($el1);
        if (el1 == null) return;
        var insertAfterThisEl = this.__getEl($insertAfterThisEl);
        if (insertAfterThisEl == null) return;
        insertAfterThisEl.after($el1);
    }

    appendAfter = function($existingNode, $newNode) {
        this.insertAfter($existingNode, $newNode);
    }

    //triple toggle,  asc, desc, none
    sortToggle = function ($el, $colNumb, $sort = "desc", 
                            $callback = null, 
                            $sortype="numeric") {

                            
                            //$descending = "bi bi-sort-alpha-down", 
                            //$ascending = "ai-sort-asc") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("span");
        btn.style.cursor = "pointer";
        btn.classList = "ez-sort-btn";
        btn.setAttribute("col_num", $colNumb);
        if ($sort == "desc") {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
            }
            btn.classList.add("active");
        }
        else if ($sort == "asc") {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-up'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-up'></i>";
            }
            btn.classList.add("active");
        }

        //none is parked on open desc (not selected);
        else {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
            }
        }

        btn.setAttribute("sort_val", $sort);

        el.appendChild(btn);

        if ($callback != null) {
            //CLICK ORDER CYYCLE::
            //NONE->DESC->ASC
            btn.addEventListener("click", async (e)=> {
                var target = e.target;
                target = ANNE.A.getFirstParentOf(target, "ez-sort-btn");
                var colNumb = target.getAttribute("col_num");
                var sort_val = target.getAttribute("sort_val");
                if (sort_val == "desc") {
                    target.setAttribute("sort_val", "asc");
                    ANNE.A.addClass(target, "active");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-up'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-up'></i>";
                    }
                    $callback(colNumb, "asc");
                }

                //asc becomes NONE.. not active.
                //show non-active desc sort.  if clicked, it becomes active to desc sort.
                else if (sort_val == "asc") {
                    target.classList.remove("active");
                    target.setAttribute("sort_val", "none");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
                    }
                    $callback(colNumb, "none");
                }

                else if (sort_val == "none") {
                    target.setAttribute("sort_val", "desc");
                    ANNE.A.addClass(target, "active");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
                    }
                    $callback(colNumb, "desc");
                }


            });

            //init sort.
            //if ($sort=="desc" || $sort=="asc") $callback( $colNumb, $sort);
        }
    }

    /**
     * simple toggle of a bool attribute with get/set
     */
    toggleElAttributeBool=function($el, $prop) {
        if ($el==null) return;
        if ($el.getAttribute($prop)==true) $el.setAttribute($prop, false);
        else $el.setAttribute($prop, true);
    }

    toggleBtn = function ($el, $nObj, $class = "", $callback = null, $on = true, $lbl = "", $ton = "toggled-on", $toff = "toggled-off") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = "toggle-btn " + $class;
        if ($on == true) btn.classList.add($ton);
        else btn.classList.add($toff);
        btn.setAttribute("tnid", $nObj.temp_nid);
        if ($lbl.length) btn.innerHTML = $lbl;
        else btn.innerHTML = nObj.bestNameHTML();
        el.appendChild(btn);

        btn.addEventListener("click", async (e)=> {
            var target = e.target;

            //if this is off we are turning it on.
            //if it is on.. it stays on even when clicked again.
            if (target.classList.contains($toff)) {
                target.classList.remove($toff);
                target.classList.add($ton);
            }
            else {
                target.classList.remove($ton);
                target.classList.add($toff);

                //callback.
                if ($callback != null && typeof $callback === "function") $callback(e);
            }
        });

        return btn;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $masterObj 
     * @param {*} $eltype 
     */
    makeToggleSet = function($el, $masterObj, $eltype="div") {
        var ats = new anne_toggle_set($masterObj);
        this.makeToggleSetGUI($el, ats, $eltype);
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $class 
     * @returns 
     */
    geAllSibs = function($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent_el = el.parentNode;
        if (parent_el==null) return null;
        return parent_el.querySelectorAll("."+$class);
    }


    /**
     * Creates divs from array of anne_toggle_objs.
     * Will allow for toggle of each div via a toggle btn.
     * @param {*} $el 
     * @param {anne_toggle_set} ats 
     * @param {*} $eltype 
     * @returns {anne_toggle_set}
     */
    makeToggleSetGUI = function($el, $ats, $eltype="div") {
        var el = this.__getEl($el);
        if (el == null) return;

        var el_btn_row = ANNE.A.ezEl($el, "div", "tg-btns", "", "");
        var el_toggle_divs = ANNE.A.ezEl($el, "div", "tg-toggle-divs", "", "");

        for (var i=0; i < $ats.atos.length; i++) {
            const ato = $ats.atos[i];
            console.log("btn tnid "+ato.btn_tnid);
            
            ato.el_toggle_divs = el_toggle_divs;

            //put new button in btn row
            var btn = ANNE.A.ezButton(el_btn_row, "", ato.btn_text, ato.btn_class);
            ato.btn_el = btn;
            if (ato.btn_tnid!="") btn.setAttribute("tnid", ato.btn_tnid);

            //put new div in the .tg-set div.
            var newEl = ANNE.A.ezEl(el_toggle_divs, $eltype, ato.div_class+" tg-set", ato.text, ato.id, ato.title, ato.style);
            ato.div_el = newEl;
            btn.setAttribute("btn_tnid", ato.btn_tnid);

            
            btn.addEventListener("click", async($e)=>{
                if (ato.div_el.classList.contains("active")==true) {
                    ato.div_el.classList.remove("active");
                    if (ato.cb!=null) ato.cb(ato, false);
                }
                else {
                    //should get all tg-set divs
                    //var els = ANNE.A.geAllSibs(ato.div_el, "tg-set");
                    for (var e = 0; e < el_toggle_divs.children.length; e++) {
                        el_toggle_divs.children[e].remove("active");
                    }
                    ato.div_el.classList.add("active");
            
                    if (ato.cb!=null) ato.cb(ato, true)
                };
            });
        }
        return $ats;
    }


    /**
     * simple toggle set of buttons
     */
    toggleBtnSet = function ($el, $nObjs, $classes, $callback, $defaultNid = null, $lbloverride = []) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($nObjs.length == 0) return;
        var elset = document.createElement("div");
        elset.classList = "toggle-btn-set";
        el.appendChild(elset);
        for (var n = 0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            var btnclass = $classes[n];
            var btn = document.createElement("button");
            btn.classList = "toggle-btn " + btnclass;
            if ($defaultNid && nObj.nid == $defaultNid) btn.classList.add("toggled-on");
            else btn.classList.add("toggled-off");
            btn.setAttribute("nid", nObj.nid);
            if ($lbloverride.length) btn.innerHTML = $lbloverride[n];
            else btn.innerHTML = nObj.text;
            elset.appendChild(btn);
            btn.addEventListener("click", async (e)=> {
                var target = e.target;
                //if this is off we are turning it on.
                //if it is on.. it stays on even when clicked again.
                if (target.classList.contains("toggled-off")) {
                    //turn off the curren toggled btn.
                    var btnset = ANNE.A.getFirstParentOf(target.parentNode, "toggle-btn-set", true);
                    if (btnset != null) {
                        var onbtn = btnset.querySelector(".toggled-on");
                        if (onbtn != null) {
                            onbtn.classList.remove("toggled-on");
                            onbtn.classList.add("toggled-off");
                        }
                    }
                    //turn on this one.
                    target.classList.remove("toggled-off");
                    target.classList.add("toggled-on");

                    //callback.
                    if ($callback != null && typeof $callback === "function") $callback(e);
                }
            });
        }
        return elset;
    }

    wireShowHideToggler=function($el_to_click, $el_showhide) {
        var el = this.__getEl($el_to_click);
        if (el == null) return;

		var elshowhide = null;
		if (el.classList.contains("hide")==true) {
			elshowhide = ANNE.A.setHTML($el_showhide, "show"); 
		}
		else {
			elshowhide = ANNE.A.setHTML($el_showhide, "hide"); 
		}

		if (elshowhide && elshowhide.getAttribute("showlink_id")==undefined) {
			elshowhide.setAttribute("showlink_id", el.id);
		}
		
        el.addEventListener("click", (e)=>{
            var newval = ANNE.A.toggle($el_showhide);
            //self fixing if things get whacky
            if (newval=="show") {
                if (el.innerHTML === 'show') el.innerHTML = 'hide';
            }
            else if(newval=="hide") {
                el.innerHTML = 'show';
            }
        });
    }

    toggleShowHideLink=function($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var el = document.getElementById('toggleLink');
        if (el.innerHTML === 'hide') el.innerHTML = 'show';
        else el.innerHTML = 'hide';
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $idToToggle 
     * @param {*} $addText 
     * @param {*} $defaultOpen 
     * @param {*} $ss 
     * @returns 
     */
    toggleMinMaxBtn = function ($el, $idToToggle, $addText = "", $defaultOpen = true, $ss = null, $addClass="") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("div");
        btn.setAttribute("toggle_divid", $idToToggle);
        btn.classList = "ezMinMax hover";
        if ($addClass!="") btn.classList.add($addClass)
        if ($defaultOpen == false) {
            btn.innerHTML = "<span class='ai-right-triangle'></span>" + $addText;
            ANNE.A.hide($idToToggle);
        }
        else {
            btn.innerHTML = "<span class='ai-down-triangle'></span>" + $addText;
            ANNE.A.fadeIn($idToToggle);
        }
        btn.style.cursor = "pointer";
        btn.style.display = "inline-block";
        el.append(btn);
        const ss = $ss;
        btn.addEventListener("click", async (e)=> {
            var target = e.target;
            target = ANNE.A.getFirstParentOf(target, "ezMinMax");
            var idToToggle = target.getAttribute("toggle_divid");
            if (idToToggle == undefined || idToToggle == null) return;
            var classz = ANNE.A.toggleWithFade(idToToggle);
            if (classz == "hide") {
                btn.innerHTML = "<span class='ai-right-triangle'></span>" + $addText;
            }
            else if (classz == "show") {
                btn.innerHTML = "<span class='ai-down-triangle'></span>" + $addText;
                var ssEl = this.__getEl(ss);

                //var percentagePosition = (el.offsetTop / ssEl.scrollHeight) * 100;
                //ssEl.scrollTop = (ssEl.scrollHeight * percentagePosition) / 100;
                //console.log("show section");
                //var scrollHeight = ssEl.scrollHeight;
                //ssEl.scrollTop = scrollHeight;
            }
        });

        return btn;
    }


    __getEl = function ($el) {
        if (typeof $el === "string") {
            if ($el[0] == ".") return document.querySelector($el);

            //legacy references to jquery will have # in front of them.
            else if ($el[0] == "#") return document.getElementById($el.substring(1));

            return document.getElementById($el);
        }
        return $el;
    }

    delayThen = async ($dataparam, $delayms = 300, $callback = null) => {
        await UTILS.sleep($delayms);
        if ($callback != null) $callback($dataparam);
    }

    //we may add a delay to this..
    fadeIn = function ($el, $ms = 500, $ease = "power1.in", $displayProp = "") {
        const el = this.__getEl($el);
        if (el == null) return null;
		el.style.opacity=0;
        if (el.style==undefined || el.style==null) {
            console.log("el style problem on el id ="+$el)
            return;
        }
        gsap.to(el, {
            opacity:1,
            duration:$ms/1000,
            ease:$ease,
            onComplete:function() {
                el.style.display = $displayProp;
                el.classList.add("show");
                el.classList.remove("hide");
            }
        });

        return el;
    }

    fadeInPop = function ($el, $addglowclass = "", $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("show");
        el.classList.remove("hide");
        el.style.opacity = 1;
        el.style.display = $displayProp;
        el.style.transition = "all 2000ms cubic-bezier(.39,.54,0,1.6);";
        if ($addglowclass != "") el.classList.add($addglowclass);
        return el;
    }


    fadeInDelayFadeOut = async ($el, $delayms, $inms = 500, $intt = "ease-in", $outms = 500, $outtt = "ease-out") => {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("show");
        el.classList.remove("hide");
        el.style.opacity = 1;
        el.style.display = "";
        el.style.transition = "all " + $innms + "ms " + $intt + ";";
        await UTILS.sleep($inms + $delayms);
        el.classList.add("hide");
        el.classList.remove("show");
        el.style.opacity = 0;
        el.style.display = "none";
        el.style.transition = "all " + $outms + "ms " + $outtt + ";";
        return el;
    }


    /**
     * This is blocking if it is awaited on!
     * Don't await it!
     */
    fadeOut = async ($el, $ms = 500, $delay = 0, $ease = "power1.out", $onComplete=null) => {
        const el = this.__getEl($el);
        if (el == null) return null;
        
        if (el.style==undefined || el.style==null) {
            console.log("el style problem on el id ="+$el)
            return el;
        }

        //max 2 sec delay to avoid accidents
        if ($delay > 0) await UTILS.sleep();

        gsap.to(el, {
            delay: Math.min(2, $delay/1000),
            opacity:0,
            duration:$ms/1000,
            ease:$ease,
            onComplete:function() {
                el.style.display = "none";
                el.classList.add("hide");
                el.classList.remove("show");
                if ($onComplete!=null) $onComplete();
            }
        });

        return el;
    }


    /**
     * Tired of doing this.. so I will do it here.
     */
    formatSpanStrings = function ($el, $classes, $vars, $str, $bSHOW = true) {
        var el = this.__getEl($el);
        if ($classes.length == 0) return $str;
        if ($classes.length != $vars.length) {
            console.log("(formatSpanStrings) classes array and vars array are different lengths.  This may not do what you think it should do.");
        }
        for (var c = 0; c < $classes.length; c++) {
            var clz = $classes[c];
            var i = $str.indexOf("%");
            if (i >= 0) $str = $str.replace("%", "<span class='" + clz + "'>" + $vars[c] + "</span>");
        }
        el.innerHTML = $str;
        if ($bSHOW == true) ANNE.A.show(el);
    }

    getBottom = function ($dfc = ".smart-scroll-framer") {
        return parseInt(document.querySelector($dfc).getBoundingClientRect().bottom);
    }

    getElIndex = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return -1;
        return Array.prototype.indexOf.call(el.parentElement.children, el);
    }

    //because i forget names.
    getNearestParentOf = function($el, $stopClass, $bINCLUDESELF = true) {
        return this.getFirstParentOf($el, $stopClass, $bINCLUDESELF);
    }

    getFirstParentOf = function ($el, $stopClass, $bINCLUDESELF = true) {
        if ($stopClass == "") return null;
        var el = this.__getEl($el);
        if (el == null) return null;
        var safety = 0;
        if ($bINCLUDESELF == true) {
            if (el.classList.contains($stopClass) == true) return el;
        }
        while (el.parentNode != null) {
            if (safety > 20) return null;
            el = el.parentNode;
            if (el == null || el == undefined) return null;
            if (el.nodeName == "#document") return null;
            if (el.classList.contains($stopClass) == true) return el;
            safety++;
        }
        return null;
    }

    getFirstParentAttr = function ($el, $stopClass, $getAttribute) {
        if ($getAttribute == "") return null;
        var el = this.getFirstParentOf($el, $stopClass);
        if (el == null) return null;
        return el.getAttribute($getAttribute);
    }


    getHTML = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.innerHTML;
    }

    getValue = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.value;
    }

    /**
     * input (w/ inputid) marked with class="rqd" wired to element with mod-el=inputid of rqd class el
     * displays helper msg
     * DOM wires rqd element with a .modxx ptr to "mod-el" span element
     * uses css::before to display color/status in the mod-el span.
     */
    applyRQD = function ($el, $bVALID = true, $msg = "required") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("rqd") == true && el.modxx != undefined) {
            if ($bVALID) {
                el.modxx.style['color'] = "#00db00";
                el.modxx.innerHTML = 'ok';
            }
            else {
                el.modxx.style['color'] = "red";
                el.modxx.innerHTML = $msg;
            }
        }
        return el;
    }


    hasClass = function ($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.classList.contains($class);
    }


    hide = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;

        if (el.getAttribute("showlink_id")!=undefined) {
			var showlink = document.getElementById(el.getAttribute("showlink_id"));
			if (showlink!=null) {
				showlink.innerHTML = "show";
			}
		}

        if (el.classList==undefined || el.classList==null) {
            alert("what in the world???");
            return;
        }
        if (el.classList.contains("hide") == false) el.classList.add("hide");
        if (el.classList.contains("show") == true) el.classList.remove("show");
        el.style.display = "none";
        return el;
    }
    hideParent = function ($el, $bHIDESELFTOO = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent_el = el.parentNode;
        this.hide(parent_el);
        if ($bHIDESELFTOO == true) this.hide(el);
    }

    hideMulti = function ($arrDivids) {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            el.classList.add("hide");
            el.classList.remove("show");
            el.style.display = "none";
        }
        return el;//return the last el
    }
    emptyAndHideMulti = function ($arrDivids) {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            el.classList.add("hide");
            el.classList.remove("show");
            el.style.display = "none";
            el.innerHTML="";
        }
        return el;//return the last el
    }

    showAll = function($arr) {
        var el = null;
        for (var a=0; a < $arr.length; a++) {
            el = ANNE.A.show($arr[a]);
        }
        return el;
    }

    hideAll = function ($arr) {
        var el = null;
        for (var a=0; a < $arr.length; a++) {
            el = ANNE.A.hide($arr[a]);
        }
        return el;
    }

    jsonSyntaxHighlight = function ($json) {
        if ($json == null) return "";
        if (typeof $json != 'string') {
            $json = JSON.stringify($json, undefined, 2);
        }
        $json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return $json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) {
                if (/:$/.test(match)) {
                    cls = 'key';
                } else {
                    cls = 'string';
                }
            } else if (/true|false/.test(match)) {
                cls = 'boolean';
            } else if (/null/.test(match)) {
                cls = 'null';
            }
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }



    masterThumbImg = function ($el, $beNOBJ, $nObj, $imgClass = "", $bNOIMGIMG = false) {
        var el = this.__getEl($el);
        if (el == null) return null;

        var img = document.createElement("img");
        if ($imgClass != "") img.classList = $imgClass;

        //get the thumb for this if we have it,
        //falls back on class pic if we don't have thumb for this.
        var url = MEM.__GET_DEFAULT_IMG_NEURON($beNOBJ, $nObj);
        if (url == "") {
            if ($bNOIMGIMG == true) img.setAttribute("src", "imgs/unknown.png");
        }
        else img.setAttribute("src", url);
        el.append(img);
        return img;
    }


    setElBorder = function ($el, $nObj, $pendingHanai = null) {
        var el = this.__getEl($el);
        if (el == null) return;

        if ($pendingHanai != null) el.classList.add("hanai-border");
        else if ($nObj.ghost == true) el.classList.add("ghost-border");
        else if ($nObj.freeze == true) el.classList.add("freeze-border");
        else if ($nObj.owned_by == USER._userNObj.nid || $nObj.nid == USER._userNObj.nid) el.classList.add("self-owned-border");
        else if (ANNE.isANNE($nObj.owned_by)==true) el.classList.add("anne-owned-border");
        else el.classList.add("other-owned-border");
    }

    setElBG = function ($el, $nObj, $pendingHanai = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($pendingHanai != null) el.classList.add("hanai-bg");
        else if ($nObj.ghost == true) el.classList.add("ghost-bg");
        else if ($nObj.freeze == true) el.classList.add("freeze-bg");
        else if ($nObj.owned_by == USER._userNObj.nid || $nObj.nid == USER._userNObj.nid) el.classList.add("self-owned-border");
        else if (ANNE.isANNE($nObj.owned_by)==true) el.classList.add("anne-owned-bg");
        else el.classList.add("other-owned-bg");
    }

    getImgBorder = function ($nObj, $postfix="") {
        if ($postfix!="") {
            if ($postfix.substring(0,1)!="-") $postfix="-"+$postfix;
        }
        if ($nObj == null) return "";
        if ($nObj.ghost == true) return "ghost-img"+$postfix;
        else if ($nObj.owned_by == USER._userNObj.nid) return "self-owned-img"+$postfix;
        else if (ANNE.isANNE($nObj.owned_by)==true) return "anne-owned-img"+$postfix;
        else return "other-owned-img"+$postfix;
    }


    injectPartFail = function ($x) {
        console.log("newNeuronFail");
        console.log($x);
    }


    /**
     * Update SIF options from outside
     * @param {*} $el_datalist 
     * @param {*} $nObjs 
     * @param {*} $bEMPTY 
     */
    addSIF_opts_externalupdate=function($el_datalist, $nObjs, $bEMPTY=true) {
        var el_datalist = this.__getEl($el_datalist);
        if ($bEMPTY==true) ANNE.A.empty(el_datalist);
        for (var n=0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            var option = document.createElement("option");
            option.setAttribute("value", nObj.bestSelectDisplay("", true, true));
            option.setAttribute("temp_nid", nObj.temp_nid);
            option.setAttribute("text", nObj.text); //for willy wonka datalist using value as lbl 
            el_datalist.append(option);
        }
    }
    

    /**
     * This is intended for lists and lookups that have potentially GINORMOUS result sets.
     * We don't want the entire universe of possibilities in a data object in the gui.
     * A neuron class with 100000 options would have 100k elements in a datalist.  This is silly.
     * 
     * Here we default to what the user has been doing in the app. (context of neurons)
     * 
     * When they click that local memory AND ANNODE memory is searched.
     * 
     * @param {Element or String} $el 
     * @param {Function} $search_cb 
     * @param {int} $seachCharMin 
     * @returns 
     */
    makeANNE_SIF=function($el, $search_locl_only_cb, $search_cb=null, $selected_cb=null, 
                                $defaultNOBJS=[], $searchCharMin=2, $bDODELETE=true, $findMoreClass="") {
        const el = this.__getEl($el);
        const local_cb = $search_locl_only_cb;
        const selected_cb = $selected_cb;

        if (el == null) return;
        if (el.type!="text" && el.nodeName!="INPUT"){
            console.log("to make an ANNE SIF we need to start with an input field!");
            return;
        }

        if(typeof $search_cb!=="function") {
            console.log("to make an ANNE SIF we need a search cb!");
            return;
        }

        el.classList.add("ANNE-SIF");

        const searchCharMin = $searchCharMin;

        el.setAttribute("autocomplete", "off");
        
        var bcr = el.getBoundingClientRect();
        var el_datalist = document.createElement("datalist");
        el_datalist.setAttribute("id", el.id+"__list");
        el.parentElement.append(el_datalist);
        el.setAttribute("list", el.id+"__list");       //tie input to data list.

        const btn = document.createElement("button");  
        btn.classList="btn-yellow-sq";
        if ($findMoreClass!="") btn.classList.add($findMoreClass);
        btn.setAttribute("id", el.id+"_sif");             
        btn.innerHTML = "find more";
        //btn.style.position="absolute";
        //btn.style['margin-left'] = (bcr.right-70)+"px";
        btn.setAttribute("list_id", el.id+"__list");
        el.parentNode.append(btn);
        if (bcr.height>0) {
            btn.style['max-height']=Math.floor(bcr.height)+2+"px";
            btn.style['height']=Math.floor(bcr.height)+2+"px";
        }
        else {
            btn.style['max-height']=32+"px";
            btn.style['height']=32+"px";
        }
        btn.style['max-width']='70px';
        btn.style['width']='70px';
        ANNE.A.hide(btn);
        
        const btn2 = document.createElement("button");  

        if ($bDODELETE) {
            btn2.classList="btn-red-sq";   
            btn2.setAttribute("id", el.id+"_sif_x");             
            btn2.innerHTML = "[x]";
            //btn2.style.position="absolute";
            //btn.style['margin-left'] = (bcr.right-40)+"px";
            btn2.setAttribute("list_id", el.id+"__list");
            el.parentNode.append(btn2);
            if (bcr.height>0) {
                btn2.style['max-height']=Math.floor(bcr.height)+2+"px";
                btn2.style['height']=Math.floor(bcr.height)+2+"px";
            }
            else {
                btn2.style['max-height']=32+"px";
                btn2.style['height']=32+"px";
            }
            btn2.style['max-width']='40px';
            btn2.style['width']='40px';

            //clear the selection.
            btn2.addEventListener("click", function(e){
                el.value="";
                ANNE.A.hide(btn);
                ANNE.A.hide(btn2);
                if (selected_cb!=null) selected_cb(-1);
                return;
            });

            ANNE.A.hide(btn2);
        }


        function addOpts($nObjs) {
            if ($nObjs==undefined || $nObjs==null) return;
            for (var n=0; n < $nObjs.length; n++) {
                var nObj = $nObjs[n];
                var option = document.createElement("option");
                option.setAttribute("value", nObj.bestSelectDisplay("", true, true));
                option.setAttribute("temp_nid", nObj.temp_nid);
                option.setAttribute("text", nObj.text); //for willy wonka datalist using value as lbl 
                el_datalist.append(option);
            }
        }
        //add in any default tnids (like local favs)
        if ($defaultNOBJS.length) addOpts($defaultNOBJS);

        el.setAttribute("search-roots", ""); 
        el.setAttribute("local-search-roots", ""); 

        function isInRoot($val, $which="search-roots") {
            var searchRootsArr = [];
            var search_roots = el.getAttribute($which);
            if (ANNE.UTILS.isset(search_roots)==true) {
                searchRootsArr = search_roots.split(",");
                for (var s=0; s < searchRootsArr.length; s++) {
                    if (searchRootsArr[s]=="") continue;
                    if ($val.indexOf(searchRootsArr[s])==0) {
                        return true;
                    }
                }
            }
            //jam the new search root val back in.  This optimizes lookups so we don't 
            //needlessly chase extended strings of strings we've already looked up!
            searchRootsArr.push($val.trim());
            el.setAttribute($which, searchRootsArr.join(","));
            return false;
        }

        var getActualVal = function() {
            //see if it matches anything we have in the datalist.
            //if so, it is selected -- so call the callback.
            //if there are multiples matching this text the user can use the dd to selecte which one they want.
            var selected_tnid = -1;
            var lowerz = el.value.toLowerCase();
            var actualVal = el.value;

            //find the selected datalist item by its unique name.
            for (var j = 0; j < el_datalist.options.length; j++) {
                if (el_datalist.options[j].value.toLowerCase()==lowerz) {
                    selected_tnid = parseInt(el_datalist.options[j].getAttribute("temp_nid") );
                    actualVal = el_datalist.options[j].getAttribute("text");
                    break;
                }
            }
            return {actualVal:actualVal, selected_tnid:selected_tnid};
        }

        
        //CALLED ON TEXT INPUT.
        var testLocal = async($e)=>{
            //willy wonka data lists only use the value prop as display.
            //if we want a disambig display in the option then that becomes the value.
            //use extra text attr in option for real text value.
            var avObj = getActualVal();
            var actualVal = avObj.actualVal;
            var selected_tnid = avObj.selected_tnid;

            //already searched for this and it is in the widget.
            //if (actualVal.length) {//} && isInRoot(actualVal, "local-search-roots")==false) {
            if (actualVal.length>0 && local_cb!=null) {
                ANNE.A.show("loading");
                var nObjs = await local_cb(actualVal);
                ANNE.A.hide("loading");
                if (ANNE.UTILS.isset(el_datalist)==false) {
                    el_datalist = document.getElementById(btn.getAttribute("list_id"));
                    if (ANNE.UTILS.isset(el_datalist)==false) return;
                }
                ANNE.A.empty(el_datalist);
                addOpts(nObjs);
            }
            //}

            if (selected_tnid>-1 && selected_cb!=null) {
                ANNE.A.hide(btn);
                ANNE.A.show(btn2);
                selected_cb(selected_tnid);
                return;
            }
            if (actualVal.length>searchCharMin) {
                ANNE.A.hide(btn2);
                ANNE.A.show(btn);
            }
            else {
                if (selected_tnid==-1 && selected_cb!=null) {
                    ANNE.A.hide(btn2);
                    ANNE.A.hide(btn);
                    selected_cb(-1);
                    return;
                }
            }
        }
    
        
        const returnKeyPress = async($e)=> {
            $e.preventDefault();
            if ($e.keyCode === 13) {
                if (btn.classList.contains("show")==true) {
                    var event = new MouseEvent('click', {
                        'view': window,
                        'bubbles': true,
                        'cancelable': true
                    })
                    btn.dispatchEvent(event);
                }
            }
        }

        //if more than the search min, open up the anne search btn.
        //if we haven't touched this search root yet, do a local ajax style search.
        el.addEventListener("input", $e=>testLocal($e));
        el.addEventListener("keyup", $e=>returnKeyPress($e));

        //find more button
        btn.addEventListener("click", async($e)=>{
            ANNE.A.show("loading");
            ANNE.A.hide(btn);

            var avObj = getActualVal();
            var actualVal = avObj.actualVal;
            
            //already searched for this and it is in the widget.
            if (actualVal.length>0 && isInRoot(actualVal, "search-roots")==true) return;

             //add in to local-search-roots too so data list feature takes over.
            if (actualVal.length>0) isInRoot(actualVal, "local-search-roots");  
        
            var nObjs = await $search_cb($e);
            if (ANNE.UTILS.isset(el_datalist)==false) {
                el_datalist = document.getElementById(btn.getAttribute("list_id"));
            }
            if (ANNE.UTILS.isset(el_datalist)==false) return;
            ANNE.A.empty(el_datalist);
            addOpts(nObjs);
            ANNE.A.hide("loading");
        });

        return btn;
    }


    setAsGhost=function($el, $class="ghost-text") {
        var el = this.__getEl($el);
        if (el == null) return;
        el.classList.add($class);
        var span=document.createElement("span");
        span.classList = "ghost-small-blue-ico";
        el.appendChild(span);
        el.placeholder = "Ghost neuron, must send TX to ANNE to make real.";
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $cb 
     * @returns 
     */
    makeQuickDeleteBtn = function($el, $cb=null) {
        var el = this.__getEl($el);
        if (el == null) return;
        const btn = document.createElement("button");  
        btn.classList="btn-red-sq";                
        btn.innerHTML = "[x]";
        el.append(btn);

        if ($cb!=null && typeof $cb ==="function") {
            btn.addEventListener("click", e=>$cb(e));
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $injid 
     * @param {*} $paramRelon 
     * @param {*} $GUID 
     * @param {*} $page 
     * @param {*} $existingClumpNid 
     * @param {*} $lbl 
     * @param {*} $bNEWOK 
     * @returns 
     */
    makeANTFromParam = function ($el, $injid, $paramRelon, $GUID, $page = "selector_action", $existingClumpNid = "", $lbl = "", $bNEWOK = true) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($paramRelon == undefined || $paramRelon == null) return;
        var relNOBJ = MEMG.getNeuronByTempNid($paramRelon.r);
        var toNOBJ = MEMG.getNeuronByTempNid($paramRelon.t);
        var antag = document.createElement("an");
        antag.setAttribute("role", relNOBJ.text);
        antag.setAttribute("rolereln", relNOBJ.nid);
        antag.setAttribute("newok", +$bNEWOK);
        antag.setAttribute("role_class_nid", toNOBJ.nid);
        if ($lbl == "") {
            var parts2 = relNOBJ.text.split("_");
            parts2.shift();
            $lbl = parts2.join("_");
        }
        antag.setAttribute("lbl", $lbl);
        antag.setAttribute("injid", $injid);
        antag.setAttribute("pguid", $GUID);
        antag.setAttribute("clump_nid", $existingClumpNid);
        antag.setAttribute("page", $page);
        el.append(antag);
    }

    /**
     * Using clone will destroy the existing events. (that's what it is here for)
     */
    moveElement = function ($subj2bmoved, $el_destination, $bCLONE = true) {
        if ($subj2bmoved == null) return;
        var el_parent = $subj2bmoved.parentElement;
        if (el_parent == null) return;
        var el_destination = this.__getEl($el_destination);
        if (el_destination == null) return;
        el_parent.removeChild($subj2bmoved);
        if ($bCLONE == true) {
            return this.cloneAndAppend($subj2bmoved, el_destination, true);
        }
        else el_destination.appendChild($subj2bmoved);
        return $subj2bmoved;
    }

    /**
     * Take a DOM node from its parent and move it to the spot of first child (prepend)
     */
    moveToFirst = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent = el.parentNode;
        if (parent == null) return;
        parent.removeChild(el);
        parent.prepend(el);
    }
    moveToEnd = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent = el.parentNode;
        if (parent == null) return;
        parent.removeChild(el);
        parent.append(el);
    }


    /**
     */
    parseANT_sentence = function ($ant) {
        console.log("%c (parseANT_sentence)", TALK.bgpink)
    }

    /**
     * Parse the ANTag (<an/>) which can show up in a blob.
     */
    parseANTags = function ($el, $bCHASE = false, $vars = null, $extraClass = "") {
        var el = this.__getEl($el);
        if (el == null) return;
        var ants = el.getElementsByTagName("an");

        for (var a = 0; a < ants.length; a++) {
            var ant = ants[a];

            //sentence type is for parsing sentence
            var type = ant.getAttribute("type");
            if (type != undefined) {
                console.log("%c ANT TAG TYPE " + type, TALK.bgpink)
                if (type == "sentence") {
                    this.parseANT_sentence(ant);
                    continue;
                }
            }

            var nid = ant.getAttribute("nid");
            if (nid != undefined && nid != "") {
                var nObj = MEMG.getNeuronByNid(nid);
                var txt = "";
                var dnrt = ant.getAttribute("dnrt");
                if (dnrt != null && dnrt == 1) {
                    txt = ant.innerHTML;
                }

                ant.innerHTML = "";

                if (nObj != null) {
                    EXT_BTNS.makeNeuronLink(ant, nid, $extraClass, "", false, txt);

                }
                else if ($bCHASE == true) {
                    console.log("we need to go get this value");
                }
            }

            // to do. create helpful injectable gui
            if ($vars != null) {
                var varname = ant.getAttribute("varname");
                if (varname != undefined) {
                    var dyno = $vars[varname];
                    if (dyno != undefined) {
                        ant.innerHTML = dyno;
                        ant.classList = "dynamic-value";
                    }
                }
            }

        }
    }


    /**
      * Parse the ANTag (<an/>) which can show up in a blob.
      * only do vars, not text/html tags.
      */
    parseANTagsVarsOnly = function ($el, $vars = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var ants = el.getElementsByTagName("an");
        for (var a = 0; a < ants.length; a++) {
            var ant = ants[a];
            if ($vars != null) {
                var varname = ant.getAttribute("varname");
                if (varname != undefined) {
                    var dyno = $vars[varname];
                    if (dyno != undefined) {
                        ant.innerHTML = dyno;
                        ant.classList = "dynamic-value";
                    }
                }
            }
        }
    }

    stopPlay = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.pause();
        el.currentTime = 0;
    }
    pausePlay = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.pause();
    }
    stopSound = function($el) { 
        this.stopPlay($el);
    }
    startSound = function($el) { this.play($el); }
    playSound = function($el) { this.play($el); }
    play = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.play();
    }

    pointerEventsNone = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.pointerEvents="none";
        ANNE.TALK.printf_console(TALK.RED, "pointerEventsNone");
        return el;
    }
    pointerEventsAll = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.pointerEvents="all";
        ANNE.TALK.printf_console(TALK.GREEN, "pointerEventsAll");
        return el;
    }

    removeSelf = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        el.parentNode.removeChild(el);
        return el;
    }

    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     * @param {*} $prefName 
     * @param {*} $featureProp
     * @param {*} $fpLbl
     * @param {*} $fpClass
     * 
     * @returns 
     */
    neuronIdentityDisplay = async($nObj, $bEMULATE=false, $prefName="", $featureProp="did_annex", $fpLbl="ANNEX'd neurons", $fpClass="color-green")=> {
        var anneid = $nObj._anneid;
        if (anneid==undefined || anneid==null) {
            if ($bEMULATE==true) {
                anneid = AWARE.emulateANNEID($nObj);
            }
            else {
                anneid = await ANNE.API.getANNEID($nObj, false);
            }
        }
        var el_acf = document.createElement("div");
        el_acf.classList = "anneid-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donutNONE";
        el_acf.appendChild(el_donut);
        
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        if ($nObj.getParent()==null) {
            await ANNE.CHASE.queryMinBem($nObj);
        }
        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="anneid-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "acn";
        var name = $prefName;
        if (name=="") name = $nObj.bestNameHTML();
        acn2.innerHTML = name;
        if (acn2.innerText.length>25) acn2.innerText = acn2.innerText.substring(0, 25);
        //if ($prefName[$prefName.length-1]!=")") acn2.innerText+=")";
        acn.appendChild(acn2);

        //anneid
        var fp = UTILS.commas(anneid[$featureProp]);
        var acs = document.createElement("div");
        acs.classList = "acs";
        acs.classList.add("acs-"+$nObj.nid);    //find the smart-reference.. however we update that.
        acs.innerHTML = "<span class='"+$fpClass+"'>"+fp+"</span>&nbsp;"+$fpLbl;
        acn.appendChild(acs);
        
        return el_acf;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {object} $match
     * @param {object} $arr
     * @returns 
     */
    neuronRevelatianMatchDisplay = async($nObj, $match, $lvlOffsets)=> {
        var beNOBJ = $nObj.getParent();
        if (beNOBJ==null) {
            console.log("nid:"+$nObj.nid+" text:"+$nObj.tex+" needs a parent");    
            return;
        }

        //console.log($match);

        var el_acf = document.createElement("div");
        el_acf.classList = "rev-circ-full";
        el_acf.setAttribute("overlay", true);
        el_acf.setAttribute("tnid", $nObj.temp_nid);
        el_acf.setAttribute("id", "match_tnid_"+$nObj.temp_nid);
            
            var el_donut = document.createElement("div");
            el_donut.classList = "donutNONE";
            el_acf.appendChild(el_donut);
    
            var p1 = document.createElement("p");
            p1.classList="n-circ-annex";
            el_acf.appendChild(p1);
                var sp = document.createElement("span");
                sp.classList = "REV-ico";
                p1.appendChild(sp);

                var sp2 = document.createElement("span");
                sp2.style="color:color-green font-size:20px; font-weight:900";
                sp2.innerHTML="";
                p1.appendChild(sp2)

            var p2 = document.createElement("p");
            p2.classList = "n-circ-imgfix";
            el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

            var el_revcirc = document.createElement("div");
            el_revcirc.classList="rev-circ-name";
            el_acf.appendChild(el_revcirc);
            
                var acn2 = document.createElement("div");
                acn2.classList = "rev";
                var name = $nObj.bestNameHTML();
                acn2.innerHTML = name;
                el_revcirc.appendChild(acn2);

                var lbl = document.createElement("div");
                lbl.classList="revelatian-sublbl";
                
                if ($nObj.ntype==NTYPE_INST) lbl.innerHTML = beNOBJ.bestNameHTML() + "<span class='rev-tiny-ntype'>inst</span>";
                else if ($nObj.ntype==NTYPE_CLASS) lbl.innerHTML = beNOBJ.bestNameHTML()+ "<span class='rev-tiny-ntype'>class</span>";
                el_revcirc.appendChild(lbl);

                //var lvloffset0 = document.createElement("span");
                //if ($lvlOffsets[0]!=undefined) {
                   // lvloffset0.innerHTML = "<span class='rev-lvl0ffset0'>"+$lvlOffsets[0]+"</span>";
                //}
				//lbl.appendChild(lvloffset0);
                //var lvloffset1 = document.createElement("span");
                //if ($lvlOffsets[1]!=undefined) {
                   // lvloffset1.innerHTML = "<span class='rev-lvl0ffset1'>"+$lvlOffsets[1]+"</span>";
                //}
				//lbl.appendChild(lvloffset1);

                var pathways = document.createElement("div");
                pathways.setAttribute("id", "pathways_"+$nObj.temp_nid);
                pathways.classList="revelatian-pathways";
                el_revcirc.appendChild(pathways);

        // var match = {tnid:$matchPt.tnid, myl:1, paths:paths};
        // paths (arraymap) = {pathid:path, hits:1}
        return {el:el_acf, el_pathways:pathways};
    }


    /**
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     */
    neuronRevelatianDisplay = async($nObj)=> {
        var beNOBJ = $nObj.getParent();
        if (beNOBJ==null) {
            console.log("nid:"+$nObj.nid+" text:"+$nObj.tex+" needs a parent");    
            return;
        }
        var el_acf = document.createElement("div");
        el_acf.classList = "rev-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donutNONE";
        el_acf.appendChild(el_donut);
       
        var p1 = document.createElement("p");
        p1.classList="n-circ-annex";
        el_acf.appendChild(p1);
        var sp = document.createElement("span");
        sp.classList = "REV-ico";
        p1.appendChild(sp)
        var sp2 = document.createElement("span");
        sp2.style="color:color-green font-size:20px; font-weight:900";
        sp2.innerHTML="";
        p1.appendChild(sp2)
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="rev-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "rev";
        var name = $nObj.bestNameHTML();
        if (name.length>25) name = name.substring(0, 25);
        acn2.innerHTML = name;
        acn.appendChild(acn2);

        var lbl = document.createElement("div");
        lbl.classList="revelatian-sublbl";
       
        lbl.innerHTML = "<span>parent&nbsp;<b>"+beNOBJ.bestNameHTML().substring(0, 25)+"</b></span>";
        acn.appendChild(lbl);
        return el_acf;
    }


    /**
     * make and populate an html element to display goodies for an annexable neuron
     * this is short-hand used for the user to click to get to a more in-depth view.
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     */
    neuronAplexDisplay = async($nObj, $bEMULATE=false, $prefName="", $colorClass="green", $icopost="2")=> {
        var apx = $nObj._aplex;
        if (apx==null) {
            if ($bEMULATE==true) {
                apx = new aplex($nObj, $nObj.nid, null);    //3rd param applex obj
                apx.emulateAplex();
            }
            else {
                apx = await ANNE.CHASE.get_aplex_data($nObj.nid, false, "neuronAplexDisplay");
            }
        }

        var el_acf = document.createElement("div");
        el_acf.classList = "awarez-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donut";
        el_acf.appendChild(el_donut);
        if (apx.posPer==null) apx.initPercents();
        var l1 = Math.round(apx.posPer*360);
        var l2 = Math.round(apx.negPer*360);
        var varstring = "--l1:0deg "+l1+"deg; --l2:"+l1+"deg "+(l1+l2)+"deg; --l3:"+(l1+l2)+"deg 360deg;";
        el_donut.style=varstring;

        var p1 = document.createElement("p");
        p1.classList="n-circ-annex";
        el_acf.appendChild(p1);
        var sp = document.createElement("span");
        sp.classList = "annex-"+$colorClass+"-ico"+$icopost;
        p1.appendChild(sp)
        var sp2 = document.createElement("span");
        sp2.style="color:color-"+$colorClass+" font-size:20px; font-weight:900";
        sp2.classList = "annex-price-lvl";
        sp2.classList.add("apltnid-"+$nObj.temp_nid);
        sp2.innerHTML=apx.price_lvl;
        p1.appendChild(sp2)
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="awarez-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "acn";
        if ($prefName!="") {
            acn2.innerHTML = $prefName;
        }
        else acn2.innerHTML = $nObj.text;
        acn.appendChild(acn2);

        if ($nObj.sponsored_by!="") {
            var spNOBJ = MEMG.getOrMakeIfNew($nObj.sponsored_by);
            var acs = document.createElement("div");
            acs.classList = "acs";
            acs.innerHTML = spNOBJ.bestName();
            acn.appendChild(acs);
        }

        return el_acf;
    }



    /**
     * 
     * @param {*} $el 
     * @returns 
     */
    removeParent = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        var parent = el.parentNode;
        return this.removeSelf(parent);
    }

    /**
     * 
     * @param {*} $el 
     * @returns 
     */
    removeParentParent = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        var parent = el.parentNode;
        var parentParent = parent.parentNode;
        if (parentParent == null) return this.removeSelf(parent);
        return this.removeSelf(parentParent);
    }


    removeAll=function($class) {
        const elements = document.querySelectorAll('.'+$class);
        elements.forEach(element => element.remove());
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $classname 
     * @returns 
     */
    removeClass = function ($el, $classname) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.remove($classname);

        //remove a 2nd, if there.
        if (el.classList.contains($classname)==true) el.classlist.remove($classname);
        
        return el;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $childClass 
     * @returns 
     */
    removeByParentAndClass = function ($el, $childClass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var els = el.querySelectorAll("." + $childClass);
        for (var e = 0; e < els.length; e++) {
            var elchild = els[e];
            el.removeChild(elchild);
        }
    }

    /**
     * remove the $removeClass from all element children inside of $el
     * that match the selector.  Adds it back to the targetEl.   
     * This is used to turn an active off, then on for the selected item.
     */
    removeAddClassOfAllIn = function ($el, $selector, $removeClass, $targetEl) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($selector[0] != "#" && $selector[0] != ".") $selector = "." + $selector; //presume class.
        el.querySelectorAll($selector).forEach(target => {
            target.classList.remove($removeClass);
        });
        var targetEl = this.__getEl($targetEl);
        if (targetEl == null) return;
        if (targetEl != null) targetEl.classList.add($removeClass);
    }

    setRadioVals = function ($pre, $numMax, $checkedVal) {
        for (var c = 0; c < $numMax; c++) {
            var el = this.__getEl($pre + "_" + c);
            if (el == null) return null;
            if ($checkedVal == c) el.checked = true;
            else el.checked = false;
        }
        return;
    }


    setBodyBackground = function ($url) {
        var newImgEl = document.createElement('img');
        newImgEl.addEventListener('load', function () {
            document.body.style['background-image'] = "url('" + $url + "')";
            ANNE.A.removeSelf(newImgEl);

        });
        newImgEl.setAttribute('src', $url); // Loads the image.
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $ease 
     * @returns 
     */
    gs_alphaOut=function($el, $duration=1, $ease="power1.out"){
        if (ANNE.gsap==null) ANNE.A.hide($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        var restore = el.style.display;
        if (el.getAttribute("display_restore")==undefined && restore!="" && restore!="none") {
            el.setAttribute("display_restore", restore);
        }

        if (ANNE.gsap==undefined || ANNE.gsap==null) {
            ANNE.A.hide(el);
            el.style.alpha=0;
            return;
        }
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration: $duration,
            alpha: 0,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if (el.classList.contains("hide")==false) el.classList.add("hide");
                el.style.display = "none";
            }
        } );
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $duration 
     * @param {*} $ease 
     * @returns 
     */
    gs_alphaScaleOut=function($el, $duration=1, $scale=0.1, $ease="power1.out"){
        if (ANNE.gsap==null) ANNE.A.hide($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        var restore = el.style.display;
        if (el.getAttribute("display_restore")==undefined && restore!="" && restore!="none") {
            el.setAttribute("display_restore", restore);
        }
        ANNE.gsap.to(el, {
            ease:$ease,
            duration: $duration,
            transform:'scale('+$scale+')',
            alpha: 0,
            onUpdate: function() {
            },
            onComplete: async()=> {
                el.classList.add("hide");
                el.style.display = "none";
            }
        } );
    }
    

    /**
     * 
     * @param {*} $el 
     * @param {*} $ease 
     * @param {*} $displayProp 
     * @returns 
     */
    gs_alphaIn=function($el, $duration=1, $ease="power1.in", $displayProp="", ){
        if (ANNE.gsap==null) ANNE.A.show($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        el.alpha=0;
        el.classList.remove("hide");

        if (ANNE.gsap==undefined || ANNE.gsap==null) {
            ANNE.A.show(el);
            el.style.alpha=1;
            return;
        }
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration:$duration,
            alpha: 1,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if ($displayProp!="") {
                    el.style.display=$displayProp;
                }
                else {
                    var restore = el.getAttribute("display_restore");
                    if (restore!=undefined) el.style.display=restore;
                    else el.style.display="";
                    el.classList.remove("hide"); //incase another animation
                }
            }
        } );
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $duration 
     * @param {*} $ease 
     * @param {*} $displayProp 
     * @returns 
     */
    gs_alphaScaleIn=function($el, $duration=1, $ease="power1.in", $scale=1, $displayProp="", ){
        if (ANNE.gsap==null) ANNE.A.show($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        el.alpha=0;
        el.classList.remove("hide");
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration:$duration,
            transform:'scale('+$scale+')',
            alpha: 1,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if ($displayProp!="") {
                    el.style.display=$displayProp;
                }
                else {
                    var restore = el.getAttribute("display_restore");
                    if (restore!=undefined) el.style.display=restore;
                    else el.style.display="";
                }
            }
        } );
    }


	showing = function($el) {
		var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show") == true) return true;
		if (el.classList.contains("hide") == true) return false; 
		if (el.style.display == "none") return false;
		return true;
	}
	
    /**
     * 
     * @param {*} $el 
     * @param {*} $displayProp 
     * @returns 
     */
    show = function ($el, $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show") == false) el.classList.add("show");
        if (el.classList.contains("hide") == true) el.classList.remove("hide");
        if (el.classList.contains("faded-neuron") == true) el.classList.remove("faded-neuron");

		if (el.getAttribute("showlink_id")!=undefined) {
			var showlink = document.getElementById(el.getAttribute("showlink_id"));
			if (showlink!=null) {
				showlink.innerHTML = "hide";
			}
		}
		
        if ($displayProp == "") {
            var restore = el.getAttribute("display_restore");
            if (restore!=null && restore.length>0) {
				el.style.display = restore;
			}
            else if (el.style.display == "none") {
				el.style.display = "";
			}
        }
        else el.style.display = $displayProp;
        return el;
    }


    showParent = function ($el, $bSHOWSELFTOO = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent_el = el.parentNode;
        this.show(parent_el);
        if ($bSHOWSELFTOO == true) this.show(el);
        return el;
    }

    showMulti = function ($arrDivids, $displayProp = "") {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            if (el == null) continue;
            el.classList.add("show");
            el.classList.remove("hide");
            el.style.display = $displayProp;
            el.style.opacity = 1; //correct in case a fadeOut was used prior
        }
        return el;//return the last el
    }

    setAttribute = function ($el, $attribute, $value, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.setAttribute($attribute, $value);
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setDeactiveWLbl = function ($el, $prepend = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($prepend == true) var dl = ANNE.A.ezElPrepend(el, "span", "dib deactivate-lbl", "");
        else var dl = ANNE.A.ezEl(el, "span", "dib deactivate-lbl", "");
        ANNE.A.ezEl(dl, "span", "dm", "");
        return dl;
    }

    setHTML = function ($el, $html, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.innerHTML = $html;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setJSON = function ($el, $json, $bSHOW = true, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($json == undefined || $json == null) return;
        var str = this.jsonSyntaxHighlight($json);
        el.innerHTML = str;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }


    /**
     * Set key up date event.
     * All dates get handled by our doDateFormat handler
     */
    setKeyUpDate = function ($el, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($callback != null) {
            el.addEventListener('keyup', function (e) {
                MEM.doDateFormat(e);
                $callback(e);
            });
        }
        else el.addEventListener('keyup', MEM.doDateFormat);
    }

    setKeyUpTime = function ($el, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($callback != null) {
            el.addEventListener('keyup', function (e) {
                MEM.doTimeFormat(e);
                $callback(e);
            });
        }
        else el.addEventListener('keyup', MEM.doTimeFormat);
    }


    setStyle = function ($el, $jsoncss, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var keys = Object.keys($jsoncss);
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsoncss[prop];
            el.style[prop] = val;
        }
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setStyleVar = function ($el, $prop, $val) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.setProperty($prop, $val);
    }


    setValue = function ($el, $value, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($value==null) {
            el.selectedIndex=0;
        }
        else el.value = $value;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $localRef 
     */
    soundFileLink = function($el, $localRef, $bDOWAVEFORM=true) {
        var el = this.__getEl($el);
        if (el == null) return null;
    
        
        var el_btn = document.createElement("span");
        el_btn.innerHTML = "▶️";
		el_btn.style.cursor = "pointer";
		el_btn.classList = "hoverable";
        el.appendChild(el_btn);

        
        var audio_tag = document.createElement("audio");
        audio_tag.src = $localRef;
        el.append(audio_tag);

        el_btn.addEventListener('click', function() {
            audio_tag.play();
        });


        if ($bDOWAVEFORM==false) return audio_tag;
		
        var el_wf_canvas = document.createElement("canvas");
		el_wf_canvas.style["height"]="90px";
		el_wf_canvas.style["min-width"]="100px";
		
        el.append(el_wf_canvas);
        this.drawWaveform(audio_tag, el_wf_canvas);

        return audio_tag;
    }


    /**
     * 
     * @param {*} $audio_tag 
     * @param {*} $cavasForWaveform 
     */
    drawWaveform=function($audio_tag, $cavasForWaveform) {
        const audioElement = this.__getEl($audio_tag);
        if (audioElement==null) return;

        const canvas = this.__getEl($cavasForWaveform);
        if (canvas==null) return;

        const audioContext = new (window.AudioContext || window.webkitAudioContext)();
        const audioSrc = audioContext.createMediaElementSource(audioElement);
        const analyser = audioContext.createAnalyser();

        audioSrc.connect(analyser);
        analyser.connect(audioContext.destination);

        const bufferLength = analyser.frequencyBinCount;
        const dataArray = new Uint8Array(bufferLength);

        const canvasCtx = canvas.getContext('2d');

        function draw() {
        const WIDTH = canvas.width;
        const HEIGHT = canvas.height;

        analyser.getByteTimeDomainData(dataArray);

        canvasCtx.fillStyle = 'rgb(200, 200, 200)';
        canvasCtx.fillRect(0, 0, WIDTH, HEIGHT);

        canvasCtx.lineWidth = 2;
        canvasCtx.strokeStyle = 'rgb(0, 0, 0)';

        canvasCtx.beginPath();

        const sliceWidth = (WIDTH * 1.0) / bufferLength;
        let x = 0;

        for (let i = 0; i < bufferLength; i++) {
            const v = dataArray[i] / 128.0;
            const y = (v * HEIGHT) / 2;

            if (i === 0) {
            canvasCtx.moveTo(x, y);
            } else {
            canvasCtx.lineTo(x, y);
            }

            x += sliceWidth;
        }

        canvasCtx.lineTo(canvas.width, canvas.height / 2);
        canvasCtx.stroke();

        requestAnimationFrame(draw);
        }

        draw();
    }


    copyIco = function ($el, $val) {
        var el = this.__getEl($el);
        if (el) {
            ANNE.A.empty(el);
            this.addCopyIco(el, $val);
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $data 
     * @returns 
     */
    addCopyIco=function($el, $data){
        if (ANNE.UTILS==null) return;
        var spanpre=document.createElement("button");
        spanpre.classList="ico-copy";
        spanpre.title = "copy to clipboard";
        spanpre.setAttribute("data-copy", $data);
        var sp = document.createElement("i")
        sp.classList = 'bi bi-clipboard-plus';
        spanpre.append(sp);
        spanpre.addEventListener("click", (e)=>{
            ANNE.playSound("click");
            e.stopPropagation();
            e.preventDefault();
            ANNE.UTILS.copyToClipboard($data); 
        });
        var el = this.__getEl($el)
        if (el==null) return;
        el.append(spanpre);
    }


    setMaxHeights = function ($el, $maxheight = 0, $height = 0) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($maxheight > 0) el.style["max-height"] = $maxheight + "px";
        if ($height > 0) el.style["height"] = $height + "px";
    }


    //updates any class, ie, 'bal-[temp_nid]' tags in the DOM
    quickBalanceUpdate = function ($class, $html) {
        var bals = document.querySelectorAll(".bal-" + USER._userNObj.temp_nid);
        for (var b = 0; b < bals.length; b++) {
            bals[b].innerHTML = $html;
        }
    }

    //return the span
    span = function ($html, $class = "") {
        var span = document.createElement("span");
        if ($class != "") span.classList = $class;
        span.innerHTML = $html;
        return span;
    }

    /* strip all occurances of a class out from an el or its decendants */
    stripClass = function ($el, $class, $addClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var children = $el.querySelectorAll("." + $class);
        if (children == null || children.length == 0) return;
        for (var c = 0; c < children.length; c++) {
            children[c].classList.remove($class);
            if ($addClass != "") children[c].classList.add($addClass);
        }
    }

    toggleWithFade = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show")) {
            this.fadeOut($el);
            return "hide";
        }
        else {
            this.fadeIn($el);
            return "show";
        }
    }

    toggle = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show")) {
            this.hide($el);
            return "hide";
        }
        else {
            this.show($el);
            return "show";
        }
    }


    toggleClass = function ($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($class) == true) el.classList.remove($class);
        else el.classList.add($class);
    }

    /**
     * if class1 is turned on (after the fucntion) it is -1
     * if class2 is turned on (after the function) it is 1
     */
    toggleClass2 = function ($el, $class1, $class2) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($class1) == true) {
            el.classList.remove($class1);
            el.classList.add($class2);
            return 1;
        }
        else if (el.classList.contains($class2) == true) {
            el.classList.remove($class2);
            el.classList.add($class1);
            return -1;
        }
        return 0;
    }

    triggerHover = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var event = new MouseEvent('mouseover', {
            'view': window,
            'bubbles': true,
            'cancelable': true
        });
        el.dispatchEvent(event);
    }


    setMaxH = function ($el, $maxH) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style['max-height'] = parseInt($maxH) + "px";
        el.style['height'] = parseInt($maxH) + "px"
    }

    toggleByClass = function ($el, $topSelector, $targetClassSelector, $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent = el.closest("." + $topSelector);
        if (parent != null) {
            var toggleTarget = parent.querySelector("." + $targetClassSelector);
            if (toggleTarget != null) {
                if (toggleTarget.classList.contains("show")) this.hide(toggleTarget);
                else this.show(toggleTarget, $displayProp);
            }
        }
    }


    /**
     * Forgot about this one.. should try it again (lol)
     * 
     * @param {*} $el 
     * @param {*} $msg 
     * @param {*} $hovertext 
     * @param {*} $attr 
     * @param {*} $classList 
     * @returns 
     */
    tooltip = function ($el, $msg, $hovertext = "hover over me", $attr = "data-text", $classList = "tooltip") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.classList = $classList;
        span.setAttribute($attr, $msg);
        span.innerHTML = "&nbsp;&#9432;"+$hovertext;
        el.append(span);
        return span;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $event 
     * @param {*} $bHIDE 
     * @returns 
     */
    trigger = function ($el, $event, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el[$event] != undefined && typeof el[$event] === 'function') {
            el[$event]();
        }
        else {
            console.log("(anne_DOM.trigger) cannot trigger event " + $event + " on element " + el.id);
            return;
        }
        if ($bHIDE == true) this.hide(el);
    }


    updateAll = function($el, $html) {
        var els = document.querySelectorAll($el);
        if (els == null || els.length==0) return null;
        for (var e=0; e < els.length; e++) {
            e.innerHTML = $html;
        }
    }


    /**
     * 
     * @param {*} $input_el 
     * @param {*} $showbtn 
     * @param {*} $rNObj 
     * @param {*} $voNOBJ 
     * @param {*} $txt 
     * @param {*} $validc 
     * @param {*} $invalidc 
     * @returns 
     */
    validator = function ($input_el, $showbtn, $rNObj = null, $voNOBJ = null, $txt = "", $validc = "valid", $invalidc = "invalid") {
        if ($txt.length) {
            if ($rNObj != null && $rNObj.format_validator != null) {
                var valid_check_rez = $rNObj.format_validator($txt);
                if (valid_check_rez != true) {
                    ANNE.A.hide($showbtn);
                    $input_el.classList.remove($validc);
                    $input_el.classList.add($invalidc);
                    return false;
                }
                else {
                    ANNE.A.show($showbtn);
                    $input_el.classList.remove($invalidc);
                    $input_el.classList.add($validc);
                    return true;
                }
            }
            else if ($voNOBJ != null && $voNOBJ.format_validator != null) {
                var valid_check_rez = $voNOBJ.format_validator($txt);
                if (valid_check_rez != true) {
                    ANNE.A.hide($showbtn);
                    $input_el.classList.remove($validc);
                    $input_el.classList.add($invalidc);
                    return false;
                }
                else {
                    ANNE.A.show($showbtn);
                    $input_el.classList.remove($invalidc);
                    $input_el.classList.add($validc);
                    return true;
                }
            }
            else {
                ANNE.A.show($showbtn);
                $input_el.classList.remove($invalidc);
                $input_el.classList.add($validc);
                return true;
            }
        }
        else {
            ANNE.A.hide($showbtn);
            $input_el.classList.remove($validc);
            $input_el.classList.add($invalidc);
            return false;
        }

        return false;
    }


    /**
     * YES, this is dangerous, if we were loading external scripts.  We aren't.
     * We are loading scripts in our own pages.
     * 
     * @param {*} $loadIntoEl 
     * @param {*} $pgui 
     * @param {*} $url 
     * @param {*} $loadClass 
     * @returns 
    */
    load = async ($loadIntoEl, $pgui, $url, $loadClass = "load-wrapper") => {
        var loadIntoEl = typeof $loadIntoEl === "string" ? document.getElementById($loadIntoEl) : loadIntoEl = $loadIntoEl;
        this.show(loadIntoEl);
        if (loadIntoEl == null) return;
        var xhr;
        if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
        else xhr = new ActiveXObject("Microsoft.XMLHTTP");

        var tempEl = document.createElement("div");
        if ($loadClass != "") tempEl.classList = $loadClass;

        return new Promise((resolve, reject) => {
            xhr.onreadystatechange = function () {
                if (this.readyState == 4 && this.status == 200) {
                    var resp = this.response;
                    loadIntoEl.innerHTML=resp;
                    var scripts = loadIntoEl.getElementsByTagName('script');
                    for (var s = 0; s < scripts.length; s++) {
                        var newscript = document.createElement("script");
                        newscript.type = 'text/javascript';
                        newscript.innerHTML = scripts[s].innerHTML;
                        if (scripts[s].parentNode == newEl) {
                            loadIntoEl.removeChild(scripts[s]);
                        }
                        loadIntoEl.append(newscript);
                        //eval(newscript);
                    }

                    //if this has a guid, clone it.  This is crucial to not
                    //getting DOM ID conflicts when using templates and injecting 
                    //them multiple times.
                    if ($pgui._guid && ($pgui.page.multiples == true || $pgui.page.instantiate == true)) {
                        var items = loadIntoEl.querySelectorAll('.dynamic');

                        //any divs marked as class=dynamic get their ID altered.
                        //this will allow us to have multiple copies of these templates in
                        //the dom and stil be able to uniquely reference their elements by id.
                        var mod_els = [];
                        var temp_list = [];

                        for (var i = 0; i < items.length; i++) {
                            var item = items[i];
                            if (item.id != null && item.id != "") {
                                temp_list[item.id.substring(0)] = item;
                                item.id = item.id + "_" + $pgui._guid;
                            }
                            var forattr = item.getAttribute("for");
                            if (forattr != null) item.setAttribute("for", forattr + "_" + $pgui._guid);
                            var nameattr = item.getAttribute("name");
                            if (nameattr != null) item.setAttribute("name", nameattr + "_" + $pgui._guid);
                            var mod_el_attr = item.getAttribute("mod-el");
                            if (mod_el_attr != null) {
                                item.setAttribute("mod-el", mod_el_attr + "_" + $pgui._guid);
                                mod_els.push({ el: item, mod_el_attr: mod_el_attr });
                            }
                        }

                        //create .modxx and tie it to the element that was tagged with mod-el
                        //and the id of the target.
                        for (var m = 0; m < mod_els.length; m++) {
                            var me = mod_els[m];
                            var srcel = temp_list[me.mod_el_attr];
                            if (srcel == undefined) continue;
                            srcel.modxx = me.el;
                        }

                        if ($pgui.pObj && $pgui.pObj.noshow != undefined);
                        else ANNE.A.show(loadIntoEl);
                        return resolve(loadIntoEl);
                    }

                    else {
                        if ($pgui.pObj && $pgui.pObj.noshow != undefined);
                        else ANNE.A.show(loadIntoEl);
                        return resolve(loadIntoEl);
                    }
                }
            };

            xhr.open("GET", $url + "?v=" + Math.random() * 10, true);
            xhr.responseType = "";
            xhr.send();
        });
    }


    /**
     * 
     * @param {*} $loadIntoEl 
     * @param {*} $spObj 
     * @param {*} $url 
     * @param {*} $loadClass 
     * @returns 
     */
    load2 = async ($loadIntoEl, $spObj, $url, $loadClass = "load-wrapper") => {
        var loadIntoEl = typeof $loadIntoEl === "string" ? document.getElementById($loadIntoEl) : loadIntoEl = $loadIntoEl;
        this.show(loadIntoEl);
        if (loadIntoEl == null) return;
        var xhr;
        if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
        else xhr = new ActiveXObject("Microsoft.XMLHTTP");

        var tempEl = document.createElement("div");
        if ($loadClass != "") tempEl.classList = $loadClass;

        return new Promise((resolve, reject) => {
            xhr.onreadystatechange = function () {
                if (this.readyState == 4 && this.status == 200) {
                    var resp = this.response;
                    var newEl = document.createElement("div");
                    if ($loadClass != "") newEl.classList = $loadClass;
                    else newEl.classList = "pichachu";
                    newEl.innerHTML = resp;
                    loadIntoEl.append(newEl);

                    //hmm. do we want to do this?  I suppose.
                    var scripts = newEl.getElementsByTagName('script');
                    for (var s = 0; s < scripts.length; s++) {
                        var newscript = document.createElement("script");
                        newscript.type = 'text/javascript';
                        newscript.innerHTML = scripts[s].innerHTML;
                        if (scripts[s].parentNode == newEl) {
                            newEl.removeChild(scripts[s]);
                        }
                        newEl.append(newscript);
                    }

                    //if this has a spObj._guid, clone it.  This is crucial to not
                    //getting DOM ID conflicts when using templates and injecting 
                    //them multiple times.
                    if ($spObj!=null && $spObj._guid!=null && $spObj._guid!="" && $spObj._guid.length>0) {
                        var items = newEl.querySelectorAll('.dynamic');
                        var guid=$spObj._guid;

                        //any divs marked as class=dynamic get their ID altered.
                        //this will allow us to have multiple copies of these templates in
                        //the dom and stil be able to uniquely reference their elements by id.
                        var mod_els = [];
                        var temp_list = [];

                        for (var i = 0; i < items.length; i++) {
                            var item = items[i];
                            if (item.id != null && item.id != "") {
                                temp_list[item.id.substring(0)] = item;
                                item.id = item.id + "_" + guid;
                            }
                            var forattr = item.getAttribute("for");
                            if (forattr != null) item.setAttribute("for", forattr + "_" + guid);
                            var nameattr = item.getAttribute("name");
                            if (nameattr != null) item.setAttribute("name", nameattr + "_" + guid);
                            var mod_el_attr = item.getAttribute("mod-el");
                            if (mod_el_attr != null) {
                                item.setAttribute("mod-el", mod_el_attr + "_" + guid);
                                mod_els.push({ el: item, mod_el_attr: mod_el_attr });
                            }
                        }

                        //create .modxx and tie it to the element that was tagged with mod-el
                        //and the id of the target.
                        for (var m = 0; m < mod_els.length; m++) {
                            var me = mod_els[m];
                            var srcel = temp_list[me.mod_el_attr];
                            if (srcel == undefined) continue;
                            srcel.modxx = me.el;
                        }
                        return resolve(newEl);
                    }

                    else {
                        return resolve(newEl);
                    }
                }
            };

            xhr.open("GET", $url + "?v=" + Math.random() * 10, true);
            xhr.responseType = "";
            xhr.send();
        });
    }
    
}

var A = new anne_dom();
A.set_init(A);

/************** (from file ./js/base/anne_DOM.js)**/ 

// Production steps of ECMA-262, Edition 5, 15.4.4.14
// Reference: https://es5.github.io/#x15.4.4.14
if (!Array.prototype.indexOf) {
    Array.prototype.indexOf = function(searchElement, fromIndex) {
      "use strict";
      var k;
  
      // 1. Let o be the result of calling ToObject passing
      //    the this value as the argument.
      if (this == null) {
        throw new TypeError('"this" is null or not defined');
      }
  
      var o = Object(this);
  
      // 2. Let lenValue be the result of calling the Get
      //    internal method of o with the argument "length".
      // 3. Let len be ToUint32(lenValue).
      var len = o.length >>> 0;
  
      // 4. If len is 0, return -1.
      if (len === 0) {
        return -1;
      }
  
      // 5. If argument fromIndex was passed let n be
      //    ToInteger(fromIndex); else let n be 0.
      var n = fromIndex | 0;
  
      // 6. If n >= len, return -1.
      if (n >= len) {
        return -1;
      }
  
      // 7. If n >= 0, then Let k be n.
      // 8. Else, n<0, Let k be len - abs(n).
      //    If k is less than 0, then let k be 0.
      k = Math.max(n >= 0 ? n : len - Math.abs(n), 0);
  
      // 9. Repeat, while k < len
      for (; k < len; k++) {
        // a. Let Pk be ToString(k).
        //   This is implicit for LHS operands of the in operator
        // b. Let kPresent be the result of calling the
        //    HasProperty internal method of o with argument Pk.
        //   This step can be combined with c
        // c. If kPresent is true, then
        //    i.  Let elementK be the result of calling the Get
        //        internal method of o with the argument ToString(k).
        //   ii.  Let same be the result of applying the
        //        Strict Equality Comparison Algorithm to
        //        searchElement and elementK.
        //  iii.  If same is true, return k.
        if (k in o && o[k] === searchElement)
          return k;
      }
      return -1;
    };
  }


  if (!NodeList.prototype.forEach && Array.prototype.forEach) {
    NodeList.prototype.forEach = Array.prototype.forEach;
}


String.prototype.indexOf_lc = function($str) {
  var str = this.toLowerCase();
  return str.indexOf($str.toLowerCase());
};


Array.prototype.indexOf_lc = function($str, $from_index=0) {
  var str = $str.toLowerCase();
  return this.map(function(value) {
      return value.toLowerCase();
  }).indexOf(str.toLowerCase(), $from_index);
};

/************** (from file ./js/other/browserfixes.js)**/ 

//! moment.js
//! version : 2.29.4
//! authors : Tim Wood, Iskren Chernev, Moment.js contributors
//! license : MIT
//! momentjs.com

;(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
    typeof define === 'function' && define.amd ? define(factory) :
    global.moment = factory()
}(this, (function () { 'use strict';

    var hookCallback;

    function hooks() {
        return hookCallback.apply(null, arguments);
    }

    // This is done to register the method called with moment()
    // without creating circular dependencies.
    function setHookCallback(callback) {
        hookCallback = callback;
    }

    function isArray(input) {
        return (
            input instanceof Array ||
            Object.prototype.toString.call(input) === '[object Array]'
        );
    }

    function isObject(input) {
        // IE8 will treat undefined and null as object if it wasn't for
        // input != null
        return (
            input != null &&
            Object.prototype.toString.call(input) === '[object Object]'
        );
    }

    function hasOwnProp(a, b) {
        return Object.prototype.hasOwnProperty.call(a, b);
    }

    function isObjectEmpty(obj) {
        if (Object.getOwnPropertyNames) {
            return Object.getOwnPropertyNames(obj).length === 0;
        } else {
            var k;
            for (k in obj) {
                if (hasOwnProp(obj, k)) {
                    return false;
                }
            }
            return true;
        }
    }

    function isUndefined(input) {
        return input === void 0;
    }

    function isNumber(input) {
        return (
            typeof input === 'number' ||
            Object.prototype.toString.call(input) === '[object Number]'
        );
    }

    function isDate(input) {
        return (
            input instanceof Date ||
            Object.prototype.toString.call(input) === '[object Date]'
        );
    }

    function map(arr, fn) {
        var res = [],
            i,
            arrLen = arr.length;
        for (i = 0; i < arrLen; ++i) {
            res.push(fn(arr[i], i));
        }
        return res;
    }

    function extend(a, b) {
        for (var i in b) {
            if (hasOwnProp(b, i)) {
                a[i] = b[i];
            }
        }

        if (hasOwnProp(b, 'toString')) {
            a.toString = b.toString;
        }

        if (hasOwnProp(b, 'valueOf')) {
            a.valueOf = b.valueOf;
        }

        return a;
    }

    function createUTC(input, format, locale, strict) {
        return createLocalOrUTC(input, format, locale, strict, true).utc();
    }

    function defaultParsingFlags() {
        // We need to deep clone this object.
        return {
            empty: false,
            unusedTokens: [],
            unusedInput: [],
            overflow: -2,
            charsLeftOver: 0,
            nullInput: false,
            invalidEra: null,
            invalidMonth: null,
            invalidFormat: false,
            userInvalidated: false,
            iso: false,
            parsedDateParts: [],
            era: null,
            meridiem: null,
            rfc2822: false,
            weekdayMismatch: false,
        };
    }

    function getParsingFlags(m) {
        if (m._pf == null) {
            m._pf = defaultParsingFlags();
        }
        return m._pf;
    }

    var some;
    if (Array.prototype.some) {
        some = Array.prototype.some;
    } else {
        some = function (fun) {
            var t = Object(this),
                len = t.length >>> 0,
                i;

            for (i = 0; i < len; i++) {
                if (i in t && fun.call(this, t[i], i, t)) {
                    return true;
                }
            }

            return false;
        };
    }

    function isValid(m) {
        if (m._isValid == null) {
            var flags = getParsingFlags(m),
                parsedParts = some.call(flags.parsedDateParts, function (i) {
                    return i != null;
                }),
                isNowValid =
                    !isNaN(m._d.getTime()) &&
                    flags.overflow < 0 &&
                    !flags.empty &&
                    !flags.invalidEra &&
                    !flags.invalidMonth &&
                    !flags.invalidWeekday &&
                    !flags.weekdayMismatch &&
                    !flags.nullInput &&
                    !flags.invalidFormat &&
                    !flags.userInvalidated &&
                    (!flags.meridiem || (flags.meridiem && parsedParts));

            if (m._strict) {
                isNowValid =
                    isNowValid &&
                    flags.charsLeftOver === 0 &&
                    flags.unusedTokens.length === 0 &&
                    flags.bigHour === undefined;
            }

            if (Object.isFrozen == null || !Object.isFrozen(m)) {
                m._isValid = isNowValid;
            } else {
                return isNowValid;
            }
        }
        return m._isValid;
    }

    function createInvalid(flags) {
        var m = createUTC(NaN);
        if (flags != null) {
            extend(getParsingFlags(m), flags);
        } else {
            getParsingFlags(m).userInvalidated = true;
        }

        return m;
    }

    // Plugins that add properties should also add the key here (null value),
    // so we can properly clone ourselves.
    var momentProperties = (hooks.momentProperties = []),
        updateInProgress = false;

    function copyConfig(to, from) {
        var i,
            prop,
            val,
            momentPropertiesLen = momentProperties.length;

        if (!isUndefined(from._isAMomentObject)) {
            to._isAMomentObject = from._isAMomentObject;
        }
        if (!isUndefined(from._i)) {
            to._i = from._i;
        }
        if (!isUndefined(from._f)) {
            to._f = from._f;
        }
        if (!isUndefined(from._l)) {
            to._l = from._l;
        }
        if (!isUndefined(from._strict)) {
            to._strict = from._strict;
        }
        if (!isUndefined(from._tzm)) {
            to._tzm = from._tzm;
        }
        if (!isUndefined(from._isUTC)) {
            to._isUTC = from._isUTC;
        }
        if (!isUndefined(from._offset)) {
            to._offset = from._offset;
        }
        if (!isUndefined(from._pf)) {
            to._pf = getParsingFlags(from);
        }
        if (!isUndefined(from._locale)) {
            to._locale = from._locale;
        }

        if (momentPropertiesLen > 0) {
            for (i = 0; i < momentPropertiesLen; i++) {
                prop = momentProperties[i];
                val = from[prop];
                if (!isUndefined(val)) {
                    to[prop] = val;
                }
            }
        }

        return to;
    }

    // Moment prototype object
    function Moment(config) {
        copyConfig(this, config);
        this._d = new Date(config._d != null ? config._d.getTime() : NaN);
        if (!this.isValid()) {
            this._d = new Date(NaN);
        }
        // Prevent infinite loop in case updateOffset creates new moment
        // objects.
        if (updateInProgress === false) {
            updateInProgress = true;
            hooks.updateOffset(this);
            updateInProgress = false;
        }
    }

    function isMoment(obj) {
        return (
            obj instanceof Moment || (obj != null && obj._isAMomentObject != null)
        );
    }

    function warn(msg) {
        if (
            hooks.suppressDeprecationWarnings === false &&
            typeof console !== 'undefined' &&
            console.warn
        ) {
            console.warn('Deprecation warning: ' + msg);
        }
    }

    function deprecate(msg, fn) {
        var firstTime = true;

        return extend(function () {
            if (hooks.deprecationHandler != null) {
                hooks.deprecationHandler(null, msg);
            }
            if (firstTime) {
                var args = [],
                    arg,
                    i,
                    key,
                    argLen = arguments.length;
                for (i = 0; i < argLen; i++) {
                    arg = '';
                    if (typeof arguments[i] === 'object') {
                        arg += '\n[' + i + '] ';
                        for (key in arguments[0]) {
                            if (hasOwnProp(arguments[0], key)) {
                                arg += key + ': ' + arguments[0][key] + ', ';
                            }
                        }
                        arg = arg.slice(0, -2); // Remove trailing comma and space
                    } else {
                        arg = arguments[i];
                    }
                    args.push(arg);
                }
                warn(
                    msg +
                        '\nArguments: ' +
                        Array.prototype.slice.call(args).join('') +
                        '\n' +
                        new Error().stack
                );
                firstTime = false;
            }
            return fn.apply(this, arguments);
        }, fn);
    }

    var deprecations = {};

    function deprecateSimple(name, msg) {
        if (hooks.deprecationHandler != null) {
            hooks.deprecationHandler(name, msg);
        }
        if (!deprecations[name]) {
            warn(msg);
            deprecations[name] = true;
        }
    }

    hooks.suppressDeprecationWarnings = false;
    hooks.deprecationHandler = null;

    function isFunction(input) {
        return (
            (typeof Function !== 'undefined' && input instanceof Function) ||
            Object.prototype.toString.call(input) === '[object Function]'
        );
    }

    function set(config) {
        var prop, i;
        for (i in config) {
            if (hasOwnProp(config, i)) {
                prop = config[i];
                if (isFunction(prop)) {
                    this[i] = prop;
                } else {
                    this['_' + i] = prop;
                }
            }
        }
        this._config = config;
        // Lenient ordinal parsing accepts just a number in addition to
        // number + (possibly) stuff coming from _dayOfMonthOrdinalParse.
        // TODO: Remove "ordinalParse" fallback in next major release.
        this._dayOfMonthOrdinalParseLenient = new RegExp(
            (this._dayOfMonthOrdinalParse.source || this._ordinalParse.source) +
                '|' +
                /\d{1,2}/.source
        );
    }

    function mergeConfigs(parentConfig, childConfig) {
        var res = extend({}, parentConfig),
            prop;
        for (prop in childConfig) {
            if (hasOwnProp(childConfig, prop)) {
                if (isObject(parentConfig[prop]) && isObject(childConfig[prop])) {
                    res[prop] = {};
                    extend(res[prop], parentConfig[prop]);
                    extend(res[prop], childConfig[prop]);
                } else if (childConfig[prop] != null) {
                    res[prop] = childConfig[prop];
                } else {
                    delete res[prop];
                }
            }
        }
        for (prop in parentConfig) {
            if (
                hasOwnProp(parentConfig, prop) &&
                !hasOwnProp(childConfig, prop) &&
                isObject(parentConfig[prop])
            ) {
                // make sure changes to properties don't modify parent config
                res[prop] = extend({}, res[prop]);
            }
        }
        return res;
    }

    function Locale(config) {
        if (config != null) {
            this.set(config);
        }
    }

    var keys;

    if (Object.keys) {
        keys = Object.keys;
    } else {
        keys = function (obj) {
            var i,
                res = [];
            for (i in obj) {
                if (hasOwnProp(obj, i)) {
                    res.push(i);
                }
            }
            return res;
        };
    }

    var defaultCalendar = {
        sameDay: '[Today at] LT',
        nextDay: '[Tomorrow at] LT',
        nextWeek: 'dddd [at] LT',
        lastDay: '[Yesterday at] LT',
        lastWeek: '[Last] dddd [at] LT',
        sameElse: 'L',
    };

    function calendar(key, mom, now) {
        var output = this._calendar[key] || this._calendar['sameElse'];
        return isFunction(output) ? output.call(mom, now) : output;
    }

    function zeroFill(number, targetLength, forceSign) {
        var absNumber = '' + Math.abs(number),
            zerosToFill = targetLength - absNumber.length,
            sign = number >= 0;
        return (
            (sign ? (forceSign ? '+' : '') : '-') +
            Math.pow(10, Math.max(0, zerosToFill)).toString().substr(1) +
            absNumber
        );
    }

    var formattingTokens =
            /(\[[^\[]*\])|(\\)?([Hh]mm(ss)?|Mo|MM?M?M?|Do|DDDo|DD?D?D?|ddd?d?|do?|w[o|w]?|W[o|W]?|Qo?|N{1,5}|YYYYYY|YYYYY|YYYY|YY|y{2,4}|yo?|gg(ggg?)?|GG(GGG?)?|e|E|a|A|hh?|HH?|kk?|mm?|ss?|S{1,9}|x|X|zz?|ZZ?|.)/g,
        localFormattingTokens = /(\[[^\[]*\])|(\\)?(LTS|LT|LL?L?L?|l{1,4})/g,
        formatFunctions = {},
        formatTokenFunctions = {};

    // token:    'M'
    // padded:   ['MM', 2]
    // ordinal:  'Mo'
    // callback: function () { this.month() + 1 }
    function addFormatToken(token, padded, ordinal, callback) {
        var func = callback;
        if (typeof callback === 'string') {
            func = function () {
                return this[callback]();
            };
        }
        if (token) {
            formatTokenFunctions[token] = func;
        }
        if (padded) {
            formatTokenFunctions[padded[0]] = function () {
                return zeroFill(func.apply(this, arguments), padded[1], padded[2]);
            };
        }
        if (ordinal) {
            formatTokenFunctions[ordinal] = function () {
                return this.localeData().ordinal(
                    func.apply(this, arguments),
                    token
                );
            };
        }
    }

    function removeFormattingTokens(input) {
        if (input.match(/\[[\s\S]/)) {
            return input.replace(/^\[|\]$/g, '');
        }
        return input.replace(/\\/g, '');
    }

    function makeFormatFunction(format) {
        var array = format.match(formattingTokens),
            i,
            length;

        for (i = 0, length = array.length; i < length; i++) {
            if (formatTokenFunctions[array[i]]) {
                array[i] = formatTokenFunctions[array[i]];
            } else {
                array[i] = removeFormattingTokens(array[i]);
            }
        }

        return function (mom) {
            var output = '',
                i;
            for (i = 0; i < length; i++) {
                output += isFunction(array[i])
                    ? array[i].call(mom, format)
                    : array[i];
            }
            return output;
        };
    }

    // format date using native date object
    function formatMoment(m, format) {
        if (!m.isValid()) {
            return m.localeData().invalidDate();
        }

        format = expandFormat(format, m.localeData());
        formatFunctions[format] =
            formatFunctions[format] || makeFormatFunction(format);

        return formatFunctions[format](m);
    }

    function expandFormat(format, locale) {
        var i = 5;

        function replaceLongDateFormatTokens(input) {
            return locale.longDateFormat(input) || input;
        }

        localFormattingTokens.lastIndex = 0;
        while (i >= 0 && localFormattingTokens.test(format)) {
            format = format.replace(
                localFormattingTokens,
                replaceLongDateFormatTokens
            );
            localFormattingTokens.lastIndex = 0;
            i -= 1;
        }

        return format;
    }

    var defaultLongDateFormat = {
        LTS: 'h:mm:ss A',
        LT: 'h:mm A',
        L: 'MM/DD/YYYY',
        LL: 'MMMM D, YYYY',
        LLL: 'MMMM D, YYYY h:mm A',
        LLLL: 'dddd, MMMM D, YYYY h:mm A',
    };

    function longDateFormat(key) {
        var format = this._longDateFormat[key],
            formatUpper = this._longDateFormat[key.toUpperCase()];

        if (format || !formatUpper) {
            return format;
        }

        this._longDateFormat[key] = formatUpper
            .match(formattingTokens)
            .map(function (tok) {
                if (
                    tok === 'MMMM' ||
                    tok === 'MM' ||
                    tok === 'DD' ||
                    tok === 'dddd'
                ) {
                    return tok.slice(1);
                }
                return tok;
            })
            .join('');

        return this._longDateFormat[key];
    }

    var defaultInvalidDate = 'Invalid date';

    function invalidDate() {
        return this._invalidDate;
    }

    var defaultOrdinal = '%d',
        defaultDayOfMonthOrdinalParse = /\d{1,2}/;

    function ordinal(number) {
        return this._ordinal.replace('%d', number);
    }

    var defaultRelativeTime = {
        future: 'in %s',
        past: '%s ago',
        s: 'a few seconds',
        ss: '%d seconds',
        m: 'a minute',
        mm: '%d minutes',
        h: 'an hour',
        hh: '%d hours',
        d: 'a day',
        dd: '%d days',
        w: 'a week',
        ww: '%d weeks',
        M: 'a month',
        MM: '%d months',
        y: 'a year',
        yy: '%d years',
    };

    function relativeTime(number, withoutSuffix, string, isFuture) {
        var output = this._relativeTime[string];
        return isFunction(output)
            ? output(number, withoutSuffix, string, isFuture)
            : output.replace(/%d/i, number);
    }

    function pastFuture(diff, output) {
        var format = this._relativeTime[diff > 0 ? 'future' : 'past'];
        return isFunction(format) ? format(output) : format.replace(/%s/i, output);
    }

    var aliases = {};

    function addUnitAlias(unit, shorthand) {
        var lowerCase = unit.toLowerCase();
        aliases[lowerCase] = aliases[lowerCase + 's'] = aliases[shorthand] = unit;
    }

    function normalizeUnits(units) {
        return typeof units === 'string'
            ? aliases[units] || aliases[units.toLowerCase()]
            : undefined;
    }

    function normalizeObjectUnits(inputObject) {
        var normalizedInput = {},
            normalizedProp,
            prop;

        for (prop in inputObject) {
            if (hasOwnProp(inputObject, prop)) {
                normalizedProp = normalizeUnits(prop);
                if (normalizedProp) {
                    normalizedInput[normalizedProp] = inputObject[prop];
                }
            }
        }

        return normalizedInput;
    }

    var priorities = {};

    function addUnitPriority(unit, priority) {
        priorities[unit] = priority;
    }

    function getPrioritizedUnits(unitsObj) {
        var units = [],
            u;
        for (u in unitsObj) {
            if (hasOwnProp(unitsObj, u)) {
                units.push({ unit: u, priority: priorities[u] });
            }
        }
        units.sort(function (a, b) {
            return a.priority - b.priority;
        });
        return units;
    }

    function isLeapYear(year) {
        return (year % 4 === 0 && year % 100 !== 0) || year % 400 === 0;
    }

    function absFloor(number) {
        if (number < 0) {
            // -0 -> 0
            return Math.ceil(number) || 0;
        } else {
            return Math.floor(number);
        }
    }

    function toInt(argumentForCoercion) {
        var coercedNumber = +argumentForCoercion,
            value = 0;

        if (coercedNumber !== 0 && isFinite(coercedNumber)) {
            value = absFloor(coercedNumber);
        }

        return value;
    }

    function makeGetSet(unit, keepTime) {
        return function (value) {
            if (value != null) {
                set$1(this, unit, value);
                hooks.updateOffset(this, keepTime);
                return this;
            } else {
                return get(this, unit);
            }
        };
    }

    function get(mom, unit) {
        return mom.isValid()
            ? mom._d['get' + (mom._isUTC ? 'UTC' : '') + unit]()
            : NaN;
    }

    function set$1(mom, unit, value) {
        if (mom.isValid() && !isNaN(value)) {
            if (
                unit === 'FullYear' &&
                isLeapYear(mom.year()) &&
                mom.month() === 1 &&
                mom.date() === 29
            ) {
                value = toInt(value);
                mom._d['set' + (mom._isUTC ? 'UTC' : '') + unit](
                    value,
                    mom.month(),
                    daysInMonth(value, mom.month())
                );
            } else {
                mom._d['set' + (mom._isUTC ? 'UTC' : '') + unit](value);
            }
        }
    }

    // MOMENTS

    function stringGet(units) {
        units = normalizeUnits(units);
        if (isFunction(this[units])) {
            return this[units]();
        }
        return this;
    }

    function stringSet(units, value) {
        if (typeof units === 'object') {
            units = normalizeObjectUnits(units);
            var prioritized = getPrioritizedUnits(units),
                i,
                prioritizedLen = prioritized.length;
            for (i = 0; i < prioritizedLen; i++) {
                this[prioritized[i].unit](units[prioritized[i].unit]);
            }
        } else {
            units = normalizeUnits(units);
            if (isFunction(this[units])) {
                return this[units](value);
            }
        }
        return this;
    }

    var match1 = /\d/, //       0 - 9
        match2 = /\d\d/, //      00 - 99
        match3 = /\d{3}/, //     000 - 999
        match4 = /\d{4}/, //    0000 - 9999
        match6 = /[+-]?\d{6}/, // -999999 - 999999
        match1to2 = /\d\d?/, //       0 - 99
        match3to4 = /\d\d\d\d?/, //     999 - 9999
        match5to6 = /\d\d\d\d\d\d?/, //   99999 - 999999
        match1to3 = /\d{1,3}/, //       0 - 999
        match1to4 = /\d{1,4}/, //       0 - 9999
        match1to6 = /[+-]?\d{1,6}/, // -999999 - 999999
        matchUnsigned = /\d+/, //       0 - inf
        matchSigned = /[+-]?\d+/, //    -inf - inf
        matchOffset = /Z|[+-]\d\d:?\d\d/gi, // +00:00 -00:00 +0000 -0000 or Z
        matchShortOffset = /Z|[+-]\d\d(?::?\d\d)?/gi, // +00 -00 +00:00 -00:00 +0000 -0000 or Z
        matchTimestamp = /[+-]?\d+(\.\d{1,3})?/, // 123456789 123456789.123
        // any word (or two) characters or numbers including two/three word month in arabic.
        // includes scottish gaelic two word and hyphenated months
        matchWord =
            /[0-9]{0,256}['a-z\u00A0-\u05FF\u0700-\uD7FF\uF900-\uFDCF\uFDF0-\uFF07\uFF10-\uFFEF]{1,256}|[\u0600-\u06FF\/]{1,256}(\s*?[\u0600-\u06FF]{1,256}){1,2}/i,
        regexes;

    regexes = {};

    function addRegexToken(token, regex, strictRegex) {
        regexes[token] = isFunction(regex)
            ? regex
            : function (isStrict, localeData) {
                  return isStrict && strictRegex ? strictRegex : regex;
              };
    }

    function getParseRegexForToken(token, config) {
        if (!hasOwnProp(regexes, token)) {
            return new RegExp(unescapeFormat(token));
        }

        return regexes[token](config._strict, config._locale);
    }

    // Code from http://stackoverflow.com/questions/3561493/is-there-a-regexp-escape-function-in-javascript
    function unescapeFormat(s) {
        return regexEscape(
            s
                .replace('\\', '')
                .replace(
                    /\\(\[)|\\(\])|\[([^\]\[]*)\]|\\(.)/g,
                    function (matched, p1, p2, p3, p4) {
                        return p1 || p2 || p3 || p4;
                    }
                )
        );
    }

    function regexEscape(s) {
        return s.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&');
    }

    var tokens = {};

    function addParseToken(token, callback) {
        var i,
            func = callback,
            tokenLen;
        if (typeof token === 'string') {
            token = [token];
        }
        if (isNumber(callback)) {
            func = function (input, array) {
                array[callback] = toInt(input);
            };
        }
        tokenLen = token.length;
        for (i = 0; i < tokenLen; i++) {
            tokens[token[i]] = func;
        }
    }

    function addWeekParseToken(token, callback) {
        addParseToken(token, function (input, array, config, token) {
            config._w = config._w || {};
            callback(input, config._w, config, token);
        });
    }

    function addTimeToArrayFromToken(token, input, config) {
        if (input != null && hasOwnProp(tokens, token)) {
            tokens[token](input, config._a, config, token);
        }
    }

    var YEAR = 0,
        MONTH = 1,
        DATE = 2,
        HOUR = 3,
        MINUTE = 4,
        SECOND = 5,
        MILLISECOND = 6,
        WEEK = 7,
        WEEKDAY = 8;

    function mod(n, x) {
        return ((n % x) + x) % x;
    }

    var indexOf;

    if (Array.prototype.indexOf) {
        indexOf = Array.prototype.indexOf;
    } else {
        indexOf = function (o) {
            // I know
            var i;
            for (i = 0; i < this.length; ++i) {
                if (this[i] === o) {
                    return i;
                }
            }
            return -1;
        };
    }

    function daysInMonth(year, month) {
        if (isNaN(year) || isNaN(month)) {
            return NaN;
        }
        var modMonth = mod(month, 12);
        year += (month - modMonth) / 12;
        return modMonth === 1
            ? isLeapYear(year)
                ? 29
                : 28
            : 31 - ((modMonth % 7) % 2);
    }

    // FORMATTING

    addFormatToken('M', ['MM', 2], 'Mo', function () {
        return this.month() + 1;
    });

    addFormatToken('MMM', 0, 0, function (format) {
        return this.localeData().monthsShort(this, format);
    });

    addFormatToken('MMMM', 0, 0, function (format) {
        return this.localeData().months(this, format);
    });

    // ALIASES

    addUnitAlias('month', 'M');

    // PRIORITY

    addUnitPriority('month', 8);

    // PARSING

    addRegexToken('M', match1to2);
    addRegexToken('MM', match1to2, match2);
    addRegexToken('MMM', function (isStrict, locale) {
        return locale.monthsShortRegex(isStrict);
    });
    addRegexToken('MMMM', function (isStrict, locale) {
        return locale.monthsRegex(isStrict);
    });

    addParseToken(['M', 'MM'], function (input, array) {
        array[MONTH] = toInt(input) - 1;
    });

    addParseToken(['MMM', 'MMMM'], function (input, array, config, token) {
        var month = config._locale.monthsParse(input, token, config._strict);
        // if we didn't find a month name, mark the date as invalid.
        if (month != null) {
            array[MONTH] = month;
        } else {
            getParsingFlags(config).invalidMonth = input;
        }
    });

    // LOCALES

    var defaultLocaleMonths =
            'January_February_March_April_May_June_July_August_September_October_November_December'.split(
                '_'
            ),
        defaultLocaleMonthsShort =
            'Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec'.split('_'),
        MONTHS_IN_FORMAT = /D[oD]?(\[[^\[\]]*\]|\s)+MMMM?/,
        defaultMonthsShortRegex = matchWord,
        defaultMonthsRegex = matchWord;

    function localeMonths(m, format) {
        if (!m) {
            return isArray(this._months)
                ? this._months
                : this._months['standalone'];
        }
        return isArray(this._months)
            ? this._months[m.month()]
            : this._months[
                  (this._months.isFormat || MONTHS_IN_FORMAT).test(format)
                      ? 'format'
                      : 'standalone'
              ][m.month()];
    }

    function localeMonthsShort(m, format) {
        if (!m) {
            return isArray(this._monthsShort)
                ? this._monthsShort
                : this._monthsShort['standalone'];
        }
        return isArray(this._monthsShort)
            ? this._monthsShort[m.month()]
            : this._monthsShort[
                  MONTHS_IN_FORMAT.test(format) ? 'format' : 'standalone'
              ][m.month()];
    }

    function handleStrictParse(monthName, format, strict) {
        var i,
            ii,
            mom,
            llc = monthName.toLocaleLowerCase();
        if (!this._monthsParse) {
            // this is not used
            this._monthsParse = [];
            this._longMonthsParse = [];
            this._shortMonthsParse = [];
            for (i = 0; i < 12; ++i) {
                mom = createUTC([2000, i]);
                this._shortMonthsParse[i] = this.monthsShort(
                    mom,
                    ''
                ).toLocaleLowerCase();
                this._longMonthsParse[i] = this.months(mom, '').toLocaleLowerCase();
            }
        }

        if (strict) {
            if (format === 'MMM') {
                ii = indexOf.call(this._shortMonthsParse, llc);
                return ii !== -1 ? ii : null;
            } else {
                ii = indexOf.call(this._longMonthsParse, llc);
                return ii !== -1 ? ii : null;
            }
        } else {
            if (format === 'MMM') {
                ii = indexOf.call(this._shortMonthsParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._longMonthsParse, llc);
                return ii !== -1 ? ii : null;
            } else {
                ii = indexOf.call(this._longMonthsParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._shortMonthsParse, llc);
                return ii !== -1 ? ii : null;
            }
        }
    }

    function localeMonthsParse(monthName, format, strict) {
        var i, mom, regex;

        if (this._monthsParseExact) {
            return handleStrictParse.call(this, monthName, format, strict);
        }

        if (!this._monthsParse) {
            this._monthsParse = [];
            this._longMonthsParse = [];
            this._shortMonthsParse = [];
        }

        // TODO: add sorting
        // Sorting makes sure if one month (or abbr) is a prefix of another
        // see sorting in computeMonthsParse
        for (i = 0; i < 12; i++) {
            // make the regex if we don't have it already
            mom = createUTC([2000, i]);
            if (strict && !this._longMonthsParse[i]) {
                this._longMonthsParse[i] = new RegExp(
                    '^' + this.months(mom, '').replace('.', '') + '$',
                    'i'
                );
                this._shortMonthsParse[i] = new RegExp(
                    '^' + this.monthsShort(mom, '').replace('.', '') + '$',
                    'i'
                );
            }
            if (!strict && !this._monthsParse[i]) {
                regex =
                    '^' + this.months(mom, '') + '|^' + this.monthsShort(mom, '');
                this._monthsParse[i] = new RegExp(regex.replace('.', ''), 'i');
            }
            // test the regex
            if (
                strict &&
                format === 'MMMM' &&
                this._longMonthsParse[i].test(monthName)
            ) {
                return i;
            } else if (
                strict &&
                format === 'MMM' &&
                this._shortMonthsParse[i].test(monthName)
            ) {
                return i;
            } else if (!strict && this._monthsParse[i].test(monthName)) {
                return i;
            }
        }
    }

    // MOMENTS

    function setMonth(mom, value) {
        var dayOfMonth;

        if (!mom.isValid()) {
            // No op
            return mom;
        }

        if (typeof value === 'string') {
            if (/^\d+$/.test(value)) {
                value = toInt(value);
            } else {
                value = mom.localeData().monthsParse(value);
                // TODO: Another silent failure?
                if (!isNumber(value)) {
                    return mom;
                }
            }
        }

        dayOfMonth = Math.min(mom.date(), daysInMonth(mom.year(), value));
        mom._d['set' + (mom._isUTC ? 'UTC' : '') + 'Month'](value, dayOfMonth);
        return mom;
    }

    function getSetMonth(value) {
        if (value != null) {
            setMonth(this, value);
            hooks.updateOffset(this, true);
            return this;
        } else {
            return get(this, 'Month');
        }
    }

    function getDaysInMonth() {
        return daysInMonth(this.year(), this.month());
    }

    function monthsShortRegex(isStrict) {
        if (this._monthsParseExact) {
            if (!hasOwnProp(this, '_monthsRegex')) {
                computeMonthsParse.call(this);
            }
            if (isStrict) {
                return this._monthsShortStrictRegex;
            } else {
                return this._monthsShortRegex;
            }
        } else {
            if (!hasOwnProp(this, '_monthsShortRegex')) {
                this._monthsShortRegex = defaultMonthsShortRegex;
            }
            return this._monthsShortStrictRegex && isStrict
                ? this._monthsShortStrictRegex
                : this._monthsShortRegex;
        }
    }

    function monthsRegex(isStrict) {
        if (this._monthsParseExact) {
            if (!hasOwnProp(this, '_monthsRegex')) {
                computeMonthsParse.call(this);
            }
            if (isStrict) {
                return this._monthsStrictRegex;
            } else {
                return this._monthsRegex;
            }
        } else {
            if (!hasOwnProp(this, '_monthsRegex')) {
                this._monthsRegex = defaultMonthsRegex;
            }
            return this._monthsStrictRegex && isStrict
                ? this._monthsStrictRegex
                : this._monthsRegex;
        }
    }

    function computeMonthsParse() {
        function cmpLenRev(a, b) {
            return b.length - a.length;
        }

        var shortPieces = [],
            longPieces = [],
            mixedPieces = [],
            i,
            mom;
        for (i = 0; i < 12; i++) {
            // make the regex if we don't have it already
            mom = createUTC([2000, i]);
            shortPieces.push(this.monthsShort(mom, ''));
            longPieces.push(this.months(mom, ''));
            mixedPieces.push(this.months(mom, ''));
            mixedPieces.push(this.monthsShort(mom, ''));
        }
        // Sorting makes sure if one month (or abbr) is a prefix of another it
        // will match the longer piece.
        shortPieces.sort(cmpLenRev);
        longPieces.sort(cmpLenRev);
        mixedPieces.sort(cmpLenRev);
        for (i = 0; i < 12; i++) {
            shortPieces[i] = regexEscape(shortPieces[i]);
            longPieces[i] = regexEscape(longPieces[i]);
        }
        for (i = 0; i < 24; i++) {
            mixedPieces[i] = regexEscape(mixedPieces[i]);
        }

        this._monthsRegex = new RegExp('^(' + mixedPieces.join('|') + ')', 'i');
        this._monthsShortRegex = this._monthsRegex;
        this._monthsStrictRegex = new RegExp(
            '^(' + longPieces.join('|') + ')',
            'i'
        );
        this._monthsShortStrictRegex = new RegExp(
            '^(' + shortPieces.join('|') + ')',
            'i'
        );
    }

    // FORMATTING

    addFormatToken('Y', 0, 0, function () {
        var y = this.year();
        return y <= 9999 ? zeroFill(y, 4) : '+' + y;
    });

    addFormatToken(0, ['YY', 2], 0, function () {
        return this.year() % 100;
    });

    addFormatToken(0, ['YYYY', 4], 0, 'year');
    addFormatToken(0, ['YYYYY', 5], 0, 'year');
    addFormatToken(0, ['YYYYYY', 6, true], 0, 'year');

    // ALIASES

    addUnitAlias('year', 'y');

    // PRIORITIES

    addUnitPriority('year', 1);

    // PARSING

    addRegexToken('Y', matchSigned);
    addRegexToken('YY', match1to2, match2);
    addRegexToken('YYYY', match1to4, match4);
    addRegexToken('YYYYY', match1to6, match6);
    addRegexToken('YYYYYY', match1to6, match6);

    addParseToken(['YYYYY', 'YYYYYY'], YEAR);
    addParseToken('YYYY', function (input, array) {
        array[YEAR] =
            input.length === 2 ? hooks.parseTwoDigitYear(input) : toInt(input);
    });
    addParseToken('YY', function (input, array) {
        array[YEAR] = hooks.parseTwoDigitYear(input);
    });
    addParseToken('Y', function (input, array) {
        array[YEAR] = parseInt(input, 10);
    });

    // HELPERS

    function daysInYear(year) {
        return isLeapYear(year) ? 366 : 365;
    }

    // HOOKS

    hooks.parseTwoDigitYear = function (input) {
        return toInt(input) + (toInt(input) > 68 ? 1900 : 2000);
    };

    // MOMENTS

    var getSetYear = makeGetSet('FullYear', true);

    function getIsLeapYear() {
        return isLeapYear(this.year());
    }

    function createDate(y, m, d, h, M, s, ms) {
        // can't just apply() to create a date:
        // https://stackoverflow.com/q/181348
        var date;
        // the date constructor remaps years 0-99 to 1900-1999
        if (y < 100 && y >= 0) {
            // preserve leap years using a full 400 year cycle, then reset
            date = new Date(y + 400, m, d, h, M, s, ms);
            if (isFinite(date.getFullYear())) {
                date.setFullYear(y);
            }
        } else {
            date = new Date(y, m, d, h, M, s, ms);
        }

        return date;
    }

    function createUTCDate(y) {
        var date, args;
        // the Date.UTC function remaps years 0-99 to 1900-1999
        if (y < 100 && y >= 0) {
            args = Array.prototype.slice.call(arguments);
            // preserve leap years using a full 400 year cycle, then reset
            args[0] = y + 400;
            date = new Date(Date.UTC.apply(null, args));
            if (isFinite(date.getUTCFullYear())) {
                date.setUTCFullYear(y);
            }
        } else {
            date = new Date(Date.UTC.apply(null, arguments));
        }

        return date;
    }

    // start-of-first-week - start-of-year
    function firstWeekOffset(year, dow, doy) {
        var // first-week day -- which january is always in the first week (4 for iso, 1 for other)
            fwd = 7 + dow - doy,
            // first-week day local weekday -- which local weekday is fwd
            fwdlw = (7 + createUTCDate(year, 0, fwd).getUTCDay() - dow) % 7;

        return -fwdlw + fwd - 1;
    }

    // https://en.wikipedia.org/wiki/ISO_week_date#Calculating_a_date_given_the_year.2C_week_number_and_weekday
    function dayOfYearFromWeeks(year, week, weekday, dow, doy) {
        var localWeekday = (7 + weekday - dow) % 7,
            weekOffset = firstWeekOffset(year, dow, doy),
            dayOfYear = 1 + 7 * (week - 1) + localWeekday + weekOffset,
            resYear,
            resDayOfYear;

        if (dayOfYear <= 0) {
            resYear = year - 1;
            resDayOfYear = daysInYear(resYear) + dayOfYear;
        } else if (dayOfYear > daysInYear(year)) {
            resYear = year + 1;
            resDayOfYear = dayOfYear - daysInYear(year);
        } else {
            resYear = year;
            resDayOfYear = dayOfYear;
        }

        return {
            year: resYear,
            dayOfYear: resDayOfYear,
        };
    }

    function weekOfYear(mom, dow, doy) {
        var weekOffset = firstWeekOffset(mom.year(), dow, doy),
            week = Math.floor((mom.dayOfYear() - weekOffset - 1) / 7) + 1,
            resWeek,
            resYear;

        if (week < 1) {
            resYear = mom.year() - 1;
            resWeek = week + weeksInYear(resYear, dow, doy);
        } else if (week > weeksInYear(mom.year(), dow, doy)) {
            resWeek = week - weeksInYear(mom.year(), dow, doy);
            resYear = mom.year() + 1;
        } else {
            resYear = mom.year();
            resWeek = week;
        }

        return {
            week: resWeek,
            year: resYear,
        };
    }

    function weeksInYear(year, dow, doy) {
        var weekOffset = firstWeekOffset(year, dow, doy),
            weekOffsetNext = firstWeekOffset(year + 1, dow, doy);
        return (daysInYear(year) - weekOffset + weekOffsetNext) / 7;
    }

    // FORMATTING

    addFormatToken('w', ['ww', 2], 'wo', 'week');
    addFormatToken('W', ['WW', 2], 'Wo', 'isoWeek');

    // ALIASES

    addUnitAlias('week', 'w');
    addUnitAlias('isoWeek', 'W');

    // PRIORITIES

    addUnitPriority('week', 5);
    addUnitPriority('isoWeek', 5);

    // PARSING

    addRegexToken('w', match1to2);
    addRegexToken('ww', match1to2, match2);
    addRegexToken('W', match1to2);
    addRegexToken('WW', match1to2, match2);

    addWeekParseToken(
        ['w', 'ww', 'W', 'WW'],
        function (input, week, config, token) {
            week[token.substr(0, 1)] = toInt(input);
        }
    );

    // HELPERS

    // LOCALES

    function localeWeek(mom) {
        return weekOfYear(mom, this._week.dow, this._week.doy).week;
    }

    var defaultLocaleWeek = {
        dow: 0, // Sunday is the first day of the week.
        doy: 6, // The week that contains Jan 6th is the first week of the year.
    };

    function localeFirstDayOfWeek() {
        return this._week.dow;
    }

    function localeFirstDayOfYear() {
        return this._week.doy;
    }

    // MOMENTS

    function getSetWeek(input) {
        var week = this.localeData().week(this);
        return input == null ? week : this.add((input - week) * 7, 'd');
    }

    function getSetISOWeek(input) {
        var week = weekOfYear(this, 1, 4).week;
        return input == null ? week : this.add((input - week) * 7, 'd');
    }

    // FORMATTING

    addFormatToken('d', 0, 'do', 'day');

    addFormatToken('dd', 0, 0, function (format) {
        return this.localeData().weekdaysMin(this, format);
    });

    addFormatToken('ddd', 0, 0, function (format) {
        return this.localeData().weekdaysShort(this, format);
    });

    addFormatToken('dddd', 0, 0, function (format) {
        return this.localeData().weekdays(this, format);
    });

    addFormatToken('e', 0, 0, 'weekday');
    addFormatToken('E', 0, 0, 'isoWeekday');

    // ALIASES

    addUnitAlias('day', 'd');
    addUnitAlias('weekday', 'e');
    addUnitAlias('isoWeekday', 'E');

    // PRIORITY
    addUnitPriority('day', 11);
    addUnitPriority('weekday', 11);
    addUnitPriority('isoWeekday', 11);

    // PARSING

    addRegexToken('d', match1to2);
    addRegexToken('e', match1to2);
    addRegexToken('E', match1to2);
    addRegexToken('dd', function (isStrict, locale) {
        return locale.weekdaysMinRegex(isStrict);
    });
    addRegexToken('ddd', function (isStrict, locale) {
        return locale.weekdaysShortRegex(isStrict);
    });
    addRegexToken('dddd', function (isStrict, locale) {
        return locale.weekdaysRegex(isStrict);
    });

    addWeekParseToken(['dd', 'ddd', 'dddd'], function (input, week, config, token) {
        var weekday = config._locale.weekdaysParse(input, token, config._strict);
        // if we didn't get a weekday name, mark the date as invalid
        if (weekday != null) {
            week.d = weekday;
        } else {
            getParsingFlags(config).invalidWeekday = input;
        }
    });

    addWeekParseToken(['d', 'e', 'E'], function (input, week, config, token) {
        week[token] = toInt(input);
    });

    // HELPERS

    function parseWeekday(input, locale) {
        if (typeof input !== 'string') {
            return input;
        }

        if (!isNaN(input)) {
            return parseInt(input, 10);
        }

        input = locale.weekdaysParse(input);
        if (typeof input === 'number') {
            return input;
        }

        return null;
    }

    function parseIsoWeekday(input, locale) {
        if (typeof input === 'string') {
            return locale.weekdaysParse(input) % 7 || 7;
        }
        return isNaN(input) ? null : input;
    }

    // LOCALES
    function shiftWeekdays(ws, n) {
        return ws.slice(n, 7).concat(ws.slice(0, n));
    }

    var defaultLocaleWeekdays =
            'Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday'.split('_'),
        defaultLocaleWeekdaysShort = 'Sun_Mon_Tue_Wed_Thu_Fri_Sat'.split('_'),
        defaultLocaleWeekdaysMin = 'Su_Mo_Tu_We_Th_Fr_Sa'.split('_'),
        defaultWeekdaysRegex = matchWord,
        defaultWeekdaysShortRegex = matchWord,
        defaultWeekdaysMinRegex = matchWord;

    function localeWeekdays(m, format) {
        var weekdays = isArray(this._weekdays)
            ? this._weekdays
            : this._weekdays[
                  m && m !== true && this._weekdays.isFormat.test(format)
                      ? 'format'
                      : 'standalone'
              ];
        return m === true
            ? shiftWeekdays(weekdays, this._week.dow)
            : m
            ? weekdays[m.day()]
            : weekdays;
    }

    function localeWeekdaysShort(m) {
        return m === true
            ? shiftWeekdays(this._weekdaysShort, this._week.dow)
            : m
            ? this._weekdaysShort[m.day()]
            : this._weekdaysShort;
    }

    function localeWeekdaysMin(m) {
        return m === true
            ? shiftWeekdays(this._weekdaysMin, this._week.dow)
            : m
            ? this._weekdaysMin[m.day()]
            : this._weekdaysMin;
    }

    function handleStrictParse$1(weekdayName, format, strict) {
        var i,
            ii,
            mom,
            llc = weekdayName.toLocaleLowerCase();
        if (!this._weekdaysParse) {
            this._weekdaysParse = [];
            this._shortWeekdaysParse = [];
            this._minWeekdaysParse = [];

            for (i = 0; i < 7; ++i) {
                mom = createUTC([2000, 1]).day(i);
                this._minWeekdaysParse[i] = this.weekdaysMin(
                    mom,
                    ''
                ).toLocaleLowerCase();
                this._shortWeekdaysParse[i] = this.weekdaysShort(
                    mom,
                    ''
                ).toLocaleLowerCase();
                this._weekdaysParse[i] = this.weekdays(mom, '').toLocaleLowerCase();
            }
        }

        if (strict) {
            if (format === 'dddd') {
                ii = indexOf.call(this._weekdaysParse, llc);
                return ii !== -1 ? ii : null;
            } else if (format === 'ddd') {
                ii = indexOf.call(this._shortWeekdaysParse, llc);
                return ii !== -1 ? ii : null;
            } else {
                ii = indexOf.call(this._minWeekdaysParse, llc);
                return ii !== -1 ? ii : null;
            }
        } else {
            if (format === 'dddd') {
                ii = indexOf.call(this._weekdaysParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._shortWeekdaysParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._minWeekdaysParse, llc);
                return ii !== -1 ? ii : null;
            } else if (format === 'ddd') {
                ii = indexOf.call(this._shortWeekdaysParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._weekdaysParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._minWeekdaysParse, llc);
                return ii !== -1 ? ii : null;
            } else {
                ii = indexOf.call(this._minWeekdaysParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._weekdaysParse, llc);
                if (ii !== -1) {
                    return ii;
                }
                ii = indexOf.call(this._shortWeekdaysParse, llc);
                return ii !== -1 ? ii : null;
            }
        }
    }

    function localeWeekdaysParse(weekdayName, format, strict) {
        var i, mom, regex;

        if (this._weekdaysParseExact) {
            return handleStrictParse$1.call(this, weekdayName, format, strict);
        }

        if (!this._weekdaysParse) {
            this._weekdaysParse = [];
            this._minWeekdaysParse = [];
            this._shortWeekdaysParse = [];
            this._fullWeekdaysParse = [];
        }

        for (i = 0; i < 7; i++) {
            // make the regex if we don't have it already

            mom = createUTC([2000, 1]).day(i);
            if (strict && !this._fullWeekdaysParse[i]) {
                this._fullWeekdaysParse[i] = new RegExp(
                    '^' + this.weekdays(mom, '').replace('.', '\\.?') + '$',
                    'i'
                );
                this._shortWeekdaysParse[i] = new RegExp(
                    '^' + this.weekdaysShort(mom, '').replace('.', '\\.?') + '$',
                    'i'
                );
                this._minWeekdaysParse[i] = new RegExp(
                    '^' + this.weekdaysMin(mom, '').replace('.', '\\.?') + '$',
                    'i'
                );
            }
            if (!this._weekdaysParse[i]) {
                regex =
                    '^' +
                    this.weekdays(mom, '') +
                    '|^' +
                    this.weekdaysShort(mom, '') +
                    '|^' +
                    this.weekdaysMin(mom, '');
                this._weekdaysParse[i] = new RegExp(regex.replace('.', ''), 'i');
            }
            // test the regex
            if (
                strict &&
                format === 'dddd' &&
                this._fullWeekdaysParse[i].test(weekdayName)
            ) {
                return i;
            } else if (
                strict &&
                format === 'ddd' &&
                this._shortWeekdaysParse[i].test(weekdayName)
            ) {
                return i;
            } else if (
                strict &&
                format === 'dd' &&
                this._minWeekdaysParse[i].test(weekdayName)
            ) {
                return i;
            } else if (!strict && this._weekdaysParse[i].test(weekdayName)) {
                return i;
            }
        }
    }

    // MOMENTS

    function getSetDayOfWeek(input) {
        if (!this.isValid()) {
            return input != null ? this : NaN;
        }
        var day = this._isUTC ? this._d.getUTCDay() : this._d.getDay();
        if (input != null) {
            input = parseWeekday(input, this.localeData());
            return this.add(input - day, 'd');
        } else {
            return day;
        }
    }

    function getSetLocaleDayOfWeek(input) {
        if (!this.isValid()) {
            return input != null ? this : NaN;
        }
        var weekday = (this.day() + 7 - this.localeData()._week.dow) % 7;
        return input == null ? weekday : this.add(input - weekday, 'd');
    }

    function getSetISODayOfWeek(input) {
        if (!this.isValid()) {
            return input != null ? this : NaN;
        }

        // behaves the same as moment#day except
        // as a getter, returns 7 instead of 0 (1-7 range instead of 0-6)
        // as a setter, sunday should belong to the previous week.

        if (input != null) {
            var weekday = parseIsoWeekday(input, this.localeData());
            return this.day(this.day() % 7 ? weekday : weekday - 7);
        } else {
            return this.day() || 7;
        }
    }

    function weekdaysRegex(isStrict) {
        if (this._weekdaysParseExact) {
            if (!hasOwnProp(this, '_weekdaysRegex')) {
                computeWeekdaysParse.call(this);
            }
            if (isStrict) {
                return this._weekdaysStrictRegex;
            } else {
                return this._weekdaysRegex;
            }
        } else {
            if (!hasOwnProp(this, '_weekdaysRegex')) {
                this._weekdaysRegex = defaultWeekdaysRegex;
            }
            return this._weekdaysStrictRegex && isStrict
                ? this._weekdaysStrictRegex
                : this._weekdaysRegex;
        }
    }

    function weekdaysShortRegex(isStrict) {
        if (this._weekdaysParseExact) {
            if (!hasOwnProp(this, '_weekdaysRegex')) {
                computeWeekdaysParse.call(this);
            }
            if (isStrict) {
                return this._weekdaysShortStrictRegex;
            } else {
                return this._weekdaysShortRegex;
            }
        } else {
            if (!hasOwnProp(this, '_weekdaysShortRegex')) {
                this._weekdaysShortRegex = defaultWeekdaysShortRegex;
            }
            return this._weekdaysShortStrictRegex && isStrict
                ? this._weekdaysShortStrictRegex
                : this._weekdaysShortRegex;
        }
    }

    function weekdaysMinRegex(isStrict) {
        if (this._weekdaysParseExact) {
            if (!hasOwnProp(this, '_weekdaysRegex')) {
                computeWeekdaysParse.call(this);
            }
            if (isStrict) {
                return this._weekdaysMinStrictRegex;
            } else {
                return this._weekdaysMinRegex;
            }
        } else {
            if (!hasOwnProp(this, '_weekdaysMinRegex')) {
                this._weekdaysMinRegex = defaultWeekdaysMinRegex;
            }
            return this._weekdaysMinStrictRegex && isStrict
                ? this._weekdaysMinStrictRegex
                : this._weekdaysMinRegex;
        }
    }

    function computeWeekdaysParse() {
        function cmpLenRev(a, b) {
            return b.length - a.length;
        }

        var minPieces = [],
            shortPieces = [],
            longPieces = [],
            mixedPieces = [],
            i,
            mom,
            minp,
            shortp,
            longp;
        for (i = 0; i < 7; i++) {
            // make the regex if we don't have it already
            mom = createUTC([2000, 1]).day(i);
            minp = regexEscape(this.weekdaysMin(mom, ''));
            shortp = regexEscape(this.weekdaysShort(mom, ''));
            longp = regexEscape(this.weekdays(mom, ''));
            minPieces.push(minp);
            shortPieces.push(shortp);
            longPieces.push(longp);
            mixedPieces.push(minp);
            mixedPieces.push(shortp);
            mixedPieces.push(longp);
        }
        // Sorting makes sure if one weekday (or abbr) is a prefix of another it
        // will match the longer piece.
        minPieces.sort(cmpLenRev);
        shortPieces.sort(cmpLenRev);
        longPieces.sort(cmpLenRev);
        mixedPieces.sort(cmpLenRev);

        this._weekdaysRegex = new RegExp('^(' + mixedPieces.join('|') + ')', 'i');
        this._weekdaysShortRegex = this._weekdaysRegex;
        this._weekdaysMinRegex = this._weekdaysRegex;

        this._weekdaysStrictRegex = new RegExp(
            '^(' + longPieces.join('|') + ')',
            'i'
        );
        this._weekdaysShortStrictRegex = new RegExp(
            '^(' + shortPieces.join('|') + ')',
            'i'
        );
        this._weekdaysMinStrictRegex = new RegExp(
            '^(' + minPieces.join('|') + ')',
            'i'
        );
    }

    // FORMATTING

    function hFormat() {
        return this.hours() % 12 || 12;
    }

    function kFormat() {
        return this.hours() || 24;
    }

    addFormatToken('H', ['HH', 2], 0, 'hour');
    addFormatToken('h', ['hh', 2], 0, hFormat);
    addFormatToken('k', ['kk', 2], 0, kFormat);

    addFormatToken('hmm', 0, 0, function () {
        return '' + hFormat.apply(this) + zeroFill(this.minutes(), 2);
    });

    addFormatToken('hmmss', 0, 0, function () {
        return (
            '' +
            hFormat.apply(this) +
            zeroFill(this.minutes(), 2) +
            zeroFill(this.seconds(), 2)
        );
    });

    addFormatToken('Hmm', 0, 0, function () {
        return '' + this.hours() + zeroFill(this.minutes(), 2);
    });

    addFormatToken('Hmmss', 0, 0, function () {
        return (
            '' +
            this.hours() +
            zeroFill(this.minutes(), 2) +
            zeroFill(this.seconds(), 2)
        );
    });

    function meridiem(token, lowercase) {
        addFormatToken(token, 0, 0, function () {
            return this.localeData().meridiem(
                this.hours(),
                this.minutes(),
                lowercase
            );
        });
    }

    meridiem('a', true);
    meridiem('A', false);

    // ALIASES

    addUnitAlias('hour', 'h');

    // PRIORITY
    addUnitPriority('hour', 13);

    // PARSING

    function matchMeridiem(isStrict, locale) {
        return locale._meridiemParse;
    }

    addRegexToken('a', matchMeridiem);
    addRegexToken('A', matchMeridiem);
    addRegexToken('H', match1to2);
    addRegexToken('h', match1to2);
    addRegexToken('k', match1to2);
    addRegexToken('HH', match1to2, match2);
    addRegexToken('hh', match1to2, match2);
    addRegexToken('kk', match1to2, match2);

    addRegexToken('hmm', match3to4);
    addRegexToken('hmmss', match5to6);
    addRegexToken('Hmm', match3to4);
    addRegexToken('Hmmss', match5to6);

    addParseToken(['H', 'HH'], HOUR);
    addParseToken(['k', 'kk'], function (input, array, config) {
        var kInput = toInt(input);
        array[HOUR] = kInput === 24 ? 0 : kInput;
    });
    addParseToken(['a', 'A'], function (input, array, config) {
        config._isPm = config._locale.isPM(input);
        config._meridiem = input;
    });
    addParseToken(['h', 'hh'], function (input, array, config) {
        array[HOUR] = toInt(input);
        getParsingFlags(config).bigHour = true;
    });
    addParseToken('hmm', function (input, array, config) {
        var pos = input.length - 2;
        array[HOUR] = toInt(input.substr(0, pos));
        array[MINUTE] = toInt(input.substr(pos));
        getParsingFlags(config).bigHour = true;
    });
    addParseToken('hmmss', function (input, array, config) {
        var pos1 = input.length - 4,
            pos2 = input.length - 2;
        array[HOUR] = toInt(input.substr(0, pos1));
        array[MINUTE] = toInt(input.substr(pos1, 2));
        array[SECOND] = toInt(input.substr(pos2));
        getParsingFlags(config).bigHour = true;
    });
    addParseToken('Hmm', function (input, array, config) {
        var pos = input.length - 2;
        array[HOUR] = toInt(input.substr(0, pos));
        array[MINUTE] = toInt(input.substr(pos));
    });
    addParseToken('Hmmss', function (input, array, config) {
        var pos1 = input.length - 4,
            pos2 = input.length - 2;
        array[HOUR] = toInt(input.substr(0, pos1));
        array[MINUTE] = toInt(input.substr(pos1, 2));
        array[SECOND] = toInt(input.substr(pos2));
    });

    // LOCALES

    function localeIsPM(input) {
        // IE8 Quirks Mode & IE7 Standards Mode do not allow accessing strings like arrays
        // Using charAt should be more compatible.
        return (input + '').toLowerCase().charAt(0) === 'p';
    }

    var defaultLocaleMeridiemParse = /[ap]\.?m?\.?/i,
        // Setting the hour should keep the time, because the user explicitly
        // specified which hour they want. So trying to maintain the same hour (in
        // a new timezone) makes sense. Adding/subtracting hours does not follow
        // this rule.
        getSetHour = makeGetSet('Hours', true);

    function localeMeridiem(hours, minutes, isLower) {
        if (hours > 11) {
            return isLower ? 'pm' : 'PM';
        } else {
            return isLower ? 'am' : 'AM';
        }
    }

    var baseConfig = {
        calendar: defaultCalendar,
        longDateFormat: defaultLongDateFormat,
        invalidDate: defaultInvalidDate,
        ordinal: defaultOrdinal,
        dayOfMonthOrdinalParse: defaultDayOfMonthOrdinalParse,
        relativeTime: defaultRelativeTime,

        months: defaultLocaleMonths,
        monthsShort: defaultLocaleMonthsShort,

        week: defaultLocaleWeek,

        weekdays: defaultLocaleWeekdays,
        weekdaysMin: defaultLocaleWeekdaysMin,
        weekdaysShort: defaultLocaleWeekdaysShort,

        meridiemParse: defaultLocaleMeridiemParse,
    };

    // internal storage for locale config files
    var locales = {},
        localeFamilies = {},
        globalLocale;

    function commonPrefix(arr1, arr2) {
        var i,
            minl = Math.min(arr1.length, arr2.length);
        for (i = 0; i < minl; i += 1) {
            if (arr1[i] !== arr2[i]) {
                return i;
            }
        }
        return minl;
    }

    function normalizeLocale(key) {
        return key ? key.toLowerCase().replace('_', '-') : key;
    }

    // pick the locale from the array
    // try ['en-au', 'en-gb'] as 'en-au', 'en-gb', 'en', as in move through the list trying each
    // substring from most specific to least, but move to the next array item if it's a more specific variant than the current root
    function chooseLocale(names) {
        var i = 0,
            j,
            next,
            locale,
            split;

        while (i < names.length) {
            split = normalizeLocale(names[i]).split('-');
            j = split.length;
            next = normalizeLocale(names[i + 1]);
            next = next ? next.split('-') : null;
            while (j > 0) {
                locale = loadLocale(split.slice(0, j).join('-'));
                if (locale) {
                    return locale;
                }
                if (
                    next &&
                    next.length >= j &&
                    commonPrefix(split, next) >= j - 1
                ) {
                    //the next array item is better than a shallower substring of this one
                    break;
                }
                j--;
            }
            i++;
        }
        return globalLocale;
    }

    function isLocaleNameSane(name) {
        // Prevent names that look like filesystem paths, i.e contain '/' or '\'
        return name.match('^[^/\\\\]*$') != null;
    }

    function loadLocale(name) {
        var oldLocale = null,
            aliasedRequire;
        // TODO: Find a better way to register and load all the locales in Node
        if (
            locales[name] === undefined &&
            typeof module !== 'undefined' &&
            module &&
            module.exports &&
            isLocaleNameSane(name)
        ) {
            try {
                oldLocale = globalLocale._abbr;
                aliasedRequire = require;
                aliasedRequire('./locale/' + name);
                getSetGlobalLocale(oldLocale);
            } catch (e) {
                // mark as not found to avoid repeating expensive file require call causing high CPU
                // when trying to find en-US, en_US, en-us for every format call
                locales[name] = null; // null means not found
            }
        }
        return locales[name];
    }

    // This function will load locale and then set the global locale.  If
    // no arguments are passed in, it will simply return the current global
    // locale key.
    function getSetGlobalLocale(key, values) {
        var data;
        if (key) {
            if (isUndefined(values)) {
                data = getLocale(key);
            } else {
                data = defineLocale(key, values);
            }

            if (data) {
                // moment.duration._locale = moment._locale = data;
                globalLocale = data;
            } else {
                if (typeof console !== 'undefined' && console.warn) {
                    //warn user if arguments are passed but the locale could not be set
                    console.warn(
                        'Locale ' + key + ' not found. Did you forget to load it?'
                    );
                }
            }
        }

        return globalLocale._abbr;
    }

    function defineLocale(name, config) {
        if (config !== null) {
            var locale,
                parentConfig = baseConfig;
            config.abbr = name;
            if (locales[name] != null) {
                deprecateSimple(
                    'defineLocaleOverride',
                    'use moment.updateLocale(localeName, config) to change ' +
                        'an existing locale. moment.defineLocale(localeName, ' +
                        'config) should only be used for creating a new locale ' +
                        'See http://momentjs.com/guides/#/warnings/define-locale/ for more info.'
                );
                parentConfig = locales[name]._config;
            } else if (config.parentLocale != null) {
                if (locales[config.parentLocale] != null) {
                    parentConfig = locales[config.parentLocale]._config;
                } else {
                    locale = loadLocale(config.parentLocale);
                    if (locale != null) {
                        parentConfig = locale._config;
                    } else {
                        if (!localeFamilies[config.parentLocale]) {
                            localeFamilies[config.parentLocale] = [];
                        }
                        localeFamilies[config.parentLocale].push({
                            name: name,
                            config: config,
                        });
                        return null;
                    }
                }
            }
            locales[name] = new Locale(mergeConfigs(parentConfig, config));

            if (localeFamilies[name]) {
                localeFamilies[name].forEach(function (x) {
                    defineLocale(x.name, x.config);
                });
            }

            // backwards compat for now: also set the locale
            // make sure we set the locale AFTER all child locales have been
            // created, so we won't end up with the child locale set.
            getSetGlobalLocale(name);

            return locales[name];
        } else {
            // useful for testing
            delete locales[name];
            return null;
        }
    }

    function updateLocale(name, config) {
        if (config != null) {
            var locale,
                tmpLocale,
                parentConfig = baseConfig;

            if (locales[name] != null && locales[name].parentLocale != null) {
                // Update existing child locale in-place to avoid memory-leaks
                locales[name].set(mergeConfigs(locales[name]._config, config));
            } else {
                // MERGE
                tmpLocale = loadLocale(name);
                if (tmpLocale != null) {
                    parentConfig = tmpLocale._config;
                }
                config = mergeConfigs(parentConfig, config);
                if (tmpLocale == null) {
                    // updateLocale is called for creating a new locale
                    // Set abbr so it will have a name (getters return
                    // undefined otherwise).
                    config.abbr = name;
                }
                locale = new Locale(config);
                locale.parentLocale = locales[name];
                locales[name] = locale;
            }

            // backwards compat for now: also set the locale
            getSetGlobalLocale(name);
        } else {
            // pass null for config to unupdate, useful for tests
            if (locales[name] != null) {
                if (locales[name].parentLocale != null) {
                    locales[name] = locales[name].parentLocale;
                    if (name === getSetGlobalLocale()) {
                        getSetGlobalLocale(name);
                    }
                } else if (locales[name] != null) {
                    delete locales[name];
                }
            }
        }
        return locales[name];
    }

    // returns locale data
    function getLocale(key) {
        var locale;

        if (key && key._locale && key._locale._abbr) {
            key = key._locale._abbr;
        }

        if (!key) {
            return globalLocale;
        }

        if (!isArray(key)) {
            //short-circuit everything else
            locale = loadLocale(key);
            if (locale) {
                return locale;
            }
            key = [key];
        }

        return chooseLocale(key);
    }

    function listLocales() {
        return keys(locales);
    }

    function checkOverflow(m) {
        var overflow,
            a = m._a;

        if (a && getParsingFlags(m).overflow === -2) {
            overflow =
                a[MONTH] < 0 || a[MONTH] > 11
                    ? MONTH
                    : a[DATE] < 1 || a[DATE] > daysInMonth(a[YEAR], a[MONTH])
                    ? DATE
                    : a[HOUR] < 0 ||
                      a[HOUR] > 24 ||
                      (a[HOUR] === 24 &&
                          (a[MINUTE] !== 0 ||
                              a[SECOND] !== 0 ||
                              a[MILLISECOND] !== 0))
                    ? HOUR
                    : a[MINUTE] < 0 || a[MINUTE] > 59
                    ? MINUTE
                    : a[SECOND] < 0 || a[SECOND] > 59
                    ? SECOND
                    : a[MILLISECOND] < 0 || a[MILLISECOND] > 999
                    ? MILLISECOND
                    : -1;

            if (
                getParsingFlags(m)._overflowDayOfYear &&
                (overflow < YEAR || overflow > DATE)
            ) {
                overflow = DATE;
            }
            if (getParsingFlags(m)._overflowWeeks && overflow === -1) {
                overflow = WEEK;
            }
            if (getParsingFlags(m)._overflowWeekday && overflow === -1) {
                overflow = WEEKDAY;
            }

            getParsingFlags(m).overflow = overflow;
        }

        return m;
    }

    // iso 8601 regex
    // 0000-00-00 0000-W00 or 0000-W00-0 + T + 00 or 00:00 or 00:00:00 or 00:00:00.000 + +00:00 or +0000 or +00)
    var extendedIsoRegex =
            /^\s*((?:[+-]\d{6}|\d{4})-(?:\d\d-\d\d|W\d\d-\d|W\d\d|\d\d\d|\d\d))(?:(T| )(\d\d(?::\d\d(?::\d\d(?:[.,]\d+)?)?)?)([+-]\d\d(?::?\d\d)?|\s*Z)?)?$/,
        basicIsoRegex =
            /^\s*((?:[+-]\d{6}|\d{4})(?:\d\d\d\d|W\d\d\d|W\d\d|\d\d\d|\d\d|))(?:(T| )(\d\d(?:\d\d(?:\d\d(?:[.,]\d+)?)?)?)([+-]\d\d(?::?\d\d)?|\s*Z)?)?$/,
        tzRegex = /Z|[+-]\d\d(?::?\d\d)?/,
        isoDates = [
            ['YYYYYY-MM-DD', /[+-]\d{6}-\d\d-\d\d/],
            ['YYYY-MM-DD', /\d{4}-\d\d-\d\d/],
            ['GGGG-[W]WW-E', /\d{4}-W\d\d-\d/],
            ['GGGG-[W]WW', /\d{4}-W\d\d/, false],
            ['YYYY-DDD', /\d{4}-\d{3}/],
            ['YYYY-MM', /\d{4}-\d\d/, false],
            ['YYYYYYMMDD', /[+-]\d{10}/],
            ['YYYYMMDD', /\d{8}/],
            ['GGGG[W]WWE', /\d{4}W\d{3}/],
            ['GGGG[W]WW', /\d{4}W\d{2}/, false],
            ['YYYYDDD', /\d{7}/],
            ['YYYYMM', /\d{6}/, false],
            ['YYYY', /\d{4}/, false],
        ],
        // iso time formats and regexes
        isoTimes = [
            ['HH:mm:ss.SSSS', /\d\d:\d\d:\d\d\.\d+/],
            ['HH:mm:ss,SSSS', /\d\d:\d\d:\d\d,\d+/],
            ['HH:mm:ss', /\d\d:\d\d:\d\d/],
            ['HH:mm', /\d\d:\d\d/],
            ['HHmmss.SSSS', /\d\d\d\d\d\d\.\d+/],
            ['HHmmss,SSSS', /\d\d\d\d\d\d,\d+/],
            ['HHmmss', /\d\d\d\d\d\d/],
            ['HHmm', /\d\d\d\d/],
            ['HH', /\d\d/],
        ],
        aspNetJsonRegex = /^\/?Date\((-?\d+)/i,
        // RFC 2822 regex: For details see https://tools.ietf.org/html/rfc2822#section-3.3
        rfc2822 =
            /^(?:(Mon|Tue|Wed|Thu|Fri|Sat|Sun),?\s)?(\d{1,2})\s(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s(\d{2,4})\s(\d\d):(\d\d)(?::(\d\d))?\s(?:(UT|GMT|[ECMP][SD]T)|([Zz])|([+-]\d{4}))$/,
        obsOffsets = {
            UT: 0,
            GMT: 0,
            EDT: -4 * 60,
            EST: -5 * 60,
            CDT: -5 * 60,
            CST: -6 * 60,
            MDT: -6 * 60,
            MST: -7 * 60,
            PDT: -7 * 60,
            PST: -8 * 60,
        };

    // date from iso format
    function configFromISO(config) {
        var i,
            l,
            string = config._i,
            match = extendedIsoRegex.exec(string) || basicIsoRegex.exec(string),
            allowTime,
            dateFormat,
            timeFormat,
            tzFormat,
            isoDatesLen = isoDates.length,
            isoTimesLen = isoTimes.length;

        if (match) {
            getParsingFlags(config).iso = true;
            for (i = 0, l = isoDatesLen; i < l; i++) {
                if (isoDates[i][1].exec(match[1])) {
                    dateFormat = isoDates[i][0];
                    allowTime = isoDates[i][2] !== false;
                    break;
                }
            }
            if (dateFormat == null) {
                config._isValid = false;
                return;
            }
            if (match[3]) {
                for (i = 0, l = isoTimesLen; i < l; i++) {
                    if (isoTimes[i][1].exec(match[3])) {
                        // match[2] should be 'T' or space
                        timeFormat = (match[2] || ' ') + isoTimes[i][0];
                        break;
                    }
                }
                if (timeFormat == null) {
                    config._isValid = false;
                    return;
                }
            }
            if (!allowTime && timeFormat != null) {
                config._isValid = false;
                return;
            }
            if (match[4]) {
                if (tzRegex.exec(match[4])) {
                    tzFormat = 'Z';
                } else {
                    config._isValid = false;
                    return;
                }
            }
            config._f = dateFormat + (timeFormat || '') + (tzFormat || '');
            configFromStringAndFormat(config);
        } else {
            config._isValid = false;
        }
    }

    function extractFromRFC2822Strings(
        yearStr,
        monthStr,
        dayStr,
        hourStr,
        minuteStr,
        secondStr
    ) {
        var result = [
            untruncateYear(yearStr),
            defaultLocaleMonthsShort.indexOf(monthStr),
            parseInt(dayStr, 10),
            parseInt(hourStr, 10),
            parseInt(minuteStr, 10),
        ];

        if (secondStr) {
            result.push(parseInt(secondStr, 10));
        }

        return result;
    }

    function untruncateYear(yearStr) {
        var year = parseInt(yearStr, 10);
        if (year <= 49) {
            return 2000 + year;
        } else if (year <= 999) {
            return 1900 + year;
        }
        return year;
    }

    function preprocessRFC2822(s) {
        // Remove comments and folding whitespace and replace multiple-spaces with a single space
        return s
            .replace(/\([^()]*\)|[\n\t]/g, ' ')
            .replace(/(\s\s+)/g, ' ')
            .replace(/^\s\s*/, '')
            .replace(/\s\s*$/, '');
    }

    function checkWeekday(weekdayStr, parsedInput, config) {
        if (weekdayStr) {
            // TODO: Replace the vanilla JS Date object with an independent day-of-week check.
            var weekdayProvided = defaultLocaleWeekdaysShort.indexOf(weekdayStr),
                weekdayActual = new Date(
                    parsedInput[0],
                    parsedInput[1],
                    parsedInput[2]
                ).getDay();
            if (weekdayProvided !== weekdayActual) {
                getParsingFlags(config).weekdayMismatch = true;
                config._isValid = false;
                return false;
            }
        }
        return true;
    }

    function calculateOffset(obsOffset, militaryOffset, numOffset) {
        if (obsOffset) {
            return obsOffsets[obsOffset];
        } else if (militaryOffset) {
            // the only allowed military tz is Z
            return 0;
        } else {
            var hm = parseInt(numOffset, 10),
                m = hm % 100,
                h = (hm - m) / 100;
            return h * 60 + m;
        }
    }

    // date and time from ref 2822 format
    function configFromRFC2822(config) {
        var match = rfc2822.exec(preprocessRFC2822(config._i)),
            parsedArray;
        if (match) {
            parsedArray = extractFromRFC2822Strings(
                match[4],
                match[3],
                match[2],
                match[5],
                match[6],
                match[7]
            );
            if (!checkWeekday(match[1], parsedArray, config)) {
                return;
            }

            config._a = parsedArray;
            config._tzm = calculateOffset(match[8], match[9], match[10]);

            config._d = createUTCDate.apply(null, config._a);
            config._d.setUTCMinutes(config._d.getUTCMinutes() - config._tzm);

            getParsingFlags(config).rfc2822 = true;
        } else {
            config._isValid = false;
        }
    }

    // date from 1) ASP.NET, 2) ISO, 3) RFC 2822 formats, or 4) optional fallback if parsing isn't strict
    function configFromString(config) {
        var matched = aspNetJsonRegex.exec(config._i);
        if (matched !== null) {
            config._d = new Date(+matched[1]);
            return;
        }

        configFromISO(config);
        if (config._isValid === false) {
            delete config._isValid;
        } else {
            return;
        }

        configFromRFC2822(config);
        if (config._isValid === false) {
            delete config._isValid;
        } else {
            return;
        }

        if (config._strict) {
            config._isValid = false;
        } else {
            // Final attempt, use Input Fallback
            hooks.createFromInputFallback(config);
        }
    }

    hooks.createFromInputFallback = deprecate(
        'value provided is not in a recognized RFC2822 or ISO format. moment construction falls back to js Date(), ' +
            'which is not reliable across all browsers and versions. Non RFC2822/ISO date formats are ' +
            'discouraged. Please refer to http://momentjs.com/guides/#/warnings/js-date/ for more info.',
        function (config) {
            config._d = new Date(config._i + (config._useUTC ? ' UTC' : ''));
        }
    );

    // Pick the first defined of two or three arguments.
    function defaults(a, b, c) {
        if (a != null) {
            return a;
        }
        if (b != null) {
            return b;
        }
        return c;
    }

    function currentDateArray(config) {
        // hooks is actually the exported moment object
        var nowValue = new Date(hooks.now());
        if (config._useUTC) {
            return [
                nowValue.getUTCFullYear(),
                nowValue.getUTCMonth(),
                nowValue.getUTCDate(),
            ];
        }
        return [nowValue.getFullYear(), nowValue.getMonth(), nowValue.getDate()];
    }

    // convert an array to a date.
    // the array should mirror the parameters below
    // note: all values past the year are optional and will default to the lowest possible value.
    // [year, month, day , hour, minute, second, millisecond]
    function configFromArray(config) {
        var i,
            date,
            input = [],
            currentDate,
            expectedWeekday,
            yearToUse;

        if (config._d) {
            return;
        }

        currentDate = currentDateArray(config);

        //compute day of the year from weeks and weekdays
        if (config._w && config._a[DATE] == null && config._a[MONTH] == null) {
            dayOfYearFromWeekInfo(config);
        }

        //if the day of the year is set, figure out what it is
        if (config._dayOfYear != null) {
            yearToUse = defaults(config._a[YEAR], currentDate[YEAR]);

            if (
                config._dayOfYear > daysInYear(yearToUse) ||
                config._dayOfYear === 0
            ) {
                getParsingFlags(config)._overflowDayOfYear = true;
            }

            date = createUTCDate(yearToUse, 0, config._dayOfYear);
            config._a[MONTH] = date.getUTCMonth();
            config._a[DATE] = date.getUTCDate();
        }

        // Default to current date.
        // * if no year, month, day of month are given, default to today
        // * if day of month is given, default month and year
        // * if month is given, default only year
        // * if year is given, don't default anything
        for (i = 0; i < 3 && config._a[i] == null; ++i) {
            config._a[i] = input[i] = currentDate[i];
        }

        // Zero out whatever was not defaulted, including time
        for (; i < 7; i++) {
            config._a[i] = input[i] =
                config._a[i] == null ? (i === 2 ? 1 : 0) : config._a[i];
        }

        // Check for 24:00:00.000
        if (
            config._a[HOUR] === 24 &&
            config._a[MINUTE] === 0 &&
            config._a[SECOND] === 0 &&
            config._a[MILLISECOND] === 0
        ) {
            config._nextDay = true;
            config._a[HOUR] = 0;
        }

        config._d = (config._useUTC ? createUTCDate : createDate).apply(
            null,
            input
        );
        expectedWeekday = config._useUTC
            ? config._d.getUTCDay()
            : config._d.getDay();

        // Apply timezone offset from input. The actual utcOffset can be changed
        // with parseZone.
        if (config._tzm != null) {
            config._d.setUTCMinutes(config._d.getUTCMinutes() - config._tzm);
        }

        if (config._nextDay) {
            config._a[HOUR] = 24;
        }

        // check for mismatching day of week
        if (
            config._w &&
            typeof config._w.d !== 'undefined' &&
            config._w.d !== expectedWeekday
        ) {
            getParsingFlags(config).weekdayMismatch = true;
        }
    }

    function dayOfYearFromWeekInfo(config) {
        var w, weekYear, week, weekday, dow, doy, temp, weekdayOverflow, curWeek;

        w = config._w;
        if (w.GG != null || w.W != null || w.E != null) {
            dow = 1;
            doy = 4;

            // TODO: We need to take the current isoWeekYear, but that depends on
            // how we interpret now (local, utc, fixed offset). So create
            // a now version of current config (take local/utc/offset flags, and
            // create now).
            weekYear = defaults(
                w.GG,
                config._a[YEAR],
                weekOfYear(createLocal(), 1, 4).year
            );
            week = defaults(w.W, 1);
            weekday = defaults(w.E, 1);
            if (weekday < 1 || weekday > 7) {
                weekdayOverflow = true;
            }
        } else {
            dow = config._locale._week.dow;
            doy = config._locale._week.doy;

            curWeek = weekOfYear(createLocal(), dow, doy);

            weekYear = defaults(w.gg, config._a[YEAR], curWeek.year);

            // Default to current week.
            week = defaults(w.w, curWeek.week);

            if (w.d != null) {
                // weekday -- low day numbers are considered next week
                weekday = w.d;
                if (weekday < 0 || weekday > 6) {
                    weekdayOverflow = true;
                }
            } else if (w.e != null) {
                // local weekday -- counting starts from beginning of week
                weekday = w.e + dow;
                if (w.e < 0 || w.e > 6) {
                    weekdayOverflow = true;
                }
            } else {
                // default to beginning of week
                weekday = dow;
            }
        }
        if (week < 1 || week > weeksInYear(weekYear, dow, doy)) {
            getParsingFlags(config)._overflowWeeks = true;
        } else if (weekdayOverflow != null) {
            getParsingFlags(config)._overflowWeekday = true;
        } else {
            temp = dayOfYearFromWeeks(weekYear, week, weekday, dow, doy);
            config._a[YEAR] = temp.year;
            config._dayOfYear = temp.dayOfYear;
        }
    }

    // constant that refers to the ISO standard
    hooks.ISO_8601 = function () {};

    // constant that refers to the RFC 2822 form
    hooks.RFC_2822 = function () {};

    // date from string and format string
    function configFromStringAndFormat(config) {
        // TODO: Move this to another part of the creation flow to prevent circular deps
        if (config._f === hooks.ISO_8601) {
            configFromISO(config);
            return;
        }
        if (config._f === hooks.RFC_2822) {
            configFromRFC2822(config);
            return;
        }
        config._a = [];
        getParsingFlags(config).empty = true;

        // This array is used to make a Date, either with `new Date` or `Date.UTC`
        var string = '' + config._i,
            i,
            parsedInput,
            tokens,
            token,
            skipped,
            stringLength = string.length,
            totalParsedInputLength = 0,
            era,
            tokenLen;

        tokens =
            expandFormat(config._f, config._locale).match(formattingTokens) || [];
        tokenLen = tokens.length;
        for (i = 0; i < tokenLen; i++) {
            token = tokens[i];
            parsedInput = (string.match(getParseRegexForToken(token, config)) ||
                [])[0];
            if (parsedInput) {
                skipped = string.substr(0, string.indexOf(parsedInput));
                if (skipped.length > 0) {
                    getParsingFlags(config).unusedInput.push(skipped);
                }
                string = string.slice(
                    string.indexOf(parsedInput) + parsedInput.length
                );
                totalParsedInputLength += parsedInput.length;
            }
            // don't parse if it's not a known token
            if (formatTokenFunctions[token]) {
                if (parsedInput) {
                    getParsingFlags(config).empty = false;
                } else {
                    getParsingFlags(config).unusedTokens.push(token);
                }
                addTimeToArrayFromToken(token, parsedInput, config);
            } else if (config._strict && !parsedInput) {
                getParsingFlags(config).unusedTokens.push(token);
            }
        }

        // add remaining unparsed input length to the string
        getParsingFlags(config).charsLeftOver =
            stringLength - totalParsedInputLength;
        if (string.length > 0) {
            getParsingFlags(config).unusedInput.push(string);
        }

        // clear _12h flag if hour is <= 12
        if (
            config._a[HOUR] <= 12 &&
            getParsingFlags(config).bigHour === true &&
            config._a[HOUR] > 0
        ) {
            getParsingFlags(config).bigHour = undefined;
        }

        getParsingFlags(config).parsedDateParts = config._a.slice(0);
        getParsingFlags(config).meridiem = config._meridiem;
        // handle meridiem
        config._a[HOUR] = meridiemFixWrap(
            config._locale,
            config._a[HOUR],
            config._meridiem
        );

        // handle era
        era = getParsingFlags(config).era;
        if (era !== null) {
            config._a[YEAR] = config._locale.erasConvertYear(era, config._a[YEAR]);
        }

        configFromArray(config);
        checkOverflow(config);
    }

    function meridiemFixWrap(locale, hour, meridiem) {
        var isPm;

        if (meridiem == null) {
            // nothing to do
            return hour;
        }
        if (locale.meridiemHour != null) {
            return locale.meridiemHour(hour, meridiem);
        } else if (locale.isPM != null) {
            // Fallback
            isPm = locale.isPM(meridiem);
            if (isPm && hour < 12) {
                hour += 12;
            }
            if (!isPm && hour === 12) {
                hour = 0;
            }
            return hour;
        } else {
            // this is not supposed to happen
            return hour;
        }
    }

    // date from string and array of format strings
    function configFromStringAndArray(config) {
        var tempConfig,
            bestMoment,
            scoreToBeat,
            i,
            currentScore,
            validFormatFound,
            bestFormatIsValid = false,
            configfLen = config._f.length;

        if (configfLen === 0) {
            getParsingFlags(config).invalidFormat = true;
            config._d = new Date(NaN);
            return;
        }

        for (i = 0; i < configfLen; i++) {
            currentScore = 0;
            validFormatFound = false;
            tempConfig = copyConfig({}, config);
            if (config._useUTC != null) {
                tempConfig._useUTC = config._useUTC;
            }
            tempConfig._f = config._f[i];
            configFromStringAndFormat(tempConfig);

            if (isValid(tempConfig)) {
                validFormatFound = true;
            }

            // if there is any input that was not parsed add a penalty for that format
            currentScore += getParsingFlags(tempConfig).charsLeftOver;

            //or tokens
            currentScore += getParsingFlags(tempConfig).unusedTokens.length * 10;

            getParsingFlags(tempConfig).score = currentScore;

            if (!bestFormatIsValid) {
                if (
                    scoreToBeat == null ||
                    currentScore < scoreToBeat ||
                    validFormatFound
                ) {
                    scoreToBeat = currentScore;
                    bestMoment = tempConfig;
                    if (validFormatFound) {
                        bestFormatIsValid = true;
                    }
                }
            } else {
                if (currentScore < scoreToBeat) {
                    scoreToBeat = currentScore;
                    bestMoment = tempConfig;
                }
            }
        }

        extend(config, bestMoment || tempConfig);
    }

    function configFromObject(config) {
        if (config._d) {
            return;
        }

        var i = normalizeObjectUnits(config._i),
            dayOrDate = i.day === undefined ? i.date : i.day;
        config._a = map(
            [i.year, i.month, dayOrDate, i.hour, i.minute, i.second, i.millisecond],
            function (obj) {
                return obj && parseInt(obj, 10);
            }
        );

        configFromArray(config);
    }

    function createFromConfig(config) {
        var res = new Moment(checkOverflow(prepareConfig(config)));
        if (res._nextDay) {
            // Adding is smart enough around DST
            res.add(1, 'd');
            res._nextDay = undefined;
        }

        return res;
    }

    function prepareConfig(config) {
        var input = config._i,
            format = config._f;

        config._locale = config._locale || getLocale(config._l);

        if (input === null || (format === undefined && input === '')) {
            return createInvalid({ nullInput: true });
        }

        if (typeof input === 'string') {
            config._i = input = config._locale.preparse(input);
        }

        if (isMoment(input)) {
            return new Moment(checkOverflow(input));
        } else if (isDate(input)) {
            config._d = input;
        } else if (isArray(format)) {
            configFromStringAndArray(config);
        } else if (format) {
            configFromStringAndFormat(config);
        } else {
            configFromInput(config);
        }

        if (!isValid(config)) {
            config._d = null;
        }

        return config;
    }

    function configFromInput(config) {
        var input = config._i;
        if (isUndefined(input)) {
            config._d = new Date(hooks.now());
        } else if (isDate(input)) {
            config._d = new Date(input.valueOf());
        } else if (typeof input === 'string') {
            configFromString(config);
        } else if (isArray(input)) {
            config._a = map(input.slice(0), function (obj) {
                return parseInt(obj, 10);
            });
            configFromArray(config);
        } else if (isObject(input)) {
            configFromObject(config);
        } else if (isNumber(input)) {
            // from milliseconds
            config._d = new Date(input);
        } else {
            hooks.createFromInputFallback(config);
        }
    }

    function createLocalOrUTC(input, format, locale, strict, isUTC) {
        var c = {};

        if (format === true || format === false) {
            strict = format;
            format = undefined;
        }

        if (locale === true || locale === false) {
            strict = locale;
            locale = undefined;
        }

        if (
            (isObject(input) && isObjectEmpty(input)) ||
            (isArray(input) && input.length === 0)
        ) {
            input = undefined;
        }
        // object construction must be done this way.
        // https://github.com/moment/moment/issues/1423
        c._isAMomentObject = true;
        c._useUTC = c._isUTC = isUTC;
        c._l = locale;
        c._i = input;
        c._f = format;
        c._strict = strict;

        return createFromConfig(c);
    }

    function createLocal(input, format, locale, strict) {
        return createLocalOrUTC(input, format, locale, strict, false);
    }

    var prototypeMin = deprecate(
            'moment().min is deprecated, use moment.max instead. http://momentjs.com/guides/#/warnings/min-max/',
            function () {
                var other = createLocal.apply(null, arguments);
                if (this.isValid() && other.isValid()) {
                    return other < this ? this : other;
                } else {
                    return createInvalid();
                }
            }
        ),
        prototypeMax = deprecate(
            'moment().max is deprecated, use moment.min instead. http://momentjs.com/guides/#/warnings/min-max/',
            function () {
                var other = createLocal.apply(null, arguments);
                if (this.isValid() && other.isValid()) {
                    return other > this ? this : other;
                } else {
                    return createInvalid();
                }
            }
        );

    // Pick a moment m from moments so that m[fn](other) is true for all
    // other. This relies on the function fn to be transitive.
    //
    // moments should either be an array of moment objects or an array, whose
    // first element is an array of moment objects.
    function pickBy(fn, moments) {
        var res, i;
        if (moments.length === 1 && isArray(moments[0])) {
            moments = moments[0];
        }
        if (!moments.length) {
            return createLocal();
        }
        res = moments[0];
        for (i = 1; i < moments.length; ++i) {
            if (!moments[i].isValid() || moments[i][fn](res)) {
                res = moments[i];
            }
        }
        return res;
    }

    // TODO: Use [].sort instead?
    function min() {
        var args = [].slice.call(arguments, 0);

        return pickBy('isBefore', args);
    }

    function max() {
        var args = [].slice.call(arguments, 0);

        return pickBy('isAfter', args);
    }

    var now = function () {
        return Date.now ? Date.now() : +new Date();
    };

    var ordering = [
        'year',
        'quarter',
        'month',
        'week',
        'day',
        'hour',
        'minute',
        'second',
        'millisecond',
    ];

    function isDurationValid(m) {
        var key,
            unitHasDecimal = false,
            i,
            orderLen = ordering.length;
        for (key in m) {
            if (
                hasOwnProp(m, key) &&
                !(
                    indexOf.call(ordering, key) !== -1 &&
                    (m[key] == null || !isNaN(m[key]))
                )
            ) {
                return false;
            }
        }

        for (i = 0; i < orderLen; ++i) {
            if (m[ordering[i]]) {
                if (unitHasDecimal) {
                    return false; // only allow non-integers for smallest unit
                }
                if (parseFloat(m[ordering[i]]) !== toInt(m[ordering[i]])) {
                    unitHasDecimal = true;
                }
            }
        }

        return true;
    }

    function isValid$1() {
        return this._isValid;
    }

    function createInvalid$1() {
        return createDuration(NaN);
    }

    function Duration(duration) {
        var normalizedInput = normalizeObjectUnits(duration),
            years = normalizedInput.year || 0,
            quarters = normalizedInput.quarter || 0,
            months = normalizedInput.month || 0,
            weeks = normalizedInput.week || normalizedInput.isoWeek || 0,
            days = normalizedInput.day || 0,
            hours = normalizedInput.hour || 0,
            minutes = normalizedInput.minute || 0,
            seconds = normalizedInput.second || 0,
            milliseconds = normalizedInput.millisecond || 0;

        this._isValid = isDurationValid(normalizedInput);

        // representation for dateAddRemove
        this._milliseconds =
            +milliseconds +
            seconds * 1e3 + // 1000
            minutes * 6e4 + // 1000 * 60
            hours * 1000 * 60 * 60; //using 1000 * 60 * 60 instead of 36e5 to avoid floating point rounding errors https://github.com/moment/moment/issues/2978
        // Because of dateAddRemove treats 24 hours as different from a
        // day when working around DST, we need to store them separately
        this._days = +days + weeks * 7;
        // It is impossible to translate months into days without knowing
        // which months you are are talking about, so we have to store
        // it separately.
        this._months = +months + quarters * 3 + years * 12;

        this._data = {};

        this._locale = getLocale();

        this._bubble();
    }

    function isDuration(obj) {
        return obj instanceof Duration;
    }

    function absRound(number) {
        if (number < 0) {
            return Math.round(-1 * number) * -1;
        } else {
            return Math.round(number);
        }
    }

    // compare two arrays, return the number of differences
    function compareArrays(array1, array2, dontConvert) {
        var len = Math.min(array1.length, array2.length),
            lengthDiff = Math.abs(array1.length - array2.length),
            diffs = 0,
            i;
        for (i = 0; i < len; i++) {
            if (
                (dontConvert && array1[i] !== array2[i]) ||
                (!dontConvert && toInt(array1[i]) !== toInt(array2[i]))
            ) {
                diffs++;
            }
        }
        return diffs + lengthDiff;
    }

    // FORMATTING

    function offset(token, separator) {
        addFormatToken(token, 0, 0, function () {
            var offset = this.utcOffset(),
                sign = '+';
            if (offset < 0) {
                offset = -offset;
                sign = '-';
            }
            return (
                sign +
                zeroFill(~~(offset / 60), 2) +
                separator +
                zeroFill(~~offset % 60, 2)
            );
        });
    }

    offset('Z', ':');
    offset('ZZ', '');

    // PARSING

    addRegexToken('Z', matchShortOffset);
    addRegexToken('ZZ', matchShortOffset);
    addParseToken(['Z', 'ZZ'], function (input, array, config) {
        config._useUTC = true;
        config._tzm = offsetFromString(matchShortOffset, input);
    });

    // HELPERS

    // timezone chunker
    // '+10:00' > ['10',  '00']
    // '-1530'  > ['-15', '30']
    var chunkOffset = /([\+\-]|\d\d)/gi;

    function offsetFromString(matcher, string) {
        var matches = (string || '').match(matcher),
            chunk,
            parts,
            minutes;

        if (matches === null) {
            return null;
        }

        chunk = matches[matches.length - 1] || [];
        parts = (chunk + '').match(chunkOffset) || ['-', 0, 0];
        minutes = +(parts[1] * 60) + toInt(parts[2]);

        return minutes === 0 ? 0 : parts[0] === '+' ? minutes : -minutes;
    }

    // Return a moment from input, that is local/utc/zone equivalent to model.
    function cloneWithOffset(input, model) {
        var res, diff;
        if (model._isUTC) {
            res = model.clone();
            diff =
                (isMoment(input) || isDate(input)
                    ? input.valueOf()
                    : createLocal(input).valueOf()) - res.valueOf();
            // Use low-level api, because this fn is low-level api.
            res._d.setTime(res._d.valueOf() + diff);
            hooks.updateOffset(res, false);
            return res;
        } else {
            return createLocal(input).local();
        }
    }

    function getDateOffset(m) {
        // On Firefox.24 Date#getTimezoneOffset returns a floating point.
        // https://github.com/moment/moment/pull/1871
        return -Math.round(m._d.getTimezoneOffset());
    }

    // HOOKS

    // This function will be called whenever a moment is mutated.
    // It is intended to keep the offset in sync with the timezone.
    hooks.updateOffset = function () {};

    // MOMENTS

    // keepLocalTime = true means only change the timezone, without
    // affecting the local hour. So 5:31:26 +0300 --[utcOffset(2, true)]-->
    // 5:31:26 +0200 It is possible that 5:31:26 doesn't exist with offset
    // +0200, so we adjust the time as needed, to be valid.
    //
    // Keeping the time actually adds/subtracts (one hour)
    // from the actual represented time. That is why we call updateOffset
    // a second time. In case it wants us to change the offset again
    // _changeInProgress == true case, then we have to adjust, because
    // there is no such time in the given timezone.
    function getSetOffset(input, keepLocalTime, keepMinutes) {
        var offset = this._offset || 0,
            localAdjust;
        if (!this.isValid()) {
            return input != null ? this : NaN;
        }
        if (input != null) {
            if (typeof input === 'string') {
                input = offsetFromString(matchShortOffset, input);
                if (input === null) {
                    return this;
                }
            } else if (Math.abs(input) < 16 && !keepMinutes) {
                input = input * 60;
            }
            if (!this._isUTC && keepLocalTime) {
                localAdjust = getDateOffset(this);
            }
            this._offset = input;
            this._isUTC = true;
            if (localAdjust != null) {
                this.add(localAdjust, 'm');
            }
            if (offset !== input) {
                if (!keepLocalTime || this._changeInProgress) {
                    addSubtract(
                        this,
                        createDuration(input - offset, 'm'),
                        1,
                        false
                    );
                } else if (!this._changeInProgress) {
                    this._changeInProgress = true;
                    hooks.updateOffset(this, true);
                    this._changeInProgress = null;
                }
            }
            return this;
        } else {
            return this._isUTC ? offset : getDateOffset(this);
        }
    }

    function getSetZone(input, keepLocalTime) {
        if (input != null) {
            if (typeof input !== 'string') {
                input = -input;
            }

            this.utcOffset(input, keepLocalTime);

            return this;
        } else {
            return -this.utcOffset();
        }
    }

    function setOffsetToUTC(keepLocalTime) {
        return this.utcOffset(0, keepLocalTime);
    }

    function setOffsetToLocal(keepLocalTime) {
        if (this._isUTC) {
            this.utcOffset(0, keepLocalTime);
            this._isUTC = false;

            if (keepLocalTime) {
                this.subtract(getDateOffset(this), 'm');
            }
        }
        return this;
    }

    function setOffsetToParsedOffset() {
        if (this._tzm != null) {
            this.utcOffset(this._tzm, false, true);
        } else if (typeof this._i === 'string') {
            var tZone = offsetFromString(matchOffset, this._i);
            if (tZone != null) {
                this.utcOffset(tZone);
            } else {
                this.utcOffset(0, true);
            }
        }
        return this;
    }

    function hasAlignedHourOffset(input) {
        if (!this.isValid()) {
            return false;
        }
        input = input ? createLocal(input).utcOffset() : 0;

        return (this.utcOffset() - input) % 60 === 0;
    }

    function isDaylightSavingTime() {
        return (
            this.utcOffset() > this.clone().month(0).utcOffset() ||
            this.utcOffset() > this.clone().month(5).utcOffset()
        );
    }

    function isDaylightSavingTimeShifted() {
        if (!isUndefined(this._isDSTShifted)) {
            return this._isDSTShifted;
        }

        var c = {},
            other;

        copyConfig(c, this);
        c = prepareConfig(c);

        if (c._a) {
            other = c._isUTC ? createUTC(c._a) : createLocal(c._a);
            this._isDSTShifted =
                this.isValid() && compareArrays(c._a, other.toArray()) > 0;
        } else {
            this._isDSTShifted = false;
        }

        return this._isDSTShifted;
    }

    function isLocal() {
        return this.isValid() ? !this._isUTC : false;
    }

    function isUtcOffset() {
        return this.isValid() ? this._isUTC : false;
    }

    function isUtc() {
        return this.isValid() ? this._isUTC && this._offset === 0 : false;
    }

    // ASP.NET json date format regex
    var aspNetRegex = /^(-|\+)?(?:(\d*)[. ])?(\d+):(\d+)(?::(\d+)(\.\d*)?)?$/,
        // from http://docs.closure-library.googlecode.com/git/closure_goog_date_date.js.source.html
        // somewhat more in line with 4.4.3.2 2004 spec, but allows decimal anywhere
        // and further modified to allow for strings containing both week and day
        isoRegex =
            /^(-|\+)?P(?:([-+]?[0-9,.]*)Y)?(?:([-+]?[0-9,.]*)M)?(?:([-+]?[0-9,.]*)W)?(?:([-+]?[0-9,.]*)D)?(?:T(?:([-+]?[0-9,.]*)H)?(?:([-+]?[0-9,.]*)M)?(?:([-+]?[0-9,.]*)S)?)?$/;

    function createDuration(input, key) {
        var duration = input,
            // matching against regexp is expensive, do it on demand
            match = null,
            sign,
            ret,
            diffRes;

        if (isDuration(input)) {
            duration = {
                ms: input._milliseconds,
                d: input._days,
                M: input._months,
            };
        } else if (isNumber(input) || !isNaN(+input)) {
            duration = {};
            if (key) {
                duration[key] = +input;
            } else {
                duration.milliseconds = +input;
            }
        } else if ((match = aspNetRegex.exec(input))) {
            sign = match[1] === '-' ? -1 : 1;
            duration = {
                y: 0,
                d: toInt(match[DATE]) * sign,
                h: toInt(match[HOUR]) * sign,
                m: toInt(match[MINUTE]) * sign,
                s: toInt(match[SECOND]) * sign,
                ms: toInt(absRound(match[MILLISECOND] * 1000)) * sign, // the millisecond decimal point is included in the match
            };
        } else if ((match = isoRegex.exec(input))) {
            sign = match[1] === '-' ? -1 : 1;
            duration = {
                y: parseIso(match[2], sign),
                M: parseIso(match[3], sign),
                w: parseIso(match[4], sign),
                d: parseIso(match[5], sign),
                h: parseIso(match[6], sign),
                m: parseIso(match[7], sign),
                s: parseIso(match[8], sign),
            };
        } else if (duration == null) {
            // checks for null or undefined
            duration = {};
        } else if (
            typeof duration === 'object' &&
            ('from' in duration || 'to' in duration)
        ) {
            diffRes = momentsDifference(
                createLocal(duration.from),
                createLocal(duration.to)
            );

            duration = {};
            duration.ms = diffRes.milliseconds;
            duration.M = diffRes.months;
        }

        ret = new Duration(duration);

        if (isDuration(input) && hasOwnProp(input, '_locale')) {
            ret._locale = input._locale;
        }

        if (isDuration(input) && hasOwnProp(input, '_isValid')) {
            ret._isValid = input._isValid;
        }

        return ret;
    }

    createDuration.fn = Duration.prototype;
    createDuration.invalid = createInvalid$1;

    function parseIso(inp, sign) {
        // We'd normally use ~~inp for this, but unfortunately it also
        // converts floats to ints.
        // inp may be undefined, so careful calling replace on it.
        var res = inp && parseFloat(inp.replace(',', '.'));
        // apply sign while we're at it
        return (isNaN(res) ? 0 : res) * sign;
    }

    function positiveMomentsDifference(base, other) {
        var res = {};

        res.months =
            other.month() - base.month() + (other.year() - base.year()) * 12;
        if (base.clone().add(res.months, 'M').isAfter(other)) {
            --res.months;
        }

        res.milliseconds = +other - +base.clone().add(res.months, 'M');

        return res;
    }

    function momentsDifference(base, other) {
        var res;
        if (!(base.isValid() && other.isValid())) {
            return { milliseconds: 0, months: 0 };
        }

        other = cloneWithOffset(other, base);
        if (base.isBefore(other)) {
            res = positiveMomentsDifference(base, other);
        } else {
            res = positiveMomentsDifference(other, base);
            res.milliseconds = -res.milliseconds;
            res.months = -res.months;
        }

        return res;
    }

    // TODO: remove 'name' arg after deprecation is removed
    function createAdder(direction, name) {
        return function (val, period) {
            var dur, tmp;
            //invert the arguments, but complain about it
            if (period !== null && !isNaN(+period)) {
                deprecateSimple(
                    name,
                    'moment().' +
                        name +
                        '(period, number) is deprecated. Please use moment().' +
                        name +
                        '(number, period). ' +
                        'See http://momentjs.com/guides/#/warnings/add-inverted-param/ for more info.'
                );
                tmp = val;
                val = period;
                period = tmp;
            }

            dur = createDuration(val, period);
            addSubtract(this, dur, direction);
            return this;
        };
    }

    function addSubtract(mom, duration, isAdding, updateOffset) {
        var milliseconds = duration._milliseconds,
            days = absRound(duration._days),
            months = absRound(duration._months);

        if (!mom.isValid()) {
            // No op
            return;
        }

        updateOffset = updateOffset == null ? true : updateOffset;

        if (months) {
            setMonth(mom, get(mom, 'Month') + months * isAdding);
        }
        if (days) {
            set$1(mom, 'Date', get(mom, 'Date') + days * isAdding);
        }
        if (milliseconds) {
            mom._d.setTime(mom._d.valueOf() + milliseconds * isAdding);
        }
        if (updateOffset) {
            hooks.updateOffset(mom, days || months);
        }
    }

    var add = createAdder(1, 'add'),
        subtract = createAdder(-1, 'subtract');

    function isString(input) {
        return typeof input === 'string' || input instanceof String;
    }

    // type MomentInput = Moment | Date | string | number | (number | string)[] | MomentInputObject | void; // null | undefined
    function isMomentInput(input) {
        return (
            isMoment(input) ||
            isDate(input) ||
            isString(input) ||
            isNumber(input) ||
            isNumberOrStringArray(input) ||
            isMomentInputObject(input) ||
            input === null ||
            input === undefined
        );
    }

    function isMomentInputObject(input) {
        var objectTest = isObject(input) && !isObjectEmpty(input),
            propertyTest = false,
            properties = [
                'years',
                'year',
                'y',
                'months',
                'month',
                'M',
                'days',
                'day',
                'd',
                'dates',
                'date',
                'D',
                'hours',
                'hour',
                'h',
                'minutes',
                'minute',
                'm',
                'seconds',
                'second',
                's',
                'milliseconds',
                'millisecond',
                'ms',
            ],
            i,
            property,
            propertyLen = properties.length;

        for (i = 0; i < propertyLen; i += 1) {
            property = properties[i];
            propertyTest = propertyTest || hasOwnProp(input, property);
        }

        return objectTest && propertyTest;
    }

    function isNumberOrStringArray(input) {
        var arrayTest = isArray(input),
            dataTypeTest = false;
        if (arrayTest) {
            dataTypeTest =
                input.filter(function (item) {
                    return !isNumber(item) && isString(input);
                }).length === 0;
        }
        return arrayTest && dataTypeTest;
    }

    function isCalendarSpec(input) {
        var objectTest = isObject(input) && !isObjectEmpty(input),
            propertyTest = false,
            properties = [
                'sameDay',
                'nextDay',
                'lastDay',
                'nextWeek',
                'lastWeek',
                'sameElse',
            ],
            i,
            property;

        for (i = 0; i < properties.length; i += 1) {
            property = properties[i];
            propertyTest = propertyTest || hasOwnProp(input, property);
        }

        return objectTest && propertyTest;
    }

    function getCalendarFormat(myMoment, now) {
        var diff = myMoment.diff(now, 'days', true);
        return diff < -6
            ? 'sameElse'
            : diff < -1
            ? 'lastWeek'
            : diff < 0
            ? 'lastDay'
            : diff < 1
            ? 'sameDay'
            : diff < 2
            ? 'nextDay'
            : diff < 7
            ? 'nextWeek'
            : 'sameElse';
    }

    function calendar$1(time, formats) {
        // Support for single parameter, formats only overload to the calendar function
        if (arguments.length === 1) {
            if (!arguments[0]) {
                time = undefined;
                formats = undefined;
            } else if (isMomentInput(arguments[0])) {
                time = arguments[0];
                formats = undefined;
            } else if (isCalendarSpec(arguments[0])) {
                formats = arguments[0];
                time = undefined;
            }
        }
        // We want to compare the start of today, vs this.
        // Getting start-of-today depends on whether we're local/utc/offset or not.
        var now = time || createLocal(),
            sod = cloneWithOffset(now, this).startOf('day'),
            format = hooks.calendarFormat(this, sod) || 'sameElse',
            output =
                formats &&
                (isFunction(formats[format])
                    ? formats[format].call(this, now)
                    : formats[format]);

        return this.format(
            output || this.localeData().calendar(format, this, createLocal(now))
        );
    }

    function clone() {
        return new Moment(this);
    }

    function isAfter(input, units) {
        var localInput = isMoment(input) ? input : createLocal(input);
        if (!(this.isValid() && localInput.isValid())) {
            return false;
        }
        units = normalizeUnits(units) || 'millisecond';
        if (units === 'millisecond') {
            return this.valueOf() > localInput.valueOf();
        } else {
            return localInput.valueOf() < this.clone().startOf(units).valueOf();
        }
    }

    function isBefore(input, units) {
        var localInput = isMoment(input) ? input : createLocal(input);
        if (!(this.isValid() && localInput.isValid())) {
            return false;
        }
        units = normalizeUnits(units) || 'millisecond';
        if (units === 'millisecond') {
            return this.valueOf() < localInput.valueOf();
        } else {
            return this.clone().endOf(units).valueOf() < localInput.valueOf();
        }
    }

    function isBetween(from, to, units, inclusivity) {
        var localFrom = isMoment(from) ? from : createLocal(from),
            localTo = isMoment(to) ? to : createLocal(to);
        if (!(this.isValid() && localFrom.isValid() && localTo.isValid())) {
            return false;
        }
        inclusivity = inclusivity || '()';
        return (
            (inclusivity[0] === '('
                ? this.isAfter(localFrom, units)
                : !this.isBefore(localFrom, units)) &&
            (inclusivity[1] === ')'
                ? this.isBefore(localTo, units)
                : !this.isAfter(localTo, units))
        );
    }

    function isSame(input, units) {
        var localInput = isMoment(input) ? input : createLocal(input),
            inputMs;
        if (!(this.isValid() && localInput.isValid())) {
            return false;
        }
        units = normalizeUnits(units) || 'millisecond';
        if (units === 'millisecond') {
            return this.valueOf() === localInput.valueOf();
        } else {
            inputMs = localInput.valueOf();
            return (
                this.clone().startOf(units).valueOf() <= inputMs &&
                inputMs <= this.clone().endOf(units).valueOf()
            );
        }
    }

    function isSameOrAfter(input, units) {
        return this.isSame(input, units) || this.isAfter(input, units);
    }

    function isSameOrBefore(input, units) {
        return this.isSame(input, units) || this.isBefore(input, units);
    }

    function diff(input, units, asFloat) {
        var that, zoneDelta, output;

        if (!this.isValid()) {
            return NaN;
        }

        that = cloneWithOffset(input, this);

        if (!that.isValid()) {
            return NaN;
        }

        zoneDelta = (that.utcOffset() - this.utcOffset()) * 6e4;

        units = normalizeUnits(units);

        switch (units) {
            case 'year':
                output = monthDiff(this, that) / 12;
                break;
            case 'month':
                output = monthDiff(this, that);
                break;
            case 'quarter':
                output = monthDiff(this, that) / 3;
                break;
            case 'second':
                output = (this - that) / 1e3;
                break; // 1000
            case 'minute':
                output = (this - that) / 6e4;
                break; // 1000 * 60
            case 'hour':
                output = (this - that) / 36e5;
                break; // 1000 * 60 * 60
            case 'day':
                output = (this - that - zoneDelta) / 864e5;
                break; // 1000 * 60 * 60 * 24, negate dst
            case 'week':
                output = (this - that - zoneDelta) / 6048e5;
                break; // 1000 * 60 * 60 * 24 * 7, negate dst
            default:
                output = this - that;
        }

        return asFloat ? output : absFloor(output);
    }

    function monthDiff(a, b) {
        if (a.date() < b.date()) {
            // end-of-month calculations work correct when the start month has more
            // days than the end month.
            return -monthDiff(b, a);
        }
        // difference in months
        var wholeMonthDiff = (b.year() - a.year()) * 12 + (b.month() - a.month()),
            // b is in (anchor - 1 month, anchor + 1 month)
            anchor = a.clone().add(wholeMonthDiff, 'months'),
            anchor2,
            adjust;

        if (b - anchor < 0) {
            anchor2 = a.clone().add(wholeMonthDiff - 1, 'months');
            // linear across the month
            adjust = (b - anchor) / (anchor - anchor2);
        } else {
            anchor2 = a.clone().add(wholeMonthDiff + 1, 'months');
            // linear across the month
            adjust = (b - anchor) / (anchor2 - anchor);
        }

        //check for negative zero, return zero if negative zero
        return -(wholeMonthDiff + adjust) || 0;
    }

    hooks.defaultFormat = 'YYYY-MM-DDTHH:mm:ssZ';
    hooks.defaultFormatUtc = 'YYYY-MM-DDTHH:mm:ss[Z]';

    function toString() {
        return this.clone().locale('en').format('ddd MMM DD YYYY HH:mm:ss [GMT]ZZ');
    }

    function toISOString(keepOffset) {
        if (!this.isValid()) {
            return null;
        }
        var utc = keepOffset !== true,
            m = utc ? this.clone().utc() : this;
        if (m.year() < 0 || m.year() > 9999) {
            return formatMoment(
                m,
                utc
                    ? 'YYYYYY-MM-DD[T]HH:mm:ss.SSS[Z]'
                    : 'YYYYYY-MM-DD[T]HH:mm:ss.SSSZ'
            );
        }
        if (isFunction(Date.prototype.toISOString)) {
            // native implementation is ~50x faster, use it when we can
            if (utc) {
                return this.toDate().toISOString();
            } else {
                return new Date(this.valueOf() + this.utcOffset() * 60 * 1000)
                    .toISOString()
                    .replace('Z', formatMoment(m, 'Z'));
            }
        }
        return formatMoment(
            m,
            utc ? 'YYYY-MM-DD[T]HH:mm:ss.SSS[Z]' : 'YYYY-MM-DD[T]HH:mm:ss.SSSZ'
        );
    }

    /**
     * Return a human readable representation of a moment that can
     * also be evaluated to get a new moment which is the same
     *
     * @link https://nodejs.org/dist/latest/docs/api/util.html#util_custom_inspect_function_on_objects
     */
    function inspect() {
        if (!this.isValid()) {
            return 'moment.invalid(/* ' + this._i + ' */)';
        }
        var func = 'moment',
            zone = '',
            prefix,
            year,
            datetime,
            suffix;
        if (!this.isLocal()) {
            func = this.utcOffset() === 0 ? 'moment.utc' : 'moment.parseZone';
            zone = 'Z';
        }
        prefix = '[' + func + '("]';
        year = 0 <= this.year() && this.year() <= 9999 ? 'YYYY' : 'YYYYYY';
        datetime = '-MM-DD[T]HH:mm:ss.SSS';
        suffix = zone + '[")]';

        return this.format(prefix + year + datetime + suffix);
    }

    function format(inputString) {
        if (!inputString) {
            inputString = this.isUtc()
                ? hooks.defaultFormatUtc
                : hooks.defaultFormat;
        }
        var output = formatMoment(this, inputString);
        return this.localeData().postformat(output);
    }

    function from(time, withoutSuffix) {
        if (
            this.isValid() &&
            ((isMoment(time) && time.isValid()) || createLocal(time).isValid())
        ) {
            return createDuration({ to: this, from: time })
                .locale(this.locale())
                .humanize(!withoutSuffix);
        } else {
            return this.localeData().invalidDate();
        }
    }

    function fromNow(withoutSuffix) {
        return this.from(createLocal(), withoutSuffix);
    }

    function to(time, withoutSuffix) {
        if (
            this.isValid() &&
            ((isMoment(time) && time.isValid()) || createLocal(time).isValid())
        ) {
            return createDuration({ from: this, to: time })
                .locale(this.locale())
                .humanize(!withoutSuffix);
        } else {
            return this.localeData().invalidDate();
        }
    }

    function toNow(withoutSuffix) {
        return this.to(createLocal(), withoutSuffix);
    }

    // If passed a locale key, it will set the locale for this
    // instance.  Otherwise, it will return the locale configuration
    // variables for this instance.
    function locale(key) {
        var newLocaleData;

        if (key === undefined) {
            return this._locale._abbr;
        } else {
            newLocaleData = getLocale(key);
            if (newLocaleData != null) {
                this._locale = newLocaleData;
            }
            return this;
        }
    }

    var lang = deprecate(
        'moment().lang() is deprecated. Instead, use moment().localeData() to get the language configuration. Use moment().locale() to change languages.',
        function (key) {
            if (key === undefined) {
                return this.localeData();
            } else {
                return this.locale(key);
            }
        }
    );

    function localeData() {
        return this._locale;
    }

    var MS_PER_SECOND = 1000,
        MS_PER_MINUTE = 60 * MS_PER_SECOND,
        MS_PER_HOUR = 60 * MS_PER_MINUTE,
        MS_PER_400_YEARS = (365 * 400 + 97) * 24 * MS_PER_HOUR;

    // actual modulo - handles negative numbers (for dates before 1970):
    function mod$1(dividend, divisor) {
        return ((dividend % divisor) + divisor) % divisor;
    }

    function localStartOfDate(y, m, d) {
        // the date constructor remaps years 0-99 to 1900-1999
        if (y < 100 && y >= 0) {
            // preserve leap years using a full 400 year cycle, then reset
            return new Date(y + 400, m, d) - MS_PER_400_YEARS;
        } else {
            return new Date(y, m, d).valueOf();
        }
    }

    function utcStartOfDate(y, m, d) {
        // Date.UTC remaps years 0-99 to 1900-1999
        if (y < 100 && y >= 0) {
            // preserve leap years using a full 400 year cycle, then reset
            return Date.UTC(y + 400, m, d) - MS_PER_400_YEARS;
        } else {
            return Date.UTC(y, m, d);
        }
    }

    function startOf(units) {
        var time, startOfDate;
        units = normalizeUnits(units);
        if (units === undefined || units === 'millisecond' || !this.isValid()) {
            return this;
        }

        startOfDate = this._isUTC ? utcStartOfDate : localStartOfDate;

        switch (units) {
            case 'year':
                time = startOfDate(this.year(), 0, 1);
                break;
            case 'quarter':
                time = startOfDate(
                    this.year(),
                    this.month() - (this.month() % 3),
                    1
                );
                break;
            case 'month':
                time = startOfDate(this.year(), this.month(), 1);
                break;
            case 'week':
                time = startOfDate(
                    this.year(),
                    this.month(),
                    this.date() - this.weekday()
                );
                break;
            case 'isoWeek':
                time = startOfDate(
                    this.year(),
                    this.month(),
                    this.date() - (this.isoWeekday() - 1)
                );
                break;
            case 'day':
            case 'date':
                time = startOfDate(this.year(), this.month(), this.date());
                break;
            case 'hour':
                time = this._d.valueOf();
                time -= mod$1(
                    time + (this._isUTC ? 0 : this.utcOffset() * MS_PER_MINUTE),
                    MS_PER_HOUR
                );
                break;
            case 'minute':
                time = this._d.valueOf();
                time -= mod$1(time, MS_PER_MINUTE);
                break;
            case 'second':
                time = this._d.valueOf();
                time -= mod$1(time, MS_PER_SECOND);
                break;
        }

        this._d.setTime(time);
        hooks.updateOffset(this, true);
        return this;
    }

    function endOf(units) {
        var time, startOfDate;
        units = normalizeUnits(units);
        if (units === undefined || units === 'millisecond' || !this.isValid()) {
            return this;
        }

        startOfDate = this._isUTC ? utcStartOfDate : localStartOfDate;

        switch (units) {
            case 'year':
                time = startOfDate(this.year() + 1, 0, 1) - 1;
                break;
            case 'quarter':
                time =
                    startOfDate(
                        this.year(),
                        this.month() - (this.month() % 3) + 3,
                        1
                    ) - 1;
                break;
            case 'month':
                time = startOfDate(this.year(), this.month() + 1, 1) - 1;
                break;
            case 'week':
                time =
                    startOfDate(
                        this.year(),
                        this.month(),
                        this.date() - this.weekday() + 7
                    ) - 1;
                break;
            case 'isoWeek':
                time =
                    startOfDate(
                        this.year(),
                        this.month(),
                        this.date() - (this.isoWeekday() - 1) + 7
                    ) - 1;
                break;
            case 'day':
            case 'date':
                time = startOfDate(this.year(), this.month(), this.date() + 1) - 1;
                break;
            case 'hour':
                time = this._d.valueOf();
                time +=
                    MS_PER_HOUR -
                    mod$1(
                        time + (this._isUTC ? 0 : this.utcOffset() * MS_PER_MINUTE),
                        MS_PER_HOUR
                    ) -
                    1;
                break;
            case 'minute':
                time = this._d.valueOf();
                time += MS_PER_MINUTE - mod$1(time, MS_PER_MINUTE) - 1;
                break;
            case 'second':
                time = this._d.valueOf();
                time += MS_PER_SECOND - mod$1(time, MS_PER_SECOND) - 1;
                break;
        }

        this._d.setTime(time);
        hooks.updateOffset(this, true);
        return this;
    }

    function valueOf() {
        return this._d.valueOf() - (this._offset || 0) * 60000;
    }

    function unix() {
        return Math.floor(this.valueOf() / 1000);
    }

    function toDate() {
        return new Date(this.valueOf());
    }

    function toArray() {
        var m = this;
        return [
            m.year(),
            m.month(),
            m.date(),
            m.hour(),
            m.minute(),
            m.second(),
            m.millisecond(),
        ];
    }

    function toObject() {
        var m = this;
        return {
            years: m.year(),
            months: m.month(),
            date: m.date(),
            hours: m.hours(),
            minutes: m.minutes(),
            seconds: m.seconds(),
            milliseconds: m.milliseconds(),
        };
    }

    function toJSON() {
        // new Date(NaN).toJSON() === null
        return this.isValid() ? this.toISOString() : null;
    }

    function isValid$2() {
        return isValid(this);
    }

    function parsingFlags() {
        return extend({}, getParsingFlags(this));
    }

    function invalidAt() {
        return getParsingFlags(this).overflow;
    }

    function creationData() {
        return {
            input: this._i,
            format: this._f,
            locale: this._locale,
            isUTC: this._isUTC,
            strict: this._strict,
        };
    }

    addFormatToken('N', 0, 0, 'eraAbbr');
    addFormatToken('NN', 0, 0, 'eraAbbr');
    addFormatToken('NNN', 0, 0, 'eraAbbr');
    addFormatToken('NNNN', 0, 0, 'eraName');
    addFormatToken('NNNNN', 0, 0, 'eraNarrow');

    addFormatToken('y', ['y', 1], 'yo', 'eraYear');
    addFormatToken('y', ['yy', 2], 0, 'eraYear');
    addFormatToken('y', ['yyy', 3], 0, 'eraYear');
    addFormatToken('y', ['yyyy', 4], 0, 'eraYear');

    addRegexToken('N', matchEraAbbr);
    addRegexToken('NN', matchEraAbbr);
    addRegexToken('NNN', matchEraAbbr);
    addRegexToken('NNNN', matchEraName);
    addRegexToken('NNNNN', matchEraNarrow);

    addParseToken(
        ['N', 'NN', 'NNN', 'NNNN', 'NNNNN'],
        function (input, array, config, token) {
            var era = config._locale.erasParse(input, token, config._strict);
            if (era) {
                getParsingFlags(config).era = era;
            } else {
                getParsingFlags(config).invalidEra = input;
            }
        }
    );

    addRegexToken('y', matchUnsigned);
    addRegexToken('yy', matchUnsigned);
    addRegexToken('yyy', matchUnsigned);
    addRegexToken('yyyy', matchUnsigned);
    addRegexToken('yo', matchEraYearOrdinal);

    addParseToken(['y', 'yy', 'yyy', 'yyyy'], YEAR);
    addParseToken(['yo'], function (input, array, config, token) {
        var match;
        if (config._locale._eraYearOrdinalRegex) {
            match = input.match(config._locale._eraYearOrdinalRegex);
        }

        if (config._locale.eraYearOrdinalParse) {
            array[YEAR] = config._locale.eraYearOrdinalParse(input, match);
        } else {
            array[YEAR] = parseInt(input, 10);
        }
    });

    function localeEras(m, format) {
        var i,
            l,
            date,
            eras = this._eras || getLocale('en')._eras;
        for (i = 0, l = eras.length; i < l; ++i) {
            switch (typeof eras[i].since) {
                case 'string':
                    // truncate time
                    date = hooks(eras[i].since).startOf('day');
                    eras[i].since = date.valueOf();
                    break;
            }

            switch (typeof eras[i].until) {
                case 'undefined':
                    eras[i].until = +Infinity;
                    break;
                case 'string':
                    // truncate time
                    date = hooks(eras[i].until).startOf('day').valueOf();
                    eras[i].until = date.valueOf();
                    break;
            }
        }
        return eras;
    }

    function localeErasParse(eraName, format, strict) {
        var i,
            l,
            eras = this.eras(),
            name,
            abbr,
            narrow;
        eraName = eraName.toUpperCase();

        for (i = 0, l = eras.length; i < l; ++i) {
            name = eras[i].name.toUpperCase();
            abbr = eras[i].abbr.toUpperCase();
            narrow = eras[i].narrow.toUpperCase();

            if (strict) {
                switch (format) {
                    case 'N':
                    case 'NN':
                    case 'NNN':
                        if (abbr === eraName) {
                            return eras[i];
                        }
                        break;

                    case 'NNNN':
                        if (name === eraName) {
                            return eras[i];
                        }
                        break;

                    case 'NNNNN':
                        if (narrow === eraName) {
                            return eras[i];
                        }
                        break;
                }
            } else if ([name, abbr, narrow].indexOf(eraName) >= 0) {
                return eras[i];
            }
        }
    }

    function localeErasConvertYear(era, year) {
        var dir = era.since <= era.until ? +1 : -1;
        if (year === undefined) {
            return hooks(era.since).year();
        } else {
            return hooks(era.since).year() + (year - era.offset) * dir;
        }
    }

    function getEraName() {
        var i,
            l,
            val,
            eras = this.localeData().eras();
        for (i = 0, l = eras.length; i < l; ++i) {
            // truncate time
            val = this.clone().startOf('day').valueOf();

            if (eras[i].since <= val && val <= eras[i].until) {
                return eras[i].name;
            }
            if (eras[i].until <= val && val <= eras[i].since) {
                return eras[i].name;
            }
        }

        return '';
    }

    function getEraNarrow() {
        var i,
            l,
            val,
            eras = this.localeData().eras();
        for (i = 0, l = eras.length; i < l; ++i) {
            // truncate time
            val = this.clone().startOf('day').valueOf();

            if (eras[i].since <= val && val <= eras[i].until) {
                return eras[i].narrow;
            }
            if (eras[i].until <= val && val <= eras[i].since) {
                return eras[i].narrow;
            }
        }

        return '';
    }

    function getEraAbbr() {
        var i,
            l,
            val,
            eras = this.localeData().eras();
        for (i = 0, l = eras.length; i < l; ++i) {
            // truncate time
            val = this.clone().startOf('day').valueOf();

            if (eras[i].since <= val && val <= eras[i].until) {
                return eras[i].abbr;
            }
            if (eras[i].until <= val && val <= eras[i].since) {
                return eras[i].abbr;
            }
        }

        return '';
    }

    function getEraYear() {
        var i,
            l,
            dir,
            val,
            eras = this.localeData().eras();
        for (i = 0, l = eras.length; i < l; ++i) {
            dir = eras[i].since <= eras[i].until ? +1 : -1;

            // truncate time
            val = this.clone().startOf('day').valueOf();

            if (
                (eras[i].since <= val && val <= eras[i].until) ||
                (eras[i].until <= val && val <= eras[i].since)
            ) {
                return (
                    (this.year() - hooks(eras[i].since).year()) * dir +
                    eras[i].offset
                );
            }
        }

        return this.year();
    }

    function erasNameRegex(isStrict) {
        if (!hasOwnProp(this, '_erasNameRegex')) {
            computeErasParse.call(this);
        }
        return isStrict ? this._erasNameRegex : this._erasRegex;
    }

    function erasAbbrRegex(isStrict) {
        if (!hasOwnProp(this, '_erasAbbrRegex')) {
            computeErasParse.call(this);
        }
        return isStrict ? this._erasAbbrRegex : this._erasRegex;
    }

    function erasNarrowRegex(isStrict) {
        if (!hasOwnProp(this, '_erasNarrowRegex')) {
            computeErasParse.call(this);
        }
        return isStrict ? this._erasNarrowRegex : this._erasRegex;
    }

    function matchEraAbbr(isStrict, locale) {
        return locale.erasAbbrRegex(isStrict);
    }

    function matchEraName(isStrict, locale) {
        return locale.erasNameRegex(isStrict);
    }

    function matchEraNarrow(isStrict, locale) {
        return locale.erasNarrowRegex(isStrict);
    }

    function matchEraYearOrdinal(isStrict, locale) {
        return locale._eraYearOrdinalRegex || matchUnsigned;
    }

    function computeErasParse() {
        var abbrPieces = [],
            namePieces = [],
            narrowPieces = [],
            mixedPieces = [],
            i,
            l,
            eras = this.eras();

        for (i = 0, l = eras.length; i < l; ++i) {
            namePieces.push(regexEscape(eras[i].name));
            abbrPieces.push(regexEscape(eras[i].abbr));
            narrowPieces.push(regexEscape(eras[i].narrow));

            mixedPieces.push(regexEscape(eras[i].name));
            mixedPieces.push(regexEscape(eras[i].abbr));
            mixedPieces.push(regexEscape(eras[i].narrow));
        }

        this._erasRegex = new RegExp('^(' + mixedPieces.join('|') + ')', 'i');
        this._erasNameRegex = new RegExp('^(' + namePieces.join('|') + ')', 'i');
        this._erasAbbrRegex = new RegExp('^(' + abbrPieces.join('|') + ')', 'i');
        this._erasNarrowRegex = new RegExp(
            '^(' + narrowPieces.join('|') + ')',
            'i'
        );
    }

    // FORMATTING

    addFormatToken(0, ['gg', 2], 0, function () {
        return this.weekYear() % 100;
    });

    addFormatToken(0, ['GG', 2], 0, function () {
        return this.isoWeekYear() % 100;
    });

    function addWeekYearFormatToken(token, getter) {
        addFormatToken(0, [token, token.length], 0, getter);
    }

    addWeekYearFormatToken('gggg', 'weekYear');
    addWeekYearFormatToken('ggggg', 'weekYear');
    addWeekYearFormatToken('GGGG', 'isoWeekYear');
    addWeekYearFormatToken('GGGGG', 'isoWeekYear');

    // ALIASES

    addUnitAlias('weekYear', 'gg');
    addUnitAlias('isoWeekYear', 'GG');

    // PRIORITY

    addUnitPriority('weekYear', 1);
    addUnitPriority('isoWeekYear', 1);

    // PARSING

    addRegexToken('G', matchSigned);
    addRegexToken('g', matchSigned);
    addRegexToken('GG', match1to2, match2);
    addRegexToken('gg', match1to2, match2);
    addRegexToken('GGGG', match1to4, match4);
    addRegexToken('gggg', match1to4, match4);
    addRegexToken('GGGGG', match1to6, match6);
    addRegexToken('ggggg', match1to6, match6);

    addWeekParseToken(
        ['gggg', 'ggggg', 'GGGG', 'GGGGG'],
        function (input, week, config, token) {
            week[token.substr(0, 2)] = toInt(input);
        }
    );

    addWeekParseToken(['gg', 'GG'], function (input, week, config, token) {
        week[token] = hooks.parseTwoDigitYear(input);
    });

    // MOMENTS

    function getSetWeekYear(input) {
        return getSetWeekYearHelper.call(
            this,
            input,
            this.week(),
            this.weekday(),
            this.localeData()._week.dow,
            this.localeData()._week.doy
        );
    }

    function getSetISOWeekYear(input) {
        return getSetWeekYearHelper.call(
            this,
            input,
            this.isoWeek(),
            this.isoWeekday(),
            1,
            4
        );
    }

    function getISOWeeksInYear() {
        return weeksInYear(this.year(), 1, 4);
    }

    function getISOWeeksInISOWeekYear() {
        return weeksInYear(this.isoWeekYear(), 1, 4);
    }

    function getWeeksInYear() {
        var weekInfo = this.localeData()._week;
        return weeksInYear(this.year(), weekInfo.dow, weekInfo.doy);
    }

    function getWeeksInWeekYear() {
        var weekInfo = this.localeData()._week;
        return weeksInYear(this.weekYear(), weekInfo.dow, weekInfo.doy);
    }

    function getSetWeekYearHelper(input, week, weekday, dow, doy) {
        var weeksTarget;
        if (input == null) {
            return weekOfYear(this, dow, doy).year;
        } else {
            weeksTarget = weeksInYear(input, dow, doy);
            if (week > weeksTarget) {
                week = weeksTarget;
            }
            return setWeekAll.call(this, input, week, weekday, dow, doy);
        }
    }

    function setWeekAll(weekYear, week, weekday, dow, doy) {
        var dayOfYearData = dayOfYearFromWeeks(weekYear, week, weekday, dow, doy),
            date = createUTCDate(dayOfYearData.year, 0, dayOfYearData.dayOfYear);

        this.year(date.getUTCFullYear());
        this.month(date.getUTCMonth());
        this.date(date.getUTCDate());
        return this;
    }

    // FORMATTING

    addFormatToken('Q', 0, 'Qo', 'quarter');

    // ALIASES

    addUnitAlias('quarter', 'Q');

    // PRIORITY

    addUnitPriority('quarter', 7);

    // PARSING

    addRegexToken('Q', match1);
    addParseToken('Q', function (input, array) {
        array[MONTH] = (toInt(input) - 1) * 3;
    });

    // MOMENTS

    function getSetQuarter(input) {
        return input == null
            ? Math.ceil((this.month() + 1) / 3)
            : this.month((input - 1) * 3 + (this.month() % 3));
    }

    // FORMATTING

    addFormatToken('D', ['DD', 2], 'Do', 'date');

    // ALIASES

    addUnitAlias('date', 'D');

    // PRIORITY
    addUnitPriority('date', 9);

    // PARSING

    addRegexToken('D', match1to2);
    addRegexToken('DD', match1to2, match2);
    addRegexToken('Do', function (isStrict, locale) {
        // TODO: Remove "ordinalParse" fallback in next major release.
        return isStrict
            ? locale._dayOfMonthOrdinalParse || locale._ordinalParse
            : locale._dayOfMonthOrdinalParseLenient;
    });

    addParseToken(['D', 'DD'], DATE);
    addParseToken('Do', function (input, array) {
        array[DATE] = toInt(input.match(match1to2)[0]);
    });

    // MOMENTS

    var getSetDayOfMonth = makeGetSet('Date', true);

    // FORMATTING

    addFormatToken('DDD', ['DDDD', 3], 'DDDo', 'dayOfYear');

    // ALIASES

    addUnitAlias('dayOfYear', 'DDD');

    // PRIORITY
    addUnitPriority('dayOfYear', 4);

    // PARSING

    addRegexToken('DDD', match1to3);
    addRegexToken('DDDD', match3);
    addParseToken(['DDD', 'DDDD'], function (input, array, config) {
        config._dayOfYear = toInt(input);
    });

    // HELPERS

    // MOMENTS

    function getSetDayOfYear(input) {
        var dayOfYear =
            Math.round(
                (this.clone().startOf('day') - this.clone().startOf('year')) / 864e5
            ) + 1;
        return input == null ? dayOfYear : this.add(input - dayOfYear, 'd');
    }

    // FORMATTING

    addFormatToken('m', ['mm', 2], 0, 'minute');

    // ALIASES

    addUnitAlias('minute', 'm');

    // PRIORITY

    addUnitPriority('minute', 14);

    // PARSING

    addRegexToken('m', match1to2);
    addRegexToken('mm', match1to2, match2);
    addParseToken(['m', 'mm'], MINUTE);

    // MOMENTS

    var getSetMinute = makeGetSet('Minutes', false);

    // FORMATTING

    addFormatToken('s', ['ss', 2], 0, 'second');

    // ALIASES

    addUnitAlias('second', 's');

    // PRIORITY

    addUnitPriority('second', 15);

    // PARSING

    addRegexToken('s', match1to2);
    addRegexToken('ss', match1to2, match2);
    addParseToken(['s', 'ss'], SECOND);

    // MOMENTS

    var getSetSecond = makeGetSet('Seconds', false);

    // FORMATTING

    addFormatToken('S', 0, 0, function () {
        return ~~(this.millisecond() / 100);
    });

    addFormatToken(0, ['SS', 2], 0, function () {
        return ~~(this.millisecond() / 10);
    });

    addFormatToken(0, ['SSS', 3], 0, 'millisecond');
    addFormatToken(0, ['SSSS', 4], 0, function () {
        return this.millisecond() * 10;
    });
    addFormatToken(0, ['SSSSS', 5], 0, function () {
        return this.millisecond() * 100;
    });
    addFormatToken(0, ['SSSSSS', 6], 0, function () {
        return this.millisecond() * 1000;
    });
    addFormatToken(0, ['SSSSSSS', 7], 0, function () {
        return this.millisecond() * 10000;
    });
    addFormatToken(0, ['SSSSSSSS', 8], 0, function () {
        return this.millisecond() * 100000;
    });
    addFormatToken(0, ['SSSSSSSSS', 9], 0, function () {
        return this.millisecond() * 1000000;
    });

    // ALIASES

    addUnitAlias('millisecond', 'ms');

    // PRIORITY

    addUnitPriority('millisecond', 16);

    // PARSING

    addRegexToken('S', match1to3, match1);
    addRegexToken('SS', match1to3, match2);
    addRegexToken('SSS', match1to3, match3);

    var token, getSetMillisecond;
    for (token = 'SSSS'; token.length <= 9; token += 'S') {
        addRegexToken(token, matchUnsigned);
    }

    function parseMs(input, array) {
        array[MILLISECOND] = toInt(('0.' + input) * 1000);
    }

    for (token = 'S'; token.length <= 9; token += 'S') {
        addParseToken(token, parseMs);
    }

    getSetMillisecond = makeGetSet('Milliseconds', false);

    // FORMATTING

    addFormatToken('z', 0, 0, 'zoneAbbr');
    addFormatToken('zz', 0, 0, 'zoneName');

    // MOMENTS

    function getZoneAbbr() {
        return this._isUTC ? 'UTC' : '';
    }

    function getZoneName() {
        return this._isUTC ? 'Coordinated Universal Time' : '';
    }

    var proto = Moment.prototype;

    proto.add = add;
    proto.calendar = calendar$1;
    proto.clone = clone;
    proto.diff = diff;
    proto.endOf = endOf;
    proto.format = format;
    proto.from = from;
    proto.fromNow = fromNow;
    proto.to = to;
    proto.toNow = toNow;
    proto.get = stringGet;
    proto.invalidAt = invalidAt;
    proto.isAfter = isAfter;
    proto.isBefore = isBefore;
    proto.isBetween = isBetween;
    proto.isSame = isSame;
    proto.isSameOrAfter = isSameOrAfter;
    proto.isSameOrBefore = isSameOrBefore;
    proto.isValid = isValid$2;
    proto.lang = lang;
    proto.locale = locale;
    proto.localeData = localeData;
    proto.max = prototypeMax;
    proto.min = prototypeMin;
    proto.parsingFlags = parsingFlags;
    proto.set = stringSet;
    proto.startOf = startOf;
    proto.subtract = subtract;
    proto.toArray = toArray;
    proto.toObject = toObject;
    proto.toDate = toDate;
    proto.toISOString = toISOString;
    proto.inspect = inspect;
    if (typeof Symbol !== 'undefined' && Symbol.for != null) {
        proto[Symbol.for('nodejs.util.inspect.custom')] = function () {
            return 'Moment<' + this.format() + '>';
        };
    }
    proto.toJSON = toJSON;
    proto.toString = toString;
    proto.unix = unix;
    proto.valueOf = valueOf;
    proto.creationData = creationData;
    proto.eraName = getEraName;
    proto.eraNarrow = getEraNarrow;
    proto.eraAbbr = getEraAbbr;
    proto.eraYear = getEraYear;
    proto.year = getSetYear;
    proto.isLeapYear = getIsLeapYear;
    proto.weekYear = getSetWeekYear;
    proto.isoWeekYear = getSetISOWeekYear;
    proto.quarter = proto.quarters = getSetQuarter;
    proto.month = getSetMonth;
    proto.daysInMonth = getDaysInMonth;
    proto.week = proto.weeks = getSetWeek;
    proto.isoWeek = proto.isoWeeks = getSetISOWeek;
    proto.weeksInYear = getWeeksInYear;
    proto.weeksInWeekYear = getWeeksInWeekYear;
    proto.isoWeeksInYear = getISOWeeksInYear;
    proto.isoWeeksInISOWeekYear = getISOWeeksInISOWeekYear;
    proto.date = getSetDayOfMonth;
    proto.day = proto.days = getSetDayOfWeek;
    proto.weekday = getSetLocaleDayOfWeek;
    proto.isoWeekday = getSetISODayOfWeek;
    proto.dayOfYear = getSetDayOfYear;
    proto.hour = proto.hours = getSetHour;
    proto.minute = proto.minutes = getSetMinute;
    proto.second = proto.seconds = getSetSecond;
    proto.millisecond = proto.milliseconds = getSetMillisecond;
    proto.utcOffset = getSetOffset;
    proto.utc = setOffsetToUTC;
    proto.local = setOffsetToLocal;
    proto.parseZone = setOffsetToParsedOffset;
    proto.hasAlignedHourOffset = hasAlignedHourOffset;
    proto.isDST = isDaylightSavingTime;
    proto.isLocal = isLocal;
    proto.isUtcOffset = isUtcOffset;
    proto.isUtc = isUtc;
    proto.isUTC = isUtc;
    proto.zoneAbbr = getZoneAbbr;
    proto.zoneName = getZoneName;
    proto.dates = deprecate(
        'dates accessor is deprecated. Use date instead.',
        getSetDayOfMonth
    );
    proto.months = deprecate(
        'months accessor is deprecated. Use month instead',
        getSetMonth
    );
    proto.years = deprecate(
        'years accessor is deprecated. Use year instead',
        getSetYear
    );
    proto.zone = deprecate(
        'moment().zone is deprecated, use moment().utcOffset instead. http://momentjs.com/guides/#/warnings/zone/',
        getSetZone
    );
    proto.isDSTShifted = deprecate(
        'isDSTShifted is deprecated. See http://momentjs.com/guides/#/warnings/dst-shifted/ for more information',
        isDaylightSavingTimeShifted
    );

    function createUnix(input) {
        return createLocal(input * 1000);
    }

    function createInZone() {
        return createLocal.apply(null, arguments).parseZone();
    }

    function preParsePostFormat(string) {
        return string;
    }

    var proto$1 = Locale.prototype;

    proto$1.calendar = calendar;
    proto$1.longDateFormat = longDateFormat;
    proto$1.invalidDate = invalidDate;
    proto$1.ordinal = ordinal;
    proto$1.preparse = preParsePostFormat;
    proto$1.postformat = preParsePostFormat;
    proto$1.relativeTime = relativeTime;
    proto$1.pastFuture = pastFuture;
    proto$1.set = set;
    proto$1.eras = localeEras;
    proto$1.erasParse = localeErasParse;
    proto$1.erasConvertYear = localeErasConvertYear;
    proto$1.erasAbbrRegex = erasAbbrRegex;
    proto$1.erasNameRegex = erasNameRegex;
    proto$1.erasNarrowRegex = erasNarrowRegex;

    proto$1.months = localeMonths;
    proto$1.monthsShort = localeMonthsShort;
    proto$1.monthsParse = localeMonthsParse;
    proto$1.monthsRegex = monthsRegex;
    proto$1.monthsShortRegex = monthsShortRegex;
    proto$1.week = localeWeek;
    proto$1.firstDayOfYear = localeFirstDayOfYear;
    proto$1.firstDayOfWeek = localeFirstDayOfWeek;

    proto$1.weekdays = localeWeekdays;
    proto$1.weekdaysMin = localeWeekdaysMin;
    proto$1.weekdaysShort = localeWeekdaysShort;
    proto$1.weekdaysParse = localeWeekdaysParse;

    proto$1.weekdaysRegex = weekdaysRegex;
    proto$1.weekdaysShortRegex = weekdaysShortRegex;
    proto$1.weekdaysMinRegex = weekdaysMinRegex;

    proto$1.isPM = localeIsPM;
    proto$1.meridiem = localeMeridiem;

    function get$1(format, index, field, setter) {
        var locale = getLocale(),
            utc = createUTC().set(setter, index);
        return locale[field](utc, format);
    }

    function listMonthsImpl(format, index, field) {
        if (isNumber(format)) {
            index = format;
            format = undefined;
        }

        format = format || '';

        if (index != null) {
            return get$1(format, index, field, 'month');
        }

        var i,
            out = [];
        for (i = 0; i < 12; i++) {
            out[i] = get$1(format, i, field, 'month');
        }
        return out;
    }

    // ()
    // (5)
    // (fmt, 5)
    // (fmt)
    // (true)
    // (true, 5)
    // (true, fmt, 5)
    // (true, fmt)
    function listWeekdaysImpl(localeSorted, format, index, field) {
        if (typeof localeSorted === 'boolean') {
            if (isNumber(format)) {
                index = format;
                format = undefined;
            }

            format = format || '';
        } else {
            format = localeSorted;
            index = format;
            localeSorted = false;

            if (isNumber(format)) {
                index = format;
                format = undefined;
            }

            format = format || '';
        }

        var locale = getLocale(),
            shift = localeSorted ? locale._week.dow : 0,
            i,
            out = [];

        if (index != null) {
            return get$1(format, (index + shift) % 7, field, 'day');
        }

        for (i = 0; i < 7; i++) {
            out[i] = get$1(format, (i + shift) % 7, field, 'day');
        }
        return out;
    }

    function listMonths(format, index) {
        return listMonthsImpl(format, index, 'months');
    }

    function listMonthsShort(format, index) {
        return listMonthsImpl(format, index, 'monthsShort');
    }

    function listWeekdays(localeSorted, format, index) {
        return listWeekdaysImpl(localeSorted, format, index, 'weekdays');
    }

    function listWeekdaysShort(localeSorted, format, index) {
        return listWeekdaysImpl(localeSorted, format, index, 'weekdaysShort');
    }

    function listWeekdaysMin(localeSorted, format, index) {
        return listWeekdaysImpl(localeSorted, format, index, 'weekdaysMin');
    }

    getSetGlobalLocale('en', {
        eras: [
            {
                since: '0001-01-01',
                until: +Infinity,
                offset: 1,
                name: 'Anno Domini',
                narrow: 'AD',
                abbr: 'AD',
            },
            {
                since: '0000-12-31',
                until: -Infinity,
                offset: 1,
                name: 'Before Christ',
                narrow: 'BC',
                abbr: 'BC',
            },
        ],
        dayOfMonthOrdinalParse: /\d{1,2}(th|st|nd|rd)/,
        ordinal: function (number) {
            var b = number % 10,
                output =
                    toInt((number % 100) / 10) === 1
                        ? 'th'
                        : b === 1
                        ? 'st'
                        : b === 2
                        ? 'nd'
                        : b === 3
                        ? 'rd'
                        : 'th';
            return number + output;
        },
    });

    // Side effect imports

    hooks.lang = deprecate(
        'moment.lang is deprecated. Use moment.locale instead.',
        getSetGlobalLocale
    );
    hooks.langData = deprecate(
        'moment.langData is deprecated. Use moment.localeData instead.',
        getLocale
    );

    var mathAbs = Math.abs;

    function abs() {
        var data = this._data;

        this._milliseconds = mathAbs(this._milliseconds);
        this._days = mathAbs(this._days);
        this._months = mathAbs(this._months);

        data.milliseconds = mathAbs(data.milliseconds);
        data.seconds = mathAbs(data.seconds);
        data.minutes = mathAbs(data.minutes);
        data.hours = mathAbs(data.hours);
        data.months = mathAbs(data.months);
        data.years = mathAbs(data.years);

        return this;
    }

    function addSubtract$1(duration, input, value, direction) {
        var other = createDuration(input, value);

        duration._milliseconds += direction * other._milliseconds;
        duration._days += direction * other._days;
        duration._months += direction * other._months;

        return duration._bubble();
    }

    // supports only 2.0-style add(1, 's') or add(duration)
    function add$1(input, value) {
        return addSubtract$1(this, input, value, 1);
    }

    // supports only 2.0-style subtract(1, 's') or subtract(duration)
    function subtract$1(input, value) {
        return addSubtract$1(this, input, value, -1);
    }

    function absCeil(number) {
        if (number < 0) {
            return Math.floor(number);
        } else {
            return Math.ceil(number);
        }
    }

    function bubble() {
        var milliseconds = this._milliseconds,
            days = this._days,
            months = this._months,
            data = this._data,
            seconds,
            minutes,
            hours,
            years,
            monthsFromDays;

        // if we have a mix of positive and negative values, bubble down first
        // check: https://github.com/moment/moment/issues/2166
        if (
            !(
                (milliseconds >= 0 && days >= 0 && months >= 0) ||
                (milliseconds <= 0 && days <= 0 && months <= 0)
            )
        ) {
            milliseconds += absCeil(monthsToDays(months) + days) * 864e5;
            days = 0;
            months = 0;
        }

        // The following code bubbles up values, see the tests for
        // examples of what that means.
        data.milliseconds = milliseconds % 1000;

        seconds = absFloor(milliseconds / 1000);
        data.seconds = seconds % 60;

        minutes = absFloor(seconds / 60);
        data.minutes = minutes % 60;

        hours = absFloor(minutes / 60);
        data.hours = hours % 24;

        days += absFloor(hours / 24);

        // convert days to months
        monthsFromDays = absFloor(daysToMonths(days));
        months += monthsFromDays;
        days -= absCeil(monthsToDays(monthsFromDays));

        // 12 months -> 1 year
        years = absFloor(months / 12);
        months %= 12;

        data.days = days;
        data.months = months;
        data.years = years;

        return this;
    }

    function daysToMonths(days) {
        // 400 years have 146097 days (taking into account leap year rules)
        // 400 years have 12 months === 4800
        return (days * 4800) / 146097;
    }

    function monthsToDays(months) {
        // the reverse of daysToMonths
        return (months * 146097) / 4800;
    }

    function as(units) {
        if (!this.isValid()) {
            return NaN;
        }
        var days,
            months,
            milliseconds = this._milliseconds;

        units = normalizeUnits(units);

        if (units === 'month' || units === 'quarter' || units === 'year') {
            days = this._days + milliseconds / 864e5;
            months = this._months + daysToMonths(days);
            switch (units) {
                case 'month':
                    return months;
                case 'quarter':
                    return months / 3;
                case 'year':
                    return months / 12;
            }
        } else {
            // handle milliseconds separately because of floating point math errors (issue #1867)
            days = this._days + Math.round(monthsToDays(this._months));
            switch (units) {
                case 'week':
                    return days / 7 + milliseconds / 6048e5;
                case 'day':
                    return days + milliseconds / 864e5;
                case 'hour':
                    return days * 24 + milliseconds / 36e5;
                case 'minute':
                    return days * 1440 + milliseconds / 6e4;
                case 'second':
                    return days * 86400 + milliseconds / 1000;
                // Math.floor prevents floating point math errors here
                case 'millisecond':
                    return Math.floor(days * 864e5) + milliseconds;
                default:
                    throw new Error('Unknown unit ' + units);
            }
        }
    }

    // TODO: Use this.as('ms')?
    function valueOf$1() {
        if (!this.isValid()) {
            return NaN;
        }
        return (
            this._milliseconds +
            this._days * 864e5 +
            (this._months % 12) * 2592e6 +
            toInt(this._months / 12) * 31536e6
        );
    }

    function makeAs(alias) {
        return function () {
            return this.as(alias);
        };
    }

    var asMilliseconds = makeAs('ms'),
        asSeconds = makeAs('s'),
        asMinutes = makeAs('m'),
        asHours = makeAs('h'),
        asDays = makeAs('d'),
        asWeeks = makeAs('w'),
        asMonths = makeAs('M'),
        asQuarters = makeAs('Q'),
        asYears = makeAs('y');

    function clone$1() {
        return createDuration(this);
    }

    function get$2(units) {
        units = normalizeUnits(units);
        return this.isValid() ? this[units + 's']() : NaN;
    }

    function makeGetter(name) {
        return function () {
            return this.isValid() ? this._data[name] : NaN;
        };
    }

    var milliseconds = makeGetter('milliseconds'),
        seconds = makeGetter('seconds'),
        minutes = makeGetter('minutes'),
        hours = makeGetter('hours'),
        days = makeGetter('days'),
        months = makeGetter('months'),
        years = makeGetter('years');

    function weeks() {
        return absFloor(this.days() / 7);
    }

    var round = Math.round,
        thresholds = {
            ss: 44, // a few seconds to seconds
            s: 45, // seconds to minute
            m: 45, // minutes to hour
            h: 22, // hours to day
            d: 26, // days to month/week
            w: null, // weeks to month
            M: 11, // months to year
        };

    // helper function for moment.fn.from, moment.fn.fromNow, and moment.duration.fn.humanize
    function substituteTimeAgo(string, number, withoutSuffix, isFuture, locale) {
        return locale.relativeTime(number || 1, !!withoutSuffix, string, isFuture);
    }

    function relativeTime$1(posNegDuration, withoutSuffix, thresholds, locale) {
        var duration = createDuration(posNegDuration).abs(),
            seconds = round(duration.as('s')),
            minutes = round(duration.as('m')),
            hours = round(duration.as('h')),
            days = round(duration.as('d')),
            months = round(duration.as('M')),
            weeks = round(duration.as('w')),
            years = round(duration.as('y')),
            a =
                (seconds <= thresholds.ss && ['s', seconds]) ||
                (seconds < thresholds.s && ['ss', seconds]) ||
                (minutes <= 1 && ['m']) ||
                (minutes < thresholds.m && ['mm', minutes]) ||
                (hours <= 1 && ['h']) ||
                (hours < thresholds.h && ['hh', hours]) ||
                (days <= 1 && ['d']) ||
                (days < thresholds.d && ['dd', days]);

        if (thresholds.w != null) {
            a =
                a ||
                (weeks <= 1 && ['w']) ||
                (weeks < thresholds.w && ['ww', weeks]);
        }
        a = a ||
            (months <= 1 && ['M']) ||
            (months < thresholds.M && ['MM', months]) ||
            (years <= 1 && ['y']) || ['yy', years];

        a[2] = withoutSuffix;
        a[3] = +posNegDuration > 0;
        a[4] = locale;
        return substituteTimeAgo.apply(null, a);
    }

    // This function allows you to set the rounding function for relative time strings
    function getSetRelativeTimeRounding(roundingFunction) {
        if (roundingFunction === undefined) {
            return round;
        }
        if (typeof roundingFunction === 'function') {
            round = roundingFunction;
            return true;
        }
        return false;
    }

    // This function allows you to set a threshold for relative time strings
    function getSetRelativeTimeThreshold(threshold, limit) {
        if (thresholds[threshold] === undefined) {
            return false;
        }
        if (limit === undefined) {
            return thresholds[threshold];
        }
        thresholds[threshold] = limit;
        if (threshold === 's') {
            thresholds.ss = limit - 1;
        }
        return true;
    }

    function humanize(argWithSuffix, argThresholds) {
        if (!this.isValid()) {
            return this.localeData().invalidDate();
        }

        var withSuffix = false,
            th = thresholds,
            locale,
            output;

        if (typeof argWithSuffix === 'object') {
            argThresholds = argWithSuffix;
            argWithSuffix = false;
        }
        if (typeof argWithSuffix === 'boolean') {
            withSuffix = argWithSuffix;
        }
        if (typeof argThresholds === 'object') {
            th = Object.assign({}, thresholds, argThresholds);
            if (argThresholds.s != null && argThresholds.ss == null) {
                th.ss = argThresholds.s - 1;
            }
        }

        locale = this.localeData();
        output = relativeTime$1(this, !withSuffix, th, locale);

        if (withSuffix) {
            output = locale.pastFuture(+this, output);
        }

        return locale.postformat(output);
    }

    var abs$1 = Math.abs;

    function sign(x) {
        return (x > 0) - (x < 0) || +x;
    }

    function toISOString$1() {
        // for ISO strings we do not use the normal bubbling rules:
        //  * milliseconds bubble up until they become hours
        //  * days do not bubble at all
        //  * months bubble up until they become years
        // This is because there is no context-free conversion between hours and days
        // (think of clock changes)
        // and also not between days and months (28-31 days per month)
        if (!this.isValid()) {
            return this.localeData().invalidDate();
        }

        var seconds = abs$1(this._milliseconds) / 1000,
            days = abs$1(this._days),
            months = abs$1(this._months),
            minutes,
            hours,
            years,
            s,
            total = this.asSeconds(),
            totalSign,
            ymSign,
            daysSign,
            hmsSign;

        if (!total) {
            // this is the same as C#'s (Noda) and python (isodate)...
            // but not other JS (goog.date)
            return 'P0D';
        }

        // 3600 seconds -> 60 minutes -> 1 hour
        minutes = absFloor(seconds / 60);
        hours = absFloor(minutes / 60);
        seconds %= 60;
        minutes %= 60;

        // 12 months -> 1 year
        years = absFloor(months / 12);
        months %= 12;

        // inspired by https://github.com/dordille/moment-isoduration/blob/master/moment.isoduration.js
        s = seconds ? seconds.toFixed(3).replace(/\.?0+$/, '') : '';

        totalSign = total < 0 ? '-' : '';
        ymSign = sign(this._months) !== sign(total) ? '-' : '';
        daysSign = sign(this._days) !== sign(total) ? '-' : '';
        hmsSign = sign(this._milliseconds) !== sign(total) ? '-' : '';

        return (
            totalSign +
            'P' +
            (years ? ymSign + years + 'Y' : '') +
            (months ? ymSign + months + 'M' : '') +
            (days ? daysSign + days + 'D' : '') +
            (hours || minutes || seconds ? 'T' : '') +
            (hours ? hmsSign + hours + 'H' : '') +
            (minutes ? hmsSign + minutes + 'M' : '') +
            (seconds ? hmsSign + s + 'S' : '')
        );
    }

    var proto$2 = Duration.prototype;

    proto$2.isValid = isValid$1;
    proto$2.abs = abs;
    proto$2.add = add$1;
    proto$2.subtract = subtract$1;
    proto$2.as = as;
    proto$2.asMilliseconds = asMilliseconds;
    proto$2.asSeconds = asSeconds;
    proto$2.asMinutes = asMinutes;
    proto$2.asHours = asHours;
    proto$2.asDays = asDays;
    proto$2.asWeeks = asWeeks;
    proto$2.asMonths = asMonths;
    proto$2.asQuarters = asQuarters;
    proto$2.asYears = asYears;
    proto$2.valueOf = valueOf$1;
    proto$2._bubble = bubble;
    proto$2.clone = clone$1;
    proto$2.get = get$2;
    proto$2.milliseconds = milliseconds;
    proto$2.seconds = seconds;
    proto$2.minutes = minutes;
    proto$2.hours = hours;
    proto$2.days = days;
    proto$2.weeks = weeks;
    proto$2.months = months;
    proto$2.years = years;
    proto$2.humanize = humanize;
    proto$2.toISOString = toISOString$1;
    proto$2.toString = toISOString$1;
    proto$2.toJSON = toISOString$1;
    proto$2.locale = locale;
    proto$2.localeData = localeData;

    proto$2.toIsoString = deprecate(
        'toIsoString() is deprecated. Please use toISOString() instead (notice the capitals)',
        toISOString$1
    );
    proto$2.lang = lang;

    // FORMATTING

    addFormatToken('X', 0, 0, 'unix');
    addFormatToken('x', 0, 0, 'valueOf');

    // PARSING

    addRegexToken('x', matchSigned);
    addRegexToken('X', matchTimestamp);
    addParseToken('X', function (input, array, config) {
        config._d = new Date(parseFloat(input) * 1000);
    });
    addParseToken('x', function (input, array, config) {
        config._d = new Date(toInt(input));
    });

    //! moment.js

    hooks.version = '2.29.4';

    setHookCallback(createLocal);

    hooks.fn = proto;
    hooks.min = min;
    hooks.max = max;
    hooks.now = now;
    hooks.utc = createUTC;
    hooks.unix = createUnix;
    hooks.months = listMonths;
    hooks.isDate = isDate;
    hooks.locale = getSetGlobalLocale;
    hooks.invalid = createInvalid;
    hooks.duration = createDuration;
    hooks.isMoment = isMoment;
    hooks.weekdays = listWeekdays;
    hooks.parseZone = createInZone;
    hooks.localeData = getLocale;
    hooks.isDuration = isDuration;
    hooks.monthsShort = listMonthsShort;
    hooks.weekdaysMin = listWeekdaysMin;
    hooks.defineLocale = defineLocale;
    hooks.updateLocale = updateLocale;
    hooks.locales = listLocales;
    hooks.weekdaysShort = listWeekdaysShort;
    hooks.normalizeUnits = normalizeUnits;
    hooks.relativeTimeRounding = getSetRelativeTimeRounding;
    hooks.relativeTimeThreshold = getSetRelativeTimeThreshold;
    hooks.calendarFormat = getCalendarFormat;
    hooks.prototype = proto;

    // currently HTML5 input type only supports 24-hour formats
    hooks.HTML5_FMT = {
        DATETIME_LOCAL: 'YYYY-MM-DDTHH:mm', // <input type="datetime-local" />
        DATETIME_LOCAL_SECONDS: 'YYYY-MM-DDTHH:mm:ss', // <input type="datetime-local" step="1" />
        DATETIME_LOCAL_MS: 'YYYY-MM-DDTHH:mm:ss.SSS', // <input type="datetime-local" step="0.001" />
        DATE: 'YYYY-MM-DD', // <input type="date" />
        TIME: 'HH:mm', // <input type="time" />
        TIME_SECONDS: 'HH:mm:ss', // <input type="time" step="1" />
        TIME_MS: 'HH:mm:ss.SSS', // <input type="time" step="0.001" />
        WEEK: 'GGGG-[W]WW', // <input type="week" />
        MONTH: 'YYYY-MM', // <input type="month" />
    };

    return hooks;

})));


/************** (from file ./js/other/moment.js)**/ 

!function(n){"use strict";function d(n,t){var r=(65535&n)+(65535&t);return(n>>16)+(t>>16)+(r>>16)<<16|65535&r}function f(n,t,r,e,o,u){return d((u=d(d(t,n),d(e,u)))<<o|u>>>32-o,r)}function l(n,t,r,e,o,u,c){return f(t&r|~t&e,n,t,o,u,c)}function g(n,t,r,e,o,u,c){return f(t&e|r&~e,n,t,o,u,c)}function v(n,t,r,e,o,u,c){return f(t^r^e,n,t,o,u,c)}function m(n,t,r,e,o,u,c){return f(r^(t|~e),n,t,o,u,c)}function c(n,t){var r,e,o,u;n[t>>5]|=128<<t%32,n[14+(t+64>>>9<<4)]=t;for(var c=1732584193,f=-271733879,i=-1732584194,a=271733878,h=0;h<n.length;h+=16)c=l(r=c,e=f,o=i,u=a,n[h],7,-680876936),a=l(a,c,f,i,n[h+1],12,-389564586),i=l(i,a,c,f,n[h+2],17,606105819),f=l(f,i,a,c,n[h+3],22,-1044525330),c=l(c,f,i,a,n[h+4],7,-176418897),a=l(a,c,f,i,n[h+5],12,1200080426),i=l(i,a,c,f,n[h+6],17,-1473231341),f=l(f,i,a,c,n[h+7],22,-45705983),c=l(c,f,i,a,n[h+8],7,1770035416),a=l(a,c,f,i,n[h+9],12,-1958414417),i=l(i,a,c,f,n[h+10],17,-42063),f=l(f,i,a,c,n[h+11],22,-1990404162),c=l(c,f,i,a,n[h+12],7,1804603682),a=l(a,c,f,i,n[h+13],12,-40341101),i=l(i,a,c,f,n[h+14],17,-1502002290),c=g(c,f=l(f,i,a,c,n[h+15],22,1236535329),i,a,n[h+1],5,-165796510),a=g(a,c,f,i,n[h+6],9,-1069501632),i=g(i,a,c,f,n[h+11],14,643717713),f=g(f,i,a,c,n[h],20,-373897302),c=g(c,f,i,a,n[h+5],5,-701558691),a=g(a,c,f,i,n[h+10],9,38016083),i=g(i,a,c,f,n[h+15],14,-660478335),f=g(f,i,a,c,n[h+4],20,-405537848),c=g(c,f,i,a,n[h+9],5,568446438),a=g(a,c,f,i,n[h+14],9,-1019803690),i=g(i,a,c,f,n[h+3],14,-187363961),f=g(f,i,a,c,n[h+8],20,1163531501),c=g(c,f,i,a,n[h+13],5,-1444681467),a=g(a,c,f,i,n[h+2],9,-51403784),i=g(i,a,c,f,n[h+7],14,1735328473),c=v(c,f=g(f,i,a,c,n[h+12],20,-1926607734),i,a,n[h+5],4,-378558),a=v(a,c,f,i,n[h+8],11,-2022574463),i=v(i,a,c,f,n[h+11],16,1839030562),f=v(f,i,a,c,n[h+14],23,-35309556),c=v(c,f,i,a,n[h+1],4,-1530992060),a=v(a,c,f,i,n[h+4],11,1272893353),i=v(i,a,c,f,n[h+7],16,-155497632),f=v(f,i,a,c,n[h+10],23,-1094730640),c=v(c,f,i,a,n[h+13],4,681279174),a=v(a,c,f,i,n[h],11,-358537222),i=v(i,a,c,f,n[h+3],16,-722521979),f=v(f,i,a,c,n[h+6],23,76029189),c=v(c,f,i,a,n[h+9],4,-640364487),a=v(a,c,f,i,n[h+12],11,-421815835),i=v(i,a,c,f,n[h+15],16,530742520),c=m(c,f=v(f,i,a,c,n[h+2],23,-995338651),i,a,n[h],6,-198630844),a=m(a,c,f,i,n[h+7],10,1126891415),i=m(i,a,c,f,n[h+14],15,-1416354905),f=m(f,i,a,c,n[h+5],21,-57434055),c=m(c,f,i,a,n[h+12],6,1700485571),a=m(a,c,f,i,n[h+3],10,-1894986606),i=m(i,a,c,f,n[h+10],15,-1051523),f=m(f,i,a,c,n[h+1],21,-2054922799),c=m(c,f,i,a,n[h+8],6,1873313359),a=m(a,c,f,i,n[h+15],10,-30611744),i=m(i,a,c,f,n[h+6],15,-1560198380),f=m(f,i,a,c,n[h+13],21,1309151649),c=m(c,f,i,a,n[h+4],6,-145523070),a=m(a,c,f,i,n[h+11],10,-1120210379),i=m(i,a,c,f,n[h+2],15,718787259),f=m(f,i,a,c,n[h+9],21,-343485551),c=d(c,r),f=d(f,e),i=d(i,o),a=d(a,u);return[c,f,i,a]}function i(n){for(var t="",r=32*n.length,e=0;e<r;e+=8)t+=String.fromCharCode(n[e>>5]>>>e%32&255);return t}function a(n){var t=[];for(t[(n.length>>2)-1]=void 0,e=0;e<t.length;e+=1)t[e]=0;for(var r=8*n.length,e=0;e<r;e+=8)t[e>>5]|=(255&n.charCodeAt(e/8))<<e%32;return t}function e(n){for(var t,r="0123456789abcdef",e="",o=0;o<n.length;o+=1)t=n.charCodeAt(o),e+=r.charAt(t>>>4&15)+r.charAt(15&t);return e}function r(n){return unescape(encodeURIComponent(n))}function o(n){return i(c(a(n=r(n)),8*n.length))}function u(n,t){return function(n,t){var r,e=a(n),o=[],u=[];for(o[15]=u[15]=void 0,16<e.length&&(e=c(e,8*n.length)),r=0;r<16;r+=1)o[r]=909522486^e[r],u[r]=1549556828^e[r];return t=c(o.concat(a(t)),512+8*t.length),i(c(u.concat(t),640))}(r(n),r(t))}function t(n,t,r){return t?r?u(t,n):e(u(t,n)):r?o(n):e(o(n))}"function"==typeof define&&define.amd?define(function(){return t}):"object"==typeof module&&module.exports?module.exports=t:n.md5=t}(this);
//# sourceMappingURL=md5.min.js.map

/************** (from file ./js/other/md5.min.js)**/ 

/**
 * We will potentially have MILLIONS of relons in IDB memory.
 * In earlier iterations of the MEM we loaded ALL relons into the hypergraph
 * and had access anywhere in the app to the neurons/relons outs/ints etc.
 * 
 * This is not possible with a big memory footprint nor efficient.
 * 
 * We now load the relons from IDB on-demand, typically when we are going to check
 * ANNE for more relons.  This ensures we only load MEM in RAM when we need it.
 * 
 * We do a default MEM load of relons with a flag of .preload table everytime 
 * any app starts.  This will only grow with basic class definitions and some other
 * relons we target to store at bootup.
 */


var IDB = {}

//these pulls from IDB are NEVER updating IDB.  We just need to run them throught the MEM processor
//to get them added in to the local RAM hypergraph.
IDB.bUPDATEMEM=false;
IDB.init = function() {
    IDB.map = new arraymap();
    IDB.map.setPrimaryKey("nid");
}

/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $bSPONSOREDTOO 
 * @returns 
 */
IDB.getAllMyStuff=async($phObj=null, $nid, $bSPONSOREDTOO)=> {
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            if (aRelonsToProcess[r].to_s!=$nid) {
                if ($bSPONSOREDTOO==false) continue;
                if (aRelonsToProcess[r].sponsored_by!=undefined && aRelonsToProcess[r].sponsored_by!=$nid) {
                    if (aRelonsToProcess[r].relon_sponsor!=undefined && aRelonsToProcess[r].relon_sponsor!=$nid) {
                        continue;
                    }
                }
            }
        }
        aRelons.push(aRelonsToProcess[r]);
    }
    
    console.log("%c IDB getAllMy$tuff RAM LOAD%c"+aRelons.length+"%c relons", TALK.bgyellow, TALK.bggreen, TALK.bgblack);
    
    if (aRelons.length==0) {
        return [];
    }

    //NOTE:when we remove the timer just return the return value and don't store it in aRelons2.
	var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, IDB.bUPDATEMEM, false, "relons", "<<from>> IDB.getAllMyStuff");

    return aRelons2;
}


/**
 * When querying for recents we pull a bunch of relons.
 * To avoid insert errors and to avoid looking up each txid in IDB one by one
 * We will load some memory here.
 * @param {*} $rels 
 * @returns 
 */
IDB.extractAndIDBFromNids=async($rels)=>{
    var nids = [];
    for (var r=0; r < $rels.length; r++) {
        if (nids.indexOf($rels[r].from_nid)==-1) {
            nids.push($rels[r].from_nid);
        }
    }
    var types=[0,1,2];
    var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", nids.join("")+types.join(""));
    await IDB.lookupNids(phObj, nids, types);
    return;
}


/**
 * use an array map of new data relons we have pulled from ANNE.
 * Some may be already in IDB memory but not in RAM.
 * 
 * Loop IDB memory to find what we have already stored in IDB.
 * If we have it in IDB, take it out of the relonsMap (but still process it into RAM)
 * 
 * The leftover .arr will be the new relons.
 *
 * @param {array_map} $relonsMap 
 */
IDB.checkForNewTxids=async($relonsMap)=>{
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        var obj = $relonsMap.get(aRelonsToProcess[r].txid);
        if (obj!=null) {
            $relonsMap.remove(aRelonsToProcess[r].txid, "txid");
            aRelons.push(obj);
            continue;
        }
    }

    if (aRelons.length==0) return [];

    //only putting the net-diff aRelons in RAM here.
    await MEMP.processNeuronsFromRelons(aRelons, false, false, "relons", "IDB.checkForNewTxids");
}


/**
 * 
 * @param {*} $relonsMap 
 */
IDB.checkForNewTxids_ONLY=async($relonsMap, $_thisRunNECROSIS=[])=>{
    //not needed, we loaded ALL mem at start
    var aRelonsToProcess = await DB.loadAll("relons");
    var rs = [];

    //process ALL RELONS
    for (var r=0; r < aRelonsToProcess.length; r++) {

        if ($relonsMap.arr.length==0) return;
            
        //in this data pull we found a neuron that is nuked.
        //tag ALL relons of that neuron as nuked.
        //we will remove them from IDB and RAM.
        if ($_thisRunNECROSIS.length) {
            if (aRelonsToProcess[r].to_type==TO_TYPE_ANNE && $_thisRunNECROSIS.indexOf(aRelonsToProcess[r].to_s)!=-1) {
                if (MEM._localRunDelete.indexOf(aRelonsToProcess[r].txid)===1) MEM._localRunDelete.push(aRelonsToProcess[r].txid);
                var obj = $relonsMap.get(aRelonsToProcess[r].txid);
                if (obj!=null) {
                    obj.idb_insert=false;
                    obj.idb_update=false;
                    obj.idb_delete=true;
                    obj.active=false;
                    obj.bad_data=true;
                }
                continue; 
            }

            else if ($_thisRunNECROSIS.indexOf(aRelonsToProcess[r].reln)!=-1) {
                if (MEM._localRunDelete.indexOf(aRelonsToProcess[r].txid)===1) MEM._localRunDelete.push(aRelonsToProcess[r].txid)
                var obj = $relonsMap.get(aRelonsToProcess[r].txid);
                if (obj!=null) {
                    obj.idb_insert=false;
                    obj.idb_update=false;
                    obj.idb_delete=true;
                    obj.active=false;
                    obj.bad_data=true;
                }    
                continue;     
            }

            else if ($_thisRunNECROSIS.indexOf(aRelonsToProcess[r].from_nid)!=-1) {
                if (MEM._localRunDelete.indexOf(aRelonsToProcess[r].txid)===1) MEM._localRunDelete.push(aRelonsToProcess[r].txid)
                var obj = $relonsMap.get(aRelonsToProcess[r].txid);
                if (obj!=null) {
                    obj.idb_insert=false;
                    obj.idb_update=false;
                    obj.idb_delete=true;
                    obj.active=false;
                    obj.bad_data=true;
                }
                continue;        
            }
        }

        //the idb relon we currently have in IDB is in our list of relons we are about to process.
        var obj = $relonsMap.get(aRelonsToProcess[r].txid);

        //we have it already in IDB mem..
        // && (obj.force_update==undefined || obj.force_update==false)
        if (obj!=null && obj.bMUSTINSERT==undefined) {
            
            obj.idb_insert = false;
            obj.idb_update = true;

            //remove from the .arr array --- WE HAVE THIS IN IDB MEMORY.
            $relonsMap.remove(aRelonsToProcess[r].txid, "txid");
            $relonsMap.arr_other.push(obj);

             //unconf shoudl be -1 now (NOT 0);

            if (obj.time!=undefined && (obj.temp==undefined || obj.temp==false)) {
                let time = parseInt(obj.time);
                if (time > HISTORY._GREATEST_LAST_KNOWN_BLOCK) HISTORY.updateGLKB_fromRelons(time);
            }

            if (obj.bMUSTINSERT!=undefined) {
                obj.dupe=false;
                obj.temp=false;
                obj.idb_insert=true;
                obj.idb_update=false
                delete obj.bMUSTINSERT;
                continue;
            }

            var dupe_robj_in_RAM = ANNE.MEMG.lookupByTXID(obj.txid);
            var chgs=0;


            //relon in mem is a temp, but we have a REAL relon now.
            if (dupe_robj_in_RAM!=null && dupe_robj_in_RAM.temp==true && obj.ghost==false) {
                if (obj.time!=undefined && obj.time==-1) {
                    if (obj.txid[0]=="z") {
                        obj.ghost=true;
                    }
                    if (bNESTED==false && ANNE.MEM_UDF.bGETDEDSATBOOTUP==true) {
                        _memPoolBootups.push(rObj.txid);
                    }
                }
                else {
                    dupe_robj_in_RAM.temp=false;
                }
            }
            

            //parent is different than what we have in IDB
            if ((obj.parent!=undefined && obj.parent!=null)
                && ((aRelonsToProcess[r].parent==undefined || aRelonsToProcess[r].parent==null)
                || obj.time!=aRelonsToProcess[r].parent)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.parent = obj.parent;
                chgs+=1;
            }

            //time is different than what we have in IDB
            if ((obj.time!=undefined && obj.time!=null)
                && ((aRelonsToProcess[r].time==undefined || aRelonsToProcess[r].time==null)
                || obj.time!=aRelonsToProcess[r].time)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.time = obj.time;
                chgs+=1;
            }
            

            //sponsored_by is different than what we have in IDB
            if ((obj.sponsored_by!=undefined && obj.sponsored_by!=null)
                &&  ((aRelonsToProcess[r].sponsored_by==undefined || aRelonsToProcess[r].sponsored_by==null)
                || obj.sponsored_by!=aRelonsToProcess[r].sponsored_by)) {
                if (dupe_robj_in_RAM!=null) {
                    if (obj.from_nidTNOBJ==null) obj.from_nidTNOBJ = MEMG.getOrMakeIfNew(obj.from_nid);
                    dupe_robj_in_RAM.sponsored_by = obj.sponsored_by;
                    if (dupe_robj_in_RAM.sponsored_by!=undefined && dupe_robj_in_RAM.sponsored_by!="") {
                        let oldSP = MEMG.getOrMakeIfNew(dupe_robj_in_RAM.sponsored_by);
                        if (oldSP) obj.from_nidTNOBJ._removeSponsoredBy(oldSP.temp_nid);
                    }
                }
                chgs+=1;
            }

            //to_s is different than what we have in IDB
            if ((obj.to_s!=undefined && obj.to_s!=null)
                &&  ((aRelonsToProcess[r].to_s==undefined || aRelonsToProcess[r].to_s==null)
                || obj.to_s!=aRelonsToProcess[r].to_s)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.to_s = obj.to_s;
                    if (obj.to_sTNOBJ==null) obj.to_sTNOBJ = MEMG.getNeuronByNid(obj.to_s);
                    if (obj.to_sTNOBJ!=null) {
                        obj.t = obj.to_sTNOBJ.temp_nid;
                        dupe_robj_in_RAM.t = obj.to_sTNOBJ.temp_nid;
                    }        
                }
                chgs+=1;
            }


            //the to_s will change on a vote (we create the annode vote relon before they vote)..
            //we tie it to the unvoted to_s but also the ref_txid pts to their anjucator relon
            //in the source neuron.  That sys_ref_txid is updated to their actual on-chain vote
            //when they do infact vote.
            if (dupe_robj_in_RAM!=null) {
                if ((dupe_robj_in_RAM.to_s==MEM.anj_vote_hasnot.nid) && ((obj.to_s!=MEM.anj_vote_hasnot.nid))) {      
                    dupe_robj_in_RAM.to_s=obj.to_s;
                    if (obj.to_sTNOBJ==null) obj.to_sTNOBJ = MEMG.getNeuronByNid(rObj.to_s);
                    if (obj.to_sTNOBJ!=null) {
                        dupe_robj_in_RAM.t = obj.to_sTNOBJ.temp_nid;
                        obj.t = obj.to_sTNOBJ.temp_nid;
                    }
                    chgs+=1;
                }
            }


            //owned_by is different than what we have in IDB
            if ((obj.owned_by!=undefined && obj.owned_by!=null)
                &&  ((aRelonsToProcess[r].owned_by==undefined || aRelonsToProcess[r].owned_by==null)
                || obj.owned_by!=aRelonsToProcess[r].owned_by)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.owned_by = obj.owned_by;
                    if (obj.from_nidTNOBJ==null) obj.from_nidTNOBJ = MEMG.getOrMakeIfNew(obj.from_nid);
                        if (dupe_robj_in_RAM.owned_by!=undefined && dupe_robj_in_RAM.owned_by!="") {
                            let oldOw = MEMG.getOrMakeIfNew(dupe_robj_in_RAM.owned_by);
                            if (oldOw) obj.from_nidTNOBJ._removeOwns(oldOw.temp_nid);
                        }
                        dupe_robj_in_RAM.owned_by=obj.owned_by;
                        obj.from_nidTNOBJ.owned_by=obj.owned_by;
                    }   
                chgs+=1;
            }

            //relon_sponsor is different than what we have in IDB
            if ((obj.relon_sponsor!=undefined && obj.relon_sponsor!=null)
                &&  ((aRelonsToProcess[r].relon_sponsor==undefined || aRelonsToProcess[r].relon_sponsor==null)
                || obj.relon_sponsor!=aRelonsToProcess[r].relon_sponsor)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.relon_sponsor = obj.relon_sponsor;
                    dupe_robj_in_RAM.relon_sponsor=obj.relon_sponsor;
                    if (rs.indexOf(dupe_robj_in_RAM.relon_sponsor)==-1) {
                        ANNE.MEMG.getOrMakeIfNew(dupe_robj_in_RAM.relon_sponsor, "", NTYPE_UNKNOWN, 0);
                        rs.push(dupe_robj_in_RAM.relon_sponsor);
                    }
                    //add to our tracking array.
                    if (ANNE.USER._userNObj!=null && obj.relon_sponsor==ANNE.USER._userNObj.nid) {
                        ANNE.USER._userNObj.addUsAsRelonSponsor(obj);
                    }
                }
                chgs+=1;
            }

            //first seen is different than what we have in IDB
            if ((obj.firstseen!=undefined && obj.firstseen!=null)
                &&  ((aRelonsToProcess[r].firstseen==undefined || aRelonsToProcess[r].firstseen==null)
                || obj.firstseen!=aRelonsToProcess[r].firstseen)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.firstseen = obj.firstseen;
                chgs+=1;
            }

            //annies are different than what we have in IDB
            if ((obj.annies!=undefined && obj.annies!=null)
                &&  ((aRelonsToProcess[r].annies==undefined || aRelonsToProcess[r].annies==null)
                || obj.annies!=aRelonsToProcess[r].annies)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.annies = obj.annies;
                chgs+=1;
            }

            //freeze is different than what we have in IDB
            if ((obj.freeze!=undefined && obj.freeze!=null)
                &&  ((aRelonsToProcess[r].freeze==undefined || aRelonsToProcess[r].freeze==null)
                || obj.freeze!=aRelonsToProcess[r].freeze)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.freeze = obj.freeze;
                chgs+=1;
            }

            //sys ref txid is different than what we have in IDB
            if ((obj.sys_ref_txid!=undefined && obj.sys_ref_txid!=null)
                &&  ((aRelonsToProcess[r].sys_ref_txid==undefined || aRelonsToProcess[r].sys_ref_txid==null)
                || obj.sys_ref_txid!=aRelonsToProcess[r].sys_ref_txid)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.sys_ref_txid = obj.sys_ref_txid;
                    dupe_robj_in_RAM.to_s = obj.to_s;
                    if (obj.to_sTNOBJ==null) obj.to_sTNOBJ = MEMG.getNeuronByNid(obj.to_s);
                    if (obj.to_sTNOBJ!=null) {
                        dupe_robj_in_RAM.t = obj.to_sTNOBJ.temp_nid;
                        obj.t = obj.to_sTNOBJ.temp_nid;
                    }
                }
                chgs+=1;
            }

            if (chgs>0 && dupe_robj_in_RAM!=null || (obj.force_update!=undefined && obj.force_update==true)) {
                obj.dupe_robj_in_mem = dupe_robj_in_RAM;
                obj.idb_update = true;
                obj.dupe=true;
                delete obj.force_update;
                continue;
            }

            if (obj.relon_sponsor!=undefined) {
                if (ANNE.USER._userNObj!=null && obj.relon_sponsor==ANNE.USER._userNObj.nid) {
                    ANNE.USER._userNObj.addUsAsRelonSponsor(obj);
                }
            }

            obj.idb_update = false;
            continue;
        }
    }
    return;
}


/**
 * @param {*} $phObj 
 * @param {*} $nid 
 * @returns 
 */
IDB.getAllPayments = async($phObj=null, $nid) =>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { 
        console.log("skip IDB, all loaded already"); return []; 
    }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons_payment");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid && aRelonsToProcess[r].to_s!=$nid) continue;
        if (aRelonsToProcess[r].type!=TYPE_INTERNAL_PAY && aRelonsToProcess[r].type!=TYPE_BROADCAST_PAY) {
            if ((aRelonsToProcess[r].annies==undefined || aRelonsToProcess[r].annies==null || aRelonsToProcess[r].annies==0)) continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons_payment", "IDB.getAllPayments");
    return aRelons2;
}



/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOuts_not = async($phObj=null, $nid, $types)=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) continue;

        //we WANT outs NOT in the $types array.  Skip those in the types array.
        if ($types.indexOf(aRelonsToProcess[r].type)!=-1) continue;
        
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

   
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.qqueryForOuts_not");
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nids 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOutsFroms_not = async($phObj=null, $nids, $types=[])=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        }
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)!=-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOutsFroms_not");	
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOuts=async($phObj=null, $nid, $types=[])=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            continue;
        }
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons2=[];
    if (aRelons.length) {
        aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOuts");	
    }
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForIns=async($phObj=null, $nid)=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    var beFroms = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].to_s!=$nid) {
            continue;
        }
        aRelons.push(aRelonsToProcess[r]);
        
        //get from nids (we atleast need names in RAM)
        if (beFroms.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            beFroms.push(aRelonsToProcess[r].from_nid)
        }
    }

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].type!=TYPE_BE) continue;
        if (beFroms.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2=[];
    if (aRelons.length) {
        aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForIns");	
    }
    return aRelons2.concat(aRelons);
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOutsAndIns=async($phObj=null, $nid, $types=[])=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            if (aRelonsToProcess[r].to_s!=$nid) {
                continue;
            }
        }
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOutsAndIns");	
    return aRelons2;
}


IDB.queryForOutsAndInsFroms=async($phObj=null, $nids, $types=[])=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            if ($nids.indexOf(aRelonsToProcess[r].to_s)==-1) {
                continue;
            }
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "queryForOutsAndForInsFroms");	
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $bem_nids 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOutsAndIns_BEM=async($phObj=null, $bem_nids, $types=[])=>{
    if ($bem_nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    if ($bem_nids.length) {
        var aRelonsToProcess = await DB.loadAll("relons");
        var aRelons = [];
        for (var r=0; r < aRelonsToProcess.length; r++) {
            if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
            if ($bem_nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
            aRelons.push(aRelonsToProcess[r]);
        }

        if (aRelons.length==0) return [];
        aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOutsAndIns_BEM");
        return aRelons;
    }
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $relns 
 * @returns 
 */
IDB.queryTosFromRelns=async($phObj=null, $relns)=>{
    if ($relns==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    var nids = [];

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].to_type!=TO_TYPE_ANNE) continue;
        if ($relns.indexOf(aRelonsToProcess[r].reln)==-1) {
            if ($relns.indexOf(aRelonsToProcess[r].to_s)==-1 || aRelonsToProcess[r].reln!=MEM.ec_is.nid) {
                continue;
            }
            nids.push(aRelonsToProcess[r].to_s);
        }
        else {
            nids.push(aRelonsToProcess[r].from_nid);
        }

        if (nids.indexOf(aRelonsToProcess[r].reln)!=-1) continue;
        
    }

    //get all those outs.
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].type==TYPE_AWARENESS) continue;
        if (nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryTosFromRelons");
    return aRelons;
}


/**
 * This only gets the direct parent..
 * Not BEM and not the creative alt-be parent lines.
 * @param {*} $phObj 
 * @param {*} $parent_nid 
 * @returns 
 */
IDB.getChldren=async($phObj=null, $parent_nid, $ntype=null)=>{
    if ($parent_nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($ntype!=null && aRelonsToProcess[r].ntype!=$ntype) continue;
        if (aRelonsToProcess[r].type!=TYPE_BE) continue;
        if (aRelonsToProcess[r].to_type!=TO_TYPE_ANNE) continue;
        if (aRelonsToProcess[r].to_s!=$parent_nid) continue;
        if (aRelonsToProcess[r].r!=MEM.inst.temp_nid && aRelonsToProcess[r].r!=MEM.class.temp_nid) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.getChildren");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $bem_nids 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOuts_BEM=async($phObj=null, $bem_nids, $types=[])=>{
    if ($bem_nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    if ($bem_nids.length==0) return [];

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        if ($bem_nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOuts_BEM");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @returns 
 */
IDB.queryForCircuitOutsAndIns=async($phObj=null, $nid)=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            if (aRelonsToProcess[r].to_s!=$nid )continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForCircuitOutsAndIns");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $text 
 * @param {*} $class_nid 
 * @param {*} $wvs 
 * @returns 
 */
IDB.queryForText_BEM=async($phObj=null, $text, $class_nid, $wvs)=>{
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var lowerText = $text.toLowerCase();
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    var classNOBJ = MEMG.getNeuronByNid($class_nid);
    if (classNOBJ==null) return [];

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($wvs.length && $wvs.indexOf(aRelonsToProcess[r].owned_by)==-1) continue;
        
        if (aRelonsToProcess[r].to_type==TO_TYPE_STRING && aRelonsToProcess[r].to_s.toLowerCase()!=lowerText) continue;
        if (aRelonsToProcess[r].text!=undefined && aRelonsToProcess[r].text.toLowerCase()!=lowerText) continue;
         
        if (MEMG.isBesOfs(classNOBJ, aRelonsToProcess[r].bes)==false) continue;
        if (classNOBJ.is()==false) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForTExt_BEM");
    return aRelons;
}


/**
 * 
 * @param {*} $text 
 * @param {*} $phObj 
 * @param {Boolean} $bEXACT
 * @returns 
 */
IDB.queryForText=async($text, $phObj=null, $bEXACT=true)=>{
    if ($text==undefined || $text==null || typeof $text!=="string") {
        console.warn("fix this .. passed no or wrong var to IDB.queryForText");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj==null) {
        $phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForText", $text+$bEXACT+"");
        $phObj.bDIDIDB=true;
    }
    else if ($phObj.bDIDIDB==true) return [];
   
    var lowerText = $text.toLowerCase();
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];

    //this will get relons of names and aliases from IDB.  
    if ($bEXACT==true) {
        for (var r=0; r < aRelonsToProcess.length; r++) {
            if (aRelonsToProcess[r].to_type==TO_TYPE_STRING) {
                if(aRelonsToProcess[r].to_s.toLowerCase()==lowerText) {
                    aRelons.push(aRelonsToProcess[r]);
                    continue;
                }
            }
            else if (aRelonsToProcess[r].text!=undefined && aRelonsToProcess[r].text.toLowerCase()==lowerText) {
                aRelons.push(aRelonsToProcess[r]);
                continue;
            }
        }
    }

    else {
        for (var r=0; r < aRelonsToProcess.length; r++) {
            if (aRelonsToProcess[r].to_type==TO_TYPE_STRING) {
                if(aRelonsToProcess[r].to_s.toLowerCase().indexOf(lowerText)!=-1) {
                    aRelons.push(aRelonsToProcess[r]);
                    continue;
                }
            }
            else if (aRelonsToProcess[r].text!=undefined && aRelonsToProcess[r].text.toLowerCase().indexOf(lowerText)!=-1) {
                aRelons.push(aRelonsToProcess[r]);
                continue;
            }
        }
    }

    if (aRelons.length==0) return [];
 
    if (aRelons.length>1000) {
        console.log("audit spot check this..");
        console.log("should '"+$text+"' be returning so many variables?");
    }
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForText");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $text 
 * @param {*} $be_nid 
 * @param {*} $aWVs 
 * @returns 
 */
IDB.queryForTextAndBe=async($phObj=null, $text, $be_nid, $WVs)=>{
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var lowerText = $text.toLowerCase();
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].be!=$be_nid) continue;
        if (aRelonsToProcess[r].from_text.toLowerCase()!=lowerText) continue;
        if ($WVs.length && $WVs.indexOf(aRelonsToProcess[r].owned_by)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForTextAndBe");
    return aRelons;
}


IDB.queryForNid=async($phObj=null, $nid, $types=[])=>{
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    return await IDB.lookupNid($phObj=null, $nid, $types);
}

IDB.getNid=async($phObj=null, $nid, $types=[])=>{
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    return await IDB.lookupNid($phObj=null, $nid, $types);
}

/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.lookupNid=async($phObj=null, $nid, $types=[])=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    if ($phObj!=null) {
        if($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            continue;
        }
        if ($types.length && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) {
        return [];
    }

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNid");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nids 
 * @param {*} $types 
 * @param {*} $sysMsgRelns 
 * @returns 
 */
IDB.lookupNidsAndSysMsgRelns=async($phObj=null, $nids, $types=[], $sysMsgRelns)=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        if ($sysMsgRelns.length>0 && $sysMsgRelns.indexOf(aRelonsToProcess[r])==-1) continue;
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNidsAndSysMsgRelns");
    return aRelons;
}



//ridiculousness.
IDB.lookupNids_fast=async($nObjs, $types=[])=> {
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { 
        console.log("skip IDB, all loaded already"); 
        return []; 
    }
    var nips = new arraymap("n");
    for (var n=0; n < $nObjs.length; n++) {
        var map2 = IDB.map.getU($nObjs[n].nid);
        if (map2==undefined) {
            map2 = {"nid":$nObjs[n].nid, im:new arraymap("f")};
            map2.im.addU({f:"lookupNids_fast", t:$types});
            IDB.map.addU(map2);
            nips.addU({n:$nObjs[n].nid, t:$types});
        }
        else {
            var rec = map2.im.get("lookupNids_fast");
            if (rec!=null) {
                if (rec.t.length==0) continue;
                if ($types.length==0) { //just get all finally.
                    nips.addU({n:$nObjs[n].nid, t:[]});
                    continue;
                }
                var dones = rec.t;
                var typestodo = [];
                for (var t=0; t < $types.length; t++) {
                    if (dones.indexOf($types[t])==-1) {
                        typestodo.push($types[t]);
                        dones.push($types[t]);
                    }
                }
                if (typestodo.length==0) continue;
            }
            nips.addU({n:$nObjs[n].nid, t:typestodo});
        }
    }
    if (nips.arr.length==0) return [];
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        var nip = nips.get(aRelonsToProcess[r].from_nid);
        if (nip==null) continue;
        if (nip.t.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }
    if (aRelons.length==0) return [];
    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {array} $nids 
 * @param {array} $types
 * @param {string} $from_msg
 * @returns 
 */
IDB.lookupNids=async($phObj=null, $nids, $types=[], $from_msg="")=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { 
        console.log("skip IDB, all loaded already"); return []; 
    }

    if($phObj==null) {
        $phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", $nids.join("")+$types.join(""));
    }
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        };
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNids "+$from_msg);
    return aRelons;
}


/** 
 * 
 * @param {*} $phObj 
 * @param {*} $nids 
 * @returns 
 */
IDB.lookupNidsGetAllOuts=async($phObj=null, $nids)=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj==null) $phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNidsGetAllOuts ", $nids);
		
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNidsGetAllOuts");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $type 
 * @param {*} $reln 
 * @param {*} $to 
 * @param {*} $to_type 
 * @returns 
 */
IDB.queryForInsWithTo=async($phObj=null, $type, $reln, $to, $to_type)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].to_s!=$to) continue;
        if (aRelonsToProcess[r].reln!=$reln) continue;
        if (aRelonsToProcess[r].type!=$type) continue;
        if (aRelonsToProcess[r].to_type!=$to_type) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
  
    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForInsWithTo");
    return aRelons;
}


/**
 * Targetted RAM load of specific nids and outs.
 * In our optimized pre-load we only take the firstseen X number of relons.
 * Any given app/use case may need to preload specific nid classes that come later.
 * @param {Array} $nids 
 */
IDB.loadIDBNidOuts=async($nids)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }
    
    if (aRelons.length==0) return [];

    
    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidOuts");
    return aRelons;
}


/**
 * 
 * @param {*} $txids 
 */
IDB.lookupTxids = async($txids)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($txids.indexOf(aRelonsToProcess[r].txid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
        if ($txids.length==1) break;    //mini-time save, we have what we came for.
    }
    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupTxids");
}


/**
 * 
 * @param {*} $nid 
 * @param {*} $skip_type 
 */
IDB.loadIDBNidOuts_single=async($nid, $skip_type=TYPE_AWARENESS)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].type==$skip_type) continue;
        if (aRelonsToProcess[r].from_nid!=$nid) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidOuts_single");
}



/**
 * Targeted load of specific ins to nids     
 * In our optimized pre-load we only take the firstseen X number of relons.
 * Any given app/use case may need to preload specific nid classes that come later.
 * @param {Array} $nids 
 */
IDB.loadIDBNidIns=async($nids)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].to_s==-1)) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidIns");
}

IDB.loadIDBNidIns_single=async($nid)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.loadIDBNidIns_single ", $nid);
    if (phObj!=null) {
        if (phObj.bDIDIDB==true) return [];
        phObj.bDIDIDB=true;
    }
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nid!=aRelonsToProcess[r].to_s) continue;
        aRelons.push(aRelonsToProcess[r]);
    }
    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidIns");
}


/**
 * 
 * @param {*} $nid 
 */
IDB.testForNid=async($nid)=> {
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var yep = [];
    var rels = await DB.loadAll("relons");
    for (var r=0; r < rels.length; r++) {
        if (rels[r].from_nid==$nid) {
            yep.push(rels[r]);
        }
    }
    console.log("in relons table", yep);
    ANNE.TALK.printRelons(yep);
}


/**
 * Fix IDB memory after finding out run-time (in TX broadcast)
 * that a neuron has been updated with a owner/sponsor.
 * 
 * This comes from a broadcast hanai error message that informs us the 
 * neuron is already owned by ANNE.  (meaning a previous hanai already went through and the GUI is wrong)
 * 
 * @param {*} $nid 
 * @param {*} $owned_by 
 * @param {*} $sponsored_by 
 * @returns 
 */
IDB.updateAllOwnedBySponsoredBy=async($nid, $owned_by, $sponsored_by)=> {

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) continue;

        if (aRelonsToProcess[r].owned_by==undefined 
            || aRelonsToProcess[r].owned_by==null 
            || aRelonsToProcess[r].owned_by=="") continue;
        
        aRelonsToProcess[r].owned_by = $owned_by;
        aRelonsToProcess[r].sponsored_by = $sponsored_by;
        aRelons.push(aRelonsToProcess[r]);
    }

    await ANNE.DB.___updateAll(aRelons);
    return;
}

/************** (from file ./js/db/IDB.js)**/ 

//used for collabs and ghosts.
class history_cg  {
    constructor(){
        this._aHistory = []; 
        this.pullid = 1; 
        this._lastMyTxHeight = 0;   
        this._lastTxIndex = 0;
        this.my_last_block_index = 0;
    }

    getNextPullID = function () {
        var pullid = this.pullid;
        this.pullid +=1;
        return pullid;
    }

    /**                                                  
     * 
     * @author {Mr.Scatman}
     */
    processPullHistory = function ($aFullPullHistory=null) {
        var maxpullid = 0;
        if ($aFullPullHistory!=undefined && $aFullPullHistory!=null) {
            this._aHistory = $aFullPullHistory;
            for (var h=0; h < this._aHistory.length; h++) {
                var hObj = this._aHistory[h];
                if (hObj.pullid > maxpullid) maxpullid = hObj.pullid;
                hObj.bDIDIDB = false;
            }
        }

        this.pullid = maxpullid + 1;
    }

    
    /**
     * For functions where we get the latest height in the API return
     * we can manually update this without looping/checking
     * Much cleaner.
     */
     manuallyUpdatePullHistory=async($phObj, $lastBlock=null)=>{
        if ($lastBlock==null) $lastBlock = ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK+1;
        $phObj.lastBlock = $lastBlock;

        if ($phObj.pullid==0) {
            $phObj.pullid = HISTORY.getNextPullID();    
            console.log("%c ((inserting new history record with pullid:"+$phObj.pullid+" "+$phObj.functionName+" target:"+$phObj.functionName+"))", 'background: white; color: green');
            var done = await ANNE.DB_local.___update($phObj, "fullPullHistory");   
        }

        //if we had new info, update it.
        else {
            var done = ANNE.DB_local.___update($phObj, "fullPullHistory");    //update the localDB record
        }
    }

    

    /**
     * Get the last block we found results using this exact query.
     * 
     * Try to save processing/pull attempts by looking for this pull obj.  
     * We will still try to pull from the blockchain but with a min block  
     * number from the last known spot we already pulled in data for this  
     * type of pull.                                                       
     * 
     * @author {Mr.Scatman}
     */
    getLastBlockFromPullHistory = function($prefix, $functionName, $target, $extra=null) {
        var extra="";
        if ($extra!=null) extra=$extra;

        if (this._aHistory.length==0) {
            console.log("%c NO HISTORY ---- U SHOULD SEE THIS ONCE for "+$functionName+" target:"+$target+" "+$extra, 'background: white; color: red');
            return this.__makeDefaultPullHistoryObj($prefix, $functionName, $target, $extra);
        }

        for (var h=0; h < this._aHistory.length; h++) {
            var phObj = this._aHistory[h];
            if (phObj.prefix!=$prefix || phObj.functionName!=$functionName || phObj.target!=$target) continue;
            if (extra!="") {
                if (phObj.extra!=$extra) {
                    continue;
                }
            }
            console.log("%c history - we found a pull history object for function: "+$functionName+" target:"+$target, 'background: green; color: white');
            console.log(phObj);
            return phObj;
        }

        //we had history in our  history object, but not for THIS entry.
        //console.warn("temp new HISTORY object (will save if we find results) for ", $prefix, $functionName, $target);
        return this.__makeDefaultPullHistoryObj($prefix, $functionName, $target, $extra);
    }


    /**
     * IS THIS BEING USED????
     * 
     * We just need to get the highest height so we know
     * how to optimize our transaction pulls on the ANNE ID page.
     * 
     * We will do pulls to get the TXs list anytime we need it.
     * We do not need it hanging around in memory as we already have
     * a high enough local-RAM mem burden with relons.
     */
    preProcessStoredConfirmedTXs = function($stored_txs){
        for (var c=0; c < $stored_txs.length; c++) {
            if ($stored_txs[c].height > this._lastMyTxHeight) this._lastMyTxHeight=$stored_txs[c].height;
        }
        this._lastTxIndex = $stored_txs.length;
    }




    /**
     * DEPRECATED
     * we should emulate our transaction format so we can
     * process mined blocks in the payouts display with the same code.
     */
    processBlocks = function($my_blocks) {
        console.log("%c processBlocks "+$my_blocks.length, TALK.bgredbig);
        console.log($my_blocks);
        
        //calc actual block reward based on % via block number algo.

        var blocks = [];  //timestamp, scoopnum ?
        for (var b=0; b < $my_blocks.length; b++) {
            var block = $my_blocks[b];
            var emulated = {};
            emulated.blockid = "blockid_"+block.block;
            emulated.txid = "";
            emulated.type = "block";
            emulated.annies = parseInt(block.blockRewardNQT);
            emulated.fee = 0;
            emulated.height = block.height;
            emulated.confirmations = 1;
            
            emulated.miningObj = ANNE.reward(emulated.annies, block.height, block.totalFeeNQT);
            //console.log(emulated.miningObj);

            emulated.net_chg = parseInt(emulated.miningObj.mining_reward)+parseInt(emulated.miningObj.fees);

            emulated.anne_app_user = ANNE.USER._userNObj.nid;
            emulated.recipient = ANNE.USER._userNObj.nid;
            emulated.sender=null;
            emulated.downs = 0;
            emulated.sent_tx=false;
            emulated.attach=false;

            blocks.push(emulated);
        }

        if ($my_blocks.length) {
            USER.bMINER=true;
            USER._blocksMined = blocks;
        }

        this.my_last_block_index = 0;
        return blocks;
    }

    /**
     * This will preset our optmizer value for height
     * so we can properly filter (start index) the TXs when pulling them from the ANNODE
     */
    getAndProcessOURTXS_bootup=async()=>{
        var ourTXs = await this.getAndProcessOURTXS();
        this.preProcessStoredConfirmedTXs(ourTXs);
    }


    /**
     */
     getAndProcessOURTXS = async()=> {
        var aTXS =  await DB_local.__getStoredTxs();
        if (aTXS==null || aTXS.length==0) return [];
        var ours = [];
        for (var t=0; t< aTXS.length; t++) {
            if (aTXS[t].anne_app_user!=USER._userNObj.nid) continue;
            ours.push(aTXS[t]);
        }
        return ours;
    }


    /**
     * @author {Mr.Scatman}
     */ 
    __makeDefaultPullHistoryObj=function($prefix, $functionName, $target, $extra=null) {
        var phObj = new Object();
        phObj.prefix = $prefix;
        phObj.functionName = $functionName;
        phObj.target = $target;
        phObj.extra = $extra;
        phObj.lastBlock = ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK;
        phObj.pullid = 0;               //we will set this to the next one in line if we are going to insert the record
        phObj.bDONETHISSESSION = false; //for ANNE API optimzation
        phObj.bDIDIDB = false;          //for RAM optimization.
        this._aHistory.push(phObj);
        return phObj;
    }
}

const HISTORY_LOCAL = new history_cg ();

/************** (from file ./js/db/history_local.js)**/ 

class govenor {
    constructor($nObj) {
        this.target_nObj=$nObj;
        this.dr_restrict_reln_tnids=[];  //delegated params restricts.        
        this.target_ntype = $nObj.ntype;
        this.action=govenor.COLLAB_REQ_NO_COLLAB_RIGHTS;
        this.bBADGARDEN = false;
        this._applied_for_identity_nid = 0;

        //set with respect to the identity in question
        this.bANDY = false;             //can approve collab request
        this.bANDY_SUPER = false;       //can set new admins and approve any type of collab request.
        this.hanai_admin=false;         //can appprove hanai of instances of this neuron.
        
        //tallied for *all* andy admins on this neuron (not just the identity in question)
        this.andy_admins = [];      //an array of objects with .user_nid and .andy_right
        this.hanai_admins = [];     //an array of NIDS

        this._target_reln_temp_nid=-1;
    }

    static COLLAB_REQ_AUTOREJECT = 0;
    static COLLAB_NOCOLLABIDENTITY = 1
    static COLLAB_REQ_NO_COLLAB_RIGHTS = 2;
    static COLLAB_REQ_OK = 4;                //collab request is ok, let it through
    static DELEGATE_NOW = 5;                 //delegation is approved
    static SELF_OWNED = 6; 
    static THIS_IS_YOU = 7;  
    static actionText = [   "No collab", 
                            "No collab on identities", 
                            "No collab rights", 
                            "[xx]", 
                            "You have collab rights", 
                            "You have delegation rights", 
                            "This is your neuron", 
                            "This is your identity"];

    static auditGT() {
        for (var n=0; n < MEM._aNeurons.length; n++) { 
            if (MEM._aNeurons[n]._gt!=null) { 
                var gtNOBJ = MEMG.getNeuronByNid(MEM._aNeurons[n]._gt.t);
                console.log(MEM._aNeurons[n].text, "** "+gtNOBJ.text);
            } 
        }
    }

    /**
     * No collab on sys msg, awareness, brain (firing), episodic or experience relons  
     * 
     * @param {int} - relon type of relon we are attempting to collab on
     * 
     * @author {Mr.Scatman}
     */
     static isCollabDimension ($type) {
        if ($type==TYPE_SYS) return true;     //sys dimension
        if ($type==TYPE_BE) return true;      //be dimension
        if ($type==TYPE_PARAM) return true;   //param dimension
        if ($type==TYPE_HAS) return true;     //has dimension
        if ($type==TYPE_ROLE) return true;    //points to a role inside of a clump
        if ($type==TYPE_CLUMP) return true;   //points to various clump types
        return false;
    }
        

    addHanaiAdmin=function($identity) {
        if (this.hanai_admins.indexOf($identity)==-1) {
            this.hanai_admins.push($identity);
        }
    }

    isHanaiAdmin=function($identity) {
        if (this.hanai_admins.indexOf($identity)!==-1) return true;
        return false;
    }
    


    /**
     * Apply the known garden rules to assess for collab and delegation.
     * 
     * The default garden allows for COLLAB requests and presumes  
     * communual requests are ok.

     * Default garden will still honor the PARAM RESTRICTIONS.
     * Default garden will DEFAULT to requiring request reln's to meet PARAM requirements.
     * This enables loose rules on who can attempt to collab but still 
     * narrows the desired relns.
     *
     * Default garden does NOT honor let or deny.  (since there is no gardne)
     * LET or DENY only work w/ specified garden type
     * Default garden DOES NOT allow for delegation.  
     * SET THE GARDEN TO TELL ANNE TO PAY ATTENTION TO DELEGATION and LET/DENY lists.
     *
     * @param {string} - nid long string of the identity we want to analzye this for.$identity_nid
     * @author {Mr.Scatman}
     * 
     */
    applygardenRulesForIdentity=async($identity_nid)=> {
        if (this._applied_for_identity_nid==$identity_nid) return;

        //reset identity specific vars
        this.dr_restrict_reln_tnids=[];
        this.bANDY = false;             
        this.hanai_admin=false;
        this.action = govenor.COLLAB_REQ_NO_COLLAB_RIGHTS;
        
        this._applied_for_identity_nid = $identity_nid;
        if (this.target_nObj.nid==$identity_nid) {
            this.action = govenor.THIS_IS_YOU;
            this.target_nObj.collab_ACTION = this.action;
            return;
        }


//check for them..
//if we've pulled or just created this.. 
//no need to tax the annode API.
//we have to have the 0/1 outs.
//await ANNE.CHASE.lookupMinNid(this.target_nObj.nid);
        
        //EXIT, no collab on identities.
        if (MEM.isIdentity(this.target_nObj)==true) {
            this.action = govenor.COLLAB_NOCOLLABIDENTITY;
            this.target_nObj.collab_ACTION = this.action;
            return;
        }

        
        //this is owned by you.  Fire away.
        if (this.target_nObj.owned_by==$identity_nid) {
            this.action = govenor.SELF_OWNED;
            this.target_nObj.collab_ACTION = this.action;

            //still want to set hanai rights flags to know if
            //the hanai will be auto approved (and properly display that)
            await this.analyzeDelegationRights($identity_nid);
            return;
        }
        

        await this.analyzeDelegationRights($identity_nid);

        //*************AUDITS **************//
        if (this.target_nObj.ntype==NTYPE_CLASS) {
            if (this.target_nObj._gt==null) {
                if (this.target_nObj.bes.length) {
                    var parentNOBJ = MEMG.getNeuronByTempNid(this.target_nObj.bes[0]);
                    if (parentNOBJ._gt!=undefined && parentNOBJ._gt!=null) {
                        //console.log("%cfix this, _gt should have been set on "+this.target_nObj.text, TALK.bgredbig);
                        this.target_nObj.setGtRelon(parentNOBJ._gt);
                    }
                }
            }
        }
        if (this.target_nObj._gti==null) {
            if (this.target_nObj.bes.length) {
                var parentNOBJ = MEMG.getNeuronByTempNid(this.target_nObj.bes[0]);
                if (parentNOBJ._gti!=undefined && parentNOBJ._gti!=null) {
                    //console.log("%cfix this, _gti should have been set on inst "+this.target_nObj.text, TALK.bgredbig);
                    this.target_nObj.setGtiRelon(parentNOBJ._gti);
                }
            }
        }
        //*************AUDITS **************//

        var GT = null;
        if (this.target_ntype==NTYPE_CLASS) GT = this.target_nObj._gt;
        else if (this.target_ntype==NTYPE_INST) GT = this.target_nObj._gti;
        else {
            console.log("no ntype on this neuron. Cannot collab/delegate");
            this.action = govenor.COLLAB_REQ_AUTOREJECT;
            this.target_nObj.collab_ACTION = this.action;
            return;
        }


        if (GT==null) {
            //only way to get collab req reject in no-garden type is for the class
            //to have a param restriction and this request doesn't meet it.
            console.log("** no garden type ** for ('"+this.target_nObj.text+"') applying defaults.  Collab ok.  Will check params restricts.");
            this.action = govenor.COLLAB_REQ_OK;

            if (this.bANDY==true) this.action = govenor.DELEGATE_NOW;

            this.target_nObj.collab_ACTION = this.action;
            return;
        }


        //Communal garden - Honors owner set delegation to allow collab approvals.  
        //Deny entry will OVERRIDE delegated rights
        //Defaults to BLACKLIST mode for denys (ignores those on the LET list).
        //Anyone can make collab relon requests.
        //Anyone can make keyless children (that they own)
        //gui - Neurons owned by the owner or anyone not in the deny list will be displayed.
        else if (GT.t==MEM.gt_communal.temp_nid) {

            //user has been denied.
            if (this.target_nObj._wv_deny!=null && this.target_nObj._wv_deny.length>0  && this.target_nObj._wv_deny.indexOf($identity_nid)!=-1) {
                this.action = govenor.COLLAB_REQ_AUTOREJECT;
                this.bANDY = false;
                console.log("INVALID: garden type communual, user in deny list.");
                this.target_nObj.collab_ACTION = govenor.COLLAB_REQ_AUTOREJECT;
                return;
            }

            else {
                //console.log("garden type communual, user not in deny list.");
                this.action = govenor.COLLAB_REQ_OK;
            }
        }

        
        //Owner collab - Honors owner set delegation to allow collab approvals.
        //Defaults to WHITELIST mode for LETS.
        //Only those listed in LETS can make collab relon request and keyless children (that they own).
        //gui - Only those neurons owned by the owner, or the LETS, will be displayed.
        else if (GT.t==MEM.gt_owner_collab.temp_nid) {

            //no lets, no one gets in.  sorry charley.
            if (this.target_nObj._wv_let==null || this.target_nObj._wv_let.length==0) {
                this.action = govenor.COLLAB_REQ_NO_COLLAB_RIGHTS;
                this.target_nObj.collab_ACTION = govenor.COLLAB_REQ_NO_COLLAB_RIGHTS;
                 console.log("INVALID: garden type owner collab, user not in let list.");
            }

            //user is not in let list.  Deny.
            else if (this.target_nObj._wv_let!=null && this.target_nObj._wv_let.length>0 
                && this._wv_let.indexOf(this.target_nObj.identity_nid)==-1) {
                this.action = govenor.COLLAB_REQ_NO_COLLAB_RIGHTS;
                this.target_nObj.collab_ACTION = govenor.COLLAB_REQ_NO_COLLAB_RIGHTS;
                console.log("INVALID: garden type owner collab, user not in let list.");
                return;
            }

            //user is in the let list.. so they have access to collab on this.
            else this.action = govenor.COLLAB_REQ_OK;
        }

        //Owner Only - No delegation, no LET, no DENY, noone can make collab relons,
        //and no one can make keyless children.
        //TO DO, enforce this on keyless children
        //gui - In guis only neurons created by the owner will be queried and displayed.
        else if (GT.t==MEM.gt_owner_only.temp_nid) {
            this.bANDY = false;        //to override any permitted params.
            this.action = govenor.COLLAB_REQ_AUTOREJECT;
            this.target_nObj.collab_ACTION = govenor.COLLAB_REQ_AUTOREJECT;    
        }
        else {
            console.log("unhandled garden type");
            this.bBADGARDEN = true;
            this.bANDY=false;
        }
        
        if (this.bANDY==true) {
            this.action = govenor.DELEGATE_NOW;
        }

        
        this.target_nObj.collab_ACTION = this.action;
        return;
    }


    /**
     * analyze delegation for this user considering this neuron class
     * and the DR rights set for it.
     * 
     * Garden type is not considered here.
     */
     analyzeDelegationRights=async($identity_nid)=> {

        //TO DO: check that we have chased 0/1 for both parent and child.
        //if not, chase it.parent_temp_nid

        if (this.target_nObj.delegation==null) this.target_nObj.delegation=[];

        var beline_dg_rights = MEMG.getBELINEDelegation(this.target_nObj);
        //TALK.printRelons(beline_dg_rights);
        
        var target_nid = this.target_nObj.nid;

        //now we have all delegation relons ready to analyze.
        for (var d=0; d < beline_dg_rights.length; d++) {
            var rObj = beline_dg_rights[d];
            var from_nid = rObj.from_nid;

            //we handle permitted dr relns (specific to user + neuron) in a clump syntax.
            //we snag these here tho we won't use them until analyzing a specific
            //relon for collab/delegation.
            if (rObj.type==TYPE_CLUMP) {
                var clumpNOBJ = MEMG.getNeuronByTempNid(rObj.t);
                if (clumpNOBJ==null) continue;
                if (clumpNOBJ.bPROCESSEDCLUMP==undefined) {
                    await clumpNOBJ.processActionNeuron();
                }
                if (clumpNOBJ.validateEquals('role_affected_thing', $identity_nid)==false) continue;
                if (clumpNOBJ.validateEquals('role_thing', rObj.from_nid)==false) continue;
                
                //only our permitted reln needs to use the full clump syntax.
                //the rest are single semantic relon shorthand clumps.  (a->b->c).
                //we use a clump when we need more than a basic subj/predicate/obj syntax.
                if (clumpNOBJ.validateEquals('role_action', MEM.dr_permitted_reln.nid)==false) continue;
                
                //for the permission clump the role_theme_thing will be the relns
                //that are approved for delegation
                var paramNOBJS = clumpNOBJ.extractRoles(MEM.role_theme_thing);

                //dr_restrict_reln_tnids
                for (var p=0; p < paramNOBJS.length; p++) {
                    var param = paramNOBJS[p];
                    if (this.dr_restrict_reln_tnids.indexOf(param.temp_nid)==-1) {
                        this.dr_restrict_reln_tnids.push(param.temp_nid);
                    }
                }
                continue;
            }
                     
            //for non super, the target neuron class must have the same owner
            //as the rights issuer.
            if (rObj.r==MEM.dr_andy_defn.temp_nid) {
                if (from_nid!=target_nid) continue;
                else if (rObj.to_s==$identity_nid) {
                    this.bANDY=true;
                }
                this.andy_admins.push( {andy_right:rObj.r, user_nid:rObj.to_s});
            }
            else if (rObj.r==MEM.dr_andy_super_all.temp_nid) {
                var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                if (fromNOBJ.owned_by!=this.target_nObj.owned_by) continue;
                else if (rObj.to_s==$identity_nid) {
                    this.bANDY=true;
                    this.bANDY_SUPER=true;
                }
                this.andy_admins.push( {andy_right:rObj.r, user_nid:rObj.to_s});
            }

            //new super super (all rights to the neuron in one go)
            else if (rObj.r==MEM.dr_super_super.temp_nid) {
                var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                this.addHanaiAdmin(rObj.to_s);

                this.andy_admins.push( {andy_right:rObj.r, user_nid:rObj.to_s});
                if (fromNOBJ.owned_by!=this.target_nObj.owned_by) {

                    //if this is a merge, and the neuron is an instance..
                    if (this._target_reln_temp_nid==MEM.CMD_MERGE_NEURON.temp_nid && this.target_nObj.ntype==NTYPE_INST) {
                        //var targetBE = MEMG.getNeuronByTempNid(this.target_nObj.bes[0]);

                        //the super has access to this.
                        if (fromNOBJ.owned_by!=this.target_nObj.owned_by && rObj.to_s==$identity_nid) {
                            this.bANDY=true;
                            this.bANDY_SUPER=true;
                            this.hanai_admin=true;
                        }
                    }
                    else {
                        if (fromNOBJ.owned_by!=this.target_nObj.owned_by && rObj.to_s==$identity_nid) {
                            this.bANDY=true;
                            this.bANDY_SUPER=true;
                            this.hanai_admin=true;
                        }
                    }
                    continue;
                }
                else if (rObj.to_s==$identity_nid) {
                    this.bANDY=true;
                    this.bANDY_SUPER=true;
                    this.hanai_admin=true;
                }
            }

            else if (this.target_ntype==NTYPE_CLASS) {
                //ANDY super class rights DO cascade in the protocol rules.
                //they cascade only to same-owned neurons as the class that granted
                //the rights.
                if (rObj.r==MEM.dr_andy_super_class.temp_nid) {
                    var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                    if (fromNOBJ.owned_by!=this.target_nObj.owned_by) continue;
                    else if (rObj.to_s==$identity_nid) this.bANDY=true;
                    this.andy_admins.push( {andy_right:rObj.r, user_nid:rObj.to_s});
                }
                //for hanai approvals.
                else if (rObj.r==MEM.dr_andy_hanai_in.temp_nid) {
                    this.addHanaiAdmin(rObj.to_s);
                    if (rObj.to_s==$identity_nid) this.hanai_admin=true;
                }
            }
            
            else if (this.target_ntype==NTYPE_INST) {
                if (rObj.r==MEM.dr_andy_super_insts.temp_nid) {
                    var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                    
                    if (fromNOBJ.owned_by!=this.target_nObj.owned_by) continue;
                    else if (rObj.to_s==$identity_nid) {
                        this.bANDY=true;
                    }
                    this.andy_admins.push( {andy_right:rObj.r, user_nid:rObj.to_s});
                }

                //for hanai approvals. (makes no sense for NTYPE_INST)
                else if (rObj.r==MEM.dr_andy_hanai_in.temp_nid) {
                    this.addHanaiAdmin(rObj.to_s);
                    if (rObj.to_s==$identity_nid) this.hanai_admin=true;
                }
            }
        }

        
    }


    /**
     * Quick check for a valid collab response reln
     */
    static isCollabRespReln = function($tnid_action_taken) {
        if (MEM.collabRespRelns.indexOf($tnid_action_taken)==-1) return false;
        return true;
    }


    /**
     * Is this a valid Collab response?
     * 
     * An admin has taken action.  Is the action valid to broadcast to the chain?
     * 
     */
    isValidCollabResponse=async($admin_nid, $tnid_action_taken, $target_reln_temp_nid)=> {
        if (this._applied_for_identity_nid==null) {
         alert("MUST APPLY GARDEN RULES FOR THIS USER, FIRST");
            await this.applygardenRulesForIdentity($admin_nid);
        }

        if (this._applied_for_identity_nid!=$admin_nid) {
            alert("garden set up for a different user.  This should not happen");
            return false;
        }

        //is the action a known collab response action?
        if (govenor.isCollabRespReln($tnid_action_taken)==false) {
            alert("not a valid collab action");
            return false;
        };

        //does this admin have access to delegate on this?
        if (this.action < govenor.DELEGATE_NOW) {
            console.log($admin_nid+" has no delegation rights to nid:'"+this.target_nObj.nid+"' "+this.target_nObj.text);
            return false;
        }

        //NOTE: move this to the top of this function once we vet the stuffs above it are working anyhow.
        if (this.action==govenor.SELF_OWNED || $admin_nid==this.target_nObj.owned_by) {
            console.log('ur the owner of this.. u can approve collab requests');
            return true;
        }

        if (this.dr_restrict_reln_tnids.length) {  //must be an array of temp_nid values
            var targetRelnNOBJ = MEMG.getNeuronByTempNid($target_reln_temp_nid);
            if (this.dr_restrict_reln_tnids.indexOf($target_reln_temp_nid)==-1) {
                console.log($admin_nid+" has delegation rights to nid:'"+this.target_nObj.nid+"' "+this.target_nObj.text+" but NOT to param "+targetRelnNOBJ.text);
                return false;
            }
            else {
                console.log($admin_nid+" has delegation rights to nid:'"+this.target_nObj.nid+"' "+this.target_nObj.text+", param "+targetRelnNOBJ.text);
                return true;
            }
        }
        
        console.log($admin_nid+" has delegation rights to nid:'"+this.target_nObj.nid+"' "+this.target_nObj.text+", no param restrictions set in the garden");
        return true;
    }


    /**
     * Get the summation of extra garden relns (param restrictions) for neuron
     * up its be line.
     * 
     * @author {Mr.Scatman}
     */
    getEgrsSum = function() {
        var egrs = this.target_nObj.egrs_tnids;
        if (egrs==undefined || egrs==null) egrs=[];
        var gens = 0;
        var parent_temp_nid = this.target_nObj.bes[0];

        //climb the be line and snag all egrs_tnids.
        while (gens++<15) {
            var pNOBJ = MEMG.getNeuronByTempNid(parent_temp_nid);
            if (pNOBJ==null) break;
            if (pNOBJ.egrs_tnids==undefined || pNOBJ.egrs_tnids==null || pNOBJ.egrs_tnids.length==0) continue;
            for (var p=0; p < pNOBJ.egrs_tnids.length; p++) {
                if (egrs.indexOf(pNOBJ.egrs_tnids[p])==-1) egrs.push(pNOBJ.egrs_tnids[p]);
            }
            if (pNOBJ.be_lvl==0) break;
        }
        //console.log(egrs);
        return egrs;
    }


    /**
     * Is this a valid Collab request?
     * 
     * With garden rules applied for this identity, assess if the collab request
     * is valid in consideration of the RELN being used in the collab.
     * 
     * returns a helpful TXB.TXTTYPE for generating the proper ANNE TRANSACTION.
     * 
     * @author {Mr.Scatman}
     * 
     */
    isValidCollabReq=async($identity_nid, $target_reln_temp_nid)=> {
        this._target_reln_temp_nid = $target_reln_temp_nid;
        
        if (this._applied_for_identity_nid==null || this._applied_for_identity_nid==0) {
            await this.applygardenRulesForIdentity($identity_nid);
        }
        if (this._applied_for_identity_nid!=$identity_nid) {
            alert("garden set up for a different user.  This should not happen");
            return 0;
        }

        if (this.action < govenor.COLLAB_REQ_OK) return 0;
        if (this.target_nObj.nid==USER._userNObj.nid) return TXB.TXTYPE_SELF_RELON;

        //move this to the top of this function once we vet the stuffs above it are working anyhow.
        if (this.action==govenor.SELF_OWNED || this.target_nObj.owned_by==$identity_nid) {
            return TXB.TXTYPE_CHILD_RELON;
        }
        if (this.action>=govenor.COLLAB_REQ_OK) {
            //we consider THIS neuron and its parent extra garden relns.
            //we do not inherit these from entire parental line.. that is a big 
            //burden on the ANNODE to validate.. There are ways to build hybrid solutions
            //to handle such graph-like queries at the protocol layer but I opt
            //to NOT go that road.. Garden rules are inherited ONE GENERATION...
            //if a class is sub-classed and that class is sub-classed then the garden
            //rules must be specified again.. (as they may be different anyhow)
            //and another admin may want to set different params.
            var egrs = this.getEgrsSum();

            //are restrictions set to delegating only certain params?
            if (egrs.length) {
                //we have egrs but this isn't one of them.
                if (egrs.indexOf($target_reln_temp_nid)==-1) return 0;
            }
        }

        //check if auto delegation is set for this. 
        //will inform the user ...
        if (this.action==govenor.DELEGATE_NOW) {

            //are there restrictions to what param are delegated?
            if (this.dr_restrict_reln_tnids.length) {
                
                //sorry, not this one.
                if (this.dr_restrict_reln_tnids.indexOf($target_reln_temp_nid)==-1) return TXB.TXTYPE_COLLAB_REQ;

                //good param! delegation possible!
                else return TXB.TXTYPE_COLLAB_REQ_EXP_DEL;
            }

            //delgation allowed, no restrictions on granted delegation rights to this.
            else return TXB.TXTYPE_COLLAB_REQ_EXP_DEL;
        }

        return TXB.TXTYPE_COLLAB_REQ;
    }


    /**
     * 
     */
    printRights=function() {
        for (var h =0; h < this.hanai_admins.length; h++) {
            var nObj = MEMG.getNeuronByNid(this.hanai_admins[h]);
            console.log("%c HANAI RIGHTS:%c"+nObj.bestNameHTML(), ANNE.TALK.bgpink, ANNE.TALK.bgyellow);
        }
        for (var a =0; a < this.andy_admins.length; a++) {
            var nObj = MEMG.getNeuronByNid(this.andy_admins[a].user_nid);
            var rightsNOBJ = MEMG.getNeuronByTempNid(this.andy_admins[a].andy_right);

            console.log("%c ANDY RIGHTS:%c"+rightsNOBJ.text+" %c"+nObj.bestNameHTML(), ANNE.TALK.bgblue, ANNE.TALK.bggreen, ANNE.TALK.bgyellow);
        }

        this.bANDY = false;             //can approve collab request
        this.bANDY_SUPER = false;       //can set new admins and approve any type of collab request.
        this.hanai_admin=false;         //can appprove hanai of instances of this neuron.
        
        console.log("bANDY:"+this.bANDY+" bANDY_SUPER:"+this.bANDY_SUPER+" hanai_admin:"+this.hanai_admin);
    }
}

/************** (from file ./js/base/govenor.js)**/ 


            //var lat = MEMG.getLat(this.nObj);
            //var long = MEMG.getLong(this.nObj);

class anne_geo {
    constructor($nObj) {
        this.nObj = $nObj;
        this.googleMap = null;
        this.lat =MEMG.getHasValue(MEM.gps_lat.temp_nid);
        this.long =MEMG.getHasValue(MEM.gps_long.temp_nid);
    }
}

class anne_role_mapper {
    constructor($cNOBJ) {
        this.conveyorNOBJ = $cNOBJ;
    }


    /**
     * 
     * @param {nlp_clump} $clump 
     * @param {string} $reln_part_nid 
     * @param {string} $debug 
     * @returns 
     */
    getMapperRoleNid = function($clump, $reln_part_nid, $debug="") {
        for (var r=0; r < this.conveyorNOBJ.outs.length; r++) {
            if (this.conveyorNOBJ.outs[r].to_s==$reln_part_nid) {
                var mapper_role_temp_nid=this.conveyorNOBJ.outs[r].r;
                var role_tnid =  MEM.role_map[mapper_role_temp_nid];
                if (role_tnid==undefined) return null;
                var auditNOBJ = MEMG.getNeuronByTempNid(role_tnid);
                console.log("(getMapperRoleNid) %c"+$debug+" %c"+auditNOBJ.text, ANNE.TALK.bgyellow, ANNE.TALK.bggreen);
                var roles = $clump.nlpRolesObj[role_tnid];
                if (roles!=undefined) {
                    var rolesValueNOBJ = roles[0].nObj;
                    if (rolesValueNOBJ==null) continue;
                    console.log("(getMapperRoleNid) roel value %c"+rolesValueNOBJ.text, ANNE.TALK.bggreen);
                    return rolesValueNOBJ;
                }
            }
        }
        return null;
    }


    /**
     * 
     * @param {nlp_clump} $clump 
     * @returns 
     */
    extrap=async($clump)=>{
        var defines = MEMG.getDefines(this.conveyorNOBJ);
        //(for die we would have deathday, and death place here)
        if (defines.length==0) return;

        //for cNOBJ die:  deathday, deathplace
        for (var d=0; d < defines.length; d++) {
            
            var dNOBJ = defines[d];
            console.log("conveyor %c"+this.conveyorNOBJ.text+" %c"+dNOBJ.text, ANNE.TALK.bgyellow, ANNE.TALK.bggreen);
            //deathday, role_when
            
            var dROLE = MEMG.getDefinedBy(dNOBJ);
            if (dROLE==null) continue;

            console.log("conveyor %c"+this.conveyorNOBJ.text+" %c"+dROLE.text, ANNE.TALK.bgyellow, ANNE.TALK.bggreen);
            
            //extract the right role  (ie, )
            if ($clump.nlpRolesObj[dROLE.temp_nid]!=undefined) {
                //var role_value = $clump.nlpRolesObj[dROLE.temp_nid].nObj;    //ie, the when..
                var fromNOBJ = this.getMapperRoleNid($clump, MEM.relon_from.nid, "mapping "+dNOBJ.text);
                if (fromNOBJ==null) continue;
    
                //not sure if we map the reln (ever??)
                //it should be the mapped object (dNOBJ)
                //ie, birthday, deathday, death place.. etc.
                //that is the reln of the TYPE_HAS.
                var relnNOBJ = this.getMapperRoleNid($clump, MEM.relon_reln.nid);
                if (relnNOBJ==null) relnNOBJ = dNOBJ;
                if (relnNOBJ==null) continue;
                
                var toNOBJ = this.getMapperRoleNid($clump, MEM.relon_to.nid);
                if (toNOBJ==null) continue;
                
                var rel = await ANNE.GHOSTDATA.new_ghost_has(fromNOBJ, relnNOBJ.nid, TO_TYPE_ANNE, toNOBJ.nid);    
                $clump.extraps.push(rel);
            }
        }
    }
}


class anne_date_range {
    constructor() {
        this.nid1 = null;
        this.nid2 = null;
        this.ad1 = null;
        this.ad2 = null;
        this.nObj1 = null;
        this.nObj2 = null;

        this.rangeNOBJ = null;      //for proper range neurons that cover ranges
        this.simpleDisplay = "";

        this.error_msg = "";

        //for parsing year month date ranges (no day)
        this.d1_str = "";
        this.d2_str = "";
    }


    //wikipedia ranges.
    //March–October 2006
    //2006–2007
    //2023–present
    parseFromString=async($str, $bMAKE=false)=>{
        var obj = ANNE.UTILS.stripParens($str);
        var str = obj.str;

        //just a start date.
        if (str.indexOf("-present")!=-1) {
            str = str.replace("-present", "");
            if (parseInt(str)>0) {
                this.ad1 = new anne_date(str, false);
                if (this.ad1!=null) {
                    await this.ad1.testDate($bMAKE);
                    this.nObj1 = this.ad1.nObj;
                    if (this.nObj1!=null) this.nid1 = this.nObj1.nid;
                }
            }
            return;
        }

        //just one part.
        var parts = str.split("–");
        if (parts.length==1) parts = str.split("-");
        if (parts==1) {
            if (parseInt(str)>0) {
                this.ad1 = new anne_date(str, false);
                if (this.ad1!=null) {
                    await this.ad1.testDate($bMAKE);
                    this.nObj1 = this.ad1.nObj;
                    if (this.nObj1!=null) {
                        this.nid1 = this.nObj1.nid;
                        this.d1_str = this.nObj1.text;
                    }
                }
            }
            return;
        }

        var part1=parts[0].trim();
        var part2=parts[0].trim();

        //parse the parts.  This is a standard range string like: 2006–2007
        if (parseInt(part1)>0 && parseInt(part2>0)) {
            this.ad1 = new anne_date(part1, false);
            if (this.ad1!=null) {
                await this.ad1.testDate($bMAKE);
                this.nObj1 = this.ad1.nObj;
                if (this.nObj1!=null) {
                    this.nid1 = this.nObj1.nid;
                    this.d1_str = this.nObj1.text;
                }
            }
            this.ad2 = new anne_date(part2, false);
            if (this.ad2!=null) {
                await this.ad2.testDate($bMAKE);
                this.nObj2 = this.ad2.nObj;
                if (this.nObj2!=null) {
                    this.nid2 = this.nObj2.nid;
                    this.d2_str = this.nObj2.text;
                }
            }
        }
        
        //same year month range
        //March–October 2006
        //["March", "October 2006"]
        else {
            var month_start = null;
            var month_end = null;
            var year_both = null;

            if (isNaN(partsInt(part1))==true) {
                var month_start = ANNE.MEM._monthMap[part1.toLowerCase()];
                if (month_start==null) {
                    this.error_msg = "Can't decipher "+$str+" with present range parsing code.";
                    return;
                }
            }

            if (part2.indexOf(" ")!=-1) {
                //["October 2006"]
                var part2_parts = part2.split(" ");
                if (part2_parts.length!=2) {
                    this.error_msg = "Can't decipher 2nd part of month range in "+$str+" with present range parsing code.";
                    return;
                }
                var part2_part1 = part2_parts[0].trim();
                var month_end = ANNE.MEM._monthMap[part2_part1.toLowerCase()];
                if (month_end==null) {
                    this.error_msg = "Can't decipher 2nd month of month range in "+$str+" with present range parsing code.";
                    return;
                }
                var part2_part1 = part2_parts[1].trim();
                var year_both = 0;
                if (isNaN(partsInt(part1))==true) {
                    this.error_msg = "Can't decipher year month range in "+$str+" with present range parsing code.";
                }
                else {
                    year_both = parseInt(part_part1);
                }

                //these are now date parts.
                //our year month format is:  JAN-1953
                this.d2_str = month_start.text.toUpperCase().substring(0,3)+"-"+year_both;
                this.ad1 = new anne_date(this.d2_str, false, true, month_start, null);
                await this.ad1.testDate($bMAKE);
                if (this.ad1.nObj!=null) {
                    this.nObj1 = this.ad2.nObj;
                    this.nid1 = this.ad1.nObj.nid;
                }

                this.d2_str = month_end.text.toUpperCase().substring(0,3)+"-"+year_both;
                this.ad2 = new anne_date(this.d2_str, false, true, month_end, null);
                await this.ad2.testDate($bMAKE);
                if (this.ad2.nObj!=null) {
                    this.nObj2 = this.ad2.nObj;
                    this.nid2 = this.ad2.nObj.nid;
                }
            }
        }

        return;
    }


    /**
     * Use this with any two date nids..
     * for custom/ad-hoc ranges.
     * @param {*} $d1_nid 
     * @param {*} $d2_nid 
     */
    loadFromDates=async($d1_nid=null, $d2_nid=null)=> {
        this.nid1 = $d1_nid;
        this.nid2 = $d2_nid;

        if (this.nid1==null && this.nid2==null) return;

        if (this.nid1!=null) this.nObj1 = MEMG.getNeuronByNid(this.nid1);
        else this.nObj1 = null;
        
        if (this.nid2!=null) this.nObj2 = MEMG.getNeuronByNid(this.nid2);
        else this.nObj2 = null;

        var queryNids = [];
        if (this.nid1 && this.nObj1 && this.nObj1.outs.length < 5 && this.nObj1.ghost==false) queryNids.push(this.nid1);
        if (this.nid2 && this.nObj2 && this.nObj2.outs.length < 5 && this.nObj2.ghost==false) queryNids.push(this.nid2);
        
        //need to get the basics for these dates.
        if (queryNids.length) {
            await ANNE.API.queryForOutsFroms_not(queryNids, false, [4], true, false);
        }

        if (this.nObj1!=null) {
            this.ad1 = new anne_date(this.nObj1.text, false);
            this.ad1.loadFromNeuron(this.nObj1, false);
        }
        if (this.nObj2!=null) {
            this.ad2 = new anne_date(this.nObj2.text, false);
            this.ad2.loadFromNeuron(this.nObj2, false);    
        }
        
        if (this.nObj2!=null) {
            if (this.ad1!=null) {
                this.simpleDisplay = "("+this.ad1.year+" - "+this.ad2.year+")";
            }
            else {
                this.simpleDisplay = "(? - "+this.ad2.year+")";
            }
        }   
        else if (this.nObj1!=null) {
            this.simpleDisplay = "("+this.ad1.year+" - )";
        }
        else this.simpleDisplay("(? - ?)");
    }

    genBornDeadDisplay=function() {
        if (this.nObj2!=null) {
            if (this.ad1!=null) {
                this.simpleDisplay = "(b."+this.ad1.year+" - d."+this.ad2.year+")";
            }
            else {
                this.simpleDisplay = "(b.? - d."+this.ad2.year+")";
            }
        }   
        else if (this.nObj1!=null) {
            this.simpleDisplay = "(b."+this.ad1.year+" - )";
        }
        else this.simpleDisplay("(b.? - d.?)");
    }
    

    /**
     * write this when needed.. we have range neurons that hold ranges..
     * @param {*} $rangeNeuron 
     */
    loadFromRangeNeuron=async($rangeNeuron)=>{
        this.rangeNOBJ = $rangeNeuron;
    }


    getYears=function() {
        if (this.ad1!=null && this.ad2!=null) {
            return this.ad2.year - this.ad1.year;
        }
    }

    getdays=function() {
        if (this.ad1!=null && this.ad2!=null) {
            return this.ad2.aed - this.ad1.aed;
        }
    }
}



class anne_date {
    constructor($dateString, $bFROMWIKI=true, $bMONTH_YEAR=false, $MY_M=null, $MY_Y=null) {
        this.dateString = $dateString;
        this.day = -1;
        this.month = -1;
        this.year = -1;
        this.era = "";

        this.new = true;
        this.bNEW = true;
        this.nObj = null;
        this.bTESTED = false;

        this.makeClassNOBJ = MEM.date_gregorian;

        this.dayNOBJ = null;
        this.monthNOBJ = null;
        this.yearNOBJ = null;
        this.eraNOBJ = null;
        this.dotyNOBJ = null;
        this.dotwNOBJ = null;
        this.anneepochday = null;
        this.daymonthNOBJ = null;
        this.monthyearNOBJ = null;
        this.aed = null;

        this.bLIKELY_MONTHYEAR = $bMONTH_YEAR;
        
        if ($bMONTH_YEAR==true) {
            this.parseMonthYearStr($dateString, $MY_M, $MY_Y);
        }
        else if($bFROMWIKI==true) {
            this.parseWikiDateStr($dateString);
        }
        else this.parseANNEDateStr($dateString);
    }


    /**
     * 
     * @param {*} $dateString 
     * @returns 
     */
    desperadoDatePartParse = function($dateString) {
        this.dateStr = $dateString;
        this.makeClassNOBJ = MEM.monthyear;

        var parts = $dateString.split(" ");
        if (parts.length==1) parts = $dateString.split("-");
        if (parts.length!=2) return;

        this.monthNOBJ = ANNE.MEM._monthMap[parts[0].toLowerCase().trim()];
        if (this.monthNOBJ==null) {  
            this.monthNOBJ = ANNE.MEM._monthMap[parts[1].toLowerCase().trim()];
            if (this.monthNOBJ==null) {
                this.monthNOBJ=null;
                this.month=-1;
                return;
            }   
            else {
                this.month=this.monthNOBJ.value;
                var yr_part = parts[0].trim();
                if (isNaN(partsInt(yr_part))==true) {
                    this.yearNOBJ=null;
                    this.year = -1;
                    return;
                }
                else {
                    this.year = partsInt(yr_part);
                    this.dateStr = this.monthNOBJ.text.toUpperCase().substring(0,3)+"-"+this.year;
                }
            }
        }

        else {
            this.month=this.monthNOBJ.value;
            var yr_part = parseInt(parts[1].trim());
            if (isNaN(yr_part)==true) {
                this.yearNOBJ=null;
                this.year = -1;
                return;
            }
            else {
                this.year = parseInt(yr_part);
                this.dateStr = this.monthNOBJ.text.toUpperCase().substring(0,3)+"-"+this.year;
            }
        }
    }


    /**
     * 
     * @param {*} $dateString 
     * @param {*} $MY_M 
     * @param {*} $MY_Y 
     * @returns 
     */
    parseMonthYearStr = function($dateString, $MY_M=null, $MY_Y=null) {
        if ($MY_M!=null) {
            this.monthNOBJ = $MY_M;
            this.month = this.monthNOBJ.value;
        }

        //must get month here or no-can-go-forward.
        else {
            return this.desperadoDatePartParse($dateString);
        }

        if ($MY_Y!=null) {
            this.yearNOBJ = $MY_Y;
            this.year = $MY_Y.value;
        }
        else {
            this.yearNOBJ=null;
            this.year = -1;
        }

        if ($dateString!=null) {
            this.dateStr = $dateString;
        }
        else if (this.monthNOBJ!=null && this.yearNOBJ!=null) {
            this.dateStr = this.monthNOBJ.text.toUpperCase().substring(0,3)+"-"+this.yearNOBJ.text;
        }
    }


    sameMonthYear = function($ad) {
        if ($ad==null) return false;
        if (this.monthyearNOBJ==null && $ad.monthyearNOBJ==null) return false;
        if (this.monthyearNOBJ.nid == $ad.monthyearNOBJ.nid) return true;
        return false;
    }
    sameDayMonth = function($ad) {
        if ($ad==null) return false;
        if (this.daymonthNOBJ==null && $ad.daymonthNOBJ==null) return false;
        if (this.daymonthNOBJ.nid == $ad.daymonthNOBJ.nid) return true;
        return false;
    }
    sameYear = function($ad) {
        if ($ad==null) return false;
        if (this.yearNOBJ==null && $ad.yearNOBJ==null) return false;
        if (this.yearNOBJ==null) {
            if (this.nObj.nid==$ad.yearNOBJ.nid) return true;
            else return false;
        }
        else if ($ad.yearNOBJ==null) {
            if (this.yearNOBJ.nid == $ad.nObj.nid) return true;
            else return false;
        }
        else if (this.yearNOBJ.nid == $ad.yearNOBJ.nid) {
            return true;
        }
        return false;
    }

    
    /**
     * 
     * @param {*} $dateNOBJ 
     * @param {*} $bPARSESTRING 
     */
    loadFromNeuron=function($dateNOBJ, $bPARSESTRING=true) {
        this.nObj = $dateNOBJ;

        if ($bPARSESTRING==true) {
            if (this.nObj.ghost==true) {
                this.parseANNEDateStr($dateNOBJ.text);
            }
            else {
                //load the relons
                //TO DO..
                this.parseANNEDateStr($dateNOBJ.text);
            }
        }

        else {
            this.dayNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.day.temp_nid);
            this.monthNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.month.temp_nid);
            if (this.monthNOBJ!=null) {
                //set this.month value from monthNOBJ
            }
            this.yearNOBJ =  ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.year.temp_nid);
            if (this.yearNOBJ!=null) {
                this.year = parseInt(this.yearNOBJ.text); // or better, use value
            }
            this.eraNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.era.temp_nid);
            this.daymonthNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.daymonth.temp_nid);
            this.monthyearNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.monthyear.temp_nid);
            this.dotyNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.doty.temp_nid);
            this.dotwNOBJ = ANNE.MEMG.getHasNObj(this.nObj.outs, ANNE.MEM.dotw.temp_nid);
            this.aed = ANNE.MEMG.getHasInt(this.nObj.outs, ANNE.MEM.ANNE_EPOCH_DAY.temp_nid);
        }
        
    }


    /**
     * preferred ANNE date string format.
     */
    getANNEDateStr = function($sep="-") {
        var annedateStr = "";
        if (this.year!=-1) annedateStr = UTILS.formatDigits(this.year, 4);
        if (this.month!=-1) annedateStr+=$sep+UTILS.formatDigits(this.month, 2);
        if (this.day!=-1) annedateStr+=$sep+UTILS.formatDigits(this.day, 2);
        if (this.era!="") annedateStr+=" "+this.era;
        //console.log("(getANNEDateStr) "+annedateStr);
        return annedateStr;
    }

    
    /**
     * test date with default worldview ANNE and the active user.
     */
    testDate = async($bMAKEGHOST=false)=> {
        if (this.dateStr==null || this.dateStr==undefined || this.dateStr=="") {
            this.nObj = null;
            return;
        }

        var bMUSTPULL=true;
        
        //just an integer... ( a single year ) 
        if (this.year>-1 && this.month==-1 && this.day==-1) {
            this.nObj = await MEMG.getOrMakeIfNewInteger(null, this.year+"");
            if (this.nObj.bNEW!=undefined) {
                this.new = this.nObj.bNEW;
                this.bNEW = this.nObj.bNEW;
            }
        }

        //a year month like "JAN-1953"
        else if (this.year>-1 && this.month>-1 && this.day==-1) {
            this.nObj = await ANNE.API.queryForMonthYear(this.dateStr, [MEM.ANNE_ACCTID, USER._userNObj.nid], bMUSTPULL);
            if (this.nObj==null) {
                await this.makeTempDate($bMAKEGHOST);
            }
        }

        //an actual date    
        else {
            //quick index search (should have don before this, but still give it a go)
            //a partial date is valid as well.
            this.nObj = await ANNE.MEMG.getDateNeuron(this.dateStr);
            if (this.nObj!=null) {
                this.bTESTED=true;
                return;
            }

            //this will return null or a single date NOBJ
            this.nObj = await ANNE.API.queryForDate(this.dateStr, [MEM.ANNE_ACCTID, USER._userNObj.nid], bMUSTPULL);
            
            console.log(this.nObj);
            if (this.nObj==null) {
                console.log("%c "+this.dateStr+" %cthis is a new date", TALK.bgblackgreen, TALK.bgblue);
                await this.makeTempDate($bMAKEGHOST);
            }
            else {
                console.log("%c "+this.dateStr+" %cthis is a known date", TALK.bgblackgreen, TALK.bggreen);
                this.new=false;
                this.bNEW=false;
            } 
         }
        
        this.bTESTED=true;
        
        return;
    }

    isAlpha=function($v) {
        const regExp = new RegExp(/^[a-z0-9]+$/i)
        return regExp.test($v);
    }


    parseANNEDateStr = function($str) {
        var str = $str.replaceAll(","," ");
        str = str.replaceAll("-", " ");
        str = str.replaceAll("/", " ");
        str = str.replace(/\s+/g, " ");
        
        var dateParts = str.split(" ");

        //assume single item is year
        if (dateParts.length==1) {
            this.year=parseInt(dateParts[0]);
        }

        //two items is either YEAR ERA (10,000 BC) or MONTHYEAR (JULY 2000)
        else if (dateParts.length==2) {
            if (this._isEra(dateParts[1])==false) {
                this.year=parseInt(dateParts[1]);
                this.month=parseInt(dateParts[0]);
                return;
            }
            else {
                this.era = dateParts[1];
                this.year = dateParts[0];
                this.dateStr = this.getANNEDateStr();
                return;
            }
        }

        //YYYY MM DD
        else if (dateParts.length==3) {
            this.year=parseInt(dateParts[0]);
            this.month=parseInt(dateParts[1]);
            this.day=parseInt(dateParts[2]);
        }

        else if (dateParts.length==4) {
            if (this._isEra(dateParts[3])==false) {
                alert("unhandled 4 part date");
                console.log(dateParts);
                return;
            }

            this.year=parseInt(dateParts[0]);
            this.month=parseInt(dateParts[1]);
            this.day=parseInt(dateParts[2]);
            this.era=parseInt(dateParts[3]);
        }
        this.dateStr = this.getANNEDateStr();
    }



    /**
     * Wiki external data uses full form month names.  We have to parse those.
     */
    parseWikiDateStr = function($str="") {
        if ($str==undefined || $str==null || $str=="") return "";
        $str = $str+"";
        
        var str = $str.replaceAll(",", " ");
        str = str.replaceAll("-", " ");
        str = str.replaceAll("/", " ");
        str = str.replace(/\s+/g, " ");
        
        var dateParts = str.split(" ");
        var again = dateParts.join("");
        if (UTILS.isNum(again)==true) return this.parseANNEDateStr($str); //all numbers


        //1935-08-20 dupe?
        
        if (dateParts.length==1) {
            this.year=parseInt(dateParts[0]);
        }

        //this could be a YEAR/MONTH  or  YEAR AD/BC
        else if (dateParts.length==2) {

            //first check for ear (like 10000 BC)
            if (this._isEra(dateParts[1])==true) {
                this.era = dateParts[1];
                this.year=parseInt(dateParts[0]);
                this.dateStr = this.getANNEDateStr();
                return;
            }


            //next check our 2 parts for a space (could be June 2000)
            this.month = ANNE.MEM._monthMap[dateParts[0]+"".toLowerCase()];
            if (this.month !=undefined) {
                if (parseInt(dateParts[1])>1600) this.year = parseInt(dateParts[1]);
            }
            else {
                var safu = dateParts[1]+"";
                this.month = ANNE.MEM._monthMap[safu.toLowerCase()];
            }

            if(this.month==undefined) {
                var part0 = parseInt(dateParts[0]);
                var part1 = parseInt(dateParts[1]);
                if (part0>31 && part1 < 13) {
                    this.year = part0;
                    this.month = part1;
                }
                else if (part1>31 && part0 < 13) {
                    this.year = part1;
                    this.month = part0;
                }
                else {
                    this.year = -1;
                    this.month = -1;
                    return;
                }
                
            }
        }

        else if (dateParts.length==3) {

            //2 parts with ERA
            if (this._isEra(dateParts[2])==true) {
                console.log(dateParts);
                return;
            }

            //3 parts, non numeric
            else {
                var parts = [
                        {found:false, part:dateParts[0]+""},
                        {found:false, part:dateParts[1]+""},
                        {found:false, part:dateParts[2]+""}]


                //named text month.
                for (var p=0; p < parts.length; p++) {
                    var part = parts[p].part.toLowerCase();
                    if (ANNE.UTILS.isNum(part)==true) continue;
                    var month_test = ANNE.MEM._monthMap[part];
                    if (month_test!=undefined) {
                        this.month = month_test.value;
                        parts.splice(p,1);
                        break;
                    }
                }

                //look for a numbered month.
                if (this.month == 0) {
                    for (var p=0; p < parts.length; p++) {
                        var month_test = ANNE.MEM._monthMap[parts[p].part.toLowerCase()];
                        if (month_test!=undefined) {
                            this.month = month_test.value;
                            parts.splice(p,1);
                            break;
                        }
                    }
                }


                //pick the bigger one for year.  Not perfect but fine for us.
                //who puts July 23, 2 for a year?
                if (this.month>0){
                    var part1 = parseInt(parts[0].part);
                    var part2 = parseInt(parts[1].part);
                    if (part1>part2 && part2<32) {
                        this.year = part1;
                        this.day = part2;
                    }
                    else if (part1<32){
                        this.year = part2;
                        this.day = part1;
                    }
                }
  
            }
        }

        if (this.year>-1 && this.month>0) {
            if (this.day==-1) {
                this.makeClassNOBJ = MEM.monthyear;
                this.dateStr = this.getShort(this.month)+"-"+this.getMin2(this.day);
            }
            else {
                this.makeClassNOBj = MEM.date_gregorian;
                this.dateStr = this.getANNEDateStr()
            }
        }
            
        else this.dateStr = this.getANNEDateStr();
    }

    getMin2=function($n) {
        if ($n+"".length==1) return "0"+$n;
        return $n;
    }

    getShort=function($m){
        if ($m==1) return "JAN";
        else if ($m==2) return "FEB";
        else if ($m==3) return "MAR";
        else if ($m==4) return "APR";
        else if ($m==5) return "MAY";
        else if ($m==6) return "JUN";
        else if ($m==7) return "JUL";
        else if ($m==8) return "AUG";
        else if ($m==9) return "SEP";
        else if ($m==10) return "OCT";
        else if ($m==11) return "NOV";
        else if ($m==12) return "DEC";
        return "";
    }

    /**
     * Make a temp date.  
     * No extra relons needed for temp dates (just the text and the be class) as they 
     * will be officially created by the ANNODE using the GLOBAL feature.  Dates are safe 
     * to make straight away by the ANNODE (with no approval/review) since bad data can be
     * tested by script (no human intervention).  Temp dates are for ghosts to use in 
     * building circuits and testing them before making them for real.
     */
    makeTempDate=async($bGHOST=false)=>{
        if ($bGHOST==true) {

            if (this.year>0 && (this.era =="" || this.era=="AD") && this.month==-1 && this.day==-1) {
                this.nObj = await MEMG.getOrMakeIfNewInteger(null, this.year+"");
            }

            else {
                this.nObj = await GHOSTDATA.makeGhostNeuron(this.makeClassNOBJ, this.dateStr, NTYPE_INST);
            }
        }

        this.new=true;
        this.bNEW=true;

        //if we don't store a temp name relon (which we don't need when submitting to ANNE)
        //our ghost won't have a name on refresh of ghost db (new session)
        //MEM.GHOSTDATA.makeNameRelon(this.nObj, this.dateStr);
    }

    
    /**
     */
     _isEra = function($part) {
        var part = $part.toLowerCase();
        if (part=="bc" || part=="b.c." || part=="bce" ||  part=="b.c.e." 
            || part=="ad" || part=="a.d.") return true;
        return "";
    }
}


class anne_person {
    constructor($name) {
        this.name = $name;
        this.first = "";
        this.middle = "";
        this.middle2 = "";
        this.last = "";
        this.suffix = "";
        this.fullName = "";

        this.nObj = null;

        this.setNameParts();
    }

    /**
     * Legal name suffixes (no titles)
     */
    suffixChk = function($test) {
        if (UTILS.isUSSuffix($test)==true) return true;
        return false;
    }


    /**
     */
    setNameParts = function() {
        var aParts = this.name.split(" ");

        this.first = "";
        this.middle = "";
        this.middle2 = "";
        this.last = "";
        this.suffix = "";

        this.first = aParts[0];
        if (aParts.length==1) {
            this.last = aParts[0];
            this.first="";
        }
        else if (aParts.length==2) {
            var isSuffix = this.suffixChk(aParts[1]);
            if (isSuffix==true) this.suffix = aParts[1];
            else this.last = aParts[1];
        }
        else if (aParts.length==3) {
            var isSuffix = this.suffixChk(aParts[2]);
            if (isSuffix==true) {
                this.suffix = aParts[2];
                this.last = aParts[1];
            }
            else {
                this.last = aParts[2];
                this.middle = aParts[1];
            }
        }
        else if (aParts.length==4) {
            var isSuffix = this.suffixChk(aParts[3]);
            if (isSuffix==true) {
                this.suffix = aParts[3];
                this.last = aParts[2];
                this.middle = aParts[1];
            }
            else {
                this.last = aParts[3];
                this.middle2 = aParts[2];
                this.middle = aParts[1];
            }
        }
        else if (aParts.length==5) {
            var isSuffix = this.suffixChk(aParts[4]);
            if (isSuffix==true) {
                this.suffix = aParts[4];
                this.last = aParts[3];
                this.middle2 = aParts[2];
                this.middle = aParts[1];
            }
            else {
                this.last = aParts[4];
                this.middle2 = aParts[3];
                this.middle = aParts[1]+" "+aParts[2];
            }
        }

        //we will store the full name as the neuron's text.
        if (this.first) this.fullName=this.first;
        if (this.middle) this.fullName+=" "+this.middle;
        if (this.middle2) this.fullName+=" "+this.middle2;
        if (this.last) this.fullName+=" "+this.last;
        if (this.suffix) this.fullName+=" "+this.suffix;

        this.fullName = this.fullName.trim();
    }
}


class anneid {
    constructor($nObj=null, $rez=null) {
        this.nObj = $nObj;
        if ($nObj==null) {
            console.warn(" nOBJ is required here");
            return;
        }
        this.nid = $nObj.nid;
        this.andy_actions = 0; 
        this.relons_sponsored = 0;  
        this.hanais = 0; 
        this.sponsored_neurons = 0; 
        this.did_annex = 0;  
        this.got_annexed = 0;  
        this.blocks_mined = 0 ;  
        this.share_mines = 0;
        this.boostybacks_won = 0; 
        this.swap_volume = 0;

        $nObj._anneid = this;
        this.bRANKS=false;

        this.total_users = 0;
        this.rank_aa = 0;
        this.rank_rs = 0;
        this.rank_ha = 0;
        this.rank_sn = 0;
        this.rank_da = 0;
        this.rank_ga = 0;
        this.rank_bm = 0;
        this.rank_sm = 0;
        this.rank_bw = 0;

        if ($rez!=null) this.processFromObj($rez);
    }


    /**
     * 
     * @param {*} $anneidObj 
     */
    processFromObj = function($anneidObj) {
        if ($anneidObj==undefined || $anneidObj==null) return;
        
        this.andy_actions = parseInt($anneidObj.andy_actions);
        this.relons_sponsored = parseInt($anneidObj.relons_sponsored);
        this.hanais = parseInt($anneidObj.hanais);
        this.sponsored_neurons = parseInt($anneidObj.sponsored_neurons);
        this.did_annex = parseInt($anneidObj.did_annex);
        this.got_annexed = parseInt($anneidObj.got_annexed);
        this.blocks_mined = parseInt($anneidObj.blocks_mined);
        this.share_mines = parseInt($anneidObj.share_mines);
        this.boostybacks_won = parseInt($anneidObj.boostybacks_won/100);
        this.swap_volume = parseInt($anneidObj.swap_volume/100);

        if ($anneidObj.rank_total_users!=undefined) {
            this.processRanksFromObj($anneidObj);
        }
    }


    /**
     */
    getRanks=async()=>{
        return ANNE.API.getANNEID_ranks(this.nObj);
    }
    

    /**
     * 
     * @param {*} $obj 
     */
    processRanksFromObj=function($obj) {
        this.bRANKS=true;
        this.total_users = parseInt($obj.rank_total_users);
        this.rank_aa = parseInt($obj.rank_andy_actions);
        this.rank_rs = parseInt($obj.rank_relons_sponsored);
        this.rank_ha = parseInt($obj.rank_hanais);
        this.rank_sn = parseInt($obj.rank_sponsored_neurons);
        this.rank_da = parseInt($obj.rank_did_annex);
        this.rank_ga = parseInt($obj.rank_got_annexed);
        this.rank_bm = parseInt($obj.rank_blocks_mined);
        this.rank_sm = parseInt($obj.rank_share_mines);
        this.rank_bw = parseInt($obj.rank_boostybacks_won);
    }


    /**
     * 
     */
    emulate=function($factor = 1000, $total_users=33) {    
        this.total_users = $total_users;
        this.andy_action = 0;
        if (Math.random()>.3) this.andy_action = Math.round($factor * Math.random());
        var f2 = Math.pow(2, Math.random());
        this.relons_sponsored = Math.round(Math.random()*$factor*100*f2);
        this.hanais = Math.round(Math.random()*$factor);
        this.sponsored_neurons = Math.round(Math.random()*$factor)*10;
        this.did_annex = Math.round(Math.random()*$factor/2);
        this.got_annexed = Math.round(Math.random()*$factor/2);

        if (Math.random()>.3) {
            this.blocks_mined = Math.round(Math.random()*$factor)*30;
        }
        if (Math.random()>.3) {
            this.share_mines = Math.round(Math.random()*$factor)*333;
        }

        if (Math.random()>.7) {
            this.boostybacks_won = Math.round(Math.random()*2000000);
        }
    
        if (Math.random()>.7) {
            this.swap_volume = Math.round(Math.random()*2000000);
        }
 
        this.rank_bm = UTILS.range(1, this.total_users);
        this.rank_sm = UTILS.range(1, this.total_users);
        this.rank_aa = UTILS.range(1, this.total_users);
        this.rank_ha = UTILS.range(1, this.total_users);
        this.rank_sn = UTILS.range(1, this.total_users);
        this.rank_rs = UTILS.range(1, this.total_users);
        this.rank_da = UTILS.range(1, this.total_users);
        this.rank_ga = UTILS.range(1, this.total_users);
        this.rank_bw = UTILS.range(1, this.total_users);
    }


    colorizeRank = function($rank) {
        var colorized = "color-yellow";
        var green_tier = .8;
        var yellow_tier = .5;
        var rper = ((this.total_users+1)-$rank)/this.total_users;        
        var hash = "<span style='font-size:.9rem'>#</span>";
        if (rper >=green_tier) colorized="<span class='color-green'>"+hash+$rank+"<span class='rank-of'>/"+this.total_users+"</span></span>";
        else if (rper >=yellow_tier) colorized="<span class='color-yellow'>"+hash+$rank+"<span class='rank-of'>/"+this.total_users+"</span></span>";
        else colorized="<span class='color-orange'>"+hash+$rank+"<span class='rank-of'>/"+this.total_users+"</span></span>";
        return colorized;
    }
}


const APLEX_SUMMARY_SPECIFIC_NEURON = 0;    //awarenesses for a specific neuron (class or inst at any level)
const APLEX_SUMMARY_NEURON_LINE = 1;        //awareness for a neuron line starting at a top neuron including children classes and insts.
const APLEX_SUMMARY_IDENTITY = 2;           //awarenesses OF a specific user 
const APLEX_SUMMARY_IDENTITY_GROUP = 3;     //awaarenesses of an identity group of some kind (any group of ANNE IDs)


/**
 * summary of aplex data at various levels in a tree structure..
 * (classes include summary of their direct value, subclasses, and instances)
 */
class aplex_sum {
    constructor($nObj, $summaryType=APLEX_SUMMARY_SPECIFIC_NEURON, $rez=null) {
        this.nid = $nObj.nid;
        this.pos = 0;
        this.neu = 0;
        this.neg = 0;
        this.fire = 0;
        this.bEMULATED=false;
        this.since_height=0;

        this.posPer = 0;
        this.negPer = 0;
        this.neuPer = 0;
        
        $nObj._aplex_sum = this;

        this.bucket_neg=[];
        this.bucket_neu=[];
        this.bucket_pos=[];

        this.id_type = 0;
        this.summary_type = $summaryType;

        this._aplex = null; //standard aplex obj (non-summary) with price_lvl etc
        if ($nObj._aplex!=undefined && $nObj._aplex!=null) {
            this._aplex = $nObj._aplex;
        }
        
        //test this with live data (not emulated)
        if ($rez!=null) {
            this.processAwarezSummaryIntoBucketsObj($rez);
        }
    }



    /**
     * 
     * @param {*} $sumObj 
     */
    processSumObj = function($sumObj) {
        this.pos = $sumObj.sum_POS;
        this.neg = $sumObj.sum_NEG;
        this.neu = $sumObj.sum_NEU;
        this.fire = $sumObj.sum_FIRE;
        var sum = this.pos+this.neg+this.neu;
        this.posPer = Math.floor(100*this.pos/sum)/100;
        this.negPer = Math.floor(100*this.neg/sum)/100;
        this.neuPer = Math.floor(100*this.neu/sum)/100;
    }


    /**
     * 
     * @param {*} $aplex 
     */
    emulateSummaryDataWithAplex = function($aplex) {
        if ($aplex==undefined || $aplex==null) return;

        this.pos = $aplex.pos;
        this.neg = $aplex.neg;
        this.neu = $aplex.neu;
        this.fire = this.pos+this.neg+this.neu;
        this.posPer = Math.floor(1000*this.pos/this.fire)/1000;
        this.negPer = Math.floor(1000*this.neg/this.fire)/1000;
        this.neuPer = Math.floor(1000*this.neu/this.fire)/1000;
        console.log("%cEMULATE DATA %c fire "+UTILS.commas(this.fire)+"%c"+UTILS.commas(this.pos)+"%c"+UTILS.commas(this.neg)+"%c"+UTILS.commas(this.neu), TALK.bgfire, TALK.bgblack, TALK.bgpos, TALK.bgneg, TALK.bgneu);
        this.bEMULATED=true;

        //how we will process them from the API when pulling a summary of buckets.
        var nidCntArr = this.emulateBucketVals();
        console.log(nidCntArr); //an array of objs with .nid and .count just like it came form the API.

        //wraap it in the same way the API returns it in its wrapper
        var emulatedRez = {"result":nidCntArr, "results":nidCntArr.length, "since_height":this.since_height, "for_nid":this.nid};

        //this is the same processor for emulated and real.
        this.processAwarezSummaryIntoBucketsObj(emulatedRez);
    }


    /**
     * 
     * @param {*} $factor 
     */
    emulateSummaryData = function($factor=1, $bFORCEEQUAL=false, $bEMULATEBUCKETS=true) {
        this.pos = Math.floor(Math.random() * 101)*$factor;
        if ( $bFORCEEQUAL==true) {
            this.neg = this.pos;
            this.neu = this.pos;
        }
        else {
            this.neg = Math.floor(Math.random() * 101)*$factor;
            this.neu = Math.floor(Math.random() * 101)*$factor;
        }
        
        this.fire = this.pos+this.neg+this.neu;
        this.posPer = Math.floor(1000*this.pos/this.fire)/1000;
        this.negPer = Math.floor(1000*this.neg/this.fire)/1000;
        this.neuPer = Math.floor(1000*this.neu/this.fire)/1000;
        console.log("%cEMULATE DATA %c fire "+UTILS.commas(this.fire)+"%c"+UTILS.commas(this.pos)+"%c"+UTILS.commas(this.neg)+"%c"+UTILS.commas(this.neu), TALK.bgfire, TALK.bgblack, TALK.bgpos, TALK.bgneg, TALK.bgneu);
        this.bEMULATED=true;

        if ($bEMULATEBUCKETS==false) return;
    
        //how we will process them from the API when pulling a summary of buckets.
        var nidCntArr = this.emulateBucketVals();
        console.log(nidCntArr); //an array of objs with .nid and .count just like it came form the API.

        //wraap it in the same way the API returns it in its wrapper
        var emulatedRez = {"result":nidCntArr, "results":nidCntArr.length, "since_height":this.since_height, "for_nid":this.nid};

        //this is the same processor for emulated and real.
        this.processAwarezSummaryIntoBucketsObj(emulatedRez);
    }




    /**
     * We're going to go the long road here
     * to emulate these as they may look coming back from the API when pulled for real.
     * 
     * When we do query for buckets we grab all the buckets in a "nid" and incriment a "count".
     * We can then organize these into a fz bucket object for easy traversal and use in the GUIS.
     */
    emulateBucketVals=function() {
        var nidCntArr = [];
        
        var seed = Math.random()*69350;
        while(this.pos>0) {
            if (Math.random() > .995) {
                if(nidCntArr[MEM.positive.nid]==undefined) nidCntArr[MEM.positive.nid]={"count":1};
                else nidCntArr[MEM.positive.nid]["count"]+=1;
            }
            else {
                var nid = AWARE.getRandPosBucketVal(seed);
                if (nid==undefined) {
                    alert("IMPOSSIBLE");
                }
                if(nidCntArr[nid]==undefined) nidCntArr[nid]={"count":1};
                else nidCntArr[nid]["count"]+=1;
            }
            this.pos=this.pos-1;
        }

        var seed = Math.random()*296.9;
        while(this.neg>0) {
            if (Math.random() > .995) {
                if(nidCntArr[MEM.negative.nid]==undefined) nidCntArr[MEM.negative.nid]={"count":1};
                else nidCntArr[MEM.negative.nid]["count"]+=1;
            }
            else {
                var nid = AWARE.getRandNegBucketVal(seed);
                if (nid==undefined) {
                    alert("IMPOSSIBLE");
                }
                if(nidCntArr[nid]==undefined) nidCntArr[nid]={"count":1};
                else nidCntArr[nid]["count"]+=1;
            }
            this.neg=this.neg-1;
        }

        var seed = Math.random()*66579;
        while(this.neu>0) {
            if (Math.random() > .995) {
                if(nidCntArr[MEM.neutral.nid]==undefined) nidCntArr[MEM.neutral.nid]={"count":1};
                else nidCntArr[MEM.neutral.nid]["count"]+=1;
            }
            else {
                var nid = AWARE.getRandNeuBucketVal(seed);
                if (nid==undefined) {
                    alert("IMPOSSIBLE");
                }
                if(nidCntArr[nid]==undefined) nidCntArr[nid]={"count":1};
                else nidCntArr[nid]["count"]+=1;
            }
            this.neu=this.neu-1;
        }

        if (this.neu!=0 || this.neg!=0 || this.pos!=0) {
            alert("BAD EMULATION");
        }
        var oldFireAudit = this.fire;
        console.log("old fire audit %c"+oldFireAudit, TALK.bgred);
        this.fire=0;

        var newFireAudit = 0;

        //use a normaal array.. not an associative.
        //that was helpful to easily find dupes.
        var keys = Object.keys(nidCntArr);
        var nidFinalArr = [];
        for (var k=0; k < keys.length; k++) {
            var nid = keys[k];
            if(nid==undefined || nid==undefined) {
                alert("IMPOSSIBLE.");
            }
            var obj = nidCntArr[nid];
            obj.nid=nid;
            newFireAudit+=parseInt(obj.count);
            nidFinalArr.push(obj);
        }
        console.log(nidFinalArr);
        console.log("new fire audit %c"+newFireAudit, TALK.bgred);

        return nidFinalArr;
    }


    /**
     * We want to recalc the gui but we already know what the updates are since we just 
     * made the AWAREZ relons ourself with this app!
     * 
     * Update the summary here.
     * We can also update the anneid summaries too.
     * @param {*} $rels 
     */
    
    updateFromLiveSend = function($rels) {

        const THIS = this;
        
        var addOrIncrPos = function($bnid) {
            for (var b=0; b < THIS.bucket_pos.length; b++) {
                if(THIS.bucket_pos[b].nid==$bnid){
                    THIS.bucket_pos[b].count+=1;
                    THIS.pos+=1;
                    THIS.fire+=1;
                    return;
                }
            }
            THIS.addPos({nid:$bnid, count:1});
        }
        var addOrIncrNeg = function($bnid) {
            for (var b=0; b < THIS.bucket_neg.length; b++) {
                if(THIS.bucket_neg[b].nid==$bnid){
                    THIS.bucket_neg[b].count+=1;
                    THIS.neg+=1;
                    THIS.fire+=1;
                    return;
                }
            }
            THIS.addNeg({nid:$bnid, count:1});
        }
        var addOrIncrNeu = function($bnid) {
            for (var b=0; b < THIS.bucket_neu.length; b++) {
                if(THIS.bucket_neu[b].nid==$bnid){
                    THIS.bucket_neu[b].count+=1;
                    THIS.neu+=1;
                    THIS.fire+=1;
                    return;
                }
            }
            THIS.addNeu({nid:$bnid, count:1});
        }

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var nOBJ = MEMG.getOrMakeIfNew(rel.reln);
            if (nOBJ.owned_by!=ANNE_ACCTID && nOBJ.owned_by!=USER._userNObj.nid) {
                console.log("we're not interested in this.");
                continue;
            }

            var bucket = AWARE.getBucket(nOBJ.nid);
            if (bucket==null || bucket.nObj==undefined || bucket.nObj==null) {
                continue;
            }

            var bucket_nid = bucket.nObj.nid;

            var pol = bucket.pol;
            if (pol==undefined || pol==null) {
                if (nOBJ!=null) pol = nOBJ.pol;
            }

            if (pol=="pos") addOrIncrPos(bucket_nid);
            else if (pol=="neg") addOrIncrNeg(bucket_nid);
            else if (pol=="neu") addOrIncrNeu(bucket_nid);
            else console.log(" unmapped, won't make the summary!", bucket);
        }


        this.posPer = Math.floor(1000*this.pos/this.fire)/1000;
        this.negPer = Math.floor(1000*this.neg/this.fire)/1000;
        this.neuPer = Math.floor(1000*this.neu/this.fire)/1000;
    }



    /**
     * We need to iteratea through a list of AWARE reln nids with counts and put them in their pos/neg/neu buckets.
     * @param {*} $rez 
     */
    processAwarezSummaryIntoBucketsObj=function($rez) {

        var rez = $rez.result;
        if (rez==undefined || rez==null) rez=[];
        
        this.since_height = $rez.since_height;

        var tempPosArr = [];
        var tempNegArr = [];
        var tempNeuArr = [];
        
        var fireAuditAgain = 0;
        var fireauditInner = 0;

        var unmapped_fire = 0;
        
        for (var r=0; r < rez.length; r++) {
            var rezObj = rez[r];
            var nid = rezObj.nid;
            
            if (fireAuditAgain!=fireauditInner) {
                console.log("debug here");
            }

            //these nids are from AWAREZ relons... and the nid was the RELN of those RELONS.
            //these must be ANNE owned or self-owned, otherwise we're not summarizing them.
            var nOBJ = MEMG.getOrMakeIfNew(nid);
            if (nOBJ.owned_by!=ANNE_ACCTID && nOBJ.owned_by!=USER._userNObj.nid) {
                console.log("we're not interested in this.");
                continue;
            }

            var count = parseInt(rezObj.count);
            fireAuditAgain+=count;

            //for the final bucket click view we want our leafs (feelz/op/buckets/polarityneurons) with specific vals.
            var doLeaf = function($arr, $bnid, $nid, $count) {
                if ($count==undefined) {
                    console.log("why no count?");
                }
                if ($arr[$bnid].leafs==undefined) {
                    $arr[$bnid].leafs=[];
                    $arr[$bnid].leafs[$nid]={"count":$count};
                }
                else {
                    if ($arr[$bnid].leafs[$nid]==undefined) {
                        $arr[$bnid].leafs[$nid]={"count":$count};
                    }
                    else {
                        $arr[$bnid].leafs[nid].count+=$count;
                    }
                }
            }

            if (nid==MEM.positive.nid) {
                if (tempPosArr[MEM.positive.nid]==undefined) tempPosArr[MEM.positive.nid]={nid:MEM.positive.nid, count:count};
                else tempPosArr[MEM.positive.nid].count+=count;
                doLeaf(tempPosArr, MEM.positive.nid, MEM.positive.nid, count);
                fireauditInner+=count;
                continue;
            }
            else if (nid==MEM.negative.nid) {
                if (tempNegArr[MEM.negative.nid]==undefined) tempNegArr[MEM.negative.nid]={nid:MEM.negative.nid, count:count};
                else tempNegArr[MEM.negative.nid].count+=count;
                doLeaf(tempNegArr, MEM.negative.nid, MEM.negative.nid, count);
                fireauditInner+=count;
                continue;
            }
            else if (nid==MEM.neutral.nid) {
                if (tempNeuArr[MEM.neutral.nid]==undefined) tempNeuArr[MEM.neutral.nid]={nid:MEM.neutral.nid, count:count};
                else tempNeuArr[MEM.neutral.nid].count+=count;
                doLeaf(tempNeuArr, MEM.neutral.nid, MEM.neutral.nid, count);
                fireauditInner+=count;
                continue;
            }

            var bucket = AWARE.getBucket(nid);
            if (bucket==null || bucket.nObj==undefined || bucket.nObj==null) {
                alert("PROBLEM, get bucket yielded no bucket.");
                continue;
            }

            var bucket_nid = bucket.nObj.nid;
            

            var pol = bucket.pol;
            if (pol==undefined || pol==null) {
                if (nOBJ!=null) pol = nOBJ.pol;
            }

            if (pol=="pos") {
                if (tempPosArr[bucket_nid]==undefined) tempPosArr[bucket_nid]={nid:bucket_nid, count:count};
                else tempPosArr[bucket_nid].count+=count;
                fireauditInner+=count;
                doLeaf(tempPosArr, bucket_nid, nid, count);
            }
            else if (pol=="neg") {
                if (tempNegArr[bucket_nid]==undefined) tempNegArr[bucket_nid]={nid:bucket_nid, count:count};
                else tempNegArr[bucket_nid].count+=count;
                fireauditInner+=count;
                doLeaf(tempNegArr, bucket_nid, nid, count);
            }
            else if (pol=="neu") {
                if (tempNeuArr[bucket_nid]==undefined) tempNeuArr[bucket_nid]={nid:bucket_nid, count:count};
                else tempNeuArr[bucket_nid].count+=count;
                fireauditInner+=count;
                doLeaf(tempNeuArr, bucket_nid, nid, count);
            }
            else {
                unmapped_fire+=count;
                console.log(" unmapped, won't make the summary!", bucket);
            }
        }

        //console.log("unmapped_fire %c"+unmapped_fire, TALK.bgfire);
        //console.log("fireauditInner %c"+fireauditInner, TALK.bgfire);
        //console.log("fireAuditAgain %c"+fireAuditAgain, TALK.bgfire);

        //now we have pos/neg/neu buckets to use that map to our bars..
        //with proper counts
        var keys = Object.keys(tempPosArr);
        for (var k=0; k < keys.length; k++) {this.addPos(tempPosArr[ keys[k] ]);}

        var keys2 = Object.keys(tempNegArr);
        for (var k=0; k < keys2.length; k++) {this.addNeg(tempNegArr[ keys2[k] ]);}
        
        var keys3 = Object.keys(tempNeuArr);
        for (var k=0; k < keys3.length; k++) {this.addNeu(tempNeuArr[ keys3[k] ]);}

        //console.log("%cold percents %c"+UTILS.commas(this.posPer)+"%c"+UTILS.commas(this.negPer)+"%c"+UTILS.commas(this.neuPer), TALK.bgblack, TALK.bgpos, TALK.bgneg, TALK.bgneu);
        if (this.pos>0 && this.fire>0) this.posPer = Math.floor(1000*this.pos/this.fire)/1000;
        if (this.neg>0 && this.fire>0) this.negPer = Math.floor(1000*this.neg/this.fire)/1000;
        if (this.neu>0 && this.fire>0) this.neuPer = Math.floor(1000*this.neu/this.fire)/1000;

        //console.log("%cEMULATE DATA %c fire "+UTILS.commas(this.fire)+"%c"+UTILS.commas(this.pos)+"%c"+UTILS.commas(this.neg)+"%c"+UTILS.commas(this.neu), TALK.bgfire, TALK.bgblack, TALK.bgpos, TALK.bgneg, TALK.bgneu);
        //console.log("%cnew percents %c"+UTILS.commas(this.posPer)+"%c"+UTILS.commas(this.negPer)+"%c"+UTILS.commas(this.neuPer), TALK.bgblack, TALK.bgpos, TALK.bgneg, TALK.bgneu);
        
        //we should have these equal.
        //this.auditPos();
        //this.auditNeg();
        //this.auditNeu();
    }

    auditPos=function() {
        var bucket_audit=0;
        for (var b=0; b < this.bucket_pos.length; b++) {
            bucket_audit+=this.bucket_pos[b].count;
        }
        console.log("pos bucket audit:"+bucket_audit+" .pos:"+this.pos);
    }
    addPos=function($obj) {
        this.bucket_pos.push($obj);
        this.pos+=$obj.count;
        this.fire+=$obj.count;
    }
    auditNeg=function() {
        var bucket_audit=0;
        for (var b=0; b < this.bucket_neg.length; b++) {
            bucket_audit+=this.bucket_neg[b].count;
        }
        console.log("neg bucket audit:"+bucket_audit+" .neg:"+this.neg);
    }
    addNeg=function($obj) {
        this.bucket_neg.push($obj);
        this.neg+=$obj.count;
        this.fire+=$obj.count;
    }
    auditNeu=function() {
        var bucket_audit=0;
        for (var b=0; b < this.bucket_neu.length; b++) {
            bucket_audit+=this.bucket_neu[b].count;
        }
        console.log("neu bucket audit:"+bucket_audit+" .neu:"+this.neu);
    }
    addNeu=function($obj) {
        this.bucket_neu.push($obj);
        this.neu+=$obj.count;
        this.fire+=$obj.count;
    }
}

const APLEX_ID_NEURON = 0;  //the majority of all awarenesses will be on neurons.
const APLEX_ID_RELON = 1;   //we can take awarenesses on relons too, by txid

class aplex {
    /**
     */
    constructor($obj, $obj_id, $aplexJsonObj=null) {
        this.id=$obj_id;
        this.id_type=APLEX_ID_NEURON; //default nid
        this.annies_in = 0;
        this.fire = 0;
        this.prev_annex_price=0;
        this.boosties=0;
        this.price_lvl=0;
        this.neg = 0;
        this.neu = 0;
        this.pos = 0;
        this.total = 0;

        //a pending annex/boosties change is in the mempool and affects this neuron.
        this.has_pending = false;   
        this.pending_sponsored_by = "";
        
        //must be calc'd.
        this.calculated_cost = 0;
        this.annex_sponsored_relons = 0;
        this.non_awareness_outs = 0;

        this.boosties = 0;
        this.fibo_factor = 0;
        this.last_fullhash = "";

        this.bANNEX_CALCULATED=false;
        this.bANNEXABLE = false; //update tfhis 
        this.pending_relon = null;
        
        if ($aplexJsonObj!=null) {
            this.processObj($aplexJsonObj);
        }
        if ($obj!=null) {
            $obj._aplex = this;

            //set the aplex_sum ptr back to the _aplex object.
            //this is for if we already had the aplex sum obj before we had the aplex obj
            if ($obj._aplex_sum!=undefined && $obj._aplex_sum._aplex==null) {
                $obj._aplex_sum._aplex = this;
            }    
        }

        this.pulled_when = Date.now()/1000;
    }

    getAgeInSecs = function() {
        if (this.pulled_when==null) {
            return 0;
        }
        return (Date.now()/1000) - this.pulled_when;
    }
    
    /**
     * 
     * @param {*} $obj 
     */
    processObj = function($obj) {
        this.id_type=$obj.id_type;
        if ($obj.annies_in==null) this.annies_in=0;
        else this.annies_in=parseInt($obj.annies_in);
        
        if ($obj.fire==null) $obj.fire=0;
        else this.fire=parseInt($obj.fire);

        if(ANNE.UTILS.isset($obj.prev_annex_price)==true) this.prev_annex_price=parseInt($obj.prev_annex_price);
        else this.prev_annex_price=0;

        if(ANNE.UTILS.isset($obj.boosties)==true) this.boosties=parseInt($obj.boosties);
        else this.boosties = 0;

        //console.log("boosties:"+this.boosties);
        if($obj.price_lvl!=undefined) this.price_lvl=$obj.price_lvl;

        //if (this.price_lvl==0) this.price_lvl=1;
        if (ANNE.UTILS.isset($obj.neg)==true) this.neg=parseInt($obj.neg);
        else this.neg=0;

        if (ANNE.UTILS.isset($obj.neu)==true) this.neu=parseInt($obj.neu);
        else this.neu=0;

        if (ANNE.UTILS.isset($obj.pos)==true) this.pos=parseInt($obj.pos);
        else this.pos=0;

        this.total=this.neg + this.neu + this.pos;

        //this val muts be set in the cmdObj when annex :)
        if (ANNE.UTILS.isset($obj.prev_hash)==true) this.last_fullhash = $obj.prev_hash;
        else this.last_full_hash = "";

        if (this.pos>0 && this.total>0) this.posPer = Math.round(1000*this.pos/this.total)/1000;
        else this.posPer=0;
        
        if (this.neg>0 && this.total>0) this.negPer = Math.round(1000*this.neg/this.total)/1000;
        else this.negPer=0;

        if (this.neu>0 && this.total>0) this.neuPer = Math.round(1000*this.neu/this.total)/1000;
        else this.neuPer = 0;

        //some free RAM vars we can get..
        //these won't be retained in INDEXDB tho since they didn't come in relons.
        if (this.id_type==APLEX_ID_NEURON) {
            var nObj = MEMG.getNeuronByNid(this.id);

            //snagged this in the API as a solid for pulling sets and having basic infos for display.
            if ($obj.sponsored_by!=undefined && nObj.sponsored_by=="") {
                nObj.addSponsoredBy($obj.sponsored_by);
            }
            if($obj.owned_by!=undefined) {
                nObj.addOwnedBy($obj.owned_by);
            }
            if ($obj.ntype!=undefined) {
                nObj.ntype = $obj.ntype;
            }
            if ($obj.origin!=undefined) {
                nObj.origin = $obj.origin;
            }
            if($obj.parent!=undefined) {
                var parentNOBJ = MEMG.getOrMakeIfNew($obj.parent);
                if (nObj.bes.indexOf(parentNOBJ.temp_nid)==-1) {
                    nObj.setBeTempNid(parentNOBJ.temp_nid);
                }
            }
        }


        //sometimes we can get these for low-cost on a process/pull of aplex so we do it 
        //in advance so we have it.
        if ($obj.sponsored_relons!=undefined) {
            this.processAnnexVars($obj);
        }

        
        if ($obj.has_pending==true) {
            this.has_pending=true;
            console.log($obj.pending);
            if ($obj.pending.result!=null) {
                this.pending_relon = $obj.pending.result;
                var sysROBJ = MEMG.lookupByTXID(this.pending_relon.sys_ref_txid);
                if (sysROBJ!=null) {
                    ANNE.TALK.printRelon(sysROBJ);
                }
            }
            console.log(this);
        }
        
    }


    /**
     * emulate the numbers for testing
     */
    emulateAplex = function($factor=10000) {
        if (Math.random()>.9) this.boosties = Math.ceil(Math.random()*10 * Math.random()*10 * 10000);
        this.price_lvl = 1+parseInt((Math.floor(Math.random()* (Math.ceil(Math.random()*5)-1))));
        this.fibo_factor = UTILS.nthFib(this.price_lvl+3);
        this.annex_sponsored_relons = Math.ceil(Math.random()*Math.random()*1000);
        this.non_awareness_outs = Math.ceil(Math.random()*Math.random()*1000);
        var spr_val = this.annex_sponsored_relons * ANNEX.RELON_SPONSOR_FIRE;  //they get 1 coin extra 
        var nao_val = this.non_awareness_outs * ANNEX.RELON_VALUE;      //20 coins per out.
        var factor_total = this.fibo_factor*(spr_val+nao_val);
        this.calculated_cost = factor_total+this.boosties;

        this.neg = Math.ceil(Math.random()*$factor);
        this.neu = Math.ceil(Math.random()*$factor);
        this.pos = Math.ceil(Math.random()*$factor);
        this.total = this.neg+this.neu+this.pos;
        this.fire = this.total;

        //just something (for display)
        if (this.price_lvl>0) {
            this.prev_annex_price = (this.calculated_cost - this.boosties)*.8;
        }

        this.annies_in = this.total*200;


        //we need an annexer..  who is the pretend sponsor of this neuron?
        //sponsored_by
        var nObj = MEMG.getNeuronByNid(this.id);

        if (nObj.is(MEM.identity)==false) {
            if (Math.random()>.5) nObj.sponsored_by = MEM.mr_scatman.nid;
            else nObj.sponsored_by = MEM.johnny5.nid;
        }

        this.initPercents();
    }


    //helpful in gui calcs.
    initPercents = function() {
        this.posPer = Math.round(1000*this.pos/this.total)/1000;
        this.negPer = Math.round(1000*this.neg/this.total)/1000;
        this.neuPer = Math.round(1000*this.neu/this.total)/1000;
    }



    /**
     * Switched to node based pricing for better alignment.
     * ANNEX PRICING works with a one block lag.
     * Pricing of relons is based on what relons were present up through the last mined block.
     * @param {*} $nObj 
     * @returns 
     */
    calculate_annex_vars = async($nObj)=>{
        if (this.bANNEX_CALCULATED==true || $nObj.nid!=this.id) return 1;

        ANNE.MEM_UDF.openSubHunt("calculate_annex_vars");
        var rez = await ANNE.API.calculateAnnexPrice($nObj.nid);
        if (rez==null) {
            console.log("ERROR -- couldn't get accurate ANNEX pricing data.");
            this.bANNEXABLE=false;
            return -1;
        }
        if (rez.nid!=$nObj.nid) {
            console.log("WRONG NEURON.");
            this.bANNEXABLE=false;
            return -1;
        }

        var rez = this.processAnnexVars(rez);
        ANNE.MEM_UDF.closeSubHunt();
        
        return rez;

    }


    /**
     * 
     * @param {*} $rez 
     * @returns 
     */
    processAnnexVars = function($rez) {
        this.boosties = parseInt($rez.boosties);
        this.calculated_cost = $rez.annex_price;
        this.annex_sponsored_relons = parseInt($rez.sponsored_relons);
        this.non_awareness_outs = parseInt($rez.non_awareness_relons);
        this.fibo_factor = parseInt($rez.fibo_factor);
        this.price_lvl = parseInt($rez.price_lvl);
        
        this.bANNEX_CALCULATED = true;

        return this.calculated_cost;
    }



    /**
     * Here we tally non awareness neurons.  (those are captured in the pos/neu/neg total).
     * We want to know how many relons are sponsored.  Those relon sponsors get paid one coin
     * when an ANNEX occurs.
     * We also want to know how many data outs this neuron has.  This tells us the value of the 
     * neuron based on the data in its definition.  We will multiple these numbers 
     * by the LEVEL factor (how high of a level this neuron is determines that factor)
     */
    tallyRelons = function($nObj) {
        var asr = 0;
        var nao = 0;
        for (var r=0; r < $nObj.outs.length; r++) {
            var rel = $nObj.outs[r];
            if (rel.ghost==true) continue;
            if (rel.temp==true) continue;
            if (rel.type==TYPE_AWARENESS) continue;
            if (rel.type==TYPE_SYS_MSG) continue;
            if(rel.relon_sponsor!=undefined && rel.relon_sponsor!="") asr+=1;
            nao+=1;
        }
        this.annex_sponsored_relons = asr;
        this.non_awareness_outs = nao;
    }

    updateFromLiveSend = function($rels) {
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var nOBJ = MEMG.getOrMakeIfNew(rel.reln);
            if (nOBJ.owned_by!=ANNE_ACCTID && nOBJ.owned_by!=USER._userNObj.nid) {
                console.log("we're not interested in this.");
                continue;
            }

            var bucket = AWARE.getBucket(nOBJ.nid);
            if (bucket==null || bucket.nObj==undefined || bucket.nObj==null) {
                continue;
            }

            var bucket_nid = bucket.nObj.nid;

            var pol = bucket.pol;
            if (pol==undefined || pol==null) {
                if (nOBJ!=null) pol = nOBJ.pol;
            }

            if (pol=="pos") {
                this.pos+=1;
                this.total+=1;
            }
            else if (pol=="neg") {
                this.neg+=1;
                this.total+=1;
            }
            else if (pol=="neu") {
                this.neu+=1;
                this.total+=1;
            }
            else console.log(" unmapped, won't make the summary!", bucket);
            this.initPercents();
        }
    }
}

/************** (from file ./js/base/misc_classes.js)**/ 


/** 
  */
class gui_stuff {
    constructor() {
        this.ANNE = null;
        this._el_stats = "anne_memstats";         
    }  


    /** 
     * @author {Mr.Scatman}
     */
    init=function($ANNE) {
        this.ANNE = $ANNE;
    }
    
    /** 
     * Set a memstats div to get memstats printed to the gui
     * 
     * @param {string}
     * @author {Mr.Scatman}
     */
    setMemStatsDivid = function($el_memstats="memstats") { this._el_stats = $el_memstats; }

    

    /** 
     * @author {Mr.Scatman}
     */
    updateMemStats=function() {
        if (this==null || this._el_stats=="" || this._el_stats==undefined) return;
        var stats = "neurons: "+MEM._aNeurons.length;
        stats=stats+"</BR>relons: "+MEM._aRelons.length;
        ANNE.A.setHTML(this._el_stats, stats);
    }



    html_href=function($display, $url, $class="") {
        if($class!="") $class='class="'+$class+'"';
        return '<a href="'+$url+'" '+$class+' target="_blank">'+$display+'</a>';
    }


    getghostIcoSolo=function($ghost_type, $size="120") {
        var ico = "";
        if ($ghost_type=="guto") ico = this.getIco("ghost-upgrade-to-other", $size, "ghost-upgrade");
        else if ($ghost_type=="guta") ico = this.getIco("ghost-upgrade-to-ANNE", $size, "ghost-upgrade");
        else if ($ghost_type=="gutm") ico = this.getIco("ghost-upgrade-to-mine", $size, "ghost-upgrade");
        else if ($ghost_type=="hanaicollab") ico = this.getIco("hanai-collab", $size, "ghost-upgrade");
        else if ($ghost_type=="boosties") ico = this.getIco("boosties", $size, "");
        else return "";
        return ico;
    }



    /** 
    * @author {Mr.Scatman}
    */
    getIco=function($type, $w=22, $class="img-ico", $style="") {
        var white = "_white";
        var ico="";
        var style = "";
        if ($style!="") style = "style ='"+$style+"'";
        else if ($type=="boosties") ico="<span title='add boosties' "+style+"><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/boosties"+white+"_ico.png' alt=''></img></span>";
        else if ($type=="TREE") ico="<span title='tree' "+style+"><img class='"+$class+"' width='22px' height='22px' src='./imgs/TREE"+white+"_ico.png' alt=''></img></span>";
        else if ($type=="DS") ico="<span title='data source' "+style+"><img class='"+$class+"' width='22px' height='22px' src='./imgs/DS"+white+"_ico.png' alt=''></img></span>";
        else if ($type=="DOSSIER") ico="<span title='dossier view' "+style+"><img class='"+$class+"' width='22px' height='22px' src='./imgs/ico_dossier"+white+".png' alt=''></img></span>";
        else if ($type=="delete"||$type=="bin") ico='<span class="ico-delete">&#128465;</span>';
        else if ($type=="add") ico = '<span class="ico-add">"&#x2295;</span>';
        else if ($type=="HANAI") ico = "<span class='ico-hanai'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/hanai64_ico.png' alt=''></img></span>";
        else if ($type=="ALEXANNDRIA") ico="<span title='library of alexanndria'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/alexanndria_ico.png' alt=''></img></span>";
        else if ($type=="REMEMBERZ") ico="<span><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/brain_ico.png' alt=''></img></span>";
        else if ($type=="ANDY") ico="<span title='ANDY rights'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/andy_ico.png' alt=''></img></span>";
        else if ($type=="ANJ") ico="<span title='anjucator'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/ANJ_ico.png' alt=''></img></span>";
        else if ($type=="MERGE") ico="<span title='Merge'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/ico_merge.png' alt=''></img></span>";
        else if ($type=="MINER") ico="<span title='miner ico'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/miner_ico.png' alt=''></img></span>";
        else if ($type=="ANNTHENAEUM") ico="<span title='Annthenauem'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/annth_ico.png' alt=''></img></span>";
        else if ($type=="TABLE") ico="<span title='table view'><img class='"+$class+"'  width='"+$w+"px' height='"+$w+"px' src='./imgs/TABLE_ico.png' alt=''></img></span>";
        else if ($type=="ANNET") {
            //console.log("GET ICO ANNET");
            ico="<span title='ANNET neuromorphic hypergraph'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/ANNET_ico.png' alt=''></img></span>";
        }
        else if ($type=="APLEX") ico="<span title='marketplace'><img class='"+$class+"'  width='"+$w+"px' height='"+$w+"px' src='./imgs/APLEX_ico.png' alt=''></img></span>";
        else if ($type=="DS") ico="<span title='data source'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/DS_ico.png' alt=''></img></span>";
        else if ($type=="TREE") ico="<span title='tree'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/TREE_ico.png' alt=''></img></span>";
        else if ($type=="SEND") ico="<span title='tree'><img class='"+$class+"' width='"+$w+"px' height='"+$w+"px' src='./imgs/SEND_ico.png' alt=''></img></span>";
        else if ($type=="RECEIPT") ico='<i class="ai-receipt"></i>';
        else if ($type=="ghost-upgrade-to-ANNE") ico="<span title='upgrade ghost data to ANNE global data' "+style+"><img class='"+$class+"' width='"+$w+"px' src='./imgs/ghost_upgrade_anne_ico.png' alt=''></img></span>";
        else if ($type=="ghost-upgrade-to-mine") ico="<span title='upgrade ghost data to your real data' "+style+"><img class='"+$class+"' width='"+$w+"px' src='./imgs/ghost_upgrade_mine_ico.png' alt=''></img></span>";
        else if ($type=="ghost-upgrade-to-other") ico="<span title='collab relon' "+style+"><img class='"+$class+"' width='"+$w+"px' src='./imgs/ghost_upgrade_other_ico.png' alt=''></img></span>";
        else if ($type=="hanai-collab") ico="<span title='hanai collab' "+style+"><img class='"+$class+"' width='"+$w+"px' src='./imgs/hanai_collab_ico.png' alt=''></img></span>";
        else if ($type=="wikipedia") ico="<span title='wikipedia' class='center-all' style='min-width:"+$w+"px'><span class='ai-wikipedia'></span></span>";
        
        else alert("WE DONT HAVE THAT ICO '"+$type+"', SORRY");

        return ico;
    }


    /** 
     * @author {Mr.Scatman}
     */
     ___buildListDisplayFor=async ($nObj,$types,$bCHASE=true,$bSLIM=false) => {
        if(ANNE.MEM_UDF!=null&&typeof ANNE.MEM_UDF._udf_buildListDisplayFor==="function") {
            return ANNE.MEM_UDF._udf_buildListDisplayFor($nObj,$types,$bCHASE,$bSLIM);
        }
        var all=MEMG.getParentLineRelons($nObj,$types);
        var aDisplay=[];
        if ($types.indexOf(TYPE_SYS)!=-1) var d1=await this.___genListDisplayFor($nObj,all.sys,$bCHASE,aDisplay,$bSLIM);
        if ($types.indexOf(TYPE_BE)!=-1) var d2=await this.___genListDisplayFor($nObj,all.bes,$bCHASE,aDisplay,$bSLIM);
        if ($types.indexOf(TYPE_PARAM)!=-1) var d3=await this.___genListDisplayFor($nObj,all.params,$bCHASE,aDisplay,$bSLIM);
        if ($types.indexOf(TYPE_HAS)!=-1) var d4=await this.___genListDisplayFor($nObj,all.has,$bCHASE,aDisplay,$bSLIM);
        return aDisplay;
    }

    /**
     * We want to return an array of value/text objects for use in a simple 
     * display DD  (most likely produced in-full by buildListDisplayFor)    
     * 
     * @author {Mr.Scatman} 
     */
     ___genListDisplayFor=async ($nObj,$arr,$bCHASE=true,$ptrArray=null,$bSLIM=false) => {
        if(ANNE.MEM_UDF!=null&&typeof ANNE.MEM_UDF._udf_genListDisplayFor==="function") {
            return await ANNE.MEM_UDF._udf_genListDisplayFor($nObj,$arr,$bCHASE,$ptrArray,$bSLIM);
        }
        var arr=[];
        for(var a=0;a<$arr.length;a++) {
            var pObj=$arr[a];
            if(pObj.type==TYPE_BE || pObj.type==TYPE_SYS) {
                if(pObj.from_nid!=$nObj.nid) continue;
            }
            //if(pObj.reln==MEM.key.nid) continue;   //no need to show the key
            if(pObj.type==TYPE_BE && pObj.reln==MEM.name.nid) continue;   //we know our name

            //if this relon is inhibited by the nObj or the from (parent line), do not include it
            if($nObj.inhibit_txs!=undefined&&$nObj.inhibit_txs.indexOf(pObj.txid)!=-1) continue;
            if(pObj.from_nid!=$nObj.nid) {
                var parentNOBJ=MEMG.getNeuronByTempNid(pObj.f);
                if(parentNOBJ.inhibit_txs!=undefined&&parentNOBJ.inhibit_txs.indexOf(pObj.txid)!=-1) continue;
            }
            arr.push(pObj);
        }


        //we need to get basic display info about everything in the pObjs
        if($bCHASE==true&&arr.length) {
            var bSKIPIFHASTEXT=true;
            var bDOEXTRAS=false;
            var aNidsList=MEMG.extractAllFromRelons(arr,bSKIPIFHASTEXT,bDOEXTRAS,[],true);
            if(aNidsList.length&&ANNE.CHASE!=null) {
                var aRelons=await ANNE.CHASE.lookupMinTNids(aList);
            }
        }

        for(var a=0;a<arr.length;a++) {
            var pObj=arr[a];
            var fNObj=MEMG.getNeuronByTempNid(pObj.f);
            var inherit="";
            var temp="";
            if(fNObj.nid!=$nObj.nid) {
                if($bSLIM==true) continue;
                inherit=" (I)";
            }
            if(pObj.temp==true) continue;

            var msg="";
            var bNOFONT=true;
            var talkmsg="";
            if(ANNE.TALK!=null) talkmsg=TALK.humanReadableRelon(pObj,$bSLIM,bNOFONT);
            talkmsg=msg+talkmsg+inherit+temp;
            if($ptrArray!=null) $ptrArray.push({"value": pObj.txid,"text": talkmsg});
        }
        return;
    }



    /** 
     * @author {Mr.Scatman}
     */
     ___madeDDfromValTextArr=function($el_dd,$aList,$msg) {
        if(ANNE.MEM_UDF!=null&&typeof ANNE.MEM_UDF._udf_madeDDfromValTextArr==="function") {
            return ANNE.MEM_UDF._udf_madeDDfromValTextArr($el_dd,$aList,$msg);
        }
        var dropdown=document.getElementById($el_dd);
        var headerO=document.createElement("option");
        headerO.setAttribute("data-divider",true);
        headerO.setAttribute("disabled",true);
        headerO.innerHTML=$msg;
        headerO.style.backgroundColor="#ccc";
        headerO.style.color="#000";
        dropdown.append(headerO);

        for(var t=0;t<$aList.length;t++) {
            var opt=document.createElement("option");
            headerO.setAttribute("value",$aList[t].value);
            headerO.innerHTML=$aList[t].text;
            document.append(opt);
        }
        dropdown.setAttribute('selectedIndex',0);
        ANNE.A.show(dropdown);
    }


    /** 
     * @author {Mr.Scatman}
     */
     _deleteAdminAction=function($action) {
        alert("what is this?")
        for (var a = 0; a < this._aActions.length; a++) {
            var action = this._aActions[a];
            if (action._nid==$action._nid) {
                this._aActions.splice(a, 1);
                return;
            }
        }
        return;
    }


    //just calc it here.  no need to do the uom functionality
    doAnniesToAnneSpanFull = function($el, $pre="", $annies=0) {
        var annecoin = $annies/100;
        return this.doAnneSpan($el, $pre, annecoin);
    }

    _convert = function ($val, $uom) {
        var post = "";
        if ($uom.unit_lbl.length) post=" "+$uom.unit_lbl;
        if ($uom.symbolHTML) post=post+$uom.symbolHTML;
        var factor = Math.ceil(1/$uom.factored);
        var total = Math.round(($uom.factored*parseFloat($val))*factor)/factor;
        var thisconv = {val:$val, total:total, post:post};
        if ($uom.conversions==undefined) $uom.conversions = [];
        $uom.conversions["_"+$val]=thisconv;
        return thisconv;
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $pre 
     * @param {*} $anne 
     */
    doAnneSpan = function($el, $pre, $annecoin=0, $post="") {
        if ($pre!="") ANNE.A.ezEl($el, "span", '', $pre);
        return ANNE.A.ezEl($el, "span", 'annecoin', UTILS.formatNumberCommas($annecoin) + this.getAnneImgHtml(16, $annecoin)+$post);
    }
    

    /**
     * Just to show how to do it...
     * @param {*} $val 
     * @param {*} $size 
     * @returns 
     */
    doAnniesToANNE_HTML = function($val, $size=16) {
        var uom = MEM.unitOfMeasure(MEM.annies);    //annies is a uo measure->uo account->uo crypto
        var thisconv = "";
        if (uom.conversions!=undefined && uom.conversions["_"+$val]!=undefined) {
            thisconv = uom.conversions["_"+$val];
        }
        else thisconv = this._convert($val, uom);
        return UTILS.formatNumber(thisconv.total) + this.getAnneImgHtml($size, thisconv.total);
    }

    /** 
     * @author {Mr.Scatman}
     */
    formatMultiTotal=function($el, $lbl="balance", $annies=0, $className="") {
        function commas(x) {
            return x.toString().replace(/\B(?=(\d{3})+(?!\d))/g,",");
        }
        var annecoin = "<span class='annecoin ibi'>"+commas((Math.round(1000*parseFloat($annies/ANNE.ANNE_TO_ANNIES)))/1000)+this.getAnneImgHtml('20px')+"</span>";
        if ($el!=null) ANNE.A.empty($el);
        ANNE.A.ezlblSpan($el, $lbl, annecoin, $className);
    }


    /** 
     * @author {Mr.Scatman}
     */
    formatANNECoin=function($annies) {
        return "<span class='annecoin-amt'>"+ANNE.UTILS.formatNumber($annies)+"</span><img title='ANNE coin' class='anneico-small' src='"+ANNE.___anneico_path+"'></img>";
    }
    formatANNIES_to_ANNECoin=function($annies) {
        return "<span class='annecoin-amt'>"+ANNE.UTILS.formatNumber($annies/ANNE.ANNE_TO_ANNIES)+"</span><img title='ANNE coin' class='anneico-small' src='"+ANNE.___anneico_path+"'></img>";
    }



    /** 
     * @author {Mr.Scatman}
     */
    getAnniesImgHtml($size=24, $val="", $class="ibi") {
        return "<img class='"+$class+"' width='"+$size+"' height='"+$size+"' title='"+$val.toLocaleString()+" (10,000 annies in 1 ANNE coin)' class='annieico-small' src='"+ANNE.___annieico_path+"'></img>";
    }


    /** 
     * @author {Mr.Scatman}
     */
    getAnneImgHtml($size=24, $val="", $class="ibi icofix") {
        return "<img class='"+$class+"' width='"+$size+"' height='"+$size+"' title='"+$val.toLocaleString()+"' class='anneico-small' src='"+ANNE.___anneico_path+"'></img>";
    }



    ___generateChildrenDD=function($el_dd, $classNObj, $defaultMsg="") {
        var nObjs = MEMG.getNObjsFromTempNidList($classNObj.children);
        nObjs.sort(ANNE.UTILS.sortByTextNoCase);
        this.___generateDDFromNObjs($el_dd,nObjs,$defaultMsg);
    }


     ___generateDDFromNObjs=function($el_dd,$nObjs,$defaultMsg="choose the data this point to") {
        if(typeof this._udf_generateDDFromNObjs==="function") {
            return this._udf_generateDDFromNObjs($el_dd,$nObjs,$defaultMsg);
        }
        var dropdown=document.getElementById($el_dd);
        ANNE.A.empty(dropdown);
        var indent="&#9494;";
        var space='&#160;';
        console.log("(generateDDFromNObjs)");
        for(var n=0;n<$nObjs.length;n++) {
            var nObj=$nObjs[n];
            var text=nObj.text;
            if(text=="") text=nObj.nid;
            if(nObj.temp_nid==ANNE.MEM.class.temp_nid) text="any class neuron";
            var spaces="";
            if(nObj.be_lvl>0) {
                for(var x=0;x<nObj.be_lvl;x++) spaces+=space;
                text=spaces+indent+text;
            }
            var opt=document.createElement("option");
            opt.setAttribute("value",nObj.nid);
            opt.innerHTML=nObj.text;
            dropdown.append(opt);
        }

        var opt=document.createElement("option");
        opt.setAttribute("value","");
        opt.innerHTML=$defaultMsg;
        dropdown.append(opt);
        dropdown.setAttribute('selectedIndex',0);
    }


    


    /**
     * PLAY A SOUND
     * 
     * Play a sound for whatever just happened.
     *
     * This helps me debug and allows another sense to be active in the experience.
     * Users can turn off sounds if they like.
     * 
     * @author {Mr.Scatman}
     */
    playSound=async($divid)=> {
        if (this._pref_volume_onoff!==-1) ANNE.A.play($divid);
    }



    /** 
     * @author {Mr.Scatman}
     */
     doDefaultImg=function($div, $imgElId, $w=360, $h=180, $imgClassAdd="") {
        ANNE.A.doImage($div, $imgElId, ANNE.___root+"/imgs/unknown.png", $w, $h, $imgClassAdd);
    }


     doDefaultImgDyno=function($div, $imgElId, $imgClassAdd="left_dyno_w") {
        ANNE.A.doImageDyno($div, $imgElId, ANNE.___root+"/imgs/unknown_.png", $imgClassAdd);
    }



    /** 
     * @author {Mr.Scatman}
     */
    ___initInstDDWithOverrideArray=function($el_dd,$defaultMsg,$nids,$bSEARCHOK=false) {
        var nObjs=MEMG.getNObjsFromTempNidList($nids);
        return this.___genDDArray($el_dd,$defaultMsg,[],[],false,false,false,false,nObjs,$bSEARCHOK);
    }


    /** 
     * DEPRECATED in the EXT
     * handler of building all DDs.. 
     * @author {Mr.Scatman}
     */
     ___genDDArray=function($el_dd,$defaultMsg,$disables=[],$adds=[],$bDOTTS=true,
        $bEWV=true,$bDISABLETOP_TT=false,$bDOTEMP=false,$nObjsOverride=[],$searchOK=true) {

        var aNObjs=[];
        var bOVERRIDEMODE=false;
        if($nObjsOverride.length) {
            aNObjs=$nObjsOverride;
            bOVERRIDEMODE=true;
        }
        else {
            aNObjs=MEMG.getAllClasses();
            bOVERRIDEMODE=false;
        }
        if(bOVERRIDEMODE==false) {
            var nonTT=new Object();
            var aOthers=[];
            nonTT["other classes"]=aOthers;
            var TT=new Object();  //array of arrays
            if($adds.length) {
                var easy=new Object();
                easy["easy access"]=$adds;
            }
            var mine=new Object();
            var aMine=[]
            mine["my classes"]=aMine;

            var anne_=new Object();
            var aANNE=[];
            anne_["ANNE classes"]=aANNE;

            var aTTs=MEMG.getSysToTypes();
            var aMineObjs=MEM.getMineNObjs();
            for(var m=0;m<aMineObjs.length;m++) {
                var nObj=aMineObjs[m];
                if(nObj.mine==true) {
                    if(nObj.bes[0]!=MEM.class.temp_nid) {
                        var beNOBJ=MEMG.getNeuronByTempNid(nObj.bes[0]);
                        if(beNOBJ!=null&&beNOBJ.mine) continue;
                    }
                    aMine.push(nObj);
                    continue;
                }
            }

            for(var i=0;i<aNObjs.length;i++) {
                var nObj=aNObjs[i];
                if($bDOTTS==false && nObj.bTERMINATING_TYPE==true) continue;
                if(nObj.temp==true && $bDOTEMP==false) continue;

                //push user's top classes (we will iterate through the children)
                if(nObj.mine==true) continue;

                if(useradopted.indexOf(nObj.temp_nid)!==-1) {
                    aAdopted.push(nObj);
                    continue;
                }
                if(MEM.ANNE.owns.indexOf(nObj.temp_nid)!=-1) {
                    if(nObj.be=="class") aANNE.push(nObj); //only push top-classes here
                    continue;
                }
            }

            if($bDOTTS==true) {
                var aTTs=MEMG.getSysToTypes();
                for(var t=0;t<aTTs.length;t++) {
                    var nObj=aTTs[t];
                    var isdefined=TT[nObj.text+"s"];
                    if(isdefined==undefined) TT[nObj.text+"s"]=[nObj];
                    else TT[nObj.SYS_ALIAS].push(nObj);
                }
            }

            var data=[];

            //user defined classes (all together)
            if($adds.length) data=this._prepSelectData(easy,$disables,false);

            if(aMine.length) {
                var datam=this._prepSelectData(mine,$disables,$bDISABLETOP_TT);
                data=data.concat(datam);
            }

            //all terminating type classes (ie, anne stuff)
            if($bDOTTS==true) {
                var data1=this._prepSelectData(TT,$disables,$bDISABLETOP_TT);
                if(data1.length) data=data.concat(data1);
            }

            if(aANNE.length) {
                var dataAnne=this._prepSelectData(anne_,$disables,$bDISABLETOP_TT);
                data=data.concat(dataAnne);
            }

            //all other classes (ie, anne stuff)
            var data2=this._prepSelectData(nonTT,$disables,$bDISABLETOP_TT);
            if(data2.length) data=data.concat(data2);
        }

        //build 
        else {
            var data=this._prepSelectDataOverride($nObjsOverride,"valid selections");
        }

        console.log(data);
        var ddel = document.getElementById($el_dd);    //gets it if it is a string for id/class or already an element.
        
        for(var d=0; d < data.length; d++) {
            var section=data[d];
            var opt = ANNE.A.createElement(ddel, "option", "option-group-heading", section.text);
            opt.setAttribute("disabled",true);
            opt.value="";
            for(var d=0; d < section.children.length; d++) {
                var dobj = section.children[d];
                var opt = ANNE.A.createElement(ddel, "option", "", dobj.text, dobj.value);
            }
        }
        ddel.value="";
        return;
    }


    /** 
     * handler of building all DDs.. 
     * @author {Mr.Scatman}
     */
     _prepSelectDataOverride=function($arr,$heading) {
        console.log("(_prepSelectDataOverride)");
        var results=[];
        if($arr.length) {
            var items=[];
            function textThenType($a,$b) {
                if($a==undefined||$a==null||$b==undefined||$b==null) return 0;
                if($a.ntype<$b.ntype) return -1;
                else if($a.ntype>$b.ntype) return 1;
                else {
                    if($a.text.toUpperCase()<$b.text.toUpperCase()) return -1;
                    else if($a.text.toUpperCase()>$b.text.toUpperCase()) return 1;
                    else return 0;
                }
            }

            function parentThenText($a,$b) {
                if($a==undefined||$a==null||$b==undefined||$b==null) return 0;
                if($a.bes[0]<$b.bes[0]) return -1;
                else if($a.bes[0]>$b.bes[0]) return 1;
                else {
                    if($a.text.toUpperCase()<$b.text.toUpperCase()) return -1;
                    else if($a.text.toUpperCase()>$b.text.toUpperCase()) return 1;
                    else return 0;
                }
            }

            //we actually want to show the parent..
            $arr.sort(parentThenText);

            function doHeading($nObj) {
                var par=$nObj.bes[0];
                if(_pdone.indexOf(par)!=-1) return;
                _pdone.push(par);
                var beNOBJ=MEMG.getNeuronByTempNid(par);
                items.push({id: par,text: '<div>'+beNOBJ.text+' ('+beNOBJ.nid+')</div>',value: beNOBJ.nid,disabled: true});
            }

            var _pdone=[];

            for(var a=0;a<$arr.length;a++) {
                var nObj=$arr[a];
                if(nObj.temp_nid==MEM.class.temp_nid) continue;
                doHeading(nObj);
                var indent="&#9494;";
                var space='&#160;';
                if(nObj.be_lvl==0) space="";
                else {
                    for(var i=0;i<nObj.be_lvl-1;i++) {space+='&#160;';}
                    space=space+indent;
                }
                items.push({id: nObj.temp_nid,text: '<div>'+space+nObj.text+' ('+nObj.nid+')</div>',value: nObj.nid});
            }
            results.push({text: $heading,children: items});
        }
        return results;
    }


    /** 
     * handler of building all DDs.. 
     * @author {Mr.Scatman}
     */
    _prepSelectData=function($arr,$disables=[],$bDISABLETTTOP=false) {
        var results=[];
        var keys=Object.keys($arr);
        for(var k=0;k<keys.length;k++) {
            var type=keys[k];
            var items=[];
            var arr=$arr[type];
            if(arr.length) {
                arr.sort(ANNE.UTILS.sortByTextNoCase);
                for(var a=0;a<arr.length;a++) {
                    var nObj=arr[a];
                    if($disables.length&&$disables.indexOf(nObj.nid)!=-1) {
                        items.push({id: nObj.temp_nid,disabled: true,text: '<div>'+nObj.text+' ('+nObj.nid+')</div>',value: nObj.nid});
                    }
                    else if($bDISABLETTTOP==true&&nObj.bTERMINATING_TYPE==true); //still do the children.
                    else items.push({id: nObj.temp_nid,text: '<div>'+nObj.text+' ('+nObj.nid+')</div>',value: nObj.nid});
                    var children=nObj.children;
                    if(children.length) this._indentChildren(children,items,1,$disables);
                }
                if(items.length) results.push({text: type,children: items});
            }
        }
        //console.log(results);
        return results;
    }



    /** 
     * helper in DDs to build intended items according to the BE dimension. 
     * @author {Mr.Scatman}
     */
    _indentChildren=function($children,$arr,$level,$disables=[]) {
        var level_max=12;
        if($children.length) {
            var children=[];    //our $children array is temp_nids and we can't sort by text until we get it!
            for(var st=0;st<$children.length;st++) {
                var temp_nid=$children[st];
                var nObj=MEMG.getNeuronByTempNid(temp_nid);
                children.push(nObj);
            }
            children.sort(ANNE.UTILS.sortByTextNoCase);
            for(var c=0;c<children.length;c++) {
                var nObj=children[c];
                if(nObj.ntype==NTYPE_INST) continue;
                if($disables.length&&$disables.indexOf(nObj.temp_nid)!=-1) continue;
                var d="";
                var indent="&#9494;";
                var space='&#160;';
                for(var i=0;i<$level-1;i++) {space+='&#160;';}
                d=space+indent+nObj.text+" ("+nObj.nid+")";
                $arr.push({id: nObj.temp_nid,text: '<div>'+d+"</div>",value: nObj.nid});
                if(nObj.children!=undefined&&$level<level_max) this._indentChildren(nObj.children,$arr,$level+1);
            }
        }
    }

    makeAdvancedExternalUrlLink=function($relon) {
        var url=MEMG.__GET_URL($relon);
        return '<a href="'+url+'" title="external url:'+url+'" target="_blank">external url&nbsp;<i class="ai-external-link-square-alt"></i></a>';
    }
    makeExternalUrlLink=function($url) {
        return '<a href="'+$url+'" title="external url:'+$url+'" target="_blank">external url&nbsp;<i class="ai-external-link-square-alt"></i></a>';
    }


    /**
     * Quick helper to get and display the class description of a neuron
     * in an already-present element with expected name.
     * @param {*} $nObj 
     * @param {*} $el_name 
     * @param {*} $ID 
     * @returns 
     */
    addClassDesc = function($nObj, $el_name="", $ID=null) {
        if ($nObj==null) return;
        if ($el_name=="") return;

        if ($ID==null) var el_class_desc = document.getElementById($el_name+"_class_desc");
        else var el_class_desc = document.getElementById($el_name+"_class_desc_"+$ID);
        if (el_class_desc==null) return;

        var desc = $nObj.getClassDesc();
        if (desc!=null && desc.length>0) {
            el_class_desc.innerHTML = desc;
            ANNE.A.showParent(el_class_desc);
            ANNE.A.show(el_class_desc);
        }
    }


    /** 
     * 
     * @author {Mr.Scatman} 
     */
    disambigHTMLStr=async($el, $nObj, $skips=[])=> {
        if ($el==undefined || $el==null) return;
        if ($nObj==null) return;

        var el = ANNE.A.empty($el);
        var disp_class = ANNE.TALK.getNeuronDisplayClass($nObj);
        
        var afile = await ANNE.IMAGES.getAnyValidImgAFILE($nObj);
        if (afile!=null) {
            var imgbtn = ANNE.A.imgButtonB64(el, "", afile, "circ-img", null);
            if (imgbtn!=null) imgbtn.setAttribute("nid", imgNOBJ.nid);
        }

        var span2 = document.createElement("span");
        el.append(span2);
        span2.classList="neuron-choice-info-part";
        span2.innerHTML = "<span><b>"+$nObj.text+"</b></span><span><b>nid:</b>"+$nObj.nid+"</span>";
        span2.classList.add(disp_class);

        if ($nObj.bes.length && $skips.indexOf("be")==-1) {
            var beNOBJ = MEMG.getNeuronByTempNid($nObj.bes[0]);
            var span3 = document.createElement("span");
            el.append(span3);
            span3.classList="neuron-choice-info-part";
            span3.innerHTML = beNOBJ.text;
            var disp_class = TALK.getNeuronDisplayClass(beNOBJ);
            span3.classList.add(disp_class);
        }

        if ($nObj.owned_by!="" && $skips.indexOf("ob")==-1) {
            var ownedBYNOBJ = MEMG.getOrMakeIfNew($nObj.owned_by);
            var span4 = document.createElement("span");
            el.append(span4);
            span4.classList="neuron-choice-info-part";
            span4.innerHTML = "<b>owner:</b>"+ownedBYNOBJ.nid;
            var disp_class = TALK.getNeuronDisplayClass(ownedBYNOBJ);
            span4.classList.add(disp_class);
        }

        if ($nObj.sponsored_by!="" && $skips.indexOf("sp")==-1) {
            var sponsoredBYNOBJ = MEMG.getOrMakeIfNew($nObj.sponsored_by);
            var span5 = document.createElement("span");
            el.append(span5);
            span5.classList="neuron-choice-info-part";
            span5.innerHTML = "<b>sponsor:</b>"+sponsoredBYNOBJ.nid;
            var disp_class = TALK.getNeuronDisplayClass(sponsoredBYNOBJ);
            span5.classList.add(disp_class);
        }

        if ($nObj.aka!=null && $skips.indexOf("aka")==-1) {
            var span = document.createElement("span");
            span.classList="neuron-choice-info-part";
            span.innnerHTML = "<b>also known as</b>"+$nObj.aka.join("/");
            el.append(span);
        }
        ANNE.A.show(el);
    }
}


var GUI_STUFF = new gui_stuff();

/************** (from file ./js/gui/gui_stuff.js)**/ 

class relon_table {
    constructor($pgui) {
        this.pgui = $pgui;

        //back ref from the pgui to this.  We can use this when looking up this relon
        //table by guid (for GUI drama)
        this.pgui.INST = this; 

        this.pObj = $pgui.pObj;
        this.ID = $pgui._guid;
        this.tbl_divid = "anne_tbl_"+this.ID;

        var init_cols = this.pObj.init_cols;
        if (init_cols==undefined) {
            init_cols = ["txid", "type", "reln", "to", "actions"];
        }
        
        this.cols = [];
        this.nid = this.pObj.p1;
        this.nObj = MEMG.getNeuronByNid(this.nid);
        this.data = this.pObj.p2;
        this.bUS = this.pObj.p3;
        
        if (this.pgui.p3!=null) this.cols = this.pgui.p3;

        //class els.
        this.el_tbl = document.getElementById(this.tbl_divid);
        this.el_newRelon_str = this.pObj.p4;
        
        this.el_rowsearch = this.el_tbl.querySelector('.row-search');
        this.el_rowheader = this.el_tbl.querySelector('.row-header');
        this.el_tbl_title = ANNE.A.setHTML("tbl_title_"+this.ID, "");
        
        this.el_tblbody = this.el_tbl.querySelector('.anne-tbl-body-overflow');

        this.el_embedded_filters = ANNE.A.hide("anne_embedded_filters_"+this.ID);
        this.el_availableCols = document.getElementById("ds_available_cols_"+this.ID);

        this.el_gardeninfo = this.el_availableCols.querySelector(".ds-let-deny-list");
        ANNE.A.hide(this.el_gardeninfo);
        this.el_colchks = this.el_availableCols.querySelector(".ds-col-checks");

        this.el_edit_btn = ANNE.A.hide('tbl_edit_btn_'+this.ID, "glow");
        this.el_save_btn = ANNE.A.hide('tbl_save_btn_'+this.ID, "glow");
        

        //txid link
        this.cols.push({
            active:init_cols.indexOf("txid")!=-1 ? true : false, 
            text:"txid", 
            ttype:"txid", 
            dataprop:"txid", 
            sortprop:"**DISPLAY**",
            displayprop:"txid", 
            drill_txid:true,
            filter:false,
            c:this.cols.length,
            style_class:"txid-cell",
            f:"f1"
        });
        

        //type 
        this.cols.push({
            active:init_cols.indexOf("type")!=-1 ? true : false, 
            text:"type", 
            ttype:"string", 
            dataprop:"type", 
            displayprop:"type",
            sortprop:"**DISPLAY**",
            drill_neuron:true, 
            filter:true,
            c:this.cols.length,
            f:"f2"
        });

        
        //reln
        this.cols.push({
            active:init_cols.indexOf("reln")!=-1 ? true : false, 
            text:"reln", 
            ttype:"anne", 
            dataprop:"reln", 
            displayprop:"reln", 
            sortprop:"**DISPLAY**",
            drill_neuron:true, 
            filter:true,
            c:this.cols.length,
            f:"f3"
        });

        
        //to type
        this.cols.push({
            active:init_cols.indexOf("to_type")!=-1 ? true : false, 
            text:"to type", 
            ttype:"string", 
            dataprop:"to_type", 
            displayprop:"to type",
            drill_neuron:true, 
            sortprop:"**DISPLAY**",
            filter:true,
            c:this.cols.length,
            f:"f1"
        });
        

        //to
        this.cols.push({
            active:init_cols.indexOf("to")!=-1 ? true : false,  
            text:"to", 
            ttype:"misc", 
            dataprop:"to_s", 
            displayprop:"to", 
            drill_neuron:true, 
            drill_txid:true, 
            sortprop:"**DISPLAY**",
            filter:true,
            c:this.cols.length,
            f:"f3"
        });

        //relon sponsor
        this.cols.push({
            active:false,  
            text:"sponsor", 
            ttype:"anne", 
            dataprop:"relon_sponsor", 
            displayprop:"relon_sponsor", 
            drill_neuron:true, 
            drill_txid:true, 
            sortprop:"**DISPLAY**",
            filter:true,
            c:this.cols.length,
            f:"f3"
        });
        
        this.cols.push({
            active:init_cols.indexOf("actions")!=-1 ? true : false,
            text:"actions", 
            ttype:"**ADMIN**", 
            dataprop:"", 
            displayprop:"",
            filter:false, 
            c:this.cols.length,
            f:"f2",
            extra_class:"actions-col-header"
        });


        this.neuronLinkType = this.pObj.neuronLinkType;//displayFunct
        if (this.neuronLinkType==undefined || this.neuronLinkType==null) this.neuronLinkType = LINKS_NEURON_TBL_VIEW; 

        this.relonLinkType = this.pObj.relonLinkType;  //relonDisplayFunctStr
        if (this.relonLinkType==undefined) this.relonLinkType = LINKS_RELON_ANNE;

        for (var c=0; c < this.cols.length; c++) {
            var colObj = this.cols[c];
            var checked = true;
            var disabled = true;
            if (colObj.active==false) {
                checked = false;
                disabled = false;
            }
            var chk = ANNE.A.ezCheckMark(this.el_colchks, "chk_"+this.ID+"_"+colObj.dataprop, "xx", colObj.text, checked, disabled, "truncate120");
            chk.setAttribute("dataprop", colObj.dataprop);

            chk.addEventListener("click", $e=> this.chk_click($e));
        }

        //one-time, on init.  Any added columns is handled int the checkmark click event.
        var col_index = 0;
        for (var c=0; c < this.cols.length; c++) {
            var col = this.cols[c];
            if (col.active==false) continue;
            this.addColumnHeader(col, col_index);
            col.col_index = col_index;
            col_index+=1;
        }

        //this should have already been set, but fix it if it isn't.
        //the goveneor will try to collab if it doesn't see it is bMINE=true.
        if (this.nObj.owned_by==USER._userNObj.nid) this.nObj.bMINE=true;

        //iterate through the data ---
        //inject new "row" with cols and data.
        this.el_tblbody.setAttribute("active-cols", parseInt(col_index));
    }

    /**
     * Init w/ async support.
     * any inits we need that are async come through this standard initASYNC
     * init function called by the NF when it sees that it exists. 
     */
    initASYNC=async($pgui)=> {

        //cleanup and setting of .mine
        for (var dd=0; dd < this.data.length; dd++) {
            if (this.data[dd].rObj==null) {
                alert("this should have never happened");
                continue;
            }
            if (this.data[dd].rObj.f==undefined) {
                alert("this should have never happened");
                var fNOBJ = MEMG.getOrMakeIfNew(this.data[dd].rObj.from_nid);
                if (fNOBJ==null) continue;
                this.data[dd].rObj.f=fNOBJ.temp_nid;
            }
            var nObj = MEMG.getNeuronByTempNid(this.data[dd].rObj.f);
            if (nObj==null) {
                alert("this should have never happened");
                console.log(this.data[dd].rObj);
                continue;
            }
        }
        
        //pass the data here even though we have it.
        await this.addDataToColsInit(this.data);
        
        //serves as a callback..
        this.pgui.addRelon = async($relon)=> {
            var rObj = {rObj:$relon, active:true};
            await this.addDataToColsInit([rObj], true);
        }     

        this.pgui.hideRid = async($rid)=> {
            var row = this.el_tblbody.querySelector(".rowrid_"+$rid);
            if (row) ANNE.A.hide(row);
        } 

        //set up table footer display
        var rowslbl = ANNE.TALK.pluralizeNounIf2(this.data.length, "row", "rows");
        ANNE.A.setHTML("tbl_summary_"+this.ID, this.data.length+" "+rowslbl);

        //initialize the final table FX DOM events.
        this.initEvents(this.tbl_divid);
        ANNE.A.hide("loading"); 


        
        if (this.pObj.misc!=null) {
            if (this.pObj.misc.dataprop!=undefined) {
                var col = this.getColByDataProp(this.pObj.misc.dataprop);
                if (col && col.sortIco!=undefined) {
                    this.defaultSort(col.sortIco, this.pObj.misc.sort);
                }
            }
        }
    }

    
    /**
     * get active and inactive here..
     */
    getOuts = function() {
        var outs = [];
        for (var d=0; d < this.data.length; d++) { outs.push(this.data[d].rObj);}
        return outs;
    }


    /**
     */
    chk_click = function($e) {
        if ($e.target.checked==true) {
            var dataprop = $e.target.getAttribute("dataprop");
            var col = this.getColByDataProp(dataprop);
            if (col==null) return;
            console.log("you checked a new column of data w/ datanid:"+dataprop);
            col.active = true;
            var col_index = parseInt(this.el_tblbody.getAttribute("active-cols"));
            this.el_tblbody.setAttribute("active-cols", parseInt((col_index+1)));
            col.col_index = col_index; 

            //sort ico and filters are wired up in the addColumnHeader..
            this.addColumnHeader(col, col_index, this.tbl_divid, this.el_rowsearch, this.el_rowheader);
            this.addColumnToRows(col); //add the data to the rows.. 
        }
        else {
            var dataprop = $e.target.getAttribute("dataprop");
            var col_index = parseInt(this.el_tblbody.getAttribute("active-cols")-1);
            this.el_tblbody.setAttribute("active-cols", parseInt((col_index)));
            var col = this.getColByDataProp(dataprop)
            if (col==null) return;
            this.deleteColumnFromRows(col); //add the data to the rows..
            this.deleteColumnHeader(col.col_index);              
            col.active=false;
        }
    }

    /**
     */
    getColByDataProp = function($dataprop) {
        for (var c=0; c < this.cols.length; c++) {
            if (this.cols[c].dataprop==$dataprop) return this.cols[c];
        }
        return null;
    }


    /**
     * This will handle any collab row for this thing that is in the DOM,
     * even if it showed up in a table or ANNET.. etc.. (or both)
     */
    static updateCollabRow=async($ID, $collab_txid, $status_text)=> {
        var el_tbl = document.getElementById("anne_tbl_"+$ID);
        if (el_tbl==null) {
            console.log("relon tbl with id "+$ID+" not found in the DOM");
            return;
        }
        var el_rows = el_tbl.querySelectorAll(".rowtxid_"+$collab_txid);
        if (el_rows==null || el_rows.length==0) return;
        if ($status_text=="d") {
            for (var e=0; e < el_rows.length; e++) {
                var el_row = el_row[e];
                ANNE.A.removeSelf(el_row);
            }
            await MEM_UDF.removeCollab($collab_txid);                   // remove collab from collab memory
        }
        else if ($status_text=="a") {
            var collab_relon = await MEM_UDF.removeCollab($collab_txid); // remove collab from collab memory
            if (collab_relon==null) return;

            //run the collab relon through the mem processor.. get a real rid.
            var relon = MEMP.processNeuronsFromRelons([collab_relon], true, true, "relons", "relon_table.updateCollabRow");
            for (var e=0; e < el_rows.length; e++) {
                var el_row = el_row[e];
                el_row.classList.remove("pending-collab");
                el_row.classList.remove("rowtxid_"+$collab_txid);
                var el_datacol = el_row.querySelector('actionstxid_'+$collab_txid);
                if (el_datacol==null) continue;
                el_datacol.classList.remove('actionstxid_'+$collab_txid);  
                el_datacol.classList.remove('andy-ico');            
                el_row.classList.add("rowrid_"+relon.rid);
                el_datacol.classList.add("actionsrid_"+relon.rid);
            }
        }
    }

    /** 
     * A required param was set which allows for another param
     * to be activated now..
     */
    static makeNewTempParamFromVarParam=async($el, $newRelon, $param_txid)=> {
        var txid = $el.getAttribute("txid");
        if (txid==null) return;
        var relon = MEMG.lookupRelonByTXID(txid);
        if (relon==null) return;

        //make sure we don't make a glorious blunder.
        //we are trying to update a row that depended on another param being set first.
        //(conditional param)
        if (relon.depends_on==undefined) return;
        if (relon.depends_on!=$param_txid) return;
        
        //chase up this el to get the first anne-tbl.  this will have the tblid we 
        //need to get the instance of anne_relon_table.  (we are currently in a static function)
        var tbl = ANNE.A.getFirstParentOf($el, "anne-tbl");
        if (tbl==null) return;

        //extract the guid.
        var widget_id = tbl.id.substring(9);    //anne_tbl_ xxxxx
        if (widget_id=="") return;

        //ge the anne_relon_table.. (the INST on the pgui at runtime creation).
        var ANNE_RELON_TBL = NF.getINSTbyGuid(widget_id);
        if (ANNE_RELON_TBL==null) return;

        var newRel = ANNE_RELON_TBL.makeCDXParamIfNotExists(ANNE_RELON_TBL.nid, $newRelon, relon);
        var dObj = {active:true, rObj:newRel};
        ANNE_RELON_TBL.data.push(dObj);

        await ANNE_RELON_TBL.addDataToColsInit([dObj], true);

        //revisit this if we are "building" a definition and not "using" one.
        ANNE_RELON_TBL.deactivateRelon(relon.txid, $el);
    }   

    deactivateRelon = function($txid, $el) {
        for (var d=0; d < this.data.length; d++) {
            var dObj = this.data[d];
            if (dObj.active==false) continue;
            if (dObj.rObj.txid==$txid) {
                dObj.active=false;
                break;
            }
        }
        ANNE.A.hide($el);
    }

    getParamVarFulfillment=function($nObj, $txid) {
        var relon = MEMG.lookupRelonByTXID($txid);
        console.log(relon);
        ANNE.TALK.printRelon(relon);
        var rNOBJ = MEMG.getNeuronByTempNid(relon.r);
        console.log("(getParamVarFulfillment) looking for relon "+rNOBJ.text);

        if (relon.type!=TYPE_PARAM) return null;

        console.log($nObj);
        console.log($nObj.text);
        for (var r=0; r < $nObj.outs.length; r++) {
            var rel = $nObj.outs[r];
            if (rel.type!=TYPE_HAS) continue;
            if (rel.r==relon.r) return rel;
        }
        return null;
    }


    /**
     * make the new param that depended on another param to exist.
     * We can do this on-demand whenever this neuron comes into "mind".
     */
    makeCDXParamIfNotExists=function($nid, $fulfillmentRelon, $relon_pat) {
        console.log("makeCDXParam "+$nid);
        var nObj = MEMG.getNeuronByNid($nid);
        var to_t = $fulfillmentRelon.t;
        var to_s = $fulfillmentRelon.to_s;
        var relPatArr = [{prop:'type', val:TYPE_PARAM}, 
                            {prop:'r', val:$relon_pat.r}, 
                            {prop:'t', val:to_t}];
        console.log(relPatArr);
        var tempParamPat = MEMG.getRelByPat(nObj.outs, relPatArr);
        if (tempParamPat==null) {
            console.log("MAKE VAR PARAM!");
            relPatArr.push({from_nid:$nid});
            console.log("make param for:"+nObj.nid);
            var RNOBJ_AUDIT = MEMG.getNeuronByNid($relon_pat.reln);
            console.log("reln audit:"+RNOBJ_AUDIT.text);
            console.log("to_s audit:"+to_s);
            var to_type = TO_TYPE_ANNE;
            var newRelon = MEM.makeTempRelon(nObj, TYPE_PARAM, $relon_pat.reln, to_type, to_s, true);
            if ($relon_pat.required==true) newRelon.required=true;
            console.log(newRelon);
            var relpatrow = document.querySelector('rowrid_'+$relon_pat.rid);
            if (relpatrow!=null) ANNE.A.hide(relpatrow);
            return newRelon;
        }
        return null;
    }


    static refreshRowFromRelonUpdate = async($relon)=> {
        var relon = MEMG.lookupRelonByTXID(txid);
        if (relon==null) return;

        var row_els = $el.querySelectorAll('[txid="'+relon.txid+'"]');

        //if we have multiple tables with this entry open, this will still work.
        for (var e=0; e < row_els.length; e++) {
            var row_el = row_els[e];

            var tbl = ANNE.A.getFirstParentOf(row_el, "anne-tbl");
            if (tbl==null) return;
    
            //extract the guid.
            var widget_id = tbl.id.substring(9);    //anne_tbl_ xxxxx

            var ANNE_RELON_TBL = NF.getINSTbyGuid(widget_id);
            if (ANNE_RELON_TBL==null) return;

            ANNE.A.removeSelf(row_el); //remove the row from the GUI TABLE.

            //add the row back to the gui table.
            var newrow = await ANNE_RELON_TBL.addDataToColsInit([relon], true);
            newrow.classList.add("just-added");
        }
    }

    

    /**
     * Add data to cols.
     * 
     * We pass the data here because we also use this function for new single adds.
     * @author {Mr.Scatman}
     */
    addDataToColsInit=async($data, $bPREPEND=false, $callback=null)=> {
        console.log("data in table:"+$data.length);
        var dataNOBJ = MEMG.getNeuronByNid(this.nid);
        if (dataNOBJ==null) {
            alert("this can't be null here");
            return;
        }

        var thisTnid = dataNOBJ.temp_nid;
        var prepends = [];
        
        for (var d=0; d < $data.length; d++) {
            var bINHERITED = false;
            var dObj = $data[d];
            if (dObj.active==false) {
                console.log("skipping relon", dObj);
                continue;   //this was set in the helpers (to hide non-active types)
            }
            var relonObj = dObj.rObj;
            if (relonObj==null) {
                console.log("relon empty, skipping");
                continue;       //we must have a relon here.
            }
            if (relonObj.f!=thisTnid) {
                bINHERITED=true;
                console.log("INHERITED");
            }

            var newrow = document.createElement("div");
            newrow.classList="row";
            if (relonObj.status_text!=undefined && relonObj.status_text=="p") {
                newrow.classList.add("pending-collab");
                newrow.classList.add("rowtxid_"+relonObj.txid);
            }
            else newrow.classList.add("rowrid_"+relonObj.rid);

            if ($bPREPEND==false) this.el_tblbody.append(newrow);
            else {
                this.el_tblbody.prepend(newrow);
                newrow.classList.add("just-added");
            }

            newrow.setAttribute("txid", relonObj.txid);
            newrow.setAttribute("type", parseInt(relonObj.type));

            //highlight the required relon.
            var bDOINGREQUIRED = false;
            var adminMsgOverride = "";

            if (relonObj.type==TYPE_PARAM) {
                if (relonObj.required!=undefined && relonObj.required==true) {
                    newrow.classList.add("relon-required");
                    bDOINGREQUIRED = true;
                }
                
                if (relonObj.depends_on!=undefined) {
                    var fulfillmentRelon = this.getParamVarFulfillment(this.nObj, relonObj.depends_on);
                    if (fulfillmentRelon==null) {
                        newrow.classList.add("relon-dependson");
                        //if we get a param fulfilled of this txid we can do something 
                        //automatcially in the DOM
                        newrow.setAttribute("dotxid", relonObj.depends_on);
                        adminMsgOverride = "do "+LINKS.makeRelonLink(relonObj.depends_on, this.relonLinkType)+" first.";
                    }

                    //for already-made relons on reload of app or
                    //revisiting this neuron for the first time this session.
                    //(we do not store the cdx param relon extrapolation.. we run-time create it on-demand)
                    else {
                        var newRel = this.makeCDXParamIfNotExists(this.nid, fulfillmentRelon, relonObj);
                        if (newRel) {
                            if (relonObj.required!=undefined) newRel.required = relonObj.required;
                            if (relonObj.prompt!=undefined) newRel.prompt = relonObj.prompt;
                            prepends.push({active:true, rObj:newRel});
                        }
                        dObj.active = false;
                        ANNE.A.removeSelf(newrow);
                        continue;
                    }
                }
            }

            for (var c=0; c < this.cols.length; c++) {
                var col = this.cols[c];
                if (col.active==false) continue;
                var datacol = document.createElement("div");
                datacol.classList = "column cell";
                datacol.classList.add(col.f);
                
                if (col.style_class!=undefined) datacol.classList.add(col.style_class);

                var dataVal = relonObj[col.dataprop];
                var display = "";
                
                if (col.dataprop) datacol.setAttribute("dp", col.dataprop);

                newrow.append(datacol);

                if (col.ttype=="**ADMIN**") {
                    datacol.setAttribute("dp", "admin");
                    if (adminMsgOverride!="") {
                        //alert("admin override: "+adminMsgOverride)
                        datacol.innerHTML = "<div class='col-span'>"+adminMsgOverride+"</div>";
                        continue;
                    }

                    //no admin actions for frozen neurons.
                    if (dataNOBJ.freeze!=undefined && dataNOBJ.freeze==1) {
                        console.log("THIS NEURON IS FROZEN (aka locked)")
                        continue;
                    }
                    
                    //EXT TO DO: must make real events not injected js ones..
                    //await this.doAdminCol(newrow, datacol, dataVal, relonObj, dataNOBJ);
                    continue;
                }

                else if (col.dataprop=="type") {
                    display = TALK.typeTalk(dataVal);
                }
                else if (col.dataprop=="to_type") {
                    display = "<span class='text-white-soft'>"+TALK.toTypeTalk(dataVal)+"</span>";
                }
                else if (col.ttype=="anne") {
                    var nObj = MEMG.getNeuronByNid(dataVal);
                    EXT_BTNS.makeNeuronLink(datacol, nObj.nid, "col-span");
                    continue;
                    //display = "<span class='col-span'>"+LINKS.makeNeuronLink(nObj, this.neuronLinkType)+"</span>";
                }
                else if (col.ttype=="misc") {
                    if (col.dataprop=="to_s" && relonObj.to_type==TO_TYPE_ANNE) {
                        var nObj = MEMG.getNeuronByNid(dataVal);
                        ANNE.MEM_UDF.makeNeuronLink(datacol, nObj.nid, "col-span");
                        continue;
                    }
                    else {
                        if (relonObj.to_type==TO_TYPE_URL) {
                            display = "<a href='"+relonObj.to_s+"' class='link-blue' target='_new'><i class='ai-external-link-alt'></i>"+relonObj.to_s+"</a>";
                        }
                        else if (relonObj.to_type==TO_TYPE_TXID) {
                            //display =  "<span class='col-span'>"+LINKS.makeRelonLink(relonObj.to_s, this.relonLinkType)+"</span>";
                            ANNE.MEM_UDF.makeRelonViewLink(datacol, relonObj.to_s);
                            continue;
                        }
                        else {
                            display = "<span class='text-whiteF-lbl'>"+relonObj.to_s+"</span>";
                        }
                    }
                }

                //this will gen the txid ico and the relevant ghost ico (if needed)
                else if (col.ttype=="txid") {
                    var inh = "";
                    if (bINHERITED==true) inh = "inherited-relon";
                    ANNE.MEM_UDF.makeRelonViewLink(datacol, relonObj.txid, false, 8, true, inh);
                    continue;
                    //display =  "<span class='col-span'>"+LINKS.makeRelonLink(relonObj.txid, this.relonLinkType)+inh+"</span>";
                }

                else if (col.display!=undeined) {
                    display = relonObj[col.display];
                }

                if (display==undefined) display = dataVal;
                
                if (dataVal!=undefined && dataVal!=null) {
                    datacol.setAttribute("dataz", dataVal);
                }
                else console.log("UT OH");
                datacol.innerHTML = display;
            }

            if (dObj.active==false) newrow.display="none";
        }

        if (prepends.length) {
            await this.addDataToColsInit(prepends, true);
        }

        if ($callback!=null && typeof $callback==="function") $callback();
        return newrow;
    }


    /**
     * @author {Mr.Scatman} 
     */
    addColumnHeader = function($col, $col_index) {
        if ($col.filter==true) {
            var div = document.createElement("div");
            div.classList = "column";
            div.classList.add($col.f);
            if ($col.extra_class!=undefined) div.classList.add($col.extra_class);
            div.setAttribute("colnum", $col_index);
            div.setAttribute("tbl", this.tbl_divid);
            this.el_rowsearch.append(div);
            var input = document.createElement("input");
            input.classList = "tbl-filter";
            input.setAttribute("type", "text");
            input.setAttribute("placeholder", "filter the "+$col.text);
            div.append(input);
            input.addEventListener('input', $e=>relon_table.DOM_filterEvent($e))
        }
        else {
            var div = document.createElement("div");
            div.classList = "column";
            div.classList.add($col.f);
            if ($col.extra_class!=undefined) div.classList.add($col.extra_class);
            div.setAttribute("colnum", $col_index);
            div.setAttribute("tbl", this.tbl_divid);
            this.el_rowsearch.append(div);
        }

        //create the header cols in the header row
        var divh = document.createElement("div");
        divh.classList = "column";
        divh.classList.add($col.f);
        divh.setAttribute("colnum", $col_index);
        divh.setAttribute("tbl", this.tbl_divid);
        this.el_rowheader.append(divh);

        var sortIco = ANNE.A.ezSortCol(divh, $col, this.tbl_divid, "asc");
        sortIco.addEventListener("click", $e=> this.sortIcoClick($e));
        $col.sortIco = sortIco;
        return input;   //RETURN THE TBL FILTER
    }

    /**
     * The sort ico has been clicked
     * @author {Mr.Scatman}
     */
    sortIcoClick = function($e) {
        var el = $e.target;
        var sortDataType = el.getAttribute("prop");
        console.log(sortDataType);

        //parent child index is the column index of this header.
        //this will inform the DOM sort what column to pay attention to.
        var col_index = ANNE.A.getFirstParentAttr(el, "column", "colnum");
        var tbl = el.getAttribute("tbl");
        var isActive = el.classList.contains("activesort");
        if (isActive==null) isActive=false;
        var sort = el.getAttribute("sort");
        console.log("current sort:"+sort+" active:"+isActive);
        ANNE.A.removeAddClassOfAllIn(tbl, ".sortico", "activesort", el);
        var sort = relon_table.toggleSortIco(el, sortDataType, isActive, sort);
        
        //var iFix = (i*2) + 1;
        relon_table.DOM_sortTbl(tbl, col_index, sortDataType, sort);
    }

    defaultSort = function($el, $sort="asc") {
        var sortDataType = $el.getAttribute("prop");
        //console.log(sortDataType);

        //parent child index is the column index of this header.
        //this will inform the DOM sort what column to pay attention to.
        var col_index = ANNE.A.getFirstParentAttr($el, "column", "colnum");
        var tbl = $el.getAttribute("tbl");
        var isActive = $el.classList.contains("activesort");
        if (isActive==null) isActive=false;
        var sort = $el.setAttribute("sort", $sort);
        //console.log("current sort:"+sort+" active:"+isActive);
        ANNE.A.removeAddClassOfAllIn(tbl, ".sortico", "activesort", $el);
        var sort = relon_table.toggleSortIco($el, sortDataType, isActive, sort);
        relon_table.DOM_sortTbl(tbl, col_index, sortDataType, sort);
    }


    /**
     * @author {Mr.Scatman} 
     */
    deleteColumnHeader = function($col_index) {
        var searchheader = this.el_rowsearch.querySelectorAll('div[colnum="'+$col_index+'"]')[0];
        searchheader.remove();
        var colheader = this.el_rowheader.querySelectorAll('div[colnum="'+$col_index+'"]')[0];
        colheader.remove();
    }


    /**
     * @author {Mr.Scatman} 
     */
    addColumnToRows = function($col) {
        for (var d=0; d < this.data.length; d++) {
            if (this.data[d].active==false) continue;
            var relonObj = this.data[d].rObj;
            var rowdiv = this.el_tblbody.querySelectorAll('div[txid="'+relonObj.txid+'"]')[0];
            if ($col.active==false) break; //this should be active
            var datacol = document.createElement("div");
            datacol.classList = "column cell";
            datacol.classList.add($col.f);
            var dataVal = relonObj[$col.dataprop];
            if ($col.ttype=="anne") {
                var nObj = MEMG.getNeuronByNid(dataVal);
                display = GUI_STUFF.makeANNENeuronLink(nObj, 64, true);
            }
            else var display = relonObj[$col.display];
            datacol.setAttribute("dataz", dataVal);
            datacol.innerHTML=display;
            rowdiv.append(datacol);
        }
    }


    /**
     * @author {Mr.Scatman} 
     */
    deleteColumnFromRows = function($col) {
        const tblbody = this.tbl.querySelector('.anne-tbl-body-overflow');
        for (var d=0; d < this.data.length; d++) {
            if (this.data[d].active==false) continue;
            var relonObj = this.data[d].rObj;
            var rowdiv = tblbody.querySelectorAll('div[txid="'+relonObj.txid+'"]')[0];
            var datacol = rowdiv.children[$col.col_index];
            if (datacol!=null) datacol.remove();
        }
    }


    /**
     * @author {Mr.Scatman} 
     */
    initEvents = function($tbldivid) {
        return;
        //these jack with the formatting too much. not worth it.
        //const fxwrapper = tbl.querySelector(".anne-tbl-fx-wrapper"); //we can use this for bigger FX
        const tbl = document.getElementById($tbldivid);
        const footer = document.querySelector(".anne-tbl-footer");
        const downs = tbl.querySelector(".anne-downloadables");
        const ups = tbl.querySelector(".anne-uploadables");

        footer.addEventListener("mouseenter", (e) => {
            tbl.style.transition = "none";
            downs.style.transition  =  "all 0.5s ease-in";
            downs.style.transform = "translateX(-180px) translateY(20px)  scale(1.4)";
            ups.style.transition  =  "all 0.5s ease-in";
            ups.style.transform = "translateX(-100px) translateY(20px) scale(1.4)";
            footer.style["height"] = "64px";
            footer.style.transition  =  "all 0.5s ease-in";
        });

        footer.addEventListener("mouseleave", (e) => {
            downs.style.transition  =  "all 0.3s ease-out";
            downs.style.transform = "translateX(0px) translateY(0px)  scale(1)";
            ups.style.transition  =  "all 0.3s ease-out";
            ups.style.transform = "translateX(0px) translateY(0px) scale(1)";
            footer.style["height"] = "32px";
            footer.style.transition  =  "all 0.3s ease-out";
        });
    }


    /**
     * @author {Mr.Scatman} 
     */
    static DOM_sortTbl=function($divid, $col_index, $sortDataType, $sortDir) {
        var tbl = document.getElementById($divid); 
        if (tbl==null) return;
        const rows = tbl.querySelectorAll(".row");
        var masterRowParent = rows[0].parentElement;
        var rowsData = Array.prototype.slice.call(rows, 0);
        var dir = 1;
        if ($sortDir=="asc") dir *= -1;

        //sort them  (keep these separated even tho they are the same.. for now)..
        //we may need to handle differently..
        if ($sortDataType=="string" || $sortDataType=="anne") {
            rowsData.sort(function(a,b) {
                if (a.children[$col_index]==undefined) return 0;
                var txta = a.children[$col_index].innerHTML;//getAttribute("dataz");//textContent;
                if (b.children[$col_index]==undefined) return 0;
                var txtb = b.children[$col_index].innerHTML;//getAttribute("dataz");
                if(txta > txtb) { return -1*dir; }
                if(txta < txtb) { return 1*dir; }
                return 0;
            });
            
        }
        else if ($sortDataType=="numeric") {
            rowsData.sort(function(a,b) {
                if (a.children[$col_index]==undefined) return 0;
                var txta = (a.children[$col_index].innerHTML);//.getAttribute("dataz"));
                if (b.children[$col_index]==undefined) return 0;
                
                var txtb = (b.children[$col_index].innerHTML);//.getAttribute("dataz"));
                if(txta > txtb) { return -1*dir; }
                if(txta < txtb) { return 1*dir; }
                return 0;
            });
        }

        else {
            rowsData.sort(function(a,b) {
                if (a.children[$col_index]==undefined) return 0;
                var txta = (a.children[$col_index].innerHTML);//.getAttribute("dataz"));
                if (b.children[$col_index]==undefined) return 0;
                
                var txtb = (b.children[$col_index].innerHTML);//.getAttribute("dataz"));
                if(txta > txtb) { return -1*dir; }
                if(txta < txtb) { return 1*dir; }
                return 0;
            });
        }

        //put them back in.
        rowsData.forEach(row => {
            masterRowParent.appendChild(row)
        });
    }


    /**
     */
    static DOM_filterEvent = async($e)=> {
        var el = $e.target;
        var text = el.value;
        await relon_table.DOM_Filter(el, text); 
    }


    /**
     */
    static DOM_Filter=async($el, $text)=> {
        const col = ANNE.A.getFirstParentOf($el, "column");
        const tbldivid = col.getAttribute("tbl");
        const childindex = col.getAttribute("colnum");
        const tbl = document.getElementById(tbldivid);
        //console.log("filter tbl:"+tbl+" col "+childindex+" text "+$text);

        //get all the rows, turn off any rows where the colIndex doesn't match the text.
        const ID = tbldivid.split("_").reverse()[0].trim();

        var max_rows = 0;
        var on_cnt = 0;

        //var pgui = NF.getPguiByGUID(ID);

        //turn on all rows
        tbl.querySelectorAll(".row").forEach(row => {
            row.style.display="flex";
            max_rows++;
        });

        const tblFilters = tbl.querySelectorAll(".tbl-filter");
        for (var t=0; t < tblFilters.length; t++) {
            var tblfilter = tblFilters[t];
            var cIndex = ANNE.A.getFirstParentAttr(tblfilter, "column", "colnum");
            
            if (tblfilter.value.length) {
                var on_cnt = relon_table.DOM_Filter_single(tblfilter, tblfilter.value, cIndex);
                
                //check the last filter (the net sum of all visible/not visible)..
                //it will return the on_cnt.. 
                if (tblFilters.length-1==t) {
                    max_rows = max_rows-on_cnt;
                }
            }
        }

        var rowslbl = ANNE.TALK.pluralizeNounIf2(on_cnt, "row", "rows");
        ANNE.A.setHTML("tbl_summary_"+ID, on_cnt+" "+rowslbl);
    }



    /**
     */
    externalDataChange = function ($tbl_el, $types) {
        var max_rows = 0;
        var on_cnt = 0;

        var tbl_el = typeof $tbl_el === "string" ? document.getElementById($tbl_el) : tbl_el = $tbl_el;
        if (tbl_el==null) return null;
        var ID = $tbl_el.id;
        
        //turn on all rows
        tbl_el.querySelectorAll(".row").forEach(row => {
            var type = parseInt(row.getAttribute("type"));
            if ($types.indexOf(type)!=-1) row.style.display="flex";
            else row.style.display="none";
            max_rows++;
        });


        const tblFilters = tbl_el.querySelectorAll(".tbl-filter");
        for (var t=0; t < tblFilters.length; t++) {
            var tblfilter = tblFilters[t];
            var cIndex = ANNE.A.getFirstParentAttr(tblfilter, "column", "colnum");
            
            if (tblfilter.value.length) {
                var on_cnt = relon_table.DOM_Filter_single(tblfilter, tblfilter.value, cIndex);
                
                //check the last filter (the net sum of all visible/not visible)..
                //it will return the on_cnt.. 
                if (tblFilters.length-1==t) {
                    max_rows = max_rows-on_cnt;
                }
            }
        }

        var rowslbl = ANNE.TALK.pluralizeNounIf2(on_cnt, "row", "rows");
        ANNE.A.setHTML("tbl_summary_"+ID, on_cnt+" "+rowslbl);
    }


    static DOM_Filter_single=function($el, $text) {
        const col = ANNE.A.getFirstParentOf($el, "column");
        const tbldivid = col.getAttribute("tbl");
        const childindex = col.getAttribute("colnum");
        const tbl = document.getElementById(tbldivid);
        console.log("filter tbl:"+tbl+" col "+childindex+" text "+$text);

        var regex = new RegExp($text, 'i');
        var bEXACT = false;
        var onCnt = 0;

        tbl.querySelectorAll(".row").forEach(row => {
            var dataz = row.children[childindex].innerHTML;
            if (row.style.display=="none") return;  //skip row, go to next.
            if (dataz==null) row.style.display="none";
            else if (bEXACT==false) {
                if(regex.test(dataz)==false) row.style.display="none";
                else {
                    row.style.display="flex";
                    onCnt+=1;
                }
            }
            else if (dataz !== $text) row.style.display="none";
            else {
                row.style.display="flex";
                onCnt+=1;
            }
        });
        return onCnt;
    }


    /**
     * toggleSortIco 
     * If bISACTIVE=false then do NOT toggle the sort.. just apply the       
     * .activesort class.  If it is active we will leave it active (no need  
     * to apply the activesort class) but we will flip the ico and the dir.
     */
    static toggleSortIco = function($el, $sortDataType, $bISACTIVE, $dir="asc") {
        //console.log("toggleSortIco sort coming set to dir:"+$dir+" active:"+$bISACTIVE);
        if ($bISACTIVE==false) {
            $el.classList.add("activesort");
            //console.log("toggleSortIco returning dir:"+$dir);
            return $dir;
        }
        if ($dir=="asc") {
            ANNE.A.setAttribute($el, "sort", "desc");
            if ($sortDataType=="string" || $sortDataType=="anne") ANNE.A.addRemoveClass($el, "ai-sort-alpha-up", "ai-sort-alpha-down"); 
            else if ($sortDataType=="number") ANNE.A.addRemoveClass($el, "ai-sort-amount-up", "ai-sort-amount-down"); 
            else ANNE.A.addRemoveClass($el, "ai-sort-up", "ai-sort-down"); 
            console.log("toggleSortIco returning dir:desc");
            return "desc";
        } 
        else {
            ANNE.A.setAttribute($el, "sort", "asc");
            if ($sortDataType=="string" || $sortDataType=="anne") ANNE.A.addRemoveClass($el, "ai-sort-alpha-down", "ai-sort-alpha-up"); 
            else if ($sortDataType=="number") ANNE.A.addRemoveClass($el, "ai-sort-amount-down", "ai-sort-amount-up"); 
            else ANNE.A.addRemoveClass($el, "ai-sort-down", "ai-sort-up"); 
            console.log("toggleSortIco returning dir:asc");
            return "asc";
        }
    }

    
    /** 
     * Display a relons table.
     * 
     * hijack the relons content table to display relons
     * These can be edited and created.
     *
     * @author {Mr.Scatman}
     */
     static initRelonsTblDisplay=async ($parent_pgui, $displaydivid, $ID="demouniqueid", $new_relon_injection_divid="new_relon_injection",$pObj=null, $part_page="") => {
        var acsEl = document.getElementById($displaydivid+'_'+$ID);
        var acceptable_outs = [];
        var nid = "";
        var targetNOBJ=null;

        //NEED TO GET ALT-INHERITED LATERAL BES.
        //either from BE_CHILDREN/BE_PARENT
        //or an alt BE..  (so any TYPE_BE alt TO_TYPE_ANNE).

        if ($parent_pgui.pObj.p1!=USER._userNObj.nid || ($pObj!=null && $pObj.hijackp1!=undefined)) {
            if ($pObj!=null && $pObj.hijackp1!=undefined) nid=$pObj.hijackp1;
            else nid = $parent_pgui.pObj.p1;

            targetNOBJ = MEMG.getNeuronByNid(nid);
            
            //if this is an identity, re-route this to our identity display which 
            //has some subtle nuances.
            if (MEM.tempBeIsA(targetNOBJ.temp_nid, MEM.identity.temp_nid, NTYPE_INST)==true) {
                var basic_types_other_identity =  [TYPE_HAS, TYPE_BE, TYPE_CLUMP, TYPE_ROLE];
                acceptable_outs = MEMG.prepForIdentity(targetNOBJ, basic_types_other_identity);
            }

            //TO DO ---
            //WE NEED TO INHERIT CLUMPS that specify HAS type info..
            //If a class defines a HAS via a clump, we need to get it!
            else {
                
                //parent line inherits
                var bADDSELF = true;
                var bEXTRAPRETURN = true;
                acceptable_outs = MEMG.getParentLineRelons(targetNOBJ, [TYPE_SYS, TYPE_BE, TYPE_PARAM, TYPE_HAS, TYPE_CLUMP, TYPE_ROLE], [], bADDSELF, bEXTRAPRETURN);
            }
        }


        //used for self-identity
        else {
            targetNOBJ = MEMG.getNeuronByNid($parent_pgui.pObj.p1);
            acceptable_outs = MEMG.prepForIdentity(targetNOBJ);
        }

        return await relon_table.buildRelonContentTbl($parent_pgui, acsEl, targetNOBJ.nid, [0,1,2,3,5,6,7,8,9], acceptable_outs, $new_relon_injection_divid, $pObj, $part_page);
    }


    /**
     * Build a relons content table.
     * 
     * @author {Mr.Scatman}
     */
    static buildRelonContentTbl=async ($parent_pgui, $domEl, $nid, $types=[0,1,2,3,5,6,7,8,9], $useOnlyTheseOuts=[], $newRelonEl="display_data", $pObj=null, $part_page="") => {
        var nObj = MEMG.getNeuronByNid($nid);
        var bUS = USER.isUsOrOwnedByUs(nObj.nid);
        var outs = $useOnlyTheseOuts;

        if (outs.length==0) outs = nObj.outs;
        if (outs.length==0) {
            var el=document.createElement("div");
            el.classList = "no-more";

            el.innerHTML='<span><i class="ai-exclamation-triangle"></i>&nbsp;There are no additional relons in the system for this neuron</span>';
            $domEl.append(el);
            return;
        }

        //do our best to have basic info for all the parts of the relons.
        var done = await MEMG.extractAllFromRelonsAndChaseMin(outs,true);

        var finalRelons=[];
        var _usedTXIDs=[];
        var types=[0,1,2,3,5,6,7,8,9];

        if ($pObj!=null && $pObj.misc!=null && $pObj.misc.types!=undefined) {
            types = $pObj.misc.types;
            console.log(types);
        }

        console.log($pObj);

        //ANNE.TALK.printRelons(outs);

        for (var r=0; r < outs.length; r++) {
            var relon = outs[r];
            if (_usedTXIDs.indexOf(relon.txid)!=-1) continue;

            //skip any PARAM relons we already have supplied a HAS for...
            //for now multiples_ok is a hack prop we hardcoded to set for easy access.
            //to do it right (dynamically) we will need to encode the multiples_ok into a relon
            //and process it -- and set the flag in the mem processing whenever we find
            //a relon pointing from some class to a TXID with the multiples_ok reln.
            //ie,   class->multiples_ok->txid 22o34u2o3oi23uo2i3u42o
            //txid 22o34u2o3oi23uo2i3u42o  =  PARAM class->annothers->class 
            if (relon.type==TYPE_PARAM && (relon._multiples_ok==undefined || relon._multiples_ok==false)) {
                var toNObj = MEMG.getNeuronByNid(relon.to_s);
                var bWEHAVEIT=false;

                //if no garden is set, this is not us, and this is the garden relon,
                //skip it.  WE can't collab on no garden classes!
                //if (bUS==false && nObj._gt==null && relon.r==MEM.garden_type.temp_nid) continue;
                
                bWEHAVEIT = MEM.hasMetClassDefn(outs, relon, toNObj);

                var relnNOBJ = MEMG.getNeuronByTempNid(relon.r);
                
                if (bWEHAVEIT == true) {
                    continue;
                }
            }

            _usedTXIDs.push(relon.txid);
            var active = false;
            if (relon.r==MEM.name.temp_nid) {
                console.log("test");
            }

            if (types.indexOf(relon.type)!=-1) active = true;
            finalRelons.push({rObj:relon, active:active});
        }

        //gets relons from collab space (MEM_UDF._collab_relons)
        var andy_relons = MEM_UDF.getAndyTblRelonsForNid($nid);
        if (andy_relons!=null) {
            finalRelons = andy_relons.concat(finalRelons);
        }

        //pass the relons we want into the RELONTBL..
        var paramsObj=new page_params(null, null, null, $nid, finalRelons, bUS, $newRelonEl);
        var div = document.createElement("div");
        div.id = "relons_tbl_"+$domEl.id;
        console.log(div.id);
        div.style['min-width']="100%";
        $domEl.append(div);
        paramsObj.injection_div = div.id;
        paramsObj.randSeed = UTILS.randomHex(4);

        //default sort -- (we could pass this through the pObj if we wanted..
        //to override this default)
        if (paramsObj.misc==null) paramsObj.misc = {sort:"asc", dataprop:"type"};
        else {
            paramsObj.misc.sort = "asc";
            paramsObj.misc.dataprop = "type";
        }

        //transfer over any important props for this load.
        //we may have things like .strip_all, .init_cols, etc.
        if($pObj!=null) {
            var keys = Object.keys($pObj);
            for(var k=0; k < keys.length; k++) {
                var key = keys[k];
                paramsObj[key] = $pObj[key];
            }
        }

        if ($part_page!="") await $parent_pgui.loadPart($part_page, paramsObj, false, true);
        else return paramsObj;
    }



    /** 
     * Delete a relon.
     * 
     * Typically invoked from clicking on a X in a display column of some 
     * kind.  What we do differs depending on what type of relon this is.  
     * 
     * Ghosts are deleted from the ghost DB..  real TXIDS create inhibits.                                                                  *
     * 
     * Inhibits of this variety must be inhibited from the from...        
     * For self identity this is easy, the logged in user is the from.      
     * For neurons they own (like a favorite list) THAT from must be used   
     * to send the inhibit in order for it to be respected/honored in ANNE. 
     * 
     * @param {string} - the long string txid of the relon we want to delete.
     * @author {Mr.Scatman}
     * 
     */
    static deleteThisRelon=async ($txid) => {
        console.log($txid);
        console.log("delete relon here");
        console.log("remove the relon from the gui");
        var rObj = MEMG.lookupRelonByTXID($txid);
        var text = "";
        if (rObj == null) {
            //chase txid.. we should have it, but let's try again anyhow.
            rObj = await CHASE.getANNETX($txid);
            if (rObj == null) return;
        }

        if (ANNE.TALK != null)  text = ANNE.TALK.printRelon(rObj,true);
        var bGHOSTLOCALDELETE = false;;
        var msg = "";
        if (rObj.ghost == true) {
            bGHOSTLOCALDELETE = true;
            msg = "Are you sure you want to delete this GHOST from local memory?";
        }
        else {
            msg = "Are you sure you want to delete this from ANNE?";
            if (rObj.temp == false) msg = msg+"  We must send a delete transaction to ANNE to delete GUI_STUFF.";
        }

        var r = confirm(text+"<BR>"+msg);
        if (r == false) return;

        //if this is temp, remove it from the Q and the present display table.
        //the user has changed their mind.
        if (rObj.temp == true) {
            ANNE.A.removeSelf('.rowrid_'+rObj.rid);     //remove the rid anywhere it is displayed

            //this is for displayed relons that were created by the user,
            //then they decided to uncreate them.
            //this shows up the option to add the same thing once again.
            if (rObj.param_rid!=undefined) ANNE.A.fadeIn(".rowrid_"+rObj.param_rid);

            //delete it from ghost db tables if this is OUR ghost.
            if (bGHOSTLOCALDELETE) {
                if(ANNE.DB_local!=null) await ANNE.DB_local.delete([rObj]);
            }
            await SENDQ.deleteItemsByTXID(rObj.txid,rObj);
            return;
        }


        //NEW METHOD IN ANY TABLE WITH DELETES (for non temp items, ie, existing in mem and real)
        //JUST FADE OUT THE RELON
        //ON UNDO, we fade back in via the ANNE._queDelete
        //ON SUCCESSFUL actual delete, we search for, and remove, any of these rows in ANNE._postProcessSentTXIDS
        ANNE.A.fadeOut('.rowrid_'+rObj.rid);

        var fromNOBJ=null;
        if (rObj.from_nid == USER._userNObj.nid) fromNOBJ = USER._userNObj;
        else {
            fromNOBJ = MEMG.getOrMakeIfNew(rObj.from_nid);
            if (ANNE.CHASE!=null) var done = await CHASE.queryForOuts(fromNOBJ,true,[0,1]);
        }
        
        //inhibit or arbor prune??  //_makeArborPrune
        await fromNOBJ._makeArborPrune($txid);
    }
}

/************** (from file ./js/gui/widgets/anne_relon_table.js)**/ 

/**
 * This is a quick method for table display..
 * Below this class is the DSTBL obj which is a different method.
 */
class neuron_table {
    constructor(spObj, $bDOHEADING=true, $heading_prop="title", $dcf=".smart-scroll-framer", $YBUF=52, $max_bottom=0) {
        this.pgui = spObj;
        this._guid = "000000";
        if (spObj!=null) this._guid=spObj._guid;
        this.el_update_cnt = null;

        this.do_heading = $bDOHEADING;
        this.heading_prop = $heading_prop;
        this.records = [];
        this.cols = {};
        this.bPAGINATOR=true;
        this.bDOPAGINATORX = false;

        this.PAGINATOR = null;
        this.default_class_framer = $dcf;
        this.YBUF = $YBUF;
        this.heading_col_class = "";
        this.tbl_row_class = "anne-table-row";
        this.neuron_heading_row_class= "neuron-tbl-heading";

        var el_page = null;
        if (this.default_class_framer!=null && this.default_class_framer!="") {
            el_page = document.querySelector(this.default_class_framer);
        }

        var el_loader = document.querySelector("load-wrapper");

        if($max_bottom!=0) this.max_bottom=$max_bottom;
        if (el_page==null || el_loader!=null) {
            console.warn("this will be blasing all records at once!, turnign off paginator NO default_class_Framer found");
            this.max_bottom =  0;
            this.bPAGINATOR=false;
        }
        else {
            var bcr = el_page.getBoundingClientRect();
            this.max_bottom = bcr.bottom;
        }

        this.rankCol = {
            neuron_link:false,
            title:"",
            class1:"mw40",
            prop:"rank",
            sort_prop:"rank",
            sort_type:"numeric"
        }

        this.bDOCELLCLASSES = false;
        this.bCOLORSQARES = false;

        this.bDOBORDER = true;

        this.el_heading_row = null;
        this.el_tbl_body = null;
        this.el_paginator = null;

        this._ROW_MAX_H = 30;
        this.primaryKey = null;

        this.cb_loaded_chk = null;
        this.data_id = null;
        this.cb_rowclicker = null;
        this.data_arr_map = null;

        this._nopagrender_cb = null;

        this.INST = null;
    }

    setINST = function($INST) { this.INST = $INST; }
    setNoPaginatorRender=function($cb) {
        this._nopagrender_cb=$cb;
    }
    setTableRowClass=function($trc) {
        this.tbl_row_class = $trc;
    }
    setNeuronHeadingClass=function($c) {
        this.neuron_heading_row_class= $c;
    }
    
    setPrimaryKey = function($primaryKey) {
        this.primaryKey = $primaryKey;
    }

    wipeTableEls = function() {
        ANNE.A.removeSelf(this.el_heading_row);
        ANNE.A.removeSelf(this.el_tbl_body);
        ANNE.A.removeSelf(this.el_paginator);
        this.el_heading_row = null;
        this.el_tbl_body = null;
        this.el_paginator = null;
    }

    /**
     * 
     * @param {*} $newObj 
     */
    newDataInTime = function($newObj) {
        console.log("table:: newDataInTime ", $newObj);
        if (this.primaryKey!=null) {
            var pk = $newObj[this.primaryKey];
            if (pk==null) {
                console.log("pk missing on obj, skipping")
            }
            if (pkcheck[ pk] !=null) {
                console.log("pk violation, skippping");
            }
            pkcheck [ pk ]  = true;
        }
        $newObj.active = true;
        $newObj.rank="#"+(this.records.length+1);
        this.records.push($newObj);

        if (this.PAGINATOR!=null) this.PAGINATOR.newDataInTime($newObj);
    }



    /**
     * For manual set we just pass the objects and add
     * an active prop
     */
    setManualCols = function($cols, $objs) {
        this.cols = $cols;

        var pkcheck = new Object();
        if ($objs.length>0) {
            for (var x=0; x < $objs.length;x++) {
                var obj = $objs[x];
                if (this.primaryKey!=null) {
                    var pk = obj[this.primaryKey];
                    if (pk==null) {
                        console.log("pk missing on obj, skipping")
                        continue;
                    }

                    if (pkcheck[ pk] !=null) {
                        console.log("pk violation, skippping");
                        continue;
                    }

                    //just turn it on, only one can use this key.
                    pkcheck [ pk ]  = true;
                }

                obj.active = true;
                obj.rank="#"+(x+1);
                this.records.push(obj);
            }
        }
    }



    /**
     * For manual set we just pass the objects and add
     * an active prop
     */
    setManualColsX = function($cols, $arrmap) {
        this.cols = $cols;
        this.data_arr_map = $arrmap;
        var pkcheck = new Object();
        if ($arrmap.arr.length>0) {
            for (var x=0; x < $arrmap.arrlength;x++) {
                var obj = $arrmap.arr[x];
                if (this.primaryKey!=null) {
                    var pk = obj[this.primaryKey];
                    if (pk==null) {
                        console.log("pk missing on obj, skipping")
                        continue;
                    }
                    if (pkcheck[ pk] !=null) {
                        console.log("pk violation, skippping");
                        continue;
                    }
                    //just turn it on, only one can use this key.
                    pkcheck [ pk ]  = true;
                }
                obj.active = true;
                obj.rank="#"+(x+1);
                this.records.push(obj);
            }
        }
    }


    setManualColsX2 = function($cols, $arrmap) {
        this.cols = $cols;
        this.data_arr_map = $arrmap;
        if ($arrmap.arr.length>0) {
            for (var x=0; x < $arrmap.arrlength;x++) {
                var obj = $arrmap.arr[x];
                obj.active = true;
                this.records.push(obj);
            }
        }
    }


    /**
     * 
     * @param {*} $objs 
     */
    updateRecordsAndRefresh = function($objs) {
        console.log("updateRecordsAndRefresh");
        this.records = [];

        var pkcheck = new Object();
        if ($objs.length>0) {
            for (var x=0; x < $objs.length;x++) {
                var obj = $objs[x];
                if (this.primaryKey!=null) {
                    var pk = obj[this.primaryKey];
                    if (pk==null) {
                        console.log("pk missing on obj, skipping")
                        continue;
                    }

                    if (pkcheck[ pk] !=null) {
                        console.log("pk violation, skippping");
                        continue;
                    }

                    //just turn it on, only one can use this key.
                    pkcheck [ pk ]  = true;
                }

                obj.active = true;
                obj.rank="#"+(x+1);
                this.records.push(obj);
            }
        }
    }



    /**
     * custom maker view cols for the maker widget
     * We could just pass these in but I prefer them here.
     */
    makerViewSubCols = function($nObjs) {
        this.setRecsFromNObjs($nObjs);
        this.cols = [{
                neuron_link:true,
                title:"name",
                prop:"nid",
            },
            {
                neuron_link:true,
                title:"owner",
                prop:"owned_by",
            },
            {
                neuron_link:true,
                title:"sponsor",
                prop:"sponsored_by",
            }
        ];
    }   
    setRecsFromNObjs = function($nObjs) {
        for (var n=0; n < $nObjs.length; n++) {
            var obj = {};
            obj.nObj = $nObjs[n];
            obj.parent = MEMG.getNeuronByTempNid(obj.nObj.bes[0]);
            obj.active = true;
            obj.rank="#"+(n+1);
            this.records.push(obj);
        }
        //console.log(this.records);
    }

    
    /**
     * 
     * @param {*} $txt 
     */
    filterNeuronByString=async($txt)=>{
        console.log("(filterNeuronByString) "+$txt);
        $txt = $txt.toLowerCase();
        for (var r=0; r < this.records.length; r++) {
            var rec =  this.records[r];
            var nObj = rec.nObj;
            if (nObj.text.toLowerCase().indexOf($txt)!==-1) rec.active=true;
            else {
                console.log("NOT ACTIVE");
                rec.active=false;
            }
        }

        if (this.bPAGINATOR==true) {
            this.PAGINATOR.renderActiveResults();
        }
        else if (this.bDOPAGINATORX==true) {
            this.PAGINATOR.renderActiveResults();
        }
        else this.renderData(this.records);
    
    }

    
    //build a common use relon table cols here (when needed)
    makeViewRelonCols = function($relons) {
        var tbl_cols = [
        ];
    }


    /**
     * 
     * @param {*} $nObjs 
     */
    makerViewAllCols = function($nObjs) {
        this.setRecsFromNObjs($nObjs);
        this.cols = [{
                neuron_link:true,
                title:"name",
                prop:"nObj.nid",
                sort_prop:"nObj.text",
                sort_type:"string",
                input_filter:this.filterNeuronByString
            },
            {
                neuron_link:true,
                title:"type",
                prop:"parent.nid",
                sort_prop:"text",
                sort_type:"string"
            },
            {
                neuron_link:true,
                title:"owner",
                prop:"nObj.owned_by",
                sort_prop:"nObj.owned_by",
                sort_type:"string"
            },
            {
                neuron_link:true,
                title:"sponsor",
                prop:"nObj.sponsored_by",
                sort_prop:"nObj.sponsored_by",
                sort_type:"string"
            }
        ];
    }


    /**
     * 
     * @param {*} $rels 
     */
    relonTblMaker = function($rels) {
        this.setRecsFromRels($rels);
        this.cols = [{
                neuron_link:false,
                title:"from",
                prop:"rel.from_txt_temp"
            },
            {
                neuron_link:false,
                title:"reln",
                prop:"rel.reln_txt_temp"
            },
            {
                neuron_link:false,
                title:"to",
                prop:"rel.to_txt_temp"
            },
            {
                relon_link:true,
                title:"id",
                prop:"rel.txid"
            }
        ];
    }


    /**
     * 
     * @param {*} $rels 
     */
    andyRelonTblMaker = async($rels)=> {
        await this.setAndyRecsFromRels($rels);
        console.log(this.records);

        this.cols = [{
                neuron_link:false,
                title:"from",
                prop:"rel.from_txt_temp"
            },
            {
                neuron_link:false,
                title:"rights",
                prop:"rel.rights",
                class1:"w120"
            },
            {
                neuron_link:false,
                title:"reln",
                prop:"rel.reln_txt_temp",
                class1:"w120-sm-white"
            },
            {
                neuron_link:false,
                title:"to",
                prop:"rel.to_txt_temp"
            },
            {
                relon_link:true,
                title:"id",
                prop:"rel.txid",
                page:"page"
            },
            {
                title:"DELETE",
                prop:"rel.txid",
                actions_cb:this.deleteLNOT,
                page:"page",
                extra_class:"actions-col-header"
            }
        ];
    }


    /**
     * 
     * @param {*} $col 
     * @param {*} $record 
     * @param {*} $propVal 
     */
    deleteLNOT=function($col, $record, $propVal) {
        //console.log($col);
        //console.log($record);
        //console.log($propVal);
        var el_delete = document.createElement("span");
        el_delete.innerHTML="delete";
        el_delete.classList = "link-yellow"
        el_delete.addEventListener("click", (e)=> {
            console.log("TO DO, delete lnot. Will need to use proof of user tech.");
        });
        return $col.appendChild(el_delete);
    }



    /**
     * 
     */
    setAndyRecsFromRels = async($rels)=> {
        var lookup_nids = [];
 
        for (var r=0; r < $rels.length; r++){
            var rel = $rels[r];
            rel.rights = ANNE.MEM_UDF.getCollabType(rel);//"collab";

            var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
            if ((fromNOBJ==null || fromNOBJ.text=="") && lookup_nids.indexOf(rel.from_nid)==-1) {
                lookup_nids.push(rel.from_nid); 
            }

            var relnNOBJ = MEMG.getNeuronByTempNid(rel.r);
            if ((relnNOBJ==null || relnNOBJ.text=="") && lookup_nids.indexOf(rel.reln)==-1) {
                lookup_nids.push(rel.reln); 
            }

            if (rel.to_type==TO_TYPE_ANNE) {
                var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
                if ((toNOBJ==null || toNOBJ.text=="") && lookup_nids.indexOf(rel.to_s)==-1) {
                    lookup_nids.push(rel.to_s); 
                }
            }
        }


        if (lookup_nids.length>0) {
            var bQUERYANNODE=true;
            //just extract the basics from IDB mem into our precious RAM.
            await ANNE.API.lookupNids(lookup_nids, [0,1], bQUERYANNODE);
        }
        

        var txids_handled = [];
       
        //loop again, now that we extracted basic infor from IDB.
        for (var r=0; r < $rels.length; r++){
            var rel = $rels[r];
            //console.log($rels[r].txid);
            
            var obj = {};

            if (txids_handled.indexOf($rels[r].txid)!=-1) continue;
            txids_handled.push($rels[r].txid);

            var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
            if (fromNOBJ==null) rel.from_txt_temp = "(unk)";
            else rel.from_txt_temp = fromNOBJ.bestNameHTML();

            var relnNOBJ = MEMG.getNeuronByTempNid(rel.r);
            if (relnNOBJ!=null) rel.reln_txt_temp = relnNOBJ.bestNameHTML();
            else rel.reln_txt_temp = "(unk)";  
            
            if (rel.to_type==TO_TYPE_ANNE) {
                var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
                if (toNOBJ!=null) rel.to_txt_temp = toNOBJ.bestNameHTML();
                else rel.to_txt_temp = "(unk)";
            }
            else rel.to_txt_temp = rel.to_s;

            obj.active=true;
            obj.rel = rel;
            obj.page = "andy";  //if set, we go back to this page after successful action.
            this.records.push(obj);
        }
    }
    

    /**
     * 
     * @param {*} $rels 
     */
    setRecsFromRels = function($rels){
        for (var r=0; r < $rels.length; r++){
            var obj = {};
            var rel = $rels[r];
            
            var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
            if (fromNOBJ==null) rel.from_txt_temp = "(unk)";
            else rel.from_txt_temp = fromNOBJ.text;
            
            var relnNOBJ = MEMG.getNeuronByTempNid(rel.r);
            if (relnNOBJ!=null) rel.reln_txt_temp = relnNOBJ.text;
            else rel.reln_txt_temp = "(unk)";
            
            if (rel.to_type==TO_TYPE_ANNE) {
                var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
                if (toNOBJ!=null) rel.to_txt_temp = toNOBJ.text;
                else rel.to_txt_temp = "(unk)";
            }
            else rel.to_txt_temp = rel.to_s;

            obj.active=true;
            obj.rel = rel;
            this.records.push(obj);
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $estimatedTBLBDYTOP 
     * @param {*} $elId_updateCnt 
     * @param {*} $bDOBORDER 
     * @param {*} $bDOPAGINATOR 
     */
    buildDOMTable=function($el, $estimatedTBLBDYTOP=-1, $elId_updateCnt=null, $bDOBORDER=true, $bDOPAGINATOR=true, $ROW_MAX_H=null) {
        
        this.bDOBORDER = $bDOBORDER;
        this.bPAGINATOR = $bDOPAGINATOR;

        if ($ROW_MAX_H!=null) {
            this._ROW_MAX_H = $ROW_MAX_H;
            if (this._ROW_MAX_H < 12) {
                this._ROW_MAX_H=12;
                alert("FIX YOUR RABLE buildDOTable call, $ROW_MAX_H too low:"+$ROW_MAX_H);
            }
        }

        var el = ANNE.A.show($el);

        if (el==null) {
            return;
        }
        
        if ($elId_updateCnt!=null && $elId_updateCnt!="") {
            this.el_update_cnt = ANNE.A.__getEl($elId_updateCnt);
        }

        if (this.do_heading==true && this.heading_prop!="") {
            this.el_heading_row = document.createElement("div");
            this.el_heading_row.classList="simple-row "+this.neuron_heading_row_class;
            this.el_heading_row.id = ANNE.UTILS.randHex(4);

            for (var c=0; c < this.cols.length; c++) {
                var colobj = this.cols[c];
                var heading = colobj[this.heading_prop];
                var headingCol = ANNE.A.ezEl(this.el_heading_row, "div", "simple-col neuron-col-heading");
                
                if (colobj.classH!=undefined) {
                    //headingCol.classList.add(colobj.classH);
                    ANNE.A.__classListParse(headingCol, colobj.classH);
                }
                else if (colobj.class1!=undefined) {
                    ANNE.A.__classListParse(headingCol, colobj.class1);
                    //headingCol.classList.add(colobj.class1);
                }

                if (this.heading_col_class!="") headingCol.classList.add(this.heading_col_class);
                
                if (this.bDOBORDER==true) {
                    //if (c==0) headingCol.classList.add("vcolborderLR");
                    //else if (c!=this.cols.length-1) headingCol.classList.add("vcolborderR");
                }

                if (colobj.input_filter!=undefined && typeof colobj.input_filter=="function") {
                    ANNE.A.ezEl(headingCol, "span", "dib", heading+"&nbsp;");
                    ANNE.A.ezInputText(headingCol, "", "anne-table-filter", "", colobj.input_filter);
                }
                else {
                    if (heading=="") heading="&nbsp;";
                    ANNE.A.ezEl(headingCol, "span", "", heading);
                }
                if (colobj.s_prop!=undefined && colobj.s_type!=undefined) {
                    if (colobj.s_type=="numeric" || colobj.s_type=="alpha") {
                        ANNE.A.sortToggle(headingCol, c, "none", this.sortNumericData, colobj.s_type);
                    }
                }
            }
            el.append(this.el_heading_row);
        }

        this.el_tbl_body = ANNE.A.ezEl(el, "div", "anne-table-body");
        this.el_tbl_body.style['width']="100%";
        this.el_tbl_body.style['min-height']="200px";
        
        var bcr2 = this.el_tbl_body.getBoundingClientRect();
        var topy = 0;

        if ($estimatedTBLBDYTOP>0 && bcr2.top==0) {
            topy=$estimatedTBLBDYTOP;
        }
        else {
            topy = bcr2.top;
        }

        var height = 400;

        //??? this is all jacked up, once again.
        if (this.max_bottom==0) {
            this.el_tbl_body.style['max-height']=""; 
            height = (bcr2.bottom - topy) - this.YBUF;
            height = Math.max(height, 200);
        }
        else {
            height = (this.max_bottom - topy) - this.YBUF;
            height = Math.max(height, 200);
            this.el_tbl_body.style['max-height']=height+"px";        
        }


        if (this.bPAGINATOR==true) {
            this.el_paginator = ANNE.A.ezEl(el, "div", "simple-row paginator-row");
            var shard = Math.floor(height/this._ROW_MAX_H);
            this.PAGINATOR = new paginator(this.el_paginator, this.pgui, this._guid, this.records, this.renderCallback, shard);
            this.PAGINATOR.render($elId_updateCnt);
        }
        else {
            this.renderNoPaginator();
        }
    }
        

    /**
     * 
     * @param {*} $cb_loaded_chk 
     * @param {*} $el 
     * @param {*} $damnheight 
     * @param {*} $elId_updateCnt 
     * @param {*} $bDOBORDER 
     * @param {*} $ROW_MAX_H 
     * @param {*} $dataid - the primary key in the arraymap 
     */
    buildDOMTable_nextX=function($cb_loaded_chk, $el, $damnheight=300, $elId_updateCnt=null, $bDOBORDER=true, $ROW_MAX_H=null, $dataid="") {
        this.cb_loaded_chk = $cb_loaded_chk;
        this.data_id = $dataid;
        this.bDOBORDER = $bDOBORDER;
        this.bDOPAGINATORX = true;
        this.bPAGINATOR = false;

        if ($ROW_MAX_H!=null) {
            this._ROW_MAX_H = $ROW_MAX_H;
            if (this._ROW_MAX_H < 12) {
                this._ROW_MAX_H=12;
                alert("FIX YOUR RABLE buildDOTable call, $ROW_MAX_H too low:"+$ROW_MAX_H);
            }
        }

        var el = ANNE.A.show($el);

        //aaaaaaaaaaaaaaaaaaaaaaaand?
        if ($elId_updateCnt!=null && $elId_updateCnt!="") {
            this.el_update_cnt = ANNE.A.__getEl($elId_updateCnt);
        }

        if (this.do_heading==true && this.heading_prop!="") {
            this.el_heading_row = document.createElement("div");
            this.el_heading_row.classList="simple-row "+this.neuron_heading_row_class;
            this.el_heading_row.id = ANNE.UTILS.randHex(4);

            for (var c=0; c < this.cols.length; c++) {
                var colobj = this.cols[c];
                var heading = colobj[this.heading_prop];
                var headingCol = ANNE.A.ezEl(this.el_heading_row, "div", "simple-col neuron-col-heading");
                
                if (colobj.classH!=undefined) {
                    //headingCol.classList.add(colobj.classH);
                    ANNE.A.__classListParse(headingCol, colobj.classH);
                }
                else if (colobj.class1!=undefined) {
                    ANNE.A.__classListParse(headingCol, colobj.class1);
                    //headingCol.classList.add(colobj.class1);
                }

                if (this.heading_col_class!="") headingCol.classList.add(this.heading_col_class);
                
                if (this.bDOBORDER==true) {
                    //if (c==0) headingCol.classList.add("vcolborderLR");
                    //else if (c!=this.cols.length-1) headingCol.classList.add("vcolborderR");
                }

                if (colobj.input_filter!=undefined && typeof colobj.input_filter=="function") {
                    ANNE.A.ezEl(headingCol, "span", "dib", heading+"&nbsp;");
                    ANNE.A.ezInputText(headingCol, "", "anne-table-filter", "", colobj.input_filter);
                }
                else {
                    if (heading=="") heading="&nbsp;";
                    ANNE.A.ezEl(headingCol, "span", "dib", heading);
                }
                if (colobj.s_prop!=undefined && colobj.s_type!=undefined) {
                    if (colobj.s_type=="numeric" || colobj.s_type=="alpha") {
                        ANNE.A.sortToggle(headingCol, c, "none", this.sortNumericData, colobj.s_type);
                    }
                }
            }
            el.append(this.el_heading_row);
        }

        this.el_tbl_body = ANNE.A.ezEl(el, "div", "anne-table-body");
        this.el_tbl_body.style['width']="100%";
        //this.el_tbl_body.style['min-height']="200px";
        
        this.el_paginator = ANNE.A.ezEl(el, "div", "simple-row paginator-row");
        var shard = Math.floor($damnheight/this._ROW_MAX_H);

        this.PAGINATOR = new paginator_nextx(this.el_paginator, this._guid, this.renderCallbackLive, shard);
        this.PAGINATOR.tbl_ptr = this;
        this.PAGINATOR.render($elId_updateCnt);
    }


    sortNumericData=async($colNumb, $sort)=>{
        //var prop = 'col_'+$colNumb+'_propVal';
        if ($sort=="none") return;
        var multiplier = 1;
        if ($sort=="desc") multiplier=-1;

        //hmmm.  only the rendered items have colvals set from renderData.
        //we need to sort  with the actual s_prop in the col (not prop value col stuff).
        console.log(this.cols);
        var col = this.cols[$colNumb];
        var prop = col.s_prop;
        
        this.records.sort( function($a, $b) {
            if (parseInt($a[prop]) > parseInt($b[prop])) return 1*multiplier;
            else if (parseInt($a[prop]) < parseInt($b[prop])) return -1*multiplier;
            return 0;
        });
        ANNE.A.empty(this.el_tbl_body);
        this.auditSort(this.records, prop);
        
        //sort_arr is now sorted... we just need to rebuild our this.records.
        if (this.bPAGINATOR==true) {
            this.PAGINATOR.renderActiveResults();
        }
        else if (this.bDOPAGINATORX==true) {
            this.PAGINATOR.renderActiveResults();
        }
        else {
            this.renderNoPaginator(this.records);
        }
    }


    auditSort = function($arr, $prop) {
        for (var a=0; a < $arr.length; a++) {
            if ($arr[a][$prop]==undefined) {
                console.log($arr[a]);
            }
            console.log("#a "+$arr[a][$prop]);
        }
    }

    renderCallback = async(spObj, $data)=> {
        this.renderData($data);
    }

    
    renderNoPaginator = async()=>{
        if (this._nopagrender_cb!=null) {
            var results = await this._nopagrender_cb();
            this.renderData(results);
            if (this.el_update_cnt==null) return;
            this.el_update_cnt.innerHTML= results.length+" of "+ANNE.UTILS.commas(this.results.length)+" total";
        }
        else {
            this.renderData(this.records);
            if (this.el_update_cnt==null) return;
            this.el_update_cnt.innerHTML= ANNE.UTILS.commas(this.records.length)+" total";
        }

    }


    /**
     * 
     * @param {*} $data 
     * @param {*} $pageObj 
     * @param {*} $dataTohighlight
     */
    renderCallbackLive = async($data, $pageObj, $dataHighlight=null)=> {
        console.log("renderCallbackLive", $data, $pageObj);
        if ($pageObj!=null && $pageObj.loaded==false) {
            var finaldata = await this.cb_loaded_chk($data, $pageObj);
            if (finaldata==null || finaldata.length==0) {
                //alert("something went wrong with the live data query");
                return;
            }
            if (finaldata!=null) this.renderData(finaldata);
        }
        else {
            if ($data!=null) this.renderData($data);
            else {
                console.log("render data in renderCallbackLive is null!");
            }
        }
    }


    /**
     * Render data handed to us by the paginator.
     */
    renderData = function($arr, $dataTohighlight=null) {
        if ($arr==null || $arr==undefined) {
            console.warn("problem with render data!. Arr = null");
            return;
        }
        ANNE.A.empty(this.el_tbl_body);
        //console.log("renderData", $arr);
        for (var r=0; r < $arr.length; r++) {
            var record = $arr[r];
            
            //full mode passes base records with active on/off for filterings.
            if (record.active!=undefined && record.active==false) continue;

            var recordRow = ANNE.A.ezEl(this.el_tbl_body, "div", "simple-row "+this.tbl_row_class);
            
            var did = null;
            if (this.data_id!=null && this.data_id!="") {
                did = record[this.data_id];
                if (did!=undefined) {
                    recordRow.setAttribute("data-id", did); //for clicks.
                }
                if ($dataTohighlight!=null && $dataTohighlight[this.data_id]==did) {
                    recordRow.classList.add("highlight-new-row");
                }
            }

            for (var c=0; c < this.cols.length; c++) {
                var colobj = this.cols[c];
                var col_el = ANNE.A.ezEl(recordRow, "div", "simple-col");

                if (did!=null && colobj.idpat!=undefined) {
                    var colid = colobj.idpat+did;
                    col_el.id = colid;
                }

                if (this.bDOBORDER==true) {
                    //if (c==0) col_el.classList.add("vcolborderLR");
                    //else col_el.classList.add("vcolborderR");
                }
                
                if (colobj.class1!=undefined) {
                    ANNE.A.__classListParse(col_el, colobj.class1);
                }
                var propVal = record['col_'+c+'_propVal'];

                if (propVal==undefined) {
                    if (colobj.prop==undefined || colobj.prop==null || colobj.prop=="") {
                        col_el.innerHTML="&nbsp;";
                        continue;
                    }
                    else if (colobj.prop.length) {
                        var props = colobj.prop.split(".");
                        propVal = this._propExtractor(record, props);
                        record['col_'+c+'_propVal'] = propVal;
                    }
                }


                //this does a cb to render the dtaa...
                if (colobj.actions_cb!=undefined && typeof colobj.actions_cb==="function") {
                    colobj.actions_cb(col_el, record, propVal, this.INST);
                }
                
                else if (propVal!==0 && (propVal==undefined || propVal==null || propVal=="")) {
                    console.log("empty prop val");
                    col_el.innerHTML="&nbsp;";
                    record['col_'+c+'_propVal'] = "";  
                }

                //if bDOCELLCLASSES is on we want to turn the col a color based
                //on the val of the col.
                else if (colobj.neuron_link==true) {
                    var btn = null;

                    var page = "dossier";
                    if (record.page!=undefined) {
                        console.log("override dossier for page '"+page+"'");
                        page = record.page;
                    }

                    if (this.bCOLORSQARES == true) {
                        btn=ANNE.EXT_BTNS.makeNeuronLink_colorSquares(col_el, propVal, "link-bigger", "", false, "", page);
                    }
                    else {
                        btn=ANNE.EXT_BTNS.makeNeuronLink(col_el, propVal, "link-bigger", "", false, "", page);
                    }

                    if (colobj.achat!=undefined) {
                        var achat_sesh_id = record.achat_sesh_id;
                        if (achat_sesh_id!=undefined) {
                            btn.setAttribute("miscp1", "achat_sesh_id");
                            btn.setAttribute("achat_sesh_id", achat_sesh_id);
                            btn.setAttribute("page", "achat");  //change page to achat.
                        }
                    }

                    if (this.bDOCELLCLASSES==true) {
                        var nOBJ=MEMG.getNeuronByNid(propVal);
                        var cellClass=EXT_BTNS.classifySpanBtn(nOBJ, null);
                        if (cellClass!="") col_el.classList.add(cellClass);
                    }
                }

                //if bDOCELLCLASSES is on we want to turn the col_el a color based
                //on the val of the col_el.
                else if (colobj.relon_link!=undefined && colobj.relon_link==true) {
                    var relon = null;

                    if (typeof ANNE.MEM_UDF.getAnyRelon==="function") relon = ANNE.MEM_UDF.getAnyRelon(propVal);
                    else relon = ANNE.MEMG.lookupByTXID(propVal);

                    var el_link = null;

                    if (relon) el_link = ANNE.MEM_UDF.makeRelonViewLink_SMART(col_el, relon, true, 10);
                    else el_link = ANNE.MEM_UDF.makeRelonViewLink(col_el, propVal, false, 10, true);
                
                    //if we have a page var in the data we want to pass that through
                    //to the link to pass that fwd thru to the page of the link.
                    if (record.page!=undefined & record.page!=null 
                        && record.page.length>0 && el_link!=undefined && el_link!=null) {
                        el_link.setAttribute("page", record.page);
                    }
                }


                else {
                    if (colobj.formatter!=undefined) {
                        if (colobj.formatter=="ezAnniesToAnneDisplay") {
                            ANNE.A.ezAnniesToAnneDisplay(col_el, propVal);
                        }
                        else if (colobj.formatter=="ezAnneDisplayApprox") {
                            ANNE.A.ezAnneDisplayApprox(col_el, propVal);
                        }
                        else if (colobj.formatter=="commas") {
                            col_el.innerHTML=ANNE.UTILS.commas(propVal);
                        }
                        else if (colobj.formatter=="blocknum") {
                            col_el.innerHTML="#"+ANNE.UTILS.commas(propVal);
                        }
                        else if (colobj.formatter=="formatPercent") {
                            col_el.innerHTML=propVal+"%";
                        }
                        else if (colobj.formatter_cb!=null) {
                            colobj.formatter_cb(col_el)
                        }
                        else {
                            console.log("unhandled formatter");
                            col_el.innerHTML = propVal+"";
                        }
                    }
                    else col_el.innerHTML = propVal+"";
                }
            }
        }
    }
    

    _propExtractor = function ($v, $props) {
        for (var p=0; p < $props.length; p++) {
            var prop = $props[p];
            if (prop==undefined || $v[prop]==undefined) {
                console.log("ut oh with propVal:"+$v+" prop:'"+prop+"'");
                break;
            } 
            else $v = $v[ prop ];
        }
        return $v;
    }


    goToPageIndex = async($pageIndex)=> {
        await this.PAGINATOR.goToPageIndex($pageIndex);
    }


    /**
     * 
     * @param {*} $handleClick 
     * @returns 
     */
    rowClicker = function($handleClick=null) {
        if ($handleClick==null) return;
        this.cb_rowclicker = $handleClick;
        this.el_tbl_body.addEventListener('click', async(event)=> {
            const targetRow = event.target.closest('.'+this.tbl_row_class);
            var els = this.el_tbl_body.querySelectorAll('.'+this.tbl_row_class);
            for (var e=0; e<els.length; e++) { els[e].classList.remove("activeblockrow");}
            if (targetRow) {
                const id = targetRow.getAttribute('data-id');
                targetRow.classList.add("activeblockrow");
                $handleClick(id);
            }
        });
    }


    /**
     * 
     * @param {*} $dataid 
     * @returns 
     */
    invokeRowClicker=async($dataid)=> {
        if (this.cb_rowclicker==null) return;
        var els = this.el_tbl_body.querySelectorAll('.'+this.tbl_row_class);
        for (var e=0; e<els.length; e++) { 
            els[e].classList.remove("activeblockrow");
        }
        for (var e=0; e<els.length; e++) { 
            if (els[e].getAttribute("data-id")==$dataid) {
                els[e].classList.add("activeblockrow");
                return await this.cb_rowclicker($dataid);
            }
        } 
    }
}


var DSTBL = {};


/**
 * use in data source widgets (no in the EXT)
 */
DSTBL.init=async(spObj)=> {

    //should have a fancier loader to say "building live data source"
    ANNE.A.show("loading"); 

    var ID = spObj._guid;
    var pObj = spObj.pObj;
    var data_src_aid = pObj.p1;
    var manualNOBJS = pObj.p2;
    var manual_col_props = pObj.p3;

    var tbl_divid = "anne_tbl_"+ID;
    var tbl = document.getElementById(tbl_divid);
  
    var nObj = ANNE.MEMG.getNeuronByNid(data_src_aid);
    if (nObj==null) {
        alert("inconceivable!");
        return;
    }

    //datasources are derrivatives of neuron data.
    //they will be used in BI, reporting, etc.
    //Traditionally, BI shops will pre-make cubes to control their
    //universe.. They will have "snapshots" ready w/ summary data that matches
    //the desired grain.  (monthly, weekly, etc)
    //We will do it live with the later option to save-off snapshots.
    //we may want a MEM.BI here tied to indexDB tbl we work with during the session...
    //as these DS will not change  much during the session and are heavy on processing.
    spObj.INST = new anne_datasource(nObj);

    if (manualNOBJS!=null && manual_col_props!=null) {
        await spObj.INST.init_manual_override(manualNOBJS, manual_col_props);
    }
    else {
        await spObj.INST.init();
    }
    
    var cols = spObj.INST.cols;
    var data = spObj.INST.data;
    console.log(cols);
    
    const rowsearch = tbl.querySelector('.row-search');
    const rowheader = tbl.querySelector('.row-header');

    ANNE.A.setHTML("tbl_title_"+ID, " "+nObj.text);

    if (spObj.INST.GEN1SRC==true) ANNE.A.hide("anne_embedded_filters_"+ID);
    else {
        var filterString = spObj.INST.getFilterStrings();
        ANNE.A.setHTML("anne_embedded_filters_"+ID, filterString, true);
    }

    const availableCols = document.getElementById("ds_available_cols_"+ID);

    var gardeninfo = availableCols.querySelector(".ds-let-deny-list");
    gardeninfo.innerHTML = "";
    await ANNE.MEMZ.ezDSGardenInfo(gardeninfo, spObj.INST.ddsNObj, LINKS_NEURON_TBL_VIEW);

    var colchks = availableCols.querySelector(".ds-col-checks");

    for (var c=0; c < spObj.INST.cols.length; c++) {
        var colObj = spObj.INST.cols[c];
        var checked = true;
        var disabled = true;
        if (colObj.active==false) {
            checked = false;
            disabled = false;
        }
        var chk = ANNE.A.ezCheckMark(colchks, "chk_"+ID+"_"+colObj.datanid, colObj.prelon, colObj.text, checked, disabled, "truncate120");
        
        chk.addEventListener("click", $e=> {
            if ($e.target.checked==true) {
                var datanid = $e.target.id.split("_").reverse()[0];
                var col = spObj.INST.getColByDatanid(datanid);
                if (col==null) return;
                console.log("you checked a new column of data w/ datanid:"+datanid);
                spObj.INST.addSingleColRuntime(col);

                var col_index = parseInt(tblbody.getAttribute("active-cols"));
                tblbody.setAttribute("active-cols", parseInt((col_index+1)));

                col.col_index = col_index; 

                //sort ico and filters are wired up in the addColumnHeader..
                DSTBL.addColumnHeader(col, col_index, tbl_divid, rowsearch, rowheader);
                DSTBL.addColumnToRows(col, spObj.INST.data, tbl); //add the data to the rows.. 

                col.bPENDING_SAVE = true;
                var bNEWPENDINGCHANGE = spObj.INST.setPendingChanges();
                if (bNEWPENDINGCHANGE==true) DSTBL.handleSaveEditBtns(spObj, ID);
            }
            else {
                var datanid = $e.target.id.split("_").reverse()[0];
                console.log("you unckecked a column of data w/ datanid:"+datanid);

                var col_index = parseInt(tblbody.getAttribute("active-cols")-1);
                tblbody.setAttribute("active-cols", parseInt((col_index)));
                
                var datanid = $e.target.id.split("_").reverse()[0];
                var col = spObj.INST.getColByDatanid(datanid);
                if (col==null) return;
                
                DSTBL.deleteColumnFromRows(col, spObj.INST.data, tbl); //add the data to the rows..
                DSTBL.deleteColumnHeader(col.col_index, rowsearch, rowheader);
                                        
                col.active=false;
                col.bPENDING_SAVE = false;

                //this checks if we have any pending saves in the Datasource...
                var bNEWPENDINGCHANGE = spObj.INST.setPendingChanges();
                if (bNEWPENDINGCHANGE==true) DSTBL.handleSaveEditBtns(spObj, ID);
            }
        });
    }

    //one-time, on init.  Any added columns is handled int the checkmark click event.
    var col_index = 0;
    for (var c=0; c < cols.length; c++) {
        var col = cols[c];
        if (col.active==false) continue;
        
        //filter events are added in this addColumnHeader function
        DSTBL.addColumnHeader(col, col_index, tbl_divid, rowsearch, rowheader);
        col.col_index = col_index;
        col_index+=1;
    }


    //iterate through the data ---
    //inject new "row" with all 3 cols and respective data.
    const tblbody = tbl.querySelector('.anne-tbl-body-overflow');
    tblbody.setAttribute("active-cols", parseInt(col_index));
    DSTBL.addDataToColsInit(data, cols, tblbody);
    
    //set up table footer display
    var rowslbl = ANNE.TALK.pluralizeNounIf2(data.length, "row", "rows");
    ANNE.A.setHTML("tbl_summary_"+ID, data.length+" "+rowslbl);

    //initialize the final table FX DOM events.
    DSTBL.initEvents(tbl_divid, spObj, ID);
    ANNE.A.hide("loading"); 
}



/*** handleSaveEditBtns ********************************** xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.handleSaveEditBtns=function(spObj, $ID) {

    //if this is a self-owned data source we can either:
    //edit, OR create a new datasource.
    if (spObj.INST.GEN1SRC==false && spObj.INST.ddsNObj.owned_by==USER._userNObj.nid) {
        //do an edit.
        var btnel = ANNE.A.fadeInPop('tbl_update_btn_'+$ID, "glow");
        ANNE.A.addRemoveClass(btnel, "hoverable-yellow", "hoverable-green");

        //do a save (either option is possible)
        var btnel = ANNE.A.fadeInPop('tbl_save_btn_'+$ID, "glow");
        ANNE.A.addRemoveClass(btnel, "hoverable-yellow", "hoverable-green");
    }

    //if this is not self-owned then the only option is to save a new datasource.
    else {
        var btnel = ANNE.A.fadeInPop('tbl_save_btn_'+$ID, "glow");
        ANNE.A.addRemoveClass(btnel, "hoverable-yellow", "hoverable-green");
    }
}



/*** addDataToColsInit *********************************** xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.addDataToColsInit=function($data, $cols, $tblBody) {
    for (var d=0; d < $data.length; d++) {
        var dataobj = $data[d].nObj;
        if (dataobj==null) continue;
        var newrow = document.createElement("div");
        newrow.classList="row";
        $tblBody.append(newrow);
        if ($data[d].active==false) {
            console.log("live skip of data obj.. not from DS, from this session changes");
            continue;
        }
        newrow.setAttribute("nid", dataobj.temp_nid);
        for (var c=0; c < $cols.length; c++) {
            var col = $cols[c];
            if (col.active==false) continue;
            var datacol = document.createElement("div");
            datacol.classList = "column cell";
            var dataVal = dataobj[col.dataprop];
            var display = dataobj[col.display];
            if (display==undefined) display=dataVal;
            datacol.setAttribute("dataz", dataVal);
            datacol.innerHTML=display;
            newrow.append(datacol);
        }
    }
}


/*** addColumnHeader ************************************* xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.addColumnHeader = function($col, $col_index, $tbl_divid, $rowsearch, $rowheader) {
    var div = document.createElement("div");
    div.classList = "column";
    div.setAttribute("colnum", $col_index);
    div.setAttribute("tbl", $tbl_divid);
    $rowsearch.append(div);
    var input = document.createElement("input");
    input.classList = "tbl-filter";
    input.setAttribute("type", "text");
    input.setAttribute("width", 300);
    input.style["min-height"]="30px";   //do this in css instead..
    input.setAttribute("placeholder", "filter the "+$col.text);
    div.append(input);
    input.addEventListener('input', $e=>DSTBL.DOM_filterEvent($e))

    //create the header cols in the header row
    var divh = document.createElement("div");
    divh.classList = "column";
    divh.setAttribute("colnum", $col_index);
    divh.setAttribute("tbl", $tbl_divid);
    $rowheader.append(divh);

    var sortIco = ANNE.A.ezSortCol(divh, $col, $tbl_divid, "asc");
    sortIco.addEventListener("click", $e=> {
        var el = $e.target;
        var sortDataType = el.getAttribute("prop");
        console.log(sortDataType);

        //parent child index is the column index of this header.
        //this will inform the DOM sort what column to pay attention to.
        var col_index = ANNE.A.getFirstParentAttr(el, "column", "colnum");
        var tbl = el.getAttribute("tbl");
        var isActive = el.classList.contains("activesort");
        if (isActive==null) isActive=false;
        var sort = el.getAttribute("sort");
        console.log("current sort:"+sort+" active:"+isActive);
        ANNE.A.removeAddClassOfAllIn(tbl, ".sortico", "activesort", el);
        var sort = DSTBL.toggleSortIco(el, sortDataType, isActive, sort);
        
        //var iFix = (i*2) + 1;
        DSTBL.DOM_sortTbl(tbl, col_index, sortDataType, sort);
    });

    //set up the default sort ico state if this table was pre-sorted.
    if ($col.default_sort!=undefined) {
        sortIco.setAttribute("sort", $col.default_sort);
        ANNE.A.removeAddClassOfAllIn(tbl, ".sortico", "activesort", sortIco);
        var isActive = sortIco.classList.add("activesort");
        var sortDataType = sortIco.getAttribute("prop");
        var sort = DSTBL.toggleSortIco(sortIco, sortDataType, isActive, sort);
    }

    return input;   //RETURN THE TBL FILTER
}


/*** deleteColumnHeader ********************************** xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.deleteColumnHeader = function($col_index, $rowsearch, $rowheader) {
    var searchheader = $rowsearch.querySelectorAll('div[colnum="'+$col_index+'"]')[0];
    searchheader.remove();
    var colheader = $rowheader.querySelectorAll('div[colnum="'+$col_index+'"]')[0];
    colheader.remove();
    //we need to either de-activate or delete any filters.
}



/*** addColumnToRows ************************************* xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.addColumnToRows = function($col, $data, $tbl) {
    const tblbody = $tbl.querySelector('.anne-tbl-body-overflow');
    for (var d=0; d < $data.length; d++) {
        if ($data[d].active==false) continue;
        var dataobj = $data[d].nObj;
        var rowdiv = tblbody.querySelectorAll('div[nid="'+dataobj.temp_nid+'"]')[0];
        if ($col.active==false) break; //this should be active
        var datacol = document.createElement("div");
        datacol.classList = "column cell";
        var dataVal = dataobj[$col.dataprop];
        var display = dataobj[$col.display];
        if (display==undefined) display=dataVal;
        datacol.setAttribute("dataz", dataVal);
        datacol.innerHTML=display;
        rowdiv.append(datacol);
    }
}


/*** deleteColumnFromRows ******************************** xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.deleteColumnFromRows = function($col, $data, $tbl) {
    const tblbody = $tbl.querySelector('.anne-tbl-body-overflow');
    for (var d=0; d < $data.length; d++) {
        if ($data[d].active==false) continue;
        var dataobj = $data[d].nObj;
        var rowdiv = tblbody.querySelectorAll('div[nid="'+dataobj.temp_nid+'"]')[0];
        //var dataVal = dataobj[$col.dataprop];
        var datacol = rowdiv.children[$col.col_index];
        if (datacol!=null) datacol.remove();
    }
}


/*** initEvents ****************************************** xx/xx/202x ***
******************************************************** Mr.Scatman ****/ 
DSTBL.initEvents = function($tbldivid, spObj, $ID) {
    //const fxwrapper = tbl.querySelector(".anne-tbl-fx-wrapper"); //we can use this for bigger FX
    const tbl = document.getElementById($tbldivid);
    const footer = document.querySelector(".anne-tbl-footer");
    const downs = tbl.querySelector(".anne-downloadables");
    const ups = tbl.querySelector(".anne-uploadables");

    footer.addEventListener("mouseenter", (e) => {
        tbl.style.transition = "none";
        downs.style.transition  =  "all 0.5s ease-in";
        downs.style.transform = "translateX(-180px) translateY(20px)  scale(1.4)";
        ups.style.transition  =  "all 0.5s ease-in";
        ups.style.transform = "translateX(-100px) translateY(20px) scale(1.4)";
        footer.style["height"] = "64px";
        footer.style.transition  =  "all 0.5s ease-in";
    });

    footer.addEventListener("mouseleave", (e) => {
        downs.style.transition  =  "all 0.3s ease-out";
        downs.style.transform = "translateX(0px) translateY(0px)  scale(1)";
        ups.style.transition  =  "all 0.3s ease-out";
        ups.style.transform = "translateX(0px) translateY(0px) scale(1)";
        footer.style["height"] = "32px";
        footer.style.transition  =  "all 0.3s ease-out";
    });

    var editBtn = ANNE.A.fadeInPop('tbl_update_btn_'+$ID, "glow");
    editBtn.addEventListener("click", (e) => {
        spObj.INST.saveEditsToExisting();
    });


    var saveBtn = ANNE.A.fadeInPop('tbl_save_btn_'+$ID, "glow");
    saveBtn.addEventListener("click", (e) => {
        spObj.INST.savePendingChangesToNew();
    });

    var jsonBtn = document.getElementById("btn_json_"+$ID);
    jsonBtn.addEventListener("click", (e)=> {
        alert("dl json");
    });

    var csvBtn = document.getElementById("btn_csv_"+$ID);
    csvBtn.addEventListener("click", (e)=> {
        alert("dl csv");
    });
    
    var excelBtn = document.getElementById("btn_excel_"+$ID);
    excelBtn.addEventListener("click", (e)=> {
        alert("dl excel");
    });
    
}


/*** DOM_sortTbl ***************************************** 0x/xx/20xx ***
******************************************************** Mr.Scatman ****/
DSTBL.DOM_sortTbl=function($divid, $col_index, $sortDataType, $sortDir) {
    var tbl = document.getElementById($divid); 
    if (tbl==null) return;

    const rows = tbl.querySelectorAll(".row");

    var masterRowParent = rows[0].parentElement;
    var rowsData = Array.prototype.slice.call(rows, 0);
    var dir = 1;
    if ($sortDir=="asc") dir *= -1;

    //sort them  (keep these separated even tho they are the same.. for now)..
    //we may need to handle differently..
    if ($sortDataType=="string" || $sortDataType=="anne") {
        rowsData.sort(function(a,b) {
            if (a.children[$col_index]==undefined) return 0;
            var txta = a.children[$col_index].getAttribute("dataz");//textContent;
            if (b.children[$col_index]==undefined) return 0;
            var txtb = b.children[$col_index].getAttribute("dataz");
            if(txta > txtb) { return -1*dir; }
            if(txta < txtb) { return 1*dir; }
            return 0;
        });
          
    }
    else if ($sortDataType=="numeric") {
        rowsData.sort(function(a,b) {
            if (a.children[$col_index]==undefined) return 0;
            var txta = (a.children[$col_index].getAttribute("dataz"));
            if (b.children[$col_index]==undefined) return 0;
            
            var txtb = (b.children[$col_index].getAttribute("dataz"));
            if(txta > txtb) { return -1*dir; }
            if(txta < txtb) { return 1*dir; }
            return 0;
        });
    }

    else {
        rowsData.sort(function(a,b) {
            if (a.children[$col_index]==undefined) return 0;
            var txta = (a.children[$col_index].getAttribute("dataz"));
            if (b.children[$col_index]==undefined) return 0;
            
            var txtb = (b.children[$col_index].getAttribute("dataz"));
            if(txta > txtb) { return -1*dir; }
            if(txta < txtb) { return 1*dir; }
            return 0;
        });
    }

    //put them back in.
    rowsData.forEach(row => {
        masterRowParent.appendChild(row)
    });
}


/*** DOM_filterEvent ************************************* 0x/xx/20xx ***
******************************************************** Mr.Scatman ****/
DSTBL.DOM_filterEvent = async($e)=> {
    var el = $e.target;
    var text = el.value;
    if (text.length && text[text.length-1]=="") return;
    await DSTBL.DOM_Filter(el, text);
}


/*** DOM_Filter ****************************************** 0x/xx/20xx ***
******************************************************** Mr.Scatman ****/
DSTBL.DOM_Filter=async(el, text)=> {
    const col = ANNE.A.getFirstParentOf(el, "column");
    const tbldivid = col.getAttribute("tbl");
    const childindex = col.getAttribute("colnum");
    const tbl = document.getElementById(tbldivid);
    console.log("filter tbl:"+tbl+" col "+childindex+" text "+text);

    //get all the rows, turn off any rows where the colIndex doesn't match the text.
    const ID = tbldivid.split("_").reverse()[0].trim();

    var max_rows = 0;
    var on_cnt = 0;

    var pgui = NF.getPguiByGUID(ID);

    //turn on all rows
    tbl.querySelectorAll(".row").forEach(row => {
        row.style.display="flex";
        max_rows++;
    });

    const tblFilters = tbl.querySelectorAll(".tbl-filter");
    for (var t=0; t < tblFilters.length; t++) {
        var tblfilter = tblFilters[t];
        var cIndex = ANNE.A.getFirstParentAttr(tblfilter, "column", "colnum");

        //this is the filter the user just edited/changed...
        if (cIndex==childindex) {
            await pgui.INST.make_unsaved_filter(tblfilter.value, cIndex);
        }

        if (tblfilter.value.length) {
            var on_cnt = DSTBL.DOM_Filter_single(pgui, tblfilter, tblfilter.value, cIndex);
            
            //check the last filter (the net sum of all visible/not visible)..
            //it will return the on_cnt.. 
            if (tblFilters.length-1==t) {
                max_rows = max_rows-on_cnt;
            }
        }
    }

    var rowslbl = ANNE.TALK.pluralizeNounIf2(on_cnt, "row", "rows");
    ANNE.A.setHTML("tbl_summary_"+ID, on_cnt+" "+rowslbl);
    if (on_cnt!=max_rows) {
        var bNEWPENDINGCHANGE = pgui.INST.setPendingChanges();
        if (bNEWPENDINGCHANGE==true) DSTBL.handleSaveEditBtns(pgui, ID);
    }
}


/*** DOM_Filter_single *********************************** 0x/xx/20xx ***
******************************************************** Mr.Scatman ****/
DSTBL.DOM_Filter_single=function(spObj, el, text, $cIndex) {
    const col = ANNE.A.getFirstParentOf(el, "column");
    const tbldivid = col.getAttribute("tbl");
    const childindex = col.getAttribute("colnum");
    const tbl = document.getElementById(tbldivid);
    console.log("filter tbl:"+tbl+" col "+childindex+" text "+text);

    //var regex = new RegExp(text, 'i');
    //var bEXACT = false;

    var onCnt = 0;

    var coldataobj = spObj.INST.cols[$cIndex];
    var filters = coldataobj.filters;
    if (filters==undefined || filters==null) filters = [];

    //these are filters we made to support the user actions of typing in the live-filter boxes
    var uf =  spObj.INST.unsaved_filters;

    if (uf.length && uf[$cIndex]!=undefined) {
        filters.push(uf[$cIndex]);
    }

    tbl.querySelectorAll(".row").forEach(row => {
        var dataz = row.children[childindex].innerHTML;//.getAttribute("dataz");
        if (row.style.display=="none") return;  //skip row, go to next.
        if (dataz==null) row.style.display="none";
        else {
            
            var test = anne_x_filter.testFiltersRuntime(dataz, filters);

            if(test==false) row.style.display="none";
            else {
                row.style.display="flex";
                onCnt+=1;
            }
        }
        //else if (dataz !== text) row.style.display="none";
        //else {
         //row.style.display="flex";
            //onCnt+=1;
       // }
    });

    return onCnt;
}


/*** toggleSortIco *************************************** 0x/xx/20xx ***
* If bISACTIVE=false then do NOT toggle the sort.. just apply the       *
* .activesort class.  If it is active we will leave it active (no need  *
* to apply the activesort class) but we will flip the ico and the dir.  *
******************************************************** Mr.Scatman ****/
DSTBL.toggleSortIco = function($el, $sortDataType, $bISACTIVE, $dir="asc") {
    console.log("toggleSortIco sort coming set to dir:"+$dir+" active:"+$bISACTIVE);
    if ($bISACTIVE==false) {
        $el.classList.add("activesort");
        console.log("toggleSortIco returning dir:"+$dir);
        return $dir;
    }
    if ($dir=="asc") {
        ANNE.A.setAttribute($el, "sort", "desc");
        if ($sortDataType=="string" || $sortDataType=="anne") ANNE.A.addRemoveClass($el, "ai-sort-alpha-up", "ai-sort-alpha-down"); 
        else if ($sortDataType=="number") ANNE.A.addRemoveClass($el, "ai-sort-amount-up", "ai-sort-amount-down"); 
        else ANNE.A.addRemoveClass($el, "ai-sort-up", "ai-sort-down"); 
        console.log("toggleSortIco returning dir:desc");
        return "desc";
    } 
    else {
        ANNE.A.setAttribute($el, "sort", "asc");
        if ($sortDataType=="string" || $sortDataType=="anne") ANNE.A.addRemoveClass($el, "ai-sort-alpha-down", "ai-sort-alpha-up"); 
        else if ($sortDataType=="number") ANNE.A.addRemoveClass($el, "ai-sort-amount-down", "ai-sort-amount-up"); 
        else ANNE.A.addRemoveClass($el, "ai-sort-down", "ai-sort-up"); 
        console.log("toggleSortIco returning dir:asc");
        return "asc";
    }
}

DSTBL.addRelon = function() {
    console.log("not relevant here");
}

DSTBL.destroyAll = function() {
    alert("do we get here"); 
}


//GENERICTBL  -- this is for anything (not just neurons or relons)
class generic_table {
    constructor(spObj) {
        this.pgui = spObj;
        this.ID = spObj._guid;
        this.tbl_obj = this.spObj.p1; 
        this.cols =  this.tbl_obj.cols;
        this.data =  this.tbl_obj.data;
        this.misc = this.spObj.misc;

        var tbl_divid = "anne_tbl_"+this.ID;
        var tbl = document.getElementById(tbl_divid);

        const rowsearch = tbl.querySelector('.row-search');
        const rowheader = tbl.querySelector('.row-header');

        ANNE.A.setHTML("tbl_title_"+this.ID, this.spObj.p2);

        //one-time, on init.  Any added columns is handled int the checkmark click event.
        var col_index = 0;
        for (var c=0; c < this.cols.length; c++) {
            var col = this.cols[c];
            if (col.active==false) continue;
            
            //filter events are added in this addColumnHeader function
            generic_table.addColumnHeader(col, col_index, tbl_divid, rowsearch, rowheader);
            col.col_index = col_index;
            col_index+=1;
        }

         //iterate through the data ---
        //inject new "row" with all 3 cols and respective data.
        const tblbody = tbl.querySelector('.anne-tbl-body-overflow');
        tblbody.setAttribute("active-cols", parseInt(col_index));
        generic_table.addDataToColsInit(this.data, this.cols, tblbody);
        
        //set up table footer display
        var rowslbl = ANNE.TALK.pluralizeNounIf2(this.data.length, "row", "rows");
        ANNE.A.setHTML("tbl_summary_"+this.ID, this.data.length+" "+rowslbl);

        //initialize the final table FX DOM events.
        generic_table.initEvents(tbl_divid, spObj, this.ID);
        ANNE.A.hide("loading"); 
    }
    
    static addDataToColsInit=function($data, $cols, $tblBody) {
        for (var d=0; d < $data.length; d++) {
            var dataobj = $data[d].nObj;
            if (dataobj==null) continue;
            var newrow = document.createElement("div");
            newrow.classList="row generic-tbl-row";
            $tblBody.append(newrow);
            if ($data[d].active==false) {
                console.log("live skip of data obj.. not from DS, from this session changes");
                continue;
            }
            newrow.setAttribute("nid", dataobj.temp_nid);
            for (var c=0; c < $cols.length; c++) {
                var col = $cols[c];
                if (col.active==false) continue;
                var datacol = document.createElement("div");
                datacol.classList = "column cell";
                
                //flex amount added to col
                if (col.f!=undefined) datacol.classList.add(col.f);

                var dataVal = dataobj[col.dataprop];
                var display = dataobj[col.displayprop];
                if (display==undefined) display=dataVal;
                datacol.setAttribute("dataz", dataVal);
                datacol.innerHTML=display;
                newrow.append(datacol);
            }
        }
    }

    static initEvents = function($tbldivid, spObj, $ID) {
        //const fxwrapper = tbl.querySelector(".anne-tbl-fx-wrapper"); //we can use this for bigger FX
        const tbl = document.getElementById($tbldivid);
        const footer = document.querySelector(".anne-tbl-footer");
        const downs = tbl.querySelector(".anne-downloadables");
        const ups = tbl.querySelector(".anne-uploadables");

        var jsonBtn = document.getElementById("btn_json_"+$ID);
        jsonBtn.addEventListener("click", (e)=> {
            alert("dl json");
        });
        var csvBtn = document.getElementById("btn_csv_"+$ID);
        csvBtn.addEventListener("click", (e)=> {
            alert("dl csv");
        });
        var excelBtn = document.getElementById("btn_excel_"+$ID);
        excelBtn.addEventListener("click", (e)=> {
            alert("dl excel");
        });
    }

    static DOM_filterEvent = async($e)=> {
        var el = $e.target;
        var text = el.value;
        if (text.length && text[text.length-1]=="") return;
        await generic_table.DOM_Filter(el, text);
    }

    static addColumnHeader = function($col, $col_index, $tbl_divid, $rowsearch, $rowheader) {
        var div = document.createElement("div");
        div.classList = "column";
        div.setAttribute("colnum", $col_index);
        div.setAttribute("tbl", $tbl_divid);
        $rowsearch.append(div);
        var input = document.createElement("input");
        input.classList = "tbl-filter";
        
        //flex amount added to col
        if ($col.f!=undefined) div.classList.add($col.f);

        input.setAttribute("type", "text");
        input.setAttribute("width", 300);
        input.style["min-height"]="30px";   //do this in css instead..
        input.setAttribute("placeholder", "filter the "+$col.text);
        div.append(input);
        input.addEventListener('input', $e=>generic_table.DOM_filterEvent($e))

        //create the header cols in the header row
        var divh = document.createElement("div");
        divh.classList = "column";

        //flex amount added to col
        if ($col.f!=undefined) divh.classList.add($col.f);

        divh.setAttribute("colnum", $col_index);
        divh.setAttribute("tbl", $tbl_divid);
        $rowheader.append(divh);

        var sortIco = ANNE.A.ezSortCol(divh, $col, $tbl_divid, "asc");
        sortIco.addEventListener("click", $e=> {
            var el = $e.target;
            var sortDataType = el.getAttribute("prop");
            console.log(sortDataType);

            //parent child index is the column index of this header.
            //this will inform the DOM sort what column to pay attention to.
            var col_index = ANNE.A.getFirstParentAttr(el, "column", "colnum");
            var tbl = el.getAttribute("tbl");
            var isActive = el.classList.contains("activesort");
            if (isActive==null) isActive=false;
            var sort = el.getAttribute("sort");
            console.log("current sort:"+sort+" active:"+isActive);
            ANNE.A.removeAddClassOfAllIn(tbl, ".sortico", "activesort", el);
            var sort = generic_table.toggleSortIco(el, sortDataType, isActive, sort);
            
            //var iFix = (i*2) + 1;
            generic_table.DOM_sortTbl(tbl, col_index, sortDataType, sort);
        });

        return input;   //RETURN THE TBL FILTER
    }

    static DOM_Filter=async(el, text)=> {
        const col = ANNE.A.getFirstParentOf(el, "column");
        const tbldivid = col.getAttribute("tbl");
        const childindex = col.getAttribute("colnum");
        const tbl = document.getElementById(tbldivid);
        console.log("filter tbl:"+tbl+" col "+childindex+" text "+text);

        //get all the rows, turn off any rows where the colIndex doesn't match the text.
        const ID = tbldivid.split("_").reverse()[0].trim();
        var max_rows = 0;
        var on_cnt = 0;
        var pgui = NF.getPguiByGUID(ID);

        //turn on all rows
        tbl.querySelectorAll(".row").forEach(row => {
            row.style.display="flex";
            max_rows++;
        });

        const tblFilters = tbl.querySelectorAll(".tbl-filter");
        for (var t=0; t < tblFilters.length; t++) {
            var tblfilter = tblFilters[t];
            var cIndex = ANNE.A.getFirstParentAttr(tblfilter, "column", "colnum");

            if (tblfilter.value.length) {
                var on_cnt = generic_table.DOM_Filter_single(pgui, tblfilter, tblfilter.value, cIndex);
                
                //check the last filter (the net sum of all visible/not visible)..
                //it will return the on_cnt.. 
                if (tblFilters.length-1==t) {
                    max_rows = max_rows-on_cnt;
                }
            }
        }

        var rowslbl = ANNE.TALK.pluralizeNounIf2(on_cnt, "row", "rows");
        ANNE.A.setHTML("tbl_summary_"+ID, on_cnt+" "+rowslbl);
    }

    static DOM_sortTbl=function($divid, $col_index, $sortDataType, $sortDir) {
        var tbl = document.getElementById($divid); 
        if (tbl==null) return;
        const rows = tbl.querySelectorAll(".row");
        var masterRowParent = rows[0].parentElement;
        var rowsData = Array.prototype.slice.call(rows, 0);
        var dir = 1;
        if ($sortDir=="asc") dir *= -1;

        //sort them  (keep these separated even tho they are the same.. for now)..
        //we may need to handle differently..
        if ($sortDataType=="string" || $sortDataType=="anne") {
            rowsData.sort(function(a,b) {
                if (a.children[$col_index]==undefined) return 0;
                var txta = a.children[$col_index].getAttribute("dataz");//textContent;
                if (b.children[$col_index]==undefined) return 0;
                var txtb = b.children[$col_index].getAttribute("dataz");
                if(txta > txtb) { return -1*dir; }
                if(txta < txtb) { return 1*dir; }
                return 0;
            });
            
        }
        else if ($sortDataType=="numeric") {
            rowsData.sort(function(a,b) {
                if (a.children[$col_index]==undefined) return 0;
                var txta = (a.children[$col_index].getAttribute("dataz"));
                if (b.children[$col_index]==undefined) return 0;
                
                var txtb = (b.children[$col_index].getAttribute("dataz"));
                if(txta > txtb) { return -1*dir; }
                if(txta < txtb) { return 1*dir; }
                return 0;
            });
        }

        else {
            rowsData.sort(function(a,b) {
                if (a.children[$col_index]==undefined) return 0;
                var txta = (a.children[$col_index].getAttribute("dataz"));
                if (b.children[$col_index]==undefined) return 0;
                
                var txtb = (b.children[$col_index].getAttribute("dataz"));
                if(txta > txtb) { return -1*dir; }
                if(txta < txtb) { return 1*dir; }
                return 0;
            });
        }

        //put them back in.
        rowsData.forEach(row => {
            masterRowParent.appendChild(row)
        });
    }
    

    static DOM_Filter_single=function(spObj, el, text, $cIndex) {
        const col = ANNE.A.getFirstParentOf(el, "column");
        const tbldivid = col.getAttribute("tbl");
        const childindex = col.getAttribute("colnum");
        const tbl = document.getElementById(tbldivid);
        console.log("filter tbl:"+tbl+" col "+childindex+" text "+text);
        var onCnt = 0;
        var coldataobj = spObj.INST.cols[$cIndex];
        var filters = coldataobj.filters;
        if (filters==undefined || filters==null) filters = [];
        tbl.querySelectorAll(".row").forEach(row => {
            var dataz = row.children[childindex].innerHTML;//.getAttribute("dataz");
            if (row.style.display=="none") return;  //skip row, go to next.
            if (dataz==null) row.style.display="none";
            else {
                var test = anne_x_filter.testFiltersRuntime(dataz, filters);
                if(test==false) row.style.display="none";
                else {
                    row.style.display="flex";
                    onCnt+=1;
                }
            }
        });

        return onCnt;
    }
}

/************** (from file ./js/gui/widgets/anne_table.js)**/ 

//rename this file anne_paginator.js for Pete's sake.

//just starting this one.. we need it.
class paginator {
    constructor ($parent, $pgui, $guid, $dataArray, $renderCallback, $shard=20) {
        this.pgui = $pgui;
        this._guid = $guid;      //yes, this is also on pgui._guid
        this.ALL_DATA = $dataArray;
        this.data = this.ALL_DATA.concat();
        this.shard = $shard;
        console.log("shard:"+this.shard);
        this.bNOTNEEDED = false;
        this.active_page_index = 0;
        this.parent=$parent;
        this.renderCallback = $renderCallback;
        this.tbl_ptr = null; //ptr back to the table. (added in x, sb added here too)

        //hmm.. altering this to PAGINATOR_INST (will see if i bust something.)
        //i now need $pgui.INST retained for the original value!
        if ($pgui!=null) $pgui.PAGINATOR_INST = this; //back ref to the paginator.. 
    }

    
    /**
     * Build data array from active records
     */
    renderActiveResults = function() {
        ANNE.A.empty(this._el);
        ANNE.A.removeSelf(this._el);
        this.active_page_index = 0;
        this._el=null;
        this.data = [];
        
        for (var d=0; d < this.ALL_DATA.length; d++) {
            if (this.ALL_DATA[d].active==true) {
                this.data.push(this.ALL_DATA[d]);
            }
        }
        
        this.render();
    }


    /**
     * Render data records
     */
    render = function($el_total=null) {
        var totalresults = this.data.length;
        
        if ($el_total!=null) this.el_total = $el_total;

        //if (this.el_total!=null) this.el_total.innerHTML = totalresults +" total";
        
        if (totalresults <= this.shard) {
            this.bNOTNEEDED = true;
            if (this.renderCallback!=null) return this.renderCallback(this.pgui, this.data);
        }

        this.bNOTNEEDED=false;
        this.pagesNeeded = Math.ceil(totalresults/this.shard);
        this._el = ANNE.A.ezEl(this.parent, "div", "anne-paginator");
        if (this._guid!=null) this._el.setAttribute("guid", this._guid);   //incase we lose our "this"
        
        this.active_page_index = 0;
        var this_paginator = this;
        var maxPages = 10;

        var pgL = ANNE.A.ezEl(this._el, "div", "paginator-btn-sides", "<span class='ai-angle-double-left'></span>");
        pgL.setAttribute("dir", "first");
        //pg.style.display="none";

        var pg2L = ANNE.A.ezEl(this._el, "div", "paginator-btn-sides", "<span class='ai-angle-left'></span>");
        pg2L.setAttribute("dir", "prev");
        //pg2.style.display="none";
        
        var pgs = Math.min(maxPages, this.pagesNeeded);

        for (var p=0; p < pgs; p++) {
            var pg = ANNE.A.ezEl(this._el, "div", "paginator-btn", p+1);
            pg.setAttribute("page_i", p);
            if (p==0) pg.classList.add("active");

            pg.addEventListener("click", $e=>{
                var pg = $e.target;
                var thispaginatorel = ANNE.A.getFirstParentOf(pg, "anne-paginator");
                ANNE.A.removeAddClassOfAllIn(thispaginatorel, ".paginator-btn", "active", pg);

                var page_index = parseInt(pg.getAttribute("page_i"));
                if (page_index == this_paginator.active_page_index) return [];

                var moving = "down";
                if (page_index>this_paginator.active_page_index) moving="up";
                var start_index = (this_paginator.shard*page_index);
                var pagedata = this_paginator.data.slice(start_index, start_index+this_paginator.shard);
                
                //shift buttons..
                //we must pass the right el in case we have multiple instances
                //of this onscreen.
                this_paginator.active_page_index = page_index;

                this_paginator.fixBtns(thispaginatorel, this_paginator.active_page_index, this_paginator.pagesNeeded, moving);

                if (this_paginator.renderCallback!=null) this_paginator.renderCallback(this_paginator.pgui, pagedata);
            });
        }

        var pgR = ANNE.A.ezEl(this._el, "div", "paginator-btn-sides", "<span class='ai-angle-right'></span>");
        pgR.setAttribute("dir", "next"); 

        var pg2R = ANNE.A.ezEl(this._el, "div", "paginator-btn-sides", "<span class='ai-angle-double-right'></span>");
        pg2R.setAttribute("dir", "last");

        if (this.pagesNeeded < 10) {
            pgR.style.display="none";
            pg2R.style.display="none";

            pgL.style.display="none";
            pg2L.style.display="none";
        }
       
        var btns = this._el.querySelectorAll(".paginator-btn-sides");
        for (var b=0; b < btns.length; b++) {
            btns[b].addEventListener("click", $e=>{
                var btn = $e.target;
                if (btn.classList.contains("paginator-btn-sides")==false) btn = btn.parentElement;
                var thispaginatorel = ANNE.A.getFirstParentOf(btn, "anne-paginator");
                var page_index = this_paginator.active_page_index;
                var dir = btn.getAttribute("dir");
                if (dir=="first")page_index = 0;
                else if (dir=="prev") page_index = page_index-1;
                else if (dir=="next") page_index = page_index+1;
                else if (dir=="last") page_index = this_paginator.pagesNeeded-1;
            
                var moving = "down";
                if (page_index>this_paginator.active_page_index) moving="up";

                var start_index = (this_paginator.shard*page_index);
                var pagedata = this_paginator.data.slice(start_index, start_index+this_paginator.shard);
                
                this_paginator.active_page_index = page_index;
                this_paginator.fixBtns(thispaginatorel, this_paginator.active_page_index, this_paginator.pagesNeeded, moving);

                var pagedata = this_paginator.data.slice(start_index, start_index+this_paginator.shard);
        
                if (this_paginator.renderCallback!=null) this_paginator.renderCallback(this_paginator.pgui, pagedata);
            });
        }


        var start_index = (this_paginator.shard*0);
        var pagedata = this_paginator.data.slice(start_index, start_index+this_paginator.shard);

        this.fixTotals(0);

        if (this.renderCallback!=null) this.renderCallback(this.pgui, pagedata);
    }

    fixTotals=function($activeIndex) {
        if (this.el_total!=null) {
            var start = parseInt($activeIndex*(this.shard))+1;
            var end = start+parseInt(this.shard)-1;
            this.el_total.innerHTML= this.data.length+" total &nbsp;<span class='lbl-rows-showing'>(showing&nbsp;"+start+"&nbsp;-&nbsp;"+end+")</span>";
        }
    }

    fixBtns  = function($el, $activeIndex, $pagesNeeded, $moving="up") {
        var btns = $el.querySelectorAll(".paginator-btn-sides");
        if (btns.length!=4) console.warn("btns "+btns.length+" should be 4!!!");
        var firstBtn = btns[0];
        var prevBtn = btns[1];
        var nextBtn = btns[2];
        var endBtn = btns[3];
        //console.log("%c(fixBtns) pages needed:"+$pagesNeeded, TALK.bgpurple);

        this.fixTotals($activeIndex);

        if ($pagesNeeded > 10) {
            var btns = $el.querySelectorAll(".paginator-btn");

            //5 left, 4 to the right.
            //try to center our page in the paginator at all times.
            if ($activeIndex > 0) {
                //console.log("%c(fixBtns) activeIndex:"+$activeIndex, TALK.bgpurple);

                if (($pagesNeeded-$activeIndex) <4) {       //close to end (right)
                    ANNE.A.addClass(endBtn, "disabled");
                    ANNE.A.addClass(nextBtn, "disabled");
                    ANNE.A.removeClass(firstBtn, "disabled");
                    ANNE.A.removeClass(prevBtn, "disabled");

                    //base offset from the right...
                    var offset = $pagesNeeded-($activeIndex+1);

                    for (var b=0; b < btns.length; b++) {
                        var newpage = ($pagesNeeded-1)-9+(b);
                        console.log("newpage:"+newpage);
                        btns[b].setAttribute("page_i", newpage);
                        btns[b].innerHTML = newpage+1;
                        if (newpage!=$activeIndex) btns[b].classList.remove("active");
                        else btns[b].classList.add("active");
                    }
                }

                //do nothing.
                else if ($activeIndex<=4) {
                    return;
                }

                else {
                    ANNE.A.removeClass(endBtn, "disabled");
                    ANNE.A.removeClass(nextBtn, "disabled");
                    ANNE.A.removeClass(firstBtn, "disabled");
                    ANNE.A.removeClass(prevBtn, "disabled");
                    
                    var minOffset = Math.max($activeIndex-4, 0);
                    if ((minOffset+10) > this.pagesNeeded)  {
                        minOffset -= (minOffset+10)-(this.pagesNeeded);
                    }

                    for (var b=0; b < btns.length; b++) {
                        var newpage = minOffset+(b);
                        //console.log("newpage:"+newpage);
                        btns[b].setAttribute("page_i", newpage);
                        btns[b].innerHTML = newpage+1;
                        btns[b].classList.remove("active");
                        if (newpage!=$activeIndex) btns[b].classList.remove("active");
                        else btns[b].classList.add("active");
                    }
                }
            }

            //reset standard position.
            else {
                ANNE.A.removeClass(endBtn, "disabled");
                ANNE.A.removeClass(nextBtn, "disabled");
                ANNE.A.addClass(firstBtn, "disabled");
                ANNE.A.addClass(prevBtn, "disabled");
                

                for (var b=0; b < btns.length; b++) {
                    var newpage = b;
                    btns[b].setAttribute("page_i", newpage);
                    btns[b].innerHTML = newpage+1;
                    btns[b].classList.remove("active");
                    if (newpage!=$activeIndex) btns[b].classList.remove("active");
                    else btns[b].classList.add("active");
                }
            }
        }

        //paginator was needed by the extra << < and > >> buttons were not since
        //we had less than 10 pages of results.
        else {
            ANNE.A.addClass(endBtn, "disabled");
            ANNE.A.addClass(nextBtn, "disabled");
            ANNE.A.addClass(firstBtn, "disabled");
            ANNE.A.addClass(prevBtn, "disabled");
            
            for (var b=0; b < btns.length; b++) {
                if (b!=$activeIndex) btns[b].classList.remove("active");
                else btns[b].classList.add("active");
            }
        }
    }
}


//rename this file anne_paginator.js for Pete's sake.

//just starting this one.. we need it.
class paginator_nextx {
    constructor ($parent, $guid, $renderCallback, $shard=20) {
        this._guid = $guid;      //yes, this is also on pgui._guid
        this.shard = $shard;
        this.bNOTNEEDED = false;
        this.active_page_index = 0;
        this.active_page = null;
        this.parent=$parent;
        this.renderCallback = $renderCallback;
        this.tbl_ptr = null; //ptr back to the table.
    }

    
    /**
     * Build data array from active records
     * NOTE:
     * This (was) for sorting.. and is not going to work like this for nextx where
     * we are looking to dynamically load data on demand in the table.
     */
    renderActiveResults = function() {
        ANNE.A.empty(this._el);
        ANNE.A.removeSelf(this._el);
        this.active_page_index = 0;
        this._el=null;
        this.render();
    }


    newDataInTime = function($newData, $datacol="height") {
        console.log("new data!", $newData);

        var last_page = this._pages.get(this.pagesNeeded-1);
        
        //we need a new page!
        if ($newData[$datacol] > last_page.end_index) {
            //new page!
            var pobj = {page_index:last_page.page_index+1, loaded:true};
            pobj.start_index = (this.shard*last_page.page_index+1);
            pobj.end_index = pobj.start_index+this.shard-1
            pobj.shard = this.shard;
            this._pages.addU(pobj);
            this.pagesNeded+=1;
            
            var bREFRESHNOW=false;
            if (last_page.page_index==this.active_page_index) {
                bREFRESHNOW=true;
            }
            this.active_page_index = last_page.page_index+1;
            this.active_page = pobj;

            if (bREFRESHNOW==true) {
                var slice = this.getPageSlice(this.active_age)
                return this.renderCallback(slice, this.active_page, $newData);
            }
        }

        else {
            //if the page we just added to is the page we're looking at...
            if (last_page.page_index==this.active_page_index) {
                var slice = this.getPageSlice(last_page);
                //we want to add the record into the table!!!
                if (slice==null) {
                    console.log("slice is null");
                }
                return this.renderCallback(slice, last_page, $newData);
            }
        }
    }


    getPageSlice=function($page){
        if ($page==undefined || $page==null) return;
        return this.tbl_ptr.data_arr_map.getSlice2($page.start_index, $page.end_index);
    }


    /**
     * Render data records
     */
    render = function($el_total=null) {
        var totalresults = this.tbl_ptr.data_arr_map.arr.length;

        if ($el_total!=null) this.el_total = ANNE.A.__getEl($el_total);

        //if (this.el_total!=null) this.el_total.innerHTML = totalresults +" total";
        
        this.bNOTNEEDED=false;
        this.pagesNeeded = Math.ceil(totalresults/this.shard);
        this._pages = new arraymap();
        this._pages.setPrimaryKey("page_index");

        //chop our pages up ready for use
        for (var p=0; p < this.pagesNeeded; p++) {
            var pobj = {page_index:p, loaded:false};
            pobj.start_index = (this.shard*p);
            pobj.end_index = pobj.start_index+this.shard-1
            pobj.shard = this.shard;
            this._pages.addU(pobj);
        }

        if (totalresults <= this.shard) {
            this.bNOTNEEDED = true;
            if (this.renderCallback!=null) {
                this.active_page_index=0;
                this.active_page = this._pages.get(this.active_page_index);
                
                var slice2 = this.getPageSlice(this.active_page)
                if (slice2==null) {
                    console.log("slice2 is null");
                }

                return this.renderCallback(slice2, pobj);
            }
        }


        this._el = ANNE.A.ezEl(this.parent, "div", "anne-paginator");
        if (this._guid!=null) this._el.setAttribute("guid", this._guid);   //incase we lose our "this"
        
        this.active_page_index = this.pagesNeeded-1;
        this.active_page = this._pages.get(this.active_page_index);
        var this_paginator = this;

        var pgL = ANNE.A.ezEl(this._el, "div", "paginator-btn-sidesx", "first");
        pgL.setAttribute("dir", "first");

        var pg2L = ANNE.A.ezEl(this._el, "div", "paginator-btn-sidesx", "prev "+this.shard);
        pg2L.setAttribute("dir", "prev");

        var pgR = ANNE.A.ezEl(this._el, "div", "paginator-btn-sidesx", "next"+this.shard);
        pgR.setAttribute("dir", "next"); 

        var pg2R = ANNE.A.ezEl(this._el, "div", "paginator-btn-sidesx", "last");
        pg2R.setAttribute("dir", "last");

        /************************************************************* */
        //first, prev, next, last.
        var btns = this._el.querySelectorAll(".paginator-btn-sidesx");
        for (var b=0; b < btns.length; b++) {
            btns[b].addEventListener("click", async($e)=>{
                var btn = ANNE.A.getFirstParentOf($e.target, "paginator-btn-sidesx");
                var thispaginatorel = ANNE.A.getFirstParentOf(btn, "anne-paginator");
                var active_page_index = this_paginator.active_page_index;
                var active_page = this_paginator._pages.get(this.active_page_index);
                var dir = btn.getAttribute("dir");

                var new_page_index = 0;
                if (dir=="first")new_page_index = 0;
                else if (dir=="prev") new_page_index = active_page_index-1;
                else if (dir=="next") new_page_index = active_page_index+1;
                else if (dir=="last") new_page_index = this_paginator.pagesNeeded-1;
            
                var moving = "down";

                //direction we move from the present location.
                if (new_page_index>this_paginator.active_page_index) moving="up";

                var new_active_page = this_paginator._pages.get(new_page_index);
                console.log("new_page_index:"+new_page_index+" new start_index:"+new_active_page.start_index);                
                this_paginator.active_page_index = new_page_index;
                this_paginator.active_page = new_active_page;

                this_paginator.fixBtns(thispaginatorel, new_active_page.page_index, this_paginator.pagesNeeded, moving);

                if (this_paginator.renderCallback!=null) {
                    var slice3 = this_paginator.getPageSlice(new_active_page);
                    if (slice3==null) {
                        console.log("slice3 is null");
                    }
                    this_paginator.renderCallback(slice3, new_active_page);
                }
            });
        }

        /************************************************************* */
        //init to last newest page.
        /************************************************************* */
        this.fixTotals(0);
        this.fixBtns(this._el, this.active_page.page_index, this.pagesNeeded);
        var slice4 = this.getPageSlice(this.active_page);
        if (this.renderCallback!=null) {
            if (slice4==null) {
                console.log("slice4 is null");
            }
            this.renderCallback(slice4, this.active_page);
        }
    }
    

    /**
     * 
     * @param {*} $activeIndex 
     */
    fixTotals=function($activeIndex) {
        if (this.el_total!=null) {
            var total = this.tbl_ptr.data_arr_map.arr.length;
            var start = parseInt($activeIndex*(this.shard));
            var end = Math.min(total-1, start+parseInt(this.shard)-1);
            this.el_total.innerHTML= "<span class='total-paginator'>"+ANNE.UTILS.commas(total-1)+"</span>&nbsp;total &nbsp;<span class='lbl-rows-showing'>(showing&nbsp;"+ANNE.UTILS.commas(start)+"&nbsp;-&nbsp;"+ANNE.UTILS.commas(end)+")</span>";
        }
    }

    
    goToPageIndex=async($pageIndex)=> {
        this.active_page_index = $pageIndex;
        this.active_page = this._pages.get(this.active_page_index);
        var slice5  = this.getPageSlice(this.active_page);
        if (slice5==null) console.log("slice5 in goToPageIndex is null");
        return await this.renderCallback(slice5, this.active_page);
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $activeIndex 
     * @param {*} $pagesNeeded 
     */
    fixBtns  = function($el, $activeIndex, $pagesNeeded) {
        var btns = $el.querySelectorAll(".paginator-btn-sidesx");
        if (btns.length!=4) console.warn("btns "+btns.length+" should be 4!!!");
        var firstBtn = btns[0];
        var prevBtn = btns[1];
        var nextBtn = btns[2];
        var endBtn = btns[3];
        this.fixTotals($activeIndex);

        ANNE.A.removeClass(endBtn, "disabled");
        ANNE.A.removeClass(nextBtn, "disabled");
        ANNE.A.removeClass(firstBtn, "disabled");
        ANNE.A.removeClass(prevBtn, "disabled");

        ANNE.A.removeClass(endBtn, "active");
        ANNE.A.removeClass(nextBtn, "active");
        ANNE.A.removeClass(firstBtn, "active");
        ANNE.A.removeClass(prevBtn, "active");

        if ($activeIndex==0 || $pagesNeeded==1) {
            ANNE.A.addClass(firstBtn, "disabled");
            ANNE.A.addClass(prevBtn, "disabled");
            ANNE.A.addClass(endBtn, "active");
            ANNE.A.addClass(nextBtn, "active");
        }
        else if ($activeIndex>0 && $activeIndex>=($pagesNeeded-1)) {
            ANNE.A.addClass(endBtn, "disabled");
            ANNE.A.addClass(nextBtn, "disabled");
            ANNE.A.addClass(firstBtn, "active");
            ANNE.A.addClass(prevBtn, "active");
        }
        else {
            ANNE.A.addClass(endBtn, "active");
            ANNE.A.addClass(nextBtn, "active");
            ANNE.A.addClass(firstBtn, "active");
            ANNE.A.addClass(prevBtn, "active");
        }
    }
}

/************** (from file ./js/gui/widgets/anne_paginator.js)**/ 

/*!
 * GSAP 3.4.0
 * https://greensock.com
 * 
 * @license Copyright 2020, GreenSock. All rights reserved.
 * Subject to the terms at https://greensock.com/standard-license or for Club GreenSock members, the agreement issued with that membership.
 * @author: Jack Doyle, jack@greensock.com
 */

!function(t,e){"object"==typeof exports&&"undefined"!=typeof module?e(exports):"function"==typeof define&&define.amd?define(["exports"],e):e((t=t||self).window=t.window||{})}(this,function(e){"use strict";function _inheritsLoose(t,e){t.prototype=Object.create(e.prototype),(t.prototype.constructor=t).__proto__=e}function _assertThisInitialized(t){if(void 0===t)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function n(t){return"string"==typeof t}function o(t){return"function"==typeof t}function p(t){return"number"==typeof t}function q(t){return void 0===t}function r(t){return"object"==typeof t}function s(t){return!1!==t}function t(){return"undefined"!=typeof window}function u(t){return o(t)||n(t)}function K(t){return(l=pt(t,at))&&ie}function L(t,e){return console.warn("Invalid property",t,"set to",e,"Missing plugin? gsap.registerPlugin()")}function M(t,e){return!e&&console.warn(t)}function N(t,e){return t&&(at[t]=e)&&l&&(l[t]=e)||at}function O(){return 0}function Y(t){var e,i,n=t[0];if(r(n)||o(n)||(t=[t]),!(e=(n._gsap||{}).harness)){for(i=dt.length;i--&&!dt[i].targetTest(n););e=dt[i]}for(i=t.length;i--;)t[i]&&(t[i]._gsap||(t[i]._gsap=new Et(t[i],e)))||t.splice(i,1);return t}function Z(t){return t._gsap||Y(yt(t))[0]._gsap}function $(t,e){var r=t[e];return o(r)?t[e]():q(r)&&t.getAttribute(e)||r}function _(t,e){return(t=t.split(",")).forEach(e)||t}function aa(t){return Math.round(1e5*t)/1e5||0}function ba(t,e){for(var r=e.length,i=0;t.indexOf(e[i])<0&&++i<r;);return i<r}function ca(t,e,r){var i,n=p(t[1]),a=(n?2:1)+(e<2?0:1),o=t[a];if(n&&(o.duration=t[1]),o.parent=r,e){for(i=o;r&&!("immediateRender"in i);)i=r.vars.defaults||{},r=s(r.vars.inherit)&&r.parent;o.immediateRender=s(i.immediateRender),e<2?o.runBackwards=1:o.startAt=t[a-1]}return o}function da(){var t,e,r=ot.length,i=ot.slice(0);for(ut={},t=ot.length=0;t<r;t++)(e=i[t])&&e._lazy&&(e.render(e._lazy[0],e._lazy[1],!0)._lazy=0)}function ea(t,e,r,i){ot.length&&da(),t.render(e,r,i),ot.length&&da()}function fa(t){var e=parseFloat(t);return(e||0===e)&&(t+"").match(nt).length<2?e:t}function ga(t){return t}function ha(t,e){for(var r in e)r in t||(t[r]=e[r]);return t}function ia(t,e){for(var r in e)r in t||"duration"===r||"ease"===r||(t[r]=e[r])}function ka(t,e){for(var i in e)t[i]=r(e[i])?ka(t[i]||(t[i]={}),e[i]):e[i];return t}function la(t,e){var r,i={};for(r in t)r in e||(i[r]=t[r]);return i}function ma(t){var e=t.parent||E,r=t.keyframes?ia:ha;if(s(t.inherit))for(;e;)r(t,e.vars.defaults),e=e.parent||e._dp;return t}function pa(t,e,r,i){void 0===r&&(r="_first"),void 0===i&&(i="_last");var n=e._prev,a=e._next;n?n._next=a:t[r]===e&&(t[r]=a),a?a._prev=n:t[i]===e&&(t[i]=n),e._next=e._prev=e.parent=null}function qa(t,e){!t.parent||e&&!t.parent.autoRemoveChildren||t.parent.remove(t),t._act=0}function ra(t){for(var e=t;e;)e._dirty=1,e=e.parent;return t}function ua(t){return t._repeat?_t(t._tTime,t=t.duration()+t._rDelay)*t:0}function wa(t,e){return(t-e._start)*e._ts+(0<=e._ts?0:e._dirty?e.totalDuration():e._tDur)}function xa(t){return t._end=aa(t._start+(t._tDur/Math.abs(t._ts||t._rts||R)||0))}function ya(t,e){var r=t._dp;return r&&r.smoothChildTiming&&t._ts&&(t._start=aa(t._dp._time-(0<t._ts?e/t._ts:((t._dirty?t.totalDuration():t._tDur)-e)/-t._ts)),xa(t),r._dirty||ra(r)),t}function za(t,e){var r;if((e._time||e._initted&&!e._dur)&&(r=wa(t.rawTime(),e),(!e._dur||gt(0,e.totalDuration(),r)-e._tTime>R)&&e.render(r,!0)),ra(t)._dp&&t._initted&&t._time>=t._dur&&t._ts){if(t._dur<t.duration())for(r=t;r._dp;)0<=r.rawTime()&&r.totalTime(r._tTime),r=r._dp;t._zTime=-R}}function Aa(t,e,r,i){return e.parent&&qa(e),e._start=aa(r+e._delay),e._end=aa(e._start+(e.totalDuration()/Math.abs(e.timeScale())||0)),function _addLinkedListItem(t,e,r,i,n){void 0===r&&(r="_first"),void 0===i&&(i="_last");var a,s=t[i];if(n)for(a=e[n];s&&s[n]>a;)s=s._prev;s?(e._next=s._next,s._next=e):(e._next=t[r],t[r]=e),e._next?e._next._prev=e:t[i]=e,e._prev=s,e.parent=e._dp=t}(t,e,"_first","_last",t._sort?"_start":0),t._recent=e,i||za(t,e),t}function Ba(t,e){return(at.ScrollTrigger||L("scrollTrigger",e))&&at.ScrollTrigger.create(e,t)}function Ca(t,e,r,i){return qt(t,e),t._initted?!r&&t._pt&&(t._dur&&!1!==t.vars.lazy||!t._dur&&t.vars.lazy)&&d!==Mt.frame?(ot.push(t),t._lazy=[e,i],1):void 0:1}function Fa(t,e,r){var i=t._repeat,n=aa(e)||0;return t._dur=n,t._tDur=i?i<0?1e10:aa(n*(i+1)+t._rDelay*i):n,t._time>n&&(t._time=n,t._tTime=Math.min(t._tTime,t._tDur)),r||ra(t.parent),t.parent&&xa(t),t}function Ga(t){return t instanceof Rt?ra(t):Fa(t,t._dur)}function Ia(t,e){var r,i,a=t.labels,s=t._recent||mt,o=t.duration()>=z?s.endTime(!1):t._dur;return n(e)&&(isNaN(e)||e in a)?"<"===(r=e.charAt(0))||">"===r?("<"===r?s._start:s.endTime(0<=s._repeat))+(parseFloat(e.substr(1))||0):(r=e.indexOf("="))<0?(e in a||(a[e]=o),a[e]):(i=+(e.charAt(r-1)+e.substr(r+1)),1<r?Ia(t,e.substr(0,r-1))+i:o+i):null==e?o:+e}function Ja(t,e){return t||0===t?e(t):e}function La(t){return(t+"").substr((parseFloat(t)+"").length)}function Oa(t,e){return t&&r(t)&&"length"in t&&(!e&&!t.length||t.length-1 in t&&r(t[0]))&&!t.nodeType&&t!==i}function Ra(t){return t.sort(function(){return.5-Math.random()})}function Sa(t){if(o(t))return t;var p=r(t)?t:{each:t},_=Ft(p.ease),m=p.from||0,g=parseFloat(p.base)||0,v={},e=0<m&&m<1,y=isNaN(m)||e,T=p.axis,b=m,w=m;return n(m)?b=w={center:.5,edges:.5,end:1}[m]||0:!e&&y&&(b=m[0],w=m[1]),function(t,e,r){var i,n,a,s,o,u,h,l,f,d=(r||p).length,c=v[d];if(!c){if(!(f="auto"===p.grid?0:(p.grid||[1,z])[1])){for(h=-z;h<(h=r[f++].getBoundingClientRect().left)&&f<d;);f--}for(c=v[d]=[],i=y?Math.min(f,d)*b-.5:m%f,n=y?d*w/f-.5:m/f|0,l=z,u=h=0;u<d;u++)a=u%f-i,s=n-(u/f|0),c[u]=o=T?Math.abs("y"===T?s:a):j(a*a+s*s),h<o&&(h=o),o<l&&(l=o);"random"===m&&Ra(c),c.max=h-l,c.min=l,c.v=d=(parseFloat(p.amount)||parseFloat(p.each)*(d<f?d-1:T?"y"===T?d/f:f:Math.max(f,d/f))||0)*("edges"===m?-1:1),c.b=d<0?g-d:g,c.u=La(p.amount||p.each)||0,_=_&&d<0?Dt(_):_}return d=(c[t]-c.min)/c.max||0,aa(c.b+(_?_(d):d)*c.v)+c.u}}function Ta(e){var r=e<1?Math.pow(10,(e+"").length-2):1;return function(t){return Math.floor(Math.round(parseFloat(t)/e)*e*r)/r+(p(t)?0:La(t))}}function Ua(u,t){var h,l,e=J(u);return!e&&r(u)&&(h=e=u.radius||z,u.values?(u=yt(u.values),(l=!p(u[0]))&&(h*=h)):u=Ta(u.increment)),Ja(t,e?o(u)?function(t){return l=u(t),Math.abs(l-t)<=h?l:t}:function(t){for(var e,r,i=parseFloat(l?t.x:t),n=parseFloat(l?t.y:0),a=z,s=0,o=u.length;o--;)(e=l?(e=u[o].x-i)*e+(r=u[o].y-n)*r:Math.abs(u[o]-i))<a&&(a=e,s=o);return s=!h||a<=h?u[s]:t,l||s===t||p(t)?s:s+La(t)}:Ta(u))}function Va(t,e,r,i){return Ja(J(t)?!e:!0===r?!!(r=0):!i,function(){return J(t)?t[~~(Math.random()*t.length)]:(r=r||1e-5)&&(i=r<1?Math.pow(10,(r+"").length-2):1)&&Math.floor(Math.round((t+Math.random()*(e-t))/r)*r*i)/i})}function Za(e,r,t){return Ja(t,function(t){return e[~~r(t)]})}function ab(t){for(var e,r,i,n,a=0,s="";~(e=t.indexOf("random(",a));)i=t.indexOf(")",e),n="["===t.charAt(e+7),r=t.substr(e+7,i-e-7).match(n?nt:W),s+=t.substr(a,e-a)+Va(n?r:+r[0],+r[1],+r[2]||1e-5),a=i+1;return s+t.substr(a,t.length-a)}function db(t,e,r){var i,n,a,s=t.labels,o=z;for(i in s)(n=s[i]-e)<0==!!r&&n&&o>(n=Math.abs(n))&&(a=i,o=n);return a}function fb(t){return qa(t),t.progress()<1&&bt(t,"onInterrupt"),t}function kb(t,e,r){return(6*(t=t<0?t+1:1<t?t-1:t)<1?e+(r-e)*t*6:t<.5?r:3*t<2?e+(r-e)*(2/3-t)*6:e)*wt+.5|0}function lb(t,e,r){var i,n,a,s,o,u,h,l,f,d,c=t?p(t)?[t>>16,t>>8&wt,t&wt]:0:xt.black;if(!c){if(","===t.substr(-1)&&(t=t.substr(0,t.length-1)),xt[t])c=xt[t];else if("#"===t.charAt(0))4===t.length&&(t="#"+(i=t.charAt(1))+i+(n=t.charAt(2))+n+(a=t.charAt(3))+a),c=[(t=parseInt(t.substr(1),16))>>16,t>>8&wt,t&wt];else if("hsl"===t.substr(0,3))if(c=d=t.match(W),e){if(~t.indexOf("="))return c=t.match(H),r&&c.length<4&&(c[3]=1),c}else s=+c[0]%360/360,o=c[1]/100,i=2*(u=c[2]/100)-(n=u<=.5?u*(o+1):u+o-u*o),3<c.length&&(c[3]*=1),c[0]=kb(s+1/3,i,n),c[1]=kb(s,i,n),c[2]=kb(s-1/3,i,n);else c=t.match(W)||xt.transparent;c=c.map(Number)}return e&&!d&&(i=c[0]/wt,n=c[1]/wt,a=c[2]/wt,u=((h=Math.max(i,n,a))+(l=Math.min(i,n,a)))/2,h===l?s=o=0:(f=h-l,o=.5<u?f/(2-h-l):f/(h+l),s=h===i?(n-a)/f+(n<a?6:0):h===n?(a-i)/f+2:(i-n)/f+4,s*=60),c[0]=~~(s+.5),c[1]=~~(100*o+.5),c[2]=~~(100*u+.5)),r&&c.length<4&&(c[3]=1),c}function mb(t){var r=[],i=[],n=-1;return t.split(kt).forEach(function(t){var e=t.match(tt)||[];r.push.apply(r,e),i.push(n+=e.length+1)}),r.c=i,r}function nb(t,e,r){var i,n,a,s,o="",u=(t+o).match(kt),h=e?"hsla(":"rgba(",l=0;if(!u)return t;if(u=u.map(function(t){return(t=lb(t,e,1))&&h+(e?t[0]+","+t[1]+"%,"+t[2]+"%,"+t[3]:t.join(","))+")"}),r&&(a=mb(t),(i=r.c).join(o)!==a.c.join(o)))for(s=(n=t.replace(kt,"1").split(tt)).length-1;l<s;l++)o+=n[l]+(~i.indexOf(l)?u.shift()||h+"0,0,0,0)":(a.length?a:u.length?u:r).shift());if(!n)for(s=(n=t.split(kt)).length-1;l<s;l++)o+=n[l]+u[l];return o+n[s]}function qb(t){var e,r=t.join(" ");if(kt.lastIndex=0,kt.test(r))return e=Ot.test(r),t[1]=nb(t[1],e),t[0]=nb(t[0],e,mb(t[1])),!0}function yb(t){var e=(t+"").split("("),r=At[e[0]];return r&&1<e.length&&r.config?r.config.apply(null,~t.indexOf("{")?[function _parseObjectInString(t){for(var e,r,i,n={},a=t.substr(1,t.length-3).split(":"),s=a[0],o=1,u=a.length;o<u;o++)r=a[o],e=o!==u-1?r.lastIndexOf(","):r.length,i=r.substr(0,e),n[s]=isNaN(i)?i.replace(St,"").trim():+i,s=r.substr(e+1).trim();return n}(e[1])]:rt.exec(t)[1].split(",").map(fa)):At._CE&&Pt.test(t)?At._CE("",t):r}function Ab(t,e){for(var r,i=t._first;i;)i instanceof Rt?Ab(i,e):!i.vars.yoyoEase||i._yoyo&&i._repeat||i._yoyo===e||(i.timeline?Ab(i.timeline,e):(r=i._ease,i._ease=i._yEase,i._yEase=r,i._yoyo=e)),i=i._next}function Cb(t,e,r,i){void 0===r&&(r=function easeOut(t){return 1-e(1-t)}),void 0===i&&(i=function easeInOut(t){return t<.5?e(2*t)/2:1-e(2*(1-t))/2});var n,a={easeIn:e,easeOut:r,easeInOut:i};return _(t,function(t){for(var e in At[t]=at[t]=a,At[n=t.toLowerCase()]=r,a)At[n+("easeIn"===e?".in":"easeOut"===e?".out":".inOut")]=At[t+"."+e]=a[e]}),a}function Db(e){return function(t){return t<.5?(1-e(1-2*t))/2:.5+e(2*(t-.5))/2}}function Eb(r,t,e){function il(t){return 1===t?1:i*Math.pow(2,-10*t)*Q((t-a)*n)+1}var i=1<=t?t:1,n=(e||(r?.3:.45))/(t<1?t:1),a=n/B*(Math.asin(1/i)||0),s="out"===r?il:"in"===r?function(t){return 1-il(1-t)}:Db(il);return n=B/n,s.config=function(t,e){return Eb(r,t,e)},s}function Fb(e,r){function ql(t){return t?--t*t*((r+1)*t+r)+1:0}void 0===r&&(r=1.70158);var t="out"===e?ql:"in"===e?function(t){return 1-ql(1-t)}:Db(ql);return t.config=function(t){return Fb(e,t)},t}var E,i,a,h,l,f,d,c,m,g,v,y,T,b,w,x,k,C,A,P,S,D,F,U={autoSleep:120,force3D:"auto",nullTargetWarn:1,units:{lineHeight:""}},I={duration:.5,overwrite:!1,delay:0},z=1e8,R=1/z,B=2*Math.PI,X=B/4,V=0,j=Math.sqrt,G=Math.cos,Q=Math.sin,J=Array.isArray,W=/(?:-?\.?\d|\.)+/gi,H=/[-+=.]*\d+[.e\-+]*\d*[e\-\+]*\d*/g,tt=/[-+=.]*\d+[.e-]*\d*[a-z%]*/g,et=/[-+=.]*\d+(?:\.|e-|e)*\d*/gi,rt=/\(([^()]+)\)/i,it=/[+-]=-?[\.\d]+/,nt=/[#\-+.]*\b[a-z\d-=+%.]+/gi,at={},st={},ot=[],ut={},ht={},lt={},ft=30,dt=[],ct="",pt=function _merge(t,e){for(var r in e)t[r]=e[r];return t},_t=function _animationCycle(t,e){return(t/=e)&&~~t===t?~~t-1:~~t},mt={_start:0,endTime:O},gt=function _clamp(t,e,r){return r<t?t:e<r?e:r},vt=[].slice,yt=function toArray(t,e){return!n(t)||e||!a&&Ct()?J(t)?function _flatten(t,e,r){return void 0===r&&(r=[]),t.forEach(function(t){return n(t)&&!e||Oa(t,1)?r.push.apply(r,yt(t)):r.push(t)})||r}(t,e):Oa(t)?vt.call(t,0):t?[t]:[]:vt.call(h.querySelectorAll(t),0)},Tt=function mapRange(e,t,r,i,n){var a=t-e,s=i-r;return Ja(n,function(t){return r+((t-e)/a*s||0)})},bt=function _callback(t,e,r){var i,n,a=t.vars,s=a[e];if(s)return i=a[e+"Params"],n=a.callbackScope||t,r&&ot.length&&da(),i?s.apply(n,i):s.call(n)},wt=255,xt={aqua:[0,wt,wt],lime:[0,wt,0],silver:[192,192,192],black:[0,0,0],maroon:[128,0,0],teal:[0,128,128],blue:[0,0,wt],navy:[0,0,128],white:[wt,wt,wt],olive:[128,128,0],yellow:[wt,wt,0],orange:[wt,165,0],gray:[128,128,128],purple:[128,0,128],green:[0,128,0],red:[wt,0,0],pink:[wt,192,203],cyan:[0,wt,wt],transparent:[wt,wt,wt,0]},kt=function(){var t,e="(?:\\b(?:(?:rgb|rgba|hsl|hsla)\\(.+?\\))|\\B#(?:[0-9a-f]{3}){1,2}\\b";for(t in xt)e+="|"+t+"\\b";return new RegExp(e+")","gi")}(),Ot=/hsl[a]?\(/,Mt=(b=Date.now,w=500,x=33,k=b(),C=k,P=A=1/240,T={time:0,frame:0,tick:function tick(){kk(!0)},wake:function wake(){f&&(!a&&t()&&(i=a=window,h=i.document||{},at.gsap=ie,(i.gsapVersions||(i.gsapVersions=[])).push(ie.version),K(l||i.GreenSockGlobals||!i.gsap&&i||{}),y=i.requestAnimationFrame),g&&T.sleep(),v=y||function(t){return setTimeout(t,1e3*(P-T.time)+1|0)},m=1,kk(2))},sleep:function sleep(){(y?i.cancelAnimationFrame:clearTimeout)(g),m=0,v=O},lagSmoothing:function lagSmoothing(t,e){w=t||1e8,x=Math.min(e,w,0)},fps:function fps(t){A=1/(t||240),P=T.time+A},add:function add(t){S.indexOf(t)<0&&S.push(t),Ct()},remove:function remove(t){var e;~(e=S.indexOf(t))&&S.splice(e,1)},_listeners:S=[]}),Ct=function _wake(){return!m&&Mt.wake()},At={},Pt=/^[\d.\-M][\d.\-,\s]/,St=/["']/g,Dt=function _invertEase(e){return function(t){return 1-e(1-t)}},Ft=function _parseEase(t,e){return t&&(o(t)?t:At[t]||yb(t))||e};function kk(e){var t,r,i=b()-C,n=!0===e;w<i&&(k+=i-x),C+=i,T.time=(C-k)/1e3,(0<(t=T.time-P)||n)&&(T.frame++,P+=t+(A<=t?.004:A-t),r=1),n||(g=v(kk)),r&&S.forEach(function(t){return t(T.time,i,T.frame,e)})}function Hl(t){return t<F?D*t*t:t<.7272727272727273?D*Math.pow(t-1.5/2.75,2)+.75:t<.9090909090909092?D*(t-=2.25/2.75)*t+.9375:D*Math.pow(t-2.625/2.75,2)+.984375}_("Linear,Quad,Cubic,Quart,Quint,Strong",function(t,e){var r=e<5?e+1:e;Cb(t+",Power"+(r-1),e?function(t){return Math.pow(t,r)}:function(t){return t},function(t){return 1-Math.pow(1-t,r)},function(t){return t<.5?Math.pow(2*t,r)/2:1-Math.pow(2*(1-t),r)/2})}),At.Linear.easeNone=At.none=At.Linear.easeIn,Cb("Elastic",Eb("in"),Eb("out"),Eb()),D=7.5625,F=1/2.75,Cb("Bounce",function(t){return 1-Hl(1-t)},Hl),Cb("Expo",function(t){return t?Math.pow(2,10*(t-1)):0}),Cb("Circ",function(t){return-(j(1-t*t)-1)}),Cb("Sine",function(t){return 1===t?1:1-G(t*X)}),Cb("Back",Fb("in"),Fb("out"),Fb()),At.SteppedEase=At.steps=at.SteppedEase={config:function config(t,e){void 0===t&&(t=1);var r=1/t,i=t+(e?0:1),n=e?1:0;return function(t){return((i*gt(0,.99999999,t)|0)+n)*r}}},I.ease=At["quad.out"],_("onComplete,onUpdate,onStart,onRepeat,onReverseComplete,onInterrupt",function(t){return ct+=t+","+t+"Params,"});var zt,Et=function GSCache(t,e){this.id=V++,(t._gsap=this).target=t,this.harness=e,this.get=e?e.get:$,this.set=e?e.getSetter:Gt},It=((zt=Animation.prototype).delay=function delay(t){return t||0===t?(this.parent&&this.parent.smoothChildTiming&&this.startTime(this._start+t-this._delay),this._delay=t,this):this._delay},zt.duration=function duration(t){return arguments.length?this.totalDuration(0<this._repeat?t+(t+this._rDelay)*this._repeat:t):this.totalDuration()&&this._dur},zt.totalDuration=function totalDuration(t){if(!arguments.length)return this._tDur;this._dirty=0;var e=this._time/this._dur||0;return Fa(this,this._repeat<0?t:(t-this._repeat*this._rDelay)/(this._repeat+1)),this._tTime?ya(this,e*t+ua(this)):this},zt.totalTime=function totalTime(t,e){if(Ct(),!arguments.length)return this._tTime;var r=this._dp;if(r&&r.smoothChildTiming&&this._ts){for(ya(this,t);r.parent;)r.parent._time!==r._start+(0<=r._ts?r._tTime/r._ts:(r.totalDuration()-r._tTime)/-r._ts)&&r.totalTime(r._tTime,!0),r=r.parent;!this.parent&&this._dp.autoRemoveChildren&&(0<this._ts&&t<this._tDur||this._ts<0&&0<t||!this._tDur&&!t)&&Aa(this._dp,this,this._start-this._delay)}return(this._tTime!==t||!this._dur&&!e||this._initted&&Math.abs(this._zTime)===R||!t&&!this._initted)&&(this._ts||(this._pTime=t),ea(this,t,e)),this},zt.time=function time(t,e){return arguments.length?this.totalTime(Math.min(this.totalDuration(),t+ua(this))%this._dur||(t?this._dur:0),e):this._time},zt.totalProgress=function totalProgress(t,e){return arguments.length?this.totalTime(this.totalDuration()*t,e):this.totalDuration()?Math.min(1,this._tTime/this._tDur):this.ratio},zt.progress=function progress(t,e){return arguments.length?this.totalTime(this.duration()*(!this._yoyo||1&this.iteration()?t:1-t)+ua(this),e):this.duration()?Math.min(1,this._time/this._dur):this.ratio},zt.iteration=function iteration(t,e){var r=this.duration()+this._rDelay;return arguments.length?this.totalTime(this._time+(t-1)*r,e):this._repeat?_t(this._tTime,r)+1:1},zt.timeScale=function timeScale(t){if(!arguments.length)return this._rts===-R?0:this._rts;if(this._rts===t)return this;var e=this.parent&&this._ts?wa(this.parent._time,this):this._tTime;return this._rts=+t||0,this._ts=this._ps||t===-R?0:this._rts,function _recacheAncestors(t){for(var e=t.parent;e&&e.parent;)e._dirty=1,e.totalDuration(),e=e.parent;return t}(this.totalTime(gt(-this._delay,this._tDur,e),!0))},zt.paused=function paused(t){return arguments.length?(this._ps!==t&&((this._ps=t)?(this._pTime=this._tTime||Math.max(-this._delay,this.rawTime()),this._ts=this._act=0):(Ct(),this._ts=this._rts,this.totalTime(this.parent&&!this.parent.smoothChildTiming?this.rawTime():this._tTime||this._pTime,1===this.progress()&&(this._tTime-=R)&&Math.abs(this._zTime)!==R))),this):this._ps},zt.startTime=function startTime(t){if(arguments.length){this._start=t;var e=this.parent||this._dp;return!e||!e._sort&&this.parent||Aa(e,this,t-this._delay),this}return this._start},zt.endTime=function endTime(t){return this._start+(s(t)?this.totalDuration():this.duration())/Math.abs(this._ts)},zt.rawTime=function rawTime(t){var e=this.parent||this._dp;return e?t&&(!this._ts||this._repeat&&this._time&&this.totalProgress()<1)?this._tTime%(this._dur+this._rDelay):this._ts?wa(e.rawTime(t),this):this._tTime:this._tTime},zt.globalTime=function globalTime(t){for(var e=this,r=arguments.length?t:e.rawTime();e;)r=e._start+r/(e._ts||1),e=e._dp;return r},zt.repeat=function repeat(t){return arguments.length?(this._repeat=t,Ga(this)):this._repeat},zt.repeatDelay=function repeatDelay(t){return arguments.length?(this._rDelay=t,Ga(this)):this._rDelay},zt.yoyo=function yoyo(t){return arguments.length?(this._yoyo=t,this):this._yoyo},zt.seek=function seek(t,e){return this.totalTime(Ia(this,t),s(e))},zt.restart=function restart(t,e){return this.play().totalTime(t?-this._delay:0,s(e))},zt.play=function play(t,e){return null!=t&&this.seek(t,e),this.reversed(!1).paused(!1)},zt.reverse=function reverse(t,e){return null!=t&&this.seek(t||this.totalDuration(),e),this.reversed(!0).paused(!1)},zt.pause=function pause(t,e){return null!=t&&this.seek(t,e),this.paused(!0)},zt.resume=function resume(){return this.paused(!1)},zt.reversed=function reversed(t){return arguments.length?(!!t!==this.reversed()&&this.timeScale(-this._rts||(t?-R:0)),this):this._rts<0},zt.invalidate=function invalidate(){return this._initted=0,this._zTime=-R,this},zt.isActive=function isActive(){var t,e=this.parent||this._dp,r=this._start;return!(e&&!(this._ts&&this._initted&&e.isActive()&&(t=e.rawTime(!0))>=r&&t<this.endTime(!0)-R))},zt.eventCallback=function eventCallback(t,e,r){var i=this.vars;return 1<arguments.length?(e?(i[t]=e,r&&(i[t+"Params"]=r),"onUpdate"===t&&(this._onUpdate=e)):delete i[t],this):i[t]},zt.then=function then(t){var i=this;return new Promise(function(e){function Zm(){var t=i.then;i.then=null,o(r)&&(r=r(i))&&(r.then||r===i)&&(i.then=t),e(r),i.then=t}var r=o(t)?t:ga;i._initted&&1===i.totalProgress()&&0<=i._ts||!i._tTime&&i._ts<0?Zm():i._prom=Zm})},zt.kill=function kill(){fb(this)},Animation);function Animation(t,e){var r=t.parent||E;this.vars=t,this._delay=+t.delay||0,(this._repeat=t.repeat||0)&&(this._rDelay=t.repeatDelay||0,this._yoyo=!!t.yoyo||!!t.yoyoEase),this._ts=1,Fa(this,+t.duration,1),this.data=t.data,m||Mt.wake(),r&&Aa(r,this,e||0===e?e:r._time,1),t.reversed&&this.reverse(),t.paused&&this.paused(!0)}ha(It.prototype,{_time:0,_start:0,_end:0,_tTime:0,_tDur:0,_dirty:0,_repeat:0,_yoyo:!1,parent:null,_initted:!1,_rDelay:0,_ts:1,_dp:0,ratio:0,_zTime:-R,_prom:0,_ps:!1,_rts:1});var Rt=function(i){function Timeline(t,e){var r;return void 0===t&&(t={}),(r=i.call(this,t,e)||this).labels={},r.smoothChildTiming=!!t.smoothChildTiming,r.autoRemoveChildren=!!t.autoRemoveChildren,r._sort=s(t.sortChildren),r.parent&&za(r.parent,_assertThisInitialized(r)),t.scrollTrigger&&Ba(_assertThisInitialized(r),t.scrollTrigger),r}_inheritsLoose(Timeline,i);var t=Timeline.prototype;return t.to=function to(t,e,r,i){return new Xt(t,ca(arguments,0,this),Ia(this,p(e)?i:r)),this},t.from=function from(t,e,r,i){return new Xt(t,ca(arguments,1,this),Ia(this,p(e)?i:r)),this},t.fromTo=function fromTo(t,e,r,i,n){return new Xt(t,ca(arguments,2,this),Ia(this,p(e)?n:i)),this},t.set=function set(t,e,r){return e.duration=0,e.parent=this,ma(e).repeatDelay||(e.repeat=0),e.immediateRender=!!e.immediateRender,new Xt(t,e,Ia(this,r),1),this},t.call=function call(t,e,r){return Aa(this,Xt.delayedCall(0,t,e),Ia(this,r))},t.staggerTo=function staggerTo(t,e,r,i,n,a,s){return r.duration=e,r.stagger=r.stagger||i,r.onComplete=a,r.onCompleteParams=s,r.parent=this,new Xt(t,r,Ia(this,n)),this},t.staggerFrom=function staggerFrom(t,e,r,i,n,a,o){return r.runBackwards=1,ma(r).immediateRender=s(r.immediateRender),this.staggerTo(t,e,r,i,n,a,o)},t.staggerFromTo=function staggerFromTo(t,e,r,i,n,a,o,u){return i.startAt=r,ma(i).immediateRender=s(i.immediateRender),this.staggerTo(t,e,i,n,a,o,u)},t.render=function render(t,e,r){var i,n,a,s,o,u,h,l,f,d,c,p,_=this._time,m=this._dirty?this.totalDuration():this._tDur,g=this._dur,v=this!==E&&m-R<t&&0<=t?m:t<R?0:t,y=this._zTime<0!=t<0&&(this._initted||!g);if(v!==this._tTime||r||y){if(_!==this._time&&g&&(v+=this._time-_,t+=this._time-_),i=v,f=this._start,u=!(l=this._ts),y&&(g||(_=this._zTime),!t&&e||(this._zTime=t)),this._repeat&&(c=this._yoyo,o=g+this._rDelay,(g<(i=aa(v%o))||m===v)&&(i=g),(s=~~(v/o))&&s===v/o&&(i=g,s--),d=_t(this._tTime,o),!_&&this._tTime&&d!==s&&(d=s),c&&1&s&&(i=g-i,p=1),s!==d&&!this._lock)){var T=c&&1&d,b=T===(c&&1&s);if(s<d&&(T=!T),_=T?0:g,this._lock=1,this.render(_||(p?0:aa(s*o)),e,!g)._lock=0,!e&&this.parent&&bt(this,"onRepeat"),this.vars.repeatRefresh&&!p&&(this.invalidate()._lock=1),_!==this._time||u!=!this._ts)return this;if(b&&(this._lock=2,_=T?g+1e-4:-1e-4,this.render(_,!0),this.vars.repeatRefresh&&!p&&this.invalidate()),this._lock=0,!this._ts&&!u)return this;Ab(this,p)}if(this._hasPause&&!this._forcing&&this._lock<2&&(h=function _findNextPauseTween(t,e,r){var i;if(e<r)for(i=t._first;i&&i._start<=r;){if(!i._dur&&"isPause"===i.data&&i._start>e)return i;i=i._next}else for(i=t._last;i&&i._start>=r;){if(!i._dur&&"isPause"===i.data&&i._start<e)return i;i=i._prev}}(this,aa(_),aa(i)))&&(v-=i-(i=h._start)),this._tTime=v,this._time=i,this._act=!l,this._initted||(this._onUpdate=this.vars.onUpdate,this._initted=1,this._zTime=t),_||!i||e||bt(this,"onStart"),_<=i&&0<=t)for(n=this._first;n;){if(a=n._next,(n._act||i>=n._start)&&n._ts&&h!==n){if(n.parent!==this)return this.render(t,e,r);if(n.render(0<n._ts?(i-n._start)*n._ts:(n._dirty?n.totalDuration():n._tDur)+(i-n._start)*n._ts,e,r),i!==this._time||!this._ts&&!u){h=0,a&&(v+=this._zTime=-R);break}}n=a}else{n=this._last;for(var w=t<0?t:i;n;){if(a=n._prev,(n._act||w<=n._end)&&n._ts&&h!==n){if(n.parent!==this)return this.render(t,e,r);if(n.render(0<n._ts?(w-n._start)*n._ts:(n._dirty?n.totalDuration():n._tDur)+(w-n._start)*n._ts,e,r),i!==this._time||!this._ts&&!u){h=0,a&&(v+=this._zTime=w?-R:R);break}}n=a}}if(h&&!e&&(this.pause(),h.render(_<=i?0:-R)._zTime=_<=i?1:-1,this._ts))return this._start=f,xa(this),this.render(t,e,r);this._onUpdate&&!e&&bt(this,"onUpdate",!0),(v===m&&m>=this.totalDuration()||!v&&_)&&(f!==this._start&&Math.abs(l)===Math.abs(this._ts)||this._lock||(!t&&g||!(v===m&&0<this._ts||!v&&this._ts<0)||qa(this,1),e||t<0&&!_||!v&&!_||(bt(this,v===m?"onComplete":"onReverseComplete",!0),!this._prom||v<m&&0<this.timeScale()||this._prom())))}return this},t.add=function add(t,e){var r=this;if(p(e)||(e=Ia(this,e)),!(t instanceof It)){if(J(t))return t.forEach(function(t){return r.add(t,e)}),ra(this);if(n(t))return this.addLabel(t,e);if(!o(t))return this;t=Xt.delayedCall(0,t)}return this!==t?Aa(this,t,e):this},t.getChildren=function getChildren(t,e,r,i){void 0===t&&(t=!0),void 0===e&&(e=!0),void 0===r&&(r=!0),void 0===i&&(i=-z);for(var n=[],a=this._first;a;)a._start>=i&&(a instanceof Xt?e&&n.push(a):(r&&n.push(a),t&&n.push.apply(n,a.getChildren(!0,e,r)))),a=a._next;return n},t.getById=function getById(t){for(var e=this.getChildren(1,1,1),r=e.length;r--;)if(e[r].vars.id===t)return e[r]},t.remove=function remove(t){return n(t)?this.removeLabel(t):o(t)?this.killTweensOf(t):(pa(this,t),t===this._recent&&(this._recent=this._last),ra(this))},t.totalTime=function totalTime(t,e){return arguments.length?(this._forcing=1,!this._dp&&this._ts&&(this._start=aa(Mt.time-(0<this._ts?t/this._ts:(this.totalDuration()-t)/-this._ts))),i.prototype.totalTime.call(this,t,e),this._forcing=0,this):this._tTime},t.addLabel=function addLabel(t,e){return this.labels[t]=Ia(this,e),this},t.removeLabel=function removeLabel(t){return delete this.labels[t],this},t.addPause=function addPause(t,e,r){var i=Xt.delayedCall(0,e||O,r);return i.data="isPause",this._hasPause=1,Aa(this,i,Ia(this,t))},t.removePause=function removePause(t){var e=this._first;for(t=Ia(this,t);e;)e._start===t&&"isPause"===e.data&&qa(e),e=e._next},t.killTweensOf=function killTweensOf(t,e,r){for(var i=this.getTweensOf(t,r),n=i.length;n--;)Lt!==i[n]&&i[n].kill(t,e);return this},t.getTweensOf=function getTweensOf(t,e){for(var r,i=[],n=yt(t),a=this._first,s=p(e);a;)a instanceof Xt?ba(a._targets,n)&&(s?(!Lt||a._initted&&a._ts)&&a.globalTime(0)<=e&&a.globalTime(a.totalDuration())>e:!e||a.isActive())&&i.push(a):(r=a.getTweensOf(n,e)).length&&i.push.apply(i,r),a=a._next;return i},t.tweenTo=function tweenTo(t,e){e=e||{};var r=this,i=Ia(r,t),n=e.startAt,a=e.onStart,s=e.onStartParams,o=Xt.to(r,ha(e,{ease:"none",lazy:!1,time:i,duration:e.duration||Math.abs((i-(n&&"time"in n?n.time:r._time))/r.timeScale())||R,onStart:function onStart(){r.pause();var t=e.duration||Math.abs((i-r._time)/r.timeScale());o._dur!==t&&Fa(o,t).render(o._time,!0,!0),a&&a.apply(o,s||[])}}));return o},t.tweenFromTo=function tweenFromTo(t,e,r){return this.tweenTo(e,ha({startAt:{time:Ia(this,t)}},r))},t.recent=function recent(){return this._recent},t.nextLabel=function nextLabel(t){return void 0===t&&(t=this._time),db(this,Ia(this,t))},t.previousLabel=function previousLabel(t){return void 0===t&&(t=this._time),db(this,Ia(this,t),1)},t.currentLabel=function currentLabel(t){return arguments.length?this.seek(t,!0):this.previousLabel(this._time+R)},t.shiftChildren=function shiftChildren(t,e,r){void 0===r&&(r=0);for(var i,n=this._first,a=this.labels;n;)n._start>=r&&(n._start+=t),n=n._next;if(e)for(i in a)a[i]>=r&&(a[i]+=t);return ra(this)},t.invalidate=function invalidate(){var t=this._first;for(this._lock=0;t;)t.invalidate(),t=t._next;return i.prototype.invalidate.call(this)},t.clear=function clear(t){void 0===t&&(t=!0);for(var e,r=this._first;r;)e=r._next,this.remove(r),r=e;return this._time=this._tTime=this._pTime=0,t&&(this.labels={}),ra(this)},t.totalDuration=function totalDuration(t){var e,r,i,n,a=0,s=this,o=s._last,u=z;if(arguments.length)return s.timeScale((s._repeat<0?s.duration():s.totalDuration())/(s.reversed()?-t:t));if(s._dirty){for(n=s.parent;o;)e=o._prev,o._dirty&&o.totalDuration(),u<(i=o._start)&&s._sort&&o._ts&&!s._lock?(s._lock=1,Aa(s,o,i-o._delay,1)._lock=0):u=i,i<0&&o._ts&&(a-=i,(!n&&!s._dp||n&&n.smoothChildTiming)&&(s._start+=i/s._ts,s._time-=i,s._tTime-=i),s.shiftChildren(-i,!1,-Infinity),u=0),a<(r=xa(o))&&o._ts&&(a=r),o=e;Fa(s,s===E&&s._time>a?s._time:a,1),s._dirty=0}return s._tDur},Timeline.updateRoot=function updateRoot(t){if(E._ts&&(ea(E,wa(t,E)),d=Mt.frame),Mt.frame>=ft){ft+=U.autoSleep||120;var e=E._first;if((!e||!e._ts)&&U.autoSleep&&Mt._listeners.length<2){for(;e&&!e._ts;)e=e._next;e||Mt.sleep()}}},Timeline}(It);ha(Rt.prototype,{_lock:0,_hasPause:0,_forcing:0});function Mb(t,e,i,a,s,u){var h,l,f,d;if(ht[t]&&!1!==(h=new ht[t]).init(s,h.rawVars?e[t]:function _processVars(t,e,i,a,s){if(o(t)&&(t=Yt(t,s,e,i,a)),!r(t)||t.style&&t.nodeType||J(t))return n(t)?Yt(t,s,e,i,a):t;var u,h={};for(u in t)h[u]=Yt(t[u],s,e,i,a);return h}(e[t],a,s,u,i),i,a,u)&&(i._pt=l=new ee(i._pt,s,t,0,1,h.render,h,0,h.priority),i!==c))for(f=i._ptLookup[i._targets.indexOf(s)],d=h._props.length;d--;)f[h._props[d]]=l;return h}var Lt,Bt=function _addPropTween(t,e,r,i,a,s,u,h,l){o(i)&&(i=i(a||0,t,s));var f,d=t[e],c="get"!==r?r:o(d)?l?t[e.indexOf("set")||!o(t["get"+e.substr(3)])?e:"get"+e.substr(3)](l):t[e]():d,p=o(d)?l?jt:Vt:Zt;if(n(i)&&(~i.indexOf("random(")&&(i=ab(i)),"="===i.charAt(1)&&(i=parseFloat(c)+parseFloat(i.substr(2))*("-"===i.charAt(0)?-1:1)+(La(c)||0))),c!==i)return isNaN(c*i)?(d||e in t||L(e,i),function _addComplexStringPropTween(t,e,r,i,n,a,s){var o,u,h,l,f,d,c,p,_=new ee(this._pt,t,e,0,1,Wt,null,n),m=0,g=0;for(_.b=r,_.e=i,r+="",(c=~(i+="").indexOf("random("))&&(i=ab(i)),a&&(a(p=[r,i],t,e),r=p[0],i=p[1]),u=r.match(et)||[];o=et.exec(i);)l=o[0],f=i.substring(m,o.index),h?h=(h+1)%5:"rgba("===f.substr(-5)&&(h=1),l!==u[g++]&&(d=parseFloat(u[g-1])||0,_._pt={_next:_._pt,p:f||1===g?f:",",s:d,c:"="===l.charAt(1)?parseFloat(l.substr(2))*("-"===l.charAt(0)?-1:1):parseFloat(l)-d,m:h&&h<4?Math.round:0},m=et.lastIndex);return _.c=m<i.length?i.substring(m,i.length):"",_.fp=s,(it.test(i)||c)&&(_.e=0),this._pt=_}.call(this,t,e,c,i,p,h||U.stringFilter,l)):(f=new ee(this._pt,t,e,+c||0,i-(c||0),"boolean"==typeof d?Jt:Qt,0,p),l&&(f.fp=l),u&&f.modifier(u,this,t),this._pt=f)},qt=function _initTween(t,e){var r,i,n,a,o,u,h,l,f,d,c,p,_=t.vars,m=_.ease,g=_.startAt,v=_.immediateRender,y=_.lazy,T=_.onUpdate,b=_.onUpdateParams,w=_.callbackScope,x=_.runBackwards,k=_.yoyoEase,O=_.keyframes,M=_.autoRevert,C=t._dur,A=t._startAt,P=t._targets,S=t.parent,D=S&&"nested"===S.data?S.parent._targets:P,F="auto"===t._overwrite,z=t.timeline;if(!z||O&&m||(m="none"),t._ease=Ft(m,I.ease),t._yEase=k?Dt(Ft(!0===k?m:k,I.ease)):0,k&&t._yoyo&&!t._repeat&&(k=t._yEase,t._yEase=t._ease,t._ease=k),!z){if(p=(l=P[0]?Z(P[0]).harness:0)&&_[l.prop],r=la(_,st),A&&A.render(-1,!0).kill(),g){if(qa(t._startAt=Xt.set(P,ha({data:"isStart",overwrite:!1,parent:S,immediateRender:!0,lazy:s(y),startAt:null,delay:0,onUpdate:T,onUpdateParams:b,callbackScope:w,stagger:0},g))),v)if(0<e)M||(t._startAt=0);else if(C&&!(e<0&&A))return void(t._zTime=e)}else if(x&&C)if(A)M||(t._startAt=0);else if(e&&(v=!1),n=ha({overwrite:!1,data:"isFromStart",lazy:v&&s(y),immediateRender:v,stagger:0,parent:S},r),p&&(n[l.prop]=p),qa(t._startAt=Xt.set(P,n)),v){if(!e)return}else _initTween(t._startAt,R);for(t._pt=0,y=C&&s(y)||y&&!C,i=0;i<P.length;i++){if(h=(o=P[i])._gsap||Y(P)[i]._gsap,t._ptLookup[i]=d={},ut[h.id]&&da(),c=D===P?i:D.indexOf(o),l&&!1!==(f=new l).init(o,p||r,t,c,D)&&(t._pt=a=new ee(t._pt,o,f.name,0,1,f.render,f,0,f.priority),f._props.forEach(function(t){d[t]=a}),f.priority&&(u=1)),!l||p)for(n in r)ht[n]&&(f=Mb(n,r,t,c,o,D))?f.priority&&(u=1):d[n]=a=Bt.call(t,o,n,"get",r[n],c,D,0,_.stringFilter);t._op&&t._op[i]&&t.kill(o,t._op[i]),F&&t._pt&&(Lt=t,E.killTweensOf(o,d,t.globalTime(0)),Lt=0),t._pt&&y&&(ut[h.id]=1)}u&&te(t),t._onInit&&t._onInit(t)}t._from=!z&&!!_.runBackwards,t._onUpdate=T,t._initted=!!t.parent},Yt=function _parseFuncOrString(t,e,r,i,a){return o(t)?t.call(e,r,i,a):n(t)&&~t.indexOf("random(")?ab(t):t},Nt=ct+"repeat,repeatDelay,yoyo,repeatRefresh,yoyoEase",Ut=(Nt+",id,stagger,delay,duration,paused,scrollTrigger").split(","),Xt=function(D){function Tween(t,e,i,n){var a;"number"==typeof e&&(i.duration=e,e=i,i=null);var o,h,l,f,d,c,_,m,g=(a=D.call(this,n?e:ma(e),i)||this).vars,v=g.duration,y=g.delay,T=g.immediateRender,b=g.stagger,w=g.overwrite,x=g.keyframes,k=g.defaults,C=g.scrollTrigger,A=g.yoyoEase,P=a.parent,S=(J(t)?p(t[0]):"length"in e)?[t]:yt(t);if(a._targets=S.length?Y(S):M("GSAP target "+t+" not found. https://greensock.com",!U.nullTargetWarn)||[],a._ptLookup=[],a._overwrite=w,x||b||u(v)||u(y)){if(e=a.vars,(o=a.timeline=new Rt({data:"nested",defaults:k||{}})).kill(),o.parent=_assertThisInitialized(a),x)ha(o.vars.defaults,{ease:"none"}),x.forEach(function(t){return o.to(S,t,">")});else{if(f=S.length,_=b?Sa(b):O,r(b))for(d in b)~Nt.indexOf(d)&&((m=m||{})[d]=b[d]);for(h=0;h<f;h++){for(d in l={},e)Ut.indexOf(d)<0&&(l[d]=e[d]);l.stagger=0,A&&(l.yoyoEase=A),m&&pt(l,m),c=S[h],l.duration=+Yt(v,_assertThisInitialized(a),h,c,S),l.delay=(+Yt(y,_assertThisInitialized(a),h,c,S)||0)-a._delay,!b&&1===f&&l.delay&&(a._delay=y=l.delay,a._start+=y,l.delay=0),o.to(c,l,_(h,c,S))}o.duration()?v=y=0:a.timeline=0}v||a.duration(v=o.duration())}else a.timeline=0;return!0===w&&(Lt=_assertThisInitialized(a),E.killTweensOf(S),Lt=0),P&&za(P,_assertThisInitialized(a)),(T||!v&&!x&&a._start===aa(P._time)&&s(T)&&function _hasNoPausedAncestors(t){return!t||t._ts&&_hasNoPausedAncestors(t.parent)}(_assertThisInitialized(a))&&"nested"!==P.data)&&(a._tTime=-R,a.render(Math.max(0,-y))),C&&Ba(_assertThisInitialized(a),C),a}_inheritsLoose(Tween,D);var t=Tween.prototype;return t.render=function render(t,e,r){var i,n,a,s,o,u,h,l,f,d=this._time,c=this._tDur,p=this._dur,_=c-R<t&&0<=t?c:t<R?0:t;if(p){if(_!==this._tTime||!t||r||this._startAt&&this._zTime<0!=t<0){if(i=_,l=this.timeline,this._repeat){if(s=p+this._rDelay,(p<(i=aa(_%s))||c===_)&&(i=p),(a=~~(_/s))&&a===_/s&&(i=p,a--),(u=this._yoyo&&1&a)&&(f=this._yEase,i=p-i),o=_t(this._tTime,s),i===d&&!r&&this._initted)return this;a!==o&&(l&&this._yEase&&Ab(l,u),!this.vars.repeatRefresh||u||this._lock||(this._lock=r=1,this.render(aa(s*a),!0).invalidate()._lock=0))}if(!this._initted){if(Ca(this,t<0?t:i,r,e))return this._tTime=0,this;if(p!==this._dur)return this.render(t,e,r)}for(this._tTime=_,this._time=i,!this._act&&this._ts&&(this._act=1,this._lazy=0),this.ratio=h=(f||this._ease)(i/p),this._from&&(this.ratio=h=1-h),!i||d||e||bt(this,"onStart"),n=this._pt;n;)n.r(h,n.d),n=n._next;l&&l.render(t<0?t:!i&&u?-R:l._dur*h,e,r)||this._startAt&&(this._zTime=t),this._onUpdate&&!e&&(t<0&&this._startAt&&this._startAt.render(t,!0,r),bt(this,"onUpdate")),this._repeat&&a!==o&&this.vars.onRepeat&&!e&&this.parent&&bt(this,"onRepeat"),_!==this._tDur&&_||this._tTime!==_||(t<0&&this._startAt&&!this._onUpdate&&this._startAt.render(t,!0,!0),!t&&p||!(_===this._tDur&&0<this._ts||!_&&this._ts<0)||qa(this,1),e||t<0&&!d||!_&&!d||(bt(this,_===c?"onComplete":"onReverseComplete",!0),!this._prom||_<c&&0<this.timeScale()||this._prom()))}}else!function _renderZeroDurationTween(t,e,r,i){var n,a,s=t.ratio,o=e<0||!e&&s&&!t._start&&t._zTime>R&&!t._dp._lock||t._ts<0||t._dp._ts<0?0:1,u=t._rDelay,h=0;if(u&&t._repeat&&(h=gt(0,t._tDur,e),_t(h,u)!==(a=_t(t._tTime,u))&&(s=1-o,t.vars.repeatRefresh&&t._initted&&t.invalidate())),t._initted||!Ca(t,e,i,r))if(o!==s||i||t._zTime===R||!e&&t._zTime){for(a=t._zTime,t._zTime=e||(r?R:0),r=r||e&&!a,t.ratio=o,t._from&&(o=1-o),t._time=0,t._tTime=h,r||bt(t,"onStart"),n=t._pt;n;)n.r(o,n.d),n=n._next;t._startAt&&e<0&&t._startAt.render(e,!0,!0),t._onUpdate&&!r&&bt(t,"onUpdate"),h&&t._repeat&&!r&&t.parent&&bt(t,"onRepeat"),(e>=t._tDur||e<0)&&t.ratio===o&&(o&&qa(t,1),r||(bt(t,o?"onComplete":"onReverseComplete",!0),t._prom&&t._prom()))}else t._zTime||(t._zTime=e)}(this,t,e,r);return this},t.targets=function targets(){return this._targets},t.invalidate=function invalidate(){return this._pt=this._op=this._startAt=this._onUpdate=this._act=this._lazy=0,this._ptLookup=[],this.timeline&&this.timeline.invalidate(),D.prototype.invalidate.call(this)},t.kill=function kill(t,e){if(void 0===e&&(e="all"),!(t||e&&"all"!==e)&&(this._lazy=0,this.parent))return fb(this);if(this.timeline){var r=this.timeline.totalDuration();return this.timeline.killTweensOf(t,e,Lt&&!0!==Lt.vars.overwrite)._first||fb(this),this.parent&&r!==this.timeline.totalDuration()&&Fa(this,this._dur*this.timeline._tDur/r),this}var i,a,s,o,u,h,l,f=this._targets,d=t?yt(t):f,c=this._ptLookup,p=this._pt;if((!e||"all"===e)&&function _arraysMatch(t,e){for(var r=t.length,i=r===e.length;i&&r--&&t[r]===e[r];);return r<0}(f,d))return fb(this);for(i=this._op=this._op||[],"all"!==e&&(n(e)&&(u={},_(e,function(t){return u[t]=1}),e=u),e=function _addAliasesToVars(t,e){var r,i,n,a,s=t[0]?Z(t[0]).harness:0,o=s&&s.aliases;if(!o)return e;for(i in r=pt({},e),o)if(i in r)for(n=(a=o[i].split(",")).length;n--;)r[a[n]]=r[i];return r}(f,e)),l=f.length;l--;)if(~d.indexOf(f[l]))for(u in a=c[l],"all"===e?(i[l]=e,o=a,s={}):(s=i[l]=i[l]||{},o=e),o)(h=a&&a[u])&&("kill"in h.d&&!0!==h.d.kill(u)||pa(this,h,"_pt"),delete a[u]),"all"!==s&&(s[u]=1);return this._initted&&!this._pt&&p&&fb(this),this},Tween.to=function to(t,e,r){return new Tween(t,e,r)},Tween.from=function from(t,e){return new Tween(t,ca(arguments,1))},Tween.delayedCall=function delayedCall(t,e,r,i){return new Tween(e,0,{immediateRender:!1,lazy:!1,overwrite:!1,delay:t,onComplete:e,onReverseComplete:e,onCompleteParams:r,onReverseCompleteParams:r,callbackScope:i})},Tween.fromTo=function fromTo(t,e,r){return new Tween(t,ca(arguments,2))},Tween.set=function set(t,e){return e.duration=0,e.repeatDelay||(e.repeat=0),new Tween(t,e)},Tween.killTweensOf=function killTweensOf(t,e,r){return E.killTweensOf(t,e,r)},Tween}(It);ha(Xt.prototype,{_targets:[],_lazy:0,_startAt:0,_op:0,_onInit:0}),_("staggerTo,staggerFrom,staggerFromTo",function(r){Xt[r]=function(){var t=new Rt,e=vt.call(arguments,0);return e.splice("staggerFromTo"===r?5:4,0,0),t[r].apply(t,e)}});function Xb(t,e,r){return t.setAttribute(e,r)}function dc(t,e,r,i){i.mSet(t,e,i.m.call(i.tween,r,i.mt),i)}var Zt=function _setterPlain(t,e,r){return t[e]=r},Vt=function _setterFunc(t,e,r){return t[e](r)},jt=function _setterFuncWithParam(t,e,r,i){return t[e](i.fp,r)},Gt=function _getSetter(t,e){return o(t[e])?Vt:q(t[e])&&t.setAttribute?Xb:Zt},Qt=function _renderPlain(t,e){return e.set(e.t,e.p,Math.round(1e4*(e.s+e.c*t))/1e4,e)},Jt=function _renderBoolean(t,e){return e.set(e.t,e.p,!!(e.s+e.c*t),e)},Wt=function _renderComplexString(t,e){var r=e._pt,i="";if(!t&&e.b)i=e.b;else if(1===t&&e.e)i=e.e;else{for(;r;)i=r.p+(r.m?r.m(r.s+r.c*t):Math.round(1e4*(r.s+r.c*t))/1e4)+i,r=r._next;i+=e.c}e.set(e.t,e.p,i,e)},$t=function _renderPropTweens(t,e){for(var r=e._pt;r;)r.r(t,r.d),r=r._next},Ht=function _addPluginModifier(t,e,r,i){for(var n,a=this._pt;a;)n=a._next,a.p===i&&a.modifier(t,e,r),a=n},Kt=function _killPropTweensOf(t){for(var e,r,i=this._pt;i;)r=i._next,i.p===t&&!i.op||i.op===t?pa(this,i,"_pt"):i.dep||(e=1),i=r;return!e},te=function _sortPropTweensByPriority(t){for(var e,r,i,n,a=t._pt;a;){for(e=a._next,r=i;r&&r.pr>a.pr;)r=r._next;(a._prev=r?r._prev:n)?a._prev._next=a:i=a,(a._next=r)?r._prev=a:n=a,a=e}t._pt=i},ee=(PropTween.prototype.modifier=function modifier(t,e,r){this.mSet=this.mSet||this.set,this.set=dc,this.m=t,this.mt=r,this.tween=e},PropTween);function PropTween(t,e,r,i,n,a,s,o,u){this.t=e,this.s=i,this.c=n,this.p=r,this.r=a||Qt,this.d=s||this,this.set=o||Zt,this.pr=u||0,(this._next=t)&&(t._prev=this)}_(ct+"parent,duration,ease,delay,overwrite,runBackwards,startAt,yoyo,immediateRender,repeat,repeatDelay,data,paused,reversed,lazy,callbackScope,stringFilter,id,yoyoEase,stagger,inherit,repeatRefresh,keyframes,autoRevert,scrollTrigger",function(t){return st[t]=1}),at.TweenMax=at.TweenLite=Xt,at.TimelineLite=at.TimelineMax=Rt,E=new Rt({sortChildren:!1,defaults:I,autoRemoveChildren:!0,id:"root",smoothChildTiming:!0}),U.stringFilter=qb;var re={registerPlugin:function registerPlugin(){for(var t=arguments.length,e=new Array(t),r=0;r<t;r++)e[r]=arguments[r];e.forEach(function(t){return function _createPlugin(t){var e=(t=!t.name&&t.default||t).name,r=o(t),i=e&&!r&&t.init?function(){this._props=[]}:t,n={init:O,render:$t,add:Bt,kill:Kt,modifier:Ht,rawVars:0},a={targetTest:0,get:0,getSetter:Gt,aliases:{},register:0};if(Ct(),t!==i){if(ht[e])return;ha(i,ha(la(t,n),a)),pt(i.prototype,pt(n,la(t,a))),ht[i.prop=e]=i,t.targetTest&&(dt.push(i),st[e]=1),e=("css"===e?"CSS":e.charAt(0).toUpperCase()+e.substr(1))+"Plugin"}N(e,i),t.register&&t.register(ie,i,ee)}(t)})},timeline:function timeline(t){return new Rt(t)},getTweensOf:function getTweensOf(t,e){return E.getTweensOf(t,e)},getProperty:function getProperty(i,t,e,r){n(i)&&(i=yt(i)[0]);var a=Z(i||{}).get,s=e?ga:fa;return"native"===e&&(e=""),i?t?s((ht[t]&&ht[t].get||a)(i,t,e,r)):function(t,e,r){return s((ht[t]&&ht[t].get||a)(i,t,e,r))}:i},quickSetter:function quickSetter(r,e,i){if(1<(r=yt(r)).length){var n=r.map(function(t){return ie.quickSetter(t,e,i)}),a=n.length;return function(t){for(var e=a;e--;)n[e](t)}}r=r[0]||{};var s=ht[e],o=Z(r),u=o.harness&&(o.harness.aliases||{})[e]||e,h=s?function(t){var e=new s;c._pt=0,e.init(r,i?t+i:t,c,0,[r]),e.render(1,e),c._pt&&$t(1,c)}:o.set(r,u);return s?h:function(t){return h(r,u,i?t+i:t,o,1)}},isTweening:function isTweening(t){return 0<E.getTweensOf(t,!0).length},defaults:function defaults(t){return t&&t.ease&&(t.ease=Ft(t.ease,I.ease)),ka(I,t||{})},config:function config(t){return ka(U,t||{})},registerEffect:function registerEffect(t){var n=t.name,i=t.effect,e=t.plugins,a=t.defaults,s=t.extendTimeline;(e||"").split(",").forEach(function(t){return t&&!ht[t]&&!at[t]&&M(n+" effect requires "+t+" plugin.")}),lt[n]=function(t,e,r){return i(yt(t),ha(e||{},a),r)},s&&(Rt.prototype[n]=function(t,e,i){return this.add(lt[n](t,r(e)?e:(i=e)&&{},this),i)})},registerEase:function registerEase(t,e){At[t]=Ft(e)},parseEase:function parseEase(t,e){return arguments.length?Ft(t,e):At},getById:function getById(t){return E.getById(t)},exportRoot:function exportRoot(t,e){void 0===t&&(t={});var r,i,n=new Rt(t);for(n.smoothChildTiming=s(t.smoothChildTiming),E.remove(n),n._dp=0,n._time=n._tTime=E._time,r=E._first;r;)i=r._next,!e&&!r._dur&&r instanceof Xt&&r.vars.onComplete===r._targets[0]||Aa(n,r,r._start-r._delay),r=i;return Aa(E,n,0),n},utils:{wrap:function wrap(e,t,r){var i=t-e;return J(e)?Za(e,wrap(0,e.length),t):Ja(r,function(t){return(i+(t-e)%i)%i+e})},wrapYoyo:function wrapYoyo(e,t,r){var i=t-e,n=2*i;return J(e)?Za(e,wrapYoyo(0,e.length-1),t):Ja(r,function(t){return e+(i<(t=(n+(t-e)%n)%n||0)?n-t:t)})},distribute:Sa,random:Va,snap:Ua,normalize:function normalize(t,e,r){return Tt(t,e,0,1,r)},getUnit:La,clamp:function clamp(e,r,t){return Ja(t,function(t){return gt(e,r,t)})},splitColor:lb,toArray:yt,mapRange:Tt,pipe:function pipe(){for(var t=arguments.length,e=new Array(t),r=0;r<t;r++)e[r]=arguments[r];return function(t){return e.reduce(function(t,e){return e(t)},t)}},unitize:function unitize(e,r){return function(t){return e(parseFloat(t))+(r||La(t))}},interpolate:function interpolate(e,r,t,i){var a=isNaN(e+r)?0:function(t){return(1-t)*e+t*r};if(!a){var s,o,u,h,l,f=n(e),d={};if(!0===t&&(i=1)&&(t=null),f)e={p:e},r={p:r};else if(J(e)&&!J(r)){for(u=[],h=e.length,l=h-2,o=1;o<h;o++)u.push(interpolate(e[o-1],e[o]));h--,a=function func(t){t*=h;var e=Math.min(l,~~t);return u[e](t-e)},t=r}else i||(e=pt(J(e)?[]:{},e));if(!u){for(s in r)Bt.call(d,e,s,"get",r[s]);a=function func(t){return $t(t,d)||(f?e.p:e)}}}return Ja(t,a)},shuffle:Ra},install:K,effects:lt,ticker:Mt,updateRoot:Rt.updateRoot,plugins:ht,globalTimeline:E,core:{PropTween:ee,globals:N,Tween:Xt,Timeline:Rt,Animation:It,getCache:Z,_removeLinkedListItem:pa}};_("to,from,fromTo,delayedCall,set,killTweensOf",function(t){return re[t]=Xt[t]}),Mt.add(Rt.updateRoot),c=re.to({},{duration:0});function hc(t,e){for(var r=t._pt;r&&r.p!==e&&r.op!==e&&r.fp!==e;)r=r._next;return r}function jc(t,a){return{name:t,rawVars:1,init:function init(t,i,e){e._onInit=function(t){var e,r;if(n(i)&&(e={},_(i,function(t){return e[t]=1}),i=e),a){for(r in e={},i)e[r]=a(i[r]);i=e}!function _addModifiers(t,e){var r,i,n,a=t._targets;for(r in e)for(i=a.length;i--;)(n=(n=t._ptLookup[i][r])&&n.d)&&(n._pt&&(n=hc(n,r)),n&&n.modifier&&n.modifier(e[r],t,a[i],r))}(t,i)}}}}var ie=re.registerPlugin({name:"attr",init:function init(t,e,r,i,n){var a,s;for(a in e)(s=this.add(t,"setAttribute",(t.getAttribute(a)||0)+"",e[a],i,n,0,0,a))&&(s.op=a),this._props.push(a)}},{name:"endArray",init:function init(t,e){for(var r=e.length;r--;)this.add(t,r,t[r]||0,e[r])}},jc("roundProps",Ta),jc("modifiers"),jc("snap",Ua))||re;Xt.version=Rt.version=ie.version="3.4.0",f=1,t()&&Ct();function Uc(t,e){return e.set(e.t,e.p,Math.round(1e4*(e.s+e.c*t))/1e4+e.u,e)}function Vc(t,e){return e.set(e.t,e.p,1===t?e.e:Math.round(1e4*(e.s+e.c*t))/1e4+e.u,e)}function Wc(t,e){return e.set(e.t,e.p,t?Math.round(1e4*(e.s+e.c*t))/1e4+e.u:e.b,e)}function Xc(t,e){var r=e.s+e.c*t;e.set(e.t,e.p,~~(r+(r<0?-.5:.5))+e.u,e)}function Yc(t,e){return e.set(e.t,e.p,t?e.e:e.b,e)}function Zc(t,e){return e.set(e.t,e.p,1!==t?e.b:e.e,e)}function $c(t,e,r){return t.style[e]=r}function _c(t,e,r){return t.style.setProperty(e,r)}function ad(t,e,r){return t._gsap[e]=r}function bd(t,e,r){return t._gsap.scaleX=t._gsap.scaleY=r}function cd(t,e,r,i,n){var a=t._gsap;a.scaleX=a.scaleY=r,a.renderTransform(n,a)}function dd(t,e,r,i,n){var a=t._gsap;a[e]=r,a.renderTransform(n,a)}function hd(t,e){var r=ae.createElementNS?ae.createElementNS((e||"http://www.w3.org/1999/xhtml").replace(/^https/,"http"),t):ae.createElement(t);return r.style?r:ae.createElement(t)}function id(t,e,r){var i=getComputedStyle(t);return i[e]||i.getPropertyValue(e.replace(Ee,"-$1").toLowerCase())||i.getPropertyValue(e)||!r&&id(t,Ne(e)||e,1)||""}function ld(){(function _windowExists(){return"undefined"!=typeof window})()&&window.document&&(ne=window,ae=ne.document,se=ae.documentElement,ue=hd("div")||{style:{}},he=hd("div"),Be=Ne(Be),qe=Be+"Origin",ue.style.cssText="border-width:0;line-height:0;position:absolute;padding:0",fe=!!Ne("perspective"),oe=1)}function md(t){var e,r=hd("svg",this.ownerSVGElement&&this.ownerSVGElement.getAttribute("xmlns")||"http://www.w3.org/2000/svg"),i=this.parentNode,n=this.nextSibling,a=this.style.cssText;if(se.appendChild(r),r.appendChild(this),this.style.display="block",t)try{e=this.getBBox(),this._gsapBBox=this.getBBox,this.getBBox=md}catch(t){}else this._gsapBBox&&(e=this._gsapBBox());return i&&(n?i.insertBefore(this,n):i.appendChild(this)),se.removeChild(r),this.style.cssText=a,e}function nd(t,e){for(var r=e.length;r--;)if(t.hasAttribute(e[r]))return t.getAttribute(e[r])}function od(e){var r;try{r=e.getBBox()}catch(t){r=md.call(e,!0)}return r&&(r.width||r.height)||e.getBBox===md||(r=md.call(e,!0)),!r||r.width||r.x||r.y?r:{x:+nd(e,["x","cx","x1"])||0,y:+nd(e,["y","cy","y1"])||0,width:0,height:0}}function pd(t){return!(!t.getCTM||t.parentNode&&!t.ownerSVGElement||!od(t))}function qd(t,e){if(e){var r=t.style;e in Se&&e!==qe&&(e=Be),r.removeProperty?("ms"!==e.substr(0,2)&&"webkit"!==e.substr(0,6)||(e="-"+e),r.removeProperty(e.replace(Ee,"-$1").toLowerCase())):r.removeAttribute(e)}}function rd(t,e,r,i,n,a){var s=new ee(t._pt,e,r,0,1,a?Zc:Yc);return(t._pt=s).b=i,s.e=n,t._props.push(r),s}function td(t,e,r,i){var n,a,s,o,u=parseFloat(r)||0,h=(r+"").trim().substr((u+"").length)||"px",l=ue.style,f=Ie.test(e),d="svg"===t.tagName.toLowerCase(),c=(d?"client":"offset")+(f?"Width":"Height"),p="px"===i,_="%"===i;return i===h||!u||Ue[i]||Ue[h]?u:("px"===h||p||(u=td(t,e,r,"px")),o=t.getCTM&&pd(t),_&&(Se[e]||~e.indexOf("adius"))?aa(u/(o?t.getBBox()[f?"width":"height"]:t[c])*100):(l[f?"width":"height"]=100+(p?h:i),a=~e.indexOf("adius")||"em"===i&&t.appendChild&&!d?t:t.parentNode,o&&(a=(t.ownerSVGElement||{}).parentNode),a&&a!==ae&&a.appendChild||(a=ae.body),(s=a._gsap)&&_&&s.width&&f&&s.time===Mt.time?aa(u/s.width*100):(!_&&"%"!==h||(l.position=id(t,"position")),a===t&&(l.position="static"),a.appendChild(ue),n=ue[c],a.removeChild(ue),l.position="absolute",f&&_&&((s=Z(a)).time=Mt.time,s.width=a[c]),aa(p?n*u/100:n&&u?100/n*u:0))))}function ud(t,e,r,i){var n;return oe||ld(),e in Le&&"transform"!==e&&~(e=Le[e]).indexOf(",")&&(e=e.split(",")[0]),Se[e]&&"transform"!==e?(n=Ge(t,i),n="transformOrigin"!==e?n[e]:Qe(id(t,qe))+" "+n.zOrigin+"px"):(n=t.style[e])&&"auto"!==n&&!i&&!~(n+"").indexOf("calc(")||(n=Ze[e]&&Ze[e](t,e,r)||id(t,e)||$(t,e)||("opacity"===e?1:0)),r&&!~(n+"").indexOf(" ")?td(t,e,n,r)+r:n}function vd(t,e,r,i){if(!r||"none"===r){var n=Ne(e,t,1),a=n&&id(t,n,1);a&&a!==r?(e=n,r=a):"borderColor"===e&&(r=id(t,"borderTopColor"))}var s,o,u,h,l,f,d,c,p,_,m,g,v=new ee(this._pt,t.style,e,0,1,Wt),y=0,T=0;if(v.b=r,v.e=i,r+="","auto"===(i+="")&&(t.style[e]=i,i=id(t,e)||i,t.style[e]=r),qb(s=[r,i]),i=s[1],u=(r=s[0]).match(tt)||[],(i.match(tt)||[]).length){for(;o=tt.exec(i);)d=o[0],p=i.substring(y,o.index),l?l=(l+1)%5:"rgba("!==p.substr(-5)&&"hsla("!==p.substr(-5)||(l=1),d!==(f=u[T++]||"")&&(h=parseFloat(f)||0,m=f.substr((h+"").length),(g="="===d.charAt(1)?+(d.charAt(0)+"1"):0)&&(d=d.substr(2)),c=parseFloat(d),_=d.substr((c+"").length),y=tt.lastIndex-_.length,_||(_=_||U.units[e]||m,y===i.length&&(i+=_,v.e+=_)),m!==_&&(h=td(t,e,f,_)||0),v._pt={_next:v._pt,p:p||1===T?p:",",s:h,c:g?g*c:c-h,m:l&&l<4?Math.round:0});v.c=y<i.length?i.substring(y,i.length):""}else v.r="display"===e&&"none"===i?Zc:Yc;return it.test(i)&&(v.e=0),this._pt=v}function xd(t){var e=t.split(" "),r=e[0],i=e[1]||"50%";return"top"!==r&&"bottom"!==r&&"left"!==i&&"right"!==i||(t=r,r=i,i=t),e[0]=Xe[r]||r,e[1]=Xe[i]||i,e.join(" ")}function yd(t,e){if(e.tween&&e.tween._time===e.tween._dur){var r,i,n,a=e.t,s=a.style,o=e.u,u=a._gsap;if("all"===o||!0===o)s.cssText="",i=1;else for(n=(o=o.split(",")).length;-1<--n;)r=o[n],Se[r]&&(i=1,r="transformOrigin"===r?qe:Be),qd(a,r);i&&(qd(a,Be),u&&(u.svg&&a.removeAttribute("transform"),Ge(a,1),u.uncache=1))}}function Cd(t){return"matrix(1, 0, 0, 1, 0, 0)"===t||"none"===t||!t}function Dd(t){var e=id(t,Be);return Cd(e)?Ve:e.substr(7).match(H).map(aa)}function Ed(t,e){var r,i,n,a,s=t._gsap||Z(t),o=t.style,u=Dd(t);return s.svg&&t.getAttribute("transform")?"1,0,0,1,0,0"===(u=[(n=t.transform.baseVal.consolidate().matrix).a,n.b,n.c,n.d,n.e,n.f]).join(",")?Ve:u:(u!==Ve||t.offsetParent||t===se||s.svg||(n=o.display,o.display="block",(r=t.parentNode)&&t.offsetParent||(a=1,i=t.nextSibling,se.appendChild(t)),u=Dd(t),n?o.display=n:qd(t,"display"),a&&(i?r.insertBefore(t,i):r?r.appendChild(t):se.removeChild(t))),e&&6<u.length?[u[0],u[1],u[4],u[5],u[12],u[13]]:u)}function Fd(t,e,r,i,n,a){var s,o,u,h=t._gsap,l=n||Ed(t,!0),f=h.xOrigin||0,d=h.yOrigin||0,c=h.xOffset||0,p=h.yOffset||0,_=l[0],m=l[1],g=l[2],v=l[3],y=l[4],T=l[5],b=e.split(" "),w=parseFloat(b[0])||0,x=parseFloat(b[1])||0;r?l!==Ve&&(o=_*v-m*g)&&(u=w*(-m/o)+x*(_/o)-(_*T-m*y)/o,w=w*(v/o)+x*(-g/o)+(g*T-v*y)/o,x=u):(w=(s=od(t)).x+(~b[0].indexOf("%")?w/100*s.width:w),x=s.y+(~(b[1]||b[0]).indexOf("%")?x/100*s.height:x)),i||!1!==i&&h.smooth?(y=w-f,T=x-d,h.xOffset=c+(y*_+T*g)-y,h.yOffset=p+(y*m+T*v)-T):h.xOffset=h.yOffset=0,h.xOrigin=w,h.yOrigin=x,h.smooth=!!i,h.origin=e,h.originIsAbsolute=!!r,t.style[qe]="0px 0px",a&&(rd(a,h,"xOrigin",f,w),rd(a,h,"yOrigin",d,x),rd(a,h,"xOffset",c,h.xOffset),rd(a,h,"yOffset",p,h.yOffset)),t.setAttribute("data-svg-origin",w+" "+x)}function Id(t,e,r){var i=La(e);return aa(parseFloat(e)+parseFloat(td(t,"x",r+"px",i)))+i}function Pd(t,e,r,i,a,s){var o,u,h=360,l=n(a),f=parseFloat(a)*(l&&~a.indexOf("rad")?De:1),d=s?f*s:f-i,c=i+d+"deg";return l&&("short"===(o=a.split("_")[1])&&(d%=h)!==d%180&&(d+=d<0?h:-h),"cw"===o&&d<0?d=(d+36e9)%h-~~(d/h)*h:"ccw"===o&&0<d&&(d=(d-36e9)%h-~~(d/h)*h)),t._pt=u=new ee(t._pt,e,r,i,d,Vc),u.e=c,u.u="deg",t._props.push(r),u}function Qd(t,e,r){var i,n,a,s,o,u,h,l=he.style,f=r._gsap;for(n in l.cssText=getComputedStyle(r).cssText+";position:absolute;display:block;",l[Be]=e,ae.body.appendChild(he),i=Ge(he,1),Se)(a=f[n])!==(s=i[n])&&"perspective,force3D,transformOrigin,svgOrigin".indexOf(n)<0&&(o=La(a)!==(h=La(s))?td(r,n,a,h):parseFloat(a),u=parseFloat(s),t._pt=new ee(t._pt,f,n,o,u-o,Uc),t._pt.u=h||0,t._props.push(n));ae.body.removeChild(he)}var ne,ae,se,oe,ue,he,le,fe,de=At.Power0,ce=At.Power1,pe=At.Power2,_e=At.Power3,me=At.Power4,ge=At.Linear,ve=At.Quad,ye=At.Cubic,Te=At.Quart,be=At.Quint,we=At.Strong,xe=At.Elastic,ke=At.Back,Oe=At.SteppedEase,Me=At.Bounce,Ce=At.Sine,Ae=At.Expo,Pe=At.Circ,Se={},De=180/Math.PI,Fe=Math.PI/180,ze=Math.atan2,Ee=/([A-Z])/g,Ie=/(?:left|right|width|margin|padding|x)/i,Re=/[\s,\(]\S/,Le={autoAlpha:"opacity,visibility",scale:"scaleX,scaleY",alpha:"opacity"},Be="transform",qe=Be+"Origin",Ye="O,Moz,ms,Ms,Webkit".split(","),Ne=function _checkPropPrefix(t,e,r){var i=(e||ue).style,n=5;if(t in i&&!r)return t;for(t=t.charAt(0).toUpperCase()+t.substr(1);n--&&!(Ye[n]+t in i););return n<0?null:(3===n?"ms":0<=n?Ye[n]:"")+t},Ue={deg:1,rad:1,turn:1},Xe={top:"0%",bottom:"100%",left:"0%",right:"100%",center:"50%"},Ze={clearProps:function clearProps(t,e,r,i,n){if("isFromStart"!==n.data){var a=t._pt=new ee(t._pt,e,r,0,0,yd);return a.u=i,a.pr=-10,a.tween=n,t._props.push(r),1}}},Ve=[1,0,0,1,0,0],je={},Ge=function _parseTransform(t,e){var r=t._gsap||new Et(t);if("x"in r&&!e&&!r.uncache)return r;var i,n,a,s,o,u,h,l,f,d,c,p,_,m,g,v,y,T,b,w,x,k,O,M,C,A,P,S,D,F,z,E,I=t.style,R=r.scaleX<0,L="deg",B=id(t,qe)||"0";return i=n=a=u=h=l=f=d=c=0,s=o=1,r.svg=!(!t.getCTM||!pd(t)),m=Ed(t,r.svg),r.svg&&(M=!r.uncache&&t.getAttribute("data-svg-origin"),Fd(t,M||B,!!M||r.originIsAbsolute,!1!==r.smooth,m)),p=r.xOrigin||0,_=r.yOrigin||0,m!==Ve&&(T=m[0],b=m[1],w=m[2],x=m[3],i=k=m[4],n=O=m[5],6===m.length?(s=Math.sqrt(T*T+b*b),o=Math.sqrt(x*x+w*w),u=T||b?ze(b,T)*De:0,(f=w||x?ze(w,x)*De+u:0)&&(o*=Math.cos(f*Fe)),r.svg&&(i-=p-(p*T+_*w),n-=_-(p*b+_*x))):(E=m[6],F=m[7],P=m[8],S=m[9],D=m[10],z=m[11],i=m[12],n=m[13],a=m[14],h=(g=ze(E,D))*De,g&&(M=k*(v=Math.cos(-g))+P*(y=Math.sin(-g)),C=O*v+S*y,A=E*v+D*y,P=k*-y+P*v,S=O*-y+S*v,D=E*-y+D*v,z=F*-y+z*v,k=M,O=C,E=A),l=(g=ze(-w,D))*De,g&&(v=Math.cos(-g),z=x*(y=Math.sin(-g))+z*v,T=M=T*v-P*y,b=C=b*v-S*y,w=A=w*v-D*y),u=(g=ze(b,T))*De,g&&(M=T*(v=Math.cos(g))+b*(y=Math.sin(g)),C=k*v+O*y,b=b*v-T*y,O=O*v-k*y,T=M,k=C),h&&359.9<Math.abs(h)+Math.abs(u)&&(h=u=0,l=180-l),s=aa(Math.sqrt(T*T+b*b+w*w)),o=aa(Math.sqrt(O*O+E*E)),g=ze(k,O),f=2e-4<Math.abs(g)?g*De:0,c=z?1/(z<0?-z:z):0),r.svg&&(M=t.getAttribute("transform"),r.forceCSS=t.setAttribute("transform","")||!Cd(id(t,Be)),M&&t.setAttribute("transform",M))),90<Math.abs(f)&&Math.abs(f)<270&&(R?(s*=-1,f+=u<=0?180:-180,u+=u<=0?180:-180):(o*=-1,f+=f<=0?180:-180)),r.x=((r.xPercent=i&&Math.round(t.offsetWidth/2)===Math.round(-i)?-50:0)?0:i)+"px",r.y=((r.yPercent=n&&Math.round(t.offsetHeight/2)===Math.round(-n)?-50:0)?0:n)+"px",r.z=a+"px",r.scaleX=aa(s),r.scaleY=aa(o),r.rotation=aa(u)+L,r.rotationX=aa(h)+L,r.rotationY=aa(l)+L,r.skewX=f+L,r.skewY=d+L,r.transformPerspective=c+"px",(r.zOrigin=parseFloat(B.split(" ")[2])||0)&&(I[qe]=Qe(B)),r.xOffset=r.yOffset=0,r.force3D=U.force3D,r.renderTransform=r.svg?tr:fe?Ke:Je,r.uncache=0,r},Qe=function _firstTwoOnly(t){return(t=t.split(" "))[0]+" "+t[1]},Je=function _renderNon3DTransforms(t,e){e.z="0px",e.rotationY=e.rotationX="0deg",e.force3D=0,Ke(t,e)},We="0deg",$e="0px",He=") ",Ke=function _renderCSSTransforms(t,e){var r=e||this,i=r.xPercent,n=r.yPercent,a=r.x,s=r.y,o=r.z,u=r.rotation,h=r.rotationY,l=r.rotationX,f=r.skewX,d=r.skewY,c=r.scaleX,p=r.scaleY,_=r.transformPerspective,m=r.force3D,g=r.target,v=r.zOrigin,y="",T="auto"===m&&t&&1!==t||!0===m;if(v&&(l!==We||h!==We)){var b,w=parseFloat(h)*Fe,x=Math.sin(w),k=Math.cos(w);w=parseFloat(l)*Fe,b=Math.cos(w),a=Id(g,a,x*b*-v),s=Id(g,s,-Math.sin(w)*-v),o=Id(g,o,k*b*-v+v)}_!==$e&&(y+="perspective("+_+He),(i||n)&&(y+="translate("+i+"%, "+n+"%) "),!T&&a===$e&&s===$e&&o===$e||(y+=o!==$e||T?"translate3d("+a+", "+s+", "+o+") ":"translate("+a+", "+s+He),u!==We&&(y+="rotate("+u+He),h!==We&&(y+="rotateY("+h+He),l!==We&&(y+="rotateX("+l+He),f===We&&d===We||(y+="skew("+f+", "+d+He),1===c&&1===p||(y+="scale("+c+", "+p+He),g.style[Be]=y||"translate(0, 0)"},tr=function _renderSVGTransforms(t,e){var r,i,n,a,s,o=e||this,u=o.xPercent,h=o.yPercent,l=o.x,f=o.y,d=o.rotation,c=o.skewX,p=o.skewY,_=o.scaleX,m=o.scaleY,g=o.target,v=o.xOrigin,y=o.yOrigin,T=o.xOffset,b=o.yOffset,w=o.forceCSS,x=parseFloat(l),k=parseFloat(f);d=parseFloat(d),c=parseFloat(c),(p=parseFloat(p))&&(c+=p=parseFloat(p),d+=p),d||c?(d*=Fe,c*=Fe,r=Math.cos(d)*_,i=Math.sin(d)*_,n=Math.sin(d-c)*-m,a=Math.cos(d-c)*m,c&&(p*=Fe,s=Math.tan(c-p),n*=s=Math.sqrt(1+s*s),a*=s,p&&(s=Math.tan(p),r*=s=Math.sqrt(1+s*s),i*=s)),r=aa(r),i=aa(i),n=aa(n),a=aa(a)):(r=_,a=m,i=n=0),(x&&!~(l+"").indexOf("px")||k&&!~(f+"").indexOf("px"))&&(x=td(g,"x",l,"px"),k=td(g,"y",f,"px")),(v||y||T||b)&&(x=aa(x+v-(v*r+y*n)+T),k=aa(k+y-(v*i+y*a)+b)),(u||h)&&(s=g.getBBox(),x=aa(x+u/100*s.width),k=aa(k+h/100*s.height)),s="matrix("+r+","+i+","+n+","+a+","+x+","+k+")",g.setAttribute("transform",s),w&&(g.style[Be]=s)};_("padding,margin,Width,Radius",function(e,r){var t="Right",i="Bottom",n="Left",o=(r<3?["Top",t,i,n]:["Top"+n,"Top"+t,i+t,i+n]).map(function(t){return r<2?e+t:"border"+t+e});Ze[1<r?"border"+e:e]=function(e,t,r,i,n){var a,s;if(arguments.length<4)return a=o.map(function(t){return ud(e,t,r)}),5===(s=a.join(" ")).split(a[0]).length?a[0]:s;a=(i+"").split(" "),s={},o.forEach(function(t,e){return s[t]=a[e]=a[e]||a[(e-1)/2|0]}),e.init(t,s,n)}});var er,rr,ir,nr={name:"css",register:ld,targetTest:function targetTest(t){return t.style&&t.nodeType},init:function init(t,e,r,i,n){var a,s,o,u,h,l,f,d,c,p,_,m,g,v,y,T=this._props,b=t.style;for(f in oe||ld(),e)if("autoRound"!==f&&(s=e[f],!ht[f]||!Mb(f,e,r,i,t,n)))if(h=typeof s,l=Ze[f],"function"===h&&(h=typeof(s=s.call(r,i,t,n))),"string"===h&&~s.indexOf("random(")&&(s=ab(s)),l)l(this,t,f,s,r)&&(y=1);else if("--"===f.substr(0,2))this.add(b,"setProperty",getComputedStyle(t).getPropertyValue(f)+"",s+"",i,n,0,0,f);else{if(a=ud(t,f),u=parseFloat(a),(p="string"===h&&"="===s.charAt(1)?+(s.charAt(0)+"1"):0)&&(s=s.substr(2)),o=parseFloat(s),f in Le&&("autoAlpha"===f&&(1===u&&"hidden"===ud(t,"visibility")&&o&&(u=0),rd(this,b,"visibility",u?"inherit":"hidden",o?"inherit":"hidden",!o)),"scale"!==f&&"transform"!==f&&~(f=Le[f]).indexOf(",")&&(f=f.split(",")[0])),_=f in Se)if(m||((g=t._gsap).renderTransform||Ge(t),v=!1!==e.smoothOrigin&&g.smooth,(m=this._pt=new ee(this._pt,b,Be,0,1,g.renderTransform,g,0,-1)).dep=1),"scale"===f)this._pt=new ee(this._pt,g,"scaleY",g.scaleY,p?p*o:o-g.scaleY),T.push("scaleY",f),f+="X";else{if("transformOrigin"===f){s=xd(s),g.svg?Fd(t,s,0,v,0,this):((c=parseFloat(s.split(" ")[2])||0)!==g.zOrigin&&rd(this,g,"zOrigin",g.zOrigin,c),rd(this,b,f,Qe(a),Qe(s)));continue}if("svgOrigin"===f){Fd(t,s,1,v,0,this);continue}if(f in je){Pd(this,g,f,u,s,p);continue}if("smoothOrigin"===f){rd(this,g,"smooth",g.smooth,s);continue}if("force3D"===f){g[f]=s;continue}if("transform"===f){Qd(this,s,t);continue}}else f in b||(f=Ne(f)||f);if(_||(o||0===o)&&(u||0===u)&&!Re.test(s)&&f in b)(d=(a+"").substr((u+"").length))!==(c=(s+"").substr(((o=o||0)+"").length)||(f in U.units?U.units[f]:d))&&(u=td(t,f,a,c)),this._pt=new ee(this._pt,_?g:b,f,u,p?p*o:o-u,"px"!==c||!1===e.autoRound||_?Uc:Xc),this._pt.u=c||0,d!==c&&(this._pt.b=a,this._pt.r=Wc);else if(f in b)vd.call(this,t,f,a,s);else{if(!(f in t)){L(f,s);continue}this.add(t,f,t[f],s,i,n)}T.push(f)}y&&te(this)},get:ud,aliases:Le,getSetter:function getSetter(t,e,r){var i=Le[e];return i&&i.indexOf(",")<0&&(e=i),e in Se&&e!==qe&&(t._gsap.x||ud(t,"x"))?r&&le===r?"scale"===e?bd:ad:(le=r||{})&&("scale"===e?cd:dd):t.style&&!q(t.style[e])?$c:~e.indexOf("-")?_c:Gt(t,e)},core:{_removeProperty:qd,_getMatrix:Ed}};ie.utils.checkPrefix=Ne,ir=_((er="x,y,z,scale,scaleX,scaleY,xPercent,yPercent")+","+(rr="rotation,rotationX,rotationY,skewX,skewY")+",transform,transformOrigin,svgOrigin,force3D,smoothOrigin,transformPerspective",function(t){Se[t]=1}),_(rr,function(t){U.units[t]="deg",je[t]=1}),Le[ir[13]]=er+","+rr,_("0:translateX,1:translateY,2:translateZ,8:rotate,8:rotationZ,8:rotateZ,9:rotateX,10:rotateY",function(t){var e=t.split(":");Le[e[1]]=ir[e[0]]}),_("x,y,z,top,right,bottom,left,width,height,fontSize,padding,margin,perspective",function(t){U.units[t]="px"}),ie.registerPlugin(nr);var ar=ie.registerPlugin(nr)||ie,sr=ar.core.Tween;e.Back=ke,e.Bounce=Me,e.CSSPlugin=nr,e.Circ=Pe,e.Cubic=ye,e.Elastic=xe,e.Expo=Ae,e.Linear=ge,e.Power0=de,e.Power1=ce,e.Power2=pe,e.Power3=_e,e.Power4=me,e.Quad=ve,e.Quart=Te,e.Quint=be,e.Sine=Ce,e.SteppedEase=Oe,e.Strong=we,e.TimelineLite=Rt,e.TimelineMax=Rt,e.TweenLite=Xt,e.TweenMax=sr,e.default=ar,e.gsap=ar;if (typeof(window)==="undefined"||window!==e){Object.defineProperty(e,"__esModule",{value:!0})} else {delete e.default}});



/************** (from file ./js/other/gsap.min.js)**/ 


//
class generic_websocket {
    constructor($ws_uri, $bKEEP_ALIVE=true) {
        this.uri = $ws_uri;
        this.bKEEP_ALIVE = $bKEEP_ALIVE;
        this.ws = null;
        this.cb_data = null;
    }
    set_data_cb = function($cb) {
        this.cb_data = $cb;
    }
    connect = async()=> {
        return new Promise((resolve, reject) => {
            this.ws = new WebSocket(this.uri); 
            this.ws.onerror = function(err) {
                console.warn("on error");
                console.log("ERROR!", err );
            };
            this.ws.onopen = function() {
                resolve(true);
            };
            this.ws.onmessage = function ($evt) {
                if (this.cb_data!=null) this.cb_data($evt);
            };
            this.ws.onclose = function() {
                if (this.bKEEP_ALIVE==true) {
                    this.connect();
                }
            };
        });
    }

    sendJSON=function($json) {
        if (this.ws==null) return;
        var msg = JSON.stringify($json);
        this.ws.send(msg);
    }
}



var SUBSCRIBER = {}

SUBSCRIBER.bANNEX_ON = false;
SUBSCRIBER.bBOOSTIES_ON = false;
SUBSCRIBER.bBLOCKS_ON = false;
SUBSCRIBER.bANTOR_ON = false;

//the two sockets we use
SUBSCRIBER.bCHAINSOCKET_ON = false;
SUBSCRIBER.bALTDATASOCKET_ON = false;

SUBSCRIBER._cb_new_block = null;
SUBSCRIBER._cb_new_boosties = null;
SUBSCRIBER._cb_new_annex = null;
SUBSCRIBER._cb_new_transaction = null;
SUBSCRIBER._cb_new_mempool = null;
SUBSCRIBER._cb_antor = null;

SUBSCRIBER.chain_ws = null;
SUBSCRIBER.altdata_ws = null;

SUBSCRIBER.el_log = null;

//an assoc array of assoc array of callbacks by [type][subtpe] for alt data.
SUBSCRIBER._cb_alt_data = [];
SUBSCRIBER._cb_alt_data_cdxfilters = [];

SUBSCRIBER.el_cd = "socket_chaindata";
SUBSCRIBER.el_ad = "socket_altdata";

SUBSCRIBER._altdata_closed_cb = null;
SUBSCRIBER._chaindata_closed_cb = null;

/**
 * subscribe to websocket for new blocks. 
 */
SUBSCRIBER.subscribeToBlocks=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_block = $cb;
        if (SUBSCRIBER.bBLOCKS_ON==false) {
            SUBSCRIBER.sendJSONToServer("onblock", null);
            SUBSCRIBER.bBLOCKS_ON = true;
        }
        resolve(true);
    });
}





SUBSCRIBER.subscribeToAntor=async($cb=null, $file_nid, $elid=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_block = $cb;
        if (SUBSCRIBER.bANTOR_ON==false) {
            var params = {filter_nid:$file_nid, filter_nid_prop:"nid"};
            
            // the socket will pass back to us the el id when the antor is 
            // ready.  We can then map to the elid we want and display the goodies
            // directly from the socket return.  We can also store this in our antor file map.
            if ($elid!=null) params.elid = $elid;

            SUBSCRIBER.sendJSONToServer("onantor", params);
            SUBSCRIBER.bANTOR_ON = true;
        }
        resolve(true);
    });
}

/**
 * We are subscribing to mempool TXs that have us as the recipient.
 * @param {*} $cb 
 * @returns 
 */
SUBSCRIBER.subscribeToOurMempool=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_mempool = $cb;
        SUBSCRIBER.sendJSONToServer("onmempool", {filter_nid:ANNE.USER._userNObj.nid, filter_nid_prop:"recipient"});
        resolve(true);
    });
}

SUBSCRIBER.subscribeToOurMinedTXs=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_transaction = $cb;
        SUBSCRIBER.sendJSONToServer("ontransaction", {filter_nid:ANNE.USER._userNObj.nid, filter_nid_prop:"recipient"});
        resolve(true);
    });
}


/**
 * subscribe to websocket for annexes (mempool or mined)
 * @param {*} $cb 
 */
SUBSCRIBER.subscribeToAnnex=async($cb=null)=> {  
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_annex = $cb;
        if (SUBSCRIBER.bANNEX_ON==false) {
            SUBSCRIBER.sendJSONToServer("onannex", null);
            SUBSCRIBER.bANNEX_ON = true;  
        }
        resolve(true);
    });
}


/**
 * subscribe to websocket for boosties (mempool or mined)
 * @param {*} $cb 
 */
SUBSCRIBER.subscribeToBoosties=async($cb=null)=> {  
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }  
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_boosties = $cb;
        if (SUBSCRIBER.bBOOSTIES_ON==false) {
            SUBSCRIBER.sendJSONToServer("onboosties", null);
            SUBSCRIBER.bBOOSTIES_ON = true;
        }
        resolve(true);
    });
}


/**
 * Working for blocks, testing for annex/boosties.
 * @returns 
 */
SUBSCRIBER._initChainSocket = async()=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==true) {
        ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-on", "socket-off");
        return true;
    }
    return new Promise((resolve, reject) => {
        SUBSCRIBER.chain_ws = new WebSocket(ANNE._annodesLS.last_socket+"chaindata"); 
        SUBSCRIBER.chain_ws.onerror = function(err) {
            console.warn("on error");
            console.log("ERROR!", err );
            SUBSCRIBER.bCHAINSOCKET_ON=false;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-off", "socket-on");
        };
        SUBSCRIBER.chain_ws.onopen = function() {
            //console.log("on open");
            SUBSCRIBER.bCHAINSOCKET_ON=true;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-on", "socket-off");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Connected to chaindata service!");
            }
            resolve(true);
        };
        
        SUBSCRIBER.chain_ws.onmessage = function (evt) {
            var json = JSON.parse(evt.data);
            var logmsghtml = SUBSCRIBER.formatLogData(json);
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog(logmsghtml);
            }
            var sub = json.subscription;
            if (sub==undefined || sub==null) {
                console.warn("Must have sub on return from socket");
                return;
            } 

            //ANNE.TALK.printf_console(ANNE.TALK.BLUE, "on chain socket msg", ANNE.TALK.GREEN, sub);
            //console.log("chaindata socket", json);
            
            if (sub=="onblock") {
                if (SUBSCRIBER._cb_new_block!=null) {
                    SUBSCRIBER._cb_new_block(json);
                }
            }
            else if (sub=="onannex") {
                if (SUBSCRIBER._cb_new_annex!=null) {
                    SUBSCRIBER._cb_new_annex(json);
                }
            }
            else if (sub=="onboosties") {
                if (SUBSCRIBER._cb_new_boosties!=null) {
                    SUBSCRIBER._cb_new_boosties(json);
                }
            }
            else if (sub=="ontransaction") {
                console.log("new ontransaction");
                if (SUBSCRIBER._cb_new_transaction!=null) {
                    console.log("send to callback");
                    SUBSCRIBER._cb_new_transaction(json);
                }
            }

            //subscribes to specific mempool transactions (typically sends we receive.)
            else if (sub=="onmempool") {
                //console.log("new onmempool");
                if (SUBSCRIBER._cb_new_mempool!=null) {
                    console.log("send to callback");
                    SUBSCRIBER._cb_new_mempool(json);
                }
                else console.log("no callback set, do nothing.");
            }
        };

        SUBSCRIBER.chain_ws.onclose = function() {
            console.log("on close");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Disconnected from chaindata service!");
            }
            SUBSCRIBER.bCHAINSOCKET_ON=false;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-off", "socket-on");
            if (SUBSCRIBER._chaindata_closed_cb!=null) SUBSCRIBER._chaindata_closed_cb();
        };
        });
}




/**
 * if the altdatasocket is not on (never turned on, errored, or turend of)
 * then reconnect it before sending subscribe.
 * @param {*} $type 
 * @param {*} $subtype
 * @param {*} $extraJson 
 * @param {*} $cb 
 * @returns 
 */
SUBSCRIBER.subscribeToAltData=async($type, $subtype, $extraJson, $cb=null)=> {  
    if (SUBSCRIBER.bALTDATASOCKET_ON==false) {
        await SUBSCRIBER._initAltDataSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) {
            if (SUBSCRIBER._cb_alt_data[$type]==undefined) {
                SUBSCRIBER._cb_alt_data[$type]=[];
            }
            if ($subtype!="") SUBSCRIBER._cb_alt_data[$type][$subtype]=$cb;
            else SUBSCRIBER._cb_alt_data[$type]["*"]=$cb;
        }
        SUBSCRIBER.sendJSONToServerAlt($type, $subtype, $extraJson);
        resolve(true);
    });
}


/**
 * 
 * @returns 
 */
SUBSCRIBER._initAltDataSocket = async()=> {
    if (SUBSCRIBER.bALTDATASOCKET_ON==true) {
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-on", "socket-off");
        return true;
    }
    return new Promise((resolve, reject) => {
        SUBSCRIBER.altdata_ws = new WebSocket(ANNE._annodesLS.last_socket+"altdata");
        SUBSCRIBER.altdata_ws.onopen = function() {
            console.log("alt data socket opened");
            SUBSCRIBER.bALTDATASOCKET_ON=true;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-on", "socket-off");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Connected to altdata service!");
            }
            resolve(true);
        };
        SUBSCRIBER.finalAltDataInit();
    });
}


/**
 * 
 * @param {*} $type 
 * @param {*} $subtype 
 * @param {*} $cdxfilter 
 */
SUBSCRIBER.setAltDataCallbackManual = function($type, $subtype, $cdxfilter) {
    if (SUBSCRIBER._cb_alt_data_cdxfilters[$type]==undefined) SUBSCRIBER._cb_alt_data_cdxfilters[$type]=new Object();
    SUBSCRIBER._cb_alt_data_cdxfilters[$type][$subtype]=$cdxfilter;
}

SUBSCRIBER.removeAltDataCDXFilter = function($type,$subtype,$cb) {
    delete SUBSCRIBER._cb_alt_data_cdxfilters[$type][$subtype];
}


/**
 * 
 */
SUBSCRIBER.finalAltDataInit=function() {
    SUBSCRIBER.altdata_ws.onmessage = function (evt) {
        var json = JSON.parse(evt.data);

        var logmsghtml = SUBSCRIBER.formatLogData(json);
        if (SUBSCRIBER.el_log!=null) {
            SUBSCRIBER.appendLog(logmsghtml);
        }
        
        var adtype = json.adtype;
        if (adtype==undefined || adtype==null) {
            console.warn("Must have adtype on return from socket", json);
            return;
        } 

        var adsubtype = json.adsubtype;
        
        ANNE.TALK.printf_console(ANNE.TALK.BLUE, "on altdata socket msg", ANNE.TALK.GREEN, adtype, ANNE.TALK.YELLOW, adsubtype);
        console.log("altdata socket", json);
            
        //filter cb (one-offs)
        if (SUBSCRIBER._cb_alt_data_cdxfilters[adtype]!=undefined 
            && adsubtype!=undefined 
            && SUBSCRIBER._cb_alt_data_cdxfilters[adtype][adsubtype]!=undefined) {
            var cdxfilter = SUBSCRIBER._cb_alt_data_cdxfilters[adtype][adsubtype];
            var filter = cdxfilter.filter;
            var cb = cdxfilter.cb;
            var results = json.data.results;
            for (var r=0; r < results.length; r++) {
                var obj = results[r];
                var propkey = filter.propkey;
                var propval = filter.propvalue;
                if (obj[propkey]==undefined) {
                    break;
                }
                if (obj[propkey]!=propval) {
                    break;
                }

                //take first one.  return all, then first match.
                cb(json, obj);
                return;
            }
        }
    
        if (adsubtype!=undefined && SUBSCRIBER._cb_alt_data[adtype]!=undefined) {
            if (SUBSCRIBER._cb_alt_data[adtype][adsubtype]!=undefined) {
                SUBSCRIBER._cb_alt_data[adtype][adsubtype](json);
            }
            else if (SUBSCRIBER._cb_alt_data[adtype]["*"]!=undefined) {
                SUBSCRIBER._cb_alt_data[adtype]["*"](json);
            }
        }
        else {
            console.log("*** no finalAltDataInit calback ***");
        }
    };

    SUBSCRIBER.altdata_ws.onclose = function() {
        console.log("on close");
        if (SUBSCRIBER.el_log!=null) {
            SUBSCRIBER.appendLog("Disconnected from chaindata service!");
        }
        SUBSCRIBER.bALTDATASOCKET_ON=false;
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-off", "socket-on");
        if (SUBSCRIBER._altdata_closed_cb!=null) SUBSCRIBER._altdata_closed_cb();

    };

    SUBSCRIBER.altdata_ws.onerror = function(err) {
        console.warn("on error");
        console.log("ERROR!", err )
        SUBSCRIBER.bALTDATASOCKET_ON = false;
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-off", "socket-on");
    };
}







/**
 * 
 * @param {*} $subscribe 
 * @param {*} $json 
 */
SUBSCRIBER.sendJSONToServer=function($subscribe, $json) {
    //console.log("SendToServer subscribe:"+$subscribe);
    if ($json!=null) $json.subscribe = $subscribe;
    else $json = {subscribe:$subscribe};
    const msg = JSON.stringify($json);
    if (SUBSCRIBER.chain_ws.readyState === WebSocket.OPEN) {
        SUBSCRIBER.chain_ws.send(msg);
    } 
    else {
        SUBSCRIBER.appendLog("ERROR: WebSocket is not open. Ready state: " + SUBSCRIBER.chain_ws.readyState);
    }
}


/**
 * 
 * @param {*} $adtype 
 * @param {*} $adsubtype 
 * @param {*} $json 
 */
SUBSCRIBER.sendJSONToServerAlt=function($adtype, $adsubtype, $json) {
    console.log("sendJSONToServerAlt adtype:"+$adtype, "adsubtype:"+$adsubtype);
    if ($json!=null) {
        $json.adtype = $adtype;
        $json.adsubtype = $adsubtype;
    }
    else $json = {adtype:$adtype, adsubtype:$adsubtype};
    const msg = JSON.stringify($json);
    if (SUBSCRIBER.altdata_ws.readyState === WebSocket.OPEN) {
        SUBSCRIBER.altdata_ws.send(msg);
    } 
    else {
        SUBSCRIBER.appendLog("ERROR: WebSocket is not open. Ready state: " + SUBSCRIBER.altdata_ws.readyState);
    }
}



/**
 * 
 * @param {*} $logText 
 */
SUBSCRIBER.appendLog=function($logText) {
    if (SUBSCRIBER.el_log==null) return; 
    var log = document.getElementById(SUBSCRIBER.el_log);
    if (log==null) return;
    log.innerHTML = log.innerHTML + $logText + "<BR>>";
}


/**
 * 
 * @param {*} $d 
 * @returns 
 */
SUBSCRIBER.formatLogData = function($d) {
    const good_status = "<span class='good'>good</span>";
    const bad_status = "<span class='bad'>bad</span>";

    var html = "";
    if ($d.status=="good") html = good_status;
    else html = bad_status;
    html = html + "<span class='subscription'>"+$d.subscription +"</span>";
    if ($d.message!=undefined) html = html + "<span class='message'>"+$d.message+"</span>";
    if ($d.data!=undefined) {
        html = html + SUBSCRIBER.jsonSyntaxHighlight($d.data);
    }
    return html;
}


/**
 * 
 * @param {*} $json 
 * @returns 
 */
SUBSCRIBER.jsonSyntaxHighlight = function ($json) {
    if ($json == null) return "";
    if (typeof $json != 'string') {
        $json = JSON.stringify($json, undefined, 2);
    }
    $json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
    return $json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
        var cls = 'number';
        if (/^"/.test(match)) {
            if (/:$/.test(match)) {
                cls = 'key';
            } else {
                cls = 'string';
            }
        } else if (/true|false/.test(match)) {
            cls = 'boolean';
        } else if (/null/.test(match)) {
            cls = 'null';
        }
        return '<span class="' + cls + '">' + match + '</span>';
    });
}

/**
 * Working for blocks, testing for annex/boosties.
 * @returns 
 */
SUBSCRIBER._initGeneric = async($ws_uri, $KEEP_ALIVE=true)=> {
    var ws_generic = new generic_websocket($ws_uri, $bKEEP_ALIVE)
    ws_generic.connect();
}

/************** (from file ./js/network/subscriber.js)**/ 

var ANNODE = {};
ANNODE.annode_details = null;

ANNODE.init_annodes = async($data, $callbackNO=null)=> {
    if ($data!=null) {
        console.log("(ANNODE.init_annodes) data passed in:", $data);
    }
    else {
        console.log("ANNODE.init_annodes no data passed in");
    }

    //default new node uri to origin of page.
    ANNE.A.setHTML("a_uri", origin);

    var cb_no = $callbackNO;
    var uri_val_selected = null;
    ANNODE.annode_details = null;

    //set a_uri on select.
    const btnTest = document.getElementById("a_btnTest");
    const btnYes = document.getElementById("a_btnYes");
    const btnNo = document.getElementById("a_btnNo");
    const a_uri = document.getElementById("a_uri");
    const a_btnDelete = document.getElementById("a_btnDelete");
    const input_name = document.getElementById("a_input_frenly_name");

    //set existin uri (same as footer did, but it is hiding)
    //a_existing_uri
    ANNODE.setExisting(ANNE.API.base_uri);

    function validateURI() {
        uri_val_selected = ANNE.A.getValue(a_uri);
        ANNE.A.setHTML("a_approve_msg", "<b class='text-white'>selected new uri:</b> "+uri_val_selected);
        if (uri_val_selected.length>3) {
            ANNE.A.show(btnTest);
        }
        else {
            ANNE.A.hide(btnTest);
        }
    }

    a_uri.addEventListener("focus", e=>{validateURI();});
    a_uri.addEventListener("input", e=>{validateURI();})

    const config = {
        defaultLbl:"select annode",
        ddclass:"annodedd",
        cb:async($annode_nid, $uri)=>{
            console.log("select new annode! "+$annode_nid+" uri:"+$uri);
            uri_val_selected = $uri;
            if (uri_val_selected.length>3) {
                ANNE.A.setHTML("a_approve_msg", "<b class='text-white'>selected new uri:</b> "+uri_val_selected, true);
                var annodeObj = ANNE.getAnnodeByURI($uri);
                if (annodeObj && annodeObj.nickname!=undefined && annodeObj.nickname.length>0) {
                    ANNE.A.show("a_frenly_name_div");
                    ANNE.A.setValue("a_input_frenly_name", annodeObj.nickname);
                }
                ANNE.A.show(btnTest);
                ANNE.A.show(a_btnDelete);
            }
            else {
                ANNE.A.setHTML("a_approve_msg", "", true);
                ANNE.A.hide(btnTest);
                ANNE.A.hide(a_btnDelete);
                ANNE.A.hide("a_frenly_name_div");
                ANNE.A.setValue("a_input_frenly_name", "");
                ANNE.A.hide("a_input_frenly_name");
            }
        }
    }
    
    await ANNE.A.makeANNODEDD_first_time("a_dd_annodes", config);

    //on yes, officially set the details of the annode in the various trackers.
    btnYes.addEventListener("click", async($e)=>{

        $e.preventDefault();

        ANNE.API.setNodeConnected(true);
        
        var frenlyname = ANNE.A.getValue("a_input_frenly_name");
        if (input_name.value!="") frenlyname = input_name.value;
        
        if (input_name.value.length>0) ANNODE.annode_details.annode_name = frenlyname;                         
        await ANNE.setAnnodeByDetails(ANNODE.annode_details, uri_val_selected);
        
        ANNODE.setExisting(ANNE.API.base_uri);

        //if redirected here from a passthrough, resume the passsthrough.
        if (ANNE.MEM_UDF!=null && ANNE.MEM_UDF._lastPage!=null) {
            ANNE.MEM_UDF.handlePageDoneDefault();
        }
        else {
            ANNE.nextStepHandler();
        }
    });

    ANNE.A.hide(btnYes);

    btnNo.addEventListener("click", (e)=>{
        ANNE.TALK.printf_console(TALK.CYAN, " user canceled changing annode.");
        if (cb_no!=null) cb_no();
    });

    btnTest.addEventListener("click", async (e)=>{
        if (uri_val_selected==null) {
            alert("you must have a val selected or a custom annode typed in.");
            return;
        }
        ANNE.A.spinnerInject(btnTest);
        
        var rez = await ANNE.API.annodeDetails(uri_val_selected);//query("annodeDetails", null, uri_val_selected);
        console.log(rez);
        

        if (rez!=null && rez.annode_nid!=undefined) {    
            ANNE.A.setHTML("a_nid", rez.annode_nid);
            ANNE.A.show("a_nid_lbl");
            ANNE.A.showParent("a_nid");
        }
        else {
            ANNE.A.setHTML("a_nid", "??");
            ANNE.A.show("a_nid_lbl");
            ANNE.A.showParent("a_nid");
        }

        ANNE.A.spinnerRemove(btnTest);
        
        //if rez successful, turn on yes button.
        if (rez!=undefined && rez!=null && rez.annode_nid!=undefined) {
            ANNODE.annode_details = rez;
            ANNODE.annode_details.val_selected = uri_val_selected;
            ANNE.A.show(btnYes);
            ANNE.A.hide(btnTest);
            ANNE.A.setHTML("test_result", "status: <b class='checkmark-item'>good</b> ");
            ANNE.A.show("a_frenly_name_div");
            ANNE.A.show("a_input_frenly_name");

            if (ANNODE.annode_details.annode_name!=undefined) {
                ANNE.A.setValue("a_input_frenly_name", ANNODE.annode_details.annode_name, true);
            }
        }
        else {
            ANNE.A.hide(btnYes); 
            ANNE.A.show(btnTest);
            ANNE.A.hide("a_frenly_name_div");
            ANNE.A.hide("a_input_frenly_name");
            ANNE.A.setValue("a_input_frenly_name", "");
        }
    });

    a_btnDelete .addEventListener("click", async(e)=>{
        var eltest = document.createElement("div");
        eltest.innerHTML="!! delete this node !!";
        ANNE.MEM_UDF.doConfirmModalWithEl(eltest, async(e)=>{
            ANNE.removeAnnode(ANNODE.annode_details.annode_nid); //should have been annode_nid :(
            ANNE.A.empty("a_dd_annodes");
            await ANNE.A.makeANNODEDD_first_time("a_dd_annodes", config);
        }, null, "are you sure?");
    });


    ANNE.A.hide(a_btnDelete);
    ANNE.A.hide(btnTest); //show when valid select is made.
}

ANNODE.setExisting=function($uri) {
    ANNE.A.setHTML("a_existing_uri", $uri);
}

ANNODE.refreshEXT=function(){
    location.reload();
}

/************** (from file ./js/tops/annode.js)**/ 

var ALTDATA = {};


ALTDATA.encryptJSON = async($json, $pubkey)=>{
    if ($string==undefined ||  $string==null) {
        console.warn("fix json");
        return;
    }
    if ($pubkey==undefind || $pubkey==null) {
        console.warn("fix pubkey");
        return;
    }

    if (ANNE.USER.priv_key=="") {
        if (ANNE.USER.seedObj==null) {
            if (ANNE.USER.unlock_pw!="" && ANNE.USER.unlock_pw!=null) {
                ANNE.USER.seedObj= await ANNE.USER.getSeedWithPW(ANNE.USER.unlock_pw);
            }
            else {
                alert("we need the unlock_pw, the priv_key, or the seedObj set to continue");
                return null;
            }
        }
    }
    else {
        var eObj = await ANNE.USER.encryptMsgForPUBKEY($json.stringify(), $pubkey, ANNE.USER.priv_key);
    }
    var eObj = await ANNE.USER.encryptMsgForPUBKEY($json.stringify(), $pubkey, aO.priv_key);


    if (eObj!=null) return {data_string:eObj.data_string, nonce_string:eObj.nonce_string};
}


/**
 * returns:
 *      {.data_string .nonce_string}
 * @param {*} $string 
 * @param {*} $pubkey 
 * @returns 
 */
ALTDATA.encryptPlain = async($string, $pubkey)=>{
    if ($string==undefined ||  $string==null) {
        console.warn("fix string");
        return;
    }
    if ($pubkey==undefined || $pubkey==null) {
        console.warn("fix pubkey");
        return;
    }
    
    if (ANNE.USER.privKey==null || ANNE.USER.privKey=="") {
        if (ANNE.USER.seedObj==null) {
            if (ANNE.USER.unlock_pw!="" && ANNE.USER.unlock_pw!=null) {
                ANNE.USER.seedObj= await ANNE.USER.getSeedWithPW(ANNE.USER.unlock_pw);
            }
            else {
                alert("we need the unlock_pw, the privKey, or the seedObj set to continue");
                return null;
            }
        }
    }
    else {
        var eObj = await ANNE.USER.encryptMsgForPUBKEY("thisisomebullshit"+$string, $pubkey, ANNE.USER.privKey);
    }
    
    var aO = ANNE.USER.prepANNEAuthObjFromSeed(ANNE.USER.seedObj.seed);
    if (aO==null) {
        alert("problem with private key of anne user account.");
        return null;
    }
    var eObj = await ANNE.USER.encryptMsgForPUBKEY($string, $pubkey, aO.priv_key);
    if (eObj!=null) return {data_string:eObj.data_string, nonce_string:eObj.nonce_string};
}

/**
 * Sign the payload for any coms that should require signed payload.
 * 
 * NOTE: if you change a com to be signMsg it potentially change the order of params
 * that hit the alt-data bash handler.  
 * @param {*} $json 
 * @returns 
 */
ALTDATA.signMsg=async($json)=>{
    var seedObj = null;
    if (USER.seedObj==null) {
        if (USER.unlock_pw=="" || USER.unlock_pw==null) {
            ANNE.USER.lockApp();
            return null;
        }
        seedObj = await USER.getSeedWithPW(USER.unlock_pw);
    }
    else seedObj = USER.seedObj;

    if (seedObj==null) {
        //this.recapBad("problem extracting key from memory.  Can't encrypt message.");
        return null;
    }
    var aO = USER.prepANNEAuthObjFromSeed(seedObj.seed);
    if (aO==null) {
        //this.recapBad("problem extracting key from memory.  Can't encrypt message.");
        return null;
    }

    if (USER.privKey=="" || USER.privKey==undefined) {
        USER.privKey = aO.sign_key;
    }

    var sig=null;
    

    //if we are signning the payload, make sure the zreq_sig flag is on.
    //this will force the alt-data to verify the signature.
    //in the bash that handled various altdata subtypes, the signature flag can be 
    //looked for an ensured to be on -- thus guaranteeing the values were not falsified.
    if ($json.zreq_sig!=true) {
        $json.zreq_sig=true;
    }


    //slip in zzzzzz_newpubkey for every request now.
    if (ANNE.USER._userLS==null || ANNE.USER.isActivated()==false) {
        $json.zzzzzz_newpubkey=ANNE.USER._userLS.pubKey;
    }

    var payloadString = JSON.stringify($json);

    //gen the signature
    try {
        var aObj = await ANNE.USER.getPrivateKeyHex();
        if (aObj==null) {
            console.warn("no private key hex found");
            return null;
        }
        sig = await ANNE.USER.signMessage(payloadString, aObj.sign_key);
        return {sig:sig, payloadstring:payloadString, pubkey:ANNE.USER._userLS.pubKey};
    }

    catch(e) {
        console.log("problem");
        return null;
    }
}

/************** (from file ./js/altdata/altdata.js)**/ 


var ADF_OTC_SWAP = {};
ADF_OTC_SWAP.ico = "<img src='imgs/anne_ico_small.png' width='12px'></img>";

ADF_OTC_SWAP.ADF_PUB_KEY = "7444e77e865970388bc20009cc8f571bfe466367319b3db5435d4fabb807e37f";
ADF_OTC_SWAP.ADF_SWAP_ACTIVE=false;
ADF_OTC_SWAP.REF = null;

ADF_OTC_SWAP.max_swap_id = 0;
ADF_OTC_SWAP.last_height = 0;
ADF_OTC_SWAP.last_complete_highest_price = null;

ADF_OTC_SWAP.mose_recent_swap = null;

ADF_OTC_SWAP.total_swap_volume = 0;
ADF_OTC_SWAP.onboards = 0;
ADF_OTC_SWAP.circulating_supply_supply = 0;
ADF_OTC_SWAP.bADMIN_ON = false;
ADF_OTC_SWAP.swaps = new arraymap();

ADF_OTC_SWAP.bch_node_online = false;
ADF_OTC_SWAP.bch_node_synced = false;
ADF_OTC_SWAP.bch_blocks = 0;
ADF_OTC_SWAP.annode_online = false;
ADF_OTC_SWAP.annode_livemode = false;
ADF_OTC_SWAP.BLOCKS_TO_COMPLETE_SWAP = 12;

ADF_OTC_SWAP.last_show_value = "showothers";    //radio to choose to show others, or only your stuff
ADF_OTC_SWAP.bHIDECOMPLETE = false;             //chk to alter filter

//node that is the primary proivder of the ADF SWAP app.
ADF_OTC_SWAP.defaultTargetNodeNid="14270667173691107670";

ADF_OTC_SWAP.subtypes=["ADD_OFFER", "TAKE_OFFER", "DELETE_OFFER", "RECHECK_OFFER", "ADF_UPDATED", "GET_OFFERS"];

ADF_OTC_SWAP.bch_usdt = 0;

ADF_OTC_SWAP.filterShow=function($value) {
    ADF_OTC_SWAP.last_show_value = $value;

    //either ours, or available.  SHOW.
    //else, no show.
    console.log(ADF_OTC_SWAP.swaps.arr);

    if ($value === "mineORavailable") {
        for (var s=0; s < ADF_OTC_SWAP.swaps.arr.length; s++) {
            var sobj = ADF_OTC_SWAP.swaps.arr[s];
            var swapid = sobj.swap_id;
            var sstatus = sobj.swap_status;
            var tnidc = sobj.take_nid_to_credit;

            if (ADF_OTC_SWAP.bHIDECOMPLETE==true && sobj.swap_status=="complete") {
                ANNE.A.hide("swap_"+swapid);
                console.log("hide swap "+swapid);
                continue;
            }

            if (tnidc==undefined || tnidc==null) tnidc="";
            //ours, show
            if (tnidc!="" && tnidc==ANNE.USER._userNObj.nid) {
                ANNE.A.show("swap_"+swapid);
                continue;
            }
            //not available, hide
            if (sstatus!="available") {
                ANNE.A.hide("swap_"+swapid);
                continue;
            }
            //show.
            ANNE.A.show("swap_"+swapid);
        }    
    } 

    //simply show all.
    else if ($value === "showothers") {
        for (var s=0; s < ADF_OTC_SWAP.swaps.arr.length; s++) {
            var sobj = ADF_OTC_SWAP.swaps.arr[s];
            if (ADF_OTC_SWAP.bHIDECOMPLETE==true && sobj.swap_status=="complete") {
                ANNE.A.hide("swap_"+sobj.swap_id);;
                continue;
            }
            ANNE.A.show("swap_"+sobj.swap_id);
        } 
    }
}


/**
 * 
 * @param {*} $ticker 
 * @returns 
 */
ADF_OTC_SWAP.getBicoPrice=async($ticker)=>{
    return new Promise((resolve, reject) => {
        ANNE.API.fetchJSON("https://www.biconomy.com/api/v1/depth?symbol="+$ticker)
        .then(data => {
            var biggestBid = 0.000000;
            for (var a=0; a < data.bids.length; a++) {
                var num = +data.bids[a][0]; 
                if (num > biggestBid) biggestBid = num;
            }

            var lowestAsk = 11111111111;
            for (var a=0; a < data.asks.length; a++) {
                var num = +data.asks[a][0]; 
                if (num < lowestAsk) lowestAsk = num;
            }

            var price = new Number(((lowestAsk - biggestBid)/ 2)+lowestAsk);
            console.log(price);
            return resolve(price);
        })
        .catch(error => {
            console.error("Failed to fetch data:", error);
            return reject();
        });
    });
}


/**
 * 
 * @param {*} $alttotal 
 * @returns 
 */
ADF_OTC_SWAP.estimateBCH_USDT=function($alttotal) {
    return "approx $"+ANNE.UTILS.commas(Math.ceil($alttotal*ADF_OTC_SWAP.bch_usdt))+ADF_OTC_SWAP.getTicker("USDT");
}



/**
 * Anytime the page is pulled up in the EXT, this is called.
 * If the socket is already running, we skip socket init.
 * 
 * We will get optimized offers via 2 filters for anne_height_last_status and our  max_swap_id known.
 */
ADF_OTC_SWAP.init = async()=> {
    ANNE.A.empty("aos_offers");

    ADF_OTC_SWAP.bch_usdt = await ADF_OTC_SWAP.getBicoPrice("BCH_USDT");
 
    if (ADF_OTC_SWAP.REF==null) {
        ADF_OTC_SWAP.REF = localStorage.getItem("aon_ref");
        if (ADF_OTC_SWAP.REF==undefined || ADF_OTC_SWAP.REF==null) ADF_OTC_SWAP.REF="";
    }

    if (ADF_OTC_SWAP.REF!=null && ADF_OTC_SWAP.REF!=undefined && ADF_OTC_SWAP.REF!="") {
        console.log("using "+ADF_OTC_SWAP.REF);
        var swapimg = document.getElementById("swapimg");
        if (swapimg!=null) {
            swapimg.setAttribute("title", "ADF OTC SWAP ref "+ADF_OTC_SWAP.REF);
        }
    }

    //ADF_OTC_SWAP.altDataSoocketDisconnected
    const rad_mineandavailable = document.getElementById("rad_mineandavailable"); 
    const rad_showothers = document.getElementById("rad_showothers");
    rad_mineandavailable.addEventListener("change", (e)=>{
        ANNE.playSound("beep");
        ADF_OTC_SWAP.filterShow(rad_mineandavailable.value);
    });
    rad_showothers.addEventListener("change", (e)=>{
        ANNE.playSound("beep");
        ADF_OTC_SWAP.filterShow(rad_showothers.value);
    });

    const chk_hidecomplete = document.getElementById("chk_hidecomplete");
    chk_hidecomplete.addEventListener("change", (e)=>{
        ANNE.playSound("beep");
        ADF_OTC_SWAP.bHIDECOMPLETE = chk_hidecomplete.checked;
        ADF_OTC_SWAP.filterShow(ADF_OTC_SWAP.last_show_value);
    });

    if (ANNE.USER._userNObj.nid=="777776301542347271" 
        || ANNE.USER._userNObj.nid == ANNE.MEM.ANNE_DEV_FUND_ACCTID
        || ANNE.USER._userNObj.nid == ADF_OTC_SWAP.defaultTargetNodeNid) {
        ANNE.A.show("aos_admin_area");
        ADF_OTC_SWAP.bADMIN_ON = true;
        var el_new = document.getElementById("btn_add_new");
        el_new.addEventListener("click", async(e)=>{
            ADF_OTC_SWAP.initNewForm();
        });

        var el_toggle = document.getElementById("btn_toggle_admin");
        el_toggle.addEventListener("click", ()=>{
            if (ADF_OTC_SWAP.bADMIN_ON==true) {
                ADF_OTC_SWAP.bADMIN_ON=false;
                ANNE.A.hide("aos_admin_area");
                ANNE.A.empty("aos_offers");
                ADF_OTC_SWAP.processSwaps(ADF_OTC_SWAP.swaps.arr);
            }
            else {
                ADF_OTC_SWAP.bADMIN_ON = true;
                ANNE.A.show("aos_admin_area");
                ANNE.A.empty("aos_offers");
                ADF_OTC_SWAP.processSwaps(ADF_OTC_SWAP.swaps.arr);
            }
        })
    }
    else {
        ANNE.A.hide("aos_admin_area");
    }
    
    if (ANNE.network==TESTNET) {
        ANNE.A.show("testnetwarning");
        var elx = document.getElementById("warning_x");
        elx.addEventListener("click", (e)=>{
            ANNE.A.removeSelf("warning_x");
            ANNE.A.removeSelf("testnetwarning");
        })
    }
    else ANNE.A.hide("testnetwarning");

    ADF_OTC_SWAP.max_swap_id = 0;
    ADF_OTC_SWAP.last_height = 0;
    ADF_OTC_SWAP.last_complete_highest_price = null;
    ADF_OTC_SWAP.total_swap_volume = 0;

    var anne_usdt = await ADF_OTC_SWAP.getBicoPrice("ANNE_USDT");
    var bch_usdt = await ADF_OTC_SWAP.getBicoPrice("BCH_USDT");
    var bchperannecoin = new Number(anne_usdt/bch_usdt).toFixed(8);
    bchperannecoin = bchperannecoin+"&nbsp;bch&nbsp;/&nbsp;"+ADF_OTC_SWAP.ico; 

    
    
//******* hack for testing to redo everytime. *******
//REMOVE WHEN LIVE.

ADF_OTC_SWAP.ADF_SWAP_ACTIVE=false;


    if (ADF_OTC_SWAP.ADF_SWAP_ACTIVE==true) {
        ADF_OTC_SWAP.processSwaps(ADF_OTC_SWAP.swaps.arr);

        //don't repull this on page refresh.
        ADF_OTC_SWAP.circulating_supply_supply = await ANNE.API.getAnnodeDetail("circulating_supply");
        ADF_OTC_SWAP.annode_mode = await ANNE.API.getAnnodeDetail("annode_mode");
        ANNE.A.setHTML("circulating_supply", ANNE.UTILS.commas(ADF_OTC_SWAP.circulating_supply_supply/100));
        var el_status = ANNE.A.setHTML("annode_status", ADF_OTC_SWAP.annode_mode.replace(" ", "-"));
        el_status.classList  = ADF_OTC_SWAP.annode_mode;
        var last_anne_height = await ANNE.API.getAnnodeDetail("height");
        ANNE.A.setHTML("anne_height", last_anne_height);
        ADF_OTC_SWAP.setBCHStatus();
    }

    else {
        ADF_OTC_SWAP.circulating_supply_supply = await ANNE.API.getAnnodeDetail("circulating_supply");
        ADF_OTC_SWAP.annode_mode = await ANNE.API.getAnnodeDetail("annode_mode");
        ANNE.A.setHTML("circulating_supply", ANNE.UTILS.commas(ADF_OTC_SWAP.circulating_supply_supply/100));
        var el_status = ANNE.A.setHTML("annode_status", ADF_OTC_SWAP.annode_mode.replace(" ", "-"));
        el_status.classList  = ADF_OTC_SWAP.annode_mode;
        var last_anne_height = await ANNE.API.getAnnodeDetail("height");
        ANNE.A.setHTML("anne_height", last_anne_height);
        await ADF_OTC_SWAP.activateADFSwap();
        await ADF_OTC_SWAP.getOffers();

        //try?
        ADF_OTC_SWAP.setBCHStatus();
    }
}


/**
 * 
 * @returns 
 */
ADF_OTC_SWAP.activateADFSwap=async()=> {
    if (ADF_OTC_SWAP.ADF_SWAP_ACTIVE==true) return;
    await ANNE.SUBSCRIBER.subscribeToAltData("ADF_SWAP", "", null, ADF_OTC_SWAP.cb_swapdata);
    ANNE.SUBSCRIBER._altdata_closed_cb = ADF_OTC_SWAP.altDataSoocketDisconnected;
    await ADF_OTC_SWAP.loadTable();
    ADF_OTC_SWAP.ADF_SWAP_ACTIVE=true;
}

//keep this thing open for pete's
ADF_OTC_SWAP.altDataSoocketDisconnected=async()=>{
    ADF_OTC_SWAP.reconnectSocket();
}


/**
 * 
 * @returns 
 */
ADF_OTC_SWAP.reconnectSocket=async()=>{
    //force resubscribe.. for testing (to be safe)
    //ANNE.SUBSCRIBER.bALTDATASOCKET_ON=false;
    if (ANNE.SUBSCRIBER.bALTDATASOCKET_ON==true) return;
    
    await ANNE.SUBSCRIBER.subscribeToAltData("ADF_SWAP", "", null, ADF_OTC_SWAP.cb_swapdata);
}


ADF_OTC_SWAP.setBCHStatus=function() {
    if (ADF_OTC_SWAP.bch_node_online==true) {
        if (ADF_OTC_SWAP.bch_node_synced==true) {
            var elbchstatus=ANNE.A.setHTML("bchnode_status", "LIVE");
            if (elbchstatus!=null) elbchstatus.classList="LIVE";
        }
        else {
            var elbchstatus=ANNE.A.setHTML("bchnode_status", "SYNC");
            if (elbchstatus!=null) elbchstatus.classList="SYNC";
        }
    }
    else {
        var elbchstatus=ANNE.A.setHTML("bchnode_status", "OFFLINE");
        if (elbchstatus!=null) elbchstatus.classList="OFFLINE";
    }
}


/**
 * 
 * @param {*} $rez 
 * @param {*} $txs 
 * @returns 
 */
ADF_OTC_SWAP.recheckHandle=function($rez, $txs=[], $msg="") {

    ANNE.A.spinnerRemove("recheck_msg_"+$rez.swap_id);
    console.warn("recheckHandle", $rez);

    if ($rez.swap_status=="complete") {

        var divOffer = document.getElementById("swap_"+$rez.swap_id);
        if (divOffer!=null) {

            //kill any popups
            ANNE.MEM_UDF.popupElCancel();
            
            //repopup the proper page
            divOffer.click();

            //then redo offer (the tile)
            ANNE.A.removeSelf(divOffer);
            ADF_OTC_SWAP.gui_addSwap($rez)
        }
    }
    else if ($rez.swap_status=="expired") {
        ANNE.playSound("toolate");
        var divOffer = document.getElementById("swap_"+$rez.swap_id);

        //just nuke the tile.
        if (divOffer!=null) {
            ANNE.A.removeSelf(divOffer);
        }
    }
    
   if ($msg!="" && $msg!=undefined) {
        var el_recheck_msg = document.getElementById("recheck_msg_"+$rez.swap_id);
        if (el_recheck_msg!=null) {
            ANNE.A.showParent(el_recheck_msg);
            ANNE.A.setHTML(el_recheck_msg, $msg);
        }
    }

    if ($txs.length==0) return;
    
    var el_known = document.getElementById("knowns_"+$rez.swap_id);
    if (el_known==null) return;

    var txids = [];
    var totalpaid = 0.00;
    for (var t=0; t < $txs.length; t++) {
        var tx = $txs[t];
        var amt = tx.amount;
        console.log("amt:"+amt);
        const amtf = Number(amt);
        totalpaid+=amtf;
        var txid = tx.txid;
        txids.push(txid);
    }
    
    totalpaid = Number(totalpaid.toFixed(8));
    console.log("total:"+totalpaid);
    
    ANNE.A.show("knowns_"+$rez.swap_id);
    ANNE.A.setHTML("known_txids_"+$rez.swap_id, txids.join("<BR>"));
    var partial = ADF_OTC_SWAP.formatBitcoinishPrice($rez.swap_ticker, totalpaid);
    ANNE.A.setHTML("partial_deposit2", partial);
    ANNE.A.setHTML("take_paid_thus_far_"+$rez.swap_id, partial);

    var t2 = $rez.annies_available/$rez.annies_per_alt;
    t2 = Number(t2.toFixed(8));
    var reminingCost = t2 - totalpaid;
    ANNE.A.copyIco("copy_take_remaining_cost_"+$rez.swap_id, reminingCost);
    const t2_string = ADF_OTC_SWAP.formatBitcoinishPrice($rez.swap_ticker, reminingCost);
    ANNE.A.setHTML("take_remaining_cost_"+$rez.swap_id, t2_string);

    ANNE.A.hide("copy_take_must_cost");
}



/**
 * all swap data from sockets comes here.
 * @param {*} $json 
 * @returns 
 */
ADF_OTC_SWAP.cb_swapdata =async($json)=>{
    //console.log($json);
    
    if (subtype=="ADF_JOMOMMA") {
        alert("msg test:"+$json.adpayload.msg);
        console.warn($json);
    }

    if ($json.adtype!="ADF_SWAP") return;
    var subtype = $json.adsubtype;
    if (subtype && subtype.length>0) console.log("subtype:"+subtype);
    if (subtype==undefined) return;
    
    if (ADF_OTC_SWAP.subtypes.indexOf(subtype)==-1) return;

    //a test from bash to blast an alt-data network com one-way (response only, not request)

    var adpayload = $json.adpayload;
    if (adpayload==undefined || adpayload==null) return;

    //we have issues vetting this with encodings and java/js diffs.
    //it is vetted on the socket-sending annode.
    //for best security, anyhow, require encrypted return.
    if ($json.RETURN_SIG!=undefined) {
        
    }


    if (adpayload.bch_node_online!=undefined && adpayload.bch_node_synced!=undefined) {
        ADF_OTC_SWAP.bch_node_online=adpayload.bch_node_online;
        ADF_OTC_SWAP.bch_node_synced=adpayload.bch_node_synced;
        ADF_OTC_SWAP.bch_blocks=adpayload.blocks;
        ADF_OTC_SWAP.setBCHStatus();
        ANNE.A.setHTML("BCH_blocks", ADF_OTC_SWAP.bch_blocks);
    }
    
    var results=adpayload.results;
    if (results==undefined || results==null) return;

    //only for TAKE_OFFER returns.
    if (subtype=="TAKE_OFFER") {
        if (results.length==0) return;
        console.log("this is a TAKE OFFER return adpayload.");
        console.log(results);
        if (results!=undefined && results.length==1) {
            var rez = results[0];
            console.log(rez);
            rez.swap_id = parseInt(rez.swap_id);
            rez.take_swap_id = parseInt(rez.take_swap_id);
            var swapOBJ = ADF_OTC_SWAP.swaps.get(rez.take_swap_id);

            //old data coming thru the socket.
            if (swapOBJ!=null) {
                if (swapOBJ.swap_status=="complete" && rez.swap_status!="complete") {
                    return;
                }
                                
                //make sure this obj is updated in ram
                swapOBJ = ADF_OTC_SWAP.updateRAM(rez, rez.take_swap_id);
            }
    
            //fix the tile
            var divOffer = document.getElementById("swap_"+rez.take_swap_id);
            if (divOffer!=null) {
                ANNE.A.removeSelf(divOffer);
                ADF_OTC_SWAP.gui_addSwap(swapOBJ);
            }

            var el_take = document.getElementById("adfswaptake_"+rez.take_swap_id);
            //a take offer on a computer that doesn't have the primary swap yet?

            if (el_take!=null) {    
                if (swapOBJ!=null) {
                    var t2 = swapOBJ.annies_available/swapOBJ.annies_per_alt;
                    t2 = Number(t2.toFixed(8))
                    const t2_string = ADF_OTC_SWAP.formatBitcoinishPrice(swapOBJ.swap_ticker.toLowerCase(), t2);
                    ANNE.A.setHTML("take_must_cost_"+swapOBJ.swap_id, t2_string);

                    if (swapOBJ.take_nid_to_credit!=ANNE.USER._userNObj.nid) {
                        //sad sound, too slow joe!
                        console.log(swapOBJ.take_nid_to_credit);
                        ANNE.A.playSound("sad");
                        ANNE.A.spinnerRemove("take_swapaddr_"+swapOBJ.swap_id);
                        ANNE.A.hideParent("take_swapaddr_"+swapOBJ.swap_id);
                        ANNE.A.setHTML("anne_modal_title", "Too slow! Someone else claimed #"+swapOBJ.swap_id+"first.");
                    }

                    else {
                        ANNE.A.copyIco("copy_take_must_cost_"+swapOBJ.swap_id, t2)
                        ANNE.A.show("copy_take_swapaddr_"+swapOBJ.swap_id);
                        if (swapOBJ.take_altaddr!=null) {
                            ANNE.A.copyIco("copy_take_swapaddr_"+swapOBJ.swap_id, swapOBJ.take_altaddr);
                            ANNE.A.setHTML("take_swapaddr_"+swapOBJ.swap_id, swapOBJ.take_altaddr)
                            ANNE.A.spinnerRemove("take_swapaddr_"+swapOBJ.swap_id);
                        }
                        ANNE.A.playSound("boomhit"); 

                        ANNE.A.setHTML("anne_modal_title", "You claimed swap #"+swapOBJ.swap_id+"!");

                        await SUBSCRIBER.subscribeToOurMempool(ANNE.MEM_UDF.ws_processOurMempool);
                        await SUBSCRIBER.subscribeToOurMinedTXs(ANNE.MEM_UDF.ws_processOurMinedTXs);
                    }
                }
            }

            else {
                console.log("swaptake not on screen anymore");
            }
        }

        return await ADF_OTC_SWAP.cb_takedata(results);
    }


    else if (subtype=="DELETE_OFFER") {

    }

    //tap into any visible items and update their value.
    else if (subtype=="RECHECK_OFFER") {
        console.log(adpayload);
        console.log(results);
        var txs = adpayload.txs;
        if (results!=undefined && results.length==1) {
            var rez = results[0];
            rez.swap_id = parseInt(rez.swap_id);
            rez.take_swap_id = parseInt(rez.take_swap_id);

            var existingRAMobj = ADF_OTC_SWAP.swaps.get(rez.swap_id);
            if (existingRAMobj==null) return null;

            ADF_OTC_SWAP.updateRAM(rez, rez.swap_id);

            if (txs!=undefined && txs.length>0) rez.txs = txs;
            var msg = "";
            if (adpayload.msg!=undefined) msg = adpayload.msg;
            ADF_OTC_SWAP.recheckHandle(rez, txs, msg);
        }
    }


    var update_offers = [];
    var new_offers = [];
    var deleted_offers = [];
    
    ANNE.TALK.printf_console(TALK.YELLOW, "results", TALK.GREEN, results.length);
    
    for (var r=0; r < results.length; r++) {
        var rez = results[r];
        if (rez.swap_id==undefined) continue;
        
        //it is possible the annode returning adpayload is on a diff network?
        //perhaps this should be injected in every item?
        if (rez.network==undefined) rez.network = ANNE.network;

        if (rez.swap_id=="NULL" || rez.swap_id==null) {
            console.warn("bad data problem in rez", rez);
            continue;
        }

        //*************** clean vars...
        rez.swap_id = ANNE.UTILS.forceInt(rez.swap_id);
        rez.ispaystream = ANNE.UTILS.forceInt(rez.ispaystream);
        rez.moonths = ANNE.UTILS.forceInt(rez.moonths);
        rez.limitedofferuntil = ANNE.UTILS.forceInt(rez.limitedofferuntil);
        
        if (rez.take_id!=undefined) rez.take_id = parseInt(rez.take_id);
        if (rez.take_swap_id!=undefined) rez.take_swap_id = parseInt(rez.take_swap_id);
        //*************** 

        var bADDCOMPLETMETAS = false;

        if (rez.swap_status=="deleted" || rez.swap_status=="expired")  {
            deleted_offers.push([rez.swap_id, rez.network]);
            console.log("we will delete this offer");
            continue;
        }

        //we have a pending mempool txid now...
        else if (rez.swap_status=="complete" || rez.swap_status=="streampaying") {
            if (rez.take_nid_to_credit!=undefined && rez.take_nid_to_credit!=null && rez.take_nid_to_credit==ANNE.USER._userNObj.nid) {
                if (subtype!="GET_OFFERS") {
                    ANNE.playSound("epicsax");
                    
                    //we should get a socket for this too, if socket was on.
                    //however, it could time out and not be on?
                    ANNE.MEM_UDF.mempoolTXs.addU({txid:rez.take_txid_anne, recipient:rez.take_nid_to_credit, amount:rez.annies_available}); 
                }
            }


            //if this is an ADF_UPDATED, and if it is ours,
            //if the take is up, close it, and pop it up again w/ complete page + new info
            if (subtype=="ADF_UPDATED") {
                bADDCOMPLETMETAS=true;  //these have been completed.
                ADF_OTC_SWAP.updateRAM(rez, rez.swap_id);

                //this is ours!
                if (rez.take_nid_to_credit!=undefined && rez.take_nid_to_credit!=null && rez.take_nid_to_credit==ANNE.USER._userNObj.nid) {
                    var divOffer = document.getElementById("swap_"+rez.swap_id);
                    if (divOffer!=null) {
                        ANNE.MEM_UDF.popupElCancel();
                        divOffer.click();
                    }
                }
            }
        }


        if (ADF_OTC_SWAP.swaps.contains(rez.swap_id)==true) {
            //if this is a GET_OFFERS, and status is available, and we already have it,
            //there is nothing to update!
            if (subtype=="GET_OFFERS" && rez.swap_status=="available") continue;

            update_offers.push(rez);
            //we added ADF_UPDATED because it needed to be before it popuped up.
            ADF_OTC_SWAP.updateRAM(rez, rez.swap_id);
        }

        else {
            //complete, but our first time seeing it.  Tally the complete metas.
            if (rez.swap_status=="complete") bADDCOMPLETMETAS=true;
            rez.network = ANNE.network;
            new_offers.push(rez);
        }

        ADF_OTC_SWAP.setMetaAppVars(rez, bADDCOMPLETMETAS);
    }

    
    //insert db record and add to screen.
    if (new_offers.length>0) {
        await ANNE.DB_local.___insert(new_offers, "adf_otc_swaps");
        ADF_OTC_SWAP.processSwaps(new_offers);
    }


    //update in db, and on-screen.
    if (update_offers.length>0) {
        await ANNE.DB_local.___updateAll(update_offers, "adf_otc_swaps");
        for (var u=0; u < update_offers.length; u++) {
            ADF_OTC_SWAP.gui_editSwap(update_offers[u]);
        }
    }

    //remove db, RAM and gui.
    if (deleted_offers.length>0) {
        console.log("we have offers to delete", deleted_offers);
        await ANNE.DB_local.___deleteWithMultiKey(deleted_offers, "adf_otc_swaps");
        ANNE.A.playSound("poof");
        for (var d=0; d < deleted_offers.length; d++) {
            console.log(deleted_offers[d]);
            ANNE.TALK.printf_console(TALK.RED, "remove swap "+deleted_offers[d][0]+" from gui and db!");
            ADF_OTC_SWAP.swaps.removeU(deleted_offers[d][0]);
            var el_offer = document.getElementById("swap_"+deleted_offers[d][0]);
            ANNE.A.removeSelf(el_offer);
        }
    }

}


/**
 * 
 * @param {*} $swap 
 * @param {*} $divOffer 
 */
ADF_OTC_SWAP.nukeIt=async($swap, $divOffer=null)=>{
    if ($swap==null) return;
    await ANNE.DB_local.___deleteWithMultiKey([[$swap.swap_id, $swap.network]], "adf_otc_swaps");
    ADF_OTC_SWAP.swaps.removeU($swap.swap_id);
    ANNE.A.playSound("poof");

    if ($divOffer==null) {
        $divOffer = document.getElementById("swap_"+$swap.swap_id);
    }
    if ($divOffer==null) return;
    ANNE.A.removeSelf($divOffer);
    
}


/**
 * 
 * @param {*} $newObj
 * @param {*} $swap_id 
 */
ADF_OTC_SWAP.updateRAM=function($newObj, $swap_id=null) {
    //update props in our ram obj.  Can't just replace the obj unless 
    //replace both .arr and map in arraymap obj
    if ($swap_id==null) {
        console.warn("fix this programmer!")
        $swap_id = $newObj.swap_id;
        if ($swap_id==undefined) $swap_id = $newObj.take_swap_id;
        if ($swap_id==undefined) {
            return;
        }
    }

    var existingRAMobj = ADF_OTC_SWAP.swaps.get($swap_id);
    if (existingRAMobj==null) return null;
    
    var keys = Object.keys($newObj);
    
    for (var k=0; k < keys.length; k++) {
        var prop = keys[k];
        
        //we may have partial return data and don't want to override our good data.
        if ($newObj[ keys[k] ]==undefined) continue;
        if ($newObj[ keys[k] ]==null) continue;
        if ($newObj[ keys[k] ]=="") continue;

        existingRAMobj[prop] = $newObj[ keys[k] ];
    }
    return existingRAMobj;
}


/**
 * @param {*} $results 
 */
ADF_OTC_SWAP.cb_takedata =async($results)=>{
    var update_offers = [];
    for (var r=0; r < $results.length; r++) {
        var rez = $results[r];
        if (isNaN(rez.take_swap_id)==true) {
            console.warn("another bash problem. WTF. NaN take_swap_id.");
            continue;
        }
        if (rez.take_swap_id !=null && rez.take_swap_id !="") {
            rez.take_swap_id = parseInt(rez.take_swap_id);
        }
        else if (rez.take_swap_id==undefined) continue;
        
        if (rez.take_id != null && rez.take_id!="") rez.take_id = parseInt(rez.take_id);
        
        if (ADF_OTC_SWAP.swaps.contains(rez.take_swap_id)==true) {
            var existing = ADF_OTC_SWAP.swaps.get(rez.take_swap_id);
            update_offers.push(existing);
            ADF_OTC_SWAP.updateRAM(rez, rez.take_swap_id);
        }
        else {
            console.log("we don't have this swapid to create a take for.");
        }
    }

    if (update_offers.length>0) {
        await ANNE.DB_local.___updateAll(update_offers, "adf_otc_swaps");
        for (var u=0; u < update_offers.length; u++) {
            ADF_OTC_SWAP.gui_editSwap(update_offers[u]);
        }
    }
}


/**
 * 
 */
ADF_OTC_SWAP.loadTable=async()=>{
    var swaps = await ANNE.DB_local.loadADFSwaps(ANNE.network);
    ADF_OTC_SWAP.swaps.setMapKey("swap_id");
    if (swaps.length>0) ADF_OTC_SWAP.processSwaps(swaps);
}


ADF_OTC_SWAP.getTicker=function($ticker){
    if ($ticker=="BCH") return "<img style='filter:grayscale(1);' width='12px' src='./imgs/coin_icos/BCH.png'></img>"
    else if ($ticker=="ANNE") return "<img width='12px' src='./imgs/coin_icos/ANNE.png'></img>"
    else if ($ticker=="USDT") return "<img width='12px' src='./imgs/coin_icos/USDT.png'></img>"
}

/**
 * 
 * @param {*} $swaps 
 */
ADF_OTC_SWAP.processSwaps=function($swaps) {   
    //console.log($swaps);
    
    $swaps.sort(function($a, $b) {
        if ($a["swap_id"] < $b["swap_id"]) return -1;
        else if ($a["swap_id"] > $b["swap_id"]) return 1;
        return 0;
    })

    for (var s=0; s < $swaps.length; s++) {
        $swaps["swap_id"] = parseInt($swaps["swap_id"]);

        if ($swaps["swap_id"]==109) {
            console.log("WHERE IS THIS COMING FROM???");
            console.log(rez);
        }

        ADF_OTC_SWAP.swaps.addU($swaps[s]);
        if (parseInt($swaps[s].swap_id) > ADF_OTC_SWAP.max_swap_id) {
            ADF_OTC_SWAP.max_swap_id = parseInt($swaps[s].swap_id);
            //ANNE.A.setHTML("aos_max_swap_id", "#"+ADF_OTC_SWAP.max_swap_id+"");
        }
        ADF_OTC_SWAP.setMetaAppVars($swaps[s]);
        ADF_OTC_SWAP.gui_addSwap($swaps[s]);
    }

}



/**
 * 
 * @param {*} $swap 
 * @param {*} $bADDCOMPLETMETAS
 */
ADF_OTC_SWAP.setMetaAppVars=function($swap, $bADDCOMPLETMETAS=true) {
    $swap.anne_height_last_status = ANNE.UTILS.forceInt($swap.anne_height_last_status);
    $swap.take_last_status_change_height = ANNE.UTILS.forceInt($swap.take_last_status_change_height);

    //anne height last status stays whatever it was when the offer was made.
    //it does not update.
    if ($swap.anne_height_last_status > ADF_OTC_SWAP.last_height) {
        ADF_OTC_SWAP.last_height = $swap.anne_height_last_status;
        
        if ($swap.swap_status=="complete" || $swap.swap_status=="streampaying") {
            ADF_OTC_SWAP.mose_recent_swap = $swap;
            ADF_OTC_SWAP.last_complete_highest_price=Number(ADF_OTC_SWAP.gui_setCostPerAnnecoin("annecoin_last_price", $swap)); 
        }
    }

    //take last status change height is actually the height that changes...
    if ($swap.take_last_status_change_height!=undefined && 
        $swap.take_last_status_change_height > ADF_OTC_SWAP.last_height) {
        ADF_OTC_SWAP.last_height = $swap.take_last_status_change_height;
        ADF_OTC_SWAP.mose_recent_swap = $swap;

        if ($swap.swap_status=="complete" || $swap.swap_status=="streampaying") {
            ADF_OTC_SWAP.mose_recent_swap = $swap;
            ADF_OTC_SWAP.last_complete_highest_price=Number(ADF_OTC_SWAP.gui_setCostPerAnnecoin("annecoin_last_price", $swap)); 
        }
    }


    if ($swap.swap_status=="complete") {
        //don't double dip count this!
        if ($bADDCOMPLETMETAS==true) {
            $swap.onboarding = ANNE.UTILS.forceInt($swap.onboarding);
            if ($swap.onboarding!=undefined && $swap.onboarding>0) {
                ADF_OTC_SWAP.onboards+=1;
                ANNE.A.setHTML("aos_onboards", ADF_OTC_SWAP.onboards);
            }
            $swap.annies_available = ANNE.UTILS.forceInt($swap.annies_available);
            ADF_OTC_SWAP.total_swap_volume+=$swap.annies_available/100;
            ANNE.A.setHTML("aos_swap_volume", ANNE.UTILS.commas(ADF_OTC_SWAP.total_swap_volume));
        }
    }
}


/**
 * @param {*} $ticker
 * @param {*} $alttotal 
 * @param {*} $sep 
 * @param {*} $bDOSATS
 * @returns 
 */
ADF_OTC_SWAP.formatBitcoinishPrice=function($ticker, $alttotal, $sep="", $bDOSATS=true) {
    const satsper = 100000000;
    var sats = Math.floor($alttotal * satsper); 
    var deci = $alttotal.toFixed(8);
    var str= deci+"&nbsp;"+$ticker+$sep;
    if ($bDOSATS==true) str = str+"<span class='sats'>("+ANNE.UTILS.commas(sats)+" sats)</span>";
    return str;
}

ADF_OTC_SWAP.takeMustCost=function($swapObj, $el) {
    var t2 = $swapObj.annies_available/$swapObj.annies_per_alt;
    t2 = Number(t2.toFixed(8));
    const t2_string = ADF_OTC_SWAP.formatBitcoinishPrice($swapObj.swap_ticker.toLowerCase(), t2);
    ANNE.A.setHTML($el, t2_string);
}


/**
 * 
 * @param {*} $swap 
 */
ADF_OTC_SWAP.gui_addSwap = function($swap) {
    //console.log("addSwap:", $swap);
    
    var el_offers = document.getElementById("aos_offers");
    if (el_offers==null) {
        console.log("swaps not showing, do something to notify?");
        return;
    }

    if ($swap.swap_status=="deleted" || $swap.swap_status=="expired") return;

    var divOffer = document.createElement("div");
    divOffer.setAttribute("id", "swap_"+$swap.swap_id);
    divOffer.setAttribute("swapid", $swap.swap_id);
    divOffer.classList = "aos-offer";
    el_offers.prepend(divOffer);

    var bOURS = false;

    if ($swap.take_nid_to_credit!=undefined && $swap.take_nid_to_credit!=null && $swap.take_nid_to_credit==ANNE.USER._userNObj.nid) {
        bOURS=true;
    }

    if ($swap.swap_status=="complete") {
        if (bOURS==true) divOffer.classList.add("aos-complete-us");
        else divOffer.classList.add("aos-complete");
    }
    else if ($swap.swap_status=="streampaying") {
        if (bOURS==true) divOffer.classList.add("aos-complete-us");
        else divOffer.classList.add("aos-complete");
    } 
    else if ($swap.swap_status=="awaiting_funds" || $swap.swap_status=="partial_fund") {
        if (bOURS==true) divOffer.classList.add("aos-pending-ours");
        else divOffer.classList.add("aos-grayout");     
    }



    var divSwapid = document.createElement("div");
    divSwapid.classList = "aos-swapid";
    if ($swap.ispaystream==1) {
        divSwapid.innerHTML = "<span class='streampay'>SP:m"+$swap.moonths+"</span><span class='sid'>"+"#"+$swap.swap_id+"</span>";
    }
    else {
        divSwapid.innerHTML = "#"+$swap.swap_id;
    }
    divOffer.appendChild(divSwapid);

    var divInner = document.createElement("div");
    divInner.classList = "aos-inner";
    
    if ($swap.istreampay && $swap.limitedofferuntil>0) {
        divInner.classList = "aos-inner";
        divInner.style['margin-top:20px!important'];
    }
    else {
        divInner.classList = "aos-inner";
    }

    divOffer.appendChild(divInner);

    var divQuantity = document.createElement("div");
    divQuantity.classList = "aos-quantity";
    divQuantity.innerHTML = ANNE.UTILS.commas($swap.annies_available/100);
    var fb = document.createElement("span");
    fb.classList = "fermatta-big";
    fb.innerHTML="𝄐";
    divQuantity.appendChild(fb);
    divInner.appendChild(divQuantity);

    var t2 = $swap.annies_available/$swap.annies_per_alt;


    var divRate = document.createElement("div");
    divRate.classList = "aos-rate";
    var costPerAnnecoin = ADF_OTC_SWAP.calc_cost_per_annecoin($swap.annies_per_alt);
    divRate.innerHTML = costPerAnnecoin+"&nbsp;"+$swap.swap_ticker.toLowerCase()+"&nbsp;/&nbsp;"+ADF_OTC_SWAP.ico; 
    divInner.appendChild(divRate);


    var divRateApprox = document.createElement("div");
    divRateApprox.classList = "aos-approx-usdt";
    divRateApprox.innerHTML = ADF_OTC_SWAP.estimateBCH_USDT(t2);
    divInner.appendChild(divRateApprox);

    var divStatus = document.createElement("div");
    divStatus.classList = "aos-status "+$swap.swap_status;
    divStatus.innerHTML = $swap.swap_status;
    divInner.appendChild(divStatus);

    
    t2 = Number(t2.toFixed(8));

    if ($swap.ispaystream==true) {
        var t2_string = ADF_OTC_SWAP.formatBitcoinishPrice($swap.swap_ticker.toLowerCase(), t2, "", false);
    }
    else var t2_string = ADF_OTC_SWAP.formatBitcoinishPrice($swap.swap_ticker.toLowerCase(), t2);
        
    if ($swap.swap_status=="available") {
        var divSwapnow = document.createElement("div");
        divSwapnow.classList = "swapnow";
        divSwapnow.innerHTML = "swap "+t2_string;
        divInner.appendChild(divSwapnow);
    }


    if (($swap.swap_status=="complete" || $swap.swap_status=="streampaying")) {
        if ($swap.take_nid_to_credit!=undefined) {
            var divSwapper = document.createElement("div");
            divSwapper.classList = "aos-swapper";
            var userNOBJ = ANNE.MEMG.getOrMakeIfNew($swap.take_nid_to_credit);
            divSwapper.innerHTML = userNOBJ.bestNameHTML();
            divInner.appendChild(divSwapper);
            divSwapper.addEventListener("click", ($e)=>{
                $e.stopPropagation();
                $e.preventDefault();
                $e.stopImmediatePropagation();
                var userNOBJ = ANNE.MEMG.getOrMakeIfNew($swap.take_nid_to_credit);
                ANNE.MEM_UDF.makeNewIdentityView(userNOBJ);
            });
            
        }
    }

    //COVERS THE LIMITEDOFFERUNTIL use case.
    if ($swap.limitedofferuntil>0 && ($swap.swap_status=="available" 
                || $swap.swap_status=="awaiting_funds" 
                || $swap.swap_status=="partial_fund")) {

        var madeAt = $swap.anne_height_last_status;
        var currentH = ANNE.HISTORY.getGLKB();
        var expires = $swap.limitedofferuntil+madeAt;
        
        //console.log("currentH:"+currentH+' expires:'+expires+" limitedoffuntil:"+$swap.limitedofferuntil+" madeAt:"+madeAt);
        var limitedofferuntil = document.createElement("div");
        var diff = expires-currentH;
        var time = ANNE.UTILS.estimateTimeByBlocksLeft(diff);
        
        //console.log("diff:"+diff);
        
        if (diff<=0) {
            limitedofferuntil.classList = "time-left-red";
            limitedofferuntil.innerHTML =  "EXPIRED. Don't send funds.";
            divOffer.classList.add("aos-expired-override");
            ADF_OTC_SWAP.nukeIt($swap, divOffer);
            return;
        }
        else if (diff<3) {
            limitedofferuntil.classList = "time-left-red";
            limitedofferuntil.innerHTML =  "limited time offer!<br><span class='time'>"+time+"</span>";
            divOffer.classList.add("aos-limitedoffer-red");
        }
        else if (diff < 7) {
            limitedofferuntil.classList = "time-left-yellow";
            limitedofferuntil.innerHTML = "limited time offer!<br><span class='time'>"+time+"</span>";
            divOffer.classList.add("aos-limitedoffer-yellow");
        }
        else {
            limitedofferuntil.classList = "time-left-green";
            limitedofferuntil.innerHTML = "limited time offer!<br><span class='time'>"+time+"</span>";
            divOffer.classList.add("aos-limitedoffer-green");
        }
        divInner.appendChild(limitedofferuntil);
    }


    //THIS COVERS THE BASIC TIME LIMITED USE CASE.
    //user must act within time line to keep the take.
    else if ($swap.swap_status=="awaiting_funds" || $swap.swap_status=="partial_fund") {
        var madeAt = ANNE.UTILS.forceInt($swap.anne_height_last_status);
        var currentH = ANNE.HISTORY.getGLKB();
        var expires = ADF_OTC_SWAP.BLOCKS_TO_COMPLETE_SWAP+madeAt;
        
        var limitedofferuntil = document.createElement("div");
        var diff = expires-currentH;

        var time = ANNE.UTILS.estimateTimeByBlocksLeft(diff);
        
        console.log("swapid:"+$swap.swap_id+" madeAt:"+madeAt+" expires:"+expires+" current:"+currentH+" diff:"+diff + "time:"+time);
        if (diff<=0) {
            limitedofferuntil.classList = "time-left-red";
            limitedofferuntil.innerHTML =  "<i class='bi bi-exclamation-triangle-fill'></i>&nbsp;EXPIRED. Don't send funds.";
            divOffer.classList.add("aos-expired-override");
            ADF_OTC_SWAP.nukeIt($swap, divOffer);
            return;
        }
        else if (diff<3) {
            limitedofferuntil.classList = "time-left-red";
            limitedofferuntil.innerHTML =  "<i class='bi-alarm'></i>&nbsp;must pay by<br>("+time+")";
        }
        else if (diff < 7) {
            limitedofferuntil.classList = "time-left-yellow";
            limitedofferuntil.innerHTML = "<i class='bi-alarm'></i>&nbsp;must pay by<br>("+time+")";
        }
        else {
            limitedofferuntil.classList = "time-left-green";
            limitedofferuntil.innerHTML = "<i class='bi-alarm'></i>&nbsp;must pay by<br>("+time+")";
        }
        divInner.appendChild(limitedofferuntil);
    }


    if (ADF_OTC_SWAP.bADMIN_ON ==true && $swap.swap_status=="available") {
        var divCancel = document.createElement("div");
        divCancel.classList = "aos-admin-cancel";
        divCancel.innerHTML = "[x]";
        divInner.appendChild(divCancel);
        divCancel.addEventListener("click", ($e)=>{
            $e.preventDefault();
            $e.stopPropagation();
            $e.stopImmediatePropagation();

            //do a popup with a confirm to delete.
            console.log("cancel swap."); 
            ANNE.A.showParent("btn_anne_modal_confirm");

            ANNE.MEM_UDF.doConfirmModalWithHTML("Are you really sure you want to delete this SWAP, dear admin?",
                "This will nuke the swap record", 
                ()=>{
                    var id = ANNE.A.getFirstParentAttr($e.target, "aos-offer", "swapid");
                    var swapObj = ADF_OTC_SWAP.swaps.get(id);
                    if (swapObj==null) return;
                    ADF_OTC_SWAP.deleteOffer(swapObj.swap_target_nid, swapObj.swap_id);
                }, 
                ()=>{
                }, 
                "CONFIRM", 
                "<img style='display:inline-block' src='../imgs/swap.png' width='32px'></img>", 
                false);

        });
    }


    divOffer.addEventListener("click", async($e)=>{
        var id = ANNE.A.getFirstParentAttr($e.target, "aos-offer", "swapid");
        var swapObj = ADF_OTC_SWAP.swaps.get(id);
        if (swapObj==null) return;
        
        if ((swapObj.swap_status=="awaiting_funds" || swapObj.swap_status=="partial_fund") && bOURS==true) {
            var page = "adfswaptaken";
            if (swapObj.swap_status=="partial_fund") page = "adfswaptaken_partial";

            await ANNE.MEM_UDF._popupWithLoad(page, "Claimed ANNE ADF OTC Swap Offer #"+swapObj.swap_id, 
                ()=>{
                    //console.log("popup confirmed");
                }, 
                ()=>{
                    //console.log("popup cancelled");
                },
                ()=>{
                    
                    //console.log("popup loaded");
                    if (swapObj.ispaystream==1) {
                        ANNE.A.setHTML("swaptaken_sp_payment", "~"+ANNE.UTILS.commas(Math.ceil(swapObj.annies_available/swapObj.moonths)/100));
                        ANNE.A.setHTML("swaptaken_payments", swapObj.moonths+" moonthly");
                        ANNE.A.hide("swaptake_nostreampay");
                        ANNE.A.show("swaptake_streampay");
                    }
                    else {
                        ANNE.A.show("swaptake_nostreampay");
                        ANNE.A.hide("swaptake_streampay");
                    }
                    
                    var el_swapdiv = document.getElementById("swapdiv");
                    el_swapdiv.id = "swapdiv_"+swapObj.swap_id;
                    var el_take_must_cost = document.getElementById("take_must_cost");
                    el_take_must_cost.id = "take_must_cost_"+swapObj.swap_id;
                    var el_take_status = document.getElementById("take_status");
                    el_take_status.id = "take_status_"+swapObj.swap_id;
                    var el_copy_take_must_cost = document.getElementById("copy_take_must_cost");
                    el_copy_take_must_cost.id = "copy_take_must_cost_"+swapObj.swap_id;
                    var el_copy_take_swapaddr = document.getElementById("copy_take_swapaddr");
                    el_copy_take_swapaddr.id="copy_take_swapaddr_"+swapObj.swap_id;
                    var el_take_swapaddr = document.getElementById("take_swapaddr");
                    el_take_swapaddr.id="take_swapaddr_"+swapObj.swap_id;

                    var el_take_time_left = document.getElementById("take_time_left");
                    el_take_time_left.id="take_time_left_"+swapObj.swap_id;

                    var el_recheck_msg = document.getElementById("recheck_msg");
                    el_recheck_msg.id="recheck_msg_"+swapObj.swap_id;
                    ANNE.A.hideParent(el_recheck_msg);

                    if (page=="adfswaptaken_partial") {
                        var el_knowns = document.getElementById("knowns_div");
                        el_knowns.id = "knowns_"+swapObj.swap_id;
                        var el_known_txids = document.getElementById("known_txids");
                        el_known_txids.id = "known_txids_"+swapObj.swap_id;
                        var el_take_paid_thus_far = document.getElementById("take_paid_thus_far");
                        el_take_paid_thus_far.id = "take_paid_thus_far_"+swapObj.swap_id;
                        var el_take_remaining_cost = document.getElementById("take_remaining_cost");
                        el_take_remaining_cost.id = "take_remaining_cost_"+swapObj.swap_id;
                        var el_copy_take_remaining_cost = document.getElementById("copy_take_remaining_cost");
                        el_copy_take_remaining_cost.id = "copy_take_remaining_cost_"+swapObj.swap_id;
                    }
                        
                    
                    var t2 = swapObj.annies_available/swapObj.annies_per_alt;
                    t2 = Number(t2.toFixed(8));
                    const t2_string = ADF_OTC_SWAP.formatBitcoinishPrice(swapObj.swap_ticker.toLowerCase(), t2);
                    ANNE.A.setHTML("take_must_cost_"+swapObj.swap_id, t2_string);
                    ANNE.A.copyIco("copy_take_must_cost_"+swapObj.swap_id, t2);

                    ANNE.A.setHTML("take_status_"+swapObj.swap_id,swapObj.swap_status);  
                    ANNE.A.addClass("take_status_"+swapObj.swap_id,swapObj.swap_status);

                    ANNE.A.setHTML("pop_swap_id", "#"+swapObj.swap_id);
                    ANNE.A.setHTML("take_swapid", "#"+swapObj.swap_id);
                    if (swapObj.ispaystream==1) {
                        ANNE.A.setHTML("take_swapid","<span class='streampay'>SP:m"+swapObj.moonths+"</span><span class='sid'>"+"#"+swapObj.swap_id+"</span>");
                    }
                    else {
                        ANNE.A.setHTML("take_swapid","#"+swapObj.swap_id);
                    }

                    if (swapObj.take_altaddr==undefined || swapObj.take_altaddr=="" || swapObj.take_altaddr==null) {
                        ANNE.A.hide("copy_take_swapaddr_"+swapObj.swap_id);
                    }
                    else {
                        ANNE.A.show("copy_take_swapaddr_"+swapObj.swap_id);
                        ANNE.A.copyIco("copy_take_swapaddr_"+swapObj.swap_id, swapObj.take_altaddr);
                        ANNE.A.setHTML("take_swapaddr_"+swapObj.swap_id, swapObj.take_altaddr);
                    }
                    
                    //if this is ours, we could show a cancel.
                    var btntake = document.getElementById("btn_takeyes");
                    ANNE.A.hide(btntake);

                    const btnRecheck = document.getElementById("btn_recheck");
                    btnRecheck.addEventListener("click", (e)=>{
                        ANNE.A.hide(btnRecheck);
                        ANNE.A.showParent("recheck_msg_"+swapObj.swap_id);
                        ANNE.A.spinnerInject("recheck_msg_"+swapObj.swap_id);
                        ADF_OTC_SWAP.recheckStatus(swapObj.swap_target_nid, swapObj.swap_id);
                    })
                    
                    ANNE.A.show(btnRecheck);

                    var elq = ANNE.A.setHTML("take_quantity",ANNE.UTILS.commas(swapObj.annies_available/100));
                    var fb = document.createElement("span");
                    fb.classList = "fermatta-big";
                    fb.innerHTML="𝄐";
                    elq.appendChild(fb);
    
                    var costPerAnnecoin = ADF_OTC_SWAP.calc_cost_per_annecoin(swapObj.annies_per_alt);
                    ANNE.A.setHTML("take_rate",costPerAnnecoin+"&nbsp;"+swapObj.swap_ticker.toLowerCase()+"&nbsp;/&nbsp;"+ADF_OTC_SWAP.ico);

                    ADF_OTC_SWAP.takeMustCost(swapObj, "take_must_cost_"+swapObj.take_id);
                    
                    if (swapObj.txs!=undefined) {
                        ADF_OTC_SWAP.recheckHandle(swapObj, swapObj.txs);
                    }

                    if (swapObj.limitedofferuntil>0) {
                        var ahls = swapObj.anne_height_last_status;
                        var expire = ahls+swapObj.limitedofferuntil;
                        var currentH = ANNE.HISTORY.getGLKB();
                        var diff = expire-currentH;
                    }
                    else {
                        var ahls = swapObj.anne_height_last_status;
                        var expire = ahls+ADF_OTC_SWAP.BLOCKS_TO_COMPLETE_SWAP;
                        var currentH = ANNE.HISTORY.getGLKB();
                        var diff = expire-currentH;
                    }

                    console.log("madeAt:"+ahls+" current:"+currentH+" expire:"+expire);

                    var time = ANNE.UTILS.estimateTimeByBlocksLeft(diff);
                    var timeinfo = "<i class='bi-alarm'>~"+diff+" more blocks ("+time+")"
                    el_take_time_left.innerHTML=timeinfo;

                    if (diff<=0) {
                        ANNE.A.play("warning");
                        el_take_time_left.classList.add("time-left-red");
                        el_take_time_left.innerHTML="<i class='bi bi-exclamation-triangle-fill'></i>&nbsp;EXPIRED. Don't send funds.";
                    }
                    else if (diff<3) {
                        ANNE.A.play("beep");
                        el_take_time_left.classList.add("time-left-red");
                    }
                    else if (diff < 7) {
                        el_take_time_left.classList.add("time-left-yellow");
                    }
                    else {
                        el_take_time_left.classList.add("time-left-green");
                    }
                    
                    
                }, "close");
    
            return;
        }

        if ((swapObj.swap_status=="complete" || swapObj.swap_status=="streampaying") && bOURS==true) {
            var page = "adfswapcomplete";
            var title = "Your completed swap";
            if (swapObj.swap_status=="streampaying") {
                page = "adfswapstreampaying";
                title = "Your streampaying swap"
            }
            ANNE.MEM_UDF._popupWithLoad(page, title,
                ()=>{}, 
                ()=>{},
                async()=>{
                    ANNE.A.setHTML("take_quantity", ANNE.UTILS.commas(swapObj.annies_available/100));
                    ANNE.A.setHTML("take_status",swapObj.swap_status);  
                    if (swapObj.ispaystream==1) {
                        ANNE.A.setHTML("take_swapid","<span class='streampay'>SP:m"+swapObj.moonths+"</span><span class='sid'>"+"#"+swapObj.swap_id+"</span>");
                        ANNE.A.setHTML("take_sp_nid", swapObj.take_sp_nid);
                        ANNE.A.show("take_sp_nid_div");
                        ANNE.A.hideParent("anne_txid");

                        ANNE.A.show("btn_view_streampay");
                        ANNE.A.hide("swaptake_nostreampay");
                        ANNE.A.show("swaptake_streampay");
                        const swap_ID = swapObj.swap_id;
                        var el_sp = document.getElementById("btn_view_streampay");
                        el_sp.addEventListener("click",  (e)=>{
                            ANNE.MEM_UDF.popupElCancel();
                            ADF_OTC_SWAP.showStreampay(swapObj.take_sp_nid, swap_ID);
                        });

                        if (page=="adfswapstreampaying") {
                            ANNE.A.setHTML("complete_block", "(actively distributing TXs per schedule)");
                        }
                        else {
                            ANNE.A.setHTML("complete_block", "final anne txid submit at block #"+ANNE.UTILS.commas(swapObj.take_last_status_change_height));
                        }
                        
                    }
                    else {
                        ANNE.A.show("swaptake_nostreampay");
                        ANNE.A.hide("swaptake_streampay");
                        ANNE.A.setHTML("take_swapid","#"+swapObj.swap_id);
                        ANNE.A.hide("take_sp_nid_div");
                        ANNE.A.showParent("anne_txid");
                        ANNE.A.setHTML("complete_block", "anne txid submit at block #"+ANNE.UTILS.commas(swapObj.take_last_status_change_height));
                    }
                    ANNE.A.setHTML("alt_ticker", swapObj.swap_ticker.toLowerCase());
                    ANNE.A.setHTML("alt_txid", swapObj.take_txid_alt);

                    //getBlockchainTransaction
                    //
                    ANNE.A.setHTML("anne_txid", swapObj.take_txid_anne);
                    if (swapObj.take_txid_anne!=null && swapObj.take_txid_anne!="") {
                        var rel = ANNE.MEMG.lookupByTXID(swapObj.take_txid_anne);
                        if (rel && rel.time>0) return;

                        var dif = ANNE.HISTORY.getGLKB()-swapObj.take_last_status_change_height;
                        console.log(dif);
                        var test = await ANNE.API.getBlockchainTransaction(swapObj.take_txid_anne);
                        console.log(test);
                        if (test && test.confirmations>0) {
                            ANNE.A.appendHTML("anne_txid", "<span class='tx-confirmed'>confirmed</span>");
                        }
                        else {
                            ANNE.A.appendHTML("anne_txid", "<span class='tx-pending'>pending</span>");
                            if (ANNE.USER._userNOb && ANNE.USER._userNObj._userLS.actived==false) {
                                ANNE.USER._userNObj.balanceObj.pendingFirstTX = test;
                                ANNE.USER._userNObj.balanceObj.scatbalance=test.amountNQT;
                                ANNE.USER._userNObj.balanceObj.pendingOnboarding=true;
                            }
                        }
                    }
                },
                "exit");
            return;
        }

        
        //if not available (???) no click.
        if (swapObj.swap_status!="available") return;

        //do a popup to confirm the user really wants to do this.
        await ANNE.MEM_UDF._popupWithLoad("adfswaptake", "Take this offer #"+swapObj.swap_id+"?", 
            ()=>{
                //console.log("popup confirmed");
            }, 
            ()=>{
                //console.log("popup cancelled");
            },
            ()=>{
                //remap these as actions we do may have data to fill them from
                //our master handler... and this is a popup.
                var el_adfswaptake = document.getElementById("adfswaptake");
                el_adfswaptake.id = "adfswaptake_"+swapObj.swap_id;
                
                var el_copy_take_swapaddr = document.getElementById("copy_take_swapaddr");
                el_copy_take_swapaddr.id = "copy_take_swapaddr_"+swapObj.swap_id;
                var el_take_swapaddr = document.getElementById("take_swapaddr");
                el_take_swapaddr.id = "take_swapaddr_"+swapObj.swap_id;
                var el_take_must_cost=document.getElementById("take_must_cost");
                el_take_must_cost.id = "take_must_cost_"+swapObj.swap_id;
                var el_copy_take_must_cost=document.getElementById("copy_take_must_cost");
                el_copy_take_must_cost.id = "copy_take_must_cost_"+swapObj.swap_id;

                ANNE.A.setHTML("take_status",swapObj.swap_status);  
                ADF_OTC_SWAP.takeMustCost(swapObj, "take_must_cost_"+swapObj.swap_id);
              

                if (swapObj.ispaystream==1) {
                    ANNE.A.hide("swaptake_nostreampay");
                    ANNE.A.show("swaptake_streampay");
                    ANNE.A.setHTML("take_swapid","<span class='streampay'>SP:m"+swapObj.moonths+"</span><span class='sid'>"+"#"+swapObj.swap_id+"</span>");
                    ANNE.A.setHTML("swaptake_payments", swapObj.moonths+" moonthly");
                    ANNE.A.setHTML("swaptake_sp_payment", "~"+ANNE.UTILS.commas(Math.ceil(swapObj.annies_available/swapObj.moonths)/100));
                    ANNE.A.setHTML("pop_swap_tickerSP",swapObj.swap_ticker.toUpperCase());
                    ANNE.A.setHTML("pop_swap_ticker2SP",swapObj.swap_ticker.toUpperCase());
                    ANNE.A.setHTML("pop_swap_idSP", "#"+swapObj.swap_id);
                    var el_sp = document.getElementById("btn_view_streampay");
                    const swap_ID = swapObj.swap_id;
                    el_sp.addEventListener("click",  (e)=>{
                        ANNE.MEM_UDF.popupElCancel();
                        ADF_OTC_SWAP.showStreampay(null, swap_ID);
                    });
                }
                else {
                    ANNE.A.setHTML("pop_swap_id", "#"+swapObj.swap_id);
                    ANNE.A.show("swaptake_nostreampay");
                    ANNE.A.hide("swaptake_streampay");
                    ANNE.A.hide("btn_view_streampay");
                    ANNE.A.setHTML("take_swapid","#"+swapObj.swap_id);
                    ANNE.A.setHTML("pop_swap_ticker",swapObj.swap_ticker.toUpperCase());
                    ANNE.A.setHTML("pop_swap_ticker2",swapObj.swap_ticker.toUpperCase());
                }
                ANNE.A.hide("copy_take_swapaddr_"+swapObj.swap_id);
                
                var btntake = document.getElementById("btn_takeyes");

                if (btntake==null) {
                    console.warn("problem, no takebutton in DOM.");
                    return;
                }
                var elq = ANNE.A.setHTML("take_quantity",ANNE.UTILS.commas(swapObj.annies_available/100));
                var fb = document.createElement("span");
                fb.classList = "fermatta-big";
                fb.innerHTML="𝄐";
                elq.appendChild(fb);


                if (swapObj.limitedofferuntil>0) {
                    var madeAt = swapObj.anne_height_last_status;
                    var el_lod = ANNE.A.show("limited_offer_div");
                    var currentH = ANNE.HISTORY.getGLKB();
                    var expires = swapObj.limitedofferuntil+madeAt;
                    var diff = expires-currentH;
                    var limitedofferuntil = document.getElementById("limitedofferuntil");
                    var time = ANNE.UTILS.estimateTimeByBlocksLeft(diff);
                    if (diff<=0) {
                        ANNE.A.play("warning");
                        limitedofferuntil.classList.add("time-left-red");
                        ANNE.A.setHTML(limitedofferuntil, "<i class='bi bi-exclamation-triangle-fill'></i>&nbsp;EXPIRED. Don't send funds.");
                        el_lod.classList.add("lto-red");
                        btntake.classList.add("disabled");
                        btntake.style['pointer-events']='none';
                    }
                    else if (diff<3) {
                        ANNE.A.play("beep");
                        limitedofferuntil.classList.add("time-left-red");
                        ANNE.A.setHTML(limitedofferuntil, "<i class='bi-alarm'></i>&nbsp;block "+ANNE.UTILS.commas(swapObj.limitedofferuntil+madeAt)+" ("+time+")");
                        el_lod.classList.add("lto-red");
                    }
                    else if (diff < 7) {
                        limitedofferuntil.classList.add("time-left-yellow");
                        ANNE.A.setHTML(limitedofferuntil, "<i class='bi-alarm'></i>&nbsp;block "+ANNE.UTILS.commas(swapObj.limitedofferuntil+madeAt)+" ("+time+")");
                        el_lod.classList.add("lto-yellow");
                    }
                    else {
                        limitedofferuntil.classList.add("time-left-green");
                        ANNE.A.setHTML(limitedofferuntil, "<i class='bi-alarm'></i>&nbsp;block "+ANNE.UTILS.commas(swapObj.limitedofferuntil+madeAt)+" ("+time+")");
                        el_lod.classList.add("lto-green");
                    }
                }

                else {
                    ANNE.A.hide("limited_offer_div");
                }


                if (ADF_OTC_SWAP.bch_node_online==false) {
                    btntake.classList.add("disabled");
                    btntake.innerHTML="BCH NODE IS OFFLINE";
                    return;
                }
                else if (ADF_OTC_SWAP.bch_node_synced==false) {
                    btntake.classList.add("disabled");
                    btntake.innerHTML="BCH NODE IS SYNCING";
                    return;
                }

                btntake.addEventListener("click", async(e)=>{
                    ANNE.A.spinnerInject("take_swapaddr_"+swapObj.swap_id);
                    ANNE.A.hide("btn_takeyes");
                    ADF_OTC_SWAP.takeOffer(swapObj.swap_target_nid, swapObj.swap_id, ANNE.USER._userNObj.nid, ANNE.USER._userLS.pubKey);
                });
            });
    });

}


/**
 * 
 * @param {*} $nidOfexistingsp 
 * @param {*} $swapid_back 
 */
ADF_OTC_SWAP.showStreampay=async($nidOfexistingsp, $swapid_back)=> {
    const spSwap = ADF_OTC_SWAP.swaps.get($swapid_back);
    var title = "Paystream Viewer";
    if ($nidOfexistingsp!=null && $nidOfexistingsp!="") {
        title = "Paystream Viewer for nid:"+$nidOfexistingsp;
    }

    ANNE.A.hide("anne_modal_widest");
    await ANNE.MEM_UDF._popupWithLoad("adfpaystream", title, 
        ()=>{
            console.log("popup confirmed "+$swapid_back);
        }, 
        ()=>{
            console.log("popup cancelled "+$swapid_back);
            var divOffer = document.getElementById("swap_"+$swapid_back);
            if (divOffer!=null) divOffer.click();
        },
        async()=>{
            ANNE.A.hide("anne_modal_widest_addendum");
            
            ANNE.A.show("anne_modal_widest");
            
            ANNE.A.setHTML("sp_header", "");
            ANNE.A.setHTML("sp_recap", "");
            ANNE.A.setHTML("sp_table", "");
            ANNE.A.setHTML("sp_relons", "");
            ANNE.A.setHTML("sp_bottom", "");
            
            var height = ANNE.API.getAnnodeProp("height");

            var initobj = new Object();
            initobj.sp_quantity = Math.ceil(spSwap.annies_available/spSwap.moonths);  //specific payment to send, per moonth.                          
            initobj.sp_qt = MEM.distro_qt_specific.nid;           //hardcoded, specific amount payment (not %))
            initobj.sp_frequency = ANNE.blocks_time.lunar_cycle;  //hardcoded, swap app does moonth cycles.  
            initobj.sp_benefactors = [ANNE.USER._userNObj.nid];
            initobj.sp_max_payments = spSwap.moonths;
            initobj.sp_min_start_block = height+10; 
            initobj.so_min_payment = 1000000; //10k coins
            initobj.sp_src = MEM.streampay_src_theFund.nid;
            initobj.sp_keep_min_balance = 0;
            initobj.sp_desired_total_funding = spSwap.annies_available;
            initobj.swap_id = spSwap.swap_id;

            var sp = new streampay_viewer($nidOfexistingsp, initobj);
            //ANNE.A.fadeIn("anne_modal_widest");
            await sp.initASYNC();
        },
        "close",
        "anne_modal_widest");
}


/**
 * 
 * @param {*} $swap 
 * @returns 
 */
ADF_OTC_SWAP.gui_deletedSwap=function($swap) {
    var el_offers = document.getElementById("aos_offers");
    if (el_offers==null) {
        console.log("swaps not showing, do something to notify?");
        return;
    }
    var el_offer = document.getElementById("swap_"+$swap.swap_id);
    ANNE.A.removeSelf(el_offer);
}


/**
 * Just delete the swap and re-add to the display on edit.
 * Easier than altering all the props.
 * @param {*} $swap 
 */
ADF_OTC_SWAP.gui_editSwap = function($swap) {
    //if swaps are showing in ext, do something.
    var el_offers = document.getElementById("aos_offers");
    if (el_offers==null) {
        console.log("swaps not showing, do something to notify?");
        return;
    }
    var el_offer = document.getElementById("swap_"+$swap.swap_id);
    if (el_offer==null) {
        ADF_OTC_SWAP.gui_addSwap($swap);
        return;
    }

    ANNE.A.removeSelf(el_offer);

    ADF_OTC_SWAP.gui_addSwap($swap);
    return;
}



/**
 * 
 * @param {*} $altsPerAnne 
 * @returns 
 */
ADF_OTC_SWAP.calc_annies_per_full_alt=function($altsPerAnne) {
    var annePerAlt = 1/$altsPerAnne;
    console.log("annePerAlt:"+annePerAlt);
    var anniesPerAlt = annePerAlt*100;
    console.log("anniesPerAlt:"+anniesPerAlt);
    return anniesPerAlt;
}

ADF_OTC_SWAP.calc_cost_per_annecoin=function($annies_per_alt) {
    var rate = (1/($annies_per_alt/100));
    return rate.toFixed(8);
}

ADF_OTC_SWAP.fixAltPrice=function($altprice, $places=8) {
    return $altprice.toFixed($places);
}

    
/**
 * Used by admin to send new offer.
 * this will create a record in the table.
 * It will then proogate_results to the rest of the alt-data network.
 * @param {string} $ticker
 * @param {*} $annecoin 
 * @param {*} $altpriceperanne
 * @param {string} $targetNodeNid 
 * @param {int} $limitedofferuntil
 * @param {int} $commish_on - convert to 0 or 1, for no/yes.  commish dyno handled elsehwere
 * @returns 
 */
ADF_OTC_SWAP.admin_enter_new_offer=async($ticker="BCH", $annecoin=10000, $altpriceperanne=0.00000200, $targetNodeNid="", $bISPAYSTREAM=false, $MOONTHS=0, $limitedofferuntil=0, $commish_on=0)=>{
    await ADF_OTC_SWAP.reconnectSocket();

    if ($commish_on < 0 || $commish_on > 50) {
        $commish_on = 0;
        alert("commish out of bounds at "+$commish_on);
        return;
    }

    if ($commish_on>0) $commish_on=1;

    if ($targetNodeNid==undefined || $targetNodeNid==null || $targetNodeNid=="") {
        $targetNodeNid=ADF_OTC_SWAP.defaultTargetNodeNid;
    }

    var anniesPerAltCoin = ADF_OTC_SWAP.calc_annies_per_full_alt($altpriceperanne);
    
    if ($limitedofferuntil==undefined || $limitedofferuntil==null) {
        $limitedofferuntil = 0;
    }

    var adpayload = {
        adid:ANNE.UTILS.randHex(12),
        annecoin:$annecoin,
        anniesPerAlt:anniesPerAltCoin,
        commish_on:$commish_on,
        ispaystream:$bISPAYSTREAM,
        limitedofferuntil:$limitedofferuntil,
        moonths:$MOONTHS,
        targetNodeNid:$targetNodeNid,
        ticker:$ticker,
        user_nid:ANNE.USER._userNObj.nid,
        zpropagate:false,
        zreq_sig:true,
        zzzzret_targetNodeNid:ANNE._annodesLS.last_nid
    }
    var pObj = await ADF_OTC_SWAP.signMsg(adpayload);
    if (pObj==null) {
        console.warn("signing problem");
        return null;
    }
    var adtype = "ADF_SWAP";
    var adsubtype = "ADD_OFFER";
    var rez = await ANNE.API.sendAltData(adtype, adsubtype, pObj.payloadstring, pObj.sig);
    return rez;
}



/**
 * 
 * @param {*} $targetNodeNid 
 * @returns 
 */
ADF_OTC_SWAP.getOffers=async($targetNodeNid)=>{
    await ADF_OTC_SWAP.reconnectSocket();
    if ($targetNodeNid==undefined || $targetNodeNid==null || $targetNodeNid=="") {
        $targetNodeNid=ADF_OTC_SWAP.defaultTargetNodeNid;
    }

    var adpayload = {
        adid:ANNE.UTILS.randHex(12),
        last_swap_id:ADF_OTC_SWAP.max_swap_id,
        targetNodeNid:$targetNodeNid,
        user_nid:ANNE.USER._userNObj.nid,
        zdistributable:true,
        zlastupdate_block:ADF_OTC_SWAP.last_height,
        zpropagate:false,
        zreq_sig:true,
        zzzzret_targetNodeNid:ANNE._annodesLS.last_nid
    }

    var pObj = await ADF_OTC_SWAP.signMsg(adpayload);
    if (pObj==null) {
        console.warn("signing problem");
        return null;
    }

    var adtype = "ADF_SWAP";
    var adsubtype = "GET_OFFERS"; //AD.ADF_SWAP.GET_OFFERS

    var rez = await ANNE.API.sendAltData(adtype, adsubtype, pObj.payloadstring, pObj.sig);
    return rez;
}


/**
 * 
 * @param {*} $targetNodeNid 
 * @param {*} $swap_id 
 * @param {*} $TAKE_NID_TO_CREDIT 
 * @param {*} $TAKE_PUBKEY 
 * @returns 
 */
ADF_OTC_SWAP.takeOffer=async($targetNodeNid, $swap_id, $TAKE_NID_TO_CREDIT, $TAKE_PUBKEY)=>{
    if ($targetNodeNid==undefined || $targetNodeNid==null || $targetNodeNid=="") {
        $targetNodeNid=ADF_OTC_SWAP.defaultTargetNodeNid;
    }
    if ($swap_id==null || $swap_id=="") {
        console.log("(getOffers) missing swap_id");
        return;
    }
    if ($TAKE_NID_TO_CREDIT==null || $TAKE_NID_TO_CREDIT=="") {
        console.log("(getOffers) missing TAKE_NID_TO_CREDIT");
        return;
    }
    if ($TAKE_PUBKEY==null || $TAKE_PUBKEY=="") {
        console.log("(getOffers) missing TAKE_PUBKEY");
        return;
    }

    await ADF_OTC_SWAP.reconnectSocket();

    //TODO: set socket for listening to payments with filter for THIS ANNE_ID.
    //we will pass the nidlongstring in the filter and when we blast the socket we
    //will make sure to use the nidlong string as well.

    if (ADF_OTC_SWAP.REF==null) ADF_OTC_SWAP.REF = "";
    //if we have nulls the value goes missing in the bash and alters the props order.

    var adpayload = {
        adid:ANNE.UTILS.randHex(12),
        ref:ADF_OTC_SWAP.REF,
        swap_id:$swap_id,
        targetNodeNid:$targetNodeNid,
        take_nid_to_cred:$TAKE_NID_TO_CREDIT,
        take_pubkey:$TAKE_PUBKEY,
        user_nid:ANNE.USER._userNObj.nid,
        zdistributable:false,
        zpropagate:true,
        zreq_sig:true,
        zzzzret_targetNodeNid:ANNE._annodesLS.last_nid
    }

    var pObj = await ADF_OTC_SWAP.signMsg(adpayload);
    if (pObj==null) {
        console.warn("signing problem");
        return null;
    }
    
    var adtype = "ADF_SWAP";
    var adsubtype = "TAKE_OFFER"; 
    var rez = await ANNE.API.sendAltData(adtype, adsubtype, pObj.payloadstring, pObj.sig);
    return rez;
}


/**
 * Let the user manually invoke the recheck in case socket was down,
 * or something was missed.
 * @param {*} $targetNodeNid 
 * @param {*} $swap_id
 * @returns 
 */
ADF_OTC_SWAP.recheckStatus=async($targetNodeNid, $swap_id)=>{
    if ($targetNodeNid==undefined || $targetNodeNid==null ||$targetNodeNid=="") {
        $targetNodeNid=ADF_OTC_SWAP.defaultTargetNodeNid;
    }
    if ($targetNodeNid==null || $targetNodeNid=="") {
        console.log("(recheckStatus) missing targetNodeNid");
        return;
    }
    if ($swap_id==null || $swap_id=="") {
        console.log("(recheckStatus) missing swap_id");
        return;
    }

    await ADF_OTC_SWAP.reconnectSocket();

    //TODO: set socket for listening to payments with filter for THIS ANNE_ID.
    //we will pass the nidlongstring in the filter and when we blast the socket we
    //will make sure to use the nidlong string as well.

    var adpayload = {
        adid:ANNE.UTILS.randHex(12),
        swap_id:$swap_id,
        targetNodeNid:$targetNodeNid,
        user_nid:ANNE.USER._userNObj.nid,
        zdistributable:false,
        zpropagate:false,
        zreq_sig:true,
        zzret_sig:true,
        zzzzret_targetNodeNid:ANNE._annodesLS.last_nid,
        zzzzzz_newpubkey:""
    }

    var pObj = await ADF_OTC_SWAP.signMsg(adpayload);
    if (pObj==null) {
        console.warn("signing problem");
        return null;
    }
    
    var adtype = "ADF_SWAP";
    var adsubtype = "RECHECK_OFFER"; 
    var rez = await ANNE.API.sendAltData(adtype, adsubtype, pObj.payloadstring, pObj.sig);
    return rez;
}


/**
 * Let the user manually invoke the recheck in case socket was down,
 * or something was missed.
 * @param {*} $targetNodeNid 
 * @param {*} $swap_id 
 * @returns 
 */
ADF_OTC_SWAP.deleteOffer=async($targetNodeNid, $swap_id)=>{
    if ($targetNodeNid==undefined || $targetNodeNid==null || $targetNodeNid=="") {
        $targetNodeNid=ADF_OTC_SWAP.defaultTargetNodeNid;
    }
    if ($targetNodeNid==null || $targetNodeNid=="") {
        console.log("(recheckStatus) missing targetNodeNid");
        return;
    }
    if ($swap_id==null || $swap_id=="") {
        console.log("(recheckStatus) missing swap_id");
        return;
    }
    await ADF_OTC_SWAP.reconnectSocket();
    var adpayload = {
        adid:ANNE.UTILS.randHex(12),
        swap_id:$swap_id,
        targetNodeNid:$targetNodeNid,
        user_nid:ANNE.USER._userNObj.nid,
        zdistributable:false,
        zpropagate:false,
        zreq_sig:true,
        zzret_sig:true,
        zzzzret_targetNodeNid:ANNE._annodesLS.last_nid
    }
    var pObj = await ADF_OTC_SWAP.signMsg(adpayload);
    if (pObj==null) {
        console.warn("signing problem");
        return null;
    }
    var adtype = "ADF_SWAP";
    var adsubtype = "DELETE_OFFER"; 
    var rez = await ANNE.API.sendAltData(adtype, adsubtype, pObj.payloadstring, pObj.sig);
    return rez;
}


/**
 * Sign the payload for any coms that should require signed payload.
 * 
 * NOTE: if you change a com to be signMsg it potentially change the order of params
 * that hit the alt-data bash handler.  
 * @param {*} $json 
 * @returns 
 */
ADF_OTC_SWAP.signMsg=async($json)=>{
    var seedObj = null;
    if (USER.seedObj==null) {
        if (USER.unlock_pw=="" || USER.unlock_pw==null) {
            if (typeof ANNE.MEM_UDF.hide3js =="function") ANNE.MEM_UDF.hide3js();
            ANNE.A.hide("top_header_stuff");
            ANNE.USER.lockApp();
            return null;
        }
        seedObj = await USER.getSeedWithPW(USER.unlock_pw);
    }
    else seedObj = USER.seedObj;

    if (seedObj==null) {
        //this.recapBad("problem extracting key from memory.  Can't sign message.");
        return null;
    }
    var aO = USER.prepANNEAuthObjFromSeed(seedObj.seed);
    if (aO==null) {
        //this.recapBad("problem extracting key from memory.  Can't sign message.");
        return null;
    }

    if (USER.privKey=="" || USER.privKey==undefined) {
        USER.privKey = aO.sign_key;
    }

    var sig=null;
    

    //if we are signning the payload, make sure the zreq_sig flag is on.
    //this will force the alt-data to verify the signature.
    //in the bash that handled various altdata subtypes, the signature flag can be 
    //looked for an ensured to be on -- thus guaranteeing the values were not falsified.
    if ($json.zreq_sig!=true) {
        $json.zreq_sig=true;
    }


    //slip in zzzzzz_newpubkey if needed
    if (ANNE.USER._userLS==null || ANNE.USER.isActivated()==false) {
        $json.zzzzzz_newpubkey=ANNE.USER._userLS.pubKey;
        console.log($json.zzzzzz_newpubkey);
    }

    var payloadString = JSON.stringify($json);

    //gen the signature
    try {
        var aObj = await ANNE.USER.getPrivateKeyHex();
        if (aObj==null) {
            console.warn("no private key hex found");
            return null;
        }
        sig = await ANNE.USER.signMessage(payloadString, aObj.sign_key);
        return {sig:sig, payloadstring:payloadString};
    }

    catch(e) {
        console.log("problem");
        return null;
    }
}


ADF_OTC_SWAP.initNewForm=async()=> {
    console.log("launch add new form.");

    await ANNE.MEM_UDF._popupWithLoad("adfswapmake", "ADMIN: Make new swap offer", 
        ()=>{
            console.log("popup confirmed");
        }, 
        ()=>{
            console.log("popup cancelled");
        },
        ()=>{
            console.log("poup loaded");

            var btnmake = document.getElementById("btn_make");
            ANNE.A.addClass(btnmake, "disabled");
            
            const el_total_payouts = document.getElementById("aa_total_payouts");
            const el_ticker = document.getElementById("aa_ticker");
            const el_altrate = document.getElementById("aa_altrate_per_annecoin");
            const el_total_cost = document.getElementById("aa_total_cost");
            const el_streampay = document.getElementById("aa_chk_streampay");
            const el_moonths = document.getElementById("aa_sp_maxpayments");
            const el_limited_time = document.getElementById("aa_sp_limited_time");
            const el_commish_per = document.getElementById("aa_sp_commish_per");

            function validateForm() {
                var totalpayout =  ANNE.UTILS.forceInt(el_total_payouts.value);
                var probs = [];
                var maxpayments =  ANNE.UTILS.forceInt(el_moonths.value);

                if (el_streampay.checked==true && (totalpayout=="" || totalpayout < maxpayments*10000)) {
                    probs.push("min total for SP must be 10k annecoins * moonths.");
                }
                else {
                    var altr = el_altrate.value;
                    if (altr=="") {
                        probs.push("rate required");
                    }
                    else {
                        altr = altr.replace(/[^0-9.]/g, '');
                        const decimalIndex = altr.indexOf('.');
                        if (decimalIndex !== -1) {
                            altr = altr.slice(0, decimalIndex + 1) + altr.slice(decimalIndex + 1).replace(/\./g, '');
                        }
                         
                        var total =  ANNE.UTILS.forceInt(totalpayout)*parseFloat(altr);
                        total = total.toFixed(8)   
                        el_total_cost.innerHTML = total;
                    }
                }

                if (el_streampay.checked==true) {
                    var maxpayments =  ANNE.UTILS.forceInt(el_moonths.value);
                    if (maxpayments<2) {
                        probs.push("moonths must be 2 or more");
                    }
                }

                var altr = el_altrate.value;
                if (altr=="" || isNaN(altr)) {
                    probs.push("alt rate must be a number");
                }

                var ticker = el_ticker.value; 
                if (ticker=="" || (ticker!="BCH" && ticker!="bch")) {
                    probs.push("only bch supported right now");
                }

                if (probs.length>0) {
                    el_total_cost.innerHTML = 0;
                    ANNE.A.setHTML("warning_msg", probs.join(" :: "));
                    ANNE.A.addClass(btnmake, "disabled");
                    ANNE.A.showParent("warning_msg");
                }
                else {
                    ANNE.A.setHTML("warning_msg", "");
                    ANNE.A.removeClass(btnmake, "disabled");
                    ANNE.A.hideParent("warning_msg");
                }
            }

            el_total_payouts.addEventListener("input", (e)=>{
                validateForm();
            });

            el_ticker.addEventListener("input", (e)=>{
                validateForm();
            });

            el_altrate.addEventListener("input", (e)=>{
                validateForm();
            });
            
            el_moonths.addEventListener("change", (e)=>{
                el_moonths.value = el_moonths.value.replace(/[^0-9]/g, '');
                validateForm();
            });

            var istreampay = 0;

            el_streampay.addEventListener("change", (e)=>{
                if (el_streampay.value=="yes") {
                    ANNE.A.showParent(el_moonths);
                    istreampay = 1;
                    validateForm();
                }
                else {
                    ANNE.A.hideParent(el_moonths);
                    istreampay = 0;
                    validateForm();
                }
            });

            
            //init default.
            btnmake.addEventListener("click", async(e)=>{
                ANNE.A.hide("btnmake");
                var total_payouts = el_total_payouts.value;
                var altr = el_altrate.value;
                var maxpayments = 0;
                if (el_streampay.checked==true) {
                    ANNE.A.showParent(el_moonths);
                    istreampay = 1;
                    maxpayments =  ANNE.UTILS.forceInt(el_moonths.value);
                    validateForm();
                }
                else {
                    ANNE.A.hideParent(el_moonths);
                    istreampay = 0;
                    validateForm();
                }

                var limitedtime = ANNE.UTILS.forceInt(el_limited_time.value);
                var commish_on = ANNE.UTILS.forceInt(el_commish_per.value);
                ADF_OTC_SWAP.admin_enter_new_offer(el_ticker.value.toUpperCase(), total_payouts, altr, ADF_OTC_SWAP.defaultTargetNodeNid, istreampay, maxpayments, limitedtime, commish_on);
                
                ANNE.MEM_UDF.popupElCancel();
            })

            validateForm();
        });
}


/**
 * 
 * @param {*} $el 
 * @param {*} $swap 
 * @returns 
 */
ADF_OTC_SWAP.gui_setCostPerAnnecoin = function($el, $swap) {
    var costPerAnnecoin = ADF_OTC_SWAP.calc_cost_per_annecoin($swap.annies_per_alt);
    var disp = costPerAnnecoin+"&nbsp;"+$swap.swap_ticker.toLowerCase()+"&nbsp;/&nbsp;"+ADF_OTC_SWAP.ico;
    ANNE.A.setHTML($el, disp);
    return costPerAnnecoin;
}



/************** (from file ./js/altdata/altdata_adf_otc_swap2.js)**/